/*!***************************************************************************

  module:		 Container_DoubleNodeIterator.hpp

  ------------------------------------------------------------------------------

  responsible:  Till Luhmann

  special area: Standard container in SAPDB

  description:  Bidirectional iterator and reverse iterator templates 
                for doubly linked list nodes

  see also:     Container_SingleNodeIterator.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef CONTAINER_DOUBLENODEITERATOR_HPP
#define CONTAINER_DOUBLENODEITERATOR_HPP


#include "Container/Container_NodeIterator.hpp"


/*!*****************************************************************************

  class:        Container_DoubleNodeIterator

  description:  
  
  Bidirectional iterator template for doubly linked list nodes.

  This iterator may be used to iterate Container_DoubleList<T> and
  Container_CircularList<T>.

  template arguments:    
  
  <class T>             Type of the elements of the list.
  <class ValueType>     Type that the dereference operator returns.
  <class NodePointer>   Type of the node pointer.

*******************************************************************************/
template <class T, class ValueType, class NodePointer>
class Container_DoubleNodeIterator :
public Container_NodeIterator<T, ValueType, NodePointer>
{

private:

    typedef Container_DoubleNodeIterator<T, ValueType, NodePointer> 
            Iterator;

    typedef Container_NodeIterator<T, ValueType, NodePointer> 
            Parent;

public:

    /*!----------------------------------------------------------------------
    function:     Container_DoubleNodeIterator()
    description:  Constructor
    arguments:    node [in]   pointer to a list node
    return value: none
    prototypes:   2
    -----------------------------------------------------------------------*/
    Container_DoubleNodeIterator()
    {}

    Container_DoubleNodeIterator
        (NodePointer                    Node)
        : 
         Parent(Node) 
    {}

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  Prefix increment; advances iterator to next node
    arguments:    none
    return value: (Iterator&) *this
    --------------------------------------------------------------------------*/
    Iterator& operator++() 
    {
        SAPDBERR_ASSERT_STATE(this->IsValid());
        this->m_Node = this->m_Node->m_Next;
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     operator--()
    description:  Prefix decrement; steps back iterator to previous node
    arguments:    none
    return value: (Iterator&) *this
    --------------------------------------------------------------------------*/
    Iterator& operator--() 
    {
        SAPDBERR_ASSERT_STATE(this->IsValid());
        this->m_Node = this->m_Node->m_Prev;
        return *this;
    } 
  
};
/*!*****************************************************************************

   endclass: Container_DoubleNodeIterator

*******************************************************************************/



/*!*****************************************************************************

  class:        Container_DoubleNodeReverseIterator

  description:  
  
  Bidirectional reverse iterator template for doubly linked list nodes.

  This iterator may be used to iterate Container_DoubleList<T> and
  Container_CircularList<T>.

  template arguments:    
  
  <class T>             Type of the elements of the list.
  <class ValueType>     Type that the dereference operator returns.
  <class NodePointer>   Type of the node pointer.

*******************************************************************************/
template <class T, class ValueType, class NodePointer>
class Container_DoubleNodeReverseIterator :
public Container_DoubleNodeIterator<T, ValueType, NodePointer>
{
private:

    typedef Container_DoubleNodeReverseIterator<T, ValueType, NodePointer> 
            ReverseIterator;

    typedef Container_DoubleNodeIterator<T, ValueType, NodePointer> 
            Parent;

public:

    /*!----------------------------------------------------------------------
    function:     Container_DoubleNodeReverseIterator()
    description:  Constructor
    arguments:    node [in]   pointer to a list node
    return value: none
    prototypes:   2
    -----------------------------------------------------------------------*/
    Container_DoubleNodeReverseIterator()
    {}

    Container_DoubleNodeReverseIterator
        (NodePointer                    Node)
        : 
         Parent(Node) 
    {}

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  Prefix increment; advances iterator to next node
    arguments:    none
    return value: (Iterator&) *this
    --------------------------------------------------------------------------*/
    ReverseIterator& operator++() 
    {
        Parent::operator--();
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     operator--()
    description:  Prefix decrement; steps back iterator to previous node
    arguments:    none
    return value: (Iterator&) *this
    --------------------------------------------------------------------------*/
    ReverseIterator& operator--() 
    {
        Parent::operator++();
        return *this;
    }

};
/*!*****************************************************************************

   endclass: Container_DoubleNodeReverseIterator

*******************************************************************************/





#endif // CONTAINER_DOUBLENODEITERATOR_HPP

