/*!******************************************************************************

  module:       Container_List.hpp

  ------------------------------------------------------------------------------

  author:       Stefan Baier, Till Luhmann

  responsible:  Stefan Baier

  special area: Standard container in SAPDB

  description:  Class declaration and definition for template list classes
                (singly linked lists)

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin LGPL
  Copyright (C) 2000 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*******************************************************************************/

#ifndef CONTAINER_LIST_HPP
#define CONTAINER_LIST_HPP

#include <memory.h>

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "Container/Container_Node.hpp"
#include "Container/Container_NodeIterator.hpp"



/*!****************************************************************************

  class:       Container_List

  description: 
  
  This class implements a doubly linked list

  The list uses raw allocators as defined in the SAPDBMem component
  for allocating memory for the list elements and the link pointers.

  As of release 7.4, no exceptions are thrown.
  Instead, SAPDB_Bool return values or invalid iterators are returned by
  the methods where appropriate.

  template arguments:   
  
  <T>           type of the elements to be held in the list

                T must provide the following methods:

                - Default constructor T()
                - SAPDB_Bool operator==(const T& t) const
                          
******************************************************************************/
template <class T>
class Container_List
{
private:

   typedef Container_Node<T>                                NodeType;

   typedef Container_List<T>                                List;

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : ValueType
    description:  Node content type
    --------------------------------------------------------------------------*/
    typedef T                                               ValueType;
    /*!-------------------------------------------------------------------------
    declaration : NodePointer
    description:  Pointer to node
    --------------------------------------------------------------------------*/
    typedef NodeType*                                       NodePointer;
    /*
    typedef SAPDB_Invalidatable
            <NodeType*, NodeType::InvalidNode>              NodePointer;
    */
    /*!-------------------------------------------------------------------------
    declaration : Iterator
    description:  List iterator
    --------------------------------------------------------------------------*/
    typedef Container_NodeIterator
            <ValueType, ValueType, NodeType* >              Iterator;
    /*!-------------------------------------------------------------------------
    declaration : ConstIterator
    description:  Constant list iterator 
    --------------------------------------------------------------------------*/
    typedef Container_NodeIterator
            <ValueType, const ValueType, const NodeType* >  ConstIterator;
    /*!-------------------------------------------------------------------------
    declaration : SizeType
    description:  Count of nodes (elements) in a list
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt                                      SizeType;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  allocator for queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDBMem_IRawAllocator                          Allocator;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Construction and destruction                                 */
    /*!-------------------------------------------------------------------------
    function:     Container_List()
    description:  Constructor
    arguments:    alloc [in] reference to raw allocator
    return value: none
    --------------------------------------------------------------------------*/
    Container_List
        (Allocator&                     alloc)
        : 
         m_Allocator(alloc), 
         m_Head(NodeType::InvalidNode),
         m_Last(NodeType::InvalidNode),
         m_NodeCount(0)
    {}

    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  Initializes a list 
    
                  - The former list content is deleted.
                  - All elements from 'list' are copied.

    arguments:    list  [in] list to be copied
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool Initialize
        (const List&              list)
    {
        return CopyList(list);
    }

    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes the whole list
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Delete();

    /*!-------------------------------------------------------------------------
    function:     ~Container_List()
    description:  Destructs the list and deallocates all elements
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    ~Container_List()
    {
        Delete();
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Construction and Destruction                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Status functions                                             */
    /*!----------------------------------------------------------------------
    function:     IsEmpty()
    description:  Checks whether list is empty
    arguments:    none
    return value: (SAPDB_Bool) true if empty
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsEmpty() const
    { 
        return m_Head == NodeType::InvalidNode; 
    }

    /*!-------------------------------------------------------------------------
    function:     GetSize()
    description:  Returns the actual node count of the list
    arguments:    none
    return:       (SizeType) node count
    --------------------------------------------------------------------------*/
    SizeType GetSize() const
    { 
        return m_NodeCount; 
    }

    /*!-------------------------------------------------------------------------
    function:     GetAllocator()
    description:  Returns the associated allocator
    arguments:    none
    return value: (Allocator&) reference to allocator
    --------------------------------------------------------------------------*/
    Allocator& GetAllocator() const
    { 
        return m_Allocator; 
    }

    /*!-------------------------------------------------------------------------
    function:     operator==()
    description:  Compares actual list with 'list'
    arguments:    list [in]     list to be compared
    return value: (SAPDB_Bool) true if both lists have the same content
    --------------------------------------------------------------------------*/
    SAPDB_Bool operator==
        (const List&                    list) const;

    /*!-------------------------------------------------------------------------
    function:     operator!=()
    description:  Compares actual list with 'list'
    arguments:    list [in]     list to be compared
    return value: (SAPDB_Bool) true if both lists have different content
    --------------------------------------------------------------------------*/
    SAPDB_Bool operator !=
        (const List&                    list) const
    { 
        return !(this->operator==(list)); 
    }
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Status functions                                          */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: List manipulation                                            */
    /*!-------------------------------------------------------------------------
    function:     InsertFront()
    description:  Inserts a new element at the beginning of the list
    arguments:    elem [in]     element to be inserted
    return value: (Iterator) iterator pointing to inserted node;
                             invalid iterator if insert not successful
    --------------------------------------------------------------------------*/
    Iterator InsertFront
        (const ValueType&               elem);

    /*!-------------------------------------------------------------------------
    function:     InsertEnd()
    description:  Inserts a new element at the end of the list

    arguments:    elem [in]     element to be inserted
    return value: (Iterator) iterator pointing to inserted node;
                             invalid iterator if insert not successful
    --------------------------------------------------------------------------*/
    Iterator InsertEnd
        (const ValueType&               elem);

    /*!-------------------------------------------------------------------------
    function:     Insert()
    description:  Inserts a new element before a given iterator position

                  - After the operation, 'pos' points to the inserted node.
                  - This method is rather slow, because in order to insert a
                    list element at a certain position, the predecessor of that
                    position has to be found.

    arguments:    pos  [in/out]  iterator insert position
                  elem [in]      element to be inserted
    return value: (Iterator) iterator pointing to inserted node;
                             invalid iterator if insert not successful
    --------------------------------------------------------------------------*/
    Iterator Insert
        (Iterator&                      pos, 
         const ValueType&               elem);

    /*!-------------------------------------------------------------------------
    function:     InsertBehind()
    description:  Inserts a new element behind a given iterator position

                  - After the operation, 'pos' points to the inserted node.
    arguments:    pos [in/out]  iterator insert position
                  elem [in]     element to be inserted
    return value: (Iterator) iterator pointing to inserted node;
                             invalid iterator if insert not successful
    --------------------------------------------------------------------------*/
    Iterator InsertBehind
        (Iterator&                      pos, 
         const ValueType&               elem);

    /*!-------------------------------------------------------------------------
    function:     DeleteFront()
    description:  Deletes the first element of the list

                  - The list must not be empty.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void DeleteFront();

    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes the element at a given iterator position

                 - This method is rather slow, because in order to delete a
                   list element at a certain position, the predecessor of that
                   position has to be found.
                  - After the operation, 'pos' points to the node behind 
                    the deleted one.
    arguments:    pos [in/out]  iterator delete position
    return value: none
    --------------------------------------------------------------------------*/
    void Delete
        (Iterator&                      pos);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: List manipulation                                         */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator functions                                           */
    /*!-------------------------------------------------------------------------
    function:     Begin()
    description:  Returns an iterator to the first list node
    arguments:    none
    return value: (Iterator/ConstIterator)
    prototypes:   2
    --------------------------------------------------------------------------*/
    Iterator Begin()
    { 
        return Iterator(m_Head); 
    }

    ConstIterator Begin() const
    { 
        return ConstIterator(m_Head); 
    }

    /*!-------------------------------------------------------------------------
    function:     Last()
    description:  Returns an iterator to the last list node
    arguments:    none
    return value: (Iterator/ConstIterator)
    prototypes:   2
    --------------------------------------------------------------------------*/
    Iterator Last()
    { 
        return Iterator(m_Last); 
    }

    ConstIterator Last() const
    { 
        return ConstIterator(m_Last); 
    }

    /*!-------------------------------------------------------------------------
    function:     End()
    description:  Returns an iterator behind the last list node
    arguments:    none
    return value: (Iterator/ConstIterator)
    prototypes:   2
    --------------------------------------------------------------------------*/
    Iterator End()
    { 
        return Iterator(NodeType::InvalidNode); 
    }

    ConstIterator End() const
    { 
        return ConstIterator(NodeType::InvalidNode); 
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator functions                                        */
    /* -----------------------------------------------------------------------*/

    /*!----------------------------------------------------------------------
    function:     NodeCountConsistent()
    description:  Checks whether actual node count is the same as m_NodeCount

                  - For test usage only.
    arguments:   none
    return:      (SAPDB_Bool) true if consistent
    -----------------------------------------------------------------------*/
    SAPDB_Bool NodeCountConsistent() const;

private:

    /*--------------------------------------------------------------------------
    function:     Container_List()
    description:  Copy constructor
    arguments:    list  [in] list to be copied
    return value: none
    --------------------------------------------------------------------------*/
    Container_List
        (const List&                    list)
    { 
        // do nothing in 7.4
    }

    /*--------------------------------------------------------------------------
    function:     operator=()
    description:  Assigns list (does not work in 7.4)

                  - The content of 'list' is copied (duplicated).
    arguments:    list [in]     list to be assigned
    return value: (DoubleList&) *this
    --------------------------------------------------------------------------*/
    List& operator=
        (const List&                    list)
    {
        // do nothing in 7.4
        return *this;
    }

    /*-----------------------------------------------------------------------
    function:     CopyList()
    description:  Performs an elementwise list copy

                  - The former content of the actual list is deleted.
                  - Argument list 'list' is copied element by element
                    into the actual list.
    arguments:    list  [in] list to be copied
    return:      (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool CopyList
        (const List&                    list);

    /*--------------------------------------------------------------------------
    function:     CreateNode()
    description:  Dynamically creates a new node
    arguments:    elem [in]     content of the new node
    return value: (NodeType*) pointer to node (0 if not successful)
    --------------------------------------------------------------------------*/
    // !!! Attention: Please do not change the following lines
    // Otherwise a compiler error on linux is generated.
    NodeType* CreateNode
        (const ValueType&               elem)
    {
      return new(m_Allocator) NodeType(elem);
    }
    // !!! Attention: Please do not change the lines above.

    /*--------------------------------------------------------------------------
    declaration : m_Allocator
    description:  Reference to underlying raw allocator
    --------------------------------------------------------------------------*/
    Allocator&                                      m_Allocator;
    /*--------------------------------------------------------------------------
    declaration : m_Head
    description:  Pointer to first node in the list
    --------------------------------------------------------------------------*/
    NodePointer                                     m_Head;
    /*--------------------------------------------------------------------------
    declaration : m_Last
    description:  Pointer to last node in the list
    --------------------------------------------------------------------------*/
    NodePointer                                     m_Last;
    /*--------------------------------------------------------------------------
    declaration : m_NodeCount
    description:  Actual node count in the list
    --------------------------------------------------------------------------*/
    SizeType                                        m_NodeCount;

};
/*!*****************************************************************************

   endclass: Container_DoubleList

*******************************************************************************/

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_List<T>::operator ==
    (const List&                    list) const
{
    // -------------------------------------------------------------------------
    // check whether lists have the same node count
    // -------------------------------------------------------------------------
    if (m_NodeCount != list.m_NodeCount)
        return false;

    // -------------------------------------------------------------------------
    // check whether both lists are empty (lists have the same node count)
    // -------------------------------------------------------------------------
    if (IsEmpty())
        return true;

    // -------------------------------------------------------------------------
    // compare lists elementwise (lists have the same node count)
    // -------------------------------------------------------------------------
    ConstIterator iter1 = Begin();
    ConstIterator iter2 = list.Begin();
    ConstIterator end   = End();
    while (iter1 != end) 
    {
        if (!(*iter1 == *iter2)) 
            return false;
        ++iter1;
        ++iter2;
    }

    return true;
}

// -----------------------------------------------------------------------------

template <class T>
typename Container_List<T>::Iterator 
Container_List<T>::InsertFront
    (const ValueType&               elem)
{
    // -------------------------------------------------------------------------
    // create a new node
    // -------------------------------------------------------------------------
    NodeType *node     = CreateNode(elem);
    if (node == 0)
        return Iterator();

    // -------------------------------------------------------------------------
    // chain in the node at the beginning of the list
    // -------------------------------------------------------------------------
    node->m_Next   = m_Head;
    if (IsEmpty ()) //List is empty
    {
        m_Last = node;
    }

    m_Head         = node;

    ++m_NodeCount;

    return Iterator(node);
}

// -----------------------------------------------------------------------------

template <class T>
inline
typename Container_List<T>::Iterator 
Container_List<T>::InsertEnd
    (const ValueType&               elem)
{
    // -------------------------------------------------------------------------
    // create a new node
    // -------------------------------------------------------------------------
    NodeType *node     = CreateNode(elem);
    if (node == 0)
        return Iterator();

    // -------------------------------------------------------------------------
    // chain in the node at the end of the list
    // -------------------------------------------------------------------------
    if (IsEmpty())
    {
        m_Head = node;
    }
    else
    {
        m_Last->m_Next = node;
    }

    m_Last          = node;
    m_Last->m_Next  = NodeType::InvalidNode;

    ++m_NodeCount;

    return Iterator(node);
}

// -----------------------------------------------------------------------------

template <class T>
typename Container_List<T>::Iterator 
Container_List<T>::Insert
    (Iterator&                      pos, 
     const ValueType&               elem)
{
    SAPDBERR_ASSERT_ARGUMENT(pos.IsValid());

    // -------------------------------------------------------------------------
    // check whether InsertFront() can be used
    // -------------------------------------------------------------------------
    if (pos == Begin()) 
    {
        // insert the new element at the beginning of the list
        if (!InsertFront(elem).IsValid())
            return Iterator();
        else
        {
            // set iterator position to new begin
            pos = Begin();
            return pos;
        }
    }
    // -------------------------------------------------------------------------
    // insert element in the middle of the list
    // -------------------------------------------------------------------------
    else
    {
        // -------------------------------------------------------------------------
        // Find the iterator position before 'pos'
        // -------------------------------------------------------------------------
        Iterator prev  = Begin();
        Iterator next = ++Begin();
        while (next != pos) 
        {
            SAPDBERR_ASSERT_STATE(next != End());      // error: 'pos' is not in list
            ++prev;
            ++next;
        }

        // -------------------------------------------------------------------------
        // create a new node
        // -------------------------------------------------------------------------
        NodeType *node = CreateNode(elem);
        if (node == 0)
            return Iterator();

        // -------------------------------------------------------------------------
        // chain the new node behind 'prev'
        // -------------------------------------------------------------------------
        node->m_Next           = prev.GetNode()->m_Next;
        prev.GetNode()->m_Next = node;

        if (pos == End ())
        {
            m_Last = node;   
        }

        // Set the iterator position.
        pos            = Iterator(node);

        ++m_NodeCount;
        return pos;
    }
}

// -----------------------------------------------------------------------------

template <class T>
typename Container_List<T>::Iterator 
Container_List<T>::InsertBehind
    (Iterator&                      pos, 
     const ValueType&               elem)
{
    SAPDBERR_ASSERT_ARGUMENT(pos.IsValid());
    SAPDBERR_ASSERT_ARGUMENT(pos != End());
    SAPDBERR_ASSERT_STATE(!IsEmpty());

    // -------------------------------------------------------------------------
    // create a new node
    // -------------------------------------------------------------------------
    NodeType *node  = CreateNode(elem);
    if (node == 0)
        return Iterator();

    Iterator prev   = pos;

    // -------------------------------------------------------------------------
    // chain the new node behind 'pos'
    // -------------------------------------------------------------------------
    node->m_Next           = prev.GetNode()->m_Next;
    prev.GetNode()->m_Next = node;
    if (pos == Last ())  // Insertion takes place at the end 
    {   
       m_Last = node;
    }

    pos                    = Iterator(node);

    ++m_NodeCount;
    return pos;
}

// -----------------------------------------------------------------------------

template <class T>
inline
void Container_List<T>::DeleteFront()
{
    SAPDBERR_ASSERT_STATE(!IsEmpty());

    NodeType* node = m_Head;
    m_Head         = m_Head->m_Next;
    if (IsEmpty ())    // Last node in the list was deleted
    {
        m_Last = m_Head;
    }

    --m_NodeCount;
    destroy(node, m_Allocator);
}

// -----------------------------------------------------------------------------

template <class T>
void Container_List<T>::Delete
    (Iterator&                      pos)
{
    SAPDBERR_ASSERT_ARGUMENT(pos.IsValid());
    SAPDBERR_ASSERT_ARGUMENT(pos != End());
    SAPDBERR_ASSERT_STATE(!IsEmpty());

    // -------------------------------------------------------------------------
    // check whether DeleteFront() can be used
    // -------------------------------------------------------------------------
    if (pos == Begin()) 
    {
        // Delete the first list element
        DeleteFront();
        // Set the iterator position.
        pos = Begin();
    }
    // -------------------------------------------------------------------------
    // delete element from the middle of the list
    // -------------------------------------------------------------------------
    else
    {
        // -------------------------------------------------------------------------
        // Find the iterator position before 'pos'
        // -------------------------------------------------------------------------
        Iterator prev = Begin();
        Iterator next = ++Begin();
        while (next != pos) 
        {
            SAPDBERR_ASSERT_STATE(next != End());      // error: 'pos' is not in list
            ++prev;
            ++next;
        }

        // -------------------------------------------------------------------------
        // chain the node out of the list
        // -------------------------------------------------------------------------
        prev.GetNode()->m_Next = pos.GetNode()->m_Next;
        if (pos == Last())
        {
            m_Last = prev.GetNode();
        }

        NodePointer p = pos.GetNode();
        destroy(p, m_Allocator);
        pos = prev.GetNode()->m_Next;
        --m_NodeCount;
    }
}

// -----------------------------------------------------------------------------

template <class T>
inline
void Container_List<T>::Delete()
{
   while (!IsEmpty()) 
   {
      DeleteFront();
   }
}

// -----------------------------------------------------------------------------

template <class T>
bool Container_List<T>::NodeCountConsistent() const
{
     ConstIterator iter    = Begin();
     ConstIterator enditer = End();
     SizeType nodecount    = 0;
     
     while (iter != enditer)
     {
          ++iter;
          ++nodecount;
     }
     return (nodecount == m_NodeCount);
}


// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_List<T>::CopyList
    (const List&                    list)
{
    // -------------------------------------------------------------------------
    // save actual list contents:
    // -------------------------------------------------------------------------
    NodeType* head      = m_Head;
    NodeType* last      = m_Last;
    SizeType  nodecount = m_NodeCount;

    // -------------------------------------------------------------------------
    // copy list elementwise:
    // -------------------------------------------------------------------------
    m_Head              = NodeType::InvalidNode;
    m_Last              = NodeType::InvalidNode;
    m_NodeCount         = 0;
    ConstIterator iter  = list.Begin();
    ConstIterator end   = list.End();
    while (iter != end)
    {
        if (!(InsertEnd(*iter).IsValid()))
        {
            Delete();
            m_Head      = head;
            m_Last      = last;
            m_NodeCount = nodecount;
            return false;
        }
        ++iter;
    }

    // -------------------------------------------------------------------------
    // delete old list contents:
    // -------------------------------------------------------------------------
    List del(m_Allocator);
    del.m_Head      = head;
    del.m_Last      = last;
    del.m_NodeCount = nodecount;
    del.Delete();
    return true;
}

// -----------------------------------------------------------------------------



#endif // CONTAINER_LIST_HPP

