/*!***************************************************************************

  module:		Container_NodeIterator.hpp

  ------------------------------------------------------------------------------

  author:       Stefan Baier, Till Luhmann

  responsible:  Till Luhmann

  special area: Standard container in SAPDB

  description:  Unidirectional iterator template for singly linked list nodes

  see also:     Container_DoubleNodeIterator.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------
 
  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef CONTAINER_NODEITERATOR_HPP
#define CONTAINER_NODEITERATOR_HPP


/*!*****************************************************************************

  class:        Container_NodeIterator

  description:  
  
  Unidirectional iterator template for singly linked list nodes

  This iterator may be used to iterate Container_List<T> 

  template arguments:    
  
  <class T>             Type of the elements of the list.
  <class ValueType>     Type that the dereference operator returns.
  <class NodePointer>   Type of the node pointer.

*******************************************************************************/
template <class T, class ValueType, class NodePointer>
class Container_NodeIterator
{

private:

    typedef Container_NodeIterator<T, ValueType, NodePointer> 
            Iterator;

public:

    /*!----------------------------------------------------------------------
    function:     Container_NodeIterator()
    description:  Constructor
    arguments:    node [in]   pointer to a list node
    return value: none
    prototypes:   2
    -----------------------------------------------------------------------*/
    Container_NodeIterator()
        : 
         m_Node(0) 
    {}

    Container_NodeIterator
        (NodePointer                    Node)
        : 
         m_Node(Node) 
    {}

    /*!-------------------------------------------------------------------------
    function:     operator*()
    description:  Dereferences node pointer
    arguments:    none
    return value: (ValueType&) reference to actual node content
    --------------------------------------------------------------------------*/
    ValueType& operator*() const
    {
        SAPDBERR_ASSERT_STATE(IsValid());
        return m_Node->m_Content;
    }

    /*!-------------------------------------------------------------------------
    function:     operator->()
    description:  Allows access to members of the node content
    arguments:    none
    return value: (ValueType*) pointer to node content
    --------------------------------------------------------------------------*/
#   if defined(_WIN32) || defined(WIN32)
#   pragma warning(disable : 4284)
#   endif
    ValueType* operator->() const
    {
        SAPDBERR_ASSERT_STATE(IsValid());
        return &(m_Node->m_Content);
    }
#   if defined(_WIN32) || defined(WIN32)
#   pragma warning(default : 4284)
#   endif

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  Prefix increment; advances iterator to next node
    arguments:    none
    return value: (Iterator&) *this
    --------------------------------------------------------------------------*/
    Iterator& operator++() 
    {
        SAPDBERR_ASSERT_STATE(IsValid());
        m_Node = m_Node->m_Next;
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     operator==()
    description:  Compares two iterators
    arguments:    iter [in]     iterator to be compared with
    return value: (bool) true if both iterators point to same node
    --------------------------------------------------------------------------*/
    bool operator==
        (const Iterator&                iter) const
    { 
        return m_Node == iter.m_Node; 
    }

    /*!-------------------------------------------------------------------------
    function:     operator!=()
    description:  Compares two iterators
    arguments:    iter [in]     iterator to be compared with
    return value: (bool) true if both iterators point to different nodes
    --------------------------------------------------------------------------*/
    bool operator!=
        (const Iterator&                iter) const
    { 
        return m_Node != iter.m_Node; 
    }

    /*--------------------------------------------------------------------------
    function:     IsValid()
    description:  Checks whether actual node is valid
    arguments:    none
    return value: (bool) true if valid
    --------------------------------------------------------------------------*/
    bool IsValid() const
    {
        return m_Node != 0;
    }

    /*--------------------------------------------------------------------------
    function:     GetNode()
    description:  Returns pointer to actual node
    arguments:    none
    return value: (NodePointer) pointer to actual node
    --------------------------------------------------------------------------*/
    const NodePointer& GetNode() const
    {
        return m_Node;
    }

protected:

    /*--------------------------------------------------------------------------
    declaration : m_Node
    description:  pointer to list node
    --------------------------------------------------------------------------*/
    NodePointer m_Node;
};
/*!*****************************************************************************

   endclass: Container_NodeIterator

*******************************************************************************/



#endif // CONTAINER_NODEITERATOR_HPP

