/*!******************************************************************************

  module:       Container_Stack.hpp

  ------------------------------------------------------------------------------

  responsible:  Stefan Baier

  special area: Standard container in SAPDB

  description:  Class declaration and definition for a template stack class

  ------------------------------------------------------------------------------

  copyright:                 Copyright by SAP AG, 1999



     ========== licence begin LGPL
     Copyright (C) 2000 SAP AG

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
     ========== licence end

*******************************************************************************/

#ifndef CONTAINER_STACK_HPP
#define CONTAINER_STACK_HPP

#include "Container/Container_Vector.hpp"


/*!****************************************************************************
  class:        Container_Stack

  description: This class implements a stack class.

               This implementation uses a vector as the underlying container 
               class. Elements can be pushed onto the stack and popped from the 
               stack. 

  template arguments:
               <class T> 
                   Type of the objects to be pushed onot the stack
                   The class T must provide the operator==
                        bool operator==(const T& t) const;
                   T doesn't have to have a default constructor. All constructing 
                   methods (constructors, Resize(), Reserve()) offer the 
                   possibility to pass an default value as an argument.

******************************************************************************/

template <class T>
class Container_Stack : private Container_Vector<T>
{
private:
    // typedef for abbreviation
    typedef Container_Vector<T> VectorType;
public:
    /*!----------------------------------------------------------------------
    declarations: Type definitions
    -----------------------------------------------------------------------*/
    typedef typename VectorType::ValueType ValueType;   // The type of the elements
    typedef typename VectorType::SizeType   SizeType;   // The range for the number of stack elements
    typedef typename VectorType::IndexType IndexType;   // The range for the index of the stack elements

    /*! enddeclarations: */

private:
    // The index of the next element that will be pushed onto the stack.
    IndexType m_NextIndex;

public:
    /*! chapter: Construction and destruction */

    /*!----------------------------------------------------------------------
    function: Constructors
    description: Construct a stack.

                 A stack can only be constructed, if an raw allocator is given,
                 thus a default constructor is not offered.
    arguments:   RawAlloc [in]
                 The raw allocator which provides the stack with memory.
    -----------------------------------------------------------------------*/
    explicit Container_Stack(SAPDBMem_IRawAllocator& RawAlloc);

    /*!----------------------------------------------------------------------
    function: Destructor
    description: Destructs the stack and frees the memory of the stack
                 elements.
    -----------------------------------------------------------------------*/
    ~Container_Stack() 
        { Delete(); }

    /*!----------------------------------------------------------------------
    function:    Initialize()
    description: Initializes a stack with a certain capacity.

    arguments:   Capacity [in]
                 The number of stack elements (must be greater or equal zero).
                 InitStack [in]
                 The initial values for the stack elements are copied from here.
    returns:     (SAPDB_Bool) true if successful
    prototypes:  2
    -----------------------------------------------------------------------*/
    SAPDB_Bool Initialize
         (SizeType Capacity);

    SAPDB_Bool Initialize
         (const Container_Stack<T>& InitStack);

    /*!----------------------------------------------------------------------
    function: Delete()
    description: Destructs the stack and frees the memory of the stack
                 elements.
    -----------------------------------------------------------------------*/
    void Delete()
        { VectorType::Delete(); }

    /*! endchapter: */

    /*! chapter: Status functions */

    /*!----------------------------------------------------------------------
    function: GetSize
    description: Returns the number of elements on the stack.
    -----------------------------------------------------------------------*/
    SizeType GetSize() const
        { return VectorType::GetSize(); }

    /*!----------------------------------------------------------------------
    function: IsEmpty
    description: Returns true, if there are no elements on the stack.
    -----------------------------------------------------------------------*/
    bool IsEmpty() const
        { return (GetSize() == 0); }

    /*!----------------------------------------------------------------------
    function: GetCapacity
    description: Returns the capacity of the stack.
    
                     The capacity is the number of stack elements that can be 
                     pushed onto the stack without reallocating memory.
    -----------------------------------------------------------------------*/
    SizeType GetCapacity() const
        { return VectorType::GetCapacity(); }

    /*!----------------------------------------------------------------------
    function: GetRawAllocator
    description: Returns the underlying raw allocator
    -----------------------------------------------------------------------*/
    SAPDBMem_IRawAllocator& GetRawAllocator() const
        { return VectorType::GetRawAllocator(); }

    /*!----------------------------------------------------------------------
    operator: ==
    description: Compares two stacks and returns true, if the content of both
                     stacks is identical.
    result value: bool
    -----------------------------------------------------------------------*/
    bool operator ==(const Container_Stack<T>& CompStack) const
        { return VectorType::operator ==(CompStack); }

    /*!----------------------------------------------------------------------
    operator: !=
    description: Compares two stacks and returns true, if the content of both
                     stacks is not identical.
    result value: bool
    -----------------------------------------------------------------------*/
    bool operator !=(const Container_Stack<T>& CompVector) const
        { return VectorType::operator !=(CompVector); }

    /*!----------------------------------------------------------------------
    function: Top
    description:  Returns a reference to the uppermost element on the stack.
    arguments:    none
    return value: T&
                  The uppermost element.
    -----------------------------------------------------------------------*/
    T& Top();

    /*! endchapter: */

    /*! chapter: Status manipulation functions */

    /*!----------------------------------------------------------------------
    function: Reserve
    description: Changes the capacity of the stack.

                 If the new capacity is greater than the actual size, the 
                 memory for the stack is reallocated.
                 ATTENTION!! After reallocation all iterators may be invalid!
    arguments:   NewCapacity [in]
                 The new number of vector elements.
    returns:     (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool Reserve(SizeType NewCapacity)
        { return VectorType::Reserve(NewCapacity); }

    /*! endchapter: */

    /*! chapter: Manipulating functions */

    /*!----------------------------------------------------------------------
    function: Push
    description: Pushes an element <Elem> onto the stack.

                 If the capacity of the stack is not sufficient to hold
                 another element, the capacity is enlarged as described
                 in method 'Reserve()'.
    arguments:   Elem [in]
    return value: (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool Push(const T& Elem);

    /*!----------------------------------------------------------------------
    function: Pop
    description:  Pops an element <Elem> from the stack.
    arguments:    none
    return value: (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool Pop();

    /*!----------------------------------------------------------------------
    function: Empty
    description:  Empties the stack without changing the capacity.
    arguments:    none
    return value: (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool Empty();

    /*! endchapter: */

};

/*! endclass: */


//********************************************************************************
// Implementation of constructors
//********************************************************************************

template <class T>
inline
Container_Stack<T>::Container_Stack(
    SAPDBMem_IRawAllocator& RawAlloc
)
 : VectorType(RawAlloc), m_NextIndex(0)
{}


template <class T>
inline
SAPDB_Bool Container_Stack<T>::Initialize(
    SizeType Capacity
)
{
    return Reserve(Capacity);
}


template <class T>
inline
SAPDB_Bool Container_Stack<T>::Initialize(
    const Container_Stack<T>& InitStack
)
{
    return VectorType::Initialize(InitStack);
}

//********************************************************************************
// Status functions 
//********************************************************************************

template <class T>
inline
T& Container_Stack<T>::Top()
{
    SAPDBERR_ASSERT_STATE(!IsEmpty());
    // Return the uppermost element.
    return operator[](m_NextIndex-1);
}


//********************************************************************************
// Manipulation functions 
//********************************************************************************

template <class T>
SAPDB_Bool Container_Stack<T>::Push(const T& Elem)
{
    if (!(GetSize() < GetCapacity())) {
        if (GetCapacity() < 4) {
            // Reserve space for at least four stack elements.
            if (!Reserve(4)) return false;
        }
        else {
            // Enlarge the capacity by 50%.
            // This is done to avoid reallocating memory to often.
            if (!Reserve(GetCapacity()+GetCapacity()/2)) return false;
        }
    }
    // Increment the size of the vector by one.
    if (!Resize(m_NextIndex+1)) return false;
    // Insert the new element.
    operator[](m_NextIndex++) = Elem;
    return true;
}


template <class T>
inline
SAPDB_Bool Container_Stack<T>::Pop()
{
    SAPDBERR_ASSERT_STATE(!IsEmpty());
    // Decrement the size of the vector by one.
    --m_NextIndex;
    return Resize(m_NextIndex);
}


template <class T>
inline
SAPDB_Bool Container_Stack<T>::Empty()
{
    m_NextIndex = 0;
    return this->Resize(0);
}


#endif // CONTAINER_STACK_HPP

