/*!******************************************************************************

  module:       Container_Vector.hpp

  ------------------------------------------------------------------------------

  author:       Stefan Baier, Till Luhmann

  responsible:  Stefan Baier

  special area: Standard container in SAPDB

  description:  Class declaration and definition for a template vector class

  see also:     

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin LGPL
  Copyright (C) 2000 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*******************************************************************************/

#ifndef CONTAINER_VECTOR_HPP
#define CONTAINER_VECTOR_HPP

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"

#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"


/*!****************************************************************************
  class:        Container_Vector

  description: This class implements a vector class.

               The vector uses raw allocators as defined in the SAPDBMem component
               for allocating memory for the vector elements. It is not possible 
               to construct a vector without supplying a raw allocator.
               Thus, the default constructor for a vector is not offered.

  template arguments:
               <class T> 
                   Type of the objects to be held in vector.
                   The class T must provide the operator==
                       SAPDB_Bool operator==(const T& t) const;
                   T doesn't have to have a default constructor. All constructing 
                   methods (constructors, Resize(), Reserve()) offer the 
                   possibility to pass an default value as an argument.
                                  
  example:     This example constructs an vector of type 'int' by using
               a 'SAPDBMem_DefaultRawAllocator'.

               void f()
               {
                   // ...

                   // Construct an vector of type 'int' with ten elements using 
                   // a default raw allocator.
                   Container_Vector<int>  MyVector(SAPDBMem_DefaultRawAllocator::Instance(), 10);

                   // ...
               }
******************************************************************************/

template <class T>
class Container_Vector
{

public:

    /*!----------------------------------------------------------------------
    declarations: Type definitions
    -----------------------------------------------------------------------*/
    typedef T            ValueType;     // The type of the elements
    typedef SAPDB_UInt   SizeType;      // The range for the number of vector elements
    typedef SAPDB_UInt   IndexType;     // The range for the index of the vector elements
    typedef T*           Iterator;      // Iterator over vector elements
    typedef const T*     ConstIterator; // Constant Iterator
    typedef SAPDBMem_IRawAllocator Allocator; // Allocator type used for vectors

    /*! enddeclarations: */

private:
    // The underlying raw allocator.
    Allocator& m_Allocator;
    // The number of vector elements that can be stored without resizing the vector
    SizeType m_Capacity;

protected:	// used in ..sortableVector
    // The number of vector elements
    SizeType m_Size;
    // The pointer to the vector elements
    T* m_ArrayPtr;

public:
    /*! chapter: Construction and destruction */

    /*!----------------------------------------------------------------------
    function: Container_Vector()
    description: Constructs a vector

                 A vector can only be constructed, if a raw allocator is given.
                 Thus, a default constructor is not offered.
    arguments:   RawAlloc [in]
                 The raw allocator providing the vector with memory.
    -----------------------------------------------------------------------*/
    Container_Vector(Allocator& RawAlloc);

    /*!----------------------------------------------------------------------
    function:    Initialize()
    description: Initializes a vector with a certain number of elements

    arguments:   Size [in]
                 The number of vector elements (must be greater or equal zero).
                 Init [in]
                 The initial value of all vector elements. By default,
                 the default constructor of <T> is used as the init value.
                 InitVector [in]
                 The initial values for the vector elements are copied from here.
    returns:     (SAPDB_Bool) true if successful
    prototypes:  3
    -----------------------------------------------------------------------*/
    SAPDB_Bool Initialize
        (SizeType Size);

    SAPDB_Bool Initialize
        (SizeType Size,
         const T& Init);

    SAPDB_Bool Initialize
        (const Container_Vector<T>& InitVector);

    /*!----------------------------------------------------------------------
    function: Clear()
    description: Destructs the vector elements. Preserves the capacity.
    -----------------------------------------------------------------------*/
    void Clear();

    /*!----------------------------------------------------------------------
    function: Delete()
    description: Destructs the vector and frees the memory of the vector 
                 elements.
    -----------------------------------------------------------------------*/
    void Delete();

    /*!----------------------------------------------------------------------
    function: ~Container_Vector()
    description: Destructs the vector and frees the memory of the vector 
                 elements.
    -----------------------------------------------------------------------*/
    ~Container_Vector();

    /*! endchapter: */

    /*! chapter: Status functions */

    /*!----------------------------------------------------------------------
    function: GetSize()
    description: Returns the number of vector elements within the vector
    -----------------------------------------------------------------------*/
    SizeType GetSize() const
        { return m_Size; }

    /*!----------------------------------------------------------------------
    function: IsEmpty()
    description: Returns true, if the number of vector elements is zero
    -----------------------------------------------------------------------*/
    SAPDB_Bool IsEmpty() const
        { return (m_Size == 0); }

    /*!----------------------------------------------------------------------
    function: IsFull()
    description: Returns true, if the number of vector elements is equal
                 to the actual capacity 
    -----------------------------------------------------------------------*/
    SAPDB_Bool IsFull() const
        { return (m_Size == m_Capacity); }

    /*!----------------------------------------------------------------------
    function: GetCapacity
    description: Returns the capacity of the vector.
    
                 The capacity is the number of vector elements that can be 
                 given as an argument to 'Resize()' without reallocating memory.
    -----------------------------------------------------------------------*/
    SizeType GetCapacity() const
        { return m_Capacity; }

    /*!----------------------------------------------------------------------
    function: GetAllocator
    description: Returns the underlying raw allocator
    -----------------------------------------------------------------------*/
    Allocator& GetAllocator() const
        { return m_Allocator; }

    /*!----------------------------------------------------------------------
    operator: ==
    description: Compares two vectors and returns true, if the content of both
                 vectors is identical.
    -----------------------------------------------------------------------*/
    SAPDB_Bool operator ==(const Container_Vector<T>& CompVector) const;

    /*!----------------------------------------------------------------------
    operator: !=
    description: Compares two vectors and returns true, if the content of both
                 vectors is not identical.
    -----------------------------------------------------------------------*/
    SAPDB_Bool operator !=(const Container_Vector<T>& CompVector) const
        { return !(this->operator==(CompVector)); }

    /*! endchapter: */

    /*! chapter: Status manipulation functions */

    /*!----------------------------------------------------------------------
    function: Resize()
    description: Changes the number of vector elements.

                 If the capacity of the vector is not large enough, the 
                 memory for the vector elements is reallocated.
                 
                 ATTENTION!! After reallocation all iterators may be invalid!
                 
    arguments:   NewSize [in]
                 The new number of vector elements.
                 Init [in]
                 An initialization value for the new vector elements.
                 If <Init> is not given, the default constructor of <T> is
                 used as initialization value.
    returns:     (SAPDB_Bool) true if successful
    prototypes:  2
    -----------------------------------------------------------------------*/
    SAPDB_Bool Resize(SizeType NewSize);

    SAPDB_Bool Resize(SizeType NewSize, const T& Init);

    /*!----------------------------------------------------------------------
    function: Reserve()
    description: Changes the capacity of the vector.

                 If the new capacity is greater than the actual capacity, memory for
                 the vector elements is reallocated.
                 
                 ATTENTION!! After reallocation all iterators may be invalid!
                             The newly reserved space is not initialized.
                 
    arguments:   NewCapacity [in]
                 The new number of vector elements.
    returns:     (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool Reserve(SizeType NewCapacity);

    /*! endchapter: */

    /*! chapter: Manipulating functions */

    /*!----------------------------------------------------------------------
    operator: [n]
    description: Returns a reference to the <n>-th vector element.
    prototypes:  2
    -----------------------------------------------------------------------*/
    T& operator[](IndexType n)
    {
        SAPDBERR_ASSERT_ARGUMENT(n < m_Size);
        return *(m_ArrayPtr + n);
    }

    const T& operator[](IndexType n) const
    {
        SAPDBERR_ASSERT_ARGUMENT(n < m_Size);
        return *(m_ArrayPtr + n);
    }

    /*!----------------------------------------------------------------------
    function: Position()
    description: Returns an iterator to the <n>-th vector element.
    prototypes:  2
    -----------------------------------------------------------------------*/
    Iterator Position(IndexType n)
    {
        SAPDBERR_ASSERT_ARGUMENT(n < m_Size);
        return m_ArrayPtr + n;
    }

    ConstIterator Position(IndexType n) const
    {
        SAPDBERR_ASSERT_ARGUMENT(n < m_Size);
        return m_ArrayPtr + n;
    }

    /*!----------------------------------------------------------------------
    function: InsertFront()
    description: Inserts an element <Elem> at the beginning of the vector.

                 The size of the vector grows by one.
                 There must be enough capacity to do this.
                 All vector elements change their index by +1.
                 Thus, all iterators are invalid after a call to 'InsertFront'.
                 This method is rather slow, because all vector elements are
                 moved by one position.
    arguments:   Elem [in]
                 The element to insert.
    -----------------------------------------------------------------------*/
    void InsertFront(const T& Elem);

    /*!----------------------------------------------------------------------
    function: InsertEnd()
    description: Inserts an element <Elem> at the end of the vector.

                 The size of the vector grows by one.
                 There must be enough capacity to do this.
                 All iterators remain valid after a call to 'InsertEnd'.
    arguments:   Elem [in]
                 The element to insert.
    returns:     true if the possibly necessary reallocation was successfull.
    -----------------------------------------------------------------------*/
    SAPDB_Bool InsertEnd(const T& Elem);

    /*!----------------------------------------------------------------------
    function: Insert()
    description: Inserts a new element at the iterator position <Pos>.

                 The size of the vector grows by one.
                 There must be enough capacity to do this.
                 All vector elements after <Pos> change their index by +1.
                 Some iterators are invalid after a call to 'Insert'.
                 This method is rather slow, because all vector elements after
                 <Pos> are moved by one position.
    arguments:   Pos [in]
                 Iterator position
                 Elem [in]
                 The element to insert.
    -----------------------------------------------------------------------*/
    void Insert(Iterator& Pos, const T& Elem);

    /*!----------------------------------------------------------------------
    function: DeleteFront()
    description: Deletes the first element of the vector.

                 The vector must not be empty.
    -----------------------------------------------------------------------*/
    void DeleteFront();

    /*!----------------------------------------------------------------------
    function: Delete()
    description:  Deletes one vector element or a range of vector elements.
    
                  The vector must not be empty. 
                  If only one iterator is given as an argument, this element is 
                  deleted. If a range is given, the range is deleted.
    arguments:    Pos [in]
                  Iterator to one element.
                  From [in]
                  Iterator to the begin of the range.
                  To [in]
                  Iterator to the end of the range.
    return value: Iterator
                  The iterator points to the vector element behind the last
                  one deleted.
    prototypes:   2
    -----------------------------------------------------------------------*/
    void Delete(const Iterator& Pos);
    void Delete(const Iterator& From, const Iterator& To);

    /*! endchapter: */


    /*! chapter: Iterator functions */

    /*!----------------------------------------------------------------------
    function: Begin()
    description: Returns an iterator to the first vector element.
    prototypes:  2
    -----------------------------------------------------------------------*/
    Iterator Begin()
    { 
        SAPDBERR_ASSERT_STATE(m_ArrayPtr != 0 || (m_Capacity == 0));
        return m_ArrayPtr; 
    }

    ConstIterator Begin() const
    { 
        SAPDBERR_ASSERT_STATE(m_ArrayPtr != 0 || (m_Capacity == 0));
        return m_ArrayPtr; 
    }

    /*!----------------------------------------------------------------------
    function: End()
    description: Returns an iterator behind the last vector element.
    prototypes:  2
    -----------------------------------------------------------------------*/
    Iterator End()
    { 
        SAPDBERR_ASSERT_STATE((m_ArrayPtr != 0) || (m_Capacity == 0));
        return m_ArrayPtr + m_Size; 
    }

    ConstIterator End() const
    { 
        SAPDBERR_ASSERT_STATE((m_ArrayPtr != 0) || (m_Capacity == 0));
        return m_ArrayPtr + m_Size; 
    }


    /*!----------------------------------------------------------------------
    function: GetIndex()
    description: Returns the index that belongs to an iterator
    return value: IndexType
    -----------------------------------------------------------------------*/
    IndexType GetIndex(const Iterator& Iter) const
    {
        SAPDBERR_ASSERT_STATE(m_ArrayPtr != 0);
        // current impl. assumes that iter is a pointer
        SAPDBERR_ASSERT_ARGUMENT(Iter >= m_ArrayPtr);
        SAPDBERR_ASSERT_ARGUMENT(Iter <= m_ArrayPtr + m_Size);
        return (IndexType)(Iter - m_ArrayPtr);
    }

    /*! endchapter: */
};

/*! endclass: */

// -----------------------------------------------------------------------------

template <class T>
inline
Container_Vector<T>::Container_Vector(
    Allocator& RawAlloc
)
 : m_Allocator(RawAlloc), m_Size(0), m_Capacity(0), m_ArrayPtr(0)
{}

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_Vector<T>::Initialize(SizeType Size)
{
    SAPDBERR_ASSERT_ARGUMENT(Size > 0);

    T* NewArrayPtr = 0; 
    if (Size > 0)
    {
        // allocate and initialize new vector:
        newarray(NewArrayPtr, Size, m_Allocator);
        if (NewArrayPtr == 0) 
            return false;
    }

    // delete old vector:
    Delete();

    // initialize vector:
    m_ArrayPtr = NewArrayPtr;
    m_Capacity = Size;
    m_Size     = Size;
    return true;
}

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_Vector<T>::Initialize(
    SizeType Size,
    const T& Init
)
{
    SAPDBERR_ASSERT_ARGUMENT(Size > 0);

    T* NewArrayPtr = 0; 
    if (Size > 0)
    {
        // allocate and initialize new vector:
        newarray(NewArrayPtr, Size, m_Allocator, Init);
        if (NewArrayPtr == 0) 
            return false;
    }

    // delete old vector:
    Delete();

    // initialize vector:
    m_ArrayPtr = NewArrayPtr;
    m_Capacity = Size;
    m_Size     = Size;
    return true;
}

// -----------------------------------------------------------------------------

template <class T>
inline
SAPDB_Bool Container_Vector<T>::Initialize(
    const Container_Vector<T>& InitVector
)
{
    SizeType Capacity = InitVector.GetCapacity();
    SizeType Size     = InitVector.GetSize();

    T* NewArrayPtr = 0; 
    if (Capacity > 0) 
    {
        // Construct the new vector elements from InitVector
        newarray(NewArrayPtr, Capacity, m_Allocator, InitVector.m_ArrayPtr);
        if (NewArrayPtr == 0) 
            return false;
    }

    // delete old vector:
    Delete();

    // Everything worked ok.
    m_ArrayPtr = NewArrayPtr;
    m_Size     = Size;
    m_Capacity = Capacity;
    return true;
}

// -----------------------------------------------------------------------------

template <class T>
inline
void Container_Vector<T>::Clear()
{
    if (m_Size > 0) 
    {
        // Destruct all used array elements.
        T* tmp = m_ArrayPtr;
        destroyarray(tmp, m_Size, m_ArrayPtr);
    }
    m_Size     = 0;
}

// -----------------------------------------------------------------------------

template <class T>
inline
void Container_Vector<T>::Delete()
{
    Clear();
    if (m_ArrayPtr != 0) 
    {
        // Deallocate the memory for the reserved capacity
        m_Allocator.Deallocate(m_ArrayPtr);
    }
    m_ArrayPtr = 0;
    m_Capacity = 0;
}

// -----------------------------------------------------------------------------

template <class T>
inline
Container_Vector<T>::~Container_Vector()
{
    Delete();
}

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_Vector<T>::operator ==(const Container_Vector<T>& CompVector) const
{
    SAPDB_Bool Result = false;

    // Compare the sizes of both vectors
    if (m_Size == CompVector.GetSize()) {
        // If the sizes match, the result is true even if the size is zero
        Result = true;
        if (m_Size > 0) {
            // if the size is greater than zero, compare the vectors element by element
            ConstIterator Iter1, Iter2;
            Iter2 = CompVector.Begin();
            for (Iter1 = Begin(); Iter1 != End(); ++Iter1) {
                if (!(*Iter1 == *Iter2)) {
                    // If two elements are not equal, the result is false.
                    // Then break the loop.
                    Result = false;
                    break;
                }
                ++Iter2;
            }
        }
    }
    return Result;
}

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_Vector<T>::Resize(SizeType NewSize)
{
    T* tmp;

    if (NewSize < m_Size) {
        // Destroy the unused part of the vector.
        tmp = m_ArrayPtr + NewSize;
        destroyarray(tmp, m_Size-NewSize, tmp);
    }
    else {
        if (NewSize > m_Capacity) {
            if (!Reserve(NewSize))
                return false;
        }
        // Initialize the new part of the vector.
        if ( NewSize > m_Size ) // PTS 1117895 UH 2002-10-18
            newarray(tmp, NewSize-m_Size, m_ArrayPtr+m_Size);
    }
    m_Size = NewSize;
    return true;
}

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_Vector<T>::Resize(SizeType NewSize, const T& Init)
{
    T* tmp;

    if (NewSize < m_Size) {
        // Destroy the unused part of the vector.
        tmp = m_ArrayPtr + NewSize;
        destroyarray(tmp, m_Size-NewSize, tmp);
    }
    else {
        if (NewSize > m_Capacity) {
            if (!Reserve(NewSize))
                return false;
        }
        // Initialize the new part of the vector.
        if ( NewSize > m_Size ) // PTS 1117895 UH 2002-10-18
            newarray(tmp, NewSize-m_Size, m_ArrayPtr+m_Size, Init);
    }
    m_Size = NewSize;
    return true;
}

// -----------------------------------------------------------------------------

template <class T>
SAPDB_Bool Container_Vector<T>::Reserve(SizeType NewCapacity)
{
    // If the reserved capacity is sufficient, do nothing.
    // The capacity never becomes smaller.
    if (NewCapacity <= m_Capacity) {
        return true;
    }
    // Give m_ArrayPtr as a hint to try to reallocate memory at the hint position.
    T* NewArrayPtr = reinterpret_cast<T*>(m_Allocator.Allocate(NewCapacity * sizeof(T), m_ArrayPtr));

    if (NewArrayPtr == 0) 
        { return false; }

    if (NewArrayPtr != m_ArrayPtr) {
        if (m_ArrayPtr != 0 && m_Size != 0) {
            // If the reallocation at the hint position didn't work...
            // ...initialize the new vector elements with the old ones,
            newarray(NewArrayPtr, m_Size, NewArrayPtr, m_ArrayPtr);
            // ...and destroy the old vector.
            destroyarray(m_ArrayPtr, m_Size, m_Allocator);
        }
        // Assign the new member.
        m_ArrayPtr = NewArrayPtr;
    }
    m_Capacity = NewCapacity;

    return true;
}

// -----------------------------------------------------------------------------

template <class T>
inline
void Container_Vector<T>::InsertFront(const T& Elem)
{
    Iterator Iter = Begin();
    Insert(Iter, Elem);
}

// -----------------------------------------------------------------------------

template <class T>
inline
SAPDB_Bool Container_Vector<T>::InsertEnd(const T& Elem)
{
    // There must be at least space for one more element.
    SAPDBERR_ASSERT_ARGUMENT(m_Size < m_Capacity);
    // Let the size grow by one and insert the element at the end
    return Resize(m_Size+1, Elem);
}

// -----------------------------------------------------------------------------

template <class T>
void Container_Vector<T>::Insert(Iterator& Pos, const T& Elem)
{
    // There must be at least space for one more element.
    SAPDBERR_ASSERT_ARGUMENT(m_Size < m_Capacity);
    // Let the size grow by one
    Resize(m_Size+1, Elem);
    // Calculate the index of iterator <Pos>
    IndexType PosIndex = GetIndex(Pos);
    // Shift all vector elements from <Pos> on one position to the end.
    for (IndexType i=m_Size-1; i>PosIndex; --i) {
        operator[](i) = operator[](i-1);
    }
    // Insert the new element
    operator[](PosIndex) = Elem;
    // Set the new iterator position.
    Pos = Position(PosIndex);
}

// -----------------------------------------------------------------------------

template <class T>
inline
void Container_Vector<T>::DeleteFront()
{
    Delete(Begin());
}

// -----------------------------------------------------------------------------

template <class T>
void Container_Vector<T>::Delete(const Iterator& Pos)
{
    SAPDBERR_ASSERT_ARGUMENT(!IsEmpty());
    SAPDBERR_ASSERT_ARGUMENT(Pos != End());
    // Calculate the index of iterator <Pos>
    IndexType PosIndex = GetIndex(Pos);
    // Shift all vector elements behind <Pos> one position.
    for (IndexType i=PosIndex+1; i<m_Size; ++i) {
        operator[](i-1) = operator[](i);
    }
    // Let the size shrink by one.
    // *Begin() is necessary to work with classes that don't provide a default constructor
    Resize(m_Size-1, *Begin());
}

// -----------------------------------------------------------------------------

template <class T>
void Container_Vector<T>::Delete(const Iterator& From, const Iterator& To)
{
    SAPDBERR_ASSERT_ARGUMENT(!IsEmpty());
    SAPDBERR_ASSERT_ARGUMENT(From != End());
    SAPDBERR_ASSERT_ARGUMENT(To != End());
    // Calculate the index of iterator positions.
    IndexType FromIndex = GetIndex(From);
    IndexType ToIndex = GetIndex(To);
    // Shift all vector elements behind <To> to <From>
    for (IndexType i=ToIndex+1; i<m_Size; ++i) {
        operator[](FromIndex+i-(ToIndex+1)) = operator[](i);
    }
    // Let the size shrink.
    // *Begin() is necessary to work with classes that don't provide a default constructor
    Resize(m_Size-(ToIndex-FromIndex+1), *Begin());
}

// -----------------------------------------------------------------------------



#endif // CONTAINER_VECTOR_HPP

