/*!**************************************************************************

  module      : Converter_BackUpPageNoContainer.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author:       AlexanderK

  special area: Converter

  description : Container which holds all page numbers that belong 
                to a back up 

  see also    : 

  copyright:    Copyright by SAP AG, 2001

  -------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/

#ifndef CONVERTER_BACKUPPAGENOCONTAINER_HPP
#define CONVERTER_BACKUPPAGENOCONTAINER_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "DataAccess/Data_Types.hpp"
#include "Converter/Converter_Types.hpp"
#include "Container/Container_Vector.hpp"
#include "Converter/Converter_FlagInfo.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"

/*!*****************************************************************************
 
   class: Converter_BackUpPageNoContainer
   
*******************************************************************************/

class Converter_BackUpPageNoContainer
{

public:

	/*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes container structures; deallocates memory

    arguments:    none                    
    return value: none
    --------------------------------------------------------------------------*/
    void Delete()
    {
        m_FlagInfoPtrVector.Delete();
        m_FlagInfoVector.Delete();
        m_NumUsedFlagInfoBlocks = 0;
		m_NumFlagInfoBlocks     = 0;
        m_PageNoCount           = 0;
    }

    /*!-------------------------------------------------------------------------
    function:     Converter_BackUpPageNoContainer()
    description:  Constructor
                    
    arguments:    alloc     [in]  allocator for container resources
    return value: none
    --------------------------------------------------------------------------*/
    Converter_BackUpPageNoContainer  (SAPDBMem_IRawAllocator&  Allocator)
    : 
     m_FlagInfoPtrVector(Allocator),
     m_FlagInfoVector(Allocator)
    {
		 Delete();
	}

    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  Initializes container
	
	arguments:    MaxPageNo                  [in] maximum page number to be stored in the container
                  NumFlagInfoBlocks          [in]   
				  NumPageNosPerFlagInfoBlock [in]  

	return value: (Converter_ReturnCode) if the memory for the container cannot be allocated 
	                                     the value Converter_MemoryAllocationFailed is returned 
										 otherwise Converter_Ok
    --------------------------------------------------------------------------*/
	Converter_ReturnCode  Initialize (SAPDB_UInt  MaxPageNo,
	                		SAPDB_UInt  NumFlagInfoBlocks,
			                SAPDB_UInt  NumPageNosPerFlagInfoBlock);

	/*!-------------------------------------------------------------------------
    function:     Insert()
    description:  inserts a page number into the container

    arguments:    PageNo                  page number to be stored in the container                    
    return value: (Converter_ReturnCode)  in case the page number is already included
	                                      the value Converter_DuplicatePageNo is returned
										  otherwise Converter_Ok.
    --------------------------------------------------------------------------*/
    Converter_ReturnCode Insert(Data_PageNo  PageNo);

	/*!-------------------------------------------------------------------------
    function:     IsEmpty()
    description:  Checks whether container is empty

	arguments:    none
	return value: (SAPDB_Bool) true if empty
    --------------------------------------------------------------------------*/
	SAPDB_Bool IsEmpty() const
    {
        return m_PageNoCount == 0;
    }
	
	/*!-------------------------------------------------------------------------
    function:     GetPageNoCount()
    description:  Returns total number of pagenos stored in container

	arguments:    none
	return value: (SAPDB_UInt) number of pagenos stored in container
    --------------------------------------------------------------------------*/
	SAPDB_UInt GetPageNoCount() const
    {
        return m_PageNoCount;
    }
	
	/*!-------------------------------------------------------------------------
    function:     Remove()
    description:  Removes a page number from the container
	
	arguments:    pageno [in] page number to be removed
    return value: (SAPDB_Bool) true if pageno found in container and removed
    --------------------------------------------------------------------------*/
    SAPDB_Bool Remove (Data_PageNo  PageNo);

private: // methods

    /*--------------------------------------------------------------------------
    function:     IsInitialized()
    description:  Checks whether container is initialized
    arguments:    none
    return value: (SAPDB_Bool) true if initialized
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsInitialized() const
    {
        return (m_FlagInfoPtrVector.GetSize() > 0) && (m_FlagInfoVector.GetSize() > 0) ;
    }

private: // members

    Container_Vector<Converter_FlagInfo::FlagInfoData *>    m_FlagInfoPtrVector;
    Container_Vector<Converter_FlagInfo::FlagInfoData>      m_FlagInfoVector;
    RTESync_InterlockedCounter<SAPDB_UInt>                  m_PageNoCount;
    SAPDB_UInt                                              m_NumUsedFlagInfoBlocks;
	SAPDB_UInt                                              m_NumFlagInfoBlocks;
	SAPDB_UInt                                              m_NumPageNosPerFlagInfoBlock;
	SAPDB_UInt                                              m_SizeOfFlagInfoBlock;
};

/*!*****************************************************************************

  endclass: Converter_BackUpPageNoContainer
  
*******************************************************************************/



#endif  /* CONVERTER_BACKUPPAGENOCONTAINER_HPP */

