/*!***************************************************************************

module:       Converter_Chain.hpp

------------------------------------------------------------------------------

responsible :  TorstenS

author      : TillL

special area: Converter

description :  Doubly linked chain of entries

              - Used in converter map and converter index.
              - Does not perform memory allocation.

see also:

------------------------------------------------------------------------------

copyright:            Copyright by SAP AG, 2000



  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_CHAIN_HPP
#define CONVERTER_CHAIN_HPP


#include "SAPDBCommon/SAPDB_Types.hpp"


/*!*****************************************************************************

   class: Converter_ChainNode

   description: Node of converter chain

                - Arguments: T         link type
                                      (must be derived from template
                                       SAPDB_Invalidatable<>)

*******************************************************************************/
template <class T>
class Converter_ChainNode
{
public:

    typedef T LinkType;

    /*!-------------------------------------------------------------------------
    function:     Converter_ChainNode()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Converter_ChainNode()
    {}

    /*!-------------------------------------------------------------------------
    function:     SetSuccessor()
    description:  Sets successor
    arguments:    succ [in]     successor to be set
    return value: none
    --------------------------------------------------------------------------*/
    void SetSuccessor( const LinkType   &succ )
    {
#ifdef SAPDB_QUICK
        if(( m_Successor.IsValid()) && ( succ.IsValid()))
            SAPDBERR_ASSERT_STATE(m_Successor != succ );
#endif
        m_Successor = succ;
    }

    /*!-------------------------------------------------------------------------
    function:     SetPredecessor()
    description:  Sets predecessor
    arguments:    pred [in]     predecessor to be set
    return value: none
    --------------------------------------------------------------------------*/
    void SetPredecessor( const LinkType &pred )
    {
        m_Predecessor = pred;
    }

    /*!-------------------------------------------------------------------------
    function:     GetPredecessor()
    description:  Returns predecessor
    arguments:    none
    return value: (const LinkType&) link to predecessor
    --------------------------------------------------------------------------*/
    const LinkType& GetPredecessor() const
    {
        return m_Predecessor;
    }

    /*!-------------------------------------------------------------------------
    function:     GetSuccessor()
    description:  Returns successor
    arguments:    none
    return value: (const LinkType&) link to successor
    --------------------------------------------------------------------------*/
    const LinkType& GetSuccessor() const
    {
        return m_Successor;
    }

    /*!-------------------------------------------------------------------------
    function:     Invalidate()
    description:  Assigns an invalid value
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Invalidate()
    {
        m_Predecessor.Invalidate();
        m_Successor.Invalidate();
    }

    /*!-------------------------------------------------------------------------
    function:     IsInvalid()
    description:  Checks whether actual node is invalid
    arguments:    none
    return value: (SAPDB_Bool) true if invalid
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsInvalid() const
    {
        return m_Predecessor.IsInvalid() && m_Successor.IsInvalid();
    }

private:

    /*--------------------------------------------------------------------------
    declaration : m_Predecessor
    description:  link to preceding node
    --------------------------------------------------------------------------*/
    LinkType                        m_Predecessor;
    /*--------------------------------------------------------------------------
    declaration : m_Successor
    description:  link to successive node
    --------------------------------------------------------------------------*/
    LinkType                        m_Successor;

};
/*!*****************************************************************************

   endclass: Converter_ChainNode

*******************************************************************************/



/*!*****************************************************************************

   class: Converter_ChainIteratorTPL

   description: (do not use directly)

                - Arguments: NodeContainer  container of nodes
                                            (must provide a method
                                             NodeType& GetNode(const LinkType&)
                                             returning the node associated with
                                             a given link)
                             NodeType       node type
                                            (must be derived from template
                                             Converter_ChainNode<>)

*******************************************************************************/
template <class NodeSequence, class NodeType>
class Converter_ChainIteratorTPL
{
public:


    typedef typename NodeType::LinkType             LinkType;
    typedef Converter_ChainIteratorTPL
    <NodeSequence, NodeType>                ChainIterator;

    /*!-------------------------------------------------------------------------
    function:     Converter_ChainIteratorTPL()
    description:  Constructor
    arguments:    nodes  [in]   reference to node sequence
                  link   [in]   reference to link in the chain
    return value: none
    --------------------------------------------------------------------------*/
    Converter_ChainIteratorTPL
    (NodeSequence&                  nodes,
     const LinkType&                link)
    :
    m_pNodes(&nodes),
    m_Pos(link)
    {}

    /*!-------------------------------------------------------------------------
    function:     operator!=()
    description:  Compares with other iterator
    arguments:    iter [in]    other
    return value: (SAPDB_Bool) true if unequal
    --------------------------------------------------------------------------*/
    SAPDB_Bool operator!=(const ChainIterator& iter) const
    {
        SAPDBERR_ASSERT_ARGUMENT(m_pNodes == iter.m_pNodes);
        return m_Pos != iter.m_Pos;
    }

    /*!-------------------------------------------------------------------------
    function:     operator*()
    description:  Dereferences the iterator
    arguments:    none
    return value: (NodeType&) reference to actual node
    --------------------------------------------------------------------------*/
    NodeType& operator*() const
    {
        SAPDBERR_ASSERT_STATE(m_Pos.IsValid());
        return (*m_pNodes)[m_Pos];
    }

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  Prefix increment; advances iterator
    arguments:    none
    return value: (ChainIterator&) this instance
    --------------------------------------------------------------------------*/
    ChainIterator& operator++()
    {
        SAPDBERR_ASSERT_STATE(m_Pos.IsValid());
        m_Pos = (*m_pNodes)[m_Pos].GetSuccessor();
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     operator--()
    description:  Prefix decrement; pushes back iterator
    arguments:    none
    return value: (ChainIterator&) this instance
    --------------------------------------------------------------------------*/
    ChainIterator& operator--()
    {
        SAPDBERR_ASSERT_STATE(m_Pos.IsValid());
        m_Pos = (*m_pNodes)[m_Pos].GetPredecessor();
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     GetPosition()
    description:  Returns actual link
    arguments:    none
    return value: (LinkType) actual link
    --------------------------------------------------------------------------*/
    LinkType GetPosition() const
    {
        return m_Pos;
    }

protected:
    /*--------------------------------------------------------------------------
    declaration : m_Nodes
    description:  link to successive node
    --------------------------------------------------------------------------*/
    NodeSequence*                  m_pNodes;

    /*--------------------------------------------------------------------------
    declaration : m_Anchor
    description:  link to successive node
    --------------------------------------------------------------------------*/
    LinkType                        m_Pos;
};
/*!*****************************************************************************

   endclass: Converter_ChainIteratorTPL

*******************************************************************************/



/*!*****************************************************************************

   class: Converter_Chain

   description: Node of converter chain

                - Arguments: see Converter_ChainIterator

*******************************************************************************/
template <class NodeSequence, class NodeType>
class Converter_Chain
{
public:

    typedef typename NodeType::LinkType                         LinkType;
    typedef SAPDB_UInt                                          NodeCount;
    typedef Converter_ChainIteratorTPL
    <NodeSequence, NodeType>                            Iterator;
    typedef Converter_ChainIteratorTPL
    <NodeSequence, const NodeType>                      ConstIterator;

    /*!-------------------------------------------------------------------------
    function:     Converter_Chain()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Chain()
    :
    m_Nodes(0),            // m_Head, m_Tail auto-initialize to "invalid"
    m_NodeCount(0)
    {}

    /*!-------------------------------------------------------------------------
    function:     Converter_Chain()
    description:  Constructor
    arguments:    nodes  [in]   reference to node sequence
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Chain( NodeSequence   &nodes )
    :
    m_Nodes(&nodes),
    m_Head(LinkType()),
    m_Tail(LinkType()),
    m_NodeCount(0)
    {}

    /*!-------------------------------------------------------------------------
    function:     Assign()
    description:  Assigns new sequence
    arguments:    nodes  [in]   reference to node sequence
    return value: none
    --------------------------------------------------------------------------*/
    void Assign( NodeSequence   &nodes )
    {
        m_Nodes = &nodes;
        Reset();
    }

    /*!-------------------------------------------------------------------------
    function:     Reset()
    description:  Initializes head and tail; sets node count to 0
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Reset()
    {
        m_Head      = LinkType();
        m_Tail      = LinkType();
        m_NodeCount = 0;
    }

    /*!-------------------------------------------------------------------------
    function:     InsertFront()
    description:  Inserts a node at the beginning of the chain
    arguments:    link [in]     link to node to be inserted
    return value: (Iterator) iterator pointing to inserted node
    --------------------------------------------------------------------------*/
    Iterator InsertFront( const LinkType    &link )
    {
        SAPDBERR_ASSERT_STATE( ! IsMember( link ));

        // -------------------------------------------------------------------------
        // get node handles
        // -------------------------------------------------------------------------
        NodeType & node     = GetNode(link);

        // -------------------------------------------------------------------------
        // chain in the node at the beginning of the list
        // -------------------------------------------------------------------------
        if (m_Head.IsValid())
        {
            NodeType & head     = GetNode(m_Head);
            node.SetSuccessor   (m_Head);
            node.SetPredecessor (link);   // points to itself
            head.SetPredecessor (link);
            m_Head = link;
        }
        else // list yet empty
        {
            node.SetSuccessor   (LinkType());
            node.SetPredecessor (link);   // points to itself
            m_Head = link;
            m_Tail = link;
        }

        ++m_NodeCount;

        SAPDBERR_ASSERT_STATE( IsMember( link ));

        return Begin();
    }

    /*!-------------------------------------------------------------------------
    function:     InsertEnd()
    description:  Inserts a node at the end of the chain
    arguments:    link [in]     link to node to be inserted
    return value: (Iterator) iterator pointing to inserted node
    --------------------------------------------------------------------------*/
    
    Iterator InsertEnd( const LinkType  &link )
    {
        SAPDBERR_ASSERT_STATE( ! IsMember( link ));

        // -------------------------------------------------------------------------
        // get node handles
        // -------------------------------------------------------------------------
        NodeType& node = GetNode(link);

        // -------------------------------------------------------------------------
        // chain in the node at the end of the list
        // -------------------------------------------------------------------------
        if (m_Tail.IsValid())
        {
            NodeType& tail = GetNode(m_Tail);

            node.SetPredecessor (m_Tail);
            node.SetSuccessor   (LinkType());
            tail.SetSuccessor   (link);
            m_Tail = link;
        }
        else // list yet empty
        {
            node.SetPredecessor (link);   // points to itself
            node.SetSuccessor   (LinkType());
            m_Tail = link;
            m_Head = link;
        }

        ++m_NodeCount;

        SAPDBERR_ASSERT_STATE( IsMember( link ));

        return Iterator( *m_Nodes, m_Tail );;
    }
    
    /*!-------------------------------------------------------------------------
    function:     Insert()
    description:  Inserts a node before a given iterator position

    arguments:    iter [in]     iterator insert position
                  link [in]     link to node to be inserted
    return value: (Iterator) iterator pointing to inserted node
    --------------------------------------------------------------------------*/
    /*
    Iterator Insert(
        const Iterator &iter,
        const LinkType &link )
    {
        SAPDBERR_ASSERT_ARGUMENT(iter != End());

        // -------------------------------------------------------------------------
        // check whether InsertFront() can be used
        // -------------------------------------------------------------------------
        if (iter.GetPosition() == m_Head)
        {
            // insert the node at the beginning of the chain
            return InsertFront(link);
        }
        else
            // -------------------------------------------------------------------------
            // check whether InsertEnd() can be used
            // -------------------------------------------------------------------------
            if (iter.GetPosition() == m_Tail)
            {
                // insert the new element at the end of the list
                return InsertEnd(link);
            }
        // -------------------------------------------------------------------------
        // insert element in the middle of the list
        // -------------------------------------------------------------------------
            else
            {
                SAPDBERR_ASSERT_STATE( ! IsMember( link ));

                // at this point, we are sure that there is a precedessor as well as
                // a successor to 'iter' in the list
                // -------------------------------------------------------------------------
                // get node handles
                // -------------------------------------------------------------------------
                NodeType& node     = GetNode(link);
                Iterator  prev     = iter; --prev;
                NodeType& prevnode = GetNode(prev);
                NodeType& nextnode = GetNode(iter);     // insert before 'iter'!

                // -------------------------------------------------------------------------
                // chain in node
                // -------------------------------------------------------------------------
                node.SetSuccessor       (prevnode.GetSuccessor());
                node.SetPredecessor     (prev.GetPosition());
                prevnode.SetSuccessor   (link);
                nextnode.SetPredecessor (link);

                ++m_NodeCount;

                SAPDBERR_ASSERT_STATE( IsMember( link ));

                return Iterator(*m_Nodes, link);
            }
    }
*/
    /*!-------------------------------------------------------------------------
    function:     RemoveFront()
    description:  Removes the first element of the chain

                  - The chain must not be empty.
    arguments:    none
    return value: (Iterator) iterator pointing to element behind removed element
                             or invalid iterator if next element doesn't exist
    --------------------------------------------------------------------------*/
    Iterator RemoveFront()
    {
        SAPDBERR_ASSERT_STATE(!IsEmpty());

        NodeType& node    = GetNode(m_Head);

        // -------------------------------------------------------------------------
        // special case: only one node in the list
        // -------------------------------------------------------------------------
        if (m_NodeCount == 1)
        {
            m_Head.Invalidate();
            m_Tail.Invalidate();
        }
        // -------------------------------------------------------------------------
        // at least 2 nodes in the list
        // -------------------------------------------------------------------------
        else
        {
            NodeType& nextnode = GetNode(node.GetSuccessor());
            m_Head             = node.GetSuccessor();
            nextnode.SetPredecessor(m_Head);    // points to itself
        }
        node.Invalidate();
        --m_NodeCount;

        return Iterator(*m_Nodes, m_Head);
    }

    /*!-------------------------------------------------------------------------
    function:     RemoveEnd()
    description:  Removes the last element of the chain

                  - The chain must not be empty.
    arguments:    none
    return value: (Iterator) invalid iterator (pointing behind tail)
    --------------------------------------------------------------------------*/
    Iterator RemoveEnd()
    {
        SAPDBERR_ASSERT_STATE(!IsEmpty());

        NodeType& node    = GetNode(m_Tail);

        // -------------------------------------------------------------------------
        // special case: only one node in the list
        // -------------------------------------------------------------------------
        if (m_NodeCount == 1)
        {
            m_Head.Invalidate();
            m_Tail.Invalidate();
        }
        // -------------------------------------------------------------------------
        // at least 2 nodes in the list
        // -------------------------------------------------------------------------
        else
        {
            NodeType& prevnode = GetNode(node.GetPredecessor());
            m_Tail             = node.GetPredecessor();
            prevnode.SetSuccessor(node.GetSuccessor());
        }
        node.Invalidate();
        --m_NodeCount;

        return Iterator(*m_Nodes, LinkType());
    }

    /*!-------------------------------------------------------------------------
    function:     Remove()
    description:  Removes the node at a given iterator position

    arguments:    link [in]   link to node to be removed
    return value: (Iterator) iterator pointing to element behind removed element
                             or invalid iterator if next element doesn't exist
    --------------------------------------------------------------------------*/
    Iterator Remove( const LinkType &link )
    {
        SAPDBERR_ASSERT_STATE( !IsEmpty() );
        SAPDBERR_ASSERT_ARGUMENT( link.IsValid() );
        SAPDBERR_ASSERT_STATE( IsMember( link ));

        // -------------------------------------------------------------------------
        // check whether RemoveFront() can be used
        // -------------------------------------------------------------------------
        if (link == m_Head)
        {
            // Remove the first chain element
            return RemoveFront();
        }
        // -------------------------------------------------------------------------
        // check whether RemoveEnd() can be used
        // -------------------------------------------------------------------------
        else
            if (link == m_Tail)
            {
                // Remove the last list element
                return RemoveEnd();
            }
        // -------------------------------------------------------------------------
        // delete element from the middle of the list
        // -------------------------------------------------------------------------
            else
            {
                // at this point, we are sure that there is a precedessor to the node
                // indicated by 'link' in the chain
                // -------------------------------------------------------------------------
                // get node handles
                // -------------------------------------------------------------------------
                NodeType& node     = GetNode(link);
                NodeType& prevnode = GetNode(node.GetPredecessor());

                // -------------------------------------------------------------------------
                // chain the node out of the list
                // -------------------------------------------------------------------------
                prevnode.SetSuccessor(node.GetSuccessor());
                if (node.GetSuccessor().IsValid())
                {
                    NodeType& nextnode = GetNode(node.GetSuccessor());
                    nextnode.SetPredecessor(node.GetPredecessor());
                }
                LinkType succ = node.GetSuccessor();
                node.Invalidate();
                --m_NodeCount;

                SAPDBERR_ASSERT_STATE( ! IsMember( link ));

                return Iterator(*m_Nodes, succ);
            }
    }

    /* -----------------------------------------------------------------------*/
    /*! chapter: Status functions                                             */
    /*!----------------------------------------------------------------------
    function:     IsEmpty()
    description:  Checks whether chain is empty
    arguments:    none
    return value: (bool) true if empty
    --------------------------------------------------------------------------*/
    bool IsEmpty() const
    {
        return m_Head.IsInvalid();
    }

    /*!-------------------------------------------------------------------------
    function:     GetNodeCount()
    description:  Returns the actual node count of the chain
    arguments:    none
    return:       (NodeCount) node count
    --------------------------------------------------------------------------*/
    NodeCount GetNodeCount() const
    {
        return m_NodeCount;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Status functions                                          */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: ConstIterator functions                                           */
    /*!-------------------------------------------------------------------------
    function:     Begin()
    description:  Returns an iterator to the first chain node
    arguments:    none
    return value: (Iterator/ConstIterator)
    prototypes:   2
    --------------------------------------------------------------------------*/
    Iterator Begin()
    {
        return Iterator(*m_Nodes, m_Head);
    }

    ConstIterator Begin() const
    {
        return ConstIterator(*m_Nodes, m_Head);
    }

    /*!-------------------------------------------------------------------------
    function:     End()
    description:  Returns an iterator behind the last chain node
    arguments:    none
    return value: (Iterator/ConstIterator)
    prototypes:   2
    --------------------------------------------------------------------------*/
    Iterator End()
    {
        return Iterator(*m_Nodes, LinkType());
    }

    ConstIterator End() const
    {
        return ConstIterator(*m_Nodes, LinkType());
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: ConstIterator functions                                   */
    /* -----------------------------------------------------------------------*/

    /*-----------------------------------------------------------------------
    function:     IsMember()
    description:  Checks whether the given node is member of the chain.
    arguments:    link [in]   link to node to be checked
    return value: (SAPDB_Bool) true if node is member else false
    ----------------------------------------------------------------------*/

    SAPDB_Bool IsMember( const LinkType   &link )
    {
        Iterator        iter        = Begin();
        Iterator        endIter     = End();
        NodeCount       loopCnt     = 0;
        const NodeCount maxEntryCnt = ( *m_Nodes ).GetSize();

        for( ; iter != endIter; ++iter, ++loopCnt )
        {
            if( link == iter.GetPosition() )
                return( SAPDB_TRUE );

            SAPDBERR_ASSERT_STATE(  loopCnt <  maxEntryCnt );
        }
        return( SAPDB_FALSE );
    }

private:

    /*-----------------------------------------------------------------------
    function:     GetNode()
    description:  Returns the requested node
    arguments:    link [in]   link to node to be requested
    return value: NodeType
    ------------------------------------------------------------------------*/

    NodeType& GetNode( const LinkType   &link )
    {
        return (*m_Nodes)[link];
    }

    /*--------------------------------------------------------------------------
    declaration : m_Nodes
    description:  link to node sequence
    --------------------------------------------------------------------------*/
    NodeSequence*                   m_Nodes;

    /*--------------------------------------------------------------------------
    declaration : m_NodeCount
    description:  count of nodes in chain
    --------------------------------------------------------------------------*/
    NodeCount                       m_NodeCount;

    /*--------------------------------------------------------------------------
    declaration : m_Head
    description:  chain head
    --------------------------------------------------------------------------*/
    LinkType                        m_Head;

    /*--------------------------------------------------------------------------
    declaration : m_Head
    description:  chain tail
    --------------------------------------------------------------------------*/
    LinkType                        m_Tail;
};
/*!*****************************************************************************

   endclass: Converter_Chain

*******************************************************************************/




#endif // CONVERTER_CHAIN_HPP
