/****************************************************************************

  module      : Converter_converter.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: converter
  
  description : implementation of all converter methods

  last changed: 2001-01-15  15:00
  see also    : 

  author:       AlexanderK

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hgg01.h"    //g01islivecache
#include "hbd20_6.h"  // bd20GetNumberOfChangedPages
#include "hbd91.h"    //garbage collector 
#include "Converter/Converter_Converter.hpp"
#include "Converter/Converter_Dump.hpp"
#include "Converter/Converter_LeafPage.hpp"
#include "Converter/Converter_Exception.hpp"
#include "Converter/Converter_Messages.hpp"
#include "Converter/Converter_BlockAddressIterator.hpp"
#include "FreeBlockManagement/FBM_Manager.hpp"
#include "IOManager/IOMan_IDataInfo.hpp"
#include "IOManager/IOMan_IConverterIO.hpp"
#include "Pager/Pager_IRestart.hpp"
#include "Pager/Pager_ISavepoint.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/RTE_Crash.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define CONV_NUM_SECURITY_PAGES 10

/*===========================================================================*
 *  INSTANCE METHODS OF THE INTERFACE CLASSES  OF THE CONVERTER CLASS        *
 *===========================================================================*/

Converter_IManager & Converter_IManager::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IMigration & Converter_IMigration::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IDump & Converter_IDump::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IPageIO & Converter_IPageIO::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_ISavePoint & Converter_ISavePoint::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IVerify & Converter_IVerify::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_ICommon & Converter_ICommon::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IBackUp & Converter_IBackUp::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IPageNoMapping & Converter_IPageNoMapping::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IPageNoManager & Converter_IPageNoManager::Instance()
{
    return Converter_Converter::Instance();
}

/*---------------------------------------------------------------------------*/

Converter_IEvent & Converter_IEvent::Instance()
{
    return Converter_Converter::Instance();
}


/*===========================================================================*
 *  METHODS OF THE CONVERTER CLASS                                           *
 *===========================================================================*/

Converter_Converter* Converter_Converter::m_Instance = NULL;

/*---------------------------------------------------------------------------*/

Converter_Converter::Converter_Converter( SAPDBMem_IRawAllocator  &Allocator )
        :
        m_Allocator( UTF8( "Converter_Manager" ), Allocator),
        m_MapBlockingSize( 8 ),
        m_TempMapControl( m_Allocator, Data_PageRecoveryMode( Data_Dynamic, Data_NotRecoverable )),
        m_PermDynamicMapControl( m_Allocator, Data_PageRecoveryMode( Data_Dynamic, Data_Recoverable )),
        m_PermStaticMapControl( m_Allocator, Data_PageRecoveryMode( Data_Static, Data_Recoverable )),
        m_Index (m_Allocator),
        m_Active( false ),
        m_InMaintenanceMode( false ),
        m_SavepointIsRequested( false ),
        m_DBFullWaitQueue (CONV_SUSPEND_HANDLE_DB_FULL),
        m_FreePageNoWaitQueue (CONV_SUSPEND_FREE_PAGENO_WAIT_DURING_SVP),
        m_ExpandConverterWaitQueue(CONV_SUSPEND_EXPAND_WAIT_DURING_SVP),
        m_DBUsageMonitor (sp31ei_db_filling_above_limit, sp31ei_db_filling_below_limit),
        m_VolumeList(NULL)
{
    /* the exceeding of the following limits will always be written */
    /* into the knldiag. all values are defined in units of 0.01 %  */
    m_DBUsageMonitor.EnableEventing();
    const SAPDB_Bool bOverFlowEvent          = true;
    const SAPDB_Bool bWriteAlwaysIntoKnldiag = true;

    m_DBUsageMonitor.AddEvent(8000, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);
    m_DBUsageMonitor.AddEvent(9000, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);
    m_DBUsageMonitor.AddEvent(9500, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);
    m_DBUsageMonitor.AddEvent(9600, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);
    m_DBUsageMonitor.AddEvent(9700, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);
    m_DBUsageMonitor.AddEvent(9800, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);
    m_DBUsageMonitor.AddEvent(9900, bOverFlowEvent, sp31ep_high, bWriteAlwaysIntoKnldiag);

    for (int Limit = 500; Limit < 10000; Limit+=500)
    {
        m_DBUsageMonitor.AddEvent(Limit, bOverFlowEvent,   sp31ep_low, ! bWriteAlwaysIntoKnldiag);
        m_DBUsageMonitor.AddEvent(Limit, ! bOverFlowEvent, sp31ep_low, ! bWriteAlwaysIntoKnldiag);
    }
}

/*---------------------------------------------------------------------------*/

Converter_Converter&
Converter_Converter::CreateInstance ()
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::CreateInstance", Converter_Trace, 5 );

    if (m_Instance == NULL)
    {
        m_Instance = new (RTEMem_Allocator::Instance())
                     Converter_Converter( RTEMem_Allocator::Instance());
        if (NULL == m_Instance)
        {
            Converter_Exception NoMoreMemoryToCreateConverter (__CONTEXT__, CONVERTER_NO_MEMORY_TO_BUILD_CONVERTER);
            RTE_Crash (NoMoreMemoryToCreateConverter);
        }
    }
    return *m_Instance;
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::Restart (const tsp00_TaskId        TaskId,
                              const Converter_Version  &ConverterVersion,
                              const IOMan_BlockAddress &ConverterRootBlockAddress,
                              const SAPDB_Bool          bIncConverterVersion,
                              const SAPDB_Bool          bLastSaveDataSuccessful,
                              const Data_PageNo         MaxDynamicPageNo,
                              const Data_PageNo         MaxStaticPageNo,
                              const SAPDB_UInt          NumAllPages,
                              const tsp00_RegionId      BaseRegionId,
                              const SAPDB_UInt          NumConverterRegions,
                              const SAPDB_UInt          ConverterPageSize,
                              const SAPDB_UInt1         DevNoBitCount)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::Restart", Converter_Trace, 5 );

    /* initialize static member of the converter page defining */
    /* the number of bits used to address a device             */
    Converter_Page::SetDevNoBitCount  (DevNoBitCount);

    if(
        ( ! m_TempMapControl.Initialize( TaskId, NumAllPages,
                                         BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
        ||
        ( ! m_PermDynamicMapControl.Initialize( TaskId, SAPDB_MAX( MaxDynamicPageNo, Data_PageNo( NumAllPages )),
                                                BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
        ||
        (! m_PermStaticMapControl.Initialize( TaskId, SAPDB_MAX( MaxStaticPageNo, Data_PageNo( NumAllPages )),
                                              BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
    )
    {
        Converter_Exception MapInitFailed(__CONTEXT__, CONVERTER_MAP_INIT_FAILED1,
                                          SAPDB_ToString(NumAllPages,      _T_d),
                                          SAPDB_ToString(MaxDynamicPageNo, _T_d),
                                          SAPDB_ToString(MaxStaticPageNo,  _T_d));
        RTE_Crash (MapInitFailed);
    }

    InitializeBase( NumAllPages );

    m_ConverterVersion          = ConverterVersion;
    m_ConverterRootBlockAddress = ConverterRootBlockAddress;
    m_LastSaveDataSuccessful    = bLastSaveDataSuccessful;

    /* get access to the page allocator to get the memory for the pages reading next */
    Converter_PageFrameAllocator PageFrameAllocator (TaskId);

    /* the parent index is created starting with the data stored on the ConverterRootBlock */
    m_Index.Restart (PageFrameAllocator, TaskId, ConverterRootBlockAddress);

    /* create per device lists of all converter leaf nodes */
    Container_Vector<Converter_Index::BlockVector> volumeList(m_Allocator);
    m_Index.CreateInputQueues(volumeList, ConverterPageSize);
    m_VolumeList = &volumeList;

    /* read all converter leaf pages referenced by the converter tree */
    Pager_IRestart::GetInstance().ReadConverter(TaskId);

    m_VolumeList = NULL;

    SAPDBERR_ASSERT_STATE( m_Index.Verify()); // PTS 1113242 TS 2001-12-21

    if (bIncConverterVersion)
        ++m_ConverterVersion;
    else
        m_InMaintenanceMode = true;

    if( m_MaxPermUsed < GetNumberOfUsedPages() )
        m_MaxPermUsed = GetNumberOfUsedPages();

    SAPDBTRACE_WRITELN( Converter_Trace, 5, "PermPagesUsed: " << m_NumPermPagesUsed );
    SAPDBTRACE_WRITELN( Converter_Trace, 5, "UpdatedDynamicPages: " <<
                        m_PermDynamicMapControl.NumUpdatedPermPages( TaskId ));
    SAPDBTRACE_WRITELN( Converter_Trace, 5, "UpdatedStaticPages: " <<
                        m_PermStaticMapControl.NumUpdatedPermPages( TaskId ));
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::ReadPages(
    const tsp00_TaskId  taskId,
    const SAPDB_UInt    pagerIndex,
    SAPDB_UInt          &totalPages,
    SAPDB_UInt          &totalIO )
{

    SAPDBTRACE_ROUTINE_DEBUG( "Converter::ReadPages", Converter_Trace, 5 );

    SAPDBERR_ASSERT_STATE (0 != m_VolumeList);

    // check if Pager has something to do ..
    if ((*m_VolumeList).GetSize() <= pagerIndex){
        return;
    }

    /* get access to the page allocator to get the memory for the pages reading next */
    Converter_PageFrameAllocator PageFrameAllocator (taskId);

    Converter_Index::BlockVector::Iterator iter = (*m_VolumeList)[pagerIndex].Begin();
    Converter_Index::BlockVector::Iterator iterEnd = (*m_VolumeList)[pagerIndex].End();

    for (; iter < iterEnd; iter++)
    {
        SAPDB_UInt NumPermPagesUpdated          = 0;
        SAPDB_UInt NumRemainingPermPagesUpdated = 0;

        Converter_LeafPage  LeafPage(PageFrameAllocator.New( true ));

        IOMan_IConverterIO::GetInstance().ReadConverterPage(taskId, iter->GetBlockAddress(), LeafPage);
        totalPages++;
        totalIO++;

        Converter_PageEntryIterator   PageEntryIterator (LeafPage);

        /* check that this page is really a permanent page */
        SAPDBERR_ASSERT_STATE (LeafPage.ReadPageRecoveryMode().PageIsRecoverable());

        /* write the address of the parent index page on the leaf page according the the iterator */
        LeafPage.WriteParentReference (iter->GetParentReference());

        SAPDB_UInt FreeEntryCount = 0;

        /* mark all used blocks referenced by the current converter leaf pages as       */
        /* occupied in the FBM. Moreover the RepairIterator sets all reserved converter */
        /* leaf page entries to free                                                    */
        for (PageEntryIterator=LeafPage.Begin(); PageEntryIterator!=LeafPage.End(); ++PageEntryIterator)
        {
            if(!PageEntryIterator.EntryInUse())
            {
                ++FreeEntryCount;
            }
            else
            {
                if (PageEntryIterator.GetPackedBlockAddress().IsReserved())
                {
                    LeafPage.ReleaseEntry(PageEntryIterator.GetPosition());
                    ++FreeEntryCount;
                }
                else
                {

                    /* increment number of all used pages */
                    m_NumPermPagesUsed.Increment();

                    /* write a message if the number of used pages exceeds a critical value */
                    CheckFillingOfDataBase();

                    /* increment the number of pages which are changed since the last save data */
                    if (LeafPage.SavePagesPending(PageEntryIterator.GetPosition()) && !m_LastSaveDataSuccessful)
                        // ++ NumPermPagesUpdated; PTS 1121818 TS 2003-04-28
                        ++NumRemainingPermPagesUpdated;

                    if (LeafPage.SavePagesRequested(PageEntryIterator.GetPosition()))
                        // ++ NumRemainingPermPagesUpdated; PTS 1121818 TS 2003-04-28
                        ++ NumPermPagesUpdated;
                }
            }
        }
        Converter_BlockAddressIterator elements = LeafPage.elements();
        if( SAPDB_TRUE != FBM_Manager::Instance().SetBlockStateToOccupied(
                    taskId, elements))
        {

            Converter_Exception errMsg( __CONTEXT__, CONVERTER_BAD_ENTRY_STATE,
                    (( Data_Dynamic == LeafPage.ReadPageRecoveryMode().GetAddressingMode())
                    ? "dynamic" : "static"),
                    SAPDB_ToString( LeafPage.GetPageNo( elements.GetPosition()), _T_d ));

            RTE_Crash( errMsg );

        }

        if (Data_Static == LeafPage.ReadPageRecoveryMode().GetAddressingMode())
        {
            m_PermStaticMapControl.IntegratePage (taskId, LeafPage,
                                                  FreeEntryCount, NumPermPagesUpdated, NumRemainingPermPagesUpdated);
        }
        else // Data_Dynamic
        {
            m_PermDynamicMapControl.IntegratePage (taskId, LeafPage,
                                                   FreeEntryCount, NumPermPagesUpdated, NumRemainingPermPagesUpdated);
        }
    }

}

/*---------------------------------------------------------------------------*/
void
Converter_Converter::Create (const tsp00_TaskId        TaskId,
                             const SAPDB_UInt          NumAllPages,
                             const tsp00_RegionId      BaseRegionId,
                             const SAPDB_UInt          NumConverterRegions,
                             const SAPDB_UInt          ConverterPageSize,
                             const SAPDB_UInt1         DevNoBitCount)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::Create", Converter_Trace, 5 );

    /* initialize static member of the converter page defining */
    /* the number of bits used to address a device             */
    Converter_Page::SetDevNoBitCount  (DevNoBitCount);

    if(
        ( ! m_TempMapControl.Initialize( TaskId, NumAllPages,
                                         BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
        ||
        ( ! m_PermDynamicMapControl.Initialize( TaskId, NumAllPages,
                                                BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
        ||
        (! m_PermStaticMapControl.Initialize( TaskId, NumAllPages,
                                              BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
    )

    {
        Converter_Exception MapInitFailed (__CONTEXT__, CONVERTER_MAP_INIT_FAILED2,
                                           SAPDB_ToString(NumAllPages, _T_d));
        RTE_Crash (MapInitFailed);
    }

    InitializeBase( NumAllPages );
    m_ConverterVersion.SetToInitialVersion(); // corresponds to version 1

    /* the parent index is created */
    Converter_PageFrameAllocator PageFrameAllocator (TaskId);
    m_Index.Create (PageFrameAllocator);
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::CreateForRestore (tsp00_TaskId              TaskId,
                                       const Converter_Version  &ConverterVersion,
                                       const SAPDB_Bool          bLastSaveDataSuccessful,
                                       const Data_PageNo         MaxDynamicPageNo,
                                       const Data_PageNo         MaxStaticPageNo,
                                       const SAPDB_UInt          NumAllPages,
                                       const tsp00_RegionId      BaseRegionId,
                                       const SAPDB_UInt          NumConverterRegions,
                                       const SAPDB_UInt          ConverterPageSize,
                                       const SAPDB_UInt1         DevNoBitCount)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::CreateForRestore", Converter_Trace, 5 );

    /* initialize static member of the converter page defining */
    /* the number of bits used to address a device             */
    Converter_Page::SetDevNoBitCount  (DevNoBitCount);

    if(
        ( ! m_TempMapControl.Initialize( TaskId, NumAllPages,
                                         BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
        ||
        ( ! m_PermDynamicMapControl.Initialize( TaskId, SAPDB_MAX(MaxDynamicPageNo,Data_PageNo(NumAllPages)),
                                                BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
        ||
        (! m_PermStaticMapControl.Initialize( TaskId, SAPDB_MAX(MaxStaticPageNo,Data_PageNo(NumAllPages)),
                                              BaseRegionId, NumConverterRegions, m_MapBlockingSize, ConverterPageSize ))
    )
    {
        Converter_Exception MapInitFailed(__CONTEXT__, CONVERTER_MAP_INIT_FAILED1,
                                          SAPDB_ToString(NumAllPages,      _T_d),
                                          SAPDB_ToString(MaxDynamicPageNo, _T_d),
                                          SAPDB_ToString(MaxStaticPageNo,  _T_d));
        RTE_Crash (MapInitFailed);
    }

    InitializeBase( NumAllPages );
    m_InMaintenanceMode      = true;
    m_RestoreDataActive      = true;
    m_ConverterVersion       = ConverterVersion;
    m_LastSaveDataSuccessful = bLastSaveDataSuccessful;

    /* the parent index is created */
    Converter_PageFrameAllocator PageFrameAllocator (TaskId);
    m_Index.Create (PageFrameAllocator);
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::Shutdown (const tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::Shutdown", Converter_Trace, 5 );

    if( ! m_Active )
        return;

    m_NumPermPagesUsed   = 0;
    m_NumTempPagesUsed   = 0;
    m_NumPermConvPages   = 0;
    m_NumTempConvPages   = 0;
    m_MaxPermUsed        = 0;
    m_NumAllPages        = 0;
    m_SecuritySpaceLimit = 0;
    m_Active             = false;
    m_InMaintenanceMode  = false;
    m_SavePagesIsActiv   = false;
    m_SaveDataIsActiv    = false;
    m_RestoreDataActive  = false;
    m_RestorePagesActive = false;
    m_ColdVerifyIsActive = false;
    m_ConverterVersion.Invalidate();

    m_DBFullWaitQueue.ResumeAllWaitingTasks(TaskId);
    m_FreePageNoWaitQueue.ResumeAllWaitingTasks(TaskId);

    m_TempMapControl.Shutdown( TaskId );
    m_PermDynamicMapControl.Shutdown( TaskId );
    m_PermStaticMapControl.Shutdown( TaskId );

    Converter_PageFrameAllocator PageFrameAllocator (TaskId);
    m_Index.Shutdown(PageFrameAllocator);
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::FlushParallel (tgg00_TransContext  &Trans,
                                    Converter_Version   &ConverterVersion,
                                    Data_PageNo         &MaxStaticPageNo,
                                    Data_PageNo         &MaxDynamicPageNo,
                                    IOMan_BlockAddress  &ConverterRootBlockAddress,
                                    SAPDB_UInt          &NumPagesTotal,
                                    SAPDB_UInt          &NumIOTotal)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::FlushParallel", Converter_Trace, 5 );

    SAPDB_UInt  NumPages = 0;
    SAPDB_UInt  NumIO    = 0;


    if( ! m_Active )
    {
        Trans.trError_gg00 = e_shutdown;
        return;
    }

    // MapControl remembers parentReferences of empty ConverterLeafPages
    // between converter flushes, now it's time to drop these pages
    // from the index

    m_PermStaticMapControl.RemoveEmptyLeaves(Trans.trTaskId_gg00, m_Index);
    m_PermDynamicMapControl.RemoveEmptyLeaves(Trans.trTaskId_gg00, m_Index);

    Pager_ISavepoint::GetInstance().FlushConverterParallel(Trans.trTaskId_gg00, NumPages, NumIO);
    NumPagesTotal += NumPages;
    NumIOTotal    += NumIO;

    MaxStaticPageNo             = m_MaxUsedStaticPageNo;
    MaxDynamicPageNo            = m_MaxUsedDynamicPageNo;

    m_ConverterRootBlockAddress = m_Index.Flush (Trans.trTaskId_gg00, NumPages, NumIO);
    NumPagesTotal += NumPages;
    NumIOTotal    += NumIO;

    ConverterRootBlockAddress   = m_ConverterRootBlockAddress;
    ConverterVersion            = m_ConverterVersion;
}


/*---------------------------------------------------------------------------*/

void
Converter_Converter::WritePages(
    const tsp00_TaskId  taskId,
    const SAPDB_UInt    iSection,
    SAPDB_UInt          &totalPages,
    SAPDB_UInt          &totalIO )
{
    totalPages = 0;
    totalIO    = 0;

    m_PermStaticMapControl.FlushSection  (taskId, m_Index, iSection, totalPages, totalIO);
    m_PermDynamicMapControl.FlushSection (taskId, m_Index, iSection, totalPages, totalIO);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::FlushSerial (
    const tsp00_TaskId  taskId,
    Converter_Version   &ConverterVersion,
    Data_PageNo         &MaxStaticPageNo,
    Data_PageNo         &MaxDynamicPageNo,
    IOMan_BlockAddress  &ConverterRootBlockAddress)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::FlushSerial", Converter_Trace, 5 );

    SAPDB_UInt  NumPages = 0;
    SAPDB_UInt  NumIO    = 0;

    if( ! m_Active )
    {
        return( false );
    }

    // MapControl remembers parentReferences of empty ConverterLeafPages
    // between converter flushes, now it's time to drop these pages
    // from the index

    m_PermStaticMapControl.RemoveEmptyLeaves( taskId, m_Index );
    m_PermDynamicMapControl.RemoveEmptyLeaves( taskId, m_Index);

    MaxStaticPageNo             = m_PermStaticMapControl.Flush( taskId, m_Index );
    MaxDynamicPageNo            = m_PermDynamicMapControl.Flush( taskId, m_Index );

    m_ConverterRootBlockAddress = m_Index.Flush( taskId, NumPages, NumIO);
    ConverterRootBlockAddress   = m_ConverterRootBlockAddress;
    ConverterVersion            = m_ConverterVersion;

    return( true );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::Expand (
    const tsp00_TaskId     TaskId,
    const IOMan_DeviceNo   &DevNo,
    const SAPDB_UInt        DevSize,
    const SAPDB_UInt        NumPagesAdded)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::Expand", Converter_Trace, 5 );

    if( ! m_Active )
        return false;

    if( ! ExpandMap( TaskId, Data_PageRecoveryMode( Data_Dynamic, Data_NotRecoverable ),  NumPagesAdded ))
        return false;

    if( ! ExpandMap( TaskId, Data_PageRecoveryMode( Data_Dynamic, Data_Recoverable ), NumPagesAdded ))
        return false;

    if( ! ExpandMap( TaskId, Data_PageRecoveryMode( Data_Static, Data_Recoverable ), NumPagesAdded ))
        return false;

    if(! FBM_Manager::Instance().AddVolume( TaskId, DevNo, DevSize ))
        return false;

    m_NumAllPages +=NumPagesAdded;
    m_DBUsageMonitor.Rescale (m_NumAllPages);

    // PTS 1124331 TS 2003-09-23
    m_SecuritySpaceLimit = ( m_NumAllPages - SAPDB_MIN( m_NumAllPages/25, 4000));

    m_DBFullWaitQueue.ResumeAllWaitingTasks(TaskId);

    return true;
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::BeginSaveData (tsp00_TaskId  TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::BeginSaveData", Converter_Trace, 5 );

    SAPDBERR_ASSERT_STATE (! BackUpIsActive());
    SAPDBERR_ASSERT_STATE (false == m_ColdVerifyIsActive);
    SAPDBERR_ASSERT_STATE (0     == m_NumStatConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumDynConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumPagesForBackUp);

    m_SaveDataIsActiv = true;
    SAPDB_UInt NumPagesForBackUp   = 0;

    m_PermStaticMapControl.MarkPermPagesForSaveData (TaskId, m_LastSaveDataSuccessful,
            m_NumStatConvPagesForBackUp, NumPagesForBackUp );

    m_NumPagesForBackUp = NumPagesForBackUp;

    m_PermDynamicMapControl.MarkPermPagesForSaveData (TaskId, m_LastSaveDataSuccessful,
            m_NumDynConvPagesForBackUp, NumPagesForBackUp );

    m_NumPagesForBackUp         += NumPagesForBackUp;

    FBM_Manager::Instance().BeginReadingBlocksMarkedForBackUp (TaskId);
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::BeginSavePages (tsp00_TaskId  TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::BeginSavePages", Converter_Trace, 5 );

    SAPDBERR_ASSERT_STATE (! BackUpIsActive());
    SAPDBERR_ASSERT_STATE (false == m_ColdVerifyIsActive);
    SAPDBERR_ASSERT_STATE (0     == m_NumStatConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumDynConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumPagesForBackUp);

    m_SavePagesIsActiv           = true;
    SAPDB_UInt NumPagesForBackUp = 0;

    m_PermStaticMapControl.MarkPermPagesForSavePages (TaskId,
            m_LastSaveDataSuccessful, m_NumStatConvPagesForBackUp, NumPagesForBackUp, Version() );

    m_NumPagesForBackUp = NumPagesForBackUp;

    m_PermDynamicMapControl.MarkPermPagesForSavePages (TaskId,
            m_LastSaveDataSuccessful, m_NumDynConvPagesForBackUp, NumPagesForBackUp, Version());

    m_NumPagesForBackUp += NumPagesForBackUp;

    FBM_Manager::Instance().BeginReadingBlocksMarkedForBackUp (TaskId);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::CheckBackUp( const tsp00_TaskId taskId ) const
{

    return ((FBM_Manager::Instance().NumBlocksMarkedForBackup( taskId ) == 0) &&
            m_PermDynamicMapControl.BackUpPageNoListIsEmpty() &&
            m_PermStaticMapControl.BackUpPageNoListIsEmpty());
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::ResetDataPageAfterBackUp (
    const tsp00_TaskId          &TaskId,
    const IOMan_BlockAddress    &BlockAddress,
    const Data_PageNo           &PageNo,
    const Data_AddressingMode   &AddressingMode)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::ResetDataPageAfterBackUp", Converter_Trace, 5 );

    SAPDB_Bool retCode = SAPDB_FALSE;

    if( m_Active )
    {
        if( Data_Dynamic == AddressingMode )
        {
            retCode = m_PermDynamicMapControl.RemovePageNoFromBackUpPageNoList( TaskId, PageNo );
        }
        else if( Data_Static == AddressingMode )
        {
            retCode = m_PermStaticMapControl.RemovePageNoFromBackUpPageNoList( TaskId, PageNo );
        }

        if( SAPDB_TRUE == retCode ){
            FBM_Manager::Instance().RestoreBlockStateMarkedForBackup( TaskId, BlockAddress );
        }
    }
    return( retCode );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::EndSave  (const tsp00_TaskId &TaskId,
                               SAPDB_Bool          bBackUpSuccessfullyFinished)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::EndSave", Converter_Trace, 5 );

    if( ! m_Active )
        return;

    m_NumDynConvPagesForBackUp  = 0;
    m_NumStatConvPagesForBackUp = 0;
    m_NumPagesForBackUp         = 0;

    if (m_SaveDataIsActiv)
    {
        m_SaveDataIsActiv = false;

        if (!bBackUpSuccessfullyFinished)
            m_LastSaveDataSuccessful = false;
        else
        {
            m_LastSaveDataSuccessful = true;

            m_PermDynamicMapControl.ResetCounterOfPagesStillMarkedForSavePages(TaskId);
            m_PermStaticMapControl.ResetCounterOfPagesStillMarkedForSavePages(TaskId);
        }
    }
    else
    {
        m_SavePagesIsActiv = false;

        m_PermDynamicMapControl.ClearUsedPageNoList(TaskId);
        m_PermStaticMapControl.ClearUsedPageNoList(TaskId);
    }

    /* remove all flags and clear all lists which indicate */
    /* that a page is marked for the current back up       */
    if ( !bBackUpSuccessfullyFinished)
        FBM_Manager::Instance().RestoreAllBlockStatesMarkedForBackup (TaskId);
    else
    {
        SAPDBERR_ASSERT_STATE( CheckBackUp( TaskId ));
    }

    /* release all resources occupied by the back up pageno container */
    m_PermDynamicMapControl.ClearBackUpPageNoList();
    m_PermStaticMapControl.ClearBackUpPageNoList();
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::EndSavepoint( const tsp00_TaskId   taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::EndSavepoint", Converter_Trace, 5 );

    ++m_ConverterVersion;

    m_PermStaticMapControl.ReSetSavepointActiveFlag( taskId );
    m_PermDynamicMapControl.ReSetSavepointActiveFlag( taskId );

    if( m_MaxPermUsed < GetNumberOfUsedPages() )
        m_MaxPermUsed = GetNumberOfUsedPages();

    m_SavepointIsRequested = false;
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::FreePageNosAfterSavepoint( const tsp00_TaskId &TaskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::FreePageNosAfterSavepoint", Converter_Trace, 5 );

    /* release all PageNos stored in the FreePageNoList */
    m_PermStaticMapControl.FreePageNosAfterSavepoint( TaskId );
    m_PermDynamicMapControl.FreePageNosAfterSavepoint( TaskId );

    /* resume all tasks which were suspended during a savepoint since there    */
    /* was no more space to insert the pageno to release into a freepagenolist */
    if( ! m_FreePageNoWaitQueue.IsEmpty( TaskId ))
        m_FreePageNoWaitQueue.ResumeAllWaitingTasks( TaskId );

    /* resume tasks which want to expand the converter and collidated with the savepoint */
    if( ! m_ExpandConverterWaitQueue.IsEmpty( TaskId ))
        m_ExpandConverterWaitQueue.ResumeAllWaitingTasks( TaskId );

    // resume all tasks suspended because a FBM overflow occured
    if( ! m_DBFullWaitQueue.IsEmpty( TaskId ))
        m_DBFullWaitQueue.ResumeAllWaitingTasks( TaskId );
}

/*---------------------------------------------------------------------------*/

IOMan_BlockAddress
Converter_Converter::GetDevicePositionForReading (
    const tsp00_TaskId          taskId,
    const Data_PageNo           pageNo,
    const Data_PageRecoveryMode &pageRecoveryMode)
{
    if( ! m_Active )
        return IOMan_BlockAddress();

    if( pageRecoveryMode.PageIsRecoverable() )
    {
        if( pageRecoveryMode.UseDynamicPageAddressing() )
            return( m_PermDynamicMapControl.GetBlockAddress( taskId, pageNo ));
        else
            return( m_PermStaticMapControl.GetBlockAddress( taskId, pageNo ));
    }
    else
        return( m_TempMapControl.GetBlockAddress ( taskId, pageNo ));

}

/*---------------------------------------------------------------------------*/

IOMan_BlockAddress
Converter_Converter::GetDevicePositionForWriting (
    const tsp00_TaskId          taskId,
    tgg00_BasisError            &trError,
    const Data_PageNo           &PageNo,
    const Data_PageRecoveryMode &PageRecoveryMode,
    const Converter_Version     &PageConverterVersion)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::GetDevicePositionForWriting", Converter_Trace, 5 );

    if( ! m_Active )
    {
        trError = e_shutdown;
        return IOMan_BlockAddress();
    }

    const SAPDB_Bool bIsPerm = PageRecoveryMode.PageIsRecoverable();

    if ((PageConverterVersion == m_ConverterVersion) || (PageConverterVersion.IsValid() && ( ! bIsPerm )))
    {
        // the currently occupied block can be reused

        if( bIsPerm )
        {
            if( PageRecoveryMode.UseDynamicPageAddressing() )
                return( m_PermDynamicMapControl.GetBlockAddress( taskId, PageNo ));
            else
                return( m_PermStaticMapControl.GetBlockAddress( taskId, PageNo ));
        }
        else
            return( m_TempMapControl.GetBlockAddress ( taskId, PageNo ));
    }
    else // old block is not reusable
    {
        SAPDBERR_ASSERT_STATE (PageConverterVersion.IsInvalid() || (PageConverterVersion < m_ConverterVersion));

        if( bIsPerm )
        {
            if( PageRecoveryMode.UseDynamicPageAddressing() )
                return( m_PermDynamicMapControl.GetNewPositionForWrite( taskId, PageNo,
                        PageConverterVersion ));
            else
                return( m_PermStaticMapControl.GetNewPositionForWrite( taskId, PageNo,
                        PageConverterVersion ));
        }
        else
            return( m_TempMapControl.GetNewPositionForWrite( taskId, PageNo ));
    }
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::SetDevicePositionForWriting (
    const tsp00_TaskId          taskId,
    tgg00_BasisError            &trError,
    const Converter_Version     &PageConverterVersion,
    const Data_PageNo           &PageNo,
    const Data_PageRecoveryMode &PageRecoveryMode,
    const IOMan_BlockAddress    &BlockAddress)
{
    if( ! m_Active )
    {
        trError = e_shutdown;
        return;
    }

    if( PageRecoveryMode.PageIsRecoverable() )
    {
        if( PageRecoveryMode.UseDynamicPageAddressing() )
            m_PermDynamicMapControl.SetPositionForWriting( taskId, m_ConverterVersion,
                    PageConverterVersion, PageNo, BlockAddress );
        else
            m_PermStaticMapControl.SetPositionForWriting( taskId, m_ConverterVersion,
                    PageConverterVersion, PageNo, BlockAddress );
    }
    else
        m_TempMapControl.SetPositionForWriting ( taskId, PageConverterVersion, PageNo, BlockAddress );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::RequestSpecificiedNewStaticPageNo  (
    tgg00_TransContext &Trans,
    const Data_PageNo   PageNo)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::RequestSpecificiedNewStaticPageNo", Converter_Trace, 5 );

    // PTS 1116298 UH 2002-06-21
    // Added the wait interface and the loop because RequestSpecifiedNewPageNo()
    // has to wait for a savepoint.
    // It may happen that there exists a pending Free for the requested pageno.

    SAPDB_Bool bWaitUntilSavepointFinished  = false;

    do
    {
        if( ! m_Active )
        {
            Trans.trError_gg00 = e_shutdown;
            return;
        }

        if( !m_PermStaticMapControl.RequestSpecifiedNewPageNo( Trans.trTaskId_gg00,
                PageNo, m_FreePageNoWaitQueue, bWaitUntilSavepointFinished ))
        {
            Trans.trError_gg00 =  e_page_no_in_use;
            SAPDBTRACE_WRITELN( Converter_Trace, 5, "PageNo: " <<  PageNo << "not available" );
            return;
        }

        if( bWaitUntilSavepointFinished )
            m_FreePageNoWaitQueue.Wait( Trans.trTaskId_gg00 );
    }
    while( bWaitUntilSavepointFinished );

    m_NumPermPagesUsed.Increment();

    CheckFillingOfDataBase ();

    SAPDBTRACE_WRITELN( Converter_Trace, 7, "PermPagesUsed: " << m_NumPermPagesUsed );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::FreeTempPageNo(
    const tsp00_TaskId  taskId,
    const Data_PageNo   pageNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::FreeTempPageNo", Converter_Trace, 5 );

    SAPDBTRACE_WRITELN( Converter_Trace, 5, "Free temp PageNo: " << pageNo );

    if( ! m_Active )
        return( SAPDB_FALSE );

    m_TempMapControl.FreePageNo( taskId, pageNo );

    m_NumTempPagesUsed.Decrement();
    SAPDBTRACE_WRITELN( Converter_Trace, 7, "TempPagesUsed: " << m_NumTempPagesUsed );

    CheckFillingOfDataBase();

    if( ! m_DBFullWaitQueue.IsEmpty( taskId )){
        m_DBFullWaitQueue.ResumeAllWaitingTasks( taskId );
    }
    return( SAPDB_TRUE );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::HandleDBFull (
    tgg00_TransContext  &Trans,
    SAPDB_UInt          NumDataPagesRequested,
    const SAPDB_Bool    bAddDeviceIsPossible)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::HandleDBFull", Converter_Trace, 5 );

    tgg00_BasisError    &trError = Trans.trError_gg00;

    NumDataPagesRequested += CONV_NUM_SECURITY_PAGES; // add some pages for security reasons

    while (m_Active)
    {
        const SAPDB_Int4 numChangedPages = NumChangedPages();

        //const SAPDB_Bool pageNosExhausted =
        //    GetNumberOfUsedPages() + NumDataPagesRequested >= GetNumberOfAllPages();
       // PTS 1124331 TS 2003-09-23
        const SAPDB_Bool pageNosExhausted = ! IsSpaceAvailable( NumDataPagesRequested );

        const SAPDB_Bool blocksExhausted =
            ! FBM_Manager::Instance().IsSpaceAvailable( numChangedPages + NumDataPagesRequested );

        if( ! pageNosExhausted && ! blocksExhausted )
            return; // comman way to leave this method

        if(( blocksExhausted ) && SavepointIsPossible() && FBM_Manager::Instance().FreeAfterSVPExist() )
            StartSavepoint( Trans, numChangedPages ); // Do not check necessity of SVP

        if( g01restart_local )
        {

            if((  pageNosExhausted ) || ( blocksExhausted && ! m_SavepointIsRequested ))
            {
                Converter_Exception errMsg( __CONTEXT__, CONVERTER_DB_FULL_DURING_RESTART );
                RTE_Message( errMsg );

                trError = e_no_more_perm_space; // PTS 1111138 TS 2001-07-27
                return;
            }
        }

        if(( g01is_livecache() ) && ( ! SavepointNeeded() ))
            bd91StartOMSGarbageCollection( Trans, false );

        if( ! TransactionIsSuspendable( Trans ))
            return; // the task is qualified to proceed with his job

        const SAPDB_Bool bIsRollback = Trans.trState_gg00.includes( tsRollback_egg00 );

        if
        (
            (
                (
                    (SAPDB_UInt(Trans.trTempCount_gg00) > m_NumTempPagesUsed/20) &&
                    (m_NumTempPagesUsed                 > m_NumAllPages/100    ) &&
                    (SAPDB_UInt(Trans.trTempCount_gg00) > 100                  )
                )
                ||
                ( ! bAddDeviceIsPossible )
            )
            &&
            ( ! bIsRollback )
        )
        {
            trError = e_no_more_temp_space; // force rollback but do not suspend task

            if (!bAddDeviceIsPossible)
            {
                Converter_Exception NoAddDataDevspace (__CONTEXT__, CONVERTER_NO_ADD_DATA_DEVICE,
                                                       SAPDB_ToString( Trans.trTaskId_gg00, _T_d ));
                RTE_Message (NoAddDataDevspace);
            }
            else
            {
                Converter_Exception ForceRollback(__CONTEXT__,CONVERTER_FORCE_ROLLBACK_DUE_TO_DB_FULL,
                                                  SAPDB_ToString( Trans.trTaskId_gg00, _T_d ),
                                                  SAPDB_ToString(m_NumAllPages, _T_d),
                                                  SAPDB_ToString(m_NumTempPagesUsed, _T_d),
                                                  SAPDB_ToString(Trans.trTempCount_gg00, _T_d));
                RTE_Message (ForceRollback);
            }
            return;
        }

        if( blocksExhausted )
        {
            Converter_Exception errMsg( __CONTEXT__,CONVERTER_SUSPEND_TASK_DUE_TO_FBM_FULL,
                                        SAPDB_ToString( Trans.trTaskId_gg00, _T_d ),
                                        SAPDB_ToString( numChangedPages + NumDataPagesRequested, _T_d ),
                                        SAPDB_ToString( FBM_Manager::Instance().GetNumberOfUsedBlocks(), _T_d ),
                                        SAPDB_ToString( FBM_Manager::Instance().GetNumberOfFreeBlocks(), _T_d ));
            RTE_Message( errMsg );

            SAPDBTrace_Stream   traceStream( &Converter_Trace.GetTracePrinter() );

            traceStream << "*** " << errMsg.DateTime( SAPDB_TRUE ) << " "
            << errMsg.Type( SAPDB_TRUE ) << " "
            << errMsg.ID() << " "
            << errMsg.Component() << " " << NewLine
            << "*** " << errMsg.Message() << NewLine;

        }
        else // pageNosExhausted
        {
            Converter_Exception errMsg( __CONTEXT__,CONVERTER_SUSPEND_TASK_DUE_TO_DB_FULL,
                                        SAPDB_ToString( Trans.trTaskId_gg00, _T_d ),
                                        SAPDB_ToString( GetNumberOfUsedPages(), _T_d ),
                                        SAPDB_ToString( NumDataPagesRequested, _T_d ),
                                        SAPDB_ToString( GetNumberOfAllPages(), _T_d ));
            RTE_Message( errMsg );

            SAPDBTrace_Stream   traceStream( &Converter_Trace.GetTracePrinter() );

            traceStream << "*** " << errMsg.DateTime( SAPDB_TRUE ) << " "
            << errMsg.Type( SAPDB_TRUE ) << " "
            << errMsg.ID() << " "
            << errMsg.Component() << " " << NewLine
            << "*** " << errMsg.Message() << NewLine;
        }

        m_DBFullWaitQueue.InsertTaskAndWait( Trans.trTaskId_gg00 );

        if(( Trans.trRteCommPtr_gg00->to_cancel ) && ( ! bIsRollback ))
        {
            Converter_Exception errMsg( __CONTEXT__,CONVERTER_INFO_CANCELLED_DURING_DB_FULL,
                                        SAPDB_ToString( Trans.trTaskId_gg00, _T_d ));
            RTE_Message( errMsg );

            SAPDBTrace_Stream   traceStream( &Converter_Trace.GetTracePrinter() );

            traceStream << "*** " << errMsg.DateTime( SAPDB_TRUE ) << " "
            << errMsg.Type( SAPDB_TRUE ) << " "
            << errMsg.ID() << " "
            << errMsg.Component() << " " << NewLine
            << "*** " << errMsg.Message() << NewLine;

            trError  = e_cancelled;
            return;
        }
    }
    if( ! m_Active )
        trError = e_shutdown;

    return;
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::BeginColdVerify( const tsp00_TaskId    taskId )
{
    SAPDBERR_ASSERT_STATE (! BackUpIsActive());
    SAPDBERR_ASSERT_STATE (false == m_ColdVerifyIsActive);
    SAPDBERR_ASSERT_STATE (0     == m_NumStatConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumDynConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumPagesForBackUp);

    m_InMaintenanceMode  = true;
    m_ColdVerifyIsActive = true;

    m_PermStaticMapControl.InitializeBackUpPageNoList ( taskId );
    m_PermDynamicMapControl.InitializeBackUpPageNoList( taskId );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::EndColdVerify(
    const tsp00_TaskId  taskId,
    Data_PageCount      &NumStaticPermDataPages,
    Data_PageCount      &NumDynamicPermDataPages,
    Data_PageCount      &NumPermConvPages,
    Data_PageCount      &NumReleaseBlocks )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::EndColdVerify", Converter_Trace, 5 );

    SAPDBERR_ASSERT_STATE (! BackUpIsActive());
    SAPDBERR_ASSERT_STATE (true == m_ColdVerifyIsActive);
    SAPDBERR_ASSERT_STATE (0     == m_NumStatConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumDynConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumPagesForBackUp);

    SAPDB_UInt NumPermStaticConvPages   = 0;
    SAPDB_UInt NumPermDynamicConvPages  = 0;
    SAPDB_UInt NumStaticTempDataPages   = 0;
    SAPDB_UInt NumDynamictempDataPages  = 0;
    SAPDB_UInt numStaticReleasedBlocks  = 0;
    SAPDB_UInt numDynamicReleasedBlocks = 0;

    m_PermStaticMapControl.ReleaseUnusedPageNos( taskId, NumStaticPermDataPages,
            NumPermStaticConvPages, numStaticReleasedBlocks );

    m_PermDynamicMapControl.ReleaseUnusedPageNos( taskId, NumDynamicPermDataPages,
            NumPermDynamicConvPages, numDynamicReleasedBlocks );

    NumPermConvPages = NumPermStaticConvPages + NumPermDynamicConvPages;
    NumReleaseBlocks = numStaticReleasedBlocks + numDynamicReleasedBlocks;

    m_NumPermPagesUsed = NumStaticPermDataPages + NumDynamicPermDataPages;

    m_PermStaticMapControl.ClearBackUpPageNoList  ();
    m_PermDynamicMapControl.ClearBackUpPageNoList ();

    m_ColdVerifyIsActive = false;
    m_InMaintenanceMode  = false;;
}


/*---------------------------------------------------------------------------*/

void
Converter_Converter::EndFailedColdVerify ( const tsp00_TaskId  TaskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::EndFailedColdVerify", Converter_Trace, 5 );

    SAPDBERR_ASSERT_STATE (! BackUpIsActive());
    SAPDBERR_ASSERT_STATE (true == m_ColdVerifyIsActive);
    SAPDBERR_ASSERT_STATE (0     == m_NumStatConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumDynConvPagesForBackUp);
    SAPDBERR_ASSERT_STATE (0     == m_NumPagesForBackUp);

    m_PermStaticMapControl.EndFailedColdVerify  (TaskId);
    m_PermDynamicMapControl.EndFailedColdVerify (TaskId);

    m_ColdVerifyIsActive = false;
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::GetDbSpaceStatistics (tsp00_TaskId        TaskId,
        tgg00_DbSpaceInfo  &DbSpaceInfo)
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::GetDbSpaceStatistics", Converter_Trace, 5 );

    DbSpaceInfo.db_total_pages    = IOMan_IDataInfo::GetInstance().TotalDataPages();
    DbSpaceInfo.db_perm_pages     = IOMan_IDataInfo::GetInstance().TotalDataPages();
    DbSpaceInfo.db_real_perm_used = m_NumPermPagesUsed + m_NumPermConvPages;
    DbSpaceInfo.db_perm_percent   = (tsp00_Int2)((DbSpaceInfo.db_real_perm_used * 100.0)/DbSpaceInfo.db_total_pages);
    DbSpaceInfo.db_temp_pages     = m_NumTempPagesUsed;
    DbSpaceInfo.db_temp_percent   = (tsp00_Int2)((m_NumTempPagesUsed*100.0)/DbSpaceInfo.db_total_pages);
    DbSpaceInfo.db_used_pages     = GetNumberOfUsedPages ();
    DbSpaceInfo.db_updated_pages  = m_PermStaticMapControl.NumUpdatedPermPages(TaskId) +
                                    m_PermDynamicMapControl.NumUpdatedPermPages(TaskId);
    DbSpaceInfo.db_max_perm_used  = m_MaxPermUsed;
    DbSpaceInfo.db_used_blocks    = FBM_Manager::Instance().GetNumberOfUsedBlocks();
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::IsDBFull(
    const tsp00_TaskId taskId ) const
{
    if( ! m_DBFullWaitQueue.IsEmpty( taskId ))
        return( SAPDB_TRUE ); // task is suspended because of db full

    const SAPDB_UInt numDataPagesRequested = CONV_NUM_SECURITY_PAGES * 2;// see HandledbFull

    //if( GetNumberOfUsedPages() + numDataPagesRequested >= GetNumberOfAllPages())
    //    return( SAPDB_TRUE ); // pageNos exhausted
   
    // PTS 1124331 TS 2003-09-23
    if( ! IsSpaceAvailable( numDataPagesRequested ))
        return( SAPDB_TRUE ); // pageNos exhausted

    if( ! FBM_Manager::Instance().IsSpaceAvailable( NumChangedPages() + numDataPagesRequested ))
        return( SAPDB_TRUE ); // blocks in FBM exhausted

    return( SAPDB_FALSE );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::SecuritySpaceReached() const
{
    // PTS 1124331 TS 2003-09-23

    if( ! IsSpaceAvailable( CONV_NUM_SECURITY_PAGES ))
        return( SAPDB_TRUE );

    if( ! FBM_Manager::Instance().IsSpaceAvailable( NumChangedPages()))
        return( SAPDB_TRUE );
    
    return( SAPDB_FALSE );
}

/*---------------------------------------------------------------------------*/

Converter_GarbageCollectionReturnCode
Converter_Converter::GarbageCollectionNeeded() const // PTS 1116157 FF 2002-06-14
{
    if( ! SavepointIsPossible() )
        return( Converter_GCNotWanted );

    if( SavepointNeeded() ) // Savepoint will come immediately started by TimeOut Task
        return( Converter_GCNotWanted );

    // data base is filled up to 90 percent inclusive the changed pages of the data cache

    if( FBM_Manager::Instance().GarbageCollectionNeeded( NumChangedPages() ))
        return( Converter_GCNeeded );
    else
        return( Converter_GCNotNeeded );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::StartSavepoint( tgg00_TransContext &trans )
{
    if( ! SavepointNeeded() )
        return; // savepoint is not needed

    StartSavepoint( trans, NumChangedPages() );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::SavepointNeeded() const
{
    if( ! SavepointIsPossible() )
        return( false );

    if( FBM_Manager::Instance().SavepointNeeded() ) // more blocks in state free after SVP than free
        return( true );

    if( NumChangedPages() > (( FBM_Manager::Instance().GetNumberOfFreeBlocks()) / 3 ))
        return( true );

    return( false );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::Dump(
    const tsp00_TaskId  TaskId,
    Kernel_Dump         &Dump,
    const SAPDB_Bool    bEnterRegion )
{
    struct ConverterManager   CM;

    CM.dmpVersion                         = m_ConverterVersion;
    CM.dmpNumPermPagesUsed                = m_NumPermPagesUsed;
    CM.dmpNumTempPagesUsed                = m_NumTempPagesUsed;
    CM.dmpNumAllPages                     = m_NumAllPages;
    CM.dmpNumPagesForBackup               = m_NumPagesForBackUp;
    CM.dmpNumStatConvPagesForBackup       = m_NumStatConvPagesForBackUp;
    CM.dmpNumDynConvPagesForBackup        = m_NumDynConvPagesForBackUp;
    CM.dmpNumPermConvPages                = m_NumPermConvPages;
    CM.dmpNumTempConvPages                = m_NumTempConvPages;
    CM.dmpMapBlockingSize                 = m_MapBlockingSize;
    CM.dmpConverterIsActive               = m_Active;
    CM.dmpSavePagesIsActive               = m_SavePagesIsActiv;
    CM.dmpSaveDataIsActive                = m_SaveDataIsActiv;
    CM.dmpRestoreDataIsActive             = m_RestoreDataActive;
    CM.dmpRestorePagesIsActive            = m_RestorePagesActive;
    CM.dmpColdVerifyIsActive              = m_ColdVerifyIsActive;
    CM.dmpLastSaveDataSuccessful          = m_LastSaveDataSuccessful;
    CM.dmpSavepointIsRequested            = m_SavepointIsRequested;

    if( m_ConverterRootBlockAddress.IsValid() )
    {
        CM.dmpRootDevNo   = m_ConverterRootBlockAddress.GetDeviceNo();
        CM.dmpRootBlockNo = m_ConverterRootBlockAddress.GetBlockNo();
    }
    else
    {
        CM.dmpRootDevNo   = IOMan_DeviceNo();
        CM.dmpRootBlockNo = IOMan_BlockNo();
    }

    Dump.InsertEntry( Kernel_Dump::DmpConverterManager, Kernel_DumpPage::Entry( &CM, sizeof( CM )));

    m_Index.Dump( Dump );

    m_TempMapControl.Dump( TaskId, Dump, bEnterRegion );
    m_PermDynamicMapControl.Dump( TaskId, Dump, bEnterRegion );
    m_PermStaticMapControl.Dump( TaskId, Dump, bEnterRegion );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::SetBlockAddressForMigratedPage (
    const tsp00_TaskId          taskId,
    const IOMan_BlockAddress    &block,
    const Data_PageNo           &pageNo,
    const SAPDB_Bool            bSetToOccupied )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::SetBlockAddressForMigratedPage", Converter_Trace, 5 );

    SAPDBERR_ASSERT_STATE( m_RestoreDataActive );

    // At this time the corresponding converter page entry is invalid (initialized),
    // because the converter is build up the current migration step.

    m_PermDynamicMapControl.SetBlockAddressOfRestoredPage(
        taskId, pageNo, block, m_RestoreDataActive);

    m_NumPermPagesUsed.Increment();

    CheckFillingOfDataBase();

    if( SAPDB_TRUE != bSetToOccupied )
        return;

    // After updating the converter page entry, the block is set to occupied within
    // the FBM. Note that the given block must not specify one of the reserved blocks
    // in the data devspaces.

    if( SAPDB_TRUE != FBM_Manager::Instance().SetBlockStateToOccupied( taskId, block ))
    {
        Converter_Exception errMsg( __CONTEXT__, CONVERTER_BAD_ENTRY_STATE,
                                    "dynamic", SAPDB_ToString( pageNo, _T_d ));
        RTE_Crash( errMsg );
    }
}

/*---------------------------------------------------------------------------*/

IOMan_BlockAddress
Converter_Converter::DetermineAndSetBlockAddressForMigratedPage (
    const tsp00_TaskId  taskId,
    const Data_PageNo   &pageNo )
{
    const SAPDB_Bool         bSetToOccupied = SAPDB_FALSE;
    const IOMan_BlockAddress newBlock       = FBM_Manager::Instance().GetFreeBlock( taskId );

    // Note that the newBlock is set to occupied by the GetFreeBlock
    // method and therefore not set again to occupupied by the subsequent call.

    SetBlockAddressForMigratedPage( taskId, newBlock, pageNo, bSetToOccupied );

    return( newBlock );
}

//----------------------------------------------------------------------------
//  private methods
//----------------------------------------------------------------------------

void
Converter_Converter::InitializeBase( const SAPDB_UInt NumAllPages )
{
    m_NumPermPagesUsed                 = 0;
    m_NumTempPagesUsed                 = 0;
    m_NumPagesForBackUp                = 0;
    m_NumStatConvPagesForBackUp        = 0;
    m_NumDynConvPagesForBackUp         = 0;
    m_NumPermConvPages                 = 0;
    m_NumTempConvPages                 = 0;
    m_MaxPermUsed                      = 0;
    m_MaxUsedDynamicPageNo             = 0;
    m_MaxUsedStaticPageNo              = 0;
    m_NumAllPages                      = NumAllPages;
 // PTS 1124331 TS 2003-09-23
    m_SecuritySpaceLimit               = ( m_NumAllPages - SAPDB_MIN( m_NumAllPages/25, 4000));
    m_Active                           = true;
    m_InMaintenanceMode                = false;
    m_SavepointIsRequested             = false;
    m_SavePagesIsActiv                 = false;
    m_SaveDataIsActiv                  = false;
    m_RestoreDataActive                = false;
    m_RestorePagesActive               = false;
    m_ColdVerifyIsActive               = false;
    m_LastSaveDataSuccessful           = true;
    m_DBUsageMonitor.Rescale (m_NumAllPages);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::ExpandMap (
    const tsp00_TaskId           taskId,
    const Data_PageRecoveryMode  &recoveryMode,
    const SAPDB_UInt             numPagesAdded)
{
    Converter_ReturnCode    retCode;
    const SAPDB_UInt        newAllPages = m_NumAllPages + numPagesAdded;

    do
    {
        if( recoveryMode.PageIsRecoverable() )
        {
            if( recoveryMode.UseDynamicPageAddressing() ){
                retCode = m_PermDynamicMapControl.Expand( taskId, newAllPages,
                          m_ExpandConverterWaitQueue );
            }
            else{
                retCode = m_PermStaticMapControl.Expand( taskId, newAllPages,
                          m_ExpandConverterWaitQueue );
            }
        }
        else{
            retCode = m_TempMapControl.Expand( taskId, newAllPages );
        }

        switch( retCode )
        {
        case Converter_Ok:
            return true;
        case Converter_SavePointActive:
            m_ExpandConverterWaitQueue.Wait( taskId );
            break;
        default:
            {
                Converter_Exception ExpandFailed(__CONTEXT__,CONVERTER_EXPAND_FAILED,
                                                 SAPDB_ToString(m_NumAllPages, _T_d),
                                                 SAPDB_ToString(newAllPages, _T_d));
                RTE_Message (ExpandFailed);
            }
            return false;
        }
    }
    while(m_Active);

    return false;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Converter_Converter::FreePermPageNo(
    const tsp00_TaskId          taskId,
    Converter_PermMapControl    &mapControl,
    const Data_PageNo           pageNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Converter::FreePermPageNo", Converter_Trace, 5 );

    SAPDBTRACE_WRITELN( Converter_Trace, 5, "Free perm PageNo: " << pageNo );

    if( ! m_Active ){
        return( SAPDB_FALSE );
    }

    SAPDB_Bool           bWaitUntilSavepointFinished  = false;
    SAPDB_Bool           bOldBlockIsUsableImmediately = false;

    do
    {
        mapControl.FreePageNo( taskId, pageNo,
                               m_FreePageNoWaitQueue, bWaitUntilSavepointFinished,
                               bOldBlockIsUsableImmediately );

        if( bWaitUntilSavepointFinished )
        {
            m_FreePageNoWaitQueue.Wait( taskId );

            if( !m_Active ){
                return( SAPDB_FALSE );
            }
        }
    }
    while( bWaitUntilSavepointFinished );

    m_NumPermPagesUsed.Decrement();

    SAPDBTRACE_WRITELN( Converter_Trace, 7, "PermPagesUsed: " << m_NumPermPagesUsed );

    CheckFillingOfDataBase();

    if(( bOldBlockIsUsableImmediately ) && ( ! m_DBFullWaitQueue.IsEmpty( taskId ) )){
        m_DBFullWaitQueue.ResumeAllWaitingTasks( taskId );
    }
    return( SAPDB_TRUE );
}

/*---------------------------------------------------------------------------*/

void
Converter_Converter::StartSavepoint(
    tgg00_TransContext &trans,
    const SAPDB_Int4    numChangedPages )
{
    Converter_Exception errMsg( __CONTEXT__, CONVERTER_SAVEPOINT_NEEDED,
                                SAPDB_ToString( numChangedPages, _T_d ),
                                SAPDB_ToString( FBM_Manager::Instance().GetNumberOfUsedBlocks(), _T_d ),
                                SAPDB_ToString( FBM_Manager::Instance().GetNumberOfFreeBlocks(), _T_d ));
    RTE_Message( errMsg );

    m_SavepointIsRequested = true;

    Log_SavepointManager.StartSavepoint( trans, Log_SVPReasonConverter );
}

/*---------------------------------------------------------------------------*/

SAPDB_Int4
Converter_Converter::NumChangedPages() const
{
    return( bd20GetNumberOfChangedPages() + GetNumberOfChangedConvPages());
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
