/*!**************************************************************************

  module      : Converter_Converter.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author      : AlexanderK

  special area: converter 
  description : general converter header file


  last changed: 2001-01-10  13:36
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_CONVERTER_HPP
#define CONVERTER_CONVERTER_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Converter/Converter_IBackUp.hpp"
#include "Converter/Converter_IDump.hpp"
#include "Converter/Converter_IPageIO.hpp"
#include "Converter/Converter_ISavePoint.hpp"
#include "Converter/Converter_IManager.hpp"
#include "Converter/Converter_IMigration.hpp"
#include "Converter/Converter_ICommon.hpp"
#include "Converter/Converter_IPageNoMapping.hpp"
#include "Converter/Converter_IPageNoManager.hpp"
#include "Converter/Converter_IVerify.hpp"
#include "Converter/Converter_IEvent.hpp"
#include "Converter/Converter_Index.hpp"
#include "Converter/Converter_MapControl.hpp" 
#include "Converter/Converter_TempMapControl.hpp" 
#include "Converter/Converter_PermMapControl.hpp" 
#include "Converter/Converter_PermStaticMapControl.hpp" 
#include "Converter/Converter_PageFrameAllocator.hpp"
#include "Converter/Converter_Types.hpp"
#include "Converter/Converter_WaitQueue.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "RunTime/MemoryManagement/RTEMem_AllocatorWrapper.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"
#include "DataAccess/Data_Types.hpp"
#include "KernelCommon/Kernel_Event.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define CONV_SUSPEND_HANDLE_DB_FULL                 197
#define CONV_SUSPEND_FREE_PAGENO_WAIT_DURING_SVP    236
#define CONV_SUSPEND_EXPAND_WAIT_DURING_SVP         237

#define CONV_INCREMENT_VERSION                      true

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


/*!*****************************************************************************

   class: Converter_Converter

*******************************************************************************/

class Converter_Converter : public Converter_IBackUp,
            public Converter_IDump,
            public Converter_IPageIO,
            public Converter_ISavePoint,
            public Converter_IManager,
            public Converter_IMigration,
            public Converter_IPageNoManager,
            public Converter_IPageNoMapping,
            public Converter_ICommon,
            public Converter_IVerify,
            public Converter_IEvent
{
    friend class Converter_PageFrameAllocator;

public:


    /* -----------------------------------------------------------------------*/
    /*! chapter: Singelton                                                    */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     CreateInstance() 
    description:  static method to create a singelton of the type Converter_Converter
    arguments:    none
    return value: (Converter_Converter&) reference to the instance of Converter_Converter 
    --------------------------------------------------------------------------*/
    static        Converter_Converter& CreateInstance ();

    /*!-------------------------------------------------------------------------
    function:     Instance() 
    description:  returns the reference to the sigelton instance of Converter_Converter
    arguments:    none
    return value: (Converter_Converter&) reference to the instance of Converter_Converter 
    --------------------------------------------------------------------------*/
    static Converter_Converter & Instance  ()
    {
        SAPDBERR_ASSERT_STATE (m_Instance != NULL);
        return *m_Instance;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Singelton                                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Common                                                       */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Version()
    description:  Returns the actual converter version

    arguments:    none
    return value: (Converter_Version)
    --------------------------------------------------------------------------*/
    virtual Converter_Version Version () const
    {
        return m_ConverterVersion;
    }

    /*!-------------------------------------------------------------------------
    function:     BackUpIsActive()
    description:  Checks whether a backup is active

     arguments:    none
     return value: (SAPDB_Bool) true if active
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool BackUpIsActive () const
    {
        return (m_SaveDataIsActiv || m_SavePagesIsActiv);
    }

    /*!-------------------------------------------------------------------------
    function:     GetNumberOfAllPages()
    description:  Returns the numbers of all  pages 

      - returns the number of the maximum number of pages which
      can be accomodated in the database 
      
        arguments:    none
        return value: number of pages which can be stored in the database
    --------------------------------------------------------------------------*/
    virtual Data_PageCount GetNumberOfAllPages () const
    {
        return  m_NumAllPages;
    }

    /*!-------------------------------------------------------------------------
    function:     GetNumberOfUsedPages()
    description:  Returns the numbers of all  currently used pages 

      - returns  the number of all pages
        which are currently occupied by data and converter pages
                  
    arguments:     none
    return value:  number of pages currently used
    --------------------------------------------------------------------------*/
    virtual Data_PageCount GetNumberOfUsedPages () const
    {
        return (m_NumPermPagesUsed +  m_NumPermConvPages + m_NumTempPagesUsed);
    }

    /*!-------------------------------------------------------------------------
    function:     GetDbSpaceStatistics()
    description:  returns a number of data describing the filling state of the converter 
                  
    arguments:     TaskId      [in] task id
                   DbSpaceInfo [out] collection of state data of the converter
    return value:  none
    --------------------------------------------------------------------------*/
    virtual void GetDbSpaceStatistics (tsp00_TaskId        TaskId,
                                       tgg00_DbSpaceInfo  &DbSpaceInfo);

    /*!-------------------------------------------------------------------------
    function:     UsedConverterPages()
    description:  Returns the numbers of all currently used converter pages 

    arguments:     none
    return value:  number of converter pages (perm and temp) currently used
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt UsedConverterPages() const
    {
        return( m_NumPermConvPages + m_NumTempConvPages );
    }

    /*!-------------------------------------------------------------------------
    function:     IsDBFull()
    description:  True is returned, if the data area is (nearly) full.
    arguments:    TaskId [in] task id
    return value: (bool)
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool IsDBFull(
        const tsp00_TaskId taskId ) const;

    /*!-------------------------------------------------------------------------
    function:     SecuritySpaceReached()
    description:  If the database filling is nearly full and therefore a 
                  sensible work is not possible this method returns a true
                  else false.
    arguments:    none
    return value: true if security is reached and database nearly full else false
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool SecuritySpaceReached() const;

    /*!-------------------------------------------------------------------------
    function:     GarbageCollectionNeeded()
    description:  If the database filling is about 90 percent and the converter
                  is online and not in maintenance mode, for example cold verify, 
                  diagnose repair or save data cold, this method returns an enum 
                  to  signalize that garbage collector support is needed. The
                  value Converter_GCNeeded determines that a garbage collection 
                  is needed.
    arguments:    none
    return value: Converter_GarbageCollectionReturnCode
    --------------------------------------------------------------------------*/
    virtual Converter_GarbageCollectionReturnCode GarbageCollectionNeeded() const;

    /*!-------------------------------------------------------------------------
    function:     StartSavepoint()
    description:  If the number of changed converter pages and changed pages
                  within data cache is nearly to the number of free blocks within
                  FBM a savepoint is started. As long as this requested 
                  savepoint is finished this method will not request a new savepoint.
                  This will be done only if the converter is online and not 
                  in maintenance mode.
    arguments:    trans [in] transaction context
    return value: none
    --------------------------------------------------------------------------*/
    void StartSavepoint( tgg00_TransContext &trans );

    /*!-------------------------------------------------------------------------
    function:     SavepointNeeded()
    description:  If the number of changed converter pages and changed pages
                  within data cache is nearly to the number of free blocks within
                  FBM a savepoint is needed. This will be done only if the 
                  converter is online and not in maintenance mode.
    arguments:    none
    return value: true if a savepoint is needed; else false
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool SavepointNeeded() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Common                                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Manager                                                      */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Restart
    description:  Restart the  converter

                  - The internal structures of the converter are created.
                  - Converter pages are read from devices. The converter root
                    page can be found at address 'rootaddr'.
                  - The parameters 'MaxDynamicPageNo', 'MaxStaticPageNo' and  'NumAllPages' are used 
                    to determine the initial size of the  converter.
                  - 'ConverterVersion' is the converter version of the database to be
                    recovered. If 'bIncConverterVersion' is true, the converter version
                    is incremented after a successful restart.
                  

    arguments:    TaskId                    [in] own Task ID
                  ConverterVersion          [in] converter version from restart record
                  ConverterRootBlockAddress [in] i/o address of converter root page
                  bIncConverterVersion      [in] indicates whether version must be
                                                 incremented after the restart
                  bLastSaveDataSuccessful   [in] indicates whether the last save data finished successfully
                  MaxDynamicPageNo          [in] maximun used dynamic page number of the restarted database
                  MaxStaticPageNo           [in] maximun used static page number of the restarted database 
                  NumAllPages               [in] maximum number of pages which can be stored in the data base. this
                                                 corresponds to the sum of size of all devices
                  BaseRegionId              [in] see next parameter
                  NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                 converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                 The more regions are supplied the better parallel accesses can be handled
                  ConverterPageSize         [in]  total size of a converter page in byte
                  DevNoBitCount             [in] defines the number of bits used in each converter 
                                                 page entry to address a device   

    return value: none
    --------------------------------------------------------------------------*/
    virtual void Restart (const tsp00_TaskId        TaskId,
                          const Converter_Version  &ConverterVersion,
                          const IOMan_BlockAddress &ConverterRootBlockAddress,
                          const SAPDB_Bool          bIncConverterVersion,
                          const SAPDB_Bool          bLastSaveDataSuccessful,
                          const Data_PageNo         MaxDynamicPageNo,
                          const Data_PageNo         MaxStaticPageNo,
                          const SAPDB_UInt          NumAllPages,
                          const tsp00_RegionId      BaseRegionId,
                          const SAPDB_UInt          NumConverterRegions,
                          const SAPDB_UInt          ConverterPageSize,
                          const SAPDB_UInt1         DevNoBitCount);


    /*!-------------------------------------------------------------------------
    function:     ReadPages()
    description:  This method is used to read a converter leaf page from a device
    arguments:    TaskId     [in] own Task ID
                  pagerIndex [in] logical id of the calling pager task
                  totalPages [out] number of read pages
                  totalIO    [out] number of io's made to read the pages
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadPages(
        const tsp00_TaskId  taskId,
        const SAPDB_UInt    pagerIndex,
        SAPDB_UInt          &totalPages,
        SAPDB_UInt          &totalIO );

    /*!-------------------------------------------------------------------------
    function:     RestartForRestore
    description:  Restart the converter for restore pages.

                  - The internal structures of the converter are restarted
                  - Converter pages are read from devices. The converter root
                    page can be found at address 'rootaddr'.
                  - The parameters 'MaxDynamicPageNo', 'MaxStaticPageNo' and  'NumAllPages' are used 
                    to determine the initial size of the  converter.
                  - 'ConverterVersion' is the converter version of the database to be
                    recovered. If 'bIncConverterVersion' is true, the converter version
                    is incremented after a successful restart.
                  

    arguments:    TaskId                    [in] own Task ID
                  ConverterVersion          [in] converter version from restart record
                  ConverterRootBlockAddress [in] i/o address of converter root page
                  bLastSaveDataSuccessful   [in] indicates whether the last save data finished successfully
                  MaxDynamicPageNo          [in] maximun used dynamic page number of the restarted database
                  MaxStaticPageNo           [in] maximun used static page number of the restarted database 
                  NumAllPages               [in] maximum number of pages which can be stored in the data base. this
                                                 corresponds to the sum of size of all devices
                  BaseRegionId              [in] see next parameter
                  NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                 converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                 The more regions are supplied the better parallel accesses can be handled
                  ConverterPageSize         [in]  total size of a converter page in byte
                  DevNoBitCount             [in] defines the number of bits used in each converter 
                                                 page entry to address a device   

    return value: none
    --------------------------------------------------------------------------*/
    virtual void RestartForRestore (
        const tsp00_TaskId        TaskId,
        const Converter_Version  &ConverterVersion,
        const IOMan_BlockAddress &ConverterRootBlockAddress,
        const SAPDB_Bool          bLastSaveDataSuccessful,
        const Data_PageNo         MaxDynamicPageNo,
        const Data_PageNo         MaxStaticPageNo,
        const SAPDB_UInt          NumAllPages,
        const tsp00_RegionId      BaseRegionId,
        const SAPDB_UInt          NumConverterRegions,
        const SAPDB_UInt          ConverterPageSize,
        const SAPDB_UInt1         DevNoBitCount)
    {
        Restart( TaskId, ConverterVersion, ConverterRootBlockAddress, ! CONV_INCREMENT_VERSION,
                 bLastSaveDataSuccessful, MaxDynamicPageNo, MaxStaticPageNo, NumAllPages,
                 BaseRegionId, NumConverterRegions, ConverterPageSize, DevNoBitCount);

        m_InMaintenanceMode  = true;
        m_RestorePagesActive = true;
    }

    /*!-------------------------------------------------------------------------
     function:     Create
     description:  Creates an empty converter for a new data base (needed for activate serverdb)

            - The internal structures of the converter are created.
            - The parameter 'NumAllPages' is used to determine the initial size of the  converter.
           
     arguments:    TaskId                    [in] task id
                   NumAllPages               [in] maximum number of pages which can be stored in the data base. this
                                                  corresponds to the sum of size of all devices
                   BaseRegionId              [in] see next parameter
                   NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                  converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                  The more regions are supplied the better parallel accesses can be handled
                   ConverterPageSize         [in] total size of a converter page in byte
                   DevNoBitCount             [in] defines the number of bits used in each converter 
                                                  page entry to address a device  
     return value: none
     --------------------------------------------------------------------------*/
    /* creates an empty converter (needed for activate serverdb) */
    virtual void Create (
        const tsp00_TaskId       TaskId,
        const SAPDB_UInt         NumAllPages,
        const tsp00_RegionId     BaseRegionId,
        const SAPDB_UInt         NumConverterRegions,
        const SAPDB_UInt         ConverterPageSize,
        const SAPDB_UInt1        DevNoBitCount);

    /*!-------------------------------------------------------------------------
    function:     CreateForRestore
    description:  Creates an empty converter for restore data

                  - The internal structures of the converter are created and
                    initialized.
                  - The parameters 'MaxDynamicPageNo', 'MaxStaticPageNo' and  'NumAllPages' 
                    are used to determine the initial size of the  converter.
                  - 'ConverterVersion' is the converter version of the database to be
                    recovered.

    arguments:    TaskId                    [in] task id
                  ConverterVersion          [in] converter version from restart record
                  bLastSaveDataSuccessful   [in] indicates whether the last save data finished successfully
                  MaxDynamicPageNo          [in] maximun used dynamic page number of the restarted database
                  MaxStaticPageNo           [in] maximun used static page number of the restarted database 
                  NumAllPages               [in] maximum number of pages which can be stored in the data base. this                                              corresponds to the sum of size of all device
                  BaseRegionId              [in] see next parameter
                  NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                 converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                 The more regions are supplied the better parallel accesses can be handled
                  ConverterPageSize         [in] total size of a converter page in byte
                  DevNoBitCount             [in] defines the number of bits used in each converter 
                                                 page entry to address a device             
    return value: none
    --------------------------------------------------------------------------*/
    virtual void  CreateForRestore (tsp00_TaskId              TaskId,
                                    const Converter_Version  &ConverterVersion,
                                    const SAPDB_Bool          bLastSaveDataSuccessful,
                                    const Data_PageNo         MaxDynamicPageNo,
                                    const Data_PageNo         MaxStaticPageNo,
                                    const SAPDB_UInt          NumAllPages,
                                    const tsp00_RegionId      BaseRegionId,
                                    const SAPDB_UInt          NumConverterRegions,
                                    const SAPDB_UInt          ConverterPageSize,
                                    const SAPDB_UInt1         DevNoBitCount);

    /*!-------------------------------------------------------------------------
    function:     Shutdown
    description:  Shuts the converter down

                  - All resources held by the converter are released and all 
                    member data are reseted to their initial values         
                  - All tasks waiting for the converter are resumed.

    arguments:    TaskId [in,out]    task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void Shutdown (const tsp00_TaskId TaskId);

    /*!-------------------------------------------------------------------------
     function:     FlushSerial
     description:  Saves converter to devspace (sequential writing)

                   - All converter pages which have been changed since the last
                     savepoint and which refer to permanent data pages are
                     written out to data devspaces.
                   - The routine returns several parameters describing the actual
                     state of the converter.
                   - If the flush is triggered by a SAVEPOINT the returned parameters are not 
                     needed since they were already inserted into the restart record by 
                     the Converter_ISavePoint::EndSavePoint method.
                     If the converter is not active the return value is false;

     arguments:    TaskId                    [in]   current task id
                   ConverterVersion          [out]  new valid converter version
                   ConverterRootBlockAddress [out]  Block address of the converter root page (to be stored in the restart record
                   MaxDynamicPageNo          [out]  maximun used dynamic page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (dynamic map) after a restart.
                   MaxStaticPageNo           [out]  maximun used static page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (static map) after a restart. 
                                                    (see Converter_IManager::Restart)
     return value: SAPDB_Bool
     --------------------------------------------------------------------------*/
    /* flushes all changed permanent page frames held by the converter are flushed to a device */
    virtual SAPDB_Bool FlushSerial (
        const tsp00_TaskId  taskId,
        Converter_Version   &ConverterVersion,
        Data_PageNo         &MaxStaticPageNo,
        Data_PageNo         &MaxDynamicPageNo,
        IOMan_BlockAddress  &ConverterRootBlockAddress);

    /*!-------------------------------------------------------------------------
     function:     FlushParallel
     description:  Saves converter to devspace (parallel writing)

                   - All converter pages which have been changed since the last
                     savepoint and which refer to permanent data pages are
                     written out to data devspaces.
                   - The routine returns several parameters describing the actual
                     state of the converter.

     arguments:    Trans [in,out]     transaction context
                   ConverterVersion          [out]  new valid converter version
                   ConverterRootBlockAddress [out]  Block address of the converter root page (to be stored in the restart record
                   MaxDynamicPageNo          [out]  maximun used dynamic page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (dynamic map) after a restart.
                   MaxStaticPageNo           [out]  maximun used static page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (static map) after a restart. 
                                                    (see Converter_IManager::Restart)
     return value: none
     --------------------------------------------------------------------------*/
    /* flushes all changed permanent page frames held by the converter are flushed to a device */
    virtual void FlushParallel (tgg00_TransContext  &Trans,
                                Converter_Version   &ConverterVersion,
                                Data_PageNo         &MaxStaticPageNo,
                                Data_PageNo         &MaxDynamicPageNo,
                                IOMan_BlockAddress  &ConverterRootBlockAddress,
                                SAPDB_UInt          &NumPages,
                                SAPDB_UInt          &NumIO);

    /*!-------------------------------------------------------------------------
     function:     WritePages
     description:  Flush static and dynamic changed converter leaf pages of one 
                   section to devspace
     arguments:    TaskId     [in]  task id
                   iSection   [in]  converter section to flush
                   totalPages [out] number of changed pages
                   totalIO    [out] number of io's made to flush the pages
     return value: none
     --------------------------------------------------------------------------*/

    virtual void WritePages(
        const tsp00_TaskId  taskId,
        const SAPDB_UInt    iSection,
        SAPDB_UInt          &totalPages,
        SAPDB_UInt          &totalIO );

    /*!-------------------------------------------------------------------------
    function:     Expand()
    description:  Expands the converter

                  - Expands the converter to to manage NumPagesAdded more data pages 
                  - The routine is applicable with database in both cold and
                    warm state.
                  - The internal structures of the converter are adapted.

    arguments:    TaskId        [in] current task id
                  DevNo         [in] logical identifier for the added data devspace
                  DevSize       [in] size of the added data devspace in pages.
                  NumPagesAdded [in] number of pages added and usable by the data base (by assigning a new device)
    return value: none
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Expand (
        const tsp00_TaskId    TaskId,
        const IOMan_DeviceNo &DevNo,
        const SAPDB_UInt     DevSize,
        const SAPDB_UInt     NumPagesAdded);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Manager                                                   */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Back Up                                                      */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     BeginSaveData
    description:  Starts a total backup (also called save data)

          - All currently occupied blocks of the database are marked within the FBM 
            to be included in the back up

    arguments:    TaskId  [in]   task id

    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginSaveData (tsp00_TaskId TaskId);

    /*!-------------------------------------------------------------------------
    function:     BeginSavePages
    description:  Starts a differential backup (also called save pages)
        
          - All currently occupied blocks of the database which are marked as changed since the last 
          total backup are marked within the FBM  to be included in the back up.
          - A list off all page numbers used in the current data base is created
          
    arguments:    TaskId  [in]   task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginSavePages (tsp00_TaskId TaskId);

    /*!-------------------------------------------------------------------------
    function:     EndSave
    description:  Ends the backup of data or pages

          - The back up active flag in the converter is reset
          - If  the current back up  did not finish successfully all
            blocks still marked as to be written to the back up files are unmarked.

    arguments:    TaskId [in]                      current task id 
                  bBackUpSuccessfullyFinished [in] indicates whether the current back
                                                   up finished successfully
    return value: none
    --------------------------------------------------------------------------*/
    virtual void  EndSave  (const tsp00_TaskId &TaskId,
                            SAPDB_Bool          bBackUpSuccessfullyFinished);

    /*!-------------------------------------------------------------------------
    function:     GetNumPagesForBackUp
    description:  Returns the number of data pages marked for back up

    arguments:    none
    return value: Number of blocks marked for back up
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetNumPagesForBackUp()
    {
        return m_NumPagesForBackUp;
    }

    /*!-------------------------------------------------------------------------
    function:     GetNumConvPagesForBackUp
    description:  Returns the number of converter pages required to address all 
                  static data pages 

       - Returns the number of converter pages required to address all 
                 data pages which belong to the current backup

    arguments:    none
    return value: Number of converter pages required to address all 
                  data pages which belong to the current backup
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetNumConvPagesForBackUp()
    {
        return (m_NumStatConvPagesForBackUp + m_NumDynConvPagesForBackUp);
    }

    /*!-------------------------------------------------------------------------
    function:     CheckBackUp
    description:  Checks if the current back up finished successfully

                 - Checks if all pages marked as to be included in 
                   the back up were fetched by the back up manager

    arguments:    TaskId [in]  current task id 
    return value: true if the back was successful else false
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool CheckBackUp( const tsp00_TaskId taskId ) const;

    /*!-------------------------------------------------------------------------
    function:     ResetDataPageAfterBackUp
    description:  Signals the converter that a given page was written into the back up file 

                 - Resets the flag in the FBM which displays that the given page still has to be
                   written to the back up file 
                 - Removes the page number from the list of page numbers still to inlclude in the back up 

    arguments:    TaskId           [in]  current task id 
                  BlockAddress     [in]  block address for which the back up file is to remove in the FBM
                  PageNo           [in]  page number which is to remove from 
                                         the list of pages still to write to the back up device
                  AddressingMode   [in]  addressing mode of the page for which the back up flags have to be removed 
    return value: true if the page number was still in the list of pages still to write 
                  to the back up device and if the corresponding block was still marked for back up 
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool  ResetDataPageAfterBackUp (
        const tsp00_TaskId          &TaskId,
        const IOMan_BlockAddress    &BlockAddress,
        const Data_PageNo           &PageNo,
        const Data_AddressingMode   &AddressingMode);

    /*!-------------------------------------------------------------------------
    function:     SetBlockAddressOfRestoredPage
    description:  Sets a block address in a converter entry 

                 - after a page was restored this command updates the converter entry, i.e the  
                   block address of the restored page is updated in the converter entry

    arguments:    TaskId           [in]  current task id 
                  BlockAddress     [in]  new block address which hast to be inserted in the converter entry
                                         of the page defined by PageNo und PageRecoveryMode
                  PageNo           [in]  page number for which the converter entry has to be updated 
                  AddressingMode   [in]  addressing mode of the page for which the back up flags have to be removed 
    return value: none
    --------------------------------------------------------------------------*/
    virtual void SetBlockAddressOfRestoredPage (
        const tsp00_TaskId          &TaskId,
        const IOMan_BlockAddress    &BlockAddress,
        const Data_PageNo           &PageNo,
        const Data_AddressingMode   &AddressingMode)
    {
        SAPDBERR_ASSERT_STATE (m_RestoreDataActive || m_RestorePagesActive);

        if( Data_Dynamic == AddressingMode )
        {
            m_PermDynamicMapControl.SetBlockAddressOfRestoredPage(
                TaskId, PageNo, BlockAddress, m_RestoreDataActive );
        }
        else if( Data_Static == AddressingMode )
        {
            m_PermStaticMapControl.SetBlockAddressOfRestoredPage(
                TaskId, PageNo, BlockAddress, m_RestoreDataActive );
        }
    }

    /*!-------------------------------------------------------------------------
    function:     GetUsedStaticPageNoContainer
    description:  Returns the container of used static page numbers

                 - returns the container containing the used static 
                   page numbers included in the current back up (save pages) 

    arguments:    none
    return value: (Converter_UsedPageNoContainer) reference to the used static page no container
    --------------------------------------------------------------------------*/
    virtual Converter_UsedPageNoContainer & GetUsedStaticPageNoContainer()
    {
        return m_PermStaticMapControl.GetUsedPageNoContainer();
    }

    /*!-------------------------------------------------------------------------
    function:     GetUsedDynamicPageNoContainer
    description:  Returns the container of used dynamic page numbers

                 - returns the container containing the used dynamic
                   page numbers included in the current back up (save pages) 

    arguments:    none
    return value: (Converter_UsedPageNoContainer) reference to the used dynamic page no container
    --------------------------------------------------------------------------*/
    virtual Converter_UsedPageNoContainer & GetUsedDynamicPageNoContainer()
    {
        return m_PermDynamicMapControl.GetUsedPageNoContainer();
    }

    /*!-------------------------------------------------------------------------
    function:     ClearUsedPageNoContainers
    description:  Clears the containers with used static and dynamic page numbers 

                 - removes all page numbers stored in the containers of
                   used dynamic and static page numbers which belong to a back up (save pages) 

    arguments:    TaskId [in]  task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void ClearUsedPageNoContainers(const tsp00_TaskId TaskId)
    {
        LockUsedPageNoContainers(); //  PTS 1125487 2003-11-17 

        m_PermStaticMapControl.ClearUsedPageNoList (TaskId);
        m_PermDynamicMapControl.ClearUsedPageNoList (TaskId);

        UnlockLockUsedPageNoContainers();
    }

    /*!-------------------------------------------------------------------------
    function:     BeginRestoreUsedPageNos()
    description:  Starts recovery of used pagenos 

      - This method must be called prior to the first call to
        RestoreUsedPageNos().
      - Allocates memory!
      
    arguments:    none
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool BeginRestoreUsedPageNos()
    {
        return (m_PermStaticMapControl.BeginRestoreUsedPageNos() &&
                m_PermDynamicMapControl.BeginRestoreUsedPageNos());
    }

    /*!-------------------------------------------------------------------------
    function:     RestoreUsedPageNos
    description:  Releases page numbers which do not belong to a restored data base 

                 - for all page number which are not marked as used at the UsedPageNoBitMapPage 
                   the corresponding converter entry is set to free and the block is released
                 - for all page numbers which are marked as used in the UsedPageNoBitMapPage but not
                   in the converter the correspondung page entry is reserved 
                 - all converter pages which contain any used page number are marked as used  
                   (see also FreePageNosNotInConverterBitMapPages)

    arguments:   TaskId                [in] task id   
                 UsedPageNoBitMapPage  [in] page which stores the page numbers of used pages in a bit map 
    return value: none
    --------------------------------------------------------------------------*/
    virtual void RestoreUsedPageNos (const tsp00_TaskId          TaskId,
                                     const Converter_BitMapPage &UsedPageNoBitMapPage)
    {
        SAPDB_Int DiffNumPermPagesUsed = 0;

        if( Data_Dynamic == UsedPageNoBitMapPage.ReadPageAddressingMode() )
        {
            m_PermDynamicMapControl.RestoreUsedPageNos( TaskId, UsedPageNoBitMapPage, DiffNumPermPagesUsed );
            m_NumPermPagesUsed.Increment( DiffNumPermPagesUsed );
        }
        else if( Data_Static == UsedPageNoBitMapPage.ReadPageAddressingMode())
        {
            m_PermStaticMapControl.RestoreUsedPageNos( TaskId, UsedPageNoBitMapPage, DiffNumPermPagesUsed );
            m_NumPermPagesUsed.Increment( DiffNumPermPagesUsed );
        }
    }

    /*!-------------------------------------------------------------------------
    function:     FinishRestoreUsedPageNos
    description:  Releases page numbers which do not belong to a restored data base 
                 
                 - after all UsedPageNoBitMapPage of a back up are read this function 
                   scans the converter and removes all those converter leaf pages (and their entries)
                   which are not marked as used. (see also FreeUnusedPageNo)
                   This means all those converter entries which were neither marked as used nor as 
                   free in the  UsedPageNoBitMapPage's  (means not at all mentioned) are removed                  

    arguments:   TaskId                [in] task id   
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FinishRestoreUsedPageNos (const tsp00_TaskId TaskId)
    {
        SAPDB_Int DiffNumPermPagesUsed = 0;

        m_PermStaticMapControl.FinishRestoreUsedPageNos (TaskId, DiffNumPermPagesUsed);
        m_NumPermPagesUsed.Increment(DiffNumPermPagesUsed);

        m_PermDynamicMapControl.FinishRestoreUsedPageNos (TaskId, DiffNumPermPagesUsed);
        m_NumPermPagesUsed.Increment(DiffNumPermPagesUsed);

        CheckFillingOfDataBase();
    }


    /*!-------------------------------------------------------------------------
    function:     LockUsedPageNoContainers
    description:  locks the UsedPageNoContainers

      - the lock can be  used to allow only one task access to the used
      page no container
      
        arguments:    none
        return value: none
    --------------------------------------------------------------------------*/
    virtual void LockUsedPageNoContainers()
    {
        m_UsedPageNoContainerLock.Lock();
    }


    /*!-------------------------------------------------------------------------
    function:     UnlockLockUsedPageNoContainers
    description:  locks the UsedPageNoContainers

      - the lock can be  used to allow only one task access to the used
      page no container
      
        arguments:    none
        return value: none
    --------------------------------------------------------------------------*/
    virtual void UnlockLockUsedPageNoContainers()
    {
        m_UsedPageNoContainerLock.Unlock();
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Back Up                                                   */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: Savepoint                                                    */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     BeginSavepoint
    description:  Sets all savepoint activ flags in map control 

          - After all save point activ flags are set no more blocks can be set to the state 
            free_after_save_point. Instead pages wich become free are stored in a FreePageNoList.
            After the save point is finished successfully all page numbers which are in the FreePageNoList
            are released and the corresponding blocks are set to free or free_after_savepoint 
            (see method) FreePageNosAfterSavepoint.

    arguments:    TaskId                   [in]   current task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void  BeginSavepoint (const tsp00_TaskId &TaskId)
    {
        m_MaxUsedStaticPageNo  = m_PermStaticMapControl.SetSavepointActiveFlag(TaskId);
        m_MaxUsedDynamicPageNo = m_PermDynamicMapControl.SetSavepointActiveFlag(TaskId);
    }

    /*!-------------------------------------------------------------------------
    function:     EndSavepoint
    description:  Increases the converter version and resets all savepoint activ flags 


    arguments:    taskId [in]   current task id
    return value: none
    --------------------------------------------------------------------------*/
    /* increase the converter version and reset savepoint activ flags */
    virtual void EndSavepoint( const tsp00_TaskId   taskId );

    /*!-------------------------------------------------------------------------
    function:     FreePageNosAfterSavepoint
    description:  Release all PageNos stored in the FreePageNoList  

          - After the save point is finished successfully all page numbers which are in the FreePageNoList
            are released and the corresponding blocks are set to free or free_after_savepoint 

    arguments:    TaskId                   [in]   current task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FreePageNosAfterSavepoint (const tsp00_TaskId &TaskId);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Savepoint                                                 */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: PageNo Mapping on block addresses                            */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetDevicePositionForReading()
    description:  Returns actual i/o address of data page with given page number 

                  - The actual i/o address of a given page number is read from
                    the converter and returned.
                  - If the corresponding converter entry contains an invalid
                    i/o address or if 'id' denotes a free pageid, the program
                    aborts.

    arguments: taskId             [in] current task id
               pageNo             [in] page number for which the corresponding block address is requested  
               pageRecoveryMode   [in] page recovery mode of the page for which a deviced position is wanted 
    return value: (IOMan_BlockAddress) requested address
    --------------------------------------------------------------------------*/
    virtual IOMan_BlockAddress GetDevicePositionForReading (
        const tsp00_TaskId          taskId,
        const Data_PageNo           pageNo,
        const Data_PageRecoveryMode &pageRecoveryMode);

    /*!-------------------------------------------------------------------------
    function:     GetDevicePositionForWriting()
    description:  Returns an i/o address where the data page can be written to 

                  - depending on the Recoverability of the page an its converter version
                    either the either the address of the currently occupied block is returned
                    or e new block address is supplied and the currently used block is set to 
                    'free after save point'
              
    arguments: taskId           [in]      identification of calling task
               trError          [out]     current error code
               PageNo           [in]      page number for which the corresponding block address is requested  
               PageRecoveryMode [in]      recovery mode of the page to write
               PageConverterVersion [in]  converter version of the page to write
    return value: (IOMan_BlockAddress) requested address
    --------------------------------------------------------------------------*/
    virtual IOMan_BlockAddress GetDevicePositionForWriting (
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        const Data_PageNo           &PageNo,
        const Data_PageRecoveryMode &PageRecoveryMode,
        const Converter_Version     &PageConverterVersion);

    /*!-------------------------------------------------------------------------
    function:     SetDevicePositionForWriting()
    description:  Sets a new  i/o address into the converter entry of a given page 

                  - set a new  i/o address into the converter entry of a given page 
                  - The block occuied so far is either set to 'free' for temp pages or
                    set to 'free after save point' for permanent pages
              
    arguments: taskId               [in]  identification of calling task
               trError              [out]     current error code
               PageNo               [in]  page number for which the corresponding block address is requested  
               PageRecoveryMode     [in]  recovery mode of the page to write
               PageConverterVersion [in]  converter version of the page to write
               BlockAddress         [in]  new block address which is 
                                          to insert in the corresponding converter entry
    return value: none
    --------------------------------------------------------------------------*/
    virtual void SetDevicePositionForWriting (
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        const Converter_Version     &PageConverterVersion,
        const Data_PageNo           &PageNo,
        const Data_PageRecoveryMode &PageRecoveryMode,
        const IOMan_BlockAddress    &BlockAddress);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: PageNo Mapping on block addresses                         */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: PageNo management                                            */
    /* -----------------------------------------------------------------------*/
    /*!-------------------------------------------------------------------------
    function:     RequestNewTempPageNo()
    description:  Returns an unused temporary PageNo 

                  - After this operation the supplied page number is "used".
                  - the temp counter in TransContext is increased.
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag
    arguments:    taskId     [in] identification of calling task
                  newPageNo [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool RequestNewTempPageNo(
        const tsp00_TaskId  taskId,
        Data_PageNo         &newPageNo )
    {
        if( RequestNewPageNo( taskId, m_TempMapControl, newPageNo ))
        {
            m_NumTempPagesUsed.Increment();
            SAPDBTRACE_WRITELN( Converter_Trace, 7, "TempPagesUsed: " << m_NumTempPagesUsed );
            return( SAPDB_TRUE );
        }
        return( SAPDB_FALSE );
    }

    /*!-------------------------------------------------------------------------
    function:     RequestNewPermDynamicPageNo()
    description:  Returns an unused perm,dynamic PageNo 

                  - After this operation the supplied page number is "used".
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag
    arguments:    taskId     [in] identification of calling task
                  newPageNo [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool RequestNewPermDynamicPageNo (
        const tsp00_TaskId        taskId,
        Data_PageNo               &newPageNo)
    {
        if( RequestNewPageNo( taskId, m_PermDynamicMapControl, newPageNo ))
        {
            m_NumPermPagesUsed.Increment();
            SAPDBTRACE_WRITELN( Converter_Trace, 7, "PermPagesUsed: " << m_NumPermPagesUsed );
            return( SAPDB_TRUE );
        }
        return( SAPDB_FALSE );
    }

    /*!-------------------------------------------------------------------------
    function:     RequestNewStaticPageNo()
    description:  Returns an unused static PageNo 

                  - After this operation the supplied page number is "used".
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag
    arguments:    taskId     [in] identification of calling task
                  newPageNo [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool RequestNewStaticPageNo (
        const tsp00_TaskId        taskId,
        Data_PageNo               &newPageNo)
    {
        if( RequestNewPageNo( taskId, m_PermStaticMapControl, newPageNo ))
        {
            m_NumPermPagesUsed.Increment();
            SAPDBTRACE_WRITELN( Converter_Trace, 7, "PermPagesUsed: " << m_NumPermPagesUsed );
            return( SAPDB_TRUE );
        }
        return( SAPDB_FALSE );
    }

    /*!-------------------------------------------------------------------------
    function:     RequestSpecificiedNewStaticPageNo()
    description:  Checks whether specified pageid is still free and returns it

                 - After this operation the supplied page number is "used".
                 - In contrast to method 'RequestNewStaticPageNo()', not any page number
                   at choice, but a specific page number is requested.
                 - If the specified page number is already used, the error
                   e_multiple_converter_entry  is set in the TransContext.
                 - If one of the database fill percent limits 80, 90, 95, 96,
                   97, 98 and 99 is reached, a message is written to the knldiag

    arguments:    Trans            [in/out] transaction context,in case the page number is already in use
                                           the error state Trans.trError_gg00 is set to  e_page_no_in_use 
                 PageNo           [in]     requested page number 
    return value: none
    --------------------------------------------------------------------------*/
    virtual void RequestSpecificiedNewStaticPageNo  (
        tgg00_TransContext &Trans,
        const Data_PageNo   PageNo);

    /*!-------------------------------------------------------------------------
    function:     FreeTempPageNo()
    description:  Puts back the given  temporary page number to the "pool" of unused numbers.

    arguments:    taskId   [in] taskId
                  pageNo   [in] given page to be set to free

    return value: SAPDB_Bool true if pageNo could set to free
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool FreeTempPageNo(
        const tsp00_TaskId  taskId,
        const Data_PageNo   pageNo );

    /*!-------------------------------------------------------------------------
    function:     FreePermDynamicPageNo()
    description:  Puts back the given  permanent, dynamic page number to the "pool" 
                  of unused numbers.

    arguments:    taskId   [in] taskId
                  pageNo   [in] given page to be set to free

    return value: SAPDB_Bool true if pageNo could set to free
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool FreePermDynamicPageNo(
        const tsp00_TaskId  taskId,
        const Data_PageNo   pageNo )
    {
        return( FreePermPageNo( taskId, m_PermDynamicMapControl, pageNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     FreePermStaticPageNo()
    description:  Puts back the given  static page number to the "pool" 
                  of unused numbers.

    arguments:    taskId   [in] taskId
                  pageNo   [in] given page to be set to free

    return value: SAPDB_Bool true if pageNo could set to free
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool FreePermStaticPageNo(
        const tsp00_TaskId  taskId,
        const Data_PageNo   pageNo )
    {
        return( FreePermPageNo( taskId, m_PermStaticMapControl, pageNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     HandleDBFull()
    description:  Checks if the required count of pages is available.

                  - If the required count of pages is available this function returns true
                  - If the required count of pages is not available, the task is suspended
                    and waits until the wanted number of pages is available. After the task is resumed
                    and the pages are available the fundtion returns true. 
                  - Sometimes tasks cannot be suspended since this could cause dead locks. In this
                    case false is returned and sometimes even the error e_no_more_temp_space is set
                    to trigger a rollback.

    note:         Note this function does not reserve any number of pages for this tasks, i.e.
                  the information that a particular number is available  is dirty and may be 
                  even false, when this task really tries to get wanted number of free pages. 
    arguments:    Trans                [in/out] transaction context
                  NumDataPagesRequested    [in] number of pages wanted 
                  bAddDeviceIsPossible     [in] displays whether a new device can be added to the 
                                                data base to gain free pages

    return value: none
    --------------------------------------------------------------------------*/
    virtual void HandleDBFull (
        tgg00_TransContext   &Trans,
        SAPDB_UInt           NumDataPagesRequested,
        const SAPDB_Bool     bAddDeviceIsPossible);

    /*!-------------------------------------------------------------------------
    function:     MaxTempPageNo()
    description:  Returns the maximum possible temporary page number

    arguments:    none
    return value: (Data_PageNo) maximum possible temporary page number
    --------------------------------------------------------------------------*/
    virtual Data_PageNo MaxTempPageNo() const
    {
        return m_TempMapControl.MaxPageNo();
    }

    /*!-------------------------------------------------------------------------
    function:     MaxPermDynamicPageNo()
    description:  Returns the maximum possible perm, dynamic page number

    arguments:    none
    return value: (Data_PageNo) maximum possible perm, dynamic page number
    --------------------------------------------------------------------------*/
    virtual Data_PageNo MaxPermDynamicPageNo() const
    {
        return m_PermDynamicMapControl.MaxPageNo();
    }

    /*!-------------------------------------------------------------------------
    function:     MaxPermStaticPageNo()
    description:  Returns the maximum possible static page number

      arguments:    none
      return value: (Data_PageNo) maximum possible static page number
    --------------------------------------------------------------------------*/
    virtual Data_PageNo MaxPermStaticPageNo() const
    {
        return m_PermStaticMapControl.MaxPageNo();
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: PageNo management                                         */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: ColdVerify                                                   */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     BeginColdVerify
    description:  Prepares converter verification in "cold" database state

    arguments:    taskId[in]   task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginColdVerify(  const tsp00_TaskId    taskId );

    /*!-------------------------------------------------------------------------
    function:     EndColdVerify
    description:  Finishes converter verification in "cold" database state

        - All pages which are not marked as used (see method MarkPageNoAsUsed)
          are released          

        - The routine returns several parameters describing the actual
          state of the converter and of the database.

    arguments:    TaskId                  [in]   task id
                  NumStaticPermDataPages  [out]  number of permanent data pages used
                                                 with static page numbers
                  NumDynamicPermDataPages [out]  number of permanent data pages used
                                                 with dynamic page numbers
                  NumPermConPages         [out]  number of converter pages used for
                                                 permanent data pages
                  NumReleaseBlocks        [out]  number of released blocks, which were
                                                 not accessed during verify
    return value: none
    --------------------------------------------------------------------------*/

    virtual void EndColdVerify (
        const tsp00_TaskId  taskId,
        Data_PageCount      &NumStaticPermDataPages,
        Data_PageCount      &NumDynamicPermDataPages,
        Data_PageCount      &NumPermConvPages,
        Data_PageCount      &NumReleaseBlocks );

    /*!-------------------------------------------------------------------------
    function:     EndFailedColdVerify
    description:  Finishes a failed converter verification

        - The converter is restored to the state it had before the verification 
          was started by the BeginColdVerify

    arguments:    TaskId         [in]  task id
    return value: none
    --------------------------------------------------------------------------*/

    virtual void EndFailedColdVerify (const tsp00_TaskId  TaskId);

    /*!-------------------------------------------------------------------------
    function:     MarkPageNoAsUsed
    description:  Marks a page in the converter as used 
    arguments:    TaskId         [in]   task id
                  PageNo         [int]    page number to be marked as used
                  AddressingMode [in]     addressing mode of the page to mark 

    return value: none
    --------------------------------------------------------------------------*/

    virtual void MarkPageNoAsUsed (
        const tsp00_TaskId          taskId,
        const Data_PageNo           PageNo,
        const Data_AddressingMode   AddressingMode)
    {
        if( Data_Dynamic == AddressingMode )
            m_PermDynamicMapControl.MarkPageNoAsUsed( taskId, PageNo );
        else if( Data_Static == AddressingMode )
            m_PermStaticMapControl.MarkPageNoAsUsed( taskId, PageNo );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: ColdVerify                                                */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Event                                                        */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     AddEvent()
    description:  defines an event, i.e. whenever the occupation of the database passes
                  the given RelativeThreshold an event is triggered.

    arguments:    RelativeThreshold [in] relative filling level of the data base, whenever this critical 
                                         value is passed an event is triggered
                                         critical value in 0.01 percent with respect to the total number of pages of the data base
                  bOverflow         [in] if this is true the event is triggerd if the filling exceeds the
                                         critical value else an event is triggered if the filling falls below
                                         the critical value.
                  Priority          [in] priority of the event                       
                                         
    return value: (SAPDB_Bool) true if the value could be added successfully
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddEvent(const SAPDB_Int                RelativeThreshold,
                                const SAPDB_Bool               bOverflow,
                                const tsp31_event_prio_Param   Priority)
    {
        return m_DBUsageMonitor.AddEvent(RelativeThreshold, bOverflow, Priority);
    }

    /*!-------------------------------------------------------------------------
    function:     DelEvent()
    description:  removes an event from the set of events

    arguments:    RelativeThreshold        [in] defines the critical value to be removed. The value is 
                                                given in units of 0.01 percent with respect to the base unit
                  bOverflow                [in] displays whether the event of exceeding or falling below is to remove

    return value: (SAPDB_Bool) true if the event set is not empty
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool DelEvent  (const SAPDB_Int    RelativeThreshold,
                                  const SAPDB_Bool   bOverflow)
    {
        return m_DBUsageMonitor.DelEvent(RelativeThreshold, bOverflow);
    }

    /*!-------------------------------------------------------------------------
    function:     GetEventSet()
    description:  puts all events into an external structure in order to show them in a system view

      - The ShortEventDesc is filled with all events in order to show the events in a system view 

    arguments:    ShortEventDesc     [out] structure to fill
                  bGetOverflowEvents [in]  if true  ShortEventDesc is filled with overflow events
                                           if false ShortEventDesc is filled with underflow events
    return value: none
    --------------------------------------------------------------------------*/

    virtual void GetEventSet  (tsp31_short_event_desc &ShortEventDesc,
                               SAPDB_Bool              bGetOverflowEvents)
    {
        m_DBUsageMonitor.GetEventSet(ShortEventDesc, bGetOverflowEvents);
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Event                                                     */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Dump                                                         */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Dump()
    description:  This method is used to dump the main memory structures of the
                  component converter.
    arguments:    TaskId       [in]     identification of the calling task
                  Dump         [in/out] kernel dump file
                  bEnterRegion [in]     use the converter regions
    return value: none
    --------------------------------------------------------------------------*/

    virtual void Dump(
        const tsp00_TaskId  TaskId,
        Kernel_Dump         &Dump,
        const SAPDB_Bool    bEnterRegion );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Dump                                                      */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Data base migration                                          */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     SetBlockAddressForMigratedPage
    description:  Set converter entry and block state in FBM for the given
                  page to occupied. The page is identified by the pageNo and
                  the current position on the data devspace is defined by block.
                  Precondtion is that the addressing mode of the given page no 
                  is dynamic. Additional it is not allowed to give a page id
                  with a block, that specifies one of the reserved block within
                  the data devspaces. (IOMan_InfoPage or RestartPage).
    arguments:    taskId         [in]  current task id 
                  block          [in]  new block address which has to be inserted 
                                       in the converter entry and the FBM of the 
                                       page defined by pageNo.
                  pageNo         [in]  page number for which the converter entry 
                                       has to be updated 
                  bSetToOccupied [in] 
    return value: none
    --------------------------------------------------------------------------*/

    virtual void SetBlockAddressForMigratedPage (
        const tsp00_TaskId          taskId,
        const IOMan_BlockAddress    &block,
        const Data_PageNo           &pageNo,
        const SAPDB_Bool            bSetToOccupied = SAPDB_TRUE );

    /*!-------------------------------------------------------------------------
    function:     DetermineAndSetBlockAddressForMigratedPage
    description:  This method uses SetBlockAddressForMigratedPage, therefore
                  the same restrictions are valid. The new block address used
                  by the given page will be determined by this method and 
                  returned to the caller. If no fre block is available or any
                  error occur the system will execute an emergency shutdown.
    arguments:    taskId [in]  current task id 
                  pageNo [in]  page number of the converter page entry to
                               be updated.
    return value: IOMan_BlockAddress
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockAddress DetermineAndSetBlockAddressForMigratedPage (
        const tsp00_TaskId  taskId,
        const Data_PageNo   &pageNo );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Data base migration                                       */
    /* -----------------------------------------------------------------------*/

private:
    /* -----------------------------------------------------------------------*/
    /*! chapter: private methods                                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Converter_Converter()
    description:  default constructor
    arguments:    Allocator  [in] Allocator for memory allocation
    return value: none
    --------------------------------------------------------------------------*/

    Converter_Converter( SAPDBMem_IRawAllocator  &Allocator );

    /*!-------------------------------------------------------------------------
    function:     InitializeBase()
    description:  initialize primitive members of the converter

    arguments:    NumAllPages [in] maximum number of pages which can be stored in the data base. this
                                   corresponds to the sum of size of all devices 
    return value: none
    --------------------------------------------------------------------------*/

    void InitializeBase( const SAPDB_UInt NumAllPages );

    /*!-------------------------------------------------------------------------
    function:     ExpandMap
    description:  Expand a particular map of the converter

    arguments:   taskId        [in] task id
                 recoveryMode  [in] recovery mode of the map to expand
                 numPagesAdded [in] value by which the map is to expand
    return value: Returns true if the expansion succeded
    --------------------------------------------------------------------------*/

    SAPDB_Bool ExpandMap (
        const tsp00_TaskId           taskId,
        const Data_PageRecoveryMode  &recoveryMode,
        const SAPDB_UInt             numPagesAdded);


    /*!-------------------------------------------------------------------------
    function:     TransactionIsSuspendable
    description:  Investigates if a transaction can be suspended if the data base is full 

    arguments:    Trans           [in] transaction context
    return value: Returns true if the task can be suspended without causing deadlocks 
    --------------------------------------------------------------------------*/
    SAPDB_Bool  TransactionIsSuspendable (tgg00_TransContext &Trans) const
    {
        return
            (
                (0 == Trans.trBdExclFileLocks_gg00                  ) &&
                (! Trans.trState_gg00.includes(tsDbFullNowait_egg00))
            );
    }

    /*!-------------------------------------------------------------------------
    function:     CheckFillingOfDataBase()
    description:  Triggers events or entries into the knldiag if the filling of the data base passes critical values

    arguments:    none
    return value: none 
    --------------------------------------------------------------------------*/
    void CheckFillingOfDataBase()
    {
        m_DBUsageMonitor.Watch (GetNumberOfUsedPages());
    }

    /*!-------------------------------------------------------------------------
    function:     IncrementPermConvPageCount()
    description:  Increment number of converter pages used for permanent 
                  data page numbers.
    arguments:    none
    return value: none 
    --------------------------------------------------------------------------*/

    void IncrementPermConvPageCount()
    {
        m_NumPermConvPages.Increment();
    }

    /*!-------------------------------------------------------------------------
    function:     DecrementPermConvPageCount()
    description:  Decrement number of converter pages used for permanent 
                  data page numbers.
    arguments:    none
    return value: none 
    --------------------------------------------------------------------------*/

    void DecrementPermConvPageCount()
    {
        m_NumPermConvPages.Decrement();
    }

    /*!-------------------------------------------------------------------------
    function:     IncrementTempConvPageCount()
    description:  Increment number of converter pages used for temporary 
                  data page numbers.
    arguments:    none
    return value: none 
    --------------------------------------------------------------------------*/

    void IncrementTempConvPageCount()
    {
        m_NumTempConvPages.Increment();
    }

    /*!-------------------------------------------------------------------------
    function:     DecrementTempConvPageCount()
    description:  Decrement number of converter pages used for temporary 
                  data page numbers.
    arguments:    none
    return value: none 
    --------------------------------------------------------------------------*/

    void DecrementTempConvPageCount()
    {
        m_NumTempConvPages.Decrement();
    }

    /*!-------------------------------------------------------------------------
    function:     GetNumberOfChangedConvPages()
    description:  Returns the number of converter pages marked as changed
                  in dynamic and static map.
    arguments:    none
    return value: (Data_PageCount)
    --------------------------------------------------------------------------*/

    Data_PageCount GetNumberOfChangedConvPages() const
    {
        return( m_PermDynamicMapControl.GetNumberOfChangedConvPages() +
                m_PermStaticMapControl.GetNumberOfChangedConvPages() );
    }

    /*!-------------------------------------------------------------------------
    function:     SavepointIsPossible()
    description:  Returns true if it's possible to execute a savepoint.
                  If the converter is not active or in maintenance or a savepoint
                  is still requested no new savepoint is possible.
    arguments:    none
    return value: true if a savepoint is executable
    --------------------------------------------------------------------------*/

    SAPDB_Bool SavepointIsPossible() const
    {
        if( ! m_Active ) // converter is offline
            return( false );

        if( m_InMaintenanceMode ) // converter is in special mode
            return( false );

        if( m_SavepointIsRequested ) // SVP is still requested
            return( false );

        return( true );
    }

    /*!-------------------------------------------------------------------------
    function:     StartSavepoint()
    description:  Start savepoint without checking any conditions
    arguments:    trans           [in] transaction context
                  numChangedPages [in] number of changed data and converter pages
    return value: none
    --------------------------------------------------------------------------*/

    void StartSavepoint(
        tgg00_TransContext &trans,
        const SAPDB_Int4    numChangedPages );

    /*!-------------------------------------------------------------------------
    function:     NumChangedPages()
    description:  Returns the number of changed data and converter pages.
    arguments:    none
    return value: SAPDB_Int4
    --------------------------------------------------------------------------*/

    SAPDB_Int4 NumChangedPages() const;

    /*--------------------------------------------------------------------------
    function:     FreePermPageNo()
    description:  Puts back the given permanent page number to the "pool" of unused numbers.
    arguments:    taskId               [in] task identification
    arguments:    mapControl       [in|out] corresponding map control
    arguments:    pageNo               [in] number of the page to release
    arguments:   (SAPDB_Bool) true if successfully executed
    --------------------------------------------------------------------------*/

    SAPDB_Bool FreePermPageNo(
        const tsp00_TaskId          taskId,
        Converter_PermMapControl    &mapControl,
        const Data_PageNo           pageNo );

    /*!-------------------------------------------------------------------------
    function:     RequestNewPageNo()
    description:  Request for unsused pageId's for the given map control.
    arguments:    taskId         [in] task id
    arguments:    mapControl [in|out] corresponding map control
    arguments:    newPageNo     [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/

    SAPDB_Bool RequestNewPageNo(
        const tsp00_TaskId      taskId,
        Converter_MapControl    &mapControl,
        Data_PageNo             &newPageNo )
    {
        if( ! m_Active )
            return( SAPDB_FALSE );

        if( m_NumAllPages > GetNumberOfUsedPages() + 1 )
        {
            if( mapControl.RequestNewPageNo( taskId, newPageNo ))
            {
                CheckFillingOfDataBase();
                SAPDBTRACE_WRITELN( Converter_Trace, 5, "New PageNo: " << newPageNo );
                return( SAPDB_TRUE );
            }
        }
        SAPDBTRACE_WRITELN( Converter_Trace, 5, "No PageNo available" );
        return( SAPDB_FALSE );
    }

    /*
       @brief   Checks if the given number of pages is addressable with the converter.
                This check includes a security limit of four percent, but not more 
                than 4.000 pages.
       @param   numPagesRequested [in] number of requested pages
       @return  true is given number of pages is addressable
     */

    // PTS 1124331 TS 2003-09-23
    SAPDB_Bool IsSpaceAvailable(  const SAPDB_Int4 numPagesRequested ) const
    {
        return(( numPagesRequested + GetNumberOfUsedPages() ) < m_SecuritySpaceLimit );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: private methods                                           */
    /* -----------------------------------------------------------------------*/


private:

    static Converter_Converter               *m_Instance;

    RTEMem_AllocatorWrapper                   m_Allocator; // Don't remove this from top => init list!

    const SAPDB_UInt                          m_MapBlockingSize;

    Converter_TempMapControl                  m_TempMapControl;
    Converter_PermMapControl                  m_PermDynamicMapControl;
    Converter_PermStaticMapControl            m_PermStaticMapControl;

    Converter_Index                           m_Index;

    Converter_Version                         m_ConverterVersion;
    IOMan_BlockAddress                        m_ConverterRootBlockAddress;

    RTESync_InterlockedCounter<SAPDB_UInt>    m_NumPermPagesUsed;
    RTESync_InterlockedCounter<SAPDB_UInt>    m_NumTempPagesUsed;
    RTESync_InterlockedCounter<SAPDB_UInt>    m_NumPermConvPages;
    RTESync_InterlockedCounter<SAPDB_UInt>    m_NumTempConvPages;

    RTESync_Spinlock                          m_UsedPageNoContainerLock;

    SAPDB_UInt                                m_MaxPermUsed;
    SAPDB_UInt                                m_NumAllPages;
    SAPDB_UInt                                m_SecuritySpaceLimit;

    SAPDB_UInt                                m_NumPagesForBackUp;
    SAPDB_UInt                                m_NumStatConvPagesForBackUp;
    SAPDB_UInt                                m_NumDynConvPagesForBackUp;

    Data_PageNo                               m_MaxUsedDynamicPageNo;
    Data_PageNo                               m_MaxUsedStaticPageNo;

    SAPDB_Bool                                m_Active;
    SAPDB_Bool                                m_InMaintenanceMode;
    SAPDB_Bool                                m_SavepointIsRequested;
    SAPDB_Bool                                m_SavePagesIsActiv;
    SAPDB_Bool                                m_SaveDataIsActiv;
    SAPDB_Bool                                m_RestoreDataActive;
    SAPDB_Bool                                m_RestorePagesActive;
    SAPDB_Bool                                m_LastSaveDataSuccessful;
    SAPDB_Bool                                m_ColdVerifyIsActive;

    Converter_WaitQueue                       m_DBFullWaitQueue;
    Converter_WaitQueue                       m_FreePageNoWaitQueue;
    Converter_WaitQueue                       m_ExpandConverterWaitQueue;
    Kernel_EventSet                           m_DBUsageMonitor;

    Container_Vector<Converter_Index::BlockVector>* m_VolumeList;
};

/*!*****************************************************************************

   endclass: Converter_Converter

*******************************************************************************/

#endif  /* CONVERTER_CONVERTER_HPP */
