  /*!***************************************************************************

  module:       Converter_Entry.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       TillL

  special area: Converter

  description:  Base class for entry of converter index or map
                
  see also:     

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

  *****************************************************************************/


#ifndef CONVERTER_ENTRY_HPP
#define CONVERTER_ENTRY_HPP


#include "Converter/Converter_Types.hpp"
#include "Converter/Converter_PageFrameAllocator.hpp"


/*!*****************************************************************************

   class: Converter_Entry

   description: Entry of converter parent index

   arguments:   Page : converter page type (leaf page or index page)

*******************************************************************************/
template <class Page>
class Converter_Entry
{
public:

    /*!-------------------------------------------------------------------------
    declaration : PageEntryCount
    description:  converter page entry count
    --------------------------------------------------------------------------*/
    typedef Converter_PageEntryCount        PageEntryCount;

    /*!-------------------------------------------------------------------------
    function:     Converter_Entry()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Entry()
    {}

    /*!-------------------------------------------------------------------------
    function:     AssignPage()
    description:  Assigns page handler
    arguments:    page           [in]  page handler
                  freeentrycount [in]  number of free entries on the page
    return value: none
    --------------------------------------------------------------------------*/
    void AssignPage
        (Page&                          page,
         PageEntryCount                 freeentrycount)
    {
        SAPDBERR_ASSERT_ARGUMENT(freeentrycount == 0 || page.IsAssigned());
        m_Page               = page;
        m_FreePageEntryCount = freeentrycount;
    }

    /*!-------------------------------------------------------------------------
    function:     ReleasePageFrame()
    description:  Releases page frame from page handler
    arguments:    framealloc [in] page frame allocator for dallocation
    return value: none
    --------------------------------------------------------------------------*/
    void ReleasePageFrame
        (Converter_PageFrameAllocator  &framealloc)
    {
        framealloc.Free(m_Page.GetFrame(), m_Page.ReadPageRecoveryMode().PageIsRecoverable());
        m_FreePageEntryCount = 0;
    }

    /*!-------------------------------------------------------------------------
    function:     GetPage()
    description:  Returns reference to associated page
    arguments:    none
    return value: (const Page& / Page&) reference to page
    prototypes:   2
    --------------------------------------------------------------------------*/
    Page& GetPage()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_Page;
    }

    const Page& GetPage() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_Page;
    }

    /*!-------------------------------------------------------------------------
    function:     PageIsAssigned()
    description:  Checks whether a page has been assigned to this entry
    arguments:    none
    return value: (SAPDB_Bool) true if assigned
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIsAssigned() const
    {
        return m_Page.IsAssigned();
    }

    /*!-------------------------------------------------------------------------
    function:     PageIsEmpty()
    description:  Checks whether actual page is empty
    arguments:    none
    return value: (SAPDB_Bool) true if empty
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIsEmpty() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        SAPDBERR_ASSERT_STATE(m_Page.HasFreeEntryAvailable() ||
                              (m_FreePageEntryCount == 0));

        return m_FreePageEntryCount == m_Page.GetEntryCount();
    }

    /*!-------------------------------------------------------------------------
    function:     PageIsFull()
    description:  Checks whether actual page is full
    arguments:    none
    return value: (SAPDB_Bool) true if full
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIsFull() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        SAPDBERR_ASSERT_STATE(m_Page.HasFreeEntryAvailable() ||
                              (m_FreePageEntryCount == 0));

        return m_FreePageEntryCount == 0;
    }

    /*!-------------------------------------------------------------------------
    function:     SetFreePageEntryCount()
    description:  Sets actual free page entry count
    arguments:    count [in]   free page entry count
    return value: none
    --------------------------------------------------------------------------*/
    void SetFreePageEntryCount
        (PageEntryCount                 count)
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        m_FreePageEntryCount = count;
    }

    /*!-------------------------------------------------------------------------
    function:     GetFreePageEntryCount()
    description:  Returns actual free page entry count
    arguments:    none
    return value: (PageEntryCount) free page entry count
    --------------------------------------------------------------------------*/
    PageEntryCount GetFreePageEntryCount() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_FreePageEntryCount;
    }

    /*!-------------------------------------------------------------------------
    function:     IncFreePageEntryCount()
    description:  Increments free page entry count
    arguments:    inc [in]   increment
    return value: none
    --------------------------------------------------------------------------*/
    void IncFreePageEntryCount
        (PageEntryCount                 inc = 1)
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        m_FreePageEntryCount += inc;
    }

    /*!-------------------------------------------------------------------------
    function:     DecFreePageEntryCount()
    description:  Decrements free page entry count
    arguments:    dec [in]   decrement
    return value: none
    --------------------------------------------------------------------------*/
    void DecFreePageEntryCount
        (PageEntryCount                 dec = 1)
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(m_FreePageEntryCount >= dec);
        m_FreePageEntryCount -= dec;
    }

protected:

    /*--------------------------------------------------------------------------
    declaration : m_Page
    description:  pointer to converter page
    --------------------------------------------------------------------------*/
    Page                       m_Page;

private:
    /*--------------------------------------------------------------------------
    declaration : m_FreePageEntryCount
    description:  number of free page entries on associated page
    --------------------------------------------------------------------------*/
    PageEntryCount             m_FreePageEntryCount;

};
/*!*****************************************************************************

   endclass: Converter_Entry

*******************************************************************************/





#endif // CONVERTER_ENTRY_HPP
