/*!*****************************************************************************

  module:       Converter_FlagInfo.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       AlexanderK

  special area: 

  description:  
                
  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef CONVERTER_FLAGINFO_HPP
#define CONVERTER_FLAGINFO_HPP

#include "DataAccess/Data_Types.hpp"

/*!*****************************************************************************

   class: Converter_FlagInfo

   description: 

*******************************************************************************/
class Converter_FlagInfo 
{
public:	

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration: FlagInfoData
    description: Flag info data
    
                 You cannot change this type without performing additional tasks!
                 If you change it to 8 bytes, critically check all flag info routines.
                 There may arise alignment problems.
    --------------------------------------------------------------------------*/

    typedef SAPDB_UInt4                         FlagInfoData;

	/* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

private:
    /*!-------------------------------------------------------------------------
    function:     FlagInfoBitPos()
    description:  Returns bit position for flag of specified entry
    within the corresponding flag info block
    arguments:    EntryNo [in]  entry no
    return value: (SAPDB_UInt)
    --------------------------------------------------------------------------*/
    static SAPDB_UInt FlagInfoBitPos (const SAPDB_UInt&   EntryNo) 
	{
		// this is meant:
		// return EntryNo % (SAPDB_BITS_PER_BYTE * sizeof(FlagInfo));
		
		// this is faster:
		return EntryNo & 0x1F;
	}
	
public:

    /*!-------------------------------------------------------------------------
    declaration: FlagInfoBitMask
    description: Single bit bitmask for every bit in a FlagInfo block
    --------------------------------------------------------------------------*/
    static FlagInfoData FlagInfoBitMask[];

    /*!-------------------------------------------------------------------------
    function:     FlagInfoBlockNo()
    description:  Returns sequence number of the flag info block where a flag
	of the specified entry is stored
    arguments:    EntryNo [in]  entry no
    return value: (SAPDB_UInt)
    --------------------------------------------------------------------------*/
    static SAPDB_UInt FlagInfoBlockNo (const SAPDB_UInt&  EntryNo) 
	{
		// this is meant:
		// return EntryNo / (SAPDB_BITS_PER_BYTE * sizeof(FlagInfo));
		
		// this is faster:
		return EntryNo >> 5;
	};

	/*!-----------------------------------------------------------------------
    function:     Converter_FlagInfo()
    description:  

    arguments:    EntryNo [in]  entry no
    return value: (SAPDB_UInt)
    --------------------------------------------------------------------------*/
	Converter_FlagInfo (FlagInfoData * FlagInfoDataBlockPtr,
		                SAPDB_UInt     BlockOffset)
	{
		m_FlagInfoDataPtr = FlagInfoDataBlockPtr + FlagInfoBlockNo(BlockOffset);
	}

	/*!-------------------------------------------------------------------------
    function:     FlagIsSet()
    description:  

    arguments:    EntryNo [in]  entry no
    return value: (SABDB_Bool)
    --------------------------------------------------------------------------*/
    SAPDB_Bool FlagIsSet (SAPDB_UInt EntryNo) const
	{
	  return ((*m_FlagInfoDataPtr & FlagInfoBitMask[FlagInfoBitPos(EntryNo)]) != 0);
	}

	/*!-------------------------------------------------------------------------
    function:     SetFlag()
    description:  

    arguments:    EntryNo [in]  entry no
    return value: none
    --------------------------------------------------------------------------*/
	void SetFlag (SAPDB_UInt EntryNo)
	{
		*m_FlagInfoDataPtr |= FlagInfoBitMask[FlagInfoBitPos(EntryNo)];
	}

	/*!-------------------------------------------------------------------------
    function:     DelFlag()
    description:  

    arguments:    EntryNo [in]  entry no
    return value: none
    --------------------------------------------------------------------------*/
    void DelFlag (SAPDB_UInt EntryNo)
	{
		*m_FlagInfoDataPtr &= ~FlagInfoBitMask[FlagInfoBitPos(EntryNo)];
	}

	/*!-------------------------------------------------------------------------
    function:     FlagInfoDataPtr()
    description:  

    arguments:    none
    return value: (FlagInfoData*)
    --------------------------------------------------------------------------*/
	FlagInfoData* FlagInfoDataPtr()
	{
		return m_FlagInfoDataPtr;
	}

private:

	FlagInfoData * m_FlagInfoDataPtr;

};

/*!*****************************************************************************

   endclass: Converter_FlagInfo

*******************************************************************************/

#endif // CONVERTER_FLAGINFO_HPP
            