/*!**************************************************************************

  module      : Converter_IBackUp.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author:       AlexanderK

  special area: converter
  description : abstract back up interface class  of the converter


  last changed: 2001-01-10  11:26
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_IBACKUP_HPP
#define CONVERTER_IBACKUP_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h" 
#include "ggg00.h"
#include "DataAccess/Data_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "Converter/Converter_UsedPageNoContainer.hpp"
#include "Converter/Converter_BitMapPage.hpp"


/*!*****************************************************************************

   class: Converter_IBackUp

*******************************************************************************/
class Converter_IBackUp
{
public:
    /*!-------------------------------------------------------------------------
    function:     Converter_IBackUp
    description:  returns the backup interface of the converter

          - This static function returns the interface of the converter which
		    allows access to all methods defined in this interface class 

    arguments:    none
    return value: (Converter_IBackUp) instance of the converter 
    --------------------------------------------------------------------------*/
	static   Converter_IBackUp & Instance();

   /*!-------------------------------------------------------------------------
    function:     BeginSaveData
    description:  Starts a total backup (also called save data)

          - All currently occupied blocks of the database are marked within the FBM 
            to be included in the back up

    arguments:    TaskId  [in]   task id

    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginSaveData (tsp00_TaskId TaskId) = 0;
    
    /*!-------------------------------------------------------------------------
    function:     BeginSavePages
    description:  Starts a differential backup (also called save pages)
        
          - All currently occupied blocks of the database which are marked as changed since the last 
          total backup are marked within the FBM  to be included in the back up.
          - A list off all page numbers used in the current data base is created
          
    arguments:    TaskId  [in]   task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginSavePages (tsp00_TaskId TaskId) = 0;

	/*!-------------------------------------------------------------------------
    function:     EndSave
    description:  Ends the backup of data or pages

          - The back up active flag in the converter is reset
		  - If  the current back up  did not finish successfully all
			blocks still marked as to be written to the back up files are unmarked.

    arguments:    TaskId [in]                      current task id 
                  bBackUpSuccessfullyFinished [in] indicates whether the current back
	                                               up finished successfully
    return value: none
    --------------------------------------------------------------------------*/
	virtual void  EndSave  (const tsp00_TaskId &TaskId,  
				            SAPDB_Bool          bBackUpSuccessfullyFinished) = 0;     

	/*!-------------------------------------------------------------------------
    function:     GetNumPagesForBackUp
    description:  Returns the number of of pages marked for back up

    arguments:    none
    return value: Number of blocks marked for back up
    --------------------------------------------------------------------------*/
	virtual SAPDB_UInt GetNumPagesForBackUp() = 0 ;

	/*!-------------------------------------------------------------------------
    function:     GetNumConvPagesForBackUp
    description:  Returns the number of converter pages required to address all 
                  static data pages 

       - Returns the number of converter pages required to address all 
                 data pages which belong to the current backup

    arguments:    none
    return value: Number of converter pages required to address all 
                  data pages which belong to the current backup
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetNumConvPagesForBackUp() = 0;

    /*!-------------------------------------------------------------------------
    function:     CheckBackUp
    description:  Checks if the current back up finished successfully

                 - Checks if all pages marked as to be included in 
				   the back up were fetched by the back up manager

    arguments:    TaskId [in]  current task id 

    return value: true if the back was successful else false
    --------------------------------------------------------------------------*/
	virtual SAPDB_Bool CheckBackUp( const tsp00_TaskId taskId ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     ResetDataPageAfterBackUp
    description:  Signals the converter that a given page was written into the back up file 

                 - Resets the flag in the FBM which displays that the given page still has to be
				   written to the back up file 
                 - Removes the page number from the list of page numbers still to inlclude in the back up 

    arguments:    TaskId           [in]  current task id 
                  BlockAddress     [in]  block address for which the back up file is to remove in the FBM
                  PageNo           [in]  page number which is to remove from 
				                         the list of pages still to write to the back up device
                  AddressingMode   [in]  addressing mode of the page for which the back up flags have to be removed 
    return value: true if the page number was still in the list of pages still to write 
	              to the back up device and if the corresponding block was still marked for back up 
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool  ResetDataPageAfterBackUp (const tsp00_TaskId          &TaskId,
			               					      const IOMan_BlockAddress    &BlockAddress,
						        			      const Data_PageNo           &PageNo,
								    		      const Data_AddressingMode   &AddressingMode) = 0;

 /*!-------------------------------------------------------------------------
    function:     SetBlockAddressOfRestoredPage
    description:  Sets a block address in a converter entry 

                 - after a page was restored this command updates the converter entry, i.e the  
                   block address of the restored page is updated in the converter entry

    arguments:    TaskId           [in]  current task id 
                  BlockAddress     [in]  new block address which hast to be inserted in the converter entry
				                         of the page defined by PageNo und PageRecoveryMode
                  PageNo           [in]  page number for which the converter entry has to be updated 
                  AddressingMode   [in]  addressing mode of the page for which the back up flags have to be removed 
    return value: none
    --------------------------------------------------------------------------*/
	virtual void SetBlockAddressOfRestoredPage (const tsp00_TaskId          &TaskId,
											    const IOMan_BlockAddress    &BlockAddress,
											    const Data_PageNo           &PageNo,
											    const Data_AddressingMode   &AddressingMode) = 0;

    /*!-------------------------------------------------------------------------
    function:     GetUsedStaticPageNoContainer
    description:  Returns the container of used static page numbers

                 - returns the container containing the used static 
				   page numbers included in the current back up (save pages) 

    arguments:    none
    return value: reference to the used static page no container
    --------------------------------------------------------------------------*/
    virtual Converter_UsedPageNoContainer & GetUsedStaticPageNoContainer() = 0;

    /*!-------------------------------------------------------------------------
    function:     GetUsedDynamicPageNoContainer
    description:  Returns the container of used dynamic page numbers

                 - returns the container containing the used dynamic
				   page numbers included in the current back up (save pages) 

    arguments:    none
    return value: reference to the used dynamic page no container
    --------------------------------------------------------------------------*/
    virtual Converter_UsedPageNoContainer & GetUsedDynamicPageNoContainer() = 0;

    /*!-------------------------------------------------------------------------
    function:     ClearUsedPageNoContainers
    description:  Clears the containers with used static and dynamic page numbers 

                 - removes all page numbers stored in the containers of
                   used dynamic and static page numbers which belong to a back up (save pages) 

    arguments:    TaskId [in]  task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void ClearUsedPageNoContainers(const tsp00_TaskId TaskId) = 0;

    /*!-------------------------------------------------------------------------
	function:     BeginRestoreUsedPageNos()
	description:  Starts recovery of used pagenos 
	
	  - This method must be called prior to the first call to
	    RestoreUsedPageNos().
	  - Allocates memory!
	  
	arguments:    none
	return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
	virtual SAPDB_Bool BeginRestoreUsedPageNos() = 0;

	/*!-------------------------------------------------------------------------
    function:     RestoreUsedPageNos
    description:  Releases page numbers which do not belong to a restored data base 

                 - for all page number which are not marked as used at the UsedPageNoBitMapPage 
                   the corresponding converter entry is set to free and the block is released
                 - for all page numbers which are marked as used in the UsedPageNoBitMapPage but not
				   in the converter the correspondung page entry is reserved 
                 - all converter pages which contain any used page number are marked as used  
                   (see also FreePageNosNotInConverterBitMapPages)

    arguments:   TaskId                [in] task id   
	             UsedPageNoBitMapPage  [in] page which stores the page numbers of used pages in a bit map 
    return value: none
    --------------------------------------------------------------------------*/
    virtual void RestoreUsedPageNos (const tsp00_TaskId          TaskId,
		                             const Converter_BitMapPage &UsedPageNoBitMapPage) = 0;
	
	/*!-------------------------------------------------------------------------
    function:     FinishRestoreUsedPageNos
    description:  Releases page numbers which do not belong to a restored data base 
                 
                 - after all UsedPageNoBitMapPage of a back up are read this function 
                   scans the converter and removes all those converter leaf pages (and their entries)
                   which are not marked as used. (see also FreeUnusedPageNo)
                   This means all those converter entries which were neither marked as used nor as 
                   free in the  UsedPageNoBitMapPage's  (means not at all mentioned) are removed                  

    arguments:   TaskId                [in] task id   
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FinishRestoreUsedPageNos (const tsp00_TaskId TaskId) = 0;

		/*!-------------------------------------------------------------------------
    function:     LockUsedPageNoContainers
    description:  locks the UsedPageNoContainers
	
	  - the lock can be  used to allow only one task access to the used
	  page no container
	  
		arguments:    none
		return value: none
    --------------------------------------------------------------------------*/
	virtual void LockUsedPageNoContainers() = 0;
	
	/*!-------------------------------------------------------------------------
    function:     UnlockLockUsedPageNoContainers
    description:  locks the UsedPageNoContainers
	
	  - the lock can be  used to allow only one task access to the used
	  page no container
	  
		arguments:    none
		return value: none
    --------------------------------------------------------------------------*/
	virtual void UnlockLockUsedPageNoContainers() = 0;
};

/*!*****************************************************************************

   endclass: Converter_IBackUp

*******************************************************************************/


#endif  /* CONVERTER_IBACKUP_HPP */
