/*!**************************************************************************

  module      : Converter_ICommon.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author:       AlexanderK

  special area: converter
  description : very common simple functions of the converter


  last changed: 2001-01-15  10:15
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/

#ifndef CONVERTER_ICOMMON_HPP
#define CONVERTER_ICOMMON_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "ggg00.h"
#include "Converter/Converter_Types.hpp"
#include "DataAccess/Data_Types.hpp"

/*!*****************************************************************************

   class: Converter_ICommon

*******************************************************************************/

class Converter_ICommon 
{
	
public:
    /*!-------------------------------------------------------------------------
    function:     Converter_ICommon
    description:  returns the common interface of the converter

      - This static function returns the interface of the converter which
        allows access to all methods defined in this interface class 
  
    arguments:    none
    return value: (Converter_ICommon) instance of the converter 
    --------------------------------------------------------------------------*/
	static  Converter_ICommon & Instance();
	
    /*!-------------------------------------------------------------------------
    function:     Version()
    description:  Returns the actual converter version
	
	arguments:    none
	return value: (Converter_Version)
    --------------------------------------------------------------------------*/
    virtual Converter_Version Version () const = 0;
	
    /*!-------------------------------------------------------------------------
    function:     BackUpIsActive()
    description:  Checks whether converter backup is active
	
	arguments:    none
	return value: (SAPDB_Bool) true if active
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool BackUpIsActive() const = 0;
	
	/*!-------------------------------------------------------------------------
    function:     GetNumbersOfAllPages()
    description:  Returns the numbers of all  pages 
	
	  - returns the number of the maximum number of pages which
	  can be accomodated in the database 
	  
	arguments:    none
	return value: number of pages which can be stored in the database
    --------------------------------------------------------------------------*/
	virtual Data_PageCount GetNumberOfAllPages () const = 0;
	
	/*!-------------------------------------------------------------------------
    function:     GetNumberOfUsedPages()
    description:  Returns the numbers of all  currently used pages 
	
	  - returns  the number of all pages
				  which are currently occupied by data and converter pages
				  
	arguments:     none
	return value:  number of pages currently used
    --------------------------------------------------------------------------*/
	virtual Data_PageCount GetNumberOfUsedPages () const = 0;
	
	/*!-------------------------------------------------------------------------
    function:     GetDbSpaceStatistics()
    description:  returns a number of data describing the filling state of the converter 
	
    arguments:     TaskId      [in] task id
	               DbSpaceInfo [out] collection of state data of the converter
	return value:  none
    --------------------------------------------------------------------------*/
	virtual void GetDbSpaceStatistics (tsp00_TaskId        TaskId,
		                               tgg00_DbSpaceInfo  &DbSpaceInfo) = 0;
    /*!-------------------------------------------------------------------------
    function:     UsedConverterPages()
    description:  Returns the numbers of all currently used converter pages 
    
    arguments:     none
    return value:  number of converter pages (perm and temp) currently used
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt UsedConverterPages() const = 0;
     
    /*!-------------------------------------------------------------------------
    function:     IsDBFull()
    description:  True is returned, if the data area is (nearly) full.
    arguments:    TaskId [in] task id
    return value: (bool)
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool IsDBFull( 
        const tsp00_TaskId taskId ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     SecuritySpaceReached()
    description:  If the database filling is nearly full and therefore a 
                  sensible work is not possible this method returns a true
                  else false.
    arguments:    none
    return value: true if security is reached and database nearly full else false
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool SecuritySpaceReached() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GarbageCollectionNeeded()
    description:  If the database filling is about 90 percent and the converter
                  is online and not in maintenance mode, for example cold verify, 
                  diagnose repair or save data cold, this method returns an enum 
                  to  signalize that garbage collector support is needed. The
                  value Converter_GCNeeded determines that a garbage collection 
                  is needed.
    arguments:    none
    return value: Converter_GarbageCollectionReturnCode
    --------------------------------------------------------------------------*/
    virtual Converter_GarbageCollectionReturnCode GarbageCollectionNeeded() const = 0;
    
    /*!-------------------------------------------------------------------------
    function:     MaxTempPageNo()
    description:  Returns the maximum possible temporary page number

    arguments:    none
    return value: (Data_PageNo) maximum possible temporary page number
    --------------------------------------------------------------------------*/
    virtual Data_PageNo MaxTempPageNo() const = 0;

    /*!-------------------------------------------------------------------------
    function:     MaxPermDynamicPageNo()
    description:  Returns the maximum possible perm, dynamic page number

    arguments:    none
    return value: (Data_PageNo) maximum possible perm, dynamic page number
    --------------------------------------------------------------------------*/
    virtual Data_PageNo MaxPermDynamicPageNo() const = 0;
    
    /*!-------------------------------------------------------------------------
    function:     MaxPermStaticPageNo()
    description:  Returns the maximum possible static page number

      arguments:    none
      return value: (Data_PageNo) maximum possible static page number
    --------------------------------------------------------------------------*/
     virtual Data_PageNo MaxPermStaticPageNo() const = 0;

    /*!-------------------------------------------------------------------------
    function:     StartSavepoint()
    description:  If the number of changed converter pages and changed pages
                  within data cache is nearly to the number of free blocks within
                  FBM a savepoint is started. As long as this requested 
                  savepoint is finished this method will not request a new savepoint.
                  This will be done only if the converter is online and not 
                  in maintenance mode.
    arguments:    trans [in] transaction conect
    return value: none
    --------------------------------------------------------------------------*/
    virtual void StartSavepoint( tgg00_TransContext &trans ) = 0;

    /*!-------------------------------------------------------------------------
    function:     SavepointNeeded()
    description:  If the number of changed converter pages and changed pages
                  within data cache is nearly to the number of free blocks within
                  FBM a savepoint is needed. This will be done only if the 
                  converter is online and not in maintenance mode.
    arguments:    none
    return value: true if a savepoint is needed; else false
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool SavepointNeeded() const = 0;
};

/*!*****************************************************************************

   endclass: Converter_ICommon

*******************************************************************************/

#endif //CONVERTER_ICOMMON_HPP