/*!**************************************************************************

  module      : Converter_IManager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author:       AlexanderK

  special area: converter 
  description : abstract interface class to the converter
                containing general functions to manage the converter 

  last changed: 2001-01-10  15:45
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_IMANAGER_HPP
#define CONVERTER_IMANAGER_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "ggg00.h"   // PASCAL: GG_basic_constants_and_tyype

#include "Converter/Converter_Types.hpp"
#include "DataAccess/Data_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"

/*!*****************************************************************************

   class: Converter_IManager

*******************************************************************************/

class Converter_IManager
{
public:
    /*!-------------------------------------------------------------------------
       function:     Converter_IManager
       description:  returns the manager interface of the converter

                     - This static function returns the interface of the converter which
                     allows access to all methods defined in this interface class 

       arguments:    none
       return value: (Converter_IManager) instance of the converter 
       --------------------------------------------------------------------------*/
    static  Converter_IManager & Instance();

    /*!-------------------------------------------------------------------------
    function:     Restart
    description:  Creates converter for restarting an existing database

                  - The internal structures of the converter are created.
                  - Converter pages are read from devices. The converter root
                    page can be found at address 'rootaddr'.
                  - The parameters 'MaxDynamicPageNo', 'MaxStaticPageNo' and  'NumAllPages' are used 
                    to determine the initial size of the  converter.
                  - 'ConverterVersion' is the converter version of the database to be
                    recovered. If 'bIncConverterVersion' is true, the converter version
                    is incremented after a successful restart.
                  

    arguments:    Trans                     [in,out] transaction context
                  ConverterVersion          [in] converter version from restart record
                  ConverterRootBlockAddress [in] i/o address of converter root page
                  bIncConverterVersion      [in] indicates whether version must be
                                                 incremented after the restart
                  bLastSaveDataSuccessful   [in] indicates whether the last save data finished successfully
                  MaxDynamicPageNo          [in] maximun used dynamic page number of the restarted database
                  MaxStaticPageNo           [in] maximun used static page number of the restarted database 
                  NumAllPages               [in] maximum number of pages which can be stored in the data base. this
                                                 corresponds to the sum of size of all devices
                  BaseRegionId              [in] see next parameter
                  NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                 converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                 The more regions are supplied the better parallel accesses can be handled
                  ConverterPageSize         [in]  total size of a converter page in byte
                  DevNoBitCount             [in] defines the number of bits used in each converter 
                                              page entry to address a device   

    return value: none
    --------------------------------------------------------------------------*/
    virtual void Restart (
        const tsp00_TaskId        TaskId,
        const Converter_Version  &ConverterVersion,
        const IOMan_BlockAddress &ConverterRootBlockAddress,
        const SAPDB_Bool          bIncConverterVersion,
        const SAPDB_Bool          bLastSaveDataSuccessful,
        const Data_PageNo         MaxDynamicPageNo,
        const Data_PageNo         MaxStaticPageNo,
        const SAPDB_UInt          NumAllPages,
        const tsp00_RegionId      BaseRegionId,
        const SAPDB_UInt          NumConverterRegions,
        const SAPDB_UInt          ConverterPageSize,
        const SAPDB_UInt1         DevNoBitCount) = 0;

    /*!-------------------------------------------------------------------------
    function:     RestartForRestore
    description:  Restart the converter for restore pages.

                  - The internal structures of the converter are restarted
                  - Converter pages are read from devices. The converter root
                    page can be found at address 'rootaddr'.
                  - The parameters 'MaxDynamicPageNo', 'MaxStaticPageNo' and  'NumAllPages' are used 
                    to determine the initial size of the  converter.
                  - 'ConverterVersion' is the converter version of the database to be
                    recovered. If 'bIncConverterVersion' is true, the converter version
                    is incremented after a successful restart.
                  

    arguments:    TaskID                    [in] own Task ID
                  ConverterVersion          [in] converter version from restart record
                  ConverterRootBlockAddress [in] i/o address of converter root page
                  bLastSaveDataSuccessful   [in] indicates whether the last save data finished successfully
                  MaxDynamicPageNo          [in] maximun used dynamic page number of the restarted database
                  MaxStaticPageNo           [in] maximun used static page number of the restarted database 
                  NumAllPages               [in] maximum number of pages which can be stored in the data base. this
                                                 corresponds to the sum of size of all devices
                  BaseRegionId              [in] see next parameter
                  NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                 converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                 The more regions are supplied the better parallel accesses can be handled
                  ConverterPageSize         [in]  total size of a converter page in byte
                  DevNoBitCount             [in] defines the number of bits used in each converter 
                                              page entry to address a device   

    return value: none
    --------------------------------------------------------------------------*/
    virtual void RestartForRestore (
        const tsp00_TaskId        TaskId,
        const Converter_Version  &ConverterVersion,
        const IOMan_BlockAddress &ConverterRootBlockAddress,
        const SAPDB_Bool          bLastSaveDataSuccessful,
        const Data_PageNo         MaxDynamicPageNo,
        const Data_PageNo         MaxStaticPageNo,
        const SAPDB_UInt          NumAllPages,
        const tsp00_RegionId      BaseRegionId,
        const SAPDB_UInt          NumConverterRegions,
        const SAPDB_UInt          ConverterPageSize,
        const SAPDB_UInt1         DevNoBitCount) = 0;

    /*!-------------------------------------------------------------------------
     function:     Create
     description:  Creates an empty converter for a new data base (needed for activate serverdb)

                   - The internal structures of the converter are created.
                   - The parameter 'NumAllPages' is used to determine the initial size of the  converter.
           
     arguments:    TaskId                    [in] task id
                   NumAllPages               [in] maximum number of pages which can be stored in the data base. this
                                                  corresponds to the sum of size of all devices
                   BaseRegionId              [in] see next parameter
                   NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                  converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                  The more regions are supplied the better parallel accesses can be handled
                   ConverterPageSize         [in] total size of a converter page in byte
          DevNoBitCount             [in] defines the number of bits used in each converter 
                                               page entry to address a device  
     return value: none
     --------------------------------------------------------------------------*/
    /* creates an empty converter (needed for activate serverdb) */
    virtual void Create (
        const tsp00_TaskId       TaskId,
        const SAPDB_UInt         NumAllPages,
        const tsp00_RegionId     BaseRegionId,
        const SAPDB_UInt         NumConverterRegions,
        const SAPDB_UInt         ConverterPageSize,
        const SAPDB_UInt1        DevNoBitCount) = 0;

    /*!-------------------------------------------------------------------------
    function:     CreateForRestore
    description:  Creates an empty converter for restoring existing database

                  - The internal structures of the converter are created and
                    initialized.
                  - The parameters 'MaxDynamicPageNo', 'MaxStaticPageNo' and  'NumAllPages' 
                    are used to determine the initial size of the  converter.
                  - 'ConverterVersion' is the converter version of the database to be
                    recovered.

    arguments:    TaskId                    [in] task id
                  ConverterVersion          [in] converter version from restart record
                  bLastSaveDataSuccessful   [in] indicates whether the last save data finished successfully
                  MaxDynamicPageNo          [in] maximun used dynamic page number of the restarted database
                  MaxStaticPageNo           [in] maximun used static page number of the restarted database 
                  NumAllPages               [in] maximum number of pages which can be stored in the data base. this                                              corresponds to the sum of size of all device
                  BaseRegionId              [in] see next parameter
                  NumConverterRegions       [in] The converter is devided into  NumConverterRegions 'little'
                                                 converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                                 The more regions are supplied the better parallel accesses can be handled
                  ConverterPageSize         [in] total size of a converter page in byte
      DevNoBitCount             [in] defines the number of bits used in each converter 
                                              page entry to address a device             
    return value: none
    --------------------------------------------------------------------------*/
    virtual void  CreateForRestore (
        tsp00_TaskId              TaskId,
        const Converter_Version  &ConverterVersion,
        const SAPDB_Bool          bLastSaveDataSuccessful,
        const Data_PageNo         MaxDynamicPageNo,
        const Data_PageNo         MaxStaticPageNo,
        const SAPDB_UInt          NumAllPages,
        const tsp00_RegionId      BaseRegionId,
        const SAPDB_UInt          NumConverterRegions,
        const SAPDB_UInt          ConverterPageSize,
        const SAPDB_UInt1         DevNoBitCount) = 0;

    /*!-------------------------------------------------------------------------
    function:     Shutdown
    description:  Shuts the converter down

                  - All resources held by the converter are released and all 
                    member data are reseted to their initial values         
                  - All tasks waiting for the converter are resumed.

    arguments:    TaskId [in,out]    task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void Shutdown (const tsp00_TaskId TaskId) = 0;

    /*!-------------------------------------------------------------------------
     function:     FlushSerial
     description:  Saves converter to devspace (sequential writing)

                   - All converter pages which have been changed since the last
                     savepoint and which refer to permanent data pages are
                     written out to data devspaces.
                   - The routine returns several parameters describing the actual
                     state of the converter.
                   - If the flush is triggered by a SAVEPOINT the returned parameters are not 
                   needed since they were already inserted into the restart record by 
                   the Converter_ISavePoint::EndSavePoint method.
                   If the converter is not active the return value is false;

     arguments:    TaskId                    [in]   current task id
                   ConverterVersion          [out]  new valid converter version
                   ConverterRootBlockAddress [out]  Block address of the converter root page (to be stored in the restart record
                   MaxDynamicPageNo          [out]  maximun used dynamic page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (dynamic map) after a restart.
                   MaxStaticPageNo           [out]  maximun used static page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (static map) after a restart. 
                                                    (see Converter_IManager::Restart)
     return value: SAPDB_Bool
     --------------------------------------------------------------------------*/
    virtual SAPDB_Bool FlushSerial (
        const tsp00_TaskId  taskId,
        Converter_Version   &ConverterVersion,
        Data_PageNo         &MaxStaticPageNo,
        Data_PageNo         &MaxDynamicPageNo,
        IOMan_BlockAddress  &ConverterRootBlockAddress) = 0;

    /*!-------------------------------------------------------------------------
       function:     Expand()
       description:  Expands the converter

                     - Expands the converter to to manage NumPagesAdded more data pages 
                     - The routine is applicable with database in both cold and
                       warm state.
                     - The internal structures of the converter are adapted.

       arguments:    TaskId        [in] current task id
                     DevNo         [in] logical identifier for the added data devspace
                     DevSize       [in] size of the added data devspace in pages.
                     NumPagesAdded [in] number of pages added and usable by the data base (by assigning a new device)
       return value: none
       --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Expand (
        const tsp00_TaskId   TaskId,
        const IOMan_DeviceNo &DevNo,
        const SAPDB_UInt     DevSize,
        const SAPDB_UInt     NumPagesAdded) = 0;
};

/*!*****************************************************************************

   endclass: Converter_IManager

*******************************************************************************/

#endif  /* CONVERTER_IMANAGER_HPP */
