/*!**************************************************************************

  module      : Converter_PageNoManagement.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author:       AlexanderK

  special area: converter
  description : 


  last changed: 2001-01-15  10:33
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_PAGENOMANAGEMENT_HPP
#define CONVERTER_PAGENOMANAGEMENT_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"

#include "DataAccess/Data_Types.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: Converter_IPageNoManager

*******************************************************************************/

class Converter_IPageNoManager
{
public:
    /*!-------------------------------------------------------------------------
    function:     Converter_IPageNoManager
    description:  returns the page number manager interface of the converter

          - This static function returns the interface of the converter which
            allows access to all methods defined in this interface class 

    arguments:    none
    return value: (Converter_IPageNoManager) instance of the converter 
    --------------------------------------------------------------------------*/
    static Converter_IPageNoManager & Instance();

    /*!-------------------------------------------------------------------------
    function:     RequestNewTempPageNo()
    description:  Returns an unused temporary PageNo 

                  - After this operation the supplied page number is "used".
                  - the temp counter in TransContext is increased.
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag
    arguments:    taskId           [in] identification of calling task
                  newPageNo        [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool RequestNewTempPageNo(
        const tsp00_TaskId  taskId,
        Data_PageNo         &newPageNo ) = 0;
    
    /*!-------------------------------------------------------------------------
    function:     RequestNewPermDynamicPageNo()
    description:  Returns an unused perm,dynamic PageNo 

                  - After this operation the supplied page number is "used".
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag
    arguments:    taskId     [in] identification of calling task
                  newPageNo [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool RequestNewPermDynamicPageNo (
        const tsp00_TaskId        taskId,
        Data_PageNo               &newPageNo) = 0;

    /*!-------------------------------------------------------------------------
    function:     RequestNewStaticPageNo()
    description:  Returns an unused static PageNo for the required page recoverability.

                  - After this operation the supplied page number is "used".
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag
    arguments:    taskId           [in] identification of calling task
                  newPageNo        [out] new pageNo
    return value: (SAPDB_Bool) true if unused pageNo is available
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool RequestNewStaticPageNo (
        const tsp00_TaskId        taskId, 
        Data_PageNo               &newPageNo) = 0;
    
     /*!-------------------------------------------------------------------------
    function:     RequestSpecificiedNewStaticPageNo()
    description:  Checks whether specified pageid is still free and returns it

                  - After this operation the supplied page number is "used".
                  - In contrast to method 'RequestNewStaticPageNo()', not any page number
                    at choice, but a specific page number is requested.
                  - If the specified page number is already used, the error
                    e_multiple_converter_entry  is set in the TransContext.
                  - If one of the database fill percent limits 80, 90, 95, 96,
                    97, 98 and 99 is reached, a message is written to the knldiag

    arguments:    Trans            [in/out] transaction context
    return value: (Data_PageNo) requested page number 
    --------------------------------------------------------------------------*/
    virtual void RequestSpecificiedNewStaticPageNo  (tgg00_TransContext &Trans, 
                                                     const Data_PageNo   PageNo) = 0;
    
    /*!-------------------------------------------------------------------------
    function:     FreeTempPageNo()
    description:  Puts back the given  temporary page number to the "pool" of unused numbers.

    arguments:    taskId   [in] taskId
                  pageNo   [in] given page to be set to free

    return value: SAPDB_Bool true if pageNo could set to free
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool FreeTempPageNo(
        const tsp00_TaskId  taskId,
        const Data_PageNo   pageNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     FreePermDynamicPageNo()
    description:  Puts back the given  permanent, dynamic page number to the "pool" 
                  of unused numbers.

    arguments:    taskId   [in] taskId
                  pageNo   [in] given page to be set to free

    return value: SAPDB_Bool true if pageNo could set to free
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool FreePermDynamicPageNo(
        const tsp00_TaskId  taskId,
        const Data_PageNo   pageNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     FreePermStaticPageNo()
    description:  Puts back the given  static page number to the "pool" 
                  of unused numbers.

    arguments:    taskId   [in] taskId
                  pageNo   [in] given page to be set to free

    return value: SAPDB_Bool true if pageNo could set to free
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool FreePermStaticPageNo(
        const tsp00_TaskId  taskId,
        const Data_PageNo   pageNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     HandleDBFull()
    description:  Checks if the required count of pages is available.

                  - If the required count of pages is available this function returns true
                  - If the required count of pages is not available, the task is suspended
                    and waits until the wanted number of pages is available. After the task is resumed
                    and the pages are available the fundtion returns true. 
                  - Sometimes tasks cannot be suspended since this could cause dead locks. In this
                    case false is returned and sometimes even the error e_no_more_temp_space is set
                    to trigger a rollback.

    note:         Note this function does not reserve any number of pages for this tasks, i.e.
                  the information that a particular number is available  is dirty and may be 
                  even false, when this task really tries to get wanted number of free pages. 
    arguments:    Trans                 [in/out] transaction context
                  NumDataPagesRequested [in] number of pages wanted 
                  bAddDeviceIsPossible  [in] displays whether a new device can be added to the 
                                             data base to gain free pages

    return value: none
    --------------------------------------------------------------------------*/
    virtual void HandleDBFull (
        tgg00_TransContext  &Trans, 
        SAPDB_UInt          NumDataPagesRequested,
        const SAPDB_Bool    bAddDeviceIsPossible) = 0;
};

/*!*****************************************************************************

   endclass: Converter_IPageNoManager

*******************************************************************************/

#endif  /* CONVERTER_PAGENOMANAGEMENT_HPP */
