/*!**************************************************************************

  module      : Converter_IPageNoMapping.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author      : AlexanderK

  special area: converter 

  description : abstract interface class to those converter methods which concern                
                mapping of page numbers onto their physical device position  


  last changed: 2001-01-11  12:47
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_IPAGENOMAPPING_HPP
#define CONVERTER_IPAGENOMAPPING_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "DataAccess/Data_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "Converter/Converter_Types.hpp"


/*!*****************************************************************************

   class: Converter_IPageNoMapping

*******************************************************************************/



class Converter_IPageNoMapping
{
public:

    /*!-------------------------------------------------------------------------
    function:     Converter_IPageNoMapping
    description:  returns the page number mapping interface of the converter

                  - This static function returns the interface of the converter which
                  allows access to all methods defined in this interface class 

    arguments:    none
    return value: (Converter_IPageNoMapping) instance of the converter 
    --------------------------------------------------------------------------*/
    static  Converter_IPageNoMapping & Instance();

    /*!-------------------------------------------------------------------------
    function:     GetDevicePositionForReading()
    description:  Returns actual i/o address of data page with given page number 

                  - The actual i/o address of a given page number is read from
                    the converter and returned.
                  - If the corresponding converter entry contains an invalid
                    i/o address or if 'id' denotes a free pageid, the program
                    aborts.

    arguments: taskId           [in] current task id
               pageNo           [in] page number for which the corresponding block address is requested  
               pageRecoveryMode [in] page recovery mode of the page for which a deviced position is wanted 
    return value: (IOMan_BlockAddress) requested address
    --------------------------------------------------------------------------*/
  virtual IOMan_BlockAddress GetDevicePositionForReading (
            const tsp00_TaskId              taskId,
            const Data_PageNo               pageNo,
            const Data_PageRecoveryMode    &pageRecoveryMode) = 0;

    /*!-------------------------------------------------------------------------
    function:     GetDevicePositionForWriting()
    description:  Returns an i/o address where the data page can be written to 

                  - depending on the Recoverability of the page an its converter version
                    either the either the address of the currently occupied block is returned
                    or e new block address is supplied and the currently used block is set to 
                    'free after save point'
              
    arguments: taskId           [in]      identification of calling task
               trError          [out]     current error code
               PageNo           [in]      page number for which the corresponding block address is requested  
               PageRecoveryMode [in]      recovery mode of the page to write
               PageConverterVersion [in]  converter version of the page to write
    return value: (IOMan_BlockAddress) requested address
    --------------------------------------------------------------------------*/
    virtual IOMan_BlockAddress GetDevicePositionForWriting (
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        const Data_PageNo           &PageNo,
        const Data_PageRecoveryMode &PageRecoveryMode,
        const Converter_Version     &PageConverterVersion) = 0;

    /*!-------------------------------------------------------------------------
    function:     SetDevicePositionForWriting()
    description:  Sets a new  i/o address into the converter entry of a given page 

                  - set a new  i/o address into the converter entry of a given page 
                  - The block occuied so far is either set to 'free' for temp pages or
                    set to 'free after save point' for permanent pages
              
    arguments: taskId               [in]  identification of calling task
               trError              [out]     current error code
               PageNo               [in]  page number for which the corresponding block address is requested  
               PageRecoveryMode     [in]  recovery mode of the page to write
               PageConverterVersion [in]  converter version of the page to write
               BlockAddress         [in]  new block address which is 
                                          to insert in the corresponding converter entry
    return value: none
    --------------------------------------------------------------------------*/
    virtual void SetDevicePositionForWriting (
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        const Converter_Version     &ConverterVersion,
        const Data_PageNo           &PageNo,
        const Data_PageRecoveryMode &PageRecoveryMode,
        const IOMan_BlockAddress    &BlockAddress) = 0;
};

/*!*****************************************************************************

   endclass: Converter_IPageNoMapping

*******************************************************************************/

#endif  /* CONVERTER_IPAGENOMAPPING_HPP */
