/*!**************************************************************************

  module      : Converter_ISavePoint.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author      : AlexanderK

  special area: converter 
  description : abstract savepoint interface class to the converter 


  last changed: 2001-01-10  14:49
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_ISAVEPOINT_HPP
#define CONVERTER_ISAVEPOINT_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "Converter/Converter_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"

/*!*****************************************************************************

   class: Converter_ISavePoint

*******************************************************************************/

class Converter_ISavePoint
{
public:
    /*!-------------------------------------------------------------------------
    function:     Converter_ISavePoint
    description:  returns the savepoint interface of the converter

          - This static function returns the interface of the converter which
            allows access to all methods defined in this interface class 

    arguments:    none
    return value: (Converter_ISavePoint) instance of the converter 
    --------------------------------------------------------------------------*/
    static Converter_ISavePoint & Instance();

    /*!-------------------------------------------------------------------------
    function:     BeginSavepoint
    description:  Sets all savepoint activ flags in map control 

          - After all save point activ flags are set no more blocks can be set to the state 
            free_after_save_point. Instead pages wich become free are stored in a FreePageNoList.
            After the save point is finished successfully all page numbers which are in the FreePageNoList
            are released and the corresponding blocks are set to free or free_after_savepoint 
            (see method) FreePageNosAfterSavepoint.

    arguments:    TaskId                   [in]   current task id
    return value: none
    exceptions:   none
    --------------------------------------------------------------------------*/
    virtual void  BeginSavepoint (const tsp00_TaskId &TaskId) = 0;


    /*!-------------------------------------------------------------------------
    function:     EndSavepoint
    description:  Increases the converter version and resets all savepoint activ flags 
  

    arguments:    taskId [in]   current task id
    return value: none
    exceptions:   none
    --------------------------------------------------------------------------*/
    /* increase the converter version and reset savepoint activ flags */
    virtual void EndSavepoint( const tsp00_TaskId   taskId ) = 0;
    
    /*!-------------------------------------------------------------------------
    function:     FreePageNosAfterSavepoint
    description:  Release all PageNos stored in the FreePageNoList  

          - After the save point is finished successfully all page numbers which are in the FreePageNoList
            are released and the corresponding blocks are set to free or free_after_savepoint 

    arguments:    TaskId                   [in]   current task id
    return value: none
    exceptions:   none
    --------------------------------------------------------------------------*/
    virtual void FreePageNosAfterSavepoint (const tsp00_TaskId &TaskId) = 0;

    /*!-------------------------------------------------------------------------
     function:     FlushParallel
     description:  Saves converter to devspace (parallel writing)

                   - All converter pages which have been changed since the last
                     savepoint and which refer to permanent data pages are
                     written out to data devspaces.
                   - The routine returns several parameters describing the actual
                     state of the converter.

     arguments:    Trans [in,out]     transaction context
                   ConverterVersion          [out]  new valid converter version
                   ConverterRootBlockAddress [out]  Block address of the converter root page (to be stored in the restart record
                   MaxDynamicPageNo          [out]  maximun used dynamic page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (dynamic map) after a restart.
                   MaxStaticPageNo           [out]  maximun used static page number. This number has
                                                    to be stored in the restart page in order the calculate the 
                                                    right size of the converter (static map) after a restart. 
                                                    (see Converter_IManager::Restart)
     return value: none
     --------------------------------------------------------------------------*/
    /* flushes all changed permanent page frames held by the converter are flushed to a device */
    virtual void FlushParallel (tgg00_TransContext  &Trans,
                        Converter_Version   &ConverterVersion,
                        Data_PageNo         &MaxStaticPageNo,
                        Data_PageNo         &MaxDynamicPageNo,
                        IOMan_BlockAddress  &ConverterRootBlockAddress,
                        SAPDB_UInt          &NumPages,
                        SAPDB_UInt          &NumIO) = 0;

};

/*!*****************************************************************************

   endclass: Converter_ISavePoint

*******************************************************************************/

#endif  /* CONVERTER_ISAVEPOINT_HPP */
