/*!**************************************************************************

  module      : Converter_IVerify.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author      : AlexanderK

  special area: converter
  description : abstract interface class to the verify function of the converter 


  last changed: 2001-01-29  18:48
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_IVERIFY_HPP
#define CONVERTER_IVERIFY_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "DataAccess/Data_Types.hpp"

/*!*****************************************************************************

   class: Converter_IVerify

*******************************************************************************/

class Converter_IVerify
{
public:
    /*!-------------------------------------------------------------------------
    function:     Converter_IVerify
    description:  returns the verify interface of the converter

          - This static function returns the interface of the converter which
            allows access to all methods defined in this interface class 

    arguments:    none
    return value: (Converter_IVerify) instance of the converter 
    --------------------------------------------------------------------------*/

    static  Converter_IVerify & Instance();

    /*!-------------------------------------------------------------------------
    function:     BeginColdVerify
    description:  Prepares converter verification in "cold" database state

    arguments:    taskId[in]   task id
    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginColdVerify(  const tsp00_TaskId    taskId ) = 0;

     /*!-------------------------------------------------------------------------
    function:     EndColdVerify
    description:  Finishes converter verification in "cold" database state

        - All pages which are not marked as used (see method MarkPageNoAsUsed)
          are released          

        - The routine returns several parameters describing the actual
          state of the converter and of the database.

    arguments:    TaskId                  [in]   task id
                  NumStaticPermDataPages  [out]  number of permanent data pages used
                                                 with static page numbers
                  NumDynamicPermDataPages [out]  number of permanent data pages used
                                                 with dynamic page numbers
                  NumPermConPages         [out]  number of converter pages used for
                                                 permanent data pages
                  NumReleaseBlocks        [out]  number of released blocks, which were
                                                 not accessed during verify
    return value: none
    --------------------------------------------------------------------------*/

    virtual void EndColdVerify (
        const tsp00_TaskId  taskId,
        Data_PageCount      &NumStaticPermDataPages,
        Data_PageCount      &NumDynamicPermDataPages,
        Data_PageCount      &NumPermConvPages,
        Data_PageCount      &NumReleaseBlocks ) = 0;

    /*!-------------------------------------------------------------------------
    function:     EndFailedColdVerify
    description:  Finishes a failed converter verification

        - The converter is restored to the state it had before the verification 
          was started by the BeginColdVerify

    arguments:    TaskId         [in]  task id
    return value: none
    --------------------------------------------------------------------------*/

    virtual void EndFailedColdVerify (const tsp00_TaskId  TaskId) = 0;

    /*!-------------------------------------------------------------------------
    function:     MarkPageNoAsUsed
    description:  Marks a page in the converter as  used 
    arguments:    TaskId         [in] task id
                  PageNo         [in] page number to be marked as used
                  AddressingMode [in] addressing mode of the page to mark 

    return value: none
    --------------------------------------------------------------------------*/
    
    virtual void MarkPageNoAsUsed (
        const tsp00_TaskId          taskId,
        const Data_PageNo           PageNo,
        const Data_AddressingMode   AddressingMode) = 0;
};

/*!*****************************************************************************

   endclass: Converter_IVerify

*******************************************************************************/

#endif  /* CONVERTER_IVERIFY_HPP */
