/*!*****************************************************************************

  module:       Converter_Index.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       TillL

  special area: Converter

  description:  Parent index for management of converter index pages
                
  see also:     

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef CONVERTER_INDEX_HPP
#define CONVERTER_INDEX_HPP


#include "Container/Container_Vector.hpp"
#include "Container/Container_SortableVector.hpp"
#include "Converter/Converter_Chain.hpp"
#include "Converter/Converter_IndexEntry.hpp"
#include "Converter/Converter_LeafPage.hpp"
#include "Converter/Converter_PageFrameAllocator.hpp"
#include "Converter/Converter_LeafReference.hpp"
#include "KernelCommon/Kernel_Dump.hpp"

/*--------------------------------------------------------------------------
declaration : Converter_IndexVector
description:  buffer which contains converter index entries (== the index column)
--------------------------------------------------------------------------*/
typedef Container_Vector<Converter_IndexEntry>
Converter_IndexVector;


/*!*****************************************************************************

   class: Converter_IndexIterator

*******************************************************************************/
class Converter_IndexIterator
{
    friend class Converter_Index;

public:

    /*!-------------------------------------------------------------------------
    function:     Converter_IndexIterator()
    description:  Constructor
    arguments:    indexiter    [in] iterator at start position in index column
                  indexenditer [in] iterator behind end position in index column
    return value: none
    --------------------------------------------------------------------------*/

    Converter_IndexIterator
    (const Converter_IndexVector::Iterator  &indexiter,
     const Converter_IndexVector::Iterator  &indexenditer)
    :
    m_EntryIter    (indexiter),
    m_EntryEndIter (indexenditer),
    m_PageIter     ((*indexiter).GetPage().Begin()),
    m_PageEndIter  ((*indexiter).GetPage().End())
    {}

    /*!-------------------------------------------------------------------------
    function:     operator!=()
    description:  Compares with other iterator
    arguments:    iter [in]    other
    return value: (SAPDB_Bool) true if unequal
    --------------------------------------------------------------------------*/
    SAPDB_Bool operator!=(const Converter_IndexIterator &iter) const
    {
        return (m_EntryIter != iter.m_EntryIter) ||
               (m_PageIter  != iter.m_PageIter);
    }

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  Prefix increment; advances iterator
    arguments:    none
    return value: (Converter_IndexIterator&) this instance
    --------------------------------------------------------------------------*/
    Converter_IndexIterator& operator++()
    {
        SAPDBERR_ASSERT_STATE(CanAdvance());

        // jump to next converter entry on the current page
        ++m_PageIter;

        // check if we have reached index page end
        if( m_PageIter == m_PageEndIter )
        {
            // jump to next converter index page
            ++m_EntryIter;
            while(CanAdvance() && !((*m_EntryIter).PageIsAssigned())) {
                ++m_EntryIter;
            }

            // check if we have not reached last index entry
            if (CanAdvance())
            {
                // reset page iterators
                Converter_IndexPage &indexpage = (*m_EntryIter).GetPage();
                m_PageIter                     = indexpage.Begin();
                m_PageEndIter                  = indexpage.End();
            }
        }
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     CanAdvance()
    description:  Checks whether iterator can advance with operator++()
    arguments:    none
    return value: (SAPDB_Bool) true if iter can advance
    --------------------------------------------------------------------------*/
    SAPDB_Bool CanAdvance() const
    {
        return m_EntryIter != m_EntryEndIter;
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockAddress()
    description:  Returns the block address the iterator points to
    arguments:    none
    return value: (IOMan_BlockAddress) actual block address
    --------------------------------------------------------------------------*/
    IOMan_BlockAddress GetBlockAddress() const
    {
        return m_PageIter.GetBlockAddress();
    }

    /*!-------------------------------------------------------------------------
    function:     EntryInUse()
    description:  Checks whether actual entry is in use
    arguments:    none
    return value: (SAPDB_Bool) true if entry is in use
    --------------------------------------------------------------------------*/
    SAPDB_Bool EntryInUse() const
    {
        return m_PageIter.EntryInUse();
    }

protected:

    Converter_IndexVector::Iterator m_EntryIter;
    Converter_IndexVector::Iterator m_EntryEndIter;
    Converter_PageEntryIterator     m_PageIter;
    Converter_PageEntryIterator     m_PageEndIter;
};
/*!*****************************************************************************

   endclass: Converter_IndexIterator

*******************************************************************************/

/*!*****************************************************************************

   class: Converter_Index

   description: Converter parent index

     -Note, that the index tree consists always of two index an one leaf level

*******************************************************************************/
class Converter_Index
{

public:

    /*!-------------------------------------------------------------------------
    declaration : IndexEntry
    description:  a index entry object (== slot of the index column)
    --------------------------------------------------------------------------*/
    typedef Converter_IndexEntry                            IndexEntry;
    /*!-------------------------------------------------------------------------
    declaration : IndexEntryCount
    description:  any number of index entries
    --------------------------------------------------------------------------*/
    typedef Converter_IndexEntryCount                       IndexEntryCount;
    /*!-------------------------------------------------------------------------
    declaration : IndexEntryNo
    description:  index of a certain index entry
    --------------------------------------------------------------------------*/
    typedef Converter_IndexEntryNo                          IndexEntryNo;
    /*!-------------------------------------------------------------------------
    declaration : LeafPage
    description:  a converter leaf page handler object
    --------------------------------------------------------------------------*/
    typedef Converter_LeafPage                              LeafPage;
    /*!-------------------------------------------------------------------------
    declaration : IndexPage
    description:  a converter index page handler object
    --------------------------------------------------------------------------*/
    typedef Converter_IndexPage                             IndexPage;
    /*!-------------------------------------------------------------------------
    declaration : PageEntryCount
    description:  any number of page entries on a converter page
    --------------------------------------------------------------------------*/
    typedef Converter_PageEntryCount                        PageEntryCount;
    /*!-------------------------------------------------------------------------
    declaration : PageEntryNo
    description:  index of a certain page entry on a converter leaf page
    --------------------------------------------------------------------------*/
    typedef Converter_PageEntryNo                           PageEntryNo;
    /*!-------------------------------------------------------------------------
    declaration : IndexIterator
    description:  standard iterator for the whole index
    --------------------------------------------------------------------------*/
    typedef Converter_IndexIterator                         IndexIterator;

private:

    /*!-------------------------------------------------------------------------
    declaration : PageEntryIterator
    description:  iterator for converter pages
    --------------------------------------------------------------------------*/
    typedef Converter_PageEntryIterator                     PageEntryIterator;

    /*--------------------------------------------------------------------------
    declaration : EntryChain
    description:  doubly linked chain of map entries
    --------------------------------------------------------------------------*/
    typedef Converter_Chain
    <Converter_IndexVector, Converter_IndexEntry>   EntryChain;

public:

    /*!-------------------------------------------------------------------------
    declaration : EntryIterator
    description:  a converter index column iterator
    --------------------------------------------------------------------------*/
    typedef Converter_IndexVector::Iterator                 EntryIterator;

    /*!-------------------------------------------------------------------------
    declaration : PartiallyUsedEntryIterator
    description:  iterator of partially used entries
    --------------------------------------------------------------------------*/
    typedef EntryChain::Iterator                            PartiallyUsedEntryIterator;

    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  resource type used for dynamic memory allocation
    --------------------------------------------------------------------------*/
    typedef Converter_IndexVector::Allocator                Allocator;

    /*--------------------------------------------------------------------------
    declaration : BlockVector
    description:  queue of converter block numbers
    --------------------------------------------------------------------------*/
    typedef Container_SortableVector<Converter_LeafReference> BlockVector;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Converter_Index()
    description:  Default constructor
                    
    arguments:    alloc [in]  allocator for dynamic index resources
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Index
    (Allocator&                     alloc)
    :
    m_EntryVector                  (alloc),
    m_PartiallyUsedEntryChain      (m_EntryVector),
    m_FullEntryChain               (m_EntryVector),
    m_EmptyEntryChain              (m_EntryVector),
    m_Allocator                    (alloc)
    {}

    /*!-------------------------------------------------------------------------
    function:     Create()
    description:  Builds the initial index consisting of an empty root page

    arguments:    frameresource    [in] resource of page frames
    return value: none
    --------------------------------------------------------------------------*/
    void Create
    (Converter_PageFrameAllocator&       frameresource);

    /*!-------------------------------------------------------------------------
    function:     Restart()
    description:  Builds up the index 
        
                  - The index is initialized and all index pages are loaded
                    into memory. 

    arguments:    frameresource [in] resource of page frames
                  TaskId        [in] own task id
                  rootaddr      [in] block address of the converter root
    return value: none
    --------------------------------------------------------------------------*/
    void Restart
    (Converter_PageFrameAllocator&  frameresource,
     const tsp00_TaskId&            TaskId,
     const IOMan_BlockAddress&      rootaddr);

    /*!-------------------------------------------------------------------------
    function:     CreateInputQueues()
    description:  create sorted queues of converter pages, one queue per volume 
        
    arguments:    volumeList        [in,out] list of converter page queues
                  ConverterPageSize [in] size of one ConverterPage
    return value: none
    --------------------------------------------------------------------------*/
    void CreateInputQueues(Container_Vector<BlockVector> &volumeList,
                           const SAPDB_UInt               ConverterPageSize);

    /*!-------------------------------------------------------------------------
    function:     Shutdown()
    description:  Deletes the index and frees all pages
        
                  - The index is deleted; memory is deallocated,

    arguments:    frameresource    [in] resource of page frames
    return value: none
    --------------------------------------------------------------------------*/
    void Shutdown
    (Converter_PageFrameAllocator&  frameresource);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Miscellaneous                                                */
    /*!-------------------------------------------------------------------------
    function:     Flush()
    description:  Flushes all index pages to device
        
                  - This method is used at the end of a savepoint.

    arguments:    TaskId [in]   own task id
                  NumPages [out] number of pages written
                  NumIO    [out] number of clustered IO transactions
    return value: (IOMan_BlockAddress) address of the index root page
    --------------------------------------------------------------------------*/

    IOMan_BlockAddress Flush
    (const tsp00_TaskId&    TaskId,
           SAPDB_UInt&      NumPages,
           SAPDB_UInt&      NumIO);

    /*!-------------------------------------------------------------------------
    function:     Dump()
    description:  Dumps the index entries into the kernel dump file
    arguments:    dump [in/out] kernel dump file
    return value: none
    --------------------------------------------------------------------------*/

    void Dump( Kernel_Dump  &dump );

    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Checks the structure of the parent index, including the
                  free counter and the parent index chains.
    arguments:    none
    return value: (SAPDB_Bool) true is paranet index is okay else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool Verify() const; // PTS 1113242 TS 2001-12-21

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Miscellaneous                                             */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Index access                                                 */
    /*!-------------------------------------------------------------------------
    function:     Reserve()
    description:  Provides capacity to hold at least the requested number of
                  index entries
                    
                  - This method may require reallocation of the index.

    arguments:    size [in]     requested index capacity
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool Reserve
    (IndexEntryCount                size)
    {
        return m_EntryVector.Resize(size);
    }

    /*!-------------------------------------------------------------------------
    function:     IntegratePage()
    description:  Updates index data on given converter leaf page

    arguments:    taksId [in] taskId    
    arguments:    page   [in]  converter leaf page handler

    return value: none
    --------------------------------------------------------------------------*/
    void IntegratePage(
        const tsp00_TaskId taskId,
        LeafPage&           page );

    /*!-------------------------------------------------------------------------
    function:     GetParentReference()
    description:  returns the parent reference associated with the given iterator

    arguments:    IndexIterator  [in]  iterator pointing to a particular entry in on a index page

    return value: Converter_Page:ParentReference : parent reference associated with the given iterator
    --------------------------------------------------------------------------*/
    Converter_ParentReference GetParentReference (const Converter_IndexIterator &indexIterator) const
    {
        return (Converter_ParentReference (m_EntryVector.GetIndex(indexIterator.m_EntryIter),
                                 indexIterator.m_PageIter.GetPosition()));
    }

    /*!-------------------------------------------------------------------------
    function:     RemoveEmptyLeaves()
    description:  remove all references to empty leafPages from index

    arguments:    taskId     [in,out]  taskId
    arguments:    parentRefs [in,out]  list of references to empty leafPages in one mapSection

    return value: none
    --------------------------------------------------------------------------*/
    void RemoveEmptyLeaves (
             const tsp00_TaskId                          taskId,
             Container_Vector<Converter_ParentReference> &parentRefs);
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Index access                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterators                                                    */
    /*!-------------------------------------------------------------------------
    function:     GetIterator()
    description:  Returns iterator at first level1 index page entry

                  - In contrast to other iterators, this iterator has a
                    method 'CanAdvance()' to determine whether the iterator
                    has reached the last page entry.
    arguments:    none
    return value: (IndexIterator)
    --------------------------------------------------------------------------*/
    IndexIterator GetIterator()
    {
        return IndexIterator(Level1Begin(), Level1End());
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterators                                                 */
    /* -----------------------------------------------------------------------*/



private:

    /*-------------------------------------------------------------------------
    function:     RequestFreeParentReference()
    description:  Reserves a free parent reference on an index page
                    
                  - An unused (free) entry is reserved and returned.
                  - If necessary, a new index page is allocated by 'frameresource'
    arguments:    taskId  [in] taskId
    return value: (Converter_ParentReference) reserved parent reference
    --------------------------------------------------------------------------*/
    Converter_ParentReference RequestFreeParentReference( const tsp00_TaskId taskId );

    /*-------------------------------------------------------------------------
    function:     Level1Begin()
    description:  Returns iterator at first level1 index entry
    arguments:    none
    return value: (EntryIterator)
    --------------------------------------------------------------------------*/
    EntryIterator Level1Begin()
    {
        return m_EntryVector.Position(Level1StartEntryNo());
    }

    /*-------------------------------------------------------------------------
    function:     Level1End()
    description:  Returns iterator behind last level1 index entry
    arguments:    none
    return value: (EntryIterator)
    --------------------------------------------------------------------------*/
    EntryIterator Level1End()
    {
        return m_EntryVector.End();
    }


    void SetRootPage(IndexPage &page);

    IndexEntryNo InsertPage (IndexPage  &page);

    static PageEntryNo RootEntryNo()
    {
        return 0;
    }

    static PageEntryNo Level1StartEntryNo()
    {
        return 1;
    }

    IndexEntry& GetRootEntry()
    {
        return m_EntryVector[RootEntryNo()];
    }

    IndexEntry& GetEntry
    (IndexEntryNo                   entryno)
    {
        return m_EntryVector[entryno];
    }


private:

    /*--------------------------------------------------------------------------
    declaration : m_EntryVector
    description:  array of entries
    --------------------------------------------------------------------------*/
    Converter_IndexVector           m_EntryVector;
    /*--------------------------------------------------------------------------
    declaration:  m_PartiallyUsedEntryChain
    description:  doubly linked chains of partially used index entries 
    --------------------------------------------------------------------------*/
    EntryChain                      m_PartiallyUsedEntryChain;
    /*--------------------------------------------------------------------------
    declaration:  m_FullEntryChain
    description:  doubly linked chains of completely filled index entries
    --------------------------------------------------------------------------*/
    EntryChain                      m_FullEntryChain;
    /*--------------------------------------------------------------------------
    declaration:  m_EmptyEntryChain
    description:  doubly linked chains of empty index entries
    --------------------------------------------------------------------------*/
    EntryChain                      m_EmptyEntryChain;

    /*--------------------------------------------------------------------------
    declaration : m_Allocator
    description:  allocator for IOQueues
    --------------------------------------------------------------------------*/
    Allocator&               m_Allocator;

};
/*!*****************************************************************************

   endclass: Converter_Index

*******************************************************************************/

#endif // CONVERTER_INDEX_HPP
