/*!*****************************************************************************

  module:       Converter_IndexPage.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       TillL

  special area: Converter

  description:  Handler class for converter index pages

  see also:     Kernel_IPage.html 
                Kernel_Page.html
                Kernel_Page74.html
                Converter_Page.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef CONVERTER_INDEXPAGE_HPP
#define CONVERTER_INDEXPAGE_HPP

#include "Converter/Converter_Page.hpp"


/*!*****************************************************************************

   class: Converter_IndexPage

   description:

   Specialized converter page class for index pages (concrete class)

   Call static method 'SetDevNoBitCount()' once prior to first usage
   of this class.

*******************************************************************************/
class Converter_IndexPage : public Converter_Page
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Converter_IndexPage()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Converter_IndexPage();

    /*!-------------------------------------------------------------------------
    function:     Converter_IndexPage()
    description:  Constructor; assigns frame

                  Does not change frame contents.
    arguments:    frame [in]    page frame
    return value: none
    --------------------------------------------------------------------------*/
    Converter_IndexPage
        (const PageFrame&               frame);
   
    /*!-------------------------------------------------------------------------
    function:     Converter_IndexPage()
    description:  Constructor; assigns frame and initializes frame contents
    arguments:    frame   [in]  page frame
                  version [in]  converter version
                  parent  [in]  reference to parent in converter tree parent index
    return value: none
    --------------------------------------------------------------------------*/
    Converter_IndexPage
        (const PageFrame&               frame, 
         const Version&                 version, 
         const Converter_ParentReference&         parent);
   
    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame   [in]  page frame
                  version [in]  converter version
                  parent  [in]  reference to parent in converter tree parent index
    return value: none
    --------------------------------------------------------------------------*/
    void SetAndInitializeFrame
        (const PageFrame&               frame, 
         const Version&                 version, 
         const Converter_ParentReference&         parent);
   
    /*!-------------------------------------------------------------------------
    function:     InitializeFrame()
    description:  Initializes frame contents
    arguments:    version [in]  converter version
                  parent  [in]  reference to parent in converter tree parent index
    return value: none
    --------------------------------------------------------------------------*/
    void InitializeFrame
        (const Version&                 version, 
         const Converter_ParentReference&         parent);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /*!-------------------------------------------------------------------------
    function:     GetEntryCount()
    description:  Returns total count of entries on converter index page.
    arguments:    none
    return value: (Converter_PageEntryCount) entry count
    --------------------------------------------------------------------------*/
    virtual PageEntryCount GetEntryCount() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Protected global functions                                   */
    /*!-------------------------------------------------------------------------
    function:     InitializeFlagInfo()
    description:  Initializes all entry flags on the page with 0's.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    virtual void InitializeFlagInfo();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Protected global functions                                */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     InUseFlagsOffset()
    description:  Returns offset on page where entire in-use flags block begins
    arguments:    none
    return value: (OffsetType) offset
    --------------------------------------------------------------------------*/
    virtual OffsetType InUseFlagsOffset() const;

    /*-------------------------------------------------------------------------
    function:     CalcEntryCount()
    description:  Returns total count of entries on converter index page.
                  
                  - Preconditions: as = 4 or 8, hs = multiple of 4
                  - usablesize (us) = pagesize(ps) - headersize(hs) - trailersize(ts)
                  - entrysize(es)   = addrsize(as) + flagsize(fs) (no off-cut)
                  - as              = 4 bytes
                  - fs              = 1/8 (1 bit)
                  - es              = as + 1/8 = 33/8
    arguments:    none
    return value: (PageEntryCount) entry count
    --------------------------------------------------------------------------*/
    PageEntryCount CalcEntryCount() const;
    
};
/*!*****************************************************************************

   endclass: Converter_IndexPage

*******************************************************************************/


//------------------------------------------------------------------------------

inline Converter_IndexPage::Converter_IndexPage()
{}

//------------------------------------------------------------------------------

inline Converter_IndexPage::Converter_IndexPage
    (const PageFrame&               frame)
    : 
     Converter_Page(frame)
{}

//------------------------------------------------------------------------------

inline Converter_IndexPage::Converter_IndexPage
    (const PageFrame&               frame, 
     const Version&                 version, 
     const Converter_ParentReference&         parent)
    : 
     Converter_Page(frame)
{
    InitializeFrame(version, parent);
}

//------------------------------------------------------------------------------

inline void 
    Converter_IndexPage::SetAndInitializeFrame
    (const PageFrame&               frame, 
     const Version&                 version, 
     const Converter_ParentReference&         parent)
{
    SetFrame(frame);
    InitializeFrame(version, parent);
}

//------------------------------------------------------------------------------

inline void 
    Converter_IndexPage::InitializeFrame
    (const Version&                 version, 
     const Converter_ParentReference&         parent)
{
	Converter_Page::InitializeFrame(version, parent);
    WriteConverterPageType(Converter_IndexPageType);
    SAPDBERR_ASSERT_ARGUMENT (Verify());
}

//------------------------------------------------------------------------------

inline Converter_IndexPage::PageEntryCount 
    Converter_IndexPage::GetEntryCount() const
{
    static PageEntryCount count = CalcEntryCount();
    return count;
}

//------------------------------------------------------------------------------

inline void 
    Converter_IndexPage::InitializeFlagInfo()
{

    Converter_FlagInfo::FlagInfoData* flaginfoptr    = GetFlagInfo(InUseFlagsOffset(), 0).FlagInfoDataPtr();
	Converter_FlagInfo::FlagInfoData* flaginfoendptr = GetFlagInfo(InUseFlagsOffset(), GetEntryCount() -1).FlagInfoDataPtr();    
    
	while (flaginfoptr <= flaginfoendptr)
    {
        *flaginfoptr = 0;
        ++flaginfoptr;
    }
}
//------------------------------------------------------------------------------

inline Converter_IndexPage::OffsetType 
    Converter_IndexPage::InUseFlagsOffset() const
{
    static OffsetType offs 
           = sizeof(PersistentHeader) + GetEntryCount() * sizeof(PersistentEntry);
    return offs;
}

//------------------------------------------------------------------------------


#endif // CONVERTER_INDEXPAGE_HPP
