/*!**************************************************************************

  module      : Converter_MapControl.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: converter

  description : manages the parallel access to the structure mapping the page numbers onto their
                physical block address

  last changed: 2001-01-10  18:13

  see also    : 

  author      : AlexanderK

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_MAPCONTROL_HPP
#define CONVERTER_MAPCONTROL_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

#include "Converter/Converter_Map.hpp"
#include "Converter/Converter_MapSectionData.hpp"
#include "Container/Container_Vector.hpp"
#include "KernelCommon/Kernel_Common.hpp"



/*===========================================================================*
 *  FORWARD DECLARATION                                                      *
 *===========================================================================*/

class Data_PageRecoveryMode;
class Kernel_Dump;
class SAPDBMem_IRawAllocator;


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define MAX_CONVERTER_SECTIONS                  64


/*!*****************************************************************************

   class: Converter_MapControl

*******************************************************************************/

class Converter_MapControl
{
public:

    /// Defines abbreviation for a section data

    typedef Container_Vector<Converter_MapSectionData>  SectionData;

    /// Defines abbreviation for a section data iterator

    typedef SectionData::Iterator                       SectionDataIterator;

public:

    /*!-------------------------------------------------------------------------
    enddeclarations: Type definitions
    --------------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Manager                                                      */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Initialize
    description:  sets all members of the map control according to the input parameters

         
    arguments:    taskId              [in]  task id
                  maxDataPnoInMap     [in]  maximum page number to manage by the converter
                  baseRegionId        [in]  see next parameter
                  numConverterRegions [in]  the converter is devided into  NumConverterRegions 'little'
                                            converter each protected by an own region with BaseRegionId <=RegionId<BaseRegionId+NumConverterRegions  
                                            The more regions are supplied the better parallel accesses can be handled
                  clusterSize         [in]
                  converterPageSize   [in]  total size of a converter page in byte
    return value: (SAPDB_Bool) returns true if the initilization was successful
    --------------------------------------------------------------------------*/

    SAPDB_Bool  Initialize(
        const tsp00_TaskId   taskId,
        const SAPDB_UInt     maxDataPnoInMap,
        const tsp00_RegionId baseRegionId,
        const SAPDB_UInt     numConverterRegions,
        const SAPDB_UInt     clusterSize,
        const SAPDB_UInt     converterPageSize );

    /*!-------------------------------------------------------------------------
    function:     Shutdown
    description:  sets all members of the map control to their initial values

                  - All resources held by members of the map control are returned.
                  - Resets all map section data and checks if all free pno lists are empty 
         
    arguments:    taskId [in] taskid
    return value: none
    --------------------------------------------------------------------------*/

    void  Shutdown( const tsp00_TaskId   taskId );

    /*!-------------------------------------------------------------------------
    function:     Expand
    description:  expand the map control to accomadate NewNumAllPages  data pages
         
                  - Before the map is expanded all map sections are locked exclusively

    arguments:    taskId         [in] task id
                  newNumAllPages [in] after this command succeded the map control 
                                      should manage 
                                      the minimum number of NewNumAllPages pages 

    return value: (Converter_ReturnCode)
                   Converter_Ok                     : Expand finished successfully
                   Converter_MemoryAllocationFailed : Expand failed due to a lack of memory
    --------------------------------------------------------------------------*/

    Converter_ReturnCode  Expand(
        const tsp00_TaskId   taskId,
        const SAPDB_UInt     newNumAllPages );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Manager                                                   */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: PageNo management                                            */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     RequestNewPageNo()
    description:  Reserves a new page number.
         
                  - The returned page number is marked as used.

    arguments:    taskId [in]   task id
                  pageNo [out]  requested new page number     
    return value: (SAPDB_Bool) returns true if  a page could be returned if the 
                               data base is already full and no page could be 
                               supplied false is returned
    --------------------------------------------------------------------------*/

    SAPDB_Bool  RequestNewPageNo(
        const tsp00_TaskId  taskId,
        Data_PageNo         &pageNo );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: PageNo management                                         */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Common                                                       */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     MaxPageNo()
    description:  Returns the maximum possible dynamic page number
    arguments:    none
    return value: (Data_PageNo) maximum possible dynamic page number
    --------------------------------------------------------------------------*/

    Data_PageNo MaxPageNo() const{
        return m_Map.GetHighestPossiblePageNo();
    }

    /*!-------------------------------------------------------------------------
    function:     GetRecoveryMode()
    description:  Returns the recovery mode of the data pages
                  managed by this control strucure.
    arguments:    none
    return value: (Data_PageRecoveryMode&)
    --------------------------------------------------------------------------*/

    const Data_PageRecoveryMode& GetRecoveryMode() const{
        return( m_RecoveryMode );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Common                                                    */
    /* -----------------------------------------------------------------------*/
    /* -----------------------------------------------------------------------*/
    /*! chapter: Dump                                                         */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Dump()
    description:  Dumps MapControl and the Map into the kernel dump file

    arguments:    taskId       [in] identification of the calling task
                  dump     [in/out] kernel dump file
                  bEnterRegion [in] use the converter regions
    return value: none
    --------------------------------------------------------------------------*/

    void Dump(
        const tsp00_TaskId  taskId,
        Kernel_Dump         &dump,
        const SAPDB_Bool    bEnterRegion );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Dump                                                      */
    /* -----------------------------------------------------------------------*/

protected:

    // Constructor

    Converter_MapControl(
        SAPDBMem_IRawAllocator        &allocator,
        const Data_PageRecoveryMode   &recMode )
    :
    m_NextSectionToSearchForNewPageNo( 0 ),
    m_RecoveryMode( recMode ),
    m_Map( allocator ),
    m_SectionData( allocator )
    {}

    // Determine section id for given page no

    SAPDB_UInt  CalculateSectionNo( const Data_PageNo pageNo ) const{
        return( m_Map.CalcSectionNo( pageNo ));
    }

    // Return sectionId of the section pointed by the given iterator.
    // It's assumed that the iterator is valid.

    SAPDB_UInt  GetSectionNo( const SectionDataIterator &iter ) const{
        return( m_SectionData.GetIndex( iter ));
    }

    // Number of used sections, i.e. number of configured converter latches.

    SAPDB_UInt GetSectionCount() const{
        return( m_SectionData.GetSize());
    }

protected:

    // Defines the sections number which is used next to get a free page no.

    SAPDB_UInt  m_NextSectionToSearchForNewPageNo;

    // Defines the recovery mode of this map control. Generally needed
    // for trace and diagnostic output.

    Data_PageRecoveryMode   m_RecoveryMode;

    /// handle for the converter map which contains the converter leaf pages

    Converter_Map   m_Map;

    // list of section item containing the section specific informations

    SectionData m_SectionData;

};

/*!*****************************************************************************

   endclass: Converter_MapControl

*******************************************************************************/


#endif  /* CONVERTER_MAPCONTROL_HPP */
