  /*!***************************************************************************

  module:       Converter_MapEntry.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       TillL

  special area: Converter

  description:  Entry of converter map
                
  see also:     

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

  *****************************************************************************/


#ifndef CONVERTER_MAPENTRY_HPP
#define CONVERTER_MAPENTRY_HPP


#include "Converter/Converter_Chain.hpp"
#include "Converter/Converter_Entry.hpp"
#include "Converter/Converter_LeafPage.hpp"
#include "Converter/Converter_PageFrameAllocator.hpp"


/*!*****************************************************************************

   class: Converter_MapEntry

   description: Entry of converter map

*******************************************************************************/
class Converter_MapEntry : 
public Converter_ChainNode <Converter_MapEntryNo>, 
public Converter_Entry     <Converter_LeafPage>
{
public:

    /*!-------------------------------------------------------------------------
    declaration : LeafPage
    description:  a converter leaf page
    --------------------------------------------------------------------------*/
    typedef Converter_LeafPage                  LeafPage;

    /*!-------------------------------------------------------------------------
    function:     Converter_MapEntry()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Converter_MapEntry()
        :
        m_PageIOPending (false)
    {}

    /*!-------------------------------------------------------------------------
    function:     AssignPage()
    description:  Assigns page handler
    arguments:    page           [in]  page handler
                  freeentrycount [in]  number of free entries on the page
    return value: none
    --------------------------------------------------------------------------*/
    void AssignPage
        (LeafPage&                      page,
         PageEntryCount                 freeentrycount)
    {
        Converter_Entry<LeafPage>::AssignPage(page, freeentrycount);
        m_PageIOPending = false;
    }

    /*!-------------------------------------------------------------------------
    function:     ReleasePageFrame()
    description:  Releases page frame from page handler
    arguments:    framealloc [in] page frame allocator for dallocation
    return value: none
    --------------------------------------------------------------------------*/
    void ReleasePageFrame
        (Converter_PageFrameAllocator  &framealloc)
    {
        Converter_Entry<LeafPage>::ReleasePageFrame(framealloc);
        m_PageIOPending = false;
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageIOPending()
    description:  Sets page IO pending state
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void SetPageIOPending()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        m_PageIOPending = true;
    }

    /*!-------------------------------------------------------------------------
    function:     ClearPageIOPending()
    description:  Clears page IO pending state
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void ClearPageIOPending()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        m_PageIOPending = false;
    }

    /*!-------------------------------------------------------------------------
    function:     PageIOPending()
    description:  Checks whether page IO is pending
    arguments:    none
    return value: (SAPDB_Bool) true if pending
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIOPending() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_PageIOPending;
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageRecoveryMode()
    description:  Writes recovery mode from associated leaf page 
    arguments:    rec [in]   page recovery mode
    return value: none
    --------------------------------------------------------------------------*/
    void SetPageRecoveryMode
        (const Data_PageRecoveryMode&   rec)
    {
        m_Page.WritePageRecoveryMode(rec);
    }

    /*!-------------------------------------------------------------------------
    function:     GetPageRecoveryMode()
    description:  Reads recovery mode from associated leaf page 
    arguments:    none
    return value: (Data_PageRecoveryMode) page recovery mode
    --------------------------------------------------------------------------*/
    Data_PageRecoveryMode GetPageRecoveryMode() const
    {
        return m_Page.ReadPageRecoveryMode();
    }

    /*!-------------------------------------------------------------------------
    function:     PageIsRecoverable()
    description:  Checks whether associated page is recoverable
    arguments:    none
    return value: (SAPDB_Bool) true if recoverable
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIsRecoverable() const
    {
        return m_Page.ReadPageRecoveryMode().PageIsRecoverable();
    }

private:

    /*--------------------------------------------------------------------------
    declaration : m_PageIOPending
    description:  indicates whether associated page is currently in IO state
    --------------------------------------------------------------------------*/
    SAPDB_Bool                      m_PageIOPending;
};
/*!*****************************************************************************

   endclass: Converter_MapEntry

*******************************************************************************/





#endif // CONVERTER_PAGEIDMAPENTRY_HPP
