/*!*****************************************************************************

  module:       Converter_Page.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       TillL

  special area: Converter

  description:  Base class for converter page handlers
                
  see also:     Kernel_IPage.html 
                Kernel_Page.html
                Kernel_Page74.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef CONVERTER_PAGE_HPP
#define CONVERTER_PAGE_HPP

#include "DataAccess/Data_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "KernelCommon/Kernel_Page74.hpp"
#include "Converter/Converter_Types.hpp"
#include "Converter/Converter_ParentReference.hpp"
#include "Converter/Converter_FlagInfo.hpp"


class Converter_PageEntryIterator;
class Converter_PageEntryRepairIterator;


/*!*****************************************************************************

   class: Converter_Page

   description: 
   
   Virtual base class for converter page handlers

   Defines the following pure virtual methods:

   - GetEntryCount()
   - InUseFlagsOffset()
   - InitializeFlagInfo()

   Call static method 'SetDevNoBitCount()' once prior to first usage
   of this class.

*******************************************************************************/
class Converter_Page : public Kernel_Page74
{
    friend class Converter_PageEntryIterator;
    friend class Converter_PageEntryRepairIterator;
    friend class Converter_Map;

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : PageEntryCount
    description:  any number of converter page entries
    --------------------------------------------------------------------------*/
    typedef Converter_PageEntryCount            PageEntryCount;
    /*!-------------------------------------------------------------------------
    declaration : PageEntryNo
    description:  index of a certain page entry
    --------------------------------------------------------------------------*/
    typedef Converter_PageEntryNo               PageEntryNo;
    /*!-------------------------------------------------------------------------
    declaration : IndexEntryCount
    description:  any number of converter index entries
    --------------------------------------------------------------------------*/
    typedef Converter_IndexEntryCount           IndexEntryCount;
    /*!-------------------------------------------------------------------------
    declaration : IndexEntryNo
    description:  index of a certain converter index entry
    --------------------------------------------------------------------------*/
    typedef Converter_IndexEntryNo              IndexEntryNo;
    /*!-------------------------------------------------------------------------
    declaration : Version
    description:  converter version
    --------------------------------------------------------------------------*/
    typedef Converter_Version                   Version;
    /*!-------------------------------------------------------------------------
    declaration : BlockAddress
    description:  block i/o address
    --------------------------------------------------------------------------*/
    typedef IOMan_BlockAddress                  BlockAddress;
    /*!-------------------------------------------------------------------------
    declaration : BitCount
    description:  device number bit count
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt1                         BitCount;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/
protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Protected constructors and initialization                    */
    /*!-------------------------------------------------------------------------
    function:     Converter_Page()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Page();

    /*!-------------------------------------------------------------------------
    function:     Converter_Page()
    description:  Constructor; assigns frame

                  - Does not change frame contents.
    arguments:    frame [in]    page frame
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Page
    (const PageFrame&               frame);

    /*!-------------------------------------------------------------------------
    function:     Converter_Page()
    description:  Constructor; assigns frame and initializes frame contents
    arguments:    frame         [in] page frame
                  version       [in] converter version
                  parent        [in] reference to parent in converter tree parent 
                                     index
    return value: none
    --------------------------------------------------------------------------*/
    Converter_Page
    (const PageFrame&               frame,
     const Version&                 version,
     const Converter_ParentReference&         parent);

    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame   [in]   page frame
                  version [in]   converter version
                  parent  [in]   reference to parent in converter tree parent 
                                 index
    return value: none
    --------------------------------------------------------------------------*/
    void SetAndInitializeFrame
    (const PageFrame&               frame,
     const Version&                 version,
     const Converter_ParentReference&         parent);

    /*!-------------------------------------------------------------------------
    public function: InitializeFrame()
    description:  Initializes frame contents
    arguments:    version [in]   converter version
                  parent  [in]   reference to parent in converter tree parent 
                                 index
    return value: none
    --------------------------------------------------------------------------*/
    void InitializeFrame
    (const Version&                 version,
     const Converter_ParentReference&         parent);
public:

    /*!-------------------------------------------------------------------------
    public function: SetDevNoBitCount()
    description:  Initializes static member 'device number bit count'
    arguments:    devnobitcount [in] number of bits used for persistent device
                                     number representation
    return value: none
    --------------------------------------------------------------------------*/
    static void SetDevNoBitCount
    (BitCount                       devnobitcount);

    /*!-------------------------------------------------------------------------
    function:     ReadConverterPageType()
    description:  Reads converter page type from frame
    arguments:    frame [in] page frame
    return value: (Converter_PageType) page type
    --------------------------------------------------------------------------*/
    static Converter_PageType ReadConverterPageType(
        const PageFrame   &frame)
    {
        return PersistentPageType::ConvertPersistentToInternal
               (reinterpret_cast<const PersistentHeader*>
                (frame.GetPtrAt(0, sizeof(PersistentHeader)))->m_Info);
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /*!-------------------------------------------------------------------------
    function:     WriteVersion()
    description:  Writes converter version to frame
    arguments:    version [in]  converter version
    return value: none
    --------------------------------------------------------------------------*/
    void WriteVersion
    (const Version&                 version);

    /*!-------------------------------------------------------------------------
    function:     ReadVersion()
    description:  Reads converter version from frame
    arguments:    none
    return value: (Version) converter version
    --------------------------------------------------------------------------*/
    Version ReadVersion() const;

    /*!-------------------------------------------------------------------------
    function:     WriteBlockAddress()
    description:  Writes block i/o address of the current converter page to frame
    arguments:    address [in]       block i/o address
    return value: none
    --------------------------------------------------------------------------*/
    void WriteBlockAddress
    (const BlockAddress&            address);

    /*!-------------------------------------------------------------------------
    function:     ReadBlockAddress()
    description:  Reads block i/o address of the current converter page from frame
    arguments:    none
    return value: (BlockAddress) block address in internal format
    --------------------------------------------------------------------------*/
    BlockAddress ReadBlockAddress() const;

    /*!-------------------------------------------------------------------------
    function:     ReadConverterPageType()
    description:  Reads converter page type from frame
    arguments:    none
    return value: (Converter_PageType) page type
    --------------------------------------------------------------------------*/
    Converter_PageType ReadConverterPageType() const;

    /*!-------------------------------------------------------------------------
    function:     ReadDeviceNoBitCount()
    description:  Reads the count of bits used for storage
                  of logical device numbers from frame
    arguments:    none
    return value: (BitCount) bit count
    --------------------------------------------------------------------------*/
    BitCount ReadDeviceNoBitCount() const;

    /*!-------------------------------------------------------------------------
    function:     WriteParentReference()
    description:  Writes reference to parent in converter parent index to frame

    arguments     parent [in]   reference to parent
    return value: none
    --------------------------------------------------------------------------*/
    void WriteParentReference
    (const Converter_ParentReference&         parent);

    /*!-------------------------------------------------------------------------
    function:     ReadParentReference()
    description:  Reads reference to parent in converter parent index from frame

    arguments:    none
    return value: (Converter_ParentReference) reference to parent
    --------------------------------------------------------------------------*/
    Converter_ParentReference ReadParentReference() const;

    /*!-------------------------------------------------------------------------
    function:     HasFreeEntryAvailable()
    description:  Checks whether there is at least one free entry on the page
    arguments:    none
    return value: (SAPDB_Bool) true if free entry available
    --------------------------------------------------------------------------*/
    SAPDB_Bool HasFreeEntryAvailable() const;

    /*!-------------------------------------------------------------------------
    function:     GetEntryCount()
    description:  Returns total count of entries on converter page
    arguments:    none
    return value: (PageEntryCount) entry count
    --------------------------------------------------------------------------*/
    virtual PageEntryCount GetEntryCount() const = 0;

    /*!-------------------------------------------------------------------------
    function:     FlagInfoPtr()
    description:  Returns pointer to block on frame where flag info on specified
                  entry is stored
    arguments:    flagblockoffs [in]  additional page offset 
                                      (sizeof(FlagInfo) aligned)
                  entryno       [in]  entry number
    return value: ((const)FlagInfo*)
    prototypes:   2
    --------------------------------------------------------------------------*/
    Converter_FlagInfo GetFlagInfo
    (OffsetType                     flagblockoffs,
     const PageEntryNo&             entryno);

    Converter_FlagInfo GetFlagInfo
    (OffsetType                     flagblockoffs,
     const PageEntryNo&             entryno) const;

    /*!-------------------------------------------------------------------------
    function:     DetermineFreeEntryCount()
    description:  Determines total count of free entries on converter page
    arguments:    none
    return value: (PageEntryCount) entry count
    --------------------------------------------------------------------------*/
    PageEntryCount DetermineFreeEntryCount() const;

    /*!-------------------------------------------------------------------------
    function:     ReintegrateReservedEntries()
    description:  Scans whole page for reserved entries and rechains them into
                  the free entry chain

                  - This routine is useful when recovering a database crash.
                  - There may be entries on a page which have been reserved
                    and marked as used, but are in fact unused. These must be
                    rechained and marked as unused.
    arguments:    none
    return value: (PageEntryCount) number of reserved page entries found
    --------------------------------------------------------------------------*/
    PageEntryCount ReintegrateReservedEntries();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Kernel_IPage interface                                       */
    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       (SAPDB_Bool) true if verify successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Kernel_IPage interface                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterators                                                    */
    /*!-------------------------------------------------------------------------
    function:     Begin()
    description:  Returns iterator at first entry on page
    arguments:    none
    return value: (Converter_PageEntryIterator)
    --------------------------------------------------------------------------*/
    Converter_PageEntryIterator Begin();

    /*!-------------------------------------------------------------------------
    function:     End()
    description:  Returns iterator behind last entry on page
    arguments:    none
    return value: (Converter_PageEntryIterator)
    --------------------------------------------------------------------------*/
    Converter_PageEntryIterator End();

    /*!-------------------------------------------------------------------------
    function:     RepairBegin()
    description:  Returns repair iterator at first entry on page
    arguments:    none
    return value: (Converter_PageEntryRepairIterator)
    --------------------------------------------------------------------------*/
    Converter_PageEntryRepairIterator RepairBegin();

    /*!-------------------------------------------------------------------------
    function:     RepairEnd()
    description:  Returns repair iterator behind last entry on page
    arguments:    none
    return value: (Converter_PageEntryRepairIterator)
    --------------------------------------------------------------------------*/
    Converter_PageEntryRepairIterator RepairEnd();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterators                                                 */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Entry access                                                 */
    /*!-------------------------------------------------------------------------
    function:     WriteBlockAddress()
    description:  Writes to entry on frame: i/o block address
    arguments:    entryno   [in]  entry position
                  addr      [in]  i/o block address to be written
                  oldaddr   [out] old i/o address found at the entry
    return value: none
    --------------------------------------------------------------------------*/
    virtual void WriteBlockAddress
    (const PageEntryNo&             entryno,
     const BlockAddress&            addr,
     BlockAddress&                  oldaddr);

    /*!-------------------------------------------------------------------------
    function:     ReadBlockAddress()
    description:  Reads from entry on frame: i/o block address
    arguments:    entryno  [in]  entry position
                  bladdr   [out] requested i/o address
    return value: (BlockAddress) requested i/o address
    --------------------------------------------------------------------------*/
    BlockAddress ReadBlockAddress
    (const PageEntryNo&             entryno,
     BlockAddress&                  bladdr);

    /*!-------------------------------------------------------------------------
    function:     ReserveEntry()
    description:  Reserves a free page entry and returns its entry number
    arguments:    none
    return value: (PageEntryNo) entry number
    --------------------------------------------------------------------------*/
    PageEntryNo ReserveEntry();

    /*!-------------------------------------------------------------------------
    function:     ReserveSpecifiedEntry()
    description:  Reserves a specified page entry and returns its entry number

                  - Entry 'entryno' must be unused (free).
    arguments:    entryno    [in]  entry position
    return value: (PageEntryNo) entry number
    --------------------------------------------------------------------------*/
    PageEntryNo ReserveSpecifiedEntry
    (PageEntryNo                    entryno);

    /*!-------------------------------------------------------------------------
    function:     ReleaseEntry()
    description:  Chains in free page entry
    arguments:    entryno    [in]  entry number
    return value: none
    --------------------------------------------------------------------------*/
    void ReleaseEntry
    (PageEntryNo                    entryno);

    /*!-------------------------------------------------------------------------
    function:     EntryInUse()
    description:  Checks whether specified entry is in use
    arguments:    entryno    [in]  entry number
    return value: (SAPDB_Bool) true if entry is in use
    --------------------------------------------------------------------------*/
    SAPDB_Bool EntryInUse
    (const PageEntryNo&             entryno) const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Entry access                                              */
    /* -----------------------------------------------------------------------*/

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#else
private:
#endif

    /*--------------------------------------------------------------------------
    declaration: Persistent types
    description: Persistent type definitions with compile time size check
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent         < Version,                  MultiUse1, 4 >           PersistentVersion;
    typedef SAPDB_Persistent         < IOMan_PackedBlockAddress, SAPDB_UInt4, 4 >         PersistentBlockAddress;
    typedef SAPDB_Persistent         < PageEntryNo,              SAPDB_UInt2, 2 >         PersistentPageEntryNo;
    typedef SAPDB_Persistent         < IndexEntryNo,             SAPDB_UInt2, 2 >         PersistentIndexEntryNo;
    typedef SAPDB_PersistentBitRange < Converter_PageType,       SAPDB_UInt1, 1, 0, 1 >   PersistentPageType;
    typedef SAPDB_PersistentBitRange < SAPDB_UInt1,              SAPDB_UInt1, 1, 1, 4 >   PersistentDevNoBitCount;


    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within kernel page header
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):
    // PersistentVersion reuses MultiUse1 in Kernel_Page74::PersistentHeader
    typedef Kernel_Page74::PersistentHeader                 PH1;
    typedef PersistentBlockAddress::PersistentFormat        PH2;
    typedef SAPDB_UInt4                                     PH3;
    typedef PersistentPageEntryNo::PersistentFormat         PH4;
    typedef SAPDB_UInt1                                     PH5;
    typedef SAPDB_UInt1                                     PH6;
    typedef PersistentIndexEntryNo::PersistentFormat        PH7;
    typedef PersistentPageEntryNo::PersistentFormat         PH8;

protected:

    struct PersistentHeader
    {
        PH1     m_KernelPageHeader;
        PH2     m_BlockAddress;
        PH3     m_Unused1;
        PH4     m_FreeEntryAnchorNo;
        PH5     m_Info;
        PH6     m_Unused2;
        PH7     m_ParentIndexEntryNo;
        PH8     m_ParentPageEntryNo;
    };

    /*--------------------------------------------------------------------------
    declaration: ConverterPagePersistentHeader
    description: Compile time size check for persistent header
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent
    < PersistentHeader,
    PersistentHeader,
    sizeof(PH1) +
    sizeof(PH2) +
    sizeof(PH3) +
    sizeof(PH4) +
    sizeof(PH5) +
    sizeof(PH6) +
    sizeof(PH7) +
    sizeof(PH8) >
    CheckedPersistentHeader;

private:

    /*--------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    return value: (PersistentHeader*) header pointer
    prototypes:   2
    --------------------------------------------------------------------------*/
    PersistentHeader* PersistentHeaderPtr();

    const PersistentHeader* PersistentHeaderPtr() const;

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access (protected)                                     */
    /*!-------------------------------------------------------------------------
    declaration: PersistentEntry
    description: Internal union type for handling of entries.

                 The member m_BlockAddress is unused but required for correct 
                 entry alignment on the page.
    --------------------------------------------------------------------------*/
    union PersistentEntry
    {
        struct
        {
            PersistentPageEntryNo::PersistentFormat m_Predecessor;
            PersistentPageEntryNo::PersistentFormat m_Successor;
        }
        m_Free;                                 // used with free entry chain
        PersistentBlockAddress::PersistentFormat
        m_BlockAddress;                         // used with block addresses
    };

    /*!-------------------------------------------------------------------------
    function:     PersistentEntryPtr()
    description:  Returns pointer to page entry on frame
    arguments:    entryno [in]    entry position
    return value: (PersistentEntry*) header pointer
    prototypes:   2
    --------------------------------------------------------------------------*/
    PersistentEntry* PersistentEntryPtr
    (PageEntryNo                        entryno);

    const PersistentEntry* PersistentEntryPtr
    (PageEntryNo                        entryno) const;


private:
    /*--------------------------------------------------------------------------
    declaration: ConverterPagePersistentEntry
    description: Compile time size check for persistent entry
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent
    < PersistentEntry,
    PersistentEntry,
    sizeof(PersistentBlockAddress::PersistentFormat)
    >
    ConverterPagePersistentEntry;

protected:

    /*!-------------------------------------------------------------------------
    function:     GetEntrySize()
    description:  Returns size of page entry in bytes
    arguments:    none
    return value: (SizeType) entry size
    --------------------------------------------------------------------------*/
    static SizeType GetEntrySize();

    /*!-------------------------------------------------------------------------
       function:     GetHeaderSize()
       description:  Returns size of page header in bytes
       arguments:    none
       return value: (SizeType) header size
       --------------------------------------------------------------------------*/
    virtual SizeType GetHeaderSize() const;

    /*!-------------------------------------------------------------------------
    function:     WriteConverterPageType()
    description:  Writes converter page type to frame
    arguments:    type [in]   converter page type
    return value: none
    --------------------------------------------------------------------------*/
    void WriteConverterPageType
    (const Converter_PageType&      type);

    /*!-------------------------------------------------------------------------
    function:     InUseFlagsOffset()
    description:  Returns offset on page where entire in-use flags block begins
    arguments:    none
    return value: (OffsetType) offset
    --------------------------------------------------------------------------*/
    virtual OffsetType InUseFlagsOffset() const = 0;

    /*!-------------------------------------------------------------------------
    function:     InitializeFlagInfo()
    description:  Initializes all entry flags on the page with 0's.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    virtual void InitializeFlagInfo() = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access (protected)                                  */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Entry access (protected)                                     */
    /*!-------------------------------------------------------------------------
    function:     Reserve()
    description:  Performs additional tasks when reserving entries

                  - Implementation detail
    arguments:    entryno    [in]  entry position
    return value: none
    --------------------------------------------------------------------------*/
    virtual void Reserve
    (const PageEntryNo&             entryno);

    /*!-------------------------------------------------------------------------
    function:     Release()
    description:  Performs additional tasks when releasing entries

                  - Implementation detail
    arguments:    entryno    [in]  entry position
    return value: none
    --------------------------------------------------------------------------*/
    virtual void Release
    (const PageEntryNo&             entryno);

    /*!-------------------------------------------------------------------------
    function:     SetBlockAddressAsReserved()
    description:  Sets block address of specified entry to "reserved"

    arguments:    entryno    [in]  entry position
    return value: none
    --------------------------------------------------------------------------*/
    void SetBlockAddressAsReserved
    (const PageEntryNo&             entryno);

    /*!-------------------------------------------------------------------------
    function:     SetBlockAddressAsInvalid()
    description:  Sets block address of specified entry to "invalid"

    arguments:    entryno    [in]  entry position
    return value: none
    --------------------------------------------------------------------------*/
    void SetBlockAddressAsInvalid
    (const PageEntryNo&             entryno);

    /*!-------------------------------------------------------------------------
    function:     SetEntryInUse()
    description:  Sets the "InUse"-flag of the specified entry
    arguments:    entryno [in]  entry no.
    return value: none
    --------------------------------------------------------------------------*/
    void SetEntryInUse
    (const PageEntryNo&             entryno);

    /*!-------------------------------------------------------------------------
    function:     ClearEntryInUse()
    description:  Clears the "InUse"-flag of the specified entry
    arguments:    entryno [in]  entry no.
    return value: none
    --------------------------------------------------------------------------*/
    virtual void ClearEntryInUse
    (const PageEntryNo&             entryno);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Entry access (protected)                                  */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     BuildFreeEntryChain()
    description:  Constructs the chain of free entries on a "fresh" converter page
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void BuildFreeEntryChain();

    /*--------------------------------------------------------------------------
    function:     ConvertUInt4ToBlockAddress()
    description:  Converts uint4 to block io address (persistent format)
    arguments:    addr [in]          block io address
                  devnobitcount [in] number of bits used for persistent device
                                     number representation
    return value: (BlockAddress) block io address
    --------------------------------------------------------------------------*/
    static BlockAddress ConvertUInt4ToBlockAddress
    (SAPDB_UInt4                    persaddr,
     SAPDB_UInt                     devnobitcount);

    /*--------------------------------------------------------------------------
    function:     WriteDeviceNoBitCount()
    description:  Writes number of bits used for device number to frame

    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void WriteDeviceNoBitCount();

    /*--------------------------------------------------------------------------
    function:     WriteFreeEntryAnchor()
    description:  Writes entry position of free entry anchor to frame
    arguments:    entryno [in]   entry position
    return value: none
    --------------------------------------------------------------------------*/
    void WriteFreeEntryAnchor
    (const PageEntryNo&                 entryno);

    /*--------------------------------------------------------------------------
    function:     ReadFreeEntryAnchor()
    description:  Reads entry position of free entry anchor from frame
    arguments:    none
    return value: (PageEntryNo) entry position
    --------------------------------------------------------------------------*/
    PageEntryNo ReadFreeEntryAnchor() const;

    /*--------------------------------------------------------------------------
    declaration : DevNoBitCount
    description:  number of bits used for persistent representation of 
                  device numbers
    --------------------------------------------------------------------------*/
    static BitCount DevNoBitCount;

};
/*!*****************************************************************************

   endclass: Converter_Page

*******************************************************************************/



/*!*****************************************************************************

   class: Converter_PageEntryIterator

   description: 
   
   Steps over all entries on a page 
   
   This iterator makes no difference between used and unused entries

*******************************************************************************/

class Converter_PageEntryIterator
{
public:

    /*!-------------------------------------------------------------------------
    function:     Converter_PageEntryIterator()
    description:  Constructor
    arguments:    page [in] reference to converter page
                  pos  [in] entry number defining start position on the page
                            (default: 0)
    return value: none
    --------------------------------------------------------------------------*/
    Converter_PageEntryIterator
    (Converter_Page&                page,
     Converter_PageEntryNo          pos   = 0)
    :
    m_pPage(&page),
    m_Pos(pos),
    m_BitCount(page.ReadDeviceNoBitCount())
    {}

    /*!-------------------------------------------------------------------------
    function:     operator=()
    description:  Assigns other iterator
    arguments:    iter [in]    iterator to be assigned
    return value: none
    --------------------------------------------------------------------------*/
    Converter_PageEntryIterator& operator=
    (const Converter_PageEntryIterator& iter)
    {
        //if (iter != *this) // PTS 1113242 TS 2001-12-20
        //{
        m_pPage          = iter.m_pPage;
        m_Pos            = iter.m_Pos;
        m_BitCount       = iter.m_BitCount;
        //}
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     operator!=()
    description:  Compares with other iterator
    arguments:    iter [in]    other
    return value: (SAPDB_Bool) true if unequal
    --------------------------------------------------------------------------*/
    SAPDB_Bool operator!=(const Converter_PageEntryIterator &iter) const
    {
        SAPDBERR_ASSERT_ARGUMENT(m_pPage == iter.m_pPage);

        return m_Pos != iter.m_Pos;
    }

    /*!-------------------------------------------------------------------------
    function:     operator==()
    description:  Compares with other iterator
    arguments:    iter [in]    other
    return value: (SAPDB_Bool) true if equal
    --------------------------------------------------------------------------*/
    SAPDB_Bool operator==(const Converter_PageEntryIterator &iter) const
    {
        SAPDBERR_ASSERT_ARGUMENT(m_pPage == iter.m_pPage);

        return m_Pos == iter.m_Pos;
    }


    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  Prefix increment; advances iterator
    arguments:    none
    return value: (Converter_PageEntryIterator&) this instance
    --------------------------------------------------------------------------*/
    Converter_PageEntryIterator& operator++()
    {
        ++m_Pos;
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     GetPackedBlockAddress()
    description:  Returns the packed block address the iterator points to
    arguments:    none
    return value: (IOMan_PackedBlockAddress) actual block address in packed format
    --------------------------------------------------------------------------*/
    IOMan_PackedBlockAddress GetPackedBlockAddress() const
    {
        return( (*m_pPage).PersistentEntryPtr(m_Pos)->m_BlockAddress );
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockAddress()
    description:  Returns the block address the iterator points to
    arguments:    none
    return value: (IOMan_BlockAddress) actual block address
    --------------------------------------------------------------------------*/
    IOMan_BlockAddress GetBlockAddress() const
    {
        return( Unpack((*m_pPage).PersistentEntryPtr(m_Pos)->m_BlockAddress,
                      (SAPDB_UInt)m_BitCount));
    }

    /*!-------------------------------------------------------------------------
    function:     EntryInUse()
    description:  Checks whether actual entry is in use
    arguments:    none
    return value: (SAPDB_Bool) true if entry is in use
    --------------------------------------------------------------------------*/
    SAPDB_Bool EntryInUse() const
    {
        return( (*m_pPage).EntryInUse(m_Pos));
    }

    /*!-------------------------------------------------------------------------
    function:     GetPosition()
    description:  Returns actual entry number iter points to
    arguments:    none
    return value: (Converter_PageEntryNo) actual entry number
    --------------------------------------------------------------------------*/
    Converter_PageEntryNo GetPosition() const
    {
        return m_Pos;
    }

protected:
    Converter_Page             *m_pPage; // PTS 1113242 TS 2001-12-21
    Converter_PageEntryNo       m_Pos;
    SAPDB_UInt1                 m_BitCount;
};
/*!*****************************************************************************

   endclass: Converter_PageEntryIterator

*******************************************************************************/


// #if defined(SUN)

template<>
inline SAPDB_UInt1 SAPDB_Persistent
<Converter_PageType, SAPDB_UInt1, 1>::ConvertInternalToPersistent
(const Converter_PageType&   type)
{
    return type;
}

// -----------------------------------------------------------------------------

template<>
inline SAPDB_UInt1 SAPDB_PersistentBitRange
<Converter_PageType, SAPDB_UInt1, 1, 0, 1>::ConvertInternalToPersistent
(const Converter_PageType&   type)
{
    return SAPDB_Persistent<Converter_PageType, SAPDB_UInt1, 1>
           ::ConvertInternalToPersistent(type);
}

// #endif

//------------------------------------------------------------------------------
//  Converter_Page
//------------------------------------------------------------------------------

inline Converter_Page::Converter_Page()
{}

//------------------------------------------------------------------------------

inline Converter_Page::Converter_Page
(const PageFrame&               frame)
        :
        Kernel_Page74(frame)
{}

//------------------------------------------------------------------------------

inline Converter_Page::Converter_Page
(const PageFrame&               frame,
 const Version&                 version,
 const Converter_ParentReference&         parent)
        :
        Kernel_Page74(frame)
{
    InitializeFrame(version, parent);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::SetAndInitializeFrame
(const PageFrame&               frame,
 const Version&                 version,
 const Converter_ParentReference&         parent)
{
    SetFrame(frame);
    InitializeFrame(version, parent);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::InitializeFrame
(const Version&                 version,
 const Converter_ParentReference&         parent)
{
    Kernel_Page74::InitializeFrame(Kernel_ConverterPageType);
    WriteVersion(version);
    WriteBlockAddress(BlockAddress());
    PersistentHeaderPtr()->m_Unused1 = 0;  // PTS 1125335 TS 2003-12-03
    WriteDeviceNoBitCount();
    PersistentHeaderPtr()->m_Unused2 = 0; // PTS 1125335 TS 2003-12-03
    WriteParentReference(parent);
    BuildFreeEntryChain();

    SAPDBERR_ASSERT_ARGUMENT (Converter_Page::Verify());
}

//------------------------------------------------------------------------------

inline void
Converter_Page::SetDevNoBitCount
(BitCount                       devnobitcount)
{
    DevNoBitCount = devnobitcount;
}

//------------------------------------------------------------------------------

inline void
Converter_Page::WriteVersion
(const Version&                 version)
{
    WriteMultiUse1(version);
}

//------------------------------------------------------------------------------

inline Converter_Page::Version
Converter_Page::ReadVersion() const
{
    return ReadMultiUse1();
}

//------------------------------------------------------------------------------

inline void
Converter_Page::WriteBlockAddress
(const BlockAddress&            address)
{
    PersistentHeaderPtr()->m_BlockAddress =
        PersistentBlockAddress::PersistentFormat(Pack(address, DevNoBitCount));
}

//------------------------------------------------------------------------------

inline Converter_Page::BlockAddress
Converter_Page::ReadBlockAddress() const
{
    return Unpack(PersistentHeaderPtr()->m_BlockAddress, DevNoBitCount);
}

//------------------------------------------------------------------------------

inline Converter_PageType
Converter_Page::ReadConverterPageType() const
{
    return PersistentPageType::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_Info);
}

//------------------------------------------------------------------------------

inline Converter_Page::BitCount
Converter_Page::ReadDeviceNoBitCount() const
{
    return PersistentDevNoBitCount::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_Info);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::WriteParentReference
(const Converter_ParentReference&         parent)
{
    PersistentHeaderPtr()->m_ParentIndexEntryNo =
        PersistentIndexEntryNo::ConvertInternalToPersistent(parent.GetIndexEntryNo());
    PersistentHeaderPtr()->m_ParentPageEntryNo  =
        PersistentPageEntryNo::ConvertInternalToPersistent(parent.GetPageEntryNo());
}

//------------------------------------------------------------------------------

inline Converter_ParentReference
Converter_Page::ReadParentReference() const
{
    return Converter_ParentReference
           (PersistentIndexEntryNo::ConvertPersistentToInternal
            (PersistentHeaderPtr()->m_ParentIndexEntryNo),
            PersistentPageEntryNo::ConvertPersistentToInternal
            (PersistentHeaderPtr()->m_ParentPageEntryNo));
}

//------------------------------------------------------------------------------

inline SAPDB_Bool
Converter_Page::HasFreeEntryAvailable() const
{
    return ReadFreeEntryAnchor().IsValid();
}

//------------------------------------------------------------------------------

inline void
Converter_Page::WriteBlockAddress
(const PageEntryNo&             entryno,
 const BlockAddress&            addr,
 BlockAddress&                  oldaddr)
{
    PersistentEntry* pentry = PersistentEntryPtr(entryno);

    if (IOMan_PackedBlockAddress (pentry->m_BlockAddress).IsReserved())
        oldaddr.Invalidate();
    else
        oldaddr = Unpack(pentry->m_BlockAddress, DevNoBitCount);
    pentry->m_BlockAddress =
        PersistentBlockAddress::PersistentFormat(Pack(addr, DevNoBitCount));
}

//------------------------------------------------------------------------------

inline Converter_Page::BlockAddress
Converter_Page::ReadBlockAddress
(const PageEntryNo&             entryno,
 BlockAddress&                  bladdr)
{
    if (EntryInUse(entryno))
        bladdr = Unpack(PersistentEntryPtr(entryno)->m_BlockAddress, DevNoBitCount);
    else
        bladdr.Invalidate();
    return bladdr;
}

//------------------------------------------------------------------------------

inline Converter_Page::PersistentHeader*
Converter_Page::PersistentHeaderPtr()
{
    return reinterpret_cast<PersistentHeader*>
           (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}

//------------------------------------------------------------------------------

inline const Converter_Page::PersistentHeader*
Converter_Page::PersistentHeaderPtr() const
{
    return reinterpret_cast<const PersistentHeader*>
           (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}

//------------------------------------------------------------------------------

inline Converter_Page::PersistentEntry*
Converter_Page::PersistentEntryPtr
(PageEntryNo                        entryno)
{
    SAPDBERR_ASSERT_ARGUMENT(entryno < GetEntryCount());

    return reinterpret_cast<PersistentEntry*>
           (m_Frame.GetPtrAt
            (GetHeaderSize() + entryno * sizeof(PersistentEntry),
             sizeof(PersistentEntry)));
}

//------------------------------------------------------------------------------

inline const Converter_Page::PersistentEntry*
Converter_Page::PersistentEntryPtr
(PageEntryNo                        entryno) const
{
    SAPDBERR_ASSERT_ARGUMENT(entryno < GetEntryCount());

    return reinterpret_cast<const PersistentEntry*>
           (m_Frame.GetPtrAt
            (GetHeaderSize() + entryno * sizeof(PersistentEntry),
             sizeof(PersistentEntry)));
}

//------------------------------------------------------------------------------

inline Converter_Page::SizeType
Converter_Page::GetEntrySize()
{
    return sizeof(PersistentEntry);
}

//------------------------------------------------------------------------------

inline Converter_Page::SizeType
Converter_Page::GetHeaderSize() const
{
    return sizeof(PersistentHeader);
}

//------------------------------------------------------------------------------


inline void
Converter_Page::WriteConverterPageType
(const Converter_PageType&      type)
{
    PersistentPageType pt(type);
    pt.AssignPersistentValueTo(PersistentHeaderPtr()->m_Info);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::Reserve
(const PageEntryNo&             entryno)
{
    SetBlockAddressAsReserved(entryno);
    SetEntryInUse(entryno);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::Release
(const PageEntryNo&             entryno)
{
    ClearEntryInUse(entryno);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::SetBlockAddressAsReserved
(const PageEntryNo&             entryno)
{
    // set block address on frame to "reserved"
    IOMan_PackedBlockAddress addr;
    addr.SetAsReserved();
    PersistentEntryPtr(entryno)->m_BlockAddress =
        PersistentBlockAddress::ConvertInternalToPersistent(addr);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::SetBlockAddressAsInvalid
(const PageEntryNo&             entryno)
{
    // set block address on frame to "invalid"
    PersistentEntryPtr(entryno)->m_BlockAddress =
        PersistentBlockAddress::ConvertInternalToPersistent
        (IOMan_PackedBlockAddress()); // implicitely within constructor
}

//------------------------------------------------------------------------------

inline SAPDB_Bool
Converter_Page::EntryInUse
(const PageEntryNo&             entryno) const
{
    return	GetFlagInfo(InUseFlagsOffset(), entryno).FlagIsSet(entryno);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::SetEntryInUse
(const PageEntryNo&             entryno)
{
    GetFlagInfo(InUseFlagsOffset(), entryno).SetFlag(entryno);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::ClearEntryInUse
(const PageEntryNo&             entryno)
{
    GetFlagInfo(InUseFlagsOffset(), entryno).DelFlag(entryno);
}

//------------------------------------------------------------------------------

inline Converter_Page::BlockAddress
Converter_Page::ConvertUInt4ToBlockAddress
(SAPDB_UInt4                    persaddr,
 SAPDB_UInt                     devnobitcount)
{
    if (persaddr != SAPDB_MAX_UINT4)
    {
        return BlockAddress(
                   IOMan_DeviceNo
                   (persaddr & ~(SAPDB_MAX_UINT4 << devnobitcount)),
                   IOMan_BlockNo
                   ((persaddr & (SAPDB_MAX_UINT4 << devnobitcount))
                    >> devnobitcount));
    }
    else
        return BlockAddress();
}

//------------------------------------------------------------------------------

inline void
Converter_Page::WriteDeviceNoBitCount()
{
    SAPDBERR_ASSERT_STATE(DevNoBitCount != 0);

    PersistentDevNoBitCount bc(DevNoBitCount);
    bc.AssignPersistentValueTo(PersistentHeaderPtr()->m_Info);
}

//------------------------------------------------------------------------------

inline void
Converter_Page::WriteFreeEntryAnchor
(const PageEntryNo&                 entryno)
{
#ifdef SAPDB_QUICK
    if( entryno.IsValid() )
    {
        SAPDBERR_ASSERT_STATE( entryno < GetEntryCount());
    }
#endif

    PersistentHeaderPtr()->m_FreeEntryAnchorNo =
        PersistentPageEntryNo::ConvertInternalToPersistent(entryno);
}

//------------------------------------------------------------------------------

inline Converter_Page::PageEntryNo
Converter_Page::ReadFreeEntryAnchor() const
{
    return PersistentPageEntryNo::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_FreeEntryAnchorNo);
}

//------------------------------------------------------------------------------

inline Converter_FlagInfo
Converter_Page::GetFlagInfo
(OffsetType                     flagblockoffs,
 const PageEntryNo&             entryno)
{
    SAPDBERR_ASSERT_ARGUMENT(flagblockoffs % sizeof(Converter_FlagInfo::FlagInfoData) == 0);

    return Converter_FlagInfo(reinterpret_cast<Converter_FlagInfo::FlagInfoData*>
                              (m_Frame.GetPtrAt(flagblockoffs, sizeof(Converter_FlagInfo::FlagInfoData))), entryno);
}

//------------------------------------------------------------------------------

inline Converter_FlagInfo
Converter_Page::GetFlagInfo
(OffsetType                     flagblockoffs,
 const PageEntryNo&             entryno) const
{
    SAPDBERR_ASSERT_ARGUMENT(flagblockoffs % sizeof(Converter_FlagInfo::FlagInfoData) == 0);

    return Converter_FlagInfo(reinterpret_cast<Converter_FlagInfo::FlagInfoData*>
                              (const_cast <unsigned char*>(m_Frame.GetPtrAt(flagblockoffs, sizeof(Converter_FlagInfo::FlagInfoData)))), entryno);
}

//------------------------------------------------------------------------------


#endif // CONVERTER_PAGE_HPP

