/*!**************************************************************************

  module      : Converter_PageFrameAllocator.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: Converter_PageFrameAllocator
  description : This module offers the facility to request and release
                memory capable for converter page I/O operations.
                
                The Converter_PageFrameAllocator class is a convenient 
                capsule of the component frame control, which is the only
                one manager for I/O capable memory.

  last changed: 2001-01-23  12:49
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_PAGEFRAMEALLOCATOR_HPP
#define CONVERTER_PAGEFRAMEALLOCATOR_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

#include "KernelCommon/Kernel_IPage.hpp"
#include "KernelCommon/Kernel_Common.hpp"


/*!*****************************************************************************

   class: Converter_PageFrameAllocator
   
   description: The Converter_PageFrameAllocator class is a convenient 
                capsule of the component frame control, which is the only
                one manager for I/O capable memory.

*******************************************************************************/

class Converter_PageFrameAllocator
{
public:

    /*!-------------------------------------------------------------------------
    function:     Converter_PageFrameAllocator()
    description:  The constructor requires a task id for the proper allocation
                  and deallocation operations of converter page frames.Note that
                  he caller has to assure, that the methods of this class never 
                  be called by another task. Contravention will increase the 
                  propability that the RTE resumes a wrong task!
    arguments:    taskId                 [in] Identification of the calling task
                  NumPermConvPageCounter [in] Reference to the shared converter
                                              member data counting the number
                                              of permanent used converter pages.
    return value: none
    --------------------------------------------------------------------------*/

    Converter_PageFrameAllocator( const tsp00_TaskId taskId )
    : m_TaskId( taskId )
    {}

    /*!-------------------------------------------------------------------------
    function:     Converter_PageFrameAllocator()
    description:  If no task id is given the task id will be determined. Note that 
                  the task id is used to get resourcen like latches. Therefore it is
                  not allowed to give this handle from one task to another.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    Converter_PageFrameAllocator() : m_TaskId( Converter_PageFrameAllocator::GetTaskId() )
    {}

    /*!-------------------------------------------------------------------------
    function:     New()
    description:  This method is used to get a piece of memory suitable to store
                  a kernel page and to make I/O. If no more memory is available
                  the componente frame control executes an emergency shutdown.
    arguments:    bIsRecoverable [in] count the allocation if the requested page 
                                      frame is for a permanent converter page.
    return value: (Kernel_IPage::PageFrame) New page frame handle
    --------------------------------------------------------------------------*/

    Kernel_IPage::PageFrame New( const SAPDB_Bool bIsRecoverable = true );

    /*!-------------------------------------------------------------------------
    function:     Free()
    description:  This method is used to release a page frame which is not needed 
                  any more by a converter page handler. The given frame will be
                  invalidated.
    arguments:    frame          [in] Handle of the page frame to release 
                  bIsRecoverable [in] count the deallocation if the given page 
                                          frame was used by a permanent converter page
    return value: none
    --------------------------------------------------------------------------*/

    void Free( Kernel_IPage::PageFrame  &frame,
               const SAPDB_Bool         bIsRecoverable = true );

private:

    static tsp00_TaskId GetTaskId();

private:
    /*-------------------------------------------------------------------------
    declaration: m_TaskId
    description: Identification of the calling task
    --------------------------------------------------------------------------*/

    const tsp00_TaskId m_TaskId;
};

/*!*****************************************************************************

   endclass: Converter_PageFrameAllocator

*******************************************************************************/


#endif  /* CONVERTER_PAGEFRAMEALLOCATOR_HPP */
