  /*!***************************************************************************

  module:       Converter_PageNoContainer.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author:       TillL

  special area: Converter

  description:  Simple container of data pagenos
                
  see also:     

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

  *****************************************************************************/


#ifndef CONVERTER_PAGENOCONTAINER_HPP
#define CONVERTER_PAGENOCONTAINER_HPP


#include "Container/Container_Vector.hpp"
#include "DataAccess/Data_Types.hpp"



/*!*****************************************************************************

   class: Converter_PageNoContainer

   description: Simple container of data page numbers

*******************************************************************************/
class Converter_PageNoContainer : 
private Container_Vector<Data_PageNo>
{
private:

    typedef Container_Vector<Data_PageNo> PageNoVector;

public:

    /*!-------------------------------------------------------------------------
    declaration : SizeType
    description:  container size in pagenos
    --------------------------------------------------------------------------*/
    typedef PageNoVector::SizeType                          SizeType;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  resource type used for dynamic memory allocation
    --------------------------------------------------------------------------*/
    typedef PageNoVector::Allocator                         Allocator;
    /*!-------------------------------------------------------------------------
    declaration : ConstIterator
    description:  r/only iterator of all pagenos in the container
    --------------------------------------------------------------------------*/
    typedef PageNoVector::ConstIterator                     ConstIterator;
        
	/*!-------------------------------------------------------------------------
    declaration : Iterator
    description:  iterator of all pagenos in the container
    --------------------------------------------------------------------------*/
    typedef PageNoVector::Iterator                          Iterator;

    /*!-------------------------------------------------------------------------
    function:     Converter_PageNoContainer()
    description:  Constructor
                    
    arguments:    alloc       [in]  allocator for container resources
    return value: none
    --------------------------------------------------------------------------*/
    Converter_PageNoContainer
        (Allocator&                     alloc)
        : 
         PageNoVector(alloc),
		 m_MinimumSize (0)
    {}

    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  Initializes container structures; allocates memory
                    
    arguments:    pnocount [in]   container size
    return value: none
    --------------------------------------------------------------------------*/
    SAPDB_Bool Initialize
        (SizeType                       pnocount)
    {
        SAPDBERR_ASSERT_STATE(GetSize() == 0);
        m_MinimumSize = pnocount;
        return Reserve(pnocount);
    }

    /*!-------------------------------------------------------------------------
    function:     Append()
    description:  Appends pageno to container
    
                  - A given pageno is inserted at the end of the container.
                  - The method fails if the container is full.

    arguments:    pageno [in]   pageno to be appended
    return value: (SAPDB_Bool) true if pageno successfully appended
    --------------------------------------------------------------------------*/
    SAPDB_Bool Append
        (const Data_PageNo&             pageno)
    {
        if (IsFull())
		{
			if (!Reserve(1+2*GetSize()))
				return false;
		}
        
		InsertEnd (pageno);
        return true;
    }

    /*!-------------------------------------------------------------------------
    function:     IsContained()
    description:  Checks whether specified pageno is in the container
    arguments:    pageno [in]   pageno to be found
    return value: (SAPDB_Bool) true if found
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsContained
        (const Data_PageNo&             pageno)
    {
		//if (IsEmpty())
		//	return false;

        ConstIterator iter    = Begin();
        ConstIterator enditer = End();
        SAPDB_Bool found      = false;

        while (iter != enditer && !found)
            found = *iter++ == pageno;

        return found;
    }
    
	
    /*!-------------------------------------------------------------------------
    function:     Clear()
    description:  Clears the contents of the container

                  - After this operation the container is empty
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Clear()
    {
		if (GetSize() > m_MinimumSize)
		{
			Delete();
			Reserve(m_MinimumSize);
		}
		else
		{
	    	PageNoVector::Clear();
		}
    }

	/*!-------------------------------------------------------------------------
    function:     Begin()
    description:  Returns an iterator refencing the first pageno in the
                  container

    arguments:    none 
    return value: (Iterator) iterator pointing on the first pageno
    --------------------------------------------------------------------------*/
	Iterator Begin()
    {
        return PageNoVector::Begin();
    }

	/*!-------------------------------------------------------------------------
    function:     UsedPageNoPageEnd()
    description:  Returns an iterator pointing behind the last pageno in
                  the container 
                  
    arguments:    none 
    return value: (Iterator) iterator pointing behind last pageno
    --------------------------------------------------------------------------*/
	Iterator End()
    {
        return PageNoVector::End();
    }

    /*!-------------------------------------------------------------------------
    function:     GetSize()
    description:  Returns the number of page numbers stored in the container
                  
    arguments:    none 
    return value: (SizeType)  number of page numbers stored in the container
    --------------------------------------------------------------------------*/
    SizeType GetSize() const
	{
		return PageNoVector::GetSize();
	}

private:

	SizeType                        m_MinimumSize;

};
/*!*****************************************************************************

   endclass: Converter_PageNoContainer

*******************************************************************************/



#endif // CONVERTER_PAGENOCONTAINER_HPP


