/*!**************************************************************************

  module      : Converter_UsedPageNoContainer.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author:       TillL

  special area: Converter

  description : Container to store the used data page numbers needed for a 
                save pages

  see also    : 

  copyright:    Copyright by SAP AG, 2001

  -------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_USEDPAGENOCONTAINER_HPP
#define CONVERTER_USEDPAGENOCONTAINER_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "DataAccess/Data_Types.hpp"
#include "Container/Container_Vector.hpp"
#include "Converter/Converter_Types.hpp"
#include "Converter/Converter_LeafPage.hpp"
#include "Converter/Converter_BitMapPage.hpp"
#include "Converter/Converter_PageFrameAllocator.hpp"


/*!*****************************************************************************

   class: Converter_UsedPageNoContainer

*******************************************************************************/

class Converter_UsedPageNoContainer
{
private:

    typedef Container_Vector<Converter_BitMapPage>          BitMapPageVector;

public:
	
    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : SizeType
    description:  container size type
    --------------------------------------------------------------------------*/
    typedef BitMapPageVector::SizeType                      SizeType;
    /*!-------------------------------------------------------------------------
    declaration : Iterator
    description:  iterator of bitmap pages (steps from page to page)
    --------------------------------------------------------------------------*/
    typedef BitMapPageVector::Iterator                      Iterator;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  resource type used for dynamic memory allocation
    --------------------------------------------------------------------------*/
    typedef BitMapPageVector::Allocator                     Allocator;

private:

    /*--------------------------------------------------------------------------
    declaration : IndexType
    description:  container index type
    --------------------------------------------------------------------------*/
    typedef BitMapPageVector::IndexType                     IndexType;
    /*--------------------------------------------------------------------------
    declaration : BitMapPageIterator
    description:  iterator which steps from slot to slot within a page
    --------------------------------------------------------------------------*/
    typedef Converter_BitMapPage::Iterator                  BitMapPageIterator;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

public:

	/*!-------------------------------------------------------------------------
    function:     Converter_UsedPageNoContainer()
    description:  Constructor

    arguments:    alloc      [in]  allocator for map resources
                  framealloc [in]  frame allocator for bitmap pages
                  addrmode   [in]  indicates whether the container should 
                                 store static or dynamic page numbers 
                  sizedelta  [in] realloc size increment if container full
    return value: none
    --------------------------------------------------------------------------*/
	Converter_UsedPageNoContainer 
        (Allocator&                     alloc,
         Data_AddressingMode            addrmode,
         SizeType                       sizedelta)
        : 
         m_PageVector(alloc),
         m_AddressingMode(addrmode),
         m_SizeDelta(sizedelta)
    {}

    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes container structures; deallocates memory and frames

    arguments:    framealloc [in]  frame allocator for bitmap pages
    return value: none
    --------------------------------------------------------------------------*/
    void Delete
        (Converter_PageFrameAllocator&  framealloc)
    {
        for (IndexType i=0; i<m_PageVector.GetSize(); i++)
            framealloc.Free(m_PageVector[i].GetFrame(), false);

        m_PageVector.Delete();
    }

	/*!-------------------------------------------------------------------------
    function:     Append()
    description:  Appends data page numbers to the container 

                  - Appends all data page numbers marked as used on 
	                the converter leaf page to the container. 

    arguments:    leafpage   [in] converter leaf page  
                  framealloc [in]  frame allocator for bitmap pages,
				  version    [in]  current converter version
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
	SAPDB_Bool Append 
        (const Converter_LeafPage&      leafpage,
         Converter_PageFrameAllocator&  framealloc,
	     Converter_Version              version);

	/*!-------------------------------------------------------------------------
    function:     Begin()
    description:  Returns an iterator referencing the first bitmap page in the
                  container

    arguments:    none 
    return value: (Iterator) iterator referencing the first container page
    --------------------------------------------------------------------------*/
	Iterator Begin()
    {
        return m_PageVector.Begin();
    }

	/*!-------------------------------------------------------------------------
    function:     End()
    description:  Returns an iterator refencing behind the last bitmap page in
                  the container 
                  
    arguments:    none 
    return value: (Iterator) iterator pointing behind the last container page
    --------------------------------------------------------------------------*/
	Iterator End()
    {
        return m_PageVector.End();
    }

	/*!-------------------------------------------------------------------------
    function:     IsEmpty()
    description:  Returns true if the container contains at least one page

    arguments:    none 
    return value: (SAPDB_Bool)true if the container contains at least one page
    --------------------------------------------------------------------------*/
	SAPDB_Bool IsEmpty()
    {
        return m_PageVector.IsEmpty();
    }

    /*!-------------------------------------------------------------------------
    function:     GetSize()
    description:  Returns the size of the container

    arguments:    none 
    return value: (SizeType)true size of the container
    --------------------------------------------------------------------------*/
	SizeType GetSize()
    {
        return m_PageVector.GetSize();
    }

    /*!-------------------------------------------------------------------------
    function:     Position()
    description:  Returns an iterator to the <n>-th converter  bit map page.
                  
    arguments:    none 
    return value: (Iterator) iterator pointing behind the last container page
    --------------------------------------------------------------------------*/
	Iterator Position(BitMapPageVector::IndexType n)
    {
        return m_PageVector.Position (n);
    }

    /*!----------------------------------------------------------------------
    function: Resize()
    description: Changes the number of vector elements.

                 If the capacity of the vector is not large enough, the 
                 memory for the vector elements is reallocated.
                 
                 ATTENTION!! After reallocation all iterators may be invalid!
                 
    arguments:   NewSize [in]  The new number of vector elements.
    returns:     (SAPDB_Bool) true if successful
    -----------------------------------------------------------------------*/
    SAPDB_Bool Resize(SizeType NewSize)
	{
        return m_PageVector.Resize(NewSize);
    }

private:

    /*--------------------------------------------------------------------------
    declaration : m_PageVector
    description:  vector of bitmap page handlers
    --------------------------------------------------------------------------*/
    BitMapPageVector                    m_PageVector;
    /*--------------------------------------------------------------------------
    declaration : m_AddressingMode
    description:  addressing mode (static/dynamic) of the map
    --------------------------------------------------------------------------*/
    Data_AddressingMode                 m_AddressingMode;
    /*--------------------------------------------------------------------------
    declaration : m_SizeDelta
    description:  vector realloc size delta
    --------------------------------------------------------------------------*/
    SizeType                            m_SizeDelta;
};

/*!*****************************************************************************

   endclass: Converter_UsedPageNoContainer

*******************************************************************************/




#endif  /* CONVERTER_USEDPAGENOCONTAINER_HPP */
