/*!**************************************************************************

  module      : Converter_WaitQueue.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  author      : AlexanderK

  special area: converter
  description : wait queue class used in the converter


  last changed: 2001-01-24  16:16
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef CONVERTER_WAITQUEUE_HPP
#define CONVERTER_WAITQUEUE_HPP


#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "gsp00.h"
#include "ggg00.h"
#include "heo56.h" //vsuspend


/*!*****************************************************************************

   class: Converter_WaitQueue

*******************************************************************************/

class Converter_WaitQueue : private Kernel_WaitStack
{
public:

    /*!-------------------------------------------------------------------------
    function:     Converter_WaitQueue()
    description:  Constructor
    arguments:    suspendReason [in]   an id for reson why the task was suspended
                                       for the reason corresponding to the id see veo00xc  
    return value: none
    --------------------------------------------------------------------------*/

    Converter_WaitQueue( const SAPDB_Int2 suspendReason ):m_SuspendReason( suspendReason ){};
     
    /*!-------------------------------------------------------------------------
    function:     ResumeAllWaitingTasks()
    description:  Resumes all task waiting in this queue 
    arguments:    taskId [in] task id  
    return value: none
    --------------------------------------------------------------------------*/
    
    void ResumeAllWaitingTasks( const tsp00_TaskId taskId )
    {
        PopAllTaskIdsAndResume( taskId );
    }

    /*!-------------------------------------------------------------------------
    function:     InsertTaskAndWait()
    description:  Inserts a task into this queue of suspended tasks and suspends the task
    arguments:    taskId [in] task id  
    return value: none
    --------------------------------------------------------------------------*/
    
    void InsertTaskAndWait( const tsp00_TaskId taskId )
    {
        InsertTask( taskId );
        Wait( taskId );
    }

    /*!-------------------------------------------------------------------------
    function:     InsertTask()
    description:  Inserts a task into this queue without suspending the task.
    arguments:    taskId [in] task id  
    return value: none
    --------------------------------------------------------------------------*/
    
    void InsertTask( const tsp00_TaskId taskId )
    {
        PushTaskId( taskId );
    }

    /*!-------------------------------------------------------------------------
    function:     Wait()
    description:  The calling task is suspended..
    arguments:    taskId [in] task id  
    return value: none
    --------------------------------------------------------------------------*/
    
    void Wait( const tsp00_TaskId taskId ) const
    {
        vsuspend( taskId, cgg_suspend_lvl_0, m_SuspendReason );
    }

    /*!-------------------------------------------------------------------------
    function:     IsEmpty()
    description:  Returns true if the wait queue is empty else false.
    arguments:    taskId [in] task id  
    return value: (bool)
    --------------------------------------------------------------------------*/
    
    bool IsEmpty( const tsp00_TaskId taskId ) const
    {
       return( Kernel_WaitStack::IsEmpty( taskId ));
    }

private:

    SAPDB_Int2 m_SuspendReason;
};

/*!*****************************************************************************

   endclass: Converter_WaitQueue

*******************************************************************************/






#endif  /* CONVERTER_WAITQUEUE_HPP */
