/*!
  @file           DBMCli_EventListener.cpp
  @author         Bernd Vorsprach - bernd.vorsprach@sap.com
  @brief          Listener for database kernel events

\if EMIT_LICENCE
    ========== licence begin LGPL
    Copyright (C) 2003 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
\endif
*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
#include "SAPDB/DBM/Cli/DBMCli_EventListener.hpp"

/*! @brief Constructor  */
DBMCli_Event::DBMCli_Event
    (DBMCli_Result & oResult )
    : m_sName(""),
      m_oPriority(prio_unknown),
      m_oDate(),
      m_oTime(),
      m_nValue1(0),
      m_nValue2(0),
      m_nCount(0),
      m_sText(""),
      m_sDescription("")
{
  DBMCli_String sLine;
  
  oResult.SetPos(0);

  while (oResult.GetLine(sLine)) {
    SetProperty(sLine);
  } // end while
} // end DBMCli_Event::DBMCli_Event

/*! @brief private member  */
void DBMCli_Event::SetProperty
    ( const DBMCli_String & sProperty )
{
  int nAssign = sProperty.Find("=");
  if (nAssign > 0) {
    DBMCli_String sName  = sProperty.Left(nAssign);
    DBMCli_String sValue = sProperty.Mid(nAssign + 1);
    
    sName.Trim();
    sValue.Trim();
    
    if (sName.Compare("NAME")        == 0) {
      m_sName = sValue;
    } else if (sName.Compare("PRIORITY")    == 0) {
      if (sValue.Compare("LOW") == 0) {
        m_oPriority = prio_low;
      } else if (sValue.Compare("MEDIUM") == 0) {
        m_oPriority = prio_medium;
      } else if (sValue.Compare("HIGH") == 0) {
        m_oPriority = prio_high;
      } else  {
        m_oPriority = prio_unknown;
      } // end if
    } else if (sName.Compare("DATE")        == 0) {
      m_oDate.Set(sValue, DT_NUM_DATE_FMT);
    } else if (sName.Compare("TIME")        == 0) {
      m_oTime.Set(sValue, DT_NUM_TIME_FMT);
    } else if (sName.Compare("VALUE1")      == 0) {
      m_nValue1      = atoi(sValue);
    } else if (sName.Compare("VALUE2")      == 0) {
      m_nValue2      = atoi(sValue);
    } else if (sName.Compare("COUNT")       == 0) {
      m_nCount       = atoi(sValue);
    } else if (sName.Compare("TEXT")        == 0) {
      m_sText        = sValue;
    } else if (sName.Compare("DESCRIPTION") == 0) {
      m_sDescription = sValue;
    } // end if
  } // end if

} // end DBMCli_Event::SetProperty

/*! @brief Constructor  */
DBMCli_EventListener::DBMCli_EventListener
    ( const DBMCli_String  & sServer,
      const DBMCli_String  & sDatabase,
      const DBMCli_String  & sUserPwd,
      SAPDBErr_MessageList & oMessageList)
    : DBMCli_Database(sServer, sDatabase, sUserPwd, oMessageList)
{

} // end DBMCli_EventListener::DBMCli_EventListener

/*! @brief public member  */
bool DBMCli_EventListener::Run
     ( )
{
  m_oMessageList.ClearMessageList();

  while (DBMCli_Session::Execute("event_wait", m_oMessageList) ) {
    DBMCli_Result & oResult = DBMCli_Session::GetResult();
    if (!oResult.Check(m_oMessageList)) {
      // generate error and return
      return false;
    } // end if
    DBMCli_Event oEvent(oResult);

    if (!EventProc(oEvent)) {
      return true;
    } // end if

    m_oMessageList.ClearMessageList();
  } // end while

  return m_oMessageList.IsEmpty();
} // end DBMCli_EventListener::Run

/*! @brief public member  */
bool DBMCli_EventListener::EventProc
    ( const DBMCli_Event & oEvent )
{
  return true;
} // end DBMCli_EventListener::EventProc
