/*!
  -----------------------------------------------------------------------------
  module: error.cpp
  -----------------------------------------------------------------------------
  responsible:  MartinR
  special area: DBanalyzer - The performance analyse tool for SAPDB
  description:  dbanalyzer error handling
  version:      7.4 and higher 
  -----------------------------------------------------------------------------

  copyright: by SAP AG, 2002

    ========== licence begin LGPL
    Copyright (C) 2002 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

  -----------------------------------------------------------------------------
*/

#include "SAPDB/DBanalyzer/error.hpp"
#include "SAPDB/DBanalyzer/protocol.hpp"

DBA_error::DBA_error ()
{
  m_nNumber = DBA_OK;
  m_nValues = 0;
  m_sExtText = "";
  m_bReconnect = false;
  m_bIgnore    = false;
}

DBA_error::~DBA_error ()
{
}

DBA_void DBA_error::Clear ()
{
  m_nNumber = DBA_OK;
  m_nValues = 0;
  m_sExtText = "";
  m_bReconnect = false;
  m_bIgnore    = false;
}

DBA_void DBA_error::SetError (const DBA_errornum nNumber)
{
  m_nNumber = nNumber;
  m_nValues = 0;
  m_sExtText = "";
}

DBA_void DBA_error::SetError(const DBA_errornum nNumber, 
                             const DBA_string & sValue1)
{
  m_nNumber = nNumber;
  m_nValues = 1;
  m_sValue1 = sValue1;
  m_sExtText = "";
}

DBA_void DBA_error::SetError(const DBA_errornum nNumber, 
                             const DBA_string & sValue1,
                             const DBA_string & sValue2)
{
  m_nNumber = nNumber;
  m_nValues = 2;
  m_sValue1 = sValue1;
  m_sValue2 = sValue2;
  m_sExtText = "";
}

DBA_void DBA_error::SetError(const DBA_errornum nNumber, 
                             const DBA_string & sValue1,
                             const DBA_string & sValue2,
                             const DBA_string & sValue3)
{
  m_nNumber = nNumber;
  m_nValues = 3;
  m_sValue1 = sValue1;
  m_sValue2 = sValue2;
  m_sValue3 = sValue3;
  m_sExtText = "";
}

DBA_void DBA_error::SetExtText(const DBA_string & sExtText)
{
  m_sExtText = sExtText;
}

DBA_string DBA_error::GetErrorMsg()
{
  DBA_string sError;

  switch (m_nNumber) {
    case DBA_OK:                sError = DBA_ERRTXT_OK; break;         
    case DBA_ERR:               sError = DBA_ERRTXT_ERR; break;
    case DBA_ERR_ARGS:          sError = DBA_ERRTXT_ARGS; break;    
    case DBA_ERR_OPENFILE:      sError = DBA_ERRTXT_OPENFILE; break;
    case DBA_ERR_CONNECT:       sError = DBA_ERRTXT_CONNECT; break;
    case DBA_ERR_EXECUTE:       sError = DBA_ERRTXT_EXECUTE; break;
    case DBA_ERR_NOMEM:         sError = DBA_ERRTXT_NOMEM; break;
    case DBA_ERR_PARSE:         sError = DBA_ERRTXT_PARSE; break;
    case DBA_ERR_CALC:          sError = DBA_ERRTXT_CALC; break;
    case DBA_ERR_INVCONFIGKEY:  sError = DBA_ERRTXT_INVCONFIGKEY; break;
    case DBA_ERR_TYPE:          sError = DBA_ERRTXT_TYPE; break;
    case DBA_WRN_CALC:          sError = DBA_WRNTXT_CALC; break;
    case DBA_ERR_CREATEDIR:     sError = DBA_ERRTXT_CREATEDIR; break;
    case DBA_ERR_ACTIVE:        sError = DBA_ERRTXT_ACTIVE; break;
    case DBA_ERR_OTHERDB:       sError = DBA_ERRTXT_OTHERDB; break;
    case DBA_ERR_UNKNOWNOBJ:    sError = DBA_ERRTXT_UNKNOWNOBJ; break;
    case DBA_ERR_DUPLICATE:     sError = DBA_ERRTXT_DUPLICATE; break;
    case DBA_ERR_FORMTAB:       sError = DBA_ERRTXT_FORMTAB; break;
    case DBA_ERR_KILL:          sError = DBA_ERRTXT_KILL; break;
    case DBA_ERR_UNKNOWNVAL:    sError = DBA_ERRTXT_UNKNOWNVAL; break;
    case DBA_ERR_NOTACTIVE:     sError = DBA_ERRTXT_NOTACTIVE; break;
    default:                    sError = DBA_ERRTXT_UNKNOWN; break;
  }

  if (m_nValues > 0) {
    DBA_char   sBuffer[1024];

    switch (m_nValues) {
      case 1:
        sprintf(sBuffer, sError.CharPtr(), m_sValue1.CharPtr());
        break;  

      case 2:
        sprintf(sBuffer, sError.CharPtr(), m_sValue1.CharPtr(),
                                           m_sValue2.CharPtr());
        break;  
      case 3:
        sprintf(sBuffer, sError.CharPtr(), m_sValue1.CharPtr(),
                                           m_sValue2.CharPtr(),
                                           m_sValue3.CharPtr());
        break;  
      default:
        sprintf(sBuffer, sError.CharPtr());
        break;  
    }
    sError = DBA_string(sBuffer);
  }

  return sError;
}

//
// DBA_error::ShowErrorMsg
//
DBA_void DBA_error::ShowErrorMsg(DBA_protocol * pProtocol, DBA_bool bInfo) 
{
  DBA_string sClass;

  if (bInfo) {
    sClass = "INFO";
  } else {
    sClass = (m_nNumber == DBA_WRN_CALC) ? "WARNING" : "ERROR";
  } // end if

  DBA_string sError = GetErrorMsg();
  
  if (pProtocol != NULL) {
    pProtocol->WriteError( sClass, m_nNumber, sError, m_sExtText);
  } else {
    printf("%s %d: %s\n", sClass.CharPtr(), (int) m_nNumber, sError.CharPtr());
    if (!m_sExtText.Empty()) {
      printf("%s\n\n", m_sExtText.CharPtr());
    } else {
      printf("\n");
    } // end if
  } // end if
} // end DBA_error::ShowErrorMsg

DBA_void DBA_error::SetReconnect(const DBA_bool bReconnect)
{
  m_bReconnect = bReconnect;
}

DBA_bool DBA_error::GetReconnect()
{
  return m_bReconnect;
}

DBA_void DBA_error::SetIgnore(const DBA_bool bIgnore)
{
  m_bIgnore = bIgnore;
}

DBA_bool DBA_error::GetIgnore()
{
  return m_bIgnore;
}

