/*!***************************************************************************

  module      : Data_ChainFixSizeSpace.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-22
  last changed: 2000-11-22 12:00
  copyright   : Copyright by SAP AG, 2000
  description : defines class Data_ChainFixSizeSpace



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef Data_ChainFixSizeSpace_HPP
#define Data_ChainFixSizeSpace_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "DataAccess/Data_Exceptions.hpp"
#include "DataAccess/Data_Messages.hpp"
#include "DataAccess/Data_PageFixSizeSpace.hpp"
#include "DataAccess/Data_Chain.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define FIRST_RECORD true

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Data_ChainFixSizeSpace
    description: A persistent page chain handler, based on data pages.
 */
template <class PAGE> class Data_ChainFixSizeSpace : public Data_Chain<PAGE>
{

public:

    /*! -----------------------------------------------------------------------
    declaration:    internal types
    */
    typedef SAPDB_Byte*                        RecordSpace;
    typedef Data_PageFixSizeSpace::Iterator    RecordIterator;
    typedef Data_ChainIterator<PAGE>           PageIterator;
    typedef Data_PageFixSizeSpace::RecordCount RecordId;

public:

    /*---------------------------------------------------------------------------*/
    /*! class:       Iterator
        description: Iterator for records in the chain.
     */
    class Iterator
    {
    friend
        class Data_ChainFixSizeSpace<PAGE>;
    
    public:
    
        /*! -----------------------------------------------------------------------
        function:    Iterator
        arguments:   PAM       [in] The AccessManager to use by the Iterator.
        description: The iterater is not initialized and at first invalid.       
        */
        Iterator (Data_IPageAccessManager& PAM)
        : m_PAM        (PAM),
          m_PageIter   (PAM)
        {}

        /*! -----------------------------------------------------------------------
        function:    IsValid
        returns:     true, if the iterator points to a record.
        */
        bool IsValid() const
        {
            return m_PageIter.IsValid()
                   &&
                   m_RecordIter.IsValid();
        }

        /*! -----------------------------------------------------------------------
        function:    Invalidate
        arguments:   isOK [in] if set to false the pages are released without update
        description: This invalidates all members.
        */
        void Invalidate(bool isOK = true)
        {
            m_RecordIter.Invalidate();
            m_PageIter.Invalidate();
        }

        /*! -----------------------------------------------------------------------
        function:    GetPosition
        arguments:   PageRef [out]
        description: This returns a page reference of the position, where 
                     the iterator points to.
        */
        void GetPosition (Data_PageNo &pno,
                          RecordId    &recordid)
        {
            // PTS 1115056 UH 2002-03-27 changed the assertion to return value
            if ( IsValid() )
            {
                pno      = (*m_PageIter).PageNo();
                recordid = m_RecordIter.GetRecordId();
            }
            else
            {
                pno      = Data_PageNo();
                recordid = 0;
            }
        }

        /*! -----------------------------------------------------------------------
        function:    SetPosition
        arguments:   PageRef [in]
        returns:     true, if position is reached.
        description: The iterator is set to the given position.
        */
        bool SetPosition (Data_PageNo     pno,
                          RecordId        recordid,
                          Data_AccessMode accessmode)
        {
            Assign (pno, true, accessmode);
            if ( ! m_PageIter.IsValid() )
                return false;
            (*m_PageIter).SetToRecord (recordid, m_RecordIter);
            return m_RecordIter.IsValid();
        }

        /*! -----------------------------------------------------------------------
        function:    Break
        description: The iterator releases its resources and saves current position.
        */
        void Break()
        {
            GetPosition ( m_BreakInfo.lastPno,
                          m_BreakInfo.lastrecordid );
            m_BreakInfo.lastAccessMode = GetPage().AccessMode();
            Invalidate();
        }

        /*! -----------------------------------------------------------------------
        function:    Continue
        returns:     (bool) false, if the saved position could not be set.
        description: The iterator takes and sets the saved position.
        */
        bool Continue()
        {
            bool result = SetPosition ( m_BreakInfo.lastPno,
                                        m_BreakInfo.lastrecordid,
                                        m_BreakInfo.lastAccessMode );
            m_BreakInfo = BreakInfo();
            return result;
        }

        /*! -----------------------------------------------------------------------
        function:    GetPageIterator
        description: Gives the current page the iterator points to.
                     The iterator must be valid.
        returns:     pageiter
        */
        PageIterator& GetPageIterator()
        {
            return m_PageIter;
        }
        
        /*! -----------------------------------------------------------------------
        function:    GetPage
        description: Gives the current page the iterator points to.
                     The iterator must be valid.
        returns:     <PAGE>
        */
        PAGE& GetPage()
        {
            SAPDBERR_ASSERT_STATE( m_PageIter.IsValid() );
            return *m_PageIter;
        }
        
        /*! -----------------------------------------------------------------------
        function:    GetRecordId
        returns:     SAPDB_UInt, MAX_UINT if invalid
        */
        RecordId GetRecordId () const
        {
            if ( ! IsValid() )
                return SAPDB_MAX_UINT4;
            return m_RecordIter.GetRecordId();
        }

        /*! -----------------------------------------------------------------------
        function:    SetRecordId
        arguments:   recordid
        */
        void SetRecordId (RecordId recordid)
        {
            SAPDBERR_ASSERT_STATE( m_PageIter.IsValid() );
            m_RecordIter.SetRecordId(recordid);
        }

        /*! -----------------------------------------------------------------------
        function:    operator *
        returns:     RecordSpace
        description: This returns the record, which is currently referenced.
        */
        RecordSpace operator * ()
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return *m_RecordIter;
        }

        /*! -----------------------------------------------------------------------
        function:    operator 
        description: If the iterator is valid it is set to the next record.
        */
        Iterator& operator ++ ()
        {
            if( ! IsValid() )
                return *this;

            ++m_RecordIter;
            
            if ( ! m_RecordIter.IsValid() )
            {
                m_RecordIter.Invalidate();

                ++m_PageIter;

                if ( m_PageIter.IsValid() )
                    (*m_PageIter).Begin (m_RecordIter);
            }
            return *this;
        }

        /*! -----------------------------------------------------------------------
        function:    operator --
        description: If the iterator is valid it is set to the previous record.
        */
        Iterator& operator -- ()
        {
            if( ! IsValid() )
                return *this;

            --m_RecordIter;
            
            if ( ! m_RecordIter.IsValid() )
            {
                --m_PageIter;

                if ( m_PageIter.IsValid() )
                    (*m_PageIter).Begin (m_RecordIter);
            }
            return *this;
        }

        /*! -----------------------------------------------------------------------
        function:    WriteToTrace
        arguments:   title [in]
        */
        void WriteToTrace (const char* title = 0)
        {
            Kernel_VTrace trace;
            if ( title != 0 )
                trace << title << ": ";
            else
                trace << "fixiter: ";

            if ( m_PageIter.IsValid() )
                trace << "@" << (*m_PageIter).PageNo();
            else
                trace << "@invalid";
                
            trace << "." << m_RecordIter.GetPosition()
                  << "[" << m_RecordIter.GetRecordId() << "]";
            trace << ", break@" << m_BreakInfo.lastPno << "["
                  << m_BreakInfo.lastrecordid << "]";
            trace << (!IsValid()?" (not valid)":"");
        }
        
    protected:
    
        /*! -----------------------------------------------------------------------
        function:    Assign
        arguments:   PageNo           [in]
                     SetToFirstRecord [in]
                     AccessMode       [in]
        description: this is used by the container.
                     if an error occurred, the iterator is invalid.
        */
        void Assign (Data_PageNo     PageNo,
                     bool            SetToFirstRecord,
                     Data_AccessMode AccessMode)
        {
            if ( m_PageIter.IsValid() )
            {
                if ( PageNo != (*m_PageIter).PageNo()
                     ||
                     (*m_PageIter).AccessMode() != AccessMode )
                {
                    Invalidate();
                    if ( ! m_PAM.GetPage ( (*m_PageIter),
                                           PageNo,
                                           (*m_PageIter).RecoveryMode(),
                                           AccessMode) )
                        Invalidate();
                }
            }
            else
            {
                if ( ! m_PAM.GetPage ( (*m_PageIter),
                                       PageNo,
                                       (*m_PageIter).RecoveryMode(),
                                       AccessMode) )
                    Invalidate();
            }
            if ( m_PageIter.IsValid() )
            {
                if ( SetToFirstRecord )
                    (*m_PageIter).Begin (m_RecordIter);
                else
                    (*m_PageIter).End (m_RecordIter);
            }
        }
        
    private:
    
        /*! -----------------------------------------------------------------------
        variable:    m_PAM
        */
        Data_IPageAccessManager& m_PAM;

        /*! -----------------------------------------------------------------------
        variable:    m_RecordIter
        */
        RecordIterator m_RecordIter;
        
        /*! -----------------------------------------------------------------------
        variable:    m_PageIter
        */
        PageIterator m_PageIter;

    private:
    
        /*! -----------------------------------------------------------------------
        declaration: BreakInfo
        */
        struct BreakInfo
        {
            Data_PageNo     lastPno;
            RecordId        lastrecordid;
            Data_AccessMode lastAccessMode;
            BreakInfo()
            {
                lastPno.Invalidate();
                lastrecordid   = 0;
                lastAccessMode = Data_ForRead;
            }
        };

        /*! -----------------------------------------------------------------------
        variable:    m_BreakInfo
        description: this is used by Break() and Continue()
        */
        BreakInfo m_BreakInfo;

    };
    /*! endclass: Iterator */

public:

    /*! ---------------------------------------------------------------------------
    function:    Data_ChainFixSizeSpace
    arguments:   PAM             [in/out]
                 RootId          [in] this id defines the beginning of the page chain
                 FixRecordLength [in]
    description: if the rootid is valid, the given fix-record-size is checked.
    */
    Data_ChainFixSizeSpace (Data_IPageAccessManager &PAM,
                            Data_PageId             &RootId,
                            Data_RecordLength        FixRecordLength)
    : Data_Chain<PAGE>  (PAM, RootId),
      m_FixRecordLength (FixRecordLength)
    {
        #ifdef SAPDB_QUICK
        CheckFixRecordLength(); // PTS 1114647 for better performance only check in quick
        #endif
    }

    /*! ---------------------------------------------------------------------------
    function:    Data_ChainFixSizeSpace
    arguments:   PAM             [in/out]
                 RootId          [in] this id defines the beginning of the page chain
                 LastId          [in] this id defines the ending of the page chain
                 FixRecordLength [in]
    description: if the rootid is valid, the given fix-record-size is checked.
    */
    Data_ChainFixSizeSpace (Data_IPageAccessManager &PAM,
                            Data_PageId             &RootId,
                            Data_PageNo              LastPageNo,
                            Data_RecordLength        FixRecordLength)
    : Data_Chain<PAGE>  (PAM, RootId, LastPageNo),
      m_FixRecordLength (FixRecordLength)
    {
        #ifdef SAPDB_QUICK
        CheckFixRecordLength(); // PTS 1114647 for better performance only check in quick
        #endif
    }

    /*! ---------------------------------------------------------------------------
    function:    Create
    arguments:   RootPageIter    [out]
    returns:     false if no space is available
    description: A new Chain is created. The PageIter points to the root.
    */
    bool Create (PageIterator &RootPageIter)
    {
        if ( ! Data_Chain<PAGE>::Create (RootPageIter) )
            return false; // PTS 1121659 UH 2003-04-30
        
        PAGE &page = *RootPageIter;
        
        page.SetFixRecordLength ( m_FixRecordLength );
        page.SetLastPageNo      ( page.PageNo()     );
        return true; // PTS 1121659 UH 2003-04-30
    }

    /*! ---------------------------------------------------------------------------
    function:    Create
    description: A new Chain is created.
    */
    bool Create ()
    {
        PageIterator RootPageIter (this->GetPageAccessManager());
        return Create (RootPageIter); // PTS 1121659 UH 2003-04-30
    }

    /*! ---------------------------------------------------------------------------
    function:    Set
    arguments:   RecordIterator [in/out]
                 pageno         [in]
                 AccessMode     [in]
    description: The RecordIterator is set to the first record of the given page.
    */
    void Set (Iterator&       iter,
              Data_PageNo     pageno,
              Data_AccessMode accessmode,
              bool            toFirstRecord)
    {
        iter.Assign( pageno, toFirstRecord, accessmode);
    }

    /*! ---------------------------------------------------------------------------
    function:    Begin
    arguments:   RecordIterator [out]
                 AccessMode     [in]
    description: The RecordIterator is set to the first record in the chain.
    */
    void Begin (Iterator&       iter,
                Data_AccessMode accessmode)
    {
        Set ( iter, this->RootId().PageNo(), accessmode, true);
    }

    /*! ---------------------------------------------------------------------------
    function:    ReserveSpace
    arguments:   RecordIter      [out] A RecordIterator, which points to the newly created space.
    description: Space is reserved in the chain.
                 The record iterator points to the new space.
                 If no space was allocated, the iterator is invalid.
                 The Chain is locked here but must explicitely Unlock()'ed after the
                 space has been written !!
    */
    bool ReserveSpace (Iterator& iter,
                       bool      useForStructureChange = true)
    {
        if ( ! this->m_PAM.CheckSpace (1) ) // PTS 1115170 UH 2002-04-09
            return false; // PTS 1121659 UH 2003-04-30

        if ( ! this->Lock (useForStructureChange?Data_ForStructureChange:Data_ForUpdate) )
        {
            RTE_Crash( SAPDBErr_Exception(
                       __FILE__, __LINE__,
                       SAPDBERR_ASSERT_STATE_FAILED,
                       "Data_ChainFixSizeSpace::ReserveSpace: Lock()") );
        }

        Data_PageNo   last     = this->LastPageNo();
        PageIterator &pageiter = iter.m_PageIter;
        
        if ( ! GetPage (last, Data_ForUpdate, pageiter) )
        {    
            Kernel_VTrace() << "FixChain.reserve: root: " << this->RootId().PageNo()
                            << ", last: " << last << NewLine;
            RTE_Crash( SAPDBErr_Exception(
                       __FILE__, __LINE__,
                       SAPDBERR_ASSERT_STATE_FAILED,
                       "Data_ChainFixSizeSpace::ReserveSpace: Get last page") );
        }
        
        (*pageiter).ReserveSpace (iter.m_RecordIter);
        
        if ( ! iter.IsValid() )
        {
            PAGE newPage;
                        
            (*pageiter).Append (newPage, last);

            newPage.SetFixRecordLength (m_FixRecordLength);

            ReleasePage (pageiter);
            
            pageiter.Set(newPage);

            (*pageiter).ReserveSpace (iter.m_RecordIter);
            
            if ( this->m_RootPage.IsAssigned() )
            {
                this->m_RootPage.SetLastPageNo ((*pageiter).PageNo());
                this->m_RootPage.SetChanged();
            }
            
        }
        // PTS 1127083 UH 2004-01-14 removed Unlock()

        return true; // PTS 1121659 UH 2003-04-30
    }

    /*! ---------------------------------------------------------------------------
    function:    RemoveSpace
    arguments:   RecordIter [in] A RecordIterator, which points to the space to remove.
    description: Space is removed in the chain.

                 The iterator is set to the next valid position, if there is one.
                 A page is only removed, if it is not the root or the last page.
    */
    void RemoveSpace (Iterator &iter)
    {
        PAGE &page = iter.GetPage();

        SAPDBERR_ASSERT_STATE( page.GetRecordCount() > 0 );

        page.DecrementRecordCount();

        if ( page.GetRecordCount() > 0
             ||
             this->RootId().PageNo() == page.PageNo() )
            ++iter;
        else
        {
            PageIterator &pageiter = iter.GetPageIterator();

            Data_PageNo prevPageNo = page.PrevPageNo();

            if ( prevPageNo.IsInvalid() )
            {
                Kernel_VTrace() << "FixChain: root: " << this->RootId().PageNo()
                                << ", pageno: " << page.PageNo() << NewLine;
                RTE_Crash( SAPDBErr_Exception(
                           __FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Data_ChainFixSizeSpace::Removespace: prev pno is invalid.") );
            }

            pageiter.Invalidate();

            if ( ! this->Lock (Data_ForStructureChange) )
            {
                Kernel_VTrace() << "FixChain.remove: root: " << this->RootId().PageNo() << NewLine;
                RTE_Crash( SAPDBErr_Exception(
                           __FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Data_ChainFixSizeSpace::Removespace: Lock()") );
            }
            
            if ( ! GetPage(prevPageNo, Data_ForUpdate, pageiter) )
            {
                Kernel_VTrace() << "FixChain.remove: root: " << this->RootId().PageNo()
                                << ", prevpageno: " << prevPageNo << NewLine;
                RTE_Crash( SAPDBErr_Exception(
                           __FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Data_ChainFixSizeSpace::Removespace: get prev page") );
            }

            Delete (pageiter);

            if ( this->m_RootPage.IsAssigned() )
            {
                Data_PageNo NextPageNo = (*pageiter).NextPageNo();
                ReleasePage (pageiter);
                if ( NextPageNo.IsValid() )
                    (void) GetPage (NextPageNo, Data_ForUpdate, pageiter);
            }
            else
                ++pageiter;

            if ( ! pageiter.IsValid() )
            {
                // The removed page was the last page in the chain.

                if ( this->m_RootPage.IsAssigned() )
                {
                    this->m_RootPage.SetLastPageNo (prevPageNo);
                    this->m_RootPage.SetChanged();
                }

                iter.Invalidate();
            }
            else
            {            
                // The page iter points now to the next page after the removed page.

                (*pageiter).SetPrevPageNo(prevPageNo);
                (*pageiter).Begin (iter.m_RecordIter);
            }
            this->UnLock();
        }
    }

    /*! ---------------------------------------------------------------------------
    function:    GetSpaceSize
    returns:     size in bytes
    */
    Data_RecordLength GetSpaceSize () const
    {
        return m_FixRecordLength;
    }

private: // methods
    
    /*! ---------------------------------------------------------------------------
    function:    CheckFixRecordLength
    */
    void CheckFixRecordLength()
    {
        if ( this->RootId().IsValid() )
        {
            PageIterator RootPageIter (this->GetPageAccessManager());

            Data_Chain<PAGE>::Begin (RootPageIter, Data_ForRead);

            if ( ! RootPageIter.IsValid() )
            {
                this->InfoMessage("ChainFixSizeSpace");
                RTE_Crash(SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,"accessing root"));
            }

            if ( (*RootPageIter).GetFixRecordLength() != m_FixRecordLength )
            {
                this->InfoMessage("ChainFixSizeSpace");
                RTE_Crash(Data_Exception(__CONTEXT__, DATA_ERROR_CHAIN_FIX_SIZE_MISMATCH,
                                         SAPDB_ToString(m_FixRecordLength),
                                         SAPDB_ToString((*RootPageIter).GetFixRecordLength()) ));
            }
        }
    }

private: // members

    /*! -----------------------------------------------------------------------
    variable:  m_FixRecordLength
    */
    const Data_RecordLength m_FixRecordLength;

};
/*! endclass: Data_ChainFixSizeSpace */

#undef FIRST_RECORD

#endif // Data_ChainFixSizeSpace_HPP
