/*!***************************************************************************

  module      : Data_ChainSplitSpaceForwardRead.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-22
  last changed: 2000-11-22 12:00
  copyright   : Copyright by SAP AG, 2000
  description : Defines classes to organize Split records over page chains.
  


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef Data_ChainSplitSpaceForwardRead_HPP
#define Data_ChainSplitSpaceForwardRead_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "DataAccess/Data_Exceptions.hpp"
#include "DataAccess/Data_Chain.hpp"
#include "DataAccess/Data_PageSplitSpace.hpp"
#include "DataAccess/Data_SplitSpace.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define FIRST_FREE   true

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Data_ChainSplitSpaceForwardRead
    description: This chain of Split record supports forward reading.
 */
template <class PAGE>
class Data_ChainSplitSpaceForwardRead : public Data_Chain<PAGE>
{

public:

    /*! -----------------------------------------------------------------------
    declaration:    RecordSpace
    */
    typedef Data_SplitSpace<PAGE> RecordSpace;

public:

    /*---------------------------------------------------------------------------*/
    /*! class:       Iterator
        description: Iterator for records in the chain.
     */
    class Iterator : public Data_IBreakable // PTS 1114994 2002-03-22
    {

    public:
    
        /*! -----------------------------------------------------------------------
        declaration:    Container
        */
        typedef Data_ChainSplitSpaceForwardRead<PAGE> Container;

    friend
        class Data_ChainSplitSpaceForwardRead<PAGE>;
    
    public:
    
        /*! -----------------------------------------------------------------------
        function:    Iterator
        arguments:   PAM       [in] The AccessManager to use by the Iterator.
                     Allocator [in] Is used for the Split space.
        description: The iterater is not initialized and at first invalid.       
        */
        Iterator (Data_IPageAccessManager& PAM,
                  SAPDBMem_IRawAllocator&  Allocator)
        : m_PAM                   (PAM),
          m_RecordIterator        (),
          m_RecordSpace           (Allocator),
          m_RecordSpaceIsComplete (false)
        {}

        /*! -----------------------------------------------------------------------
        function:    ~Iterator
        description: deallocates everything
        */
        ~Iterator ()
        {
            Delete();
        }

        /*! -----------------------------------------------------------------------
        function:    Initialize
        arguments:   AccessMode [in]
        returns:     false, if the iterator could not be used
        description: The iterater is initialized.
        */
        bool Initialize(Data_AccessMode AccessMode = Data_ForUpdate)
        {
            bool initialized = m_RecordSpace.Initialize();
            if ( initialized )
                m_RecordSpace.SetAccessMode(AccessMode);
            return initialized;
        }

        /*! -----------------------------------------------------------------------
        function:    Delete
        description: The resources of the iterater are deleted.
        */
        void Delete()
        {
            Invalidate();
            m_RecordSpace.Delete();
        }

        /*! -----------------------------------------------------------------------
        function:    IsValid
        returns:     true, if a part of a split space is available.
                     if the part is complete must be tested outside.
        */
        bool IsValid() const
        {
            return m_RecordSpace.IsAssigned() && m_RecordIterator.IsValid();
        }

        /*! -----------------------------------------------------------------------
        function:    Invalidate
        arguments:   isOK [in] if set to false the pages are released without update
        description: This releases all allocated internal members.
        */
        void Invalidate(bool isOK = true)
        {
            SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::Iterator::Invalidate", DataChain_Trace, 5);
            m_RecordSpaceIsComplete = false;
            m_RecordIterator.Invalidate();
            // PTS 1115979 UH 2002-05-29
            // PTS 1117126 UH 2002-08-07 removed UndoLastReservedSpace
            m_RecordSpace.Deassign(isOK);
        }

        /*! -----------------------------------------------------------------------
        function:    SetPosition
        arguments:   pno, offset
        returns:     true, if successfull
        description: ...
        */
        bool SetPosition (Data_PageNo     pno,
                          Data_PageOffset offset,
                          Data_AccessMode accessmode)
        {
            return Assign (pno, offset, accessmode);
        }

        /*! -----------------------------------------------------------------------
        function:    GetPosition
        arguments:   PageRef [out]
        description: This returns a page reference of the position, where 
                     the iterator points to.
        */
        void GetPosition (Data_PageNo     &pno,
                          Data_PageOffset &offset)
        {
            // PTS 1115056 UH 2002-03-27 changed the assertion to return value
            if ( IsValid() )
            {
                PAGE       &page = m_RecordSpace.GetPage(0);
                SAPDB_UInt  length;
                SAPDB_Byte *valueptr;
                m_RecordSpace.GetPart (0, length, valueptr);
    
                pno    = page.PageNo();
                offset = page.GetOffset(valueptr, length);
            }
            else
            {
                pno    = Data_PageNo();
                offset = 0;
            }
        }

        /*! -----------------------------------------------------------------------
        function:    Break
        description: This releases all resources (pages in datacache)
        */
        virtual bool Break() // PTS 1114994 2002-03-22 returns bool
        {
            GetPosition ( m_BreakInfo.lastPno,
                          m_BreakInfo.lastOffset );
            m_BreakInfo.lastAccessMode = GetCurrentPage().AccessMode();
            Invalidate();
            return true;
        }
        
        /*! -----------------------------------------------------------------------
        function:    Continue
        description: This resumes the state of a previous Break().
        */
        virtual bool Continue()
        {
            if ( ! SetPosition ( m_BreakInfo.lastPno,
                                 m_BreakInfo.lastOffset,
                                 m_BreakInfo.lastAccessMode ) )
            {
                WriteToTrace ("Continue failed");
            	return false; // PTS 1115056 UH 2002-03-27 changed the assertion to return value
            }
            // assign the record space again
            (void) this->operator * ();
             m_BreakInfo = BreakInfo();
             return true; // PTS 1115056 UH 2002-03-27 changed the assertion to return value
        }
        
        /*! -----------------------------------------------------------------------
        function:    GetCurrentPage
        returns:     page
        description: This returns the handle of the current page.
        */
        PAGE& GetCurrentPage ()
        {
            return m_RecordSpace.GetCurrentPage();
        }

        /*! -----------------------------------------------------------------------
        function:    operator *
        returns:     const Data_SplitSpace<PAGE>&
        description: This returns the record, which is currently referenced.
        */
        Data_SplitSpace<PAGE>& operator * ()
        {
            if ( ! m_RecordSpaceIsComplete )
                AssignRecordspace (m_RecordIterator);
            return m_RecordSpace;
        }

        /*! -----------------------------------------------------------------------
        function:    GetSize
        returns:     Data_SplitRecordLength
        description: Without assigning the complete record space.
                     The complete size is returned.
                     0 is returned, if the iterator is not valid.
        */
        Data_SplitRecordLength GetSize()
        {
            // PTS 1117126 UH 2002-08-07 new
            if ( ! IsValid() )
                return 0;
            return m_RecordIterator.RestLength();
        }

        /*! -----------------------------------------------------------------------
        function:    operator ++
        description: The iterator is set to the previous record and
                     the internal record space is assigned.
        */
        Iterator& operator ++ ()
        {
            SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::Iterator::++", DataChain_Trace, 5);

            SAPDBERR_ASSERT_STATE( IsValid() );

            // reorder and reassign the split record space
            m_RecordSpace.Reassign ();
            m_RecordSpaceIsComplete = false;
            
            SAPDBERR_ASSERT_STATE( m_RecordSpace.CurrentPart() == 0 );
            
            PAGE& Page = m_RecordSpace.GetCurrentPage();
            Page.Use (m_RecordIterator);

            ++m_RecordIterator;
            
            if ( ! m_RecordIterator.IsValid() )
            {
                Data_PageNo NextPageNo = Page.NextPageNo();

                if ( NextPageNo.IsValid() )
                {
                    const Data_PageNo auxPageNo = Page.PageNo();

                    m_PAM.ReleasePage (Page);

                    if ( ! m_PAM.GetPage (Page, NextPageNo) )
                    {
                        m_RecordSpace.WriteToTrace("++");
                        RTE_Crash( Data_Exception(__CONTEXT__,
                                                  DATA_CHAIN_IS_INCONSISTENT,
                                                  SAPDB_ToString(auxPageNo),
                                                  SAPDB_ToString(Page.PageNo()) ));
                    }

                    Page.Begin (m_RecordIterator);
                }
                else
                    m_RecordSpace.Deassign ();
            }
            return *this;
        }

        /*! -----------------------------------------------------------------------
        function:    WriteToTrace
        arguments:   title [in]
        */
        virtual void WriteToTrace (const char* title = NULL) const
        {
            Kernel_VTrace() << "forwReadChainIter: Recordspace: "
                            << (m_RecordSpaceIsComplete?"complete":"not complete");
            Kernel_VTrace() << "BreakInfo: " << m_BreakInfo.lastPno << "." << m_BreakInfo.lastOffset
                            << ", accessmode: " << m_BreakInfo.lastAccessMode;
            m_RecordSpace.WriteToTrace(title);
        }
        
    private: // methods
    
        /*! -----------------------------------------------------------------------
        function:    Assign
        arguments:   PageNo         [in]
                     bToFirstRecord [in]
        returns:     true, if the given PageNo was found.
        description: The record space is newly assigned to the specified page
                     and dependent on bFirstFree the record is set to the first
                     record in the page or to the first free position.
        */
        bool Assign (Data_PageNo     PageNo,
                     Data_PageOffset Offset,
                     Data_AccessMode AccessMode)
        {
            SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::Iterator::Assign", DataChain_Trace, 5);

            if ( m_RecordSpace.IsAssigned() )
            {
                if ( m_RecordSpace.GetCurrentPage().PageNo() == PageNo 
                     &&
                     m_RecordSpace.GetCurrentPage().AccessMode() == AccessMode )
                    m_RecordSpace.Reassign();
                else
                {
                    m_RecordSpace.Deassign();
                    m_RecordSpace.SetAccessMode(AccessMode);
                    if ( ! m_PAM.GetPage ( m_RecordSpace.UseNextPart(),
                                           PageNo,
                                           AccessMode) )
                        return false;
                }
            }
            else
            {
                m_RecordSpace.SetAccessMode(AccessMode);
                if ( ! m_PAM.GetPage ( m_RecordSpace.UseNextPart(),
                                       PageNo,
                                       AccessMode) )
                    return false;
            }       
            SAPDBERR_ASSERT_STATE( m_RecordSpace.CurrentPart() == 0 );
            
            if ( Offset == 0 )
                m_RecordSpace.GetCurrentPage().Begin (m_RecordIterator);
            else if ( Offset == SAPDB_MAX_UINT2 )
                m_RecordSpace.GetCurrentPage().End (m_RecordIterator);
            else
                m_RecordSpace.GetCurrentPage().Use (m_RecordIterator, Offset);

            m_RecordSpaceIsComplete = false;
            
            return true;

        }

        /*! -----------------------------------------------------------------------
        function:    AssignRecordSpace
        arguments:   RecordIterator [in/out]
        description: The internal record space of the Split record is initialized.
        */
        void AssignRecordspace(Data_SplitSpaceForwardReadIterator& RecordIter)
        {
            SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::Iterator::AssignRecordSpace", DataChain_Trace, 5);

            if ( ! RecordIter.IsValid()
                 ||
                 RecordIter.PageIsEmpty() )
                 return;

            Data_SplitRecordLength TotalRestLength = RecordIter.RestLength();

            while ( TotalRestLength > 0 )
            {
                if ( m_RecordSpace.CurrentPart() > 0 )
                    TotalRestLength -= (RecordIter.Length() - sizeof(Data_SplitRecordLength));
                else
                    TotalRestLength -= RecordIter.Length();

                m_RecordSpace.SetPart (RecordIter.Length() - sizeof(Data_SplitRecordLength),
                                       *RecordIter);

                if ( TotalRestLength > 0 )
                {
                    ++RecordIter;

                    if ( ! RecordIter.IsValid() )
                    {
                        Data_PageNo NextPageNo = m_RecordSpace.GetCurrentPage().NextPageNo();

                        if ( ! NextPageNo.IsValid() )
                        {
                            m_RecordSpace.WriteToTrace("AssignRecordSpace");
                            RTE_Crash( Data_Exception(__CONTEXT__,
                                                      DATA_CHAIN_IS_INCONSISTENT,
                                                      SAPDB_ToString(m_RecordSpace.GetCurrentPage().PageNo()),
                                                      SAPDB_ToString(NextPageNo) ));
                        }

                        if ( ! m_PAM.GetPage (m_RecordSpace.UseNextPart(), NextPageNo) )
                        {
                            m_RecordSpace.WriteToTrace("AssignRecordSpace");
                            RTE_Crash( Data_Exception(__CONTEXT__,
                                                      DATA_CHAIN_IS_INCONSISTENT,
                                                      SAPDB_ToString(m_RecordSpace.GetCurrentPage().PageNo()),
                                                      SAPDB_ToString(NextPageNo) ));
                        }

                        m_RecordSpace.GetCurrentPage().Begin (RecordIter);
                    }
                }
            }
            if ( TotalRestLength != 0 )
            {
                m_RecordSpace.WriteToTrace("AssignRecordSpace");
                RTE_Crash( Data_Exception(__CONTEXT__,
                                          DATA_CHAIN_IS_INCONSISTENT,
                                          "TotalRestLength",
                                          SAPDB_ToString(TotalRestLength) ));
            }

            m_RecordSpaceIsComplete = true;
            
            SAPDBTRACE_IF (DataChain_Trace, 6, m_RecordSpace.WriteToTrace("AssignRecordSpace"));
        }
        
        /*! -----------------------------------------------------------------------
        function:    SetAssigned
        arguments:   IsAssigned [in]
        description: Set the isAssigned flag. Is Needed for Reserve.
        */
        void SetAssigned (bool IsAssigned)
        {
            m_RecordSpaceIsComplete = IsAssigned;
        }

        /*! -----------------------------------------------------------------------
        function:    GetRecordSpace
        returns:     RecordSpace& to internal RecordSpace.
        */
        RecordSpace &GetRecordSpace()
        {
            return m_RecordSpace;
        }
        
    private: // member
    
        /*! -----------------------------------------------------------------------
        variable:    m_PAM
        */
        Data_IPageAccessManager& m_PAM;

        /*! -----------------------------------------------------------------------
        variable:    m_RecordIterator
        */
        Data_SplitSpaceForwardReadIterator m_RecordIterator;
        
        /*! -----------------------------------------------------------------------
        variable:    m_RecordSpace
        description: This refers to the space of the current record.
        */
        RecordSpace m_RecordSpace;

        /*! -----------------------------------------------------------------------
        variable:    m_RecordSpaceIsComplete
        description: if this is true, the record space is correctly assigned.
        */
        bool m_RecordSpaceIsComplete;

        /*! -----------------------------------------------------------------------
        declaration: BreakInfo
        */
        struct BreakInfo
        {
            Data_PageNo     lastPno;
            Data_PageOffset lastOffset;
            Data_AccessMode lastAccessMode;
            BreakInfo()
            {
                lastPno.Invalidate();
                lastOffset     = 0;
                lastAccessMode = Data_ForRead;
            }
        };

        /*! -----------------------------------------------------------------------
        variable:    m_BreakInfo
        description: this is used by Break() and Continue()
        */
        BreakInfo m_BreakInfo;
    };
    /*! endclass: Iterator */

public:

    /*! ---------------------------------------------------------------------------
    function:    Data_ChainSplitSpaceForwardRead
    arguments:   RootId [in] this id defines the beginning of the page chain
    description: The parameter are stored.

    */
    Data_ChainSplitSpaceForwardRead (Data_IPageAccessManager &PAM,
                                      Data_PageId             &RootId)
    : Data_Chain<PAGE> (PAM, RootId)
    {}

    /*! ---------------------------------------------------------------------------
    function:    Data_ChainSplitSpaceForwardRead
    arguments:   RootId [in] this id defines the beginning of the page chain
                 LastPageNo [in]
    description: The parameter are stored.

    */
    Data_ChainSplitSpaceForwardRead (Data_IPageAccessManager &PAM,
                                     Data_PageId             &RootId,
                                     Data_PageNo              LastPageNo)
    : Data_Chain<PAGE> (PAM, RootId, LastPageNo)
    {}

    /*! ---------------------------------------------------------------------------
    function:    Create
    arguments:   RootPageIter [out]
    description: A new Chain is created. The PageIter points to the root.
    */
    bool Create (Data_ChainIterator<PAGE>& RootPageIter)
    {
        return Data_Chain<PAGE>::Create (RootPageIter); // PTS 1121659 UH 2003-04-30
    }

    /*! ---------------------------------------------------------------------------
    function:    Create
    description: A new Chain is created.
    */
    bool Create ()
    {
        Data_ChainIterator<PAGE> RootPageIter (this->GetPageAccessManager());
        return Create (RootPageIter); // PTS 1121659 UH 2003-04-30
    }

    /*! ---------------------------------------------------------------------------
    function:    CheckSpace
    returns:     bool
    description: true is returned if ReserveSpace() will find enough pages.
                 The task can be supended !
    */
    bool CheckSpace()
    {
		// PTS 1115593 UH 2002-05-07 new
        SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::Checkspace", DataChain_Trace, 5);
        return this->m_PAM.CheckSpace (1); // PTS 1115170 UH 2002-04-09
    }


    /*! ---------------------------------------------------------------------------
    function:    ReserveSpace
    arguments:   WantedSize          [in]  The amount of space needed.
                 FirstPartMiniumSize [in]  if set to 0, no minimum is given
                 RecordIterator      [out] A RecordIterator, which points to the knewly created space.
	returns:     SAPDB_UInt (addedPages)
    description: Space is reserved in the chain. The record iterator points to the Split space.
    */
    bool ReserveSpace (Data_SplitRecordLength  WantedSize,
                       Data_RecordLength       FirstPartMiniumSize,
                       Iterator&               RecordIterator,
                       SAPDB_UInt             &numberOfAddedPages ) // PTS 1121659 UH 2003-04-30
    {
        SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::ReserveSpace", DataChain_Trace, 5);
        
        SAPDBERR_ASSERT_STATE( WantedSize >= FirstPartMiniumSize );

        SAPDBTRACE_WRITELN (DataChain_Trace, 6, "Wanted: " << WantedSize << \
                                           ", FirstPartMin: " << FirstPartMiniumSize);

        // PTS 1115593 UH 2002-05-07

        if ( ! this->m_PAM.CheckSpace ( WantedSize/PAGE::MaxSpaceSize() ) ) // PTS 1121659 UH 2003-04-30
            return false; // PTS 1121659 UH 2003-04-30


        numberOfAddedPages = 0;
        // prepare the record space

        RecordIterator.Invalidate();
        
        if ( ! RecordIterator.Assign( this->LastPageNo(), SAPDB_MAX_UINT2, Data_ForUpdate ) )
        {
            this->InfoMessage("ChainSplitSpaceForwardRead");
            RTE_Crash( Data_Exception(__CONTEXT__,
                                      DATA_CHAIN_IS_INCONSISTENT,
                                      "-",
                                      SAPDB_ToString(this->LastPageNo()) ));
        }

        bool AllFits;
        bool FirstPartFits;
        
        RecordIterator.GetRecordSpace().GetCurrentPage().
            SpaceInfo (WantedSize, FirstPartMiniumSize, FirstPartFits, AllFits);
                                          
        if ( ! AllFits )
        {
            RecordIterator.Invalidate();
            if ( ! RecordIterator.Assign( this->LastPageNo(), SAPDB_MAX_UINT2, Data_ForStructureChange ) )
            {
                this->InfoMessage("ChainSplitSpaceForwardRead");
                RTE_Crash( Data_Exception(__CONTEXT__,
                                          DATA_CHAIN_IS_INCONSISTENT,
                                          "-",
                                          SAPDB_ToString(this->LastPageNo()) ));
            }
        }
        
        // fill the internal record space of the record iterator

        Data_RecordLength   CurrentReservedSize = 0;
        SAPDB_Byte         *CurrentSpace;

        do
        {
            RecordIterator.GetRecordSpace().GetCurrentPage().
                ReserveSpace (WantedSize, FirstPartMiniumSize,
                              CurrentReservedSize, CurrentSpace);

            SAPDBTRACE_WRITELN (DataChain_Trace, 6, \
                                "ReserveSpace(" << RecordIterator.GetRecordSpace().CurrentPart() \
                                << "): pno/CurrReserved/CurrReserved: " \
                                << RecordIterator.GetRecordSpace().GetCurrentPage().PageNo() << "/" \
                                << CurrentReservedSize << "/" << CurrentReservedSize);

            if ( CurrentReservedSize > 0 )
            {
                SAPDBERR_ASSERT_STATE( CurrentSpace != NULL );
                RecordIterator.GetRecordSpace().SetPart (CurrentReservedSize, CurrentSpace);
            }

            if ( WantedSize > 0 )
            {
                RecordIterator.GetRecordSpace().SetNewPageWasAppended();
                
                if ( CurrentReservedSize > 0 )
                {
					// PTS 1117126 UH 2002-08-07
					// UseNextPart() calls Resize() which reallocates the memory
					// So before UseNextPart() no reference into the memory of Space
                    // may be saved.
                    RecordSpace     &Space             = RecordIterator.GetRecordSpace();
                    const SAPDB_Int  partOfCurrentPage = Space.CurrentPart();
                    PAGE            &NextPage          = Space.UseNextPart();
                    PAGE            &CurrentPage       = Space.GetPage(partOfCurrentPage);

                    CurrentPage.Append (NextPage, this->LastPageNo());
                    NextPage.Begin(RecordIterator.m_RecordIterator);
                }
                else
                {
                    // the space was not big enough for FirstPartMiniumSize
                    PAGE& CurrentPage = RecordIterator.GetRecordSpace().GetCurrentPage();
                    PAGE  OldPage;
                    
                    OldPage.Assign (CurrentPage);
                    OldPage.Append (CurrentPage, this->LastPageNo());
                    
                    CurrentPage.Begin(RecordIterator.m_RecordIterator);
                }
                ++numberOfAddedPages;
            }
        }
        while ( WantedSize > 0 );

        RecordIterator.SetAssigned (true);

        SAPDBTRACE_IF (DataChain_Trace, 6, RecordIterator.GetRecordSpace().WriteToTrace("ReserveSpace"));

        return true;
    }

    /*! ---------------------------------------------------------------------------
    function:    UndoReserveSpace
    arguments:   RecordIterator      [in] The RecordIterator used by ReserveSpace
    returns:     none
    description: All changes done be ReserveSpace() are undone.
    */
    void UndoReserveSpace (Iterator& RecordIterator)
    {
		// PTS 1117126 UH 2002-08-07 new
        SAPDBTRACE_METHOD_DEBUG ("Data_ChainSplitSpaceForwardRead::UndoReserveSpace", DataChain_Trace, 5);
        SAPDBERR_ASSERT_STATE ( RecordIterator.IsValid() );

        this->LastPageNo() = (*RecordIterator).GetPage(0).PageNo();
    }

    /*! ---------------------------------------------------------------------------
    function:    Begin
    arguments:   RecordIterator [out]
    description: The RecordIterator is set to the first record in the chain.
                 The internal RecordSpace is assigned.
    */
    void Begin (Iterator&       RecordIterator,
                Data_AccessMode AccessMode)
    {
        if ( ! RecordIterator.Assign( this->RootId().PageNo(), 0, AccessMode) )
        {
            this->InfoMessage("ChainSplitSpaceForwardRead");
            RTE_Crash( Data_Exception(__CONTEXT__,
                                      DATA_CHAIN_IS_INCONSISTENT,
                                      "-",
                                      SAPDB_ToString(this->RootId().PageNo()) ));
        }
    }

};
/*! endclass: Data_ChainSplitSpaceForwardRead */

#undef FIRST_FREE

#endif // Data_ChainSplitSpaceForwardRead_HPP
