/*!***************************************************************************

  module      : Data_IPageAccessManager.hpp
  special area: DataAccess
  last changed: 2000-11-20
  last changed: 2000-11-20  12:00
  copyright   : Copyright by SAP AG, 2000
  responsible : UweH
  description : defines an interface to the Datacache



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#ifndef DATA_IPAGE_ACCESS_MANAGER_HPP
#define DATA_IPAGE_ACCESS_MANAGER_HPP

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "DataAccess/Data_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  FORWARD DECLARATION *
 *===========================================================================*/

/* --------------------------------------------------------------------------- */
class Data_Page;

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! ---------------------------------------------------------------------------
class:       Data_IPageAccessManager
description: this is an Interface to PageAccessManagers
*/
class Data_IPageAccessManager
{
public:

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    arguments:   Page         [out]
                 PageNo       [in]
                 RecoveryMode [in]
                 AccessMode   [in]
    returns:     true, if page could be found, else false
	description: The specified page is given in memory with the given request kind.

    The pointer to the frame in memory is given and a generic hint
    */
    virtual bool GetPage (Data_Page                   &Page,
                          const Data_PageNo&           PageNo,
                          const Data_PageRecoveryMode &RecoveryMode,
                          Data_AccessMode              AccessMode) = 0;

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    arguments:   Page         [in/out]
                 PageNo       [in]
                 AccessMode   [in]
    returns:     true, if page could be found, else false
	description: All parameters needed for getting the page are taken from Page.
    */
    virtual bool GetPage (Data_Page                   &Page,
                          const Data_PageNo           &PageNo,
                          Data_AccessMode              AccessMode) = 0;

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    arguments:   Page         [in/out]
                 PageNo       [in]
    returns:     true, if page could be found, else false
	description: All parameters needed for getting the page are taken from Page.
    */
    virtual bool GetPage (Data_Page                   &Page,
                          const Data_PageNo           &PageNo) = 0;

    /*! ---------------------------------------------------------------------------
    function:    ReleasePage
    arguments:   Page [in/out]
	description: The specified page is released out of a cache.
    */
    virtual void ReleasePage (Data_Page& Page) = 0;

    /*! ---------------------------------------------------------------------------
    function:    CheckSpace
    arguments:   pagecount [in]
    returns:     bool
	description: Before allocating new pages with NewPage this can be used to
                 determine if there is space enough.
                 pagecount is the amount of pages which should be allocated.
                 true is returned if space is available.
                 false is returned and the BasisError is set if there is not enough space.
    */
    virtual bool CheckSpace (SAPDB_UInt pagecount ) const = 0;

    /*! ---------------------------------------------------------------------------
    function:    NewPage
    arguments:   Page   [out]
                 Mode   [in]
	description: A new page with a unique PageId is given
    */
    virtual void NewPage (Data_Page&            Page,
                          Data_PageRecoveryMode Mode = Data_PageRecoveryMode()) = 0;

    /*! ---------------------------------------------------------------------------
    function:    FreePage
    arguments:   Page   [out]
	description: A frame with a given Page is removed. The Page may be reused.
    */
    virtual void FreePage (Data_Page& Page) = 0;

    /*! ---------------------------------------------------------------------------
    function:    FreePageNo
    arguments:   PageNo       [in]
                 RecoveryMode [in]
	description: An identified frame is removed.
    */
    virtual void FreePageNo (const Data_PageNo&           PageNo,
                             const Data_PageRecoveryMode& RecoveryMode) = 0;

    /*! ---------------------------------------------------------------------------
    function:    Invalidate
	description: This resets this handle
    */
    virtual void Invalidate() = 0;

    /*! ---------------------------------------------------------------------------
    function:    IsValid
    returns:     bool
	description: If this handle is invalid false is returned.
    */
    virtual bool IsValid() const = 0;

};
/*! endclass: Data_IPageAccessManager */


#endif // DATA_IPAGE_ACCESS_MANAGER_HPP
