/*!***************************************************************************

  module      : Data_Page.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-20
  last changed: 2000-11-20 12:00
  copyright   : Copyright by SAP AG, 2000
  description : defines class Data_Page



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef DATA_PAGE_HPP
#define DATA_PAGE_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"

#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"

#include "DataAccess/Data_BasePage.hpp"
#include "DataAccess/Data_Types.hpp"
#include "DataAccess/Data_IPageAccessManager.hpp"

#include "gbd00.h" // tbd_node

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Data_Page
    description: Access is offered to a basic database page.
    
                 This page can be accessed by the datacache.
 */
class Data_Page : public Data_BasePage
{
    friend class Data_PageAccessManager;
public:

    /*! ---------------------------------------------------------------------------
    function:    Data_Page
    description: default contructor
                 all members are initialized with harmless values
    */
    Data_Page ()
    : m_AccessMode     (Data_ForRead),
      m_ReleaseMode    (Data_Unchanged),
      m_LruChainInfo   (Data_LruNormal),
      m_pAccessManager (NULL)
    {}

    /*! ---------------------------------------------------------------------------
    function:    Data_Page
    arguments:   page [in/out]
    description: copy constructor
                 this is used by container_vector, which is used by Data_SplitSpace
                 Assign() is used intern.
    */
    Data_Page (Data_Page& page)
    {
        Assign(page);
    }

    /*! ---------------------------------------------------------------------------
    function:    ~Data_Page
    description: default destructor
                 Deassign() is called
    */
    ~Data_Page()
    {
        Deassign();
    }

    /*! ---------------------------------------------------------------------------
    function:    Assign
    arguments:   OldPage [in/out]
    description: OldPage moves to current page.
    
                 OldPage must be assigned. This must not be assigned.
                 After operation the assertions changed with each other.
    */
    virtual void Assign (Data_Page& OldPage)
    {
        SAPDBERR_ASSERT_STATE(  OldPage.IsAssigned());
        SAPDBERR_ASSERT_STATE(! IsAssigned()        );

        m_AccessMode     = OldPage.m_AccessMode;
        m_ReleaseMode    = OldPage.m_ReleaseMode;
        m_LruChainInfo   = OldPage.m_LruChainInfo;
        m_pAccessManager = OldPage.m_pAccessManager;

        SetRecoveryMode(OldPage.RecoveryMode());
        SetHint (OldPage.GetHint());
        SetFrame(OldPage.GetFrame());

        OldPage.m_pAccessManager = NULL;
        OldPage.SetHint (NULL);
        OldPage.ReleaseFrame();

        SAPDBERR_ASSERT_STATE(! OldPage.IsAssigned());
        SAPDBERR_ASSERT_STATE(  IsAssigned()        );
    }

    /*! ---------------------------------------------------------------------------
    function:    Swap
    arguments:   OtherPage [in/out]
    description: Both page handler swap their members.
    */
    virtual void Swap (Data_Page& Other)
    {
        Data_Page AuxPage;
        AuxPage.Assign (Other);
        Other.Assign   (*this);
        Assign         (AuxPage);
    }

    /*! ---------------------------------------------------------------------------
    function:    Clone
    arguments:   OtherPage [in/out]
    description: Both page handler point to the same page.
                 A cloned page must be destroyed with DeClone()
                 A cloned page has no PAM, so a DeAssign will not release it.
                 Instead of it saves (!)
    */
    virtual void Clone (Data_Page& Other)
    {
        SAPDBERR_ASSERT_STATE(  Other.IsAssigned() );
        SAPDBERR_ASSERT_STATE(! IsAssigned()       );

        m_AccessMode     = Other.m_AccessMode;
        m_ReleaseMode    = Other.m_ReleaseMode;
        m_LruChainInfo   = Other.m_LruChainInfo;
        m_pAccessManager = Other.m_pAccessManager;

        SetRecoveryMode( Other.RecoveryMode() );
        SetFrame       ( Other.GetFrame()     );
        SetHint        ( NULL                 ); // this indicates no action on release

        SAPDBERR_ASSERT_STATE( Other.PageNo() == PageNo() );
        SAPDBERR_ASSERT_STATE( Other.IsAssigned()         );
        SAPDBERR_ASSERT_STATE( IsAssigned()               );
    }

    /*! ---------------------------------------------------------------------------
    function:    WriteToTrace
    description: writes the main characteristica of the page handle to the knltrace
    arguments:   title string
    returns:     none
    */
    virtual void WriteToTrace (const SAPDB_Char *title = NULL) const;

    /*! ---------------------------------------------------------------------------
    function:    Message
    description: writes the main characteristica of the page handle to the knldiag
    arguments:   none
    returns:     none
    */
    virtual void Message (const SAPDB_Char *title = NULL) const;

    /*! ---------------------------------------------------------------------------
    function:    SetUnchanged
    */
    void SetUnChanged()
    {
        m_ReleaseMode = Data_Unchanged;
    }

    /*! ---------------------------------------------------------------------------
    function:    SetChanged
    */
    void SetChanged()
    {
        m_ReleaseMode = Data_Changed;
    }
    /*! ---------------------------------------------------------------------------
    function:    SetLruChainInfo
    */
    void SetLruChainInfo( const Data_LRUChainInfo   lruChainInfo )
    {
        m_LruChainInfo = lruChainInfo;
    }

    /*! ---------------------------------------------------------------------------
    function:    Deassign
    arguments:   isOK [in]
    description: The referenced page frame is released in the cache.
                 If isOK is true, then release for update, else release without
                 changes become persistent.

    POST: NOT IsAssigned
    */
    void Deassign (bool isOK = true)
    {
        if ( IsAssigned()
             &&
             m_pAccessManager != NULL
             &&
             GetHint() != NULL )
        {
            if ( ! isOK )
                SetUnChanged();
            m_pAccessManager->ReleasePage (*this);
            SAPDBERR_ASSERT_STATE(!IsAssigned());
            return;
        }
        
        m_pAccessManager = NULL;
        SetHint (NULL);
        ReleaseFrame();
        
        SAPDBERR_ASSERT_STATE(!IsAssigned());
    }

    /*! ---------------------------------------------------------------------------
    function:    ReleaseMode
    returns:     Data_ReleaseMode
    */
    Data_ReleaseMode ReleaseMode () const
    {
        return m_ReleaseMode;
    }

    /*! ---------------------------------------------------------------------------
    function:    AccessMode
    returns:     Data_AccessMode
    */
    Data_AccessMode AccessMode () const
    {
        return m_AccessMode;
    }

    /*! ---------------------------------------------------------------------------
    function:    SetAccessMode
    arguments:   AccessMode [in]
    description: This sets explicity the access mode.
    
                 PRE: NOT IsAssigned
    */
    void SetAccessMode (Data_AccessMode AccessMode)
    {
        m_AccessMode = AccessMode;
    }

    /*! ---------------------------------------------------------------------------
    function:    PageId
    arguments:   PageId [out]
    */
    void PageId (Data_PageId& PageId) const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        PageId.SetPageNo          (NodePtr()->nd_id());
        PageId.SetPageRecoveryMode( RecoveryMode() );
    }
    
    /*! ---------------------------------------------------------------------------
    function:    RootPageNo
    returns:     PageNo
    */
    Data_PageNo RootPageNo () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return STATIC_CAST(Data_PageNo,NodePtr()->nd_root());
    }

    /*! ---------------------------------------------------------------------------
    function:    PageNo
    returns:     PageNo
    */
    Data_PageNo PageNo () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());

        return STATIC_CAST(Data_PageNo, NodePtr()->nd_id());
    }

    /*! ---------------------------------------------------------------------------
    function:    Successor
    returns:     PageNo
    */
    Data_PageNo Successor () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());

        return STATIC_CAST(Data_PageNo, NodePtr()->nd_right());
    }

    /*! ---------------------------------------------------------------------------
    function:    SetRootPageNo
    arguments:   PageNo [in]
    */
    void SetRootPageNo (const Data_PageNo& RootPageNo)
    {
        SAPDBERR_ASSERT_STATE( IsAssigned()                            );
        SAPDBERR_ASSERT_STATE( Data_ForUpdate          == m_AccessMode ||
                               Data_ForStructureChange == m_AccessMode );

        NodePtr()->nd_root() = (tsp00_PageNo)RootPageNo;
    }

    /*! ---------------------------------------------------------------------------
    function:    InitializeFrame
    */
    virtual void InitializeFrame ()
    {
        // many parts of the Kernel72_Page frame were initialized in vbd13.cpp
        /*
        NodePtr()->nd_id()     = SAPDB_MAX_INT4;
        NodePtr()->nd_root()   = SAPDB_MAX_INT4;
        NodePtr()->nd_bottom() = MinFreeOffset();
        */
    }

    /*! ---------------------------------------------------------------------------
    function:    ReferenceIsValid
    returns:     true, if the reference points within the page frame.
    */
    bool ReferenceIsValid (SAPDB_Byte* Reference) const
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );

        return Reference >= FramePtr()
               &&
               Reference <= FramePtr() + FrameSize();
    }

    /*! ---------------------------------------------------------------------------
    function:    GetOffset
    arguments:   Ptr [in]
    returns:     offset in page
    description: PRECONDITION: pointer must point into page.
    */
    Data_PageOffset GetOffset (SAPDB_Byte* Reference) const
    {
        if ( ReferenceIsValid(Reference) )
            return Data_PageOffset(Reference - FramePtr());
        else
            return SAPDB_MAX_UINT4;
    }

    /*! ---------------------------------------------------------------------------
    function:    IsEmpty
    returns:     bool
    description: returns true if nothing was written into the page.
    */
    bool IsEmpty ()
    {
        return NodePtr()->nd_bottom() == sizeof (tbd_node_header)+1;
    }
    
    /*! ---------------------------------------------------------------------------
    function:    FreeSpace
    returns:     Size
    description: The size of freely usable space is returned.
    
                 This method uses nd_bottom to calculate the size.
    */
    Data_RecordLength FreeSpace ()
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        return MaxSpaceOffset() - FirstFreeOffset(FramePtr());
    }

    /*! ---------------------------------------------------------------------------
    function:    SpaceUsed
    returns:     Size
    description: The methods returns the occupancy in bytes of the page
    
                 This method uses nd_bottom to calculate the occupancy.
    */
    Data_RecordLength SpaceUsed () const
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );

        return( NodePtr()->nd_bottom() - BODY_BEG_BD00 );
    }

    /*! ---------------------------------------------------------------------------
    function:    ReserveSpace
    arguments:   Size [in]
    description: Reserve the given Size amount of space in the page body.
    
                 PRE: page is assigned
                 PRE: AccessMode is Data_ForUpdate or Data_ForStructureChange
                 PRE: Size is less or equal to the free space size.
                 This method uses nd_bottom to reserve the size.
    */
    SAPDB_Byte* ReserveSpace (Data_RecordLength Size)
    {
        SAPDBERR_ASSERT_STATE    ( IsAssigned()                               );
        SAPDBERR_ASSERT_STATE    ( Data_ForUpdate          == m_AccessMode || \
                                   Data_ForStructureChange == m_AccessMode );
        SAPDBERR_ASSERT_ARGUMENT ( Size <= FreeSpace()                        );

        NodePtr()->nd_bottom() += Size;
        return FramePtr() + (NodePtr()->nd_bottom()-1) - Size;
    }

    /*! ---------------------------------------------------------------------------
    function:    SetFirstFreeOffset
    arguments:   offset [in]
    description: The first free offset in the page is set to a new value.
    
                 PRE: page is assigned
                 PRE: AccessMode is Data_ForUpdate or Data_ForStructureChange
                 This method uses nd_bottom.
    */
    void SetFirstFreeOffset (Data_PageOffset offset)
    {
        SAPDBERR_ASSERT_STATE    ( IsAssigned()                            );
        SAPDBERR_ASSERT_STATE    ( Data_ForUpdate          == m_AccessMode || \
                                   Data_ForStructureChange == m_AccessMode );

        NodePtr()->nd_bottom() = offset+1; // this is a position not an offset
    }

    /*! ---------------------------------------------------------------------------
    function:    FrameSize
    returns:     size in bytes of a frame
    */
    static SAPDB_UInt FrameSize ()
    {
        return sizeof (tbd_node);
    }

    /*! ---------------------------------------------------------------------------
    function:    MaxSpaceSize
    returns:     size in bytes
    description: The maximum size of bytes which can be written with any contents
    */
    static SAPDB_UInt MaxSpaceSize ()
    {
        // PTS 1121659 UH 2003-04-30 new
        return   sizeof (tbd_node) /*really dirty, but "m_Frame.Capacity()" cannot be used now */
               - sizeof (tbd_node_header)
               - sizeof (PersistentChecksumTrailer);
    }

    /*! ---------------------------------------------------------------------------
    function:    FirstFreeOffset
    arguments:   pPage [in]
    returns:     Offset
    description: Returns the offset, of the first free usable byte of space.
    */
    static Data_PageOffset FirstFreeOffset (const SAPDB_Byte* pPage)
    {
        const SAPDB_Int4 &bottom = (reinterpret_cast <const tbd_node*>(pPage))->nd_bottom();
        SAPDBERR_ASSERT_STATE ( bottom >= 1 );
        return static_cast <Data_PageOffset> (bottom-1);
    }

    /*! ---------------------------------------------------------------------------
    function:    MinSpaceOffset
    returns:     Offset
    description: The minimum offset from page begin, which can be used as free space.
    */
    static Data_PageOffset MinSpaceOffset ()
    {
        return sizeof (tbd_node_header);
    }

    /*! ---------------------------------------------------------------------------
    function:    MaxSpaceOffset
    returns:     Offset
    description: The maximum offset from page begin, which can be used as free space.
    */
    static Data_PageOffset MaxSpaceOffset ()
    {
        return sizeof(tbd_node) /*really dirty, but "m_Frame.Capacity()" cannot be used now */
               - sizeof(PersistentChecksumTrailer);
    }

protected:

    /*! ---------------------------------------------------------------------------
    function:    FramePtr
    */
    SAPDB_Byte * FramePtr()
    {
        // return REINTERPRET_CAST (SAPDB_Byte*, &(m_Frame[0]));
        return REINTERPRET_CAST (SAPDB_Byte*, m_Frame.GetPtrAt(0, 0));
    }

    /*! ---------------------------------------------------------------------------
    function:    FramePtr
    */
    const SAPDB_Byte * FramePtr() const
    {
        // return REINTERPRET_CAST (SAPDB_Byte*, &(m_Frame[0]));
        return REINTERPRET_CAST (const SAPDB_Byte*, m_Frame.GetPtrAt(0, 0));
    }

    /*! ---------------------------------------------------------------------------
    function:    NodePtr
    */
    tbd_node* NodePtr()
    {
        return REINTERPRET_CAST (tbd_node*, FramePtr());
    }

    /*! ---------------------------------------------------------------------------
    function:    NodePtr
    */
    const tbd_node* NodePtr() const
    {
        return REINTERPRET_CAST (const tbd_node*, FramePtr());
    }

    /*! ---------------------------------------------------------------------------
    function:    SetNodePtr
    */
    void SetNodePtr(tbd_node* pNode)
    {
        m_Frame.Assign (pNode, sizeof(tbd_node));
    }

private: // forbidden methods

    Data_Page& operator = (Data_Page& old)
    {
        // FORBIDDEN TO USE BECAUSE THERE SHOULD ONLY EXIST ONE HANDLE
        // FOR ONE DATA PAGE IN DATA CACHE
        // EXCEPTION: Clone() for Lock()
        return *this;
    }

protected:

    /*! ---------------------------------------------------------------------------
    variable:    m_AccessMode
    */
    Data_AccessMode m_AccessMode;

    /*! ---------------------------------------------------------------------------
    variable:    m_ReleaseMode
    */
    Data_ReleaseMode m_ReleaseMode;

    /*! ---------------------------------------------------------------------------
    variable:    Data_LRUChainInfo
                 specifies how the page will be positioned within the data cache 
                 LRU List Data_lruNormal leaves the page at the current position 
                 assumed the page is not in the last 10 percent of the LRU list. 
                 In this case the page will be rechained to the begining. 
                 Data_lruMiddle puts the page in front of the last 10 percent of 
                 the LRU list. Data_lruLast inserts the page at the end.
    */
    Data_LRUChainInfo m_LruChainInfo;

    /*! ---------------------------------------------------------------------------
    variable:    m_pAccessManager
    */
    Data_IPageAccessManager* m_pAccessManager;
};
/*! endclass: Data_Page */

#endif // DATA_PAGE_HPP
