/*!***************************************************************************

  module      : Data_PageAccessManager.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-20
  last changed: 2000-11-20  12:00
  copyright   : Copyright by SAP AG, 2000
  description : encapsulates the page access manager (Data_Cache)



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef DATA_PAGE_ACCESS_MANAGER_HPP
#define DATA_PAGE_ACCESS_MANAGER_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h" // tgg00_TransContext
#include "gbd00.h" // tbd_current_tree

#include "DataAccess/Data_IPageAccessManager.hpp"
#include "DataAccess/Data_Page.hpp"

#include "File/File_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! ---------------------------------------------------------------------------
class:       Data_PageAccessManager 
description: encapsulates the page access manager (Data_Cache)
*/
class Data_PageAccessManager : public Data_IPageAccessManager
{
public:
    /*! ---------------------------------------------------------------------------
    function:    Data_PageAccessManager
    arguments:   transcontext [in/out]
                 filetype     [in]
                 rootpageno   [in]
    description: This initializes a context to access pages of a special type.
    */
    Data_PageAccessManager (tgg00_TransContext  &TransContext,
                            File_Type           FileType,
                            Data_PageNo         RootPageNo);

    /*! ---------------------------------------------------------------------------
    function:    Data_PageAccessManager
    arguments:   current [in/out]
    description: This initializes a context to access pages for a given file
    */
    Data_PageAccessManager ( tbd_current_tree   &current )
    :m_Current( current )
    {}

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    arguments:   Page         [out]
                 PageNo       [in]
                 RecoveryMode [in]
                 AccessMode   [in]
    returns:     true, if page could be found, else false
    description: The specified page is given in memory with the given request kind.

                 The pointer to the frame in memory is given and a generic hint
                 On any error from vbd13 false is returned.
    */
    virtual bool GetPage (Data_Page                   &Page,
                          const Data_PageNo&           PageNo,
                          const Data_PageRecoveryMode &RecoveryMode,
                          Data_AccessMode              AccessMode);

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    arguments:   Page         [in/out]
                 PageNo       [in]
                 AccessMode   [in]
    returns:     true, if page could be found, else false
    description: All parameters needed for getting the page are taken from Page.
                 On any error from vbd13 false is returned.
    */
    virtual bool GetPage (Data_Page                   &Page,
                          const Data_PageNo           &PageNo,
                          Data_AccessMode              AccessMode);

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    arguments:   Page         [in/out]
                 PageNo       [in]
    returns:     true, if page could be found, else false
    description: All parameters needed for getting the page are taken from Page.
                 On any error from vbd13 false is returned.
    */
    virtual bool GetPage (Data_Page                   &Page,
                          const Data_PageNo           &PageNo);

    /*! ---------------------------------------------------------------------------
    function:    ReleasePage
    arguments:   Page [in/out]
    description: The specified page is released out of a cache.

    On any error from vbd13 execution is aborted.
    */
    virtual void ReleasePage (Data_Page &Page);

    /*! ---------------------------------------------------------------------------
    function:    CheckSpace
    arguments:   pagecount [in]
    returns:     bool
    description: Before allocating new pages with NewPage this can be used to
                 determine if there is space enough.
                 pagecount is the amount of pages which should be allocated.
                 true is returned if space is available.
                 false is returned and the BasisError is set if there is not enough space.
    */
    virtual bool CheckSpace (SAPDB_UInt pagecount ) const;

    /*! ---------------------------------------------------------------------------
    function:    NewPage
    arguments:   Page   [out]
                 Mode   [in]
    description: A new page with a unique PageId is given

    On any error from vbd13 execution is aborted.
    */
    virtual void NewPage (Data_Page&            Page,
                          Data_PageRecoveryMode Mode = Data_PageRecoveryMode());

    /*! ---------------------------------------------------------------------------
    function:    FreePage
    arguments:   Page   [out]
    description: A frame with a given Page is removed. The Page may be reused.

    On any error from vbd13 execution is aborted.
    */
    virtual void FreePage (Data_Page& Page);

    /*! ---------------------------------------------------------------------------
    function:    FreePageNo
    arguments:   PageNo       [in]
                 RecoveryMode [in]
    description: An identified frame is removed.

    On any error from vbd13 execution is aborted.
    */
    virtual void FreePageNo (const Data_PageNo&           PageNo,
                             const Data_PageRecoveryMode& RecoveryMode);

    /*! ---------------------------------------------------------------------------
    function:    Invalidate
    description: This resets this handle
    */
    virtual void Invalidate();

    /*! ---------------------------------------------------------------------------
    function:    IsValid
    returns:     bool
    description: If this handle is invalid false is returned.
    */
    virtual bool IsValid() const;

private:

    /*! ---------------------------------------------------------------------------
    function:    RegisterToPage
    arguments:   Page [in/out]
                 Hint [in]
                 Mode [in]
    description: The Accessmanagers can save themself here,
                 so the page destructor can do a release.
    */
    void RegisterToPage (Data_Page&                   Page,
                         tbd_node_ptrs&               Nptrs,
                         const Data_PageRecoveryMode& RecoveryMode,
                         Data_AccessMode              Mode);

    /*! ---------------------------------------------------------------------------
    variable:    m_Current
    description: this is used by the vbd13 Interface.
    */
    tbd_current_tree m_Current;
};
/*! endclass: Data_PageAccessManager */


#endif // DATA_PAGE_ACCESS_MANAGER_HPP
