/*!***************************************************************************

  module      : Data_PageFixSizeSpace.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-22
  last changed: 2000-11-22 12:00
  copyright   : Copyright by SAP AG, 2000
  description : Defines classes Data_PageFixSizeSpace,  

    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef Data_PageFixSizeSpace_HPP
#define Data_PageFixSizeSpace_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "DataAccess/Data_ChainPage.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define ALIGNMENT 4

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Data_PageFixSizeSpace
    description: The page can store Split records.
 */
class Data_PageFixSizeSpace : public Data_ChainPage
{

public:

    /*! ---------------------------------------------------------------------------
    declaration: some types
    */
    typedef SAPDB_UInt2 FixRecordLength;
    typedef SAPDB_UInt4 RecordCount;

public:

    /*! ---------------------------------------------------------------------------
    class:       Iterator
    description: This iterator points to records in a page Data_PageFixSizeSpace.
    */
    class Iterator
    {
    friend
        class Data_PageFixSizeSpace;
        
    public:

        /*! -----------------------------------------------------------------------
        function:    Iterator
        */
        Iterator ()
        : m_pFrame       (NULL),
          m_RecordOffset (0),
          m_RecordSize   (0)
          
        {}

        /*! -----------------------------------------------------------------------
        function:    IsValid
        returns:     true, if it can be dereferenced.
        */
        bool IsValid() const
        {
            return m_pFrame                    != NULL                         &&
                   m_RecordOffset              >= Data_Page::MinSpaceOffset () &&
                   m_RecordOffset+m_RecordSize <= Data_Page::MaxSpaceOffset () &&       // PTS 1114647 UH 2002-03-05
                   m_RecordOffset+m_RecordSize <= Data_Page::FirstFreeOffset(m_pFrame); // PTS 1114647 UH 2002-03-05
        }

        /*! -----------------------------------------------------------------------
        function:    Invalidate
        description: resets all internal members. IsValid will return false.
        */
        void Invalidate()
        {
            m_pFrame       = NULL;
            m_RecordOffset = 0;
            m_RecordSize   = 0;
        }

        /*! -----------------------------------------------------------------------
        function:    GetPosition
        returns:     Data_PageOffset of current position
        */
        Data_PageOffset GetPosition () const
        {
            return m_RecordOffset;
        }

        /*! -----------------------------------------------------------------------
        function:    GetRecordId
        returns:     SAPDB_UInt, MAX_UINT if invalid
        */
        RecordCount GetRecordId () const
        {
            if ( ! IsValid() )
                return SAPDB_MAX_UINT4;
            SAPDBERR_ASSERT_STATE( (m_RecordOffset - Data_Page::MinSpaceOffset()) % m_RecordSize == 0 );
            return ((m_RecordOffset-Data_Page::MinSpaceOffset()) / m_RecordSize)+1;
        }

        /*! -----------------------------------------------------------------------
        function:    SetRecordId
        arguments:   recordid
        */
        void SetRecordId (RecordCount recordid)
        {
            m_RecordOffset = Data_PageFixSizeSpace::Offset (recordid, m_RecordOffset);
            SAPDBERR_ASSERT_STATE( IsValid() );
        }

        /*! -----------------------------------------------------------------------
        function:    operator *
        returns:     SAPDB_Byte*
        description: This returns the record, which is currently referenced.

        If the iterator is not valid NULL is returned.
        */
        SAPDB_Byte* operator * ()
        {
            if ( ! IsValid() )
                return NULL;
            else
                return m_pFrame + m_RecordOffset;
        }

        /*! -----------------------------------------------------------------------
        function:    operator ++
        description: Sets the iterator to the next record in page.
                     If it points after end or before begin it is invalid.
        */
        Iterator& operator ++ ()
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            m_RecordOffset += m_RecordSize;
            return *this;
        }

        /*! -----------------------------------------------------------------------
        function:    operator ++
        description: Sets the iterator to the next record in page.
                     If it points after end or before begin it is invalid.
        */
        Iterator& operator -- ()
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            m_RecordOffset -= m_RecordSize;
            return *this;
        }

    protected: // member

        /*! -----------------------------------------------------------------------
        variable:    m_pFrame
        description: A page pointer.
        */
        SAPDB_Byte* m_pFrame;
        /*! -----------------------------------------------------------------------
        variable:    m_RecordOffset
        description: This refers to the current record: (m_pFrame + m_RecordOffset)
        */
        Data_PageOffset m_RecordOffset;
        /*! -----------------------------------------------------------------------
        variable:    m_RecordSize
        description: The fix record length in the page.
        */
        Data_RecordLength m_RecordSize;
    };
    /*! endclass: Data_SplitSpaceIterator */



public:

    /*! ---------------------------------------------------------------------------
    function:    InitializeFrame
    */
    virtual void InitializeFrame ()
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        Data_ChainPage::InitializeFrame();
        NodePtr()->nd_left() = SAPDB_MAX_INT4;
        SetFixRecordLength (0);
        SetRecordCount     (0);
    }

    /*! ---------------------------------------------------------------------------
    function:    AlignRecordLength
    arguments:   Length [in/out]
    description: The RecordSize is properly aligned.
    */
    static FixRecordLength AlignRecordSize (FixRecordLength Length)
    {
        return ((Length-1) & ~(ALIGNMENT-1)) + ALIGNMENT;
    }
    
    /*! -----------------------------------------------------------------------
    function:    GetFixRecordLength
    returns:     length of record in the page
    */
    FixRecordLength GetFixRecordLength () const
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        return NodePtr()->nd_filler2();
    }

    /*! -----------------------------------------------------------------------
    function:    SetFixRecordLength
    arguments:   Sets the length of all records in the page.
                 The given length is aligned before.
    */
    void SetFixRecordLength (FixRecordLength Length)
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        FixRecordLength* aux
            = REINTERPRET_CAST(FixRecordLength*,&(NodePtr()->nd_filler2()));
        *aux = AlignRecordSize(Length);
    }

    /*! -----------------------------------------------------------------------
    function:    GetRecordCount
    returns:     the actual record count.
    */
    RecordCount GetRecordCount () const
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        return NodePtr()->nd_record_cnt();
    }

    /*! -----------------------------------------------------------------------
    function:    DecrementRecordCount
    */
    void DecrementRecordCount ()
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        --(NodePtr()->nd_record_cnt());
        if ( GetRecordCount() == 0 )
            SetFirstFreeOffset (Data_Page::MinSpaceOffset());
    }

    /*! -----------------------------------------------------------------------
    function:    PrevPageNo
    returns:     PageNo
    description: The page id of the prev page is returned.
    
                 If no prev page is available, PageId.IsValid() returns false.
    */
    Data_PageNo PrevPageNo ()
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return STATIC_CAST(Data_PageNo,NodePtr()->nd_left());
    }

    /*! -----------------------------------------------------------------------
    function:    SetPrevPageNo
    arguments:   PageNo [in]
    */
    void SetPrevPageNo (const Data_PageNo& PageNo)
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        NodePtr()->nd_left() = (tsp00_PageNo)PageNo;
    }

    /*! -----------------------------------------------------------------------
    function:    Append
    description: NewPage [out]
    */
    void Append (Data_PageFixSizeSpace &NewPage,
                 Data_PageNo&           LastPageNo)
    {
        Data_ChainPage::Append (NewPage, LastPageNo);

        NewPage.SetPrevPageNo (PageNo());

        if ( NewPage.NextPageNo().IsValid() )
        {
            // THIS<->NEW<->NEXT
            Data_PageFixSizeSpace NextPage;
            
            (void) m_pAccessManager->GetPage (NextPage,
                                              NewPage.NextPageNo());
            NextPage.SetPrevPageNo( NextPageNo() );
        }
    }

    /*! ---------------------------------------------------------------------------
    function:    SetToRecord
    arguments:   RecordId [in]
                 Iterator [out]
    description: The iterator is set to the given record on this page.
                 If RecordId is not in the valid range, then the iterator becomes
                 invalid.
    */
    void SetToRecord (RecordCount RecordId,
                      Iterator&   Iter)
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        SAPDBERR_ASSERT_STATE( Iter.m_pFrame == FramePtr() );
        Iter.m_RecordOffset = Offset (RecordId, GetFixRecordLength());
    }

    /*! ---------------------------------------------------------------------------
    function:    Begin
    arguments:   Iterator [in/out]
    description: The iterator is set to first record on this page.
    */
    void Begin (Iterator& Iter)
    {
        Iter.m_pFrame     = FramePtr();
        Iter.m_RecordSize = GetFixRecordLength();
        SetToRecord (1, Iter);
    }
    
    /*! ---------------------------------------------------------------------------
    function:    End
    arguments:   Iterator [in/out]
    description: The iterator is set to last record on this page.
    */
    void End (Iterator& Iter)
    {
        Iter.m_pFrame       = FramePtr();
        Iter.m_RecordSize   = GetFixRecordLength();
        Iter.m_RecordOffset = FirstFreeOffset(Iter.m_pFrame) - GetFixRecordLength();
    }
    
    /*! ---------------------------------------------------------------------------
    function:    ReserveSpace
    arguments:   Iter [out] - points to the new space.
    */
    void ReserveSpace (Iterator& Iter)
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() && GetFixRecordLength() > 0 );

        Iter.Invalidate();
        
        const FixRecordLength LengthToReserve = GetFixRecordLength();
        
        if ( LengthToReserve  > FreeSpace() )
            return;

        RecordCount aux = GetRecordCount();
        ++aux;
        SetRecordCount(aux);

        Data_Page::ReserveSpace (LengthToReserve);

        End (Iter);
    }

    /*! ---------------------------------------------------------------------------
    function:    Offset
    arguments:   RecordId [in]
    returns:     Offset to given RecordId.
    */
    static Data_PageOffset Offset (RecordCount     RecordId,
                                   FixRecordLength RecordLength)
    {
        return Data_Page::MinSpaceOffset() + (RecordLength * (RecordId-1));
    }
    
private: // methods

    /*! -----------------------------------------------------------------------
    function:    SetRecordCount
    arguments:   RecordCount [in]
    */
    void SetRecordCount (RecordCount Count)
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );
        RecordCount* aux
            = REINTERPRET_CAST(RecordCount*,&(NodePtr()->nd_record_cnt()));
        *aux = Count;
    }
};
/*! endclass: Data_PageFixSizeSpace */

/*===========================================================================*
 *  UNDEFINES                                                                *
 *===========================================================================*/

#undef ALIGNMENT

#endif // Data_PageFixSizeSpace_HPP
