/*!***************************************************************************

  module      : Data_PageSplitSpace.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-22
  last changed: 2000-11-22 12:00
  copyright   : Copyright by SAP AG, 2000
  description : Defines classes Data_PageSplitSpace,
                                Data_SplitSpaceIterator,
                                Data_SplitSpaceBackwardReadIterator,
                                Data_SplitSpaceForwardReadIterator
  


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef Data_PageSplitSpace_HPP
#define Data_PageSplitSpace_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "DataAccess/Data_ChainPage.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define ALIGNMENT 4

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! ---------------------------------------------------------------------------
class:       Data_SplitSpaceIterator
description: This iterator points to records in a page Data_PageSplitSpace.

There can be only three states for the iterator:
1. It is invalid: then it does not point to a valid offset within min and max.
   It cannot be dereferenced.
2. If page is empty it can be dereferenced, but it returns NULL
3. If page is empty it points to a record.
*/
class Data_SplitSpaceIterator
{
friend class Data_PageSplitSpace;

public:

    /*! -----------------------------------------------------------------------
    function:    Data_SplitSpaceIterator
    */
    Data_SplitSpaceIterator ()
    : m_RecordOffset (0),
      m_pFrame       (NULL)
    {}

    /*! -----------------------------------------------------------------------
    function:    IsValid
    returns:     true, if it can be dereferenced.
    */
    bool IsValid() const
    {
        return m_pFrame       != NULL                        &&
               m_RecordOffset >= Data_Page::MinSpaceOffset() &&
               m_RecordOffset <= Data_Page::MaxSpaceOffset() &&
               ! PageIsEmpty();
    }

    /*! -----------------------------------------------------------------------
    function:    Invalidate
    description: resets all internal members. IsValid will return false.
    */
    void Invalidate()
    {
        m_pFrame       = NULL;
        m_RecordOffset = 0;
    }
    
    /*! -----------------------------------------------------------------------
    function:    GetPosition
    returns:     Data_PageOffset of current position
    */
    Data_PageOffset GetPosition () const
    {
        return m_RecordOffset;
    }

    /*! -----------------------------------------------------------------------
    function:    PageIsEmpty
    returns:     true, if no record is on the page.
    */
    bool PageIsEmpty() const
    {
        return Data_Page::FirstFreeOffset(m_pFrame) == Data_Page::MinSpaceOffset();
    }

    /*! -----------------------------------------------------------------------
    function:    RestLength
    returns:     length of all Split parts of the record from current on.
    */
    Data_SplitRecordLength RestLength ()
    {
        SAPDBERR_ASSERT_STATE( IsValid() && ! PageIsEmpty() );
        return LengthValue (m_RecordOffset);
    }

    /*! -----------------------------------------------------------------------
    function:    operator *
    returns:     SAPDB_Byte*
    description: This returns the record, which is currently referenced.

    If the iterator is not valid NULL is returned.
    */
    virtual SAPDB_Byte* operator * () = 0;

    /*! -----------------------------------------------------------------------
    function:    Length
    returns:     Length of current part of the Split record.
    */
    virtual Data_RecordLength Length () = 0;

    /*! -----------------------------------------------------------------------
    function:    LengthIsValid
    returns:     true, if the determined value of Length() fits into the current page
    */
    virtual bool LengthIsValid() = 0;

private: // functions

    /*! -----------------------------------------------------------------------
    function:    Begin
    arguments:   pPage  [in] a pointer to a frame
    description: The internal frame pointer is assigned and the internal record pointer is set.
    
    The Offset is set to the first record.
    If the page is empty, the iterator points to the first free space.
    In this case the iterator dereferences to NULL.
    */
    void Begin (SAPDB_Byte* pPage)
    {
        m_pFrame       = pPage;
        m_RecordOffset = Data_Page::MinSpaceOffset();
    }

    /*! -----------------------------------------------------------------------
    function:    End
    arguments:   pPage  [in] a pointer to a frame
    description: The internal frame pointer is assigned and the internal record pointer is set.
    
    The Offset is set to the last record.
    If the page is empty, the iterator points to the first free space.
    In this case the iterator dereferences to NULL.
    */
    void End (SAPDB_Byte* pPage)
    {
        m_pFrame       = pPage;
        m_RecordOffset = Data_Page::FirstFreeOffset(pPage);
    }

    /*! -----------------------------------------------------------------------
    function:    ReAssign
    arguments:   pPage  [in] a new frame pointer
    description: Only the frame pointer is newly assigned.
    */
    void ReAssign (SAPDB_Byte* pPage)
    {
        m_pFrame = pPage;
    }

    /*! -----------------------------------------------------------------------
    function:    ReAssign
    arguments:   pPage  [in] a new frame pointer
                 Offset [in] the new offset in the new frame.
    */
    void ReAssign (SAPDB_Byte*     pPage,
                   Data_PageOffset Offset)
    {
        m_pFrame       = pPage;
        m_RecordOffset = Offset;
        SAPDBERR_ASSERT_STATE( IsValid() );
    }

    /*! -----------------------------------------------------------------------
    function:    LengthValue
    arguments:   Offset [in]
    returns:     the value of .
    */
    virtual Data_SplitRecordLength LengthValue (Data_PageOffset Offset) = 0;

protected: // member

    /*! -----------------------------------------------------------------------
    variable:    m_pFrame
    description: A page pointer.
    */
    SAPDB_Byte* m_pFrame;
    
    /*! -----------------------------------------------------------------------
    variable:    m_RecordOffset
    description: This refers to the current record: (m_pFrame + m_RecordOffset)
    */
    Data_PageOffset m_RecordOffset;
};
/*! endclass: Data_SplitSpaceIterator */

/*! ---------------------------------------------------------------------------
class:       Data_SplitSpaceBackwardReadIterator
description: This iterator points to records in this page.

There can be only three states for the iterator:
1. It is invalid: then it does not point to a valid offset within min and max.
   It cannot be dereferenced.
2. If page is empty it can be dereferenced, but it returns NULL
3. If page is empty it points to a record.
*/
class Data_SplitSpaceBackwardReadIterator : public Data_SplitSpaceIterator
{
public:

    /*! -----------------------------------------------------------------------
    function:    operator *
    returns:     SAPDB_Byte*
    description: This returns the record, which is currently referenced.

    If the iterator is not valid NULL is returned.
    */
    virtual SAPDB_Byte* operator * ()
    {
        SAPDBERR_ASSERT_STATE( IsValid() );
        if ( PageIsEmpty() )
            return NULL;
        else
            return m_pFrame + m_RecordOffset - Length();
    }

    /*! -----------------------------------------------------------------------
    function:    Length
    returns:     Length of current part of the Split record.
    */
    virtual Data_RecordLength Length ()
    {
        SAPDBERR_ASSERT_STATE( IsValid() && ! PageIsEmpty() );
        
        const Data_SplitRecordLength RestRecordLength
            = RestLength();
            
        const Data_SplitRecordLength RestSpaceInPage
            = m_RecordOffset - Data_Page::MinSpaceOffset();

        if ( RestSpaceInPage >= RestRecordLength )
            return RestRecordLength;
        else
            return RestSpaceInPage;
    }

    /*! -----------------------------------------------------------------------
    function:    LengthIsValid
    returns:     true, if the determined value of Length() fits into the current page
    */
    virtual bool LengthIsValid()
    {
        // PTS 1113270 UH 2001-12-28 new
        
        if ( ! IsValid() )
            return false;
        
        const SAPDB_UInt testoffset = m_RecordOffset - Length();
        
        return testoffset >= Data_Page::MinSpaceOffset()
               &&
               testoffset <= Data_Page::MaxSpaceOffset();
    }

    /*! -----------------------------------------------------------------------
    function:    operator --
    description: The iterator is set to the previous record.
    */
    Data_SplitSpaceBackwardReadIterator& operator -- ()
    {
        if ( m_RecordOffset - Length() <= Data_Page::MinSpaceOffset() )
            Invalidate();
        else
            m_RecordOffset -= Length();
        return *this;
    }

private:

    /*! -----------------------------------------------------------------------
    function:    LengthValue
    arguments:   Offset [in]
    returns:     Length of given record.
    */
    virtual Data_SplitRecordLength LengthValue (Data_PageOffset Offset)
    {
        return *(REINTERPRET_CAST(Data_SplitRecordLength*,(m_pFrame + Offset - sizeof (Data_SplitRecordLength))));
    }
};
/*! endclass: Data_SplitSpaceBackwardReadIterator */


/*! ---------------------------------------------------------------------------
class:       Data_SplitSpaceForwardReadIterator
description: This iterator points to records in this page.

There can be only three states for the iterator:
1. It is invalid: then it does not point to a valid offset within min and max.
   It cannot be dereferenced.
2. If page is empty it can be dereferenced, but it returns NULL
3. If page is empty it points to a record.
*/
class Data_SplitSpaceForwardReadIterator : public Data_SplitSpaceIterator
{
public:

    /*! -----------------------------------------------------------------------
    function:    operator *
    returns:     SAPDB_Byte*
    description: This returns the record, which is currently referenced.

    If the iterator is not valid NULL is returned.
    */
    virtual SAPDB_Byte* operator * ()
    {
        SAPDBERR_ASSERT_STATE( IsValid() );
        if ( PageIsEmpty() )
            return NULL;
        else
            return m_pFrame + m_RecordOffset + sizeof(Data_SplitRecordLength);
    }

    /*! -----------------------------------------------------------------------
    function:    Length
    returns:     Length of current part of the Split record.
    */
    virtual Data_RecordLength Length ()
    {
        SAPDBERR_ASSERT_STATE( IsValid() && ! PageIsEmpty() );
        const Data_SplitRecordLength RestLen = RestLength();

        if ( m_RecordOffset + RestLen >= Data_Page::MaxSpaceOffset() )
            return Data_Page::MaxSpaceOffset() - m_RecordOffset;
        
        return RestLen;
    }

    /*! -----------------------------------------------------------------------
    function:    LengthIsValid
    returns:     true, if the determined value of Length() fits into the current page
    */
    virtual bool LengthIsValid()
    {
        // PTS 1113270 UH 2001-12-28 new
        
        if ( ! IsValid() )
            return false;
        
        const SAPDB_UInt testoffset = m_RecordOffset + Length();
        
        return testoffset >= Data_Page::MinSpaceOffset()
               &&
               testoffset <= Data_Page::MaxSpaceOffset();
    }

    /*! -----------------------------------------------------------------------
    function:    operator ++
    description: The iterator is set to the next record.
    */
    Data_SplitSpaceForwardReadIterator& operator ++ ()
    {
        if ( m_RecordOffset + Length() >= Data_Page::FirstFreeOffset(m_pFrame) )
            Invalidate();
        else
            m_RecordOffset += Length();
        return *this;
    }

private:

    /*! -----------------------------------------------------------------------
    function:    LengthValue
    arguments:   Offset [in]
    returns:     Length of given record.
    */
    virtual Data_SplitRecordLength LengthValue (Data_PageOffset Offset)
    {
        return *(REINTERPRET_CAST(Data_SplitRecordLength*,(m_pFrame + Offset)));
    }

};
/*! endclass: Data_SplitSpaceForwardReadIterator */

/*---------------------------------------------------------------------------*/
/*! class:       Data_PageSplitSpace
    description: The page can store Split records.
 */
class Data_PageSplitSpace : public Data_ChainPage
{

public:

    /*!-------------------------------------------------------------------------
    function:     ~Data_PageSplitSpace()
    description:  This destructor can call verify on error abort.
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/
    ~Data_PageSplitSpace();
    
    /*! ---------------------------------------------------------------------------
    function:    InitializeFrame
    */
    virtual void InitializeFrame ()
    {
        Data_ChainPage::InitializeFrame();
        NodePtr()->nd_left() = SAPDB_MAX_INT4;
    }

    /*! ---------------------------------------------------------------------------
    function:    AlignRecordLength
    arguments:   Length [in/out]
    description: The RecordSize is properly aligned.
    */
    static Data_SplitRecordLength AlignRecordSize (Data_SplitRecordLength Length)
    {
        return ((Length-1) & ~(ALIGNMENT-1)) + ALIGNMENT;
    }
    
    /*! -----------------------------------------------------------------------
    function:    PrevPageNo
    returns:     PageNo
    description: The page id of the prev page is returned.
    
                 If no prev page is available, PageId.IsValid() returns false.
    */
    Data_PageNo PrevPageNo ()
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return STATIC_CAST(Data_PageNo,NodePtr()->nd_left());
    }

    /*! -----------------------------------------------------------------------
    function:    SetPrevPageNo
    arguments:   PageNo [in]
    */
    void SetPrevPageNo (const Data_PageNo& PageNo)
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        NodePtr()->nd_left() = (tsp00_PageNo)PageNo;
    }

    /*! -----------------------------------------------------------------------
    function:    Append
    description: NewPage [out]
    */
    void Append (Data_PageSplitSpace& NewPage,
                 Data_PageNo&         LastPageNo)
    {
        Data_ChainPage::Append (NewPage, LastPageNo);

        NewPage.SetPrevPageNo (PageNo());

        if ( NewPage.NextPageNo().IsValid() )
        {
            // THIS<->NEW<->NEXT
            Data_PageSplitSpace NextPage;
            
            (void) m_pAccessManager->GetPage (NextPage,
                                              NewPage.NextPageNo());
            NextPage.SetPrevPageNo( NextPageNo() );
        }
    }

    /*! ---------------------------------------------------------------------------
    function:    Use
    arguments:   Iterator [in/out]
    description: The iterator should now use this page with its current settings.
    */
    void Use (Data_SplitSpaceIterator& Iter)
    {
        Iter.ReAssign (FramePtr());
    }
    
    /*! ---------------------------------------------------------------------------
    function:    Use
    arguments:   Iterator        [in/out]
                 Data_PageOffset [in]
    description: The iterator should now use this page with its current settings.
    */
    void Use (Data_SplitSpaceIterator& Iter,
              Data_PageOffset          Offset)
    {
        Iter.ReAssign (FramePtr(), Offset);
    }
    
    /*! ---------------------------------------------------------------------------
    function:    Begin
    arguments:   Iterator [in/out]
    description: The iterator is set to first record on this page.
    */
    void Begin (Data_SplitSpaceIterator& Iter)
    {
        Iter.Begin (FramePtr());
    }
    
    /*! ---------------------------------------------------------------------------
    function:    End
    arguments:   Iterator [in/out]
    description: The iterator is set to last record on this page.
    */
    void End (Data_SplitSpaceIterator& Iter)
    {
        Iter.End (FramePtr());
    }
    
    /*! ---------------------------------------------------------------------------
    function:    SpaceInfo
    arguments:   WantedSize    [in]
                 MinimumSize   [in]
                 FirstPartFits [out]
                 AllFits       [out]
    description: This is used to get the correct page locks.
    */
    void SpaceInfo (Data_SplitRecordLength  WantedSize,
                    Data_RecordLength       MinimumSize,
                    bool                   &FirstPartFits,
                    bool                   &AllFits);
    
    /*! ---------------------------------------------------------------------------
    function:    UndoLastReservedSpace
    arguments:   LastReservedSize [in] - the size which was reserved before
    description: The changes done by the last reserve are un-done.
    */
    void UndoLastReservedSpace (Data_RecordLength LastReservedSize)
    {
        // PTS 1115979 UH 2002-05-29 new
        if ( 0 == LastReservedSize ) return;
        

        SetFirstFreeOffset ( Data_Page::FirstFreeOffset(this->FramePtr())
                             -
							 // PTS 1117126 UH 2002-08-07 added alignment
                             (AlignRecordSize(LastReservedSize)
                              + sizeof(Data_SplitRecordLength)) );
    }

    /*! ---------------------------------------------------------------------------
    function:    Truncate
    arguments:   Iterator [in/out]
    description: The given iterator points to the new first free position.
                 Everything from this position on is deleted.
    */
    void Truncate (Data_SplitSpaceIterator& iter)
    {
        Data_Page::SetFirstFreeOffset( iter.GetPosition() );
    }
};
/*! endclass: Data_PageSplitSpace */

/*---------------------------------------------------------------------------*/
/*! class:       Data_PageSplitSpaceBackwardRead
 */
class Data_PageSplitSpaceBackwardRead : public Data_PageSplitSpace
{

public:

    /*!-------------------------------------------------------------------------
    function:     WriteToTrace()
    description:  writes important values from this page to the knltrace
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/
    virtual void WriteToTrace (const SAPDB_Char *title = 0) const;

    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       true if verify successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const;

    /*! ---------------------------------------------------------------------------
    function:    ReserveSpace
    arguments:   ReservedSize        [in]     - the size which is already reserved starting with 0
                 WantedSize          [in]     - the size which is wanted
                 MinimumSize         [in]     - the size which must be available
                 CurrentReservedSize [in/out] - the size which is allocated
                 CurrentSpace              [out]    - the pointer to the reserved space
    description: The given WantedSize is used to allocate space.
                 If CurrentReservedSize is 0 then no space is available in current page.
    precondition: page is assigned
    */
    void ReserveSpace (Data_SplitRecordLength &ReservedSize,
                       Data_SplitRecordLength &WantedSize,
                       Data_RecordLength       MinimumSize,
                       Data_RecordLength      &CurrentReservedSize,
                       SAPDB_Byte*            &CurrentSpace);

    /*! -----------------------------------------------------------------------
    function:    GetOffset
    arguments:   partptr    [in]
                 partlength [in]
    description: Return the offset of the given part ptr
                 This depends on the page type (backward-read/forward-read)
    */
    Data_PageOffset GetOffset (SAPDB_Byte* partptr,
                               SAPDB_UInt  partlength) const;
};
/*! endclass: Data_PageSplitSpaceBackwardRead */

/*---------------------------------------------------------------------------*/
/*! class:       Data_PageSplitSpaceForwardRead
 */
class Data_PageSplitSpaceForwardRead : public Data_PageSplitSpace
{

public:

    /*!-------------------------------------------------------------------------
    function:     WriteToTrace()
    description:  writes important values from this page to the knltrace
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/
    virtual void WriteToTrace (const SAPDB_Char *title = 0) const;

    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       true if verify successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const;

    /*! ---------------------------------------------------------------------------
    function:    ReserveSpaceForwardRead
    arguments:   ReservedSize         [in]     - the size which is already reserved starting with 0
                 WantedSize           [in]     - the size which is wanted
                 FirstPartMiniumSize  [in]     - if set to 0, no minimum is given
                 CurrentReservedSize  [in/out] - the size which is allocated
                 CurrentSpace         [out]    - the pointer to the reserved space
    description: The given WantedSize is used to allocate space.
                 If CurrentReservedSize is 0 then no space is available in current page.
    precondition: page is assigned
    */
    void ReserveSpace (Data_SplitRecordLength &WantedSize,
                       Data_RecordLength       FirstPartMiniumSize,
                       Data_RecordLength      &CurrentReservedSize,
                       SAPDB_Byte*            &CurrentSpace);

    /*! -----------------------------------------------------------------------
    function:    GetOffset
    arguments:   partptr    [in]
                 partlength [in]
    description: Return the offset of the given part ptr
                 This depends on the page type (backward-read/forward-read)
    */
    Data_PageOffset GetOffset (SAPDB_Byte* partptr,
                               SAPDB_UInt  partlength) const;

};
/*! endclass: Data_PageSplitSpaceBackwardRead */

/*===========================================================================*
 *  UNDEFINES                                                                *
 *===========================================================================*/

#undef ALIGNMENT

#endif // Data_PageSplitSpace_HPP
