/****************************************************************************

  module      : Data_PrimTreeStatistic.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: File statistic
  description : Implementation of table statistics and update statistic column

  last changed: 2002-04-24
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hbd01_1.h"
#include "hbd05.h"
#include "hbd30.h"
#include "hgg01_3.h"
#include "hgg04.h"
#include "hgg10.h"
#include "ggg90.h"
#include "heo50.h"
#include <math.h>   // sqrt
#include <stdlib.h> // rand


#include "Converter/Converter_IPageNoManager.hpp"
#include "Container/Container_Vector.hpp"
#include "DataAccess/Data_PrimTreeStatistic.hpp"
#include "DataAccess/Data_PageAccessManager.hpp"
#include "DataAccess/Data_RandomTreeAccess.hpp"
#include "DataAccess/Data_ScanTreeAccess.hpp"
#include "DataAccess/Data_TraverseTreeAccess.hpp"
#include "DataAccess/Data_Messages.hpp"
#include "DataAccess/Data_Exceptions.hpp"
#include "IOManager/IOMan_IDataInfo.hpp"
#include "KernelCommon/Kernel_FileIO.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "RunTime/RTE_Message.hpp"
#include "SAPDBCommon/SAPDB_ToString.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

inline SAPDB_Bool
IsColumnUnique( const tgg00_StEntryAddr   pColumnStackEntry )
{
    return
        (( st_varkey == ( pColumnStackEntry )->etype())     &&
         (1          == ( pColumnStackEntry )->epos()))
        ||
        (( op_unique      == ( pColumnStackEntry )->eop())  ||
         ( op_unique_desc == ( pColumnStackEntry )->eop()));
}

/*---------------------------------------------------------------------------*/

inline SAPDBMem_IRawAllocator&
GetAllocator( tgg00_TransContext &trans )
{
    return( *( reinterpret_cast< SAPDBMem_IRawAllocator* >( trans.trAllocator_gg00 )));
}

/*---------------------------------------------------------------------------*/

inline SAPDB_Real8
myAbsolute( SAPDB_Real8 myValue )
{
    if( myValue < 0 )
        return( -1.0 * myValue );
    else
        return( myValue );
}

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

class TempStatisticRecord
{
public:

    TempStatisticRecord( tgg00_TransContext  &trans )
    :
    m_Allocator( GetAllocator( trans )),
    m_TrError( trans.trError_gg00 ),
    m_pRec( NULL ),
    m_Capacity( 0 )
    {}

    TempStatisticRecord(
        SAPDBMem_IRawAllocator  &allocator,
        tgg00_BasisError        &trError )
    :
    m_Allocator( allocator ),
    m_TrError( trError ),
    m_pRec( NULL ),
    m_Capacity( 0 )
    {}

    ~TempStatisticRecord(){ Delete(); }

    SAPDB_Bool Initialize( const SAPDB_UInt4 capacity );

    void AssignSpecialKey( SAPDB_Int2 columnId );

    void AssignKey(
        SAPDB_Int2     columnId,
        SAPDB_Int2     keyLength,
        SAPDB_Byte    *pKey );


    void AssignRecord( tgg00_Rec    *pRecord )
    {
        SAPDBERR_ASSERT_STATE( 0 == m_Capacity );
        m_pRec = pRecord;
    }

    tgg00_RecPtr GetRecordPtr() const
    {
        SAPDBERR_ASSERT_STATE( NULL != m_pRec );
        return( m_pRec );
    }

    SAPDB_UInt4 GetFrequency() const
    {
        SAPDBERR_ASSERT_STATE( NULL != m_pRec );
        return( m_pRec->recPno_gg00().gg90GetInt4() );
    }

    SAPDB_Int2 GetColumnId() const;

private:

    void InitRecordHeader();

    void Delete();

private:

    SAPDBMem_IRawAllocator  &m_Allocator;
    tgg00_BasisError        &m_TrError;
    tgg00_Rec               *m_pRec;
    SAPDB_UInt4             m_Capacity;
};

/*---------------------------------------------------------------------------*/

class EstimateDistinctValues
{

public:

    // Estimation algorithm is from Thomas Theodor Koetter (SAP Labs Berlin)

    EstimateDistinctValues(
        SAPDBMem_IRawAllocator  &allocator,
        const SAPDB_Real8       numberOfSampledRecords,
        const SAPDB_Real8       numberOfEstimatedRecords )
    :
    m_ColumnList( allocator ),
    m_SampleRecordCount( numberOfSampledRecords ),
    m_EstimatedRecordCount( numberOfEstimatedRecords )
    {}

    ~EstimateDistinctValues(){ m_ColumnList.Delete(); }

    SAPDB_Bool Initialize( const SAPDB_UInt columnCount );

    void AddValue(
        const SAPDB_UInt    columnId,
        const SAPDB_UInt4   frequency );

    SAPDB_Real8 Calculate(
        const SAPDB_UInt  columnId,
        const SAPDB_Int4  numberOfDistinctValues ) const;

private:

    struct ColumnDistinctValueItem
    {
        SAPDB_Real8 cdviTotalProbability;
        SAPDB_Real8 cdviRecCntWithLimitedFrequency;
        SAPDB_Real8 cdviRecCntWithLowFrequency;
        SAPDB_Real8 cdviDistinctRecCntWithLowFrequency;
#       ifdef SAPDB_QUICK
        SAPDB_UInt4 cdviFrequency[11];
#endif
    };

    typedef Container_Vector< ColumnDistinctValueItem > ColumnList;

private:

    ColumnList  m_ColumnList;

    const SAPDB_Real8   m_SampleRecordCount;
    const SAPDB_Real8   m_EstimatedRecordCount;
};

/*==========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

void
Data_PrimTreeStatistic::CalculateColumnStatistic(
    tbd_current_tree            &tempCurrent,
    const SAPDB_Int2            columnCount,
    SAPDB_Int2                  columnFirstPosition,
    SAPDB_Int2                  columnLastPosition,
    const tgg00_StEntryAddr     pColumnDescription,
    Data_PageCount              &numberOfLeaves,
    Data_PageCount              &numberOfRecords,
    tgg00_ColumnDistinctValues  &distinctValues )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::CalculateColumnStatistic", DataTree_Trace, 5 );

    const SAPDB_Bool                bCalculate = true;
    TempStatisticRecord             tempRecord( m_Trans );
    Data_SerializableScanTreeAccess scanTreeAccessor( m_Current );

    if( !tempRecord.Initialize( MAX_RECLEN_GG00 ))
        return; // e_no_more_memory is set

    numberOfLeaves    = 0;
    numberOfRecords   = 0;

    columnFirstPosition -= POS_OFF_DIFF_BD00; // necessary because of pascal and C++ mix
    columnLastPosition  -= POS_OFF_DIFF_BD00; // and the usage as an array index

    for( SAPDB_Int2 columnIndex = 0; columnIndex < columnCount; columnIndex++ )
        distinctValues[ columnIndex ] = 0;

    while( e_ok == m_TrError )
    {
        Data_PrimPage &primPage = scanTreeAccessor.GetLeafNode();

        if(( e_ok != m_TrError ) || (! primPage.IsAssigned()))
            break; // common loop termination

        if( ! ColumnStatisticForSinglePage( bCalculate, tempCurrent, tempRecord,
                                            primPage, columnCount, columnFirstPosition,
                                            columnLastPosition, pColumnDescription,
                                            distinctValues ))
            return;

        if( DataBaseFull( 1 ))
            return;

        if( CommandCanceled())
            return;

        numberOfLeaves  += 1;
        numberOfRecords += primPage.GetRecordCount();
    }
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::SampleColumnStatistic(
    tbd_current_tree            &tempCurrent,
    const Data_PageCount        leafNodesToSample,
    const SAPDB_Int2            columnCount,
    SAPDB_Int2                  columnFirstPosition,
    SAPDB_Int2                  columnLastPosition,
    const tgg00_StEntryAddr     pColumnDescription,
    Data_PageCount              &numberOfLeaves,
    Data_PageCount              &numberOfRecords,
    tgg00_ColumnDistinctValues  &distinctValues )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::SampleColumnStatistic", DataTree_Trace, 5 );

    SAPDBERR_ASSERT_STATE( 0 != leafNodesToSample );

    const SAPDB_Bool bCalculate             = true;
    const SAPDB_Bool bNoDuplicates          = true;
    SAPDB_Real8      sampledNumberOfRecords = 0;
    SAPDB_Real8      sampledNumberOfLeaves  = 0;

    TempStatisticRecord     tempRecord( m_Trans );
    Data_RandomTreeAccess   randomTreeAccessor( m_Trans, m_Current, leafNodesToSample, ! bNoDuplicates );

    columnFirstPosition -= POS_OFF_DIFF_BD00; // necessary because of pascal and C++ mix
    columnLastPosition  -= POS_OFF_DIFF_BD00; // and the usage as an array index

    for( SAPDB_Int2 columnIndex = 0; columnIndex < columnCount; columnIndex++ )
        distinctValues[ columnIndex ] = 0;

    if( ! tempRecord.Initialize( MAX_RECLEN_GG00 ))
        return; // e_no_more_memory is set

    if( ! randomTreeAccessor.Initialize( leafNodesToSample ))
        return;

    while( e_ok == m_TrError )
    {
        Data_PrimPage &primPage = randomTreeAccessor.GetLeafNode();

        if(( e_ok != m_TrError ) || (! primPage.IsAssigned()))
            break; // common loop termination

        if( ! ColumnStatisticForSinglePage( ! bCalculate, tempCurrent, tempRecord,
                                            primPage, columnCount, columnFirstPosition,
                                            columnLastPosition, pColumnDescription,
                                            distinctValues ))
            return;

        if( DataBaseFull( 1 ))
            return;

        if( CommandCanceled())
            return;

        sampledNumberOfLeaves  += 1.0;
        sampledNumberOfRecords += primPage.GetRecordCount();
    }

    if( e_ok == m_TrError )
    {
        // estimate records and leaf nodes
        const SAPDB_Real8 averageNumberOfRecords   = sampledNumberOfRecords / sampledNumberOfLeaves;
        const SAPDB_Real8 estimatedNumberOfLeaves  = randomTreeAccessor.GetNumberOfEstimatedLeafNodes();
        const SAPDB_Real8 estimatedNumberOfRecords = averageNumberOfRecords * estimatedNumberOfLeaves;

        numberOfRecords    = static_cast< Data_PageCount >( estimatedNumberOfRecords );
        numberOfLeaves     = static_cast< Data_PageCount >( estimatedNumberOfLeaves );

        if( g01newupdstat )
        {
            // New algorithmen from T.Koetter

            // estimate distinct values for given columns
            // Note that the first parameter sampledNumberOfLeaves specifies the numbers of
            // records used by the prior ColumnStatisticForSinglePage method call. In case of
            // an estimation only one record of each page is used!

            EstimateNumberOfDistinctValues( sampledNumberOfLeaves, estimatedNumberOfRecords,
                                            columnCount, tempCurrent, distinctValues );

            //EstimateNumberOfDistinctValues( sampledNumberOfRecords, estimatedNumberOfRecords,
            //                                columnCount, tempCurrent, distinctValues );
        }
        else
        {
            // Old algorithmen from H.Bayen resp. K.-D. Reinsch

            EstimateNumberOfDistinctValues( estimatedNumberOfRecords, sampledNumberOfRecords,
                                            columnCount, distinctValues );
        }
    }
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::CalculateRecordsAndLeaves(
    Data_PageCount  &numberOfLeaves,
    Data_PageCount  &numberOfRecords )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::CalculateRecordsAndLeaves", DataTree_Trace, 5 );

    Data_SerializableScanTreeAccess scanTreeAccessor( m_Current );

    numberOfLeaves    = 0;
    numberOfRecords   = 0;

    while( e_ok == m_TrError )
    {
        Data_PrimPage &primPage = scanTreeAccessor.GetLeafNode();

        if(( e_ok != m_TrError ) || (! primPage.IsAssigned()))
            break; // common loop termination

        if( CommandCanceled())
            return;

        numberOfLeaves  += 1;
        numberOfRecords += primPage.GetRecordCount();
    }
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::SampleRecordsAndLeaves(
    const Data_PageCount    leafNodesToSample,
    Data_PageCount          &numberOfLeaves,
    Data_PageCount          &numberOfRecords )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::SampleRecordsAndLeaves", DataTree_Trace, 5 );

    SAPDB_Real8 sampledNumberOfRecords = 0;
    SAPDB_Real8 sampledNumberOfLeaves  = 0;

    Data_RandomTreeAccess   randomTreeAccessor( m_Trans, m_Current, leafNodesToSample );

    if( ! randomTreeAccessor.Initialize( leafNodesToSample ))
        return;

    while( e_ok == m_TrError )
    {
        Data_PrimPage &primPage = randomTreeAccessor.GetLeafNode();

        if(( e_ok != m_TrError ) || (! primPage.IsAssigned()))
            break; // common loop termination

        if( CommandCanceled())
            return;

        sampledNumberOfLeaves  += 1.0;
        sampledNumberOfRecords += primPage.GetRecordCount();
    }

    if( e_ok == m_TrError )
    {
        // estimate records and leaf nodes
        const SAPDB_Real8 averageNumberOfRecords   = sampledNumberOfRecords / sampledNumberOfLeaves;
        const SAPDB_Real8 estimatedNumberOfLeaves  = randomTreeAccessor.GetNumberOfEstimatedLeafNodes();
        const SAPDB_Real8 estimatedNumberOfRecords = averageNumberOfRecords * estimatedNumberOfLeaves;

        numberOfRecords    = static_cast< Data_PageCount >( estimatedNumberOfRecords );
        numberOfLeaves     = static_cast< Data_PageCount >( estimatedNumberOfLeaves );
    }
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::CalculateTableStatistic(
    const SAPDB_UInt        numberOfLongColumns,
    Kernel_TableStatistic   &tableStatistic )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::SampleRecordsAndLeaves", DataTree_Trace, 5 );

    Data_TraverseTreeAccess   traverseTreeAccessor( m_Current );

    while( e_ok == m_TrError )
    {
        Data_PrimPage &primPage = traverseTreeAccessor.GetNode();

        if(( e_ok != m_TrError ) || (! primPage.IsAssigned()))
            break; // common loop termination

        if( LEAF_LEVEL_BD00 == primPage.GetLevel() )
        {
            if( 0 == numberOfLongColumns )
                TableStatisticForSingleLeafPage( primPage, tableStatistic );
            else
                TableStatisticForSingleLeafPageWithLongColumns(
                    numberOfLongColumns, primPage, tableStatistic );
        }
        else
            TableStatisticForSingleNonLeafPage( primPage, tableStatistic );

        if( CommandCanceled())
            return;
    }

    if( e_ok == m_TrError )
        tableStatistic.CalculateResult();
}

/*---------------------------------------------------------------------------*/
/*                             private methods                               */
/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_PrimTreeStatistic::ColumnStatisticForSinglePage(
    const SAPDB_Bool1            bCalculate,
    tbd_current_tree             &tempCurrent,
    TempStatisticRecord          &tempRecord,
    Data_PrimPage                &primPage,
    const SAPDB_Int2             columnCount,
    const SAPDB_Int2             columnFirstPosition,
    const SAPDB_Int2             columnLastPosition,
    const tgg00_StEntryAddr      pColumnDescription,
    tgg00_ColumnDistinctValues   &distinctValues )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::ColumnStatisticForSinglePage", DataTree_Trace, 6 );

    const Data_PageCount    recordCount        = primPage.GetRecordCount();
    SAPDB_Int4              currLength;
    SAPDB_Int4              currPosition;
    SAPDB_Int2              currColumnPosition = columnFirstPosition;
    SAPDB_Int2              currColumnId       = 0;
    Data_RecordIndex        startRecordIndex   = FIRST_REC_INDEX_BD00;
    Data_RecordIndex        stopRecordIndex    = recordCount;

    if(( g01newupdstat ) && ( ! bCalculate ) && ( 0 < recordCount ))
    {
        // determine one random record for estimation
        startRecordIndex = rand() % recordCount -1;
        stopRecordIndex  = startRecordIndex +1;
    }

    for( ; currColumnPosition <= columnLastPosition; currColumnPosition++, currColumnId++ )
    {
        if( IsColumnUnique( pColumnDescription + currColumnPosition ))
        {
            // do not test the uniqueness of a unique column
            distinctValues[ currColumnId ] += recordCount;
            continue;
        }
        tgg00_VarColPosList varColumnPositionList;

        // take a look at the not unique columns and count duplicate values
        for( Data_RecordIndex recordIndex = startRecordIndex; recordIndex < stopRecordIndex; recordIndex++ )
        {
            tgg00_Rec* pPrimRecord = primPage.GetRecord( recordIndex );

            varColumnPositionList.vpl_last = -1;

            g04locate_col( *( pColumnDescription + currColumnPosition ),
                           pPrimRecord, varColumnPositionList, currPosition, currLength );

            if(( 0 == currLength ) || // empty column
                    ( csp_undef_byte == pPrimRecord->recBuf_gg00()[ currPosition - POS_OFF_DIFF_BD00 ] ))
                tempRecord.AssignSpecialKey( currColumnId );
            else
                tempRecord.AssignKey( currColumnId, currLength,
                                      &pPrimRecord->recBuf_gg00()[ currPosition - POS_OFF_DIFF_BD00 ] );
            if( e_ok == m_TrError )
            {
                b30tadd_temp_primary( !bCalculate, *(tempRecord.GetRecordPtr()), tempCurrent );
            }
            else // PTS 1128292 TS 2004-03-05
            {
                Data_Exception errMsg( __CONTEXT__, DATA_ERR_PRIM_STATISTIC,
                                       SAPDB_ToString( m_TrError, _T_d ),
                                       SAPDB_ToString( recordIndex, _T_d ),
                                       SAPDB_ToString( primPage.PageNo(), _T_d ),
                                       SAPDB_ToString( primPage.RootPageNo(), _T_d ));
                RTE_Message( errMsg );

                Kernel_FileIO   dumpFile( KERNEL_DATA_FILE, KERNEL_COR_FILE_EXTENSION,
                                          SAPDB_ToString( primPage.PageNo(), _T_d ));
                dumpFile.Write( primPage );
                return( SAPDB_FALSE );
            }
            if( e_ok == m_TrError )
                distinctValues[ currColumnId ] +=1;
            else if( e_duplicate_key == m_TrError )
                m_TrError = e_ok;
            else
                return( SAPDB_FALSE );
        }
    }
    return( e_ok == m_TrError );
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::TableStatisticForSingleLeafPage(
    Data_PrimPage           &primPage,
    Kernel_TableStatistic   &tableStatistic )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::TableStatisticForSingleLeafPage", DataTree_Trace, 6 );

    const Data_PageCount    recordCount = primPage.GetRecordCount();

    tableStatistic.AddLeafCovering( primPage.GetCovering() );
    tableStatistic.AddRecordCount( primPage.GetRecordCount() );

    for( Data_RecordIndex recordIndex = FIRST_REC_INDEX_BD00; recordIndex < recordCount; recordIndex++ )
    {
        tgg00_Rec*  pPrimRecord = primPage.GetRecord( recordIndex );

        tableStatistic.AddRecord( pPrimRecord->recLen_gg00(), pPrimRecord->recKeyLen_gg00() );
    }
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::TableStatisticForSingleLeafPageWithLongColumns(
    const SAPDB_UInt        numberOfLongColumns,
    Data_PrimPage           &primPage,
    Kernel_TableStatistic   &tableStatistic )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::TableStatisticForSingleLeafPageWithLongColumns", DataTree_Trace, 6 );

    const Data_PageCount    recordCount = primPage.GetRecordCount();
    tgg00_FileId            blobFileId  = b01niltree_id;
    SAPDB_Int4              blobLength;
    SAPDB_Int4              blobPages;
    SAPDB_Int4              blobVersion;

    blobFileId.fileType_gg00().clear();
    blobFileId.fileType_gg00().addElement( ftsByteStr_egg00 );
    blobFileId.fileType_gg00().addElement( ftsConcurrent_egg00 );
    blobFileId.fileTfn_gg00().becomes( tfnColumn_egg00 );

    tableStatistic.AddLeafCovering( primPage.GetCovering() );
    tableStatistic.AddRecordCount( primPage.GetRecordCount() );

    for( Data_RecordIndex recordIndex = FIRST_REC_INDEX_BD00; recordIndex < recordCount; recordIndex++ )
    {
        tgg00_Rec*  pPrimRecord        = primPage.GetRecord( recordIndex );
        SAPDB_UInt  currColumnPosition = cgg_rec_key_offset + pPrimRecord->recKeyLen_gg00();

        tableStatistic.AddRecord( pPrimRecord->recLen_gg00(), pPrimRecord->recKeyLen_gg00() );

        for( SAPDB_UInt currColumn = 0; currColumn < numberOfLongColumns;
                currColumn++, currColumnPosition+= cgg_col_fd_size )
        {
            if( csp_undef_byte == pPrimRecord->recBuf_gg00()[ currColumnPosition ] )
                continue;

            blobFileId.fileRoot_gg00() = NIL_PAGE_NO_GG00;

            g10mv( "D-STAT", 5,
                   sizeof( pPrimRecord->recBuf_gg00() ), sizeof( blobFileId.fileName_gg00()),
                   pPrimRecord->recBuf_gg00(), currColumnPosition + 1 + POS_OFF_DIFF_BD00, // do not copy define byte
                   &blobFileId.fileName_gg00(), 2 + POS_OFF_DIFF_BD00,
                   sizeof( tgg00_Surrogate ), m_TrError );

            if( e_ok != m_TrError )
                return;

            b05length_bytestr( m_Trans, blobFileId, blobLength, blobPages, blobVersion );

            if( e_ok == m_TrError )
                tableStatistic.AddLongColumn( blobLength, blobPages );
            else
            {
                if( e_file_not_found == m_TrError ) // column is assumed to be in short column file
                    m_TrError = e_ok;
                else
                    return;
            }
        }
    }
}

/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::TableStatisticForSingleNonLeafPage(
    Data_PrimPage            &primPage,
    Kernel_TableStatistic    &tableStatistic )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::TableStatisticForSingleNonLeafPage", DataTree_Trace, 6 );

    const Data_PageCount    recordCount = primPage.GetRecordCount();

    if( primPage.RootPageNo() == primPage.PageNo() )
        tableStatistic.AddRootCovering( primPage.GetCovering(), primPage.GetLevel() );
    else
        tableStatistic.AddIndexCovering( primPage.GetCovering() );

    for( Data_RecordIndex recordIndex = FIRST_REC_INDEX_BD00; recordIndex < recordCount; recordIndex++ )
    {
        tgg00_Rec*  pPrimRecord = primPage.GetRecord( recordIndex );

        tableStatistic.AddSeparator( pPrimRecord->recKeyLen_gg00() );
    }
}


/*---------------------------------------------------------------------------*/

void
Data_PrimTreeStatistic::EstimateNumberOfDistinctValues(
    const SAPDB_Real8            numberOfRecords,
    const SAPDB_Real8            estimatedNumberOfRecords,
    const SAPDB_Int2             columnCount,
    tbd_current_tree             &tempCurrent,
    tgg00_ColumnDistinctValues   &distinctValues )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::EstimateNumberOfDistinctValues", DataTree_Trace, 6 );

    SAPDB_Int2              currColumnId = 0;
    SAPDB_UInt4             frequency    = 0;
    Data_ScanTreeAccess     scanTreeAccessor( tempCurrent );
    TempStatisticRecord     tempRecord( m_Trans );
    EstimateDistinctValues  estimateDistinctValues( GetAllocator( m_Trans ), numberOfRecords, estimatedNumberOfRecords );

    if( ! estimateDistinctValues.Initialize( columnCount ))
    {
        m_TrError = e_no_more_memory;
        return;
    }
#   ifdef SAPDB_QUICK
    if( DataTree_Trace.TracesLevel( 6 ))
    {
        Kernel_VTrace() << "ESTIMATE DISTINCT VALUES " << NewLine <<
        "#sampled records (n): " << numberOfRecords <<
        " estimated records (N): " << estimatedNumberOfRecords << NewLine;
    }
#   endif

    while( e_ok == m_TrError )
    {
        Data_PrimPage &tempPage = scanTreeAccessor.GetLeafNode();

        if(( e_ok != m_TrError ) || (! tempPage.IsAssigned()))
            break; // common loop termination

        const Data_PageCount recordCount = tempPage.GetRecordCount();

        for( Data_RecordIndex recordIndex = FIRST_REC_INDEX_BD00; recordIndex < recordCount; recordIndex++ )
        {
            tgg00_Rec* pTempRecord = tempPage.GetRecord( recordIndex );

            tempRecord.AssignRecord( pTempRecord );

            currColumnId  = tempRecord.GetColumnId();
            frequency     = tempRecord.GetFrequency();

            estimateDistinctValues.AddValue( currColumnId, frequency );
        }
    }

    SAPDBERR_ASSERT_STATE( estimatedNumberOfRecords >= numberOfRecords );

    for( currColumnId = 0; currColumnId < columnCount; currColumnId++ )
    {
        distinctValues[ currColumnId ] = static_cast< SAPDB_Int4 >
                                         ( estimateDistinctValues.Calculate( currColumnId, distinctValues[ currColumnId ] ));
#   ifdef SAPDB_QUICK
        if( DataTree_Trace.TracesLevel( 6 ))
        {
            Kernel_VTrace() << "Distinct value: " << distinctValues[ currColumnId ] << NewLine;
        }
#   endif
    }
}

/*---------------------------------------------------------------------------*/
void
Data_PrimTreeStatistic::EstimateNumberOfDistinctValues(
    const SAPDB_Real8           estimatedNumberOfRecords,
    const SAPDB_Real8           sampledNumberOfRecords,
    const SAPDB_Int2            columnCount,
    tgg00_ColumnDistinctValues  &distinctValues )
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_PrimTreeStatistic::EstimateDistinctValues", DataTree_Trace, 6 );

    if( DataTree_Trace.TracesLevel( 6 ))
    {
        Kernel_VTrace() << "ESTIMATE DISTINCT VALUES " << NewLine <<
        " sampled records : "   << sampledNumberOfRecords   <<
        " estimated records : " << estimatedNumberOfRecords << NewLine;
    }

    for( SAPDB_Int2 columnNo = 0; columnNo < columnCount; columnNo++  )
    {
        SAPDB_Real8 distinctInSample = distinctValues[ columnNo ];

        distinctValues[ columnNo ] = EstimateDistinctValuesPerColumn(
                                         estimatedNumberOfRecords,
                                         sampledNumberOfRecords,
                                         distinctInSample );

        if( DataTree_Trace.TracesLevel( 6 ))
        {
            Kernel_VTrace() << "columnNo: " << columnNo << NewLine <<
            " sampled distinct : "   << distinctInSample <<
            " estimated distinct : " << distinctValues[ columnNo ] << NewLine;
        }
    }
}

/*---------------------------------------------------------------------------*/

SAPDB_Int4
Data_PrimTreeStatistic::EstimateDistinctValuesPerColumn(
    const SAPDB_Real8   estimatedNumberOfRecords,
    const SAPDB_Real8   sampledNumberOfRecords,
    const SAPDB_Real8   distinctInSample )
{
    // The input variables should meet the following unequations:
    // 0 <= distinctInSample <= sampledNumberOfRecords <= estimatedNumberOfRecords

    if( distinctInSample <= 1 )
    {
        if( distinctInSample < 0 ){
            return( 0 );
        }
        else
            return( (SAPDB_Int4) distinctInSample );
    }

    if( sampledNumberOfRecords < distinctInSample ){
        return( (SAPDB_Int4) distinctInSample );
    }

    if( estimatedNumberOfRecords < sampledNumberOfRecords ){
        return( (SAPDB_Int4) distinctInSample );
    }

    // Start destinct value determination

    const SAPDB_Real8 qpr = distinctInSample / sampledNumberOfRecords;

    if( sampledNumberOfRecords <= 1.05 * distinctInSample ){
        return( (SAPDB_Int4) (0.5 + qpr * estimatedNumberOfRecords ));
    }

    const SAPDB_Real8   atol = sqrt( 2.2204460492503131e-16 );
    const SAPDB_Real8   qps  = (distinctInSample - 1.0) / (sampledNumberOfRecords - 1.0);
    const SAPDB_Real8   qpp  = (sampledNumberOfRecords - 1.0) / sampledNumberOfRecords;

    SAPDB_Real8         rtj  = (distinctInSample - 1.0) /  distinctInSample;
    SAPDB_Real8         rtjn = 0.0;
    SAPDB_Real8         h    = 0.0;
    tsp00_NumError      numError;

    for( SAPDB_UInt loopCount = 0; loopCount < 40; loopCount++ )
    {
        if( rtj <= 0 )
            h = 0;
        else
        { // TODO
            h = ( sampledNumberOfRecords - 1.0 ) * vln( rtj, numError );  // < 0
            h = vexp( h, numError);                                       // 0 <= h < 1
        }

        rtjn = qpp * ( h * rtj - qps ) / ( h - qpr );

        if( myAbsolute( rtjn - rtj ) > atol * myAbsolute( 1.0 - rtj ))
            rtj = rtjn;
        else
            break;
    }

    if(( 0.0 >= rtjn ) || ( rtjn >= 1.0 ))
    {
        return( (SAPDB_Int4) (0.5 + qpr * estimatedNumberOfRecords ));
    }

    h = ( rtjn - 1.0 ) * estimatedNumberOfRecords;
    h = 1.0 - vexp( h, numError );

    if( h > (( 1.0 - rtjn ) * qpr * estimatedNumberOfRecords ))
    {
        return( (SAPDB_Int4) (0.5 + qpr * estimatedNumberOfRecords ));
    }
    return( (SAPDB_Int4) (0.5 + h / ( 1.0 - rtjn )));
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_PrimTreeStatistic::DataBaseFull( const Data_PageCount  numPagesRequested )
{
    Converter_IPageNoManager::Instance().HandleDBFull( m_Trans, numPagesRequested,
            IOMan_IDataInfo::GetInstance().AddDataDevSpacePossible());

    return( e_ok != m_TrError );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_PrimTreeStatistic::CommandCanceled()
{
    if( e_ok != m_TrError )
        return( true );

    if( ! m_Trans.trRteCommPtr_gg00->to_cancel )
        return( false );

    m_TrError = e_cancelled;

    Data_Exception errMsg( __CONTEXT__, DATA_PRIM_STATISTIC_CANCELED,
                           SAPDB_ToString( m_Current.curr_tree_id.fileRoot_gg00(), _T_d ),
                           SAPDB_ToString( m_Trans.trTaskId_gg00, _T_d ));
    RTE_Message( errMsg );

    return( true );
}

/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (CODE)                                           *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES (CODE)                                                     *
 *===========================================================================*/

SAPDB_Bool
TempStatisticRecord::Initialize( const SAPDB_UInt4 capacity )
{
    if(( cgg_rec_key_offset + MAX_KEYLEN_GG00 ) <= capacity )
    {
        if( capacity > m_Capacity )
        {
            if( 0 != m_Capacity )
                Delete();
            m_pRec = reinterpret_cast< tgg00_RecPtr > (m_Allocator.Allocate( capacity ));
            if( NULL != m_pRec )
                m_Capacity = capacity;
        }
    }
    if( NULL == m_pRec )
    {
        m_TrError = e_no_more_memory;
        return( false );
    }
    InitRecordHeader();
    return( true );
}

/*---------------------------------------------------------------------------*/

void
TempStatisticRecord::AssignSpecialKey( SAPDB_Int2 columnId )
{
    SAPDBERR_ASSERT_STATE( NULL != m_pRec );
    SAPDBERR_ASSERT_STATE( 0 != m_Capacity );

    m_pRec->recLen_gg00()          = sizeof( columnId ) + cgg_rec_key_offset;
    m_pRec->recKeyLen_gg00()       = sizeof( columnId );
    m_pRec->recPno_gg00().becomes( 1 );

    // Copy logical column identifier to the begining of the key.
    // This is necessary to make equal values of different columns
    // in a temporary destination file distinguishable

    g10mv( "D-STAT", 1,
           sizeof( columnId ), m_Capacity,
           &columnId, POS_OFF_DIFF_BD00,
           &( m_pRec->recBody_gg00()), POS_OFF_DIFF_BD00,
           sizeof( columnId ), m_TrError );

    if( e_move_error == m_TrError )         // PTS 1128292 TS 2004-03-05
        m_TrError = e_data_page_corrupted;

}

/*---------------------------------------------------------------------------*/

void
TempStatisticRecord::AssignKey(
    SAPDB_Int2    columnId,
    SAPDB_Int2    keyLength,
    SAPDB_Byte    *pKey )
{
    SAPDBERR_ASSERT_STATE( NULL != m_pRec );
    SAPDBERR_ASSERT_STATE( 0 != m_Capacity );

    if( keyLength > MAX_KEYLEN_GG00 )
        keyLength = MAX_KEYLEN_GG00; // TODO LATER handle columns longer than 1024

    m_pRec->recLen_gg00()          = cgg_rec_key_offset + sizeof( columnId ) + keyLength;
    m_pRec->recKeyLen_gg00()       =  sizeof( columnId ) + keyLength;
    m_pRec->recPno_gg00().becomes( 1 );

    // Copy logical column identifier to the begining of the key.
    // This is necessary to make equal values of different columns
    // in a temporary destination file distinguishable

    g10mv( "D-STAT", 2,
           sizeof( columnId ), m_Capacity,
           &columnId, POS_OFF_DIFF_BD00,
           &( m_pRec->recBody_gg00()), POS_OFF_DIFF_BD00,
           sizeof( columnId ), m_TrError );

    g10mv( "D-STAT", 3,
           keyLength, m_Capacity,
           pKey, POS_OFF_DIFF_BD00,
           &( m_pRec->recBody_gg00()[ sizeof( columnId ) ]), POS_OFF_DIFF_BD00,
           keyLength, m_TrError );

    if( e_move_error == m_TrError )
        m_TrError = e_data_page_corrupted;

}
/*---------------------------------------------------------------------------*/

SAPDB_Int2
TempStatisticRecord::GetColumnId() const
{
    SAPDBERR_ASSERT_STATE( NULL != m_pRec );

    SAPDB_Int2  columnId;

    g10mv( "D-STAT", 4,
           m_pRec->recKeyLen_gg00(), sizeof( columnId ),
           &( m_pRec->recBody_gg00()), POS_OFF_DIFF_BD00,
           &columnId, POS_OFF_DIFF_BD00,
           sizeof( columnId ), m_TrError );

    return( columnId );
}

/*---------------------------------------------------------------------------*/

void
TempStatisticRecord::InitRecordHeader()
{
    if( NULL == m_pRec)
        return;

    m_pRec->recLen_gg00()    = 0;
    m_pRec->recKeyLen_gg00() = 0;
    m_pRec->recPno_gg00().becomes( 0 );
}

/*---------------------------------------------------------------------------*/

void
TempStatisticRecord::Delete()
{
    if(( NULL != m_pRec ) && ( 0 < m_Capacity ))
        m_Allocator.Deallocate( m_pRec );

    m_pRec     = NULL;
    m_Capacity = 0;
}
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

SAPDB_Bool
EstimateDistinctValues::Initialize( const SAPDB_UInt    columnCount )
{
    if( ! m_ColumnList.Initialize( columnCount ))
        return( false );

    ColumnList::Iterator iter    = m_ColumnList.Begin();
    ColumnList::Iterator endIter = m_ColumnList.End();

    while( iter != endIter )
    {
        ( *iter ).cdviTotalProbability               = 0; // SUM( P(i) * I( n*P(i)*(1-P(i))<9.0 ))
        ( *iter ).cdviRecCntWithLimitedFrequency     = 0; // SUM( P(i)*I( n*P(i)*(1-p(i)) < 9.0) * I(m(i)>2))
        ( *iter ).cdviRecCntWithLowFrequency         = 0; // n~
        ( *iter ).cdviDistinctRecCntWithLowFrequency = 0; // m~
#       ifdef SAPDB_QUICK
        for( SAPDB_UInt i = 0; i < 11; i++ )
            ( *iter ).cdviFrequency[i] = 0;
#       endif
        iter++;
    }
    return( true );
}

/*---------------------------------------------------------------------------*/

void
EstimateDistinctValues::AddValue(
    const SAPDB_UInt    columnId,
    const SAPDB_UInt4   frequency )
{
    SAPDBERR_ASSERT_STATE( m_ColumnList.GetSize() > columnId  );

    const SAPDB_Real8 probability = frequency / m_SampleRecordCount;

    if(( m_SampleRecordCount * probability * ( 1.0 - probability )) < 9.0 )
    {
        m_ColumnList[ columnId ].cdviTotalProbability += probability;

        if( 2 < frequency )
            m_ColumnList[ columnId ].cdviRecCntWithLimitedFrequency += probability;

        m_ColumnList[ columnId ].cdviRecCntWithLowFrequency +=frequency; // n~
        m_ColumnList[ columnId ].cdviDistinctRecCntWithLowFrequency++;   // m~

#       ifdef SAPDB_QUICK
        if(( 0 < frequency ) && (11 > frequency ))
            m_ColumnList[ columnId ].cdviFrequency[ frequency ]++;
#endif
    }
}

/*---------------------------------------------------------------------------*/

SAPDB_Real8
EstimateDistinctValues::Calculate(
    const SAPDB_UInt  columnId,
    const SAPDB_Int4  numberOfDistinctValues ) const
{
    SAPDBERR_ASSERT_STATE( m_ColumnList.GetSize() > columnId  );

    /* r = SUM( P(i) * I( n*P(i)*(1-P(i))<9.0)) - 0.5*SUM( P(i)*I( n*P(i)*(1-p(i)) < 9.0) * I(m(i)>2))   */
    const SAPDB_Real8 tempResult1 =
        m_ColumnList[ columnId ].cdviTotalProbability - 0.5 * m_ColumnList[ columnId ].cdviRecCntWithLimitedFrequency;

    /* m~ DIV n~ */

    SAPDB_Real8 tempResult2 = 0;
    if( 0 < m_ColumnList[ columnId ].cdviRecCntWithLowFrequency )
        tempResult2 = ( m_ColumnList[ columnId ].cdviDistinctRecCntWithLowFrequency / m_ColumnList[ columnId ].cdviRecCntWithLowFrequency );

    /* M = (m - m~) + N * r * (m~ DIV n~) */
    const SAPDB_Real8 result =
        ( numberOfDistinctValues - m_ColumnList[ columnId ].cdviDistinctRecCntWithLowFrequency ) +
        m_EstimatedRecordCount * tempResult1 * tempResult2;

#   ifdef SAPDB_QUICK
    if( DataTree_Trace.TracesLevel( 6 ))
    {
        Kernel_VTrace() <<
        "ColumnId: " << columnId << NewLine <<
        " m  : "  << numberOfDistinctValues <<
        " m~ : " << m_ColumnList[ columnId ].cdviDistinctRecCntWithLowFrequency <<
        " n~ : " << m_ColumnList[ columnId ].cdviRecCntWithLowFrequency <<
        " r  : " << tempResult1 << NewLine <<
        " n(1) : " << m_ColumnList[ columnId ].cdviFrequency[1]   << NewLine <<
        " n(2) : " << m_ColumnList[ columnId ].cdviFrequency[2]   << NewLine <<
        " n(3) : " << m_ColumnList[ columnId ].cdviFrequency[3]   << NewLine <<
        " n(4) : " << m_ColumnList[ columnId ].cdviFrequency[4]   << NewLine <<
        " n(5) : " << m_ColumnList[ columnId ].cdviFrequency[5]   << NewLine <<
        " n(6) : " << m_ColumnList[ columnId ].cdviFrequency[6]   << NewLine <<
        " n(7) : " << m_ColumnList[ columnId ].cdviFrequency[7]   << NewLine <<
        " n(8) : " << m_ColumnList[ columnId ].cdviFrequency[8]   << NewLine <<
        " n(9) : " << m_ColumnList[ columnId ].cdviFrequency[9]   << NewLine <<
        " n(10) : " << m_ColumnList[ columnId ].cdviFrequency[10] << NewLine;
    }
#   endif

    return( result );
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
