/****************************************************************************

  module      : Data_ScanTreeAccess.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area:
  description : 

  last changed: 2002-04-18
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hbd13.h"
#include "hbd50.h"
#include "hbd75.h"
#include "heo56.h"
#include "hgg05.h"

#include "DataAccess/Data_ScanTreeAccess.hpp"
#include "DataAccess/Data_Messages.hpp"
#include "DataAccess/Data_Exceptions.hpp"
#include "KernelCommon/Kernel_FileIO.hpp"
#include "RunTime/RTE_Message.hpp"
#include "SAPDBCommon/SAPDB_ToString.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/


/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/



Data_ScanTreeAccess::~Data_ScanTreeAccess()
{
    m_Page.Deassign( false ); // release the page for read

    if( r_lock_leaf == m_Current.curr_lockstate )
        b75luread_unlock_leaf( m_Current, m_Current.curr_node_locked );
    else
        if( r_lock_index == m_Current.curr_lockstate )
            b75iuread_unlock_index( m_Current, m_Current.curr_lvl_1_locked );
}

/*---------------------------------------------------------------------------*/

Data_PrimPage&
Data_ScanTreeAccess::GetLeafNode()
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_ScanTreeAccess::GetLeafNode", DataTree_Trace, 5 );

    SAPDBERR_ASSERT_STATE( e_ok == m_Current.curr_trans->trError_gg00 );

    if( m_Page.IsAssigned() )
    {
        if( GetNextLeafNode() )
            return( m_Page );
    }
    else if( GetLeftMostLeafNode() )
        return( m_Page );

    // error handling
    if( m_Page.IsAssigned() )
        m_PageAccessManager.ReleasePage( m_Page );

    return( m_Page ); // return empty page to indicate termination of iterator
}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

Data_PrimPage&
Data_SerializableScanTreeAccess::GetLeafNode()
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_SerializableScanTreeAccess::GetLeafNode", DataTree_Trace, 5 );

    SAPDBERR_ASSERT_STATE( e_ok == m_Current.curr_trans->trError_gg00 );

    if( m_Page.IsAssigned() ) // subsequent access
    {
        if(( m_NumberOfReadLeafNodes > MAX_NODE_SCAN_BD00 ) &&
                ( b75tree_request( m_Current.curr_trans->trTaskId_gg00,
                                   m_Current.curr_tree_id.fileRoot_gg00() )))
        {
            if( ! Break() )
                return( m_Page );
            if( ! Continue() ) // go to the old leaf page and then to the successor
                return( m_Page );
        }
        if( Data_SerializableScanTreeAccess::GetNextLeafNode() )
            return( m_Page );
    }
    else if( Data_SerializableScanTreeAccess::GetLeftMostLeafNode() ) // first access to the given file
        return( m_Page );

    // error handling
    if( m_Page.IsAssigned() )
        m_PageAccessManager.ReleasePage( m_Page );

    return( m_Page ); // return empty page to indicate termination of iterator
}

/*---------------------------------------------------------------------------*/
/*                            private methods                                */
/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_ScanTreeAccess::GetLeftMostLeafNode()
{
    SAPDBERR_ASSERT_STATE( e_ok == m_Current.curr_trans->trError_gg00 );
    SAPDBERR_ASSERT_STATE( ! m_Page.IsAssigned() );

    SAPDB_Int4  leftMostPageNo  = m_Current.curr_tree_id.fileRoot_gg00();
    SAPDB_Bool1 bRootIsLeaf     = false; // caused by Pascal/C++ mix

    b50first_leaf( leftMostPageNo, r_lock_leaf, bRootIsLeaf, m_Current );

    if(( e_ok == m_Current.curr_trans->trError_gg00 ) &&
            ( m_PageAccessManager.GetPage( m_Page, leftMostPageNo  )))
        return( true );
    else
        return( false );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_ScanTreeAccess::GetNextLeafNode()
{
    SAPDBERR_ASSERT_STATE( m_Page.IsAssigned() );
    SAPDBERR_ASSERT_STATE( e_ok == m_Current.curr_trans->trError_gg00 );

    const Data_PageNo nextPageNo = m_Page.Successor();

    m_PageAccessManager.ReleasePage( m_Page );

    ReplaceLeafPageLock( nextPageNo );

    if(( nextPageNo.IsValid() ) && ( m_PageAccessManager.GetPage( m_Page, nextPageNo )))
        return( true );
    else
        return( false );
}

/*---------------------------------------------------------------------------*/

void
Data_ScanTreeAccess::ReplaceLeafPageLock( const Data_PageNo nextPageNo )
{
    if(
        ( e_ok != m_Current.curr_trans->trError_gg00 ) ||
        ( r_lock_leaf != m_Current.curr_lockstate    ) ||
        ( nextPageNo.IsInvalid()                     )
    )
        return;
    if(
        ( m_Current.curr_tree_id.fileType_gg00().includes( ftsDynamic_egg00 )) &&
        ( m_Current.curr_right_bound == m_Current.curr_node_locked           )
    )
    {
        b50cross_subtree_boundary( m_Current, m_Current.curr_node_locked, nextPageNo );

        if( e_ok ==  m_Current.curr_trans->trError_gg00 )
            b75llread_lock_leaf( m_Current, nextPageNo, 27 );
    }
    else
        b75lulread_unlocklock_leaf( m_Current, m_Current.curr_node_locked,
                                    m_Current.curr_lvl_1_locked, nextPageNo,
                                    m_Current.curr_lvl_1_locked, 28 );
}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_SerializableScanTreeAccess::Break()
{
    tgg00_BasisError    &trError = m_Current.curr_trans->trError_gg00;
    const tsp00_TaskId   &taskId  = m_Current.curr_trans->trTaskId_gg00;

    SAPDBERR_ASSERT_STATE( m_Page.IsAssigned() );
    SAPDBERR_ASSERT_STATE( e_ok == trError );

    if( NULL == m_pReEntranceKey )
    {
        // This is the first interruption of the iterator, therefore no
        // memory is available to store the re-entrance key. After the
        // break, the memory will not deallocated because it could be reused
        // for the new break, if it occures. The allocation and deallocation
        // of memory is more expensive than holding the allocated memory
        // until the end of the scan.
        m_pReEntranceKey = reinterpret_cast< SAPDB_Byte* >( m_Allocator.Allocate( MAX_RECLEN_GG00 ));
    }
    if( NULL == m_pReEntranceKey )
    {
        trError = e_no_more_memory;
        return( false );
    }

    tgg00_Rec* pRecord = m_Page.GetLastRecord();
    SAPDB_Byte* pKey   = reinterpret_cast< SAPDB_Byte* >( pRecord ) + cgg_rec_key_offset;

    gg05KeyAssign( pKey, pRecord->recKeyLen_gg00(),
                   m_pReEntranceKey, m_ReEntranceKeyLength, trError );

    if( e_ok != trError )
    {
        Data_Exception errMsg( __CONTEXT__, DATA_ERR_SCAN_TREE,
                               SAPDB_ToString( trError, _T_d ),
                               SAPDB_ToString( m_Page.GetRecordCount()-1, _T_d ),
                               SAPDB_ToString( m_Page.PageNo(), _T_d ),
                               SAPDB_ToString( m_Page.RootPageNo(), _T_d ));
        RTE_Message( errMsg );

        Kernel_FileIO   dumpFile( KERNEL_DATA_FILE, KERNEL_COR_FILE_EXTENSION,
                                  SAPDB_ToString( m_Page.PageNo(), _T_d ));
        dumpFile.Write( m_Page );
    }

    m_PageAccessManager.ReleasePage( m_Page );

    switch( m_Current.curr_lockstate )
    {
    case r_lock_leaf:
        b75luread_unlock_leaf( m_Current, m_Current.curr_node_locked );
        break;
    case r_lock_index:
        b75iuread_unlock_index( m_Current, m_Current.curr_lvl_1_locked );
        break;
    case r_lock_tree:
        b75turead_unlock_tree( m_Current );
        break;
    default:;
    }

    m_NumberOfReadLeafNodes = 0;

    vsleep(taskId, 0 ); // reschedule

    return( e_ok == trError );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_SerializableScanTreeAccess::Continue()
{
    tgg00_BasisError &trError = m_Current.curr_trans->trError_gg00;

    SAPDBERR_ASSERT_STATE( ! m_Page.IsAssigned() );
    SAPDBERR_ASSERT_STATE( e_ok == m_Current.curr_trans->trError_gg00 );
    SAPDBERR_ASSERT_STATE( no_bd_lock == m_Current.curr_lockstate );
    SAPDBERR_ASSERT_STATE( NULL != m_pReEntranceKey );
    SAPDBERR_ASSERT_STATE( 0 != m_ReEntranceKeyLength );

    tbd_subtree_info    subTreeInfo;
    SAPDB_Int4          dummy;
    SAPDB_Int4          dummyPageNo;
    tbd_neighbors       neighbors;
    tbd_node_ptrs       pNodes;

    subTreeInfo.sti_is_del_op = false;
    pNodes.np_ptr()   = NULL;
    pNodes.np_cbptr() = NULL;

    if( m_Current.curr_tree_id.fileType_gg00().includes( ftsConcurrent_egg00 ))
        b75tlread_lock_tree( m_Current, 29 );

    b50find_leaf( m_pReEntranceKey, m_ReEntranceKeyLength, r_lock_leaf,
                  nr_for_read, subTreeInfo, dummy, dummy, dummyPageNo, pNodes,
                  neighbors, m_Current );

    if( e_ok != trError )
    {
        if( NULL != pNodes.np_ptr() )
            b13r_release_node( pNodes, m_Current, lru_normal );
    }
    else
    {
        Data_PageNo destinationPageNo = pNodes.np_ptr()->nd_id();

        b13r_release_node( pNodes, m_Current, lru_normal );

        m_PageAccessManager.GetPage( m_Page, destinationPageNo  );
    }

    return( e_ok == trError );
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/

