/*!***************************************************************************

  module      : Data_SplitSpace.hpp
  responsible : UweH
  special area: DataAccess
  created     : 2000-11-22
  last changed: 2000-11-22 12:00
  copyright   : Copyright by SAP AG, 2000
  description : defines class Data_SplitSpace



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef Data_SplitSpace_HPP
#define Data_SplitSpace_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/SAPDB_EnhancedTypes.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "Container/Container_Vector.hpp"
#include "DataAccess/Data_Types.hpp"
#include "DataAccess/Data_Exceptions.hpp"
#include "DataAccess/Data_Messages.hpp"
#include "DataAccess/Data_ISplitSpace.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define DATA_SPLIT_SPACE_INIT_PART_COUNT 3
    
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL FUNCTIONS                                                       *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Data_SplitSpace
    description: This manages space which is not located in one page.
 */
template <class T> class Data_SplitSpace : public Data_ISplitSpace
{
public:

    /*! ---------------------------------------------------------------------------
    declaration: PartNo
    */
    typedef SAPDB_Invalidatable<SAPDB_Int, -1> PartNo;
    /*! enddeclaration: */

    /*! ---------------------------------------------------------------------------
    function:    Data_SplitSpace
    description: By default, two parts are supported.
    */
    Data_SplitSpace (SAPDBMem_IRawAllocator& Allocator)
    : m_Parts              (Allocator),
      m_NumPagesAppended (0)
    {}
    
    /*! ---------------------------------------------------------------------------
    function:    ~Data_SplitSpace
    */
    ~Data_SplitSpace()
    {
        Deassign();
        Delete();
    }
    
    /*! ---------------------------------------------------------------------------
    function:    Initialize
    returns:     false, if the object cannot be initialized.
    */
    bool Initialize ()
    {
        return m_Parts.Initialize(DATA_SPLIT_SPACE_INIT_PART_COUNT);
    }
    
   /*! ---------------------------------------------------------------------------
    function:    Delete
    */
    void Delete()
    {
        if ( m_Parts.GetCapacity() > 0 )
            m_Parts.Delete();
    }
    
     /*! ---------------------------------------------------------------------------
    function:    CurrentPart
    returns:     the current part
    */
    virtual SAPDB_Int CurrentPart () const
    {
        return m_CurrentPart;
    }

    /*! ---------------------------------------------------------------------------
    function:    GetPart
    arguments:   Part   [in] 
                 Length [out]
                 Value  [out]
    */
    virtual void GetPart (SAPDB_Int    Part,
                          SAPDB_UInt  &Length,
                          SAPDB_Byte* &Value) const
    {
        if ( Part > m_CurrentPart )
        {
			// PTS 1118382 UH 2002-10-17
            Value  = 0;
            Length = 0;
        }
        else
        {
            Value  = m_Parts [Part].Value;
            Length = m_Parts [Part].Length;
        }
    }
    
    /*! ---------------------------------------------------------------------------
    function:    Length
    returns:     SAPDB_UInt
    description: The sum of length of all used parts of the space is returned in bytes.
    */
    virtual SAPDB_UInt Length() const
    {
        SAPDB_UInt Result = 0;
        
        for ( SAPDB_Int i = 0; i <= m_CurrentPart; ++i )
        {
            Result += m_Parts[i].Length;
        }
        return Result;
    }

    /*! ---------------------------------------------------------------------------
    function:    UseNextPart
    returns:     the reference to the page.
    description: This part must not be used before.
                 ! BE CAREFULL                                                     !
                 ! A Resize is done internally and therefore all pointers          !
                 ! to the current SplitSpace can be become invalid after this call !
    */
    T& UseNextPart()
    {
        SAPDB_UInt newPart;

        if ( m_CurrentPart.IsInvalid() )
            m_CurrentPart = 0;
        else
            ++m_CurrentPart;
	
        newPart = m_CurrentPart;

        if ( newPart >= m_Parts.GetSize() )
        {
            bool setAccessMode = m_Parts.GetSize() > 0;

            if ( ! m_Parts.Resize(m_Parts.GetSize()+DATA_SPLIT_SPACE_INIT_PART_COUNT) )
                // PTS 1117126 UH 2002-08-07 added if-clause
                RTE_Crash( Data_Exception(__CONTEXT__,DATA_SPLIT_SPACE_RESIZE,
                           SAPDB_ToString(m_Parts.GetSize())) );
            if ( setAccessMode )
                SetAccessMode(m_Parts[0].Page.AccessMode());
        }

        SAPDBERR_ASSERT_STATE ( ! m_Parts[m_CurrentPart].Page.IsAssigned() );

        return m_Parts[m_CurrentPart].Page;
    }

    /*! ---------------------------------------------------------------------------
    function:    GetCurrentPage
    returns:     the reference to the page.
    description: This part must not be used before.
    */
    const T& GetCurrentPage () const
    {
        return m_Parts[m_CurrentPart].Page;
    }

    /*! ---------------------------------------------------------------------------
    function:    GetCurrentPage
    returns:     the reference to the page.
    description: This part must not be used before.
    */
    T& GetCurrentPage ()
    {
        return m_Parts[m_CurrentPart].Page;
    }

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    returns:     the reference to the page.
    description: This part must not be used before.
    */
    T& GetPage (SAPDB_Int part)
    {
        return m_Parts[part].Page;
    }

    /*! ---------------------------------------------------------------------------
    function:    GetPage
    returns:     the reference to the page.
    description: This part must not be used before.
    */
    const T& GetPage (SAPDB_Int part) const
    {
        return m_Parts[part].Page;
    }

    /*! ---------------------------------------------------------------------------
    function:    AppendPart
    description: A complete part is appended to the part list.
    */
    void AppendPart (T          &Page,
                     SAPDB_UInt  Length,
                     SAPDB_Byte *Value)
    {
        T& InternalPage = UseNextPart();
        InternalPage.Assign (Page);
        SetPart (Length, Value);
    }

    /*! ---------------------------------------------------------------------------
    function:    SetPart
    arguments:   Length [in]
                 Value  [in]
    description: The description of a new space is set to to current part.
    */
    void SetPart (SAPDB_UInt  Length,
                  SAPDB_Byte* Value)
    {
        SAPDBERR_ASSERT_ARGUMENT( m_Parts[m_CurrentPart].Page.ReferenceIsValid (Value) );

        m_Parts [m_CurrentPart].Value  = Value;
		m_Parts [m_CurrentPart].Length = Length;
    }
    
    /*! ---------------------------------------------------------------------------
    function:    IsAssigned
    returns:     true, if first part page is assigned.
    */
    bool IsAssigned() const
    {
        return m_CurrentPart.IsValid();
    }
    
    /*! ---------------------------------------------------------------------------
    function:    Reassign
    arguments:   bKeepFirstPart
    description: From the current used part the page is moved to part 0.
                 if is not assigned, nothing changes.
    */
    void Reassign(bool bKeepFirstPart = false)
    {
        if ( ! IsAssigned() || 0 == m_CurrentPart ) return;

        if ( bKeepFirstPart )
        {
            for ( SAPDB_Int i = 1; i <= m_CurrentPart; ++i )
                m_Parts[i].Page.Deassign();
        }
        else
        {
            m_Parts[0].Page.Deassign();
            m_Parts[0].Page.Assign (m_Parts[m_CurrentPart].Page);

            for ( SAPDB_Int i = m_CurrentPart-1; i > 0; --i )
                m_Parts[i].Page.Deassign();
        }
        m_CurrentPart      = 0;
        m_NumPagesAppended = 0;
        SAPDBERR_ASSERT_STATE( IsAssigned() );
    }

    /*! ---------------------------------------------------------------------------
    function:    Deassign
    arguments:   isOK [in]
    description: All parts are released, if ok, then release and update pages.
    */
    void Deassign (bool isOK = true)
    {
        SAPDB_Int PartNo = 0;
        while ( PartNo <= m_CurrentPart )
        {
            m_Parts[PartNo].Page.Deassign(isOK);
            ++PartNo;
        }
        m_CurrentPart.Invalidate();
        m_NumPagesAppended = 0;
    }
    
    /*! -----------------------------------------------------------------------
    function:    SetAccessMode
    description: If IsAssigned , all pages are set to AccessMode.
    */
    void SetAccessMode (Data_AccessMode AccessMode)
    {
        if ( m_Parts.GetSize() == 0 )
            (void) Initialize();
            
        for ( SAPDB_UInt i = 0; i < m_Parts.GetSize(); ++i )
			// PTS 1117126 UH 2002-08-07
            m_Parts[i].Page.SetAccessMode (AccessMode);
    }

    /*! -----------------------------------------------------------------------
    function:    CheckAccessMode
    description: returns true, if all pages have the same given access mode.
    */
    bool CheckAccessMode (Data_AccessMode AccessMode) const
    {
        bool result = true;
        for ( SAPDB_UInt i = 0; i < m_Parts.GetSize(); ++i )
            result &= (m_Parts[i].Page.AccessMode() == AccessMode);
        return result;
    }

    /*! ---------------------------------------------------------------------------
    function:    Reorder
    description: max part is moved to 0 and vice versa
    */
    void Reorder()
    {
        SAPDBERR_ASSERT_STATE( IsAssigned() );

        Part AuxPart;
        
        for ( SAPDB_Int src = 0, dst = m_CurrentPart;
              src < dst;
              ++src, --dst )
        {
            AuxPart.Page.Assign (m_Parts[src].Page);
            AuxPart.Length = m_Parts[src].Length;
            AuxPart.Value  = m_Parts[src].Value;
            
            m_Parts[src].Page.Assign (m_Parts[dst].Page);
            m_Parts[src].Length = m_Parts[dst].Length;
            m_Parts[src].Value  = m_Parts[dst].Value;
            
            m_Parts[dst].Page.Assign (AuxPart.Page);
            m_Parts[dst].Length = AuxPart.Length;
            m_Parts[dst].Value  = AuxPart.Value;
        }
    }
    
    /*! ---------------------------------------------------------------------------
    function:    WriteToTrace
    */
    void WriteToTrace (const char * Title) const
    {
        if ( ! IsAssigned() )
        {
            Kernel_VTrace() << Title << ": not assigned.";
            return;
        }
        Kernel_VTrace() << Title;
        for ( PartNo part = 0; part < m_CurrentPart; ++part )
            if ( m_Parts[part].Page.IsAssigned() )
                Kernel_VTrace() << "part(" << part << ") pno:" << m_Parts[part].Page.PageNo()
                                << ", offset: " << m_Parts[part].Page.GetOffset (m_Parts[part].Value,m_Parts[part].Length)
                                << ", length: " << m_Parts[part].Length
                                << ", numPagesAppended: " << m_NumPagesAppended;
            else
                Kernel_VTrace() << "part(" << part << ") pno: page not assigned"
                                << ", offset: page not assigned"
                                << ", length: " << m_Parts[part].Length
                                << ", numPagesAppended: " << m_NumPagesAppended;
    }
    
    /*! ---------------------------------------------------------------------------
    function:    SetNewPageWasAppended
    description: Increments the internal counter which counts new allocated pages.
    */
    void SetNewPageWasAppended()
    {
        ++m_NumPagesAppended;
    }
    
    /*! ---------------------------------------------------------------------------
    function:    NewPageWasAppended
    returns:     bool
    description: Returns true if the internal counter for new allocated pages
                 is greater than 0.
    */
    bool NewPageWasAppended()
    {
        return m_NumPagesAppended > 0;
    }
    
    /*! ---------------------------------------------------------------------------
    function:    GetNumAdditionalPagesAppended
    returns:     SAPDB_UInt
    */
    SAPDB_UInt GetNumAdditionalPagesAppended()
    {
        SAPDB_UInt parts = m_CurrentPart.IsValid() ? m_CurrentPart+1 : 0;
        return parts - m_NumPagesAppended;
    }
    
private:

    /*! ---------------------------------------------------------------------------
    declaration: Part
    description: Consists of the Page and the Length, Value within this Page
    */
    typedef struct
    {
        T           Page;
        SAPDB_UInt  Length;
        SAPDB_Byte* Value;
    } Part;
    /*! endDeclaration: Part */
    
    /*! ---------------------------------------------------------------------------
    variable:    m_CurrentPart
    */
    PartNo m_CurrentPart;
    
    /*! ---------------------------------------------------------------------------
    variable:    m_Parts
    */
    Container_Vector<Part> m_Parts;

    /*! ---------------------------------------------------------------------------
    variable:    m_NumPagesAppended
    */
    // PTS 1115979 UH 2002-05-29 changed from bool to int
    SAPDB_UInt m_NumPagesAppended;
};
/*! endclass: Data_SplitSpace */


#endif // Data_SplitSpace_HPP
