/****************************************************************************

  module      : Data_TraverseTreeAccess.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area:
  description : 

  last changed: 2001-01-20
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg90.h"

#include "hbd50.h"
#include "hbd75.h"

#include "DataAccess/Data_TraverseTreeAccess.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

Data_TraverseTreeAccess::~Data_TraverseTreeAccess()
{
    m_Page.Deassign( false ); // release the page for read

    if( r_lock_leaf == m_Current.curr_lockstate )
        b75luread_unlock_leaf( m_Current, m_Current.curr_node_locked );
    else
        if( r_lock_index == m_Current.curr_lockstate )
            b75iuread_unlock_index( m_Current, m_Current.curr_lvl_1_locked );
}

/*---------------------------------------------------------------------------*/

Data_PrimPage&
Data_TraverseTreeAccess::GetNode()
{
    SAPDBTRACE_ROUTINE_DEBUG( "Data_TraverseTreeAccess::GetNode", DataTree_Trace, 5 );

    if( ! m_Page.IsAssigned() )
    {
        SetToRootNode();
        return( m_Page );
    }

    const Data_PageNo currentPageNo = m_Page.PageNo();
    const Data_PageNo nextPageNo    = m_Page.Successor();
    const Data_Level  currentLevel  = m_Page.GetLevel();

    m_PageAccessManager.ReleasePage( m_Page );

    if( nextPageNo.IsValid() )
    {
        if( IsTreeLockUsed() )
        {
            if( currentLevel == LEAF_LEVEL_BD00 )
                ReplaceLeafNodeLock( nextPageNo );
            else if(( IsDynamicTreeLockUsed() ) && ( currentLevel == FIRST_INDEX_LEVEL_BD00 ))
                ReplaceFirstLevelIndexNodeLock( currentPageNo, nextPageNo );
        }
        if( e_ok == m_TrError )
        {
            m_PageAccessManager.GetPage( m_Page, nextPageNo );
            return( m_Page );
        }
    }
    else if( LEAF_LEVEL_BD00 < currentLevel )
    {
        if( GetLeftMostNodeAtNextLevel( currentLevel -1 ))
            return( m_Page );
    }
    // error handling
    if( m_Page.IsAssigned() )
        m_PageAccessManager.ReleasePage( m_Page );

    return m_Page;
}

/*---------------------------------------------------------------------------*/
/*                            private methods                                */
/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_TraverseTreeAccess::SetToRootNode()
{
    SAPDBERR_ASSERT_STATE( ! m_Page.IsAssigned() );

    const Data_PageNo   rootPageNo = m_Current.curr_tree_id.fileRoot_gg00();
    Data_AccessMode     accessMode = Data_ForRead;

    // Root node access with correct tree lock
    if( IsTreeLockUsed() )
    {
        if( IsDynamicTreeLockUsed() )
            accessMode = Data_ForFirstIndexOrLeafAccess;
        else
            accessMode = Data_ForLeafAccess;
    }
    m_PageAccessManager.GetPage( m_Page, rootPageNo, accessMode );
    
    if(( e_root_is_leaf == m_TrError ) || ( e_root_is_first_index == m_TrError))
    {
        if( IsDynamicTreeLockUsed() )
            b75ilread_lock_index( m_Current, rootPageNo, 218 );

        if( e_root_is_leaf == m_TrError )
            b75llread_lock_leaf( m_Current, rootPageNo, 219 );

        m_TrError = e_ok; // reset error code

        m_PageAccessManager.GetPage( m_Page, rootPageNo, Data_ForRead );
    }
    m_Page.SetAccessMode( Data_ForRead );
    return( e_ok == m_TrError );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
Data_TraverseTreeAccess::GetLeftMostNodeAtNextLevel( const Data_Level   stopLevel )
{
    SAPDBERR_ASSERT_STATE( ! m_Page.IsAssigned() );

    if( IsTreeLockUsed() )
        b75reset_lock( m_Current, m_Current.curr_node_locked );

    SetToRootNode();

    while( e_ok == m_TrError )
    {
        const tgg00_Rec     *pSeparator  = m_Page.GetFirstRecord();
        const Data_PageNo   nextPageNo   = pSeparator->recPno_gg00().gg90GetInt4();
        const Data_Level    currentLevel = m_Page.GetLevel();

        m_PageAccessManager.ReleasePage( m_Page );

        if( IsTreeLockUsed() )
        {
            if(( SECOND_INDEX_LEVEL_BD00 == currentLevel ) && ( IsDynamicTreeLockUsed()))
                b75ilread_lock_index( m_Current, nextPageNo, 205 );

            if( FIRST_INDEX_LEVEL_BD00 == currentLevel )
                b75llread_lock_leaf( m_Current, nextPageNo, 206 );
        }
        if( !m_PageAccessManager.GetPage( m_Page, nextPageNo ))
            return( false );

        if( stopLevel ==  m_Page.GetLevel() )
            return( true );
    }
    return( false );
}

/*---------------------------------------------------------------------------*/

void
Data_TraverseTreeAccess::ReplaceLeafNodeLock( const Data_PageNo nextPageNo )
{
    SAPDBERR_ASSERT_STATE( ! m_Page.IsAssigned() );
    SAPDBERR_ASSERT_STATE( nextPageNo.IsValid() );
    SAPDBERR_ASSERT_STATE( r_lock_leaf == m_Current.curr_lockstate );

    if(
        ( m_Current.curr_tree_id.fileType_gg00().includes( ftsDynamic_egg00 )) &&
        ( m_Current.curr_right_bound == m_Current.curr_node_locked           )
    )
    {
        b50cross_subtree_boundary( m_Current, m_Current.curr_node_locked, nextPageNo );

        b75llread_lock_leaf( m_Current, nextPageNo, 220 );
    }
    else
        b75lulread_unlocklock_leaf( m_Current, m_Current.curr_node_locked,
                                    m_Current.curr_lvl_1_locked, nextPageNo,
                                    m_Current.curr_lvl_1_locked, 215 );
}

/*---------------------------------------------------------------------------*/

void
Data_TraverseTreeAccess::ReplaceFirstLevelIndexNodeLock(
    const Data_PageNo currentPageNo,
    const Data_PageNo nextPageNo )
{
    SAPDBERR_ASSERT_STATE( ! m_Page.IsAssigned() );
    SAPDBERR_ASSERT_STATE( nextPageNo.IsValid() );
    SAPDBERR_ASSERT_STATE( r_lock_index == m_Current.curr_lockstate );

    b75iuiread_unlocklock_index( m_Current, currentPageNo, nextPageNo, 208 );
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
