  /*!***************************************************************************

  module:       Data_Types.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH
  author:          UweH,TillL

  special area: DataAccess

  description:  Base types of this component

                - Data_PageNo:              Logical page numbers
                - Data_PageRecoveryMode:    Page recovery modes 
                                            (temp, perm, static, dynamic)
                - Data_PageId:              == PageNo + RecoveryMode
                
  see also:     

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

  *****************************************************************************/


#ifndef DATA_TYPES_HPP
#define DATA_TYPES_HPP


#include "ggg91.h"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/SAPDB_EnhancedTypes.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDBCommon/Diagnostics/SAPDBDiag_CheckTopic.hpp"

/* external defined trace topic */
extern SAPDBTrace_Topic     DataChain_Trace;
extern SAPDBTrace_Topic     DataPam_Trace;
extern SAPDBTrace_Topic     DataTree_Trace;
extern SAPDBTrace_Topic     DataPagePrim_Trace;
extern SAPDBTrace_Topic     DataPageInv_Trace;
extern SAPDBTrace_Topic     DataPageLog_Trace;

extern SAPDBDiag_CheckTopic DataChain_Check;
extern SAPDBDiag_CheckTopic DataPam_Check;
extern SAPDBDiag_CheckTopic DataTree_Check;
extern SAPDBDiag_CheckTopic DataPagePrim_Check;
extern SAPDBDiag_CheckTopic DataPageInv_Check;
extern SAPDBDiag_CheckTopic DataPageLog_Check;

/*! declaration: AccessMode
    description: Indicates how the page is cached within the PageAccessManager.

    Data_ForRead:            the page is read locked
    Data_ForIntend           the page is locked intending an update
    Data_ForUpdate           the page is locked exclusively
    Data_ForStructureChange  like ForUpdate, but the savepoint is serialized with this lock
    Data_ForFirstIndexOrLeafAccess return if the root page is leaf or on first index level
    Data_ForLeafAccess       return if the root is leaf
 */
enum Data_AccessMode
{
    Data_ForRead = 0,
    Data_ForIntend = 1,
    Data_ForUpdate = 2,
    Data_ForStructureChange = 3,
    Data_ForFirstIndexOrLeafAccess = 4,
    Data_ForLeafAccess = 5,
    Data_AccessMode_DONT_USE = 6
};

/*! declaration: Data_AccessModeStrings */
extern const char * Data_AccessModeStrings [];

/*! declaration: ReleaseMode
    description: indicates how the persistent page is released in the PageAccessManager
 */
enum Data_ReleaseMode
{
    Data_Changed = 0,
    Data_Unchanged = 1,
    Data_ReleaseMode_DONOT_USE = 2
};

/*! declaration: Data_ReleaseModeStrings */
extern const char * Data_ReleaseModeStrings [];

/*! declaration: Data_LRUChainInfo
    description: indicates how the persistent page is chained within the data cache
                 LRU List. Data_LruNormal leaves the page at the current position
                 assumed the page is not in the last 10 percent of the LRU list. In 
                 this case the page will be rechained to the begining. Data_LruMiddle
                 puts the page in front of the last 10 percent of the LRU list and
                 Data_LruLast inserts the page at the end.
 */
enum Data_LRUChainInfo
{
    Data_LruNormal = 0, 
    Data_LruMiddle = 1,
    Data_LruLast = 2,
    Data_LRUChainInfo_DONT_USE = 3
};

/*! declaration: Data_LRUChainInfoStrings */
extern const char * Data_LRUChainInfoStrings [];

/*!-----------------------------------------------------------------------------
declaration: Data_AddressingMode
description: page addressing mode within the converter
------------------------------------------------------------------------------*/
enum Data_AddressingMode
{
    Data_Static = 0,            // == physical
    Data_Dynamic = 1,           // == logical
    Data_AddressingModes =  2 // count
};

/*! declaration: Data_LRUChainInfoStrings */
extern const char * Data_AddressingModeStrings [];

/*!-----------------------------------------------------------------------------
declaration: Data_Recoverability
description: recoverability of a data page
------------------------------------------------------------------------------*/
enum Data_Recoverability
{
    Data_Recoverable = 0,       // == permanent
    Data_NotRecoverable = 1,    // == temporary
    Data_RecoveryModes = 2 // count
};

/*! declaration: Data_LRUChainInfoStrings */
extern const char * Data_RecoverabilityStrings [];

/*!-----------------------------------------------------------------------------
declaration: Data_PageCount
description: count of pages
------------------------------------------------------------------------------*/
typedef SAPDB_UInt4 Data_PageCount;

/*!-----------------------------------------------------------------------------
declaration: Data_PageNo
description: handles a logical page number (former: pno)
------------------------------------------------------------------------------*/
typedef SAPDB_Invalidatable<SAPDB_Int4, SAPDB_MAX_INT4> Data_PageNo;


/*!*****************************************************************************

    class: Data_PageRecoveryMode

    description: Handles the attributes which define the recovery mode of a data page

     Defines an enumeration type Modes which offers the supported modes

    example:

     Data_PageRecoveryMode recoverymode(Data_Static,  Data_NotRecoverable);
     Data_PageRecoveryMode recoverymode(Data_Static,  Data_Recoverable);
     Data_PageRecoveryMode recoverymode(Data_Dynamic, Data_Recoverable);
     ...
     if (recoverymode.UseDynamicPageAddressing()) ...
     if (recoverymode.UseStaticPageAddressing()) ...
     if (recoverymode.PageIsRecoverable()) ...
     ...

*******************************************************************************/
class Data_PageRecoveryMode
{
public:

    /*!-------------------------------------------------------------------------
    function:     Data_PageRecoveryMode()
    description:  Default Constructor;  optionally sets mode
    arguments:    addrmode [in]  addressing mode (default: Data_Dynamic)
                  recmode  [in]  recovery mode (default: Data_Recoverable)  
    return value: none
    --------------------------------------------------------------------------*/
    Data_PageRecoveryMode
        (Data_AddressingMode            addrmode = Data_Dynamic,
         Data_Recoverability            recmode  = Data_Recoverable)
        : 
        m_AddressingMode(addrmode),
        m_Recoverability(recmode)
    {}
   

    /*!-------------------------------------------------------------------------
    function:     SetMode()
    description:  sets mode
    arguments:    addrmode [in]  addressing mode
                  recmode  [in]  recovery mode   
    return value: none
    --------------------------------------------------------------------------*/
    void SetMode  (Data_AddressingMode            addrmode,
                   Data_Recoverability            recmode)
	{
		m_AddressingMode = addrmode;
        m_Recoverability = recmode;
	}

    /*!-------------------------------------------------------------------------
    function:     PageIsRecoverable()
    description:  Checks whether page is recoverable
    arguments:    none
    return value: (SAPDB_Bool)
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIsRecoverable() const
    {
        return m_Recoverability == Data_Recoverable;
    }

    /*!-------------------------------------------------------------------------
    function:     UseStaticPageAddressing()
    description:  Checks whether static page addressing is to be used
    arguments:    none
    return value: (SAPDB_Bool)
    --------------------------------------------------------------------------*/
    SAPDB_Bool UseStaticPageAddressing() const
    {
        return m_AddressingMode == Data_Static;
    }

    /*!-------------------------------------------------------------------------
    function:     UseDynamicPageAddressing()
    description:  Checks whether dynamic page addressing is to be used
    arguments:    none
    return value: (SAPDB_Bool)
    --------------------------------------------------------------------------*/
    SAPDB_Bool UseDynamicPageAddressing() const
    {
        return m_AddressingMode == Data_Dynamic;
    }

    /*!-------------------------------------------------------------------------
    function:     GetAddressingMode()
    description:  Returns page addressing mode
    arguments:    none
    return value: (Data_AddressingMode)
    --------------------------------------------------------------------------*/
    Data_AddressingMode GetAddressingMode() const
    {
        return m_AddressingMode;
    }

    /*!-------------------------------------------------------------------------
    function:     GetRecoverability()
    description:  Returns page recoverability
    arguments:    none
    return value: (Data_Recoverability)
    --------------------------------------------------------------------------*/
    Data_Recoverability GetRecoverability() const
    {
        return m_Recoverability;
    }

    /*!-------------------------------------------------------------------------
    function:     operator ==
    --------------------------------------------------------------------------*/
    bool operator == (const Data_PageRecoveryMode& Other) const
    {
        return Other.m_AddressingMode == m_AddressingMode
               &&
               Other.m_Recoverability == m_Recoverability;
    }

    /*!-------------------------------------------------------------------------
    function:     operator !=
    --------------------------------------------------------------------------*/
    bool operator != (const Data_PageRecoveryMode& Other) const
    {
        return ! this->operator == (Other);
    }
    
private:
    Data_AddressingMode     m_AddressingMode;
    Data_Recoverability     m_Recoverability;

};
/*!*****************************************************************************

   endclass: Data_PageRecoveryMode

*******************************************************************************/




/*!*****************************************************************************

   class: Data_PageId

   description:
   - Handles a logical page id
   - A page id is described by logical page number (former: pno)
     and page recovery mode.

*******************************************************************************/
class Data_PageId
{
public:

    /*!-------------------------------------------------------------------------
    function:     Data_PageId()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Data_PageId()
    {}

    /*!-------------------------------------------------------------------------
    function:     Data_PageId()
    description:  Constructor, assigns page recovery mode (pageno is undefined)
    arguments:    mode [in]  page recovery mode
    return value: none
    --------------------------------------------------------------------------*/
    Data_PageId
        (const Data_PageRecoveryMode&   mode)
        : 
        m_RecoveryMode(mode)
    {}

    /*!-------------------------------------------------------------------------
    function:     Data_PageId()
    description:  Constructor, assigns page number and recovery mode
    arguments:    pno [in]   page number
                  mode [in]  page recovery mode
    return value: none
    --------------------------------------------------------------------------*/
    Data_PageId
        (const Data_PageNo&             pno, 
         const Data_PageRecoveryMode&   mode)
        : 
        m_PageNo(pno), 
        m_RecoveryMode(mode)
    {}

    /*!-------------------------------------------------------------------------
    function:     operator=()
    description:  Assignment operator
    arguments:    other [in] page id to be assigned
    return value: *this
    --------------------------------------------------------------------------*/
    Data_PageId& operator=
        (const Data_PageId&             other)
    {
        m_PageNo       = other.m_PageNo;
        m_RecoveryMode = other.m_RecoveryMode;
        return *this;
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageNo()
    description:  Sets logical page number
    arguments:    pno [in]   page number
    return value: none
    --------------------------------------------------------------------------*/
    void SetPageNo
        (const Data_PageNo&             pno)
    {
        m_PageNo = pno;
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageRecoveryMode()
    description:  Sets page recovery mode
    arguments:    kind [in]   page recovery mode
    return value: none
    --------------------------------------------------------------------------*/
    void SetPageRecoveryMode
        (const Data_PageRecoveryMode&   mode)
    {
        m_RecoveryMode = mode;
    }

    /*!-------------------------------------------------------------------------
    function:     PageNo()
    description:  Returns const reference to pageno member
    arguments:    none
    return value: (const Data_PageNo&)
    --------------------------------------------------------------------------*/
    const Data_PageNo& PageNo() const
    {
        return m_PageNo;
    }

    /*!-------------------------------------------------------------------------
    function:     PageRecoveryMode()
    description:  Returns const reference to page recovery mode member
    arguments:    none
    return value: (const Data_PageRecoveryMode&)
    --------------------------------------------------------------------------*/
    const Data_PageRecoveryMode& PageRecoveryMode() const
    {
        return m_RecoveryMode;
    }

    /*!-------------------------------------------------------------------------
    function:     Invalidate()
    description:  Invalidates the PageId
    arguments:    none
    --------------------------------------------------------------------------*/
    void Invalidate()
    {
        m_PageNo.Invalidate();
    }

    /*!-------------------------------------------------------------------------
    function:     IsValid()
    description:  returns true, if PageId is valid
    returns:      bool
    --------------------------------------------------------------------------*/
    bool IsValid() const
    {
        return m_PageNo.IsValid();
    }
private:
    /*!-------------------------------------------------------------------------
    declaration: m_PageNo
    description: logical page number
    --------------------------------------------------------------------------*/
    Data_PageNo            m_PageNo;

    /*!-------------------------------------------------------------------------
    declaration: m_RecoveryMode
    description: page recovery mode
    --------------------------------------------------------------------------*/
    Data_PageRecoveryMode  m_RecoveryMode;

};
/*!*****************************************************************************

   endclass: Data_PageId

*******************************************************************************/

/*!-----------------------------------------------------------------------------
declaration: Data_PageOffset
description: page offset in bytes within one page
------------------------------------------------------------------------------*/
typedef SAPDB_UInt4 Data_PageOffset;

/*!-----------------------------------------------------------------------------
declaration: Data_RecordLength
description: the length of a field within a page
------------------------------------------------------------------------------*/
typedef Data_PageOffset Data_RecordLength;

/*!-----------------------------------------------------------------------------
declaration: Data_SplitRecordLength
description: the length of a split field.
------------------------------------------------------------------------------*/
typedef SAPDB_UInt4 Data_SplitRecordLength;

/*!-----------------------------------------------------------------------------
declaration: Data_Level
description: the depth of a b*tree. Note that leaf pages are of 
             level LEAF_LEVEL_BD00.
------------------------------------------------------------------------------*/
typedef SAPDB_UInt1 Data_Level;

/*!-----------------------------------------------------------------------------
declaration: Data_RecordIndex
description: specifies a slot within the pointer list which contains the
             Data_PageOffset of a record. Note that the record with the 
             smallest key has always a Data_RecordIndex of FIRST_REC_INDEX_BD00.
------------------------------------------------------------------------------*/
typedef SAPDB_UInt4 Data_RecordIndex;

/*!-----------------------------------------------------------------------------
declaration: Data_RecordCount
description: specifies a number if records
------------------------------------------------------------------------------*/

typedef SAPDB_UInt4 Data_RecordCount;

#endif // DATA_TYPES_HPP
