/*!**************************************************************************

  module      : FrameCtrl_Manager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: FrameCtrl_Manager
  description : 


  last changed: 2001-02-05  17:19
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef FRAMECTRL_MANAGER_HPP
#define FRAMECTRL_MANAGER_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

// Interfaces
#include "FrameControl/FrameCtrl_ICommon.hpp"
#include "FrameControl/FrameCtrl_IConverter.hpp"
#include "FrameControl/FrameCtrl_IIOManager.hpp"
#include "FrameControl/FrameCtrl_ILog.hpp"
#include "FrameControl/FrameCtrl_IDataCache.hpp"

#include "FrameControl/FrameCtrl_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "KernelCommon/Kernel_IPage.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: FrameCtrl_Manager

   description: This class is 
   
              
*******************************************************************************/

class FrameCtrl_Manager :
            public FrameCtrl_ICommon,
            public FrameCtrl_IConverter,
            public FrameCtrl_IDataCache,
            public FrameCtrl_IIOManager,
            public FrameCtrl_ILog
{
    friend class FrameCtrl_ICommon;
    friend class FrameCtrl_IConverter;
    friend class FrameCtrl_IDataCache;
    friend class FrameCtrl_IIOManager;
    friend class FrameCtrl_ILog;

public:

    /*!-------------------------------------------------------------------------
    function:     CreateInstance
    description:  This method creates the singleton frame control.It is assumed 
                  that this method will be callled only one time by the timeout 
                  task and therefore no synchonization mechanism is used.
    arguments:    none
    return value: true means instance could be created; else false
    --------------------------------------------------------------------------*/

    static bool CreateInstance();

    /*!-------------------------------------------------------------------------
    function:     GetFreeFrameCount()
    description:  This method is used to get the number of frames currently 
                  available and therefore requestable by the new methods.
    arguments:    none       
    return value: SAPDB_UInt Number of unused frames
    --------------------------------------------------------------------------*/

    virtual SAPDB_UInt GetFreeFrameCount() const;

    /*!-------------------------------------------------------------------------
    function:     NewConvFrame()
    description:  This method is used to provide a converter page frame.
                  page. If no more memory is available the component frame control 
                  executes an emergency shutdown.
    arguments:    taskId [in] Identification of calling task
    return value: (Kernel_IPage::PageFrame) New page frame handle
    --------------------------------------------------------------------------*/

    virtual Kernel_IPage::PageFrame NewConvFrame( tsp00_TaskId taskId )
    {
        return( NewFrame( FrameCtrl_ConverterKey, taskId ));
    }

    /*!-------------------------------------------------------------------------
    function:     FreeConvFrame()
    description:  This method is used to release a converter page frame.
    arguments:    taskId [in] Identification of calling task
                  frame  [in] Handle of the page frame to release         
    return value: none
    --------------------------------------------------------------------------*/

    virtual void FreeConvFrame(
        tsp00_TaskId            taskId,
        Kernel_IPage::PageFrame &frame )
    {
        Free( FrameCtrl_ConverterKey, taskId, frame );
    }

    /*!-------------------------------------------------------------------------
    function:     NewIOManFrame()
    description:  This method is used to provide a io manager page frame.
                  If no more memory is available the component frame control 
                  executes an emergency shutdown.
    arguments:    taskId [in] Identification of calling task
    return value: (Kernel_IPage::PageFrame) New page frame handle
    --------------------------------------------------------------------------*/

    virtual Kernel_IPage::PageFrame NewIOManFrame( tsp00_TaskId taskId )
    {
        return( NewFrame( FrameCtrl_IOManagementKey, taskId ));
    }

    /*!-------------------------------------------------------------------------
    function:     FreeIOManFrame()
    description:  This method is used to release a io manager page frame.
    arguments:    taskId [in] Identification of calling task
                  frame  [in] Handle of the page frame to release         
    return value: none
    --------------------------------------------------------------------------*/

    virtual void FreeIOManFrame(
        tsp00_TaskId            taskId,
        Kernel_IPage::PageFrame &frame )
    {
        Free( FrameCtrl_IOManagementKey, taskId, frame );
    }

    /*!-------------------------------------------------------------------------
    function:     NewLogFrame()
    description:  This method is used to provide a log page frame.
                  If no more memory is available the component frame control 
                  executes an emergency shutdown.
    arguments:    taskId [in] Identification of calling task
    return value: (Kernel_IPage::PageFrame) New page frame handle
    --------------------------------------------------------------------------*/

    virtual Kernel_IPage::PageFrame NewLogFrame( tsp00_TaskId taskId )
    {
        return( NewFrame( FrameCtrl_LogKey, taskId ));
    }

    /*!-------------------------------------------------------------------------
    function:     FreeLogFrame()
    description:  This method is used to release a log page frame.
    arguments:    taskId [in] Identification of calling task
                  frame  [in] Handle of the page frame to release         
    return value: none
    --------------------------------------------------------------------------*/

    virtual void FreeLogFrame(
        tsp00_TaskId            taskId,
        Kernel_IPage::PageFrame &frame )
    {
        Free( FrameCtrl_LogKey, taskId, frame );
    }

    /*!-------------------------------------------------------------------------
    function:     NewDataFrame()
    description:  This method is used to provide a data page frame.
                  page. If no more memory is available the component frame control 
                  executes an emergency shutdown.
    arguments:    taskId [in] Identification of calling task
    return value: (Kernel_IPage::PageFrame) New page frame handle
    --------------------------------------------------------------------------*/

    virtual Kernel_IPage::PageFrame NewDataFrame( tsp00_TaskId taskId )
    {
        return( NewFrame( FrameCtrl_DataCacheKey, taskId ));
    }

    /*!-------------------------------------------------------------------------
    function:     NewAWEDataFrame()
    description:  This method is used to provide a data page frame managed
                  by the Address Windowing Extension (AWE) technic of microsoft.
                  If no more memory is available the component frame control 
                  executes an emergency shutdown. 
                  All page frames provided by this call are still associated with
                  a physical page and therefore member of the AWE Map Window.
                  To free the obtained page frame the common FreeDataFrame method 
                  could be used.
    arguments:    taskId     [in ] Identification of calling task
                  frame      [out] New page frame handle
                  aweBlockNo [out] Internal AWE handle to manage AWE page frames
    return value: none
    --------------------------------------------------------------------------*/

    virtual void NewAWEDataFrame(
        tsp00_TaskId            taskId,
        Kernel_IPage::PageFrame &frame,
        SAPDB_UInt4             &aweBlockNo )
    {
        New( FrameCtrl_DataCacheKey, taskId, frame, aweBlockNo );
    }

    /*!-------------------------------------------------------------------------
    function:     FreeDataFrame()
    description:  This method is used to release a data page frame.
    arguments:    taskId [in] Identification of calling task
                  frame  [in] Handle of the page frame to release         
    return value: none
    --------------------------------------------------------------------------*/

    virtual void FreeDataFrame(
        tsp00_TaskId            taskId,
        Kernel_IPage::PageFrame &frame )
    {
        Free( FrameCtrl_DataCacheKey, taskId, frame );
    }

protected:

    // // protected and not private to avoid warnings on linux!
    
    FrameCtrl_Manager()
    :
    m_Capacity( 0 ),
    m_PageSize( 0 ),
    m_CheckedOutByConverter( 0 ),
    m_CheckedOutByDataCache( 0 ),
    m_CheckedOutByLogQueue( 0 ),
    m_CheckedOutByIOManagement( 0 )
    {}
    
    void Initialize( const SAPDB_UInt PageSize )
    {
        m_Capacity = GetFreeFrameCount();
        m_PageSize = PageSize;
    }

private:

    // Copy and assignment operator are not supported

    FrameCtrl_Manager& operator=( const FrameCtrl_Manager &manager );

    FrameCtrl_Manager( const FrameCtrl_Manager &manager );

    static FrameCtrl_Manager& GetInstance(){ return( *m_Instance ); }

    void MaintainInternalCounter(
        const FrameCtrl_RegistrationKey registrationKey,
        const SAPDB_Bool                bIncrementCounter = true );

    Kernel_IPage::PageFrame NewFrame(
        const FrameCtrl_RegistrationKey   registrationKey,
        const tsp00_TaskId                taskId )
    {
        SAPDB_UInt4             dummyBlockNo;
        Kernel_IPage::PageFrame frame;

        New( registrationKey, taskId, frame, dummyBlockNo );
        return( frame );
    }

    void New(
        const FrameCtrl_RegistrationKey   registrationKey,
        const tsp00_TaskId                taskId,
        Kernel_IPage::PageFrame           &frame,
        SAPDB_UInt4                       &aweBlockNo );

    void Free(
        const FrameCtrl_RegistrationKey   registrationKey,
        const tsp00_TaskId                taskId,
        Kernel_IPage::PageFrame           &frame );

private:

    /*-------------------------------------------------------------------------
    declaration: m_Instance
    description: Reference to the proper FrameControl Manager
    --------------------------------------------------------------------------*/

    static FrameCtrl_Manager    *m_Instance;

    /*-------------------------------------------------------------------------
    declaration: m_PageSize
    description: Size in bytes of the pages managed by the page allocator 
    --------------------------------------------------------------------------*/

    SAPDB_UInt m_PageSize;

    /*-------------------------------------------------------------------------
    declaration: m_Capacity
    description: Initial number of frames stored within the page allocator 
    --------------------------------------------------------------------------*/

    SAPDB_UInt m_Capacity;

    /*-------------------------------------------------------------------------
    declaration: m_CheckedOutByConverter
    description: Number of page frames checked out by the component converter
    --------------------------------------------------------------------------*/

    SAPDB_UInt m_CheckedOutByConverter;

    /*-------------------------------------------------------------------------
    declaration: m_CheckedOutByDataCache
    description: Number of page frames checked out by the component data cache
    --------------------------------------------------------------------------*/

    SAPDB_UInt m_CheckedOutByDataCache;

    /*-------------------------------------------------------------------------
    declaration: m_CheckedOutByLogQueue
    description: Number of page frames checked out by the component log queue
    --------------------------------------------------------------------------*/

    SAPDB_UInt m_CheckedOutByLogQueue;

    /*-------------------------------------------------------------------------
    declaration: m_CheckedOutByIOManagement
    description: Number of page frames checked out by the component I/O Management
    --------------------------------------------------------------------------*/

    SAPDB_UInt m_CheckedOutByIOManagement;

};

/*!*****************************************************************************

   endclass: FrameCtrl_Manager

*******************************************************************************/

#endif  /* FRAMECTRL_MANAGER_HPP */
