/*!**************************************************************************

  module      : FBM_Manager.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  auhtor      : AlexanderK

  special area: FreeBlockManagement (DDBM)
  description : 


  last changed: 2000-03-10  11:00
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin  GPL
    Copyright (C) 2002 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/

#include <new.h>                                  // SYSTEM: required for placement new
#include "hgg08.h"                                // region identifier
#include "Container/Container_Vector.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "FreeBlockManagement/FBM_Manager.hpp"
#include "FreeBlockManagement/FBM_DataDevice.hpp"
#include "FreeBlockManagement/FBM_Dump.hpp"
#include "FreeBlockManagement/FBM_Types.hpp"
#include "FreeBlockManagement/FBM_Exception.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"

/*===========================================================================*
 *  DEFINES                                                                   *
 *===========================================================================*/

#define FBM_MEM_INITIAL_SIZE            (8192 * 8)
#define FBM_MEM_SUPPLEMENT_SIZE         (8192 * 8)

/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* public Methodes of FBM_Manager                                            */
/*---------------------------------------------------------------------------*/

FBM_Manager&
FBM_Manager::CreateInstance (tgg00_BasisError TrError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::CreateInstance", FBM_Trace, 5);

    if (m_Instance == NULL)
    {
        m_Instance = new (RTEMem_Allocator::Instance())
                     FBM_Manager(RTEMem_Allocator::Instance());
        if (NULL == m_Instance)
            TrError = e_sysbuf_storage_exceeded;
    }
    return *m_Instance;
}

/*---------------------------------------------------------------------------*/

FBM_Manager::FBM_Manager(SAPDBMem_IRawAllocator &Allocator)
        :
        m_Allocator( UTF8( "FBM_Manager" ), Allocator ),
        m_DataDevice(m_Allocator)
{
    ResetMembersToNullAndZero();
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::Restart (
    const tsp00_TaskId                TaskId,
    const SAPDB_Int4                  MaxNumDev,
    const IOMan_ReservedBlockAddress  &ReservedBlocks)

{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::Restart", FBM_Trace, 5);

    FBM_SynchObject SynchronizeFBM( TaskId ); // PTS 1125910 2003-12-03

    /* check that the FBM_Manager is not already restarted */
    SAPDBERR_ASSERT_ARGUMENT (m_MaxNumDev  == FBM_UNDEFINED &&
                              m_NumDev                     == 0                   &&
                              m_TotalNumBlocks             == 0                   &&
                              m_TotalNumBlocksFree         == 0                   &&
                              m_TotalNumBlocksFreeAfterSVP == 0                   &&
                              m_SavepointIsRequested       == false               &&
                              m_NextDevToUseForNewBlock    == FBM_UNDEFINED       &&
                              m_NextDevToUseForFragBlock   == FBM_UNDEFINED);

    m_MaxNumDev      = MaxNumDev;
    m_ReservedBlocks = ReservedBlocks;

    /* construct array containing the number of free pages for each section */
    SAPDBTRACE_WRITELN( FBM_Trace, 5, "MaxNumDev: " << m_MaxNumDev);

    m_Active = m_DataDevice.Initialize (m_MaxNumDev, FirstDevNo());
    return m_Active;
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::Shutdown (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::Shutdown", FBM_Trace, 5);

    /* this method releases all memory resources */

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* release memory allocated by the FBM_Manager and its member */
    m_DataDevice.Shutdown();

    /* reset all member variables, usage of the FBM_Manager leads to errors now */
    ResetMembersToNullAndZero();
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::AddVolume(
    const tsp00_TaskId  TaskId,
    const SAPDB_Int2    DevNo,
    const SAPDB_Int4    DevSize)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::AddVolume", FBM_Trace, 5);

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevNo: " << DevNo << " DevSize: "  << DevSize);

    /* check validity of the device number */
    if ((DevNo < FirstDevNo()) || (DevNo > MaxDevNo()))
    {
        /* write error message into knldiag and vtrace */
        FBM_WrongDevNo WrongDevNo(__CONTEXT__, DevNo, FirstDevNo(), MaxDevNo());
        RTE_Message (WrongDevNo);
        return false;
    }
    {
        FBM_SynchObject SynchronizeFBM( TaskId ); // PTS 1125910 2003-12-03

        /* initialize device description and fill it with start values */
        if (!m_DataDevice [DevNo].AssignDevice (DevSize))
        {
            FBM_NoMoreMemoryForNewDevice NoMoreMemoryForNewDevice (__CONTEXT__, DevNo,DevSize);
            RTE_Message (NoMoreMemoryForNewDevice);
            return false;
        }

        /* update global block counters */
        m_NumDev             ++;
        m_TotalNumBlocks     += DevSize;
        m_TotalNumBlocksFree += DevSize;

        // Security limit is 4 percent of all blocks and not more as 4000 blocks
        m_SecuritySpaceLimit = ( m_TotalNumBlocks - SAPDB_MIN( m_TotalNumBlocks/25, 4000));

        /* reserve blocks on the device for example for the restart record */
        IOMan_ReservedBlockAddressIterator Iter( m_ReservedBlocks, DevNo );

        for( Iter.Begin(); !Iter.End(); ++Iter )
        {
            m_TotalNumBlocksFree --;
            m_DataDevice [DevNo].SetBlockStateToOccupied( Iter.GetBlockAddress().GetBlockNo());
        }

        // After marking the reserved blocks the devspace is usable
        m_NextDevToUseForNewBlock = DevNo;
    }
    return true;
}

/*---------------------------------------------------------------------------*/
void
FBM_Manager::SetClusterStateToFree (tsp00_TaskId              TaskId,
                                    const IOMan_ClusterAddress &ClusterAddress)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetClusterStateToFree", FBM_Trace, 5);

    IOMan_BlockCount pageCnt = ClusterAddress.GetBlockCount();

    for (SAPDB_Int4 loop=0; loop < pageCnt; loop++){
        SetBlockStateToFree(TaskId,ClusterAddress.GetBlockAddress(loop));
    }
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::SetBlockStateToFree (tsp00_TaskId              TaskId,
                                  const IOMan_BlockAddress &BlockAddress)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToFree", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetFree: DevNo: " << DevNo << " BlockNo: " << BlockNo);

    /* check that the block containing the restart record is never set to free */
    SAPDBERR_ASSERT_STATE ( ! m_ReservedBlocks.IsMember( BlockAddress ));

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    /* check that the current state is  occupied */
    switch (Device.GetBlockState (BlockNo))
    {
    case BlockState_Occupied:
        /* set block state to free and update all auxilliary structures */
        m_TotalNumBlocksFree ++;
        Device.SetBlockStateToFree (BlockNo);
        return;
    case BlockState_BackUp:
        if (Device.GetBlockStateAfterBackup (BlockNo) == BlockState_BackUp)
        {
            /* set  back up block state to free */
            Device.SetBlockStateAfterBackup (BlockNo, BlockState_Free);
            return;
        }
    }
    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace               */
    FBM_IllegalTransition ErrorMessage (__CONTEXT__, DevNo, BlockNo,
                                        Device.GetBlockState (BlockNo),
                                        Device.GetBlockStateAfterBackup (BlockNo),
                                        BlockState_Free,
                                        BlockState_Free);
    RTE_Crash(ErrorMessage);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
FBM_Manager::SetBlockStateToOccupied (
    const tsp00_TaskId        TaskId,
    IOMan_IBlockAddressIterator &BlockIterator,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToOccupied", FBM_Trace, 5);

    FBM_SynchObject SynchronizeFBM (TaskId);

    while (BlockIterator.hasMoreElements()){
        IOMan_BlockAddress BlockAddress = BlockIterator.getNextElement();

        if (!SetBlockStateToOccupiedUnsynched ( BlockAddress, bAbortIfErrorOccured))
		{
			return SAPDB_FALSE;
		}
    }

    return SAPDB_TRUE;
}

SAPDB_Bool
FBM_Manager::SetBlockStateToOccupied(
    const tsp00_TaskId        TaskId,
    const IOMan_BlockAddress &BlockAddress,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    FBM_SynchObject SynchronizeFBM (TaskId);
    return SetBlockStateToOccupiedUnsynched ( BlockAddress, bAbortIfErrorOccured);
}

SAPDB_Bool
FBM_Manager::SetBlockStateToOccupiedUnsynched (
    const IOMan_BlockAddress &BlockAddress,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToOccupied", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetOcc: DevNo: " << DevNo << " BlockNo: " << BlockNo );

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    /* check that the current state is free */
    if (Device.GetBlockState (BlockNo) == BlockState_Free)
    {
        SAPDBERR_ASSERT_STATE (Device.GetBlockStateAfterBackup (BlockNo) == BlockState_Free);

        m_TotalNumBlocksFree --;

        /* set block state to occupied and update all auxilliary structures */
        Device.SetBlockStateToOccupied (BlockNo);
        return( SAPDB_TRUE );
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace               */
    FBM_IllegalTransition errMsg (__CONTEXT__, DevNo, BlockNo,
                                  Device.GetBlockState (BlockNo),
                                  Device.GetBlockStateAfterBackup (BlockNo),
                                  BlockState_Occupied,
                                  BlockState_Free);

    if( FBM_ABORT_IF_ERROR == bAbortIfErrorOccured )
        RTE_Crash( errMsg );
    else
        RTE_Message( errMsg );

    return( SAPDB_FALSE );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
FBM_Manager::SetBlockStateToBackup (
    const tsp00_TaskId        TaskId,
    const IOMan_BlockAddress &BlockAddress,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToBackup", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetBackU DevNo " << DevNo << " BlockNo: " << BlockNo );

    /* check that the block containing the restart record is never set to backup */
    SAPDBERR_ASSERT_STATE ( ! m_ReservedBlocks.IsMember( BlockAddress ));

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    /* check that no more blocks are marked after reading of the marked blocks has started     */
    /* which is displayed by the iterator m_ActBlockNoForBackup indicating the last read block */
    SAPDBERR_ASSERT_STATE (Device.m_ActBlockNoForBackup == FBM_UNDEFINED);

    /* check that the current state is  occupied */
    if (Device.GetBlockState (BlockNo) == BlockState_Occupied)
    {
        Device.SetBlockStateToBackUp(BlockNo);
        return( SAPDB_TRUE );
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace */
    FBM_IllegalTransition errMsg (__CONTEXT__, DevNo, BlockNo,
                                  Device.GetBlockState (BlockNo),
                                  Device.GetBlockStateAfterBackup (BlockNo),
                                  BlockState_BackUp,
                                  BlockState_Occupied);

    if( FBM_ABORT_IF_ERROR == bAbortIfErrorOccured )
        RTE_Crash( errMsg );
    else
        RTE_Message( errMsg );

    return( SAPDB_FALSE );
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::SetBlockStateToFreeAfterSVP (
    tsp00_TaskId             TaskId,
    const IOMan_BlockAddress &BlockAddress)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToFreeAfterSVP", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetFSVP: DevNo: " << DevNo << " BlockNo: " << BlockNo );

    /* check that the block containing the restart record is never set to fsvp */
    SAPDBERR_ASSERT_STATE ( ! m_ReservedBlocks.IsMember( BlockAddress ));

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    /* check that the current state is  occupied */
    switch (Device.GetBlockState (BlockNo))
    {
    case BlockState_Occupied:
        Device.SetBlockStateToFreeAfterSVP (BlockNo);
        ++m_TotalNumBlocksFreeAfterSVP;
        return;
    case BlockState_BackUp:
        {
            switch (Device.GetBlockStateAfterBackup (BlockNo))
            {
            case BlockState_Occupied:
                Device.SetBlockStateAfterBackup (BlockNo, BlockState_FreeAfterSVP);
                return;
            case BlockState_BackUp:
                ++m_TotalNumBlocksFreeAfterSVP;
                Device.SetBlockStateMarkedForCompressionToFreeAfterSVP (BlockNo);
                return;
            }
        }
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace */
    FBM_IllegalTransitionToFreeAfterSVP ErrorMessage (__CONTEXT__, DevNo, BlockNo,
            Device.GetBlockState (BlockNo),
            Device.GetBlockStateAfterBackup (BlockNo));
    RTE_Crash(ErrorMessage);
}

/*---------------------------------------------------------------------------*/

IOMan_ClusterAddress
FBM_Manager::GetMultFreeBlocks(
    const tsp00_TaskId  taskId,
    const SAPDB_Int4    NumFreeBlocksWanted,
    tgg00_BasisError   &trError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetMultFreeBlocks", FBM_Trace, 5);

    IOMan_BlockAddress      block;
    IOMan_ClusterAddress    cluster;
    SAPDB_Bool              bTriggerSavepoint = false;
    SAPDB_Int4              NumFreeBlocksSupplied = 0;

    if( ! m_Active )
    {
        trError = e_shutdown;
        return cluster;
    }

    /* This following block is needed to leave the FBM region before entering the     */
    /* log region in the block below. This is done to avoid a dead lock. The FBM      */
    /* region is left when the FBM_SynchObject is destructed at the end of this block */
    {
        FBM_SynchObject SynchronizeFBM( taskId );

        /* get number of a device containing free blocks */
        const SAPDB_Int2 DevNo = GetDeviceWithFreeBlocks (NumFreeBlocksWanted);

        if (DevNo == FBM_UNDEFINED)
        {
            /* no free block could be found */
            FBM_Exception NoMoreSpace(__CONTEXT__, FBM_NO_MORE_FREE_BLOCKS);
            RTE_Message(NoMoreSpace);
            trError = e_no_more_perm_space;
            return cluster;
        }

        block.SetAddress( DevNo, m_DataDevice [DevNo].GetMultFreeBlocks( NumFreeBlocksWanted,
                          NumFreeBlocksSupplied ));
        cluster.SetCluster( block, NumFreeBlocksSupplied );

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "Free DevNo  " << cluster.GetDeviceNo()
                            << " Free BlockNo: " << cluster.GetBlockNo( 0 ));

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "Wanted Num: " << NumFreeBlocksWanted
                            << " Suppl.Num: " << NumFreeBlocksSupplied );

        /* decrease the counter of free blocks and check if there are too */
        /* few free blocks which means that a savepoint must be triggered */
        m_TotalNumBlocksFree -= NumFreeBlocksSupplied;
        if( SavepointNeeded() )
        {
            bTriggerSavepoint      = true;
            m_SavepointIsRequested = true;
        }
    }

    if( bTriggerSavepoint )
        TriggerSavepoint( taskId );

    return cluster;
}


/*---------------------------------------------------------------------------*/

IOMan_BlockAddress
FBM_Manager::GetFreeBlock( const tsp00_TaskId  taskId )

{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetFreeBlock", FBM_Trace, 5);

    SAPDB_Bool         bTriggerSavepoint = false;
    IOMan_BlockAddress BlockAddress;

    /* This following block is needed to leave the FBM region before entering the     */
    /* log region in the block below. This is done to avoid a dead lock. The FBM      */
    /* region is left when the FBM_SynchObject is destructed at the end of this block */
    {
        FBM_SynchObject SynchronizeFBM( taskId );

        /* get number of a device containing free blocks */
        const SAPDB_Int4 cNumFreeBlocksWanted = 1;
        const SAPDB_Int4 DevNo = GetDeviceWithFreeBlocks (cNumFreeBlocksWanted);

        if (DevNo == FBM_UNDEFINED)
        {
            /* no free block could be found */
            FBM_Exception NoMoreSpace(__CONTEXT__, FBM_NO_MORE_FREE_BLOCKS);
            RTE_Crash(NoMoreSpace);
        }

        BlockAddress.SetAddress (DevNo, m_DataDevice[DevNo].GetFreeBlock());

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "Free DevNo: " << BlockAddress.GetDeviceNo()
                            << " Free BlockNo: " << BlockAddress.GetBlockNo() );

        /* decrease the counter of free blocks and check if there are too */
        /* few free blocks which means that a savepoint must be triggered */
        -- m_TotalNumBlocksFree;

        if( SavepointNeeded() )
        {
            bTriggerSavepoint      = true;
            m_SavepointIsRequested = true;
        }
    }

    if( bTriggerSavepoint )
        TriggerSavepoint( taskId );

    return BlockAddress;
}

/*---------------------------------------------------------------------------*/

SAPDB_Int2
FBM_Manager::GetDeviceWithFreeBlocks (const SAPDB_Int2 NumRequestedBlocks)
{

    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetDeviceWithFreeBlocks", FBM_Trace, 5);

    /* search a data device which has still NumRequestedBlocks free blocks */
    SAPDB_Int2  SearchBegDevNo     = m_NextDevToUseForNewBlock;
    SAPDB_Int2  iDevNo             = SearchBegDevNo;
    SAPDB_Int4  DevWithFreeBlocks  = FBM_UNDEFINED;
    SAPDB_Bool  DeviceFound        = false;

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "FirstDevNo: " << FirstDevNo()
                        << " NumDevs: " << m_NumDev );

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SearchBegDev: " << SearchBegDevNo
                        << " BlocksWanted: " << NumRequestedBlocks );

    /* loop over all devices until a device with NumRequestedBlocks free blocks is found */
    do
    {
        if (m_DataDevice[iDevNo].NumBlocksFree() < NumRequestedBlocks)
        {
            SAPDBTRACE_WRITELN( FBM_Trace, 5, "Device Full: " << iDevNo);
        }
        else
        {
            DeviceFound       = true;
            DevWithFreeBlocks = iDevNo;

            /* if this device contains too few used blocks this device is choosen imediately.   */
            /* Otherwise this device is skipped this time and the search is continued. However, */
            /* m_NumBlocksToAddUntilOptIsReached is changed so that this device is choosen      */
            /* next time it is looked at                                                        */
            if (m_DataDevice[iDevNo].m_NumBlocksToAddUntilOptIsReached > 0)
                break;
            else
                m_DataDevice[iDevNo].m_NumBlocksToAddUntilOptIsReached = 1;
        }

        /* goto next device */
        if ((++iDevNo) > LastDevNo()) iDevNo = FirstDevNo();
    }
    while (iDevNo != SearchBegDevNo) ;

    /* return device or trow exception if there is no device with enough free blocks */
    if (!DeviceFound)
        DevWithFreeBlocks = FBM_UNDEFINED;
    else
    {
        /* set device where to start the next search */
        if (DevWithFreeBlocks == LastDevNo())
            m_NextDevToUseForNewBlock = FirstDevNo();
        else
            m_NextDevToUseForNewBlock = DevWithFreeBlocks+1;
    }

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "Free_Device: " << DevWithFreeBlocks);

    return (DevWithFreeBlocks) ;
}


/*---------------------------------------------------------------------------*/

SAPDB_Int4
FBM_Manager::NumBlocksMarkedForBackup( const tsp00_TaskId taskId ) const
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::NumBlocksMarkedForBackup", FBM_Trace, 5);

    SAPDB_Int4 NumAllBlocksMarkedForBackUp = 0;

    FBM_SynchObject SynchronizeFBM( taskId ); // PTS 1125910 2003-12-03

    for (SAPDB_Int4 iDevNo=FirstDevNo(); iDevNo<=LastDevNo(); ++iDevNo)
    {
#       ifdef SAPDB_SLOW 
        if (m_DataDevice [iDevNo].m_NumBlocksMarkedForBackup > 0)
            SAPDBTRACE_WRITELN( FBM_Trace, 5, "bup blocks  " <<  m_DataDevice [iDevNo].m_NumBlocksMarkedForBackup);
#       endif

        NumAllBlocksMarkedForBackUp += m_DataDevice [iDevNo].m_NumBlocksMarkedForBackup;
    }
    return NumAllBlocksMarkedForBackUp;
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreAllBlockStatesMarkedForBackup (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreAllBlockStatesMarkedForBackup", FBM_Trace, 5);

    SAPDB_Int4      iDevNo;
    SAPDB_Int4      NumBlocksRestoredToFreeAfterSVP;
    FBM_DataDevice *pDevice;

    FBM_SynchObject SynchronizeFBM (TaskId);

    for (iDevNo=FirstDevNo(), pDevice = & m_DataDevice [FirstDevNo()];
            iDevNo<=LastDevNo();
            iDevNo++, pDevice++)
    {
        if (pDevice->m_NumBlocksMarkedForBackup>0)
        {
            pDevice->RestoreAllBlockStatesMarkedForBackup(NumBlocksRestoredToFreeAfterSVP);
            m_TotalNumBlocksFreeAfterSVP += NumBlocksRestoredToFreeAfterSVP;
        }
    }
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreAllBlockStatesMarkedForCompression (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreAllBlockStatesMarkedForCompression", FBM_Trace, 5);

    SAPDB_Int4      iDevNo;
    FBM_DataDevice *pDevice;

    FBM_SynchObject SynchronizeFBM (TaskId);

    for (iDevNo=FirstDevNo(), pDevice = & m_DataDevice [FirstDevNo()];
            iDevNo<=LastDevNo();
            iDevNo++, pDevice++)
    {
        if (pDevice->m_DeviceIsInCompressingMode)
            pDevice->RestoreAllBlockStatesMarkedForCompression(iDevNo);
    }
}
/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreBlockStateMarkedForBackup (tsp00_TaskId             TaskId,
        const IOMan_BlockAddress &BlockAddress)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreBlockStateMarkedForBackup", FBM_Trace, 5);

    const SAPDB_UInt DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "RestoreB Dev: " << DevNo
                        << " BlockNo: " <<BlockNo );

    SAPDB_Bool bRestoredStateIsFreeAfterSVP;

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    if (Device.GetBlockState(BlockNo) == BlockState_BackUp )
    {
        Device.RestoreBlockStateMarkedForBackup (BlockNo, bRestoredStateIsFreeAfterSVP);
        if (bRestoredStateIsFreeAfterSVP)
            ++m_TotalNumBlocksFreeAfterSVP;
        return;
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace               */
    FBM_IllegalRestoreAfterBackup ErrorMesage (__CONTEXT__, DevNo, BlockNo,
            Device.GetBlockState(BlockNo),
            Device.GetBlockStateAfterBackup(BlockNo));
    RTE_Crash(ErrorMesage);
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::SetAllBlocksMarkedAsFreeAfterSVPToFree (tsp00_TaskId TaskId)
{
    /* go through all registered devices an set the states of all */
    /* blocks which are marked as free after savepoint to free    */

    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetAllBlocksMarkedAsFreeAfterSVPToFree", FBM_Trace, 5);

    FBM_SynchObject SynchronizeFBM (TaskId);

    SAPDB_Int4      iDevNo;
    FBM_DataDevice *pDevice;

    for (iDevNo=FirstDevNo(), pDevice = & m_DataDevice [FirstDevNo()];
            iDevNo<=LastDevNo();
            iDevNo++, pDevice++ )
    {

        if (pDevice->m_NumBlocksFreeAfterSVP>0)
        {
            /* update global counters of all blocks  free and free after save point */
            m_TotalNumBlocksFreeAfterSVP -= pDevice->m_NumBlocksFreeAfterSVP;
            m_TotalNumBlocksFree         += pDevice->m_NumBlocksFreeAfterSVP;

            pDevice->SetAllBlocksMarkedAsFreeAfterSVPToFree();
        }
    }

    /* update counters which control which devices are prefered when free blocks are   */
    /* requested  (getfreeblock) depending on the number of used blocks on each device */
    ChangeDeviceUsagePreferences ();

    /* reset the trigger flag,i.e. from now the routine GetFreeBlocks */
    /* can trigger again a savepoint if there are too few blocks     */
    m_SavepointIsRequested = false;

    /* check that the global counter of all blocks which are free are svp is zero now */
    SAPDBERR_ASSERT_STATE (m_TotalNumBlocksFreeAfterSVP == 0)
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreBlockStateMarkedForCompression (tsp00_TaskId       TaskId,
        const SAPDB_Int2   DevNo,
        const SAPDB_Int4   BlockNo,
        SAPDB_Bool        &bBlockInUse)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreBlockStateMarkedForCompression", FBM_Trace, 5);

    /* this method checks if a given block is in the state  [backup,backup] */
    /* which means that it is marked for compression and still in use then  */
    /* the original state is restored, the block can also be in the state   */
    /* [freeaftersavepouint, free] which means that there is nothing to do  */

    /* initialize return value */
    bBlockInUse = false;

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    const FBM_BlockState BlockState            = Device.GetBlockState           (BlockNo);
    const FBM_BlockState BlockStateAfterBackup = Device.GetBlockStateAfterBackup(BlockNo);

    switch (BlockState)
    {
    case  BlockState_BackUp:
        {
            switch (BlockStateAfterBackup)
            {
            case BlockState_BackUp:
                /* restore the state to occupied*/
                Device.SetBlockStateMarkedForCompressionToOccupied (BlockNo);
                bBlockInUse = true;
                return;
            case BlockState_Free:
                /* restore state to free */
                Device.SetBlockStateMarkedForCompressionToFree (BlockNo);
                return;
            }
        }
        break;
    case BlockState_FreeAfterSVP:
        SAPDBERR_ASSERT_STATE (BlockStateAfterBackup == BlockState_Free);
        return;
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace               */
    FBM_IllegalRestoreAfterCompression ErrorMesage (__CONTEXT__, DevNo, BlockNo, BlockState, BlockStateAfterBackup);
    RTE_Crash(ErrorMesage);
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::MarkFragmentedAndUnbalancedBlocks (
    tsp00_TaskId       TaskId,
    SAPDB_Int4         DeviceCompressionLimit,
    SAPDB_Int4         DeviceBalanceLimit,
    SAPDB_Bool1       &bFragmentedAndUnbalancedBlocksFound,
    tgg00_BasisError  &TrError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::MarkFragmentedAndUnbalancedBlocks", FBM_Trace, 5);

    /* this routine checks if the devices are too fragmented or unbalanced that the  */
    /* datawriters have to become activ in order to balncae and compress the devices */
    /* All blocks which need to be moved since there are too many blocks on a device */
    /* or in order to compress a device are marked with the state BlockState_BackUp  */
    /* misused for this purpose. This means that during a backup no defragmentation  */
    /* and balancing can be performed                                                */

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "compression: " << DeviceCompressionLimit
                        << " balance: " << DeviceBalanceLimit );

    /* initialize return value */
    bFragmentedAndUnbalancedBlocksFound = false;

    /* check input value */
    if (DeviceCompressionLimit<=0) return;

    double          MaxFragmentationFactor = (100.0 /(float ) DeviceCompressionLimit);
    SAPDB_Bool      bChangeDeviceUsage     = false ;
    SAPDB_Int4      NumBlocksOcc;        // occupied blocks are used blocks minus the blocks wich become free after SVP
    SAPDB_Int4      NumBlocksToMarkForRestructuring;

    FBM_SynchObject SynchronizeFBM (TaskId);

    SAPDB_Int4 iDevNo;

    /* check that no back up is running, if so restructuring is impossible */
    for (iDevNo=FirstDevNo(); iDevNo<=LastDevNo(); ++iDevNo)
    {
        if (m_DataDevice [iDevNo].m_NumBlocksMarkedForBackup > 0)
            return;
    }

    /* calculate the average number of used blocks per        */
    /* device and derive the maximum number of blocks which   */
    /* is tolerated before blocks are pushed from this device */
    SAPDB_Int4 OptimimumNumBlocksOcc     = OptimumNumOccBlocksPerDevice ();
    SAPDB_Int4 MaxNumUsedBlocksPerDevice = 1 + OptimimumNumBlocksOcc * (100 + (DeviceBalanceLimit>>1)) / 100;

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "OptNumBlocks: " << OptimimumNumBlocksOcc
                        << " MaxNumBlocks: " << MaxNumUsedBlocksPerDevice );

    FBM_DataDevice *pDevice;

    for (iDevNo=FirstDevNo(), pDevice = & m_DataDevice [FirstDevNo()];
            iDevNo<=LastDevNo();
            iDevNo++, pDevice++ )
    {
        NumBlocksOcc = pDevice->m_NumBlocksUsed - pDevice->m_NumBlocksFreeAfterSVP;

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "SearchDevNo: " << iDevNo
                            << " DevNo: " << iDevNo
                            << " Occ Blocks: " << NumBlocksOcc );

        /* check if the device is overcrowded */
        if ( NumBlocksOcc > MaxNumUsedBlocksPerDevice )
        {
            /* mark blocks for  diplacement from their current device */
            NumBlocksToMarkForRestructuring = NumBlocksOcc - OptimimumNumBlocksOcc ;
            if (NumBlocksToMarkForRestructuring > 0)
            {
                SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevNo: " << iDevNo
                                    << " MarkedBlocks: " << NumBlocksToMarkForRestructuring );

                pDevice->MarkFragmentedBlocks (0, NumBlocksToMarkForRestructuring) ;
                bFragmentedAndUnbalancedBlocksFound |= (pDevice->m_NumBlocksMarkedForBackup > 0);
                bChangeDeviceUsage = true; ;
            }
        }
        else
        {
            /* check if the device is too fragmented */
            if (pDevice->FragmentationFactor() > MaxFragmentationFactor )
            {
                /* blocks causing the fragmentation are marked */
                SAPDBTRACE_WRITELN( FBM_Trace, 5, "Compress DevNo: " << iDevNo);

                pDevice->MarkFragmentedBlocks (pDevice->m_NumBlocksUsed, pDevice->m_NumBlocks) ;
                bFragmentedAndUnbalancedBlocksFound |= (pDevice->m_NumBlocksMarkedForBackup > 0);
            }
        }
    }

    /* update counters which control which devices are prefered when free blocks are   */
    /* requested  (getfreeblock) depending on the number of used blocks on each device */
    if (bChangeDeviceUsage) ChangeDeviceUsagePreferences ();
}


/*---------------------------------------------------------------------------*/

void
FBM_Manager::GetLastFragmentedBlock  (tsp00_TaskId      TaskId,
                                      SAPDB_Int2       &DevWithFragBlocks,
                                      SAPDB_Int4       &LastFragBlockNo)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetLastFragmentedBlocks", FBM_Trace, 5);

    FBM_SynchObject SynchronizeFBM (TaskId);

    if (m_NextDevToUseForFragBlock != FBM_UNDEFINED)
    {
        SAPDB_Int2  SearchBegDevNo  = m_NextDevToUseForFragBlock;
        SAPDB_Int2  iDevNo          = SearchBegDevNo;
        SAPDB_Bool  FragBlockFound  = false ;
        do
        {
            SAPDBTRACE_WRITELN( FBM_Trace, 5, "FragSearchDevNo: " << iDevNo);

            m_DataDevice [iDevNo].GetLastFragmentedBlock (FragBlockFound, LastFragBlockNo);

            if (!FragBlockFound)
            {
                /* goto next device if no fragmented block could be found on the current device */
                if ((++iDevNo) > LastDevNo()) iDevNo = FirstDevNo();
            }
            else
            {
                DevWithFragBlocks = iDevNo ;

                /* set device where to start the next search */
                if (DevWithFragBlocks == LastDevNo())
                    m_NextDevToUseForFragBlock = FirstDevNo();
                else
                    m_NextDevToUseForFragBlock = DevWithFragBlocks+1;

                SAPDBTRACE_WRITELN( FBM_Trace, 5, "Frag Dev: " << DevWithFragBlocks
                                    << " Frag Block: " << LastFragBlockNo );
                return;
            }
        }
        while (iDevNo != SearchBegDevNo) ;

        /* no fragmented block found */
        m_NextDevToUseForFragBlock = FBM_UNDEFINED;
    }

    DevWithFragBlocks          = FBM_UNDEFINED;
    LastFragBlockNo            = FBM_UNDEFINED;

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "No Fragented Blocks Found");

}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::GetNextBlocksForBackUp (const tsp00_TaskId taskId,
                                     const SAPDB_Int2   DevNo,
                                     const SAPDB_Int4   MaxNumBlocksWanted,
                                     SAPDB_Int4        &SuppliedNumBlocks,
                                     SAPDB_Int4        &BlockNo,
                                     tgg00_BasisError  &trError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetNextBlocksForBackUp", FBM_Trace, 5);

    if( ! m_Active )
    {
        trError = e_shutdown;
        return;
    }
    /* this routine supplies a set of SuppliedNumBlocks  */
    /* neighbouring blocks  marked for backup            */
    /* see method BeginReadingBlocksMarkedForBackUp      */

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevNo: " << DevNo);


    FBM_SynchObject SynchronizeFBM( taskId );

    /* check device number and get a reference to the current device */
    FBM_DataDevice & Device = m_DataDevice [DevNo];

    /* get blocks marked for backup */
    Device.GetNextBlocksForBackUp (MaxNumBlocksWanted, SuppliedNumBlocks, BlockNo);

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "num_found: " << SuppliedNumBlocks
                        << " BlockNo: " << BlockNo );
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::BeginReadingBlocksMarkedForBackUp (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::BeginReadingBlocksMarkedForBackUp", FBM_Trace, 5);

    /* this routine sets the  position (iterator)                                     */
    /* m_ActBlockNoForBackup from where to start to look for blocks marked for backup */
    /* first position on the device. All blocks marked for backup can than be fetched */
    /* by calling GetNextBlocksForBackUp until this function returns no more blocks   */

    FBM_SynchObject SynchronizeFBM (TaskId);

    for (int iDevNo=FirstDevNo(); iDevNo<=LastDevNo(); ++iDevNo)
    {
        SAPDBERR_ASSERT_STATE (m_DataDevice [iDevNo].m_ActBlockNoForBackup == FBM_UNDEFINED);
        if (m_DataDevice [iDevNo].m_NumBlocksMarkedForBackup > 0)
            m_DataDevice [iDevNo].m_ActBlockNoForBackup = 0;
    }
}
/*---------------------------------------------------------------------------*/

void
FBM_Manager::Dump(
    const tsp00_TaskId    taskId,
    Kernel_Dump           &dump ) const
{
    struct FBMManager   fbmMan;

    FBM_SynchObject SynchronizeFBM( taskId );

    fbmMan.dmpActive                     = m_Active;
    fbmMan.dmpSvpIsTriggered             = m_SavepointIsRequested;
    fbmMan.dmpFiller1                    = 0;
    fbmMan.dmpMaxNumDev                  = m_MaxNumDev;
    fbmMan.dmpNumDev                     = m_NumDev;
    fbmMan.dmpSecurityLimit              = m_SecuritySpaceLimit;
    fbmMan.dmpTotalNumBlocksFreeAfterSVP = m_TotalNumBlocksFreeAfterSVP;
    fbmMan.dmpTotalNumBlocksFree         = m_TotalNumBlocksFree;
    fbmMan.dmpTotalNumBlocks             = m_TotalNumBlocks;
    fbmMan.dmpTotalUsedBlocks            = m_TotalNumBlocks - m_TotalNumBlocksFree
                                           - m_TotalNumBlocksFreeAfterSVP;
    fbmMan.dmpNextDevToUseForNewBlock    = m_NextDevToUseForNewBlock;
    fbmMan.dmpNextDevToUseForFragBlock   = m_NextDevToUseForFragBlock;

    dump.InsertEntry( Kernel_Dump::DmpFBMManager,
                      Kernel_DumpPage::Entry( &fbmMan, sizeof( fbmMan )));

    if( m_Active )
    {
        for( SAPDB_Int4 devNo = FirstDevNo(); devNo <= LastDevNo(); ++devNo )
            m_DataDevice[ devNo ].Dump( dump, devNo );
    }
}

/*---------------------------------------------------------------------------*/
/* private Methodes of FBM_Manager                                           */
/*---------------------------------------------------------------------------*/

SAPDB_Int4
FBM_Manager::OptimumNumOccBlocksPerDevice()
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::OptimumNumOccBlocksPerDevice", FBM_Trace, 5);

    /* This routine tries to calculate the average occupation of all devices with used blocks        */
    /* which is also the optimum occupation with respect to maximum parallel reading of blocks       */
    /* The value OptimimumNumBlocksOcc does not mean that all devices must accomodate this           */
    /* number of occupied blocks since this would not be possible for devices which are smaller than */
    /* AverageNumBlocksForBalance. Instead  OptimimumNumBlocksOcc means that the optimum             */
    /* distribution of blocks is that all devices which are smaller than AverageNumBlocksForBalance  */
    /* are completely filled and all devices larger contain OptimimumNumBlocksOcc blocks             */

    SAPDB_Int4 TotalNumBlocksOcc  = m_TotalNumBlocks - m_TotalNumBlocksFree - m_TotalNumBlocksFreeAfterSVP;

    SAPDB_Int4                      OptimimumNumBlocksOcc;
    SAPDB_Int2                      NumDevNotFullAfterBalancing = m_NumDev;
    SAPDB_Bool                      bRepeatCalculation;
    Container_Vector<SAPDB_Bool>    DevSizeIsLessThanAvarageUsage(m_Allocator);

    /* initialze auxilliary structure */
    DevSizeIsLessThanAvarageUsage.Resize(m_MaxNumDev, false); // TODO retcode

    do
    {
        bRepeatCalculation    = false;
        OptimimumNumBlocksOcc = TotalNumBlocksOcc / NumDevNotFullAfterBalancing ;

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "devs_to_chck: " << NumDevNotFullAfterBalancing
                            << " avg_per_dev: " << OptimimumNumBlocksOcc );

        SAPDB_Int4      iDevNo;
        FBM_DataDevice *pDevice;

        for (iDevNo=FirstDevNo(), pDevice = & m_DataDevice [FirstDevNo()];
                iDevNo<=LastDevNo();
                iDevNo++, pDevice++ )
        {
            if (!DevSizeIsLessThanAvarageUsage [iDevNo] )
            {
                if ( pDevice->m_NumBlocks < OptimimumNumBlocksOcc )
                {
                    DevSizeIsLessThanAvarageUsage [iDevNo] = true;
                    TotalNumBlocksOcc -= pDevice->m_NumBlocks;
                    NumDevNotFullAfterBalancing--;
                    bRepeatCalculation = true;

                    SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevTooSmall: " << iDevNo
                                        << " DevSize: " << pDevice->m_NumBlocks );
                }
            }
        }
    }
    while (bRepeatCalculation);

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "AvgPerDevNo: " << OptimimumNumBlocksOcc);

    return ( OptimimumNumBlocksOcc ) ;
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::ChangeDeviceUsagePreferences ()
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::ChangeDeviceUsagePreferences", FBM_Trace, 5);

    /* update counters which control which devices are prefered when free blocks are   */
    /* requested  (getfreeblock) depending on the number of used blocks on each device */
    /* the prefrences are calculated to get optimum parallel read accesses that means  */
    /* all devices should accomodate the same number of blocks                         */
    SAPDB_Int4 MaxNumBlocksUsed;
    SAPDB_Int4 iDevNo;

    /* search maximum number of blocks on one device */
    MaxNumBlocksUsed = m_DataDevice [FirstDevNo()].m_NumBlocksUsed;
    for (iDevNo=FirstDevNo() + 1; iDevNo<=LastDevNo(); ++iDevNo)
    {
        if (MaxNumBlocksUsed < m_DataDevice [iDevNo].m_NumBlocksUsed)
            MaxNumBlocksUsed = m_DataDevice [iDevNo].m_NumBlocksUsed;
    }

    /* calculate the number of used blocks to add to each */
    /* device that all devices have the same occupation   */
    FBM_DataDevice * pDevice;
    for (iDevNo=FirstDevNo(), pDevice = & m_DataDevice [FirstDevNo()];
            iDevNo<=LastDevNo();
            iDevNo++, pDevice++)
    {
        pDevice->m_NumBlocksToAddUntilOptIsReached = (MaxNumBlocksUsed - pDevice->m_NumBlocksUsed);

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevNo: " << iDevNo
                            << " OptAddNumBlo: " << pDevice->m_NumBlocksToAddUntilOptIsReached );
    }

    m_NextDevToUseForFragBlock = FirstDevNo();
}

/*---------------------------------------------------------------------------*/

FBM_Manager*
FBM_Manager::m_Instance = NULL;

/*---------------------------------------------------------------------------*/

void
FBM_Manager::ResetMembersToNullAndZero()
{
    m_Active                     = false;
    m_MaxNumDev                  = FBM_UNDEFINED;
    m_SecuritySpaceLimit         = 0;
    m_TotalNumBlocksFreeAfterSVP = 0;
    m_TotalNumBlocksFree         = 0;
    m_TotalNumBlocks             = 0;
    m_NumDev                     = 0;
    m_SavepointIsRequested       = false;
    m_NextDevToUseForNewBlock    = FBM_UNDEFINED;
    m_NextDevToUseForFragBlock   = FBM_UNDEFINED;
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::TriggerSavepoint( const tsp00_TaskId taskId )
{
    tgg00_TransContext  trans;

    trans.trTaskId_gg00 = taskId;
    trans.trSessionId_gg00.becomes( cgg_nil_session );
    trans.trTransId_gg00.becomes( cgg_nil_trans );
    trans.trWriteTransId_gg00.becomes( cgg_nil_trans );
    trans.trSubtransId_gg00 = cgg_zero_subtrans;

    Log_SavepointManager.StartSavepoint( trans, Log_SVPReasonFBM );
}
