/*!**************************************************************************

  module      : FBM_Manager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  auhtor      : AlexanderK

  special area: FreeBlockManagement (FBM)
  description : General header file of the FBM


  last changed: 2001-01-19  15:35
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef FBM_MANAGER_HPP
#define FBM_MANAGER_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "ggg00.h"
#include "heo55k.h"                                // RTE   : vbegexcl,venexcl 
#include "hgg08.h"                                 // region identifier
#include "FreeBlockManagement/FBM_DataDevice.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_ReservedBlockAddress.hpp"
#include "IOManager/IOMan_IBlockAddressIterator.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "KernelCommon/Kernel_Dump.hpp"
#include "RunTime/MemoryManagement/RTEMem_AllocatorWrapper.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  CLASS DECLARATIONS                                                       *
 *===========================================================================*/

/*!*****************************************************************************
 
   class: FBM_Manager
   
*******************************************************************************/

class FBM_Manager
{


public:

    /*!-------------------------------------------------------------------------
    function:     CreateInstance() 
    description:  static method to create a singelton of the type FBM_Manager
    arguments:    TrError      [out] returns the error e_sysbuff_overflow in case
                                     that there is not enough memory to create an instance
                                     of FBM_Manager otherwise e_ok is returned
    return value: (FBM_Manager&) reference to the instance of FBM_Manager 
    --------------------------------------------------------------------------*/
    static        FBM_Manager& CreateInstance (tgg00_BasisError TrError);

    /*!-------------------------------------------------------------------------
    function:     Instance() 
    description:  returns the reference to the sigelton instance of FBM_Manager
    arguments:    none
    return value: (FBM_Manager&) reference to the instance of FBM_Manager 
    --------------------------------------------------------------------------*/
    static FBM_Manager& Instance  ()
    {
        return *m_Instance;
    }

    /*!-------------------------------------------------------------------------
    function:     Restart() 
    description:  restarts the FBM
    arguments:    TaskId         [in]  task id
                  MaxNumDev      [in]  maximum number of devices which can be handled 
                                       by the FBM_Manager
                  ReservedBlocks [in]  collection of reserved block address
    return value: true if successful executed else false
    --------------------------------------------------------------------------*/
    bool Restart(
        const tsp00_TaskId                  TaskId,
        const SAPDB_Int4                    MaxNumDev,
        const IOMan_ReservedBlockAddress    &ReservedBlocks );

    /*!-------------------------------------------------------------------------
    function:     AddVolume() 
    description:  add a volume to the FBM

    arguments:    TaskId         [in]  task id
                  DevNo          [in]  device number
                  DevSize        [in]  number of pages the device can accomodate 
    return value: true if successful executed else false
    --------------------------------------------------------------------------*/
    bool AddVolume(
        const tsp00_TaskId                  TaskId,
        const SAPDB_Int2                    DevNo,
        const SAPDB_Int4                    DevSize);

    /*!-------------------------------------------------------------------------
    function:     SetBlockStateToBackup()
    description:  Sets the state of block to 'marked for backup'

    - If the state could not be changed successfully the kernel aborts, if
      the given parameter bAbortIfErrorOccured is set to SAPDB_TRUE.

    arguments:    TaskId               [in] task id
                  BlockAddress         [in] I/O adress of the block to change  
                  bAbortIfErrorOccured [in] terminate kernel if an error occure.
                                            default is set to no. Therefore
                                            the calling method has to make the
                                            error handlig.
    return value: (SABDB_Bool) SAPDB_TRUE if operation was executed successfully 
    --------------------------------------------------------------------------*/
    SAPDB_Bool SetBlockStateToBackup (
        const tsp00_TaskId        TaskId,
        const IOMan_BlockAddress &BlockAddress,
        const SAPDB_Bool          bAbortIfErrorOccured = ! FBM_ABORT_IF_ERROR );

    /*!
       @brief          Sets the state of block to 'occupied'
       @param          TaskId [in] task id
       @param          BlockIterator [in] iterator of I/O adresses with the blocks to change
       @param          bAbortIfErrorOccured [in] terminate kernel if an error occure.
                                        default is set to no. Therefore
                                        the calling method has to make the
                                        error handlig.
       @return         (SAPDB_Bool) SAPDB_TRUE if operation was executed successful

       - If the state could not be changed successfully the kernel aborts if
      the given parameter bAbortIfErrorOccured is set to SAPDB_TRUE.
     */
    SAPDB_Bool SetBlockStateToOccupied(
        const tsp00_TaskId                 TaskId,
              IOMan_IBlockAddressIterator &BlockIterator,
        const SAPDB_Bool                   bAbortIfErrorOccured = ! FBM_ABORT_IF_ERROR);


    /*!
       @brief          Sets the state of block to 'occupied'
       @param          TaskId [in] task id
       @param          BlockAddress [in] I/O adress of the block to change
       @param          bAbortIfErrorOccured [in] terminate kernel if an error occure.
                                        default is set to no. Therefore
                                        the calling method has to make the
                                        error handlig.
       @return         (SAPDB_Bool) SAPDB_TRUE if operation was executed successful

    - If the state could not be changed successfully the kernel aborts if
      the given parameter bAbortIfErrorOccured is set to SAPDB_TRUE.

    arguments:    TaskId               [in] task id
                  BlockAddress         [in] I/O adress of the block to change  
                  bAbortIfErrorOccured [in] terminate kernel if an error occure.
                                            default is set to no. Therefore
                                            the calling method has to make the
                                            error handlig.
    return value: (SABDB_Bool) SAPDB_TRUE if operation was executed successful  
    --------------------------------------------------------------------------*/
    SAPDB_Bool SetBlockStateToOccupied (
        const tsp00_TaskId        TaskId,
        const IOMan_BlockAddress &BlockAddress,
        const SAPDB_Bool          bAbortIfErrorOccured = ! FBM_ABORT_IF_ERROR);


    /*!-------------------------------------------------------------------------
    function:     SetClusterStateToFree()
    description:  Sets the state of cluster to 'free'

      - If the state could not be changed successfully the kernel aborts

    arguments:    TaskId       [in] task id
                  Cluster      [in] I/O adress and length of the cluster to change  
    return value: none 
    --------------------------------------------------------------------------*/
    void SetClusterStateToFree (
        tsp00_TaskId              TaskId,
        const IOMan_ClusterAddress &ClusterAddress);

    /*!-------------------------------------------------------------------------
    function:     SetBlockStateToFree()
    description:  Sets the state of block to 'free'

      - If the state could not be changed successfully the kernel aborts

    arguments:    TaskId       [in] task id
                  BlockAddress [in] I/O adress of the block to change  
    return value: none 
    --------------------------------------------------------------------------*/
    void SetBlockStateToFree (
        tsp00_TaskId              TaskId,
        const IOMan_BlockAddress &BlockAddress);

    /*!-------------------------------------------------------------------------
    function:     SetBlockStateToFreeAfterSVP()
    description:  Sets the state of block to 'free after completion of the next save point'

        - If the state could not be changed successfully the kernel aborts

    arguments:    TaskId       [in] task id
                  BlockAddress [in] I/O adress of the block to change 
    return value: none 
    --------------------------------------------------------------------------*/
    void SetBlockStateToFreeAfterSVP (
        tsp00_TaskId              TaskId,
        const IOMan_BlockAddress &BlockAddress);

    /*!-------------------------------------------------------------------------
    function:     Shutdown()
    description:  all memory resources are released and members set to their initial values
    arguments:    TaskId       [in] task id
    return value:none
    --------------------------------------------------------------------------*/
    void Shutdown (tsp00_TaskId TaskId);

    /*!-------------------------------------------------------------------------
    function:     GetMultFreeBlocks()
    description:  Supplies a set of free and neighbouring blocks 

                - Supplies a set of free and neighbouring blocks and sets these blocks
                  to the state 'occupied'

    arguments:    TaskId                [in]  task id
                  NumFreeBlocksWanted   [in]  wanted number of blocks
                  TrError               [out] errors state of the function. if everything worked fine it is e_ok
    return value: (IOMan_ClusterAddress) address of the supplied cluster
    --------------------------------------------------------------------------*/
    IOMan_ClusterAddress GetMultFreeBlocks(
        const tsp00_TaskId  taskId,
        const SAPDB_Int4    NumFreeBlocksWanted,
        tgg00_BasisError   &trError);

    /*!-------------------------------------------------------------------------
    function:     GetFreeBlock()
    description:  Supplies a free block 

                - Supplies a single free block and marks this block as 'occupied'
                  if no free block could be found than an emergency shutdown will
                  be executed.

    arguments:    TaskId  [in]  task id
                  TrError [out] errors state of the function. if everything worked 
                                fine it is e_ok
    return value: (IOMan_BlockAddress) block address of the supplied block
    --------------------------------------------------------------------------*/
    IOMan_BlockAddress GetFreeBlock( const tsp00_TaskId  taskId );

    /*!-------------------------------------------------------------------------
    function:     GetLastFragmentedBlock()
    description:  Returns the address of a the the last fragmented block 

      - returns the address of the best block to move in order to reduce the
        fragmentation and to obtain the same occupation of all devices

    arguments:    TaskId                 [in]   task id
                  DevWithFragBlocks      [out]  device number of the block to move
                  LastFragBlockNo        [out]  device offset of the block to move

    return value: none
    --------------------------------------------------------------------------*/
    void GetLastFragmentedBlock(
        tsp00_TaskId      TaskId,
        SAPDB_Int2       &DevWithFragBlocks,
        SAPDB_Int4       &LastFragBlockNo);

    /*!-------------------------------------------------------------------------
    function:     SetAllBlocksMarkedAsFreeAfterSVPToFree()
    description:  All Blocks marked as 'free after savepoint' are released
    arguments:    TaskId       [in] task id
    return value: none
    --------------------------------------------------------------------------*/
    void SetAllBlocksMarkedAsFreeAfterSVPToFree (tsp00_TaskId TaskId);

    /*!-------------------------------------------------------------------------
    function:     RestoreAllBlockStatesMarkedForBackup()
    description:  Removes for all blocks the flags indicating that this block must be saved

      - The original state of the blocks is restored.

    arguments:    TaskId       [in] task id
    return value: none
    --------------------------------------------------------------------------*/
    void RestoreAllBlockStatesMarkedForBackup   (tsp00_TaskId       TaskId);

    /*!-------------------------------------------------------------------------
    function:     RestoreBlockStateMarkedForBackup()
    description:  Removes the flags indicating that a block must be saved

      - The original state of the block is restored.
      - If the state could not be restored successfully since the block to be restored
        was not at all marked for backup the kernel aborts

    arguments:    TaskId       [in] task id
                  BlockAddress [in] address of the block to restore 
    return value: true if the block was really marked for back up
    --------------------------------------------------------------------------*/
    void RestoreBlockStateMarkedForBackup (
        tsp00_TaskId             TaskId,
        const IOMan_BlockAddress &BlockAddress);

    /*!-------------------------------------------------------------------------
    function:     RestoreAllBlockStatesMarkedForBackup()
    description:  Removes all flags indicating blocks must be moved to compress the device

      - Removes for all blocks the flags indicating that a block must be moved to compress the device
      - The original state of the blocks is restored.

    arguments:    TaskId       [in] task id
    return value: none
    --------------------------------------------------------------------------*/
    void RestoreAllBlockStatesMarkedForCompression (tsp00_TaskId    TaskId);

    /*!-------------------------------------------------------------------------
    function:     RestoreBlockStateMarkedForCompression()
    description:  Removes the flags indicating that a block must be moved to compress the device

      - The original state of the block is restored.
      - If the state could not be restored successfully since the block to be restored
        was not at all marked for moving the kernel aborts

    arguments:    TaskId       [in] task id
                  BlockAddress [in] address of the block to restore 
    return value: true if the block was really marked for back up
    --------------------------------------------------------------------------*/
    void RestoreBlockStateMarkedForCompression  (
        tsp00_TaskId       TaskId,
        const SAPDB_Int2   DevNo,
        const SAPDB_Int4   BlockNo,
        SAPDB_Bool        &bBlockInUse);

    /*!-------------------------------------------------------------------------
    function:     NumBlocksMarkedForBackup()
    description:  Returns the number of all blocks marked for back up
    arguments:    taskId [in] task id
    return value: number of all blocks marked for back up
    --------------------------------------------------------------------------*/
    SAPDB_Int4 NumBlocksMarkedForBackup( const tsp00_TaskId taskId ) const;

    /*!-------------------------------------------------------------------------
    function:     MarkFragmentedAndUnbalancedBlocks()
    description:  this function marks all blocks which must be moved to get compressed and
                  equally occupied devices 

    arguments:    TaskId                 [in] task id
                  DeviceCompressionLimit [in] if DeviceCompressionLimit < (LastBlockNo(device)/NumAllBlocks (device) - 1) * 100%
                                              then the device has to compressed
                  DeviceBalanceLimit     [in] if  DeviceBalanceLimit < NumAllBlocks(device)/(sum_(device) NumAllBlocks(device))*(sum_(device))-1)*100%  
                                              then blocks of this device have to be moved to another device
                  bFragmentedAndUnbalancedBlocksFound [out] returns true if one of the conditions above is fulfilled for any device
                  TrError                [out] errors state of the function. if everything worked fine it is e_ok
    return value: none
    --------------------------------------------------------------------------*/
    void  MarkFragmentedAndUnbalancedBlocks(
        tsp00_TaskId       TaskId,
        SAPDB_Int4         DeviceCompressionLimit,
        SAPDB_Int4         DeviceBalanceLimit,
        SAPDB_Bool1       &bFragmentedAndUnbalancedBlocksFound,
        tgg00_BasisError  &TrError);

    /*!-------------------------------------------------------------------------
    function:     BeginReadingBlocksMarkedForBackUp()
    description:  The iterator supplying the blocks marked for back up is set to the first block

          - This function has to be called prior to the call of the function 
            GetNextBlocksForBackUp which returns each time it is executed another 
            block marked for backup. The function  BeginReadingBlocksMarkedForBackUp
            initializes this iterator and sets it onto the very first block marked for backup

    arguments   : TaskId             [in]  task id
    return value: none
    --------------------------------------------------------------------------*/
    void BeginReadingBlocksMarkedForBackUp      (tsp00_TaskId TaskId);

    /*!-------------------------------------------------------------------------
    function:     GetNextBlocksForBackUp()
    description:  Returns the address of a set of neighbouring blocks marked for back up

          - Each time this function is called it returns another set of
            blocks marked for backup. The function  BeginReadingBlocksMarkedForBackUp
            initializes this iterator and sets it onto the very first block marked for backup.
            After the initialization this function returns by each call an other block
            untill all marked blocks where returned. That there are no more blocks left is displayed
            by a return value 0 for the parameter SuppliedNumBlocks 


    arguments   : TaskId             [in]  task id
                  DevNo              [in]  number of the device from which the block is wanted
                  MaxNumBlocksWanted [in]  maximum number of neighbouring blocks supplied 
                                           i.e.  SuppliedNumBlocks <= MaxNumBlocksWanted
                  SuppliedNumBlocks  [out] number of adjacent marked blocks found
                  BlockNo            [out] address of the first block of the supplied set
                  TrError            [out] errors state of the function. if everything worked 
                                           fine it is e_ok
    return value: none
    --------------------------------------------------------------------------*/
    void  GetNextBlocksForBackUp  (
        const tsp00_TaskId taskId,
        const SAPDB_Int2   DevNo,
        const SAPDB_Int4   MaxNumBlocksWanted,
        SAPDB_Int4        &SuppliedNumBlocks,
        SAPDB_Int4        &BlockNo,
        tgg00_BasisError  &trError);

    /*!-------------------------------------------------------------------------
    function:     Dump()
    description:  inserts all important memory structures of the FBM into the 
                  kernel dump file
    arguments:    taskId [in]     identification of the calling task
                  dump   [in/out] kernel dump file
    return value: none
    --------------------------------------------------------------------------*/
    void Dump(
        const tsp00_TaskId  taskId,
        Kernel_Dump         &dump ) const;

    /*!-------------------------------------------------------------------------
    function:     NumBlocksUsed()
    description:  Returns the number of all used blocks on a device

       - Returns the number of all used blocks on a device. A block is considered to be used
         if it is not in the state free.

    arguments:    DevNo   [in] number of the device for which the number of used blocks is requested  
    return value: number of all blocks used
    --------------------------------------------------------------------------*/

    SAPDB_Int4 NumBlocksUsed(const SAPDB_Int2 DevNo) const
    {
        SAPDBERR_ASSERT_ARGUMENT ((DevNo >=FirstDevNo()) && (DevNo <= LastDevNo()));

        return (m_DataDevice [DevNo].m_NumBlocksUsed);
    }

    /*!-------------------------------------------------------------------------
    function:     GetNumberOfFreeBlocks()
    description:  Returns the number of blocks in state free. Note that the
                  blocks in state free after savepoint are treated as blocks
                  in state occupied, because the are not available at this 
                  moment.
    arguments:    none
    return value: (SAPDB_Int4) Number of free blocks
    --------------------------------------------------------------------------*/

    SAPDB_Int4 GetNumberOfFreeBlocks() const
    {
        return( m_TotalNumBlocksFree );
    }

    /*!-------------------------------------------------------------------------
    function:     GetNumberOfUsedBlocks()
    description:  Returns the number of all used blocks. Used means all blocks
                  not in state free. Note that blocks in state free after savepoint 
                  are handled as occupied blocks because there are not available
                  at this moment.
    arguments:    none
    return value: (SAPDB_Int4) Number of used blocks
    --------------------------------------------------------------------------*/

    SAPDB_Int4 GetNumberOfUsedBlocks() const
    {
        return( m_TotalNumBlocks - GetNumberOfFreeBlocks() );
    }

    /*!-------------------------------------------------------------------------
    function:     IsSpaceAvailable()
    description:  Checks if the given number of blocks is storable within the
                  data volumes.
    arguments:    numBlocksRequested [in] number of blocks to be written into
                                          the data volumes.
    return value: true is given number of blocks is storable within data
    --------------------------------------------------------------------------*/

    bool IsSpaceAvailable( const SAPDB_Int4 numBlocksRequested ) const
    {
        return(( numBlocksRequested + GetNumberOfUsedBlocks() ) < m_SecuritySpaceLimit );
    }

    /*!-------------------------------------------------------------------------
    function:     GarbageCollectionNeeded()
    description:  If the data volume filling is about 90 percent and the fbm
                  is in online mode, this method returns true to signalize that 
                  garbage collector support is needed.
    arguments:    numChangedPages [in] number of changed pages within data cache
                                       and converter
    return value: true if garbage collector support is needed; else false
    --------------------------------------------------------------------------*/

    bool GarbageCollectionNeeded( const SAPDB_Int4 numChangedPages ) const
    {
        if( ! m_Active )
            return( false );

        if( m_SavepointIsRequested )
            return( false );

        // data base filling is under 90 percent inclusive the changed pages of the data cache
        if(( GetNumberOfUsedBlocks() + numChangedPages ) < ( m_TotalNumBlocks * 0.9 ))
            return( false );

        return( true );
    }

    /*!-------------------------------------------------------------------------
    function:     SavepointNeeded()
    description:  If the number of blocks in state free after savepoint is greater
                  than the number of blocks in state free a savepoint is needed.
    arguments:    none
    return value: true if a savepoint is needed; else false
    --------------------------------------------------------------------------*/

    bool SavepointNeeded()
    {
        if( !m_Active )
            return( false );

        if( m_SavepointIsRequested )
            return( false );

        if( m_TotalNumBlocksFree > m_TotalNumBlocksFreeAfterSVP )
            return( false );

        return( true );
    }

    /*!-------------------------------------------------------------------------
    function:     FreeAfterSVPExist()
    description:  Returns true if the number of blocks in state
                  free after savepoint is greater than zero.
    arguments:    none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool FreeAfterSVPExist() const
    {
        return( m_TotalNumBlocksFreeAfterSVP );
    }

private:

    FBM_Manager(SAPDBMem_IRawAllocator &Allocator);

    SAPDB_Int4 FirstDevNo() const
    {
        return 1;
    }

    SAPDB_Int4 MaxDevNo() const
    {
        return (FirstDevNo() + m_MaxNumDev - 1);
    }

    SAPDB_Int4 LastDevNo() const
    {
        return (FirstDevNo() + m_NumDev - 1);
    }

    SAPDB_Int2 GetDeviceWithFreeBlocks(const SAPDB_Int2 NumRequestedBlocks);

    SAPDB_Int4 OptimumNumUsedBlocksPerDevice  ();

    void ChangeDeviceUsagePreferences();

    SAPDB_Int4 OptimumNumOccBlocksPerDevice();

    void ResetMembersToNullAndZero();

    void TriggerSavepoint( const tsp00_TaskId taskId );

    /*!
       @brief          Sets the state of block to 'occupied' - simple without locking/synchronizing
       @param          BlockAddress [in] I/O adress of the block to change
       @param          bAbortIfErrorOccured [in] terminate kernel if an error occure.
                                        default is set to no. Therefore
                                        the calling method has to make the
                                        error handlig.
       @return         (SAPDB_Bool) SAPDB_TRUE if operation was executed successful

       - If the state could not be changed successfully the kernel aborts if
      the given parameter bAbortIfErrorOccured is set to SAPDB_TRUE.
     */
    SAPDB_Bool SetBlockStateToOccupiedUnsynched(
        const IOMan_BlockAddress &BlockAddress,
        const SAPDB_Bool          bAbortIfErrorOccured = ! FBM_ABORT_IF_ERROR);


private:

    class FBM_SynchObject
    {
    private:
        tsp00_TaskId m_TaskId;
    public:
        FBM_SynchObject(tsp00_TaskId TaskId) :m_TaskId(TaskId)
        {
            vbegexcl (m_TaskId, g08fbm);
        };

        ~FBM_SynchObject()
        {
            vendexcl (m_TaskId, g08fbm);
        };
    };

private:

    RTEMem_AllocatorWrapper     m_Allocator;
    SAPDB_Int4                  m_MaxNumDev;
    SAPDB_Int4                  m_SecuritySpaceLimit;
    SAPDB_Int4                  m_TotalNumBlocksFreeAfterSVP;
    SAPDB_Int4                  m_TotalNumBlocksFree;
    SAPDB_Int4                  m_TotalNumBlocks;
    SAPDB_Int4                  m_NumDev;
    SAPDB_Int4                  m_NextDevToUseForNewBlock;
    SAPDB_Int4                  m_NextDevToUseForFragBlock;
    FBM_DataDeviceArray         m_DataDevice;
    static FBM_Manager          *m_Instance;
    SAPDB_Bool                  m_SavepointIsRequested;
    SAPDB_Bool                  m_Active;
    IOMan_ReservedBlockAddress  m_ReservedBlocks;

};

/*!*****************************************************************************

  endclass: FBM_Manager
  
*******************************************************************************/

#endif //FBM_MANAGER_HPP
