/*!*****************************************************************************

  module:       IOMan_BlockAddress.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author      : TillL

  special area: IOManager

  description:  Class which handles a logical block address for i/o
                
  see also:     

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*******************************************************************************/

#ifndef IOMAN_BLOCKADDRESS_HPP
#define IOMAN_BLOCKADDRESS_HPP


#include "IOManager/IOMan_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"

/*!*****************************************************************************

    class: IOMan_PackedBlockAddress

    description: Logical block address for i/o in packed format

    - Packed format is an unsigned 4-byte integer.

*******************************************************************************/
class IOMan_PackedBlockAddress : 
public SAPDB_Invalidatable<SAPDB_UInt4, SAPDB_MAX_UINT4>
{
    typedef SAPDB_Invalidatable<SAPDB_UInt4, SAPDB_MAX_UINT4>     Super;

public:
    
    /*!-------------------------------------------------------------------------
    declaration:  PackedValue
    description:  address value in packed format
    --------------------------------------------------------------------------*/
    typedef Super::ValueType                                    PackedValue;

    /*!-------------------------------------------------------------------------
    function:     IOMan_PackedBlockAddress()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    IOMan_PackedBlockAddress()
    {}
   
    /*!-------------------------------------------------------------------------
    function:     IOMan_PackedBlockAddress()
    description:  Constructor; sets address from an unsigned 4-byte integer

    arguments:    addr [in]    packed address to be set
    return value: none
    --------------------------------------------------------------------------*/
    IOMan_PackedBlockAddress
        (PackedValue                     addr)
        : 
         Super(addr)
    {}
   
    /*!-------------------------------------------------------------------------
    function:     SetAsReserved()
    description:  Assigns a special value indicating the reserved state
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void SetAsReserved()
    {
        m_Value = ReservedValue();
    }

    /*!-------------------------------------------------------------------------
    function:     IsReserved()
    description:  Checks whether actual value is in reserved state
    arguments:    none
    return value: (SAPDB_Bool) true if reserved
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsReserved() const
    {
        return m_Value == ReservedValue();
    }

private:

    static PackedValue ReservedValue()
    {
        static PackedValue val = SAPDB_MAX_UINT4 - 1;
        return val;
    }

};
/*!*****************************************************************************

   endclass: IOMan_PackedBlockAddress

*******************************************************************************/




/*!*****************************************************************************

   class: IOMan_BlockAddress

   description: Defines a handle to a logical block address for i/o

   - A logical block address is described by a device number and a 
     block position on that device.

*******************************************************************************/
class IOMan_BlockAddress
{
public:

    /*!-------------------------------------------------------------------------
    function:     IOMan_BlockAddress
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    IOMan_BlockAddress()
    {}
   
    /*!-------------------------------------------------------------------------
    function:     IOMan_BlockAddress
    description:  Constructor; sets block address
    arguments:    devno    [in]   device number
                  blockpos [in]   block position on device
    return value: none
    --------------------------------------------------------------------------*/
    IOMan_BlockAddress
        (const IOMan_DeviceNo   &devno, 
         const IOMan_BlockNo    &blockpos)
        : 
         m_DeviceNo(devno), 
         m_BlockPos(blockpos)
    {}
   
    /*!-------------------------------------------------------------------------
    function:     SetAddress()
    description:  Assigns block address
    arguments:    devno    [in]   device number
                  blockpos [in]   block position on device
    return value: none
    --------------------------------------------------------------------------*/
    void SetAddress
        (const IOMan_DeviceNo   &devno, 
         const IOMan_BlockNo    &blockpos)
    {
        m_DeviceNo = devno;
        m_BlockPos = blockpos;
    }

    /*!-------------------------------------------------------------------------
    function:     Pack()
    description:  Converts block io address to packed format
    arguments:    addr [in]          block address in unpacked format
                  devnobitcount [in] number of bits used for persistent device
                                     number representation
    return value: (IOMan_PackedBlockAddress) address in packed format
    --------------------------------------------------------------------------*/
    friend IOMan_PackedBlockAddress Pack
        (const IOMan_BlockAddress&      addr,
         SAPDB_UInt                     devnobitcount);

    /*!-------------------------------------------------------------------------
    function:     Unpack()
    description:  Converts packed format to unpacked format
    arguments:    packedaddr [in]    block address in packed format
                  devnobitcount [in] number of bits used for persistent device
                                     number representation
    return value: (IOMan_BlockAddress) block address in unpacked format
    --------------------------------------------------------------------------*/
    friend IOMan_BlockAddress Unpack
        (const IOMan_PackedBlockAddress&  persaddr,
         SAPDB_UInt                       devnobitcount);

    /*!-------------------------------------------------------------------------
    function:     GetDeviceNo()
    description:  Returns device number of this block address
    arguments:    none
    return value: (IOMan_DeviceNo) device number
    --------------------------------------------------------------------------*/
    const IOMan_DeviceNo& GetDeviceNo() const
    {
        SAPDBERR_ASSERT_STATE(IsValid());

        return m_DeviceNo;
    }

    /*!-------------------------------------------------------------------------
    function:     SetBlockNo(IOMan_BlockNo &BlockPos)
    description:  Sets block pos of this block address
    arguments:    new block pos
    return value: none
    --------------------------------------------------------------------------*/
    void SetBlockNo(const IOMan_BlockNo &BlockPos)
    {
        m_BlockPos = BlockPos;
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockNo()
    description:  Returns block pos of this block address
    arguments:    none
    return value: (IOMan_BlockNo) block position on device
    --------------------------------------------------------------------------*/
    const IOMan_BlockNo& GetBlockNo() const
    {
        SAPDBERR_ASSERT_STATE(IsValid());

        return m_BlockPos;
    }

    /*!-------------------------------------------------------------------------
    function:     Invalidate()
    description:  Assigns an invalid block address
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Invalidate()
    {
        m_DeviceNo.Invalidate();
    }

    /*!-------------------------------------------------------------------------
    function:     IsValid()
    description:  Checks whether actual block address is valid
    arguments:    none
    return value: (bool)
    --------------------------------------------------------------------------*/
    bool IsValid() const
    {
        return m_DeviceNo.IsValid();
    }

private:

    /*--------------------------------------------------------------------------
    declaration: m_DeviceNo
    description: logical device number
    --------------------------------------------------------------------------*/
    IOMan_DeviceNo  m_DeviceNo;

    /*--------------------------------------------------------------------------
    declaration: m_BlockPositionOnDevice
    description: logical block offset on device
    --------------------------------------------------------------------------*/
    IOMan_BlockNo   m_BlockPos;

};
/*!*****************************************************************************

   endclass: IOMan_BlockAddress

*******************************************************************************/


inline IOMan_PackedBlockAddress Pack
    (const IOMan_BlockAddress&      addr,
     SAPDB_UInt                     devnobitcount)
{
    if (!addr.IsValid())
        return IOMan_PackedBlockAddress();  // returns invalid address

    SAPDBERR_ASSERT_ARGUMENT
        (SAPDB_UInt2( addr.GetDeviceNo()) < (1UL << devnobitcount));

    SAPDBERR_ASSERT_ARGUMENT
        (SAPDB_UInt4( addr.GetBlockNo()) < (1UL << (SAPDB_BITS_PER_UINT4 
                                                    - devnobitcount)));

    return ( (addr.GetBlockNo() << devnobitcount) | addr.GetDeviceNo());
}

// --------------------------------------------------------------------------

inline IOMan_BlockAddress Unpack
    (const IOMan_PackedBlockAddress&  persaddr,
     SAPDB_UInt                       devnobitcount)
{
    if (persaddr.IsInvalid() || persaddr.IsReserved())
        return IOMan_BlockAddress();        // returns invalid address

    return IOMan_BlockAddress(
               IOMan_DeviceNo
                  (persaddr & ~(SAPDB_MAX_UINT4 << devnobitcount)),
               IOMan_BlockNo
                  ((persaddr & (SAPDB_MAX_UINT4 << devnobitcount)) 
                                >> devnobitcount));
}

// --------------------------------------------------------------------------


#endif // IOMAN_BLOCKADDRESS_HPP

