/*!*****************************************************************************

  module:       IOMan_ClusterAddress.hpp

  ------------------------------------------------------------------------------

  responsible:  TorstenS

  author      : TillL

  special area: IOManager

  description:  Class handles a cluster of logical block i/o addresses.

  see also:     

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


#ifndef IOMAN_CLUSTERADDRESS_HPP
#define IOMAN_CLUSTERADDRESS_HPP


#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"

/*!*****************************************************************************

   class: IOMan_ClusterAddress

   description: Handles a cluster of logical block i/o addresses.

   - In a cluster, all blocks are contiguous

*******************************************************************************/

class IOMan_ClusterAddress
{
public:

    /*-------------------------------------------------------------------------
    declaration: BlockIndex
    description: Block index number within cluster
    ---------------------------------------------------------------------------*/

    typedef SAPDB_UInt2 BlockIndex;

    /*!-------------------------------------------------------------------------
    function:     IOMan_ClusterAddress()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_ClusterAddress()
        : m_BlockCount(0)
    {}
   
    /*!-------------------------------------------------------------------------
    function:     IOMan_ClusterAddress()
    description:  Constructor
    arguments:    startaddr  [in]   block address of first block
                  blockcount [in]   count of blocks in cluster
    return value: none
    --------------------------------------------------------------------------*/
    
    IOMan_ClusterAddress(
        const IOMan_BlockAddress    &StartAddr, 
        IOMan_BlockCount            BlockCount )
        :m_StartAddr( StartAddr ), 
         m_BlockCount( BlockCount )
    {}

    /*!-------------------------------------------------------------------------
    function:     SetCluster()
    description:  Assigns cluster
    arguments:    startaddr  [in]   block address of first block
                  blockcount [in]   count of blocks in cluster
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetCluster(
        const IOMan_BlockAddress    &StartAddr, 
        IOMan_BlockCount            BlockCount )
    {
        m_StartAddr  = StartAddr;
        m_BlockCount = BlockCount;
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockCount()
    description:  Returns count of blocks in the cluster
    arguments:    none
    return value: (IOMan_BlockCount) block count
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetBlockCount() const
    {
        return m_BlockCount;
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockAddress()
    description:  Returns block address of requested block in cluster
    arguments:    Index [in]   block number within cluster (first block = 0)
    return value: (IOMan_BlockAddress) block address
    --------------------------------------------------------------------------*/

    IOMan_BlockAddress GetBlockAddress( BlockIndex Index ) const
    {
        SAPDBERR_ASSERT_ARGUMENT( Index < m_BlockCount );

        return IOMan_BlockAddress(
                     m_StartAddr.GetDeviceNo(), 
                     m_StartAddr.GetBlockNo() + Index );
    }

    /*!-------------------------------------------------------------------------
    function:     GetDeviceNo()
    description:  Returns device number of actual cluster address
    arguments:    none
    return value: (IOMan_DeviceNo) device number
    --------------------------------------------------------------------------*/

    const IOMan_DeviceNo& GetDeviceNo() const
    {
        return m_StartAddr.GetDeviceNo();
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockNo()
    description:  Returns block pos of requested block in cluster
    arguments:    Index [in]   block number within cluster (first block = 0)
    return value: (IOMan_BlockNo) block position on device
    --------------------------------------------------------------------------*/

    IOMan_BlockNo GetBlockNo( BlockIndex Index ) const
    {
        SAPDBERR_ASSERT_ARGUMENT( Index < m_BlockCount );

        return IOMan_BlockNo
                   ( m_StartAddr.GetBlockNo() + Index );
    }

private:

    /*!-------------------------------------------------------------------------
    declaration: m_StartAddr
    description: block address of first block in cluster
    --------------------------------------------------------------------------*/
    IOMan_BlockAddress  m_StartAddr;

    /*!-------------------------------------------------------------------------
    declaration: m_BlockCount
    description: count of blocks in sequence
    --------------------------------------------------------------------------*/
    IOMan_BlockCount m_BlockCount;

};

/*!*****************************************************************************

   endclass: IOMan_ClusterAddress

*******************************************************************************/





#endif // IOMAN_CLUSTERADDRESS_HPP

