/*!**************************************************************************

  module      : IOMan_DataControl.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : This module is used to handle a multitude of data volumes.


  last changed: 2001-02-01  12:45
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_DATACONTROL_HPP
#define IOMAN_DATACONTROL_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

#include "Container/Container_Vector.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_DataVolume.hpp"
#include "IOManager/IOMan_PageFrameAllocator.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define IOMAN_FIRST_DATA_VOLUME  1

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


/*!*****************************************************************************

   class: IOMan_DataControl

   description: This module is used to handle a multitude of data volumes.
   
              
*******************************************************************************/

class IOMan_DataControl
{
public:

    /*!-------------------------------------------------------------------------
    declarations: Type definitions
    --------------------------------------------------------------------------*/

    typedef Container_Vector<IOMan_DataVolume>  DataVolumes;
    typedef DataVolumes::Iterator               Iterator;
    typedef DataVolumes::ConstIterator          ConstIterator;

    /*!-------------------------------------------------------------------------
    enddeclarations: Type definitions
    --------------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     IOMan_DataControl()
    description:  Constructor
    arguments:    pageAllocator [in] Allocator for page frames
                  allocator     [in] Allocator
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_DataControl(
        IOMan_PageFrameAllocator    &pageAllocator,
        SAPDBMem_IRawAllocator      &allocator )
    :
    m_PageAllocator( pageAllocator ),
    m_UsedVolumes( 0 ),
    m_ConfiguredVolumes( 0 ),
    m_TotalDataPages( 0 ),
    m_Volumes( allocator )
    {}

    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  This method is used to initialize the data control structures,
                  i.e. to get the memory via the allocator.
    arguments:    taskId           [in] Identification of the calling task
                  maxDataVolumes [in] Maximun number of data volumes.
    return value: true if the initialization is succesful; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool Initialize(
        const tsp00_TaskId  taskId,
        const SAPDB_Int     maxDataVolumes );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Volume administration                                        */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     CloseAll()
    description:  This method is used close all data volumes which are still
                  in mode open.
    arguments:    taskId [in] Identification of the calling task
    return value: none
    --------------------------------------------------------------------------*/

    void CloseAll( const tsp00_TaskId taskId );

    /*!-------------------------------------------------------------------------
    function:     CreateAll()
    description:  This method is used to create, format and open all data 
                  volumes being defined within the data base configuration file.
    arguments:    taskId        [in] Identification of the calling task
                  pDBIdentifier [in] Unique data base identifier
    return value: true means successful execution
    --------------------------------------------------------------------------*/

    SAPDB_Bool CreateAll(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier );

    /*!-------------------------------------------------------------------------
    function:     OpenAll()
    description:  This method is used to open all data volumes being configued
                  in the data base configuration file.
    arguments:    taskId [in] Identification of the calling task
    return value: true means successful execution
    --------------------------------------------------------------------------*/

    SAPDB_Bool OpenAll( const tsp00_TaskId taskId );

    /*!-------------------------------------------------------------------------
    function:     AddDevSpace()
    description:  This method is used to add a new data volume to the data
                  base volume set in warm as in cold mode too.
    arguments:    taskId  [in] Identification of the calling task.
                  bIsCold [in] Is data base state db_cold (opposite db_warm)
                  devName [in] Name of the new volume
                  devSize [in] Size of the new devsppace in pages
                  devNo   [in] Logical indentifier of the volume
    return value: true means that add data volume was successful
    --------------------------------------------------------------------------*/

    SAPDB_Bool AddDevSpace(
        const tsp00_TaskId      taskId,
        const SAPDB_Bool        bIsCold,
        tsp00_VFilename         &devName,
        const IOMan_BlockCount  devSize,
        const IOMan_DeviceNo    devNo );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Volume administration                                     */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Volume access                                                */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Get()
    description:  This method returns a volume handler for the given logical
                  identifier. There is not check whether the given logical
                  is valid.
    arguments:    devNo [in] Logical volume identifier
    return value: (IOMan_DataVolume) 
    --------------------------------------------------------------------------*/

    IOMan_DataVolume & Get( const IOMan_DeviceNo devNo )
    {
        return( m_Volumes[ devNo - IOMAN_FIRST_DATA_VOLUME ] );
    }

    const IOMan_DataVolume & Get( const IOMan_DeviceNo devNo ) const
    {
        return( m_Volumes[ devNo - IOMAN_FIRST_DATA_VOLUME ] );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Volume access                                             */
    /* -----------------------------------------------------------------------*/
    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator access                                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Begin()
    description:  This method is used to set an iterator to the first data
                  volume. There is no check whether the first volume is valid.
    arguments:    none
    return value: Iterator/ConstIterator
    --------------------------------------------------------------------------*/

    Iterator Begin()
    {
        return( m_Volumes.Begin() );
    }

    ConstIterator Begin() const
    {
        return( m_Volumes.Begin() );
    }

    /*!-------------------------------------------------------------------------
    function:     End()
    description:  This method is used to get an iterator pointing behind the 
                  last valid data volume. Crazy isnt't it?
    arguments:    none
    return value: Iterator/ConstIterator
    --------------------------------------------------------------------------*/

    Iterator End()
    {
        return( m_Volumes.End() );
    }

    ConstIterator End() const
    {
        return( m_Volumes.End() );
    }

    /*!-------------------------------------------------------------------------
    function:     GetFirstDataDevSpaceNo()
    description:  This method returns the identifier of the first data volume.
                  It's not guaranteed that the returned volume identifier 
                  specifies a volume which is online.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetFirstDataDevSpaceNo() const
    {
        return( IOMAN_FIRST_DATA_VOLUME );
    }

    /*!-------------------------------------------------------------------------
    function:     GetNextDataDevSpaceNo()
    description:  This method returns the next identifier of the data volume
                  following the given data volume. It's not guaranteed that
                  the returned volume identifier specifies a volume which
                  is online.
    arguments:    devNo [in] Logical volume identifier
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetNextDataDevSpaceNo( const IOMan_DeviceNo devNo ) const
    {
        return( devNo + 1 );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator access                                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Configuration and statistic information                      */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     ConfiguredDevSpaces()
    description:  This method returns the number of data volumes configured 
                  within the data base configuration file, i.e the number of 
                  data volumes corresponding to a entry within the configuration 
                  file. This number is also available if not all data volumes 
                  are in mode open.
    arguments:    none
    return value: (SAPDB_Int) 
    --------------------------------------------------------------------------*/

    SAPDB_Int ConfiguredDevSpaces() const
    {
        return( m_ConfiguredVolumes );
    }

    /*!-------------------------------------------------------------------------
    function:     GetDevSpaceNo()
    description:  This method is used to determine the logical identifier for
                  a given data volume name. It is not needed that the data 
                  volumes are in mode open.
    arguments:    devName [in]  Name of the searched volume
                  devNo   [out] Logical identifier of the volume
    return value: true means volume was found; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetDevSpaceNo(
        tsp00_VFilename   &devName,
        IOMan_DeviceNo    &devNo ) const;

    /*!-------------------------------------------------------------------------
    function:     MaxDevSpaces()
    description:  This method returns the maximun number of configurable data 
                  volumes. This value is stored within the data base configuration
                  file.
    arguments:    none
    return value: (SAPDB_Int) Maximun number of configurable data volumes.
    --------------------------------------------------------------------------*/

    SAPDB_Int MaxDevSpaces() const
    {
        return( m_Volumes.GetCapacity() );
    }

    /*!-------------------------------------------------------------------------
    function:     TotalUsableDataPages()
    description:  This method returns the sum of all pages storable within all 
                  configured data volumes. This value doesn't include the pages
                  needed for administrativ purpose, i.e IOMan_InfoPage.
                  This value is only available if all
                  data volumes are in mode open.
    arguments:    none
    return value: SAPDB_Int4 
    --------------------------------------------------------------------------*/

    SAPDB_Int4  TotalUsableDataPages() const
    {
        return( m_TotalDataPages );
    }

    /*!-------------------------------------------------------------------------
    function:     TotalDataPages()
    description:  This method returns the sum of all pages storable within all 
                  configured data volumes. This value includes all pages of
                  all data volumes without taking amount of pages reserved 
                  for administrative purposes. This value is only available if 
                  all data volumes are in mode open.
    arguments:    none
    return value: SAPDB_Int4 
    --------------------------------------------------------------------------*/

    SAPDB_Int4 TotalDataPages() const
    {
        return( TotalUsableDataPages() + UsedDevSpaces() );
    }

    /*!-------------------------------------------------------------------------
    function:     UsedDevSpaces()
    description:  This method returns the number of data volumes which are 
                  currently in mode open. Usualy all configured data volumes 
                  are still in mode open.
    arguments:    none
    return value: (SAPDB_Int) 
    --------------------------------------------------------------------------*/

    SAPDB_Int UsedDevSpaces() const
    {
        return( m_UsedVolumes );
    }

    /*!-------------------------------------------------------------------------
    function:     IsDataDevSpaceNoValid()
    description:  This method checks the validity of the given data volume 
                  identifier.
    arguments:    devNo [in] Logical volume identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDataDevSpaceNoValid( const IOMan_DeviceNo devNo ) const
    {
        if(( IOMAN_FIRST_DATA_VOLUME > devNo ) || ( devNo > MaxDevSpaces() ))
            return( false );
        else
            return( true );
    }

    /*!-------------------------------------------------------------------------
    function:     IsDataDevSpaceOnline()
    description:  This method returns whether the given data volume is in
                  state online.
    arguments:    devNo [in] Logical volume identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDataDevSpaceOnline( const IOMan_DeviceNo devNo ) const
    {
        if( ! IsDataDevSpaceNoValid( devNo ) )
            return( false );

        return( Get( devNo ).IsOnline() );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Configuration and statistic information                   */
    /* -----------------------------------------------------------------------*/

private:

    void CreateLinkage(
        const tsp00_TaskId  taskId,
        IOMan_DataVolume  &prevVolume,
        IOMan_DataVolume  &volume );

    const SAPDB_Byte *DetermineLinkage(
        const IOMan_DataVolume    &volume,
        IOMan_DataVolume          **pPrevVolume,
        IOMan_DataVolume          **pNextVolume );

    // Copy and assignment operator are not supported

    IOMan_DataControl& operator=( const IOMan_DataControl &DataControl );

    IOMan_DataControl( const IOMan_DataControl &DataControl );

private:

    /*-------------------------------------------------------------------------
    declaration: m_PageAllocator
    description: Allocator for page frames. The page frame size is definded by
                 the data base configuration file.
    --------------------------------------------------------------------------*/

    IOMan_PageFrameAllocator    &m_PageAllocator;

    /*-------------------------------------------------------------------------
    declaration: m_UsedVolumes
    description: The number of data volumes which are currently in mode
                 open. Usualy all configured data volumes are still in
                 mode open.
    --------------------------------------------------------------------------*/

    SAPDB_Int   m_UsedVolumes;

    /*-------------------------------------------------------------------------
    declaration: m_ConfiguredVolumes
    description: The number of data volumes configured within the data
                 base configuration file, i.e the number of data volumes
                 corresponding to a entry within the configuration file.
                 This number is also available if not all data volumes are
                 in mode open.
    --------------------------------------------------------------------------*/

    SAPDB_Int4  m_ConfiguredVolumes;

    /*-------------------------------------------------------------------------
    declaration: m_TotalDataPages
    description: The sum of all pages storable within all configured
                 data volumes. This value is only available if all
                 data volumes are in mode open.
    --------------------------------------------------------------------------*/

    SAPDB_Int4  m_TotalDataPages;

    /*-------------------------------------------------------------------------
    declaration: m_Volumes
    description: Collection of data volumes. 
    --------------------------------------------------------------------------*/

    DataVolumes   m_Volumes;

};

/*!*****************************************************************************

   endclass: IOMan_DataControl

*******************************************************************************/


#endif  /* IOMAN_DATACONTROL_HPP */
