/****************************************************************************

  module      : IOMan_IConverterIO.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for read and write access of converter pages
                on the data devspaces

  last changed: 2001-03-15  19:43
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_ICONVERTERIO_HPP
#define IOMAN_ICONVERTERIO_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "ggg00.h"

#include "Converter/Converter_Page.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_IConverterIO

   description: This interface is used for read and write access of converter
                pages on the data devspaces.
              
*******************************************************************************/

class IOMan_IConverterIO
{
public:

    /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_IConverterIO interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_ILogIO interface
    --------------------------------------------------------------------------*/

    static IOMan_IConverterIO& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     ReadConverterPage()
    description:  This method is used to read one converter page synchronous from 
                  a data devspace into the given memory. The caller is blocked as 
                  long as the read io is not completed. The position of the page 
                  is given. The correctness of the read data page will be checked. 
                  In case of errors detected by the check method of the 
                  page the read io will be repeated. If the error is still present 
                  after three read operations, the provided buffer is dumped into 
                  the rundirectory of the kernel. Each failed read operation is 
                  logged within the knldiag and knldiag error file. In any case of 
                  an error an emergency shutdown is executed.
    arguments:    taskId  [in] Identification of the calling task.
                  block   [in] Position of the wanted page at the data devspace
                  page    [in] Handler for the wanted page.
    return value: none. 
    --------------------------------------------------------------------------*/

    virtual void ReadConverterPage(
        tsp00_TaskId                taskId,
        const IOMan_BlockAddress    &block,
        Converter_Page              &page ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WriteConverterPage()
    description:  This method is used to write one converter page on one data 
                  devspace. The caller is blocked as long as the write io is not 
                  completed. The destination of the converter page will be requested
                  from the componenet FBM and the old block address is released 
                  similar to the block addresses of data pages. Before the proper 
                  write is executed the correctness of the given pages will be 
                  checked. In case of an error the corrupted page is dumped into 
                  the kernel rundirectory and an emergency shutdown is executed.
                  In the other case the converter page is prepared for the io, i.e. 
                  the the checksum is calculated. 
    arguments:    taskId  [in]  Identification of the calling task.
                  page    [in]  Handler of the given page
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteConverterPage(
        const tsp00_TaskId  taskId,
        Converter_Page      &page ) = 0;
};

/*!*****************************************************************************

   endclass: IOMan_IConverterIO

*******************************************************************************/



#endif  /* IOMAN_ICONVERTERIO_HPP */
