/*!**************************************************************************

  module      : IOMan_IDataIO.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for read and write access on the data devspaces


  last changed: 2001-03-15  17:59
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_IDATAIO_HPP
#define IOMAN_IDATAIO_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "ggg00.h"   // PASCAL: GG_basic_constants_and_tyype

#include "DataAccess/Data_BasePage.hpp"
#include "DataAccess/Data_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_DataPages.hpp"
#include "KernelCommon/Kernel_RestartPage.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

//#define IOManIDataIO    IOMan_IDataIO.GetInstance()


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_IDataIO

   description: This interface is used for read and write access on the data
                devspaces.
              
*******************************************************************************/

class IOMan_IDataIO
{
public:

    /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_IDataIO interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_ILogIO interface
    --------------------------------------------------------------------------*/

    static IOMan_IDataIO& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     ReadDataPage()
    description:  This method is used to read one data page synchronous from a data 
                  devspace into the given memory. The caller is blocked as long as
                  the read io is not completed The position of the page will be determined
                  by the converter. The correctness of the read data page will be
                  checked. In case of errors detected by the check method of the 
                  page the read io will be repeated. If the error is still present 
                  after three read operations, the provided buffer is dumped into 
                  the rundirectory of the kernel. Each failed read operation is 
                  logged within the knldiag and knldiag error file. In case of 
                  fatal errors of the IO subsystem an emergency shutdown is executed.
    arguments:    taskId   [in]  Identification of the calling task.
                  trError  [out] Current error state
                  page     [in]  Handler for the wanted page.
                  pageNo   [in]  Id of the wanted page
    return value: true means page succesful read; else false. 
                  In case of a bad page the error is set to e_bad_datapage.
    --------------------------------------------------------------------------*/

    virtual bool ReadDataPage(
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        Data_BasePage               &Page,
        const Data_PageNo           &PageNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReadDataPagesForSave()
    description:  This method is used to read data pages for a save data via 
                  a separate io thread insulated from the common kernel io threads.
                  The correctness  of the page won't be  checked. The caller is 
                  blocked as long as the write io is not completed.
    arguments:    taskId       [in] Identification of the calling task.
                  asynIOHandle [in] Handle for the asynchronous io
                  pages        [in] Handler of the pages to be written
                  cluster      [in] Destination address of the pages to be written
    return value: true means write was successfull
    --------------------------------------------------------------------------*/

    virtual bool ReadDataPagesForSave(
        const tsp00_TaskId          taskId,
        const SAPDB_Int4            asynIOHandle,
        IOMan_DataPages             &pages,
        const IOMan_ClusterAddress  &cluster ) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReadDataPageDirect()
    description:  This method is used to read one data page synchronous from a data 
                  devspace into the given memory. The caller is blocked as long as
                  the read io is not completed. The position of the wanted page is
                  given as a block address and therefore not searched in the component
                  converter is neccessary. The validity of the given block address 
                  is check on an abstract level, i.e. logical device number and block
                  number are in a valid range. The correctness of the read data page 
                  will be checked. In case of errors detected by the check method of 
                  the page the read io will be repeated. If the error is still present 
                  after three read operations, the provided buffer is dumped into 
                  the rundirectory of the kernel. Each failed read operation is 
                  logged within the knldiag and knldiag error file. In case of 
                  fatal errors of the IO subsystem an emergency shutdown is executed.
    arguments:    TaskId  [in] Identification of the calling task.
                  Page  [in] Handler for the wanted page.
                  Block [in] Position of the wanted page at the data devspace
    return value: true means page succesful read; else false. 
                  In case of a bad page the error is set to e_bad_datapage.
    --------------------------------------------------------------------------*/

    virtual bool ReadDataPageDirect(
        tsp00_TaskId                TaskId,
        Data_BasePage               &Page,
        const IOMan_BlockAddress    &Block ) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReadRestartPage()
    description:  This method is used to read the restart page synchronous from 
                  a defined position of an data devspace. The caller is blocked 
                  as long as the read io  is not completed. The correctness of 
                  the read restart page will be checked. In case of errors detected 
                  by the check method of the page the read io will be repeated. 
                  If the error is still present  after three read operations, 
                  the provided buffer is dumped into the rundirectory of the kernel. 
                  Each failed read operation is logged within the knldiag and 
                  knldiag error file. In any case of errors an emergency shutdown 
                  is executed.
    arguments:    TaskId [in] Identification of the calling task.
                  Page   [in] Handler for the wanted page.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadRestartPage(
        tsp00_TaskId        TaskId,
        Kernel_RestartPage  &Page ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WriteDataPage()
    description:  This method is used to write one data page on one data devspace.
                  For more details see WriteDataPages! The difference to the
                  WriteDataPages call is, that the caller hasn't reserved a block 
                  address a prior. Therefore in the following first the componente 
                  converter specialy free block management has to contacted to get 
                  a block address. Note that this will increase the time the data 
                  page is locked in the data cache!
    arguments:    taskId  [in]  Identification of the calling task.
                  trError [out] Current error state
                  page    [in]  Handler of the given pages
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteDataPage(
        const tsp00_TaskId   taskId,
        tgg00_BasisError    &trError,        
        Data_BasePage       &Page ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WriteDataPages()
    description:  This method is used to write a number of data pages synchronous
                  into one data devspace. The caller is blocked as long as the write 
                  io is not completed. The maximum number of writable data pages is
                  limited by a database configuration parameter. Additional to the
                  writeable data pages the caller has to provide a cluster address,
                  which specifies the write destination on a data devspace. Before 
                  the proper write is executed the correctness of the given pages 
                  will be checked. In case of an error the corrupted page is dumped 
                  into the kernel rundirectory and an emergency shutdown is executed.
                  In the other case the data page is prepared for the io, i.e. the 
                  converter version is updated and the checksum is calculated. 
                  Additional the corresponding converter entry is updated and the proper
                  write operation is executed. Note that the writeable data pages not 
                  necessarily stored in a contiguous memory block, because this method
                  uses scatter/gather io. The caller has to insure that the given 
                  cluster has sufficient capacity to store all given data pages within
                  the data devspace. Pages without an corresponding block address won't
                  be written.
    arguments:    taskId  [in]  Identification of the calling task.
                  trError [out] Current error state
                  pages   [in] Handler of the given pages
                  cluster [in] Write destination of the data pages in the data devspace.
                               The caller has to insure that the given block addresses 
                               are available.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteDataPages(
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        IOMan_DataPages             &pages,
        const IOMan_ClusterAddress  &cluster ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WriteDataPagesForRestore()
    description:  This method is used to write data pages during a restore via 
                  a separate io thread insulated from the common kernel io threads.
                  The correctness  of the page won't be  checked. The caller is 
                  blocked as long as the write io is not completed.
    arguments:    taskId       [in] Identification of the calling task.
                  asynIOHandle [in] Handle for the asynchronous io
                  pages        [in] Handler of the pages to be written
                  cluster      [in] Destination address of the pages to be written
    return value: true means write was successfull
    --------------------------------------------------------------------------*/

    virtual bool WriteDataPagesForRestore(
        const tsp00_TaskId          taskId,
        const SAPDB_Int4            asynIOHandle,
        IOMan_DataPages             &pages,
        const IOMan_ClusterAddress  &cluster ) = 0;



    /*!-------------------------------------------------------------------------
    function:     WriteRestartPage()
    description:  This method is used to write the restart page synchronous onto 
                  a fixed position of a data devspace Before the proper write is 
                  executed the correctness of the  given page will be checked. In 
                  case of an error the corrupted page is dumped into the kernel 
                  rundirectory and an emergency shutdown is executed. In the other 
                  case the log info page is prepared for the io, for example the 
                  checksum is calculated. In any case of errors an emergency 
                  shutdown is executed. 
    arguments:    TaskId [in] Identification of the calling task.
                  Page   [in] Handler of the given pages
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteRestartPage(
        tsp00_TaskId          TaskId,
        Kernel_RestartPage    &Page ) = 0;


};

/*!*****************************************************************************

   endclass: IOMan_IDataIO

*******************************************************************************/


#endif  /* IOMAN_IDATAIO_HPP */
