/*!**************************************************************************

  module      : IOMan_IDataInfo.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for data devspace statistics.


  last changed: 2001-03-16  9:21
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_IDATAINFO_HPP
#define IOMAN_IDATAINFO_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_IDataInfo

   description: This interface is used for data devspace statistics.
              
*******************************************************************************/

class IOMan_IDataInfo
{
public:

    /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_IDataInfo interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_IDataInfo interface
    --------------------------------------------------------------------------*/

    static IOMan_IDataInfo& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     AddDataDevSpacePossible()
    description:  This method indicates whether a add data devspace is possible.
    arguments:    none
    return value: true means that add data devspace is possible
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddDataDevSpacePossible() const = 0;

    /*!-------------------------------------------------------------------------
    function:     ConfiguredDataDevSpaces()
    description:  This method provides the number of configured data devspaces.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int ConfiguredDataDevSpaces() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetDataDevSpaceNo()
    description:  This method returns the logical identifier of for the given
                  data devspace name. If the devspace name doesn't correspond to
                  a valid devspace the return value is false.
    arguments:    DevName [in]  Name of a data devspace
                  DevNo   [out] Logical devspace identifier
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetDataDevSpaceNo(
        tsp00_VFilename &devName,
        IOMan_DeviceNo  &devNo ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetDataDevName()
    description:  This method returns the name of the data devspace identified
                  by the given logical identifier.
    arguments:    devNo   [in ] Logical devspace identifier
                  devName [out] Name of a data devspace
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetDataDevName(
        IOMan_DeviceNo    devNo,
        tsp00_VFilename   &devName ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetDataDevSpaceSize()
    description:  This method provides the configured size in pages of the 
                  specified data devspace. It is assumed that the devspace 
                  identifier is a valid.
    arguments:    DevNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetDataDevSpaceSize( const IOMan_DeviceNo devNo ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetDataDevSpaceUsableSize()
    description:  This method provides the usable size in pages of the 
                  specified data devspace. It is assumed that the devspace 
                  identifier is valid.
    arguments:    DevNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetDataDevSpaceUsableSize( const IOMan_DeviceNo devNo ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     TotalUsableDataPages()
    description:  This method returns the sum of all pages storable within all 
                  configured data devspaces. This value doesn't include the pages
                  needed for administrativ purpose, i.e IOMan_InfoPage.
                  This value is only available if all
                  data devspaces are in mode open.
    arguments:    none
    return value: SAPDB_Int4 
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4 TotalUsableDataPages() const = 0;

    /*!-------------------------------------------------------------------------
    function:     TotalDataPages()
    description:  This method returns the sum of all pages storable within all 
                  configured data devspaces. This value includes all pages of
                  all data devspaces without taking amount of pages reserved 
                  for administrative purposes. This value is only available if 
                  all data devspaces are in mode open.
    arguments:    none
    return value: SAPDB_Int4 
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4 TotalDataPages() const = 0;

    /*!-------------------------------------------------------------------------
    function:     UsedDataDevSpaces()
    description:  This method provides the number of active data devspaces. 
                  If the I/O management isn't restarted the initial value of 
                  one is returned.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int UsedDataDevSpaces() const = 0;

};

/*!*****************************************************************************

   endclass: IOMan_IDataInfo

*******************************************************************************/


#endif  /* IOMAN_IDATAINFO_HPP */
