/*!**************************************************************************

  module      : IOMan_IDataManager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for data devspace administration.


  last changed: 2001-03-15  18:53
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_IDATAMANAGER_HPP
#define IOMAN_IDATAMANAGER_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "ggg00.h"   // PASCAL: GG_basic_constants_and_tyype


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_IDataManager

   description: This interface is used for data devspace administration purpose.
              
*******************************************************************************/

class IOMan_IDataManager
{
public:

        /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_IDataManager interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_IDataManager interface
    --------------------------------------------------------------------------*/

    static IOMan_IDataManager& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     CloseAllDataDevSpaces
    description:  This method is used to bring all data devspaces being in mode
                  open in mode close.
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void CloseAllDataDevSpaces( const tsp00_TaskId  taskId ) = 0;

    /*!-------------------------------------------------------------------------
    function:     CreateAllDataDevSpaces
    description:  This method is used to create, format and open all data 
                  devspaces given within the data base configuration file.
                  Additional it is checked whether the sum of all data pages
                  within all data devspaces fulfils the minimal space requirements 
                  of  a sapdb data base. After checking the data base size
                  the component FBM is generated and initialized. Any kind
                  of an error will be returned as an basis error within the
                  given trans context.
    arguments:    taskId        [in] Identification of the calling task.
                  pDBIdentifier [in] Unique data base identifier
    return value: true means successful creation; else false
    --------------------------------------------------------------------------*/

    virtual bool CreateAllDataDevSpaces( 
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier ) = 0;

    /*!-------------------------------------------------------------------------
    function:     OpenAllDataDevSpaces()
    description:  This method is used to bring all data devspaces specified
                  within the data base configuration file from mode close into
                  mode open. It is assumed that all data devspaces are generated
                  with the method CreateAllDataDevSpaces. Additional it is checked
                  whether the sum of the pages within all attached data devspaces
                  is equal to the sum stored within the restart record. This is 
                  done to validate if the data base configuration was changed in a
                  not offical way. If no errors occur the component FBM is 
                  generated and restarted.
    arguments:    taskId [in] Identification of the calling task.
    return value: true means succeful execution; else false
    --------------------------------------------------------------------------*/

    virtual bool OpenAllDataDevSpaces( const tsp00_TaskId   taskId ) = 0;

    /*!-------------------------------------------------------------------------
    function:     OpenOneDataDevSpace
    description:  This method is used to bring one data devspave identified by
                  the given DevNo from mode closed into mode open. If the data
                  devspace is not stored in the data base configuration file or 
                  the open method fails an error is returned.
    arguments:    taskId [in] Identification of the calling task.
                  devNo  [in] Logical devspace identifier 
    return value: true means succeful execution; else false
    --------------------------------------------------------------------------*/
    
    virtual bool OpenOneDataDevSpace( 
        const tsp00_TaskId    taskId, 
        const IOMan_DeviceNo  devNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     AddDataDevSpace()
    description:  This method is used to add a new data devspace to the data
                  base devspace set in warm as in cold mode too.
    arguments:    taskId  [in] Identification of the calling task.
                  bIsCold [in] Is data base state db_cold (opposite db_warm)
                  devName [in] Name of the new devspace
                  devSize [in] Size of the new devsppace in pages
                  devNo   [in] Logical indentifier of the devspace
    return value: true means that add data devspace was successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddDataDevSpace(
        const tsp00_TaskId      taskId,
        const SAPDB_Bool        bIsCold,
        tsp00_VFilename         &devName,
        const IOMan_BlockCount  devSize,
        const IOMan_DeviceNo    devNo ) = 0;
};

/*!*****************************************************************************

   endclass: IOMan_ILogManager

*******************************************************************************/


#endif  /* IOMAN_IDATAMANAGER_HPP */
