/*!**************************************************************************

  module      : IOMan_IDataVolumeInfo.hpp.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management

  description : Interface for data volume informations 

    
  last changed: 2002-08-16
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_IDATA_VOLUME_INFO_HPP
#define IOMAN_IDATA_VOLUME_INFO_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!***************************************************************************

   class: IOMan_DataVolumeInfo

   description: This class is used to get informations about the current
                data volume. Note that a data volume could be in state
                offline and therefore same informations could be not available!

******************************************************************************/


class IOMan_DataVolumeInfo
{
    friend class IOMan_DataVolumeInfoIterator;

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : IOMan_DataVolumeInfo()
    description : construct a handle for the given data volume. 
    arguments   : volumeNo [in] logical identifier of the data volume
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_DataVolumeInfo( const IOMan_DeviceNo volumeNo ):m_VolumeNo( volumeNo )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Information access                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetVolumeNo()
    description:  This method returns the logical data volume identifier
                  of the volume.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetVolumeNo() const
    {
        return( m_VolumeNo );
    }

    /*!-------------------------------------------------------------------------
    function:     GetVolumeNo()
    description:  This method returns whether the data volume is in state open.
    arguments:    none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsVolumeOnline() const;

    /*!-------------------------------------------------------------------------
    function:     GetDataDevName()
    description:  This method returns the name of the data volume.
    arguments:    devName [out] Name of a data volume
    return value: true if volume name was found
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetVolumeName( tsp00_VFilename   &volumeName ) const;

    /*!-------------------------------------------------------------------------
    function:     GetVolumeSize()
    description:  This method provides the configured size in pages of the data volume. 
    arguments:    none
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetVolumeSize() const;

    /*!-------------------------------------------------------------------------
    function:     GetUsableVolumeSize()
    description:  This method provides the usable size in pages of the data volume.
    arguments:    none
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetUsableVolumeSize() const;

    /*!-------------------------------------------------------------------------
    function:     GetUsedVolumeSize()
    description:  This method provides the used size in pages of the data volume.
    arguments:    none
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetUsedVolumeSize() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Information access                                        */
    /* -----------------------------------------------------------------------*/

private:

    void SetVolumeNo( const IOMan_DeviceNo volumeNo )
    {
        m_VolumeNo = volumeNo;
    }

    // no default constructor available;
    IOMan_DataVolumeInfo();

private:

    /*-------------------------------------------------------------------------
    declaration: m_VolumneNo
    description: logical identifier of the data volume
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo  m_VolumeNo;
};

/*!***************************************************************************

   endclass: IOMan_DataVolumeInfo

******************************************************************************/

/*!***************************************************************************

   class: IOMan_DataVolumeInfoIterator

   description: This class is used to iterator over all data volumes.
                Note that this iterator wil return volume which are not 
                in state online. Therefore you have to call the method
                IsVolumeOnline first.
            
******************************************************************************/

class IOMan_DataVolumeInfoIterator
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : IOMan_DataVolumeInfoIterator()
    description : constructor; Sets the iterator to the first data volume.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_DataVolumeInfoIterator();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator access                                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : Begin()
    description : Sets the iterator to the first logical data volume.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    void Begin();

    /*!-------------------------------------------------------------------------
    function    : End()
    description : Checks wheather the iterator is still valid.
    arguments   : none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool End() const;

    /*!-------------------------------------------------------------------------
    function    : operator++()
    description : Sets the iterator the the logical next data volume. The is
                  no check whether the volume is available.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    void operator++();

    /*!-------------------------------------------------------------------------
    function    : operator*()
    description : myDescription
    arguments   : Returns a reference to the data volume info. First check
                  whether the data volume is online.
    return value: IOMan_DataVolumeInfo&
    --------------------------------------------------------------------------*/

    const IOMan_DataVolumeInfo& operator*() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator access                                           */
    /* -----------------------------------------------------------------------*/

private:

    /*-------------------------------------------------------------------------
    declaration: m_VolumeInfo
    description: myText
    --------------------------------------------------------------------------*/

    IOMan_DataVolumeInfo  m_VolumeInfo;
};

/*!***************************************************************************

   endclass: IOMan_DataVolumeInfoIterator

******************************************************************************/

#endif  /* IOMAN_IDATA_VOLUME_INFO_HPP */
