/*!**************************************************************************

  module      : IOMan_ILogIO.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for read and write access on the archive log.


  last changed: 2001-03-14  16:14
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_ILOGIO_HPP
#define IOMAN_ILOGIO_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"   // PASCAL: GG_basic_constants_and_tyype

#include "IOManager/IOMan_LogPages.hpp"
#include "Logging/Log_ClusterAddress.hpp"
#include "Logging/Log_InfoPage.hpp"
#include "Logging/Log_Page.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_ILogIO

   description: This interface is used for read and write access on the archive log.
              
*******************************************************************************/

class IOMan_ILogIO
{
public:

    /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_ILogIO interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_ILogIO interface
    --------------------------------------------------------------------------*/

    static IOMan_ILogIO& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     ReadLogInfoPage()
    description:  This method is used to read the log info page synchronous from 
                  a given position of the archive log. The caller is blocked as 
                  long as the read io is not completed. The correctness of the read 
                  log info page will be checked. In case of errors detected by the 
                  check method of the page the read io will be repeated. If the 
                  error is still present after three read operations, the provided 
                  buffer is dumped into the rundirectory of the kernel. Each failed 
                  read operation is logged within the knldiag and knldiag error file. 
                  In any case of errors an emergency shutdown is executed and the
                  volume is set to bad.
    arguments:    TaskId  [in] Identification of the calling task.
                  Page    [in] Handler for the wanted page.
                  Address [in] Position of the wanted page within an given archive log
    return value: none
    --------------------------------------------------------------------------*/
    
    virtual void ReadLogInfoPage(
        tsp00_TaskId              TaskId,
        Log_InfoPage              &Page,
        const Log_ClusterAddress  &Address ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WriteLogInfoPage()
    description:  This method is used to write the log info page synchronous onto 
                  a given position of the archive log. Before the proper write is 
                  executed the correctness of the given page will be checked. In 
                  case of an error the corrupted page is dumped into the kernel 
                  rundirectory and an emergency shutdown is executed. In the other 
                  case the log info page is prepared for the io, for example the 
                  checksum is calculated. In any case of errors an emergency 
                  shutdown is executed. 
    arguments:    TaskId  [in] Identification of the calling task.
                  Page    [in] Handler of the given pages
                  Address [in] Position of the given page within an given archive log
    return value: none
    --------------------------------------------------------------------------*/
    
    virtual void WriteLogInfoPage(
        tsp00_TaskId              TaskId,
        Log_InfoPage              &Page,
        const Log_ClusterAddress  &Address) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReadLogPage()
    description:  This method is used to read a log page synchronous from the
                  archive log The caller is blocked as long as the read io is 
                  not completed. The position of the wanted log page is given 
                  as a log address  which will be transformed into a devspace 
                  specific block address. The correctness of the read log info 
                  page will be checked. In case of errors detected by the check 
                  method of the page the read io will be repeated. If the error 
                  is still present  after three read operations, the provided 
                  buffer is dumped into the rundirectory of the kernel. Each 
                  failed read operation is logged within the knldiag and knldiag 
                  error file. In any case of errors an emergency shutdown is 
                  executed and the volume is set to bad.
    arguments:    TaskId  [in] Identification of the calling task.
                  Page    [in] Handler for the wanted page.
                  Address [in] Position of the wanted page within an given archive log
    return value: none
    --------------------------------------------------------------------------*/
    
    virtual void ReadLogPage(
        tsp00_TaskId              TaskId,
        Log_Page                  &Page,
        const Log_ClusterAddress  &Address ) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReadLogPages()
    description:  This method is used to read a couple of log pages synchronous 
                  from the archive log.The caller is blocked as long as the read 
                  io is  not completed. The position of the wanted log page is given 
                  as a log address  which will be transformed into a devspace 
                  specific block address. The correctness of the read log info 
                  page will be checked. In case of errors detected by the check 
                  method of the page the read io will be repeated. If the error 
                  is still present  after three read operations, the provided 
                  buffer is dumped into the rundirectory of the kernel. Each 
                  failed read operation is logged within the knldiag and knldiag 
                  error file. In any case of errors an emergency shutdown is 
                  executed and the volume is set to bad.
                  NOTE: that it is assumed that the given page frames(!) contained
                  in the page handler are in strictly ascending order within the 
                  memory!
    arguments:    taskId  [in] Identification of the calling task.
                  page    [in] Handler for the wanted pages.
                  address [in] Position of the wanted pages within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadLogPages(
        const tsp00_TaskId          taskId,
        IOMan_LogPages              &pages,
        const Log_ClusterAddress    &address ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WriteLogPages()
    description:  This method is used to write a number of log pages synchronous
                  into one archive log. The caller is blocked as long as the write 
                  io is not completed. The maximum number of writable log pages is
                  limited by a database configuration parameter. Additional to the
                  writeable log pages the caller has to provide a log address, which 
                  specifies the write destination on an archive log. The given log
                  address will be transformed into a log devspace specific cluster 
                  address. Perhaps the I/O Management splits the order into separate 
                  pieces conditional on the number of log devspaces belonging to one 
                  archive log. Before the proper write is executed the correctness of 
                  the given pages will be checked. In case of an error the corrupted 
                  page is dumped into the kernel rundirectory and an emergency shutdown 
                  is executed.In the other case the page is prepared for the io, for
                  example the checksum is calculated. Note that the writeable log pages 
                  not necessarily stored in a contiguous memory block, because this 
                  method uses scatter/gather io.In any case of errors an emergency 
                  shutdown is executed. 
    arguments:    TaskId  [in] Identification of the calling task.
                  Pages   [in] Handler of the given pages
                  Address [in] Position for the given pages within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    
    virtual void WriteLogPages(
        tsp00_TaskId              TaskId,
        IOMan_LogPages            &Pages,
        const Log_ClusterAddress  &Address ) = 0;
};

/*!*****************************************************************************

   endclass: IOMan_ILogIO

*******************************************************************************/


#endif  /* IOMAN_ILOGIO_HPP */
