/****************************************************************************

  module      : IOMan_ILogInfo.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for log devspace statistics.


  last changed: 2001-05-10  16:52
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_ILOGINFO_HPP
#define IOMAN_ILOGINFO_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_ILogInfo

   description: This interface is used for log devspace statistics.
              
*******************************************************************************/

class IOMan_ILogInfo
{
public:

    /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_ILogInfo interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_ILogInfo interface
    --------------------------------------------------------------------------*/

    static IOMan_ILogInfo& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     ConfiguredLogDevSpaces()
    description:  This method provides the number of configured log devspaces.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int ConfiguredLogDevSpaces() const = 0;

    /*!-------------------------------------------------------------------------
    function:     IsLogMirrored()
    description:  This method provides whether the log area consists of mirrored
                  log volumes..
    arguments:    none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsLogMirrored() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetBadLogVolume()
    description:  This method is used to determine the name of a log volume
                  which is marked as bad resp. not accessable, because it's
                  removed. This method is only available in mode admin and
                  only if log mode dual is configured.
    arguments:    taskId     [in]  Identification of the calling task.
                  badDevName [out] Name of the bad log volume
    return value: true means bad log volme found
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool GetBadLogVolume(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &badDevName ) = 0;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevSpaceNo()
    description:  This method returns the logical identifier of for the given
                  log devspace name. If the devspace name doesn't correspond to
                  a valid devspace the return value is false.
    arguments:    devName [in]  Name of a data devspace
                  devNo   [out] Logical devspace identifier
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetLogDevSpaceNo(
        tsp00_VFilename     &devName,
        IOMan_DeviceNo      &devNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevName()
    description:  This method returns the name of the log devspace identified
                  by the given logical identifier.
    arguments:    devNo   [in ] Logical devspace identifier
                  devName [out] Name of a data devspace
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetLogDevName(
        const IOMan_DeviceNo        devNo,
        tsp00_VFilename             &devName ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevSpaceSize()
    description:  This method provides the configured size in pages of the 
                  specified log devspace. It is assumed that the devspace 
                  identifier is valid.
    arguments:    devNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetLogDevSpaceSize( const IOMan_DeviceNo devNo ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevSpaceSize()
    description:  This method provides the usable size in pages of the 
                  specified log devspace. It is assumed that the devspace 
                  identifier is valid.
    arguments:    devNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetLogDevSpaceUsableSize( const IOMan_DeviceNo devNo ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     TotalUsableLogPages()
    description:  This method returns the total number of pages storable within
                  all log devspaces  This number is only available if all log devspaces
                  are in mode open, else zero is returned.
    arguments:    none
    return value: SAPDB_Int4
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4 TotalUsableLogPages() const = 0;
};

/*!*****************************************************************************

   endclass: IOMan_ILogInfo

*******************************************************************************/


#endif  /* IOMAN_ILOGINFO_HPP */
