/*!**************************************************************************

  module      : IOMan_ILogManager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Interface for archive log administration and log writer registration.


  last changed: 2001-03-14  15:08
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_ILOGMANAGER_HPP
#define IOMAN_ILOGMANAGER_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

#include "IOManager/IOMan_Types.hpp"
#include "IOManager/IOMan_LogPages.hpp"

/*===========================================================================*
*  DEFINES                                                                  *
*===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

  class: IOMan_ILogManager

  description: This interface is used for archive log administration purpose.
             
*******************************************************************************/

class IOMan_ILogManager
{
public:

    /*!-------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the IOMan_ILogManager interface
                  of component IOManager. It is assumed that the IOManager singleton 
                  is available! No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_ILogManager interface
    --------------------------------------------------------------------------*/

    static IOMan_ILogManager& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     CloseAllLogDevSpaces
    description:  This method is used to bring all archive logs being in mode
                  open in mode close.
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void CloseAllLogDevSpaces( const tsp00_TaskId  taskId ) = 0;

    /*!-------------------------------------------------------------------------
    function:     CreateAllLogDevSpaces
    description:  This method is used to create, format and open all archive
                  logs specified within the data base configuration file.
    arguments:    TaskId [in] Identification of the calling task.
                  pDBIdentifier [in] Unique data base identifier
    return value: true means successful creation; else false
    --------------------------------------------------------------------------*/

    virtual bool CreateAllLogDevSpaces(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier ) = 0;

    /*!-------------------------------------------------------------------------
    function:     OpenAllArchiveLogs()
    description:  This method is used to bring all archive logs specified
                  within the data base configuration file from mode close into
                  mode open. It is assumed that all archive logs are generated
                  with the method CreateAllArchiveLogs. 
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    virtual bool OpenAllArchiveLogs( const tsp00_TaskId taskId ) = 0;

    /*!-------------------------------------------------------------------------
    function:     OpenOneLogDevSpace
    description:  This method is used to bring one log devspave identified by
                  the given DevNo from mode closed into mode open. If the log
                  devspace is not stored in the data base configuration file or 
                  the open method fails an error is returned.
    arguments:    taskId [in] Identification of the calling task.
                  devNo  [in] Logical devspace identifier 
    return value: true means succeful execution; else false
    --------------------------------------------------------------------------*/

    virtual bool OpenOneLogDevSpace(
        const tsp00_TaskId    taskId,
        const IOMan_DeviceNo  devNo ) = 0;

    /*!-------------------------------------------------------------------------
    function:     AddLogDevSpace()
    description:  This method is used to add a new log devspace to the given
                  archive log in warm as in cold mode too. The proper
                  activation of the new devspace must be done by the Log_IOManager.
    arguments:    taskId      [in]  Identification of the calling task.
                  bIsCold     [in]  Is data base state db_cold (opposite db_warm)
                  devName     [in]  Name of the new devspace
                  mirrDevName [in]  Name of the new mirrored devspace (log mode dual only)
                  devSize     [in]  Size of the new devsppace in pages
                  devNo       [in]  Logical indentifier of the devspace
                  page        [in]  (nil)-Log pages used for devspace formating
                  newLogSize  [out] New size in pages of the archive log
    return value: true means that add log devspace was successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddLogDevSpace(
        const tsp00_TaskId          taskId,
        const SAPDB_Bool            bIsCold,
        tsp00_VFilename             &devName,
        tsp00_VFilename             &mirrDevName,
        const IOMan_BlockCount      devSize,
        const IOMan_DeviceNo        devNo,
        IOMan_LogPages              &pages,
        IOMan_BlockCount            &newLogSize ) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReIntegrateBadLogVolume()
    description:  This method is used to repair a corrupted log volume by copy
                  the intact log volume of the same log device onto the corrupted.
                  If the given log volume isn't bad nothing happens. This
                  feature is available in log mode dual only.
    arguments:    taskId      [in] Identification of the calling task.
                  badDevName  [in] Name of the bad log volume to be reintegrated
    return value: true means that reintegrate log volume was successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool ReIntegrateBadLogVolume(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &badDevName ) = 0;
};

/*!*****************************************************************************

   endclass: IOMan_ILogManager

*******************************************************************************/

#endif  /* IOMAN_ILOGMANAGER_HPP */
