/*!**************************************************************************

  module      : IOMan_ILogVolumeInfo.hpp.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management

  description : Interface for log volume informations 

    
  last changed: 2002-08-16
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_ILOG_VOLUME_INFO_HPP
#define IOMAN_ILOG_VOLUME_INFO_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!***************************************************************************

   class: IOMan_LogVolumeInfo

   description: This class is used to get informations about the current
                log volume. Note that a log volume could be in state
                offline and therefore same informations could be not available!

******************************************************************************/


class IOMan_LogVolumeInfo
{
    friend class IOMan_LogVolumeInfoIterator;

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : IOMan_LogVolumeInfo()
    description : construct a handle for the given log volume. 
    arguments   : volumeNo [in] logical identifier of the log volume
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogVolumeInfo( const IOMan_DeviceNo volumeNo ):m_VolumeNo( volumeNo )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Information access                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetVolumeNo()
    description:  This method returns the logical log volume identifier
                  of the volume.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetVolumeNo() const
    {
        return( m_VolumeNo );
    }

    /*!-------------------------------------------------------------------------
    function:     GetVolumeNo()
    description:  This method returns whether the log volume is in state open.
    arguments:    none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsVolumeOnline() const;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevName()
    description:  This method returns the name of the log volume.
    arguments:    devName [out] Name of a log volume
    return value: true if volume name was found
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetVolumeName( tsp00_VFilename   &volumeName ) const;

    /*!-------------------------------------------------------------------------
    function:     GetVolumeSize()
    description:  This method provides the configured size in pages of the log volume. 
    arguments:    none
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetVolumeSize() const;

    /*!-------------------------------------------------------------------------
    function:     GetUsableVolumeSize()
    description:  This method provides the usable size in pages of the log volume.
    arguments:    none
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetUsableVolumeSize() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Information access                                        */
    /* -----------------------------------------------------------------------*/

private:

    void SetVolumeNo( const IOMan_DeviceNo volumeNo )
    {
        m_VolumeNo = volumeNo;
    }

    // no default constructor available;
    IOMan_LogVolumeInfo();

private:

    /*-------------------------------------------------------------------------
    declaration: m_VolumneNo
    description: logical identifier of the log volume
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo  m_VolumeNo;
};

/*!***************************************************************************

   endclass: IOMan_LogVolumeInfo

******************************************************************************/

/*!***************************************************************************

   class: IOMan_LogVolumeInfoIterator

   description: This class is used to iterator over all log volumes.
                Note that this iterator wil return volume which are not 
                in state online. Therefore you have to call the method
                IsVolumeOnline first.
            
******************************************************************************/

class IOMan_LogVolumeInfoIterator
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : IOMan_LogVolumeInfoIterator()
    description : constructor; Sets the iterator to the first log volume.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogVolumeInfoIterator();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/


    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator access                                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : Begin()
    description : Sets the iterator to the first logical log volume.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    void Begin();

    /*!-------------------------------------------------------------------------
    function    : End()
    description : Checks wheather the iterator is still valid.
    arguments   : none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool End() const;

    /*!-------------------------------------------------------------------------
    function    : operator++()
    description : Sets the iterator the the logical next log volume. The is
                  no check whether the volume is available.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    void operator++();

    /*!-------------------------------------------------------------------------
    function    : operator*()
    description : myDescription
    arguments   : Returns a reference to the log volume info. First check
                  whether the log volume is online.
    return value: IOMan_LogVolumeInfo&
    --------------------------------------------------------------------------*/

    const IOMan_LogVolumeInfo& operator*() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator access                                           */
    /* -----------------------------------------------------------------------*/

private:

    /*-------------------------------------------------------------------------
    declaration: m_VolumeInfo
    description: myText
    --------------------------------------------------------------------------*/

    IOMan_LogVolumeInfo  m_VolumeInfo;
};

/*!***************************************************************************

   endclass: IOMan_LogVolumeInfoIterator

******************************************************************************/

#endif  /* IOMAN_ILOG_VOLUME_INFO_HPP */
