/*!**************************************************************************

  module      : IOMan_InfoPage.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : Handler for a IOMan_InfoPage.
  
                A page handled by this class is stored at the begining of 
                each data base devspace. It contains a various number of 
                administration informations of the current devspace, for 
                example the logical device number and the size. This 
                informations are used to decide whether the devspace is
                usable and in which state the devspace is.

  last changed: 2001-07-02  16:36
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_INFOPAGE_HPP
#define IOMAN_INFOPAGE_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "KernelCommon/Kernel_Page74.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define IOMAN_INVALID_LINKAGE   0
#define IOMAN_DB_IDENT_LENGTH   132

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/



/*!*****************************************************************************

   class: IOMan_InfoPage

   description: Handler for IOMan_InfoPage. A page handled by this class is
                stored at the begining of each data base devspace. It contains
                a various number of administration informations of the current
                devspace, for example the logical device number and the size.
                This informations are used to decide whether the devspace is
                usable and in which state the devspace is.
              
*******************************************************************************/

class IOMan_InfoPage : public Kernel_Page74
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     IOMan_InfoPage()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_InfoPage()
    {}

    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame         [in] Page frame
                  devNo         [in] Logical devspace identifier
                  nextDevNo     [in] Logical devspace identifier of the logical 
                                     next devspace
                  capacity      [in] Capacity of the devspace in pages
                  blockSize     [in] BlockSize in bytes
                  pDBIdentifier [in] Pointer to the unique identifier
    return value: none
    --------------------------------------------------------------------------*/

    void SetAndInitializeFrame(
        const PageFrame         &frame,
        const IOMan_DeviceNo    devNo,
        const IOMan_BlockCount  capacity,
        const SAPDB_UInt4       blockSize,
        const SAPDB_Byte       *pDBIdentifier )
    {
        SetFrame( frame );
        InitializeFrame( devNo, capacity, blockSize, pDBIdentifier );
    }

    /*!-------------------------------------------------------------------------
    function:     InitializeFrame()
    description:  Initializes frame contents
    arguments:    devNo         [in] Logical devspace identifier
                  capacity      [in] Capacity of the devspace in pages
                  blockSize     [in] BlockSize in bytes
                  pDBIdentifier [in] Pointer to the unique identifier
    return value: none
    --------------------------------------------------------------------------*/

    void InitializeFrame(
        const IOMan_DeviceNo    devNo,
        const IOMan_BlockCount  capacity,
        const SAPDB_UInt4       blockSize,
        const SAPDB_Byte       *pDBIdentifier )
    {
        Kernel_Page74::InitializeFrame( Kernel_IOManInfoPageType );
        WriteMultiUse1( devNo );

        const SAPDB_UInt2 HeaderSize  = Kernel_Page74::GetHeaderSize();
        const SAPDB_UInt2 TrailerSize = Kernel_Page74::GetTrailerSize();

        // set the body of the frame to a defined value
        memset( m_Frame.GetPtrAt( HeaderSize, m_Frame.Capacity() - HeaderSize ),
                0x00, m_Frame.Capacity() - TrailerSize - HeaderSize );

        SetPrevLogicalDevNo( IOMAN_INVALID_LINKAGE ),
        SetNextLogicalDevNo( IOMAN_INVALID_LINKAGE );
        SetCapacity( capacity );
        SetBlockSize( blockSize );
        SetBadBlockNo( 0 ),
        SetBadResetCount( 0 ),
        SetDBIdentifier( pDBIdentifier );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /* -----------------------------------------------------------------------*/

    /*-------------------------------------------------------------------------
    function:     SetNextLogicalDevNo()
    description:  This method is used to write the logical device number of 
                  the next logical devspace into the info page. It is assumed
                  that the info page is accessable.
    arguments:    nextDevNo [in] Logical devspace identifier of the logical 
                                 next devspace
    return:       none
    --------------------------------------------------------------------------*/

    void SetNextLogicalDevNo( const IOMan_DeviceNo    nextDevNo )
    {
        PersistentHeaderPtr()->m_NextLogicalDevNo = nextDevNo;
    }

    /*-------------------------------------------------------------------------
    function:     SetPrevLogicalDevNo()
    description:  This method is used to write the logical device number of 
                  the previous logical devspace into the info page. It is 
                  assumed that the info page is accessable.
    arguments:    prevDevNo [in] Logical devspace identifier of the logical 
                                 next devspace
    return:       none
    --------------------------------------------------------------------------*/

    void SetPrevLogicalDevNo( const IOMan_DeviceNo    prevDevNo )
    {
        PersistentHeaderPtr()->m_PrevLogicalDevNo = prevDevNo;
    }

    /*-------------------------------------------------------------------------
    function:     SetBadBlockNo()
    description:  This method is used to write the block number of the bad
                  block of the current devspace into the info page. It is 
                  assumed that the info page is accessable.
    arguments:    blockNo [in] block number of the bad block
    return:       none
    --------------------------------------------------------------------------*/

    void SetBadBlockNo( const IOMan_BlockNo   blockNo )
    {
        PersistentHeaderPtr()->m_BadBlockNo = blockNo;
    }

    /*-------------------------------------------------------------------------
    function:     IncrementBadResetCount()
    description:  This method is used to increment the bad reset count. This 
                  indicates how often a bad devspace was reseted without 
                  performing a data base restore.It is assumed that the 
                  info page is accessable.
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/

    void IncrementBadResetCount()
    {
       PersistentHeaderPtr()->m_BadResetCount +=1;
    }

    /*!-------------------------------------------------------------------------
    function:     GetLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the devspace the current page is stored in. This number must 
                  be equal to the corresponding entry within the data base 
                  configuration file. It is assumed that the info page is accessable.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetLogicalDevNo() const
    {
        return ReadMultiUse1();
    }

    /*!-------------------------------------------------------------------------
    function:     GetNextLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the devspace which is the next in logical order. This number
                  must be equal to the logical next devspace within the data 
                  base configuration file. It is assumed that the info page is 
                  accessable.
    arguments:    none
    return:       IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetNextLogicalDevNo() const
    {
        return( PersistentHeaderPtr()->m_NextLogicalDevNo );
    }

    /*!-------------------------------------------------------------------------
    function:     GetPrevLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the devspace which is the predecessor in logical order. This 
                  number must be equal to the logical predecessor devspace within 
                  the data base configuration file. It is assumed that the info 
                  page is accessable.
    arguments:    none
    return:       IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetPrevLogicalDevNo() const
    {
        return( PersistentHeaderPtr()->m_PrevLogicalDevNo );
    }

    /*!-------------------------------------------------------------------------
    function:     GetCapacity()
    description:  This method is used to read the capacity of the devspace, The
                  capacity is given in blocks and the block size is fixed for a
                  devspace and not changeable. This value must be equal to the 
                  corresponding entry within the data base configuration file.
                  It is assumed that the info page is accessable.
    arguments:    none
    return:       IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetCapacity() const
    {
        return( PersistentHeaderPtr()->m_Capacity );
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockSize()
    description:  This method is used to read the block size of the devspace, 
                  The block size is fixed and not changeable. The default is
                  8192 bytes. It is assumed that the info page is accessable.
    arguments:    none
    return:       SAPDB_UInt4
    --------------------------------------------------------------------------*/

    SAPDB_UInt4 GetBlockSize() const
    {
        return( PersistentHeaderPtr()->m_BlockSize );
    }

    /*-------------------------------------------------------------------------
    function:     GetBadBlockNo()
    description:  This method is used to read the number of the block which
                  is responsible for setting the devspace to bad. It is assumed 
                  that the info page is accessable.
    arguments:    blockNo [in] block number of the bad block
    return:       SAPDB_UInt4
    --------------------------------------------------------------------------*/

    SAPDB_UInt4 GetBadBlockNo() const
    {
        return( PersistentHeaderPtr()->m_BadBlockNo );
    }

    /*!-------------------------------------------------------------------------
    function:     GetDBIdentifier()
    description:  This method is used to read the unique identifier of the 
                  devspace. All devspaces of the data base have the same 
                  DBIdentifier! It is assumed that the info page is accessable.
    arguments:    none
    return value: (const SAPDB_Byte*) pointer to (not null terminated) string
    --------------------------------------------------------------------------*/

    const SAPDB_Byte* GetDBIdentifier() const
    {
        return( PersistentHeaderPtr()->m_DBIdent );
    }

    /*!-------------------------------------------------------------------------
    function:     GetDBIdentifierSize()
    description:  This method is used to get the size of the DBIdentifier.
                  It is assumed that the info page is accessable.
    arguments:    none
    return value: SAPDB_UInt4 size in bytes
    --------------------------------------------------------------------------*/

    SAPDB_UInt4 GetDBIdentifierSize() const
    {
        SAPDBERR_ASSERT_STATE( IOMAN_DB_IDENT_LENGTH ==
                               sizeof( PersistentHeaderPtr()->m_DBIdent ));

        return( sizeof( PersistentHeaderPtr()->m_DBIdent ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsPrevLogicalDevNoValid()
    description:  This method is used to validate if a predecessor is defined.
                  It is assumed that the info page is accessable.
    arguments:    none
    return value: (bool) true if predecessor devspace is definded
    --------------------------------------------------------------------------*/

    bool IsPrevLogicalDevNoValid() const
    {
        return( GetPrevLogicalDevNo() != IOMAN_INVALID_LINKAGE );
    }

    /*!-------------------------------------------------------------------------
    function:     IsNextLogicalDevNoValid()
    description:  This method is used to validate if a successor is defined.
                  It is assumed that the info page is accessable.
    arguments:    none
    return value: (bool) true if successor devspace is definded
    --------------------------------------------------------------------------*/

    bool IsNextLogicalDevNoValid() const
    {
        return( GetNextLogicalDevNo() != IOMAN_INVALID_LINKAGE );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#else
private:
#endif

    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within IOMan_InfoPage header
    --------------------------------------------------------------------------*/

    // Must do this to make the source compileable
    typedef Kernel_Page74::PersistentHeader             PH1;    // page74 header

private:

    struct PersistentHeader
    {
        PH1             m_KernelPage74Header;
        SAPDB_UInt4     m_PrevLogicalDevNo;
        SAPDB_UInt4     m_NextLogicalDevNo;
        SAPDB_UInt4     m_Capacity;
        SAPDB_UInt4     m_BlockSize;
        SAPDB_UInt4     m_BadBlockNo;
        SAPDB_UInt4     m_BadResetCount;
        SAPDB_UInt4     m_Filler1;
        SAPDB_UInt4     m_Filler2;
        SAPDB_UInt4     m_Filler3;
        SAPDB_UInt4     m_Filler4;
        SAPDB_Byte      m_DBIdent[IOMAN_DB_IDENT_LENGTH];
    };

    /*--------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    return value: (PersistentHeader*) header pointer
    prototypes:   2
    --------------------------------------------------------------------------*/

    PersistentHeader* PersistentHeaderPtr()
    {
        return( reinterpret_cast< PersistentHeader* >
                ( m_Frame.GetPtrAt( 0, sizeof( PersistentHeader ))));
    }

    const PersistentHeader* PersistentHeaderPtr() const
    {
        return( reinterpret_cast< const PersistentHeader* >
                ( m_Frame.GetPtrAt( 0, sizeof( PersistentHeader ))));
    }

    /*--------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/

    virtual Kernel_Page::SizeType GetHeaderSize() const
    {
        return( sizeof( PersistentHeader ));
    }

    /*-------------------------------------------------------------------------
    function:     SetCapacity()
    description:  This method is used to write the capacity of the devspace
                  into the info page. The capacity is given in blocks.
    arguments:    capacity [in] Capacity of the devspace in pages
    return:       none
    --------------------------------------------------------------------------*/

    void SetCapacity( const IOMan_BlockCount  capacity )
    {
        PersistentHeaderPtr()->m_Capacity = capacity;
    }

    /*--------------------------------------------------------------------------
    function:     SetBlockSize()
    description:  This method is used to write the block size of the devspace
                  into the info page.
    arguments:    blockSize [in] BlockSize in bytes
    return:       none
    --------------------------------------------------------------------------*/

    void SetBlockSize( const SAPDB_UInt4    blockSize )
    {
        PersistentHeaderPtr()->m_BlockSize = blockSize;
    }

    /*--------------------------------------------------------------------------
    function:     SetBadResetCount()
    description:  This method is used to set the bad reset count.
    arguments:    resetCount [in] number of reset attemps
    return:       none
    --------------------------------------------------------------------------*/

    void SetBadResetCount( const SAPDB_UInt4    resetCount )
    {
        PersistentHeaderPtr()->m_BadResetCount = resetCount;
    }

    /*--------------------------------------------------------------------------
    function:     SetDBIdentifier()
    description:  This method is used to write the unique identfier of the 
                  devspace into the info page.
    arguments:    pDBIdentifier [in] Pointer to the unique identifier
    return value: none
    --------------------------------------------------------------------------*/

    void SetDBIdentifier(
        const SAPDB_Byte    *pDBIdentifier )
    {
        memcpy( PersistentHeaderPtr()->m_DBIdent, pDBIdentifier, GetDBIdentifierSize() );
    }
};

/*!*****************************************************************************

    endclass: IOMan_InfoPage

*******************************************************************************/


#endif  /* IOMAN_INFOPAGE_HPP */
