/*!**************************************************************************

  module      : IOMan_LogControl.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : This module is used to handle a multitude of log devices.


  last changed: 2001-03-07  12:18
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_LOG_CONTROL_HPP
#define IOMAN_LOG_CONTROL_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Container/Container_Vector.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_LogDevice.hpp"
#include "IOManager/IOMan_LogPages.hpp"
#include "IOManager/IOMan_LogDevices.hpp"
#include "IOManager/IOMan_PageFrameAllocator.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "Logging/Log_Types.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


/*!*****************************************************************************

   class: IOMan_LogControl

   description: This class is used to handle a multitude of log devices.
              
*******************************************************************************/

class IOMan_LogControl
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     IOMan_LogControl()
    description:  Default constructor. 
                  Creates a manager for all log devices.
    arguments:    pageAllocator  [in] Allocator for page frames
                  allocator      [in] Allocator
                  bIsLogMirrored [in] Is the log mode dual configured
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogControl(
        IOMan_PageFrameAllocator    &pageAllocator,
        SAPDBMem_IRawAllocator      &allocator,
        const SAPDB_Bool            bIsLogMirrored )
    :
    m_PageAllocator( pageAllocator ),
    m_bIsLogMirrored( bIsLogMirrored ),
    m_TotalLogPages( 0 ),
    m_ConfiguredDevices( 0 ),
    m_LastAccessedDevice( 0 ),
    m_UsedDevices( 0 ),
    m_LogDevices( allocator, bIsLogMirrored )
    {}

    /*!------------------------------------------------------------------------
    function:     Initialize()
    description:  This method is used to initialize all log devices.
    arguments:    taskId         [in] Identification of the calling task
                  maxLogDevices  [in] Maximum number of addressable log devices.
    return value: in case of an error false is returned else true
    --------------------------------------------------------------------------*/

    SAPDB_Bool Initialize(
        const tsp00_TaskId  taskId,
        const SAPDB_Int     maxLogDevices );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Log device access                                            */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Get()
    description:  This method is used to get access to a log device which is 
                  dentified by the given device number.
    arguments:    devNo [in] logical identifier of a log device
    return value: IOMan_ILogDevice
    --------------------------------------------------------------------------*/

    IOMan_ILogDevice& Get( const IOMan_DeviceNo devNo )
    {
        return( m_LogDevices.Get( devNo ));
    }

    const IOMan_ILogDevice & Get( const IOMan_DeviceNo devNo ) const
    {
        return( m_LogDevices.Get( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetBlockAddress()
    description:  This method is used to transform the logical address given by 
                  the component logging into a physical block address which consider
                  the partition of the log in separate log devices.
    arguments:    offset [in] Logical offset within the log
    return value: IOMan_BlockAddress
    --------------------------------------------------------------------------*/

    IOMan_BlockAddress GetBlockAddress( const Log_RawDeviceOffset offset );

    /*!-------------------------------------------------------------------------
    function:     GetClusterAddress()
    description:  This method is used to transform the logical address given by 
                  the component logging into a physical block address which consider
                  the partition of the log in separate log devices.
                  The returned cluster address contains the correspondimg block
                  address and furthermore the block count which is possible reduced
                  in comparison to the given BlockCount. In this case a second 
                  request is needed to handle the rest of the proper log address.
    arguments:    offset     [in] Logical offset within the log
                  blockCount [in] Number of blocks to read or write from log
    return value: IOMan_ClusterAddress
    --------------------------------------------------------------------------*/

    IOMan_ClusterAddress GetClusterAddress(
        const Log_RawDeviceOffset   offset,
        IOMan_BlockCount            blockCount );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Log device access                                         */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator access                                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetFirstLogDevSpaceNo()
    description:  This method returns the identifier of the first log device.
                  It's not guaranteed that the returned device number 
                  specifies a device which is online.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetFirstLogDevSpaceNo() const
    {
        return( m_LogDevices.FirstDeviceNo() );
    }

    /*!-------------------------------------------------------------------------
    function:     GetNextLogDevSpaceNo()
    description:  This method returns the next identifier of the log device
                  following the given log device. It's not guaranteed that
                  the returned device identifier specifies a device which
                  is online.
    arguments:    devNo [in] Logical device identifier
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetNextLogDevSpaceNo( const IOMan_DeviceNo devNo ) const
    {
        return( devNo + 1 );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator access                                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Administration                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     CloseAll()
    description:  This method is used to detach all configured log devices, 
                  which are in mode open.
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    void CloseAll( const tsp00_TaskId taskId );

    /*!-------------------------------------------------------------------------
    function:     CreateAll()
    description:  This method is used to create all configured log devices,
                  i.e. each log devices is generated and formated. At the end 
                  all log devices are in mode open.
    arguments:    taskId        [in] Identification of the calling task.
                  pDBIdentifier [in] Unique data base identifier
    return value: In case of an error false is returned else true.
    --------------------------------------------------------------------------*/

    SAPDB_Bool CreateAll(
        const tsp00_TaskId taskId,
        const SAPDB_Byte   *pDBIdentifier );

    /*!-------------------------------------------------------------------------
    function:     OpenAll()
    description:  This method is used to attach all configured log devices,
                  which are not still in mode open.
    arguments:    taskId [in] Identification of the calling task.
    return value: In case of an error false is returned else true.
    --------------------------------------------------------------------------*/

    SAPDB_Bool OpenAll( const tsp00_TaskId taskId );

    /*!-------------------------------------------------------------------------
    function:     AddDevSpace()
    description:  This method is used to add a new log devices to the 
                  log in warm as in cold mode too. The proper activation 
                  of the new device must be done by the Log_IOManager.
    arguments:    taskId      [in] Identification of the calling task.
                  devName     [in] Name of the new volume
                  mirrDevName [in] Name of the mirrored volume (log mode dual only)
                  devSize     [in] Size of the new device in pages
                  devNo       [in] Logical identifier of the device
                  pages       [in] (nil)-Log pages used for device formating
    return value: true means that add log device was successful
    --------------------------------------------------------------------------*/

    SAPDB_Bool AddDevSpace(
        const tsp00_TaskId          taskId,
        tsp00_VFilename             &devName,
        tsp00_VFilename             &mirrDevName,
        const IOMan_BlockCount      devSize,
        const IOMan_DeviceNo        devNo,
        IOMan_LogPages              &pages );

    /*!-------------------------------------------------------------------------
    function:     ReIntegrateBadLogVolume()
    description:  This method is used to repair a corrupted log volume by copy
                  the intact log volume of the same log device onto the corrupted.
                  If the given log volume isn't bad nothing happens. This
                  feature is available in log mode dual only.
    arguments:    taskId      [in] Identification of the calling task.
                  badDevName  [in] Name of the bad log volume to be reintegrated
    return value: true means that reintegrate log volume was successful
    --------------------------------------------------------------------------*/

    SAPDB_Bool ReIntegrateBadLogVolume(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &badDevName );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Administration                                            */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Configuration and statistic information                      */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetBadLogVolume()
    description:  This method is used to determine the name of a log volume
                  which is marked as bad resp. not accessable, because it's
                  removed. This method is only available in mode admin and
                  only if log mode dual is configured.
    arguments:    taskId     [in]  Identification of the calling task.
                  badDevName [out] Name of the bad log volume
    return value: true means bad log volme found
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetBadLogVolume(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &badDevName );

    /*!-------------------------------------------------------------------------
    function:     GetDevSpaceNo()
    description:  This method is used to determine the logical identifier for
                  a given log devices name. It is not needed that the log 
                  devices are in mode open.
    arguments:    devName [in] 
                  devNo   [out]
    return value: true means devices was found; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetDevSpaceNo(
        tsp00_VFilename &devName,
        IOMan_DeviceNo  &devNo );

    /*!-------------------------------------------------------------------------
    function:     ConfiguredDevSpaces()
    description:  This method returns the number of all configured log devices 
                  This number is also available if not all log devices are 
                  in mode open.
    arguments:    none
    return value: (SAPDB_Int) 
    --------------------------------------------------------------------------*/

    SAPDB_Int ConfiguredDevSpaces() const
    {
        return( m_ConfiguredDevices );
    }

    /*!-------------------------------------------------------------------------
    function:     IsLogMirrored()
    description:  This method provides whether the log area consists of mirrored
                  log volumes.
    arguments:    none
    return value: (SAPDB_Bool)
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsLogMirrored() const
    {
        return( m_bIsLogMirrored );
    }

    /*!-------------------------------------------------------------------------
    function:     TotalUsableLogPages()
    description:  This method returns the sum of all pages storable within all 
                  configured log devices. This  value is only available if all 
                  log devices are in mode open.
    arguments:    none
    return value: (SAPDB_Int4) 
    --------------------------------------------------------------------------*/

    SAPDB_Int4  TotalUsableLogPages() const
    {
        return( m_TotalLogPages );
    }
    /*!-------------------------------------------------------------------------
    function:     IsLogDevSpaceNoValid()
    description:  This method checks the validity of the given log device number 
    arguments:    devNo [in] Logical device identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsLogDevSpaceNoValid( const IOMan_DeviceNo devNo ) const
    {
        return( m_LogDevices.IsDeviceNoValid( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsLogDevSpaceOnline()
    description:  This method returns whether the given log device is in
                  state online.
    arguments:    devNo [in] Logical device identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsLogDevSpaceOnline( const IOMan_DeviceNo devNo ) const
    {
        if( ! m_LogDevices.IsDeviceNoValid( devNo ) )
            return( false );

        return( Get( devNo ).IsOnline() );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Configuration and statistic information                   */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     CreateLinkage()
    description:  This method is used to create a linkage between the given
                  log devices. The link anchor is stored within the IOManInfo
                  page of the devspaces.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    void CreateLinkage(
        const tsp00_TaskId  taskId,
        IOMan_ILogDevice    &prevDevice,
        IOMan_ILogDevice    &currDevice );

    /*--------------------------------------------------------------------------
    function:     DetermineLinkage()
    description:  This method is used to determine the logical predecessor and
                  successor of the device to be added. Neither a predecessor
                  nor a successor could be found the return DBIdentifier is NULL.
                  The DBIdentifier is a unqiue log identifier, which is part of
                  each log devspace and checked every restart.
    arguments:    none
    return value: (SAPDB_Byte*) 
    --------------------------------------------------------------------------*/

    const SAPDB_Byte *DetermineLinkage(
        const IOMan_ILogDevice  &currDevice,
        IOMan_ILogDevice        **pPrevDevice,
        IOMan_ILogDevice        **pNextDevice );

    /*--------------------------------------------------------------------------
    function:     WriteLogPages()
    description:  This method is used to write a couple of log pages to the
                  given log.
    arguments:    none
    return value: (SAPDB_Byte*) 
    --------------------------------------------------------------------------*/

    void WriteLogPages(
        const tsp00_TaskId  taskId,
        IOMan_ILogDevice    &device,
        IOMan_LogPages      &pages );

    // Copy and assignment operator are not supported

    IOMan_LogControl& operator=( const IOMan_LogControl &logControl );

    //    IOMan_LogControl( const IOMan_LogControl &logControl ); TODO

private:

    /*-----------------------------------------------------------------------------
    declaration: m_PageAllocator
    description: Allocator for page frames. The page frame size is definded by
                 the data base configuration file.
    ------------------------------------------------------------------------------*/

    IOMan_PageFrameAllocator    &m_PageAllocator;

    /*-----------------------------------------------------------------------------
    declaration: m_bIsLogMirrored
    description: specifies whether log mode dual is configured.
    ------------------------------------------------------------------------------*/

    const SAPDB_Bool  m_bIsLogMirrored;

    /*-----------------------------------------------------------------------------
    declaration: m_TotalLogPages
    description: Is the maximum number of pages storable in all log devices.
    ------------------------------------------------------------------------------*/

    IOMan_BlockCount m_TotalLogPages;

    /*-----------------------------------------------------------------------------
    declaration: m_ConfiguredDevices
    description: Is the number of configured log devices.
    ------------------------------------------------------------------------------*/

    SAPDB_UInt m_ConfiguredDevices;

    /*-----------------------------------------------------------------------------
    declaration: m_LastAccessedDevice
    description: Is the logical identifier for the log device which was involved
                 in the last read or write access. Probable this will be the
                 destination of the next operation too.
    ------------------------------------------------------------------------------*/

    IOMan_DeviceNo  m_LastAccessedDevice;

    /*-----------------------------------------------------------------------------
    declaration: m_UsedDevices
    description: Is the current number of log devices which are in mode open.
    ------------------------------------------------------------------------------*/

    SAPDB_UInt  m_UsedDevices;

    /*-------------------------------------------------------------------------
    declaration: m_Logs
    description: Collection of log devices.
    --------------------------------------------------------------------------*/

    IOMan_LogDevices    m_LogDevices;
};

/*!*****************************************************************************

   endclass: IOMan_LogControl

*******************************************************************************/


#endif  /* IOMAN_LOG_CONTROL_HPP */
