/*!**************************************************************************

  module      : IOMan_LogDevice.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management

  description : Internal interface for log devices in particular 
                single and dual log devices.

  last changed: 2002-09-02

  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_ILOG_DEVICE_HPP
#define IOMAN_ILOG_DEVICE_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"  

#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_InfoPage.hpp"
#include "IOManager/IOMan_KernelPages.hpp"
#include "IOManager/IOMan_LogVolume.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "KernelCommon/Kernel_IPage.hpp"
#include "Logging/Log_Types.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!***************************************************************************

   class: IOMan_ILogDevice

   description: This class is used as a log device interface. It describes
                the properties of the log devices in particular the single
                and dual log device. A single log device consists of one
                log volume and a dual log device of two volumes.
            
******************************************************************************/


class IOMan_ILogDevice
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Administration                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Initialize()  
    description:  This method is used to set the unique log device identifier
                  and to give the I/O capable memory for the IOManInfoPage for
                  each log volume.
    arguments:    devNo     [in] Logical device identifier
                  frame     [in] Frame for the IOManInfoPage of the log volume
                  mirrFrame [in] Frame for the IOManInfoPage of the mirrored log
                                 volume (log mode dual only)
    return value: none
    --------------------------------------------------------------------------*/

    virtual void Initialize(
        const IOMan_DeviceNo            devNo,
        const Kernel_IPage::PageFrame   &frame,
        const Kernel_IPage::PageFrame   &mirrFrame ) = 0;

    /*!-------------------------------------------------------------------------
    function:     Close()
    description:  This method is used to bring the log device from open mode into 
                  close mode, presumed it's in open mode. The device is detached. 
    arguments:    taskId [in] Identification of the calling task. 
    return value: none
    --------------------------------------------------------------------------*/

    virtual void Close( const tsp00_TaskId taskId ) = 0;

    /*!-------------------------------------------------------------------------
    function:     Create()
    description:  This method is used to create and format a log device.
                  If the creation is finished successful the method provides true 
                  else false. It is assumed and tested by an assertion that the 
                  device is in closed mode.
    arguments:    taskId        [in] Identification of the calling task. 
                  pDBIdentifier [in] Unique data base identifier
    return value: (IOMan_ReturnCode) IOMan_Okay means create was successful
    --------------------------------------------------------------------------*/

    virtual IOMan_ReturnCode Create(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier ) = 0;

    /*!-------------------------------------------------------------------------
    function:     Open()
    description:  This method is used to bring the log device from close mode into
                  open mode, presumed it isn't in open mode already. The device 
                  is attached. 
    arguments:    taskId      [in] Identification of the calling task.
                  logCapacity [in] Current size in pages of log devices opened so far.
    return value: (IOMan_ReturnCode) IOMan_Okay means open was successful
    --------------------------------------------------------------------------*/

    virtual IOMan_ReturnCode Open(
        const tsp00_TaskId      taskId,
        const IOMan_BlockCount  logCapacity ) = 0;

    /*!-------------------------------------------------------------------------
    function:     Add()
    description:  This method is used to inform the RTE about the presence of another
                  new log device. It's required and checked whether the given device 
                  is in mode offline. Furhtermore the added device is in mode closed.
    arguments:    devName     [in] name of the original device to be added
                  mirrDevName [in] name of the mirrored device (log mode dual only)
                  devSize     [in] capacity of the device to be added
    return value: true if add was successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool Add(
        tsp00_VFilename         &devName,
        tsp00_VFilename         &mirrDevName,
        const IOMan_BlockCount  devSize ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     Del()
    description:  This method is used to inform the RTE about the removing of 
                  the log device. It's required and checked whether the given 
                  device is in mode offline.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    virtual void Del() const = 0;

    /*-------------------------------------------------------------------------
    function:     SetNextLinkage()
    description:  This method is used to write the logical device number of 
                  the next logical device into the info page of the
                  current device.
    arguments:    taskId     [in] Identification of the calling task.
                  nextDevice [in] device handler
    return:       (SAPDB_Bool) linkage is created successfully
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool SetNextLinkage(
        const tsp00_TaskId      taskId,
        const IOMan_ILogDevice  &nextDevice) = 0;

    /*-------------------------------------------------------------------------
    function:     SetPrevLinkage()
    description:  This method is used to write the logical device number of 
                  the previous logical device into the info page of the
                  current device.
    arguments:    taskId     [in] Identification of the calling task.
                  prevDevice [in] device handler
    return:      (SAPDB_Bool) linkage is created successfully
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool SetPrevLinkage(
        const tsp00_TaskId      taskId,
        const IOMan_ILogDevice  &prevDevice ) = 0;

    /*!-------------------------------------------------------------------------
    function:     IsLinkageValid()
    description:  This method is used to validate the linkage between the currrent
                  and the given log device. 
    arguments:    taskId [in] Identification of the calling task.
                  device [in] device handler
    return value: (SAPDB_Bool) true if linkage is valid; else false
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsLinkageValid(
        const tsp00_TaskId      taskId,
        const IOMan_ILogDevice  &device ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     IsPrevLinkageValid()
    description:  This method is used to validate if a logical predecessor 
                  device is definded.
    arguments:    none.
    return value: (SAPDB_Bool) true if predecessor device is definded
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsPrevLinkageValid() const = 0;

    /*!-------------------------------------------------------------------------
    function:     IsNextLinkageValid()
    description:  This method is used to validate if a logical successor 
                  device is definded.
    arguments:    none
    return value: (SAPDB_Bool) true if successor device is definded
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsNextLinkageValid() const = 0;

    /*!-------------------------------------------------------------------------
    function:     IsDBIdentifierValid()
    description:  This method is used to compare the DBIdentifier of the 
                  currrent and the given device. If the identifiers are equal 
                  true is returned else false.
    arguments:    taskId [in] Identification of the calling task.
                  device [in] device handler
    return value: (SAPDB_Bool) true if identifiers are equal; else false
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsDBIdentifierValid(
        const tsp00_TaskId      taskId,
        const IOMan_ILogDevice  &device ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     UnSetBad()
    description:  This method is used to remove the bad marker from the device.
                  A following open call won't break down caused by the bad flag.
    arguments:    taskId [in] Identification of the calling task.
    return value: (IOMan_ReturnCode) IOMan_Okay means unset was successful
    --------------------------------------------------------------------------*/

    virtual IOMan_ReturnCode UnSetBad( const tsp00_TaskId   taskId ) = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Administration                                            */
    /* -----------------------------------------------------------------------*/
    /* -----------------------------------------------------------------------*/
    /*! chapter: Common information                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetCapacity()
    description:  This method is used to get the device capacity in blocks. 
                  This is the  maximum number of blocks storable in the device. 
                  Note that there are two kinds of capacity, the complete and 
                  the usable capacity. The complete capacity is equal to the 
                  value within the data base configuration file whereas the 
                  usable capacity the number of blocks defines which are realy 
                  available to store data pages. The difference is caused by 
                  IOMan_InfoPage which is stored within each device and therefore 
                  responsible for the reduced amount of blocks. The default 
                  parameter bReduceCapacity determines which value is requested. 
                  The value bReduceCapacity == true returns the reduced number 
                  of blocks, which is the proper amount of blocks for example 
                  usable by the converter. The block size is provided by the 
                  method GetBlockSize.
    arguments:    bReduceCapacity [in] Return the reduced capacity
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetCapacity(
        const SAPDB_Bool bReduceCapacity = true ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetDBIdentifier()
    description:  This method is used to get the unique identifier of the 
                  device. 
    arguments:    none
    return value: (const SAPDB_Byte*) pointer to (not null terminated) string
    --------------------------------------------------------------------------*/

    virtual const SAPDB_Byte* GetDBIdentifier() const = 0;

    /*!-------------------------------------------------------------------------
     function:     GetLogicalDevNo()
     description:  This method is used to get the logical identifier of the
                   device. All devices of the same device type are counted 
                   begining by ONE(!).
     arguments:    none
     return value: logical device id
     --------------------------------------------------------------------------*/

    virtual IOMan_DeviceNo  GetLogicalDevNo() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetName()
    description:  This method is used to get the configured name of the volume.
                  If no valid volume name is found false is returned and the
                  incoming volume name is not touched. Else the configured 
                  volume name is stored within the incoming parameter. Is is 
                  assumed and tested by an assertion that the device is in open 
                  mode. If log mode dual is defined the parameter bIsMirror
                  defines whether name of the original or the mirrored volume 
                  is requested. If no log mode dual is configured this parameter 
                  will beignored.
    arguments:    devName   [out] device name
                  bIsMirror [in]  name belongs to the mirror
    return value: true means success; else false
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool GetName(
        tsp00_VFilename     &devName,
        const SAPDB_Bool    bIsMirror = false ) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetTypeName()
    description:  This method is used to get the type for example data or log
                  of the current device.
    arguments:    none
    return value: device type
    --------------------------------------------------------------------------*/

    virtual SAPDB_Char* GetTypeName() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetNextLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the log device which is the next in logical order. This number
                  must be equal to the logical next log device within the data 
                  base configuration file.
    arguments:    none
    return:       IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    virtual IOMan_DeviceNo GetNextLogicalDevNo() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetPrevLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the log device which is the predecessor in logical order. 
                  This number must be equal to the logical predecessor log 
                  device within the data base configuration file.
    arguments:    none
    return:       IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    virtual IOMan_DeviceNo GetPrevLogicalDevNo() const = 0;

    /*!-------------------------------------------------------------------------
    function:     IsConfigured ()
    description:  This method is used to check whether the current log device 
                  exists, i.e. is stored in the data base configuration file.
    arguments:    none
    return value: true means is known as a valid log device; else false
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsConfigured() const = 0;

    /*!-------------------------------------------------------------------------
    function:     IsOnline()
    description:  This method is used to determine whether the log device 
                  is in open mode.
    arguments:    none
    return value: true means is in open mode; else false
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsOnline() const = 0;

    /*!-------------------------------------------------------------------------
    function:     FirstOffset()
    description:  This method provides the first logical block number of the
                  log device. For example if a log consist of two 
                  log devices with the capacity of 100 pages at each case. 
                  The FirstOffset of the first device is zero and of the 
                  second is 100.
    arguments:    none
    return value: Log_RawDeviceOffset
    --------------------------------------------------------------------------*/

    virtual Log_RawDeviceOffset FirstOffset() const = 0;

    /*!-------------------------------------------------------------------------
    function:     LastOffset()
    description:  This method provides the last logical block number of the
                  log device. For example if a log consist of two 
                  log devices with the capacity of 100 pages at each case. 
                  The LastOffset of the first device is 99 and of the 
                  second is 199.
    arguments:    none
    return value: Log_RawDeviceOffset
    --------------------------------------------------------------------------*/

    virtual Log_RawDeviceOffset LastOffset() const = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Common information                                        */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Input/Output                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     ReadPage()
    description:  This method is used to read one log page from a log device 
                  into the given memory. The correctness  of the read page will 
                  be checked as far as three times if an error has occured. Each 
                  error is loged in the knldiag file and the io buffer is dumped.
                  In any case of an error for example checksum or I/O subsystem
                  errors an emergency shutdown is executed.
    arguments:    taskId [in] Identification of the calling task.
                  page   [in] Handler of the wanted Page.
                  block  [in] Position of the wanted page within the device.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadPage(
        const tsp00_TaskId         taskId,
        Kernel_IPage               &page,
        const IOMan_BlockAddress   &block ) = 0;
    
    /*!-------------------------------------------------------------------------
    function:     ReadPages()
    description:  This method is used to read a couple of log pages from a log 
                  device into the given memory. The correctness  of the read 
                  pages will be checked as far as three times if an error has 
                  occured. Each error is loged in the knldiag file and the io 
                  buffer is dumped.In any case of an error for example checksum 
                  or I/O subsystem errors an emergency shutdown is executed.
    arguments:    taskId     [in] Identification of the calling task.
                  pages      [in] Handler of the pages to be read
                  cluster    [in] Address of the pages to be read
                  startIndex [in] Specifies the first page in the page collection 
                                  which is relevant for the following I/O. Usually
                                  the startIndex is zero. Only in case of read
                                  orders splited over more than one log device
                                  the startIndex is greater than zero.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadPages(
        const tsp00_TaskId         taskId,
        IOMan_IKernelPages         &pages,
        const IOMan_ClusterAddress &cluster,
        const SAPDB_UInt           startIndex ) = 0;

    /*!-------------------------------------------------------------------------
    function:     ReadSimple()
    description:  This method is used to read a page (default) from a device 
                  into the given memory. The correctness  of the read page won't
                  be checked. In case of a fatal errror during the read operation
                  caused by the IO subsystem the error is logged in the knldiag 
                  file and the return code is false. The device is marked as bad.
    arguments:    taskId     [in] Identification of the calling task.
                  pFrame     [in] Memory where the read page will be stored.
                  blockNo    [in] Position of the wanted page within the device.
                  blockCount [in] Number of pages to be read in one step
                                  Default is one page
    return value: true if the read operation is successful; else false.
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool ReadSimple(
        const tsp00_TaskId      taskId,
        const tsp00_PageAddr    pFrame,
        const IOMan_BlockNo     blockNo,
        const IOMan_BlockCount  blockCount = 1 ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WritePage()
    description:  This method is used to write one log page from given memory into
                  a log device. Before the proper io is executed the correctness
                  of the given page will be checked. Each error is loged in the 
                  knldiag file and the io buffer is dumped. In any case of error 
                  an emergency shutdown is executed.
    arguments:    taskId [in] Identification of the calling task.
                  page   [in] Handler of the wanted Page.
                  block  [in] Position of the wanted page within the device.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WritePage(
        const tsp00_TaskId          taskId,
        Kernel_IPage                &page,
        const IOMan_BlockAddress    &block ) = 0;

    /*!-------------------------------------------------------------------------
    function:     WritePages()
    description:  This method is used to write log pages from given memory into
                  a log device. Before the proper io is executed the correctness of  
                  the given page will be checked. Each error is loged in the 
                  knldiag file and the io buffer is dumped. In any case of error 
                  an emergency shutdown is executed.
    arguments:    taskId      [in] Identification of the calling task.
                  pages       [in] Handler of the pages to be written
                  cluster     [in] Destination address of the pages to be written
                  startIndex  [in] Specifies the first page in the page collection 
                                   which is relevant for the following I/O. Usually
                                   the startIndex is zero. Only in case of write
                                   orders splited over more than one log device
                                   the startIndex is greater than zero.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WritePages(
        const tsp00_TaskId          taskId,
        IOMan_IKernelPages          &pages,
        const IOMan_ClusterAddress  &cluster,
        const SAPDB_UInt            startIndex = IOMAN_FIRST_INDEX ) = 0;
   
    /*!-------------------------------------------------------------------------
    function:     WriteSimple()
    description:  This method is used to write a page onto a device 
                  The correctness  of the page won't be  checked. In case of a 
                  fatal errror during the write operation caused by the IO subsystem
                  the error is logged in the knldiag file and the return code is 
                  false. The device is marked as bad.
    arguments:    taskId     [in] Identification of the calling task.
                  page       [in] Handler of the given page.
                  blockNo    [in] Position for the page within the device.
    return value: true if the write operation is successful; else false.
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool WriteSimple(
        const tsp00_TaskId      taskId,
        const tsp00_PageAddr    pFrame,
        const IOMan_BlockNo     blockNo ) = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Input/Output                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Volume access                                                */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetPrimaryVolume()
    description:  In case of a single log device this method is used to
                  access the log volume. In case of a dual log device the
                  access to the primary log volume is returned.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    virtual IOMan_LogVolume&  GetPrimaryVolume() = 0;

    virtual const IOMan_LogVolume&  GetPrimaryVolume() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetSecondaryVolume()
    description:  In case of a single log device this method is used to
                  access the log volume as GetPrimaryVolume In case of a 
                  dual log device the access to the secondary log volume 
                  is returned.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    virtual IOMan_LogVolume&  GetSecondaryVolume() = 0;

    virtual const IOMan_LogVolume&  GetSecondaryVolume() const = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Volume access                                             */
    /* -----------------------------------------------------------------------*/

protected:

    // Do not create a pure object of this
    IOMan_ILogDevice(){}
};

/*!***************************************************************************

   endclass: IOMan_ILogDevice

******************************************************************************/

#endif  /* IOMAN_ILOG_DEVICE_HPP */
