/*!**************************************************************************

  module      : IOMan_LogDevices.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management

  description : This module is used to handle a collecton of log devices
                in particular single and dual logs.

  last changed: 2002-09-02

  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_LOG_DEVICES_HPP
#define IOMAN_LOG_DEVICES_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Container/Container_Vector.hpp"
#include "IOManager/IOMan_LogDevice.hpp"
#include "IOManager/IOMan_DualLogDevice.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class IOMan_LogDeviceIterator; // forward declaraion

/*!***************************************************************************

   class: IOMan_LogDevices

   description: This class is used to handle a collecton of log devices
                in particular single or dual logs. Note  that a dual log
                consists of original and mirrored log devspaces whereby a
                single log contains only of a single log devspace.
            
******************************************************************************/


class IOMan_LogDevices
{
public:

    friend class IOMan_LogDeviceIterator;

    /*!-------------------------------------------------------------------------
    declarations: Type definitions
    --------------------------------------------------------------------------*/

    typedef Container_Vector< IOMan_ILogDevice* >   LogDevices;
    typedef LogDevices::Iterator                    Iterator;
    typedef LogDevices::ConstIterator               ConstIterator;

    /*!-------------------------------------------------------------------------
    enddeclarations: Type definitions
    --------------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : IOMan_LogDevices()
    description : Constructor
                  Creates a collection of log devices.
    arguments   : allocator [in] Allocator
                  bIsLogMirrored  [in] Is log mode dual configured
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogDevices(
        SAPDBMem_IRawAllocator  &allocator,
        const SAPDB_Bool        bIsLogMirrored )
    :
    m_LogDevices( allocator ),
    m_bIsLogMirrored( bIsLogMirrored )
    {}

    /*!------------------------------------------------------------------------
    function:     Initialize()
    description:  This method is used to initialize all log devices.
    arguments:    taskId          [in] Identification of the calling task
                  maxLogDevSpaces [in] Maximum number of addressable log devspaces.
    return value: in case of an error false is returned else true
    --------------------------------------------------------------------------*/

    SAPDB_Bool Initialize(
        const tsp00_TaskId  taskId,
        const SAPDB_Int     maxLogDevSpaces );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: direct access                                                */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Get()
    description:  This method is used to get access to a log device which is 
                 identified by the given number.
    arguments:    devNo [in] logical identifier of a log device
    return value: IOMan_ILogDevice
    --------------------------------------------------------------------------*/

    IOMan_ILogDevice& Get( const IOMan_DeviceNo devNo )
    {
        return( *( m_LogDevices[ devNo - IOMan_LogDevices::FirstDeviceNo() ] ));
    }

    const IOMan_ILogDevice& Get( const IOMan_DeviceNo devNo ) const
    {
        return( *( m_LogDevices[ devNo - IOMan_LogDevices::FirstDeviceNo() ] ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetDualDevice()
    description:  This method is used to get access to a dual log device 
                  which is  identified by the given number. This operation
                  is available in log mode dual only.
    arguments:    devNo [in] logical identifier of a dual log device
    return value: IOMan_DualLogDevice
    --------------------------------------------------------------------------*/

    IOMan_DualLogDevice& GetDualDevice( const IOMan_DeviceNo devNo )
    {
        SAPDBERR_ASSERT_STATE( m_bIsLogMirrored );

        return( *( reinterpret_cast< IOMan_DualLogDevice* >
                   (m_LogDevices[ devNo - IOMan_LogDevices::FirstDeviceNo() ])));
    }

    const IOMan_DualLogDevice& GetDualDevice( const IOMan_DeviceNo devNo ) const
    {
        SAPDBERR_ASSERT_STATE( m_bIsLogMirrored );

        return( *( reinterpret_cast< IOMan_DualLogDevice* >
                   (m_LogDevices[ devNo - IOMan_LogDevices::FirstDeviceNo() ])));
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: direct access                                             */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Configuration and statistic information                      */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    function:     FirstDeviceNo()
    description:  This static method returns the number of the first logical
                  log device.
    arguments:    none
    return value: SAPDB_UInt
    --------------------------------------------------------------------------*/

    static IOMan_DeviceNo FirstDeviceNo()
    {
        return( 1 ); // IOMAN_FIRST_LOG_DEVSPACE
    }

    /*!-------------------------------------------------------------------------
    function:     IsDeviceNoValid()
    description:  This method checks the validity of the given log device number. 
    arguments:    devNo [in] Logical devspace identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDeviceNoValid( const IOMan_DeviceNo devNo ) const
    {
        if(
            ( IOMan_LogDevices::FirstDeviceNo() > devNo )
            ||
            ( SAPDB_UInt( devNo ) > MaxDevices())
        )
            return( false );
        else
            return( true );
    }

    /*!-------------------------------------------------------------------------
    function:     MaxDevices()
    description:  This method returns the maximun number of configurable log 
                  devices. This value is stored within the data base configuration
                  file.
    arguments:    none
    return value: (SAPDB_Int) Maximun number of configurable log devices.
    --------------------------------------------------------------------------*/

    SAPDB_UInt MaxDevices() const
    {
        return( m_LogDevices.GetCapacity() );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Configuration and statistic information                   */
    /* -----------------------------------------------------------------------*/

private:

    /*-------------------------------------------------------------------------
    declaration: m_LogDevices
    description: Contains a collection of pointers referencing either single
                 or dual log devices. Note that one log device consists of at 
                 least one log devspace. If log mode dual is defined each log 
                 device consists of two log devspaces (original and mirror).

    --------------------------------------------------------------------------*/

    LogDevices  m_LogDevices;

    /*-----------------------------------------------------------------------------
    declaration: m_bIsLogMirrored
    description: specifies whether log mode dual is configured.
    ------------------------------------------------------------------------------*/

    const SAPDB_Bool  m_bIsLogMirrored;
};

/*!***************************************************************************

   endclass: IOMan_LogDevices

******************************************************************************/

/*!***************************************************************************

   class: IOMan_LogDeviceIterator

   description: This class to get access to all log device via an iterator.
            
******************************************************************************/


class IOMan_LogDeviceIterator
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     IOMan_LogDeviceIterator()
    description:  Constructor
    arguments:    logDevices [in] Collection of log devices
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogDeviceIterator( IOMan_LogDevices   &logDevices )
    :m_LogDevices( logDevices )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator access                                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Begin()
    description:  This method is used to set the iterator to the first log
                  device. There is no check whether this device is valid.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void Begin()
    {
        m_DevNo = IOMan_LogDevices::FirstDeviceNo();
    }

    /*!-------------------------------------------------------------------------
    function:     End()
    description:  This method is used to report whether the current device 
                  pointed by the iterator is valid.
    arguments:    none
    return value: true means that the current device is not valid and that 
                  the end of the log device set is reached; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool End() const
    {
        return( SAPDB_UInt( m_DevNo ) > m_LogDevices.MaxDevices() );
    }

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  This method is used to set the iterator to the next log device.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void operator++()
    {
        ++m_DevNo;
    }

    /*!-------------------------------------------------------------------------
    function:     operator++()
    description:  This method is used to set the iterator to the next log device.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogDeviceIterator& operator=( const IOMan_LogDeviceIterator &iter )
    {
        m_DevNo = iter.m_DevNo;

        return( *this );
    }

    /*!-------------------------------------------------------------------------
    function:     operator*()
    description:  This method is used to get access to a log device pointed 
                  by the iterator.
    arguments:    none
    return value: IOMan_ILogDevice&
    --------------------------------------------------------------------------*/

    IOMan_ILogDevice& operator*()
    {
        return( m_LogDevices.Get( m_DevNo ));
    }

    const IOMan_ILogDevice& operator*() const
    {
        return( m_LogDevices.Get( m_DevNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetDualDevice()
    description:  This method is used to get access to a dual log device 
                  pointed by the iterator. This operation is available in 
                  log mode dual only.
    arguments:    none
    return value: IOMan_DualLogDevice&
    --------------------------------------------------------------------------*/

    IOMan_DualLogDevice& GetDualDevice()
    {
        return( m_LogDevices.GetDualDevice( m_DevNo ));
    }

    const IOMan_DualLogDevice& GetDualDevice( const IOMan_DeviceNo devNo ) const
    {
        return( m_LogDevices.GetDualDevice( m_DevNo ));
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator access                                           */
    /* -----------------------------------------------------------------------*/

private:

    // Copy operator is not supported

    IOMan_LogDeviceIterator( const IOMan_LogDeviceIterator &iter );

private:

    /*-------------------------------------------------------------------------
    declaration: m_LogDevices
    description: Reference of the log device collection this iterator works on.
    --------------------------------------------------------------------------*/

    IOMan_LogDevices  &m_LogDevices;

    /*-------------------------------------------------------------------------
    declaration: m_DevNo
    description: Identifier of the current log device the iterator works on.
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo  m_DevNo;
};

/*!***************************************************************************

   endclass: IOMan_LogDeviceIterator

******************************************************************************/

#endif  /* IOMAN_LOG_DEVICES_HPP */
