/****************************************************************************

  module      : IOMan_LogVolume.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : This module is used to handle a log volume.


  last changed: 2001-03-13  9:07
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "heo56.h"      // RTE   : Vsleep
#include "hgg01_3.h"    // PASCAL: g01page_size

#include "IOManager/IOMan_LogVolume.hpp"
#include "IOManager/IOMan_ReservedBlockAddress.hpp"
#include "KernelCommon/Kernel_FileIO.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

void
IOMan_LogVolume::ReadPage(
    const tsp00_TaskId         taskId,
    Kernel_IPage               &page,
    const IOMan_BlockAddress   &block )
{
    SAPDBERR_ASSERT_STATE( block.GetDeviceNo() == m_LogicalDevNo );

    if( ! IOMan_Volume::Read( taskId, page, block.GetBlockNo()))
    {
        SetBad( taskId, block.GetBlockNo() );

        IOMan_Exception errMsg( __CONTEXT__, IOMAN_EMERGENCY_MSG1 );
        RTE_Crash( errMsg );
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_LogVolume::ReadPages(
    const tsp00_TaskId         taskId,
    IOMan_IKernelPages         &pages,
    const IOMan_ClusterAddress &cluster,
    const SAPDB_UInt           startIndex )
{
    SAPDBERR_ASSERT_STATE( cluster.GetDeviceNo() == m_LogicalDevNo );

    IOMan_IKernelPagesIterator          iter( pages );
    IOMan_ClusterAddress::BlockIndex    index = 0;

#ifdef SAPDB_QUICK
    {
        // Danger, because it's assumed that the given page frames are
        // in strictly ascending order within the memory!
        IOMan_IKernelPagesIterator  prevIter( pages );

        for( prevIter.Begin(), iter.Begin( 1 ); !iter.End(); ++prevIter, ++iter )
            if( ( *prevIter ).GetFramePtr() + g01page_size  != ( *iter ).GetFramePtr() )
            {
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "IOMan_LogVolume::ReadPages") );
            }
    }
#endif

    for( SAPDB_UInt readCount = 1; IOMAN_MAX_READ_RETRIES >= readCount; ++readCount )
    {
        if(! ReadSimple( taskId, *( pages.GetFrameList( startIndex )),
                         cluster.GetBlockNo( 0 ),
                         cluster.GetBlockCount() ))
        {
            IOMan_Exception errMsg( __CONTEXT__, IOMAN_EMERGENCY_MSG1 );
            RTE_Crash( errMsg );
        }

        for( index = 0, iter.Begin( startIndex ); index < cluster.GetBlockCount(); ++iter, ++index )
        {
            if( ( *iter ).CheckAfterRead() )
                continue;

            IOMan_Exception errMsg( __CONTEXT__, IOMAN_BAD_PAGE,
                                    GetTypeName(), SAPDB_ToString( m_LogicalDevNo, _T_d ),
                                    SAPDB_ToString( cluster.GetBlockNo( index ), _T_d ));

            Kernel_FileIO   dumpFile( GetTypeName(), KERNEL_BAD_FILE_EXTENSION,
                                      SAPDB_ToString( m_LogicalDevNo, _T_d ),
                                      SAPDB_ToString( cluster.GetBlockNo( index ), _T_d ));
            dumpFile.Write( *iter );
            dumpFile.Close();

            RTE_Message( errMsg );

            vsleep( taskId, IOMAN_RETRY_SLEEP_TIME );
            break;
        }
        if( index >= cluster.GetBlockCount())
            return;     // usual case
        //else read the page collection again, because an error has occured
    }

    SetBad( taskId, cluster.GetBlockNo( 0 ) + index );

    IOMan_Exception errMsg( __CONTEXT__, IOMAN_EMERGENCY_MSG1 );
    RTE_Crash( errMsg );
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
