/*!**************************************************************************

  module      : IOMan_LogVolume.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management

  description : This module is used to handle a log volume.


  last changed: 2001-03-07  18:26

  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_LOG_VOLUME_HPP
#define IOMAN_LOG_VOLUME_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "vsp002.h"  // PASCAL: IO Types

#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_Volume.hpp"
#include "IOManager/IOMan_KernelPages.hpp"
#include "IOManager/IOMan_Exception.hpp"
#include "IOManager/IOMan_Messages.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "Logging/Log_Types.hpp"
#include "RunTime/RTE_Crash.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_LogVolume

   description: This class is used to handle log volumes.
              
*******************************************************************************/

class IOMan_LogVolume : public IOMan_Volume
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     IOMan_LogVolume()  
    description:  Constructor. 
                  This method is used to set the member data to an initial state, 
                  in particular the type to log. Additional the constructor of 
                  the derived IOMan_Volume class is called. Note that the 
                  logical devo won't be set.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogVolume()
    :
    IOMan_Volume( tsp2_dev_type::fromConst( sp2dt_log )),
    m_FirstOffset( 0 ),
    m_LastOffset( 0 )
    {}

    /*!-------------------------------------------------------------------------
    function:     Initialize()  
    description:  This method is used to set the individual logical log volume
                  identifier.
    arguments:    devNo [in] Logical volume identifier
                  frame [in] Frame for the IOManInfoPage of the log volume
    return value: none
    --------------------------------------------------------------------------*/

    void Initialize(
        const IOMan_DeviceNo            devNo,
        const Kernel_IPage::PageFrame   &frame)
    {
        IOMan_Volume::Initialize( devNo, frame );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Administration                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Open()
    description:  This method is used to bring a log volume from close mode into
                  open mode. For more informations see IOMan_Volume::Open().
    arguments:    taskId      [in] Identification of the calling task.
                  logCapacity [in] Current size in pages of log devices opened so far.
    return value: (IOMan_ReturnCode) IOMan_Okay means open was successful
    --------------------------------------------------------------------------*/

    IOMan_ReturnCode Open(
        const tsp00_TaskId      taskId,
        const IOMan_BlockCount  logCapacity )
    {
        IOMan_ReturnCode RC = IOMan_Volume::Open( taskId );

        if( IOMan_Okay == RC )
        {
            m_FirstOffset = logCapacity;
            m_LastOffset  = m_FirstOffset +  GetCapacity() -1;
        }
        return( RC );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Administration                                            */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Common information                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     FirstOffset()
    description:  This method provides the first logical block number of the
                  log volume. For example if a log device consist of two 
                  log volume with the capacity of 100 pages at each case. 
                  The FirstOffset of the first volume is zero and of the 
                  second is 100.
    arguments:    none
    return value: Log_RawDeviceOffset
    --------------------------------------------------------------------------*/

    Log_RawDeviceOffset FirstOffset() const
    {
        return( m_FirstOffset );
    }

    /*!-------------------------------------------------------------------------
    function:     LastOffset()
    description:  This method provides the last logical block number of the
                  log volume. For example if a log device consist of two 
                  log volume with the capacity of 100 pages at each case. 
                  The LastOffset of the first volume is 99 and of the 
                  second is 199.
    arguments:    none
    return value: Log_RawDeviceOffset
    --------------------------------------------------------------------------*/

    Log_RawDeviceOffset LastOffset() const
    {
        return( m_LastOffset );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Common information                                        */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Input/Output                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     ReadPage()
    description:  This method is used to read one log page from a log volume 
                  into the given memory. The correctness  of the read page will 
                  be checked as far as three times if an error has occured. Each 
                  error is loged in the knldiag file and the io buffer is dumped.
                  In any case of an error for example checksum or I/O subsystem
                  errors an emergency shutdown is executed and the volume is set
                  to bad.
    arguments:    taskId [in] Identification of the calling task.
                  page   [in] Handler of the wanted Page.
                  block  [in] Position of the wanted page within the volume.
    return value: none
    --------------------------------------------------------------------------*/

    void ReadPage(
        const tsp00_TaskId         taskId,
        Kernel_IPage               &page,
        const IOMan_BlockAddress   &block );

    /*!-------------------------------------------------------------------------
    function:     ReadPages()
    description:  This method is used to read a couple of log pages from a log 
                  volume into the given memory. The correctness  of the read 
                  pages will be checked as far as three times if an error has 
                  occured. Each error is loged in the knldiag file and the io 
                  buffer is dumped.In any case of an error for example checksum 
                  or I/O subsystem errors an emergency shutdown is executed and
                  the volume is set to bad.
    arguments:    taskId     [in] Identification of the calling task.
                  pages      [in] Handler of the pages to be read
                  cluster    [in] Address of the pages to be read
                  startIndex [in] Specifies the first page in the page collection 
                                  which is relevant for the following I/O. Usually
                                  the startIndex is zero. Only in case of read
                                  orders splited over more than one log volume
                                  the startIndex is greater than zero.
    return value: none
    --------------------------------------------------------------------------*/

    void ReadPages(
        const tsp00_TaskId         taskId,
        IOMan_IKernelPages         &pages,
        const IOMan_ClusterAddress &cluster,
        const SAPDB_UInt           startIndex );

    /*!-------------------------------------------------------------------------
    function:     WritePage()
    description:  This method is used to write one log page from given memory into
                  a log volume. Before the proper io is executed the correctness
                  of the given page will be checked. Each error is loged in the 
                  knldiag file and the io buffer is dumped. In any case of error 
                  an emergency shutdown is executed.
    arguments:    taskId [in] Identification of the calling task.
                  page   [in] Handler of the wanted Page.
                  block  [in] Position of the wanted page within the volume.
    return value: none
    --------------------------------------------------------------------------*/

    void WritePage(
        const tsp00_TaskId          taskId,
        Kernel_IPage                &page,
        const IOMan_BlockAddress    &block )
    {
        SAPDBERR_ASSERT_STATE( block.GetDeviceNo() == m_LogicalDevNo );

        IOMan_Volume::Write( taskId, page, block.GetBlockNo());
    }

    /*!-------------------------------------------------------------------------
    function:     WritePages()
    description:  This method is used to write log pages from given memory into
                  a volume. Before the proper io is executed the correctness of  
                  the given page will be checked. Each error is loged in the 
                  knldiag file and the io buffer is dumped. In any case of error 
                  an emergency shutdown is executed.
    arguments:    taskId      [in] Identification of the calling task.
                  pages       [in] Handler of the pages to be written
                  cluster     [in] Destination address of the pages to be written
                  startIndex  [in] Specifies the first page in the page collection 
                                   which is relevant for the following I/O. Usually
                                   the startIndex is zero. Only in case of write
                                   orders splited over more than one log volume
                                   the startIndex is greater than zero.
    return value: none
    --------------------------------------------------------------------------*/

    void WritePages(
        const tsp00_TaskId          taskId,
        IOMan_IKernelPages          &pages,
        const IOMan_ClusterAddress  &cluster,
        const SAPDB_UInt            startIndex)
    {
        SAPDBERR_ASSERT_STATE( cluster.GetDeviceNo() == m_LogicalDevNo );

        IOMan_Volume::WriteVector( taskId, pages, cluster, startIndex );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Input/Output                                              */
    /* -----------------------------------------------------------------------*/

protected:

    /*!-------------------------------------------------------------------------
    function:     IOMan_LogVolume()  
    description:  Constructor. 
                  This method is used to set the member data to an initial state, 
                  in particular the type to log. Additional the constructor of 
                  the derived IOMan_Volume class is called. Note that the 
                  logical devo won't be set.
    arguments:    bIsSecondaryVolume [in] specifies whether this log volume is
                                          a secondary volume. This is only 
                                          possible if log mode dual is configured.
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_LogVolume( const SAPDB_Bool bIsSecondaryVolume )
    :
    IOMan_Volume( bIsSecondaryVolume ?
                    (tsp2_dev_type::fromConst( sp2dt_mirr_log )):
                    (tsp2_dev_type::fromConst( sp2dt_log ))),
    m_FirstOffset( 0 ),
    m_LastOffset( 0 )
    {}

private:

    // Assignment operator is not supported

    IOMan_LogVolume& operator=( const IOMan_LogVolume &volume );

    // copy constructor is needed for newarray operation during initialization
    //IOMan_LogVolume( const IOMan_LogVolume &volume ); 

private:

    /*-------------------------------------------------------------------------
    declaration: m_FirstOffset
    description: This defines the logical first block number of an log volume.
                 For example if a log device consist of two log volume with 
                 the capacity of 100 pages at each case. The m_FirstOffset of 
                 the first volume is zero and of the second is 100.
    --------------------------------------------------------------------------*/

    Log_RawDeviceOffset m_FirstOffset;

    /*-------------------------------------------------------------------------
    declaration: m_LastOffset
    description: This defines the logical last block number of an log volume.
                 For example if a log device consist of two log volume with 
                 the capacity of 100 pages at each case. The m_LastOffset of 
                 the first volume is 99 and of the second is 199.
    --------------------------------------------------------------------------*/

    Log_RawDeviceOffset m_LastOffset;
};

/*!*****************************************************************************

   endclass: IOMan_LogVolume

*******************************************************************************/

#endif  /* IOMAN_LOG_VOLUME_HPP */
