/****************************************************************************

  module      : IOMan_Manager.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management 
  description : 


  last changed: 2001-01-12  13:19
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hbd73.h"
#include "hkb57.h"
#include "hgg01_3.h"
#include "vsp002.h"

#include "FreeBlockManagement/FBM_Types.hpp"
#include "FreeBlockManagement/FBM_Manager.hpp"
#include "Converter/Converter_IPageNoMapping.hpp"
#include "Converter/Converter_ICommon.hpp"
#include "IOManager/IOMan_Exception.hpp"
#include "IOManager/IOMan_Manager.hpp"
#include "IOManager/IOMan_Messages.hpp"
#include "IOManager/IOMan_Migration.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_FileIO.hpp"
#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccess.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"
#include "SAPDBCommon/SAPDB_ToString.hpp"
#include "Trace/Trace_DevSpaceIOEntry.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define TRUE                            1
#define IOMAN_READ_IO                   1

#define IOMAN_MIN_REQUIRED_DATA_PAGES   2000 // minimum number of data pages

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

IOMan_Manager*  IOMan_Manager::m_Instance = NULL;

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/


/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/

inline
SAPDB_Bool IsMirroredLogConfigured()
{

    RTEConf_Parameter::String   paramValue;
    SAPDBErr_MessageList        errMsg;

    if( ! RTEConf_ParameterAccess::Instance()->GetString(
                UTF8( "LOG_MIRRORED" ), paramValue, errMsg ))
        RTE_Crash( errMsg );

    if( 0 == strcmp( "NO", ( SAPDB_Char* ) paramValue ))
        return( SAPDB_FALSE ); // LOG_MIRRORED == NO
    else
        return( SAPDB_TRUE ); // LOG_MIRRORED == YES
}

/*---------------------------------------------------------------------------*/

SAPDB_UInt
GetMaxConfiguredVolumes( const RTEConf_Parameter::Name paramName )
{
    RTEConf_Parameter::Integer  paramValue;
    SAPDBErr_MessageList        errMsg;

    if( ! RTEConf_ParameterAccess::Instance()->GetInteger(
                paramName, paramValue, errMsg ))
        RTE_Crash( errMsg );

    return(( SAPDB_UInt ) paramValue );
}

/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

IOMan_Manager::IOMan_Manager( SAPDBMem_IRawAllocator  &allocator )
        :
        m_Allocator( UTF8( "IOMan_Manager" ), allocator ),
        m_PageAllocator(),
        m_DataControl( m_PageAllocator, m_Allocator ),
        m_LogControl( m_PageAllocator, m_Allocator, IsMirroredLogConfigured()),
        m_ReservedBlocks()
{};

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::CreateInstance( const tsp00_TaskId  taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::CreateInstance", IOMan_Trace, 5 );

    if ( NULL == m_Instance )
    {
        m_Instance = new( RTEMem_Allocator::Instance() )
                     IOMan_Manager( RTEMem_Allocator::Instance() );

        if (NULL == m_Instance ) return( false );

        if (! m_Instance->Initialize( taskId,
                                      GetMaxConfiguredVolumes( UTF8( "MAXDATADEVSPACES" )),
                                      GetMaxConfiguredVolumes( UTF8( "MAXARCHIVELOGS" ))))
            return( false );
    }
    return( true );
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::Initialize(
    const tsp00_TaskId  taskId,
    const SAPDB_UInt    maxDataDevSpaces,
    const SAPDB_UInt    maxLogDevSpaces )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::Initialize", IOMan_Trace, 5 );

    if(! m_DataControl.Initialize( taskId, maxDataDevSpaces ))
        return( false );

    if(! m_LogControl.Initialize( taskId, maxLogDevSpaces ))
        return( false );

    return( true );
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::CloseAllLogDevSpaces( const tsp00_TaskId  taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::CloseAllLogDevspaces", IOMan_Trace, 5 );

    m_LogControl.CloseAll( taskId );
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::CloseAllDataDevSpaces( const tsp00_TaskId  taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::CloseAllDataDevSpaces", IOMan_Trace, 5 );

    m_DataControl.CloseAll( taskId );

    FBM_Manager::Instance().Shutdown( taskId );
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::CreateAllLogDevSpaces(
    const tsp00_TaskId   taskId,
    const SAPDB_Byte    *pDBIdentifier )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::CreateAllLogDevSpaces", IOMan_Trace, 5 );

    return( m_LogControl.CreateAll( taskId, pDBIdentifier ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::CreateAllDataDevSpaces(
    const tsp00_TaskId  taskId,
    const SAPDB_Byte    *pDBIdentifier )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::CreateAllDataDevSpaces", IOMan_Trace, 5 );

    // Create, open and format all configured data devspaces
    if(! m_DataControl.CreateAll( taskId, pDBIdentifier ))
        return( false );

    // Check data devspace size in regard to nonsense configuration
    if( IOMAN_MIN_REQUIRED_DATA_PAGES > m_DataControl.TotalDataPages() )
    {
        IOMan_Exception errMsg( __CONTEXT__, IOMAN_MIN_DATA_PAGES,
                                SAPDB_ToString( IOMAN_MIN_REQUIRED_DATA_PAGES, _T_d ),
                                SAPDB_ToString( m_DataControl.TotalUsableDataPages(), _T_d ));
        RTE_Crash( errMsg );
    }
    return( RestartAndInitializeFBM( taskId ));
}

/*---------------------------------------------------------------------------*/

IOMan_BlockCount
IOMan_Manager::GetDataDevSpaceSize( const IOMan_DeviceNo devNo ) const
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::GetDataDevSpaceSize", IOMan_Trace, 5 );

    return( m_DataControl.Get( devNo ).GetCapacity( false ));
}
/*---------------------------------------------------------------------------*/

IOMan_BlockCount
IOMan_Manager::GetDataDevSpaceUsableSize( const IOMan_DeviceNo devNo ) const
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::GetDataDevSpaceUsableSize", IOMan_Trace, 5 );

    const IOMan_BlockAddress  block = m_ReservedBlocks.GetRestartPageAddress();

    if(( block.IsValid() ) && ( devNo == block.GetDeviceNo() ))
        return( m_DataControl.Get( devNo ).GetCapacity( true ) -1 );
    else
        return( m_DataControl.Get( devNo ).GetCapacity( true ) );
}

/*---------------------------------------------------------------------------*/

IOMan_BlockCount
IOMan_Manager::GetLogDevSpaceSize( const IOMan_DeviceNo devNo ) const
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::GetLogDevSpaceSize", IOMan_Trace, 5 );

    return( m_LogControl.Get( devNo ).GetCapacity( false ));
}

/*---------------------------------------------------------------------------*/

IOMan_BlockCount
IOMan_Manager::GetLogDevSpaceUsableSize( const IOMan_DeviceNo   devNo ) const
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::GetLogDevSpaceUsableSize", IOMan_Trace, 5 );

    if( 1 == devNo )  // log info page
        return( m_LogControl.Get( devNo ).GetCapacity( true ) -1);
    else
        return( m_LogControl.Get( devNo ).GetCapacity( true ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::OpenAllArchiveLogs( const tsp00_TaskId taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::OpenAllArchiveLogs", IOMan_Trace, 5 );

    return( m_LogControl.OpenAll( taskId ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::OpenAllDataDevSpaces( const tsp00_TaskId taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::OpenAllDataDevSpaces", IOMan_Trace, 5 );

    if(! m_DataControl.OpenAll( taskId ))
        return( false );

    kb57RestoreRestartRecord( taskId );

    return( RestartAndInitializeFBM( taskId ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::OpenOneDataDevSpace(
    const tsp00_TaskId    taskId,
    const IOMan_DeviceNo  devNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::OpenOneDataDevSpace", IOMan_Trace, 5 );

    return( IOMan_Okay == ( m_DataControl.Get( devNo ).Open( taskId )));
}

/*---------------------------------------------------------------------------*/
bool
IOMan_Manager::OpenOneLogDevSpace(
    const tsp00_TaskId    taskId,
    const IOMan_DeviceNo  devNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::OpenOneLogDevSpace", IOMan_Trace, 5 );

    return( IOMan_Okay == ( m_LogControl.Get( devNo ).Open( taskId, 0 )));
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_Manager::AddDataDevSpace(
    const tsp00_TaskId      taskId,
    const SAPDB_Bool        bIsCold,
    tsp00_VFilename         &devName,
    const IOMan_BlockCount  devSize,
    const IOMan_DeviceNo    devNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::AddDataDevSpace", IOMan_Trace, 5 );

    bool retCode = true;

    if( bIsCold )
        retCode = m_DataControl.OpenAll( taskId );

    if( retCode )
        retCode = m_DataControl.AddDevSpace( taskId, bIsCold, devName, devSize, devNo );

    if( bIsCold )
        m_DataControl.CloseAll( taskId );

    return( retCode );
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_Manager::AddLogDevSpace(
    const tsp00_TaskId          taskId,
    const SAPDB_Bool            bIsCold,
    tsp00_VFilename             &devName,
    tsp00_VFilename             &mirrDevName,
    const IOMan_BlockCount      devSize,
    const IOMan_DeviceNo        devNo,
    IOMan_LogPages              &pages,
    IOMan_BlockCount            &newLogSize )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::AddLogDevSpace", IOMan_Trace, 5 );

    bool retCode = true;

    if( bIsCold )
        retCode = m_LogControl.OpenAll( taskId );

    if( retCode )
    {
        retCode = m_LogControl.AddDevSpace( taskId, devName, mirrDevName,
                                            devSize, devNo, pages );

        if( retCode )
            newLogSize = m_LogControl.TotalUsableLogPages();
    }

    if( bIsCold )
        m_LogControl.CloseAll( taskId ); // PTS 1112474 TS 2001-11-07

    return( retCode );
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::ReadConverterPage(
    tsp00_TaskId                taskId,
    const IOMan_BlockAddress    &block,
    Converter_Page              &page )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadConverterPage", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "ConverterPage: " << block.GetDeviceNo() <<\
                        "/" << block.GetBlockNo() );

    bool bOkay = m_DataControl.Get( block.GetDeviceNo() ).ReadPage( taskId, page, block );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_ReadConverterPageEntry( auxTrans, block );
    }

    if( TRUE == bOkay )
    {
        // Check given block address against the block address
        // within the obtained buffer
        IOMan_BlockAddress  pageBlock = page.ReadBlockAddress();

        if(
            ( block.GetDeviceNo() == pageBlock.GetDeviceNo() ) &&
            ( block.GetBlockNo() == pageBlock.GetBlockNo() )
        )
        {
            FBM_Manager::Instance().SetBlockStateToOccupied( taskId, block, FBM_ABORT_IF_ERROR );
            bd73ConverterIOStatistic( IOMAN_READ_IO );
            return;
        }
    }

    // Abnormal termination
    IOMan_Exception errMsg( __CONTEXT__, IOMAN_WRONG_CONV_PAGE,
                            KERNEL_DATA_FILE, SAPDB_ToString( block.GetDeviceNo(), _T_d ),
                            SAPDB_ToString( block.GetBlockNo(), _T_d ));

    Kernel_FileIO   DumpFile( KERNEL_DATA_FILE, KERNEL_BAD_FILE_EXTENSION,
                              SAPDB_ToString( block.GetDeviceNo(), _T_d ),
                              SAPDB_ToString( block.GetBlockNo(), _T_d ));
    DumpFile.Write( page );
    DumpFile.Close();

    RTE_Crash( errMsg );
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::ReadDataPage(
    const tsp00_TaskId          taskId,
    tgg00_BasisError            &trError,
    Data_BasePage               &page,
    const Data_PageNo           &pageNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadDataPage", IOMan_Trace, 5 );


    IOMan_BlockAddress     block =
        Converter_IPageNoMapping::Instance().GetDevicePositionForReading( taskId, pageNo,
                page.RecoveryMode());

    if (!block.IsValid())
    {
        trError = e_no_converter_entry;
        return false;
    }

    if(! m_DataControl.Get( block.GetDeviceNo() ).ReadPage( taskId, page, block ))
    {
        trError = e_bad_datapage;

        IOMan_Exception errMsg( __CONTEXT__, IOMAN_BAD_PAGENO,
                                SAPDB_ToString( page.PageNo(), _T_d ));
        RTE_Message( errMsg );
        return( false );
    }

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_ReadDataPageEntry( auxTrans, page.ReadPageType2(), page.PageNo(), block );
    }

    // Check given pno against the pno within the obtained buffer
    if( pageNo != page.PageNo() )
    {
        trError = e_bad_datapage;

        IOMan_Exception errMsg( __CONTEXT__, IOMAN_WRONG_PAGENO,
                                SAPDB_ToString( pageNo, _T_d ), KERNEL_DATA_FILE,
                                SAPDB_ToString( block.GetDeviceNo(), _T_d ),
                                SAPDB_ToString( block.GetBlockNo(), _T_d ));
        RTE_Message( errMsg );

        Kernel_FileIO   dumpFile( KERNEL_DATA_FILE, KERNEL_BAD_FILE_EXTENSION,
                                  SAPDB_ToString( pageNo, _T_d ));
        dumpFile.Write( page );
        return( false );
    }
    return( true );
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::ReadDataPageDirect(
    tsp00_TaskId               TaskId,
    Data_BasePage              &Page,
    const IOMan_BlockAddress   &Block )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadDataPageDirect", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "BlockAddress: " <<
                        Block.GetDeviceNo() << "/" << Block.GetBlockNo());

    if(! m_DataControl.Get( Block.GetDeviceNo() ).ReadPage( TaskId, Page, Block ))
        return( false );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext AuxTrans = BuildAuxTrans( TaskId );
        Trace_ReadDataPageEntry( AuxTrans, Page.ReadPageType2(), Page.PageNo(), Block );
    }
    return( true );
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::ReadLogInfoPage(
    tsp00_TaskId              taskId,
    Log_InfoPage              &page,
    const Log_ClusterAddress  &address )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadLogInfoPage", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5,
                        " LogOffset: " << address.GetOffset() <<
                        " BlockCount: " << address.GetBlockCount() );

    SAPDBERR_ASSERT_STATE( 1 == address.GetBlockCount() );
    SAPDBERR_ASSERT_STATE( address.GetOffset().IsValid() );

    IOMan_BlockAddress  block = m_LogControl.GetBlockAddress( address.GetOffset() );

    m_LogControl.Get( block.GetDeviceNo() ).ReadPage( taskId, page, block );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_ReadLogPageEntry( auxTrans, pt2LogInfo_egg00, page.ReadPageID(), block );
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::ReadLogPage(
    tsp00_TaskId             taskId,
    Log_Page                 &page,
    const Log_ClusterAddress &address )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadLogPage", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5,
                        " LogOffset: " << address.GetOffset() <<
                        " BlockCount: " << address.GetBlockCount() );

    SAPDBERR_ASSERT_STATE( 1 == address.GetBlockCount() );
    SAPDBERR_ASSERT_STATE( address.GetOffset().IsValid() );

    IOMan_BlockAddress  block = m_LogControl.GetBlockAddress( address.GetOffset() );

    m_LogControl.Get( block.GetDeviceNo() ).ReadPage( taskId, page, block );

    bd73LogIOStatistic( IOMAN_READ_IO );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_ReadLogPageEntry( auxTrans, pt2Nil_egg00, page.ReadLastWriterIOSequenceNo(), block );
    }

}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::ReadLogPages(
    const tsp00_TaskId          taskId,
    IOMan_LogPages              &pages,
    const Log_ClusterAddress    &address )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadLogPages", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5,
                        " LogOffset: " << address.GetOffset() <<
                        " BlockCount: " << address.GetBlockCount() );

    SAPDBERR_ASSERT_STATE( pages.GetPageCount() == address.GetBlockCount() );
    SAPDBERR_ASSERT_STATE( 0 < pages.GetPageCount() );
    SAPDBERR_ASSERT_STATE( address.GetOffset().IsValid() );

    Log_DeviceBlockCount pagesToRead = address.GetBlockCount();
    SAPDB_UInt           startIndex  = 0;
    IOMan_ClusterAddress cluster;

    do
    {
        cluster = m_LogControl.GetClusterAddress( address.GetOffset() + startIndex, pagesToRead );
        m_LogControl.Get( cluster.GetDeviceNo() ).ReadPages( taskId, pages, cluster, startIndex );

        if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
        {
            tgg00_TransContext               auxTrans = BuildAuxTrans( taskId );
            SAPDB_UInt                       traceNo  = ( cluster.GetBlockCount() > 1 )? 1:0;
            IOMan_ClusterAddress::BlockIndex index    = 0;
            IOMan_LogPagesIterator           iter( pages );

            for( iter.Begin( startIndex ); !iter.End() && index < cluster.GetBlockCount();
                    ++iter, ++index, ++traceNo )
            {
                Trace_ReadLogPageEntry( auxTrans, pt2Nil_egg00, ( *iter ).ReadLastWriterIOSequenceNo(),
                                        cluster.GetBlockAddress( index ), traceNo );
            }
        }
        pagesToRead -= cluster.GetBlockCount();
        startIndex  += cluster.GetBlockCount();
        bd73LogIOStatistic( IOMAN_READ_IO );
    }
    while( 0 < pagesToRead );
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::ReadRestartPage(
    tsp00_TaskId         TaskId,
    Kernel_RestartPage   &Page )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadRestartPage", IOMan_Trace, 5 );


    IOMan_BlockAddress  Block = m_ReservedBlocks.GetRestartPageAddress();

    if(! m_DataControl.Get( Block.GetDeviceNo() ).ReadPage( TaskId, Page, Block ))
    {
        IOMan_Exception ErrMsg( __CONTEXT__, IOMAN_BAD_RESTART_PAGE,
                                KERNEL_DATA_FILE, SAPDB_ToString( Block.GetDeviceNo(), _T_d ),
                                SAPDB_ToString( Block.GetBlockNo(), _T_d ));

        Kernel_FileIO   DumpFile( KERNEL_DATA_FILE, KERNEL_BAD_FILE_EXTENSION,
                                  SAPDB_ToString( Block.GetDeviceNo(), _T_d ),
                                  SAPDB_ToString( Block.GetBlockNo(), _T_d ));

        DumpFile.Write( Page );
        DumpFile.Close();

        RTE_Crash( ErrMsg );
    }

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext AuxTrans = BuildAuxTrans( TaskId );
        Trace_ReadDataPageEntry( AuxTrans, Page.ReadPageType2(), Page.GetPageID(), Block );
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::WriteConverterPage(
    const tsp00_TaskId  taskId,
    Converter_Page      &page )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteConverterPage", IOMan_Trace, 5 );


    IOMan_BlockAddress  block = page.ReadBlockAddress();

    if( block.IsValid() )
    {
        FBM_Manager::Instance().SetBlockStateToFreeAfterSVP( taskId, block );

        SAPDBTRACE_WRITELN( IOMan_Trace, 5, "Old Address: " <<
                            (( Converter_IndexPageType == page.ReadConverterPageType() )?
                             "IndexPage: " : "LeafPage: ") << block.GetDeviceNo() << "/" <<
                            block.GetBlockNo() );
    }

    block = FBM_Manager::Instance().GetFreeBlock( taskId );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "New Address: " << block.GetDeviceNo() << "/" <<
                        block.GetBlockNo() );

    page.WriteBlockAddress( block );
    page.WriteVersion(Converter_ICommon::Instance().Version ());

    m_DataControl.Get( block.GetDeviceNo() ).WritePage( taskId, page, block );

    bd73ConverterIOStatistic( ! IOMAN_READ_IO );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_WriteConverterPageEntry( auxTrans, block );
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::WriteDataPage(
    const tsp00_TaskId   taskId,
    tgg00_BasisError    &trError,
    Data_BasePage       &page )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteDataPage", IOMan_Trace, 5 );


    // Update the corresponding  converter entries, before the
    // converter version on the data pages will be changed because
    // the converter needs the old converter version to make decisions
    // about updating the old blockaddres.

    IOMan_BlockAddress block =
        Converter_IPageNoMapping::Instance().GetDevicePositionForWriting( taskId,
                trError, page.PageNo(), page.RecoveryMode(), page.ConverterVersion());

    // Store current converter version within the page image to note
    // the savepoint version when the page was written the last time
    // onto the data devspace.
    page.SetConverterVersion( Converter_ICommon::Instance().Version () );

    if( e_ok != trError )
        return;

    m_DataControl.Get( block.GetDeviceNo() ).WritePage( taskId, page, block );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_WriteDataPageEntry( auxTrans, page.ReadPageType2(), page.PageNo(), block );
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::WriteDataPages(
    const tsp00_TaskId         taskId,
    tgg00_BasisError           &trError,
    IOMan_DataPages            &pages,
    const IOMan_ClusterAddress &cluster )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteDataPages", IOMan_Trace, 5 );

    SAPDBERR_ASSERT_STATE( pages.GetPageCount() <= SAPDB_UInt( cluster.GetBlockCount()) );

    // Update the corresponding  converter entries, before the
    // converter version on the data pages will be changed because
    // the converter needs the old converter version to make decisions
    // about updating the old blockaddres.

    IOMan_ClusterAddress::BlockIndex    index = 0;
    IOMan_DataPagesIterator             iter( pages );

    for( iter.Begin(); !iter.End(); ++iter, ++index )
    {
        Converter_IPageNoMapping::Instance().SetDevicePositionForWriting (
            taskId, trError, ( *iter ).ConverterVersion(), ( *iter ).PageNo(),
            ( *iter ).RecoveryMode(), cluster.GetBlockAddress( index ));

        // Store current converter version within the page image to note
        // the savepoint version when the page was written the last time
        // onto the data devspace.

        ( *iter ).SetConverterVersion( Converter_ICommon::Instance().Version () );
    }

    if( e_ok != trError )
        return;

    m_DataControl.Get( cluster.GetDeviceNo() ).WritePages( taskId, pages, cluster );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        SAPDB_UInt  traceNo  = ( pages.GetPageCount() > 1 )? 1:0;
        IOMan_ClusterAddress::BlockIndex  index = 0;

        for (iter.Begin(); !iter.End(); ++iter, ++index, ++traceNo)
        {
            tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
            Trace_WriteDataPageEntry( auxTrans, ( *iter ).ReadPageType2(),
                                      ( *iter ).PageNo(), cluster.GetBlockAddress( index ), traceNo );
        }
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::WriteLogInfoPage(
    tsp00_TaskId              taskId,
    Log_InfoPage              &page,
    const Log_ClusterAddress  &address )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteLogInfoPage", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5,
                        " LogOffset: " << address.GetOffset() <<
                        " BlockCount: " << address.GetBlockCount() );

    SAPDBERR_ASSERT_STATE( 1 == address.GetBlockCount() );
    SAPDBERR_ASSERT_STATE( address.GetOffset().IsValid() );

    IOMan_BlockAddress  block = m_LogControl.GetBlockAddress( address.GetOffset() );

    m_LogControl.Get( block.GetDeviceNo() ).WritePage( taskId, page, block );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext auxTrans = BuildAuxTrans( taskId );
        Trace_WriteLogPageEntry( auxTrans, pt2LogInfo_egg00, page.ReadPageID(), block );
    }
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::WriteLogPages(
    tsp00_TaskId              taskId,
    IOMan_LogPages            &pages,
    const Log_ClusterAddress  &address )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteLogPages", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5,
                        " LogOffset: " << address.GetOffset() <<
                        " BlockCount: " << address.GetBlockCount() );

    SAPDBERR_ASSERT_STATE( pages.GetPageCount() == address.GetBlockCount() );
    SAPDBERR_ASSERT_STATE( 0 < pages.GetPageCount() );
    SAPDBERR_ASSERT_STATE( address.GetOffset().IsValid() );

    Log_DeviceBlockCount pagesToWrite = address.GetBlockCount();
    SAPDB_UInt           startOffset  = 0;
    IOMan_ClusterAddress cluster;

    do
    {
        cluster = m_LogControl.GetClusterAddress( address.GetOffset() + startOffset, pagesToWrite );
        m_LogControl.Get( cluster.GetDeviceNo() ).WritePages( taskId, pages, cluster, startOffset );

        if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
        {
            tgg00_TransContext               auxTrans = BuildAuxTrans( taskId );
            SAPDB_UInt                       traceNo  = ( cluster.GetBlockCount() > 1 )? 1:0;
            IOMan_ClusterAddress::BlockIndex index    = 0;
            IOMan_LogPagesIterator           iter( pages );

            for( iter.Begin( startOffset );
                    !iter.End() && index < cluster.GetBlockCount();
                    ++iter, ++index, ++traceNo )
            {
                Trace_WriteLogPageEntry( auxTrans, pt2Nil_egg00, ( *iter ).ReadLastWriterIOSequenceNo(),
                                         cluster.GetBlockAddress( index ), traceNo );
            }
        }
        pagesToWrite -= cluster.GetBlockCount();
        startOffset  += cluster.GetBlockCount();
        bd73LogIOStatistic( ! IOMAN_READ_IO );
    }
    while( 0 < pagesToWrite );
}

/*---------------------------------------------------------------------------*/

void
IOMan_Manager::WriteRestartPage(
    tsp00_TaskId        TaskId,
    Kernel_RestartPage  &Page )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteRestartPage", IOMan_Trace, 5 );

    IOMan_BlockAddress  Block = m_ReservedBlocks.GetRestartPageAddress();

    m_DataControl.Get( Block.GetDeviceNo() ).WritePage( TaskId, Page, Block );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext AuxTrans = BuildAuxTrans( TaskId );
        Trace_WriteDataPageEntry( AuxTrans, Page.ReadPageType2(), Page.GetPageID(), Block );
    }
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::EnableBadDevSpace(
    const tsp00_TaskId  taskId,
    tsp00_VFilename     &devName )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::EnableBadDevSpace", IOMan_Trace, 5 );

    IOMan_DeviceNo  searchedDevNo;

    if( GetDataDevSpaceNo( devName, searchedDevNo ))
    {
        return( IOMan_Okay == m_DataControl.Get( searchedDevNo ).UnSetBad( taskId ));
    }
    else if( GetLogDevSpaceNo( devName, searchedDevNo ))
    {
        return( IOMan_Okay == m_LogControl.Get( searchedDevNo ).UnSetBad( taskId ));
    }
    else
        return( false );
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::ReadPageDirectFromDataDevSpace(
    const tsp00_TaskId          taskId,
    const IOMan_BlockAddress    &block,
    const tsp00_PageAddr        pPage )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadPageDirectFromDataDevSpace", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "BlockAddress: " <<
                        block.GetDeviceNo() << "/" << block.GetBlockNo());

    return( m_DataControl.Get( block.GetDeviceNo() ).ReadSimple(
                taskId, pPage, block.GetBlockNo() ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::WritePageDirectToDataDevSpace(
    const tsp00_TaskId          taskId,
    const IOMan_BlockAddress    &block,
    const tsp00_PageAddr        pPage )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WritePageDirectToDataDevSpace", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "BlockAddress: " <<
                        block.GetDeviceNo() << "/" << block.GetBlockNo());

    return( m_DataControl.Get( block.GetDeviceNo() ).WriteSimple(
                taskId, pPage, block.GetBlockNo() ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::WritePageDirectToArchiveLog(
    const tsp00_TaskId          taskId,
    const IOMan_BlockAddress    &block,
    const tsp00_PageAddr        pPage )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WritePageDirectToArchiveLog", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "BlockAddress: " <<
                        block.GetDeviceNo() << "/" << block.GetBlockNo());

    return( m_LogControl.Get( block.GetDeviceNo() ).WriteSimple(
                taskId, pPage, block.GetBlockNo() ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::ReadPageDirectFromArchiveLog(
    const tsp00_TaskId          taskId,
    const IOMan_BlockAddress    &block,
    const tsp00_PageAddr        pPage )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadPageDirectFromArchiveLog", IOMan_Trace, 5 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 5, "BlockAddress: " <<
                        block.GetDeviceNo() << "/" << block.GetBlockNo());

    return( m_LogControl.Get( block.GetDeviceNo() ).ReadSimple(
                taskId, pPage, block.GetBlockNo() ));
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::ReadDataPagesForSave(
    const tsp00_TaskId          taskId,
    const SAPDB_Int4            asynIOHandle,
    IOMan_DataPages             &pages,
    const IOMan_ClusterAddress  &cluster )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::ReadDataPagesForSave", IOMan_Trace, 5 );

    SAPDBERR_ASSERT_STATE( pages.GetPageCount() <= SAPDB_UInt( cluster.GetBlockCount()) );

    //FBM_Manager::Instance().GetNextBlocksForBackUp (Trans.trTaskId_gg00, DevNo,
    //  MaxNumBlocksWanted, SuppliedNumBlocks, BlockNo);      TODO

    bool retCode = m_DataControl.Get( cluster.GetDeviceNo() ).AsynReadWrite(
                       taskId, asynIOHandle, pages, cluster );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext                  auxTrans = BuildAuxTrans( taskId );
        SAPDB_UInt                          traceNo  = ( pages.GetPageCount() > 1 ) ? 1:0;
        IOMan_DataPagesIterator             iter( pages );
        IOMan_ClusterAddress::BlockIndex    index = 0;

        for( iter.Begin(); !iter.End(); ++iter, ++index, ++traceNo)
            Trace_ReadDataPageEntry( auxTrans, ( *iter ).ReadPageType2(), ( *iter ).PageNo(),
                                     cluster.GetBlockAddress( index ), traceNo );
    }
    return( retCode );
}

/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::WriteDataPagesForRestore(
    const tsp00_TaskId          taskId,
    const SAPDB_Int4            asynIOHandle,
    IOMan_DataPages             &pages,
    const IOMan_ClusterAddress  &cluster )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_Manager::WriteDataPagesForRestore", IOMan_Trace, 5 );

    SAPDBERR_ASSERT_STATE( pages.GetPageCount() <= SAPDB_UInt( cluster.GetBlockCount()) );

    // set the blocks of the given cluster to occupied
    // Converter_IBackUp::Instance().SetBlockAddressOfRestoredPage (
    //                                taskId, BlockAddress, PageNo, AddressingMode); TODO

    bool retCode = m_DataControl.Get( cluster.GetDeviceNo() ).AsynReadWrite(
                       taskId, asynIOHandle, pages, cluster );

    if( g01vtrace.vtrAll_gg00 || g01vtrace.vtrIoTrace_gg00 )
    {
        tgg00_TransContext                  auxTrans = BuildAuxTrans( taskId );
        SAPDB_UInt                          traceNo  = ( pages.GetPageCount() > 1 ) ? 1:0;
        IOMan_DataPagesIterator             iter( pages );
        IOMan_ClusterAddress::BlockIndex    index = 0;

        for( iter.Begin(); !iter.End(); ++iter, ++index, ++traceNo)
            Trace_WriteDataPageEntry( auxTrans, ( *iter ).ReadPageType2(), ( *iter ).PageNo(),
                                      cluster.GetBlockAddress( index ), traceNo );
    }
    return( retCode );
}

/*---------------------------------------------------------------------------*/


void
IOMan_Manager::MigrateDataBase(
    const tsp00_TaskId  taskId,
    tgg00_BasisError    &trError )
{
    if( IOMan_Migration::IsMigrationNeeded() )
    {
        IOMan_Migration migManager( taskId, m_PageAllocator, m_Allocator );

        if( ! migManager.Initialize( m_DataControl.MaxDevSpaces() ))
        {
            trError = e_no_more_memory;
        }
        if( ! migManager.ExecuteMigration( m_DataControl, m_LogControl ))
        {
            // only one error could occur.
            trError = e_not_prepared_for_migration;
        }
    }
}

/*---------------------------------------------------------------------------*/
/*----------------------------- Private Methods -----------------------------*/
/*---------------------------------------------------------------------------*/

bool
IOMan_Manager::RestartAndInitializeFBM( const tsp00_TaskId   taskId )
{
    if(! FBM_Manager::Instance().Restart( taskId, m_DataControl.MaxDevSpaces(), m_ReservedBlocks ))
        return( false );

    for( IOMan_DeviceNo devNo = 1; devNo <= m_DataControl.UsedDevSpaces(); ++devNo )
    {
        if(! FBM_Manager::Instance().AddVolume( taskId, devNo,
                m_DataControl.Get( devNo ).GetCapacity( false )))
            return( false );
    }
    return( true );
}

/*---------------------------------------------------------------------------*/

tgg00_TransContext
IOMan_Manager::BuildAuxTrans( tsp00_TaskId TaskId ) const
{
    tgg00_TransContext  Trans;

    Trans.trTaskId_gg00 = TaskId;
    Trans.trSessionId_gg00.becomes( cgg_nil_session );
    Trans.trTransId_gg00.becomes( cgg_nil_trans );
    Trans.trWriteTransId_gg00.becomes( cgg_nil_trans );
    Trans.trSubtransId_gg00 = cgg_zero_subtrans;

    return( Trans );
}

/*---------------------------------------------------------------------------*/
/*-------------- Implementation of Interface Instance methods ---------------*/
/*---------------------------------------------------------------------------*/

IOMan_IDataInfo&
IOMan_IDataInfo::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_IDataIO&
IOMan_IDataIO::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_IDataManager&
IOMan_IDataManager::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_IDiagnose&
IOMan_IDiagnose::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_ILogInfo&
IOMan_ILogInfo::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_ILogIO&
IOMan_ILogIO::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_ILogManager&
IOMan_ILogManager::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_IConverterIO&
IOMan_IConverterIO::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*---------------------------------------------------------------------------*/

IOMan_IMigration&
IOMan_IMigration::GetInstance()
{
    return( IOMan_Manager::GetInstance() );
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
