/*!**************************************************************************

  module      : IOMan_Manager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : This module is used to handle all devspaces of the kernel.

                The IOMan_Manager class offers the possibility to manage 
                data and log devspaces. This contains the facility to create,
                open, close, read and write from and to devspaces.

  last changed: 2001-01-12  13:18
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_MANAGER_HPP
#define IOMAN_MANAGER_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "ggg00.h"   // PASCAL: GG_basic_constants_and_tyype

// Interfaces
#include "IOManager/IOMan_IConverterIO.hpp"
#include "IOManager/IOMan_IDataInfo.hpp"
#include "IOManager/IOMan_IDataIO.hpp"
#include "IOManager/IOMan_IDataManager.hpp"
#include "IOManager/IOMan_IDiagnose.hpp"
#include "IOManager/IOMan_ILogInfo.hpp"
#include "IOManager/IOMan_ILogManager.hpp"
#include "IOManager/IOMan_ILogIO.hpp"
#include "IOManager/IOMan_IMigration.hpp"

#include "Converter/Converter_Page.hpp"
#include "DataAccess/Data_BasePage.hpp"
#include "DataAccess/Data_Types.hpp"
#include "FreeBlockManagement/FBM_Manager.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_DataControl.hpp"
#include "IOManager/IOMan_ConverterPages.hpp"
#include "IOManager/IOMan_DataPages.hpp"
#include "IOManager/IOMan_LogControl.hpp"
#include "IOManager/IOMan_LogPages.hpp"
#include "IOManager/IOMan_PageFrameAllocator.hpp"
#include "IOManager/IOMan_ReservedBlockAddress.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "KernelCommon/Kernel_RestartPage.hpp"
#include "Logging/Log_ClusterAddress.hpp"
#include "Logging/Log_InfoPage.hpp"
#include "Logging/Log_Page.hpp"
#include "RunTime/MemoryManagement/RTEMem_AllocatorWrapper.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_Manager

   description: This class is used to handle all devspaces of the kernel.
   
              
*******************************************************************************/

class IOMan_Manager : public IOMan_IConverterIO,
            public IOMan_IDataInfo,
            public IOMan_IDataIO,
            public IOMan_IDataManager,
            public IOMan_IDiagnose,
            public IOMan_ILogInfo,
            public IOMan_ILogIO,
            public IOMan_ILogManager,
            public IOMan_IMigration
{
    friend class IOMan_IConverterIO;
    friend class IOMan_IDataInfo;
    friend class IOMan_IDataIO;
    friend class IOMan_IDataManager;
    friend class IOMan_IDiagnose;
    friend class IOMan_ILogInfo;
    friend class IOMan_ILogIO;
    friend class IOMan_ILogManager;
    friend class IOMan_IMigration;

public:

    /*--------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the component IOManager.
                  It is assumed that the IOManager singleton is available! 
                  No CreateInstance call will be done.
    arguments:    none
    return value: Access to the IOMan_Manager singleton
    --------------------------------------------------------------------------*/

    static IOMan_Manager& GetInstance(){ return( *m_Instance ); }

    /*!-------------------------------------------------------------------------
    function:     CreateInstance
    description:  This method creates the singleton IOMananger.It is assumed 
                  that this method will be callled only one time by the timeout 
                  task and therefore no synchonization mechanism is used.
    arguments:    taskId [in] Identification of the calling task.
    return value: true means instance could be created; else false
    --------------------------------------------------------------------------*/

    static bool CreateInstance( const tsp00_TaskId  taskId );

    /*!-------------------------------------------------------------------------
    function:     AddDataDevSpace()
    description:  This method is used to add a new data devspace to the data
                  base devspace set in warm as in cold mode too.
    arguments:    taskId  [in] Identification of the calling task.
                  bIsCold [in] Is data base state db_cold (opposite db_warm)
                  devName [in] Name of the new devspace
                  devSize [in] Size of the new devsppace in pages
                  devNo   [in] Logical indentifier of the devspace
    return value: true means that add data devspace was successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddDataDevSpace(
        const tsp00_TaskId      taskId,
        const SAPDB_Bool        bIsCold,
        tsp00_VFilename         &devName,
        const IOMan_BlockCount  devSize,
        const IOMan_DeviceNo    devNo );

    /*!-------------------------------------------------------------------------
    function:     AddLogDevSpace()
    description:  This method is used to add a new log devspace to the given
                  archive log in warm as in cold mode too. The proper
                  activation of the new devspace must be done by the Log_IOManager.
    arguments:    taskId      [in]  Identification of the calling task.
                  bIsCold     [in]  Is data base state db_cold (opposite db_warm)
                  devName     [in]  Name of the new devspace
                  mirrDevName [in]  Name of the new mirrored devspace (log mode dual only)
                  devSize     [in]  Size of the new devsppace in pages
                  devNo       [in]  Logical indentifier of the devspace
                  page        [in]  (nil)-Log pages used for devspace formating
                  newLogSize  [out] New size in pages of the archive log
    return value: true means that add log devspace was successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddLogDevSpace(
        const tsp00_TaskId          taskId,
        const SAPDB_Bool            bIsCold,
        tsp00_VFilename             &devName,
        tsp00_VFilename             &mirrDevName,
        const IOMan_BlockCount      devSize,
        const IOMan_DeviceNo        devNo,
        IOMan_LogPages              &pages,
        IOMan_BlockCount            &newLogSize );

    /*!-------------------------------------------------------------------------
    function:     AddDataDevSpacePossible()
    description:  This method indicates whether a add data devspace is possible.
    arguments:    none
    return value: true means that add data devspace is possible
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool AddDataDevSpacePossible() const
    {
        return( m_DataControl.ConfiguredDevSpaces() <  m_DataControl.MaxDevSpaces());
    }

    /*!-------------------------------------------------------------------------
    function:     CloseAllLogDevSpaces
    description:  This method is used to bring all archive logs being in mode
                  open in mode close.
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void CloseAllLogDevSpaces( const tsp00_TaskId  taskId );

    /*!-------------------------------------------------------------------------
    function:     CloseAllDataDevSpaces
    description:  This method is used to bring all data devspaces being in mode
                  open in mode close.
    arguments:    TaskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void CloseAllDataDevSpaces( const tsp00_TaskId  taskId );

    /*!-------------------------------------------------------------------------
    function:     ConfiguredDataDevSpaces()
    description:  This method provides the number of configured data devspaces.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int ConfiguredDataDevSpaces() const
    {
        return( m_DataControl.ConfiguredDevSpaces() );
    }

    /*!-------------------------------------------------------------------------
    function:     ConfiguredLogDevSpaces()
    description:  This method provides the number of configured log devspaces
                  of all archive logs.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int ConfiguredLogDevSpaces() const
    {
        return( m_LogControl.ConfiguredDevSpaces() );
    }

    /*!-------------------------------------------------------------------------
    function:     IsLogMirrored()
    description:  This method provides whether the log area consists of mirrored
                  log volumes..
    arguments:    none
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool IsLogMirrored() const 
    {
        return( m_LogControl.IsLogMirrored() );
    }

    /*!-------------------------------------------------------------------------
    function:     GetDataDevSpaceNo()
    description:  This method returns the logical identifier of for the given
                  data devspace name. If the devspace name doesn't correspond to
                  a valid devspace the return value is false.
    arguments:    devName [in]  Name of a data devspace
                  devNo   [out] Logical devspace identifier
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetDataDevSpaceNo(
        tsp00_VFilename   &devName,
        IOMan_DeviceNo    &devNo ) const
    {
        return( m_DataControl.GetDevSpaceNo( devName, devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetDataDevName()
    description:  This method returns the name of the data devspace identified
                  by the given logical identifier.
    arguments:    devNo   [in ] Logical devspace identifier
                  devName [out] Name of a data devspace
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetDataDevName(
        IOMan_DeviceNo    devNo,
        tsp00_VFilename   &devName ) const
    {
        return( m_DataControl.Get( devNo ).GetName( devName ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetBadLogVolume()
    description:  This method is used to determine the name of a log volume
                  which is marked as bad resp. not accessable, because it's
                  removed. This method is only available in mode admin and
                  only if log mode dual is configured.(GET BAD LOG VOLUME)
    arguments:    taskId     [in]  Identification of the calling task.
                  badDevName [out] Name of the bad log volume
    return value: true means bad log volme found
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool GetBadLogVolume(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &badDevName )
    {
        return( m_LogControl.GetBadLogVolume( taskId, badDevName ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetLogDevSpaceNo()
    description:  This method returns the logical identifier of for the given
                  log devspace name. If the devspace name doesn't correspond to
                  a valid devspace the return value is false.
    arguments:    devName [in]  Name of a data devspace
                  devNo   [out] Logical devspace identifier
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetLogDevSpaceNo(
        tsp00_VFilename     &devName,
        IOMan_DeviceNo      &devNo )
    {
        return( m_LogControl.GetDevSpaceNo( devName, devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     GetLogDevName()
    description:  This method returns the name of the log devspace identified
                  by the given logical identifier.
    arguments:    devNo   [in ] Logical devspace identifier
                  devName [out] Name of a data devspace
    return value: true if devspace name was found
    --------------------------------------------------------------------------*/

    virtual bool GetLogDevName(
        const IOMan_DeviceNo        devNo,
        tsp00_VFilename             &devName ) const
    {
        return( m_LogControl.Get( devNo ).GetName( devName ));
    }

    /*!-------------------------------------------------------------------------
    function:     CreateAllLogDevSpaces
    description:  This method is used to create, format and open all archive
                  logs specified within the data base configuration file.
    arguments:    taskId       [in] Identification of the calling task.
                  pDBIdentifier [in] Unique data base identifier
    return value: true means successful creation; else false
    --------------------------------------------------------------------------*/

    virtual bool CreateAllLogDevSpaces(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier );

    /*!-------------------------------------------------------------------------
    function:     CreateAllDataDevSpaces
    description:  This method is used to create, format and open all data 
                  devspaces specified within the data base configuration file.
                  Additional it is checked whether the sum of all data pages
                  within all data devspaces fulfils the minimal space requirements 
                  of  a sapdb data base. After checking the data base size
                  the component FBM is generated and initialized. Any kind
                  of an error will be returned as an basis error within the
                  given trans context.
    arguments:    taskId        [in] Identification of the calling task.
                  pDBIdentifier [in] Unique data base identifier
    return value: true means successful creation; else false
    --------------------------------------------------------------------------*/

    virtual bool CreateAllDataDevSpaces(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier );

    /*!-------------------------------------------------------------------------
    function:     GetDataDevSpaceSize()
    description:  This method provides the configured size in pages of the 
                  specified data devspace. It is assumed that the devspace 
                  identifier is a valid.
    arguments:    DevNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetDataDevSpaceSize( const IOMan_DeviceNo devNo ) const;

    /*!-------------------------------------------------------------------------
    function:     GetDataDevSpaceUsableSize()
    description:  This method provides the usable size in pages of the 
                  specified data devspace. It is assumed that the devspace 
                  identifier is valid.
    arguments:    DevNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetDataDevSpaceUsableSize( const IOMan_DeviceNo devNo ) const;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevSpaceSize()
    description:  This method provides the configured size in pages of the 
                  specified log devspace. It is assumed that the devspace 
                  identifier is valid.
    arguments:    devNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetLogDevSpaceSize( const IOMan_DeviceNo devNo ) const;

    /*!-------------------------------------------------------------------------
    function:     GetLogDevSpaceSize()
    description:  This method provides the usable size in pages of the 
                  specified log devspace. It is assumed that the devspace 
                  identifier is valid.
    arguments:    devNo [in] Logical devspace identifier
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    virtual IOMan_BlockCount GetLogDevSpaceUsableSize( const IOMan_DeviceNo devNo ) const;

    /*!-------------------------------------------------------------------------
    function:     OpenAllArchiveLogs()
    description:  This method is used to bring all archive logs specified
                  within the data base configuration file from mode close into
                  mode open. It is assumed that all archive logs are generated
                  with the method CreateAllArchiveLogs. 
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    virtual bool OpenAllArchiveLogs( const tsp00_TaskId taskId );

    /*!-------------------------------------------------------------------------
    function:     OpenAllDataDevSpaces()
    description:  This method is used to bring all data devspaces specified
                  within the data base configuration file from mode close into
                  mode open. It is assumed that all data devspaces are generated
                  with the method CreateAllDataDevSpaces. Additional it is checked
                  whether the sum of the pages within all attached data devspaces
                  is equal to the sum stored within the restart record. This is 
                  done to validate if the data base configuration was changed in a
                  not offical way. If no errors occur the component FBM is 
                  generated and restarted.
    arguments:    taskId [in] Identification of the calling task.
    return value: true means succeful execution; else false
    --------------------------------------------------------------------------*/

    virtual bool OpenAllDataDevSpaces( const tsp00_TaskId   taskId );

    /*!-------------------------------------------------------------------------
    function:     OpenOneDataDevSpace
    description:  This method is used to bring one data devspave identified by
                  the given DevNo from mode closed into mode open. If the data
                  devspace is not stored in the data base configuration file or 
                  the open method fails an error is returned.
    arguments:    taskId [in] Identification of the calling task.
                  devNo  [in] Logical devspace identifier 
    return value: true means succeful execution; else false
    --------------------------------------------------------------------------*/

    virtual bool OpenOneDataDevSpace(
        const tsp00_TaskId    taskId,
        const IOMan_DeviceNo  devNo );

    /*!-------------------------------------------------------------------------
    function:     OpenOneLogDevSpace
    description:  This method is used to bring one log devspave identified by
                  the given DevNo from mode closed into mode open. If the log
                  devspace is not stored in the data base configuration file or 
                  the open method fails an error is returned.
    arguments:    taskId [in] Identification of the calling task.
                  devNo  [in] Logical devspace identifier 
    return value: true means succeful execution; else false
    --------------------------------------------------------------------------*/

    virtual bool OpenOneLogDevSpace(
        const tsp00_TaskId    taskId,
        const IOMan_DeviceNo  devNo );

    /*!-------------------------------------------------------------------------
    function:     ReadConverterPage()
    description:  This method is used to read one converter page synchronous from 
                  a data devspace into the given memory. The caller is blocked as 
                  long as the read io is not completed. The position of the page 
                  is given. The correctness of the read data page will be checked. 
                  In case of errors detected by the check method of the 
                  page the read io will be repeated. If the error is still present 
                  after three read operations, the provided buffer is dumped into 
                  the rundirectory of the kernel. Each failed read operation is 
                  logged within the knldiag and knldiag error file. In any case of 
                  an error an emergency shutdown is executed.
    arguments:    taskId  [in] Identification of the calling task.
                  block   [in] Position of the wanted page at the data devspace
                  page    [in] Handler for the wanted page.
    return value: none. 
    --------------------------------------------------------------------------*/

    virtual void ReadConverterPage(
        tsp00_TaskId                taskId,
        const IOMan_BlockAddress    &block,
        Converter_Page              &page );

    /*!-------------------------------------------------------------------------
    function:     ReadDataPage()
    description:  This method is used to read one data page synchronous from a data 
                  devspace into the given memory. The caller is blocked as long as
                  the read io is not completed The position of the page will be determined
                  by the converter. The correctness of the read data page will be
                  checked. In case of errors detected by the check method of the 
                  page the read io will be repeated. If the error is still present 
                  after three read operations, the provided buffer is dumped into 
                  the rundirectory of the kernel. Each failed read operation is 
                  logged within the knldiag and knldiag error file. In case of 
                  fatal errors of the IO subsystem an emergency shutdown is executed.
    arguments:    taskId   [in]  Identification of the calling task.
                  trError  [out] Current error state
                  page     [in]  Handler for the wanted page.
                  pageNo   [in]  Id of the wanted page
    return value: true means page succesful read; else false. 
                  In case of a bad page the error is set to e_bad_datapage.
    --------------------------------------------------------------------------*/

    virtual bool ReadDataPage(
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        Data_BasePage               &Page,
        const Data_PageNo           &PageNo );

    /*!-------------------------------------------------------------------------
    function:     ReadDataPagesForSave()
    description:  This method is used to read data pages for a save data via 
                  a separate io thread insulated from the common kernel io threads.
                  The correctness  of the page won't be  checked. The caller is 
                  blocked as long as the write io is not completed.
    arguments:    taskId       [in] Identification of the calling task.
                  asynIOHandle [in] Handle for the asynchronous io
                  pages        [in] Handler of the pages to be written
                  cluster      [in] Destination address of the pages to be written
    return value: true means write was successfull
    --------------------------------------------------------------------------*/

    virtual bool ReadDataPagesForSave(
        const tsp00_TaskId          taskId,
        const SAPDB_Int4            asynIOHandle,
        IOMan_DataPages             &pages,
        const IOMan_ClusterAddress  &cluster );

    /*!-------------------------------------------------------------------------
    function:     ReadDataPageDirect()
    description:  This method is used to read one data page synchronous from a data 
                  devspace into the given memory. The caller is blocked as long as
                  the read io is not completed. The position of the wanted page is
                  given as a block address and therefore not searched in the component
                  converter is neccessary. The validity of the given block address 
                  is check on an abstract level, i.e. logical device number and block
                  number are in a valid range. The correctness of the read data page 
                  will be checked. In case of errors detected by the check method of 
                  the page the read io will be repeated. If the error is still present 
                  after three read operations, the provided buffer is dumped into 
                  the rundirectory of the kernel. Each failed read operation is 
                  logged within the knldiag and knldiag error file. In case of 
                  fatal errors of the IO subsystem an emergency shutdown is executed.
    arguments:    TaskId  [in] Identification of the calling task.
                  Page  [in] Handler for the wanted page.
                  Block [in] Position of the wanted page at the data devspace
    return value: true means page succesful read; else false. 
                  In case of a bad page the error is set to e_bad_datapage.
    --------------------------------------------------------------------------*/

    virtual bool ReadDataPageDirect(
        tsp00_TaskId                TaskId,
        Data_BasePage               &Page,
        const IOMan_BlockAddress    &Block );

    /*!-------------------------------------------------------------------------
    function:     ReadLogInfoPage()
    description:  This method is used to read the log info page synchronous from 
                  a given position of the archive log. The caller is blocked as 
                  long as the read io is not completed. The correctness of the read 
                  log info page will be checked. In case of errors detected by the 
                  check method of the page the read io will be repeated. If the 
                  error is still present after three read operations, the provided 
                  buffer is dumped into the rundirectory of the kernel. Each failed 
                  read operation is logged within the knldiag and knldiag error file. 
                  In any case of errors an emergency shutdown is executed and the
                  volume is set to bad.
    arguments:    TaskId  [in] Identification of the calling task.
                  Page    [in] Handler for the wanted page.
                  Address [in] Position of the wanted page within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadLogInfoPage(
        tsp00_TaskId              TaskId,
        Log_InfoPage              &Page,
        const Log_ClusterAddress  &Address );

    /*!-------------------------------------------------------------------------
    function:     ReadLogPage()
    description:  This method is used to read a log page synchronous from the
                  archive log The caller is blocked as long as the read io is 
                  not completed. The position of the wanted log page is given 
                  as a log address  which will be transformed into a devspace 
                  specific block address. The correctness of the read log info 
                  page will be checked. In case of errors detected by the check 
                  method of the page the read io will be repeated. If the error 
                  is still present  after three read operations, the provided 
                  buffer is dumped into the rundirectory of the kernel. Each 
                  failed read operation is logged within the knldiag and knldiag 
                  error file. In any case of errors an emergency shutdown is 
                  executed and the volume is set to bad.
    arguments:    TaskId  [in] Identification of the calling task.
                  Page    [in] Handler for the wanted page.
                  Address [in] Position of the wanted page within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadLogPage(
        tsp00_TaskId              TaskId,
        Log_Page                  &Page,
        const Log_ClusterAddress  &Address );

    /*!-------------------------------------------------------------------------
    function:     ReadLogPages()
    description:  This method is used to read a couple of log pages synchronous 
                  from the archive log.The caller is blocked as long as the read 
                  io is  not completed. The position of the wanted log page is given 
                  as a log address  which will be transformed into a devspace 
                  specific block address. The correctness of the read log info 
                  page will be checked. In case of errors detected by the check 
                  method of the page the read io will be repeated. If the error 
                  is still present  after three read operations, the provided 
                  buffer is dumped into the rundirectory of the kernel. Each 
                  failed read operation is logged within the knldiag and knldiag 
                  error file. In any case of errors an emergency shutdown is 
                  executed and the volume is set to bad.
                  NOTE: that it is assumed that the given page frames(!) contained
                  in the page handler are in strictly ascending order within the 
                  memory!
    arguments:    taskId  [in] Identification of the calling task.
                  page    [in] Handler for the wanted pages.
                  address [in] Position of the wanted pages within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadLogPages(
        const tsp00_TaskId          taskId,
        IOMan_LogPages              &pages,
        const Log_ClusterAddress    &address );

    /*!-------------------------------------------------------------------------
    function:     ReadRestartPage()
    description:  This method is used to read the restart page synchronous from 
                  a defined position of an data devspace. The caller is blocked 
                  as long as the read io  is not completed. The correctness of 
                  the read restart page will be checked. In case of errors detected 
                  by the check method of the page the read io will be repeated. 
                  If the error is still present  after three read operations, 
                  the provided buffer is dumped into the rundirectory of the kernel. 
                  Each failed read operation is logged within the knldiag and 
                  knldiag error file. In any case of errors an emergency shutdown 
                  is executed.
    arguments:    TaskId [in] Identification of the calling task.
                  Page   [in] Handler for the wanted page.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void ReadRestartPage(
        tsp00_TaskId        TaskId,
        Kernel_RestartPage  &Page );

    /*!-------------------------------------------------------------------------
    function:     TotalUsableDataPages()
    description:  This method returns the sum of all pages storable within all 
                  configured data devspaces. This value doesn't include the pages
                  needed for administrativ purpose, i.e IOMan_InfoPage.
                  This value is only available if all
                  data devspaces are in mode open.
    arguments:    none
    return value: SAPDB_Int4 
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4 TotalUsableDataPages() const
    {
        return( m_DataControl.TotalUsableDataPages() );
    }

    /*!-------------------------------------------------------------------------
    function:     TotalDataPages()
    description:  This method returns the sum of all pages storable within all 
                  configured data devspaces. This value includes all pages of
                  all data devspaces without taking amount of pages reserved 
                  for administrative purposes. This value is only available if 
                  all data devspaces are in mode open.
    arguments:    none
    return value: SAPDB_Int4 
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4 TotalDataPages() const
    {
        return( m_DataControl.TotalDataPages() );
    }

    /*!-------------------------------------------------------------------------
    function:     TotalUsableLogPages()
    description:  This method returns the sum of all pages storable within all 
                  configured archive logs. This value is only available if all 
                  archive logs are in mode open.
    arguments:    none
    return value: (SAPDB_Int4) 
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4  TotalUsableLogPages() const
    {
        return( m_LogControl.TotalUsableLogPages() );
    }

    /*!-------------------------------------------------------------------------
    function:     UsedDataDevSpaces()
    description:  This method provides the number of active data devspaces. 
                  If the I/O management isn't restarted the initial value of 
                  one is returned.
    arguments:    none
    return value: SAPDB_Int
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int UsedDataDevSpaces() const
    {
        return( m_DataControl.UsedDevSpaces() );
    }

    /*!-------------------------------------------------------------------------
    function:     WriteConverterPage()
    description:  This method is used to write one converter page on one data 
                  devspace. The caller is blocked as long as the write io is not 
                  completed. The destination of the converter page will be requested
                  from the componenet FBM and the old block address is released 
                  similar to the block addresses of data pages. Before the proper 
                  write is executed the correctness of the given pages will be 
                  checked. In case of an error the corrupted page is dumped into 
                  the kernel rundirectory and an emergency shutdown is executed.
                  In the other case the converter page is prepared for the io, i.e. 
                  the the checksum is calculated. 
    arguments:    taskId  [in]  Identification of the calling task.
                  page    [in]  Handler of the given page
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteConverterPage(
        const tsp00_TaskId  taskId,
        Converter_Page      &page );

    /*!-------------------------------------------------------------------------
    function:     WriteDataPage()
    description:  This method is used to write one data page on one data devspace.
                  For more details see WriteDataPages! The difference to the
                  WriteDataPages call is, that the caller hasn't reserved a block 
                  address a prior. Therefore in the following first the componente 
                  converter specialy free block management has to contacted to get 
                  a block address. Note that this will increase the time the data 
                  page is locked in the data cache!
    arguments:    taskId  [in]  Identification of the calling task.
                  trError [out] Current error state
                  page    [in]  Handler of the given pages
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteDataPage(
        const tsp00_TaskId   taskId,
        tgg00_BasisError    &trError,        
        Data_BasePage       &page );

    /*!-------------------------------------------------------------------------
    function:     WriteDataPages()
    description:  This method is used to write a number of data pages synchronous
                  into one data devspace. The caller is blocked as long as the write 
                  io is not completed. The maximum number of writable data pages is
                  limited by a database configuration parameter. Additional to the
                  writeable data pages the caller has to provide a cluster address,
                  which specifies the write destination on a data devspace. Before 
                  the proper write is executed the correctness of the given pages 
                  will be checked. In case of an error the corrupted page is dumped 
                  into the kernel rundirectory and an emergency shutdown is executed.
                  In the other case the data page is prepared for the io, i.e. the 
                  converter version is updated and the checksum is calculated. 
                  Additional the corresponding converter entry is updated and the proper
                  write operation is executed. Note that the writeable data pages not 
                  necessarily stored in a contiguous memory block, because this method
                  uses scatter/gather io. The caller has to insure that the given 
                  cluster has sufficient capacity to store all given data pages within
                  the data devspace. Pages without an corresponding block address won't
                  be written.
    arguments:    taskId  [in]  Identification of the calling task.
                  trError [out] Current error state
                  pages   [in] Handler of the given pages
                  cluster [in] Write destination of the data pages in the data devspace.
                               The caller has to insure that the given block addresses 
                               are available.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteDataPages(
        const tsp00_TaskId          taskId,
        tgg00_BasisError            &trError,
        IOMan_DataPages             &pages,
        const IOMan_ClusterAddress  &cluster );

    /*!-------------------------------------------------------------------------
    function:     WriteDataPagesForRestore()
    description:  This method is used to write data pages during a restore via 
                  a separate io thread insulated from the common kernel io threads.
                  The correctness  of the page won't be  checked. The caller is 
                  blocked as long as the write io is not completed.
    arguments:    taskId       [in] Identification of the calling task.
                  asynIOHandle [in] Handle for the asynchronous io
                  pages        [in] Handler of the pages to be written
                  cluster      [in] Destination address of the pages to be written
    return value: true means write was successfull
    --------------------------------------------------------------------------*/

    virtual bool WriteDataPagesForRestore(
        const tsp00_TaskId          taskId,
        const SAPDB_Int4            asynIOHandle,
        IOMan_DataPages             &pages,
        const IOMan_ClusterAddress  &cluster );

    /*!-------------------------------------------------------------------------
    function:     WriteLogInfoPage()
    description:  This method is used to write the log info page synchronous onto 
                  a given position of the archive log. Before the proper write is 
                  executed the correctness of the given page will be checked. In 
                  case of an error the corrupted page is dumped into the kernel 
                  rundirectory and an emergency shutdown is executed. In the other 
                  case the log info page is prepared for the io, for example the 
                  checksum is calculated. In any case of errors an emergency 
                  shutdown is executed. 
    arguments:    TaskId  [in] Identification of the calling task.
                  Page    [in] Handler of the given pages
                  Address [in] Position of the given page within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteLogInfoPage(
        tsp00_TaskId              TaskId,
        Log_InfoPage              &Page,
        const Log_ClusterAddress  &Address );

    /*!-------------------------------------------------------------------------
    function:     WriteLogPages()
    description:  This method is used to write a number of log pages synchronous
                  into one archive log. The caller is blocked as long as the write 
                  io is not completed. The maximum number of writable log pages is
                  limited by a database configuration parameter. Additional to the
                  writeable log pages the caller has to provide a log address, which 
                  specifies the write destination on an archive log. The given log
                  address will be transformed into a log devspace specific cluster 
                  address. Perhaps the I/O Management splits the order into separate 
                  pieces conditional on the number of log devspaces belonging to one 
                  archive log. Before the proper write is executed the correctness of 
                  the given pages will be checked. In case of an error the corrupted 
                  page is dumped into the kernel rundirectory and an emergency shutdown 
                  is executed.In the other case the page is prepared for the io, for
                  example the checksum is calculated. Note that the writeable log pages 
                  not necessarily stored in a contiguous memory block, because this 
                  method uses scatter/gather io.In any case of errors an emergency 
                  shutdown is executed. 
    arguments:    TaskId  [in] Identification of the calling task.
                  Pages   [in] Handler of the given pages
                  Address [in] Position for the given pages within an given archive log
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteLogPages(
        tsp00_TaskId              TaskId,
        IOMan_LogPages            &Pages,
        const Log_ClusterAddress  &Address );

    /*!-------------------------------------------------------------------------
    function:     WriteRestartPage()
    description:  This method is used to write the restart page synchronous onto 
                  a fixed position of a data devspace Before the proper write is 
                  executed the correctness of the  given page will be checked. In 
                  case of an error the corrupted page is dumped into the kernel 
                  rundirectory and an emergency shutdown is executed. In the other 
                  case the log info page is prepared for the io, for example the 
                  checksum is calculated. In any case of errors an emergency 
                  shutdown is executed. 
    arguments:    TaskId [in] Identification of the calling task.
                  Page   [in] Handler of the given pages
    return value: none
    --------------------------------------------------------------------------*/

    virtual void WriteRestartPage(
        tsp00_TaskId          TaskId,
        Kernel_RestartPage    &Page );

    /*!-------------------------------------------------------------------------
    function:     EnableBadDevSpace()
    description:  This method is used to set a devspace from bad to okay.
    arguments:    taskId  [in] Identification of the calling task.
                  devName [in] Name of the devspace to be enabled
    return value: true means devspace was successful enabled
    --------------------------------------------------------------------------*/

    virtual bool EnableBadDevSpace(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &devName );

    /*!-------------------------------------------------------------------------
    function:     ReadPageDirectFromDataDevSpace()
    description:  This method is used to read a page from the given block address.
                  Note that NO consistent check is made.
    arguments:    taskId  [in] Identification of the calling task.
                  block   [in] Destination position for the given page frame
                  pPage   [in] Pointer to the page frame to be written
    return value: true means write was successful; else false
    --------------------------------------------------------------------------*/

    virtual bool ReadPageDirectFromDataDevSpace(
        const tsp00_TaskId          taskId,
        const IOMan_BlockAddress    &block,
        const tsp00_PageAddr        pPage );

    /*!-------------------------------------------------------------------------
    function:     WritePageDirectToDataDevSpace()
    description:  This method is used to write a given page into the given
                  data devspace. Note that NO consistent check is made and NO
                  checksum is calculated!
    arguments:    taskId  [in] Identification of the calling task.
                  block   [in] Destination position for the given page frame
                  pPage   [in] Pointer to the page frame to be written
    return value: true means write was successful; else false
    --------------------------------------------------------------------------*/

    virtual bool WritePageDirectToDataDevSpace(
        const tsp00_TaskId          taskId,
        const IOMan_BlockAddress    &block,
        const tsp00_PageAddr        pPage );

    /*!-------------------------------------------------------------------------
    function:     ReadPageDirectFromArchiveLog()
    description:  This method is used to read a page from the given block address.
                  from the archive log.Note that NO consistent check is made.
    arguments:    taskId  [in] Identification of the calling task.
                  block   [in] Destination position for the given page frame
                  pPage   [in] Pointer to the page frame to be written
    return value: true means write was successful; else false
    --------------------------------------------------------------------------*/

    virtual bool ReadPageDirectFromArchiveLog(
        const tsp00_TaskId          taskId,
        const IOMan_BlockAddress    &block,
        const tsp00_PageAddr        pPage );

    /*!-------------------------------------------------------------------------
    function:     WritePageDirectToArchiveLog()
    description:  This method is used to write a given page into the given
                  log devspace. Note that NO consistent check is made and NO
                  checksum is calculated!
    arguments:    taskId  [in] Identification of the calling task.
                  block   [in] Destination position for the given page frame
                  pPage   [in] Pointer to the page frame to be written
    return value: true means write was successful; else false
    --------------------------------------------------------------------------*/

    virtual bool WritePageDirectToArchiveLog(
        const tsp00_TaskId          taskId,
        const IOMan_BlockAddress    &block,
        const tsp00_PageAddr        pPage );

    /*!-------------------------------------------------------------------------
    function:     MigrateDataBase()
    description:  This method is used to migrate a non liveCache data base
                  from release 7.2/7.3 to release 7.4 inplace. This includes
                  the migration of the converter and the scattering over all
                  data devspaces, the creation of the IOManInfoPages and the
                  initialization of the log devspaces. In case of a fatal error
                  the system will execute and emergency shutdown. In case of a
                  softer error a base error ist set.
    arguments:    taskId  [in]  Identification of the calling task.
                  trError [out] Current error state
    return value: none
    --------------------------------------------------------------------------*/

    virtual void MigrateDataBase(
        const tsp00_TaskId  taskId,
        tgg00_BasisError    &trError );

    /*!-------------------------------------------------------------------------
    function:     GetFirstDataDevSpaceNo()
    description:  This method returns the identifier of the first data devspace.
                  It's not guaranteed that the returned devspace identifier 
                  specifies a devspace which is online.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetFirstDataDevSpaceNo() const
    {
        return( m_DataControl.GetFirstDataDevSpaceNo());
    }

    /*!-------------------------------------------------------------------------
    function:     GetNextDataDevSpaceNo()
    description:  This method returns the next identifier of the data devspace
                  following the given data devspace. It's not guaranteed that
                  the returned devspace identifier specifies a devspace which
                  is online.
    arguments:    devNo [in] Logical devspace identifier
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetNextDataDevSpaceNo( const IOMan_DeviceNo devNo ) const
    {
        return( m_DataControl.GetNextDataDevSpaceNo( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsDataDevSpaceNoValid()
    description:  This method checks the validity of the given data devspace 
                  identifier.
    arguments:    devNo [in] Logical devspace identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDataDevSpaceNoValid( const IOMan_DeviceNo devNo ) const
    {
        return( m_DataControl.IsDataDevSpaceNoValid( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsDataDevSpaceOnline()
    description:  This method returns whether the given data devspace is in
                  state online.
    arguments:    devNo [in] Logical devspace identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDataDevSpaceOnline( const IOMan_DeviceNo devNo ) const
    {
        return( m_DataControl.IsDataDevSpaceOnline( devNo ));
    }
    /*!-------------------------------------------------------------------------
    function:     GetFirstLogDevSpaceNo()
    description:  This method returns the identifier of the first log devspace.
                  It's not guaranteed that the returned devspace identifier 
                  specifies a devspace which is online.
    arguments:    none
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetFirstLogDevSpaceNo() const
    {
        return( m_LogControl.GetFirstLogDevSpaceNo());
    }

    /*!-------------------------------------------------------------------------
    function:     GetNextLogDevSpaceNo()
    description:  This method returns the next identifier of the log devspace
                  following the given data devspace. It's not guaranteed that
                  the returned devspace identifier specifies a devspace which
                  is online.
    arguments:    devNo [in] Logical devspace identifier
    return value: IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetNextLogDevSpaceNo( const IOMan_DeviceNo devNo ) const
    {
        return( m_LogControl.GetNextLogDevSpaceNo( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsLogDevSpaceNoValid()
    description:  This method checks the validity of the given log devspace 
                  identifier.
    arguments:    devNo [in] Logical devspace identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsLogDevSpaceNoValid( const IOMan_DeviceNo devNo ) const
    {
        return( m_LogControl.IsLogDevSpaceNoValid( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsLogDevSpaceOnline()
    description:  This method returns whether the given log devspace is in
                  state online.
    arguments:    devNo [in] Logical devspace identifier
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsLogDevSpaceOnline( const IOMan_DeviceNo devNo ) const
    {
        return( m_LogControl.IsLogDevSpaceOnline( devNo ));
    }

    /*!-------------------------------------------------------------------------
    function:     ReIntegrateBadLogVolume()
    description:  This method is used to repair a corrupted log volume by copy
                  the intact log volume of the same log device onto the corrupted.
                  If the given log volume isn't bad nothing happens. This
                  feature is available in log mode dual only.
                  (RESTORE LOG VOLUME 'bad volume name')
    arguments:    taskId      [in] Identification of the calling task.
                  badDevName  [in] Name of the bad log volume to be reintegrated
    return value: true means that reintegrate log volume was successful
    --------------------------------------------------------------------------*/

    SAPDB_Bool ReIntegrateBadLogVolume(
        const tsp00_TaskId  taskId,
        tsp00_VFilename     &badDevName )
    {
        return( m_LogControl.ReIntegrateBadLogVolume( taskId, badDevName ));
    }

private:

    // Copy and assignment operator are not supported

    IOMan_Manager& operator=( const IOMan_Manager &Manager );

    IOMan_Manager( const IOMan_Manager &Manager );

    IOMan_Manager( SAPDBMem_IRawAllocator  &allocator );

    bool Initialize(
        const tsp00_TaskId  taskId,
        const SAPDB_UInt    maxDataDevSpaces,
        const SAPDB_UInt    maxArchiveLogs );

    bool RestartAndInitializeFBM( const tsp00_TaskId   taskId );

    tgg00_TransContext BuildAuxTrans( tsp00_TaskId TaskId ) const;

private:

    /*-------------------------------------------------------------------------
    declaration: m_Instance
    description: Pointer to the IOMananger singleton.
    --------------------------------------------------------------------------*/

    static IOMan_Manager    *m_Instance;

    /*-------------------------------------------------------------------------
    declaration: m_Allocator
    description: Base allocator used for all dependent memory allocation of
                 this class. Note that this member data has to be the first 
                 member data defined in this class, because the sequence
                 of the execution of the initialization lists depends on it!
    --------------------------------------------------------------------------*/

    RTEMem_AllocatorWrapper  m_Allocator;

    /*-------------------------------------------------------------------------
    declaration: m_PageAllocator
    description: Allocator for page frames. The page frame size is definded by
                 the data base configuration file. Note that this member data 
                 has to be the second member data defined in this class, because 
                 the sequence of the execution of the initialization lists 
                 depends on it!
    --------------------------------------------------------------------------*/

    IOMan_PageFrameAllocator    m_PageAllocator;

    /*-------------------------------------------------------------------------
    declaration: m_DataControl
    description: Manager istance for all data devspaces.
    --------------------------------------------------------------------------*/

    IOMan_DataControl   m_DataControl;

    /*-------------------------------------------------------------------------
    declaration: m_LogControl
    description: Manager instance for all archive logs.
    --------------------------------------------------------------------------*/

    IOMan_LogControl    m_LogControl;

    /*-------------------------------------------------------------------------
    declaration: m_ReservedBlocks
    description: Collection of reserved block addresses like restart page
    --------------------------------------------------------------------------*/

    IOMan_ReservedBlockAddress  m_ReservedBlocks;
};

/*!*****************************************************************************

   endclass: IOMan_Manager

*******************************************************************************/

#endif  /* IOMAN_MANAGER_HPP */
