/*!**************************************************************************

  module      : IOMan_ReservedBlockAddress.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management 
  description : The class IOMan_ReservedBlockAddress is used to store all 
                reserved block addresses on one central place and to make 
                them accessable.

                The class IOMan_ReservedBlockAddressIterator is used to 
                iterate over a given collection of reserved block addresses.

  last changed: 2001-07-05  18:01
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef IOMAN_RESERVEDBLOCKADDRESS_HPP
#define IOMAN_RESERVEDBLOCKADDRESS_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "IOManager/IOMan_BlockAddress.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "KernelCommon/Kernel_Common.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define IOMAN_INFO_BLOCK_NO             0 // first block within each devspace
#define IOMAN_RESTART_REC_DEVNO         1
#define IOMAN_RESTART_REC_BLOCKNO       1

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_ReservedBlockAddress

   description: This class is used to store all reserved block addresses on
                one central place and to make them accessable.
                 
*******************************************************************************/

class IOMan_ReservedBlockAddress
{
    friend class IOMan_ReservedBlockAddressIterator;
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    function:     IOMan_ReservedBlockAddress
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_ReservedBlockAddress()
    :
    m_InfoBlockNo( IOMAN_INFO_BLOCK_NO ),
    m_RestartBlock( IOMAN_RESTART_REC_DEVNO, IOMAN_RESTART_REC_BLOCKNO )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Access to the reserved block addresses                       */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    function:     GetRestartPageAddress
    description:  This method is used to get the block address of the restart page.
    arguments:    none
    return value: IOMan_BlockAddress of the restart page.
    --------------------------------------------------------------------------*/

    IOMan_BlockAddress GetRestartPageAddress() const
    {
        return( m_RestartBlock );
    }

    /*--------------------------------------------------------------------------
    function:     GetInfoPageAddress
    description:  This method is used to get the block address of the devspace
                  specific IOMan_InfoPage. The devspace is defined by the devNo.
    arguments:    devNo [in] Logical devspace identifier 
    return value: IOMan_BlockAddress of the devspace specific IOMan_InfoPage
    --------------------------------------------------------------------------*/

    IOMan_BlockAddress GetInfoPageAddress( const IOMan_DeviceNo devNo ) const
    {
        return( IOMan_BlockAddress( devNo, IOMAN_INFO_BLOCK_NO ));
    }

    /*--------------------------------------------------------------------------
    function:     GetInfoPageBlockNo
    description:  This method is used to get the block address of the devspace
                  specific IOMan_InfoPage. The devspace is defined by the devNo.
    arguments:    devNo [in] Logical devspace identifier 
    return value: (IOMan_BlockNo) of the devspace specific IOMan_InfoPage
    --------------------------------------------------------------------------*/

    static IOMan_BlockNo GetInfoPageBlockNo( const IOMan_DeviceNo devNo )
    {
        return( IOMAN_INFO_BLOCK_NO );
    }

    /*--------------------------------------------------------------------------
    function:     IsMember
    description:  This method is used to check whether a given block address
                  specifies the location of a reserved block.
    arguments:    block [in] Block address to be checked
    return value: true if given block address is system used block; else false
    --------------------------------------------------------------------------*/

    bool IsMember( const IOMan_BlockAddress   &block ) const
    {
        if( ! block.IsValid() )
            return( false );

        if( m_InfoBlockNo == block.GetBlockNo() )
            return( true );

        if(
            ( m_RestartBlock.GetDeviceNo() == block.GetDeviceNo() ) &&
            ( m_RestartBlock.GetBlockNo() == block.GetBlockNo() )
        )
        {
            return( true );
        }
        return( false );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Access to the reserved block addresses                    */
    /* -----------------------------------------------------------------------*/

private:

    /*------------------------------------------------------------------------------
    declaration: m_InfoBlockNo
    description: Is the block number of the IOMan_InfoPage stored within each devspace.
    ------------------------------------------------------------------------------*/

    IOMan_BlockNo m_InfoBlockNo;

    /*------------------------------------------------------------------------------
    declaration: m_RestartBlock
    description: Is the block address of the restart page.
    ------------------------------------------------------------------------------*/

    IOMan_BlockAddress  m_RestartBlock;
};

/*!*****************************************************************************

   endclass: IOMan_ReservedBlockAddress

*******************************************************************************/

/*!*****************************************************************************

   class: IOMan_ReservedBlockAddressIterator

   description: This class is used to iterate over a given collection of 
                reserved block addresses.
                 
*******************************************************************************/

class IOMan_ReservedBlockAddressIterator
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    function:     IOMan_ReservedBlockAddressIterator
    description:  Constructor
    arguments:    reservedAddr [in] Reference to the block address collection
                  devNo        [in] Logical device number specifies the devspace
                                    on which the iterator ha sto work.
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_ReservedBlockAddressIterator(
        const IOMan_ReservedBlockAddress    &reservedAddr,
        const IOMan_DeviceNo                devNo )
    :
    m_ReservedAddr( reservedAddr ),
    m_DevNo( devNo ),
    m_CurrentAddr( IOMan_InfoPageAddr )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator methods                                             */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    function:     operator ++
    description:  This method is used to set the iterator to the next reserved
                  block address within the specified devspace.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void operator ++()
    {
        switch( m_CurrentAddr )
        {
        case IOMan_InfoPageAddr:
            if( m_DevNo == m_ReservedAddr.GetRestartPageAddress().GetDeviceNo() )
                m_CurrentAddr = IOMan_RestartPageAddr;
            else
                m_CurrentAddr = IOMan_EndAddr;
            break;
        case IOMan_RestartPageAddr:
            m_CurrentAddr = IOMan_EndAddr;
            break;
        default:
            m_CurrentAddr = IOMan_EndAddr;
        }
    }

    /*--------------------------------------------------------------------------
    function:     GetBlockAddress
    description:  This method returns the current reserved block address the 
                  iterator points to.
    arguments:    none
    return value: IOMan_BlockAddress
    --------------------------------------------------------------------------*/

    IOMan_BlockAddress GetBlockAddress() const
    {
        if( m_CurrentAddr == IOMan_InfoPageAddr )
            return( m_ReservedAddr.GetInfoPageAddress( m_DevNo ));
        else
            return( m_ReservedAddr.GetRestartPageAddress());
    }

    /*--------------------------------------------------------------------------
    function:     Begin
    description:  This method is used to set the iterator to the first reserved
                  block address within th edevspace.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void Begin()
    {
        m_CurrentAddr = IOMan_InfoPageAddr;
    }

    /*--------------------------------------------------------------------------
    function:     End
    description:  This method is used to determine the end of the reserved
                  block address collection. I fthe end is reached the method
                  returns true else false.
    arguments:    none
    return value: bool
    --------------------------------------------------------------------------*/

    bool End() const
    {
        return( IOMan_EndAddr == m_CurrentAddr );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator methods                                          */
    /* -----------------------------------------------------------------------*/

private:

    /*------------------------------------------------------------------------------
    declaration: IOMan_ReservedBlockAddressEnum
    description: Enumeration containing all known reserved block addresses.
    ------------------------------------------------------------------------------*/

    typedef enum
    {
        IOMan_InfoPageAddr = 0,
        IOMan_RestartPageAddr,
        IOMan_EndAddr
    }
    IOMan_ReservedBlockAddressEnum;

    /*------------------------------------------------------------------------------
    declaration: m_ReservedAddr
    description: Is a reference to the collection of all reserved.block addresses
                 of one devspace.
    ------------------------------------------------------------------------------*/

    const IOMan_ReservedBlockAddress  &m_ReservedAddr;

    /*------------------------------------------------------------------------------
    declaration: m_DevNo
    description: Is the logical device number of the current devspace on which the
                 iterator works.
    ------------------------------------------------------------------------------*/

    IOMan_DeviceNo  m_DevNo;

    /*------------------------------------------------------------------------------
    declaration: m_CurrentAddr
    description: Is the current reserved block address the iterator points on.
    ------------------------------------------------------------------------------*/

    IOMan_ReservedBlockAddressEnum  m_CurrentAddr;
};

/*!*****************************************************************************

   endclass: IOMan_ReservedBlockAddressIterator

*******************************************************************************/

#endif  /* IOMAN_RESERVEDBLOCKADDRESS_HPP */
