/*!**************************************************************************

  module      : IOMan_Volume.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: I/O Management
  description : This module is used to handle volumes.


  last changed: 2001-01-02  16:00
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

    ========== licence begin  GPL
    Copyright (C) 2001 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef IOMAN_VOLUME_HPP
#define IOMAN_VOLUME_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "vsp002.h"  // PASCAL: IO Types

#include "IOManager/IOMan_ClusterAddress.hpp"
#include "IOManager/IOMan_Types.hpp"
#include "IOManager/IOMan_InfoPage.hpp"
#include "IOManager/IOMan_KernelPages.hpp"
#include "KernelCommon/Kernel_Common.hpp"
#include "KernelCommon/Kernel_IPage.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define IOMAN_READ_ONE_BLOCK        1
#define IOMAN_WRITE_ONE_BLOCK       1

#define IOMAN_MAX_READ_RETRIES      3
#define IOMAN_RETRY_SLEEP_TIME      2

#define IOMAN_SUPPRESS_ERR_MSG      SAPDB_TRUE


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: IOMan_Volume

   description: This class is used to handle volumes.
                It offers an amount of common (!) methods to manage any type 
                of volume. This class is only the base of the proper volume
                class implementations like data, log or trace volume. Therefore
                this class contains only a protected constructor to force the
                programer for example to derive a data volume class from this 
                one. 
                
*******************************************************************************/

class IOMan_Volume
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Administration                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Close()
    description:  This method is used to bring the volume from open mode into 
                  close mode, presumed it's in open mode. The volume is detached. 
    arguments:    taskId [in] Identification of the calling task. 
    return value: none
    --------------------------------------------------------------------------*/

    void Close( const tsp00_TaskId taskId );

    /*!-------------------------------------------------------------------------
    function:     Create()
    description:  This method is used to create and format a volume.
                  If the creation is finished successful the method provides true 
                  else false. It is assumed and tested by an assertion that the 
                  volume is in closed mode.
    arguments:    taskId        [in] Identification of the calling task. 
                  pDBIdentifier [in] Unique data base identifier
    return value: (IOMan_ReturnCode) IOMan_Okay means create was successful
    --------------------------------------------------------------------------*/

    IOMan_ReturnCode Create(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier );

    /*!-------------------------------------------------------------------------
    function:     Open()
    description:  This method is used to bring the volume from close mode into
                  open mode, presumed it isn't in open mode already. For security
                  reasons the IOManInfoPage is checked against the configuration.
    arguments:    taskId        [in] Identification of the calling task.
                  bCheckBadFlag [in] check whether the volume is set to bad and
                                     therefore an open is not allowed.(default: true)
    return value: (IOMan_ReturnCode) IOMan_Okay means open was successful
    --------------------------------------------------------------------------*/

    IOMan_ReturnCode Open(
        const tsp00_TaskId  taskId,
        const bool          bCheckBadFlag = true );

    /*!-------------------------------------------------------------------------
    function:     Add()
    description:  This method is used to inform the RTE about the presence of another
                  new volume. It's required and checked whether the given volume 
                  is in mode offline. Furhtermore the added volume is in mode closed.
    arguments:    devName [in] name of the volume to be added
                  devSize [in] capacity of the volume to be added
    return value: true if add was successful
    --------------------------------------------------------------------------*/

    SAPDB_Bool Add(
        tsp00_VFilename         &devName,
        const IOMan_BlockCount  devSize ) const;

    /*!-------------------------------------------------------------------------
    function:     Del()
    description:  This method is used to inform the RTE about the removing of 
                  the volume. It's required and checked whether the given 
                  volume is in mode offline.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void Del() const;

    /*-------------------------------------------------------------------------
    function:     SetNextLinkage()
    description:  This method is used to write the logical device number of 
                  the next logical volume into the info page of the
                  current volume.
    arguments:    taskId     [in] Identification of the calling task.
                  nextVolume [in] Volume handler
    return:       (SAPDB_Bool) linkage is created successfully
    --------------------------------------------------------------------------*/

    SAPDB_Bool SetNextLinkage(
        const tsp00_TaskId  taskId,
        const IOMan_Volume  &nextVolume);

    /*-------------------------------------------------------------------------
    function:     SetPrevLinkage()
    description:  This method is used to write the logical device number of 
                  the previous logical volume into the info page of the
                  current volume.
    arguments:    taskId     [in] Identification of the calling task.
                  prevVolume [in] Volume handler
    return:      (SAPDB_Bool) linkage is created successfully
    --------------------------------------------------------------------------*/

    SAPDB_Bool SetPrevLinkage(
        const tsp00_TaskId  taskId,
        const IOMan_Volume  &prevVolume );

    /*!-------------------------------------------------------------------------
    function:     IsLinkageValid()
    description:  This method is used to validate the linkage between the currrent
                  and the given volume. 
    arguments:    taskId [in] Identification of the calling task.
                  volume [in] volume handler
    return value: (SAPDB_Bool) true if linkage is valid; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsLinkageValid(
        const tsp00_TaskId  taskId,
        const IOMan_Volume  &volume ) const;

    /*!-------------------------------------------------------------------------
    function:     IsPrevLinkageValid()
    description:  This method is used to validate if a logical predecessor 
                  volume is definded.
    arguments:    none.
    return value: (SAPDB_Bool) true if predecessor volume is definded
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsPrevLinkageValid() const;

    /*!-------------------------------------------------------------------------
    function:     IsNextLinkageValid()
    description:  This method is used to validate if a logical successor 
                  volume is definded.
    arguments:    none
    return value: (SAPDB_Bool) true if successor volume is definded
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsNextLinkageValid() const;

    /*!-------------------------------------------------------------------------
    function:     IsDBIdentifierValid()
    description:  This method is used to compare the DBIdentifier of the 
                  currrent and the given volume. If the identifiers are equal 
                  true is returned else false.
    arguments:    taskId [in] Identification of the calling task.
                  volume [in] Volume handler
    return value: (SAPDB_Bool) true if identifiers are equal; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDBIdentifierValid(
        const tsp00_TaskId  taskId,
        const IOMan_Volume  &volume ) const
    {
        if(( !m_Online ) || (! volume.IsOnline()))
            return( false );

        return( IsDBIdentifierValid( taskId, volume.GetDBIdentifier() ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsDBIdentifierValid()
    description:  This method is used to compare the given DBIdentifier with 
                  the DBIdentifier of the current volume. If the identifiers 
                  are equal true is returned else false.
    arguments:    taskId        [in] Identification of the calling task.
                  pDBIdentifier [in] DBIdentifier
    return value: (SAPDB_Bool) true if identifiers are equal; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsDBIdentifierValid(
        const tsp00_TaskId  taskId,
        const SAPDB_Byte    *pDBIdentifier ) const;

    /*!-------------------------------------------------------------------------
    function:     UnSetBad()
    description:  This method is used to remove the bad marker from the volume.
                  A following open call won't break down caused by the bad flag.
    arguments:    taskId [in] Identification of the calling task.
    return value: (IOMan_ReturnCode) IOMan_Okay means unset was successful
    --------------------------------------------------------------------------*/

    IOMan_ReturnCode UnSetBad( const tsp00_TaskId   taskId );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Administration                                            */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Common information                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetBlockSize()
    description:  This method is used to get the size in bytes of one block
                  managed by volumes. Note that all blocks within the same 
                  volume have the same size.
    arguments:    none
    return value: Size of one block of the current volume in bytes
    --------------------------------------------------------------------------*/

    SAPDB_UInt4 GetBlockSize() const
    {
        return( m_BlockSize );
    }

    /*!-------------------------------------------------------------------------
    function:     GetCapacity()
    description:  This method is used to get the volume capacity in blocks. 
                  This is the  maximum number of blocks storable in the volume. 
                  Note that there are two kinds of capacity, the complete and 
                  the usable capacity. The complete capacity is equal to the 
                  value within the data base configuration file whereas the 
                  usable capacity the number of blocks defines which are realy 
                  available to store data pages. The difference is caused by 
                  IOMan_InfoPage which is stored within each volume and therefore 
                  responsible for the reduced amount of blocks. The default 
                  parameter bReduceCapacity determines which value is requested. 
                  The value bReduceCapacity == true returns the reduced number 
                  of blocks, which is the proper amount of blocks for example 
                  usable by the converter. The block size is provided by the 
                  method GetBlockSize.
    arguments:    bReduceCapacity [in] Return the reduced capacity
    return value: IOMan_BlockCount
    --------------------------------------------------------------------------*/

    IOMan_BlockCount GetCapacity( const SAPDB_Bool bReduceCapacity = true ) const
    {
        if( bReduceCapacity )
            return( m_Capacity -1 ); // Reduce by one block for the IOMan_InfoPage
        else
            return( m_Capacity );
    }

    /*!-------------------------------------------------------------------------
    function:     GetDBIdentifier()
    description:  This method is used to get the unique identifier of the 
                  volume. 
    arguments:    none
    return value: (const SAPDB_Byte*) pointer to (not null terminated) string
    --------------------------------------------------------------------------*/

    const SAPDB_Byte* GetDBIdentifier() const
    {
        SAPDBERR_ASSERT_STATE( m_Online );

        return( m_InfoPage.GetDBIdentifier() );
    }

    /*!-------------------------------------------------------------------------
     function:     GetLogicalDevNo()
     description:  This method is used to get the logical identifier of the
                   volume. All volumes of the same volume type are counted 
                   begining by ONE(!).
     arguments:    none
     return value: logical volume id
     --------------------------------------------------------------------------*/

    IOMan_DeviceNo  GetLogicalDevNo() const
    {
        return( m_LogicalDevNo );
    }

    /*!-------------------------------------------------------------------------
    function:     GetNextLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the volume which is the next in logical order. This number
                  must be equal to the logical next volume within the data 
                  base configuration file.
    arguments:    none
    return:       IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetNextLogicalDevNo() const
    {
        SAPDBERR_ASSERT_STATE( m_Online );

        return( m_InfoPage.GetNextLogicalDevNo() );
    }

    /*!-------------------------------------------------------------------------
    function:     GetPrevLogicalDevNo()
    description:  This method is used to read the logical device number of
                  the volume which is the predecessor in logical order. This 
                  number must be equal to the logical predecessor volume within 
                  the data base configuration file.
    arguments:    none
    return:       IOMan_DeviceNo
    --------------------------------------------------------------------------*/

    IOMan_DeviceNo GetPrevLogicalDevNo() const
    {
        SAPDBERR_ASSERT_STATE( m_Online );

        return( m_InfoPage.GetPrevLogicalDevNo() );
    }

    /*!-------------------------------------------------------------------------
    function:     GetName()
    description:  This method is used to get the configured name of the volume.
                  If no valid volume name is found false is returned and the
                  incoming volume name is not touched. Else the configured 
                  volume name is stored within the incoming parameter. Is is 
                  assumed and tested by an assertion that the volume is in open 
                  mode.
    arguments:    devName [out] volume name
    return value: true means success; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetName( tsp00_VFilename &DevName ) const;

    /*!-------------------------------------------------------------------------
    function:     GetNameWithTermination()
    description:  Same conditions as the method above but in this case the
                  returned volume name is terminted by '\0'.
    arguments:    devName [out] volume name
    return value: true means success; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetNameWithTermination( tsp00_VFilename &devName ) const;

    /*!-------------------------------------------------------------------------
    function:     GetTypeName()
    description:  This method is used to get the type for example data or log
                  of the current volume.
    arguments:    none
    return value: volume type
    --------------------------------------------------------------------------*/

    SAPDB_Char* GetTypeName() const
    {
        switch( m_Type )
        {
        case sp2dt_log:
            return(( SAPDB_Char* )"Log" );
        case sp2dt_mirr_log:
            return(( SAPDB_Char* )"MLog" );
        case sp2dt_data:
            return(( SAPDB_Char* )"Data" );
        case sp2dt_trace:
            return(( SAPDB_Char* )"Trace" );
        default:
            return(( SAPDB_Char* )"UNDEF" );
        }
    }

    /*!-------------------------------------------------------------------------
    function:     IsConfigured ()
    description:  This method is used to check whether the current volume 
                  exists, i.e. is stored in the data base configuration file.
    arguments:    none
    return value: true means is known as a valid volume; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsConfigured() const;

    /*!-------------------------------------------------------------------------
    function:     IsOnline()
    description:  This method is used to determine whether the volume is in open
                  mode.
    arguments:    none
    return value: true means is in open mode; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsOnline() const
    {
        return( m_Online );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Common information                                        */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Input/Output                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Read()
    description:  This method encapsulate the ReadSimple method. Additional it
                  offers the possibility to read a page three times if the check
                  method of the page detect an error. Each error is loged in the 
                  knldiag file and the io buffer is dumped. In this case the 
                  method provides a false as return code. In case of fatal
                  errors of the IO subsystem an emergency shutdown is executed!!
    arguments:    taskId  [in] Identification of the calling task.
                  page    [in] Handler of the wanted Page.
                  blockNo [in] Position of the wanted page within the volume.
    return value: true if the read operation is successful; else false.
    --------------------------------------------------------------------------*/

    SAPDB_Bool Read(
        const tsp00_TaskId  taskId,
        Kernel_IPage        &Page,
        const IOMan_BlockNo blockNo );

    /*!-------------------------------------------------------------------------
    function:     ReadSimple()
    description:  This method is used to read a page (default) from a volume 
                  into the given memory. The correctness  of the read page won't
                  be checked. In case of a fatal errror during the read operation
                  caused by the IO subsystem the error is logged in the knldiag 
                  file and the return code is false. The volume is marked as bad.
    arguments:    taskId     [in] Identification of the calling task.
                  pFrame     [in] Memory where the read page will be stored.
                  blockNo    [in] Position of the wanted page within the volume.
                  blockCount [in] Number of pages to be read in one step
                                  Default is one page
    return value: true if the read operation is successful; else false.
    --------------------------------------------------------------------------*/

    SAPDB_Bool ReadSimple(
        const tsp00_TaskId      taskId,
        const tsp00_PageAddr    pFrame,
        const IOMan_BlockNo     blockNo,
        const IOMan_BlockCount  blockCount = IOMAN_READ_ONE_BLOCK );

    /*!-------------------------------------------------------------------------
    function:     Write()
    description:  This method encapsulate the WriteSimple method. Additional it
                  offers the possibility to check the page before it is written 
                  into the volume. If during the check an error occurs an 
                  emergency shutdown is executed. For more informations see 
                  WriteSimple.
    arguments:    taskId     [in] Identification of the calling task.
                  pFrame     [in] Memory of the writeable pages.
                  blockNo    [in] Position for the page within the volume.
    return value: none
    --------------------------------------------------------------------------*/

    void Write(
        const tsp00_TaskId  taskId,
        Kernel_IPage        &Page,
        const IOMan_BlockNo blockNo );

    /*!-------------------------------------------------------------------------
    function:     WriteSimple()
    description:  This method is used to write a page onto a volume 
                  The correctness  of the page won't be  checked. In case of a 
                  fatal errror during the write operation caused by the IO subsystem
                  the error is logged in the knldiag file and the return code is 
                  false. The volume is marked as bad.
    arguments:    taskId     [in] Identification of the calling task.
                  page       [in] Handler of the given page.
                  blockNo    [in] Position for the page within the volume.
    return value: true if the write operation is successful; else false.
    --------------------------------------------------------------------------*/

    SAPDB_Bool WriteSimple(
        const tsp00_TaskId      taskId,
        const tsp00_PageAddr    pFrame,
        const IOMan_BlockNo     blockNo );

    /*!-------------------------------------------------------------------------
    function:     WriteVector()
    description:  This method is used to write a number of pages onto a volume 
                  Additional it offers the possibility to check the page before 
                  it is written into the volume. In case of a fatal errror 
                  during the write operation caused by the IO subsystem the 
                  error is logged in the knldiag file and the return code is 
                  false. The volume is marked as bad.
    arguments:    taskId      [in] Identification of the calling task.
                  pages       [in] Handler of the pages to be written
                  cluster     [in] Destination address of the pages to be written
                  startIndex  [in] Specifies the first page in the page collection 
                                   which is relevant for the following I/O. Default
                                   the startIndex is zero. Only in case of write
                                   orders splited over more than one volume
                                   the startIndex is greater than zero.
    return value: none
    --------------------------------------------------------------------------*/

    void WriteVector(
        const tsp00_TaskId          taskId,
        IOMan_IKernelPages          &pages,
        const IOMan_ClusterAddress  &cluster,
        const SAPDB_UInt            startIndex = IOMAN_FIRST_INDEX);

    /*!-------------------------------------------------------------------------
    function:     AsynReadWrite()
    description:  This method is used to read or write via a separate io thread
                  insulated from the common kernel io threads. The correctness  
                  of the page won't be  checked.
    arguments:    taskId       [in] Identification of the calling task.
                  asynIOHandle [in] Handle for the asynchronous io
                  pages        [in] Handler of the pages to be written
                  cluster      [in] Destination address of the pages to be written
    return value: true means write was successfull
    --------------------------------------------------------------------------*/

    SAPDB_Bool AsynReadWrite(
        const tsp00_TaskId          taskId,
        const SAPDB_Int4            asynIOHandle,
        IOMan_IKernelPages          &pages,
        const IOMan_ClusterAddress  &cluster );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Input/Output                                              */
    /* -----------------------------------------------------------------------*/

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     IOMan_Volume()
    description:  Constructor. This method is used to set the member data to an 
                  initial state. The volume type and the logical identifier is 
                  set to the given value and at this point non changeable. The 
                  methods Create and Open are used to activate the volume. This 
                  contructor is protected to force the programer to use the derived 
                  data or log volume class constructor based on this one.
    arguments:    volumeType [in] Type of the volume
    return value: none
    --------------------------------------------------------------------------*/

    IOMan_Volume( const tsp2_dev_type &volumeType )
    :
    m_Online( false ),
    m_Type( volumeType ),
    m_Capacity( 0 ),
    m_LogicalDevNo( 0 ),
    m_PhysicalDevNo( 0 ),
    m_BlockSize( 0 ),
    m_InfoPage()
    {}

    /*!-------------------------------------------------------------------------
    function:     Initialize()  
    description:  This method is used to set the individual logical volume
                  identifier.
    arguments:    devNo [in] Logical volume identifier
                  frame [in] Frame for the IOManInfoPage of the volume
    return value: none
    --------------------------------------------------------------------------*/

    void Initialize(
        const IOMan_DeviceNo            devNo,
        const Kernel_IPage::PageFrame   &frame )
    {
        m_LogicalDevNo = devNo;
        m_InfoPage.SetFrame( frame );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Administration                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     OpenSimple()
    description:  This method is used to bring the volume from close mode into
                  open mode, presumed it isn't in open mode already. Note that
                  no security check is made and the IOMan_InfoPage is NOT read!
    arguments:    taskId          [in] Identification of the calling task.
                  bSuppressErrMsg [in] no message into knldiag if attach failed
    return value: (IOMan_ReturnCode) IOMan_Okay means open was successful
    --------------------------------------------------------------------------*/

    IOMan_ReturnCode OpenSimple(
        const tsp00_TaskId taskId,
        const SAPDB_Bool   bSuppressErrMsg );

    /*!-------------------------------------------------------------------------
    function:     SetBad()
    description:  This method is used to mark an open volume as bad. If the 
                  given block number is valid it is written into the zero page 
                  of the volume. Note that the method won't close the volume. 
                  Each following open of the volume will cause an error.
    arguments:    taskId  [in] Identification of the calling task.
                  blockNo [in] BlockNo to be inserted into the zero page.
    return value: none
    --------------------------------------------------------------------------*/

    void SetBad(
        const tsp00_TaskId  taskId,
        const IOMan_BlockNo blockNo );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Administration                                            */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Input/Output                                                 */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    function:     WriteInfoPage()  
    description:  This method is used to write the IOMan_InfoPage to the 
                  predefined position of within the current volume. In case
                  of an error the system aborts.
    arguments:    taskId [in] Identification of the calling task.
    return value: none
    --------------------------------------------------------------------------*/

    void WriteInfoPage( const tsp00_TaskId  taskId );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Input/Output                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Common information                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetPhysicalDevNo()
    description:  This method is used to get the physical identifier of the
                  volume. This is needed by the RTE Interface to identify a 
                  volume.
    arguments:    none
    return value: physical volume id
    --------------------------------------------------------------------------*/

    SAPDB_Int4 GetPhysicalDevNo() const
    {
        return( m_PhysicalDevNo );
    }

    /*!-------------------------------------------------------------------------
    function:     IsBad()
    description:  This method is used to determine if the volume is set to bad.
    arguments:    none
    return value: true means is bad; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsBad() const
    {
        return( 0 != m_InfoPage.GetBadBlockNo() );
    }

    /*!-------------------------------------------------------------------------
    function:     IsBlockNoMapable()
    description:  This method is used to determine whether a given block number
                  is addressable via the volume. i.e. the block number is in
                  the valid number range of the volume.
    arguments:    blockNo [in] BlockNo to be examined
    return value: true means is in volume scope; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsBlockNoMapable( const IOMan_BlockNo blockNo ) const
    {
        return(( 0 <= blockNo ) && ( blockNo < m_Capacity ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsVolumeConfigurationValid()  
    description:  This method is used to check the correctness of the 
                  IOMan_InfoPage in relation the the configuration file.
    arguments:    none
    return value: true if no error is detected; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool IsVolumeConfigurationValid() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Common information                                        */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     Attach()
    description:  This method is used to attach the current volume.
    arguments:    taskId          [in] Identification of the calling task.
                  bSuppressErrMsg [in] if set to true no error message
                                       is written into knldiag in case
                                       of an open error. (Default is false)
    return value: true means attach was successful; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool Attach(
        const tsp00_TaskId taskId,
        const SAPDB_Bool   bSuppressErrMsg = false );

    /*--------------------------------------------------------------------------
    function:     GetConfiguration()
    description:  This method is used to get the configuration information
                  of the current volume from the data base configuration 
                  file. If no configuration is available the return value 
                  is false.
    arguments:    capacity  [out] volume size in pages
                  blockSize [out] Page size 
    return value: true means is known as a valid volume; else false
    --------------------------------------------------------------------------*/

    SAPDB_Bool GetConfiguration(
        IOMan_BlockCount    &capacity,
        SAPDB_UInt4         &blockSize ) const;

    /*--------------------------------------------------------------------------
    function:     WriteVectorSimple()
    description:  This method is used to write a number of pages onto a volume 
                  The correctness  of the page won't be  checked. In case of a 
                  fatal errror during the write operation caused by the IO subsystem
                  the error is logged in the knldiag file and the return code is 
                  false. The volume is marked as bad.
    arguments:    taskId     [in] Identification of the calling task.
                  pFrame     [in] Memory of the writeable pages.
                  blockNo    [in] Position for the first page within the volume.
                  blockCount [in] Number of pages to write
    return value: true if the write operation is successful; else false.
    --------------------------------------------------------------------------*/

    SAPDB_Bool WriteVectorSimple(
        const tsp00_TaskId      taskId,
        tsp00_PageAddr          *pFrame,
        const IOMan_BlockNo     blockNo,
        const IOMan_BlockCount  blockCount );

private:

    // Assignment operator is not supported

    IOMan_Volume& operator=( const IOMan_Volume &volume );

    // copy constructor is needed for newarray operation during initialization
    //IOMan_Volume( const IOMan_Volume &volume );

protected:

    /*-----------------------------------------------------------------------------
    declaration: m_Online
    description: Indicates whether the volume is in open mode and with that 
                 accessable for read and write operations.
    ------------------------------------------------------------------------------*/

    SAPDB_Bool    m_Online;

    /*-----------------------------------------------------------------------------
    declaration: m_Type
    description: Determines the type of the volume. Common volume types are 
                 data and log.
    ------------------------------------------------------------------------------*/

    const tsp2_dev_type  m_Type;

    /*-----------------------------------------------------------------------------
    declaration: m_Capacity
    description: Is the maximum number of blocks storable in the volume. The 
                 block size is defined by the member data m_BlockSize.
    ------------------------------------------------------------------------------*/

    IOMan_BlockCount m_Capacity;

    /*-----------------------------------------------------------------------------
    declaration: m_LogicalDevNo
    description: Is the logical identifier for an volume. The m_LogicalDevNo is
                 a 'virtual' identifier for a volume. All volumes of the same
                 volume type are counted begining by ONE(!).
    ------------------------------------------------------------------------------*/

    IOMan_DeviceNo  m_LogicalDevNo;

    /*-----------------------------------------------------------------------------
    declaration: m_PhysicalDevNo
    description: Is the physical identifier for an volume. The m_PhysicalDevNo is
                 needed by the RTE Interface to identify a volume.
    ------------------------------------------------------------------------------*/

    SAPDB_Int4  m_PhysicalDevNo;

    /*-----------------------------------------------------------------------------
    declaration: m_BlockSize
    description: Is the size in bytes of one block within the volume. The default
                 configured block size is 8192 bytes.
    ------------------------------------------------------------------------------*/

    SAPDB_UInt4 m_BlockSize;

    /*-----------------------------------------------------------------------------
    declaration: m_InfoPage
    description: Is a page handler for the IOManInfoPage which is stored at the
                 begining of each volume. It contains administrative informations
                 which has to be check at each restart of the io management.
    ------------------------------------------------------------------------------*/

    IOMan_InfoPage  m_InfoPage;
};

/*!*****************************************************************************

   endclass: IOMan_Volume

*******************************************************************************/

#endif  /* IOMAN_VOLUME_HPP */

