/*!*****************************************************************************

  module:       Kernel_DateTime.hpp

  ------------------------------------------------------------------------------

  responsible:  Till Luhmann

  special area: SAPDB Kernel

  description:  Kernel date time support

  see also:     

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef KERNEL_DATETIME_HPP
#define KERNEL_DATETIME_HPP

#include "ggg00.h"
#include "heo52.h"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/SAPDB_EnhancedTypes.hpp"


/*!-----------------------------------------------------------------------------
declaration : Kernel_Date
description:  date value (e.g. 20010213 == 13.02.2001)
------------------------------------------------------------------------------*/
#define KERNEL_INVALID_DATE 0
typedef SAPDB_Invalidatable<SAPDB_UInt4, KERNEL_INVALID_DATE> 
        Kernel_Date;

/*!-----------------------------------------------------------------------------
declaration : Kernel_Time
description:  time value (e.g. 00154857 == 15:48:57)
------------------------------------------------------------------------------*/
#define KERNEL_INVALID_TIME 0
typedef SAPDB_Invalidatable<SAPDB_UInt4, KERNEL_INVALID_TIME> 
        Kernel_Time;

/*!*****************************************************************************

   class: Kernel_DateTime

   description: Support for date time conversion
   

*******************************************************************************/
class Kernel_DateTime
{
public:

    /*!-------------------------------------------------------------------------
    function:     GetSystemDateTime()
    description:  Returns actual system date and time in numeric format

    arguments:    date [out]  actual date
                  time [out]  actual time
    return value: none
    --------------------------------------------------------------------------*/
    static void GetSystemDateTime
        (Kernel_Date&                   date,
         Kernel_Time&                   time)
    {
        tsp00_Date  sdate;
        tsp00_Time  stime;
        SAPDB_UInt4 p;
        SAPDB_Int   i;

        vdattime(sdate, stime);

        date = 0;
        p    = 1;
        for(i = DATE_MXSP00-1; i >= 0; --i)
        {
            date += (sdate[i] - '0') * p;
            p    *= 10;
        }

        time = 0;
        p    = 1;
        for(i = DATE_MXSP00-1; i >= 0; --i)
        {
            time += (stime[i] - '0') * p;
            p    *= 10;
        }
    }
    /*!-------------------------------------------------------------------------
    function:     OlderThan()
    description:  compares two tuples of (date/time) in numeric format
                  PTS 1111525 mb 2001-01-15

    arguments:    dateLeftOp  [in]  date of left operand
                  timeLeftOp  [in]  time of left operand
                  dateRightOp [in]  date of right operand
                  timeRightOp [in]  time of right operand
    return value: bool (LeftOp olderThan RightOp)
    --------------------------------------------------------------------------*/
    static bool OlderThan
        (const Kernel_Date                   dateLeftOp,
         const Kernel_Time                   timeLeftOp,
         const Kernel_Date                   dateRightOp,
         const Kernel_Time                   timeRightOp)
    {
        if (dateLeftOp < dateRightOp)
        {
            return true;
        }
        else if (dateLeftOp == dateRightOp)
        {
            return timeLeftOp < timeRightOp;
        }
        else
        {
            return false;
        }
    }
};
/*!*****************************************************************************

    endclass: Kernel_DateTime

*******************************************************************************/


#endif // KERNEL_DATETIME_HPP
