/*!**************************************************************************

  module      : Kernel_EventSet.hpp

  -------------------------------------------------------------------------

  responsible : d025251

  special area: 
  description : 


  last changed: 2001-02-14  17:36
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef KERNEL_EVENTSET_HPP
#define KERNEL_EVENTSET_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "vsp0031.h"
#include "SAPDBCommon/SAPDB_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define MAX_NUM_EVENTS              101
#define SINGLE_VALUE_ACTIVATE_DIFF  0.25  // in % 
#define HUNDERT_PERCENT             100

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class Kernel_Event
{
	friend class Kernel_EventSet;

private: // members
	SAPDB_Int        m_AbsolutThreshold;   
	SAPDB_Int        m_RelativeThreshold;   
    tsp31_event_prio m_OverFlowPriority;
    tsp31_event_prio m_UnderFlowPriority;
	SAPDB_Bool       m_ReportOverFlow;
	SAPDB_Bool       m_ReportUnderFlow;
	SAPDB_Bool       m_WriteAlwaysIntoKnldiag;


private: // methods 
    Kernel_Event ();
	void    Initialize();   // PTS 1102939 AK 31-05-99
};

/*!*****************************************************************************

   class:        Kernel_EventSet
   description:  Using this class one can watch a parameter. Whenever the method Monitor
                 is called it is checked whether the parameter passed predefined values.
				 If this is the case an event is triggered which can be writing a message
				 into the knldiag or inserting an event into the event queue. 

*******************************************************************************/

class Kernel_EventSet
{	
public:  // methods

	/*!-------------------------------------------------------------------------
    function:     AddEvent()
    description:  adds an event to the set of events
	
	  - With this function an event can be defined which is triggered when by calling
	    the watch-function it is realized that the input value has passed  the given threshold

    arguments:    RelativeThreshold        [in] critical value in 0.01 percent with respect to the basis unit defined 
	                                            in the constructor or via rescale. If this value is passed by the parameter to watch and 
                                                bOverflow is true an event is triggered, i.e. an event is  inserted
									            into the event queue, if bReportAlwaysIntoKnldiag is true an event is always
									            written to the knldiag too, even is the eventing is disabled
                  bOverflow                [in] defines if an event should be triggerd of the value to be monitored exceeds
				                                or falls below the threshold
                  Priority                 [in] priority of the event
				  bWriteAlwaysIntoKnldiag  [in] if this is true the event is always reported in the knldiag 
				                                even if the eventing is disabled.

    return value: (SAPDB_Bool) true if the event set is not empty
    --------------------------------------------------------------------------*/
	SAPDB_Bool AddEvent      (const SAPDB_Int                RelativeThreshold, 
		                      const SAPDB_Bool               bOverflow, 
							  const tsp31_event_prio_Param   Priority,
							  const SAPDB_Bool               bWriteAlwaysIntoKnldiag = false);

	/*!-------------------------------------------------------------------------
    function:     DelEvent()
    description:  removes an event from the set of events

    arguments:    RelativeThreshold        [in] defines the critical value to be removed. The value is 
	                                            given in units of 0.01 percent with respect to the base unit
                  bOverflow                [in] displays whether the event of exceeding or falling below is to remove

    return value: (SAPDB_Bool) true if the event set is not empty
    --------------------------------------------------------------------------*/
	SAPDB_Bool DelEvent      (const SAPDB_Int    RelativeThreshold,
		                      const SAPDB_Bool   bOverflow);
	
	/*!-------------------------------------------------------------------------
    function:     Watch()
    description:  monitors a parameter

      - Whenever this function is cold it is checked wheteher the parameter has passed any of the
	    thresholds defined in the event set since the command was called the last time. If this is the case
		the events corresponding to the passed values are triggered.

    arguments:    Value [in] value (given in its absolut size) which is to watch
    return value: none
    --------------------------------------------------------------------------*/
	void Watch (const SAPDB_Int Value)
	{	
		if (Value > m_LastValue)
			LookForOverFlowEvent  (Value);
		else if (Value < m_LastValue)
			LookForUnderFlowEvent(Value);
		
		m_LastValue = Value;
	}

	/*!-------------------------------------------------------------------------
    function:     GetEventSet()
    description:  puts all events into an external structure in order to show them in a system view

      - The ShortEventDesc is filled with all events in order to show the events in a system view 

    arguments:    ShortEventDesc     [out] structure to fill
                  bGetOverflowEvents [in]  if true  ShortEventDesc is filled with overflow events
				                           if false ShortEventDesc is filled with underflow events
    return value: none
    --------------------------------------------------------------------------*/
	void GetEventSet  (tsp31_short_event_desc &ShortEventDesc,
		               SAPDB_Bool              bGetOverflowEvents);

	/*!-------------------------------------------------------------------------
    function:     DisableEventing()
    description:  disables eventing

     - After the eventing is disabled, no events are triggered but passing critical values
	   is reported into the knldiag file for those critical values, for which this was explicitely 
	   ordered (see AddEvent)

    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
	void DisableEventing()
	{
		m_EventReportingEnabled = false;
	}

	/*!-------------------------------------------------------------------------
    function:     EnableEventing()
    description:  enables eventing,i.e each time a critical value is passed an event is triggered

    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
	void EnableEventing()
	{
		m_EventReportingEnabled = true;
	}

	/*!-------------------------------------------------------------------------
    function:     Rescale()
    description:  rescales all event thresholds with repsect to the new base unit
  
    arguments:    NewBaseValue [in] new base unit which defines the unit for all relative thresholds    
    return value: none
    --------------------------------------------------------------------------*/
    void Rescale (SAPDB_UInt    NewBaseValue);

	/*!-------------------------------------------------------------------------
    function:     Kernel_EventSet()
    description:  constructor


    arguments:    OverFlowEventId      [in] id of the event when the monitored parameter exceeds a critical value 
	              UnderFlowEventId     [in] id of the event when the monitored parameter falls below a critical value 
                  BaseValue             [in] all events added to the set must define their critical value
				                            in percent with repect to this unit
    return value: none
    --------------------------------------------------------------------------*/
	Kernel_EventSet (const tsp31_event_ident_Param  OverFlowEventId, 
	        	     const tsp31_event_ident_Param  UnderFlowEventId,
					 SAPDB_UInt                     BaseValue = 100);
	
private: // methods
	SAPDB_Bool AddEmptyEvent         (const SAPDB_Int   Index, const SAPDB_Int  RelativeThreshold);
	SAPDB_Bool DelEventIfItIsEmpty   (const SAPDB_Int   Index);
	SAPDB_Int  FindEventIndex        (const SAPDB_Int   RelativeThreshold, SAPDB_Bool &bFound);
	void LookForOverFlowEvent        (const SAPDB_Int   Value);
	void LookForUnderFlowEvent       (const SAPDB_Int   Value);
	void ReportEvent                 (const tsp31_event_ident_Param  EventId, 
		                              const tsp31_event_prio_Param   EventPriority, 
				                      const Kernel_Event            &Event);
	
private: // members
	SAPDB_Int               m_IndexOfFirstActiveEvent;
	SAPDB_Int               m_LastValue;               
	SAPDB_Int               m_BaseValue;
	SAPDB_Int               m_NumEvents;
	SAPDB_Int               m_NumSingleEvents;
	SAPDB_Int               m_ReActivateEventDiff;    
	SAPDB_Int               m_UnitOfRelativeThreshold;
	SAPDB_Bool              m_EventReportingEnabled;
	tsp31_event_ident       m_OverFlowSingleEventId; 
	tsp31_event_ident       m_UnderFlowSingleEventId;

	Kernel_Event      m_Event  [MAX_NUM_EVENTS - 1];
};

/*!*****************************************************************************

   endclass: Kernel_EventSet

*******************************************************************************/

/*===========================================================================*
 *  DEFINITION OF INLINE CLASS METHODS                                       *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/* CLASS Kernel_Event                                                  */
/*---------------------------------------------------------------------------*/

inline void
Kernel_Event::Initialize ()                             
{	
	m_AbsolutThreshold       = 0;
	m_RelativeThreshold      = 0;
	m_OverFlowPriority.becomes(sp31ep_nil);
	m_UnderFlowPriority.becomes(sp31ep_nil);
	m_ReportOverFlow         = false; 
	m_ReportUnderFlow        = false;
	m_WriteAlwaysIntoKnldiag = false;
};

/*---------------------------------------------------------------------------*/

inline 
Kernel_Event::Kernel_Event ()
{
	Initialize();                                        
};

/*---------------------------------------------------------------------------*/
/* CLASS Kernel_EventSet                                                     */
/*---------------------------------------------------------------------------*/

inline
Kernel_EventSet::Kernel_EventSet (const tsp31_event_ident_Param  OverFlowEventId, 
								              const tsp31_event_ident_Param  UnderFlowEventId,
						                      SAPDB_UInt                     BaseValue)
:
m_NumEvents               (0), 
m_NumSingleEvents         (0),
m_IndexOfFirstActiveEvent (0),
m_LastValue               (0),
m_UnitOfRelativeThreshold (10000),
m_EventReportingEnabled   (false), 
m_BaseValue               (BaseValue)
{
	m_ReActivateEventDiff  = SAPDB_Int ((m_UnitOfRelativeThreshold*SINGLE_VALUE_ACTIVATE_DIFF)/HUNDERT_PERCENT);
	m_OverFlowSingleEventId.becomes (OverFlowEventId);    
	m_UnderFlowSingleEventId.becomes(UnderFlowEventId);
}

/*---------------------------------------------------------------------------*/

inline void
Kernel_EventSet::LookForOverFlowEvent   (const SAPDB_Int Value)
{		
	if (m_IndexOfFirstActiveEvent >= m_NumEvents) return ;
	Kernel_Event & CurrEvent = m_Event [m_IndexOfFirstActiveEvent];
	if(Value >  CurrEvent.m_AbsolutThreshold + m_ReActivateEventDiff)
	{
		++ m_IndexOfFirstActiveEvent;
		if (CurrEvent.m_ReportOverFlow )
			ReportEvent (m_OverFlowSingleEventId, CurrEvent.m_OverFlowPriority, CurrEvent);
	}
}

/*---------------------------------------------------------------------------*/

inline void
Kernel_EventSet::LookForUnderFlowEvent  (const SAPDB_Int Value)
{	
	if (0 == m_IndexOfFirstActiveEvent) return;
	const Kernel_Event & CurrEvent = m_Event [m_IndexOfFirstActiveEvent - 1];
	if(Value < CurrEvent.m_AbsolutThreshold - m_ReActivateEventDiff)
	{
		--m_IndexOfFirstActiveEvent;
		if (CurrEvent.m_ReportUnderFlow)
		    ReportEvent (m_UnderFlowSingleEventId, CurrEvent.m_UnderFlowPriority, CurrEvent);
	}
}

/*---------------------------------------------------------------------------*/

#endif  /* KERNEL_EVENTSET_HPP */
