/*!*****************************************************************************

  module:       Kernel_IPage.hpp

  ------------------------------------------------------------------------------

  responsible:  Till Luhmann

  special area: SAPDB Kernel

  description:  Common page handler interface
                
  see also:     

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/

#ifndef KERNEL_IPAGE_HPP
#define KERNEL_IPAGE_HPP


#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/Fields/SAPDBFields_RAMBuffer.hpp"


/*!*****************************************************************************

   class: Kernel_IPage

   description: 
   
   Interface class for all page handler classes in the kernel.

   Pure virtual methods defined by Kernel_IPage

   - CheckAfterRead ()
   - DumpTo()
   - GetFramePtr()
   - PrepareForWrite()
   - Verify()
   - GetHeaderSize()
   - GetTrailerSize()


*******************************************************************************/
class Kernel_IPage
{

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : PageFrame
    description:  Memory frame which contains the page
    --------------------------------------------------------------------------*/
    typedef SAPDBFields_RAMBuffer                           PageFrame;

    /*!-------------------------------------------------------------------------
    declaration : PageFramePtr
    description:  Pointer to memory block containing the proper page
    --------------------------------------------------------------------------*/
    typedef SAPDBFields_RAMBuffer::Pointer                  PageFramePtr;

    /*!-------------------------------------------------------------------------
    declaration: SizeType
    description: size of memory block on a page in bytes
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt2                                     SizeType;

    /*!-------------------------------------------------------------------------
    declaration: OffsetType
    description: offset on a page in bytes
    --------------------------------------------------------------------------*/
    typedef SizeType                                        OffsetType;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Interface                                                    */
    /*!-------------------------------------------------------------------------
    function:     CheckAfterRead()
    description:  Prepares page for input from device

                  - Performs parity check on the frame contents
                  - Recalculates parity from frame contents and compares
                    the result with parity bytes kept in page header and trailer
                  - Applicable immediately after a page is read from device
    arguments:    none
    return:       (SAPDB_Bool) true if check successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool CheckAfterRead () const = 0;

   
    /*!-------------------------------------------------------------------------
    function:     DumpTo()
    description:  Dumps frame contents to destination frame
    arguments:    dest [in]   dest frame
    return value: none
    --------------------------------------------------------------------------*/
    virtual void DumpTo 
        (PageFrame&                     dest) const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetFramePtr()
    description:  Returns the pointer to the proper page frame
    arguments:    none
    return value: (PageFramePtr)
    --------------------------------------------------------------------------*/
    virtual PageFramePtr GetFramePtr() = 0;

    /*!-------------------------------------------------------------------------
    function:     PrepareForWrite()
    description:  Prepares page for output to device

                  - Reads and check parity information of frame contents
                  - Updates parity information with actual frame contents
                  - Applicable immediately before a page is written to device
    arguments:    none
    return:       (SAPDB_Bool) true if check successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool PrepareForWrite() = 0;

    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       (SAPDB_Bool) true if verify successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Interface                                                 */
    /* -----------------------------------------------------------------------*/

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Interface (protected)                                        */
    /*!-------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/
    virtual SizeType GetHeaderSize() const = 0;

    /*!-------------------------------------------------------------------------
    function:     GetTrailerSize()
    description:  Returns size of page trailer
    arguments:    none
    return value: (SizeType) trailer size
    --------------------------------------------------------------------------*/
    virtual SizeType GetTrailerSize() const = 0;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Interface (protected)                                     */
    /* -----------------------------------------------------------------------*/

};
/*!*****************************************************************************

   endclass: Kernel_IPage

*******************************************************************************/


#endif // KERNEL_IPAGE_HPP


