  /*!***************************************************************************

  module:       Kernel_OpMsg.hpp

  ------------------------------------------------------------------------------

  responsible:  ThomasA

  special area: Kernel | Debug

  description:  Stream to write formatted knldiag entries

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000

    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

  *****************************************************************************/


#ifndef KERNEL_OPMSG_HPP
#define KERNEL_OPMSG_HPP

#include "ggg00.h"
#include "gsp03.h"
#include "SAPDBCommon/SAPDB_OStream.hpp"

/*!*********************************************************************

  Class: Kernel_OpMsg
  description: stream interface to knldiag.

    Stream objects are instantiated on the stack rather than
    shared between tasks.

    A new line is written to the vtrace when

    - a '\n' character is found in the stream buffer
    - the buffer size is exceeded
    - the stream object is destructed

    Warning: For performance reasons, strings should be formatted using
    the classes <Class>SAPDB_LeftJ</> and <Class>SAPDB_RightJ</>. Unlike
    <Class>SAPDB_ToString</>, these classes will not copy the string, but
    rather keep only the pointer.

    Warning: Never use multiple temporary Kernel_VTrace () objects in
    the same scope. They will be destructed (and flushed) in the reverse
    order, thus showing all output lines in the reverse order.

  example:
    // write two lines to the vtrace
    Kernel_VTrace () << "line1" << NewLine << line2;
    // the second new line is implicit when the stream is destructed

 **********************************************************************/
class Kernel_OpMsg
    : public SAPDB_OStream
{
protected:
    Kernel_OpMsg (tsp3_priority_Enum prio,tsp3_msg_type_Enum type, int msgNo, const char* label);
    ~Kernel_OpMsg ();
private:
    /*******************************************************************

      Class: LineBuffer

     ******************************************************************/
    class LineBuffer
        : public SAPDB_StreamBuffer
    {
    public:
        LineBuffer (tsp3_priority_Enum prio,tsp3_msg_type_Enum type, int msgNo, const char* label);
        virtual void Overflow ();
    private:
		char               m_txt[40];
		tsp3_priority_Enum m_prio;
        tsp3_msg_type_Enum m_type;
        int                m_no;
        tsp00_C8           m_label;
    };

    /* EndClass: LineBuffer */

private:
    LineBuffer    m_buf;
};

/*! EndClass: Kernel_OpMsg */

/*!*********************************************************************

  Class: Kernel_OpError
  description: stream interface to write errors into knldiag and knldiag.err

  example:
    // write two lines to knldiag.err
    Kernel_OpError (csp3_no_more_memory, csp3_n_memory) << "line1" << NewLine << line2;
    // the second new line is implicit when the stream is destructed

 **********************************************************************/

class Kernel_OpError : public Kernel_OpMsg
{
public :
	Kernel_OpError(int msgNo, const char* label)
		: Kernel_OpMsg(sp3p_knldiag, sp3m_error, msgNo, label)
	{
	}
};

/*!*********************************************************************

  Class: Kernel_OpInfo
  description: stream interface to write info into knldiag

 **********************************************************************/

class Kernel_OpInfo : public Kernel_OpMsg
{
public :
	Kernel_OpInfo(const char* label)
		: Kernel_OpMsg(sp3p_knldiag, sp3m_info, csp3_msg, label)
	{
	}
};

/*! EndClass: Kernel_OpError */

/*!*********************************************************************

  Class: Kernel_OpWarning
  description: stream interface to write warnings into knldiag 

  example:
    // write two lines to knldiag.err
    Kernel_OpWarning (csp3_no_more_memory, csp3_n_memory) << "line1" << NewLine << line2;
    // the second new line is implicit when the stream is destructed

 **********************************************************************/

class Kernel_OpWarning : public Kernel_OpMsg
{
public :
	Kernel_OpWarning(int msgNo, const char* label)
		: Kernel_OpMsg(sp3p_knldiag, sp3m_warning, msgNo, label)
	{
	}
};

/*! EndClass: Kernel_OpWarning */

#endif // KERNEL_OPMSG_HPP
