/*!*****************************************************************************

  module:       Kernel_Page.hpp

  ------------------------------------------------------------------------------

  responsible:  Till Luhmann

  special area: SAPDB Kernel

  description:  Kernel page handler

                - Kernel_Page is a handle to a frame containing a page 
                  in memory.

  see also:     Kernel_IPage.html 

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef KERNEL_PAGE_HPP
#define KERNEL_PAGE_HPP

#include "ggg00.h"

#include "SAPDBCommon/SAPDB_PersistentTypes.hpp"
#include "KernelCommon/Kernel_IPage.hpp"


/*!*****************************************************************************

   class: Kernel_Page

   description: 
   
   Virtual base class for all page handler classes in the kernel.

   Defines frame handling and access to some header fields.

   Yet undefined pure virtual methods inherited from Kernel_IPage

   - CheckAfterRead()
   - GetFramePtr()
   - PrepareForWrite()
   - Verify()
   - GetHeaderSize()
   - GetTrailerSize()

*******************************************************************************/
class Kernel_Page : public Kernel_IPage
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : MultiUse1
    description:  Field for multiple usage by derived classes
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt4                                     MultiUse1;

    /*!-------------------------------------------------------------------------
    declaration : MultiUse2
    description:  Field for multiple usage by derived classes
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt1                                     MultiUse2;

    /*!-------------------------------------------------------------------------
    declaration : PageType
    description:  Constants for different page categories
    --------------------------------------------------------------------------*/
    enum PageType 
    {
        Kernel_KernelPageType               = ptNil_egg00,      // these are
        Kernel_DataPageType                 = ptData_egg00,     // 7.2 page types
        Kernel_DumpPageType                 = ptDump_egg00,
        Kernel_FBMPageType                  = ptFbm_egg00,
        Kernel_LogPageType                  = ptLog_egg00,
        Kernel_SavePageType                 = ptSave_egg00,
        Kernel_SysPageType                  = ptSys_egg00,
        Kernel_VTracePageType               = ptVtrace_egg00,
        Kernel_ConverterPageType            = ptConverter_egg00,
        Kernel_ConverterBitMapPageType      = ptConverterBitMap_egg00,
        Kernel_LogInfoPageType              = ptLogInfo_egg00,
        Kernel_IOManInfoPageType            = ptIOManInfo_egg00 // change ReadPageTypeName() too
    };

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Kernel_Page
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Kernel_Page();
   
    /*!-------------------------------------------------------------------------
    function:     Kernel_Page
    description:  Constructor; assigns frame

                  - Does not change frame contents.
    arguments:    frame [in]    kernel page frame
    return value: none
    --------------------------------------------------------------------------*/
    Kernel_Page
        (const PageFrame&               frame);
   
    /*!-------------------------------------------------------------------------
    function:     SetFrame()
    description:  Assigns frame; does not change frame contents 
    arguments:    frame [in]   kernel page frame
    return value: none
    --------------------------------------------------------------------------*/
    void SetFrame
        (const PageFrame&               frame);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /*!-------------------------------------------------------------------------
    function:     ReleaseFrame()
    description:  Deassigns page frame
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void ReleaseFrame();

    /*!-------------------------------------------------------------------------
    function:     GetFrame()
    description:  Returns reference to actual frame
    arguments:    none
    return value: (const PageFrame&) reference to frame
    --------------------------------------------------------------------------*/
    const PageFrame& GetFrame() const
    {
        return m_Frame;
    }

    PageFrame& GetFrame()
    {
        return m_Frame;
    }

    /*!-------------------------------------------------------------------------
    function:     GetFramePtr()
    description:  Returns the pointer to the proper page frame
    arguments:    none
    return value: (PageFramePtr)
    --------------------------------------------------------------------------*/
    virtual PageFramePtr GetFramePtr()
    {
        return( m_Frame.GetPtrAt( 0, m_Frame.Capacity() ));
    }

    /*!-------------------------------------------------------------------------
    function:     IsAssigned()
    description:  Checks whether the page frame is assigned
    arguments:    none
    return value: (SAPDB_Bool)
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsAssigned() const;

    /*!-------------------------------------------------------------------------
    function:     ReadPageType()
    description:  Reads page type from frame
    arguments:    none
    return value: (PageType) page type
    --------------------------------------------------------------------------*/
    PageType ReadPageType() const;

    /*!-------------------------------------------------------------------------
    function:     ReadPageTypeName()
    description:  Reads page type from frame
    arguments:    none
    return value: page type as string with null termination
    --------------------------------------------------------------------------*/

    const SAPDB_Char* ReadPageTypeName() const
    {
        switch( ReadPageType())
        {
        case Kernel_KernelPageType:
            return ( SAPDB_Char* )"nil";
        case Kernel_DataPageType:
            return ( SAPDB_Char* )"data";
        case Kernel_DumpPageType:
            return ( SAPDB_Char* )"dump";
        case Kernel_FBMPageType:
            return ( SAPDB_Char* )"fbm";
        case Kernel_LogPageType:
            return ( SAPDB_Char* )"log";
        case Kernel_SavePageType:
            return ( SAPDB_Char* )"save";
        case Kernel_SysPageType:
            return ( SAPDB_Char* )"sys";
        case Kernel_VTracePageType:
            return ( SAPDB_Char* )"vtrace";
        case Kernel_ConverterPageType:
            return ( SAPDB_Char* )"converter";
        case Kernel_ConverterBitMapPageType:
            return ( SAPDB_Char* )"conv bit map";
        case Kernel_LogInfoPageType:
            return ( SAPDB_Char* )"log info";
        case Kernel_IOManInfoPageType:
            return ( SAPDB_Char* )"ioman info";
        default:
            return ( SAPDB_Char* )"UNDEF";
        }
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Kernel_IPage interface                                       */
    /*!-------------------------------------------------------------------------
    function:     DumpTo()
    description:  Dumps frame contents to destination frame
    arguments:    dest [in]   dest frame
    return value: none
    --------------------------------------------------------------------------*/
    virtual void DumpTo
        (PageFrame&                     dest) const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Kernel_IPage interface                                    */
    /* -----------------------------------------------------------------------*/

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#else
protected:
#endif

    /*--------------------------------------------------------------------------
    declaration: Persistent types
    description: Persistent type definitions with compile time size check
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent         <MultiUse1,        SAPDB_UInt4, 4> 
                                      PersistentMultiUse1;
    typedef SAPDB_PersistentBitRange <PageType,         SAPDB_Byte,  1> 
                                      PersistentPageType;
    typedef SAPDB_Persistent         <MultiUse2,        SAPDB_UInt1, 1> 
                                      PersistentMultiUse2;

    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within kernel page header
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):
    typedef PersistentMultiUse1::PersistentFormat           PH1;
    typedef PersistentPageType::PersistentFormat            PH2;
    typedef PersistentMultiUse2::PersistentFormat           PH3;
    typedef SAPDB_UInt2                                     PH4;

protected:

    struct PersistentHeader
    {
        PH1     m_MultiUse1;    // different usage: converter version, data pno, etc.
        PH2     m_PageType;
        PH3     m_MultiUse2;    // 7.4: parity algorithm; 7.3: page type 2
        PH4     m_Filler;
    };
    
    /*--------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/
    virtual SizeType GetHeaderSize() const;

private:

    /*--------------------------------------------------------------------------
    declaration: KernelPagePersistentHeader
    description: Compile time size check for header
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent 
        < PersistentHeader, 
          PersistentHeader,
          sizeof(PH1) +
          sizeof(PH2) +
          sizeof(PH3) +
          sizeof(PH4) >
        CheckedPersistentHeader;

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access (protected)                                     */
    /*!-------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    prototypes:   2
    return value: (PersistentHeader*) header pointer
    --------------------------------------------------------------------------*/
    PersistentHeader* PersistentHeaderPtr();

    const PersistentHeader* PersistentHeaderPtr() const;

    /*!-------------------------------------------------------------------------
    function:     WriteMultiUse1()
    description:  Writes multi use value to frame
    arguments:    value [in]  value to be set
    return value: none
    --------------------------------------------------------------------------*/
    void WriteMultiUse1
        (MultiUse1                      value);

    /*!-------------------------------------------------------------------------
    function:     ReadMultiUse1()
    description:  Reads multi use value from frame
    arguments:    none
    return value: (MultiUse1) multi use value
    --------------------------------------------------------------------------*/
    MultiUse1 ReadMultiUse1() const;

    /*!-------------------------------------------------------------------------
    function:     WriteMultiUse2()
    description:  Writes multi use value to frame
    arguments:    value [in]  value to be set
    return value: none
    --------------------------------------------------------------------------*/
    void WriteMultiUse2
        (MultiUse2                      value);

    /*!-------------------------------------------------------------------------
    function:     ReadMultiUse2()
    description:  Reads multi use value from frame
    arguments:    none
    return value: (MultiUse2) multi use value
    --------------------------------------------------------------------------*/
    MultiUse2 ReadMultiUse2() const;

    /*!-------------------------------------------------------------------------
    function:     WritePageType()
    description:  Writes page type to frame
    arguments:    type [in]  Page type to be set
    return value: none
    --------------------------------------------------------------------------*/
    void WritePageType
        (PageType                       type);

    /*!-------------------------------------------------------------------------
    function:     SetHint()
    description:  Sets pointer to associated hint
    arguments:    hint [in]   hint pointer to be set
    return value: none
    --------------------------------------------------------------------------*/
    void SetHint
        (void*                          hint);

    /*!-------------------------------------------------------------------------
    function:     GetHint()
    description:  Returns pointer to associated hint
    arguments:    none
    return value: (void *) pointer to hint
    --------------------------------------------------------------------------*/
    void* GetHint() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access (protected)                                  */
    /* -----------------------------------------------------------------------*/

    /*--------------------------------------------------------------------------
    declaration: m_Frame
    description: handle to frame which contains actual page
    --------------------------------------------------------------------------*/
    PageFrame m_Frame;

private:

    /*--------------------------------------------------------------------------
    declaration: m_Hint
    description: Hint pointer to additional data (for i/o purposes)
    --------------------------------------------------------------------------*/
    void *m_Hint;

};
/*!*****************************************************************************

    endclass: Kernel_Page

*******************************************************************************/


//------------------------------------------------------------------------------

inline Kernel_Page::Kernel_Page()
    : 
    m_Hint(0)
{}

//------------------------------------------------------------------------------

inline Kernel_Page::Kernel_Page
    (const PageFrame&               frame)
    : 
    m_Frame(frame), 
    m_Hint (0)
{}

//------------------------------------------------------------------------------

inline void Kernel_Page::SetFrame
    (const PageFrame&               frame)
{
    m_Frame = frame;
}

//------------------------------------------------------------------------------

inline void Kernel_Page::ReleaseFrame()
{
    m_Frame.Deassign();
}

//------------------------------------------------------------------------------

inline SAPDB_Bool Kernel_Page::IsAssigned
    () const
{
    return m_Frame.IsAssigned();
}

//------------------------------------------------------------------------------

inline Kernel_Page::PageType Kernel_Page::ReadPageType
    () const
{
    return PersistentPageType::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_PageType);
}

//------------------------------------------------------------------------------

inline void Kernel_Page::DumpTo
    (PageFrame&                     dest) const 
{
    dest.SetFieldAt(0, m_Frame.ToField());
}

//------------------------------------------------------------------------------

inline Kernel_Page::SizeType Kernel_Page::GetHeaderSize
    () const
{
    return sizeof(PersistentHeader);
}

//------------------------------------------------------------------------------

inline Kernel_Page::PersistentHeader* Kernel_Page::PersistentHeaderPtr
    ()
{
    return 
    reinterpret_cast<PersistentHeader*>
    (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}


//------------------------------------------------------------------------------

inline const Kernel_Page::PersistentHeader* Kernel_Page::PersistentHeaderPtr
    () const
{
    return 
    reinterpret_cast<const PersistentHeader*>
        (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}


//------------------------------------------------------------------------------

inline void Kernel_Page::WriteMultiUse1
    (MultiUse1                         value)
{
    PersistentHeaderPtr()->m_MultiUse1 = 
    PersistentMultiUse1::ConvertInternalToPersistent(value);
}

//------------------------------------------------------------------------------

inline Kernel_Page::MultiUse1 Kernel_Page::ReadMultiUse1
    () const
{
    return PersistentMultiUse1::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_MultiUse1);
}

//------------------------------------------------------------------------------

inline void Kernel_Page::WriteMultiUse2
    (MultiUse2                         value)
{
    PersistentHeaderPtr()->m_MultiUse2 = 
    PersistentMultiUse2::ConvertInternalToPersistent(value);
}

//------------------------------------------------------------------------------

inline Kernel_Page::MultiUse2 Kernel_Page::ReadMultiUse2
    () const
{
    return PersistentHeaderPtr()->m_MultiUse2;
}

//------------------------------------------------------------------------------

inline void Kernel_Page::WritePageType
    (PageType                       type)
{
    PersistentHeaderPtr()->m_PageType = 
    PersistentPageType::ConvertInternalToPersistent(type);
}

//------------------------------------------------------------------------------

inline void Kernel_Page::SetHint
    (void *hint)
{
    m_Hint = hint;
}

//------------------------------------------------------------------------------

inline void* Kernel_Page::GetHint
    () const
{
    return m_Hint;
}

//------------------------------------------------------------------------------




#endif // KERNEL_PAGE_HPP
