/*!*****************************************************************************

  module:       Kernel_Page72.hpp

  ------------------------------------------------------------------------------

  responsible:  Till Luhmann
                TorstenS

  special area: SAPDB Kernel

  description:  page handler base class
                
                The class kernel_page72 is used to encapsulate the old
                pages defined with pascal strucures.
                
  see also:     Kernel_IPage.html 
                Kernel_Page.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef KERNEL_PAGE72_HPP
#define KERNEL_PAGE72_HPP

#include "KernelCommon/Kernel_Page.hpp"
#include "KernelCommon/Kernel_Exception.hpp"
#include "KernelCommon/Kernel_Messages.hpp"  
#include "RunTime/RTE_Message.hpp"
#include "SAPDBCommon/SAPDB_PersistentTypes.hpp"


/*!*****************************************************************************

   class: Kernel_Page72

   description: 
   
   Base class for all 7.2 page handler classes.

   Specializes Kernel_Page in using the "multi use field 1" as a "page id"
   and the "multi use field 2" for storing the "page type 2" identifier.

   Yet undefined pure virtual methods inherited from Kernel_IPage

   - CheckAfterRead ()
   - PrepareForWrite()

*******************************************************************************/

class Kernel_Page72 : public Kernel_Page
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : PageID
    description:  individual page identifier
    --------------------------------------------------------------------------*/

    typedef SAPDB_Int4                                      PageID;

    /*!-------------------------------------------------------------------------
    declaration : PageType2
    description:  Constants for different secondary page categories
    --------------------------------------------------------------------------*/
    
    typedef tgg00_PageType2_Enum                            PageType2;

    /*!-------------------------------------------------------------------------
    declaration : PageCheck
    description:  Constants for different page check algorithms
    --------------------------------------------------------------------------*/

    typedef tgg00_PageCheckType_Enum                        PageCheck;

    /*!-------------------------------------------------------------------------
    declaration : PageMode
    description:  Constants for detail information about the page
    --------------------------------------------------------------------------*/

    typedef tgg00_PageMode__Enum                            PageMode;

    /*!-------------------------------------------------------------------------
    declaration : PageChecksum
    description:  Result of parity algorithmen
    --------------------------------------------------------------------------*/

    typedef SAPDB_Int4                                      PageChecksum;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Kernel_Page72
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    
    Kernel_Page72(){}
   
    /*!-------------------------------------------------------------------------
    function:     Kernel_Page72
    description:  Constructor; assigns frame

                  - Does not change frame contents.
    arguments:    frame [in]    kernel page frame
    return value: none
    --------------------------------------------------------------------------*/
    
    Kernel_Page72( const PageFrame &frame): Kernel_Page( frame ){}

    /*!-------------------------------------------------------------------------
    function:     Kernel_Page72
    description:  Constructor; assigns frame and initializes frame contents
    arguments:    frame [in]   kernel page frame
                  type  [in]   page type
                  type2 [in]   page type 2
    return value: none
    --------------------------------------------------------------------------*/
    
    Kernel_Page72(
        const PageFrame &frame, 
        PageType        type,
        PageType2       type2
        )
        :Kernel_Page( frame ) 
    {
        InitializeFrame( type, type2 );
    }
   
    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame [in]   kernel page frame
                  type  [in]   page type
                  type2 [in]   page type 2
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetAndInitializeFrame(
        const PageFrame    &frame, 
        PageType           type,
        PageType2          type2)
    {
        Kernel_Page::SetFrame( frame );
        InitializeFrame( type, type2 );
    }

    /*!-------------------------------------------------------------------------
    function:     InitializeFrame()
    description:  Initializes frame contents
    arguments:    type  [in]   page type
                  type2 [in]   page type 2
    return value: none
    --------------------------------------------------------------------------*/
    
    void InitializeFrame(
        PageType    type,
        PageType2   type2 )
    {
        WritePageType( type );
        SetPageType2( type2 );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Kernel_IPage interface                                       */
    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       (SAPDB_Bool) true if verify successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const
    {
        // dummy TODO
        return true;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Kernel_IPage interface                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetPageID()
    description:  Reads page id from frame
    arguments:    none
    return value: (PageID) page id
    --------------------------------------------------------------------------*/
    
    PageID GetPageID() const
    {
        return PersistentPageID::ConvertPersistentToInternal
            (PersistentHeaderPtr()->m_PageID);
    }

    /*!-------------------------------------------------------------------------
    function:     ReadPageType2()
    description:  Reads parity check algorithm number from frame
    arguments:    none
    return value: (PageType2) algorithm number
    --------------------------------------------------------------------------*/
    
    PageType2 ReadPageType2() const
    {
        return PersistentPageType2::ConvertPersistentToInternal
            (PersistentHeaderPtr()->m_PageType2);
    }

    /*!-------------------------------------------------------------------------
    function:     ResetPageType2()
    description:  Changes page type2 in frame
    arguments:    type2 [in]  page type2 to be set
    return value: none
    --------------------------------------------------------------------------*/
    
    void ResetPageType2( PageType2  type2 )
    {
        PersistentHeaderPtr()->m_PageType2 = 
            PersistentPageType2::ConvertInternalToPersistent( type2 );
        CopyHeaderToTrailer();
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#else
private:
#endif

    /*--------------------------------------------------------------------------
    declaration: Persistent types
    description: Persistent type definitions with compile time size check
    --------------------------------------------------------------------------*/
    
    typedef SAPDB_Persistent         <PageID,           MultiUse1, 4> 
                                      PersistentPageID;
    
    typedef SAPDB_PersistentBitRange <PageType2,        MultiUse2, 1> 
                                      PersistentPageType2;

    typedef SAPDB_PersistentBitRange <PageCheck,        SAPDB_UInt1, 1>
                                      PersistentPageCheck;

    typedef SAPDB_Persistent         <PageMode,         SAPDB_UInt1, 1>
                                      PersistentPageMode;

    typedef SAPDB_Persistent         <PageChecksum,     SAPDB_UInt4, 4> 
                                      PersistentPageChecksum;

    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within kernel page header
    --------------------------------------------------------------------------*/

    // Must do this redefinition to avoid redundancy (compiler reason):
    typedef PersistentPageID::PersistentFormat              PH1;
    typedef Kernel_Page::PH2                                PH2;
    typedef PersistentPageType2::PersistentFormat           PH3;
    typedef PersistentPageCheck::PersistentFormat           PH4;
    typedef PersistentPageMode::PersistentFormat            PH5;

protected:

    struct PersistentHeader
    {
        PH1     m_PageID;
        PH2     m_PageType;
        PH3     m_PageType2;
        PH4     m_PageCheck;
        PH5     m_PageMode;
    };

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#endif

    /*--------------------------------------------------------------------------
    declaration: PersistentChecksumTrailer
    description: Persistent layout of data within expand kernel page trailer
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):

    typedef PersistentPageChecksum::PersistentFormat        PT1;

    typedef struct {
        PH1     m_PageId;
        PH2     m_TrailerPageType;
        PH3     m_TrailerPageType2;
        PH4     m_TrailerPageCheck;
        PH5     m_TrailerPageMode;
    } PersistentTrailer;

protected:

    struct PersistentChecksumTrailer
    {
        PT1                  m_PageChecksum;
        PersistentTrailer    m_TrailerPageMode;
    };

private:

    /*--------------------------------------------------------------------------
    declaration: KernelPagePersistentHeader
    description: Compile time size check for header
    --------------------------------------------------------------------------*/
    
    typedef SAPDB_Persistent 
        < PersistentHeader, 
          PersistentHeader,
          sizeof(PH1) +
          sizeof(PH2) +
          sizeof(PH3) +
          sizeof(PH4) +
          sizeof(PH5)
        >
        CheckedPersistentHeader;

    /*--------------------------------------------------------------------------
    declaration: KernelPagePersistentTrailer
    description: Compile time size check for trailer
    --------------------------------------------------------------------------*/
    
    typedef SAPDB_Persistent 
        < PersistentChecksumTrailer, 
          PersistentChecksumTrailer,
          sizeof(PT1) +
          sizeof(PH1) +
          sizeof(PH2) +
          sizeof(PH3) +
          sizeof(PH4) +
          sizeof(PH5)
         >
        CheckedPersistentTrailer;

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access (protected)                                     */
    /*!-------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/
    
    virtual SizeType GetHeaderSize() const
    {
        return sizeof( PersistentHeader );
    }

    /*!-------------------------------------------------------------------------
    function:     GetTrailerSize()
    description:  Returns size of page trailer, without the checksum value
    arguments:    none
    return value: (SizeType) trailer size
    --------------------------------------------------------------------------*/
    
    virtual SizeType GetTrailerSize() const
    {
        return sizeof( PersistentTrailer );
    }

    /*!-------------------------------------------------------------------------
    function:     CheckHeaderTrailer()
    description:  compares the header with the trailer
    arguments:    none
    return value: (SAPDB_Bool) true if header equal trailer
    --------------------------------------------------------------------------*/
    
    SAPDB_Bool CheckHeaderTrailer() const
    {
        if( memcmp( Kernel_Page72::PersistentHeaderPtr(), 
            Kernel_Page72::PersistentTrailerPtr(), 
            Kernel_Page72::GetHeaderSize())
            )
        {
            Kernel_Exception ErrMsg( __CONTEXT__, KERNEL_ERR_HEADER_TRAILER_EQUALITY );
            RTE_Message( ErrMsg );

            return( SAPDB_FALSE );
        }
        return( SAPDB_TRUE );
    }

    /*!-------------------------------------------------------------------------
    function:     CopyHeaderToTrailer()
    description:  copy the header onto the trailer
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    
    void CopyHeaderToTrailer()
    {
        memcpy( Kernel_Page72::PersistentTrailerPtr(),
            Kernel_Page72::PersistentHeaderPtr(),
            Kernel_Page72::GetHeaderSize());
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageID()
    description:  Writes page id to frame
    arguments:    pageid [in]  Page id to be set
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetPageID( PageID    pageid )
    {
        PersistentHeaderPtr()->m_PageID = 
            PersistentPageID::ConvertInternalToPersistent(pageid);
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageType2()
    description:  Writes page type2 to frame
    arguments:    type2 [in]  page type2 to be set
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetPageType2( PageType2  type2 )
    {
        PersistentHeaderPtr()->m_PageType2 = 
            PersistentPageType2::ConvertInternalToPersistent( type2 );
    }

    /*!-------------------------------------------------------------------------
    function:     GetPageCheck()
    description:  Reads page check algorithmen id from frame
    arguments:    none
    return value: (PageCheck) page check algorithmen id
    --------------------------------------------------------------------------*/

    PageCheck GetPageCheck() const
    {
        return PersistentPageCheck::ConvertPersistentToInternal
            (PersistentHeaderPtr()->m_PageCheck);
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageCheck()
    description:  Writes page check algorithmen id to frame
    arguments:    check [in]  page check algorithmen id to be set
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetPageCheck( PageCheck  check )
    {
        PersistentHeaderPtr()->m_PageCheck = 
            PersistentPageCheck::ConvertInternalToPersistent( check );
    }
 
    /*!-------------------------------------------------------------------------
    function:     GetPageMode()
    description:  Reads page mode from frame
    arguments:    none
    return value: (PageMode) page mode
    --------------------------------------------------------------------------*/

    PageMode GetPageMode() const
    {
        return PersistentPageMode::ConvertPersistentToInternal
            (PersistentHeaderPtr()->m_PageMode);
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageMode()
    description:  Writes page mode to frame
    arguments:    mode [in]  page mode to be set
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetPageMode( PageMode  mode )
    {
        PersistentHeaderPtr()->m_PageMode = 
            PersistentPageMode::ConvertInternalToPersistent( mode );
    }
     /*!-------------------------------------------------------------------------
    function:     GetPageChecksum()
    description:  Reads page checksum value from frame
    arguments:    none
    return value: (PageChecksum) page checksum
    --------------------------------------------------------------------------*/

    PageChecksum GetPageChecksum() const
    {
        return PersistentPageChecksum::ConvertPersistentToInternal
            (PersistentChecksumTrailerPtr()->m_PageChecksum);
    }

    /*!-------------------------------------------------------------------------
    function:     SetPageMode()
    description:  Writes page mode to frame
    arguments:    mode [in]  page mode to be set
    return value: none
    --------------------------------------------------------------------------*/
    
    void SetPageChecksum( PageChecksum  Checksum )
    {
        PersistentChecksumTrailerPtr()->m_PageChecksum = 
            PersistentPageChecksum::ConvertInternalToPersistent( Checksum );
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access (protected)                                  */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    prototypes:   2
    return value: (PersistentHeader*) header pointer
    --------------------------------------------------------------------------*/
    
    PersistentHeader* PersistentHeaderPtr()
    {
        return 
            REINTERPRET_CAST( PersistentHeader*,
            ( m_Frame.GetPtrAt( 0, sizeof( PersistentHeader ))));
    }
    
    const PersistentHeader* PersistentHeaderPtr() const
    {
        return 
            REINTERPRET_CAST( const PersistentHeader*,
            ( m_Frame.GetPtrAt( 0, sizeof(PersistentHeader ))));
    }

    /*!-------------------------------------------------------------------------
    function:     PersistentTrailerPtr()
    description:  Returns pointer to page trailer (without checksum) on frame
    arguments:    none
    prototypes:   2
    return value: (PersistentTrailer*) trailer pointer
    --------------------------------------------------------------------------*/
    
    PersistentTrailer* PersistentTrailerPtr()
    {
        return 
            REINTERPRET_CAST( PersistentTrailer*,
            ( m_Frame.GetPtrAt
            ( m_Frame.Capacity() - sizeof( PersistentTrailer ), 
            sizeof( PersistentTrailer ))));
    }
    
    const PersistentTrailer* PersistentTrailerPtr() const
    {
        return 
            REINTERPRET_CAST( const PersistentTrailer*,
            ( m_Frame.GetPtrAt
            ( m_Frame.Capacity() - sizeof( PersistentTrailer ), 
            sizeof( PersistentTrailer ))));
    }

     /*!-------------------------------------------------------------------------
    function:     PersistentChecksumTrailerPtr()
    description:  Returns pointer to page trailer (without checksum) on frame
    arguments:    none
    prototypes:   2
    return value: (PersistentTrailer*) trailer pointer
    --------------------------------------------------------------------------*/
    
    PersistentChecksumTrailer* PersistentChecksumTrailerPtr()
    {
        return 
            REINTERPRET_CAST( PersistentChecksumTrailer*,
            ( m_Frame.GetPtrAt
            ( m_Frame.Capacity() - sizeof( PersistentChecksumTrailer ), 
            sizeof( PersistentChecksumTrailer ))));
    }
    
    const PersistentChecksumTrailer* PersistentChecksumTrailerPtr() const
    {
        return 
            REINTERPRET_CAST( const PersistentChecksumTrailer*,
            ( m_Frame.GetPtrAt
            ( m_Frame.Capacity() - sizeof( PersistentChecksumTrailer ), 
            sizeof( PersistentChecksumTrailer ))));
    }
};

/*!*****************************************************************************

    endclass: Kernel_Page72

*******************************************************************************/


#endif // KERNEL_PAGE72_HPP
