/*!*****************************************************************************

  module:       Kernel_Page74.hpp

  ------------------------------------------------------------------------------

  responsible : TorstenS

  special area: SAPDB Kernel

  description : 7.4 page handler base class
                
  see also    : Kernel_IPage.html 
                Kernel_Page.html
 
  author     :  TillL

  copyright  : (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef KERNEL_PAGE74_HPP
#define KERNEL_PAGE74_HPP


#include "KernelCommon/Kernel_Page.hpp"
#include "KernelCommon/Kernel_Exception.hpp"
#include "KernelCommon/Kernel_Messages.hpp"  
#include "RunTime/RTE_Message.hpp"


/*!*****************************************************************************

   class: Kernel_Page74

   description: 
   
   Base class for all 7.4 page handler classes.

   Specializes Kernel_Page in using the "multi use field 2" for storing the 
   "parity algorithm" identifier.

   Yet undefined pure virtual methods inherited from Kernel_IPage

   - CheckAfterRead ()
   - PrepareForWrite()

*******************************************************************************/

class Kernel_Page74 : public Kernel_Page
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : ParityAlgorithm
    description:  Internal page parity check algorithm number
    --------------------------------------------------------------------------*/

    enum ParityAlgorithm
    {
        Kernel_CheckSumAlgorithmen    = 1,
        Kernel_HeaderTrailerAlgorithm = 2
    };

    /*!-------------------------------------------------------------------------
    declaration : Parity
    description:  Internal page parity check result type
    --------------------------------------------------------------------------*/

    typedef SAPDB_UInt2 Parity;

    /*!-------------------------------------------------------------------------
    declaration : MultiUseT1
    description:  Trailer field for multiple usage by derived classes
    --------------------------------------------------------------------------*/

    typedef SAPDB_UInt2 MultiUseT1;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Kernel_Page74
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    Kernel_Page74()
    {}

    /*!-------------------------------------------------------------------------
    function:     Kernel_Page74
    description:  Constructor; assigns frame

                  - Does not change frame contents.
    arguments:    frame [in]    kernel page frame
    return value: none
    --------------------------------------------------------------------------*/

    Kernel_Page74( const PageFrame   &frame ):Kernel_Page( frame )
    {}

    /*!-------------------------------------------------------------------------
    function:     Kernel_Page74
    description:  Constructor; assigns frame and initializes frame contents
    arguments:    frame [in]   kernel page frame
                  type  [in]   page type
                  algo  [in]   parity algorithm id (default: 1)
    return value: none
    --------------------------------------------------------------------------*/

    Kernel_Page74(
        const PageFrame &frame,
        PageType        type,
        ParityAlgorithm algo = Kernel_CheckSumAlgorithmen ):Kernel_Page( frame )
    {
        InitializeFrame( type, algo );
    }

    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame [in]   kernel page frame
                  type  [in]   page type
                  algo  [in]   parity algorithm id (default: 1)
    return value: none
    --------------------------------------------------------------------------*/

    void SetAndInitializeFrame(
        const PageFrame &frame,
        PageType        type,
        ParityAlgorithm algo = Kernel_CheckSumAlgorithmen )
    {
        Kernel_Page::SetFrame( frame );
        InitializeFrame( type, algo );
    }


    /*!-------------------------------------------------------------------------
    function:     InitializeFrame()
    description:  Initializes frame contents
    arguments:    type  [in]   page type
                  algo  [in]   parity algorithm id (default: 1)
    return value: none
    --------------------------------------------------------------------------*/

    void InitializeFrame(
        PageType        type,
        ParityAlgorithm algo = Kernel_CheckSumAlgorithmen )
    {
        InitializeParityRaster();
        InitFiller();
        WritePageType( type );
        WriteParityAlgorithm( algo );
        UpdateParity();
    }


    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Kernel_IPage interface                                       */
    /*!-------------------------------------------------------------------------
    function:     CheckAfterRead()
    description:  Prepares page for input from device

                  - Performs parity check on the frame contents
                  - Recalculates parity from frame contents and compares
                    the result with parity bytes kept in page header and trailer
                  - Applicable immediately after a page is read from device
    arguments:    none
    return:       (SAPDB_Bool) true if check successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool CheckAfterRead() const
    {
        const ParityAlgorithm   algorithm  = ReadParityAlgorithm();
        const Parity            parity     = CalcParity();

        switch( algorithm )
        {
        case Kernel_CheckSumAlgorithmen:
            {
                if(( parity == ReadParityFromHeader()) && ( parity == ReadParityFromTrailer()))
                    return true;
                break;
            }
        case Kernel_HeaderTrailerAlgorithm:
            {
                if( ReadParityFromHeader() == ReadParityFromTrailer() )
                    return true;
                break;
            }
        default:
            ;
        }

        Kernel_Exception errMsg( __CONTEXT__, KERNEL_ERR_PAGE_BAD_INPUT,
                                 SAPDB_ToString( algorithm, _T_d ),
                                 SAPDB_ToString( parity, _T_d ),
                                 SAPDB_ToString( ReadParityFromHeader(), _T_d ),
                                 SAPDB_ToString( ReadParityFromTrailer(), _T_d ));
        RTE_Message( errMsg );

        return false;
    }

    /*!-------------------------------------------------------------------------
    function:     PrepareForWrite()
    description:  Prepares page for output to device

                  - Reads and check parity information of frame contents
                  - Updates parity information with actual frame contents
                  - Applicable immediately before a page is written to device
    arguments:    none
    return:       (SAPDB_Bool) true if check successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool PrepareForWrite()
    {
        if( ReadParityFromHeader() != ReadParityFromTrailer() )
        {
            Kernel_Exception errMsg( __CONTEXT__, KERNEL_ERR_PAGE_BAD_OUTPUT,
                                     SAPDB_ToString( ReadParityFromHeader(), _T_d ),
                                     SAPDB_ToString( ReadParityFromTrailer(), _T_d ));
            RTE_Message( errMsg );

            return false;
        }
        UpdateParity();
        return true;
    }

    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       (SAPDB_Bool) true if verify successful
    --------------------------------------------------------------------------*/

    virtual SAPDB_Bool Verify() const
    {
        if( ReadParityFromHeader() != ReadParityFromTrailer() )
        {
            Kernel_Exception errMsg( __CONTEXT__, KERNEL_ERR_PAGE_BAD_VERIFY,
                                     SAPDB_ToString( ReadParityFromHeader(), _T_d ),
                                     SAPDB_ToString( ReadParityFromTrailer(), _T_d ));
            RTE_Message( errMsg );

            return false;
        }
        return true;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Kernel_IPage interface                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /*!-------------------------------------------------------------------------
    function:     ReadParityAlgorithm()
    description:  Reads parity check algorithm number from frame
    arguments:    none
    return value: (ParityAlgorithm) algorithm number
    --------------------------------------------------------------------------*/

    ParityAlgorithm ReadParityAlgorithm() const
    {
        return PersistentParityAlgorithm::ConvertPersistentToInternal
               (PersistentHeaderPtr()->m_Algorithm);
    }

    /*!-------------------------------------------------------------------------
    function:     CalcParity()
    description:  Calculates and returns parity of frame contents

                  - Does not perform any changes on the frame contents
                  - Uses intrinsic parity algorithm of this class
    arguments:    none
    return value: (Parity)
    --------------------------------------------------------------------------*/

    virtual Parity CalcParity() const;

    /*--------------------------------------------------------------------------
    function:     ReadParityFromHeader()
    description:  Reads parity value kept in page header from frame
    arguments:    none
    return value: (Parity) parity value
    --------------------------------------------------------------------------*/

    Parity ReadParityFromHeader() const
    {
        return PersistentHeaderPtr()->m_Parity;
    }

    /*--------------------------------------------------------------------------
    function:     ReadParityFromTrailer()
    description:  Reads parity value kept in page trailer from frame
    arguments:    none
    return value: (Parity) parity value
    --------------------------------------------------------------------------*/

    Parity ReadParityFromTrailer() const
    {
        return PersistentTrailerPtr()->m_Parity;
    }

    /*!-------------------------------------------------------------------------
    function:     WriteMultiUseT1()
    description:  Writes trailer multi use value 1 to frame
    arguments:    value [in]  value to be set
    return value: none
    --------------------------------------------------------------------------*/

    void WriteMultiUseT1( MultiUseT1    value )
    {
        PersistentTrailerPtr()->m_MultiUseT1 =
            PersistentMultiUseT1::ConvertInternalToPersistent(value);
    }

    /*!-------------------------------------------------------------------------
    function:     ReadMultiUseT1()
    description:  Reads trailer multi use value from frame
    arguments:    none
    return value: (MultiUseT1) multi use value
    --------------------------------------------------------------------------*/

    MultiUseT1 ReadMultiUseT1() const
    {
        return PersistentMultiUseT1::ConvertPersistentToInternal
               (PersistentTrailerPtr()->m_MultiUseT1);
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#else
private:
#endif

    /*--------------------------------------------------------------------------
    declaration: Persistent types
    description: Persistent type definitions with compile time size check
    --------------------------------------------------------------------------*/

    
    typedef SAPDB_PersistentBitRange< ParityAlgorithm,  SAPDB_UInt1, 1 > PersistentParityAlgorithm;

    typedef SAPDB_Persistent< Parity, SAPDB_UInt2, 2 > PersistentParity;

    typedef SAPDB_Persistent< MultiUseT1, SAPDB_UInt2, 2 > PersistentMultiUseT1;

    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within kernel page header
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):
    typedef Kernel_Page::PH1                                PH1;

    typedef Kernel_Page::PH2                                PH2;

    typedef PersistentParityAlgorithm::PersistentFormat     PH3;

    typedef PersistentParity::PersistentFormat              PH4;

protected:

    struct PersistentHeader
    {
        PH1     m_MultiUse1;
        PH2     m_PageType;
        PH3     m_Algorithm;
        PH4     m_Parity;
    };


#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#endif

    /*--------------------------------------------------------------------------
    declaration: PersistentTrailer
    description: Persistent layout of data within kernel page trailer
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):
    typedef PersistentMultiUseT1::PersistentFormat          PT1;

    typedef PersistentParity::PersistentFormat              PT2;

protected:

    struct PersistentTrailer
    {
        PT1     m_MultiUseT1;
        PT2     m_Parity;
    };

private:

    /*--------------------------------------------------------------------------
    declaration: KernelPagePersistentHeader
    description: Compile time size check for header
    --------------------------------------------------------------------------*/

    typedef SAPDB_Persistent
    < PersistentHeader,
    PersistentHeader,
    sizeof(PH1) +
    sizeof(PH2) +
    sizeof(PH3) +
    sizeof(PH4)
    >
    CheckedPersistentHeader;

    /*--------------------------------------------------------------------------
    declaration: KernelPagePersistentTrailer
    description: Compile time size check for trailer
    --------------------------------------------------------------------------*/

    typedef SAPDB_Persistent
    < PersistentTrailer,
    PersistentTrailer,
    sizeof(PT1) +
    sizeof(PT2) >
    CheckedPersistentTrailer;

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access (protected)                                     */
    /*!-------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/

    virtual SizeType GetHeaderSize() const
    {
        return sizeof(PersistentHeader);
    }

    /*!-------------------------------------------------------------------------
    function:     GetTrailerSize()
    description:  Returns size of page trailer
    arguments:    none
    return value: (SizeType) trailer size
    --------------------------------------------------------------------------*/

    virtual SizeType GetTrailerSize() const
    {
        return sizeof(PersistentTrailer);
    }

    /*!-------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    prototypes:   2
    return value: (PersistentHeader*) header pointer
    --------------------------------------------------------------------------*/

    PersistentHeader* PersistentHeaderPtr()
    {
        return
            reinterpret_cast<PersistentHeader*>
            (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
    }

    const PersistentHeader* PersistentHeaderPtr() const
    {
        return
            reinterpret_cast<const PersistentHeader*>
            (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
    }

    /*!-------------------------------------------------------------------------
    function:     PersistentTrailerPtr()
    description:  Returns pointer to page trailer on frame
    arguments:    none
    prototypes:   2
    return value: (PersistentTrailer*) trailer pointer
    --------------------------------------------------------------------------*/

    PersistentTrailer* PersistentTrailerPtr()
    {
        return
            reinterpret_cast<PersistentTrailer*>
            (m_Frame.GetPtrAt
             (m_Frame.Capacity() - sizeof(PersistentTrailer),
              sizeof(PersistentTrailer)));
    }

    const PersistentTrailer* PersistentTrailerPtr() const
    {
        return
            reinterpret_cast<const PersistentTrailer*>
            (m_Frame.GetPtrAt
             (m_Frame.Capacity() - sizeof(PersistentTrailer),
              sizeof(PersistentTrailer)));
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access (protected)                                  */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     UpdateParity()
    description:  Calculates parity depending on frame contents

                  - Writes parity to frame
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void UpdateParity()
    {
        Parity value = CalcParity();

        PersistentHeader*  h = PersistentHeaderPtr();
        PersistentTrailer* t = PersistentTrailerPtr();

        h->m_Parity     = PersistentParity::ConvertInternalToPersistent(value);
        t->m_Parity     = PersistentParity::ConvertInternalToPersistent(value);
    }

    /*--------------------------------------------------------------------------
    function:     WriteParityAlgorithm()
    description:  Writes multi use value to frame
    arguments:    algo  [in]   parity algorithm id
    return value: none
    --------------------------------------------------------------------------*/

    void WriteParityAlgorithm( ParityAlgorithm  algo )
    {
        PersistentHeaderPtr()->m_Algorithm =
            PersistentParityAlgorithm::ConvertInternalToPersistent( algo );
    }

    /*--------------------------------------------------------------------------
    function:     InitFiller()
    description:  Writes zeros to unused header and trailer parts
    --------------------------------------------------------------------------*/

    void InitFiller()
    {
        WriteMultiUse1( 0 );
    }

    /*--------------------------------------------------------------------------
    function:     InitializeParityRaster()
    description:  - Initalizes page frame with bitmap
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/

    void InitializeParityRaster();

};
/*!*****************************************************************************

    endclass: Kernel_Page74

*******************************************************************************/


#endif // KERNEL_PAGE74_HPP
