/*!*********************************************************************

  module:       Kernel_TraceTopics.cpp

  ----------------------------------------------------------------------

  responsible:  DanielD

  special area: Kernel Tracing

  description:  contains instances of kernel trace topics

  ----------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000

    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

  *********************************************************************/

#include "gsp00.h"
#include "ggg00.h"
#include "hgg941.h"
#include "heo51.h"

#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"
#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"
#include "SAPDBCommon/Diagnostics/SAPDBDiag_CheckTopic.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Stream.hpp"
#include "Trace/Trace_Entry.hpp"
#include "KernelCommon/Kernel_TraceTopics.hpp"

/* ------------------------------------------------------------------ */

// SAPDBDiag_CheckTopic Demo_Switch    ("DEMO");

/* ------------------------------------------------------------------ */

SAPDBDiag_CheckTopic DataChain_Check     ("DATACHAIN");    // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic DataPam_Check       ("DATAPAM");      // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic DataTree_Check      ("DATATREE");     // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic DataPagePrim_Check  ("DATAPAGEPRIM"); // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic DataPageInv_Check   ("DATAPAGEINV");  // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic DataPageLog_Check   ("DATAPAGELOG");  // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic LogPage_Check       ("LOGPAGE");      // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic LogVolume_Check     ("LOGVOLUME");    // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic LogHistory_Check    ("LOGHISTORY");   // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic LogTrans_Check      ("LOGTRANS");     // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic LogAction_Check     ("LOGACTION");    // PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic ConverterPage_Check ("CONVERTERPAGE");// PTS 1117693 UH 2002-09-05
SAPDBDiag_CheckTopic Converter_Check     ("CONVERTER");
SAPDBDiag_CheckTopic FrameCtrl_Check     ("FRAMECTRL");
SAPDBDiag_CheckTopic IOMan_Check         ("IOMAN");
SAPDBDiag_CheckTopic Memory_Check        ("MEMORY");
SAPDBDiag_CheckTopic Devspace_Check      ("DEVSPACE");
SAPDBDiag_CheckTopic Runtime_Check       ("RUNTIME");
SAPDBDiag_CheckTopic AllocatorCheck      ("ALLOCATOR");
SAPDBDiag_CheckTopic FBM_Check           ("FBM");
SAPDBDiag_CheckTopic Pager_Check         ("PAGER");


/* ------------------------------------------------------------------ */

SAPDBDiag_CheckTopic * const Kernel_AllCheckSwitches [] = {
    &DataChain_Check,    // PTS 1117693 UH 2002-09-05
    &DataPam_Check,      // PTS 1117693 UH 2002-09-05
    &DataTree_Check,     // PTS 1117693 UH 2002-09-05
    &DataPagePrim_Check, // PTS 1117693 UH 2002-09-05
    &DataPageInv_Check,  // PTS 1117693 UH 2002-09-05
    &DataPageLog_Check,  // PTS 1117693 UH 2002-09-05
    &LogVolume_Check,    // PTS 1117693 UH 2002-09-05
    &LogHistory_Check,   // PTS 1117693 UH 2002-09-05
    &LogTrans_Check,     // PTS 1117693 UH 2002-09-05
    &LogAction_Check,    // PTS 1117693 UH 2002-09-05
    &LogPage_Check,      // PTS 1117693 UH 2002-09-05
    &ConverterPage_Check,// PTS 1117693 UH 2002-09-05
    &Converter_Check,
    &FrameCtrl_Check,
    &IOMan_Check,
    &Memory_Check,
    &Devspace_Check,
    &Runtime_Check,
    &AllocatorCheck,
	&FBM_Check,
    &Pager_Check,
    0
};

static inline int GetCheckSwitchCount ()
{
    static int switchCount = -1;

    if (switchCount == -1) {
        for (switchCount = 0; Kernel_AllCheckSwitches [switchCount] != 0; ++switchCount) {
            // counting is enough
        }
    }
    return switchCount;
}

/* ------------------------------------------------------------------ */

SAPDBTrace_Topic DataChain_Trace            ("DATACHAIN");     // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic DataPam_Trace              ("DATAPAM");       // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic DataTree_Trace             ("DATATREE");      // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic DataPagePrim_Trace         ("DATAPAGEPRIM");  // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic DataPageInv_Trace          ("DATAPAGEINV");   // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic DataPageLog_Trace          ("DATAPAGELOG");   // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic LogVolume_Trace            ("LOGVOLUME");     // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic LogHistory_Trace           ("LOGHISTORY");    // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic LogTrans_Trace             ("LOGTRANS");      // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic LogAction_Trace            ("LOGACTION");     // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic LogPage_Trace              ("LOGPAGE");       // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic ConverterPage_Trace        ("CONVERTERPAGE"); // PTS 1117693 UH 2002-09-05
SAPDBTrace_Topic Converter_Trace            ("CONVERTER");
SAPDBTrace_Topic FrameCtrl_Trace            ("FRAMECTRL");
SAPDBTrace_Topic IOMan_Trace                ("IOMAN");
SAPDBTrace_Topic Memory_Trace               ("MEMORY");
SAPDBTrace_Topic Devspace_Trace             ("DEVSPACE");
SAPDBTrace_Topic Runtime_Trace              ("RUNTIME");
SAPDBTrace_Topic AllocatorTrace             ("ALLOCATOR");
SAPDBTrace_Topic FBM_Trace          		("FBM");
SAPDBTrace_Topic Pager_Trace        		("PAGER");
SAPDBTrace_Topic ConsoleThread              ("CONSTHREAD");
SAPDBTrace_Topic ConsoleCommunication       ("CONSCOMMUNICATION");
SAPDBTrace_Topic SysView_Trace              ("SYSVIEW");


/* ------------------------------------------------------------------ */

SAPDBTrace_Topic * const Kernel_AllTraceTopics [] = {
    &DataChain_Trace,         // PTS 1117693 UH 2002-09-05
    &DataPam_Trace,           // PTS 1117693 UH 2002-09-05
    &DataTree_Trace,          // PTS 1117693 UH 2002-09-05
    &DataPagePrim_Trace,      // PTS 1117693 UH 2002-09-05
    &DataPageInv_Trace,       // PTS 1117693 UH 2002-09-05
    &DataPageLog_Trace,       // PTS 1117693 UH 2002-09-05 
    &LogVolume_Trace,         // PTS 1117693 UH 2002-09-05
    &LogHistory_Trace,        // PTS 1117693 UH 2002-09-05
    &LogTrans_Trace,          // PTS 1117693 UH 2002-09-05
    &LogAction_Trace,         // PTS 1117693 UH 2002-09-05
    &LogPage_Trace,           // PTS 1117693 UH 2002-09-05
    &ConverterPage_Trace,     // PTS 1117693 UH 2002-09-05
    &Converter_Trace,
    &FrameCtrl_Trace,
    &IOMan_Trace,
    &Memory_Trace,
    &Devspace_Trace,
    &Runtime_Trace,
    &AllocatorTrace,
    &FBM_Trace,
    &Pager_Trace,
    &ConsoleThread,
	&ConsoleCommunication,
    &SysView_Trace,
    0
};

static inline int GetTraceTopicCount ()
{
    static int switchCount = -1;

    if (switchCount == -1) {
        for (switchCount = 0; Kernel_AllTraceTopics [switchCount] != 0; ++switchCount) {
            // counting is enough
        }
    }
    return switchCount;
}

/* ------------------------------------------------------------------ */

/*!*********************************************************************

  Class: VtracePrinter

 **********************************************************************/
class VtracePrinter : public SAPDBTrace_LinePrinter
{
public:
    /* function: contructor
     */
    VtracePrinter ();

    /* function: PrintLn
       description: writes a line to the output medium
     */
    virtual void PrintLn(const SAPDB_Char * buf, int buflen);

private:
    tgg00_TransContext trans;
};

/*! EndClass: VtracePrinter */

/* ------------------------------------------------------------------ */

/* function: contructor
 */
VtracePrinter::VtracePrinter ()
{
    this->trans.trSessionId_gg00.becomes    (cgg_nil_session);
    this->trans.trTransId_gg00.becomes      (cgg_nil_trans);
    this->trans.trWriteTransId_gg00.becomes (cgg_nil_trans);
    this->trans.trSubtransId_gg00           = cgg_zero_subtrans;
}

/* ------------------------------------------------------------------ */

/* function: PrintLn
   description: writes a line to the output medium
 */
void VtracePrinter::PrintLn(const SAPDB_Char * buf, int buflen)
{
    vgetpid (this->trans.trTaskId_gg00);
    /* This prevents crash if trace called from outside UKT code */
    if ( 0 != this->trans.trTaskId_gg00 )
    {
        Trace_CommonEntry ( this->trans, tgg00_Debug::fromConst(ta),
                 tgg00_VtraceType::fromConst(td_none),
                 reinterpret_cast<const tsp00_Byte *> (buf), buflen);
    }
}

/* ------------------------------------------------------------------ */

static VtracePrinter mainVtracePrinter;

/*! --------------------------------------------------------------------
function:    Kernel_MultiSwitch
description: parse and exefcutes a line with switch instructions

    The line is of the form "<id> [<int>] [<id> [<int>] ] ...

    An id '*' serves as a wildcard for all topics.

    interface in vgg941.
*/
externPascal void Kernel_MultiSwitch (const char*  TopicStr, /* c20 */
                                      int          TopicStrLen)
{
    SAPDBDiag_CheckTopic::MultiSwitch (Kernel_AllCheckSwitches, TopicStr, TopicStrLen);
    SAPDBTrace_Topic::MultiSwitch (Kernel_AllTraceTopics, TopicStr, TopicStrLen);
}

/*! --------------------------------------------------------------------
function:    Kernel_CheckSwitch
description: parse and exefcutes a line with switch instructions

    The line is of the form "<id> [<int>] [<id> [<int>] ] ...

    An id '*' serves as a wildcard for all topics.

    interface in vgg941.
*/
externPascal void Kernel_CheckSwitch (const char*  TopicStr, /* c20 */
                                      int          TopicStrLen)
{
    SAPDBDiag_CheckTopic::MultiSwitch (Kernel_AllCheckSwitches, TopicStr, TopicStrLen);
}

/*! --------------------------------------------------------------------
function:    Kernel_TraceSwitch
description: parse and exefcutes a line with switch instructions

    The line is of the form "<id> [<int>] [<id> [<int>] ] ...

    An id '*' serves as a wildcard for all topics.

    interface in vgg941.
*/
externPascal void Kernel_TraceSwitch (const char*  TopicStr, /* c20 */
                                      int          TopicStrLen)
{
    SAPDBTrace_Topic::MultiSwitch (Kernel_AllTraceTopics, TopicStr, TopicStrLen);
}

/*! --------------------------------------------------------------------
function:    Kernel_InitTrace
description: sets the printer for all trace topics
*/

externPascal void Kernel_InitTrace ()
{
    SAPDBTrace_Topic::InitGroup (Kernel_AllTraceTopics, &mainVtracePrinter);
}

/*! --------------------------------------------------------------------
function:     Kernel_DiagInfo
description:
 */
pasbool Kernel_DiagInfo (
    int          requested,
    SAPDB_Char * kindOut,
    int          maxKindLen,
    SAPDB_Char * topicKeyOut,
    int          maxTopicKey,
    int        & levelOut)
{
    pasbool ok = false;
    int relativeRequested;

    int switchCount = GetCheckSwitchCount();
    relativeRequested = requested;
    if ((0 <= relativeRequested) && (relativeRequested < switchCount)) {
        Kernel_AllCheckSwitches [relativeRequested]->GetInfo (
            kindOut, maxKindLen, topicKeyOut, maxTopicKey, levelOut);
        return true;
    }
    int traceCount = GetTraceTopicCount();
    relativeRequested = requested - switchCount;
    if ((0 <= relativeRequested) && (relativeRequested < traceCount)) {
        Kernel_AllTraceTopics [relativeRequested]->GetInfo (
            kindOut, maxKindLen, topicKeyOut, maxTopicKey, levelOut);
        return true;
    }
    return ok;
}

/* ------------------------------------------------------------------ */

static void fillIdentifier (
    char * identifier,
    int size)
{
    int len = static_cast<int> (strlen (identifier));
    if (len < size) {
        memset (identifier + len, ' ', size - len);
    }
}

/*! --------------------------------------------------------------------
function:     Kernel_DiagInfop
description:  gets information about a specific topic

    use as:

    int index = 0;
    while (Kernel_DiagInfop (index, kindOut, topicKeyOut, levelOut)) {
        ...
        ++index;
    }

    interface in vgg941.

arguments:
    requested [IN]    an index into the array of topics
    kindOut   [OUT]   "Vtrace" or "Check"
    topicKeyOut [OUT] the string used to switch the topic
    levelOut [OUT]    the current level
 */
externPascal pasbool Kernel_DiagInfop (
    int                requested,
    tsp00_Identifier * kindOut,
    tsp00_Sname      * topicKeyOut,
    int              & levelOut)
{
    pasbool foundInfo;

    foundInfo = Kernel_DiagInfo (requested,
        kindOut->asCharp (), sizeof (tsp00_Identifier),
        topicKeyOut->asCharp (), sizeof (tsp00_Sname),
        levelOut);
    if (foundInfo) {
        fillIdentifier (*kindOut, sizeof (tsp00_Identifier));
        fillIdentifier (*topicKeyOut, sizeof (tsp00_Sname));
    }
    return foundInfo;
}

/*--------------------------------------------------------------------*/

