
/*!**************************************************************************

    module      : Kernel_WaitQueue.hpp

    -------------------------------------------------------------------------

    responsible : N.N.

    special area: KernelCommon
    description : implementation of methods to handle wait stacks containing task ids

    last changed: 2001-01-19

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2001 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "gsp00.h"
#include "ggg00.h"
#include "hgg08.h"
#include "heo51.h"   // vgetpid
#include "heo55k.h"  // vbegexcl
#include "heo56.h"   // vresume
#include "gsp00_0.h" // ROUTINE_DBG_MSP00

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define NOT_IMPLEMENTED   SAPDBERR_ASSERT_STATE (false)

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

Kernel_WaitStackGlob* Kernel_WaitStackGlob::m_pInstance = NULL;

/*===========================================================================*
*  METHODS of Kernel_TaskStack                                               *
*============================================================================*/

void Kernel_TaskStack::InsertAtBottom (Kernel_TaskStackItem *pItem)
{ 
    ROUTINE_DBG_MSP00 ("Kernel_TaskStack::InsertAtBottom");

    ++m_Count;

    if ( this->IsEmpty() )
    {
        pItem->PushItem (m_First);
        return;
    }
    
    {
        Kernel_TaskStackItem *pCurr = m_First;
        
        while (NULL != pCurr->GetNextItem())
        {
            pCurr = pCurr->GetNextItem();
        }
        
        pItem->AppendItem (pCurr); 
    }
}

/*===========================================================================*
*  METHODS of Kernel_WaitStackGlob                                           *
*============================================================================*/

void Kernel_WaitStackGlob::AllocateInstance (bool &IsOk)
{  
    ROUTINE_DBG_MSP00 ("Kernel_WaitStackGlob::AllocateInstance");
    
    SAPDBMEM_STATIC_RAW_ALLOCATE (Space, sizeof(Kernel_WaitStackGlob));
    
    if (NULL == m_pInstance)
    {
        m_pInstance = new (Space) Kernel_WaitStackGlob ();
    }
    
    IsOk = (NULL != m_pInstance);
}

// --------------------------------------------------------------------------------

void Kernel_WaitStackGlob::AllocateStackElements (tsp00_TaskId  CallingTaskId,
                                                  SAPDB_UInt    MaxTasks,
                                                  bool         &IsOk)
{  
    ROUTINE_DBG_MSP00 ("Kernel_WaitStackGlob::AllocateStackElements");
    
    Kernel_SyncWaitStack SyncWaitStack (CallingTaskId);
    
    IsOk = true;
    
    if (this->m_MaxTasks > 0)
    {
        IsOk = false;
        return;
    }
    
    this->m_MaxTasks = MaxTasks;
    
    if (NULL == this->m_TaskStackArray)
    {
        newarray (this->m_TaskStackArray, MaxTasks, RTEMem_Allocator::Instance());
        
        if (NULL == this->m_TaskStackArray)
        {
            IsOk = false;
            return;
        }
    }
    
    if (NULL == this->m_PredefinedWaitStack)
    {
        newarray (this->m_PredefinedWaitStack, wstLastEnum, RTEMem_Allocator::Instance());
        
        if (NULL == this->m_PredefinedWaitStack)
        {
            IsOk = false;
            return;
        }
    }
    
    for (SAPDB_UInt i = 0; i < MaxTasks; ++i)
    {
        this->m_FreeList.Push ( this->m_TaskStackArray+i );
    }
}

// --------------------------------------------------------------------------------

bool Kernel_WaitStackGlob::IsEmptyFreeList (tsp00_TaskId CallingTaskId)
{
    ROUTINE_DBG_MSP00 ("Kernel_WaitStackGlob::IsEmptyFreeList");
    
    Kernel_SyncWaitStack SyncWaitStack (CallingTaskId);
    
    return ( this->m_FreeList.IsEmpty() );
}

// --------------------------------------------------------------------------------


Kernel_WaitStackGlob::Kernel_SyncWaitStack::Kernel_SyncWaitStack (tsp00_TaskId CallingTaskId)
{
    this->m_CallingTaskId = CallingTaskId;
    vbegexcl (this->m_CallingTaskId, g08TaskWait);
}

// --------------------------------------------------------------------------------

Kernel_WaitStackGlob::Kernel_SyncWaitStack::~Kernel_SyncWaitStack ()
{
    vendexcl (this->m_CallingTaskId, g08TaskWait);
}

/*===========================================================================*
*  METHODS of Kernel_WaitStack                                               *
*============================================================================*/

Kernel_WaitStack::~Kernel_WaitStack()
{
    ROUTINE_DBG_MSP00 ("~Kernel_WaitStack");

    if ( ! this->m_WaitStack.IsEmpty() )
    {
        tsp00_TaskId MyTaskId;
        vgetpid (MyTaskId);
        {
            Kernel_TaskStack::Kernel_PopItemState       PopItemState;
            Kernel_TaskStackItem                       *pItem;
            Kernel_WaitStackGlob::Kernel_SyncWaitStack  SyncWaitStack (MyTaskId);
            
            do
            {
                this->m_WaitStack.Pop (pItem, PopItemState);
                Kernel_WaitStackGlob::Instance().PutIntoFreeList (pItem);
            }
            while (Kernel_TaskStack::pstAnyItemLeft == PopItemState);
        }
    }
}

// --------------------------------------------------------------------------------

void Kernel_WaitStack::InsertTaskIdAtBottom (tsp00_TaskId TaskId)

{
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::InsertTaskIdAtBottom");
    
    Kernel_WaitStackGlob::Kernel_SyncWaitStack SyncWaitStack (TaskId);
    
    Kernel_TaskStackItem *pItem = Kernel_WaitStackGlob::Instance().GetFromFreeList ();
    
    SAPDBERR_ASSERT_STATE (NULL != pItem);
    
    pItem->Init (TaskId);
    
    this->m_WaitStack.InsertAtBottom (pItem);
}

// --------------------------------------------------------------------------------

bool Kernel_WaitStack::IsEmpty (tsp00_TaskId CallingTaskId) const
{
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::IsEmpty");
    
    Kernel_WaitStackGlob::Kernel_SyncWaitStack SyncWaitStack (CallingTaskId);
    
    return (this->m_WaitStack.IsEmpty());
}

// --------------------------------------------------------------------------------

SAPDB_UInt Kernel_WaitStack::GetCountOfWaitingTasks () const
{
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::GetCountOfWaitingTasks");
    // no region    
    return (this->m_WaitStack.GetSize());
}

// --------------------------------------------------------------------------------

void Kernel_WaitStack::PopAllTaskIdsAndResume (tsp00_TaskId CallingTaskId)
{   
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::PopAllTaskIdsAndResume");
    
    Kernel_WaitStackGlob::Kernel_SyncWaitStack  SyncWaitStack (CallingTaskId);
    Kernel_TaskStack::Kernel_PopItemState       PopItemState;

    Kernel_TaskStackItem *pItem;
    tsp00_TaskId          WaitingTaskId;
    
    if (this->m_WaitStack.IsEmpty()) return;
    
    do
    {
        this->m_WaitStack.Pop (pItem, PopItemState);
        WaitingTaskId = pItem->GetTaskId();
        Kernel_WaitStackGlob::Instance().PutIntoFreeList (pItem);
        
        //=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        vresume (WaitingTaskId, cgg_suspend_lvl_0);
        //=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        
    }
    while (Kernel_TaskStack::pstAnyItemLeft == PopItemState);
}

// --------------------------------------------------------------------------------

void Kernel_WaitStack::PopAnyTaskIdAndResume (tsp00_TaskId CallingTaskId)
{   
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::PopAnyTaskIdAndResume");
    
    tsp00_TaskId                           WaitingTaskId;
    Kernel_TaskStack::Kernel_PopItemState  DummyState;
    
    this ->PopTaskId (CallingTaskId, WaitingTaskId, DummyState);
    
    if (cgg_nil_pid != WaitingTaskId)
    {
        //=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        vresume (WaitingTaskId, cgg_suspend_lvl_0);
        //=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    }
}
// --------------------------------------------------------------------------------

void Kernel_WaitStack::PopListOfTaskIds (tsp00_TaskId  CallingTaskId,
                                         tsp00_Int4    ListBufSize,
                                         void         *ListBuf,
                                         tsp00_Int4   &ListBufUsedLen)
{   
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::PopListOfTaskIds");
    
    Kernel_WaitStackGlob::Kernel_SyncWaitStack  SyncWaitStack (CallingTaskId);
    
    ListBufUsedLen = 0;
    
    if ((! this->m_WaitStack.IsEmpty()) && (sizeof(tsp00_TaskId) <= ListBufSize))
    {
        Kernel_TaskStack::Kernel_PopItemState   PopItemState;
        Kernel_TaskStackItem                   *pItem;

        tsp00_Int4  TaskIdCnt = 0;
        tsp00_Int4  MaxTaskIdCnt = ListBufSize / sizeof (tsp00_TaskId);
        
        do
        {
            this->m_WaitStack.Pop (pItem, PopItemState);
            
            STATIC_CAST (tsp00_TaskId*, ListBuf) [TaskIdCnt] = pItem->GetTaskId();
            TaskIdCnt += 1;
            
            Kernel_WaitStackGlob::Instance().PutIntoFreeList (pItem);
        }
        while ((Kernel_TaskStack::pstAnyItemLeft == PopItemState) && (TaskIdCnt < MaxTaskIdCnt) );

        ListBufUsedLen = TaskIdCnt * sizeof (tsp00_TaskId);
    }
}

// --------------------------------------------------------------------------------

void Kernel_WaitStack::PopTaskId (tsp00_TaskId                           CallingTaskId,
                                  tsp00_TaskId                          &PopTaskId,
                                  Kernel_TaskStack::Kernel_PopItemState &PopItemState)
{   
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::PopTaskId");
    
    Kernel_WaitStackGlob::Kernel_SyncWaitStack  SyncWaitStack (CallingTaskId);

    Kernel_TaskStackItem *pItem;
    
    if ( this->m_WaitStack.IsEmpty() )
    {
        PopTaskId    = cgg_nil_pid;
        PopItemState = Kernel_TaskStack::pstStackEmpty;
        return;
    }
    
    this->m_WaitStack.Pop (pItem, PopItemState);
    PopTaskId = pItem->GetTaskId();
    Kernel_WaitStackGlob::Instance().PutIntoFreeList (pItem);
}

// --------------------------------------------------------------------------------

void Kernel_WaitStack::PushTaskId (tsp00_TaskId TaskId)

{
    ROUTINE_DBG_MSP00 ("Kernel_WaitStack::PushTaskId");
    
    Kernel_WaitStackGlob::Kernel_SyncWaitStack SyncWaitStack (TaskId);
    
    Kernel_TaskStackItem *pItem = Kernel_WaitStackGlob::Instance().GetFromFreeList ();
    
    SAPDBERR_ASSERT_STATE (NULL != pItem);
    
    if (NULL != pItem)
    {
        pItem->Init (TaskId);
        
        this->m_WaitStack.Push (pItem);
    }
}

/*===========================================================================*
*  END OF CODE                                                               *
*============================================================================*/
