
/*!**************************************************************************

    module      : Kernel_WaitQueue.hpp

    -------------------------------------------------------------------------

    responsible : N.N.

    special area: KernelCommon
    description : defines the classes to handle wait stacks containing task ids

    last changed: 2001-01-19

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef KERNEL_WAIT_QUEUE_H
#define KERNEL_WAIT_QUEUE_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "gsp00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

//=======================================================================================
/*! class:       Kernel_TaskStackItem

    description: item of a task wait stack
*/
class Kernel_TaskStackItem
{
    
private:
    tsp00_TaskId          m_TaskId;
    Kernel_TaskStackItem *m_Next;
    
public:
    //--------------------------------------------------------------Kernel_TaskStackItem---
    /*! function:    Kernel_TaskStackItem ctor
    
        description: empty stack
    */
    Kernel_TaskStackItem ():
          m_TaskId (0),
          m_Next   (NULL)
    { }
      
    //--------------------------------------------------------------Kernel_TaskStackItem---
    /*! function:    AppendItem
    
        description: puts this item behind the other item
      
        arguments:   Next [in] the pointer to the other item
    */
    void AppendItem (Kernel_TaskStackItem *OtherItem)
    {
        if (NULL != OtherItem)
        {
            OtherItem->m_Next = this;
        }
    }
    
    //--------------------------------------------------------------Kernel_TaskStackItem---
    /*! function:    GetNextItem
    
       return value: the pointer to the following stack item
    */
    Kernel_TaskStackItem* GetNextItem ()
    { return (m_Next); }

    //--------------------------------------------------------------Kernel_TaskStackItem---
    /*! function:    GetTaskId
    
        return value: the task id contained in the item
    */
    tsp00_TaskId GetTaskId ()
    { return (m_TaskId); }
    
    //--------------------------------------------------------------Kernel_TaskStackItem---
    /*! function:    Init
    
        description: inits the item and assigns a TaskId
      
        arguments:   TaskId [in]
    */
    void Init (tsp00_TaskId TaskId)
    {
        this->m_TaskId = TaskId;
        this->m_Next   = NULL;
    }

    //--------------------------------------------------------------Kernel_TaskStackItem---
    /*! function:    PushItem
    
        description: puts this item on top of a stack
      
        arguments:   Next [in] the pointer to the first stack item 
    */
    void PushItem (Kernel_TaskStackItem* &FirstStackItem)
    {
        this->m_Next   = FirstStackItem;
        FirstStackItem = this;
    }
    
};
/*! endclass: Kernel_TaskStackItem */

//=======================================================================================
/*! class:       Kernel_TaskStack

    description: a stack containing task ids 
*/
class Kernel_TaskStack
{

public:
    
    typedef enum
    {        
            pstStackEmpty,
            pstAnyItemLeft,
            pstLastItem
    } Kernel_PopItemState;

private:
    Kernel_TaskStackItem *m_First;
    SAPDB_UInt            m_Count;
    
public:
    
    //--------------------------------------------------------------Kernel_TaskStack---
    /*! function:    Kernel_TaskStack ctor
    
        description: empty stack
    */
    Kernel_TaskStack (): m_First (NULL), m_Count (0)
    { }

    //--------------------------------------------------------------Kernel_TaskStack---
    
    /*! function:    InsertAtBottom
    
        description: Inserts an item at the bottom of the stack
        
        arguments:   pItem [in] the pointer to the item to be inserted
    */
    void InsertAtBottom (Kernel_TaskStackItem *pItem);

    //--------------------------------------------------------------Kernel_TaskStack---
    /*! function:     IsEmpty
    
        return value: true, if the stack is empty
    */
    bool IsEmpty () const
    { return (NULL == m_First); }
    
    //--------------------------------------------------------------Kernel_TaskStack---
    /*! function:     GetSize
    
        return value: size
    */
    SAPDB_UInt GetSize () const
    { return m_Count; }
    
    //--------------------------------------------------------------Kernel_TaskStack---
    /*! function:     Pop ()
    
        return value: the pointer to the removed first stack item
    */
    Kernel_TaskStackItem *Pop ()
    { 
        Kernel_TaskStackItem *pItem = m_First;
        if (NULL != m_First) { m_First = m_First->GetNextItem(); }
        --m_Count;
        return (pItem);
    }
    
    //--------------------------------------------------------------Kernel_TaskStack---
    /*! function:    Pop (pItem, PopRetCode)
    
        description: removes an item from the stack
        
        arguments:   pElement [out] the pointer to the removed stack item
                     PopItemState [out] indicates an empty stack or any/last stack item 
    */
    void Pop (
        Kernel_TaskStackItem* &pItem,
        Kernel_PopItemState   &PopItemState)
    { 
        if ( this->IsEmpty() )
        {
            pItem = NULL;
            PopItemState = pstStackEmpty;
            return;
        }
        
        pItem        = m_First;
        m_First      = m_First->GetNextItem();
        PopItemState = (NULL == m_First) ? pstLastItem : pstAnyItemLeft;
        --m_Count;
    }

    //--------------------------------------------------------------Kernel_TaskStack---
    /*! function:    Push
    
        description: Inserts an item into the stack

        arguments:   pItem [in] the pointer to the item to be inserted
    */
    void Push (Kernel_TaskStackItem *pItem)
    { pItem->PushItem (m_First); ++m_Count; }
};
/*! endclass: Kernel_TaskStack */

//=======================================================================================
/*! class:       Kernel_WaitStack

    description: defines a stack which contains waiting tasks
*/
class Kernel_WaitStack
{
private:
    
    Kernel_TaskStack m_WaitStack;
    
public:  

    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    Kernel_WaitStack ctor
    
        description: empty stack
    */
    Kernel_WaitStack (): m_WaitStack ()
    { }
    
    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    Kernel_WaitStack dtor
    
        description: releases the remaining stack items
    */
    ~Kernel_WaitStack();

    //--------------------------------------------------------------Kernel_WaitStack---    
    /*! function:    InsertTaskIdAtBottom
    
        description: Inserts a task id at the bottom of the wait stack
        
        arguments:    TaskId [in]
    */
    void InsertTaskIdAtBottom (tsp00_TaskId TaskId);

    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:     IsEmpty
    
        return value: true, if the wait stack is empty
    */
    bool IsEmpty (tsp00_TaskId CallingTaskId) const;
    
    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:     GetCountOfWaitingTasks
    
        return value: number of waiting tasks, if the wait stack is empty
    */
    SAPDB_UInt GetCountOfWaitingTasks () const;
    
    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    PopAllTaskIdsAndResume
    
        description: removes all task ids from the wait stack and performes resume for these tasks 
        
        arguments:   CallingTaskId [in] the task id calling this function
    */
    void PopAllTaskIdsAndResume (tsp00_TaskId CallingTaskId);

    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    PopAnyTaskIdAndResume
    
        description: removes a task id from the wait stack and performes resume for the task
        
        arguments:   CallingTaskId [in] the task id calling this function
    */
    void PopAnyTaskIdAndResume (tsp00_TaskId CallingTaskId);

    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    PopListOfTaskIds
    
        description: creates a list of task ids removed from the stack
        
        arguments:   CallingTaskId  [in]  the task id calling this function
                     ListBufSize    [in]  the allocated byte size of the task list buffer
                     ListBuf        [out] the pointer to the task list buffer
                     ListBufUsedLen [out] length of the buffer filled with task ids  
    */
    void PopListOfTaskIds (
        tsp00_TaskId  CallingTaskId,
        tsp00_Int4    ListBufSize,
        void         *ListBuf,
        tsp00_Int4   &ListBufUsedLen);
    
    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    PopTaskId
    
        description: gets a task id from the wait stack
        
        arguments:   CallingTaskId [in]  the task id calling this function
                     PopTaskId     [out] the task id received from the wait stack
                     PopItemState  [out] indicates an empty stack or any/last stack item 
    */
    void PopTaskId (
        tsp00_TaskId                           CallingTaskId,
        tsp00_TaskId                          &PopTaskId,
        Kernel_TaskStack::Kernel_PopItemState &PopItemState);
    
    //--------------------------------------------------------------Kernel_WaitStack---
    /*! function:    PushTaskId
    
        description: inserts a task id into the wait stack 

        arguments:   TaskId [in]
    */
    void PushTaskId (tsp00_TaskId TaskId);
};
/*! endclass: Kernel_WaitStack */

//=======================================================================================
/*! class:       Kernel_WaitStackGlob

    description: task stack and predefined wait stacks (singleton)
*/
class Kernel_WaitStackGlob
{
    friend class Kernel_WaitStack;

public:

typedef enum
{
        wstAutoSaveRead,
        wstAutoSaveWrite,
        wstBackupRead,
        wstBackupWrite,
        wstCheckpoint,
        wstLogFull,
        wstSavepoint,
        wstSaveptPrepare,
        wstSurrogate,
        wstLastEnum
} Kernel_WaitStackType;

private:
    
    static Kernel_WaitStackGlob *m_pInstance;   // internal global instance
    
    Kernel_TaskStackItem *m_TaskStackArray;
    Kernel_WaitStack     *m_PredefinedWaitStack;
    SAPDB_UInt            m_MaxTasks;
    Kernel_TaskStack      m_FreeList;
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    Kernel_WaitStackGlob ctor
    
        description: private ctor only used once during initialization
    */
    Kernel_WaitStackGlob ():
        m_TaskStackArray      (NULL),
        m_PredefinedWaitStack (NULL),
        m_MaxTasks            (0),
        m_FreeList            ()
    { }
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:     GetFromFreeList
    
        return value: the pointer to an item received from the free list        
    */

    Kernel_TaskStackItem *GetFromFreeList ()
    { return ( m_FreeList.Pop() ); }

    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    PutIntoFreeList
    
        description: inserts an item into the free list

        arguments:   pFreeItem [in] pointer to the free item
    */
    void PutIntoFreeList (Kernel_TaskStackItem *pFreeItem)
    {
        if (NULL != pFreeItem)
        {
            pFreeItem->Init (0);
            m_FreeList.Push (pFreeItem);
        }
    }

    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! class:       Kernel_SyncWaitStack

        description: class to get exclusive access to the wait stack
    */
    
    class Kernel_SyncWaitStack
    {
    private:
        tsp00_TaskId m_CallingTaskId;
        
    public:
        Kernel_SyncWaitStack (tsp00_TaskId CallingTaskId);
        
        ~Kernel_SyncWaitStack ();
    };
    /*! endclass: Kernel_SyncWaitStack */

    //--------------------------------------------------------------Kernel_WaitStackGlob--- 
    
public:
    
    /*! function:    AllocateInstance
    
        description: Allocates the WaitStackGlob singleton
      
        arguments:   IsOk [out]
    */
    static void AllocateInstance (bool &IsOk);
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    AllocateStackElements
    
        description: Allocates the wait stack entries and the predefined wait stacks 
      
        arguments:   CallingTaskId [in]
                     MaxTasks      [in]
                     IsOk          [out]
    */
    void AllocateStackElements (
        tsp00_TaskId  CallingTaskId,
        SAPDB_UInt    MaxTasks,
        bool         &IsOk);

    //--------------------------------------------------------------Kernel_WaitStack---
    
    /*! function:    InsertTaskIdAtBottom
    
        description: Inserts a task id at the bottom of the specified wait stack
        
        arguments:   TaskId        [in] 
                     WaitStackType [in] specifies the predefined wait stack 
    */
    void InsertTaskIdAtBottom (
        tsp00_TaskId         TaskId,
        Kernel_WaitStackType WaitStackType)
    {
        m_PredefinedWaitStack [WaitStackType].InsertTaskIdAtBottom (TaskId);
    }

    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:     Instance
    
        return value: the reference to the global instance of the singleton
    */
    static Kernel_WaitStackGlob &Instance ()
    { return (*m_pInstance); }

    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:     IsEmpty
    
        return value: true, if the specified wait stack is empty
    */
    bool IsEmpty (
        tsp00_TaskId         CallingTaskId,
        Kernel_WaitStackType WaitStackType)
    {
        return (m_PredefinedWaitStack [WaitStackType].IsEmpty(CallingTaskId));
    }
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
   
    /*! function:     IsEmptyFreeList
    
        return value: true, if the free list is empty
    */
    bool IsEmptyFreeList (tsp00_TaskId CallingTaskId);
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    PopAllTaskIdsAndResume
    
        description: removes all task ids from the wait stack and performes resume for these tasks 
        
        arguments:   CallingTaskId [in] the task id calling this function
                     WaitStackType [in] specifies the predefined wait stack 
    */
    void PopAllTaskIdsAndResume (
        tsp00_TaskId         CallingTaskId,
        Kernel_WaitStackType WaitStackType)
    {
        m_PredefinedWaitStack [WaitStackType].PopAllTaskIdsAndResume (CallingTaskId);
    }
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    PopAnyTaskIdAndResume
    
        description: removes a task id from the wait stack and performes resume for the task
        
        arguments:   CallingTaskId [in] the task id calling this function
                     WaitStackType [in] specifies the predefined wait stack 
    */
    void PopAnyTaskIdAndResume (
        tsp00_TaskId         CallingTaskId,
        Kernel_WaitStackType WaitStackType)
    {
        m_PredefinedWaitStack [WaitStackType].PopAnyTaskIdAndResume (CallingTaskId);
    }
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    PopListOfSaveptTaskIds
    
        description: creates a list of task ids removed from the savepoint wait stack
        
        arguments:   CallingTaskId  [in]  the task id calling this function
                     ListBufSize    [in]  the allocated byte size of the task list buffer
                     ListBuf        [out] the pointer to the task list buffer
                     ListBufUsedLen [out] length of the buffer filled with task ids  
    */
    void PopListOfSaveptTaskIds (
        tsp00_TaskId  CallingTaskId,
        tsp00_Int4    ListBufSize,
        void         *ListBuf,
        tsp00_Int4   &ListBufUsedLen)
    {
        m_PredefinedWaitStack [wstSavepoint].
            PopListOfTaskIds (CallingTaskId, ListBufSize, ListBuf, ListBufUsedLen);
    }
    

    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:    PopTaskId
    
        description: gets a task id from the specified wait stack
        
        arguments:   CallingTaskId [in]  the task id calling this function
                     WaitStackType [in]  specifies the predefined wait stack 
                     PopTaskId     [out] the task id received from the wait stack
                     PopItemState  [out] indicates an empty stack or any/last stack item 
    */
    void PopTaskId (
        tsp00_TaskId                            CallingTaskId,
        Kernel_WaitStackType                    WaitStackType,
        tsp00_TaskId                           &PopTaskId,
        Kernel_TaskStack::Kernel_PopItemState  &PopItemState)
    {
        m_PredefinedWaitStack [WaitStackType].PopTaskId (CallingTaskId, PopTaskId, PopItemState);
    }
    
    //--------------------------------------------------------------Kernel_WaitStackGlob---
    /*! function:     PushTaskId
    
        description: inserts a task id into the specified wait stack 

        arguments:    TaskId        [in]
                      WaitStackType [in] specifies one of the predefined wait stacks 
    */
    void PushTaskId (
        tsp00_TaskId         TaskId,
        Kernel_WaitStackType WaitStackType)
    {
        m_PredefinedWaitStack [WaitStackType].PushTaskId (TaskId);
    }
};
/*! endclass: Kernel_WaitStackGlob */

/*===========================================================================*
*  DEFINITION OF INLINE METHODS                                              *
*===========================================================================*/

#endif  /* KERNEL_WAIT_QUEUE_H */
