/*!***************************************************************************

  module      : LVCSim_Client.hpp

  -------------------------------------------------------------------------

  responsible : IvanS

  special area: liveCache Simulator
  description : Main include for liveCache simulator test programs. Do not
	include this file in real COM objects! Use it only for your test programs
	that use COM DLLs.

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#ifndef	LVCSIM_HPP
#define	LVCSIM_HPP

// compatibility defines

/// Interface version, lC independent (major * 1000 + minor)
#define LVCSIM_INTERFACE_VERSION    1000
/// We have symbol refresh in debug allocator
#define LVCSIM_HAS_DA_SYMREFRESH    


// Win32 problems:

#undef	LVCSIM_API

#ifdef _WIN32

#pragma warning( disable : 4275 )
#pragma warning( disable : 4273 )
#pragma warning( disable : 4786 )
#pragma warning( disable : 4995 )
#pragma warning( disable : 4624 )

#define	_AFXDLL

#ifdef	LVCSIM_DLL_VMAKE_BUILD
#define LVCSIM_API		__declspec(dllexport)
#define	LVCSIM_EXP_API	__declspec(dllexport)
#else
#define LVCSIM_API		__declspec(dllimport)
#define	LVCSIM_EXP_API	__declspec(dllexport)
#endif

#else
#define	LVCSIM_API
#define	LVCSIM_EXP_API
#endif

#if defined _WIN32 || defined LINUX || defined AIX || defined SUN || defined HP_IA64
#define	LVCSIM_STD	std::
#else
#define	LVCSIM_STD
#endif

#ifdef	HPUX
#define __HPACC_USING_MULTIPLIES_IN_FUNCTIONAL
#endif

#define SAPDB_LVCSIM_EXTRAS		// include extra functionality from livecache.h

#include <livecacheversion.h>

#ifndef	LVCSIM_GLOBFNC_HPP
#define	LVCSIM_GLOBFNC_HPP

#ifndef	_WIN32
#include <stddef.h>
#endif
#include <stdlib.h>

class	OmsHandle;

#ifndef LVCSIM_DEBUG_ALLOC_ESCAPE_ENUM
#define LVCSIM_DEBUG_ALLOC_ESCAPE_ENUM
enum LVCSim_DebugAllocatorEscapeEnum {
    LVCSIM_DA_CLEAR_LEAKS,  ///< mark all memory blocks as no leaks
    LVCSIM_DA_ENTER_NOLEAK, ///< mark all block allocated between this
    LVCSIM_DA_EXIT_NOLEAK,  ///< and this as no leaks
    LVCSIM_DA_CHECK_LEAKS,  ///< check for leaks and dump leaks into trace/on screen
    LVCSIM_DA_CHECK_MEMORY, ///< check memory consistency
    LVCSIM_DA_MESSAGES_ON,  ///< turn on memory messages on-screen (e.g., leaks; can be nested)
    LVCSIM_DA_MESSAGES_OFF, ///< turn off memory messages on-screen
	LVCSIM_DA_ALLOCERR_ON,	///< turn on allocator error messages
	LVCSIM_DA_ALLOCERR_OFF,	///< turn off allocator error messages
    LVCSIM_DA_SYMREFRESH,   ///< refresh symbol table for stack traces
    LVCSIM_DA_STACKTRACE    ///< dump stack trace to the dump file
};
#endif



class LVCSim_Thread;                            // class representing thread in simulator
class LVCSim_Mutex;                             // class representing mutex in simulator
class LVCSim_Semaphore;                         // class representing semaphore in simulator
typedef unsigned long (*LVCSim_ThreadFunc) (void *arg);  // thread main function 

typedef enum {
    LVCSim_ThreadCreateNormal = 0,
    LVCSim_ThreadCreateSuspended = 1,
    LVCSim_ThreadCreateDetached = 2
} LVCSim_ThreadCreateFlags;

typedef enum {
    LVCSim_ThreadOK = 0,
    LVCSim_ThreadNotOK = 1
} LVCSim_ThreadErrorCode;

/*!
	class:	LVCSim

	description:	liveCache Simulator global functions
*/
class LVCSIM_API LVCSim {
public:
    /*!
    	function: Attach
    
    	description: Set up liveCache sink for a handle. Since COM objects are
    		derived from a handle, you can simply instantiate a COM object.
			Attach will be then called automatically by simulator. If you
			instantiate OmsHandle as such, you MUST call Attach to attach it
			to a sink. This automatically attaches the handle to the current
            session in multithreading program.
    */
    static void Attach(OmsHandle &obj);
    
    /*!
    	function: Attach
    
    	description: This form of Attach attaches the handle in obj
    		to a new sink. This handle then represents pseudo-new session
            in a single-threaded program.
    */
    static void AttachNewSink(OmsHandle &obj);

protected:
    /*!
    	function: Detach
    
    	description: Detach a COM object from liveCache sink. When all objects
    		using particular sink are detached, the transaction belonging to this
    		sink is rolled back (if not explicitly confirmed by Commit).
    		This sink can be then reused in concurrent thread (in fact, it will
    		be deallocated, but the semantics is the same).
    */
    static void Detach(OmsHandle &obj);

    static void *GetCurrentSink();
    
#if (defined(WIN32) && !defined(_WIN64)) || defined(LINUX)
    static void SetStackBase(void *ptr);
#endif

public:
    /*!
    	function: Init
    
    	description: Init internal structures. The simulator is in unknown state
            before Init is called. Result of calling any other functions before
            Init is undefined.
    */
    static void Init();
    
    /*!
    	function: Reset
    
    	description: Reset internal state of the liveCache simulator. Be sure that
    		no transactions are running. This command resets the simulator to the
    		state after Init.
    */
    static void Reset();
    
    /*!
    	function: Commit
    
    	description: Commit current transaction. Used to simulate end of transaction
    		in test program (after calling COM object method).
    */
    static void Commit(OmsHandle &obj);
    
    /*!
    	function: Rollback
    
    	description: Rollback current transaction. Used to simulate end of transaction
    		in test program (after calling COM object method).
    */
    static void Rollback(OmsHandle &obj, bool endSession = false);

    /*!
        function: COMEnd

        description: Call this function after call of a COM routine to check allocator
            emptyness.
    */
    static void COMEnd(OmsHandle &obj);

	// stream support

	/*!
		function:	RegisterStream

		description:	Register stream into internal stream table

		returns:	new stream ID
	*/
	static int RegisterStream(class LVCSim_StreamIOBase *stream);

	/*!
		function:	UnregisterStream

		description:	Unregisters given stream
	*/
	static void UnregisterStream(int id);

	/*!
		function:	GetStream

		description:	Returns stream for given ID
	*/
	static class LVCSim_StreamIOBase *GetStream(int id);


	// old simulator functionality:

	/*!
		function:		SetDumpCount

		description:	set how many times does omsDump write HTML dump file
	*/
	static void SetDumpCount(int cnt = 9999999);

	static void FlushVtrace();

	/*!
		function:		SaveFullState

		description:	Write current simulator state into file. All transactions
						must be commited or rolled back at the time of this call!!!
	*/
	static void SaveFullState(const char *fname);

	/*!
		function:		RestoreFullState

		description:	Read simulator state from a file. All transactions
						must be commited or rolled back at the time of this call!!!
	*/
	static void RestoreFullState(const char *fname);

	static void DumpMemoryStat();

	/*!
		function:		SetTraceFile

		description:	set trace file for vtrace
	*/
	static void SetTraceFile(const char *fname);

	/*!
		function:	SetTraceLevel

		description:	Set/reset trace level

		parameters:
				level  [in] trace level name (OMS_ERROR, OMS_INTERFACE,
							OMS_CONTAINERDIR, OMS_MEMORY, OMS_VAROBJECT,
							OMS_SESSION, OMS_LOCK)
				enable [in]	true to enable given level, false to disable
	*/
	static void SetTraceLevel(const char *level, bool enable = true);

	static void SetTraceLevel(long level);

	/*!
		function:	CheckMemoryLeaks

		description: Dump all memory leaks and check all blocks in freelist

        returns: true, when memory leaks found
    */
    static bool CheckMemoryLeaks();

	/*!
		function:	CheckMemory

		description: Check all memory allocations and deallocated memory

        returns: true, when error found
    */
    static bool CheckMemory();

	/*!
		function:	DebugAllocatorEscape

		description: Debug allocator service function

        returns: true, when error found
    */
    static bool DebugAllocatorEscape(LVCSim_DebugAllocatorEscapeEnum what);



    // Multithreading support

    static LVCSim_Thread *ThreadBegin(size_t stacksize, LVCSim_ThreadCreateFlags flags,    
        LVCSim_ThreadFunc threadProc, void *arg = NULL, int *err = NULL, char *errbuf = NULL, size_t bufsize = 0);

    static void ThreadEnd(int return_code);
    
    static void ThreadYield();

    static LVCSim_Thread *ThreadGetCurrent();
    
    static void ThreadKill(LVCSim_Thread *thread, int *err = NULL, 
        char *errbuf = NULL, size_t bufsize = 0);

    static void ThreadCancel(LVCSim_Thread *thread, int return_code, 
        int *err = NULL, char *errbuf = NULL, size_t bufsize = 0);

    static void ThreadForget(LVCSim_Thread *thread);

    static void ThreadSuspend(LVCSim_Thread *thread, int *err = NULL, 
        char *errbuf = NULL, size_t bufsize = 0);

    static void ThreadResume(LVCSim_Thread *thread, int *err = NULL, 
        char *errbuf = NULL, size_t bufsize = 0);

    static int ThreadJoin(LVCSim_Thread *thread, int *err = NULL, 
        char *errbuf = NULL, size_t bufsize = 0);

    static bool ThreadIsSuspended(LVCSim_Thread *thread);

    static bool ThreadIsAlive(LVCSim_Thread *thread, int *err = NULL, 
        char *errbuf = NULL, size_t bufsize = 0);

    
    static LVCSim_Mutex *MutexCreate();

    static void MutexDestroy(LVCSim_Mutex *mutex);

    static void MutexBegin(LVCSim_Mutex *mutex);
    
    static bool MutexTryBegin(LVCSim_Mutex *mutex); // true on success
    
    static void MutexEnd(LVCSim_Mutex *mutex);


    static void TLSCreate(int &key, int *err = NULL, char *errbuf = NULL, size_t bufsize = 0);

    static void TLSDestroy(int key, int *err = NULL, char *errbuf = NULL, size_t bufsize = 0);

    static void TLSSet(int key, void *val, int *err = NULL, char *errbuf = NULL, size_t bufsize = 0);

    static void *TLSGet(int key);


    static LVCSim_Semaphore *SemCreate(int initval = 0, int *err = NULL, char *errbuf = NULL, size_t bufsize = 0);

    static void SemDestroy(LVCSim_Semaphore *sem);

    static void SemSignal(LVCSim_Semaphore *sem);

    static void SemWait(LVCSim_Semaphore *sem);

    static bool SemTimedWait(LVCSim_Semaphore *sem, int seconds);   // true on success
    
    
    static int GetStackTrace(void** &trace, size_t TRACE_SIZE, size_t TRACE_SKIP);

protected:
	// definition of extensions for liveCache simulator in Oms
	friend class	LVCSim_liveCacheSink;
    friend class    LVCSim_KernelInterface;
	friend class	OmsHandle;
	friend class	OmsAbstractObject;
    friend class    LVCSim_Initer;

	static void setName(const struct OmsObjectId &oid, const char *name);
	static const char *getName(const struct OmsObjectId &oid);
	static void omsSimDump(OmsHandle &obj, int param);
};

/*! endclass: LVCSim */



extern
void omsFree(void* p);

/*!
	class:	LVCSim_StreamIOBase

	description:	This is the base class for all different stream classes with different
		input and output behavior. You must define LVCSim_StreamIO<T> for a given storage
		class T. Simulator defines the template LVCSim_StreamStorage as the base for table
		storage. In order to declare a test handle, simply instantiate
		LVCSim_StreamHandle.
		
		Then, you can use something like:

	code:
		LVCSim_StreamStorage<T> myStorage;
		LVCSim_StreamHandle	handle(&myStorage);	// or OmsTypeStreamHandle handle(&myStorage);
		...
		myStorage.whatever_is_append(my_data);	// for input stream parameter
		...
		myCOMInstance.MYMETHOD(..., &handle, ...)
		...
		my_data = myStorage.whatever_is_read();	// for output stream parameter
*/

class LVCSIM_API LVCSim_StreamIOBase {
private:
	int	m_refcnt;

public:
	LVCSim_StreamIOBase() : m_refcnt(1)
	{
	}

	virtual ~LVCSim_StreamIOBase();
	virtual size_t getRowSize() const = 0;
	virtual const void *readRow() = 0;
	virtual void writeRow(const void *data) = 0;

	inline int ref() 
	{ 
		return ++m_refcnt; 
	}

	inline int unref() 
	{ 
		int cnt = --m_refcnt;
        if (cnt == 0) {
            this->~LVCSim_StreamIOBase();
            omsFree(this);
        }
		return cnt;
	}
};

/*!	endclass: LVCSim_StreamIOBase */


#endif	// LVCSIM_GLOBFNC_HPP


#include <lcbasetypes.h>
#include <livecache.h>
#include <livecachesql.h>

#define	LVCSIM	74030402			// LVC Simulator API version


#ifndef	LVCSIM_ONLY_GLOBALS

// Compatibility classes for old simulator

#include <iostream.h>

LVCSIM_API inline ostream& operator<<(ostream &os, const OmsObjectId &r)
{
	// TODO: return os << makeStr(r);
	return os;
}

LVCSIM_API inline ostream& operator<<(ostream &os, const DbpError &r)
{
	return os << r.dbpDBError() << ": " << r.dbpErrorTextUnicode();
}

// Compatibility macros for old simulator

//#define	LCSIM			LVCSIM
//#define	LCSIM_EXPORT	LVCSIM_API
#define	lCStr				OmsSimDumpCollector

#define	enableFullHtmlDump	LVCSim::SetDumpCount
#define	flushVtrace			LVCSim::FlushVtrace
#define	resetLcSim			LVCSim::Reset
#define	writeCheckpoint(X)	LVCSim::SaveFullState(X)
#define	readCheckpoint(X)	LVCSim::RestoreFullState(X)
#define	openVtrace(X)		LVCSim::SetTraceFile(X)

// TODO:
#define	dumpLCMemoryStat	LVCSim::DumpMemoryStat
#define	enableVtrace(X)		LVCSim::SetTraceLevel(X)


class LVCSim_FoolNewClass
{
};

inline void *operator new(size_t sz, LVCSim_FoolNewClass *p)
{
    return p;
}

inline void operator delete(void *ptr, LVCSim_FoolNewClass *p)
{
}

/*!
	class:			LVCSim_StreamStorage

	description:	Default stream storage.
*/
template <class T>
class LVCSim_StreamStorage {
private:
	class Bucket {
	public:
		T		m_data;
		Bucket	*m_next;

		Bucket(const T *_data) : m_data(*_data), m_next(NULL) 
		{
		}

		~Bucket()
		{
            if (m_next) {
			    m_next->~Bucket();
                omsFree(m_next);
            }
		}
	};

	Bucket	*last;
	Bucket	*first;
	Bucket	*toFree;

public:
	LVCSim_StreamStorage() : last(NULL), first(NULL), toFree(NULL)
	{
	}

	~LVCSim_StreamStorage()
	{
        if (toFree) {
            toFree->~Bucket();
            omsFree(toFree);
        }
        if (first) {
            first->~Bucket();
            omsFree(first);
        }
	}

	const T *readRow()
	{
        if (toFree) {
            toFree->~Bucket();
            omsFree(toFree);
        }
		if (first) {
			toFree = first;
			first = first->m_next;
			toFree->m_next = NULL;
			if (!first) last = NULL;
			return &toFree->m_data;
		} else {
			toFree = NULL;
			return NULL;
		}
	}

	void writeRow(const T *m_data)
	{
		// append after last
        Bucket *p = new(reinterpret_cast<LVCSim_FoolNewClass*>(
                omsMalloc(sizeof(Bucket)))) Bucket(m_data);
		if (!last) {
			last = first = p;
		} else {
			last->m_next = p;
			last = p;
		}
	}
};

/*!	endclass: LVCSim_StreamStorage */



// default simulator stream IO handler
template <class T>
class LVCSim_StreamIO : public LVCSim_StreamIOBase {
private:
	LVCSim_StreamStorage<T>	*storage;

public:
	LVCSim_StreamIO(LVCSim_StreamStorage<T> *data)
	{
		storage = data;
	}

	virtual size_t getRowSize() const
	{
		return sizeof(T);
	}

	virtual const void *readRow()
	{
		return storage->readRow();
	}

	virtual void writeRow(const void *data)
	{
		storage->writeRow((const T*) data);
	}
};


/*!
	class:	LVCSim_StreamHandleBase

	description:	This is the overridden handle used for simulating streams
		in simulator. 
*/

class LVCSim_StreamHandleBase : public OmsTypeStreamHandle {
private:
	LVCSim_StreamIOBase* &base()
	{
		char *addr = reinterpret_cast<char*>(&colDesc[1]);
#if defined(HPUX) || defined(OSF1) || defined(_WIN64)
		addr += 4;
#endif
		return *reinterpret_cast<LVCSim_StreamIOBase**>(addr);
	}

protected:
	void registerStream(LVCSim_StreamIOBase *stream)
	{
		ABAPTabId = LVCSim::RegisterStream(stream);
		base() = stream;
		size = (unsigned short) stream->getRowSize();
		memberCount = 1;
		filler = 0;
		colDesc[0].length = size;
		colDesc[0].memberType = 0;
		colDesc[0].offset = 0;
		colDesc[0].dec = 0;
	}

public:
	LVCSim_StreamHandleBase()
	{
		base() = NULL;
	}

	explicit LVCSim_StreamHandleBase(const LVCSim_StreamHandleBase &x)
	{
		base() = NULL;
		*this = x;
	}

	~LVCSim_StreamHandleBase()
	{
		if (base())
		if (base()->unref() == 0) {
			LVCSim::UnregisterStream(ABAPTabId);
		}
	}

	LVCSim_StreamHandleBase &operator =(const LVCSim_StreamHandleBase &x)
	{
		if (base())
		if (base()->unref() == 0) {
			LVCSim::UnregisterStream(ABAPTabId);
		}
		memmove(this, &x, sizeof(x));
		if (base())	base()->ref();
		return *this;
	}
};

/*! endclass: LVCSim_StreamHandleBase */


/*!
	class:	LVCSim_StreamHandle

	description:	This is the overridden handle used for simulating streams
		in simulator. 
*/

class LVCSim_StreamHandle : public LVCSim_StreamHandleBase {
public:
	LVCSim_StreamHandle() : LVCSim_StreamHandleBase()
	{
	}

	explicit LVCSim_StreamHandle(const LVCSim_StreamHandle &x) 
		: LVCSim_StreamHandleBase(x)
	{
	}

	/*!
		function:		LVCSim_StreamHandle
		
		description:	This constructor creates a handle for given stream storage.
	*/
	template <class T>
	inline LVCSim_StreamHandle(LVCSim_StreamStorage<T> *source)
	{
		// register new stream with given source
		registerStream(new(reinterpret_cast<LVCSim_FoolNewClass*>(
            omsMalloc(sizeof(LVCSim_StreamIO<T>))))
            LVCSim_StreamIO<T>(source));
	}
};

/*! endclass: LVCSim_StreamHandle */


/*!
	class:	LVCSim_ABAPTabHandleBase

	description:	This is the overridden handle used for simulating ABAP tables
		in simulator. 
*/

class LVCSim_ABAPTabHandleBase : public OmsTypeABAPTabHandle {
private:
	LVCSim_StreamIOBase* &base()
	{
		char *addr = reinterpret_cast<char*>(&colDesc[1]);
#if defined(HPUX) || defined(OSF1) || defined(_WIN64)
		addr += 4;
#endif
		return *reinterpret_cast<LVCSim_StreamIOBase**>(addr);
	}

protected:
	void registerStream(LVCSim_StreamIOBase *stream)
	{
		ABAPTabId = LVCSim::RegisterStream(stream);
		base() = stream;
		rowSize = (unsigned short) stream->getRowSize();
		colCount = 1;
		rowCount = 1;
		colDesc[0].length = rowSize;
		colDesc[0].abap_type = 0;
		colDesc[0].offset = 0;
		colDesc[0].dec = 0;
	}

public:
	LVCSim_ABAPTabHandleBase()
	{
		base() = NULL;
	}

	explicit LVCSim_ABAPTabHandleBase(const LVCSim_ABAPTabHandleBase &x)
	{
		base() = NULL;
		*this = x;
	}

	~LVCSim_ABAPTabHandleBase()
	{
		if (base())
		if (base()->unref() == 0) {
			LVCSim::UnregisterStream(ABAPTabId);
		}
	}

	LVCSim_ABAPTabHandleBase &operator =(const LVCSim_ABAPTabHandleBase &x)
	{
		if (base())
		if (base()->unref() == 0) {
			LVCSim::UnregisterStream(ABAPTabId);
		}
		memmove(this, &x, sizeof(x));
		if (base()) base()->ref();
		return *this;
	}
};

/*! endclass: LVCSim_ABAPTabHandleBase */


/*!
	class:	LVCSim_ABAPTabHandle

	description:	This is the overridden handle used for simulating ABAP tables
		in simulator. 
*/

class LVCSim_ABAPTabHandle : public LVCSim_ABAPTabHandleBase {
public:
	LVCSim_ABAPTabHandle() : LVCSim_ABAPTabHandleBase()
	{
	}

	explicit LVCSim_ABAPTabHandle(const LVCSim_ABAPTabHandle &x) 
		: LVCSim_ABAPTabHandleBase(x)
	{
	}

	/*!
		function:		LVCSim_ABAPTabHandle
		
		description:	This constructor creates a handle for given stream storage.
	*/
	template <class T>
	inline LVCSim_ABAPTabHandle(LVCSim_StreamStorage<T> *source)
	{
		// register new stream with given source
		registerStream(new(reinterpret_cast<LVCSim_FoolNewClass*>(
            omsMalloc(sizeof(LVCSim_StreamIO<T>))))
            LVCSim_StreamIO<T>(source));
	}
};

/*! endclass: LVCSim_ABAPTabHandle */


class LVCSIM_API LVCSim_SqlHandle : public SqlHandle {
public:
	LVCSim_SqlHandle();
	~LVCSim_SqlHandle();
};

#ifdef	SAPDB_LVCSIM_DEPRECATED
// #define to allow old test programs run without change
#define	toHtmlShort				omsSimDumpToHtmlShort
#define	toHtmlLong				omsSimDumpToHtmlLong
#define	setName					omsSimSetObjectName
#define	getName					omsSimGetObjectName
#define	omsDump					omsSimDump

#ifdef _WIN32
#define LCSIM_USE_THREADS
#else
#define LCSIM_USE_THREADS
#define USE_POSIX_THREADS
#endif	// _WIN32

#ifdef USE_POSIX_THREADS
#include <signal.h>
#include <pthread.h>
#endif

#endif	// SAPDB_LVCSIM_DEPRECATED
#endif	// LVCSIM_ONLY_GLOBALS

#endif
