/*!***************************************************************************

  module      : LVCSim.cpp

  -------------------------------------------------------------------------

  responsible : IvanS

  special area: liveCache Simulator
  description : Global functions for simulator

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/


#include "LVCSimulator/LVCSim_Internal.hpp"
#include "LVCSimulator/LVCSim_liveCacheSink.hpp"
#include "LVCSimulator/LVCSim_KernelInterface.hpp"
#include "LVCSimulator/LVCSim_OIDAllocator.hpp"
#include "LVCSimulator/LVCSim_DebugAllocator.hpp"

#define	LC_DBPROC_PROTO
#include "gco04.h"
#include "heo07.h"
#include "heo670.h"
#include "hsp77.h"

void *operator new(size_t sz) THROW_NEW_GEO573
{
    return LVCSim_KernelInterface::Instance()->Allocate(sz);
}

void operator delete(void *ptr) THROW_DELETE_GEO573
{
    LVCSim_KernelInterface::Instance()->Deallocate(ptr);
}

void *operator new[](size_t sz) THROW_NEW_GEO573
{
    return LVCSim_KernelInterface::Instance()->AllocateArray(sz);
}

void operator delete[](void *ptr) THROW_DELETE_GEO573
{
    LVCSim_KernelInterface::Instance()->DeallocateArray(ptr);
}


#if defined(OSF1)
#ifdef __MODEL_ANSI 
extern "C" void *__7__nw__FUl(size_t Size) { 
#else  //__MODEL_ARM 
extern "C" void *__nw__XUl(size_t Size) { 
#endif
  return LVCSim_KernelInterface::Instance()->Allocate(Size);
}
#endif


#ifndef WIN32
#include <iostream.h>
#endif


externCpp
void co10_ResetOMSForSimulator();


// WARNING!!! This is only temporary solution! It reads the indirect pointer
// to the sink (m_session). This pointer must be as the first in OmsHandle.

#define	GET_SINK(omshandle) (*((LVCSim_liveCacheSink***)	\
	(((char*) &omshandle) + sizeof(DbpBase))))
#define	GET_CANCEL(omshandle) (*((bool**)	\
	(((char*) &omshandle) + sizeof(DbpBase) + 3 * sizeof(void*))))



static LVCSim_KernelInterface	*kernelInterface;
static LVCSim_liveCacheSink		*defaultSink;
static double                   defaultSink_space[sizeof(LVCSim_liveCacheSink) / sizeof(double) + 1];
OMS_LibOmsInterface				*omsInterface = NULL;

static tsp00_Int4               tls_session_key = 0;

#ifdef OSF1
static bool stackTraceEnabled = false;
#endif



void LVCSim::Init()
{
#ifdef OSF1
    stackTraceEnabled = true;
#endif
#ifndef WIN32
    cout << "Initializing simulator V " << LVCSIM << endl;
    LVCSim::DebugAllocatorEscape(LVCSIM_DA_CLEAR_LEAKS);
#endif
}

void LVCSim::Reset()
{
	//fprintf(stderr, "WARNING: Reset not yet fully implemented!\n");
    LVCSim::DebugAllocatorEscape(LVCSIM_DA_ENTER_NOLEAK);
	defaultSink->reset();
	co10_ResetOMSForSimulator();
	omsInterface->InitLibOms(defaultSink);
    LVCSim::DebugAllocatorEscape(LVCSIM_DA_EXIT_NOLEAK);
}


void LVCSim::Attach(OmsHandle &obj)
{
    LVCSim_liveCacheSink *session = reinterpret_cast<LVCSim_liveCacheSink*>(
        sqlgettls(tls_session_key));
    if (!session) {
        session = new LVCSim_liveCacheSink;
        teo07_ThreadErr terr;
        tsp00_ErrTextc ttext;
        sqlsettls(tls_session_key, session, ttext, &terr);
		if (terr != THR_OK_EO07) {
			throw DbpError(e_terminate, (char*)ttext);
		}
    }
    session->AddRef();
	if (!GET_SINK(obj)) {
		// this is a simple handle, we must give it a sink pointer
		bool *cancelFlag;
		GET_SINK(obj) = defaultSink->getSinkPtr(&cancelFlag);
		GET_CANCEL(obj) = cancelFlag;
	}
	if (*GET_SINK(obj)) Detach(obj);
	*GET_SINK(obj) = session;
	co90AttachDefaultContext(STATIC_CAST(OmsHandle*, &obj));
}

void LVCSim::AttachNewSink(OmsHandle &obj)
{
	if (!GET_SINK(obj)) {
		// this is a simple handle, we must give it a sink pointer
		bool *cancelFlag;
		GET_SINK(obj) = defaultSink->getSinkPtr(&cancelFlag);
		GET_CANCEL(obj) = cancelFlag;
	}
	if (*GET_SINK(obj)) Detach(obj);
	*GET_SINK(obj) = new LVCSim_liveCacheSink;
    (*GET_SINK(obj))->AddRef();
	co90AttachDefaultContext(STATIC_CAST(OmsHandle*, &obj));
}

void *LVCSim::GetCurrentSink()
{
    LVCSim_liveCacheSink *session = reinterpret_cast<LVCSim_liveCacheSink*>(
        sqlgettls(tls_session_key));
    if (!session) {
        session = new LVCSim_liveCacheSink;
        teo07_ThreadErr terr;
        tsp00_ErrTextc ttext;
        sqlsettls(tls_session_key, session, ttext, &terr);
		if (terr != THR_OK_EO07) {
            delete session;
            return NULL;
		}
    }
    return session;
}

extern HRESULT co90ReleaseDefaultContext(void* oms_handle);

void LVCSim::Detach(OmsHandle &obj)
{
  if ((*GET_SINK(obj))->GetRefCount() == 1) {
    // rollback current transaction + deallocate session - last user
    LVCSim::Rollback(obj, true);
    co90ReleaseDefaultContext(&obj);

    teo07_ThreadErr terr;
    tsp00_ErrTextc ttext;
    sqlsettls(tls_session_key, NULL, ttext, &terr);
    if (terr != THR_OK_EO07) {
      throw DbpError(e_terminate, (char*)ttext);
    }
  } else {
    co90ReleaseDefaultContext(&obj);
  }
  if ((*GET_SINK(obj))->Release() == 0) {
    teo07_ThreadErr terr;
    tsp00_ErrTextc ttext;
    sqlsettls(tls_session_key, NULL, ttext, &terr);
    if (terr != THR_OK_EO07) {
      throw DbpError(e_terminate, (char*)ttext);
    }
  }
  defaultSink->returnSinkPtr(GET_SINK(obj));
  GET_SINK(obj) = NULL;
}

void LVCSim::COMEnd(OmsHandle &obj)
{
    (*GET_SINK(obj))->UserAllocatorCtl(OmsCallbackInterface::BlockCOMRoutine, NULL);
}

void LVCSim::Commit(OmsHandle &obj)
{
/*
	struct {
		tgg00_MessType	m1;
		tgg00_MessType2	m2;
	} msg;
	int	result_len = sizeof(msg);

	msg.m1.becomes(m_commit);
	msg.m2.becomes(mm_open);
	co90SysRequest(&obj, NULL, &result_len, (unsigned char*) &msg);

	short err = *((short*) &msg);
	if (err != 0) {
		throw;
	}
*/
	short err = obj.omsCommit(false);
	if (err != 0) {
		(*GET_SINK(obj))->SimRollback();
		throw DbpError(err);
	}
	(*GET_SINK(obj))->SimCommit();
}

void LVCSim::Rollback(OmsHandle &obj, bool endSession)
{
/*
	struct {
		tgg00_MessType	m1;
		tgg00_MessType2	m2;
	} msg;
	int	result_len = sizeof(msg);

	msg.m1.becomes(m_rollback);
	if (endSession) msg.m2.becomes(mm_close);
	else msg.m2.becomes(mm_open);
	co90SysRequest(&obj, NULL, &result_len, (unsigned char*) &msg);

	short err = *((short*) &msg);
	if (err != 0) {
		throw;
	}
*/
	obj.omsRollback(endSession, false);
	(*GET_SINK(obj))->SimRollback();
}



// TODO: move into static sink functions and reset properly on LC reset

#include <map>

typedef LVCSIM_STD map<int, LVCSim_StreamIOBase *>  StreamDirMap;
static RTESync_Spinlock                             streamLock;
static int                                          streamID = 1;
static StreamDirMap                                 *streamDir = NULL;
static double streamDir_space[sizeof(LVCSIM_STD map<int,void*>) / sizeof(double) + 1];

int LVCSim::RegisterStream(class LVCSim_StreamIOBase *stream)
{
	RTESync_LockedScope	lck(streamLock);

	if (!streamDir) streamDir = new(&streamDir_space) LVCSIM_STD map<int, LVCSim_StreamIOBase *>();
	streamDir->insert(LVCSIM_STD pair<const int,LVCSim_StreamIOBase *>(streamID, stream));

	return streamID++;
}

void LVCSim::UnregisterStream(int id)
{
	RTESync_LockedScope	lck(streamLock);

    if (streamDir) {
     	LVCSIM_STD map<int, LVCSim_StreamIOBase *>::iterator i = streamDir->find(id);
	    if (i == streamDir->end()) return;
        streamDir->erase(i);
        if (streamDir->size() == 0) {
            streamDir->~StreamDirMap();
            streamDir = NULL;
        }
    }
}

class LVCSim_StreamIOBase *LVCSim::GetStream(int id)
{
	RTESync_LockedScope	lck(streamLock);

    if (!streamDir) return NULL;
	LVCSIM_STD map<int, LVCSim_StreamIOBase *>::iterator i = streamDir->find(id);
	if (i == streamDir->end()) return NULL;
	else return i->second;
}



void LVCSim::setName(const struct OmsObjectId &oid, const char *name)
{
	// set name of given OID
	// unused - will go out
}

const char *LVCSim::getName(const struct OmsObjectId &oid)
{
	// get name of given OID
	// unused - will go out
	return NULL;
}

static int	dump_count	= 1<<30;

void LVCSim::SetDumpCount(int cnt)
{
	dump_count = cnt;
}

void LVCSim::omsSimDump(OmsHandle &obj, int param)
{
	if (dump_count <= 0) return;

	// dump the state into a HTML file
	(*GET_SINK(obj))->omsSimDump(obj, param);
	--dump_count;

}

void LVCSim::FlushVtrace()
{
	// NOP - vtrace is flushed always automatically
}

void LVCSim::SaveFullState(const char *fname)
{
	LVCSim_CheckpointFileWriter o(fname);
	LVCSim_liveCacheSink::writeToStream(o);
}

void LVCSim::RestoreFullState(const char *fname)
{
	//fprintf(stderr, "WARNING: Restore checkpoint not yet fully implemented!\n");
	defaultSink->reset();
	co10_ResetOMSForSimulator();

	LVCSim_CheckpointFileReader i(fname);
	LVCSim_liveCacheSink::readFromStream(i);

	omsInterface->InitLibOms(defaultSink);
}

void LVCSim::DumpMemoryStat()
{
	// TODO: dump simulator memory status
}

void LVCSim::SetTraceFile(const char *fname)
{
	if (!fname) {
		LVCSim_liveCacheSink::SetVTraceFile(stderr);
	} else {
		FILE *f = fopen(fname, "w");
		if (!f) {
			fprintf(stderr, "Error opening trace file '%s' for output, "
				"tracing to console\n", fname);
			LVCSim_liveCacheSink::SetVTraceFile(stderr);
		} else {
			LVCSim_liveCacheSink::SetVTraceFile(f);
		}
	}
}

void LVCSim::SetTraceLevel(const char *level, bool enable)
{
	tsp00_KnlIdentifier lvl;
	lvl.c2p(level);
	omsInterface->SetTraceLevel(lvl, enable);
}

void LVCSim::SetTraceLevel(long level)
{
	// TODO: set OMS trace level numerically
}


bool LVCSim::CheckMemoryLeaks()
{
    return LVCSim_DebugAllocator::Escape(LVCSIM_DA_CHECK_LEAKS);
}

bool LVCSim::CheckMemory()
{
    return LVCSim_DebugAllocator::Escape(LVCSIM_DA_CHECK_MEMORY);
}

bool LVCSim::DebugAllocatorEscape(LVCSim_DebugAllocatorEscapeEnum what)
{
    return LVCSim_DebugAllocator::Escape(what);
}




class LVCSim_ThreadParams {
public:
    LVCSim_ThreadFunc   func;           // thread main function
    void                *arg;           // function argument
    size_t              stacksize;      // stack size
    void                *top;           // stack highest address
    void                *bottom;        // stack lowest address
};

static int tls_stack_key = 0;

#if (defined(WIN32) && !defined(_WIN64)) || (defined(LINUX) && !defined(IA64))
void LVCSim::SetStackBase(void *ptr)
{
    int err;
    char errbuf[64];    
    TLSCreate(tls_stack_key, &err, errbuf, sizeof(errbuf));
    if (err != LVCSim_ThreadOK) {
    	fprintf(stderr, "Error creating thread stack info: %s\n", errbuf);
		exit(1);
    }

    LVCSim_ThreadParams *sinfo = (LVCSim_ThreadParams*) malloc(sizeof(LVCSim_ThreadParams));
    sinfo->top = (void*) ((((int) ptr) + 4095) & ~4095);
    sinfo->bottom = ((char*) sinfo->top) - 1024*1024;
    sinfo->stacksize = 1024*1024;
    sinfo->arg = NULL;
    sinfo->func = NULL;
    TLSSet(tls_stack_key, sinfo, &err, errbuf, sizeof(errbuf));
    if (err != LVCSim_ThreadOK) {
    	fprintf(stderr, "Error setting thread stack info: %s\n", errbuf);
		exit(1);
    }
}
#endif

extern "C" void *LVCSim_ThreadMainFunction(void *arg)
{
    LVCSim_ThreadParams *p = (LVCSim_ThreadParams*) arg;
#if (defined(WIN32) && !defined(_WIN64)) || (defined(LINUX) && !defined(IA64))
    p->top = (void*) ((((int) &p) + 4095) & ~4095);
    p->bottom = ((char*) p->top) - p->stacksize;
#endif

    int err;
    char errbuf[64];        
    LVCSim::TLSSet(tls_stack_key, p, &err, errbuf, sizeof(errbuf));
    if (err != LVCSim_ThreadOK) {
    	fprintf(stderr, "Error setting thread stack info: %s\n", errbuf);
    }

    long retcode = 0;
    try {
        retcode = (p->func)(p->arg);
    } catch (const DbpError &e) {
    	fprintf(stderr, "***** Unhandled DbpError %d exception in thread\n", e.dbpError());
        retcode = -1;
    } catch (...) {
    	fprintf(stderr, "***** Unhandled exception in thread\n");
        retcode = -2;
    }

    LVCSim::ThreadEnd((int) retcode);
    return NULL;
}


LVCSim_Thread *LVCSim::ThreadBegin(size_t stacksize, LVCSim_ThreadCreateFlags flags,    
    LVCSim_ThreadFunc threadProc, void *arg, int *err, char *errbuf, size_t bufsize)
{
    teo07_Thread thread;
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    LVCSim_ThreadParams *p = reinterpret_cast<LVCSim_ThreadParams*>(malloc(sizeof(LVCSim_ThreadParams)));
    p->func = threadProc;
    p->arg = arg;
    p->stacksize = stacksize;
    sqlbeginthread(stacksize, LVCSim_ThreadMainFunction, p, flags, &thread, ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
        free(p);
    }
    return reinterpret_cast<LVCSim_Thread*>(thread);
}

void LVCSim::ThreadEnd(int return_code)
{
    LVCSim_ThreadParams *p = reinterpret_cast<LVCSim_ThreadParams*>(TLSGet(tls_stack_key));
    if (p) {
        free(p);
    }

    LVCSim_liveCacheSink *session = reinterpret_cast<LVCSim_liveCacheSink*>(
        sqlgettls(tls_session_key));
    if (session) {
        session->AddRef();
        if (session->Release()) {
            // release all open handles
            int count = 1;
            while (session->Release()) {
                ++count;
            }
	        fprintf(stderr, "WARNING: %d handles still open at thread termination, sink %p\n",
                count, session);
        }
    }
    sqlendthread((tsp00_Int4) return_code);
}
    
void LVCSim::ThreadYield()
{
    sqlyieldthread();
}

LVCSim_Thread *LVCSim::ThreadGetCurrent()
{
    return reinterpret_cast<LVCSim_Thread*>(sqlgetthread());
}
    
void LVCSim::ThreadKill(LVCSim_Thread *thread, int *err, char *errbuf, size_t bufsize)
{
    if (!thread) {
        sp77sprintf(errbuf, (int) bufsize, "Invalid NULL thread handle");
        *err = LVCSim_ThreadNotOK;
        return;
    }
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    sqlkillthread(reinterpret_cast<teo07_Thread>(thread), ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

void LVCSim::ThreadCancel(LVCSim_Thread *thread, int return_code, 
    int *err, char *errbuf, size_t bufsize)
{
    if (!thread) {
        sp77sprintf(errbuf, (int) bufsize, "Invalid NULL thread handle");
        *err = LVCSim_ThreadNotOK;
        return;
    }
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    sqlcancelthread(reinterpret_cast<teo07_Thread>(thread), return_code, ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

void LVCSim::ThreadForget(LVCSim_Thread *thread)
{
    if (thread) {
        sqlforgetthread(reinterpret_cast<teo07_Thread>(thread));
    }
}

void LVCSim::ThreadSuspend(LVCSim_Thread *thread, int *err, 
    char *errbuf, size_t bufsize)
{
    if (!thread) {
        sp77sprintf(errbuf, (int) bufsize, "Invalid NULL thread handle");
        *err = LVCSim_ThreadNotOK;
        return;
    }
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    sqlsuspendthread(reinterpret_cast<teo07_Thread>(thread), ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

void LVCSim::ThreadResume(LVCSim_Thread *thread, int *err, 
    char *errbuf, size_t bufsize)
{
    if (!thread) {
        sp77sprintf(errbuf, (int) bufsize, "Invalid NULL thread handle");
        *err = LVCSim_ThreadNotOK;
        return;
    }
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    sqlresumethread(reinterpret_cast<teo07_Thread>(thread), ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

int LVCSim::ThreadJoin(LVCSim_Thread *thread, int *err, 
    char *errbuf, size_t bufsize)
{
    if (!thread) {
        sp77sprintf(errbuf, (int) bufsize, "Invalid NULL thread handle");
        *err = LVCSim_ThreadNotOK;
        return -1;
    }
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    tsp00_Int4 status = -1;
    sqljointhread(reinterpret_cast<teo07_Thread>(thread), &status, ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
    return status;
}

bool LVCSim::ThreadIsSuspended(LVCSim_Thread *thread)
{
    if (thread) {
        return sqlIsThreadSuspended(reinterpret_cast<teo07_Thread>(thread)) != 0;
    } else {
        return false;
    }
}

bool LVCSim::ThreadIsAlive(LVCSim_Thread *thread, int *err, 
    char *errbuf, size_t bufsize)
{
    if (!thread) {
        sp77sprintf(errbuf, (int) bufsize, "Invalid NULL thread handle");
        *err = LVCSim_ThreadNotOK;
        return false;
    }
#ifdef WIN32
    DWORD exit_code;
    GetExitCodeThread(*reinterpret_cast<HANDLE*>(thread), &exit_code );
    return exit_code == STILL_ACTIVE;
#else
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    int status = sqlthreadalive(reinterpret_cast<teo07_Thread>(thread), ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
    return status != FALSE;
#endif
}

    
LVCSim_Mutex *LVCSim::MutexCreate()
{
    teo07_Mutex *mutex = new teo07_Mutex;
    sqlcreatemutex(mutex);
    return reinterpret_cast<LVCSim_Mutex*>(mutex);
}

void LVCSim::MutexDestroy(LVCSim_Mutex *mutex)
{
    sqldestroymutex(reinterpret_cast<teo07_Mutex*>(mutex));
    delete reinterpret_cast<teo07_Mutex*>(mutex);
}

void LVCSim::MutexBegin(LVCSim_Mutex *mutex)
{
    sqlbeginmutex(reinterpret_cast<teo07_Mutex*>(mutex));
}
    
bool LVCSim::MutexTryBegin(LVCSim_Mutex *mutex)
{
    return sqltrybeginmutex(reinterpret_cast<teo07_Mutex*>(mutex)) != 0;
}
    
void LVCSim::MutexEnd(LVCSim_Mutex *mutex)
{
    sqlendmutex(reinterpret_cast<teo07_Mutex*>(mutex));
}


void LVCSim::TLSCreate(int &key, int *err, char *errbuf, size_t bufsize)
{
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    tsp00_Int4 tkey;
    sqlcreatetls(&tkey, ttext, &terr);
    key = tkey;
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

void LVCSim::TLSDestroy(int key, int *err, char *errbuf, size_t bufsize)
{
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    sqldestroytls((tsp00_Int4) key, ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

void LVCSim::TLSSet(int key, void *val, int *err, char *errbuf, size_t bufsize)
{
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    sqlsettls((tsp00_Int4) key, val, ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
}

void *LVCSim::TLSGet(int key)
{
    return sqlgettls(key);
}


LVCSim_Semaphore *LVCSim::SemCreate(int initval, int *err, char *errbuf, size_t bufsize)
{
    teo07_ThreadErr terr;    
    tsp00_ErrTextc ttext;
    teo07_ThreadSemaphore tsem = NULL;
    sqlcreatesem(&tsem, (tsp00_Int4) initval, ttext, &terr);
    if (terr == THR_OK_EO07) {
        if (err) *err = LVCSim_ThreadOK;
        if (errbuf) *errbuf = 0;
    } else {
        if (err) *err = LVCSim_ThreadNotOK;
        if (errbuf) {
            strncpy(errbuf, ttext.asCharp(), bufsize);
            errbuf[bufsize - 1] = 0;
        }
    }
    return reinterpret_cast<LVCSim_Semaphore*>(tsem);
}

void LVCSim::SemDestroy(LVCSim_Semaphore *sem)
{
    sqldestroysem(reinterpret_cast<teo07_ThreadSemaphore>(sem));
}

void LVCSim::SemSignal(LVCSim_Semaphore *sem)
{
    sqlsignalsem(reinterpret_cast<teo07_ThreadSemaphore>(sem));
}

void LVCSim::SemWait(LVCSim_Semaphore *sem)
{
    sqlwaitsem(reinterpret_cast<teo07_ThreadSemaphore>(sem));
}

bool LVCSim::SemTimedWait(LVCSim_Semaphore *sem, int seconds)   // true on success
{
    teo07_ThreadErr terr;    
    sqltimedwaitsem(reinterpret_cast<teo07_ThreadSemaphore>(sem), seconds, &terr);
    return terr == THR_OK_EO07;
}


int LVCSim::GetStackTrace(void** &trace, size_t traceSize, size_t traceSkip)
{
#ifdef OSF1
    if (!stackTraceEnabled) {
        return 0;
    }
#endif
#if (defined(WIN32) && !defined(_WIN64)) || (defined(LINUX) && !defined(IA64))
    LVCSim_ThreadParams *sinfo = (LVCSim_ThreadParams*) TLSGet(tls_stack_key);
    if (!sinfo) {
        // cannot create stack trace - no stack info
        void **trace2 = trace;
        while (traceSize--) 
            *(trace2++) = 0;
        return 0;
    }
    register void *top = sinfo->top, *bottom = sinfo->bottom;
	register void **traceStart;
#ifdef LINUX
    asm("movl %%ebp, %0" : "=r" (traceStart));
#else
	__asm {
		mov traceStart, ebp
	}
#endif
	size_t level = 0;
	int skip = traceSkip;
    while (level < traceSize && traceStart >= bottom && traceStart < top) {
        if (--skip <= 0)
            trace[level++] = traceStart[1];
    	traceStart = (void**) *traceStart;
  	}
    size_t level2 = level;
    while (level2 < traceSize)
        trace[level2++] = NULL;
#else
    int level = eo670_FillCallArray(traceSize + traceSkip, trace);
    trace += traceSkip;
#endif
    return level;
}



// Initialization

externC void* co10_GetInterface();

class LVCSim_Initer {
public:
	LVCSim_Initer()
	{
#if (defined(WIN32) && !defined(_WIN64)) || (defined(LINUX) && !defined(IA64))
        int variableonstack;
        LVCSim::SetStackBase(&variableonstack);
#endif
		teo07_Thread currentThread;
		tsp00_ErrTextc errtext;
		teo07_ThreadErr err;

        LVCSim::DebugAllocatorEscape(LVCSIM_DA_ENTER_NOLEAK);
		sqlinit_multithreading(&currentThread, errtext, &err);
		if (err != THR_OK_EO07) {
			fprintf(stderr, "Error initializing threads: %s\n", ((char*)errtext));
			exit(1);
		}
        sqlcreatetls(&tls_session_key, errtext, &err);
		if (err != THR_OK_EO07) {
			fprintf(stderr, "Error getting session TLS key: %s\n", ((char*)errtext));
			exit(1);
		}

        kernelInterface = LVCSim_KernelInterface::Instance();
		omsInterface = (OMS_LibOmsInterface*) co10_GetInterface();
        AdviseSimulatorSink();
         
		defaultSink = new(&defaultSink_space) LVCSim_liveCacheSink(false);
		defaultSink->init();
		omsInterface->InitLibOms(defaultSink);
        LVCSim::DebugAllocatorEscape(LVCSIM_DA_EXIT_NOLEAK);
	}

	~LVCSim_Initer()
	{
#ifdef OSF1
        stackTraceEnabled = false;
#endif
		kernelInterface = NULL;
		defaultSink->done();
		defaultSink->~LVCSim_liveCacheSink();
		omsInterface->UnAdviseKernelInterface();
		kernelInterface->~LVCSim_KernelInterface();
        if (LVCSim::CheckMemoryLeaks()) {
        	printf("Memory leaks found. Detail info in mem_trace.txt.\n"
                "Press Enter...\n");
            getchar();
        }
	}
private:
    void AdviseSimulatorSink() const
    {
        const bool hostIsKernel = true;
        omsInterface->AdviseKernelInterface(*kernelInterface, !hostIsKernel);
    }
};

static LVCSim_Initer	initer_instance;