/*!***************************************************************************

  module      : LVCSim_ObjectHandle.cpp

  -------------------------------------------------------------------------

  responsible : IvanS

  special area: liveCache Simulator
  description : Internal object handle 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "LVCSimulator/LVCSim_ObjectHandle.hpp"
#include "LVCSimulator/LVCSim_OIDAllocator.hpp"
#include "LVCSimulator/LVCSim_liveCacheSink.hpp"
#include "LVCSimulator/LVCSim_ObjectTree.hpp"





#define	MAX_OBJ_LOCKS	59

// Locker class for concurrent object access

class ObjLockNoGC {
private:
	static RTESync_Spinlock	locks[MAX_OBJ_LOCKS];
protected:
    RTESync_Spinlock        *m_spinlock;
public:
	ObjLockNoGC(LVCSim_ObjectHandle *ref, LVCSim_SID sid)
		: m_spinlock(&locks[ref->hash() % MAX_OBJ_LOCKS])
	{
        m_spinlock->Lock();
	}
    ~ObjLockNoGC()
    {
        if (m_spinlock) {
            m_spinlock->Unlock();
        }
    }
};

class ObjLock : public ObjLockNoGC {
public:
	ObjLock(LVCSim_ObjectHandle *ref, LVCSim_SID sid)
		: ObjLockNoGC(ref, sid)
	{
   		// optimization & GC - called before lock itself
        if (ref->m_lock == NULL && !ref->isCleared()) {
        	LVCSim_TID tid = LVCSim_liveCacheSink::getOldestTID();
            if (ref->m_tid_fmax < tid) {
                ref->GC(tid, sid, m_spinlock);
            }
        }
	}
};

RTESync_Spinlock	ObjLockNoGC::locks[MAX_OBJ_LOCKS];



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

void LVCSim_ObjectHandle::GC(LVCSim_TID tid, LVCSim_SID sid, RTESync_Spinlock* &spinlock)
{
	if (m_tid_max < tid) {
		// the whole object is not valid anymore
        m_lock = sid;
        spinlock->Unlock();
        spinlock = NULL;
		delete2(LVCSim_liveCacheSink::getContainer(getContainer()), sid);
		return;
	}

	// remove all too old m_versions
	LVCSim_ObjectVerHandle *v = m_versions;
	LVCSim_ObjectVerHandle *last = NULL;
	while (v) {
		if (v->m_tid_min < tid) {
			// we do not need older versions
            if (last) {
                m_tid_fmax = last->m_tid_min;
            } else {
                m_tid_fmax = m_tid_max;
            }
			last = v->m_next;
			v->m_next = NULL;
            m_tid_min = v->m_tid_min;
			delete last;
			return;
		}
		last = v;
		v = v->m_next;
	}

	// we shouldn't get here
    LVCSIM_ASSERT(false);
}

short LVCSim_ObjectHandle::acquireLock(LVCSim_SID sid)	// m_lock the object
{
	if (m_lock == sid) return e_ok;
    if (!isVisibleNoLock(sid)) {
        return e_object_not_found;
    }

	return e_object_not_locked;	// we can't m_lock ourselves - app must do it self
}

bool LVCSim_ObjectHandle::isVisibleNoLock(LVCSim_SID sid) const
{
    if (m_lock == sid) {
        // locked by us, we see it, if not yet deleted
        return (m_tid_max == LVCSim_TID_MAX);
    }

    if (!sid->sees(m_tid_min)) {
        // we do not see minimum TID
        return false;
    }
    if (sid->sees(m_tid_max)) {
        // we see maximum TID, ie, deleted
        return false;
    }
    // else there is a version that we see
    return true;
}

bool LVCSim_ObjectHandle::isVisible(LVCSim_SID sid)
{
	ObjLock lck(this, sid);
	return isVisibleNoLock(sid);
}

short LVCSim_ObjectHandle::isVisibleForChangeNoLock(LVCSim_SID sid) const
{
    if (m_lock == sid) {
        // already locked by us
        if (m_tid_max == LVCSim_TID_MAX) {
            return e_ok;
        } else {
            // already deleted - we see it not anymore
            return e_object_not_found;
        }
    }
    if (sid->hasReadView()) {
        if (!isVisibleNoLock(sid)) {
            return e_object_not_found;
        }
    }
    if (m_lock != sid && m_lock != NULL) {
    	// TODO: now wait until the m_lock is possible, then retest
    	return e_request_timeout;
    }
    if (m_tid_max != LVCSim_TID_MAX || m_versions->m_tid_min > sid->getCommitTime()) {
        return e_object_dirty;
    }
    if (sid->hasReadView()) {
        if (!sid->sees(m_versions->m_tid_min)) {
            return e_object_dirty;
        }
    }
    return e_ok;
}

bool LVCSim_ObjectHandle::isVisibleForChange(LVCSim_SID sid)
{
	ObjLock lck(this, sid);
	return isVisibleForChangeNoLock(sid) == e_ok;
}

bool LVCSim_ObjectHandle::isNewable(LVCSim_SID sid)
{
	ObjLockNoGC lck(this, sid);

    if (m_lock == sid) {
        // if we deleted the last version, we may insert object with the same key
        return (m_tid_max != LVCSim_TID_MAX);
    }

    if (sid->hasReadView()) {
        if (!sid->sees(m_tid_min)) {
            // we do not see minimum TID - created after us
            return false;
        }
        if (sid->sees(m_tid_max)) {
            // we see maximum TID, ie, deleted -> we may create new
            return true;
        }
    } else {
        // no read view, use only max time for comparison
        if (sid->getCommitTime() > m_tid_max) {
            // we work after the TX that deleted the object, check if commited
            if (LVCSim_liveCacheSink::checkCommited(m_tid_max)) {
                return true;
            }
        }
    }
    // else there is a version (not last) that we see or some other problem
    return false;
}

short LVCSim_ObjectHandle::lockObject(LVCSim_SID sid, size_t updateLen, bool ignoreView)
{
	ObjLock lck(this, sid);

    if (ignoreView) {
        if (m_lock == sid) {
            // already locked by us
            if (m_tid_max == LVCSim_TID_MAX) {
                return e_ok;
            } else {
                // already deleted - we see it not anymore
                return e_object_not_found;
            }
        }
        if (sid->hasReadView()) {
            if (!isVisibleNoLock(sid)) {
                return e_object_not_found;
            }
        }
        if (m_lock != sid && m_lock != NULL) {
            // TODO: now wait until the m_lock is possible, then retest
            return e_request_timeout;
        }
    } else {
        short err = isVisibleForChangeNoLock(sid);
        if (err != e_ok) {
            return err;
        }
    }

    // now we know, we either have the lock already, or noone has the lock
	if (!m_lock) {
        // not yet locked, lock it
	    m_lockChain = sid->m_lockChain;
		sid->m_lockChain = this;
   		m_lock = sid;
    }
	if (updateLen) {
        // simulate update on the object by creating a new version
		m_versions = new(updateLen) LVCSim_ObjectVerHandle(
            sid->getCommitTime(),
			m_versions->m_data.fix, updateLen, false, m_versions);
	}

	return e_ok;	// locked
}


short LVCSim_ObjectHandle::unlockObject(LVCSim_SID sid)
{
	ObjLock lck(this, sid);
	if (m_lock == sid) {
        // test if modified
        if (m_tid_max != LVCSim_TID_MAX || 
            (m_versions && m_versions->m_tid_min == sid->getCommitTime())) {
            return e_object_dirty;
        }
		// must remove from m_chain - TODO: this is too expensive
		LVCSim_ObjectHandle *last = NULL, *cur = sid->m_lockChain;
		while (cur && cur != this) {
			last = cur;
			cur = cur->m_lockChain;
		}
		if (cur) {
			if (last) last->m_lockChain = cur->m_lockChain;
			else sid->m_lockChain = cur->m_lockChain;
			cur->m_lockChain = NULL;
		} else {
			LVCSIM_THROW("unlocking not locked object???");
		}
		m_lock = NULL;
		return e_ok;
	}
    if (!isVisibleNoLock(sid)) {
        return e_object_not_found;
    }
	return e_object_not_locked;
}


LVCSim_ObjectVerHandle *LVCSim_ObjectHandle::lookupObject(LVCSim_SID sid, int &hist)
{
	ObjLock lck(this, sid);

    if (!isVisibleNoLock(sid)) {
        return NULL;
    }
	LVCSim_TID tid = sid->m_readView;
	LVCSim_ObjectVerHandle *v = m_versions;
	while (v) {
		if (v->m_tid_min <= tid && sid->sees(v->m_tid_min)) {
			// ok, we see this object
			return v;
		}
		v = v->m_next;
		++hist;
	}

	return NULL;	// visible version not found??? shouldn't happen
}


short LVCSim_ObjectHandle::update(LVCSim_SID sid, const void *data, size_t len,
	bool varsize)
{
	ObjLock lck(this, sid);

	short e = acquireLock(sid);
    if (e != e_ok) {
        return e;
    }

    if (m_tid_max != LVCSim_TID_MAX) {
        e = e_object_dirty;
    } else {
        LVCSim_TID ctime = sid->getCommitTime();
        if (m_versions) {
            if (m_tid_fmax == LVCSim_TID_MAX) {
                m_tid_fmax = ctime;
            }
        } else {
            m_tid_fmax = LVCSim_TID_MAX;
        }
		m_versions = new(len + (varsize ? sizeof(size_t) : 0)) 
			LVCSim_ObjectVerHandle(ctime, data, len, varsize, m_versions);
    }
	return e;
}

short LVCSim_ObjectHandle::remove(LVCSim_SID sid)
{
	ObjLock lck(this, sid);

	short e = acquireLock(sid);
    if (e != e_ok) {
        return e;
    }

    if (m_tid_max == LVCSim_TID_MAX) {
        m_tid_max = sid->getCommitTime();
        if (m_tid_fmax == LVCSim_TID_MAX) {
            m_tid_fmax = m_tid_max;
        }
    } else {
		e = e_object_dirty;
    }
	return e;
}

void LVCSim_ObjectHandle::doEndTxLL(LVCSim_SID sid, bool commit, char operation)
{
    if (commit) {
        // no need to do anything
        return;
    }

   	ObjLockNoGC lck(this, sid);

    switch (operation) {
    case o_delete:
        // set max time back
        if (m_versions->m_next) {
            // there is more than one version - find the oldest
            LVCSim_ObjectVerHandle *vh = m_versions;
            while (vh->m_next) {
                m_tid_fmax = vh->m_tid_min;
                vh = vh->m_next;
            }
        } else {
            // only single version
            m_tid_fmax = LVCSim_TID_MAX;
        }
        m_tid_max = LVCSim_TID_MAX;
        break;

    case o_new:
        // clear - no need for this object anymore, it was just created in this TX
        clear();
        break;

    case o_update:
        // remove last added version
        {
            LVCSim_ObjectVerHandle *vh = m_versions, *vd = vh;
            m_tid_fmax = LVCSim_TID_MAX;
            m_versions = vh = vh->m_next;
            while (vh->m_next) {
                m_tid_fmax = vh->m_tid_min;
                vh = vh->m_next;
            }
            vd->m_next = NULL;
            delete vd;
        }
        break;

    default:
        LVCSIM_ASSERT(false);
    }
}

bool LVCSim_ObjectHandle::isCleared() const
{
    return (m_tid_min == LVCSim_TID_MAX && m_tid_max == 0 && m_tid_fmax == 0);
}

void LVCSim_ObjectHandle::clear()
{
	m_tid_min = LVCSim_TID_MAX;
	m_tid_max = m_tid_fmax = 0;

	LVCSim_ObjectVerHandle *v = m_versions;
	m_versions = NULL;
	//m_lockChain = NULL;
	delete v;
}

LVCSim_ObjectHandle::~LVCSim_ObjectHandle()
{
	clear();

    m_lock = NULL;
}

void *LVCSim_ObjectHandle::operator new(size_t sz, LVCSim_ContainerMeta *cont)
{
	return cont->allocate(sz);
}

void *LVCSim_ObjectHandle::operator new(size_t sz, LVCSim_OIDAllocator *alloc, unsigned int oid)
{
	return alloc->map(oid);
}

void LVCSim_ObjectHandle::operator delete(void *ptr, LVCSim_ContainerMeta *cont)
{
	if (ptr) cont->deallocate(ptr);
}

void LVCSim_ObjectHandle::operator delete(void *ptr, LVCSim_OIDAllocator *alloc, unsigned int oid)
{
	// nothing to do, this is only to suppress warning
}

void LVCSim_ObjectHandle::delete2(LVCSim_ContainerMeta *cont, LVCSim_SID sid)
{
	// do key delete here as well
    if (cont->m_tree) {
        cont->m_tree->delKey(this, sid);
    }

	ObjLockNoGC lck(this, sid);
    clear();

    m_lock = NULL;
	cont->deallocate(this);
}


LVCSim_ObjectHandle::LVCSim_ObjectHandle(LVCSim_CheckpointReader &i, size_t objsize) 
	: m_lock(NULL), m_node(NULL), m_lockChain(NULL)
{
	i.read(this, sizeof(LVCSim_OIDObject));
	i.read(&m_tid_max, sizeof(m_tid_max));
	i.read(&m_tid_min, sizeof(m_tid_min));
	i.read(&m_tid_fmax, sizeof(m_tid_fmax));
	//i.read(&m_operation, 1);

	size_t size = 0;
	i.read(&size, sizeof(size));
	if (size) {
		m_versions = new(size) LVCSim_ObjectVerHandle(i, objsize);
	} else m_versions = NULL;
}

void LVCSim_ObjectHandle::writeToStream(LVCSim_CheckpointWriter &o, size_t objsize) const
{
	o.write(this, sizeof(LVCSim_OIDObject));
	o.write(&m_tid_max, sizeof(m_tid_max));
	o.write(&m_tid_min, sizeof(m_tid_min));
	o.write(&m_tid_fmax, sizeof(m_tid_fmax));
	//o.write(&m_operation, 1);

	if (m_versions) {
		m_versions->writeToStream(o, objsize);
	} else {
		size_t size = 0;
		o.write(&size, sizeof(size));
	}
}
