/*!***************************************************************************

  module      : LVCSim_ObjectHandle.hpp

  -------------------------------------------------------------------------

  responsible : IvanS

  special area: liveCache Simulator
  description : Internal object handle 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002



    ========== licence begin LGPL
    Copyright (C) 2002 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef LVCSIM_OBJECTHANDLE_HPP
#define LVCSIM_OBJECTHANDLE_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

//#include "SAPDBCommon/SAPDB_Types.hpp"
#include "LVCSimulator/LVCSim_ObjectVerHandle.hpp"
#include "LVCSimulator/LVCSim_CheckpointIO.hpp"
#include "LVCSimulator/LVCSim_liveCacheSink.hpp"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! 
  Class: LVCSim_ObjectHandle
  
  Description: Defines the control block of a persistent object in memory.
	This handle is allocated for each persistent object exactly once,
	independent of how many past versions exist. OIDs in simulator are actually
	logical pointers to objects of this type.
*/

class LVCSim_ObjectHandle : public LVCSim_OIDObject
{
private:
	short acquireLock(LVCSim_SID sid);	// lock the object
	void clear();						// clear metadata before delete
    bool isCleared() const;             // check, if already cleared

protected:
	friend	class LVCSim_liveCacheSink;
	friend	class ObjLock;
	friend	class LVCSim_ObjectTree;
	friend	class LVCSim_ObjectTreeNode;

	LVCSim_SID				m_lock;		// sink ID of modifying transaction
    LVCSim_ObjectHandle     *m_lockChain;   // lock chain
	LVCSim_TID				m_tid_max;	// maximum TX time (to detect deleted)
	LVCSim_TID				m_tid_min;	// minimum TX time (to detect new)
	LVCSim_TID				m_tid_fmax;	// maximum TX time of oldest version

	class LVCSim_ObjectTreeNode	*m_node;	// containing object tree node, if any

	LVCSim_ObjectVerHandle	*m_versions;    // object version chain

	//char					m_operation;    // operation performed on object

	bool isVisibleNoLock(LVCSim_SID sid) const;			// test for visibility
	short isVisibleForChangeNoLock(LVCSim_SID sid) const;// test for visibility for changes
	void GC(LVCSim_TID tid, LVCSim_SID sid,			    // garbage collection
        RTESync_Spinlock* &spinlock);

public:
	typedef enum {
		o_new, o_update, o_delete
	} OpType;

	LVCSim_ObjectHandle(LVCSim_CheckpointReader &i, size_t objsize);
	LVCSim_ObjectHandle(OmsTypeOid *oid, tgg91_PageRef *ref, LVCSim_SID tx)
	    :
		m_lock(tx), m_lockChain(tx->m_lockChain),
        m_tid_max(LVCSim_TID_MAX), m_tid_min(tx->m_commitTime), m_tid_fmax(LVCSim_TID_MAX),
		m_node(NULL), m_versions(NULL)
	{
		activateGeneration();
		getOid(oid);
		getRef(ref);
        tx->m_lockChain = this;
	}
	
	~LVCSim_ObjectHandle();

	/*!
	  function:		lookupObject

	  description:	looks up the object for consistent view identified by
		transaction time in tid.

	  arguments:	
		sid [in] sink ID of modifying transaction
		histCount [in/out] increased for each history read

	  return value:	pointer to object version header or NULL if the object is invisible
		in this consistent view
	 */
	LVCSim_ObjectVerHandle *lookupObject(LVCSim_SID sid, int &histCount);

	/*!
	  function:		isVisible

	  description:	test if this object is visible in current transaction

	  arguments:
		sid [in] sink ID of modifying transaction

	  returns:		true, if the object is visible
	 */
	bool isVisible(LVCSim_SID sid);

	/*!
	  function:		isVisibleForChange

	  description:	test if this object is visible for change in current transaction

	  arguments:
		sid [in] sink ID of modifying transaction

	  returns:		true, if the object is visible
	 */
	bool isVisibleForChange(LVCSim_SID sid);

	/*!
	  function:		isNewable

	  description:	test if this object's key can be reused

	  arguments:
		sid [in] sink ID of modifying transaction

	  returns:		true, if the object is not visible and it's key can be reused
	 */
	bool isNewable(LVCSim_SID sid);

	/*!
	  function:		lockObject

	  description:	try to exclusively lock this object

	  arguments:
		sid [in] sink ID of modifying transaction
		updateLen [in] if nonzero, simulate update (updateLen is then object size)

	  return value:	e_ok, if locked successfully, error otherwise
	 */
	short lockObject(LVCSim_SID sid, size_t updateLen = 0, bool ignoreView = false);

	/*!
	  function:		unlockObject

	  description:	unlock the object

	  arguments:	none

	  return value:	e_ok, if unlocked successfully, error otherwise
	 */
	short unlockObject(LVCSim_SID sid);

	/*!
	  function:		isLocked

	  description:	get locked status

	  arguments:	none

	  return value:	true, if the object is locked, false otherwise
	 */
	inline bool isLocked() const
	{
		return m_lock != NULL;
	}

	/*!
	  function:		update

	  description:	update last object version, if possible, or create new version

	  arguments:
		sid		[in]	sink ID for current transaction
		data	[in]	pointer to new data
		len		[in]	data size
		varsize	[in]	true if variable-sized object

	  returns: e_XX error code
	 */
	short update(LVCSim_SID sid, const void *data, size_t len, bool varsize);

	/*!
	  function:		remove

	  description:	logically delete the object

	  arguments:
		sid		[in]	sink ID for current transaction

	  returns: e_XX error code
	 */
	short remove(LVCSim_SID sid);


	void doEndTxLL(LVCSim_SID sid, bool commit, char operation);	// do commit/rollback for this item (LL)


	static void *operator new(size_t sz, LVCSim_ContainerMeta *cont);
	static void operator delete(void *ptr, LVCSim_ContainerMeta *cont);
	static void *operator new(size_t sz, LVCSim_OIDAllocator *alloc, unsigned int oid);
	static void operator delete(void *ptr, LVCSim_OIDAllocator *alloc, unsigned int oid);
	void delete2(LVCSim_ContainerMeta *cont, LVCSim_SID sid);

	void writeToStream(LVCSim_CheckpointWriter &o, size_t objsize) const;
};


#endif
