
/*!**************************************************************************

    module      : Log_ActionCopyUpdBlob.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to handle log entries for binary large objects

    last changed: 2001-03-13

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_COPY_UPD_BLOB_H
#define LOG_ACTION_COPY_UPD_BLOB_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"

#include "ggg00.h"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


//===================================================================================
/*! class:       Log_ActionCopyBlob

    description: before/after image action that copies a part of a blob to another blob

                 If the new length of the blob is less than the old length, the blob is truncated.
                 In case of before image, the source object consists of an auxiliary file.
                 The auxiliary file has to be destroyed at the end of the execution phase.
*/
class Log_ActionCopyBlob: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename SourceFilename;
        tgg00_Filename DestinFilename;
        SAPDB_Int4     DestinVersion;
        SAPDB_UInt4    CopyLength;
        SAPDB_UInt4    SourceOffset;
        SAPDB_UInt4    DestinOffset;
        SAPDB_UInt4    NewTruncatedDestinLen;
    } CopyBlobEntry;
    
    CopyBlobEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:    Log_ActionCopyBlob ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionCopyBlob ();

    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:    Log_ActionCopyBlob ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   SourceFilename      [in]
                     DestinFilename      [in]
                     DestinVersion       [in]
                     CopyLength          [in]
                     SourceOffset        [in]
                     DestinOffset        [in]
                     NewTruncatedBlobLen [in] if greater 0, the blob will be truncated to this length
    */
    Log_ActionCopyBlob (
        const tgg00_Filename &SourceFilename,
        const tgg00_Filename &DestinFilename,
        SAPDB_Int4            DestinVersion,
        SAPDB_UInt4           CopyLength,
        SAPDB_UInt4           SourceOffset,
        SAPDB_UInt4           DestinOffset,
        SAPDB_UInt4           NewTruncatedBlobLen);

    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:    Execute
    
        description: copies the blob and truncates the destination blob if necessary

                     If the source object consists of an auxiliary file, the file will be destroyed.
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:    ExecuteAction
    
        description: copies the blob and truncates the destination blob if necessary

                     If the source object consists of an auxiliary file, the file will be destroyed.
      
        arguments:   TransContext         [in out]
                     pSourceCommandFileId [in out] in case of undo/redo the file id has to be omitted
                     pDestinCommandFileId [in out] in case of undo/redo the file id has to be omitted
    */
    virtual void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pSourceCommandFileId = NULL,
        tgg00_FileId       *pDestinCommandFileId = NULL) const;

    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }

    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionCopyBlob---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }

};
/*! endclass: Log_ActionCopyBlob */


//===================================================================================
/*! class:       Log_ActionCopySameBlob

    description: after image action that copies a part of a blob into the same blob

                 The auxiliary file will be destroyed at the end of the execution phase.
*/
class Log_ActionCopySameBlob: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename BlobFilename;
        tgg00_Filename AuxBlobFilename;
        SAPDB_Int4     NewBlobVersion;
        SAPDB_UInt4    CopyLength;
        SAPDB_UInt4    SourceOffset;
        SAPDB_UInt4    DestinOffset;
    } CopyBlobEntry;
    
    CopyBlobEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:    Log_ActionCopySameBlob ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionCopySameBlob ();

    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:    Log_ActionCopySameBlob ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   BlobFilename    [in]
                     AuxBlobFilename [in]
                     NewBlobVersion  [in]
                     CopyLength      [in]
                     SourceOffset    [in]
                     DestinOffset    [in]
    */
    Log_ActionCopySameBlob (
        const tgg00_Filename &BlobFilename,
        const tgg00_Filename &AuxBlobFilename,
        SAPDB_Int4            NewBlobVersion,
        SAPDB_UInt4           CopyLength,
        SAPDB_UInt4           SourceOffset,
        SAPDB_UInt4           DestinOffset);

    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:    Execute
    
        description: redo the copy of a blob into the same blob using an auxiliary blob.
        
                     The blob is truncated depened on NewTruncatedBlobLen.
                     The copy actions are supressed, if the blob shows the NewBlobVersion.
                     The copy from the blob to the auxiliary blob is supressed, if the auxiliary
                     blob has the correct length already.
                     The auxiliary file will be destroyed at the end of the execution phase.
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const;

    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:    ExecuteAction
    
        description: copies the blob into itself using an auxiliary blob

                     The auxiliary file will be destroyed at the end of the execution phase.
      
        arguments:   TransContext [in out]
                     BlobId       [in out]
                     AuxBlobId    [in]     the auxiliary file will be created by this function
    */
    virtual void ExecuteAction (
        tgg00_TransContext   &TransContext,
        tgg00_FileId         &BlobId,
        tgg00_FileId         &AuxBlobId) const;

    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }

    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionCopySameBlob---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }

};
/*! endclass: Log_ActionCopySameBlob */


//===================================================================================
/*! class:       Log_ActionExpandBlob

    description: defines an after image action that expands a binary large object by a specified byte


*/
class Log_ActionExpandBlob: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename BlobFilename;
        SAPDB_UInt4    NewBlobLength;
        SAPDB_Byte     ExpandByte;
    } ExpandBlobEntry;
    
    ExpandBlobEntry m_Entry;
    
public:
    
    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    Log_ActionExpandBlob ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionExpandBlob ();

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    Log_ActionExpandBlob ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   BlobFilename  [in]
                     NewBlobLength [in]
                     ExpandByte    [in]
    */
    Log_ActionExpandBlob (
        const tgg00_Filename &BlobFilename,
        SAPDB_UInt4           NewBlobLength,
        SAPDB_Byte            ExpandByte);

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    Execute
    
        description: expands the binary large object using the specified expand byte
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    ExecuteAction
    
        description: expands the binary large object using the specified expand byte
      
        arguments:   TransContext       [in out]
                     pCommandBlobFileId [in out] in case of redo the file id has to be omitted
    */
    virtual void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandBlobFileId = NULL) const;

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionExpandBlob */


//===================================================================================
/*! class:       Log_ActionTruncateBlob

    description: defines a before or after image action that truncates a binary large object
*/
class Log_ActionTruncateBlob: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename BlobFilename;
        SAPDB_Int4     BlobVersion;
        SAPDB_UInt4    NewTruncatedBlobLen;
    } TruncateBlobEntry;
    
    TruncateBlobEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionTruncateBlob---
    /*! function:    Log_ActionTruncateBlob ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionTruncateBlob ();

    //--------------------------------------------------------------Log_ActionTruncateBlob---
    /*! function:    Log_ActionTruncateBlob ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   BlobFilename        [in]
                     BlobVersion         [in] 
                     NewTruncatedBlobLen [in] the blob will be truncated to this length
    */
    Log_ActionTruncateBlob (
        const tgg00_Filename &BlobFilename,
        SAPDB_Int4            BlobVersion,
        SAPDB_UInt4           NewTruncatedBlobLen);

    //--------------------------------------------------------------Log_ActionTruncateBlob---
    /*! function:    Execute
    
        description: truncates the binary large object
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    ExecuteAction
    
        description: truncates the binary large object
      
        arguments:   TransContext       [in out]
                     pCommandBlobFileId [in out] in case of undo/redo the file id has to be omitted
    */
    virtual void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandBlobFileId = NULL) const;

    //--------------------------------------------------------------Log_ActionTruncateBlob---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }

    //--------------------------------------------------------------Log_ActionTruncateBlob---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionTruncateBlob---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }

};
/*! endclass: Log_ActionTruncateBlob */


//===================================================================================
/*! class:       Log_ActionUpdateBlob

    description: defines an after image action that updates a binary large object
*/
class Log_ActionUpdateBlob: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        SAPDB_UInt4    ImageLength;
        SAPDB_UInt4    BufLength;
        SAPDB_UInt4    InsertOffset;
        tgg00_Filename BlobFilename;
    } UpdateBlobEntryHeader;
    
    UpdateBlobEntryHeader   m_EntryHeader;
    const SAPDB_Byte       *m_pBuf;
    SAPDB_Byte             *m_pBufForRead;
    SAPDBMem_IRawAllocator *m_pBufAllocator;

public:
    
    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:    Log_ActionUpdateBlob ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionUpdateBlob ();

    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:    Log_ActionUpdateBlob ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   BlobFilename [in]
                     InsertOffset [in]
                     pBuf         [in]
                     BufLength    [in]
    */
    Log_ActionUpdateBlob (
        const tgg00_Filename &BlobFilename,
        SAPDB_UInt4           InsertOffset,
        const SAPDB_Byte*     pBuf,
        SAPDB_UInt4           BufLength);

    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:    ~Log_ActionUpdateBlob
    
        description: deallocation
    */
    ~Log_ActionUpdateBlob ();

    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:    Execute
    
        description: writes the buffer contents to the binary large object
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionExpandBlob---
    /*! function:    ExecuteAction
    
        description: writes the buffer contents to the binary large object
      
        arguments:   TransContext       [in out]
                     pCommandBlobFileId [in out] in case of redo the file id has to be omitted
    */
    virtual void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandBlobFileId = NULL) const;

    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return (m_EntryHeader.ImageLength);
    }

    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionUpdateBlob---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const;
    
    /*  --------------------------------------------------------------------------------
    function:    WriteToTrace
    arguments:   Title [in]
    description: The action is written to the trace file
    ----------------------------------------------------------------------------------*/
    virtual void WriteToTrace (const char * Title) const
    {
        if ( Title != NULL )
            Kernel_VTrace() << Title;
        Kernel_VTrace() << "ImageLength: "    << m_EntryHeader.ImageLength
                        << ", BufLength: "    << m_EntryHeader.BufLength
                        << ", InsertOffset: " << m_EntryHeader.InsertOffset;
        // Kernel_VTrace() << tgg00_Filename BlobFilename;
    }
};
/*! endclass: Log_ActionUpdateBlob */


/*==========================================================================*
*  DEFINITION OF INLINE METHODS                                             *
*==========================================================================*/

inline Log_ActionCopyBlob::Log_ActionCopyBlob ():
Log_ReadWriteActionImage (Log_CopyBlob)
{
}

//---------------------------------------------------------------------------

inline Log_ActionCopySameBlob::Log_ActionCopySameBlob ():
Log_ReadWriteActionImage (Log_CopySameBlob)
{
}

//---------------------------------------------------------------------------

inline Log_ActionExpandBlob::Log_ActionExpandBlob ():
Log_ReadWriteActionImage (Log_ExpandBlob)
{
}

//---------------------------------------------------------------------------

inline Log_ActionTruncateBlob::Log_ActionTruncateBlob ():
Log_ReadWriteActionImage (Log_TruncateBlob)
{
}

//---------------------------------------------------------------------------

inline Log_ActionUpdateBlob::Log_ActionUpdateBlob ():
Log_ReadWriteActionImage (Log_UpdateBlob)
{
}

#endif  /* LOG_ACTION_COPY_UPD_BLOB_H */
