
/*!**************************************************************************

    module      : Log_ActionCreateDropFile.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to handle log entries for files

    last changed: 2001-03-28

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_CREATE_DROP_FILE_H
#define LOG_ACTION_CREATE_DROP_FILE_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_InvDescMap.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"

#include "ggg91.h"
#include "ggg00.h"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


//===================================================================================
/*! class:       Log_ActionChangeFileType

    description: defines a class to handle the before or after image of change file type
*/
class Log_ActionChangeFileType: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename    Filename;
        tgg00_FiletypeSet OldFiletypeSet;
        tgg00_FiletypeSet NewFiletypeSet;
    } ChangeFileTypeEntry;
    
    ChangeFileTypeEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:    Log_ActionChangeFileType ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionChangeFileType ();

    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:    Log_ActionChangeFileType ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   Filename       [in]
                     OldFiletypeSet [in]
                     NewFiletypeSet [in]
    */
    Log_ActionChangeFileType (
        const tgg00_Filename &Filename,
        tgg00_FiletypeSet     OldFiletypeSet,
        tgg00_FiletypeSet     NewFiletypeSet);

    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:    Execute
    
        description: undo/redo only: changes the file type set 
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:    ExecuteAction
    
        description: changes the file type set
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of undo/redo the file id has to be omitted 
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandFileId = NULL) const;
    
    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }

    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionChangeFileType---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionChangeFileType */


//===================================================================================
/*! class:       Log_ActionCreateFile

    description: defines a class to handle the after image of create file
*/
class Log_ActionCreateFile: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename    Filename;
        tgg91_FileVersion FileVersion;
        tgg00_FiletypeSet FiletypeSet;
    } CreateFileEntry;
    
    CreateFileEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:    Log_ActionCreateFile ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionCreateFile ();

    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:    Log_ActionCreateFile ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   Filename    [in]
                     FileVersion [in]
                     FiletypeSet [in]
    */
    Log_ActionCreateFile (
        const tgg00_Filename &Filename,
        tgg91_FileVersion     FileVersion,
        tgg00_FiletypeSet     FiletypeSet);
    
    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:    Execute
    
        description: redo only: creates a new file
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:    ExecuteAction
    
        description: creates a new file
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of redo the file id has to be omitted
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandFileId = NULL) const;

    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }
    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionCreateFile---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionCreateFile */


//===================================================================================
/*! class:       Log_ActionCreateIndex

    description: defines a class to handle the after image of create index
*/
class Log_ActionCreateIndex: public Log_ReadWriteActionImage
{
    
private:
    
    typedef struct
    {
        tgg00_Filename InvFilename;
        SAPDB_UInt2    InvDescMapLen;
        tgg00_Tfn      PrimTfn;
        SAPDB_UInt1    PrimTfnNo;
    } CreateIndexEntry;
    
    CreateIndexEntry      m_EntryHeader;
    Log_InvDescMap        m_InvDescMapForRead;
    const Log_InvDescMap *m_pInvDescMap;
    
public:
    
    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    Log_ActionCreateIndex ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionCreateIndex ();

    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    Log_ActionCreateIndex ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   InvFilename [in]
                     pInvDescMap [in] 
                     PrimTfn     [in]
                     PrimTfnNo   [in]
    */
    Log_ActionCreateIndex (
        const tgg00_Filename &InvFilename,
        const Log_InvDescMap *pInvDescMap,
        tgg00_Tfn             PrimTfn,
        SAPDB_UInt1           PrimTfnNo);
    
    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    Execute
    
        description: redo only: creates an empty index file with the state 'not accessible'
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteSetNotAccessible (TransContext);
    }
    
    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    ExecuteAction
    
        description: creates a new index, in case of redo the pointers have to be ommitted 
      
        arguments:   TransContext         [in out]
                     pCommandPermFileId   [in out]
                     pCommandInvFileId    [in out]
                     pCommandIndexCounter [in out] the structure is initialized by this function
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandPermFileId   = NULL,
        tgg00_FileId       *pCommandInvFileId    = NULL,
        tgg00_IndexCounter *pCommandIndexCounter = NULL) const;
    
    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    ExecuteSetNotAccessible
    
        description: creates an empty index file with the state 'not accessible'
      
        arguments:   TransContext [in out]
    */
    void ExecuteSetNotAccessible (tgg00_TransContext &TransContext) const;

    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    inline virtual SAPDB_UInt GetPersistentLength () const;

    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    RebuildIndex
    
        description: redo only: creates a new index
      
        arguments:   TransContext [in out]
    */
    void RebuildIndex (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }
    
    //--------------------------------------------------------------Log_ActionCreateIndex---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const;
    
};
/*! endclass: Log_ActionCreateIndex */


//===================================================================================
/*! class:       Log_ActionCreateTable

    description: defines a class to handle the after image of create table
*/
class Log_ActionCreateTable: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename    Filename;
        tgg91_FileVersion FileVersion;
        tgg00_FiletypeSet FiletypeSet;
        SAPDB_Bool1       NoTreeIndex;
        SAPDB_Int2        StrColCnt;
        SAPDB_Int2        VarColCntOff;
    } CreateTableEntry;
    
    CreateTableEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:    Log_ActionCreateTable ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionCreateTable ();

    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:    Log_ActionCreateTable ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   Filename     [in]
                     FileVersion  [in]
                     FiletypeSet  [in]
                     NoTreeIndex  [in] 
                     StrColCnt    [in]
                     VarColCntOff [in]
    */
    Log_ActionCreateTable (
        const tgg00_Filename &Filename,
        tgg91_FileVersion     FileVersion,
        tgg00_FiletypeSet     FiletypeSet,
        SAPDB_Bool1           NoTreeIndex,
        SAPDB_Int2            StrColCnt,
        SAPDB_Int2            VarColCntOff);
    
    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:    Execute
    
        description: redo only: creates a new table
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:    ExecuteAction
    
        description: creates a new table
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of redo the file id has to be omitted
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandFileId = NULL) const;
    
    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }
    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionCreateTable---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionCreateTable */


//===================================================================================
/*! class:       Log_ActionDropFile

    description: defines a class to handle the before image of any create except create object
*/
class Log_ActionDropFile: public Log_ReadWriteActionImage
{

private:
    
    tgg00_Filename m_Filename;

public:
    
    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:    Log_ActionDropFile ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionDropFile ();

    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:    Log_ActionDropFile ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   Filename [in]
    */
    inline Log_ActionDropFile (const tgg00_Filename &Filename);
        
    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:    Execute
    
        description: undo only: drops the file
      
        arguments:   TransContext [in out] 
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }
        
    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:    ExecuteAction
    
        description: drops the file
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of undo the file id has to be omitted
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandFileId = NULL) const;
    
    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Filename)) );
    }

    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionDropFile---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Filename, sizeof (m_Filename), IsOk);
    }
    
};
/*! endclass: Log_ActionDropFile */


//===================================================================================
/*! class:       Log_ActionRenameFile

    description: defines a class to handle the before or after image of rename file
*/
class Log_ActionRenameFile: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Filename    OldFilename;
        tgg00_Filename    NewFilename;
        tgg91_FileVersion NewFileVersion;

    } RenameFileEntry;
    
    RenameFileEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:    Log_ActionRenameFile ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionRenameFile ();

    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:    Log_ActionRenameFile ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   OldFilename    [in] 
                     NewFilename    [in]
                     NewFileVersion [in]
    */
    Log_ActionRenameFile (
        const tgg00_Filename &OldFilename,
        const tgg00_Filename &NewFilename,           
        tgg91_FileVersion     NewFileVersion);
    
    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:    Execute
    
        description: undo/redo only: renames the file
      
        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }
    
    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:    ExecuteAction
    
        description: renames the file
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of undo/redo the file id has to be omitted
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pOldCommandFileId = NULL,
        tgg00_FileId       *pNewCommandFileId = NULL) const;
    
    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }
    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionRenameFile---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionRenameFile */


//===================================================================================
/*! class:       Log_ActionSuccFileVersion

    description: defines a class to handle the before or after image of succ file version
*/
class Log_ActionSuccFileVersion: public Log_ReadWriteActionImage
{
    
private:
    
    typedef struct
    {
        tgg00_Filename    Filename;
        SAPDB_Int4        NewLongCnt;
        tgg91_FileVersion NewFileVersion;
    } SuccFileVersionEntry;
    
    SuccFileVersionEntry m_Entry;
    
public:
    
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:    Log_ActionSuccFileVersion ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionSuccFileVersion ();
    
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:    Log_ActionSuccFileVersion ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
      
        arguments:   Filename       [in]
                     NewFileVersion [in]
                     NewLongCnt     [in]
    */
    Log_ActionSuccFileVersion (
        const tgg00_Filename &Filename,
        tgg91_FileVersion     NewFileVersion,
        SAPDB_Int4            NewLongCnt);
    
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:    Execute
    
        description: undo/redo only: changes the file version and the number of long columns
      
        arguments:   TransContext [in out] 
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }
    
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:    ExecuteAction
    
        description: changes the file version and the number of long columns
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of undo/redo the file id has to be omitted
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandFileId = NULL) const;
    
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionSuccFileVersion---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionSuccFileVersion */


//===================================================================================
/*! class:       Log_ActionUpdFileStatistic

    description: defines a class to handle the before or after image of update file statistic
*/
class Log_ActionUpdFileStatistic: public Log_ReadWriteActionImage
{
    
private:
    
    typedef struct
    {
        tgg00_Filename Filename;
        SAPDB_Int4     LeafNodes;
    } UpdFileStatEntry;
    
    UpdFileStatEntry m_Entry;
    
public:
    
    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:    Log_ActionUpdFileStatistic ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionUpdFileStatistic ();

    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:    Log_ActionUpdFileStatistic ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   Filename  [in] 
                     LeafNodes [in]
    */
    Log_ActionUpdFileStatistic (
        const tgg00_Filename &Filename,
        SAPDB_Int4            LeafNodes);
    
    
    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:    Execute
    
        description: undo/redo only: changes the number of tree leaf nodes
      
        arguments:   TransContext   [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }
    
    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:    ExecuteAction
    
        description: changes the number of tree leaf nodes
      
        arguments:   TransContext   [in out]
                     pCommandFileId [in out] in case of undo/redo the file id has to be omitted
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       *pCommandFileId = NULL) const;
    
    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }
    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionUpdFileStatistic---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionUpdFileStatistic */


/*==========================================================================*
*  DEFINITION OF INLINE METHODS                                             *
*==========================================================================*/

inline Log_ActionChangeFileType::Log_ActionChangeFileType ():
Log_ReadWriteActionImage (Log_ChangeFileType)
{
}

//---------------------------------------------------------------------------

inline Log_ActionCreateFile::Log_ActionCreateFile ():
Log_ReadWriteActionImage (Log_CreateFile)
{
}

//---------------------------------------------------------------------------

inline Log_ActionCreateIndex::Log_ActionCreateIndex ():
Log_ReadWriteActionImage (Log_CreateIndex),
m_InvDescMapForRead      (),
m_pInvDescMap            (NULL)
{
}

//---------------------------------------------------------------------------

inline SAPDB_UInt Log_ActionCreateIndex::GetPersistentLength () const
{
    return ( Log_AlignedImageLen (sizeof(this->m_EntryHeader))
        +    Log_AlignedImageLen (this->m_pInvDescMap->GetPersistentLength()) );
}

//---------------------------------------------------------------------------

inline Log_ActionCreateTable::Log_ActionCreateTable ():
Log_ReadWriteActionImage (Log_CreateTable)
{
}

//---------------------------------------------------------------------------

inline Log_ActionDropFile::Log_ActionDropFile ():
Log_ReadWriteActionImage (Log_CreateFile)
{
}

//---------------------------------------------------------------------------

inline Log_ActionDropFile::Log_ActionDropFile (const tgg00_Filename &Filename):
Log_ReadWriteActionImage (Log_CreateFile)
{
    this->m_Filename = Filename;
}

//---------------------------------------------------------------------------

inline Log_ActionRenameFile::Log_ActionRenameFile ():
Log_ReadWriteActionImage (Log_RenameFile)
{
}

//---------------------------------------------------------------------------

inline Log_ActionSuccFileVersion::Log_ActionSuccFileVersion ():
Log_ReadWriteActionImage (Log_SuccFileVersion)
{
}

//---------------------------------------------------------------------------

inline Log_ActionUpdFileStatistic::Log_ActionUpdFileStatistic ():
Log_ReadWriteActionImage (Log_UpdFileStatistic)
{
}

#endif  /* LOG_ACTION_CREATE_DROP_FILE_H */
