/*!**************************************************************************

  module      : Log_ActionDeleteAllObjects.hpp
  special area: Logging
  responsible : UweH
  last changed: 2000-09-25  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines class Log_ActionDeleteAllObjects



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_ActionDeleteAllObjects_H
#define Log_ActionDeleteAllObjects_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"  // tgg00_TransContext, tgg00_FileId
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Container/Container_Vector.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_IAction.hpp"
#include "liveCache/LVC_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

class Log_AfterImage;
class Log_BeforeImage;
class Data_SplitSpaceWriter;
class Data_SplitSpaceReader;
 
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_ActionDeleteAllObjects
----------------------------------------------------------------------------------*/

class Log_ActionDeleteAllObjects : public Log_IAction
{

public:

    /*! --------------------------------------------------------------------------------
    function:    SetFileno
    arguments:   fileno [in]
    ----------------------------------------------------------------------------------*/
    void SetFileno (tgg00_ObjFileNo fileno)
    {
        m_Head.fileno = fileno;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetFileno
    returns:     fileno
    ----------------------------------------------------------------------------------*/
    tgg00_ObjFileNo GetFileno () const
    {
        return m_Head.fileno;
    }
    /*! --------------------------------------------------------------------------------
    function:    AddDeletedObject
    arguments:   oid     [in]
                 version [in]
    description: The given values for a deleted object are saved to be written to the log later.
    ----------------------------------------------------------------------------------*/
    bool AddDeletedObject (const tgg92_KernelOid &oid,
                           const tgg91_PageRef   &version)
    {
        return m_Objects.InsertEnd(ObjectDescription(oid,version));
    }
    /*! --------------------------------------------------------------------------------
    function:    AddedObjectsExists
    returns:     bool
    description: if objects are added to this action, true is returned.
    ----------------------------------------------------------------------------------*/
    bool AddedObjectsExists()
    {
        return ! m_Objects.IsEmpty();
    }
    /*! --------------------------------------------------------------------------------
    function:    Reset
    description: All stored objects are removed.
                 All resources are released.
    ----------------------------------------------------------------------------------*/
    void Reset()
    {
        m_Objects.Delete();
        m_Head.fileno     = 0;
        m_Head.numObjects = 0;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetMinimalLengthToCopyEntry
    ----------------------------------------------------------------------------------*/
    static SAPDB_UInt GetMinimalLengthToCopyEntry()
    {
        return sizeof (Head);
    }
    /*! --------------------------------------------------------------------------------
    function:    GetMinimalPersistentLength
    ----------------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetMinimalPersistentLength() const
    {
        return sizeof (Head);
    }
    /*! --------------------------------------------------------------------------------
    function:    GetPersistentLength
    ----------------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetPersistentLength() const;
    /*! --------------------------------------------------------------------------------
    function:    WritePersistentFormat
    arguments:   Writer    [in]
                 isOK      [out]
    description: The Space must be initialized with the value given by GetPersistentLength().
    ----------------------------------------------------------------------------------*/
    virtual void WritePersistentFormat (Data_SplitSpaceWriter &Writer,
                                        bool                  &isOK) const;
    /*! --------------------------------------------------------------------------------
    function:    ReadPersistentFormat
    arguments:   Reader    [in]
                 Allocator [in/out]
                 isOK      [out]
    description: all internal components are initialized with the reader
    ----------------------------------------------------------------------------------*/
    virtual void ReadPersistentFormat (Data_SplitSpaceReader  &Reader,
                                       SAPDBMem_IRawAllocator &Allocator,
                                       bool                   &isOK);
    /*  --------------------------------------------------------------------------------
    function:    WriteToTrace
    arguments:   Title [in]
    ----------------------------------------------------------------------------------*/    
    virtual void WriteToTrace (const char * Title) const;
    /*! --------------------------------------------------------------------------------
    function:    Redo
    arguments:   TransContext [in/out]
    description: The action is re-done.
    ----------------------------------------------------------------------------------*/
    void Redo (tgg00_TransContext   &Trans,
               const Log_AfterImage &AfterImage);
    /*! --------------------------------------------------------------------------------
    function:    Undo
    arguments:   TransContext [in/out]
    description: The action is rollbacked.
    ----------------------------------------------------------------------------------*/
    void Undo (tgg00_TransContext    &Trans,
               const Log_BeforeImage &BeforeImage)
    {
        Trans.trError_gg00 = e_not_implemented;
    }
    
private:

    /*! --------------------------------------------------------------------------------
    declaration:    Head
    ----------------------------------------------------------------------------------*/
    struct Head
    {
        tgg00_ObjFileNo fileno;     // 4
        SAPDB_UInt4     numObjects; // 4
        Head()
        {
            fileno     = 0;
            numObjects = 0;
        }
    };
    /*! --------------------------------------------------------------------------------
    declaration:    ObjectDescription
    ----------------------------------------------------------------------------------*/
    struct ObjectDescription
    {
        tgg92_KernelOid oid;     // 8
        tgg91_PageRef   version; // 6
        SAPDB_UInt2     filler;  // 2
        ObjectDescription()
        {
            oid.gg92SetNil();
            version.gg91SetNilRef();
            filler = 0;
        }
        ObjectDescription(const tgg92_KernelOid &newoid,
                          const tgg91_PageRef   &newversion)
        {
            oid     = newoid;
            version = newversion;
            filler  = 0;
        }
    };
    /*! --------------------------------------------------------------------------------
    declaration:    ObjectDescriptions
    ----------------------------------------------------------------------------------*/
    typedef Container_Vector<ObjectDescription> ObjectDescriptions;
    /*! --------------------------------------------------------------------------------
    variable:    m_Head
    ----------------------------------------------------------------------------------*/
    Head m_Head;
    /*! --------------------------------------------------------------------------------
    variable:    m_Objects
    ----------------------------------------------------------------------------------*/
    ObjectDescriptions m_Objects;
};
/*! endclass: Log_ActionDeleteAllObjects */

#endif  /* Log_ActionDeleteAllObjects_H */
