
/*!**************************************************************************

    module      : Log_ActionInsDelRec.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to handle log entries for insert/delete record

    last changed: 2001-02-20

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_INS_DEL_REC_H
#define LOG_ACTION_INS_DEL_REC_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_InvDescMap.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"

#include "ggg00.h"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


//===================================================================================
/*! class:       Log_ActionInsDelRec

    description: defines a class to handle action entries for records or keys
*/
class Log_ActionInsDelRec: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        SAPDB_UInt4    ImageLen;
        SAPDB_UInt4    KeyOrRecLen;
        SAPDB_UInt4    InvDescLen;
        tgg00_Filename PrimFilename;
    } InsDelRecEntryHeader;
    
    InsDelRecEntryHeader    m_EntryHeader;

    const tgg00_Rec        *m_pRec;
    const Log_InvDescMap   *m_pInvDescMap;
    bool                    m_bKeyOnly;
    
    tgg00_Rec              *m_pRecForRead;
    Log_InvDescMap          m_InvDescMapForRead;

    SAPDBMem_IRawAllocator *m_pRecAllocator;
    
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*  function:    InitHeader
    */
    void InitHeader ();

public:
    
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    Log_ActionInsDelRec ctor
       
        description: creates a key or record action depending on before or after image
    
        arguments:   ActionType    [in]
                     IsBeforeImage [in]
    */
    Log_ActionInsDelRec (
        Log_ActionType ActionType,
        bool           IsBeforeImage);

    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    ~Log_ActionInsDelRec
    
        description: deallocation
    */
    ~Log_ActionInsDelRec ();
 
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    Execute
    
        description: deletes or inserts the record and its secondary keys

        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const;
    
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:     GetPersistentLength
    
        return value: he persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return (this->m_EntryHeader.ImageLen);
    }

    /// this length is used for redo copy
    static SAPDB_UInt GetMinimalLengthToCopyEntry()
    {
        return sizeof(InsDelRecEntryHeader);
    }

    /// The action is written to the trace file.
    virtual void WriteToTrace (const char * Title) const;
    
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    Init
    
        description: initializes a key action entry
       
        arguments:   PrimFilename [in] 
                     pKey         [in] pointer to the key
                     pInvDescMap  [in] NULL or pointer to the initialized inv description map
    */
    void Init (
        const tgg00_Filename &PrimFilename,
        const tgg00_Lkey     *pKey,
        const Log_InvDescMap *pInvDescMap);

    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    Init
    
        description: initializes a record action entry
       
        arguments:   PrimFilename [in] 
                     pRec         [in] pointer to the record
                     pInvDescMap  [in] NULL or pointer to the initialized inv description map
    */
    void Init (
        const tgg00_Filename &PrimFilename,
        const tgg00_Rec      *pRec,
        const Log_InvDescMap *pInvDescMap);
    
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    MaterializeRecord

        description: writes the record to the specified space

        arguments:   DestinSize [in] the allocated size of the destination object
                     pDestin    [in] pointer to the destination object
                     Error      [out]
    */
    void MaterializeRecord (
        SAPDB_UInt4       DestinSize,
        void             *pDestin,
        tgg00_BasisError &Error) const;

    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionInsDelRec---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const;
    
};
/*! endclass: Log_ActionInsDelRec */

#endif  /* LOG_ACTION_INS_DEL_REC_H */
