
/*!**************************************************************************

    module      : Log_ActionModifyTable.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to handle the after image of modify table

    last changed: 2001-03-30

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_MODIFY_TABLE_H
#define LOG_ACTION_MODIFY_TABLE_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_InvDescMap.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"

#include "ggg00.h"


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


//===================================================================================
/*! class:       Log_ActionModifyTable

    description: defines a class to handle the after image of modify table
*/
class Log_ActionModifyTable: public Log_ReadWriteActionImage
{
    
private:
    
    typedef struct
    {
        SAPDB_UInt4    ImageLen;
        SAPDB_UInt4    InvDescMapLen;
        tgg00_Filename SourceFilename;
        tgg00_Filename DestinFilename;
    } ModifyTableEntryHeader;
    
    ModifyTableEntryHeader  m_EntryHeader;
    
    const Log_InvDescMap   *m_pSourceInvDescMap;
    const Log_InvDescMap   *m_pDestinInvDescMap;
    
    Log_InvDescMap          m_SourceInvDescMapForRead;
    Log_InvDescMap          m_DestinInvDescMapForRead;
    
public:
    
    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:    Log_ActionModifyTable ctor
    
        description: this constructor is designed for a further call of ReadImagePersistent
    */
    Log_ActionModifyTable ();

    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:    Log_ActionModifyTable ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   SourceFilename    [in] 
                     DestinFilename    [in]
                     pSourceInvDescMap [in]
                     pDestinInvDescMap [in]
    */
    Log_ActionModifyTable (
        const tgg00_Filename &SourceFilename,
        const tgg00_Filename &DestinFilename,
        const Log_InvDescMap *pSourceInvDescMap,
        const Log_InvDescMap *pDestinInvDescMap);

    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:    Execute
    
        description: modifies the table (redo only)

        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }

    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:    ExecuteAction
    
        description: modifies the table
      
        arguments:   TransContext         [in out]
                     pSourceCommandFileId [in out] in case of redo the source file id has to be omitted
                     pIndexCommandFileId  [in out] in case of redo the index file id has to be omitted
                     pDestinCommandFileId [in out] in case of redo the destination file id has to be omitted
                     IsSorted             [in]     in case of redo this paramter has to be ommitted
                     TruncCheckPos        [in]     in case of redo checking is not necessary
                     TruncCheckCnt        [in]     in case of redo checking is not necessary
                     ResultSampleInfo     [out]    in case of redo this paramter has to be ommitted
    */
    virtual void ExecuteAction (
        tgg00_TransContext  &TransContext,
        tgg00_FileId        *pSourceCommandFileId = NULL,
        tgg00_FileId        *pIndexCommandFileId  = NULL,
        tgg00_FileId        *pDestinCommandFileId = NULL,
        bool                 IsSorted             = false,
        SAPDB_Int            TruncCheckPos        = 0,
        SAPDB_Int            TruncCheckCnt        = 0,
        tgg00_SampleInfoPtr  ResultSampleInfo     = NULL) const;

    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:     GetPersistentLength
    
        return value: the persistent image length of all components of the modify table action
    */
    SAPDB_UInt GetPersistentLength () const
    {
        return (this->m_EntryHeader.ImageLen);
    }

    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:    ReadPersistentFormat

        description: reads the image of the modify table action from the specified space

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);

    //--------------------------------------------------------------Log_ActionModifyTable---
    /*! function:    WriteImagePersistent

        description: writes the image of the modify table action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const;

};
/*! endclass: Log_ActionModifyTable */

#endif  /* LOG_ACTION_MODIFY_TABLE_H */
