/*!**************************************************************************

  module      : Log_ActionObject.hpp
  special area: Logging
  responsible : UweH
  last changed: 2000-09-25  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines class Log_ActionObject



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_ActionObject_H
#define Log_ActionObject_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"  // tgg00_TransContext, tgg00_FileId
#include "ggg92.h"  // tgg92_KernelOid, tgg92_ObjHeader

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"

#include "DataAccess/Data_PageSplitSpace.hpp"
#include "DataAccess/Data_ISplitSpace.hpp"

#include "Logging/Log_Types.hpp"
#include "Logging/Log_IAction.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

class Log_AfterImage;
class Log_BeforeImage;

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_ActionObject
description: Handle for manipulating OMS objects.
*/

class Log_ActionObject : public Log_IAction
{

public:

    /*! --------------------------------------------------------------------------------
    function:    Log_ActionObject
    ----------------------------------------------------------------------------------*/
    Log_ActionObject ()
    : m_pKey       (NULL),
      m_pBody      (NULL),
      m_pAllocator (NULL),
      m_Head()
    {}
    
    /*! --------------------------------------------------------------------------------
    function:    ~Log_ActionObject
    ----------------------------------------------------------------------------------*/
    ~Log_ActionObject ()
    {
        if ( NULL != m_pAllocator )
            destroy (m_pBody,*m_pAllocator);
    }
    /*! --------------------------------------------------------------------------------
    function:    InitNew
    arguments:   ObjFileId [in]
                 KeySize   [in]
                 pKey      [in]
    description: Initialize the new action
    ----------------------------------------------------------------------------------*/
    void InitNew (const tgg00_FileId    &fileid,
                  const tgg92_KernelOid &oid,
                  const tgg91_PageRef   &version,
                  SAPDB_Int4             keySize,
                  SAPDB_Byte            *pKey)
    {
        m_Type           = Log_NewObject;
        m_Head.fileno    = fileid.fileObjFileNo_gg00();
        m_Head.oid()     = oid;
        m_Head.keySize   = keySize;
        //memcpy (m_Head.version, &version, sizeof(m_Head.version));
        m_Head.version() = version;
        m_pKey           = pKey;
    }
    /*! --------------------------------------------------------------------------------
    function:    InitDelete
    arguments:   FileNo         [in]
                 Oid            [in]
                 WantedObjVers  [in]
    description: This is used to write.
    ----------------------------------------------------------------------------------*/
    void InitDelete (tgg00_ObjFileNo        fileno,
                     const tgg92_KernelOid &oid,
                     const tgg91_PageRef   &version)
    {
        m_Type           = Log_DeleteObject;
        m_Head.fileno    = fileno;
        m_Head.oid()     = oid;
        //memcpy (m_Head.version, &version, sizeof(m_Head.version));
        m_Head.version() = version;
    }
    /*! --------------------------------------------------------------------------------
    function:    InitInsUpd
    arguments:   ObjFileId      [in]
                 Oid            [in]
                 WantedObjVers  [in]
    description: This is used to write.
    ----------------------------------------------------------------------------------*/
    void InitInsUpd (Log_ActionType         actiontype,
                     const tgg00_FileId    &fileid,
                     const tgg92_KernelOid &oid,
                     const tgg91_PageRef   &version,
                     SAPDB_Int4             keySize,
                     SAPDB_Int4             bodySize,
                     SAPDB_Byte            *pBody)
    {
        m_Type           = actiontype;
        m_Head.fileno    = fileid.fileObjFileNo_gg00();
        m_Head.oid()     = oid;
        //memcpy (m_Head.version, &version, sizeof(m_Head.version));
        m_Head.version() = version;
        m_Head.bodySize  = bodySize - keySize;
        m_pBody          = pBody;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetMinimalLengthToCopyEntry
    ----------------------------------------------------------------------------------*/
    static SAPDB_UInt GetMinimalLengthToCopyEntry()
    {
        return sizeof (Head);
    }
    /*! --------------------------------------------------------------------------------
    function:    GetMinimalPersistentLength
    description: 
    ----------------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetMinimalPersistentLength() const
    {
        return GetMinimalLengthToCopyEntry();
    }
    /*! --------------------------------------------------------------------------------
    function:    GetPersistentLength
    description: 
    ----------------------------------------------------------------------------------*/
    virtual SAPDB_UInt GetPersistentLength() const
    {
        if ( m_Head.bodySize > 0 )
            return GetMinimalPersistentLength() + Data_PageSplitSpace::AlignRecordSize (m_Head.bodySize);
        else
            return GetMinimalPersistentLength() + Data_PageSplitSpace::AlignRecordSize (m_Head.keySize);
    }
    /*! --------------------------------------------------------------------------------
    function:    WritePersistentFormat
    arguments:   Writer    [in]
                 isOK      [out]
    description: The Space must be initialized with the value given by GetPersistentLength().
    ----------------------------------------------------------------------------------*/
    virtual void WritePersistentFormat (Data_SplitSpaceWriter &Writer,
                                        bool                  &isOK) const;
    /*! --------------------------------------------------------------------------------
    function:    ReadPersistentFormat
    arguments:   Reader    [in]
                 Allocator [in/out]
                 isOK      [out]
    description: all internal components are initialized with the reader
    ----------------------------------------------------------------------------------*/
    virtual void ReadPersistentFormat (Data_SplitSpaceReader  &Reader,
                                       SAPDBMem_IRawAllocator &Allocator,
                                       bool                   &isOK);
    /*  --------------------------------------------------------------------------------
    function:    WriteToTrace
    arguments:   Title [in]
    description: The action is written to the trace file
    ----------------------------------------------------------------------------------*/
    virtual void WriteToTrace (const char * Title) const;
    /*! --------------------------------------------------------------------------------
    function:    Redo
    arguments:   TransContext [in/out]
    description: The action is re-done.
    ----------------------------------------------------------------------------------*/
    void Redo (tgg00_TransContext   &Trans,
               const Log_AfterImage &AfterImage);
    /*! --------------------------------------------------------------------------------
    function:    Undo
    arguments:   TransContext [in/out]
                 BeforeImage  [in]
                 pageno       [in] own position
                 offset       [in] own position
    description: The action is rollbacked.
    ----------------------------------------------------------------------------------*/
    void Undo (tgg00_TransContext    &Trans,
               const Log_BeforeImage &BeforeImage,
               Data_PageNo            pageno,
               Data_PageOffset        offset);
    /*! --------------------------------------------------------------------------------
    function:    GetOid()
    ----------------------------------------------------------------------------------*/
    tgg92_KernelOid& GetOid()
    {
        return m_Head.oid();
    }
    tgg92_KernelOid GetOid() const
    {
        return m_Head.const_oid();
    }
    /*! --------------------------------------------------------------------------------
    function:    GetFileNo()
    ----------------------------------------------------------------------------------*/
    tgg00_ObjFileNo GetFileNo() const
    {
        return m_Head.fileno;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetBodySize()
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt2 GetBodySize() const
    {
        return m_Head.bodySize;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetKeySize()
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt2 GetKeySize() const
    {
        return m_Head.keySize;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetVersion()
    ----------------------------------------------------------------------------------*/
    const tgg91_PageRef& GetVersion()
    {
        return m_Head.version();
    }
    /*! --------------------------------------------------------------------------------
    function:    GetKey()
    ----------------------------------------------------------------------------------*/
    const SAPDB_Byte* GetKey() const
    {
        return m_pKey;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetBody()
    ----------------------------------------------------------------------------------*/
    const SAPDB_Byte* GetBody() const
    {
        return m_pBody;
    }
    /*! --------------------------------------------------------------------------------
    function:    ExtractOid()
	arguments:   image       [in]
	             imagelength [in]
				 oid         [out]
	description: The image must be the pointer to a persistent format of this action.
    ----------------------------------------------------------------------------------*/
    static void ExtractOid ( const SAPDB_Byte  *image,
		                     SAPDB_UInt         imagelength,
							 tgg92_KernelOid   &oid )
    {
        SAPDBERR_ASSERT_ARGUMENT( imagelength >= sizeof(Head) );
		const Head *pHead = reinterpret_cast <const Head*> (image);
		oid = pHead->const_oid();
    }

public: // must be public for the struct definition (on HP)

    /*! --------------------------------------------------------------------------------
    declaration:    PersistentObjVersion
    ----------------------------------------------------------------------------------*/
    typedef SAPDB_Byte PersistentObjVersion [sizeof(tgg91_PageRef)];

    /*! --------------------------------------------------------------------------------
    declaration:    PersistentOid
    ----------------------------------------------------------------------------------*/
    typedef SAPDB_Byte PersistentOid [sizeof(tgg92_KernelOid)];

private:

    /*! --------------------------------------------------------------------------------
    function:    GetAlignmentLength
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt GetAlignmentLength() const
    {        
        SAPDB_UInt length = GetPersistentLength();
        
        if ( m_Head.bodySize > 0 )
            return length - GetMinimalPersistentLength() - m_Head.bodySize;
        else
            return length - GetMinimalPersistentLength() - m_Head.keySize;
    }

private:

    /*! --------------------------------------------------------------------------------
    declaration:    Head
    ----------------------------------------------------------------------------------*/
    struct Head
    {
        PersistentOid        poid;     //  8 bytes
        tgg00_ObjFileNo      fileno;   //  4 bytes
        SAPDB_UInt2          bodySize; //  2 bytes
        SAPDB_UInt2          keySize;  //  2 bytes
        PersistentObjVersion pversion; //  6 bytes
        SAPDB_UInt2          filler;   //  2 bytes
        Head()
        {
            oid().gg92SetNil();
            fileno    = 0;
            bodySize  = 0;
            keySize   = 0;
            version().gg91SetNilRef();
            filler    = 0;
        }
        tgg92_KernelOid& oid()
        {
            return *(reinterpret_cast<tgg92_KernelOid*>(poid));
        }
        tgg92_KernelOid const_oid() const
        {
            return *(reinterpret_cast<const tgg92_KernelOid*>(poid));
        }
        tgg91_PageRef& version()
        {
            return *(reinterpret_cast<tgg91_PageRef*>(pversion));
        }
    };                                // 28 bytes
    /*! --------------------------------------------------------------------------------
    variable:    m_Head
    ----------------------------------------------------------------------------------*/
    Head m_Head;
    /*! --------------------------------------------------------------------------------
    variable:    m_pKey
    ----------------------------------------------------------------------------------*/
    SAPDB_Byte *m_pKey;
    /*! --------------------------------------------------------------------------------
    variable:    m_pBody
    ----------------------------------------------------------------------------------*/
    SAPDB_Byte *m_pBody;
    /*! --------------------------------------------------------------------------------
    variable:    m_pAllocator
    ----------------------------------------------------------------------------------*/
    SAPDBMem_IRawAllocator* m_pAllocator;
};
/*! endclass: Log_ActionObject */

#endif  /* Log_ActionObject_H */
