/*!**************************************************************************

  module      : Log_ActionObjectFile.hpp
  special area: Logging
  responsible : UweH
  last changed: 2000-09-25  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines class Log_ActionObjectFile



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_ActionObjectFile_H
#define Log_ActionObjectFile_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"  // tgg00_TransContext, tgg00_FileId
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_IAction.hpp"
#include "liveCache/LVC_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

class Log_AfterImage;
class Log_BeforeImage;
class Data_SplitSpaceWriter;
class Data_SplitSpaceReader;
 
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_ActionObjectFile
description: Handle for manipulating OMS objects.
*/

class Log_ActionObjectFile : public Log_IAction
{

public:

    /*! --------------------------------------------------------------------------------
    function:    InitCreation
    */
    void InitCreation (tgg00_ObjFileNo            fileno,
                       bool                       forVarObjects,
                       SAPDB_UInt                 bodyLength,
                       SAPDB_UInt                 keyLength,
                       bool                       createSeveralKeyFiles,
                       const LVC_RootPageNoArray &rootPages)
    {
        m_Head.fileno                = fileno;
        m_Head.forVarObjects         = forVarObjects;
        m_Head.createSeveralKeyFiles = createSeveralKeyFiles;
        m_Head.bodyLength            = bodyLength;
        m_Head.keyLength             = keyLength;
        memcpy (m_Head.rootPages, rootPages.GetPersistentFormat(), rootPages.GetPersistentLength());
    }

    /*! --------------------------------------------------------------------------------
    function:    InitDeletion
    */
    void InitDeletion (tgg00_ObjFileNo fileno)
    {
        m_Head.fileno = fileno;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetMinimalLengthToCopyEntry
    */
    static SAPDB_UInt GetMinimalLengthToCopyEntry()
    {
        return sizeof (Head);
    }
    
    /*! --------------------------------------------------------------------------------
    function:    GetMinimalPersistentLength
    description: 
    */
    virtual SAPDB_UInt GetMinimalPersistentLength() const
    {
        return sizeof (Head);
    }
    
    /*! --------------------------------------------------------------------------------
    function:    GetPersistentLength
    description: 
    */
    virtual SAPDB_UInt GetPersistentLength() const
    {
        return sizeof (Head);
    }
    
    /*! --------------------------------------------------------------------------------
    function:    WritePersistentFormat
    arguments:   Writer    [in]
                 isOK      [out]
    description: The Space must be initialized with the value given by GetPersistentLength().
    */
    virtual void WritePersistentFormat (Data_SplitSpaceWriter &Writer,
                                        bool                  &isOK) const;

    /*! --------------------------------------------------------------------------------
    function:    ReadPersistentFormat
    arguments:   Reader    [in]
                 Allocator [in/out]
                 isOK      [out]
    description: all internal components are initialized with the reader
    */
    virtual void ReadPersistentFormat (Data_SplitSpaceReader  &Reader,
                                       SAPDBMem_IRawAllocator &Allocator,
                                       bool                   &isOK);

    /*  --------------------------------------------------------------------------------
    function:    WriteToTrace
    arguments:   Title [in]
    */    
    virtual void WriteToTrace (const char * Title) const;

    /*! --------------------------------------------------------------------------------
    function:    Redo
    arguments:   TransContext [in/out]
    description: The action is re-done.
    */
    void Redo (tgg00_TransContext   &Trans,
               const Log_AfterImage &AfterImage);
    
    /*! --------------------------------------------------------------------------------
    function:    Undo
    arguments:   TransContext [in/out]
    description: The action is rollbacked.
    */
    void Undo (tgg00_TransContext    &Trans,
               const Log_BeforeImage &BeforeImage)
    {
        Trans.trError_gg00 = e_not_implemented;
    }
    
private:

    /*! --------------------------------------------------------------------------------
    declaration:    Header
    */
    struct Head
    {
        tgg00_ObjFileNo fileno;                //  4 bytes
        SAPDB_Bool1     forVarObjects;         //  1 bytes
        SAPDB_Bool1     createSeveralKeyFiles; //  1 bytes
        SAPDB_UInt2     bodyLength;            //  2 bytes
        SAPDB_UInt2     keyLength;             //  2 bytes
        SAPDB_Byte      rootPages[LVC_ROOT_PAGENO_ARRAY_SIZE];
        
        Head()
        {
            fileno                = 0;
            forVarObjects         = false;
            createSeveralKeyFiles = false;
            bodyLength            = 0;
            keyLength             = 0;
        }
    };
    /*! enddeclaration: */

    /*! --------------------------------------------------------------------------------
    variable:    m_Head
    */
    Head m_Head;
};
/*! endclass: Log_ActionObjectFile */

#endif  /* Log_ActionObjectFile_H */
