
/*!**************************************************************************

    module      : Log_ActionRefreshSurrogSeq.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to handle log entries for surrogates and sequences

    last changed: 2001-04-19

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_REFRESH_SURROG_SEQ_H
#define LOG_ACTION_REFRESH_SURROG_SEQ_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"

#include "gsp00.h"
#include "ggg00.h"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


//===================================================================================
/*! class:       Log_ActionRefreshSurrogate

    description: defines a class to handle the after image of refresh surrogate
*/
class Log_ActionRefreshSurrogate: public Log_ReadWriteActionImage
{

private:
    
    typedef struct
    {
        tgg00_Surrogate Surrogate;
        tgg00_Surrogate SysKey;
    } RefreshSurrogateEntry;
    
    RefreshSurrogateEntry m_Entry;

public:
    
    //--------------------------------------------------------------Log_ActionRefreshSurrogate---
    /*! function:    Log_ActionRefreshSurrogate ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionRefreshSurrogate ();

    //--------------------------------------------------------------Log_ActionRefreshSurrogate---
    /*! function:    Log_ActionRefreshSurrogate ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   Surrogate [in]
                     SysKey    [in]
    */
    Log_ActionRefreshSurrogate (
        const tgg00_Surrogate &Surrogate,
        const tgg00_Surrogate &SysKey);

    //--------------------------------------------------------------Log_ActionRefreshSurrogate---
    /*! function:    Execute
    
        description: redo only: refreshes the last surrogate and the last system key 
      
        arguments:   TransContext [in out]
    */
    void Execute (tgg00_TransContext &TransContext) const;
    
    /*  --------------------------------------------------------------------------------
    function:    ForceLogWriting
    returns:     true
    description: Because the savepoint depends on writing this entry.
                 The writer must not be suspended.
    */                         
	// PTS 1117320 UH 2002-10-28 added
    virtual bool ForceLogWriting () const
    {
        return true;
    }

    //--------------------------------------------------------------Log_ActionRefreshSurrogate---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_Entry)) );
    }

    //--------------------------------------------------------------Log_ActionRefreshSurrogate---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionRefreshSurrogate---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const
    {
        ImageWriter.PersistentWrite (&m_Entry, sizeof (m_Entry), IsOk);
    }
    
};
/*! endclass: Log_ActionRefreshSurrogate */


//===================================================================================
/*! class:       Log_ActionRefreshSequence

    description: defines a class to handle the after image of refresh sequence
*/
class Log_ActionRefreshSequence: public Log_ReadWriteActionImage
{
public:
    
    typedef struct
    {
        tgg00_Surrogate Surrogate;
        tsp00_Number    Number;
    } SequenceItem;
    
    
private:
    
    SAPDB_Int4              m_SequenceCount;
    const SequenceItem     *m_pSequenceList;
    
    SequenceItem           *m_pSequenceListForRead;
    SAPDBMem_IRawAllocator *m_pSequenceListAllocator;
    
public:
    
    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:    Log_ActionRefreshSequence ctor

        description: this constructor is designed for a further call of ReadImagePersistent
    */
    inline Log_ActionRefreshSequence ();

    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:    Log_ActionRefreshSequence ctor
    
        description: this constructor is designed for a further call of WriteImagePersistent
       
        arguments:   SequenceCount [in]
                     pSequenceList [in]
    */
    inline Log_ActionRefreshSequence (
        SAPDB_Int4  SequenceCount,
        const void *pSequenceList);

    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:    ~Log_ActionRefreshSequence
    
        description: deallocation if necessary
    */
    ~Log_ActionRefreshSequence ();

    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:    Execute
    
        description: redo only: refreshes the sequences
      
        arguments:   TransContext [in out]
    */
    void Execute (tgg00_TransContext &TransContext) const;
    

    /*  --------------------------------------------------------------------------------
    function:    ForceLogWriting
    returns:     true
    description: Because the savepoint depends on writing this entry.
                 The writer must not be suspended.
    */                         
	// PTS 1117320 UH 2002-10-28 added
    virtual bool ForceLogWriting () const
    {
        return true;
    }

    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of the image of the action
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return ( Log_AlignedImageLen (sizeof(m_SequenceCount))
            +    Log_AlignedImageLen (m_SequenceCount * sizeof(SequenceItem)) );
    }

    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);
    
    //--------------------------------------------------------------Log_ActionRefreshSequence---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const;
};
/*! endclass: Log_ActionRefreshSequence */


/*==========================================================================*
*  DEFINITION OF INLINE METHODS                                             *
*==========================================================================*/

inline Log_ActionRefreshSurrogate::Log_ActionRefreshSurrogate ():

Log_ReadWriteActionImage (Log_RefreshSurrogate)
{
}

//---------------------------------------------------------------------------

inline Log_ActionRefreshSequence::Log_ActionRefreshSequence ():

Log_ReadWriteActionImage (Log_RefreshSequence),
m_SequenceCount          (0),
m_pSequenceList          (NULL),
m_pSequenceListForRead   (NULL),
m_pSequenceListAllocator (NULL)
{
}

//---------------------------------------------------------------------------

inline Log_ActionRefreshSequence::Log_ActionRefreshSequence (SAPDB_Int4  SequenceCount,
                                                             const void *pSequenceList):
Log_ReadWriteActionImage (Log_RefreshSequence),
m_SequenceCount          (SequenceCount),
m_pSequenceList          (REINTERPRET_CAST (const SequenceItem*, pSequenceList)),
m_pSequenceListForRead   (NULL),
m_pSequenceListAllocator (NULL)
{
}

#endif  /* LOG_ACTION_REFRESH_SURROG_SEQ_H */
