/*!**************************************************************************

    module      : Log_ActionReorganizeFile.hpp

    -------------------------------------------------------------------------

    responsible : UweH

    special area: Logging
    description : defines a class to handle log entries for files

    last changed: 2001-03-28

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_REORGANIZE_FILE_H
#define LOG_ACTION_REORGANIZE_FILE_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"

#include "ggg00.h"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

//===================================================================================
/*! class:       Log_ActionCreateFile

    description: defines a class to handle the after image of create file
*/
class Log_ActionReorganizeFile: public Log_ReadWriteActionImage
{

private:
    
    struct Head
    {
        tgg00_Filename filename;       // 16
        SAPDB_UInt2    lastKeyLength;  // 2
        SAPDB_UInt2    filler1;        // 2
        SAPDB_UInt4    filler2;        // 4
        Head()
        {
            memset (filename, 0, sizeof(filename));
            filler1       = 0;
            filler2       = 0;
            lastKeyLength = 0;
        }
        Head( const tgg00_Filename &newfilename,
              SAPDB_UInt2           newlastKeyLength )
        {
            filename      = newfilename;
            lastKeyLength = newlastKeyLength;
            filler1       = 0;
            filler2       = 0;
        }
    };
    
    Head                    m_head;
    SAPDB_Byte             *m_pLastKey;
    SAPDBMem_IRawAllocator *m_pAllocator;

public:
    
    /*!-------------------------------------------------------------------------------------
    function:    Log_ActionReorganizeFile
    description: This is used for reading persistent images.
    */
    Log_ActionReorganizeFile()
    : Log_ReadWriteActionImage (Log_ReorganizeFile),
      m_head(),
      m_pLastKey(0),
      m_pAllocator(0)
    {}
    /*!-------------------------------------------------------------------------------------
    function:    Log_ActionReorganizeFile
    description: This is used for begin table load
    */
    Log_ActionReorganizeFile( const tgg00_Filename &filename,
                         SAPDB_UInt2           lastKeyLength,
                         SAPDB_Byte           *pLastKey )
    : Log_ReadWriteActionImage (Log_ReorganizeFile),
      m_head    (filename, lastKeyLength),
      m_pLastKey(pLastKey),
      m_pAllocator(0)
    {}
    /*!-------------------------------------------------------------------------------------
    function:    ~Log_ActionReorganizeFile
    description: deallocates allocated memory
    */
    ~Log_ActionReorganizeFile()
    {
        if ( m_pAllocator != 0 )
            // m_pLastKey cannot be 0
            m_pAllocator->Deallocate(m_pLastKey);
    }
    /*!-------------------------------------------------------------------------------------
    function:    ExecuteAction
    arguments:   TransContext [in/out]
                 FileId
    description: called command
    */
    void ExecuteAction (tgg00_TransContext &TransContext,
                        tgg00_FileId       *pCommandFileId = NULL) const;
    /*!-------------------------------------------------------------------------------------
    function:    Execute
    arguments:   TransContext [in/out]
    description: called by undo/redo
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {
        ExecuteAction (TransContext);
    }
    /*!-------------------------------------------------------------------------------------
    function:    GetPersistentLength
    */
    virtual SAPDB_UInt GetPersistentLength () const
    {
        return sizeof(m_head) + Log_AlignedImageLen(m_head.lastKeyLength);
    }
    /*!-------------------------------------------------------------------------------------
    function:    ReadImagePersistent
    */
    void ReadImagePersistent (Log_ActionImageReader &reader,
                              bool                  &isOk)
    {
        reader.CopyImage( reinterpret_cast<SAPDB_Byte*>(&m_head),
                          sizeof(m_head),
                          isOk );

        if ( ! isOk )
            return;
        
        if ( m_head.lastKeyLength > 0 )
            reader.PersistentRead (m_pLastKey,
                                   m_head.lastKeyLength,
                                   m_pAllocator,
                                   isOk);
    }
    /*!-------------------------------------------------------------------------------------
    function:    WriteImagePersistent
    */
    void WriteImagePersistent (Log_ActionImageWriter &writer,
                               bool                  &isOk) const
    {
        writer.PersistentWrite (&m_head, sizeof (m_head), isOk);
        if ( m_head.lastKeyLength > 0 )
            writer.PersistentWrite (m_pLastKey,
                                    m_head.lastKeyLength,
                                    isOk);
    }
};
/*! endclass: Log_ActionReorganizeFile */

/*==========================================================================*
*  DEFINITION OF INLINE METHODS                                             *
*==========================================================================*/

#endif  /* LOG_ACTION_REORGANIZE_FILE_H */
