
/*!**************************************************************************

    module      : Log_ActionUpdateRec.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to handle log entries for insert/delete record

    last changed: 2001-04-12

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_ACTION_UPDATE_REC_H
#define LOG_ACTION_UPDATE_REC_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_InvDescMap.hpp"
#include "Logging/Log_ReadWriteActionImage.hpp"
#include "Logging/Log_RecordColumnMap.hpp"

#include "ggg00.h"


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


//===================================================================================
/*! class:       Log_ActionUpdateRec

    description: defines a class to handle action entries for records or keys
*/
class Log_ActionUpdateRec: public Log_RecordColumnMap
{

private:
    
    struct UpdateRecEntryHeader
    {
        SAPDB_UInt4    ImageLen;
        SAPDB_UInt2    KeyLen;
        SAPDB_UInt2    NewRecLen;
        SAPDB_UInt4    ColMapLen;
        SAPDB_UInt4    InvDescLen;
        tgg00_Filename PrimFilename;
        UpdateRecEntryHeader()
        {
            ImageLen  = 0;
        	KeyLen    = 0;
        	NewRecLen = 0;
        	ColMapLen = 0;
        	InvDescLen = 0;
        	memset (PrimFilename, 0, sizeof(PrimFilename));
        }
    };
    
    UpdateRecEntryHeader    m_EntryHeader;
    
    const tgg00_Lkey       *m_pKey;
    const Log_InvDescMap   *m_pInvDescMap;
    
    tgg00_Lkey             *m_pKeyForRead;
    Log_InvDescMap          m_InvDescMapForRead;
    
    SAPDBMem_IRawAllocator *m_pKeyAllocator;

        
    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    Init
    
        description: initializes the update action entry
       
        arguments:   PrimFilename [in] 
                     pInvDescMap  [in] NULL or pointer to the initialized inv description map
                     pKey         [in] pointer to the key
                     pNewRecord   [in] pointer to the new record of the after image
                     NullValLen   [in] number of bytes for not existing NULL-value columns
    */
    void Init (
        const tgg00_Filename &PrimFilename,
        const Log_InvDescMap *pInvDescMap,
        const tgg00_Lkey     *pKey,
        const tgg00_Rec      *pNewRecord,
        const SAPDB_UInt2     NullValLen);

public:
    
    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    Log_ActionUpdateRec ctor
    
        description: empty action entry 
    */
    Log_ActionUpdateRec ();

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    ~Log_ActionUpdateRec
    
        description: deallocates the private field and inv description map
    */
    ~Log_ActionUpdateRec ();

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    CreateNewRecAndInitAfterBeforeImage

        description: creates the new record using the old record and the column descriptions
         
                     This function applies to the after image and creates a corresponding before image.
                     In case of a key update, the new record is created but the before image is incomplete.

        arguments:   PrimFilename [in] 
                     pInvDescMap  [in]     NULL or pointer to the initialized inv description map
                     pOldRecord   [in]     pointer to the existing record
                     pNewRecord   [out]    pointer to the record space reserved to build the new record
                     NewRecSize   [in]     size of the record space
                     BeforeImage  [in out] corresponding before image
                     Error        [out]
    */
    void CreateNewRecAndInitAfterBeforeImage (
        const tgg00_Filename   &PrimFilename,
        const Log_InvDescMap   *pInvDescMap,
        const tgg00_Rec        *pOldRecord,
        tgg00_Rec              *pNewRecord,
        SAPDB_Int4              NewRecSize,
        Log_ActionUpdateRec    &BeforeImage,
        tgg00_BasisError       &Error);

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    Execute
    
        description: undo/redo only: updates the record with its secondary keys

        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const;

    /// The action is written to the trace file.
    virtual void WriteToTrace (const char * Title) const;

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    ExecuteAction
    
        description: updates the record with its secondary keys

        arguments:   TransContext [in out]
                     PrimFileId   [in out]
                     pOldRecord   [in]
                     pNewRecord   [in]
                     IsUndoOrRedo [in]
    */
    void ExecuteAction (
        tgg00_TransContext &TransContext,
        tgg00_FileId       &PrimFileId,
        const tgg00_Rec    *pOldRecord,
        const tgg00_Rec    *pNewRecord,
        bool                IsUndoOrRedo = false) const;

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:     GetPersistentLength
    
        return value: the persistent length of all components of the action entry
    */
    SAPDB_UInt GetPersistentLength () const
    {
        return (this->m_EntryHeader.ImageLen);
    }

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    InitAfterBeforeImageWithExistingNewRec
    
        description: initializes this after and before image using an existing new record
        
                     This function applies to the after image and creates a corresponding before image.
        
        arguments:   RawAllocator [ref]
                     PrimFilename [in] 
                     pInvDescMap  [in]     NULL or pointer to the initialized inv description map
                     pOldRecord   [in]     pointer to the existing record
                     pNewRecord   [out]    pointer to the record space reserved to build the new record
                     BeforeImage  [in out] corresponding before image
                     Error        [out]
    */
    void InitAfterBeforeImageWithExistingNewRec (
        SAPDBMem_IRawAllocator &RawAllocator,
        const tgg00_Filename   &PrimFilename,
        const Log_InvDescMap   *pInvDescMap,
        const tgg00_Rec        *pOldRecord,
        const tgg00_Rec        *pNewRecord,
        Log_ActionUpdateRec    &BeforeImage,
        tgg00_BasisError       &Error);
    
    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    ReadPersistentFormat

        description: reads the record entry from the specified space

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk);

    
    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    RemoveRedundantInvDescEntries

        description: removes after image inv desc entries, which are not contained in the record column map

                     A new inv desc map is allocated without the redundant entries. The pointer
                     to the inv desc map of the before image is replaced by this new allocated inv desc map.  

        arguments:   RawAllocator              [ref]
                     pOwnWriteAccessInvDescMap [in out] same pointer as m_pInvDescMap but without const
                     IsOk                      [out]
        
        return value: the persistent length of index entries
    */
    SAPDB_UInt4 RemoveRedundantInvDescEntries (
        SAPDBMem_IRawAllocator &RawAllocator,
        Log_InvDescMap         *pOwnWriteAccessInvDescMap,
        bool                   &IsOk);
    
    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    SetInvDescLen

        description: sets InvDescLen to the given value
        
                     used to set it in BeforeImage after RemoveRedundantInvDescEntries

        arguments:   NewInvDescLen [in]
    */
    void SetInvDescLen (SAPDB_UInt4 NewInvDescLen)
    {
        m_EntryHeader.InvDescLen = NewInvDescLen;
    };

    //--------------------------------------------------------------Log_ActionUpdateRec---
    /*! function:    WriteImagePersistent

        description: writes the image of the update action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const;
};
/*! endclass: Log_ActionUpdateRec */

#endif  /* LOG_ACTION_UPDATE_REC_H */
