/*!**************************************************************************

  module      : Log_AfterImage.hpp
  special area: Logging
  responsible : UweH
  last changed: 2000-01-02
  copyright:    Copyright by SAP AG, 2000
  description : defines class Log_AfterImage, Log_AfterImageSpace



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_AfterImage_H
#define Log_AfterImage_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "Logging/Log_Types.hpp"
#include "DataAccess/Data_ISplitSpace.hpp"
#include "Logging/Log_AfterImageSpace.hpp"
#include "Logging/Log_IImage.hpp"

#include "ggg91.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define LOG_AFTER_IMAGE_IS_NOT_DONE SAPDB_FALSE
#define LOG_AFTER_IMAGE_IS_DONE     SAPDB_TRUE

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_AfterImage
description: This defines the common part of all after images.
             It is in a persistent format.
*/

class Log_AfterImage : public Log_IImage
{
public:

    /*! --------------------------------------------------------------------------------
    function:     Log_AfterImage
    descriptions: default constructor
    */
    Log_AfterImage (Log_ActionType ActionType = Log_NoOp)
    : m_pIsDone (NULL) // PTS 1115068 UH 2002-03-27
    {
        m_Head.ActionType = ActionType;
		// PTS 1115068 UH 2002-03-27 initialization is done in ctor of m_Head
    }

    /*! --------------------------------------------------------------------------------
    function:     GetPersistentLength
    returns:      SAPDB_UInt
    descriptions: The persistent length is calculated and returned.
    */
    static SAPDB_UInt GetPersistentLength ()
    {
        return sizeof(Head);
    }

    /*  --------------------------------------------------------------------------------
    function:    IsAfterImage
    description: returns true, if image is an after image.
    ----------------------------------------------------------------------------------*/ 
    virtual bool IsAfterImage () const
    {
        return true;
    }

    /*! --------------------------------------------------------------------------------
    function:    WritePersistentFormat
    arguments:   Writer    [in]
                 isOK      [out]
    description: The Space must be initialized with the value given by GetPersistentLength().
    */
    virtual void WritePersistentFormat (Data_SplitSpaceWriter &Writer,
                                        bool                  &isOK) const
    {
        isOK = true;
        Data_SplitSpaceWriter::Result  result;
        Head                          *pHead;
        
        Writer.Reserve (sizeof(Head), REINTERPRET_CAST(SAPDB_Byte*&,pHead), result);
                        
        isOK = Data_SplitSpaceWriter::ok                 == result ||
               Data_SplitSpaceWriter::moreSpaceAvailable == result;
        if ( ! isOK ) return;

        *pHead = m_Head;
    }

    /*! --------------------------------------------------------------------------------
    function:    ReadPersistentFormat
    arguments:   Reader    [in/out]
                 isOK      [out]
    description: The Space must be initialized with the value given by GetPersistentLength().
    */
    virtual void ReadPersistentFormat (Data_SplitSpaceReader &Reader,
                                       bool                  &isOK) // PTS 1115068 UH 2002-03-27 & was missing
    {
        isOK = true;
        Data_SplitSpaceReader::Result  result;
        Head                          *pHead;

        m_Head.Init(); // PTS 1115068 UH 2002-03-27
        
        Reader.Reserve (sizeof(Head), REINTERPRET_CAST(SAPDB_Byte*&,pHead), result);
                        
        isOK = Data_SplitSpaceReader::ok                 == result ||
               Data_SplitSpaceReader::moreSpaceAvailable == result;
        if ( ! isOK ) return;

        m_Head    = *pHead;
        m_pIsDone = &(pHead->IsDone); // PTS 1115068 UH 2002-03-27

    }

    /*! --------------------------------------------------------------------------------
    function:    SetActionType
    */
    void SetActionType (Log_ActionType ActionType)
    {
        m_Head.ActionType = STATIC_CAST(SAPDB_Byte, ActionType);
    }

    /*! --------------------------------------------------------------------------------
    function:    GetActionType
    */
    virtual Log_ActionType GetActionType() const
    {
        if ( m_Head.ActionType >= Log_ActionType_DO_NOT_USE )
            return Log_NoOp;
        else
            return STATIC_CAST(Log_ActionType, m_Head.ActionType);
    }

    /*! --------------------------------------------------------------------------------
    function:    SetTransNo
    */
    void SetTransNo (tgg91_TransNo& TransNo)
    {
        m_Head.TransNo = TransNo;
    }
    
    /*! --------------------------------------------------------------------------------
    function:    GetTransNo
    */
    const tgg91_TransNo& GetTransNo() const
    {
        return m_Head.TransNo;
    }

    /*! --------------------------------------------------------------------------------
    function:    SetLastUndoSequence
    */
    void SetLastUndoSequence (Log_EntrySequence LastUndoSequence)
    {
        m_Head.LastUndoSequence = LastUndoSequence;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetLastRedoSequence
    */
    const Log_EntrySequence GetLastUndoSequence() const
    {
        return m_Head.LastUndoSequence;
    }

    /*! --------------------------------------------------------------------------------
    function:    SetSequence
    */
    void SetSequence (Log_EntrySequence Sequence)
    {
        m_Head.Sequence = Sequence;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetSequence
    */
    virtual Log_EntrySequence GetSequence() const
    {
        return m_Head.Sequence;
    }

    /*! --------------------------------------------------------------------------------
    function:    SetIsDone
    description: Within the entry it is stored, that it is re done.
    */
    void SetIsDone ()
    {
		// PTS 1115068 UH 2002-03-27
        SAPDB_Byte *pIsDone = m_pIsDone == NULL ? &m_Head.IsDone : m_pIsDone;
        *pIsDone = LOG_AFTER_IMAGE_IS_DONE;
    }

    /*! --------------------------------------------------------------------------------
    function:    IsDone
    returns:     true, if entry is already re done.
    */
    bool IsDone () const
    {
		// PTS 1115068 UH 2002-03-27
        const SAPDB_Byte *pIsDone = m_pIsDone == NULL ? &m_Head.IsDone : m_pIsDone;
        return LOG_AFTER_IMAGE_IS_DONE == *pIsDone;
    }

    /*! --------------------------------------------------------------------------------
    function:    WriteToTrace
    */
    virtual void WriteToTrace (const char *title = NULL) const
    {
        if ( title != NULL )
            Kernel_VTrace() << title;
        Kernel_VTrace() << "ActionType: " << Log_ActionTypeStrings[GetActionType()]
                        << ", IsDone: "   << (IsDone()?"yes":"no")
                        << ", Sequence: " << GetSequence()
                        << ", Transno: "  << GetTransNo().gg90GetInt4();
    }

private:

    /*! --------------------------------------------------------------------------------
    declaration: Head
    description: This is the structured header.

    ActionType    - The type of action, which is logged.
    Sequence      - This entry sequence no is local to the transaction.
    TransNo       - The id of the transaction.
    SubTransNo    - The subtransaction layer.
    */
    struct Head
    {
        SAPDB_Byte        ActionType;          // Log_ActionType 1 byte
        SAPDB_Bool1       IsDone;              //                1 byte
        SAPDB_UInt2       Filler1;             //                2 byte
        Log_EntrySequence Sequence;            //                4 bytes
        Log_EntrySequence LastUndoSequence;    //                4 bytes
        tgg91_TransNo     TransNo;             //                6 bytes
        SAPDB_UInt2       Filler2;             //                2 byte

        void Init()
        {
			// PTS 1115068 UH 2002-03-27 new
            ActionType        = Log_NoOp;
            IsDone            = LOG_AFTER_IMAGE_IS_NOT_DONE;
            Sequence          = 0;
            LastUndoSequence  = 0;
            TransNo.gg90SetNil();
            Filler1           = 0;
            Filler2           = 0;
        }

        Head()
        {
			// PTS 1115068 UH 2002-03-27 new
            Init();
        }

    };

    /*! --------------------------------------------------------------------------------
    variable:    m_Head
    */
    Head m_Head;
    /*! --------------------------------------------------------------------------------
    variable:    m_pIsDone
    */
    SAPDB_Byte *m_pIsDone;
};
/*! endclass: Log_AfterImage */

/*===========================================================================*
*  DEFINITION OF INLINE METHODS                                              *
*============================================================================*/
                                       
#endif  /* Log_AfterImage_H */
