/*!*****************************************************************************

  module:       Log_AfterImageSpace.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  author:       TillL

  special area: Logging

  description:  Log after image space handler class

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_AFTERIMAGESPACE_HPP
#define LOG_AFTERIMAGESPACE_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "Container/Container_Vector.hpp"
#include "DataAccess/Data_ISplitSpace.hpp"
#include "Logging/Log_QueueEntry.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define LOG_AFTERIMAGESPACE_PARTS_DELTA 3
    

/*!*****************************************************************************

   class: Log_AfterImageSpace

   description:
   
   Log after image space handler class

   An instance of this class holds a vector of log queue entries.
   These point to log pages on which space has been reserved for writing
   out the after image. 

*******************************************************************************/
class Log_AfterImageSpace : public Data_ISplitSpace
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : PartCount
    description:  any number of parts
    --------------------------------------------------------------------------*/
    typedef SAPDB_Int                                           PartCount;
    /*!-------------------------------------------------------------------------
    declaration : PartNo
    description:  index number of a specific part of an after image space
    --------------------------------------------------------------------------*/
    typedef PartCount                                           PartNo;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  allocator for queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDBMem_IRawAllocator                              Allocator;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Log_AfterImageSpace()
    description:  Constructor
    arguments:    alloc        [in] raw allocator to be used by after image
    return value: none
    --------------------------------------------------------------------------*/
    Log_AfterImageSpace 
        (Allocator&                     alloc)
        : 
         m_Parts(alloc)
    {}
    
    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  Initializes after image structures

                  - Performs memory allocation
    arguments:    size   [in]   part count (default: LOG_AFTERIMAGESPACE_PARTS_DELTA)
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool Initialize 
        (PartCount                      size = LOG_AFTERIMAGESPACE_PARTS_DELTA)
    {
        return m_Parts.Reserve(size);
    }
    
    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes after image structures

                  - Performs memory deallocation
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Delete()
    {
        m_Parts.Delete();
    }
    
    /*!-------------------------------------------------------------------------
    function:     ~Log_AfterImageSpace()
    description:  Destructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    ~Log_AfterImageSpace()
    {
        Delete();
    }
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Data_ISplitSpace interface                                   */
    /*!-------------------------------------------------------------------------
    function:     CurrentPart()
    description:  Returns the current part index
    arguments:    none
    return:       (SAPDB_Int) current part index
    --------------------------------------------------------------------------*/
    virtual SAPDB_Int CurrentPart() const
    {
        return (SAPDB_Int)m_Parts.GetSize() - 1;
    }

    /*!-------------------------------------------------------------------------
    function:     GetPart()
    description:  Returns length and pointer of specified part
    arguments:    part    [in]  requested part number
                  length  [out] size of part space in bytes
                  value   [out] pointer to part space
    return:       none
    --------------------------------------------------------------------------*/
    virtual void GetPart 
        (PartNo                         part,
         SAPDB_UInt&                    length,
         SAPDB_Byte*&                   value) const
    {
        value  = m_Parts[part].m_Value;
		length = m_Parts[part].m_Length;
    }
    
    /*!-------------------------------------------------------------------------
    function:     GetPartEntry()
    description:  Returns reference to queue entry associated with specified part
    arguments:    part    [in]  requested part number
    return:       (Log_QueueEntry&) reference to queue entry
    --------------------------------------------------------------------------*/
    Log_QueueEntry& GetPartEntry
        (PartNo                         part) const
    {
        return *(m_Parts[part].m_QueueEntry);
    }
    
    /*!-------------------------------------------------------------------------
    function:     GetPart()
    description:  Returns length and pointer of specified part
    arguments:    part    [in]  requested part number
                  length  [out] size of part space in bytes
                  value   [out] pointer to part space
    return:       none
    --------------------------------------------------------------------------*/
    virtual PartCount GetPartCount() const
    {
        return m_Parts.GetSize();
    }
    
    /*!-------------------------------------------------------------------------
    function:     Length()
    description:  Calculates and returns the sum length of all used space parts
    arguments:    none
    return:       (SAPDB_UInt) length in bytes
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt Length() const
    {
        SAPDB_UInt Result = 0;
        for (SAPDB_UInt i=0; i<m_Parts.GetSize(); ++i)
        {
            Result += m_Parts[i].m_Length;
        }
        return Result;
    }

    /*!-------------------------------------------------------------------------
    function:     IsAssigned
    returns:     SAPDB_UInt
    --------------------------------------------------------------------------*/
    virtual bool IsAssigned () const
    {
        return GetPartCount() > 0;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Data_ISplitSpace interface                                */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Space management                                             */
    /*!-------------------------------------------------------------------------
    function:     AppendPart()
    description:  A part is appended to the part vector
    arguments:    entry   [in]  reference to log queue entry
                  length  [in]  size of reserved space in bytes
                  value   [in]  pointer to reserved space
    return:       none
    --------------------------------------------------------------------------*/
    void AppendPart 
        (Log_QueueEntry&                entry,
         SAPDB_UInt                     length,
         SAPDB_Byte*                    value)
    {
        if (m_Parts.IsFull())
            if ( ! m_Parts.Reserve(m_Parts.GetSize() + LOG_AFTERIMAGESPACE_PARTS_DELTA) )
                // PTS 1117126 UH 2002-08-07 if-clause added
                RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,
                                              "m_Parts.Reserve() failed") );

        m_Parts.InsertEnd(Part(&entry, length, value));
    }

    /*!-------------------------------------------------------------------------
    function:     Deassign()
    description:  Releases all parts

                  - All log queue entries associated with this split space 
                    are released.
                  - The part vector is cleared.
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/
    void Deassign()
    {
        for (SAPDB_UInt i=0; i<m_Parts.GetSize(); ++i)
        {
            Log_QueueEntry& queueentry = *m_Parts[i].m_QueueEntry;
            queueentry.DecBusyTaskCount();
         }
        if ( ! m_Parts.Resize(0) )
            // PTS 1117126 UH 2002-08-07 if-clause added
            RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,
                       "m_Parts.Resize(0) failed") );

    }
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Space management                                          */
    /* -----------------------------------------------------------------------*/
    
private:

    /*--------------------------------------------------------------------------
    declaration : Part
    description:  after image part
    --------------------------------------------------------------------------*/
    struct Part
    {
        Log_QueueEntry*     m_QueueEntry;
        SAPDB_UInt          m_Length;
        SAPDB_Byte*         m_Value;

        Part() 
             :
             m_QueueEntry(0),
             m_Length(0),
             m_Value(0)
        {}

        Part(Log_QueueEntry*     queueentry,
             SAPDB_UInt          length,
             SAPDB_Byte*         value)
             :
             m_QueueEntry(queueentry),
             m_Length(length),
             m_Value(value)
        {}
    };

    /*--------------------------------------------------------------------------
    declaration : m_Parts
    description:  vector of after image parts
    --------------------------------------------------------------------------*/
    Container_Vector<Part>                      m_Parts;

};
/*!*****************************************************************************

   endclass: Log_AfterImageSpace

*******************************************************************************/


#endif // LOG_AFTERIMAGESPACE_HPP
