/*!*****************************************************************************

  module:       Log_ArchiveLog.cpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  special area: Archive log

  description:  see .hpp

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/

#include "Logging/Log_ArchiveLog.hpp"

#include "heo47.h"   // sql47_ptoc
#include "hkb38.h"   // k38history_lost_write
#include "hkb28.h"   // k28emergency_shutdown
#include "hkb57.h"   // k57save_restartrecord
#include "hkb57_1.h" // k57restartrec
#include "hgg01.h"   // g01io_block_count
#include "hgg01_3.h" // g01restart_time

#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "KernelCommon/Kernel_DateTime.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "Logging/Log_SeqTapeReader.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "IOManager/IOMan_ILogManager.hpp"
#include "IOManager/IOMan_ILogInfo.hpp"

/*!*****************************************************************************

   class: Log_ArchiveLog

*******************************************************************************/
Log_ErrorState Log_ArchiveLog::CreateForNewDatabase
    (TaskID                         taskid,
     RawDeviceOffset                rawlogbegin,
     RawDeviceOffset                rawlogend,
     DeviceBlockCount               savedist,
     SAPDB_UInt                     queuecount,
     QueueSize                      queuesize,
     SAPDB_UInt                     ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::CreateForNewDatabase", LogVolume_Trace, 5);

    if (m_Allocated)
    // this method is only allowed if nothing had been initialized before
    // (i.e. this method may be called only once
    {
        tgg00_TransContext  trans;
        trans.trError_gg00      = e_ok;
        trans.trTaskId_gg00     = m_WriterTaskSync.GetTaskID();
        trans.trAllocator_gg00  = NULL;
        trans.trLogContext_gg00 = NULL;
    	trans.trTaskId_gg00 = taskid;
    	k28emergency_shutdown(trans, m_init_config);
    }    	
    		
    SAPDBERR_ASSERT_ARGUMENT(rawlogbegin < rawlogend);

    m_RawDeviceBegin = rawlogbegin;
    m_RawDeviceEnd   = rawlogend;

    // -------------------------------------------------
    // allocate frame for info page and for buffered page:
    // -------------------------------------------------
    if (!AllocatePageFrames())
    {
        Delete();
        return Log_PageAllocationFailed;
    }

    // -------------------------------------------------
    // allocate i/o page handler:
    // -------------------------------------------------
    if (!m_IOPages.Initialize(ioblockcount))
    {
        Delete();
        return Log_RawAllocationFailed;
    }

    // -------------------------------------------------
    // create info log page
    // -------------------------------------------------

    m_LogInfoPage.EnterRegion(taskid);
 
    m_LogInfoPage.InitializeFrame
        ((Log_InfoPage::PageID)m_WriterTaskSync.GetTaskID());

    CreateInfoPage(savedist);

    // -------------------------------------------------
    // clear entire raw log device:
    // -------------------------------------------------
    m_LogInfoPage.WriteClearUntilLowerOffset ( LogicalLogBeginOffset() );
    m_LogInfoPage.WriteClearUntilUpperOffset ( LogicalLogEndOffset()   );
    m_LogInfoPage.WriteDeviceState           ( Log_DeviceStateClearing  );
    FlushLogInfoPage (taskid);

    FormatLog(taskid, LogicalLogBeginOffset(), LogicalLogEndOffset());

    Log_RawDeviceOffset invalidOffset;
    m_LogInfoPage.WriteClearUntilLowerOffset ( invalidOffset );
    m_LogInfoPage.WriteClearUntilUpperOffset ( invalidOffset );
    m_LogInfoPage.WriteDeviceState           ( Log_DeviceStateCleared  );

    // -------------------------------------------------
    // allocate and initialize log queues:
    // -------------------------------------------------
    if (!CreateLogQueues(queuecount, queuesize))
    {
        Delete();
        return Log_RawAllocationFailed;
    }
    m_LogInfoPage.WriteQueueCount ( m_LogQueues.GetSize() );
    
    FlushLogInfoPage (taskid);
 
    m_LogInfoPage.LeaveRegion(taskid);
    
    // -------------------------------------------------
    // mark dbm.knl so the recovery manager can work
    // -------------------------------------------------
    k38history_lost_write(taskid);
 
    // -------------------------------------------------
    // check if log device too small:
    // -------------------------------------------------
    Log_DeviceBlockCount reservedPages = m_LogWriter.CalculateReservedPages(queuecount*queuesize);
    if (GetDeviceSize() <= reservedPages)
    {
        Delete();
        return Log_LogDeviceTooSmall;
    }

    // -------------------------------------------------
    // check if save distance too large:
    // -------------------------------------------------
    if (GetDeviceSize() - reservedPages <= savedist)
    {
        Delete();
        return Log_LogSaveDistanceTooLarge;
    }

    // -------------------------------------------------
    // initialize log writer, ioseqno is invalid, write position is at first device block:
    // -------------------------------------------------
    if (! InitLogWriter( LogicalLogBeginOffset(), 
                         IOSequenceNo(), 
                         ioblockcount) )
    {
        Delete();
        return Log_RawAllocationFailed;
    }

    m_Allocated = true;

    // resume writer task after that archive log has been allocated
    m_WriterTaskSync.Resume();

    if ( LogVolume_Trace.TracesLevel(6) )
        WriteDeviceParametersToTrace ();

    return Log_Ok;
}
// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLog::RestartForRestoreLog
        (TaskID                         taskid,
         RawDeviceOffset                rawlogbegin,
         RawDeviceOffset                rawlogend,
         DeviceBlockCount               autosaveDistance,
         QueueSize                      queuesize,
         SAPDB_UInt                     ioblockcount)
{
    return RestartIntern (  taskid,
                            rawlogbegin,
                            rawlogend,
                            autosaveDistance,
                            queuesize,
                            ioblockcount,
                            Restart_for_restore_log );
}
// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLog::RestartIntern
    (TaskID                         taskid,
     RawDeviceOffset                rawlogbegin,
     RawDeviceOffset                rawlogend,
     DeviceBlockCount               autosaveDistance,
     QueueSize                      queuesize,
     SAPDB_UInt                     ioblockcount,
     RestartType                    restartType)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::RestartIntern", LogVolume_Trace, 5);

    Log_ErrorState  result        = Log_Ok;

    m_RawDeviceBegin = rawlogbegin;
    m_RawDeviceEnd   = rawlogend;
    m_autosaveDistance = autosaveDistance; 			// PTS 1113361 mb 2002-01-11
    if ( ! m_Allocated )
    {
        if ( ! AllocatePageFrames() )
        {
            Delete();
            return Log_PageAllocationFailed;
        }
        if (!m_IOPages.Initialize(ioblockcount))
        {
            Delete();
            return Log_RawAllocationFailed;
        }
    }

    m_LogInfoPage.EnterRegion(taskid);
    
    m_PageReader.ReadLogInfoPage ( taskid, m_LogInfoPage,
                                   Log_ClusterAddress(LogInfoPageOffset(),1) );

    bool addLogDevSpaceIsPending = rawlogend > m_LogInfoPage.ReadCycleRelatedEndOffset();

    if ( addLogDevSpaceIsPending )
        m_RawDeviceEnd = m_LogInfoPage.ReadCycleRelatedEndOffset();

    if ( Restart_for_clear_log != restartType )
    {
        IOSequenceNo    firstKnownSeqno;
        RawDeviceOffset firstKnownOffset;

        // for a normal restart/restore check if a CLEAR LOG is panding and finish it
        if (Log_DeviceStateClearing == m_LogInfoPage.ReadDeviceState()) //PTS 1111525 mb 2002-01-29 
        {
            FormatLog(taskid, LogicalLogBeginOffset(), LogicalLogEndOffset());
            m_LogInfoPage.WriteClearUntilLowerOffset ( Log_RawDeviceOffset());     // set to invalid
            m_LogInfoPage.WriteClearUntilUpperOffset ( Log_RawDeviceOffset());     // set to invalid
            
            tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();
            m_LogInfoPage.WriteLastKnownIOSequenceNo(SaveptParam.svpIOsequence_kb00);
            m_LogInfoPage.WriteLastKnownOffset(Log_RawDeviceOffset());
            m_LogInfoPage.WriteOldestNotSavedOffset(Log_RawDeviceOffset());
            m_LogInfoPage.WriteOldestNotSavedIOSequenceNo(Log_IOSequenceNo());
            m_LogInfoPage.WriteDeviceState ( Log_DeviceStateCleared );
            k38history_lost_write(taskid);
        }
        else if (Log_DeviceStatePartiallyClearing == m_LogInfoPage.ReadDeviceState()) //PTS 1111525 mb 2002-01-29 
        {
            FormatLog ( taskid, m_LogInfoPage.ReadClearUntilLowerOffset(),
                                m_LogInfoPage.ReadClearUntilUpperOffset());
            m_LogInfoPage.WriteClearUntilLowerOffset ( Log_RawDeviceOffset());     // set to invalid
            m_LogInfoPage.WriteClearUntilUpperOffset ( Log_RawDeviceOffset());     // set to invalid

            tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();
            m_LogInfoPage.WriteLastKnownIOSequenceNo(SaveptParam.svpIOsequence_kb00);
            m_LogInfoPage.WriteLastKnownOffset(SaveptParam.svpStartOffset_kb00);
            m_LogInfoPage.WriteDeviceState ( Log_DeviceStatePartiallyCleared );
            k38history_lost_write(taskid);
        }
        else if (Log_DeviceStateHistoryLost == m_LogInfoPage.ReadDeviceState()) // PTS 1117801  mb 2002-10-01 
        {
            // if HistoryLost no restore log is allowed, because no DB-Ident is set
            // and therefore no compatibility check can be done
            if (Restart_for_restore_log == restartType)
            {
                Delete();
                m_LogInfoPage.LeaveRegion(taskid);
                return Log_LogAndDataIncompatible;
            }
        }
    }

    if ( GetDeviceSize() <= m_LogWriter.CalculateReservedPages(m_LogInfoPage.ReadQueueCount()*queuesize) )
    {
        Delete();
        m_LogInfoPage.LeaveRegion(taskid);
        return Log_LogDeviceTooSmall;
    }

    if ( GetDeviceSize() <= autosaveDistance) 
    {
        Delete();
        m_LogInfoPage.LeaveRegion(taskid);
        return Log_LogSaveDistanceTooLarge;
    }

    if ( rawlogend < m_LogInfoPage.ReadCycleRelatedEndOffset() )
    {
        Delete();
        m_LogInfoPage.LeaveRegion(taskid);
        return Log_LogSizeHasDecreased;
    }
    
    if ( Restart_for_clear_log == restartType )
    {
        tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();

        RawDeviceOffset     SVPOffset;      // PTS 1119223 mb 2002-12-10 
        IOSequenceNo        SVPIOSeq = SaveptParam.svpIOsequence_kb00;

        m_LogInfoPage.WriteOldestNotSavedIOSequenceNo ( SVPIOSeq  );
        m_LogInfoPage.WriteOldestNotSavedOffset       ( SVPOffset );

        if ( SVPIOSeq.IsValid() )
            SVPIOSeq -= 1;
        m_LogInfoPage.WriteLastKnownIOSequenceNo      ( SVPIOSeq  );
        m_LogInfoPage.WriteLastKnownOffset            ( SVPOffset );
        m_LogInfoPage.ResetSavepointPart();
    }
    else // Restart_normal || Restart_for_restore_log
    {
        result = UpdateLogInfoPage ( taskid );
        if (result != Log_Ok)
        {
            Delete();
            m_LogInfoPage.LeaveRegion(taskid);
            return result;
        }
        if (Restart_for_restore_log == restartType)
        {
            // PTS 1119223 mb 2002-12-10 search log only for restore_log
            ScanLogForFirstKnown(taskid);
        }

        if (Restart_normal == restartType)
        {
            result = CheckLogAndDataCompatibility ( taskid );
            if (result != Log_Ok)
            {
                Delete();
                m_LogInfoPage.LeaveRegion(taskid);
                return result;
            }
        }

    }
    
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);
    
    if ( m_Allocated )
        ResetLogWriter (taskid);
    else
    {
        if ( ! CreateLogQueues(m_LogInfoPage.ReadQueueCount(), queuesize) )
        {
            Delete();
            return Log_RawAllocationFailed;
        }

        if ( ! InitLogWriter( m_LogInfoPage.ReadLastKnownOffset(),
                              m_LogInfoPage.ReadLastKnownIOSequenceNo(),
                              ioblockcount) )
        {
            Delete();
            return Log_RawAllocationFailed;
        }

        m_Allocated = true;

        // resume writer task after that archive log has been allocated
        m_WriterTaskSync.Resume();
    }

    if ( addLogDevSpaceIsPending )
    {
        m_LogWriter.SetDeviceEnd (rawlogend);
        RTE_Message(Log_Exception(__CONTEXT__, LOG_ADD_DEVSPACE_PENDING,
                                  SAPDB_ToString(m_LogInfoPage.ReadCycleRelatedEndOffset()),
                                  SAPDB_ToString(rawlogend) ));
    }
    
    if ( LogVolume_Trace.TracesLevel(6) )
        WriteDeviceParametersToTrace ();

    RTE_Message(Log_Exception(__CONTEXT__, LOG_DEVICE_INFO_BASIC,
                              SAPDB_ToString(m_LogInfoPage.ReadQueueCount(),_T_d),
                              Log_FlushModeStrings[m_LogInfoPage.ReadFlushMode()],
                              Log_DeviceStateStrings[m_LogInfoPage.ReadDeviceState()]));
    RTE_Message(Log_Exception(__CONTEXT__, LOG_DEVICE_INFO_OLDEST_NOT_SAVED,
                              SAPDB_ToString(m_LogInfoPage.ReadOldestNotSavedIOSequenceNo(),_T_d),
                              SAPDB_ToString(m_LogInfoPage.ReadOldestNotSavedOffset(),_T_d)));
    if (m_FirstKnownIOSeq.IsValid())
    {
        RTE_Message(Log_Exception(__CONTEXT__, LOG_DEVICE_INFO_FIRST_KNOWN,
                                  SAPDB_ToString(m_FirstKnownIOSeq),
                                  SAPDB_ToString(m_FirstKnownOffset)));
    }
    else
    {
        RTE_Message(Log_Exception(__CONTEXT__, LOG_DEVICE_INFO_UNCOMPLETE_CYCLE));
    }

    RTE_Message(Log_Exception(__CONTEXT__, LOG_RESTART_INFO,
                              SAPDB_ToString(k57restartrec->rstLastSavept_kb00().svpIOsequence_kb00,_T_d),
                              SAPDB_ToString(k57restartrec->rstLastSavept_kb00().svpStartOffset_kb00,_T_d),
                              SAPDB_ToString(m_LogInfoPage.ReadLastKnownIOSequenceNo(),_T_d),
                              SAPDB_ToString(m_LogInfoPage.ReadLastKnownOffset(),_T_d)));
    RTE_Message(Log_Exception(__CONTEXT__, LOG_CHECK_RESULT,
                              Log_ErrorStateStrings[result]));
	if ( result == Log_Ok )
		m_Restarted = true;

    return result;
}


// -----------------------------------------------------------------------------

Log_ErrorState Log_ArchiveLog::RestartForSaveLogCold 
    (TaskID                         taskid,
     RawDeviceOffset                rawlogbegin,
     RawDeviceOffset                rawlogend,
     DeviceBlockCount               autosaveDistance)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::RestartForSaveLogCold", LogVolume_Trace, 5);

    Log_ErrorState  result        = Log_Ok;
    IOSequenceNo    firstKnownSeqno;
    RawDeviceOffset firstKnownOffset;

    m_RawDeviceBegin = rawlogbegin;
    m_RawDeviceEnd   = rawlogend;
    m_autosaveDistance = autosaveDistance; 			// PTS 1113361 mb 2002-01-11

    if ( ! m_LogInfoPage.IsAssigned() )
    {
        if ( ! AllocatePageFrames() )
        {
            Delete();
            return Log_PageAllocationFailed;
        }
    }

    m_LogInfoPage.EnterRegion(taskid);
    
    m_PageReader.ReadLogInfoPage ( taskid, m_LogInfoPage,
                                   Log_ClusterAddress(LogInfoPageOffset(),1) );

    if (    m_LogInfoPage.ReadDeviceState() == Log_DeviceStateClearing
         || m_LogInfoPage.ReadDeviceState() == Log_DeviceStatePartiallyClearing ) //PTS 1111525 mb 2002-01-29
    {
		RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_SAVE_FOR_ABORTED_CLEAR_LOG));

        return Log_SaveForAbortedClearLog;
    }


    if ( GetDeviceSize() <= m_LogWriter.CalculateReservedPages(g01io_block_count()*m_LogInfoPage.ReadQueueCount()))
    {
        Delete();
        return Log_LogDeviceTooSmall;
    }

    if ( GetDeviceSize() <= autosaveDistance ) 
    {
        Delete();
        return Log_LogSaveDistanceTooLarge;
    }

    if ( rawlogend < m_LogInfoPage.ReadCycleRelatedEndOffset() )
    {
        Delete();
        return Log_LogSizeHasDecreased;
    }

    result = UpdateLogInfoPage ( taskid );
    if (result != Log_Ok)
    {
        Delete();
        return result;
    }

    ScanLogForFirstKnown( taskid );    
    
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);

    return result;
}


// -----------------------------------------------------------------------------

Log_ErrorState Log_ArchiveLog::CheckLogAndDataCompatibility    (  TaskID     taskid )
{
    if ( LogIsEmpty() ) //PTS 1113550 mb 2002-03-27
    {
            return Log_Ok;
    }

    // ----------------------------------------------------------------------
    // check if the DB_Identifier of Log and Data are compatible (PTS1113550 mb 2002-05-02)
    // ----------------------------------------------------------------------
    tsp00_Line &rstDBIdent = k57restartrec->rstDbIdent_kb00();

    if (m_LogInfoPage.ReadDBIdentifier()[0] != '\0' &&              //PTS 1113550 mb 2002-03-27
        rstDBIdent[0] != '\0')
    {
        if (memcmp(m_LogInfoPage.ReadDBIdentifier(), &rstDBIdent, sizeof(tsp00_Line)))
        {
            SAPDB_Char dataDBIdent[sizeof(tsp00_Line)+1];
            SAPDB_Char logDBIdent[sizeof(tsp00_Line)+1];
            
            sql47_ptoc(logDBIdent, (SAPDB_Char*)m_LogInfoPage.ReadDBIdentifier(), sizeof(tsp00_Line));
            sql47_ptoc(dataDBIdent, &rstDBIdent, sizeof(tsp00_Line));
            RTE_Message(Log_Exception(__FILE__,__LINE__,
                                      LOG_DBIDENT_MISMATCH,
                                      dataDBIdent,
                                      logDBIdent));
            return Log_LogAndDataIncompatible;
        }
    }
    else if (m_LogInfoPage.ReadDBIdentifier()[0] == '\0' &&
             rstDBIdent[0] != '\0')
    {
        m_LogInfoPage.WriteDBIdentifier(rstDBIdent);
    }

    // ----------------------------------------------------------------------
    // check if the restart record savepoint points within the log device.
    // ----------------------------------------------------------------------
    IOSequenceNo    lastKnownSeqNo = m_LogInfoPage.ReadLastKnownIOSequenceNo();
    tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();

    if ( (static_cast<IOSequenceNo>(SaveptParam.svpIOsequence_kb00)).IsValid() )
    {
        if ( lastKnownSeqNo.IsValid()
             &&
             SaveptParam.svpIOsequence_kb00 > lastKnownSeqNo )
        {
            RTE_Message(Log_Exception(__FILE__,__LINE__,
                                      LOG_SEQUENCE_NOT_IN_RANGE,
                                      SAPDB_ToString(SaveptParam.svpIOsequence_kb00,_T_d),
                                      SAPDB_ToString(lastKnownSeqNo,_T_d) ));
            RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_SAVEPOINT_SEQ_NOT_FOUND));
            return Log_LogAndDataIncompatible;
        }

        if ( ! CheckOffsetAndIOsequence ( taskid,
                                          SaveptParam.svpStartOffset_kb00,
                                          SaveptParam.svpIOsequence_kb00 ) )
        {
            RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_SAVEPOINT_NOT_FOUND));
            return Log_SavepointPositionNotFound;
        }
    }

    return Log_Ok;
}


// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLog::UpdateLogInfoPage (TaskID           taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::UpdateLogInfoPage", LogVolume_Trace, 5);

    if ( (Log_DeviceStateOkay        != m_LogInfoPage.ReadDeviceState() &&
          Log_DeviceStateHistoryLost != m_LogInfoPage.ReadDeviceState())
          ||  
          LogIsEmpty()  )
    {

		RTE_Message(Log_Exception(__FILE__,__LINE__, LOG_NO_INFOPAGE_UPD,
                                  Log_DeviceStateStrings[m_LogInfoPage.ReadDeviceState()],
                                  SAPDB_ToString(LogIsEmpty(),_T_b) ) );
        return Log_Ok;
    }

    // ----------------------------------------------------------------------
    // check consistency of info page and log device
    // ----------------------------------------------------------------------

    if ( ! CheckOffsetAndIOsequence ( taskid,
                                      m_LogInfoPage.ReadLastKnownOffset(),
                                      m_LogInfoPage.ReadLastKnownIOSequenceNo() ) )
    {
		RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_LASTKNOWN_NOT_FOUND));
        return Log_InconsistentLogInfoPage;
    }
    IOSequenceNo    lastKnownSeqNo   = m_BufferedPage.ReadLastWriterIOSequenceNo();
    RawDeviceOffset lastKnownOffset  = m_LogInfoPage.ReadLastKnownOffset();

    // ----------------------------------------------------------------------
    // find youngest log page and perhaps the oldest log page (cycle first)
    // ----------------------------------------------------------------------

    const SAPDB_UInt   maxloops           = m_LogInfoPage.ReadCycleRelatedEndOffset() -
                                            LogicalLogBeginOffset();
    SAPDB_UInt         loopcount          = 0;
    Log_IOSequenceNo   prevIOSequence     = lastKnownSeqNo;
    Log_IOSequenceNo   prevQueueSequence  = m_BufferedPage.ReadQueueIOSequenceNo();
    Log_ArchiveLogIterator::CheckLogDeviceResult checkresult;
    Log_ClusterAddress ioaddress (lastKnownOffset, 1);
    RawDeviceIterator  iter( LogicalLogBeginOffset(),
                             lastKnownOffset,
                             m_LogInfoPage.ReadCycleRelatedEndOffset() );

    do
    {
        ++iter; 

        ioaddress.SetOffset(iter.GetPosition());
        m_PageReader.ReadLogPage( taskid, m_BufferedPage, ioaddress );
        checkresult = Log_ArchiveLogIterator::CheckLogDeviceConsistency
                            ( Log_DeviceMinimizeSpace == m_QueueFlushMode ?
                              Log_CheckQueueSequences :
                              Log_CheckIOSequences,
                              m_BufferedPage,
                              ioaddress.GetOffset(),
                              prevIOSequence,
                              prevQueueSequence);

        if ( Log_ArchiveLogIterator::inconsistent == checkresult )
            return Log_InconsistentLogDeviceIOSequence;
            
        if ( Log_ArchiveLogIterator::endOfLogDevice == checkresult )
        {
            if ( m_BufferedPage.ReadLastWriterIOSequenceNo().IsValid() )
            {
                // PTS 1119223 mb 2002-12-10 save values direct in own members 
                m_FirstKnownIOSeq  = m_BufferedPage.ReadFirstWriterIOSequenceNo(); // PTS 1115226 mb 2002-04-15
                m_FirstKnownOffset = iter.GetPosition();
            }
        }
        else
        {
            prevIOSequence    = m_BufferedPage.ReadLastWriterIOSequenceNo();
            prevQueueSequence = m_BufferedPage.ReadQueueIOSequenceNo();
            lastKnownSeqNo    = prevIOSequence;
            lastKnownOffset   = iter.GetPosition();
            ++loopcount;
        }
    }
    while ( Log_ArchiveLogIterator::consistent == checkresult
            &&
            loopcount < maxloops );

    if ( loopcount >= maxloops )
    {
        RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_CYCLE_NOT_FOUND));
        return Log_InconsistentLogDeviceIOSequence;
    }

    // last known is found now
    m_LogInfoPage.WriteLastKnownIOSequenceNo(lastKnownSeqNo);
    m_LogInfoPage.WriteLastKnownOffset(lastKnownOffset);

    if ( LogVolume_Trace.TracesLevel(6) )
        WriteDeviceParametersToTrace ();

    // ----------------------------------------------------------------------
    // check if the log save history is consistent
    // ----------------------------------------------------------------------

    if ( m_LogInfoPage.ReadOldestNotSavedIOSequenceNo().IsValid()
         &&
         m_LogInfoPage.ReadOldestNotSavedOffset().IsValid()
         &&
         ! CheckOffsetAndIOsequence ( taskid,
                                      m_LogInfoPage.ReadOldestNotSavedOffset(),
                                      m_LogInfoPage.ReadOldestNotSavedIOSequenceNo() ) )
    {
        // additional check if a cold log save was done
        // then it is allowed, that oldestNotSaved is exactly the next offset after
        // last known
        iter.SetPosition (m_LogInfoPage.ReadLastKnownOffset());
        ++iter;
        if ( m_LogInfoPage.ReadOldestNotSavedOffset()
             !=
             iter.GetPosition() )
        {
            RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_OLDEST_NOT_SAVED_NOT_FOUND));
            return Log_InconsistentLogInfoPage;
        }
    }

    return Log_Ok;
}

// -----------------------------------------------------------------------------

void Log_ArchiveLog::ScanLogForFirstKnown(TaskID           taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ScanLogForFirstKnown", LogVolume_Trace, 5);
    // PTS 1119223 mb 2002-12-10 SetFirstKnownBeforeRestoreLog renamed and rewritten

    if ( ! (( Log_DeviceStateCleared  == m_LogInfoPage.ReadDeviceState()) ||  
            (LogIsEmpty()) ) )
    // an empty log need not to be scanned
    {
        if ( m_FirstKnownIOSeq.IsInvalid() )
        // FirstKnown can already be set by UpdateLogInfoPage, if it has 
        // been found by the way. 
        {
            // first, look at first offset of the device
            RawDeviceIterator  iter( LogicalLogBeginOffset(),
                                     LogicalLogBeginOffset(),
                                     m_LogInfoPage.ReadCycleRelatedEndOffset() );
            iter.SetPosition (LogicalLogBeginOffset());
            m_PageReader.ReadLogPage ( taskid, m_BufferedPage,
                                       Log_ClusterAddress(iter.GetPosition(),1) );
            IOSequenceNo IOSequenceOfFirstOffset = m_BufferedPage.ReadFirstWriterIOSequenceNo();

            // second, look at last offset and then at last offset of the device
            --iter;
            m_PageReader.ReadLogPage ( taskid, m_BufferedPage,
                                       Log_ClusterAddress(iter.GetPosition(),1) );
            IOSequenceNo IOSequenceOfLastOffset = m_BufferedPage.ReadFirstWriterIOSequenceNo();

            if ( IOSequenceOfLastOffset.IsInvalid() && IOSequenceOfFirstOffset.IsValid() ) // 2002-02-15 mb pts1111525 partially cleared log
            {
                // In this case the log is not written around
                m_FirstKnownIOSeq  = IOSequenceOfFirstOffset;
                m_FirstKnownOffset = LogicalLogBeginOffset();
            }
            else
            {
                // the log has been written around, but may also be partially cleared, so
                // search from lastknown forward to first valid -> this must be the smallest io-seq
                iter.SetPosition (m_LogInfoPage.ReadLastKnownOffset());
                ++iter;
                m_PageReader.ReadLogPage ( taskid, m_BufferedPage,
                                           Log_ClusterAddress(iter.GetPosition(),1) );
                while ( m_BufferedPage.ReadFirstWriterIOSequenceNo().IsInvalid() )
                {
                    ++iter;
                    m_PageReader.ReadLogPage ( taskid, m_BufferedPage,
                                               Log_ClusterAddress(iter.GetPosition(),1) );
                }
                m_FirstKnownIOSeq  = m_BufferedPage.ReadFirstWriterIOSequenceNo();
                m_FirstKnownOffset = iter.GetPosition();
            }
        }
    }
    else
    {
        m_FirstKnownIOSeq.Invalidate();
        m_FirstKnownOffset.Invalidate();
    }
}
// -----------------------------------------------------------------------------
SAPDB_Bool
Log_ArchiveLog::CheckOffsetAndIOsequence (TaskID          taskid,
                                          RawDeviceOffset offset,
                                          IOSequenceNo    seqno)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::CheckOffsetAndIOsequence", LogVolume_Trace, 5);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "offset: " << offset << ", sequence: " << seqno);
    SAPDB_Bool sequenceFound;
    if ( offset.IsInvalid() )
        sequenceFound = false;
    else
    {
        if (offset <= LogicalLogEndOffset())                    // PTS 1115870 mb 2002-05-23
        {
            m_PageReader.ReadLogPage (taskid, m_BufferedPage,Log_ClusterAddress(offset,1));
            sequenceFound = seqno >= m_BufferedPage.ReadFirstWriterIOSequenceNo()
                            &&
                            seqno <= m_BufferedPage.ReadLastWriterIOSequenceNo();
        }
        else
        {
            sequenceFound = false;
        }
    }
    if ( ! sequenceFound )
        RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_SEQUENCE_NOT_FOUND,
                     SAPDB_ToString(seqno,_T_d),
                     SAPDB_ToString(m_BufferedPage.ReadFirstWriterIOSequenceNo(),_T_d),
                     SAPDB_ToString(m_BufferedPage.ReadLastWriterIOSequenceNo(),_T_d),
                     SAPDB_ToString(offset,_T_d) ));
    
    return sequenceFound;
}
// -----------------------------------------------------------------------------
SAPDB_Bool Log_ArchiveLog::AllocatePageFrames()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::AllocatePageFrames", LogVolume_Trace, 5);

    Kernel_Page::PageFrame frame = m_FrameAlloc.New();
    if (!frame.IsAssigned())
        return false;
    m_LogInfoPage.SetFrame(frame);

    frame = m_FrameAlloc.New();
    if (!frame.IsAssigned())
        return false;
    m_BufferedPage.SetFrame(frame);

    return true;
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::CreateInfoPage
    (DeviceBlockCount               savedist)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::CreateInfoPage", LogVolume_Trace, 5);

    Kernel_Date date;
    Kernel_Time time;
    Kernel_DateTime::GetSystemDateTime(date, time);

    m_LogInfoPage.WriteLogDate(date);
    m_LogInfoPage.WriteLogTime(time);
    m_LogInfoPage.WriteLogFlushMode(m_QueueFlushMode);
    m_LogInfoPage.WriteDeviceState(Log_DeviceStateOkay);
    m_LogInfoPage.WriteQueueCount(0);
    m_LogInfoPage.WriteCycleRelatedEndOffset(m_RawDeviceEnd);

    // invalid last known ioseqno indicates that log has never been written
    m_LogInfoPage.WriteLastKnownIOSequenceNo(IOSequenceNo());           // == invalid
    m_LogInfoPage.WriteLastKnownOffset      (RawDeviceOffset());        // == invalid

    m_LogInfoPage.WriteDevspaceEnabled(true); // PTS 1113190 UH 2001-12-19

    m_LogInfoPage.WriteOldestNotSavedOffset(LogicalLogBeginOffset());

    // invalid oldest not saved ioseqno indicates that log has never been saved
    m_LogInfoPage.WriteOldestNotSavedIOSequenceNo(IOSequenceNo());      // == invalid

    m_LogInfoPage.WriteLogBackupCount(0);
    m_LogInfoPage.WriteRedoUntilDate(0);
    m_LogInfoPage.WriteRedoUntilTime(0);
    m_LogInfoPage.WriteClearUntilLowerOffset(RawDeviceOffset());        // inv.
    m_LogInfoPage.WriteClearUntilUpperOffset(RawDeviceOffset());        // inv.
    m_LogInfoPage.WriteIOSequenceNoValidAfterRedoUntil(IOSequenceNo()); // inv.

    tkb00_SaveptParam &SaveptParam = k57restartrec->rstLastSavept_kb00();
    m_LogInfoPage.WriteSavepointParameters(
        SAPDBFields_Field(
            reinterpret_cast<SAPDB_Byte*>(&SaveptParam), sizeof(SaveptParam) ) );
}
// -----------------------------------------------------------------------------
SAPDB_Bool Log_ArchiveLog::CreateLogQueues
    (SAPDB_UInt                     queuecount,
     QueueSize                      queuesize)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::CreateLogQueues", LogVolume_Trace, 5);

    for (Log_QueueID id=0; id<queuecount; id++)
    {
        Log_Queue* queue = new(m_Allocator)
            Log_Queue(id, m_QueueSyncMode, m_QueueFlushMode, m_WriterTaskSync,
                      m_Allocator, m_FrameAlloc);

        if (queue == 0)
            return false;

        QueueRing::Iterator insertIter = m_LogQueues.Insert(queue);

        if (!insertIter.IsValid())                      // alloc failed?
            return false;

        if (!queue->Initialize(queuesize))              // alloc failed?
            return false;
    }
    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "queuecount: " << queuecount << ", queuesize: " << queuesize);
    return true;
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::ResetLogWriter (TaskID          taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ResetLogWriter", LogVolume_Trace, 5);

    if ( m_LogInfoPage.ReadLastKnownOffset().IsValid() )
    {
        RawDeviceIterator iter ( LogicalLogBeginOffset(),
                                 m_LogInfoPage.ReadLastKnownOffset(),
                                 LogicalLogEndOffset());

        ++iter; // use next offset on device to begin writing
        m_LogWriter.SetState ( taskid,
                               m_LogInfoPage.ReadLastKnownIOSequenceNo(),
                               iter.GetPosition(),
                               LogicalLogEndOffset());
    }
    else
        // log was never been written
        m_LogWriter.SetState ( taskid,
                               m_LogInfoPage.ReadLastKnownIOSequenceNo(),
                               LogicalLogBeginOffset(),
                               LogicalLogEndOffset());
}
// -----------------------------------------------------------------------------
SAPDB_Bool Log_ArchiveLog::InitLogWriter (RawDeviceOffset offset,
                                          IOSequenceNo    seqno,
                                          SAPDB_UInt      ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::InitLogWriter", LogVolume_Trace, 5);

    RawDeviceIterator iter ( LogicalLogBeginOffset(),
                             LogicalLogBeginOffset(),
                             LogicalLogEndOffset());

    if ( offset.IsValid() )
    {
        iter.SetPosition(offset);
        ++iter; // use next offset on device to begin writing
    }

    return m_LogWriter.Initialize (seqno,
                                   iter,
                                   this,
                                   m_LogInfoPage,
                                   LogInfoPageOffset(),
                                   ioblockcount);
}

// -----------------------------------------------------------------------------

void Log_ArchiveLog::FormatLog				// new 2002-02-15 mb PTS 1111525
    (const TaskID                         taskid,
     const RawDeviceOffset                beginOffset,
     const RawDeviceOffset                endOffset)
{
    if (beginOffset <= endOffset)
    {
        FormatLogLinear(taskid, beginOffset, endOffset);
    }
    else
    {
        FormatLogLinear(taskid, beginOffset, LogicalLogEndOffset());
        FormatLogLinear(taskid, LogicalLogBeginOffset(), endOffset);
    }
    m_FirstKnownIOSeq.Invalidate();
    m_FirstKnownOffset.Invalidate();
}

// -----------------------------------------------------------------------------

void Log_ArchiveLog::FormatLogLinear
    (const TaskID                         taskid,
     const RawDeviceOffset                beginoffs,
     const RawDeviceOffset                endoffs)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::FormatLog", LogVolume_Trace, 5);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "beginoffs: " << beginoffs << ", endoffs: " << endoffs);

    SAPDBERR_ASSERT_RANGE(beginoffs, LogicalLogBeginOffset(), LogicalLogEndOffset());
    SAPDBERR_ASSERT_RANGE(endoffs, LogicalLogBeginOffset(), LogicalLogEndOffset());

    m_BufferedPage.InitializeFrame(Log_QueueID(), IOSequenceNo());

    RawDeviceIterator iter(LogicalLogBeginOffset(), beginoffs, LogicalLogEndOffset());
    RawDeviceIterator end (LogicalLogBeginOffset(), endoffs, LogicalLogEndOffset());


    RTE_Message( Log_Exception(__FILE__,__LINE__,
                               LOG_FORMATTING,
                               SAPDB_ToString(beginoffs,_T_d),
                               SAPDB_ToString(endoffs,_T_d) ));

    // fill i/o cluster with initalized pages:
    m_IOPages.Clear();
    while (!m_IOPages.IsFull())
        m_IOPages.Add(m_BufferedPage);

    // while blocks clusterwise as far as possible:
    while (iter.GetDistanceTo(end.GetPosition()) >= m_IOPages.GetPageCount())
    {
        m_PageWriter.WriteLogPages
            (taskid, m_IOPages, Log_ClusterAddress(iter.GetPosition(), m_IOPages.GetPageCount()));
        iter += m_IOPages.GetPageCount();
    }

    // write ultimate pages with single page i/o's:
    m_IOPages.Clear();
    m_IOPages.Add(m_BufferedPage);
    while (iter != end)
    {
        m_PageWriter.WriteLogPages
            (taskid, m_IOPages, Log_ClusterAddress(iter.GetPosition(), 1));
        ++iter;
    }
    m_PageWriter.WriteLogPages
        (taskid, m_IOPages, Log_ClusterAddress(iter.GetPosition(), 1));
}

// -----------------------------------------------------------------------------

Log_ErrorState Log_ArchiveLog::RestartForClearLog   //PTS 1111525 mb 2002-01-24
        (TaskID                         taskid,
         Log_RawDeviceOffset            rawlogbegin,
         Log_RawDeviceOffset            rawlogend,
         DeviceBlockCount               autosaveDistance,
         Log_Queue::EntryCount          queuesize,
         SAPDB_UInt                     ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::RestartForClearLog", LogVolume_Trace, 5);

    return RestartIntern ( taskid,
                           rawlogbegin,
                           rawlogend,
                           autosaveDistance,
                           queuesize,
                           ioblockcount,
                           Restart_for_clear_log );
}

// -----------------------------------------------------------------------------

void Log_ArchiveLog::ClearLogPartially
    (const TaskID                         taskid,
     const Log_RawDeviceOffset            firstOffsetToClear, 
     const Log_RawDeviceOffset            lastOffsetToClear,
     const Log_IOSequenceNo               lastKnownIOSeqAfterClear)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ClearLogPartially", LogVolume_Trace, 5);

    Log_RawDeviceIterator	succOfLast = ++Log_RawDeviceIterator(LogicalLogBeginOffset(), 
                                                                 lastOffsetToClear,
                                                                 LogicalLogEndOffset());
    if (succOfLast.GetPosition() == firstOffsetToClear)
    {
        ClearLogComplete(taskid,lastKnownIOSeqAfterClear);
    }
    else
    {
        if (lastKnownIOSeqAfterClear > m_LogInfoPage.ReadOldestNotSavedIOSequenceNo()) // PTS 1113550 mb 2002-05-16
        // current OldestNotSaved is OK => dont change it
        {
            ClearLogIntern(taskid,
                           firstOffsetToClear, lastOffsetToClear,
                           (--Log_RawDeviceIterator(LogicalLogBeginOffset(),    // LastKnownOffset is the predecessor
                                                    firstOffsetToClear,          // of the first cleared page
                                                    LogicalLogEndOffset())).GetPosition(),
                           m_LogInfoPage.ReadOldestNotSavedOffset(),                           // oldestNotSavedOffset
                           m_LogInfoPage.ReadOldestNotSavedIOSequenceNo(),    // OldestNotSavedIOSeq is next new one
                           Log_DeviceStatePartiallyClearing, Log_DeviceStatePartiallyCleared);
        }
        else
        // Page of OldestNotSaved will be cleared => reset it to the first cleared page
        {
            ClearLogIntern(taskid,
                           firstOffsetToClear, lastOffsetToClear,
                           (--Log_RawDeviceIterator(LogicalLogBeginOffset(),    // LastKnownOffset is the predecessor
                                                    firstOffsetToClear,          // of the first cleared page
                                                    LogicalLogEndOffset())).GetPosition(),
                           firstOffsetToClear,                           // oldestNotSavedOffset
                           lastKnownIOSeqAfterClear,                     // OldestNotSavedIOSeq is next new one
                           Log_DeviceStatePartiallyClearing, Log_DeviceStatePartiallyCleared);
        }
        // delete also log-entries on the last page if needed
        Log_RawDeviceOffset lastNotClearedPageOffset = firstOffsetToClear - 1;
        Log_IOSequenceNo    lastValidIOSequence = lastKnownIOSeqAfterClear - 1;
        TruncateLogPage (taskid,
                         lastValidIOSequence,
                         lastNotClearedPageOffset);
    }
}

// ----------------------------------------------------------------------------- // new PTS1113550 mb 2002-05-15
void Log_ArchiveLog::ClearLogComplete   (const TaskID                         taskid,
                                         const Log_IOSequenceNo               lastKnownIOSeqAfterClear)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ClearLogComplete", LogVolume_Trace, 5);
    ClearLogIntern(taskid, 
                   LogicalLogBeginOffset(), LogicalLogEndOffset(),
                   Log_RawDeviceOffset(),                          // LastKnownOffset = invalid
                   Log_RawDeviceOffset(),                          // OldestNotSavedOffset = invalid
                   lastKnownIOSeqAfterClear,                       // OldestNotSavedIOSeq is next new one
                   Log_DeviceStateClearing, Log_DeviceStateCleared);
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::ClearLogComplete   (const TaskID                         taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ClearLogComplete", LogVolume_Trace, 5);
    ClearLogIntern(taskid, 
                   LogicalLogBeginOffset(), LogicalLogEndOffset(),
                   Log_RawDeviceOffset(),                          // LastKnownOffset = invalid
                   Log_RawDeviceOffset(),                          // OldestNotSavedOffset = invalid
                   ++(m_LogInfoPage.ReadLastKnownIOSequenceNo()),                             // OldestNotSavedIOSeq is next new one
                   Log_DeviceStateClearing, Log_DeviceStateCleared);
}

// -----------------------------------------------------------------------------

void Log_ArchiveLog::ClearLogIntern (                      // PTS 1111525 mb 2002-01-29 new
        const TaskID                          taskid,
        const Log_RawDeviceOffset             firstOffsetToClear, 
        const Log_RawDeviceOffset             lastOffsetToClear,
        const Log_RawDeviceOffset             lastKnownOffset,
        const Log_RawDeviceOffset             oldestNotSavedOffset,
        const Log_IOSequenceNo                oldestNotSavedIOSeq,
        const Log_DeviceState                 stateBeforeFormat,
        const Log_DeviceState                 stateAfterFormat)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ClearLogIntern", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_ARGUMENT (stateAfterFormat != Log_DeviceStateOkay);
    
    m_LogInfoPage.EnterRegion(taskid);
    m_LogInfoPage.WriteClearUntilLowerOffset      ( firstOffsetToClear );
    m_LogInfoPage.WriteClearUntilUpperOffset      ( lastOffsetToClear   );
    m_LogInfoPage.WriteDeviceState                ( stateBeforeFormat );
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);

    FormatLog(taskid, firstOffsetToClear, lastOffsetToClear);

    m_LogInfoPage.EnterRegion(taskid);
    // PTS 1118120 UH 2002-10-01 begin
    const bool clearCompleteLog = firstOffsetToClear == LogicalLogBeginOffset()
                                  &&
                                  lastOffsetToClear == LogicalLogEndOffset();

    if ( clearCompleteLog )
    {
        // complete log is cleared => reset iosequence and log backup no
        m_LogInfoPage.WriteLastKnownIOSequenceNo      ( Log_IOSequenceNo()      );
        m_LogInfoPage.WriteLastKnownOffset            ( Log_RawDeviceOffset()   );
        m_LogInfoPage.WriteLogBackupCount             ( 0                       );
        m_LogInfoPage.WriteOldestNotSavedOffset       ( LogicalLogBeginOffset() );
        m_LogInfoPage.WriteOldestNotSavedIOSequenceNo ( Log_IOSequenceNo()      );
        k57restartrec->rstLastSavept_kb00().svpIOsequence_kb00  = Log_IOSequenceNo();
        k57restartrec->rstLastSavept_kb00().svpStartOffset_kb00 = Log_RawDeviceOffset();
    }
    else
    {
        m_LogInfoPage.WriteLastKnownOffset            ( lastKnownOffset );
        m_LogInfoPage.WriteOldestNotSavedOffset       ( oldestNotSavedOffset );
        m_LogInfoPage.WriteOldestNotSavedIOSequenceNo ( oldestNotSavedIOSeq );
    }
    // PTS 1118120 UH 2002-10-01 end
    m_LogInfoPage.WriteClearUntilLowerOffset ( Log_RawDeviceOffset());     // set to invalid
    m_LogInfoPage.WriteClearUntilUpperOffset ( Log_RawDeviceOffset());     // set to invalid
    m_LogInfoPage.WriteDeviceState ( stateAfterFormat );
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);

    if ( clearCompleteLog ) // PTS 1118120 UH 2002-10-01
        k57save_restartrecord (taskid);

    ResetLogWriter(taskid);
    k38history_lost_write(taskid);
}

// -----------------------------------------------------------------------------
void
Log_ArchiveLog::WriteSavepointValuesAndFlushInfoPage
    (TaskID                   taskid,
     Log_SavepointReason      reason,
     const SAPDBFields_Field& rrpart)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::WriteSavepointValuesAndFlushInfoPage", LogVolume_Trace, 5);
 
    m_LogInfoPage.EnterRegion(taskid);
    // PTS 1113190 UH 2001-12-19 do not reset history lost here
    m_LogInfoPage.WriteSavepointParameters (rrpart);
    FlushLogInfoPage (taskid);
 
    m_LogInfoPage.LeaveRegion(taskid);
    
    if ( m_LogInfoPage.ReadAutoOverwrite() ) // PTS 1120106 UH 2003-01-20
    {
        // no log backup wanted so free log automatically to last savepoint
        tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();
        if ( ! FreeLog (taskid, m_LogInfoPage.ReadOldestNotSavedIOSequenceNo(),
                                m_LogInfoPage.ReadOldestNotSavedOffset(),
                                SaveptParam.svpIOsequence_kb00,
                                SaveptParam.svpStartOffset_kb00) )
            RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED, "FreeLog"));
    }

    ClearLogFullQueueState(taskid);
    
    #ifdef SAPDB_SLOW
    if ( LogVolume_Trace.TracesLevel(6) )
        m_LogInfoPage.WriteToTrace();
    #endif
}

// -----------------------------------------------------------------------------
void
Log_ArchiveLog::WriteDeviceParametersToTrace ()
{
    Kernel_VTrace() << "LogicalLogOffsets: " << LogicalLogBeginOffset()
                    << " -> " << LogicalLogEndOffset();
    Kernel_VTrace() << "FirstKnownSeq: " << m_FirstKnownIOSeq
                    << "@" << m_FirstKnownOffset ;
    m_LogInfoPage.WriteToTrace();
    // only correct for one queue
    //QueueRing::Iterator queueIter = m_LogQueues.Anchor();
    //(**queueIter).WriteToTrace();
}
// -----------------------------------------------------------------------------
void
Log_ArchiveLog::SetToLastSavepointEntry (Log_ArchiveLogIterator &iter,
                                         IOSequenceNo           &svpSeq)
{
    tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();
    svpSeq = SaveptParam.svpIOsequence_kb00;
    RawDeviceOffset svpOff (SaveptParam.svpStartOffset_kb00); 
    if ( svpSeq.IsValid()
         &&
         (svpOff.IsInvalid()                                           // PTS 1115870 MB 2002-05-23
          ||
          SaveptParam.svpStartOffset_kb00 <= LogicalLogEndOffset())) // PTS 1115091 UH 2002-03-27
    {
        iter.SetPosition ( SaveptParam.svpStartOffset_kb00,
                           SaveptParam.svpStartEntryOffset_kb00 );
    }
	else
	{
        if (LogIsEmpty ())                        // PTS 1115870 MB 2002-05-23
        {
            // for an empty log with no savepoint: start with IOSeq ZERO
            svpSeq = 0;
            iter.SetLastKnownIOSeq(svpSeq);
            iter.SetPosition(LogicalLogBeginOffset());
        }
        else
        // there has been a savepoint written, but the offset is incorrect
        {
              RTE_Crash(Log_Exception(__FILE__,__LINE__,LOG_SAVEPOINT_NOT_FOUND));
        }
	}
}
// -----------------------------------------------------------------------------
void
Log_ArchiveLog::SetToBeginOfPage (Log_ArchiveLogIterator    &iter,
                                  Log_RawDeviceOffset        offset)
{
    iter.SetPosition (offset);
}
// -----------------------------------------------------------------------------
Log_ArchiveLogIterator 
Log_ArchiveLog::GetArchiveLogIterator (
                      SAPDBMem_IRawAllocator    &alloc,
                      tsp00_TaskId               taskId,
                const Kernel_Date                untilDate,    // PTS 1111525 mb 2001-01-15
                const Kernel_Time                untilTime,    // PTS 1111525 mb 2001-01-15
                const Log_IOSequenceNo           untilIOSeq)   // PTS 1121539 mb 2003-02-06
const
{
    Log_SeqLogDevReader    *logDeviceReader = new(alloc) 
                                              Log_SeqLogDevReader(  LogicalLogBeginOffset(),
                                                                    LogicalLogEndOffset(),
                                                                    m_PageReader);


    return Log_ArchiveLogIterator
                    (taskId,
                     logDeviceReader,
                     m_LogInfoPage.ReadLastKnownIOSequenceNo(),
                     alloc,
                     Log_FrameAllocator::Instance(),
                     Log_CheckQueueSequences,
                     untilDate,untilTime,untilIOSeq );
}

// -----------------------------------------------------------------------------
Log_ArchiveLogIterator 
Log_ArchiveLog::GetTapeIterator (
                      SAPDBMem_IRawAllocator     &alloc,
                      tgg00_TransContext         &readerTrans,
                      Log_IOSequenceNo            lastKnownOnTape,
                const Kernel_Date                 untilDate,    // PTS 1111525 mb 2001-01-15
                const Kernel_Time                 untilTime,    // PTS 1111525 mb 2001-01-15
                const Log_IOSequenceNo           untilIOSeq)   // PTS 1121539 mb 2003-02-06
const
{
    tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();
    IOSequenceNo lastIOSequence = SaveptParam.svpIOsequence_kb00;
	if (lastIOSequence.IsInvalid())
	// no savepoint known => restore log from the first possible IOSequence
	{
		lastIOSequence = 0;
	}
    Log_SeqTapeReader    *tapeReader = new(alloc) 
                                              Log_SeqTapeReader(lastIOSequence,
                                                                readerTrans);
    return Log_ArchiveLogIterator
                    (readerTrans.trTaskId_gg00,
                     tapeReader,
                     lastKnownOnTape,
                     alloc,
                     Log_FrameAllocator::Instance(),
                     Log_CheckQueueSequences,
                     untilDate,untilTime,untilIOSeq); 
}

// -----------------------------------------------------------------------------
Log_ArchiveLog::DeviceBlockCount 
Log_ArchiveLog::GetNumberOfUnsavedPages() const
{
    if (m_Allocated)
    // warm operating mode => m_LogWriter is allocated and active
    // this depends on the assumption that it is not possible
    // to change directly from WARM to COLD
    {
        RawDeviceIterator iter ( LogicalLogBeginOffset(),
                                 m_LogWriter.GetWriteOffset(),
                                 LogicalLogEndOffset());
        if ( m_LogInfoPage.ReadOldestNotSavedOffset().IsValid() )
        {
            if (iter.GetPosition() != m_LogInfoPage.ReadOldestNotSavedOffset())
            // the currently written page can never saved in warm mode
            {
                --iter;
                return iter.GetDistanceFrom(m_LogInfoPage.ReadOldestNotSavedOffset());
            }
            else
            // everything saved except the currently written page
            {
                return 0;
            }
        }
        else
        {
            return iter.GetDistanceFrom(LogicalLogBeginOffset());
        }
    }
    else
    // cold operating mode
    {
        if (m_LogInfoPage.ReadLastKnownOffset().IsValid())
        {
            RawDeviceIterator iter ( LogicalLogBeginOffset(),
                                     m_LogInfoPage.ReadLastKnownOffset(),
                                     LogicalLogEndOffset());
    
            if ( m_LogInfoPage.ReadOldestNotSavedOffset().IsValid() )                             
                return iter.GetDistanceFrom(m_LogInfoPage.ReadOldestNotSavedOffset());
            else
                return iter.GetDistanceFrom(LogicalLogBeginOffset());
        }
        else
        {
            return 0;
        }
    }
}


// -----------------------------------------------------------------------------
void Log_ArchiveLog::GetSaveLogIterator( tsp00_TaskId    		 taskId,
                                         SAPDBMem_IRawAllocator &allocator,
                                         bool                    isCold,
                                         bool              		 forAutosave,
                                         bool              		 completeLog,
                                         Log_DeviceBlockCount    blockSize,
                                         Log_SaveIterator*      &iterator,
                                         SaveInfo               &saveInfo)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::GetSaveLogIter", LogVolume_Trace, 5);
    
    iterator = NULL;
    saveInfo = okay;

    if ( m_LogInfoPage.ReadDeviceState() != Log_DeviceStateOkay ) 
    {
        RTE_Message( Log_Exception(__FILE__,__LINE__,
                                   LOG_HISTORY_LOST_NO_LOGSAVE_ALLOWED,
                                   Log_DeviceStateStrings[m_LogInfoPage.ReadDeviceState()]
                                   ));
        saveInfo = historyIsLost;
        return;
    }

    // determine first and last page to save
    Log_RawDeviceOffset startPoint = m_LogInfoPage.ReadOldestNotSavedOffset();
    Log_RawDeviceOffset endPoint;
    
    if (isCold)
    // for cold databases no log_writer exists, so use the log_info_page
    {
        endPoint=m_LogInfoPage.ReadLastKnownOffset();
    }
    else
    {
        endPoint = m_LogWriter.GetWriteOffset();
    }
    
    Log_RawDeviceIterator startIter;
    Log_RawDeviceIterator endIter;

    if ( ! endPoint.IsValid() )
    // exit if the log has no endpoint for save
    {
        SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "Log_ArchiveLog::GetSaveLogIter: No End-Offset found");
        saveInfo = noLogToSave;
        return;
    }

    bool firstSaveOfLog;
    if (startPoint.IsInvalid())
    {
        startPoint = LogicalLogBeginOffset();
        firstSaveOfLog = true;
    }
    else
    {
        firstSaveOfLog = false;
    }
    
    if ( ! isCold )
    // in warm-state then current-log-page must not saved, because it
    // may be modified
    {
        endIter = Log_RawDeviceIterator( LogicalLogBeginOffset(),
                                         endPoint,
                                         LogicalLogEndOffset());

        // check if writerposition points to the only page to be saved
        if (endIter.GetPosition() == startPoint)
        {
            saveInfo = noLogToSave;
            return;
        }
        
        // decrement enditer because this page is in use by the writer
        --endIter;
        
        if ( Log_DeviceMinimizeSpace == m_QueueFlushMode )
        {
            // check if writerposition points to the only page to be saved
            if (endIter.GetPosition() == startPoint)
            {
                saveInfo = noLogToSave;
                return;
            }
            --endIter;
        }
            
        endPoint=endIter.GetPosition();;
    }    
    
    // check if something has to be saved
    startIter = Log_RawDeviceIterator( LogicalLogBeginOffset(),
                                     startPoint,
                                     LogicalLogEndOffset());
    // if startPoint denotes the page after endpoint, then nothing has to be saved
    --startIter;    // startIter is reset later
    if (startIter.GetPosition() == endPoint)
    {
        if (!firstSaveOfLog)
        // when the log is saved first it may occur that every page is written
        // so startPoint denotes the first page on the log end endpoint the 
        // last page on the log
        {
            saveInfo = noLogToSave;
            return;
        }
    }

    // determine first and last IOSequence to be saved
    Log_IOSequenceNo startIOSeq = m_LogInfoPage.ReadOldestNotSavedIOSequenceNo();
    if (startIOSeq.IsInvalid())
    {
        startIOSeq = 0;
    }
    
    Log_DeviceBlockCount max_pages_to_save;
    if (completeLog)							// PTS1111985 mb 2001-11-16
    // save as much as possible (i.e. for save into pipes)
    {
        max_pages_to_save = LogicalLogEndOffset()-LogicalLogBeginOffset()+1;
    }
    else
    // save maximal plogSegmentSize_gg00 pages in one run
    {
    	max_pages_to_save = GetAutosaveDistance();  // PTS 1113361 mb 2002-01-11
    }
    if (max_pages_to_save == 0)
    {
  		RTE_Crash(Log_Exception(__FILE__,__LINE__,LOG_SAVEITER_INV_SEG_SIZE));
    }

    startIter = Log_RawDeviceIterator( LogicalLogBeginOffset(),
                                       startPoint,
                                       LogicalLogEndOffset());
    
    bool isCompleteSegment;                     // PTS1111987 mb 2001-10-12
    if (startIter.GetDistanceTo(endPoint)+1 > max_pages_to_save)
    {
        isCompleteSegment = true;
        endIter = startIter + (max_pages_to_save - 1); // +++ added () UH 2001-06-27
    }
    else
    {
        // For autosave only complete segments are allowed
        // so exit if not enough has to be saved
        if ( forAutosave )
        {
            saveInfo = incompleteSegment;
            return;
        }
        isCompleteSegment = false;
        endIter = Log_RawDeviceIterator(m_RawDeviceBegin,
                                        endPoint,
                                        m_RawDeviceEnd);
    }


    iterator = new (allocator) Log_SaveIterator(
                        allocator,
                        m_FrameAlloc,
                        m_PageReader,
                        startIter,
                        endIter,
                        startIOSeq,
                        blockSize,
                        isCompleteSegment);


    if ( iterator == NULL )
    {
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                      SAPDBERR_ASSERT_STATE_FAILED,
                      "GetSaveLogIter:alloc Log_SaveIterator failed"));
    }


    SAPDBTRACE_WRITELN (LogVolume_Trace, 6,"Log_ArchiveLog::GetSaveLogIter save from page " << iterator->GetStartPageOffset() << " to " << iterator->GetEndPageOffset());

    if (iterator->Initialize(taskId, GetNextBackupCount(taskId)))
    {
        if (iterator->GetEndIOSequence().IsValid() &&
            m_LogInfoPage.ReadOldestNotSavedIOSequenceNo().IsValid())
        {
        
            if (iterator->GetEndIOSequence() < m_LogInfoPage.ReadOldestNotSavedIOSequenceNo())
            // if the endioseq to be saved is smaller than the oldest_not_saved
            // then the pages defined by this iterator had already been saved
            {
                Kernel_VTrace() << "endioseq: " << iterator->GetEndIOSequence()
                                << " < " << m_LogInfoPage.ReadOldestNotSavedIOSequenceNo()
                                << NewLine;
                
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                              SAPDBERR_ASSERT_STATE_FAILED,
                              "Log_ArchiveLog::GetSaveLogIter() endioseq < OldestNotSaved"));
            }
        }
        else
        {
            if (iterator->GetEndIOSequence().IsInvalid())
            // endiosequence has to be valid at this point
            {
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                              SAPDBERR_ASSERT_STATE_FAILED,
                              "Log_ArchiveLog::GetSaveLogIter() endioseq is NOT valid"));
            }
        }
    }
    else
    {
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                      SAPDBERR_ASSERT_STATE_FAILED,
                      "GetSaveLogIter:initialize Log_SaveIterator failed"));
    }
}
// -----------------------------------------------------------------------------

SAPDB_UInt4 Log_ArchiveLog::GetNextBackupCount ( tsp00_TaskId taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::GetNextBackupCount", LogVolume_Trace, 5);

    m_LogInfoPage.EnterRegion(taskid);

        SAPDB_UInt backupcount = m_LogInfoPage.ReadLogBackupCount();
        backupcount++;
        m_LogInfoPage.WriteLogBackupCount(backupcount);
        FlushLogInfoPage(taskid);

    m_LogInfoPage.LeaveRegion(taskid);

    return backupcount;
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLog::FreeLog (tsp00_TaskId        taskid, 
                              Log_IOSequenceNo    oldLastSavedIOsequence,
                              Log_RawDeviceOffset oldLastSavedOffset,
                              Log_IOSequenceNo    newOldestNotSavedIOsequence,
                              Log_RawDeviceOffset newOldestNotSavedOffset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::FreeLog", LogVolume_Trace, 5);

    // check IOSequences
    Log_IOSequenceNo infoPageOldestNotSavedIOSeq = m_LogInfoPage.ReadOldestNotSavedIOSequenceNo();
    if (infoPageOldestNotSavedIOSeq.IsValid())
    {
        if (infoPageOldestNotSavedIOSeq != oldLastSavedIOsequence)
        {
            return false;
        }
        
        if (newOldestNotSavedIOsequence < infoPageOldestNotSavedIOSeq )
        {
            return false;
        }
    }
    
    m_LogInfoPage.EnterRegion(taskid);
    
    m_LogInfoPage.WriteOldestNotSavedIOSequenceNo(newOldestNotSavedIOsequence);
    m_LogInfoPage.WriteOldestNotSavedOffset      (newOldestNotSavedOffset);
    
    if (m_LogInfoPage.ReadLastKnownIOSequenceNo().IsValid())
    {
        if (newOldestNotSavedIOsequence > (m_LogInfoPage.ReadLastKnownIOSequenceNo()+1))
        {
            m_LogInfoPage.WriteLastKnownIOSequenceNo(newOldestNotSavedIOsequence);
            m_LogInfoPage.WriteLastKnownOffset(newOldestNotSavedOffset);
        }
    }

    FlushLogInfoPage (taskid);
    
    m_LogInfoPage.LeaveRegion(taskid);
    
    if (m_Restarted)
    {
        ClearLogFullQueueState(taskid);
    }
    if (m_Allocated)
    // Resume the WriterTask only if really neccessary 
    // (i.e. everything is running and not for save-cold)
    {
        m_WriterTaskSync.Resume();
    }

    // PTS 1114791 mb 2002-08-28
    m_pipeSaveState.waitingForExecuteFreeLogForPipe = false;

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "oldestNotSavedSeqNo:  " << newOldestNotSavedIOsequence);
    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "oldestNotSavedOffset: " << newOldestNotSavedOffset);
    
    return true;
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::TruncateLogPage (      TaskID               taskid,
                                      const Log_IOSequenceNo     newLastIOSequence,
                                      const Log_RawDeviceOffset  pageOffset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::TruncateLogPage", LogTrans_Trace, 5);
    SAPDBTRACE_WRITELN (LogVolume_Trace, 5, "TruncateLogPage until IOSequence:  " << newLastIOSequence);
    SAPDBTRACE_WRITELN (LogVolume_Trace, 5, "TruncateLogPage at Offset:  " << pageOffset);

    m_PageReader.ReadLogPage (taskid, m_BufferedPage,Log_ClusterAddress(pageOffset,1));
    m_BufferedPage.Truncate(newLastIOSequence);
    m_IOPages.Clear();
    m_IOPages.Add(m_BufferedPage);
    m_PageWriter.WriteLogPages (taskid, m_IOPages, Log_ClusterAddress(pageOffset, 1));
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::ChangeEndOffsetAfterAddDevSpace
        (tsp00_TaskId        writertaskid,
         Log_RawDeviceOffset newEndOffset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ChangeEndOffsetAfterAddDevSpace", LogVolume_Trace, 5);
    
    m_RawDeviceEnd = newEndOffset;

    m_LogInfoPage.EnterRegion(writertaskid);

    m_LogInfoPage.WriteCycleRelatedEndOffset(newEndOffset);

    FlushLogInfoPage (writertaskid);

    m_LogInfoPage.LeaveRegion(writertaskid);
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::SetNewDeviceEndOffset ( SAPDB_UInt4 newEndOffset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::SetNewDeviceEndOffset", LogVolume_Trace, 5);
    m_LogWriter.SetDeviceEnd(newEndOffset);
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::FlushLogInfoPage (tsp00_TaskId taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::FlushLogInfoPage", LogVolume_Trace, 5);
 
    if ( ! m_LogInfoPage.InRegion(taskid) )
    {
		RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
					  SAPDBERR_ASSERT_STATE_FAILED,
					  "Log_ArchiveLog::FlushLogInfoPage() not in region "));
    }
 
    Kernel_Date date;
    Kernel_Time time;
    Kernel_DateTime::GetSystemDateTime(date, time);
 
    m_LogInfoPage.WriteLogDate(date);
    m_LogInfoPage.WriteLogTime(time);
 
    m_PageWriter.WriteLogInfoPage ( taskid,
                                    m_LogInfoPage,
                                    Log_ClusterAddress(LogInfoPageOffset(),1) );
 
    #ifdef SAPDB_SLOW
    if ( LogVolume_Trace.TracesLevel(6) )
        m_LogInfoPage.WriteToTrace();
    #endif
}
// -----------------------------------------------------------------------------
void Log_ArchiveLog::UpdateInfoPageFromWriter (tsp00_TaskId        writertaskid,
                                               Log_IOSequenceNo    lastKnownSequence,
                                               Log_RawDeviceOffset lastKnownOffset)
{
    m_LogInfoPage.EnterRegion(writertaskid);
    if (    m_LogInfoPage.ReadDeviceState() == Log_DeviceStateCleared
         || m_LogInfoPage.ReadDeviceState() == Log_DeviceStatePartiallyCleared )   // PTS 1111525 mb 2002-01-29
    {
        m_LogInfoPage.WriteDeviceState ( Log_DeviceStateHistoryLost );
        m_LogInfoPage.WriteDBIdentifier(k57restartrec->rstDbIdent_kb00());  //PTS1113550 mb 2002-05-08
    }
    
    m_LogInfoPage.WriteLastKnownIOSequenceNo(lastKnownSequence);
    m_LogInfoPage.WriteLastKnownOffset      (lastKnownOffset);
    FlushLogInfoPage (writertaskid);

    m_LogInfoPage.LeaveRegion(writertaskid);
}
 
// -----------------------------------------------------------------------------
void Log_ArchiveLog::GetDeviceUsage ( SAPDB_UInt &deviceSize,
                                      SAPDB_UInt &pagesUsed,
                                      SAPDB_UInt &pagesNotSaved )
{
    deviceSize    = 0;
    pagesUsed     = 0;
    pagesNotSaved = 0;
    if ( ! m_Allocated )
        return;
        
    Log_Writer& writer = GetWriter();
    
    deviceSize = GetDeviceSize();
    pagesUsed  = deviceSize - writer.PagesFreeForUsers(writer.CalculateFreePages());

    RawDeviceIterator iter ( LogicalLogBeginOffset(),
                             writer.GetWriteOffset(),
                             LogicalLogEndOffset());

    if ( m_LogInfoPage.ReadOldestNotSavedOffset().IsValid() )                             
        pagesNotSaved = iter.GetDistanceFrom(m_LogInfoPage.ReadOldestNotSavedOffset());
    else
        pagesNotSaved = iter.GetDistanceFrom(LogicalLogBeginOffset());
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::GetQueueUsage ( SAPDB_UInt4    &queueSize,
                                     SAPDB_UInt4    &maxPagesUsedInQueue,
                                     SAPDB_UInt4    &countReserveRequests,
                                     SAPDB_UInt4    &countWaitForIO,
                                     SAPDB_UInt4    &countGroupCommit,
                                     SAPDB_UInt4    &countWaitForQueueFree,
                                     SAPDB_UInt4    &maxWaitForPageIO )
{
    queueSize             = 0;
    maxPagesUsedInQueue   = 0;
    countReserveRequests  = 0;
    countWaitForQueueFree = 0;
    maxWaitForPageIO      = 0;

    if ( ! m_Allocated )
        return;

    Log_Queue &queue = **(m_LogQueues.Anchor());
    
    queueSize             = queue.GetSize();
    maxPagesUsedInQueue   = queue.GetMaxPagesUsed(); // PTS 1115332 UH 2002-04-29 added
    countReserveRequests  = queue.GetReserveCount();
    countWaitForIO        = queue.GetWaitCount();
    countGroupCommit      = queue.GetGroupCommitCount();
    countWaitForQueueFree = queue.GetQueueFullSuspendCount();
    maxWaitForPageIO      = queue.GetMaxCountOfWaitingTasksPerPage();
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::ResetStatistics (tsp00_TaskId taskid)
{
    Log_Queue &queue = **(m_LogQueues.Anchor());
    queue.ResetStatistics (taskid);
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::GetRestartVersion ( tsp00_TaskId      taskid,
                                         Log_IOSequenceNo &firstiosequence,
                                         SAPDB_Char       *dbident )
{
    RawDeviceOffset firstKnownOffset;
    
    GetFirstKnownBeforeRestoreLog (firstKnownOffset, firstiosequence);

    memcpy (dbident, m_LogInfoPage.ReadDBIdentifier(), sizeof(tsp00_Line));
}

// -----------------------------------------------------------------------------
void  Log_ArchiveLog::GetLogDBIdent ( SAPDB_Char       *dbident )
{
    memcpy (dbident, m_LogInfoPage.ReadDBIdentifier(), sizeof(tsp00_Line));
}

// -----------------------------------------------------------------------------
const Log_DeviceState Log_ArchiveLog::GetDeviceState () const
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::GetDeviceState", LogVolume_Trace, 5);

    return m_LogInfoPage.ReadDeviceState();
}


// -----------------------------------------------------------------------------
bool Log_ArchiveLog::GetLogInfo ( tsp00_TaskId              taskid,
                                  Kernel_Date              &date,
                                  Kernel_Time              &time,
                                  Log_FlushMode            &flushmode,
                                  Log_DeviceState          &devicestate,
                                  Log_InfoPage::QueueCount &queuecount,
                                  Log_RawDeviceOffset      &cycleendoffset,
                                  Log_RawDeviceOffset      &oldestnotsavedoffset,
                                  Log_IOSequenceNo         &oldestnotsavedsequence,
                                  Log_RawDeviceOffset      &lastknownoffset,
                                  Log_IOSequenceNo         &lastknownsequence,
                                  SAPDB_UInt4              &backupcount,
                                  Kernel_Date              &untildate,
                                  Kernel_Time              &untiltime,
                                  Log_RawDeviceOffset      &untilclearloweroffset,
                                  Log_RawDeviceOffset      &untilclearupperoffset,
                                  Log_IOSequenceNo         &untilsequenceafterredo,
                                  Log_IOSequenceNo         &savepointsequence,
                                  Log_RawDeviceOffset      &savepointdevoffset,
                                  Log_InfoPage::PageOffset &savepointoffset )
{
    if (! m_LogInfoPage.IsAssigned())
        return false;
        
    m_LogInfoPage.EnterRegion(taskid);

    date                    = m_LogInfoPage.ReadLogDate();
    time                    = m_LogInfoPage.ReadLogTime();
    flushmode               = m_LogInfoPage.ReadFlushMode();
    devicestate             = m_LogInfoPage.ReadDeviceState();
    queuecount              = m_LogInfoPage.ReadQueueCount();
    cycleendoffset          = m_LogInfoPage.ReadCycleRelatedEndOffset();
    oldestnotsavedoffset    = m_LogInfoPage.ReadOldestNotSavedOffset();
    oldestnotsavedsequence  = m_LogInfoPage.ReadOldestNotSavedIOSequenceNo();
    lastknownoffset         = m_LogInfoPage.ReadLastKnownOffset();
    lastknownsequence       = m_LogInfoPage.ReadLastKnownIOSequenceNo();
    backupcount             = m_LogInfoPage.ReadLogBackupCount();
    untildate               = m_LogInfoPage.ReadRedoUntilDate();
    untiltime               = m_LogInfoPage.ReadRedoUntilTime();
    untilclearloweroffset   = m_LogInfoPage.ReadClearUntilLowerOffset();
    untilclearupperoffset   = m_LogInfoPage.ReadClearUntilUpperOffset();
    untilsequenceafterredo  = m_LogInfoPage.ReadIOSequenceNoValidAfterRedoUntil();
    savepointsequence       = m_LogInfoPage.ReadSavepointIOSequenceNo();
    savepointdevoffset      = m_LogInfoPage.ReadSavepointEntryOffset();
    savepointoffset         = m_LogInfoPage.ReadSavepointEntryOffsetInPage();

    m_LogInfoPage.LeaveRegion(taskid);
    
    return true;
}
 
// -----------------------------------------------------------------------------
void Log_ArchiveLog::ClearLogFullQueueState(TaskID      taskid)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ClearLogFullQueueState", LogVolume_Trace, 5);
    QueueIterator queueIter = m_LogQueues.Anchor();
    QueueIterator startPoint = queueIter;
    if (!LogIsFull(taskid))
    {
        do
        {
            (**queueIter).SetLogFullState(taskid,FALSE);
            ++queueIter;
        } while (queueIter != startPoint);
    }
}

// -----------------------------------------------------------------------------
bool  Log_ArchiveLog::DiagnoseOpenDevice (TaskID taskid)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::DiagnoseOpenDevice", LogVolume_Trace, 5);
    return IOMan_ILogManager::GetInstance().OpenAllArchiveLogs(taskid);
}

// -----------------------------------------------------------------------------
bool  Log_ArchiveLog::DiagnoseOpenInfoPage (TaskID taskid,
                                            bool   openDevice)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::DiagnoseOpenInfoPage", LogVolume_Trace, 5);

    if ( openDevice && ! DiagnoseOpenDevice(taskid) )
        return false;

    const SAPDB_Int4 LogCapacity = IOMan_ILogInfo::GetInstance().TotalUsableLogPages();

    m_RawDeviceBegin = 0;
    m_RawDeviceEnd   = LogCapacity-1;

    if ( ! m_Allocated )
    {
        if ( ! AllocatePageFrames() )
        {
            Delete();
            DiagnoseCloseDevice(taskid);
            return false;
        }
    }

    m_LogInfoPage.EnterRegion(taskid);
    m_PageReader.ReadLogInfoPage ( taskid, m_LogInfoPage,
                                   Log_ClusterAddress(LogInfoPageOffset(),1) );
    m_LogInfoPage.LeaveRegion(taskid);
    return true;
}

// -----------------------------------------------------------------------------
void  Log_ArchiveLog::DiagnoseCloseDevice (TaskID taskid)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::DiagnoseCloseDevice", LogVolume_Trace, 5);
    IOMan_ILogManager::GetInstance().CloseAllLogDevSpaces(taskid);
}

// -----------------------------------------------------------------------------
   
void Log_ArchiveLog::SetLogDBIdent ( tsp00_TaskId    taskid, // PTS 1113550 mb 2002-04-30 new
                                     tsp00_Line      dbIdent)
{
    m_LogInfoPage.EnterRegion(taskid);
    m_LogInfoPage.WriteDBIdentifier (dbIdent);
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::SetDeviceState (tsp00_TaskId    taskid,
                                     Log_DeviceState state)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::SetDeviceState", LogVolume_Trace, 5);

    m_LogInfoPage.EnterRegion(taskid);
    if ( m_LogInfoPage.ReadAutoOverwrite() // PTS 1120106 UH 2003-01-20
         &&
         Log_DeviceStateOkay == state )
        m_LogInfoPage.WriteDeviceState ( Log_DeviceStateHistoryLost );
    else
        m_LogInfoPage.WriteDeviceState ( state );
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);
    
    if ( state != Log_DeviceStateOkay )
        k38history_lost_write(taskid);
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::SetDevspaceEnabled (tsp00_TaskId    taskid,
                                         bool            enabled)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::SetDevspaceEnabled", LogVolume_Trace, 5);
    m_LogInfoPage.EnterRegion(taskid);
    m_LogInfoPage.WriteDevspaceEnabled ( enabled );
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);
}

// -----------------------------------------------------------------------------
bool Log_ArchiveLog::IsDevspaceEnabled ()
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::IsDevspaceEnabled", LogVolume_Trace, 5);
    return m_LogInfoPage.ReadDevspaceEnabled ();
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::SetAutomaticOverwrite (tsp00_TaskId    taskid,
                                            bool            enabled)
{
    // PTS 1120106 UH 2003-01-20 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::SetAutomaticOverwrite", LogVolume_Trace, 5);
    m_LogInfoPage.EnterRegion(taskid);
    m_LogInfoPage.WriteAutoOverwrite ( enabled );
    FlushLogInfoPage (taskid);
    m_LogInfoPage.LeaveRegion(taskid);
}

// -----------------------------------------------------------------------------
bool Log_ArchiveLog::IsAutomaticOverwrite ()
{
    // PTS 1120106 UH 2003-01-20 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::IsAutomaticOverwrite", LogVolume_Trace, 5);
    return m_LogInfoPage.ReadAutoOverwrite ();
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::UpdateIOSequenceNoAfterRedoLog (TaskID       taskid, //PTS 1113550 mb 2002-05-15
                                                     IOSequenceNo lastProcessedIOSequence)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::UpdateIOSequenceNoAfterRedoLog", LogVolume_Trace, 5);

    if (LogIsEmpty())
    {
        if (lastProcessedIOSequence.IsValid())
        {
            m_LogInfoPage.EnterRegion(taskid);
            m_LogInfoPage.WriteOldestNotSavedIOSequenceNo(lastProcessedIOSequence);
            m_LogInfoPage.WriteLastKnownIOSequenceNo(lastProcessedIOSequence);
            FlushLogInfoPage (taskid);
            m_LogInfoPage.LeaveRegion(taskid);
        }
        else
        // restore data without restore log and then a following restart
        // or init config
        {
            m_LogInfoPage.EnterRegion(taskid);
            m_LogInfoPage.WriteOldestNotSavedIOSequenceNo(k57restartrec->rstLastSavept_kb00().svpIOsequence_kb00);
            m_LogInfoPage.WriteLastKnownIOSequenceNo(k57restartrec->rstLastSavept_kb00().svpIOsequence_kb00);
            FlushLogInfoPage (taskid);
            m_LogInfoPage.LeaveRegion(taskid);
        }
    }
    // PTS 1117079 UH 2002-08-06
    // lastProcessedIOSequence is only invalid between InitConfig and ActivateServerDb.
    // But then the first logpage should not be written with an invalid ioseq.
    if ( lastProcessedIOSequence.IsValid() )
        m_LogWriter.SetIOSequenceNoAfterRestoreLog(lastProcessedIOSequence);
}
// -----------------------------------------------------------------------------
Log_ArchiveLog::PipeHandlingResult Log_ArchiveLog::PrepareFreeLogForPipe(
                const Log_IOSequenceNo     startIOSeq,
                const Log_IOSequenceNo     endIOSeq,
                const Log_RawDeviceOffset  startOffset,
                const Log_RawDeviceOffset  endOffset    )
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::PrepareFreeLogForPipe", LogVolume_Trace, 5);

    m_pipeSaveState.waitingForExecuteFreeLogForPipe = true;
    m_pipeSaveState.firstIOSeqSaved     = startIOSeq;
    m_pipeSaveState.lastIOSeqSaved      = endIOSeq;
    m_pipeSaveState.firstPageSaved      = startOffset;
    m_pipeSaveState.lastPageSaved       = endOffset;

    return pipeOk;
}     
// -----------------------------------------------------------------------------
Log_ArchiveLog::PipeHandlingResult Log_ArchiveLog::ExecuteFreeLogForPipe(
                const tsp00_TaskId         taskid,
                const Log_IOSequenceNo     startIOSeq,
                const Log_IOSequenceNo     endIOSeq)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ExecuteFreeLogForPipe", LogVolume_Trace, 5);

    if (!m_pipeSaveState.waitingForExecuteFreeLogForPipe)
    {
        // no call to PrepareFreeLogForPipe had been done before
        RTE_Message( Log_Exception(__FILE__,__LINE__,LOG_FREEPIPE_FAILED_PROT ));
        return lastSaveWasNot2Pipe;
    }
    else
    {
        if (   m_pipeSaveState.firstIOSeqSaved != startIOSeq
            || m_pipeSaveState.lastIOSeqSaved != endIOSeq)
        {
            RTE_Message( Log_Exception(__FILE__,__LINE__,
                   LOG_FREEPIPE_FAILED_IOSEQ,
                   SAPDB_ToString(m_pipeSaveState.firstIOSeqSaved,_T_d),
                   SAPDB_ToString(startIOSeq,_T_d),
                   SAPDB_ToString(m_pipeSaveState.lastIOSeqSaved,_T_d),
                   SAPDB_ToString(endIOSeq,_T_d) ));
            return IOSequenceDoesNotMatch;
        }
        else
        {
            // PTS 1127992 mb 2004-02-27 take care of wrap around for oldestNotSaved-calculation
            RawDeviceIterator newOldestNotSaveOffset( LogicalLogBeginOffset(),
                                                      m_pipeSaveState.lastPageSaved,
                                                      LogicalLogEndOffset());
            ++newOldestNotSaveOffset;
            if (FreeLog(taskid,
                        m_pipeSaveState.firstIOSeqSaved,
                        m_pipeSaveState.firstPageSaved,
                        ++m_pipeSaveState.lastIOSeqSaved,
                        newOldestNotSaveOffset.GetPosition()))
            {
                return pipeOk;
            }
            else
            {
                RTE_Message( Log_Exception(__FILE__,__LINE__,
                       LOG_FREEPIPE_FAILED_IOSEQ,
                       SAPDB_ToString(m_pipeSaveState.firstIOSeqSaved,_T_d),
                       SAPDB_ToString(startIOSeq,_T_d),
                       SAPDB_ToString(m_pipeSaveState.lastIOSeqSaved,_T_d),
                       SAPDB_ToString(endIOSeq,_T_d) ));
                 return IOSequenceDoesNotMatch;
            }
        }
    }
}     

// -----------------------------------------------------------------------------
void Log_ArchiveLog::CheckForLogSaveAndSavepoint (tgg00_TransContext &trans)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::CheckForLogSaveAndSavepoint", LogVolume_Trace, 5);
    m_LogWriter.CheckForLogSaveAndSavepoint(trans);
}

// -----------------------------------------------------------------------------
bool Log_ArchiveLog::SuspendAndGetLastWrittenIOSequence ( TaskID            taskid,
                                                          Log_IOSequenceNo &iosequence )
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::SuspendAndGetLastWrittenIOSequence", LogVolume_Trace, 5);
    return m_LogWriter.SuspendAndGetLastWrittenIOSequence(taskid, iosequence);
}

// -----------------------------------------------------------------------------
void Log_ArchiveLog::ResumeByUser()
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLog::ResumeByUser", LogVolume_Trace, 5);
    m_LogWriter.ResumeByUser();
}

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
