/*!*****************************************************************************

  module:       Log_ArchiveLog.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH
  
  special area: Archive log

  description:  Manages an archive log 

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_ARCHIVELOG_HPP
#define LOG_ARCHIVELOG_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // tsp00_TaskId

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "Container/Container_Vector.hpp"

#include "Logging/Log_Exceptions.hpp"
#include "Logging/Log_Queue.hpp"
#include "Logging/Log_Writer.hpp"
#include "Logging/Log_RawDeviceIterator.hpp"
#include "Logging/Log_InfoPage.hpp"
#include "Logging/Log_ArchiveLogIterator.hpp"
#include "Logging/Log_SeqLogDevReader.hpp"

#include "Logging/Log_IWritingRedoEntries.hpp"
#include "Logging/Log_IReadingRedoEntries.hpp"
#include "Logging/Log_IMonitoring.hpp"
#include "Logging/Log_IEventing.hpp"
#include "Logging/Log_ISave.hpp"

#include "IOManager/IOMan_ILogIO.hpp"

/*! --------------------------------------------------------------------------------
class:       Log_DeviceMonitorValues
description: This class is used by the ArchiveLog object to show some monitoring values.
*/
class Log_DeviceMonitorValues : public Log_MonitorValues
{
};
/*! endclass: Log_DeviceMonitorValues */



/*!*****************************************************************************

   class: Log_ArchiveLog

   description:
   
   Administrates a single archive log 

*******************************************************************************/
class Log_ArchiveLog
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : QueueRing
    description:  ring of log queues
    --------------------------------------------------------------------------*/
    typedef Log_Writer::QueueRing                               QueueRing;
    /*--------------------------------------------------------------------------
    declaration : QueueIterator
    description:  iterates from queue to queue
    --------------------------------------------------------------------------*/
    typedef QueueRing::Iterator                                 QueueIterator;
    /*!-------------------------------------------------------------------------
    declaration : QueueSize
    description:  number of entries in a log queue
    --------------------------------------------------------------------------*/
    typedef Log_Queue::EntryCount                               QueueSize;
    /*!-------------------------------------------------------------------------
    declaration : QueueSyncMode
    description:  queue synchronization mode 
                  (indicates whether queue requires sync. regions)
    --------------------------------------------------------------------------*/
    typedef Log_Queue::SynchronizationMode                      QueueSyncMode;
    /*!-------------------------------------------------------------------------
    declaration : QueueFlushMode
    description:  queue flush mode
    --------------------------------------------------------------------------*/
    typedef Log_FlushMode                                       QueueFlushMode;
    /*!-------------------------------------------------------------------------
    declaration:  IOSequenceNo
    description:  i/o sequence number of log pages
    --------------------------------------------------------------------------*/
    typedef Log_IOSequenceNo                                    IOSequenceNo;
    /*!-------------------------------------------------------------------------
    declaration : RawDeviceIterator
    description:  block iterator of raw log device
    --------------------------------------------------------------------------*/
    typedef Log_RawDeviceIterator                               RawDeviceIterator;
    /*!-------------------------------------------------------------------------
    declaration:  RawDeviceOffset
    description:  raw offset on log device
    --------------------------------------------------------------------------*/
    typedef Log_RawDeviceOffset                                 RawDeviceOffset;
    /*!-------------------------------------------------------------------------
    declaration : Log_DeviceBlockCount
    description:  number of blocks on a raw log device
    --------------------------------------------------------------------------*/
    typedef Log_DeviceBlockCount                                DeviceBlockCount;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  allocator for queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDBMem_IRawAllocator                              Allocator;
    /*!-------------------------------------------------------------------------
    declaration : FrameAllocator
    description:  allocator for page frames
    --------------------------------------------------------------------------*/
    typedef Log_FrameAllocator                                  FrameAllocator;
    /*!-------------------------------------------------------------------------
    declaration : TaskID
    description:  identifier of kernel task
    --------------------------------------------------------------------------*/
    typedef tsp00_TaskId                                        TaskID;
    /*!-------------------------------------------------------------------------
    declaration : RestartType
    description:  identifies the kind of restart
    --------------------------------------------------------------------------*/
    typedef enum RestartType  {
        Restart_normal,
        Restart_for_clear_log,
        Restart_for_restore_log
    } RestartType;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Log_ArchiveLog()
    description:  Constructor
    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task 
                  alloc        [in]  raw allocator to be used
                  syncmode     [in]  log queue synchronization mode
                  flushmode    [in]  log queue flush mode
    return value: none
    --------------------------------------------------------------------------*/
    Log_ArchiveLog 
        (TaskID                         taskid,
         TaskID                         writertaskid,
         Allocator&                     alloc,
         FrameAllocator&                framealloc,
         QueueSyncMode                  syncmode,
         QueueFlushMode                 flushmode)
        :
         m_Allocator      (alloc),
         m_FrameAlloc     (framealloc),
         m_WriterTaskSync (writertaskid),
         m_PageReader     (IOMan_ILogIO::GetInstance()),
         m_PageWriter     (IOMan_ILogIO::GetInstance()),
         m_LogQueues      (alloc),
         m_LogWriter      (alloc, m_LogQueues, IOMan_ILogIO::GetInstance(), m_WriterTaskSync),
         m_IOPages        (alloc, UTF8( "Log_ArchiveLog::WriteLogPages" )),
         m_QueueSyncMode  (syncmode),
         m_QueueFlushMode (flushmode),
         m_Allocated      (false),
         m_Restarted      (false)
    {
        m_pipeSaveState.waitingForExecuteFreeLogForPipe = false;
    }
      
    /*!-------------------------------------------------------------------------
    function:     CreateForNewDatabase()
    description:  Creates archive log for new database

                  - This method is called on 'init config'.
                  - It is assumed that this method is called only once within the
                    lifetime of a database.
                  - The archive log device is initialized between the given
                    limits; all log pages are cleared.
                  - The log info page is initialized with default values.
                  - Log queues and log writer are created.

    arguments:    taskid       [in]  id of the calling task 
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  savedist     [in]  distance between autolog saves in pages
                  queuecount   [in]  number of log queues to be created
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
    return value: (Log_ErrorState) 
                  Log_Ok,
                  Log_PageAllocationFailed,
                  Log_RawAllocationFailed,
                  Log_LogDeviceTooSmall,
                  Log_LogSaveDistanceTooLarge
    --------------------------------------------------------------------------*/
    Log_ErrorState CreateForNewDatabase
        (TaskID                         taskid,
         RawDeviceOffset                rawlogbegin,
         RawDeviceOffset                rawlogend,
         DeviceBlockCount               savedist,
         SAPDB_UInt                     queuecount,
         QueueSize                      queuesize,
         SAPDB_UInt                     ioblockcount);

    /*!-------------------------------------------------------------------------
    function:     Restart()
    description:  Restarts archive log for existing database

                  - This method is called on 'restart'.
                  - The actual log parameters are retrieved from the existing
                    log info page.
                  - Log queues and log writer are allocated if necessary.

    arguments:    taskid       [in]  id of the calling task 
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
    return value: (Log_ErrorState) 
    --------------------------------------------------------------------------*/
    Log_ErrorState Restart
        (TaskID                         taskid,
         RawDeviceOffset                rawlogbegin,
         RawDeviceOffset                rawlogend,
         DeviceBlockCount               autosaveDistance,
         QueueSize                      queuesize,
         SAPDB_UInt                     ioblockcount)
    {
        return RestartIntern ( taskid,
                               rawlogbegin,
                               rawlogend,
                               autosaveDistance,
                               queuesize,
                               ioblockcount,
                               Restart_normal);
    }
    
    /*--------------------------------------------------------------------------
    function:     ClearLogPartially()
    description:  Clears existing log device between given limits
                  PTS 1111525 mb 2002-01-24

    arguments:    taskid                    [in] id of calling task
                  firstOffsetToClear        [in] lowest offset on the device
                  lastOffsetToClear         [in] highest offset on the device
                  lastKnownIOSeqAfterClear  [in] the new value of lastKnownIOSeq
    return value: none
    --------------------------------------------------------------------------*/
    void ClearLogPartially
        (const TaskID                         taskid,
         const Log_RawDeviceOffset            firstOffsetToClear, 
         const Log_RawDeviceOffset            lastOffsetToClear,
         const Log_IOSequenceNo               lastKnownIOSeqAfterClear);

    /*--------------------------------------------------------------------------
    function:     ClearLogComplete()
    description:  Clears existing log device between given limits

    arguments:    taskid                     [in] id of calling task
                  lastKnownIOSeqAfterClear   [in] the new value of lastKnownIOSeq
    return value: none
    --------------------------------------------------------------------------*/
    void ClearLogComplete
        (const TaskID                         taskid,
         const Log_IOSequenceNo               lastKnownIOSeqAfterClear);

    /*--------------------------------------------------------------------------
    function:     ClearLogComplete()
    description:  Clears existing log device between given limits
                  PTS 1111525 mb 2002-01-24

    arguments:    taskid             [in] id of calling task
    return value: none
    --------------------------------------------------------------------------*/
    void ClearLogComplete
        (const TaskID                         taskid);

    /*--------------------------------------------------------------------------
    function:     ClearLogIntern()
    description:  Clears existing log device between given limits
                  PTS 1111525 mb 2002-01-29 new
    arguments:    taskid                [in]  id of calling task
                  firstOffsetToClear    [in]  offset of first page to be cleared
                  lastOffsetToClear     [in]  offset behind last page to be cleared 
                  lastKnownOffset       [in]  new value for lastKnownOffset in the LogInfoPage
                  oldestNotSavedOffset  [in]  new value for oldestNotSaved in the LogInfoPage
                  oldestNotSavedIOSeq   [in]  new value for oldestNotSavedIOSeq in the LogInfoPage
                  stateBeforeFormat     [in] the state, which will be set during formatLog
                  stateAfterFormat      [in] the state, which will be set during formatLog
    return value: none
    --------------------------------------------------------------------------*/
void ClearLogIntern (                      // PTS 1111525 mb 2002-01-29 new
        const TaskID                          taskid,
        const Log_RawDeviceOffset             firstOffsetToClear, 
        const Log_RawDeviceOffset             lastOffsetToClear,
        const Log_RawDeviceOffset             lastKnownOffset,
        const Log_RawDeviceOffset             oldestNotSavedOffset,
        const Log_IOSequenceNo                oldestNotSavedIOSeq,
        const Log_DeviceState                 stateBeforeFormat,
        const Log_DeviceState                 stateAfterFormat);
        
    /*--------------------------------------------------------------------------
    function:     RestartForClearLog()
    description:  Intialize all components for a subsequent call to ClearLog
                  PTS 1111525 mb 2002-01-24

    arguments:    taskid             [in] id of calling task
                  rawlogbegin        [in] lowest offset on the device
                  rawlogend          [in] highest offset on the device
                  autosaveDistance   [in] number of pages to be saved by autosave
                  queuesize          [in] number of entries in the queue
                  ioblockcount       [in] number of pages in the io-vector
    return value: Log_ErrorState
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartForClearLog
        (TaskID                         taskid,
         Log_RawDeviceOffset            rawlogbegin,
         Log_RawDeviceOffset            rawlogend,
         DeviceBlockCount               autosaveDistance,
         Log_Queue::EntryCount          queuesize,
         SAPDB_UInt                     ioblockcount);

    /*!-------------------------------------------------------------------------
    function:     RestartForRestoreLog()
    description:  Restarts archive log in order to perform a 'restore log'
    arguments:    utiltaskid [in]  id of calling utility task
    return value: none
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartForRestoreLog
        (TaskID                         taskid,
         RawDeviceOffset                rawlogbegin,
         RawDeviceOffset                rawlogend,
         DeviceBlockCount               autosaveDistance,
         QueueSize                      queuesize,
         SAPDB_UInt                     ioblockcount);

    /*!-------------------------------------------------------------------------
    function:     RestartForSaveLogCold()
    description:  Restarts archive log in order to perform a 'save log' 
                  on cold database
    
    arguments:    taskid       [in]  id of the calling task 
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device

    return value: (Log_ErrorState) 
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartForSaveLogCold 
    (TaskID                         taskid,
     RawDeviceOffset                rawlogbegin,
     RawDeviceOffset                rawlogen,
     DeviceBlockCount               autosaveDistance);
    /*!-------------------------------------------------------------------------
    function:     IsRestarted()
    description:  returns true, if the logdevice is ready for writing
    arguments:    none
    return value: (bool) 
    --------------------------------------------------------------------------*/
	bool IsRestarted()
	{
		return m_Restarted;
	}
    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes log queues and log writer
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Delete()
    {
        
        if (m_LogInfoPage.IsAssigned())
        {
            Kernel_Page::PageFrame frame = m_LogInfoPage.GetFrame();
            m_FrameAlloc.Free(frame);
            m_LogInfoPage.ReleaseFrame();
        }

        if (m_BufferedPage.IsAssigned())
        {
            Kernel_Page::PageFrame frame = m_BufferedPage.GetFrame();
            m_FrameAlloc.Free(frame);
            m_BufferedPage.ReleaseFrame();
        }

        if (m_LogQueues.GetSize() > 0)
        {
            QueueIterator iter = m_LogQueues.Anchor();
            QueueIterator end  = m_LogQueues.Anchor();
            do
            {
                destroy(*iter, m_Allocator);
                ++iter;
            }
            while (iter != end);
        }

        m_IOPages.Delete();
        m_LogQueues.Delete();
        m_LogWriter.Delete();
    }

    /*!-------------------------------------------------------------------------
    function:     ~Log_ArchiveLog()
    description:  Destructor
    return value: none
    --------------------------------------------------------------------------*/
    ~Log_ArchiveLog()
    {
        Delete();
    }
    /*!-------------------------------------------------------------------------
    function:     UpdateIOSequenceNoAfterRedoLog()
    description:  This must be used, to set the IOSequence correctly after a
                  restart/restore, because at the time, when UpdateLogInfoPage
                  is called, nothing is known about how many IOSequences are
                  Redone
    arguments:    taskId                  [in]   the taskid of the calling task
                  lastProcessedIOSequence [in]   the first IOSequenceNo after the restart
    --------------------------------------------------------------------------*/
    void UpdateIOSequenceNoAfterRedoLog (TaskID       taskid, //PTS 1113550 mb 2002-05-15
                                         IOSequenceNo lastProcessedIOSequence);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterator support                                             */
    /*!-------------------------------------------------------------------------
    function:     LogicalLogBeginOffset()
    description:  Returns lowest offset on raw log device to be used for
                  cyclical log write i/o

                  - Offset is located behind the leading log info
                    page block(s) on the raw device.
                    
    arguments:    none
    return value: (RawDeviceOffset)
    --------------------------------------------------------------------------*/
    RawDeviceOffset LogicalLogBeginOffset() const
    {
        return m_RawDeviceBegin + (RawDeviceOffset)LOG_INFOBLOCKCOUNT;
    }

    /*!-------------------------------------------------------------------------
    function:     LogicalLogEndOffset()
    description:  Returns highest offset on raw log device to be used for 
                  cyclical log write i/o
    arguments:    none
    return value: (RawDeviceOffset)
    --------------------------------------------------------------------------*/
    RawDeviceOffset LogicalLogEndOffset() const
    {
        return m_RawDeviceEnd;
    }

    /*!-------------------------------------------------------------------------
    function:     LogInfoPageOffset()
    description:  Returns offset of log info page on raw device
    arguments:    none
    return value: (RawDeviceOffset)
    --------------------------------------------------------------------------*/
    RawDeviceOffset LogInfoPageOffset() const
    {
        return m_RawDeviceBegin + (RawDeviceOffset)LOG_INFOPAGEOFFSET;
    }

    /*!-------------------------------------------------------------------------
    function:     SaveBegin()
    description:  Returns iterator at position where next save must start

                  - The returned iterator is positioned at the raw device offset
                    of the oldest not saved log page.
    arguments:    none
    return value: (RawDeviceIterator)
    --------------------------------------------------------------------------*/
    RawDeviceIterator SaveBegin() const
    {
        return RawDeviceIterator
                    (LogicalLogBeginOffset(), 
                     m_LogInfoPage.ReadOldestNotSavedOffset(), 
                     LogicalLogEndOffset());
    }

    /*!-------------------------------------------------------------------------
    function:     SaveEnd()
    description:  Returns iterator at position where next save ends

                  - The returned iterator is positioned at the raw device offset
                    where the next save ends.
                  - The parameter 'maxsavesize' indicates the maximum requested
                    save size in pages. If less than 'maxsavesize' pages are
                    available for save, the iterator is positioned at the 
                    youngest page available for save.
    arguments:    maxsavesize [in]  maximum number of pages requested for save
    return value: (RawDeviceIterator)
    --------------------------------------------------------------------------*/
    RawDeviceIterator SaveEnd
        (Log_DeviceBlockCount           maxsavesize) const
    {
        IOSequenceNo writerSeqNo         = m_LogWriter.GetIOSequenceNo();
        IOSequenceNo oldestNotSavedSeqNo = m_LogInfoPage.ReadOldestNotSavedIOSequenceNo();
        SAPDB_UInt   dist                = writerSeqNo - oldestNotSavedSeqNo;

        RawDeviceIterator iter
                    (LogicalLogBeginOffset(), 
                     m_LogInfoPage.ReadOldestNotSavedOffset(), 
                     LogicalLogEndOffset());

        iter += (maxsavesize <= dist) ? maxsavesize : dist;
        return iter;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterator support                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Log_ISave interface                                          */
    /* -----------------------------------------------------------------------*/
    /*!-------------------------------------------------------------------------
    function:    GetNumberOfUnsavedPages()
    description: Returns the number of log pages that are not saved
    returns:     (DeviceBlockCount) 
    --------------------------------------------------------------------------*/
    DeviceBlockCount GetNumberOfUnsavedPages() const;
    /*!-------------------------------------------------------------------------
    declaration: SaveInfo
    description:
        okay,
        noLogToSave,
        historyIsLost,
        incompleteSegment
    --------------------------------------------------------------------------*/
    enum SaveInfo
    {
        okay,
        noLogToSave,
        historyIsLost,
        incompleteSegment
    };
    /*!-------------------------------------------------------------------------
    function:     	GetSaveLogIterator()
    description:  	returns an iterator for saving the log


    arguments:    	taskid 		[in]  	id of the calling task 
					Allocator	[in]	Allocator from what pool the memory
										for the iterator shall be located
					IsCold				Determines whether the last written
										page can be also saved or not
                    SaveInfo [out]      okay, noLogToSave, historyIsLost, incompleteSegment
                    iterator [out]      
    --------------------------------------------------------------------------*/
    void GetSaveLogIterator( tsp00_TaskId    		TaskId,
                             SAPDBMem_IRawAllocator &Allocator,
                             bool              		 IsCold,
                             bool              		 ForAutosave,
                             bool              		 completeLog,
                             Log_DeviceBlockCount	 blockSize,
                             Log_SaveIterator*      &iterator,
                             SaveInfo               &saveInfo);

    /*!-------------------------------------------------------------------------
    function:     	GetNextBackupCount()
    description:  	returns a new backup-count for numbering the log_save-files
                    This function increments the counter in the log_info-page
                    of the device and writes the log_info-page to disk

    arguments:    	taskid 		[in]  	id of the calling task
    return value:	next free log-count 	
    --------------------------------------------------------------------------*/
    SAPDB_UInt4    GetNextBackupCount(tsp00_TaskId    			TaskId);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Log_ISave interface                                       */
    /* -----------------------------------------------------------------------*/




    /* -----------------------------------------------------------------------*/
    /*! chapter: Free Log Pages                                               */
    /* -----------------------------------------------------------------------*/

public:
    /*!-------------------------------------------------------------------------
    function:     FreeLog()
    description:  This releases log pages, which are already backuped.
    arguments:    taskid                 [in]
                  oldLastSavedIOsequence [in]
                  oldLastSavedOffset     [in]
                  newLastSavedIOsequence [in]
                  newLastSavedOffset     [in]
    return value: (bool) true, if log could be freed
    --------------------------------------------------------------------------*/
    bool FreeLog (tsp00_TaskId        taskid, 
                  Log_IOSequenceNo    oldLastSavedIOsequence,
                  Log_RawDeviceOffset oldLastSavedOffset,
                  Log_IOSequenceNo    newLastSavedIOsequence,
                  Log_RawDeviceOffset newLastSavedOffset);
    /*!-------------------------------------------------------------------------
    function:     TruncateLogPage()
    description:  This releases entries in a logpage up to a specified IOSequence 
                  Attention: This call is not synchronized with other accesses
                  to m_bufferedPage or the logVolume.
    arguments:    newLastIOSequence      [in]
                  pageOffset             [in]
    return value: none
    --------------------------------------------------------------------------*/
    void TruncateLogPage (      TaskID               taskid,
                          const Log_IOSequenceNo     newLastIOSequence,
                          const Log_RawDeviceOffset  pageOffset);
    /*!-------------------------------------------------------------------------
    declaration: PipeHandlingResult
    description:                   result set for operations which has to be done additionally
                                   for saving into a pipe:
        
        logFreed                   the log from a previous save into a pipe has successfully
                                   freed
        IOSequenceDoesNotMatch     the IOSequence of the free command does not
                                   match to the last saved segment 
        lastSaveWasNot2Pipe        the last saved segment was no pipe
    --------------------------------------------------------------------------*/
    enum PipeHandlingResult
    {
        pipeOk,
        IOSequenceDoesNotMatch,
        lastSaveWasNot2Pipe,
        noFreeLogCalledInbetween
    };
    /*!-------------------------------------------------------------------------
    function:     	PrepareFreeLogForPipe()
    description:  	saves information about the saved log_segment for
                    a subsequent call to ExecuteFreeLogForPipe
    arguments:      firstIOSeqSaved     lowest IOSequence of the first saved page
                    lastIOSeqSaved      highest IOSequence of the first saved page
                    firstPageSaved      offset of the first saved page
                    lastPageSaved       offset of the last saved page
    return value:   PipeHandlingResult
    --------------------------------------------------------------------------*/
    PipeHandlingResult PrepareFreeLogForPipe(
                        const Log_IOSequenceNo     firstIOSeqSaved,
                        const Log_IOSequenceNo     lastIOSeqSaved,
                        const Log_RawDeviceOffset  firstPageSaved,
                        const Log_RawDeviceOffset  lastPageSaved    );

    /*!-------------------------------------------------------------------------
    function:     	ExecuteFreeLogForPipe()
    description:  	frees the log which had been saved by the previous
                    "save log into pipe". It is checked, that
                    the given IOSequences matches the saved values from
                    the previous PrepareFreeLogForPipe
    arguments:      taskId        id of the calling task
                    startIOSeq    lowest IOSequence of the first freed page
                    endIOSeq      highest IOSequence of the first freed page
    return value:   PipeHandlingResult
    --------------------------------------------------------------------------*/
    PipeHandlingResult ExecuteFreeLogForPipe(
                        const tsp00_TaskId         taskid,
                        const Log_IOSequenceNo     startIOSeq,
                        const Log_IOSequenceNo     endIOSeq);
private:
    /*--------------------------------------------------------------------------
    declaration : PipeSaveState
    description:  summarizes all information for the handling of saving into pipes
    --------------------------------------------------------------------------*/
    struct  PipeSaveState
    {
        bool                 waitingForExecuteFreeLogForPipe;
        Log_IOSequenceNo     firstIOSeqSaved;
        Log_IOSequenceNo     lastIOSeqSaved;
        Log_RawDeviceOffset  firstPageSaved;
        Log_RawDeviceOffset  lastPageSaved; 
    };
    
    /*--------------------------------------------------------------------------
    declaration : m_pipeSaveState
    description:  state of the last save of the log (if it was a pipe)
    --------------------------------------------------------------------------*/
    PipeSaveState m_pipeSaveState;
        
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Free Log Pages                                          */
    /* -----------------------------------------------------------------------*/

public:
    /* -----------------------------------------------------------------------*/
    /*! chapter: Device status                                                */
    /*!-------------------------------------------------------------------------
    function:     LogIsFull()
    description:  Checks whether no more log can be written

                  - Not valid for utility.

    arguments:    taskid [in]  id of the calling task 
    return value: (SAPDB_Bool) true if full
    --------------------------------------------------------------------------*/
    SAPDB_Bool LogIsFull 
        (TaskID                         taskid)
    {
        return m_LogWriter.LogIsFull();
    }

    /*!-------------------------------------------------------------------------
    function:     LogIsEmpty()
    description:  Checks whether nothing was written to the log device.
    arguments:    none
    return value: (bool) returns true if nothing was written to the log device.
    --------------------------------------------------------------------------*/
    bool LogIsEmpty ()
    {
        SAPDBERR_ASSERT_STATE (m_LogInfoPage.IsAssigned());
        SAPDBERR_ASSERT_STATE (Log_DeviceStateClearing  != m_LogInfoPage.ReadDeviceState());
        SAPDBERR_ASSERT_STATE (Log_DeviceStatePartiallyClearing  != m_LogInfoPage.ReadDeviceState());

        return m_LogInfoPage.ReadLastKnownOffset().IsInvalid();  //PTS 1113550 mb 2002-05-2
    }

    /*! --------------------------------------------------------------------------------
    function:    GetDeviceSize
    description: returns the size of the archive log device
    returns:     DeviceBlockCount
    */
    DeviceBlockCount GetDeviceSize ()
    {
        return m_RawDeviceEnd - m_RawDeviceBegin;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetDeviceUsage
    */
    void GetDeviceUsage ( SAPDB_UInt &deviceSize,
                          SAPDB_UInt &pagesUsed,
                          SAPDB_UInt &pagesNotSaved );

    /*! --------------------------------------------------------------------------------
    function:    GetQueueUsage
    */
    void GetQueueUsage ( SAPDB_UInt4    &queueSize,
                         SAPDB_UInt4    &maxPagesUsedInQueue,
                         SAPDB_UInt4    &countReserveRequests,
                         SAPDB_UInt4    &countWaitForIO,
                         SAPDB_UInt4    &countGroupCommit,
                         SAPDB_UInt4    &countWaitForQueueFree,
                         SAPDB_UInt4    &maxWaitForPageIO );

    /*! --------------------------------------------------------------------------------
    function:    GetRestartVersion
    */
    void GetRestartVersion ( tsp00_TaskId      taskid,
                             Log_IOSequenceNo &firstiosequence,
                             SAPDB_Char       *dbident );

    /*! --------------------------------------------------------------------------------
    function:    GetLogDBIdent
    */
    void GetLogDBIdent ( SAPDB_Char       *dbident );

    /*! --------------------------------------------------------------------------------
    function:    GetDeviceState
    */
    const Log_DeviceState GetDeviceState () const;

    /*! --------------------------------------------------------------------------------
    function:    ResetStatistics
    */
    void ResetStatistics (tsp00_TaskId taskid);
    
    /*! --------------------------------------------------------------------------------
    function:    SetLogDBIdent
    */
    void SetLogDBIdent ( tsp00_TaskId    taskid,
                         tsp00_Line      dbIdent);

    /*! --------------------------------------------------------------------------------
    function:    SetDeviceState
    */
    void SetDeviceState (tsp00_TaskId    taskid,
                         Log_DeviceState state);

    /*! --------------------------------------------------------------------------------
    function:    SetDevspaceEnabled
    */
    void SetDevspaceEnabled (tsp00_TaskId    taskid,
                             bool            enabled);

    /*! --------------------------------------------------------------------------------
    function:    IsDevspaceEnabled
    */
    bool IsDevspaceEnabled ();

    /*! --------------------------------------------------------------------------------
    function:    SetAutomaticOverwrite
    */
    void SetAutomaticOverwrite (tsp00_TaskId    taskid,
                                bool            enabled);

    /*! --------------------------------------------------------------------------------
    function:    IsAutomaticOverwrite
    */
    bool IsAutomaticOverwrite ();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Device status                                             */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetQueue()
    description:  Returns log queue instance that serves calling user task

    arguments:    usertaskid [in]  id of the calling user task
    return value: (Log_Queue&) reference to queue
    --------------------------------------------------------------------------*/
    Log_Queue& GetQueue
        (TaskID                         usertaskid)
    {
        SAPDBERR_ASSERT_STATE(m_LogQueues.GetSize() > 0);
        Log_QueueID quid = usertaskid % m_LogQueues.GetSize();

        QueueIterator iter = m_LogQueues.Anchor();
        while ((**iter).GetID() != quid) 
            ++iter;

        return **iter;
    }

    /*!-------------------------------------------------------------------------
    function:     GetWriter()
    description:  Returns log writer instance

    arguments:    none
    return value: (Log_Writer&) reference to writer
    --------------------------------------------------------------------------*/
    Log_Writer& GetWriter()
    {
        return m_LogWriter;
    }

    /*!-------------------------------------------------------------------------
    function:     GetWriterTaskID()
    description:  Returns id of associated writer task
    arguments:    none
    return value: (TaskID)
    --------------------------------------------------------------------------*/
    TaskID GetWriterTaskID() const
    {
        return m_WriterTaskSync.GetTaskID();
    }

    /*!-------------------------------------------------------------------------
    function:     GetLogInfo()
    returns:      false, if logging is not initialized
    arguments:    all log info page values [out]
    --------------------------------------------------------------------------*/
    bool GetLogInfo ( tsp00_TaskId              taskid,
                      Kernel_Date              &date,
                      Kernel_Time              &time,
                      Log_FlushMode            &flushmode,
                      Log_DeviceState          &devicestate,
                      Log_InfoPage::QueueCount &queuecount,
                      Log_RawDeviceOffset      &cycleendoffset,
                      Log_RawDeviceOffset      &oldestnotsavedoffset,
                      Log_IOSequenceNo         &oldestnotsavedsequence,
                      Log_RawDeviceOffset      &lastknownoffset,
                      Log_IOSequenceNo         &lastknownsequence,
                      SAPDB_UInt4              &backupcount,
                      Kernel_Date              &untildate,
                      Kernel_Time              &untiltime,
                      Log_RawDeviceOffset      &untilclearloweroffset,
                      Log_RawDeviceOffset      &untilclearupperoffset,
                      Log_IOSequenceNo         &untilsequenceafterredo,
                      Log_IOSequenceNo         &savepointsequence,
                      Log_RawDeviceOffset      &savepointdevoffset,
                      Log_InfoPage::PageOffset &savepointoffset );

    /* -----------------------------------------------------------------------*/
    /*! chapter: read after images                                            */
    /* -----------------------------------------------------------------------*/
    
    /*!-------------------------------------------------------------------------
    function:     GetArchiveLogReader()
    description:  This offers an iterator, to access afterimages.
    arguments:    allocator    [in] 
    return value: Log_ISequentialReader
    --------------------------------------------------------------------------*/
    Log_ISequentialReader *GetArchiveLogReader(SAPDBMem_IRawAllocator    &alloc)
    {
        return  new(alloc) 
                      Log_SeqLogDevReader(  LogicalLogBeginOffset(),
                                            LogicalLogEndOffset(),
                                            m_PageReader);
    }
    /*!-------------------------------------------------------------------------
    function:     GetLastKnownIOSeq()
    description:  Returns the last knonw IOSeq known on the LogDevice
    arguments:    none
    return value: Log_IOSequenceNo
    --------------------------------------------------------------------------*/
    Log_IOSequenceNo GetLastKnownIOSeq()
    {
        return  m_LogInfoPage.ReadLastKnownIOSequenceNo();
    }

    /*!-------------------------------------------------------------------------
    function:     GetArchiveLogIterator()
    description:  This offers an iterator, to access afterimages.
    arguments:    TaskId    [in] 
    return value: Log_ArchiveLogIterator
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator GetArchiveLogIterator (
                SAPDBMem_IRawAllocator    &alloc,
                tsp00_TaskId               taskId,
          const Kernel_Date                untilDate,    // PTS 1111525 mb 2001-01-15
          const Kernel_Time                untilTime,    // PTS 1111525 mb 2001-01-15
          const Log_IOSequenceNo           untilIOSeq)   // PTS 1121539 mb 2003-02-06
    const;
                                                  
    /*!-------------------------------------------------------------------------
    function:     GetTapeIterator()
    description:  This offers an iterator, to access afterimages.
    arguments:    alloc  	        [in]      Allocator for memory allocation
                  readerTrans       [in|out]  transactioncontext
                  lastKnownOnTape   [in]      Last IOSequence to be read from tape
    return value: Log_ArchiveLogIterator
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator GetTapeIterator (
                SAPDBMem_IRawAllocator     &alloc,
                tgg00_TransContext         &readerTrans,
                Log_IOSequenceNo            lastKnownOnTape,
          const Kernel_Date                 untilDate,    // PTS 1111525 mb 2001-01-15
          const Kernel_Time                 untilTime,    // PTS 1111525 mb 2001-01-15
          const Log_IOSequenceNo           untilIOSeq)    // PTS 1121539 mb 2003-02-06
    const;

    /*!-------------------------------------------------------------------------
    function:     SetToLastSavepointEntry()
    description:  The iterator is set to the last savepoint entry,
                  of the restart record.
    arguments:    iter [in/out]
    return value: none
    --------------------------------------------------------------------------*/
    void SetToLastSavepointEntry (Log_ArchiveLogIterator &iter,
                                  IOSequenceNo           &svpSeq);
    /*!-------------------------------------------------------------------------
    function:     SetToBeginOfPage()
    description:  The iterator is set to the begin of the page denoted by offset
    arguments:    iter [in]
                  offset [in]
    return value: none
    --------------------------------------------------------------------------*/
    void SetToBeginOfPage (Log_ArchiveLogIterator    &iter,
                           Log_RawDeviceOffset        offset);
    /* -----------------------------------------------------------------------*/
    /*! endchapter: read after images                                         */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: savepoint                                                    */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     WriteSavepointValuesAndFlushInfoPage()
    description:  The given bytes from the restartrecord are copied into the
                  infopage. The loginfo page is written to disk.
    arguments:    taskid [in]
                  reason [in]
                  rrpart [in]
    return value: none
    --------------------------------------------------------------------------*/
    void WriteSavepointValuesAndFlushInfoPage (TaskID                   taskid,
                                               Log_SavepointReason      reason,
                                               const SAPDBFields_Field& rrpart);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: savepoint                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     WriteDeviceParametersToTrace()
    description:  The parameters describing the log device are written to the trace.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void WriteDeviceParametersToTrace ( );
    /*!-------------------------------------------------------------------------
    function:     FlushLogInfoPage()
    description:  This write the internal info page to the devspace.
                  Date and time are updated.
    arguments:    taskid [in]
    return value: none
    --------------------------------------------------------------------------*/
    void FlushLogInfoPage (tsp00_TaskId taskid);
    /*!-------------------------------------------------------------------------
    function:     UpdateInfoPageFromWriter()
    description:  This is used by the writer to update the log info page.
    arguments:    writertaskid      [in]
                  lastknownsequence [in]
                  lastknownoffset   [in]
    return value: none
    --------------------------------------------------------------------------*/
    void UpdateInfoPageFromWriter (tsp00_TaskId        writertaskid,
                                   Log_IOSequenceNo    lastKnownSequence,
                                   Log_RawDeviceOffset lastKnownOffset);
    /*!-------------------------------------------------------------------------
    function:     ChangeEndOffsetAfterAddDevSpace()
    description:  This is used by the writer if the first offset of the added  
                  log device is to be written.
    arguments:    writertaskid   [in]
                  newEndOffset   [in]
    return value: none
    --------------------------------------------------------------------------*/
    void ChangeEndOffsetAfterAddDevSpace (tsp00_TaskId        writertaskid,
                                          Log_RawDeviceOffset newEndOffset);
    /*!-------------------------------------------------------------------------
    function:     ScanLogForFirstKnown()
    description:  Sets the value of the smallest IOSequence/Offset
                  on the LogVolume by scanning the LogDevice and
                  stores it for later retrieval by GetFirstKnownBeforeRestoreLog.
                  Must not be called
                  after Restart
    arguments:    taskid         [in]
    return value: void
    --------------------------------------------------------------------------*/
    void ScanLogForFirstKnown (TaskID           taskid);

    /*!-------------------------------------------------------------------------
    function:     GetFirstKnownBeforeRestore()
    description:  This method is used by restartForRestore to
                  set the only at this time available information about the
                  first known offset and IOSequence
    arguments:    writertaskid   [in]
                  newEndOffset   [in]
    return value: bool
    --------------------------------------------------------------------------*/
    void GetFirstKnownBeforeRestoreLog (RawDeviceOffset     &firstKnownOffset,
                                        IOSequenceNo        &firstKnownIOSeq) 
    {
        firstKnownOffset = m_FirstKnownOffset;
        firstKnownIOSeq  = m_FirstKnownIOSeq;
    }

    /*!-------------------------------------------------------------------------
    function:     SetNewDeviceEndOffset
    description:  This method is used by ADD LOG DEVSPACE
    arguments:    taskid       [in]
                  newEndOffset [in]
    --------------------------------------------------------------------------*/
    void SetNewDeviceEndOffset ( SAPDB_UInt4 newEndOffset );

    /*!-------------------------------------------------------------------------
    function:     GetAutosaveDistance
    description:  This returns the distance in pages.
    returns:      distance [pages]
    --------------------------------------------------------------------------*/
    DeviceBlockCount GetAutosaveDistance()
    {
        return m_autosaveDistance;    // PTS 1113361 mb 2002-01-11
    }

    /*--------------------------------------------------------------------------
    function:     DiagnoseOpenDevice()
    description:  This opens the log device.
    arguments:    taskid [in]
                  open
    return value: bool - true, if device is open
    --------------------------------------------------------------------------*/
    bool DiagnoseOpenDevice (TaskID taskid);

    /*--------------------------------------------------------------------------
    function:     DiagnoseCloseDevice()
    description:  This closes the log device.
    arguments:    taskid [in]
    return value: none
    --------------------------------------------------------------------------*/
    void DiagnoseCloseDevice (TaskID taskid);
    
    /*--------------------------------------------------------------------------
    function:     DiagnoseOpenInfoPage()
    description:  This opens the log device and prepares the InfoPage for reading.
    arguments:    taskid     [in]
                  openDevice [in]
    return value: bool - true, if info page can be accessed.
    --------------------------------------------------------------------------*/
    bool DiagnoseOpenInfoPage (TaskID taskid,
                               bool   openDevice);

    /*--------------------------------------------------------------------------
    function:     CheckForLogSaveAndSavepoint()
    description:  This is called by the timeout task.
                  It is checked if an autosave should be triggered or a savepoint.
    arguments:    trans [in/out]
    return value: none
    --------------------------------------------------------------------------*/
    void CheckForLogSaveAndSavepoint (tgg00_TransContext &trans);

    /*--------------------------------------------------------------------------
    function:     SuspendAndGetLastWrittenIOSequence
    arguments:    taskid     [in]
                  iosequence [out]
    returns:      Returns false if this is called more then by one task.
    description:  The logwriter is suspended until Resume() is called.
                  The returned iosequence is the last used iosequence.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    bool SuspendAndGetLastWrittenIOSequence ( TaskID            taskid,
                                              Log_IOSequenceNo &iosequence );

    /*--------------------------------------------------------------------------
    function:     ResumeByUser
    arguments:    none
    description:  The logwriter is resumed after suspended by
                  SuspendAndGetLastWrittenIOSequence.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    void ResumeByUser ();

private:

    /*!-------------------------------------------------------------------------
    function:     RestartIntern()
    arguments:    taskid       [in]  id of the calling task 
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
                  forClearLog  [in] 
    return value: (Log_ErrorState) 
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartIntern
        (TaskID                         taskid,
         RawDeviceOffset                rawlogbegin,
         RawDeviceOffset                rawlogend,
         DeviceBlockCount               autosaveDistance,
         QueueSize                      queuesize,
         SAPDB_UInt                     ioblockcount,
         RestartType                    restartType);
    
    /*--------------------------------------------------------------------------
    function:     UpdateLogInfoPage()
    description:  Determines restart parameters for the log writer by scanning 
                  the log device and the log info page.

    arguments:    taskid [in]   id of calling task
                  firstKnownSeqNo  [out]
                  firstKnownOffset [out]
    return value: (Log_ErrorState) 
    --------------------------------------------------------------------------*/
    Log_ErrorState UpdateLogInfoPage (TaskID           taskid);

    /*--------------------------------------------------------------------------
    function:     CheckLogAndDataCompatibility()
    description:  checks the Restartrecord for consistency with the actual log

    arguments:    taskid [in]   id of calling task
    return value: (Log_ErrorState) 
    --------------------------------------------------------------------------*/
    Log_ErrorState CheckLogAndDataCompatibility
                                    (TaskID           taskid );


                                      
    /*--------------------------------------------------------------------------
    function:     CheckOffsetAndIOsequence()
    description:  Returns true, if the given ioseqno is found on the page found
                  on the given offset.

    arguments:    taskid [in]   id of calling task
                  offset [in]
                  seqno  [in]
    return value: (SAPDB_Bool) 
    --------------------------------------------------------------------------*/
    SAPDB_Bool CheckOffsetAndIOsequence (TaskID          taskid,
                                         RawDeviceOffset offset,
                                         IOSequenceNo    seqno);
                                         
    /*--------------------------------------------------------------------------
    function:     AllocatePageFrames()
    description:  Allocates page frames for info page and page buffer
    return value: true if allocated
    --------------------------------------------------------------------------*/
    SAPDB_Bool AllocatePageFrames();

    /*--------------------------------------------------------------------------
    function:     CreateInfoPage()
    description:  Creates info page with default values
    arguments:    savedist [in]  log save distance
    --------------------------------------------------------------------------*/
    void CreateInfoPage
        (DeviceBlockCount               savedist);

    /*--------------------------------------------------------------------------
    function:     CreateLogQueues()
    description:  Allocates and initializes log queues
    arguments:    queuecount   [in]  number of log queues to be created
                  queuesize    [in]  number of log pages to be allocated for each queue
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool CreateLogQueues
        (SAPDB_UInt                     queuecount,
         QueueSize                      queuesize);

    /*--------------------------------------------------------------------------
    function:     InitLogWriter()
    description:  Initializes log writer
    arguments:    offset         [in] to start writing
                  seqno          [in] to use for next log page
                  ioblockcount   [in] max. count of i/o blocks for multiple write i/o
                  totalQueueSize [in] number of all pages in all queues of this writer
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool InitLogWriter ( RawDeviceOffset offset,
                               IOSequenceNo    seqno,
                               SAPDB_UInt      ioblockcount);

    /*--------------------------------------------------------------------------
    function:     ResetLogWriter()
    description:  Initializes log writer
    arguments:    taskid [in]
                  return value: none
    --------------------------------------------------------------------------*/
    void ResetLogWriter (TaskID          taskid);

    /*--------------------------------------------------------------------------
    function:     AfterSaveLog()
    description:  resumes the waiting log device writer, if necessary
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void AfterSaveLog()
    {
        m_WriterTaskSync.Resume();
    }

    /*--------------------------------------------------------------------------
    function:     FormatLog()
    description:  Clears existing log device between given limits

                  - 'beginoffs' may be greater than 'endoffs' (cycle).
    arguments:    taskid      [in]  id of calling task
                  beginOffset [in]  offset of first page to be cleared
                  endOffset   [in]  offset behind last page to be cleared 
    return value: none
    --------------------------------------------------------------------------*/
    void FormatLog
        (const TaskID                         taskid,
         const RawDeviceOffset                beginOffset,
         const RawDeviceOffset                endOffset);

    /*--------------------------------------------------------------------------
    function:     FormatLogLinear()
    description:  Clears existing log device between given limits

                  - 'beginoffs' MUST NOT be greater than 'endoffs' (cycle).
    arguments:    taskid    [in]  id of calling task
                  beginoffs [in]  offset of first page to be cleared
                  endoffs   [in]  offset behind last page to be cleared 
    return value: none
    --------------------------------------------------------------------------*/
    void FormatLogLinear
        (const TaskID                         taskid,
         const RawDeviceOffset                beginoffs,
         const RawDeviceOffset                endoffs);

    /*--------------------------------------------------------------------------
    function:     LogIsFresh()
    description:  Checks whether the log for this database is fresh

                  - This method returns true for new databases for which no
                    log has ever been written to log device.
                  - The method requires an initialized log info page to be
                    available.

    return value: (SAPDB_Bool) true if fresh
    --------------------------------------------------------------------------*/
    SAPDB_Bool LogIsFresh() const
    {
        IOSequenceNo lastKnownSeqNo = m_LogInfoPage.ReadLastKnownIOSequenceNo();
        return lastKnownSeqNo.IsInvalid();
    }
    
    /*--------------------------------------------------------------------------
    function:     ClearLogFullQueueState()
    description:  Resets the LogFull state of the log_queues if the log is no
                  longer full
    return value: none
    --------------------------------------------------------------------------*/
    void ClearLogFullQueueState(TaskID      taskid);

private:

    /*--------------------------------------------------------------------------
    declaration: m_Allocator
    description: raw allocator 
    --------------------------------------------------------------------------*/
    Allocator&                                          m_Allocator;
    /*--------------------------------------------------------------------------
    declaration : m_FrameAlloc
    description:  allocator for page frames
    --------------------------------------------------------------------------*/
    FrameAllocator&                                     m_FrameAlloc;
    /*--------------------------------------------------------------------------
    declaration: m_WriterTaskSync
    description: log writer task synchronization object
    --------------------------------------------------------------------------*/
    Log_WriterTaskSynchronizer                          m_WriterTaskSync;
    /*--------------------------------------------------------------------------
    declaration: m_LogQueues
    description: circular list of log queues of this device
    --------------------------------------------------------------------------*/
    QueueRing                                           m_LogQueues;
    /*--------------------------------------------------------------------------
    declaration: m_QueueSyncMode
    description: log queue synchronization mode
    --------------------------------------------------------------------------*/
    QueueSyncMode                                       m_QueueSyncMode;
    /*--------------------------------------------------------------------------
    declaration: m_QueueFlushMode
    description: log writer flush mode
    --------------------------------------------------------------------------*/
    QueueFlushMode                                      m_QueueFlushMode;
    /*--------------------------------------------------------------------------
    declaration: m_LogWriter
    description: log writer instance of this log device
    --------------------------------------------------------------------------*/
    Log_Writer                                          m_LogWriter;
    /*--------------------------------------------------------------------------
    declaration: m_PageReader
    description: reader for raw log device
    --------------------------------------------------------------------------*/
    IOMan_ILogIO&                                       m_PageReader;
    /*--------------------------------------------------------------------------
    declaration: m_PageWriter
    description: writer for raw log device
    --------------------------------------------------------------------------*/
    IOMan_ILogIO&                                       m_PageWriter;
    /*--------------------------------------------------------------------------
    declaration: m_LogInfoPage
    description: log infopage handler
    --------------------------------------------------------------------------*/
    Log_InfoPage                                        m_LogInfoPage;
    /*--------------------------------------------------------------------------
    declaration: m_BufferedPage
    description: buffer for one log page
    --------------------------------------------------------------------------*/
    Log_Page                                            m_BufferedPage;
    /*--------------------------------------------------------------------------
    declaration: m_IOPages
    description: collection of page frames used for write i/o requests
    --------------------------------------------------------------------------*/
    IOMan_LogPages                                      m_IOPages;
    /*--------------------------------------------------------------------------
    declaration: m_RawDeviceBegin
    description: offset on raw device where log begins
    --------------------------------------------------------------------------*/
    RawDeviceOffset                                     m_RawDeviceBegin;
    /*--------------------------------------------------------------------------
    declaration: m_RawDeviceEnd
    description: offset on raw device where log ends
    --------------------------------------------------------------------------*/
    RawDeviceOffset                                     m_RawDeviceEnd;
    /*--------------------------------------------------------------------------
    declaration : m_Allocated
    description:  indicates whether archive log structures have been allocated
                  This member-variable is also used to determine whether
                  the Log-Device is in cold or warm operating mode. This depends
                  on the assumption that the device DOES NOT CHANGE IMMEDIATLY
                  the operation mode FROM WARM TO COLD, but switches the DB to
                  offline-mode in between.
    --------------------------------------------------------------------------*/
    SAPDB_Bool                                          m_Allocated;
    /*--------------------------------------------------------------------------
    declaration : m_Restarted
    description:  If this is true, the log device is prepared to be written.
    --------------------------------------------------------------------------*/
    SAPDB_Bool                                          m_Restarted;
    /*--------------------------------------------------------------------------
    declaration : m_FirstKnownOffset
    description:  holds the offset of the first known page
                  ONLY VALID DURING RESTART
    --------------------------------------------------------------------------*/
    RawDeviceOffset                                     m_FirstKnownOffset;
    /*--------------------------------------------------------------------------
    declaration : m_FirstKnownIOSeq
    description:  holds the IOSequence of the first known page
                  ONLY VALID DURING RESTART
    --------------------------------------------------------------------------*/
    IOSequenceNo                                        m_FirstKnownIOSeq;
	
    /*--------------------------------------------------------------------------
    declaration : m_autosaveDistance
    description:  holds the size of a saved LogSegment in pages (PTS 1113361 mb 2002-01-11)
    --------------------------------------------------------------------------*/
	DeviceBlockCount               						m_autosaveDistance;
};
/*!*****************************************************************************

   endclass: Log_ArchiveLog

*******************************************************************************/

#endif  // LOG_ARCHIVELOG_HPP
