/*!*****************************************************************************

  module:       Log_ArchiveLogIterator.cpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  special area: Archive log

  description:  This iterator gives access to parts of log entries.

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "KernelCommon/Kernel_VTrace.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "Logging/Log_ArchiveLogIterator.hpp"
#include "Logging/Log_QueueRingBuffer.hpp"
#include "Logging/Log_Exceptions.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"

/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

// -----------------------------------------------------------------------------
Log_ArchiveLogIterator::~Log_ArchiveLogIterator ()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogIterator::~Log_ArchiveLogIterator", LogVolume_Trace, 5);
}

// -----------------------------------------------------------------------------
void Log_ArchiveLogIterator::Initialize()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogIterator::Initialize", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( ! m_Frame.IsAssigned() );
    m_Frame = m_FrameAlloc.New();
    m_Page.SetFrame (m_Frame);
    m_Page.InitializeFrame (Log_Page::QueueID(),
                            Log_Page::IOSequenceNo());
    m_EntryIter.SetPage (m_Page);
    m_State = Iter_ok;
    SAPDBERR_ASSERT_STATE( m_Frame.IsAssigned() );
}

// -----------------------------------------------------------------------------
void Log_ArchiveLogIterator::Delete()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogIterator::Delete", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( m_Frame.IsAssigned() );
    m_Page.SetFrame (Log_Page::PageFrame());
    m_FrameAlloc.Free(m_Frame);
    destroy(m_PageReader,m_Allocator);
    m_State = Iter_unInitialized;
}

// -----------------------------------------------------------------------------
void Log_ArchiveLogIterator::Invalidate(const IterState state)
{
    m_EntryIter.Invalidate();
    m_State = state;
}

// -----------------------------------------------------------------------------
SAPDB_Byte* Log_ArchiveLogIterator::operator * ()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogIterator::operator *", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( m_Frame.IsAssigned() );
    SAPDBERR_ASSERT_STATE( GetState() == Iter_ok );
    return *m_EntryIter;
}

// -----------------------------------------------------------------------------
Log_ArchiveLogIterator& Log_ArchiveLogIterator::operator ++ ()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogIterator::operator ++", LogVolume_Trace, 5);

    SAPDBERR_ASSERT_STATE( m_Frame.IsAssigned() );

    Log_IOSequenceNo prevIOSequence = m_EntryIter.GetIOSequence();
    ++m_EntryIter;

    if ( m_EntryIter.IsValid() )
    // the next entry is still on the current page
    {
        if (CheckUntilIOSequenceReached())
        {
             RTE_Message( Log_Exception(
                         __FILE__,__LINE__,
                         LOG_UNTIL_IOSEQ_REACHED,
                         SAPDB_ToString(prevIOSequence ,_T_d),
                         SAPDB_ToString(m_PageReader->GetPosition(),_T_d),
                         SAPDB_ToString(m_Page.ReadDate(),_T_d),
                         SAPDB_ToString(m_Page.ReadTime(),_T_d)));
        }
        return *this;
    }
    
    // store some page-information for a later knldiag-message
    const Kernel_Date           lastPageDate        = m_Page.ReadDate();
    const Kernel_Time           lastPageTime        = m_Page.ReadTime();
    const Log_RawDeviceOffset   lastReadOffset      = m_PageReader->GetPosition();
    const Log_IOSequenceNo      lastReadIOSequence  = m_Page.ReadLastWriterIOSequenceNo();
    const Log_IOSequenceNo      prevQueueSequence   = m_Page.ReadQueueIOSequenceNo();

    // check if all of the entries are read PTS 1117798 mb 2002-09-11
    if (m_LastKnownIOSequence.IsValid())
    {
        if (prevIOSequence == m_LastKnownIOSequence)
        {
            RTE_Message(Log_Exception(
                        __FILE__,__LINE__,
                        LOG_NORMAL_END_FOUND,
                        SAPDB_ToString(lastReadOffset,_T_d),
                        SAPDB_ToString(lastReadIOSequence,_T_d) ));
            Invalidate(Iter_allRead);
            return *this;
        }
    }
    
    // A new page has to be read
    Log_ISequentialReader::ReaderStatus setResult = m_PageReader->SetIOSeq(prevIOSequence+1);
    Log_ISequentialReader::ReaderStatus readResult = m_PageReader->ReadNextPage(m_TaskId, m_Page);
    if (Log_ISequentialReader::end_of_file == readResult)
    {
        RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_NORMAL_END_FOUND,
                     SAPDB_ToString(lastReadOffset,_T_d),
                     SAPDB_ToString(lastReadIOSequence,_T_d) ));
        Invalidate(Iter_allRead);
        return *this;
    }
    else if (Log_ISequentialReader::reader_ok != readResult)
    {
        RTE_Crash( Log_Exception(__FILE__, __LINE__,
                   LOG_READ_UNEXPECTED_ERROR,
                   SAPDB_ToString(m_LastKnownIOSequence,_T_d),
                   SAPDB_ToString(prevIOSequence,_T_d) ));
    }
    
    if (CheckUntilDateReached())
    {
         RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_UNTIL_DATE_REACHED,
                     SAPDB_ToString(lastPageDate,_T_d),
                     SAPDB_ToString(lastPageTime,_T_d),
                     SAPDB_ToString(lastReadOffset,_T_d),
                     SAPDB_ToString(lastReadIOSequence,_T_d) ));
        return *this;
    }

    m_Page.Begin (m_EntryIter);

    if (CheckUntilIOSequenceReached())
    {
         RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_UNTIL_IOSEQ_REACHED,
                     SAPDB_ToString(lastReadIOSequence ,_T_d),
                     SAPDB_ToString(lastReadOffset,_T_d),
                     SAPDB_ToString(lastPageDate,_T_d),
                     SAPDB_ToString(lastPageTime,_T_d)));
        return *this;
    }

    CheckLogDeviceResult checkresult =
        CheckLogDeviceConsistency ( m_CheckMode,
                                    m_Page,
                                    m_PageReader->GetPosition(),
                                    prevIOSequence,
                                    prevQueueSequence );

    if ( inconsistent == checkresult )
		RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                   SAPDBERR_ASSERT_STATE_FAILED,
                   "Log_ArchiveLogIterator::operator++() inconsistent log device"));

    if ( endOfLogDevice == checkresult )
    {
		if ( m_LastKnownIOSequence.IsValid())
		{
			if ( m_LastKnownIOSequence != prevIOSequence )
			{
				RTE_Crash( Log_Exception(__FILE__, __LINE__,
						   LOG_READ_LAST_KNOWN_NOT_FOUND,
						   SAPDB_ToString(m_PageReader->GetPosition(),_T_d),
						   SAPDB_ToString(m_LastKnownIOSequence,_T_d),
						   SAPDB_ToString(prevIOSequence,_T_d) ));
			}
		}

        RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_NORMAL_END_FOUND,
                     SAPDB_ToString(lastReadOffset,_T_d),
                     SAPDB_ToString(lastReadIOSequence,_T_d) ));
        Invalidate(Iter_allRead);
        return *this;
    }

    return *this;
}

// -----------------------------------------------------------------------------
void Log_ArchiveLogIterator::SetPosition (Log_RawDeviceOffset pageoffset,
                                          OffsetType          entryoffset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogIterator::SetPosition", LogVolume_Trace, 5);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "pageoffset: "<<pageoffset<<", entryoffset: "<<entryoffset);

    if ( ! m_Frame.IsAssigned() )
        Initialize();

    const Kernel_Date           lastPageDate        = m_Page.ReadDate();
    const Kernel_Time           lastPageTime        = m_Page.ReadTime();
    const Log_RawDeviceOffset   lastReadOffset      = m_PageReader->GetPosition();
    const Log_IOSequenceNo      lastReadIOSequence  = m_Page.ReadLastWriterIOSequenceNo();

    m_EntryIter.SetPosition    (entryoffset);
    m_PageReader->SetPosition(pageoffset);
    m_PageReader->ReadNextPage(m_TaskId, m_Page);
    if (CheckUntilDateReached())
    {
         RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_UNTIL_DATE_REACHED,
                     SAPDB_ToString(lastPageDate,_T_d),
                     SAPDB_ToString(lastPageTime,_T_d),
                     SAPDB_ToString(lastReadOffset,_T_d),
                     SAPDB_ToString(lastReadIOSequence,_T_d) ));
    }
    if (CheckUntilIOSequenceReached())
    {
         RTE_Message( Log_Exception(
                     __FILE__,__LINE__,
                     LOG_UNTIL_IOSEQ_REACHED,
                     SAPDB_ToString(lastReadIOSequence ,_T_d),
                     SAPDB_ToString(lastReadOffset,_T_d),
                     SAPDB_ToString(lastPageDate,_T_d),
                     SAPDB_ToString(lastPageTime,_T_d)));
    }
}
// -----------------------------------------------------------------------------
void Log_ArchiveLogIterator::SetPosition (Log_RawDeviceOffset pageOffset)
{
    m_Page.Begin(m_EntryIter);
    SetPosition(pageOffset,m_EntryIter.GetPosition());
}
// -----------------------------------------------------------------------------
void Log_ArchiveLogIterator::WriteToTrace (const char* title)
{
    if ( ! GetState()==Iter_ok )
    {
        Kernel_VTrace() << title << ": invalid";
        return;
    }
    Kernel_VTrace() << title
                    << ": "  << m_PageReader->GetPosition()
                    << "."   << m_EntryIter.GetPosition()
                    << " ["  << m_EntryIter.GetIOSequence()
                    << "/"   << m_EntryIter.GetQueueId()
                    << "] "  << m_EntryIter.Length()
                    << "("  << m_EntryIter.RestLength()
                    << ")";
}

// -----------------------------------------------------------------------------
Log_ArchiveLogIterator::CheckLogDeviceResult
Log_ArchiveLogIterator::CheckLogDeviceConsistency
    (Log_DeviceCheckMode            checkmode,
     const Log_Page                &page,
     const Log_RawDeviceOffset      position,
     Log_IOSequenceNo               prevIOSequence,
     Log_IOSequenceNo               prevQueueSequence)
{
    SAPDBTRACE_ROUTINE_DEBUG ("Log_ArchiveLogIterator::CheckLogDeviceConsistency", LogVolume_Trace, 5);

    if ( page.ReadLastWriterIOSequenceNo().IsInvalid()
         ||
         page.ReadFirstWriterIOSequenceNo() < prevIOSequence )
        return endOfLogDevice;

    switch ( checkmode )
    {
    case Log_CheckQueueSequences:
        if ( page.ReadQueueIOSequenceNo() != 0 // after restart the queue sequence may be resetted
             &&
             page.ReadQueueIOSequenceNo() != prevQueueSequence + 1 )
        {
            RTE_Message( Log_Exception(
                         __FILE__,__LINE__,
                         LOG_READ_QSEQ_INCONSISTENT,
                         SAPDB_ToString(position,_T_d),
                         SAPDB_ToString(prevQueueSequence+1,_T_d),
                         SAPDB_ToString(page.ReadQueueIOSequenceNo(),_T_d) ));
            return inconsistent;
        }
        if ( page.ReadFirstWriterIOSequenceNo() > prevIOSequence + 1 )
        {
            RTE_Message( Log_Exception(__FILE__, __LINE__,
                         LOG_READ_IOSEQ_INCONSISTENT,
                         SAPDB_ToString(position,_T_d),
                         SAPDB_ToString(prevIOSequence+1,_T_d),
                         SAPDB_ToString(page.ReadFirstWriterIOSequenceNo(),_T_d) ));
            return inconsistent;
        }
        return consistent;
    case Log_CheckIOSequences:
        if ( page.ReadFirstWriterIOSequenceNo() > prevIOSequence + 1 )
        {
            RTE_Message( Log_Exception(__FILE__, __LINE__,
                         LOG_READ_IOSEQ_INCONSISTENT,
                         SAPDB_ToString(position,_T_d),
                         SAPDB_ToString(prevIOSequence + 1,_T_d),
                         SAPDB_ToString(page.ReadFirstWriterIOSequenceNo(),_T_d) ));
            return inconsistent;
        }
        return consistent;
    case Log_CheckAscendingSequences:
        // +++++++++++++++++
		RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                   SAPDBERR_ASSERT_STATE_FAILED,
                   "Log_ArchiveLogIterator::CheckLogDeviceConsistency() missing check mode"));
        return inconsistent;
    default:
		RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                   SAPDBERR_ASSERT_STATE_FAILED,
                   "Log_ArchiveLogIterator::CheckLogDeviceConsistency() invalid check mode"));
        return inconsistent;
    }
}

// -----------------------------------------------------------------------------

bool Log_ArchiveLogIterator::CheckUntilDateReached()
{
    if (m_UntilDate.IsValid() && m_UntilTime.IsValid())			
    {
        if (! Kernel_DateTime::OlderThan(m_Page.ReadDate(),m_Page.ReadTime(),
                                         m_UntilDate, m_UntilTime))
        {
            Invalidate(Iter_untilReached);
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

