/*!*****************************************************************************

  module:       Log_ArchiveLogIterator.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH
  
  special area: Archive log

  description:  This iterator gives access to parts of log entries.

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_ARCHIVELOGITERATOR_HPP
#define LOG_ARCHIVELOGITERATOR_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // tsp00_TaskId
#include "IOManager/IOMan_ILogIO.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_RawDeviceIterator.hpp"
#include "Logging/Log_ClusterAddress.hpp"
#include "Logging/Log_Queue.hpp"
#include "Logging/Log_ISequentialReader.hpp"

class Log_FrameAllocator;

/*!-------------------------------------------------------------------------
class:        Log_ArchiveLogIterator
description:  This offers an iterator, to access afterimages from the archive log.
--------------------------------------------------------------------------*/
class Log_ArchiveLogIterator
{

friend class Log_ArchiveLog;

public: // types

    /*!-------------------------------------------------------------------------
    declaration:  EntryLength
    --------------------------------------------------------------------------*/
    typedef Log_Page::EntryLength        EntryLength;
    /*!-------------------------------------------------------------------------
    declaration:  IOSequenceNo
    --------------------------------------------------------------------------*/
    typedef Log_Page::IOSequenceNo       IOSequenceNo;
    /*!-------------------------------------------------------------------------
    declaration:  QueueID
    --------------------------------------------------------------------------*/
    typedef Log_Page::QueueID            QueueID;
    /*!-------------------------------------------------------------------------
    declaration:  EntryIterator
    --------------------------------------------------------------------------*/
    typedef Log_Page::Iterator           EntryIterator;
    /*!-------------------------------------------------------------------------
    declaration:  PageFrame
    --------------------------------------------------------------------------*/
    typedef Kernel_Page::PageFrame       PageFrame;
    /*!-------------------------------------------------------------------------
    declaration:  OffsetType
    --------------------------------------------------------------------------*/
    typedef Log_Page::OffsetType         OffsetType;
    /*!-------------------------------------------------------------------------
    declaration:  CheckLogDeviceResult
    --------------------------------------------------------------------------*/
    enum CheckLogDeviceResult
    {
        consistent,
        inconsistent,
        endOfLogDevice
    };

    enum IterState {	// PTS 1111525 mb 2002-02-20 restore log until: differentiate invalid-reason
        Iter_ok,             // iter has valid value
        Iter_aborted,        // readTask has been aborted
        Iter_untilReached,   // for restoreLogUntil: until-time&date reached
        Iter_allRead,        // everything has been read from this log
        Iter_unInitialized   // the iterator has not yet been initialized
    };

    
public: // methods

    /*!-------------------------------------------------------------------------
    function:     Log_ArchiveLogIterator()
    arguments:    TaskId     [in] calling task id
                  pagereader [in] used to read log pages from correct device
                  logNo      [in] archive log device no
                  begin      [in] device begin offset
                  end        [in] device end offset
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator (tsp00_TaskId            taskId,
                            Log_ISequentialReader  *pageReader,
                            Log_IOSequenceNo        lastknown,
                            SAPDBMem_IRawAllocator &alloc,
                            Log_FrameAllocator     &framealloc,
                            Log_DeviceCheckMode     checkmode,
                      const Kernel_Date             untilDate,    // PTS 1111525 mb 2001-01-15
                      const Kernel_Time             untilTime,    // PTS 1111525 mb 2001-01-15
                      const Log_IOSequenceNo        untilIOSeq)   // PTS 1121539 mb 2003-02-06
    : m_TaskId              (taskId),
      m_PageReader          (pageReader),
      m_Allocator           (alloc),
      m_FrameAlloc          (framealloc),
      m_Page                (m_Frame),
      m_EntryIter           (m_Page.GetIterator()),
      m_CheckMode           (checkmode),
      m_LastKnownIOSequence (lastknown),
      m_UntilDate           (untilDate),
      m_UntilTime           (untilTime),
      m_UntilIOSeq          (untilIOSeq),
      m_State               (Iter_unInitialized)
    {
    }
    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  there is an io page allocated to read log pages
    --------------------------------------------------------------------------*/
    void Initialize ();
    /*!-------------------------------------------------------------------------
    function:     SetPageReader()
    description:  there is an io page allocated to read log pages
    --------------------------------------------------------------------------*/
    void SetPageReader(Log_ISequentialReader  *pageReader,
                       SAPDBMem_IRawAllocator &alloc)
    {
        destroy(m_PageReader,m_Allocator);
        m_Allocator = alloc;
        m_PageReader = pageReader;
        m_State = Iter_ok;
    }
    /*!-------------------------------------------------------------------------
    function:     SetLastKnownIOSeq()
    description:  set the last known iosequence-number 
    --------------------------------------------------------------------------*/
    void SetLastKnownIOSeq(Log_IOSequenceNo lastKnownIOSeq)
    {
        m_LastKnownIOSequence=lastKnownIOSeq;
    }

    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  the resources are released
    --------------------------------------------------------------------------*/
    void Delete ();
    
    /*!-------------------------------------------------------------------------
    function:     ~Log_ArchiveLogIterator ()
    description:  the resources are released
    --------------------------------------------------------------------------*/
    ~Log_ArchiveLogIterator ();

    /*!-------------------------------------------------------------------------
    function:     GetState()
    description:  returns the state of the iterator
                  PTS 1111525 mb 2002-02-20 new for restore log until
    return value: IterState
    --------------------------------------------------------------------------*/
    const IterState GetState() const
    {
        if (Iter_ok == m_State)
        {
            if (m_EntryIter.IsValid())
            {
                return Iter_ok;
            }
            else
            {
                return Iter_unInitialized;
            }
        }
        else
        {
            return m_State;
        }
    }
    
    /*!-------------------------------------------------------------------------
    function:     Invalidate()
    arguments:    state    [in] new state of this iterator
    description:  releases all allocated resources, after that it is invalid.
    return value: none
    --------------------------------------------------------------------------*/
    void Invalidate(const IterState state);
    /*!-------------------------------------------------------------------------
    function:     operator *
    description:  dereferences to an entry or a part of an entry.
    return value: nothing
    --------------------------------------------------------------------------*/
    SAPDB_Byte* operator * ();
    /*!-------------------------------------------------------------------------
    function:     operator ++
    description:  The iterator is set to the next afterimage.
                  If the end of log is reached, the iterator becomes invalid.
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator& operator ++ ();
    /*!-------------------------------------------------------------------------
    function:    Length
    returns:     Length of current part of the Split record.
    description: This is the length, which is currently found in the actual page.
    --------------------------------------------------------------------------*/
    EntryLength Length () const
    {
        SAPDBERR_ASSERT_STATE( GetState() == Iter_ok );
        return m_EntryIter.Length();
    }
    /*!-------------------------------------------------------------------------
    function:    RestLength
    returns:     EntryLength
    description: The rest length of current entry is given. It may be longer then
                 the current space in page is.
    --------------------------------------------------------------------------*/
    EntryLength RestLength () const
    {
        SAPDBERR_ASSERT_STATE( GetState() == Iter_ok );
        return m_EntryIter.RestLength();
    }
    /*!-------------------------------------------------------------------------
    function:    GetIOSequence
    returns:     IOSequenceNo
    description: The corresponding IOSequence to the current referenced
                 logentry is returned. If IsValid() is false, the invalid
                 iosequence is returned.
    --------------------------------------------------------------------------*/
    IOSequenceNo GetIOSequence () const
    {
        return m_EntryIter.GetIOSequence();
    }
    /*!-------------------------------------------------------------------------
    function:     GetPosition()
    arguments:    none
    returns:      Log_RawDeviceOffset of the current used page (invalid for a tape)
    --------------------------------------------------------------------------*/
    Log_RawDeviceOffset GetPosition () const
    {
        return m_PageReader->GetPosition();
    }

    /*!-------------------------------------------------------------------------
    function:    GetQueueId
    returns:     QueueID
    description: This is the queue id in which this entry was written.
    --------------------------------------------------------------------------*/
    QueueID GetQueueId () const
    {
        return m_EntryIter.GetQueueId();
    }
    /*!-------------------------------------------------------------------------
    function:    GetQueueIOSequenceNo
    returns:     IOSequenceNo
    description: The QueueIoSequence of the current log page is returned.
    --------------------------------------------------------------------------*/
    IOSequenceNo GetQueueIOSequenceNo () const
    {
        return m_EntryIter.GetQueueIOSequenceNo();
	}
    /*!-------------------------------------------------------------------------
    function:    GetTimeStamp
    arguments:   date [out]
                 time [out]
    description: This is the timestamp of the last log entry of the current page.
    --------------------------------------------------------------------------*/
    void GetTimeStamp ( Kernel_Date &date,
                        Kernel_Time &time ) const
    {
        m_EntryIter.GetLastEntryTimestamp (date, time);
    }
    /*!-------------------------------------------------------------------------
    function:    WriteToTrace
    description: writes current position to the trace file.

    Format:
    
    "title: 4711.16 [756/0/76] 310(1350)"
    
    title      - given by caller
    4711       - offset on log device
    16         - offset in log page
    [756/0/76] - writer io sequence/queue number/queue io sequence
    310        - length of logentry in current page
    1350       - length of logentry in total with 4byte length-field included
    --------------------------------------------------------------------------*/
    void WriteToTrace (const char* title);

    /*!-------------------------------------------------------------------------
    function:     CheckLogDeviceConsistency()
    arguments:    many
    description:  if the criterias are not fullfilled RTE_Crash() is called.
    --------------------------------------------------------------------------*/
    static CheckLogDeviceResult CheckLogDeviceConsistency
                                          (Log_DeviceCheckMode       checkmode,
                                           const Log_Page           &page,
                                           const Log_RawDeviceOffset position,
                                           Log_IOSequenceNo          prevIOSequence,
                                           Log_IOSequenceNo          prevQueueSequence);


    /*!-------------------------------------------------------------------------
    function:     GetPosition()
    arguments:    pageoffset  [out] current offset to begin
                  entryoffset [out] offset in bytes in the logpage to a specific entry.
    --------------------------------------------------------------------------*/
    void GetPosition (Log_RawDeviceOffset &pageoffset,
                      OffsetType          &entryoffset) const
    {
        pageoffset  = m_PageReader->GetPosition();
        entryoffset = m_EntryIter.GetPosition();
    }

protected:

    /*!-------------------------------------------------------------------------
    function:     SetPosition()
    arguments:    pageoffset  [in] current offset to begin
                  entryoffset [in] offset in bytes in the logpage to a specific entry.
    description:  This is used by the ArchiveLog to set the iterator.
    --------------------------------------------------------------------------*/
    void SetPosition (Log_RawDeviceOffset  pageoffset,
                      OffsetType           entryoffset);
    /*!-------------------------------------------------------------------------
    function:     SetPosition()
    arguments:    pageoffset  [in] current offset to begin
                                   the entry-offset will be set to the begin 
                                   of the page                 
    description:  This is used by the RedoReader to set the iterator.
    --------------------------------------------------------------------------*/
    void SetPosition (Log_RawDeviceOffset  pageOffset);

private:
    /*!-------------------------------------------------------------------------
    function:     operator=(const Log_ArchiveLogIterator& rhs)
    arguments:    rhs
    description:  do not allow an assignment operation
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator& operator=(const Log_ArchiveLogIterator& rhs);

    /*!-------------------------------------------------------------------------
    function:     UntilDateReached
    arguments:    none
    returns:      bool
    description:  checks, if the specified date is reached, sets the state
                  correctly and returns TRUE, if the date has been reached
    --------------------------------------------------------------------------*/
    bool CheckUntilDateReached();

    /*!-------------------------------------------------------------------------
    function:     UntilIOSequenceReached
    arguments:    none
    returns:      bool
    description:  checks, if the specified IOsequenceNo is reached, sets the state
                  correctly and returns TRUE, if the IOSequence has been reached
                  This is defined inline, because it is called for every
                  increment of the Interator
    --------------------------------------------------------------------------*/
    inline bool CheckUntilIOSequenceReached();

    /*--------------------------------------------------------------------------
    declaration : m_TaskId
    description:  handle to read pages from log device.
    --------------------------------------------------------------------------*/
    tsp00_TaskId                   m_TaskId;
    /*--------------------------------------------------------------------------
    declaration : m_PageReader
    description:  handle to read pages sequentially from log device or tape.
    --------------------------------------------------------------------------*/
    Log_ISequentialReader         *m_PageReader;
    /*--------------------------------------------------------------------------
    declaration : m_FrameAlloc
    description:  is used to allocate frames to read log pages
    --------------------------------------------------------------------------*/
    Log_FrameAllocator            &m_FrameAlloc;
    /*--------------------------------------------------------------------------
    declaration: m_Allocator
    description: raw allocator 
    --------------------------------------------------------------------------*/
    SAPDBMem_IRawAllocator        &m_Allocator;
    /*--------------------------------------------------------------------------
    declaration : m_pFrame
    description:  a frame to hold a read log page
    --------------------------------------------------------------------------*/
    PageFrame                      m_Frame;
    /*--------------------------------------------------------------------------
    declaration : m_Page
    description:  the log page to read entries from
    --------------------------------------------------------------------------*/
    Log_Page                       m_Page;
    /*--------------------------------------------------------------------------
    declaration : m_EntryIter
    description:  points to current entry
    --------------------------------------------------------------------------*/
    EntryIterator                  m_EntryIter;
    /*--------------------------------------------------------------------------
    declaration : m_CheckMode
    description:  this is used for checking the read log pages.
    --------------------------------------------------------------------------*/
    Log_DeviceCheckMode            m_CheckMode;
    /*--------------------------------------------------------------------------
    declaration : m_LastKnownIOSequence
    description:  this is used for checking the read log pages.
    --------------------------------------------------------------------------*/
    Log_IOSequenceNo               m_LastKnownIOSequence;
    /*--------------------------------------------------------------------------
    declaration: m_UntilDate
    description: the date until a restore/restart will be done
                 PTS 1111525 mb 2001-01-18
                 
    --------------------------------------------------------------------------*/
    const Kernel_Date              m_UntilDate;

    /*--------------------------------------------------------------------------
    declaration: m_UntilTime
    description: the time until a restore/restart will be done
                 PTS 1111525 mb 2001-01-18
    --------------------------------------------------------------------------*/
    const Kernel_Time              m_UntilTime;

    /*--------------------------------------------------------------------------
    declaration: m_UntilIOSeq
    description: the IOSequence until a restore/restart will be done (inclusive
                 this IOSequence)
                 PTS 1121539 mb 2003-02-06
   --------------------------------------------------------------------------*/
    const Log_IOSequenceNo         m_UntilIOSeq;

    /*--------------------------------------------------------------------------
    declaration: m_state
    description: the current state of this iterator
                 PTS 1111525 mb 2002-02-20
    --------------------------------------------------------------------------*/
    IterState              m_State;

};

/*==========================================================================*
*  DEFINITION OF INLINE METHODS                                             *
*==========================================================================*/

inline bool Log_ArchiveLogIterator::CheckUntilIOSequenceReached()
{
    if (m_UntilIOSeq.IsValid())                  // PTS 1121539 2003-06-06
    {
        if (m_EntryIter.GetIOSequence() > m_UntilIOSeq)
        {
            Invalidate(Iter_untilReached);
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

/*!-------------------------------------------------------------------------
endclass:     Log_ArchiveLogIterator
--------------------------------------------------------------------------*/

#endif // LOG_ARCHIVELOG_HPP
