/*!*****************************************************************************

  module:       Log_ArchiveLogManager.cpp

  ------------------------------------------------------------------------------

  author      : TillL
  responsible : UweH

  special area: Archive log

  description:  see .hpp

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "Logging/Log_ArchiveLogManager.hpp"

/*!*****************************************************************************

   class: Log_ArchiveLogManager

*******************************************************************************/
Log_ArchiveLogManager* Log_ArchiveLogManager::m_Instance = 0;
// -----------------------------------------------------------------------------
SAPDB_Bool Log_ArchiveLogManager::AddArchiveLog
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_Queue::SynchronizationMode syncmode,
     Log_FlushMode                  flushmode)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::AddArchiveLog", LogVolume_Trace, 5);
 
    ArchiveLog* archlog = new(m_Allocator) 
        ArchiveLog(taskid, writertaskid, m_Allocator, m_FrameAlloc, 
                   syncmode, flushmode);

    if (archlog == 0)
        return false;

    if (!m_ArchiveLogs.Reserve(m_ArchiveLogs.GetSize() + 1))
        return false;

    m_ArchiveLogs.InsertEnd(archlog);

    return true;
}
// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLogManager::CreateArchiveLogForNewDatabase
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_RawDeviceOffset            rawlogbegin,
     Log_RawDeviceOffset            rawlogend,
     Log_DeviceBlockCount           savedist,
     SAPDB_UInt                     queuecount,
     Log_Queue::EntryCount          queuesize,
     SAPDB_UInt                     ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::CreateArchiveLogForNewDatabase", LogVolume_Trace, 5);
 
    ArchiveLog& archlog = GetArchiveLog(writertaskid);

    Log_ErrorState result = 
        archlog.CreateForNewDatabase(taskid, rawlogbegin, rawlogend, 
                                     savedist, queuecount, queuesize, 
                                     ioblockcount);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "result: " << Log_ErrorStateStrings[result]);
    return result;
}
// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLogManager::RestartArchiveLog
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_RawDeviceOffset            rawlogbegin,
     Log_RawDeviceOffset            rawlogend,
     DeviceBlockCount               autosaveDistance,
     Log_Queue::EntryCount          queuesize,
     SAPDB_UInt                     ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::RestartArchiveLog", LogVolume_Trace, 5);
 
    ArchiveLog& archlog = GetArchiveLog(writertaskid);

    Log_ErrorState result = 
        archlog.Restart(taskid, rawlogbegin, rawlogend, 
                        autosaveDistance, queuesize, ioblockcount);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "result: " << Log_ErrorStateStrings[result]);

    if ( Log_Ok == result ) // PTS 1113190 UH 2001-12-19
        m_DevspaceEnabled = archlog.IsDevspaceEnabled();

    return result;
}
// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLogManager::RestartForRestoreLog
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_RawDeviceOffset            rawlogbegin,
     Log_RawDeviceOffset            rawlogend,
     DeviceBlockCount               autosaveDistance,
     Log_Queue::EntryCount          queuesize,
     SAPDB_UInt                     ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::RestartForRestoreLog", LogVolume_Trace, 5);
 
    ArchiveLog& archlog = GetArchiveLog(writertaskid);

    Log_ErrorState result = 
        archlog.RestartForRestoreLog(taskid, rawlogbegin, rawlogend, 
                                     autosaveDistance, queuesize, ioblockcount);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "result: " << Log_ErrorStateStrings[result]);

    if ( Log_Ok == result ) // PTS 1113190 UH 2001-12-19
        m_DevspaceEnabled = archlog.IsDevspaceEnabled();

    return result;
}


// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLogManager::RestartForSaveLogCold 
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_RawDeviceOffset            rawlogbegin,
     Log_RawDeviceOffset            rawlogend,
     DeviceBlockCount               autosaveDistance)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::RestartForSaveLogCold", LogVolume_Trace, 5);
 
    ArchiveLog& archlog = GetArchiveLog(writertaskid);

    // !!!!!!!! up to now only one log-device is supported here !!!!!!!!!!!
    Log_ErrorState result = 
        archlog.RestartForSaveLogCold(taskid, rawlogbegin, rawlogend,
                                      autosaveDistance);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "result: " << Log_ErrorStateStrings[result]);
    return result;
}

// -----------------------------------------------------------------------------
Log_ErrorState Log_ArchiveLogManager::ClearArchiveLog
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_RawDeviceOffset            rawlogbegin,
     Log_RawDeviceOffset            rawlogend,
     DeviceBlockCount               autosaveDistance,
     Log_Queue::EntryCount          queuesize,
     SAPDB_UInt                     ioblockcount)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::ClearArchiveLog", LogVolume_Trace, 5);
 
    ArchiveLog& archlog = GetArchiveLog(writertaskid);

    Log_ErrorState result = 
        archlog.RestartForClearLog(taskid, rawlogbegin, rawlogend,
                                   autosaveDistance, queuesize, ioblockcount);
    if (Log_Ok==result)
    {
        archlog.ClearLogComplete(taskid);
    }

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "result: " << Log_ErrorStateStrings[result]);
    return result;
}
// -----------------------------------------------------------------------------
Log_ArchiveLogIterator  
Log_ArchiveLogManager::GetArchiveLogIterator (
            SAPDBMem_IRawAllocator    &alloc,
            tsp00_TaskId               taskId,
      const Kernel_Date                untilDate,    // PTS 1111525 mb 2001-01-15
      const Kernel_Time                untilTime,    // PTS 1111525 mb 2001-01-15
      const Log_IOSequenceNo           untilIOSeq)   // PTS 1121539 mb 2003-02-06
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::GetArchiveLogIterator", LogVolume_Trace, 5);
    // in future there must be a merger unit before the iterators of all archive logs.
    ArchiveLog& archlog = **(ArchiveLogsBegin());
    return archlog.GetArchiveLogIterator (alloc,taskId,untilDate,untilTime,untilIOSeq);
}

// -----------------------------------------------------------------------------

Log_ArchiveLogIterator 
Log_ArchiveLogManager::GetTapeIterator (
            SAPDBMem_IRawAllocator     &alloc,
            tgg00_TransContext         &readerTrans,
            Log_IOSequenceNo            lastKnownOnTape,
      const Kernel_Date                 untilDate,    // PTS 1111525 mb 2001-01-15
      const Kernel_Time                 untilTime,    // PTS 1111525 mb 2001-01-15
      const Log_IOSequenceNo            untilIOSeq)   // PTS 1121539 mb 2003-02-06
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::GetArchiveLogIterator", LogVolume_Trace, 5);

    ArchiveLog& archlog = **(ArchiveLogsBegin());
    return archlog.GetTapeIterator (alloc,readerTrans,lastKnownOnTape,untilDate,untilTime,untilIOSeq);
}
// -----------------------------------------------------------------------------

void Log_ArchiveLogManager::SetToLastSavepointEntry (Log_ArchiveLogIterator &iter,
                                                     Log_IOSequenceNo       &svpSeq)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::SetToLastSavepointEntry", LogVolume_Trace, 5);
    // in future there must be a merger unit before the iterators of all archive logs.
    ArchiveLog& archlog = **(ArchiveLogsBegin());
    archlog.SetToLastSavepointEntry (iter, svpSeq);
}
// -----------------------------------------------------------------------------
void Log_ArchiveLogManager::SetToBeginOfPage (Log_ArchiveLogIterator    &iter,
                                              Log_RawDeviceOffset        offset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::SetToLastSavepointEntry", LogVolume_Trace, 5);
    // in future there must be a merger unit before the iterators of all archive logs.
    ArchiveLog& archlog = **(ArchiveLogsBegin());
    archlog.SetToBeginOfPage (iter, offset);
}

// -----------------------------------------------------------------------------
void
Log_ArchiveLogManager::WriteSavepointValuesAndFlushInfoPage
    (TaskID                   taskid,
     Log_SavepointReason      reason,
     const SAPDBFields_Field& rrpart)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::WriteSavepointValuesAndFlushInfoPage", LogVolume_Trace, 5);
    // in future: over all archive logs.
    ArchiveLog& archlog = **(ArchiveLogsBegin());
    archlog.WriteSavepointValuesAndFlushInfoPage (taskid, reason, rrpart);
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::FreeLog (tsp00_TaskId        taskid, 
                                     Log_IOSequenceNo    oldLastSavedIOsequence,
                                     Log_RawDeviceOffset oldLastSavedOffset,
                                     Log_IOSequenceNo    newLastSavedIOsequence,
                                     Log_RawDeviceOffset newLastSavedOffset)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_ArchiveLogManager::FreeLog", LogVolume_Trace, 5);
    // in future: over all archive logs.
    ArchiveLog& archlog = **(ArchiveLogsBegin());
    return archlog.FreeLog (taskid, oldLastSavedIOsequence, oldLastSavedOffset,
                                    newLastSavedIOsequence, newLastSavedOffset);
}
// -----------------------------------------------------------------------------
void Log_ArchiveLogManager::WriteToTrace()
{
    // in future: over all archive logs.
    ArchiveLog& archlog = **(ArchiveLogsBegin());
    archlog.WriteDeviceParametersToTrace();
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::IsRestarted ()
{
    if ( m_Instance == NULL
         ||
         Instance().m_ArchiveLogs.IsEmpty() )
        return false;
    
    ArchiveLog& log = **(Instance().ArchiveLogsBegin());
    
    return log.IsRestarted();
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::SetDeviceState (tsp00_TaskId    taskid,
                                            bool            isAdminMode,
                                            Log_DeviceState state)
{
    // PTS 1113190 UH 2001-12-19 new
    Log_ArchiveLog &log = **(ArchiveLogsBegin());
    if ( ! log.DiagnoseOpenInfoPage(taskid, isAdminMode) )
        return false;
    log.SetDeviceState (taskid, state);
    if ( isAdminMode )
        log.DiagnoseCloseDevice(taskid);
    return true;
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::SetDevspace (tsp00_TaskId taskid,
                                         bool         isAdminMode,
                                         bool         on)
{
    // PTS 1113190 UH 2001-12-19 new
    if ( ! isAdminMode )
        return false;
        
    m_DevspaceEnabled   = on;
    Log_ArchiveLog &log = **(ArchiveLogsBegin());
    
    if ( ! log.DiagnoseOpenInfoPage(taskid,true) )
        return false;
        
    log.SetDevspaceEnabled (taskid, m_DevspaceEnabled);

    if ( ! m_DevspaceEnabled )
        log.SetDeviceState (taskid, Log_DeviceStateHistoryLost);
    
    log.DiagnoseCloseDevice(taskid);
    return true;
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::IsDevspaceEnabled(tsp00_TaskId taskid)
{
    if ( ! IsRestarted() ) 				// PTS 1118565 mb 2002-11-1 really read the value in admin-mode
    {
        Log_ArchiveLog &log = **(ArchiveLogsBegin());
        
        m_DevspaceEnabled = false;
        if ( log.DiagnoseOpenInfoPage(taskid,true) )
        {
            m_DevspaceEnabled = log.IsDevspaceEnabled();
            log.DiagnoseCloseDevice(taskid);
        }
    }

    return m_DevspaceEnabled;
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::SetAutomaticOverwrite (tsp00_TaskId taskid,
                                                   bool         isAdminMode,
                                                   bool         on)
{
    // PTS 1120106 UH 2003-01-20 new
    Log_ArchiveLog &log = **(ArchiveLogsBegin());
    
    if ( isAdminMode && ! log.DiagnoseOpenInfoPage(taskid,true) )
        return false;
        
    log.SetAutomaticOverwrite (taskid, on);

    if ( on )
        log.SetDeviceState (taskid, Log_DeviceStateHistoryLost);
    
    if ( isAdminMode )
        log.DiagnoseCloseDevice(taskid);

    return true;
}
// -----------------------------------------------------------------------------
bool Log_ArchiveLogManager::IsAutomaticOverwrite(tsp00_TaskId taskid,
                                                 bool         isAdminMode)
{
    // PTS 1120106 UH 2003-01-20 new
    bool result = false;
    Log_ArchiveLog &log = **(ArchiveLogsBegin());
        
    if ( isAdminMode && ! log.DiagnoseOpenInfoPage(taskid,true) )
        return false;

    result = log.IsAutomaticOverwrite();
    
    if ( isAdminMode )
        log.DiagnoseCloseDevice(taskid);

    return result;
}
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
