/*!*****************************************************************************

  module:       Log_ArchiveLogManager.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  author:       TillL

  special area: Archive log

  description:  Archive log manager singleton

                Administrates all archive logs

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_ARCHIVELOGMANAGER_HPP
#define LOG_ARCHIVELOGMANAGER_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // tsp00_TaskId

#include "RunTime/MemoryManagement/RTEMem_AllocatorWrapper.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Container/Container_Vector.hpp"

#include "Logging/Log_ArchiveLog.hpp"

#include "Logging/Log_IWritingRedoEntries.hpp"
#include "Logging/Log_IReadingRedoEntries.hpp"
#include "Logging/Log_IMonitoring.hpp"
#include "Logging/Log_IEventing.hpp"



/*!*****************************************************************************

   class: Log_ArchiveLogManager

   description:
   
   Archive log manager

   This class is designed as a singleton.

   Log_ArchiveLogManager handles all archive logs in the kernel.

*******************************************************************************/
class Log_ArchiveLogManager
{
    /*--------------------------------------------------------------------------
    declaration: ArchiveLogVector
    description: vector of archive logs
    --------------------------------------------------------------------------*/
    typedef Container_Vector<Log_ArchiveLog*>                   ArchiveLogVector;

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : ArchiveLog
    description:  coordinator of one single archive log
    --------------------------------------------------------------------------*/
    typedef Log_ArchiveLog                                      ArchiveLog;
    /*!-------------------------------------------------------------------------
    declaration : FrameAllocator
    description:  allocator for page frames
    --------------------------------------------------------------------------*/
    typedef Log_FrameAllocator                                  FrameAllocator;
    /*!-------------------------------------------------------------------------
    declaration : ArchiveLogIterator
    description:  iterator that steps from archive log to archive log
    --------------------------------------------------------------------------*/
    typedef ArchiveLogVector::Iterator                          ArchiveLogIterator;
    /*!-------------------------------------------------------------------------
    declaration : TaskID
    description:  identifier of kernel task
    --------------------------------------------------------------------------*/
    typedef tsp00_TaskId                                        TaskID;
    /*!-------------------------------------------------------------------------
    declaration : Log_DeviceBlockCount
    description:  number of blocks on a raw log device
    --------------------------------------------------------------------------*/
    typedef Log_DeviceBlockCount                                DeviceBlockCount;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*--------------------------------------------------------------------------
    function:     Log_ArchiveLogManager()
    description:  Constructor
    arguments:    alloc        [in]  raw allocator to be used
    return value: none
    --------------------------------------------------------------------------*/
    Log_ArchiveLogManager 
        (SAPDBMem_IRawAllocator&        alloc,
         FrameAllocator&                framealloc)
        :
         m_Allocator( UTF8( "Log_ArchiveLogManager" ), alloc ), 
         m_FrameAlloc(framealloc),
         m_ArchiveLogs(m_Allocator),
         m_DevspaceEnabled(true)
    {}

    /*!-------------------------------------------------------------------------
    function:     ~Log_ArchiveLogManager()
    return value: none
    --------------------------------------------------------------------------*/
    ~Log_ArchiveLogManager()
    {
        Delete();
    }

    /*!-------------------------------------------------------------------------
    function:     CreateInstance()
    description:  Creates archive log manager singleton

                  - This method may be called only once. 
    arguments:    alloc      [in]  raw allocator to be used
                  framealloc [in]  allocator for page frames to be used
    return value: none
    --------------------------------------------------------------------------*/
    static void CreateInstance
        (SAPDBMem_IRawAllocator&        alloc,
         FrameAllocator&                framealloc)
    {
        SAPDBMEM_STATIC_RAW_ALLOCATE (Space, sizeof(Log_ArchiveLogManager));
        SAPDBERR_ASSERT_STATE(m_Instance == 0);
        m_Instance = new (Space) Log_ArchiveLogManager (alloc, framealloc);
    }
         
    /*!-------------------------------------------------------------------------
    function:     Instance()
    description:  Returns reference to archive log manager singleton

                  - This method works properly only after the singleton has
                    been created with CreateInstance().
                    
    arguments:    none
    return value: (Log_ArchiveLogManager&) reference to singleton
    --------------------------------------------------------------------------*/
    static Log_ArchiveLogManager& Instance() 
    {
        SAPDBERR_ASSERT_STATE(m_Instance != 0);
        return (*m_Instance);
    }
      
    /*!-------------------------------------------------------------------------
    function:     AddArchiveLog()
    description:  Adds a new archive log
    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task
    return value: (SAPDB_Bool) true if successfully added
    --------------------------------------------------------------------------*/
    SAPDB_Bool AddArchiveLog
        (TaskID                         taskid,
         TaskID                         writertaskid,
         Log_Queue::SynchronizationMode syncmode,
         Log_FlushMode                  flushmode);
    
    /*!-------------------------------------------------------------------------
    function:     CreateArchiveLogForNewDatabase()
    description:  Creates an archive log for a new database

                  - This method prepares an archive log for a new database.
                  - The archive log object must have been added by a previous 
                    call of AddArchiveLog().
                  - 'writertaskid' specifies the archive log to be created.

    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  savedist     [in]  distance between autolog saves in pages
                  queuecount   [in]  number of log queues to be created
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
    return value: (Log_ErrorState) 
                  Log_Ok,
                  Log_PageAllocationFailed,
                  Log_RawAllocationFailed,
                  Log_LogDeviceTooSmall,
                  Log_LogSaveDistanceTooLarge
    --------------------------------------------------------------------------*/
    Log_ErrorState CreateArchiveLogForNewDatabase
        (TaskID                         taskid,
         TaskID                         writertaskid,
         Log_RawDeviceOffset            rawlogbegin,
         Log_RawDeviceOffset            rawlogend,
         DeviceBlockCount               savedist,
         SAPDB_UInt                     queuecount,
         Log_Queue::EntryCount          queuesize,
         SAPDB_UInt                     ioblockcount);
    
    /*!-------------------------------------------------------------------------
    function:     RestartArchiveLog()
    description:  Restarts an existing archive log

                  - This method sets the device parameters for an existing
                    archive log.
                  - The archive log object must have been added by a previous 
                    call of AddArchiveLog().
                  - 'writertaskid' specifies the archive log to be restarted.
                    
    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
    return value: (Log_ErrorState) 
                  Log_Ok,
                  Log_PageAllocationFailed,
                  Log_RawAllocationFailed,
                  Log_LogAndDataIncompatible,
                  Log_InconsistentLogDeviceIOSequence,
                  Log_LogSizeHasDecreased
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartArchiveLog
        (TaskID                         taskid,
         TaskID                         writertaskid,
         Log_RawDeviceOffset            rawlogbegin,
         Log_RawDeviceOffset            rawlogend,
         DeviceBlockCount               autosaveDistance,
         Log_Queue::EntryCount          queuesize,
         SAPDB_UInt                     ioblockcount);
    
    /*!-------------------------------------------------------------------------
    function:     RestartForRestoreLog()
    description:  Restarts archive log in order to perform a 'restore log'
    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
    return value: none
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartForRestoreLog
        (TaskID                         taskid,
         TaskID                         writertaskid,
         Log_RawDeviceOffset            rawlogbegin,
         Log_RawDeviceOffset            rawlogend,
         DeviceBlockCount               autosaveDistance,
         Log_Queue::EntryCount          queuesize,
         SAPDB_UInt                     ioblockcount);

    /*!-------------------------------------------------------------------------
    function:     RestartForSaveLogCold()
    description:  Restarts archive log in order to perform a 'save log' 
                  on cold database
    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
    return value: Log_ErrorState
    --------------------------------------------------------------------------*/
    Log_ErrorState RestartForSaveLogCold 
    (TaskID                         taskid,
     TaskID                         writertaskid,
     Log_RawDeviceOffset            rawlogbegin,
     Log_RawDeviceOffset            rawlogend,
     DeviceBlockCount               autosaveDistance);

    /*!-------------------------------------------------------------------------
    function:     ClearArchiveLog()
    description:  Clears the entire archive log; initializes log info page
    arguments:    taskid       [in]  id of the calling task
                  writertaskid [in]  id of the associated writer task
                  rawlogbegin  [in]  log start offset on raw log device
                  rawlogend    [in]  log end offset on raw log device
                  autosaveDistance   [in] number of pages to be saved by autosave
                  queuesize    [in]  number of log pages to be allocated for each queue
                  ioblockcount [in]  max. count of i/o blocks for multiple write i/o
    return value: (Log_ErrorState) 
                  Log_Ok,
                  Log_PageAllocationFailed,
                  Log_RawAllocationFailed,
                  Log_LogAndDataIncompatible,
                  Log_InconsistentLogDeviceIOSequence,
                  Log_LogSizeHasDecreased
    --------------------------------------------------------------------------*/
    Log_ErrorState ClearArchiveLog
        (TaskID                         taskid,
         TaskID                         writertaskid,
         Log_RawDeviceOffset            rawlogbegin,
         Log_RawDeviceOffset            rawlogend,
         DeviceBlockCount               autosaveDistance,
         Log_Queue::EntryCount          queuesize,
         SAPDB_UInt                     ioblockcount);
    
    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes manager
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Delete()
    {
        ArchiveLogIterator iter = ArchiveLogsBegin();
        ArchiveLogIterator end  = ArchiveLogsEnd();

        while (iter != end)
        {
            destroy(*iter, m_Allocator);
            ++iter;
        }

        m_ArchiveLogs.Delete();
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: single Archive Logs                                          */
    /*!-------------------------------------------------------------------------
    function:     ArchiveLogsBegin()
    description:  Returns iterator at first archive log instance
    arguments:    none
    return value: (ArchiveLogIterator)
    --------------------------------------------------------------------------*/
    ArchiveLogIterator ArchiveLogsBegin()
    {
        return m_ArchiveLogs.Begin();
    }
    
    /*!-------------------------------------------------------------------------
    function:     ArchiveLogsEnd()
    description:  Returns iterator behind last archive log instance
    arguments:    none
    return value: (ArchiveLogIterator)
    --------------------------------------------------------------------------*/
    ArchiveLogIterator ArchiveLogsEnd()
    {
        return m_ArchiveLogs.End();
    }
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: single Archve Logs                                        */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Status                                                       */
    /*!-------------------------------------------------------------------------
    function:     GetQueue()
    description:  Returns log queue instance that serves calling user task

    arguments:    usertaskid [in]  id of the calling user task
    return value: (Log_Queue*) reference to queue
    --------------------------------------------------------------------------*/
    Log_Queue* GetQueue
        (TaskID                         usertaskid)
    {
        if ( ! m_DevspaceEnabled || m_ArchiveLogs.GetSize() == 0) // PTS 1113190 UH 2001-12-19
            return NULL;
            
        SAPDBERR_ASSERT_STATE(m_ArchiveLogs.GetSize() > 0);
        ArchiveLog& archlog = *m_ArchiveLogs[usertaskid % m_ArchiveLogs.GetSize()];
        return &(archlog.GetQueue(usertaskid));
    }

    /*!-------------------------------------------------------------------------
    function:     GetWriter()
    description:  Returns log writer instance that serves calling writer task

    arguments:    writertaskid [in]  id of the calling writer task
    return value: (Log_Writer&) reference to writer
    --------------------------------------------------------------------------*/
    Log_Writer& GetWriter
        (TaskID                         writertaskid)
    {
        Log_ArchiveLog& alog = GetArchiveLog(writertaskid);
        return alog.GetWriter();
    }

    /*!-------------------------------------------------------------------------
    function:     IsRestarted()
    description:  Checks whether archive log restart is complete
    arguments:    none
    return value: (SAPDB_Bool) true if restarted
    --------------------------------------------------------------------------*/
    static bool IsRestarted ();

    /*!-------------------------------------------------------------------------
    function:     SetDeviceState()
    description:  The given device state is written to the loginfo page.
    arguments:    taskid      [in]
                  isAdminMode [in]
                  state       [in]
    return value: bool, true is successfull.
    --------------------------------------------------------------------------*/
    bool SetDeviceState (tsp00_TaskId    taskid,
                         bool            isAdminMode,
                         Log_DeviceState state);
    /*!-------------------------------------------------------------------------
    function:     SetDevspace()
    description:  If on is true then transactions get a queue to write entries.
                  Otherwise NO archivelog entry is written.
    precondition: NOT IsRestarted()
    arguments:    taskid [in]
                  on     [in] (bool)
    return value: bool, true if successfull
    --------------------------------------------------------------------------*/
    bool SetDevspace (tsp00_TaskId taskid,
                      bool         isAdminMode,
                      bool         on);

    /*!-------------------------------------------------------------------------
    function:     IsDevspaceEnabled()
    return value: bool, true if transactions write archive log entries.
    --------------------------------------------------------------------------*/
    bool IsDevspaceEnabled(tsp00_TaskId taskid);

    /*!-------------------------------------------------------------------------
    function:     SetAutomaticOverwrite()
    description:  
    arguments:    taskid [in]
                  on     [in] (bool)
    return value: bool, true if successfull
    --------------------------------------------------------------------------*/
    bool SetAutomaticOverwrite (tsp00_TaskId taskid,
                                bool         isAdminMode,
                                bool         on);

    /*!-------------------------------------------------------------------------
    function:     IsAutomaticOverwrite()
    return value: bool
    --------------------------------------------------------------------------*/
    bool IsAutomaticOverwrite(tsp00_TaskId taskid,
                              bool         isAdminMode);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Status                                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: reading afterimages                                          */
    /*------------------------------------------------------------------------*/
    
    /*!-------------------------------------------------------------------------
    function:     GetArchiveLogIterator()
    description:  This offers an iterator, to access afterimages.
    arguments:    TaskId    [in] 
    return value: Log_ArchiveLogIterator
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator GetArchiveLogIterator (      SAPDBMem_IRawAllocator    &alloc,
                                                        tsp00_TaskId               taskId,
                                                  const Kernel_Date                untilDate,    // PTS 1111525 mb 2001-01-15
                                                  const Kernel_Time                untilTime,    // PTS 1111525 mb 2001-01-15
                                                  const Log_IOSequenceNo           untilIOSeq);  // PTS 1121539 mb 2003-02-06
    /*!-------------------------------------------------------------------------
    function:     GetTapeIterator()
    description:  This offers an iterator, to access afterimages.
    arguments:    alloc  	        [in]      Allocator for memory allocation
                  readerTrans       [in|out]  transactioncontext
                  lastKnownOnTape   [in]      Last IOSequence to be read from tape
    return value: Log_ArchiveLogIterator
    --------------------------------------------------------------------------*/
    Log_ArchiveLogIterator GetTapeIterator (SAPDBMem_IRawAllocator   &alloc,
                                            tgg00_TransContext       &readerTrans,
                                                  Log_IOSequenceNo    lastKnownOnTape,
                                            const Kernel_Date         untilDate,     // PTS 1111525 mb 2001-01-15
                                            const Kernel_Time         untilTime,     // PTS 1111525 mb 2001-01-15
                                            const Log_IOSequenceNo    untilIOSeq);   // PTS 1121539 mb 2003-02-06


    /*!-------------------------------------------------------------------------
    function:     SetToLastSavepointEntry()
    description:  The iterator is set to the last savepoint entry.
    arguments:    iter [in]
    return value: none
    --------------------------------------------------------------------------*/
    void SetToLastSavepointEntry (Log_ArchiveLogIterator &iter,
                                  Log_IOSequenceNo       &svpSeq);

    /*!-------------------------------------------------------------------------
    function:     SetToBeginOfPage()
    description:  The iterator is set to the begin of the page denoted by offset
    arguments:    iter [in]
                  offset [in]
    return value: none
    --------------------------------------------------------------------------*/
    void SetToBeginOfPage (Log_ArchiveLogIterator   &iter,
                           Log_RawDeviceOffset       offset);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: reading afterimages                                       */
    /*------------------------------------------------------------------------*/
    
    /* -----------------------------------------------------------------------*/
    /*! chapter: savepoint                                                    */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     WriteSavepointValuesAndFlushInfoPage()
    description:  The given bytes from the restartrecord are copied into the
                  infopage. The loginfo page is written to disk.
    arguments:    taskid [in]
                  reason [in]
                  rrpart [in]
    return value: none
    --------------------------------------------------------------------------*/
    void WriteSavepointValuesAndFlushInfoPage (TaskID                   taskid,
                                               Log_SavepointReason      reason,
                                               const SAPDBFields_Field& rrpart);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: savepoint                                                 */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: save                                                         */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     FreeLog()
    description:  This releases log pages, which are already backuped.
    arguments:    taskid     [in]
                  oldiosequence [in]
                  oldoffset     [in]
                  newiosequence [in]
                  newoffset     [in]
    return value: (bool) true, if log could be freed
    --------------------------------------------------------------------------*/
    bool FreeLog (tsp00_TaskId        taskid, 
                  Log_IOSequenceNo    oldLastSavedIOsequence,
                  Log_RawDeviceOffset oldLastSavedOffset,
                  Log_IOSequenceNo    newLastSavedIOsequence,
                  Log_RawDeviceOffset newLastSavedOffset);
    /* -----------------------------------------------------------------------*/
    /*! endchapter: save                                                      */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     WriteToTrace()
    description:  Write all archive logs to trace.
    return value: none
    --------------------------------------------------------------------------*/
    void WriteToTrace();

    /* -----------------------------------------------------------------------*/
    /*! chapter: Information                                                  */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     GetArchiveLogCount()
    return value: SAPDB_UInt
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetArchiveLogCount()
    {
        return m_ArchiveLogs.GetSize();
    }
    /*!-------------------------------------------------------------------------
    function:     GetQueueCountPerArchiveLog()
    return value: SAPDB_UInt
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetQueueCountPerArchiveLog()
    {
        return 1;
    }
    /*!-------------------------------------------------------------------------
    function:     IsForMinimizeSpace()
    description:  returns true, if logpages can be written to a previous position.
    return value: bool
    --------------------------------------------------------------------------*/
    bool IsForMinimizeSpace()
    {
        return true;
    }
    /*!-------------------------------------------------------------------------
    function:     GetOneArchiveLog
    arguments:    logno [in]
    return value: ArchiveLog handle
    --------------------------------------------------------------------------*/
    Log_ArchiveLog& GetOneArchiveLog (SAPDB_UInt logno)
    {
        return *(m_ArchiveLogs[logno]);
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Information                                               */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    function:     GetArchiveLog()
    description:  Returns archive log for specified task

    arguments:    taskid [in]  id of the calling writer task
    return value: (Log_ArchiveLog&) reference to archive log
    --------------------------------------------------------------------------*/
    Log_ArchiveLog& GetArchiveLog
        (TaskID                         writertaskid)
    {
        SAPDBERR_ASSERT_STATE(m_ArchiveLogs.GetSize() > 0);
        
        ArchiveLogIterator iter = ArchiveLogsBegin();
        ArchiveLogIterator end  = ArchiveLogsEnd();
        while (iter != end)
        {
            ArchiveLog& archlog = **iter;
            if (archlog.GetWriterTaskID() == writertaskid)
                return archlog;
            ++iter;
        }

        // fatal error: no writer available for this writertaskid!!!
        // this code piece must not be reached!
        SAPDBERR_ASSERT_ARGUMENT(writertaskid);
        return (**iter);
    }

private:

    /*-------------------------------------------------------------------------
    declaration: m_Allocator
    description: Base allocator used for all dependent memory allocation of
                 this template class. Note that this member data has to be the
                 first member data defined in this class, because the sequence
                 of the execution of the initialization lists depends on it!
    --------------------------------------------------------------------------*/
    RTEMem_AllocatorWrapper                             m_Allocator;
    /*--------------------------------------------------------------------------
    declaration: m_Instance
    description: pointer to singleton instance 
    --------------------------------------------------------------------------*/
    static Log_ArchiveLogManager*                       m_Instance;
    /*--------------------------------------------------------------------------
    declaration: m_FrameAlloc
    description: allocator for page frames
    --------------------------------------------------------------------------*/
    FrameAllocator&                                     m_FrameAlloc;
    /*--------------------------------------------------------------------------
    declaration: m_ArchiveLogs
    description: vector of archive logs
    --------------------------------------------------------------------------*/
    ArchiveLogVector                                    m_ArchiveLogs;
    /*--------------------------------------------------------------------------
    declaration: m_DevspaceEnabled
    description: If set to false, GetQueue returns NULL. That means no log can be
                 written to the archive log device by user tasks.
    --------------------------------------------------------------------------*/
    bool                         m_DevspaceEnabled; // PTS 1113190 UH 2001-12-19
    
};
/*!*****************************************************************************

   endclass: Log_ArchiveLogManager

*******************************************************************************/



#endif  // LOG_ARCHIVELOGMANAGER_HPP


