/*!**************************************************************************

  module      : Log_History.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : The HistoryManager is a singleton.



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_HISTORY_H
#define Log_HISTORY_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h" // tsp00_TaskId
#include "ggg00.h"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "DataAccess/Data_Types.hpp"
#include "Logging/Log_HistoryFile.hpp"
#include "Logging/Log_HistoryDirectory.hpp"
#include "GarbageCollector/GC_IGarbageCollector.hpp" // PTS xxxxxxx FF 2002-02-12

class Log_UndoFile;
class tgg92_KernelOid;
class SAPDBMem_IRawAllocator;

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Log_History
    description: This is the history manager.
*/
class Log_History
{

public:
    typedef Log_HistoryDirectory::HistoryFileInfo HistoryFileInfo;

    /*! --------------------------------------------------------------------------------
    declaration: RemoveOptions
    description: Gives the granularity of history to be deleted.

    Normal:    "phase green"  Only easy to remove history is removed.
    Sensitive: "phase yellow" All history which can be is removed, this is more expensive.
    Truncate:  "phase red"    The oldest history is removed, even if it is needed by views.
    ----------------------------------------------------------------------------------*/
    enum  RemoveOptions
    {
        Normal, Sensitive, Truncate, RemoveUnused
    };
    /*! enddeclaration */

public:

    /*!-------------------------------------------------------------------------
    function:     CreateInstance
    arguments:    trans [in]
    description:  This method creates the singleton.
    returns:      true, if instance could be created
    --------------------------------------------------------------------------*/
    static bool CreateInstance (tgg00_TransContext &trans);

    /*--------------------------------------------------------------------------
    function:     DropInstance
    description:  this deallocates all allocated resources
    --------------------------------------------------------------------------*/
    static void DropInstance();

    /*--------------------------------------------------------------------------
    function:     ~Log_History
    description:  this calls Delete()
    --------------------------------------------------------------------------*/
    ~Log_History()
    {
        DropInstance();
    }

    /*! --------------------------------------------------------------------------------
    function:    Flush
    arguments:   trans   [in]
    returns:     new history directory root
    description: All data is flushed to disk. This is used by the savepoint.
    ----------------------------------------------------------------------------------*/
    Data_PageNo Flush (tgg00_TransContext &trans);

    /*! --------------------------------------------------------------------------------
    function:    Verify
    arguments:   trans [in]
                 isCold [in]
    returns:     (bool) returns true, if all files.Verify() return true.
    ----------------------------------------------------------------------------------*/
    bool Verify (tgg00_TransContext &trans,
                 bool                isCold);

    /*! --------------------------------------------------------------------------------
    function:    RegisterUndoFile
    arguments:   trans [in]
                 file  [in]
    description: A transaction calls this, if it produced before images
                 needed by other transactions to view data consistent.
    ----------------------------------------------------------------------------------*/
    void RegisterUndoFile ( tgg00_TransContext &trans,
                            Log_UndoFile       &file );

    /*! --------------------------------------------------------------------------------
    function:    GetBeforeImage [static]
    arguments:   many...
    description: this offers a before image mathing the given consistent view.
    ----------------------------------------------------------------------------------*/
    static void GetBeforeImage (tgg00_TransContext       &Trans,
                                const tgg92_KernelOid    &Oid,
                                const tgg91_PageRef      &WantedPageRef,
                                SAPDB_UInt4               BodySize,
                                SAPDB_Byte               *pObjBody,
                                bool                      bOnlyCheckExistence, // PTS 1113317 UH 2002-01-08
                                SAPDB_UInt4              &BodyLen,
                                tgg00_ObjFrameVers       &FrameVers,
                                Log_ActionType           &ActionType,
                                tgg91_TransNo            &ConsistentView,
                                tgg91_TransNo            &PrevUpdTrans,
                                tgg91_PageRef            &NextPageRef);

    /*! --------------------------------------------------------------------------------
    function:    BeforeImageExists [static]
    arguments:   TransContext [in]
    returns:     true, if the given Id references a valid before image.
    ----------------------------------------------------------------------------------*/
    static bool BeforeImageExists (tgg00_TransContext    &Trans,
                                   const tgg92_KernelOid &testoid,
                                   const tgg91_PageRef   &wanted)
    {
        SAPDB_UInt4        BodyLen;
        tgg00_ObjFrameVers FrameVers;
        Log_ActionType     ActionType = Log_NoOp;
        tgg91_TransNo      ConsistentView;
        tgg91_TransNo      PrevUpdTrans;
        tgg91_PageRef      NextPageRef;

        GetBeforeImage ( Trans, testoid, wanted, 0, NULL, true, // PTS 1113317 UH 2002-01-08
                         BodyLen, FrameVers, ActionType,
                         ConsistentView, PrevUpdTrans, NextPageRef );

        if ( e_ok != Trans.trError_gg00 ) 
        {
            if ( e_illegal_page_no != Trans.trError_gg00 )
                Trans.trError_gg00 = e_ok;
            return false;
        }
        
        return true;
    }

    /*! --------------------------------------------------------------------------------
    function:    IsHistoryUsed [static]
    arguments:   TaskId   [in]
    returns:     true, if is used.
    description: 
    ----------------------------------------------------------------------------------*/
    static bool IsHistoryUsed (tsp00_TaskId        taskid,
                               tgg92_KernelOid    &oid,
                               tgg00_ObjTransInfo &info);

    /*! --------------------------------------------------------------------------------
    function:    DeleteBeforeImage [static]
    arguments:   TransContext [in]
                 oid          [in]
                 pageref      [in]
    description: TrError is set
    ----------------------------------------------------------------------------------*/
    static void DeleteBeforeImage (tgg00_TransContext    &trans,
                                   const tgg92_KernelOid &oid,
                                   const tgg91_PageRef   &pageref);

    /*--------------------------------------------------------------------------
    function:     GetInstance
    return value: The history manager singleton is returned.
    description:  It is assumed, that the component is initialized.
                  
                  +++ THIS IS PUBLIC UNTIL THERE ARE EXISTING DIFFERENT INTERFACES +++
    --------------------------------------------------------------------------*/
    static Log_History& GetInstance()
    {
        SAPDBERR_ASSERT_STATE (m_Instance != NULL);
        return( *m_Instance );
    }

    /*--------------------------------------------------------------------------
    function:     IsInitialized
    return value: true, if GetInstance() can be used.
    --------------------------------------------------------------------------*/
    static bool IsInitialized()
    {
        return m_Instance != NULL;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetNumberOfHistoryFiles
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt GetNumberOfHistoryFiles () const
    {
        return m_Directory.GetFileCount();
    }
                   
    /*! --------------------------------------------------------------------------------
    function:    GetNumberOfMaxUsedHistoryFiles
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt GetNumberOfMaxUsedHistoryFiles () const
    {
        return m_Directory.GetMaxUsedFileCount();
    }
                   
    /*! --------------------------------------------------------------------------------
    function:    GetOldestTransNo
    arguments:   fileno [in]
    returns:     oldest transno
    ----------------------------------------------------------------------------------*/
    const tgg91_TransNo & GetOldestTransNo (SAPDB_UInt fileno) const
    {
        return m_Directory.GetOldestTransNo (fileno);
    }
                   
    /*! --------------------------------------------------------------------------------
    function:    GetInfo
    arguments:   TaskId [in]
                 fileno [in]
    returns:     info
    description: fileno identifies a history file.
                 
                 The range is between 0 and (GetNumberOfHistoryFiles() - 1)
    ----------------------------------------------------------------------------------*/
    void GetInfo ( tsp00_TaskId     taskid,
                   SAPDB_UInt       fileno,
                   HistoryFileInfo &info,
                   SAPDB_UInt      &pagecount ) const
    {
        m_Directory.GetHistoryFileInfo( taskid, fileno, info, pagecount );
    }
                   
    /*! --------------------------------------------------------------------------------
    function:    RemoveHistory
    arguments:   TaskId       [in]
                 fileno       [in]
                 RemoveOption [in]
                 anythingRemoved [out]
    description: This call removes history depending on the given options.

                 The range is between 0 and (GetNumberOfHistoryFiles() - 1)
    ----------------------------------------------------------------------------------*/
    void RemoveHistory (tgg00_TransContext        &trans,
                        SAPDB_UInt                 fileno,
                        RemoveOptions              option,
                        GC_IGarbageCollector      &IGarbColl,                    //PTS xxxxxxx FF 2002-02-12 
                        bool                      &anythingRemoved);

    /*! --------------------------------------------------------------------------------
    function:    RemoveCompleteHistory
    arguments:   trans [in/out]
    description: This method removes the complete history including the history directory.
                 The anchor in the restartrecord is deleted, even the memory is deallocated.
                 After this method returns the history management must be created new.
                 This can be used for migration prupose.
    ----------------------------------------------------------------------------------*/
    void RemoveCompleteHistory (tgg00_TransContext &trans);

    /*! --------------------------------------------------------------------------------
    function:    CheckAndRemoveUnusedHistory
    arguments:   trans     [in/out]
                 IGarbColl [in/out]
    description: This method removes history files which are not used if there
                 are any existing.
                 History files can be unused if MAXUSERTASKS has been decreased.
    ----------------------------------------------------------------------------------*/
    void CheckAndRemoveUnusedHistory ( tgg00_TransContext   &trans,
                                       GC_IGarbageCollector &IGarbColl );

    /*! --------------------------------------------------------------------------------
    function:    GetDumpInfo
    arguments:   fileno       [in]
    returns:     void*
    description: this is used by garbage collector dump bd930DumpHistoryFileDir()
    ----------------------------------------------------------------------------------*/
    const void* GetDumpInfo (SAPDB_UInt fileno)
    {
        return m_Directory.GetDumpInfo(fileno);
    }

    /*! --------------------------------------------------------------------------------
    function:    AnyHistoryExists
    returns:     bool
    description: returns true, if any history directory entry has a valid transno.
                 Dirty Read / Outside region !
    ----------------------------------------------------------------------------------*/
    bool AnyHistoryExists (); // PTS 1114877 UH 2002-03-18 new
    
#ifdef CONSTRUCTOR_MAY_NOT_BE_PRIVATE_AND_NO_FRIEND_DEFINED
protected:
#else
private:
#endif
    // forbidden to use, because this is a singleton
    Log_History (const Log_History& other);
    Log_History& operator= (const Log_History &Manager);
                     
#ifdef CONSTRUCTOR_MAY_NOT_BE_PRIVATE_AND_NO_FRIEND_DEFINED
private:
#endif

    /*--------------------------------------------------------------------------
    function:     Log_History
    description:  the constructor for the singleton
    --------------------------------------------------------------------------*/
    Log_History (SAPDBMem_IRawAllocator &allocator,
                 SAPDB_UInt              numUKTs,
                 SAPDB_UInt              maxFiles)
    : m_Directory (allocator, numUKTs, maxFiles)
    {}

    /*--------------------------------------------------------------------------
    function:     CreateHistoryDirectory
    arguments:    trans       [in]
                  histdirroot [in]
    description:  This creates Historyfiles and initializes the directory.
    --------------------------------------------------------------------------*/
    void CreateHistoryDirectory (tgg00_TransContext &trans);

    /*--------------------------------------------------------------------------
    function:     CreateHistoryFiles
    arguments:    trans       [in]
                  dirfile     [in]
                  firstFileNo [in]
                  lastFileno  [in]
    description:  This creates Historyfiles.
    --------------------------------------------------------------------------*/
	void CreateHistoryFiles (tgg00_TransContext   &trans,
							 Log_HistoryDirectory &dirfile,
							 SAPDB_UInt            firstFileNo,
							 SAPDB_UInt            lastFileNo);
    
    /*--------------------------------------------------------------------------
    function:     RestoreHistoryDirectory
    arguments:    trans       [in]
    description:  This restores the directory.
    --------------------------------------------------------------------------*/
    void RestoreHistoryDirectory (tgg00_TransContext &trans);

    /*--------------------------------------------------------------------------
    function:     UpdateHistoryFileInfo
    arguments:    trans         [in]
                  historyFileNo [in]
                  fileinfo      [in/out]
    returns:      (bool) true, if fileinfo has changed
    description:  This restores the directory.
    --------------------------------------------------------------------------*/
    bool UpdateHistoryFileInfo ( tgg00_TransContext &trans,
                                 SAPDB_UInt          historyFileNo,
                                 HistoryFileInfo    &fileinfo,
                                 SAPDB_UInt         &pagecount );
                                 
    /*--------------------------------------------------------------------------
    function:     RemoveUndoFileWithObjects
    arguments:    trans         [in]
                  root          [in]
                  last          [in]
                  isInterrupted [in] it becomes true, if the garbage collection should be aborted.
	returns:      bool - true, if UndoFile was really removed. // PTS 1113934 UH 2002-02-01
    description:  this removes an undo file an all objects which are deleted in it.
    --------------------------------------------------------------------------*/
    bool RemoveUndoFileWithObjects (tgg00_TransContext        &trans,                      // PTS 1113934 UH 2002-02-01
                                    Data_PageNo                root,
                                    Data_PageNo                last,
                                    GC_IGarbageCollector      &IGarbColl,                  // PTS xxxxxxx FF 2002-02-12 
                                    Log_HistoryFile           &historyfile,                // PTS 1113934 UH 2002-02-01
                                    Log_HistoryFile::Iterator &undoFileInfoIter);          // PTS 1113934 UH 2002-02-01

    /*--------------------------------------------------------------------------
    function:     TruncateHistory
    arguments:    trans    [in]
                  root     [in]
                  anythingRemoved [out]
    --------------------------------------------------------------------------*/
    void TruncateHistory (tgg00_TransContext        &trans,
                          SAPDB_UInt                 fileno,
                          Data_PageNo                root,
                          Data_PageNo                last,
                          GC_IGarbageCollector      &IGarbColl,                            // PTS xxxxxxx FF 2002-02-12 
                          bool                      &anythingRemoved);
                          
    /*--------------------------------------------------------------------------
    function:     DetermineOldestTransNo
    arguments:    Log_HistoryFile [in]
    returns:      oldest transno [tgg91_TransNo]
    --------------------------------------------------------------------------*/
    tgg91_TransNo DetermineOldestTransNo (Log_HistoryFile &file);

public:

    /*! --------------------------------------------------------------------------------
    declaration: HistoryDirectory
    description: This defines the directory in main memory.
    ----------------------------------------------------------------------------------*/
    class HistoryDirectory
    {
    public:
        HistoryDirectory ( SAPDBMem_IRawAllocator& allocator,
                           SAPDB_UInt              numUKTs,
                           SAPDB_UInt              maxFiles )
        : m_Directory (allocator),
          m_numUKTs   (numUKTs),
          m_maxFiles  (maxFiles)
        {}
        
        bool Initialize()
        {
            return m_Directory.Initialize (GetMaxUsedFileCount());
        }
        
        void Delete()
        {
            m_Directory.Delete();
        }
        
        ~HistoryDirectory()
        {
            Delete();
        }
        
        SAPDB_UInt GetFileCount() const
        {
            return m_Directory.GetSize();
        }

        SAPDB_UInt GetMaxUsedFileCount() const
        {
            return m_maxFiles;
        }

        SAPDB_Int GetFileNo(tsp00_TaskId taskid) const
        {
            return taskid % GetMaxUsedFileCount();
        }

        const tgg91_TransNo & GetOldestTransNo (SAPDB_UInt fileno) const
        {
            return m_Directory[fileno].persistent.oldestTransNo;
        }

        void SetHistoryFileInfo ( tsp00_TaskId           taskid,
                                  SAPDB_UInt             fileno,
                                  const HistoryFileInfo& info,
                                  SAPDB_UInt             pagecount );

        void GetHistoryFileInfo ( tsp00_TaskId     taskid,
                                  SAPDB_UInt       fileno,
                                  HistoryFileInfo &info,
                                  SAPDB_UInt      &pagecount ) const;

        void UpdateInfoAfterRegisterNewUndoFile ( tsp00_TaskId         taskid,
                                                  SAPDB_UInt           fileno,
                                                  const tgg91_TransNo &newtransno,
                                                  SAPDB_UInt           pagecount,
                                                  Data_PageNo          last );

        void UpdateOldestTransNo ( tsp00_TaskId         taskid,
                                   SAPDB_UInt           fileno,
                                   const tgg91_TransNo &newOldestTransno,
                                   Data_PageNo          last,
                                   SAPDB_UInt           removedPages );
    
        const void* GetDumpInfo (SAPDB_UInt fileno)
        {
            return &(m_Directory[fileno]);
        }
        
        /*--------------------------------------------------------------------------
        function:     GetNextUnusedFileNo
        arguments:    taskid [in]
                      fileno [in/out]
        returns:      bool
        description:  The next fileno of an unused history file is given.
                      The history directory entry is marked as used so no other
                      requester will ever get this entry again.
                      Usually fileno is set to 0 at the beginning.
                      If false is returned there exists no file which is unused.
        --------------------------------------------------------------------------*/
        bool GetNextUnusedFileNo ( tsp00_TaskId  taskid,
                                   SAPDB_UInt   &fileno );
    
        void RemoveUnusedEntries ();

        void ResetInProcess (tsp00_TaskId taskid,
                             SAPDB_UInt   fileno);

        void Invalidate (tsp00_TaskId taskid,
                         SAPDB_UInt   fileno);

        void WriteToTrace (tsp00_TaskId      taskid,
                           const SAPDB_Char *title = NULL) const;

    public:

        struct ExtendedHistoryFileInfo
        {
            HistoryFileInfo persistent;
            SAPDB_UInt4     pagecount;
            SAPDB_Bool1     inProcess;
            SAPDB_Bool1     filler1;
            SAPDB_UInt2     filler2;
            ExtendedHistoryFileInfo()
            :pagecount(0),
             inProcess(false),
             filler1(0),
             filler2(0)
            {}
        };

    private:
    
        typedef Container_Vector<ExtendedHistoryFileInfo> Directory;
        Directory        m_Directory;
        const SAPDB_UInt m_numUKTs;
        const SAPDB_UInt m_maxFiles;
    };
    
    typedef HistoryDirectory::ExtendedHistoryFileInfo ExtendedHistoryFileInfo;

    /*! --------------------------------------------------------------------------------
    variable:    m_Directory
    ----------------------------------------------------------------------------------*/
    HistoryDirectory m_Directory;
    
    /*! --------------------------------------------------------------------------------
    variable:    m_DirectoryRoot
    ----------------------------------------------------------------------------------*/
    Data_PageNo m_DirectoryRoot;
    
private:

    /*! --------------------------------------------------------------------------------
    variable:    m_Instance
    ----------------------------------------------------------------------------------*/
    static Log_History* m_Instance;

};
/*! endclass: Log_History */

/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/

#endif  /* Log_HISTORY_H */
