
/*!**************************************************************************

    module      : Log_IAction.hpp

    -------------------------------------------------------------------------

    responsible : UweH

    special area: Logging
    description : defines the class Log_IAction

    last changed: 2001-02-10

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_IAction_H
#define Log_IAction_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "ggg00.h"

class   SAPDBMem_IRawAllocator;
class   Data_SplitSpaceWriter;
class   Data_SplitSpaceReader;

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


/*! --------------------------------------------------------------------------------
class:        Log_IAction
description : Baseclass for all actions.

- Actions define an operation - something, which ist executed.
- The persistent format describes what is to be done.
- To undo an action, the inverse action must be created and written to an undo file.
*/
class Log_IAction
{

public:

    /*  --------------------------------------------------------------------------------
    function:    Log_IAction
    */                         
    Log_IAction()
    : m_Type (Log_NoOp)
    {}
    
    /*  --------------------------------------------------------------------------------
    function:    GetType
    returns:     type of action
    */                         
    Log_ActionType GetType () const
    {
        return m_Type;
    }

    /*  --------------------------------------------------------------------------------
    function:    ForceLogWriting
    returns:     bool
    description: Some special actions need to be written with no restrictions
                 like LOGFULL. These entries may return true.
    */                         
	// PTS 1117320 UH 2002-10-28 added
    virtual bool ForceLogWriting () const
    {
        return false;
    }

    /*  --------------------------------------------------------------------------------
    function:    WaitForLogWriting
    returns:     bool
    description: Some special actions need a log position. These entries may return true.
                 e.g. commit, rollback, savepoint
    */                         
	// PTS 1117320 UH 2002-10-28 added
    virtual bool WaitForLogWriting () const
    {
        return false;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetMinimalPersistentLength
    returns:     SAPDB_UInt The persistent length of the actions image,
                 which must not be split.
    */
    virtual SAPDB_UInt GetMinimalPersistentLength () const = 0;

    /*! --------------------------------------------------------------------------------
    function:    GetPersistentLength
    returns:     SAPDB_UInt The persistent length of the actions image .
    */
    virtual SAPDB_UInt GetPersistentLength () const = 0;

    /*! --------------------------------------------------------------------------------
    function:    WritePersistentFormat
    arguments:   Writer [in/out]
                 isOK   [out]
    description: Everything is written which is needed to re-execute this action.
    */
    virtual void WritePersistentFormat (Data_SplitSpaceWriter &Writer,
                                        bool                  &isOK) const = 0;

    /*! --------------------------------------------------------------------------------
    function:    ReadPersistentFormat
    arguments:   Reader [in/out]
                 isOK   [out]
    description: Everything is read which is needed to re-execute this action.
    */
    virtual void ReadPersistentFormat (Data_SplitSpaceReader  &Reader,
                                       SAPDBMem_IRawAllocator &Allocator,
                                       bool                   &isOK) = 0;
                             
    //  --------------------------------------------------------------------------------
    /*! function:    Execute
    
        description: executes the action

        arguments:   TransContext [in out]
    */
    virtual void Execute (tgg00_TransContext &TransContext) const
    {}
      
    /*  --------------------------------------------------------------------------------
    function:    WriteToTrace
    arguments:   Title [in]
    description: The action is written to the trace file
    */    
    virtual void WriteToTrace (const char * Title) const
    {}

protected:

    /*! --------------------------------------------------------------------------------
    variable:    m_Type
    */
    Log_ActionType m_Type;

};
/*! endclass: Log_IAction */
// PTS 1117320 UH 2002-10-28 added
// removed Log_NoAction
#endif  /* Log_IAction_H */
