
/*!**************************************************************************

    module      : Log_IRedoTrafficControl.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines an interface class of a singleton to synchronize the parallel redo

    last changed: 2001-03-21

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_IREDO_TRAFFIC_CONTROL_H
#define LOG_IREDO_TRAFFIC_CONTROL_H


/*===========================================================================*
*  INCLUDES                                                                  *
*============================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_Transaction.hpp"
#include "gsp00.h"
#include "ggg00.h"


/*===========================================================================*
*  STRUCTURES, TYPES, UNIONS ...                                             *
*============================================================================*/

typedef SAPDB_UInt4 Log_IoSeqUInt;


/*===========================================================================*
*  CLASSES                                                                   *
*============================================================================*/

//=======================================================================================
/*! class:       Log_IRedoTrafficControl

    description: interface class of a singleton to synchronize the parallel redo
*/
class Log_IRedoTrafficControl
{
protected:
    
    static Log_IRedoTrafficControl *m_pInstance;
    
public:

    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:     ExecuteJobs
        
        description:  Executes vacant jobs. This method is provided for redo tasks.

        argument:     RedoTaskId [in]
    */
    virtual void ExecuteJobs (tgg00_TransContext &TransContext) = 0;


    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:     Instance
    
        return value: the reference to the global instance of the singleton
    */
    static Log_IRedoTrafficControl &Instance ()
    { 
        return (*m_pInstance);
    }

    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:    InsertJob
    
        description: Inserts a job. This method is provided for the log reader task.

        arguments:   LogReaderTaskId [in]
                     TransIndex      [in]
                     EndOfTransIoSeq [in]
                     UndoOrRedo      [in]
                     checkAscendingSequence [in]
                     pTransaction    [in]
                     IsOk            [out]
    */
    virtual void InsertJob (
        tsp00_TaskId      LogReaderTaskId,
        tgg00_TransIndex  TransIndex,
        Log_IoSeqUInt     EndOfTransIoSeq,
        Log_ActionType    UndoOrRedo,
        bool              checkAscendingSequence, // PTS 1113641 UH 2002-01-25
        Log_Transaction  &Transaction,
        bool             &IsOk)          = 0;


    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:    SetEndOfLogRead
    
        description: the log reader task sets the state: end of log read

        argument:    LogReaderTaskId [in]
    */
    virtual void SetEndOfLogRead (tsp00_TaskId LogReaderTaskId) = 0;

    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:     CheckIoSeq
        description:  a redo task checks if preceding transactions are closed
        arguments:    RedoTaskId [in]
                      TransIndex [in] reference to the transaction handled by the redo task  
                      IoSeq      [in] sequence of the log entry handled by the redo task 
    */
    virtual bool CheckIoSeq ( tsp00_TaskId     RedoTaskId,
                              tgg00_TransIndex TransIndex,
                              Log_IoSeqUInt    IoSeq )     = 0;

    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:     WaitForIoSeq
        description:  a redo task waits until preceding transactions are closed 
        arguments:    RedoTaskId [in]
    */
    virtual void WaitForIoSeq (tsp00_TaskId RedoTaskId) = 0;
    /// The log reader calls this after recreation of all open transactions from last savepoint
    virtual void EnableRedoProcessing (tsp00_TaskId LogReaderTaskId) = 0;
	/// writes some basic information into the trace 
	virtual void WriteToTrace (const char *title = 0) const = 0;
	/// This returns the youngest iosequence inserted to the redo list
	virtual Log_IOSequenceNo GetLastIOSequence (tsp00_TaskId taskid) const = 0;
};
/*! endclass: Log_IRedoTrafficControl */

#endif  /* LOG_IREDO_TRAFFIC_CONTROL_H */
