/*!**************************************************************************

  module      : Log_ISave.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines the classes Log_ISave



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_ISave_H
#define Log_ISave_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h" // tsp00_TaskId, tsp00_PageNo, tsp00_Line
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_Types.hpp" 
#include "Logging/Log_SaveIterator.hpp"
/* #include "IOManager/IOMan_ILogIO.hpp" */
/* #include "KernelCommon/Kernel_Page.hpp" */

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/


/*! ----------------------------------------------------------------------
class:      	Log_ISave
description: 	This is the Interface for the save manager.
				Log_ISave works only on one LogDevice, because
				each LogDevice may be saved independent from each other.
				Log_ISave is stateless (i.e. only the actual state of the
				LogDevice is used). But state-bound objects may be retrieved
				in order to allow operations with consistent states
				(e.g. iterator for accessing the pages to be saved).
*/
class Log_ISave
{
public:
    /*! ----------------------------------------------------------------------
    function:    	GetNumberOfUnsavedPages
    returns:     	DeviceBlockCount
    description: 	Returns the number of log pages on the associated LogDevice
	 				that are actual not saved.
    */
    virtual Log_DeviceBlockCount GetNumberOfUnsavedPages () const = 0;


    /*! ----------------------------------------------------------------------
    function:    	ResetLogFull
    arguments:   	TaskId  [in]
    description: 	Checks log if log is full and if not true, wake up all
					waiting transactions.
    */
    virtual void ResetLogFull (tsp00_TaskId TaskId) = 0;


    /*! ----------------------------------------------------------------------
    function:    	FreeLog
    arguments:   	TaskId  [in]
                 	LogNo   [in]
    description: 	Releases the Pages defined by LogNo
					This method is for user purposes and performs some
					checks befor calling UnusedLogDelete
    */
    virtual void FreeLog (	tsp00_TaskId     	TaskId,
							Log_IOSequenceNo	FirstToFree,
							Log_IOSequenceNo	LastToFree,
							Log_RawDeviceOffset	LastPageToFree) = 0;

    /*! ----------------------------------------------------------------------
    function:    	CheckRestoreLog
    arguments:   	TaskId               [in]
                 	FirstLogpageFromSave [in]
                 	LastLogpageFromSave  [in]
                 	DatabaseIdentifier   [in]
    returns:     	true, if the given restore parameter can be restored.
    description: 	During Restore Log this checks, if the LogpageRange and the
		 			DatabaseIdentifier is valid.
    */
    virtual bool CheckRestoreLog (	tsp00_TaskId      	TaskId,
									Log_IOSequenceNo	FirstToSave,
									Log_IOSequenceNo	LastToSave,
									const tsp00_Line &	DatabaseIdentifier) = 0;

    /*! ----------------------------------------------------------------------
    function:    	CheckSaveLogAllowed
    returns:     	true, if current state allows log save.
					false if Archive_log does not exists or backup_history
					is lost
    description: 	-
    */
    virtual bool CheckSaveLogAllowed () = 0;

    /*! ----------------------------------------------------------------------
    function:    	BackupHistoryIsLost
    returns:     	bool
    description: 	true, if it is.
    */
    virtual bool BackupHistoryIsLost() = 0;


    /*! ----------------------------------------------------------------------
    function:    	InitSaveIter
    arguments:   	iter			  [out] the iterator, which will be
											initialized according to the
											current ArchiveLog
					TaskId            [in]
                 	IsCold            [in]  determines whether last log_page
											may be saved
    returns:     	nothing

    description: 	determines the current state of the archive_log and
					save the states in the iterator for further processing
					without concurrent state changes
    */

    virtual void InitSaveIter(	Log_SaveIterator	&iter,
								tsp00_TaskId    	TaskId,
								bool              	IsCold) = 0;


    /*! ----------------------------------------------------------------------
    function:    	InitAutoSaveIter
    arguments:   	iter			  [out] the iterator, which will be
											initialized according to the
											current ArchiveLog
    			   	TaskId            [in]
    returns:     	bool          			TRUE, if enough pages has to
											be saved

    description: 	determines the current state of the archive_log and
					save the states in the iterator for further processing
					without state changes.
    */
    virtual bool InitAutoSaveIter(	Log_SaveIterator	&iter,
									tsp00_TaskId		TaskId) = 0;


	/*! ------------------------------------------------------------------
	function:    	UnusedLogDelete
	arguments:   	TaskId                 	[in]
					WriteLogInfo           	[in]
					LastPageNumberToDelete 	[in]


	description: 	This specifies, where the archive log on the device should
					start (i.e. the pages from FirstNotSaved to
					LastPageNumberToDelete will be available for
					logging)
					If WriteLogInfo is set to TRUE then the LogInfoPage
					will be updated.
	*/

protected:
    virtual void UnusedLogDelete (tsp00_TaskId      TaskId,
                                  bool              WriteLogInfo,
                                  tsp00_PageNo      LastPageNumberToDelete) = 0;



};
/*! endclass: Log_ISave */



/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/

#endif  /* Log_ISave_H */





				































