/*!**************************************************************************

  module      : Log_ISequentialReader.hpp
  special area: Logging
  responsible : MartinB
  last changed: 2001-07-20  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines class Log_ISequentialReader



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_ISequentialReader_H
#define Log_ISequentialReader_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h" // tsp00_TaskId, tsp00_PageNo, tsp00_Line
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

 
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_ISequentialReader
description: virtual base class for implementing a reader for the ReDoReader
*/

class Log_ISequentialReader 
{

friend class Log_ArchiveLogIterator;

public:
    /*! --------------------------------------------------------------------------------
    declaration:  ReaderStatus
    */
    enum ReaderStatus {
        reader_ok,
        end_of_file,
        invalid_IOSeq,
        invalid_Offset,
        not_implemented        
    };
    
    /*!-------------------------------------------------------------------------
    function:     SetIOSeq 
    description:  sets the actual actual read position according to the given IOSequence
                  returns not_implemented as default, beacause not all devices supports
                  such a set-operation 
    arguments:    ioSeq          ioSeq, which shall be read next 
    return value: ReaderStatus
    --------------------------------------------------------------------------*/
    virtual ReaderStatus  SetIOSeq(   Log_IOSequenceNo         ioSeq)
    {
        return not_implemented;
    }
    
    /*!-------------------------------------------------------------------------
    function:     SetPosition 
    description:  sets the actual page and IOSeq according to the given offset
                  returns not_implemented as default, beacause not all devices supports
                  such a set-operation 
    arguments:    offset         offset, which shall be read next
    return value: ReaderStatus
    --------------------------------------------------------------------------*/
    virtual ReaderStatus  SetPosition(   Log_RawDeviceOffset      offset)
    {
        return not_implemented;
    }
    
    /*!-------------------------------------------------------------------------
    function:     GetPosition 
    description:  gets the actual offset of the last page read (invalid for a tape)
                  returns the InvalidValue if the device doesn't support a position
    arguments:    none
    return value: the offset of the last page read
    --------------------------------------------------------------------------*/
    virtual Log_RawDeviceOffset  GetPosition() const
    {
        return Log_RawDeviceOffset();
    }

protected:
    /*!-------------------------------------------------------------------------
    function:     ReadNextPage 
    description:  reads the actual page and then increments the actual position 
    arguments:    taskId        identification of the running task
                  logPage       location where the read page will be stored
    return value: ReaderStatus
    --------------------------------------------------------------------------*/
    virtual ReaderStatus  ReadNextPage(  tsp00_TaskId             taskId,
                                         Log_Page                &logPage) = 0;
    
};
/*! endclass: Log_ISequentialReader */

#endif  /* Log_ISequentialReader_H */
