/*!**************************************************************************

  module      : Log_IWritingRedoEntries.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines the classes Log_IWritingRedoEntries and Log_RedoEntry
  
  This is an interface between the archive log manager and a transaction.



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_IWritingRedoEntries_H
#define Log_IWritingRedoEntries_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h" // tsp00_TaskId
#include "ggg00.h" // tgg00_TransContext

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Logging/Log_AfterImage.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_WaitForFlush
description: This handle is returned from WriteEndOfTransLogentry.

             It can be used to wait for the archive log to write out all logentries (Flush)
             for the current transaction.
*/

class Log_WaitForFlush
{
public:
    /*! --------------------------------------------------------------------------------
    function:    Log_WaitForFlush
    arguments:   FlushAll [in]
    description: The FlushHandle controls the amount of log, which is written to the device now.
    
                 If FlushAll is true, by default it is not, then this handle is
                 used to force the Archive Log Manager to flush the complete log queue.
                 So every log entry is written to the device. This is used by EndOfTransaction
                 or by a savepoint.
    */
    Log_WaitForFlush (bool FlushAll = false);
};
/*! endclass: Log_WaitForFlush */

/*! --------------------------------------------------------------------------------
class:       Log_IWritingRedoEntries
description: This is the Interface for local transactions writing redo log.
*/

class Log_IWritingRedoEntries
{
public:

    /*! --------------------------------------------------------------------------------
    function:    WriteAfterImage
    arguments:   TaskId [in]
                 Action [in]
    description: The after image is written of the given action.
    */
    virtual void WriteAfterImage (tsp00_TaskId          TaskId,
                                  const Log_AfterImage &AfterImage) = 0;
                                
    /*! --------------------------------------------------------------------------------
    function:    WriteEndOfTrans
    arguments:   TransContext          [in]
                 IsCommitted           [in]
                 WaitForFlush          [out]
                 TryCheckptOrSavepoint [out]
    description: The EOT marker is written into the archive log device.
    */
    virtual void WriteEndOfTrans (tgg00_TransContext& Trans,
                                  bool                IsCommitted,
                                  Log_WaitForFlush&   WaitForFlush,
                                  bool&               TryCheckptOrSavepoint) = 0;

    /*! --------------------------------------------------------------------------------
    function:    Flush
    arguments:   TaskId       [in]
                 WaitForFlush [in]
    description: This call flushes all logentries of the transaction, which finished before.
    */
    virtual void Flush (tsp00_TaskId            TaskId,
                        const Log_WaitForFlush& WaitForFlush) = 0;

protected:

    /*! --------------------------------------------------------------------------------
    class:       ActionEndOfTrans
    description: This Entry is written at transactions end.
    */

    class ActionEndOfTrans : public Log_AfterImage
    {
    public:

        /*! --------------------------------------------------------------------------------
        function:    EndOfTransLogentry
        arguments:   Trans       [in/out]
                     IsCommitted [in]
        description: This constructs an end-of-transaction logentry.
        */
        ActionEndOfTrans (tgg00_TransContext& Trans,
                          bool                IsCommitted);
                            
    };
    /*! endclass: EndOfTransLogentry */
};
/*! endclass: Log_IWritingRedoEntries */


/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/
                                         
#endif  /* Log_IWritingRedoEntries_H */
