/*!*****************************************************************************

  module:       Log_InfoPage.hpp

  ------------------------------------------------------------------------------

  responsible:     UweH

  original author: TillL

  special area: Logging

  description:  Handler class for log info page

  see also:     Kernel_IPage.html 
                Kernel_Page.html
                Kernel_Page74.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_INFOPAGE_HPP
#define LOG_INFOPAGE_HPP


#include <memory.h>

#include "gkb00.h"
#include "hgg08.h"  // g08loginfo
#include "heo55k.h" // vbegexcl/vendexcl/visexcl
#include "KernelCommon/Kernel_Page74.hpp"
#include "KernelCommon/Kernel_DateTime.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "Logging/Log_Types.hpp"



/*!*****************************************************************************

   class: Log_InfoPage

   description: 
   
   Handler class for log info pages
                
*******************************************************************************/
class Log_InfoPage : public Kernel_Page74
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : PageID
    description:  log info page id
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt4                                     PageID;
    /*!-------------------------------------------------------------------------
    declaration : PageOffset
    description:  offset of an entry within a page
    --------------------------------------------------------------------------*/
    typedef Log_Page::OffsetType                            PageOffset;
    /*!-------------------------------------------------------------------------
    declaration : Queue Count
    description:  any number of log queues
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt2                                     QueueCount;
    /*!-------------------------------------------------------------------------
    declaration:  IOSequenceNo
    description:  i/o sequence number of log pages
    --------------------------------------------------------------------------*/
    typedef Log_IOSequenceNo                                IOSequenceNo;
    /*!-------------------------------------------------------------------------
    declaration:  RawDeviceOffset
    description:  raw offset on log device
    --------------------------------------------------------------------------*/
    typedef Log_RawDeviceOffset                             RawDeviceOffset;
    /*!-------------------------------------------------------------------------
    declaration : DeviceBlockCount
    description:  number of blocks on a raw log device
    --------------------------------------------------------------------------*/
    typedef Log_DeviceBlockCount                            DeviceBlockCount;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Log_InfoPage()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Log_InfoPage();

    /*!-------------------------------------------------------------------------
    function:     Log_InfoPage()
    description:  Constructor; assigns frame

                  - Does not change frame contents.
                  - It is assumed that the frame contains a valid log info page.

    arguments:    frame [in]    page frame
    return value: none
    --------------------------------------------------------------------------*/
    Log_InfoPage
        (const PageFrame&               frame);
   
    /*!-------------------------------------------------------------------------
    function:     Log_InfoPage()
    description:  Constructor; assigns frame and initializes frame contents

                  - It is assumed that the frame contains an empty page
                    which may be overwritten.

    arguments:    frame    [in]  page frame
                  pageid   [in]  log info page id
    return value: none
    --------------------------------------------------------------------------*/
    Log_InfoPage
        (const PageFrame&               frame, 
         const PageID&                  pageid);
   
    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame    [in]  page frame
                  pageid   [in]  log info page id
    return value: none
    --------------------------------------------------------------------------*/
    void SetAndInitializeFrame
        (const PageFrame&               frame, 
         const PageID&                  pageid);

    /*!-------------------------------------------------------------------------
    function:     InitializeFrame()
    description:  Initializes frame contents
    arguments:    pageid   [in]  log info page id
    return value: none
    --------------------------------------------------------------------------*/
    void InitializeFrame
        (const PageID&                  pageid);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Synchronization                                              */
    /*!-------------------------------------------------------------------------
    function:     EnterRegion()
    description:  enter the region for log info page changes
    arguments:    taskid
    return value: none
    --------------------------------------------------------------------------*/
    void EnterRegion(tsp00_TaskId taskid)
    {
        vbegexcl (taskid, g08loginfo);
    }
    /*!-------------------------------------------------------------------------
    function:     LeaveRegion()
    description:  leaves the region for log info page changes
    arguments:    taskid
    return value: none
    --------------------------------------------------------------------------*/
    void LeaveRegion(tsp00_TaskId taskid)
    {
        vendexcl (taskid, g08loginfo);
    }
    /*!-------------------------------------------------------------------------
    function:     InRegion()
    description:  returns true if in log info page region
    arguments:    taskid
    return value: bool
    --------------------------------------------------------------------------*/
    bool InRegion(tsp00_TaskId taskid)
    {
        return visexcl (taskid, g08loginfo);
    }
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Synchronization                                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /*!-------------------------------------------------------------------------
    function:     ReadPageID()
    description:  Reads log info page id from frame
    arguments:    none
    return value: (PageID) page id
    --------------------------------------------------------------------------*/
    PageID ReadPageID() const;

    /*!-------------------------------------------------------------------------
    function:     ReadLogDate()
    description:  Reads log date from frame
    arguments:    none
    return value: (Kernel_Date) log date
    --------------------------------------------------------------------------*/
    Kernel_Date ReadLogDate() const;

    /*!-------------------------------------------------------------------------
    function:     ReadLogTime()
    description:  Reads log time from frame
    arguments:    none
    return value: (Kernel_Time) log time
    --------------------------------------------------------------------------*/
    Kernel_Time ReadLogTime() const;

    /*!-------------------------------------------------------------------------
    function:     ReadFlushMode()
    description:  Reads log flush mode from frame
    arguments:    none
    return value: (Log_FlushMode) log mode
    --------------------------------------------------------------------------*/
    Log_FlushMode ReadFlushMode() const;

    /*!-------------------------------------------------------------------------
    function:     ReadDeviceState()
    description:  Reads redo until state from frame
    arguments:    none
    return value: (Log_DeviceState) log redo until state
    --------------------------------------------------------------------------*/
    Log_DeviceState ReadDeviceState() const;

    /*!-------------------------------------------------------------------------
    function:     ReadDevspaceEnabled()
    description:  Reads if archive log writing is enabled
    arguments:    none
    return value: (bool) true if after images are written else false
    --------------------------------------------------------------------------*/
    bool ReadDevspaceEnabled() const;

    /*!-------------------------------------------------------------------------
    function:     ReadAutoOverwrite()
    description:  Reads if the archive log can be automatically been overwritten
                  Returns true if backups are not necassary.
    arguments:    none
    return value: (bool)
    --------------------------------------------------------------------------*/
    bool ReadAutoOverwrite() const;

    /*!-------------------------------------------------------------------------
    function:     ReadQueueCount()
    description:  Reads number of log queues from frame
    arguments:    none
    return value: (QueueCount) queue count
    --------------------------------------------------------------------------*/
    QueueCount ReadQueueCount() const;

    /*!-------------------------------------------------------------------------
    function:     ReadCycleRelatedEndOffset()
    description:  Reads cycle related log end offset from frame

                  - The size of the log device and with it the end offset on device
                    may grow during log operation by adding new devspaces with the 
                    'add log devspace' command. 
                  - However, as long as no log pages have been written to the
                    new devspace(s), the cyclic sequence of i/o seqnos is 
                    consistent within the old size limits of the log device.
                  - The present value says which of the offsets (old or new one)
                    is the right one when calculating offsets within the log
                    device.

    arguments:    none
    return value: (RawDeviceOffset) cycle related offset
    --------------------------------------------------------------------------*/
    RawDeviceOffset ReadCycleRelatedEndOffset() const;

    /*!-------------------------------------------------------------------------
    function:     ReadOldestNotSavedOffset()
    description:  Reads raw offset of 'oldest not saved' page from frame

                  - The 'oldest not saved' log page is the oldest page on 
                    log device that has not (yet) been saved to tape by a 
                    'log save' operation.
                  - It is the first log page that must be saved with the next
                    'log save'.

    arguments:    none
    return value: (RawDeviceOffset) raw log device offset 
    --------------------------------------------------------------------------*/
    RawDeviceOffset ReadOldestNotSavedOffset() const;

    /*!-------------------------------------------------------------------------
    function:     ReadOldestNotSavedIOSequenceNo()
    description:  Reads writer i/o seqno of 'oldest not saved' page from frame

                  - The 'oldest not saved' log page has been marked by the 
                    log writer with this seqno.
                  - see also ReadOldestNotSavedOffset()

    arguments:    none
    return value: (IOSequenceNo) writer i/o seqno
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadOldestNotSavedIOSequenceNo() const;

    /*!-------------------------------------------------------------------------
    function:     ReadLastKnownIOSequenceNo()
    description:  Reads 'forced last known' writer i/o seqno from frame

                  - In certain cycles of a few hundred log pages, the log info
                    page is made persistent, even when there is no savepoint
                    involved (write i/o of the log info page).
                  - The 'last known' seqno is the writer i/o seqno of
                    the last log page that was made persistent.

    arguments:    none
    return value: (IOSequenceNo) writer i/o seqno
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadLastKnownIOSequenceNo() const;

    /*!-------------------------------------------------------------------------
    function:     ReadLastKnownOffset()
    description:  Reads 'forced last known' raw log device offset from frame

                  - In certain cycles of a few hundred log pages, the log info
                    page is made persistent, even when there is no savepoint
                    involved (write i/o of the log info page).
                  - The 'last known' offset is the writer offset of
                    the last log page that was made persistent.

    arguments:    none
    return value: (RawDeviceOffset) raw log device offset 
    --------------------------------------------------------------------------*/
    RawDeviceOffset ReadLastKnownOffset() const;
    
    /*!-------------------------------------------------------------------------
    function:     ReadLogBackupCount()
    description:  Reads number of log backups from frame

                  - This is the number of log save's that have 
                    been performed so far for the present database.

    arguments:    none
    return value: (SAPDB_UInt4) log backup count
    --------------------------------------------------------------------------*/
    SAPDB_UInt4 ReadLogBackupCount() const;

    /*!-------------------------------------------------------------------------
    function:     ReadRedoUntilDate()
    description:  Reads date of last successful redo until from frame
    arguments:    none
    return value: (Kernel_Date) log redo until date
    --------------------------------------------------------------------------*/
    Kernel_Date ReadRedoUntilDate() const;

    /*!-------------------------------------------------------------------------
    function:     ReadRedoUntilTime()
    description:  Reads time of last successful redo until from frame
    arguments:    none
    return value: (Kernel_Time) log redo until time
    --------------------------------------------------------------------------*/
    Kernel_Time ReadRedoUntilTime() const;

    /*!-------------------------------------------------------------------------
    function:     ReadClearUntilLowerOffset()
    description:  Reads lower offset of 'clear until' from frame

                  - With the 'clear until' command, log pages between lower and 
                    upper offset, respectively, are cleared (initialized) on the
                    log device.
                  - see also ReadClearUntilUpperOffset()

    arguments:    none
    return value: (RawDeviceOffset) lower offset of 'clear until'
    --------------------------------------------------------------------------*/
    RawDeviceOffset ReadClearUntilLowerOffset() const;

    /*!-------------------------------------------------------------------------
    function:     ReadClearUntilUpperOffset()
    description:  Reads upper offset of 'clear until' from frame

                  - With the 'clear until' command, log pages between lower and 
                    upper offset, respectively, are cleared (initialized) on the
                    log device.
                  - see also ReadClearUntilLowerOffset()

    arguments:    none
    return value: (RawDeviceOffset) upper offset of 'clear until'
    --------------------------------------------------------------------------*/
    RawDeviceOffset ReadClearUntilUpperOffset() const;

    /*!-------------------------------------------------------------------------
    function:     ReadIOSequenceNoValidAfterRedoUntil()
    description:  Reads next writer i/o seqno from frame

                  - After that a 'restore log until' is finished, the log writer 
                    is reset to this i/o seqno.

    arguments:    none
    return value: (IOSequenceNo) writer i/o seqno
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadIOSequenceNoValidAfterRedoUntil() const;

    /*!-------------------------------------------------------------------------
    function:     ReadSavepointIOSequenceNo()
    description:  Reads i/o seqno of final log page of last savepoint from frame

                  - This page contains the 'savepoint complete' log entry
                    of the last savepoint.

    arguments:    none
    return value: (IOSequenceNo) writer i/o seqno
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadSavepointIOSequenceNo() const;

    /*!-------------------------------------------------------------------------
    function:     ReadSavepointEntryOffset()
    description:  Reads offset of 'savepoint complete' log entry.
    
                  - see also ReadSavepointIOSequenceNo()
    arguments:    none
    return value: (RawDeviceOffset) entry offset within log device
    --------------------------------------------------------------------------*/
    RawDeviceOffset ReadSavepointEntryOffset() const;

    /*!-------------------------------------------------------------------------
    function:     ReadSavepointEntryOffsetInPage()
    description:  Reads offset of 'savepoint complete' entry in log page.

                  - see also ReadSavepointIOSequenceNo()
    arguments:    none
    return value: (PageOffset) entry offset within a log page
    --------------------------------------------------------------------------*/
    PageOffset ReadSavepointEntryOffsetInPage() const;

    /*!-------------------------------------------------------------------------
    function:     ReadReleaseVersion()
    description:  Reads internal kernel release version from frame

                  - This is an internal string identifying the kernel release 
                    and build no.

    arguments:    none
    return value: (const SAPDB_Byte*) pointer to (not null terminated) string
    --------------------------------------------------------------------------*/
    const SAPDB_Byte* ReadReleaseVersion() const;

    /*!-------------------------------------------------------------------------
    function:     ReadDBIdentifier()
    description:  Reads internal database identifier from frame

                  - This is an internal string identifying the actual database
                    instance.

    arguments:    none
    return value: (const SAPDB_Byte*) pointer to (not null terminated) string
    --------------------------------------------------------------------------*/
    const SAPDB_Byte* ReadDBIdentifier() const;

    /*!-------------------------------------------------------------------------
    function:     WritePageID()
    description:  Writes page id to frame
    arguments:    pageid [in]   page id
    return value: none
    --------------------------------------------------------------------------*/
    void WritePageID
        (const PageID&                  pageid);

    /*!-------------------------------------------------------------------------
    function:     WriteLogDate()
    description:  Writes log date to frame
    arguments:    date [in]  log date
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLogDate
        (const Kernel_Date&                date);

    /*!-------------------------------------------------------------------------
    function:     WriteLogTime()
    description:  Writes log time to frame
    arguments:    time [in]  log time
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLogTime
        (const Kernel_Time&                time);

    /*!-------------------------------------------------------------------------
    function:     WriteLogFlushMode()
    description:  Writes log flush mode to frame
    arguments:    mode [in]  log flush mode
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLogFlushMode
        (const Log_FlushMode&       mode);

    /*!-------------------------------------------------------------------------
    function:     WriteDeviceState()
    description:  Writes redo until state to frame
    arguments:    state [in]  log redo until state
    return value: none
    --------------------------------------------------------------------------*/
    void WriteDeviceState
        (const Log_DeviceState&      state);

    /*!-------------------------------------------------------------------------
    function:     WriteDevspaceEnabled()
    description:  Writes if devspace is enabled
    arguments:    enabled [in]  bool
    return value: none
    --------------------------------------------------------------------------*/
    void WriteDevspaceEnabled
        (bool enabled);

    /*!-------------------------------------------------------------------------
    function:     WriteAutoOverwrite()
    arguments:    enabled [in]  bool
    return value: none
    --------------------------------------------------------------------------*/
    void WriteAutoOverwrite
        (bool enabled);

    /*!-------------------------------------------------------------------------
    function:     WriteQueueCount()
    description:  Writes number of log queues to frame
    arguments:    count [in]  queue count
    return value: none
    --------------------------------------------------------------------------*/
    void WriteQueueCount
        (const QueueCount&              count);

    /*!-------------------------------------------------------------------------
    function:     WriteCycleRelatedEndOffset()
    description:  Writes cycle related log end offset to frame

                  - see ReadCycleRelatedEndOffset()

    arguments:    size [in]   log device offset
    return value: none
    --------------------------------------------------------------------------*/
    void WriteCycleRelatedEndOffset
        (const RawDeviceOffset&        size);

    /*!-------------------------------------------------------------------------
    function:     WriteOldestNotSavedOffset()
    description:  Writes raw offset 'first not saved' page to frame

                  - see ReadOldestNotSavedOffset()

    arguments:    offset [in]  raw log device offset 
    return value: none
    --------------------------------------------------------------------------*/
    void WriteOldestNotSavedOffset
        (const RawDeviceOffset&         offset);

    /*!-------------------------------------------------------------------------
    function:     WriteOldestNotSavedIOSequenceNo()
    description:  Writes writer i/o seqno of 'first not saved' page to frame

                  - see also ReadOldestNotSavedOffset()

    arguments:    seqno [in]  writer i/o seqno
    return value: none
    --------------------------------------------------------------------------*/
    void WriteOldestNotSavedIOSequenceNo
        (const IOSequenceNo&            seqno);

    /*!-------------------------------------------------------------------------
    function:     WriteLastKnownIOSequenceNo()
    description:  Writes 'forced last known' writer i/o seqno to frame

                  - see also ReadLastKnownIOSequenceNo()

    arguments:    seqno [in]  writer i/o seqno
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLastKnownIOSequenceNo
        (const IOSequenceNo&            seqno);

    /*!-------------------------------------------------------------------------
    function:     WriteLastKnownOffset()
    description:  Writes 'forced last known' writer offset to frame

                  - see also ReadLastKnownOffset()

    arguments:    offset [in]  log dev offset
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLastKnownOffset
        (const RawDeviceOffset&            offset);

    /*!-------------------------------------------------------------------------
    function:     WriteLogBackupCount()
    description:  Writes number of log backups to frame

                  - see ReadLogBackupCount()

    arguments:    count [in]  log backup count
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLogBackupCount
        (SAPDB_UInt4                    count);

    /*!-------------------------------------------------------------------------
    function:     WriteRedoUntilDate()
    description:  Writes date of last successful redo until to frame
    arguments:    date [in]  log redo until date
    return value: none
    --------------------------------------------------------------------------*/
    void WriteRedoUntilDate
        (const Kernel_Date&                date);

    /*!-------------------------------------------------------------------------
    function:     WriteRedoUntilTime()
    description:  Writes time of last successful redo until to frame
    arguments:    time [in]  log redo until time
    return value: none
    --------------------------------------------------------------------------*/
    void WriteRedoUntilTime
        (const Kernel_Time&                time);

    /*!-------------------------------------------------------------------------
    function:     WriteClearUntilLowerOffset()
    description:  Writes lower offset of 'clear until' to frame

                  - see also ReadClearUntilLowerOffset()

    arguments:    lwb [in]  lower offset of 'clear until'
    return value: none
    --------------------------------------------------------------------------*/
    void WriteClearUntilLowerOffset
        (const IOSequenceNo&            lwb);

    /*!-------------------------------------------------------------------------
    function:     WriteClearUntilUpperOffset()
    description:  Writes upper offset of 'clear until' to frame

                  - see also ReadClearUntilUpperOffset()

    arguments:    upb [in]  upper offset of 'clear until'
    return value: none
    --------------------------------------------------------------------------*/
    void WriteClearUntilUpperOffset
        (const IOSequenceNo&            upb);

    /*!-------------------------------------------------------------------------
    function:     WriteIOSequenceNoValidAfterRedoUntil()
    description:  Writes next writer i/o seqno to frame

                  - see also ReadIOSequenceNoValidAfterRedoUntil()

    arguments:    seqno [in]  writer i/o seqno
    return value: none
    --------------------------------------------------------------------------*/
    void WriteIOSequenceNoValidAfterRedoUntil
        (const IOSequenceNo&            seqno);

    /*!-------------------------------------------------------------------------
    function:     WriteSavepointParameters()
    description:  Writes content of restart record to frame

    arguments:    rrpart [in]  part of restart record to be copied
    return value: none
    --------------------------------------------------------------------------*/
    void WriteSavepointParameters (const SAPDBFields_Field& rrpart);

    /*!-------------------------------------------------------------------------
    function:     WriteDBIdentifier()
    description:  Writes the db_identifier into the peristent part of the log_InfoPage
                  New for PTS 1113550 mb 2002-04-19

    arguments:    dbIdent [in]  the new db_identifier
    return value: none
    --------------------------------------------------------------------------*/
    void WriteDBIdentifier(const tsp00_Line dbIdent);	
    /*!-------------------------------------------------------------------------
    function:     ResetSavepointPart()
    description:  The savepoint offset and sequence are set to last known
    
                  This is used during ClearLog.
    --------------------------------------------------------------------------*/
    void ResetSavepointPart();
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Kernel_IPage interface                                       */
    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  Verifies consistency of page contents
    arguments:    none
    return:       (SAPDB_Bool) true if verify successful
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Kernel_IPage interface                                    */
    /* -----------------------------------------------------------------------*/


    /*!-------------------------------------------------------------------------
    function:     WriteToTrace()
    description:  writes the values of the info page to the vtrace.
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/
    void WriteToTrace () const;
    
private:

    /*--------------------------------------------------------------------------
    declaration: Persistent types
    description: Persistent type definitions with compile time size check
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent         < PageID,       MultiUse1,   4 >         
            PersistentPageID;

    typedef SAPDB_Persistent         < Kernel_Date,     SAPDB_UInt4, 4 > 
            PersistentDate;

    typedef SAPDB_Persistent         < Kernel_Time,     SAPDB_UInt4, 4 > 
            PersistentTime;

    typedef SAPDB_Persistent         < QueueCount,   SAPDB_UInt2, 2 > 
            PersistentQueueCount;

    typedef SAPDB_Persistent         < RawDeviceOffset, SAPDB_UInt4, 4 > 
            PersistentDeviceOffset;

    typedef SAPDB_Persistent         < DeviceBlockCount, SAPDB_UInt4, 4 > 
            PersistentBlockCount;

    typedef SAPDB_Persistent         < IOSequenceNo, SAPDB_UInt4, 4 > 
            PersistentIOSequenceNo;

    typedef SAPDB_Persistent         < SAPDB_UInt1,  SAPDB_UInt1, 1 >
            PersistentLogFlushMode;

    typedef SAPDB_Persistent         < SAPDB_UInt1,  SAPDB_UInt1, 1 >   
            PersistentDeviceState;

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#endif
    
    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within kernel page header
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):
    // PersistentIOSequenceNo reuses MultiUse1 in Kernel_Page74::PersistentHeader
    typedef Kernel_Page74::PersistentHeader             PH1;    // page74 header
    typedef PersistentDate::PersistentFormat            PH2;    // log date
    typedef PersistentTime::PersistentFormat            PH3;    // log time
    typedef PersistentLogFlushMode::PersistentFormat    PH4;    // log flush mode
    typedef PersistentDeviceState::PersistentFormat     PH5;    // redo until state
    typedef PersistentQueueCount::PersistentFormat      PH6;    // queue count
    typedef PersistentDeviceOffset::PersistentFormat    PH7;    // cycle related end offset
    typedef PersistentIOSequenceNo::PersistentFormat    PH8;    // last known ioseqno
    typedef PersistentDeviceOffset::PersistentFormat    PH9;    // last known offset
    typedef PersistentDeviceOffset::PersistentFormat    PH10;   // oldest not saved offset
    typedef PersistentIOSequenceNo::PersistentFormat    PH11;   // oldest not saved ioseqno
    typedef SAPDB_Bool1                                 PH12;   // devspace enabled
    typedef SAPDB_Bool1                                 PH13;   // auto overwrite
    typedef SAPDB_UInt2                                 PH14;   // filler, 2 byte
    typedef SAPDB_UInt4                                 PH15;   // log backup count
    typedef PersistentDate::PersistentFormat            PH16;   // redo until date
    typedef PersistentTime::PersistentFormat            PH17;   // redo until time
    typedef PersistentIOSequenceNo::PersistentFormat    PH18;   // clear until lower offset
    typedef PersistentIOSequenceNo::PersistentFormat    PH19;   // clear until upper offset
    typedef PersistentIOSequenceNo::PersistentFormat    PH20;   // redo until new ioseqno
    typedef SAPDB_UInt4                                 PH21;   // filler, 4 bytes
    typedef tsp00_C64                                   PH22;   // filler, 64 bytes
    typedef struct
    {
        tkb00_SaveptParam m_SaveptParam;     // 104 all redo related parameters
        tsp00_Version     m_RelVers;         //  40 kernel release version (internal format)
        tsp00_Line        m_DBIdent;         // 132 database identifier (internal format)
    }                                        // 276 sum of bytes
                                                        PH23;   // copied from from k57restartrec


protected:

    struct PersistentHeader
    {
        //  ---- Offset 0 ----
        PH1     m_KernelPage74Header;
        PH2     m_LogDate;
        PH3     m_LogTime;
        PH4     m_LogFlushMode;
        PH5     m_DeviceState;
        PH6     m_QueueCount;
        PH7     m_CycleRelatedEndOffset;
        PH8     m_LastKnownIOSequenceNo;
        PH9     m_LastKnownOffset;
        PH10    m_OldestNotSavedOffset;
        PH11    m_OldestNotSavedIOSequenceNo;
        PH12    m_DevspaceEnabled; // PTS 1113190 UH 2001-12-19 
        PH13	m_AutoOverwrite;   // PTS 1120106 UH 2003-01-17
        PH14	m_Filler2;
        PH15    m_LogBackupCount;
        PH16    m_RedoUntilDate;
        PH17    m_RedoUntilTime;
        PH18    m_ClearUntilLowerOffset;
        PH19    m_ClearUntilUpperOffset;
        PH20    m_IOSequenceNoValidAfterRedoUntil;
        PH21    m_Filler3;
        PH22    m_Filler4;
        //  ---- Offset 80 ----
        PH23    m_RestartRecordPart;
        //  ---- Offset 356 ----
    };

private:

    /*--------------------------------------------------------------------------
    declaration: ConverterPagePersistentHeader
    description: Compile time size check for persistent header
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent
        < PersistentHeader, 
          PersistentHeader,
          sizeof(PH1) +
          sizeof(PH2) +
          sizeof(PH3) +
          sizeof(PH4) +
          sizeof(PH5) +
          sizeof(PH6) +
          sizeof(PH7) +
          sizeof(PH8) +
          sizeof(PH9) +
          sizeof(PH10) +
          sizeof(PH11) +
          sizeof(PH12) +
          sizeof(PH13) +
          sizeof(PH14) +
          sizeof(PH15) +
          sizeof(PH16) +
          sizeof(PH17) +
          sizeof(PH18) +
          sizeof(PH19) +
          sizeof(PH20) +
          sizeof(PH21)
          >
        CheckedPersistentHeader;

    typedef Kernel_Page74::PersistentTrailer   PersistentTrailer;

protected:

    /*--------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    return value: (PersistentHeader*) header pointer
    prototypes:   2
    --------------------------------------------------------------------------*/
    PersistentHeader* PersistentHeaderPtr();

    const PersistentHeader* PersistentHeaderPtr() const;

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access (protected)                                     */
    /*!-------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/
    virtual SizeType GetHeaderSize() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access (protected)                                  */
    /* -----------------------------------------------------------------------*/

};
/*!*****************************************************************************

   endclass: Log_InfoPage

*******************************************************************************/

//------------------------------------------------------------------------------

inline Log_InfoPage::Log_InfoPage()
{}

//------------------------------------------------------------------------------

inline Log_InfoPage::Log_InfoPage
    (const PageFrame&               frame)
    : 
     Kernel_Page74(frame)
{}

//------------------------------------------------------------------------------

inline Log_InfoPage::Log_InfoPage
    (const PageFrame&               frame, 
     const PageID&                  pageid)
    : 
     Kernel_Page74(frame, Kernel_LogInfoPageType)
{
    InitializeFrame(pageid);
}


//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::SetAndInitializeFrame
    (const PageFrame&               frame, 
     const PageID&                  pageid)
{
    SetFrame(frame);
    InitializeFrame(pageid);
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::InitializeFrame
    (const PageID&                  pageid)
{
    Kernel_Page74::InitializeFrame(Kernel_LogInfoPageType);
    WritePageID(pageid);
}


//------------------------------------------------------------------------------

inline Log_InfoPage::PersistentHeader* 
    Log_InfoPage::PersistentHeaderPtr()
{
    return reinterpret_cast<PersistentHeader*>
           (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}

//------------------------------------------------------------------------------

inline const Log_InfoPage::PersistentHeader* 
    Log_InfoPage::PersistentHeaderPtr() const
{
    return reinterpret_cast<const PersistentHeader*>
           (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}

//------------------------------------------------------------------------------

inline Log_InfoPage::SizeType 
    Log_InfoPage::GetHeaderSize() const
{
    return sizeof(PersistentHeader);
}

//------------------------------------------------------------------------------

inline Log_InfoPage::PageID 
    Log_InfoPage::ReadPageID() const
{
    return ReadMultiUse1();
}

//------------------------------------------------------------------------------

inline Kernel_Date 
    Log_InfoPage::ReadLogDate() const
{
    return PersistentHeaderPtr()->m_LogDate;
}

//------------------------------------------------------------------------------

inline Kernel_Time 
    Log_InfoPage::ReadLogTime() const
{
    return PersistentHeaderPtr()->m_LogTime;
}

//------------------------------------------------------------------------------

inline Log_FlushMode 
    Log_InfoPage::ReadFlushMode() const
{
    return Log_FlushMode(PersistentHeaderPtr()->m_LogFlushMode);
}

//------------------------------------------------------------------------------

inline Log_DeviceState 
    Log_InfoPage::ReadDeviceState() const
{
    return Log_DeviceState(PersistentHeaderPtr()->m_DeviceState);
}

//------------------------------------------------------------------------------

inline bool
    Log_InfoPage::ReadDevspaceEnabled() const
{
    // PTS 1113190 UH 2001-12-19
    return PersistentHeaderPtr()->m_DevspaceEnabled;
}

//------------------------------------------------------------------------------

inline bool
    Log_InfoPage::ReadAutoOverwrite() const
{
    // PTS 1120106 UH 2003-01-17 new
    return PersistentHeaderPtr()->m_AutoOverwrite;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::QueueCount 
    Log_InfoPage::ReadQueueCount() const
{
    return PersistentHeaderPtr()->m_QueueCount;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::RawDeviceOffset
    Log_InfoPage::ReadCycleRelatedEndOffset() const
{
    return PersistentHeaderPtr()->m_CycleRelatedEndOffset;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::RawDeviceOffset 
    Log_InfoPage::ReadOldestNotSavedOffset() const
{
    return PersistentHeaderPtr()->m_OldestNotSavedOffset;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::IOSequenceNo 
    Log_InfoPage::ReadOldestNotSavedIOSequenceNo() const
{
    return PersistentHeaderPtr()->m_OldestNotSavedIOSequenceNo;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::IOSequenceNo 
    Log_InfoPage::ReadLastKnownIOSequenceNo() const
{
    return PersistentHeaderPtr()->m_LastKnownIOSequenceNo;
}


//------------------------------------------------------------------------------

inline Log_InfoPage::RawDeviceOffset
    Log_InfoPage::ReadLastKnownOffset() const
{
    return PersistentHeaderPtr()->m_LastKnownOffset;
}

//------------------------------------------------------------------------------

inline SAPDB_UInt4 
    Log_InfoPage::ReadLogBackupCount() const
{
    return PersistentHeaderPtr()->m_LogBackupCount;
}

//------------------------------------------------------------------------------

inline Kernel_Date 
    Log_InfoPage::ReadRedoUntilDate() const
{
    return PersistentHeaderPtr()->m_RedoUntilDate;
}

//------------------------------------------------------------------------------

inline Kernel_Time 
    Log_InfoPage::ReadRedoUntilTime() const
{
    return PersistentHeaderPtr()->m_RedoUntilTime;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::RawDeviceOffset 
    Log_InfoPage::ReadClearUntilLowerOffset() const
{
    return PersistentHeaderPtr()->m_ClearUntilLowerOffset;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::RawDeviceOffset 
    Log_InfoPage::ReadClearUntilUpperOffset() const
{
    return PersistentHeaderPtr()->m_ClearUntilUpperOffset;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::IOSequenceNo 
    Log_InfoPage::ReadIOSequenceNoValidAfterRedoUntil() const
{
    return PersistentHeaderPtr()->m_IOSequenceNoValidAfterRedoUntil;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::IOSequenceNo
    Log_InfoPage::ReadSavepointIOSequenceNo() const
{
    return PersistentHeaderPtr()->m_RestartRecordPart.m_SaveptParam.svpIOsequence_kb00;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::RawDeviceOffset
    Log_InfoPage::ReadSavepointEntryOffset() const
{
    return PersistentHeaderPtr()->m_RestartRecordPart.m_SaveptParam.svpStartOffset_kb00;
}

//------------------------------------------------------------------------------

inline Log_InfoPage::PageOffset
    Log_InfoPage::ReadSavepointEntryOffsetInPage() const
{
    return PersistentHeaderPtr()->m_RestartRecordPart.m_SaveptParam.svpStartEntryOffset_kb00;
}

//------------------------------------------------------------------------------

inline const SAPDB_Byte* 
    Log_InfoPage::ReadReleaseVersion() const
{
    return PersistentHeaderPtr()->m_RestartRecordPart.m_RelVers + 0; // +0 is for type conversion
}

//------------------------------------------------------------------------------

inline const SAPDB_Byte* 
    Log_InfoPage::ReadDBIdentifier() const
{
    return PersistentHeaderPtr()->m_RestartRecordPart.m_DBIdent + 0; // +0 is for type conversion
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WritePageID
    (const PageID&                  pageid)
{
    WriteMultiUse1(pageid);
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteLogDate
    (const Kernel_Date&                date)
{
    PersistentHeaderPtr()->m_LogDate = date;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteLogTime
    (const Kernel_Time&                time)
{
    PersistentHeaderPtr()->m_LogTime = time;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteLogFlushMode
    (const Log_FlushMode&       mode)
{
    PersistentHeaderPtr()->m_LogFlushMode = mode;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteDeviceState
    (const Log_DeviceState&      state)
{
    #ifdef SAPDB_QUICK
    // PTS 1111525 mb 2002-01-29 new
    // the allowed previous states are commented out. By this
    // an assertion is throwed when a not allowed previous state is found
    switch (state)
    {
        case Log_DeviceStateOkay:
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateOkay              != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateHistoryLost       != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateCleared           != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateClearing          != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyCleared  != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyClearing != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE       != PersistentHeaderPtr()->m_DeviceState);
            break;

        case Log_DeviceStateHistoryLost:
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateOkay              != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateHistoryLost       != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateCleared           != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateClearing          != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyCleared  != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyClearing != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE       != PersistentHeaderPtr()->m_DeviceState);
            break;

        case Log_DeviceStateCleared:
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateOkay              != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateHistoryLost       != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateCleared           != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateClearing          != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyCleared  != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyClearing != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(ReadClearUntilLowerOffset().IsInvalid());
            SAPDBERR_ASSERT_ARGUMENT(ReadClearUntilUpperOffset().IsInvalid());
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE       != PersistentHeaderPtr()->m_DeviceState);
            break;

        case Log_DeviceStateClearing:
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateOkay              != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateHistoryLost       != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateCleared           != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateClearing          != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyCleared  != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyClearing != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(ReadClearUntilLowerOffset().IsValid());
            SAPDBERR_ASSERT_ARGUMENT(ReadClearUntilUpperOffset().IsValid());
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE       != PersistentHeaderPtr()->m_DeviceState);
            break;

        case Log_DeviceStatePartiallyCleared:
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateOkay              != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateHistoryLost       != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateCleared           != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateClearing          != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyCleared  != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyClearing != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_STATE(ReadClearUntilLowerOffset().IsInvalid());
            SAPDBERR_ASSERT_STATE(ReadClearUntilUpperOffset().IsInvalid());
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE       != PersistentHeaderPtr()->m_DeviceState);
            break;

        case Log_DeviceStatePartiallyClearing:
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateOkay              != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateHistoryLost       != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateCleared           != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStateClearing          != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyCleared  != PersistentHeaderPtr()->m_DeviceState);
            // SAPDBERR_ASSERT_ARGUMENT(Log_DeviceStatePartiallyClearing != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE       != PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_STATE(ReadClearUntilLowerOffset().IsValid());
            SAPDBERR_ASSERT_STATE(ReadClearUntilUpperOffset().IsValid());
            break;

        default:
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE == PersistentHeaderPtr()->m_DeviceState);
            SAPDBERR_ASSERT_ARGUMENT(Log_DeviceState_DO_NOT_USE != PersistentHeaderPtr()->m_DeviceState);
    }

    #endif

    PersistentHeaderPtr()->m_DeviceState = state;
}


//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteDevspaceEnabled
    (bool enabled)
{
    // PTS 1113190 UH 2001-12-19
    PersistentHeaderPtr()->m_DevspaceEnabled = enabled;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteAutoOverwrite
    (bool enabled)
{
    // PTS 1120106 UH 2003-01-17 new
    PersistentHeaderPtr()->m_AutoOverwrite   = enabled;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteQueueCount
    (const QueueCount&              count)
{
    PersistentHeaderPtr()->m_QueueCount = count;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteCycleRelatedEndOffset
    (const RawDeviceOffset&        size)
{
    PersistentHeaderPtr()->m_CycleRelatedEndOffset = size;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteOldestNotSavedOffset
    (const RawDeviceOffset&     offset)
{
    PersistentHeaderPtr()->m_OldestNotSavedOffset = offset;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteOldestNotSavedIOSequenceNo
    (const IOSequenceNo&            seqno)
{
    PersistentHeaderPtr()->m_OldestNotSavedIOSequenceNo = seqno;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteLastKnownIOSequenceNo
    (const IOSequenceNo&            seqno)
{
    PersistentHeaderPtr()->m_LastKnownIOSequenceNo = seqno;
}
//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteLastKnownOffset
        (const RawDeviceOffset&            offset)
{
    PersistentHeaderPtr()->m_LastKnownOffset = offset;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteLogBackupCount
    (SAPDB_UInt4                    count)
{
    PersistentHeaderPtr()->m_LogBackupCount = count;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteRedoUntilDate
    (const Kernel_Date&                date)
{
    PersistentHeaderPtr()->m_RedoUntilDate = date;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteRedoUntilTime
    (const Kernel_Time&                time)
{
    PersistentHeaderPtr()->m_RedoUntilTime = time;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteClearUntilLowerOffset
    (const IOSequenceNo&            lwb)
{
    PersistentHeaderPtr()->m_ClearUntilLowerOffset = lwb;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteClearUntilUpperOffset
    (const IOSequenceNo&            upb)
{
    PersistentHeaderPtr()->m_ClearUntilUpperOffset = upb;
}

//------------------------------------------------------------------------------

inline void 
    Log_InfoPage::WriteIOSequenceNoValidAfterRedoUntil
    (const IOSequenceNo&            seqno)
{
    PersistentHeaderPtr()->m_IOSequenceNoValidAfterRedoUntil = seqno;
}

//------------------------------------------------------------------------------

inline void Log_InfoPage::WriteSavepointParameters (const SAPDBFields_Field& rrpart)
{
    tkb00_SaveptParam &param = PersistentHeaderPtr()->m_RestartRecordPart.m_SaveptParam;
    SAPDBERR_ASSERT_ARGUMENT(rrpart.GetLength() == sizeof(param));
    memcpy(&param, rrpart.GetDataPtr(), rrpart.GetLength());

    if (ReadSavepointIOSequenceNo().IsValid()
         && 
        ReadLastKnownIOSequenceNo().IsValid())
    {
        if (ReadSavepointIOSequenceNo() > ReadLastKnownIOSequenceNo())
        {
            WriteLastKnownIOSequenceNo(ReadSavepointIOSequenceNo());
            WriteLastKnownOffset(ReadSavepointEntryOffset());
        }
    }
    else
    {
        WriteLastKnownIOSequenceNo(ReadSavepointIOSequenceNo());
        WriteLastKnownOffset(ReadSavepointEntryOffset());
    }
}

//------------------------------------------------------------------------------

inline void  
    Log_InfoPage::WriteDBIdentifier(const tsp00_Line dbIdent)		// new PTS 1113550 mb 2002-04-19
{
    PersistentHeaderPtr()->m_RestartRecordPart.m_DBIdent = dbIdent;
}

//------------------------------------------------------------------------------

inline void Log_InfoPage::ResetSavepointPart()
{
    tkb00_SaveptParam &param  = PersistentHeaderPtr()->m_RestartRecordPart.m_SaveptParam;
    param.svpIOsequence_kb00  = ReadLastKnownIOSequenceNo();
    param.svpStartOffset_kb00 = ReadLastKnownOffset();
}

//------------------------------------------------------------------------------

inline SAPDB_Bool
    Log_InfoPage::Verify() const
{
    return true;
}

// -----------------------------------------------------------------------------
inline void
Log_InfoPage::WriteToTrace () const
{
    Kernel_VTrace trace;
    
    if ( ! IsAssigned() )
    {
        trace << "log info page not assigned\n";
        return;
    }
    
    trace << "InfoPageId: " << ReadPageID()
          << ", Date: " << ReadLogDate()
          << ", Time: " << ReadLogTime()
          << FlushLine;
    trace << "QueueCount: " << ReadQueueCount()
          << ", State: " << Log_DeviceStateStrings[ReadDeviceState()]
          << ", Mode: " << ReadFlushMode()
          << FlushLine;
    trace << "CycleRelatedEndOffset: " << ReadCycleRelatedEndOffset();
    trace << "LastKnown-Offset: " << ReadLastKnownOffset() 
          << ", -IOSequence: " << ReadLastKnownIOSequenceNo()
          << FlushLine;
    trace << "OldestNotSaved-Offset: " << ReadOldestNotSavedOffset() 
          << ", -IOSequence: " << ReadOldestNotSavedIOSequenceNo()
          << ", LogBackupCount: " << ReadLogBackupCount()
          << FlushLine;
    trace << "LastSavepoint-Position: " << ReadSavepointEntryOffset() 
          << "." << ReadSavepointEntryOffsetInPage() 
          << ", -IOSequence: " << ReadSavepointIOSequenceNo()
          << FlushLine;
}

//------------------------------------------------------------------------------

#endif  // LOG_INFOPAGE_HPP

