/*!**************************************************************************

  module      : Log_OpenTransFile.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines the classes Log_OpenTransFile



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_OpenTransFile_H
#define Log_OpenTransFile_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "DataAccess/Data_ChainFixSizeSpace.hpp"
#include "File/File_IFile.hpp"

class Log_Transaction;

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_OpenTransPage
description: 
*/

class Log_OpenTransPage : public Data_PageFixSizeSpace
{

public:

    
    /*! --------------------------------------------------------------------------------
    function:     SetSavepointSequence
    */
    void SetSavepointSequence (Log_SavepointSequence SavepointSequence)
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        NodePtr()->ndInvRoot_bd00() = SavepointSequence; // PTS 1120695 UH 2003-10-27
    }

    /*! --------------------------------------------------------------------------------
    function:     GetSavepointSequence
    */
    Log_SavepointSequence GetSavepointSequence () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return NodePtr()->ndInvRoot_bd00(); // PTS 1120695 UH 2003-10-27
    }
    
};
/*! endclass: Log_OpenTransPage */

/*---------------------------------------------------------------------------*/
/*! class:       Log_OpenTransFile
    description: This is the Interface to Redo log manager.
*/
class Log_OpenTransFile : public File_IFile
{

public:
    
    /*! --------------------------------------------------------------------------------
    declaration: Container, Iterator, BeforeImageId
    description: These types are used with this class.
    */
    typedef Data_ChainFixSizeSpace<Log_OpenTransPage>  Container;
    typedef Container::Iterator                        Iterator;
    typedef Data_ChainIterator<Log_OpenTransPage>      PageIterator;
    /*! enddeclaration: */

public:

    /*! --------------------------------------------------------------------------------
    function:    Log_OpenTransFile
    arguments:   Trans               [in/out]
                 RootPageNo          [in]
                 SavepointSequence   [in]
    description: This constructs an Redo log handle for the given transaction.
                 If the RootPageNo is valid, an existing file is assumed.
    */
    Log_OpenTransFile (tgg00_TransContext &Trans,
                       const Data_PageNo  &Root);
              
    /*! ---------------------------------------------------------------------------
    function:    Create
    arguments:   SavepointSequence [in]
    returns:     true, if successfull
    description: A new RedoLogfile is created, the transNo is written into the root.
    */
    bool Create (Log_SavepointSequence SavepointSequence);

    /*! ---------------------------------------------------------------------------
    function:    Drop
    description: A RedoLogfile is dropped.
    */
    virtual void Drop ();

    /*! ---------------------------------------------------------------------------
    function:    Verify
    returns:     true, if the file is consistent
    */
    virtual bool Verify(tgg00_TransContext &trans,
                        bool                isCold);

    /*! ---------------------------------------------------------------------------
    function:    WriteToTrace
    description: This traces some basic information about the file to the trace.
    */
    virtual void WriteToTrace() const
    {}

    /*! ---------------------------------------------------------------------------
    function:    GetRootId
    returns:     Data_PageId - of the root
    description: This returns the Id of the root page of this file.
                 More over this specifies what kind of recovery mode pages of the file are.
    */
    virtual Data_PageId GetRootId() const
    {
        return m_Container.RootId();
    }

    /*! ---------------------------------------------------------------------------
    function:    IsCreated
    returns:     true, if a page exists for this file.
    */
    bool IsCreated () const
    {
        return m_Container.RootId().PageNo().IsValid();
    }

    /*! --------------------------------------------------------------------------------
    function:    WriteOpenTrans
    arguments:   OpenTrans [in]
    description: The given Space is written into the redo file.
    */
    void WriteOpenTrans (Log_Transaction &OpenTrans);

    /*! --------------------------------------------------------------------------------
    function:    GetFirstOpenTrans
    description: the iterator points to the first open trans in the file.
                 the given savepoint sequence is checked.
    arguments:   SavepointSequence [in]
    returns:     Iterator&
    */
    Iterator& GetFirstOpenTrans (Log_SavepointSequence SavepointSequence);

    /*! ---------------------------------------------------------------------------
    function:    GetSpaceSize
    returns:     size in bytes
    */
    Data_RecordLength GetSpaceSize () const
    {
        return m_Container.GetSpaceSize();
    }

private:

    /*! --------------------------------------------------------------------------------
    variable:    m_Container
    description: The container for this file.
    */
    Container m_Container;

    /*! --------------------------------------------------------------------------------
    variable:    m_Iterator
    description: The iterator for the container.
    */
    Iterator m_Iterator;

};
/*! endclass: Log_OpenTransFile */


/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/

/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/
                                         
#endif  /* Log_Redo_H */
