/*!*****************************************************************************

  module:       Log_Page.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH
  
  special area: Logging

  description:  Handler for archive log pages
                
  see also:     Kernel_IPage.html 
                Kernel_Page.html
                Kernel_Page74.html

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_PAGE_HPP
#define LOG_PAGE_HPP


#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "KernelCommon/Kernel_DateTime.hpp"
#include "KernelCommon/Kernel_Page74.hpp"
#include "DataAccess/Data_Types.hpp"
#include "Logging/Log_Types.hpp"

/*--------------------------------------------------------------------------
declaration : LOG_PAGE_ALIGNMENT
description:  alignment for log page entries
--------------------------------------------------------------------------*/
#define LOG_PAGE_ALIGNMENT          4

/*--------------------------------------------------------------------------
declaration : LOG_PAGE_FIRST_OFFSET
--------------------------------------------------------------------------*/
#define LOG_PAGE_FIRST_OFFSET       1

/*!*****************************************************************************

   class: Log_Page

   description: 
   
   Archive log page handler class (concrete class)

*******************************************************************************/
class Log_Page : public Kernel_Page74
{

public:
    
    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : IOSequenceNoc
    description:  i/o sequence number of log pages
    --------------------------------------------------------------------------*/
    typedef Log_IOSequenceNo                    IOSequenceNo;
    /*!-------------------------------------------------------------------------
    declaration : QueueID
    description:  identifier for log queues
    --------------------------------------------------------------------------*/
    typedef Log_QueueID                         QueueID;
    /*!-------------------------------------------------------------------------
    declaration : OffsetCount
    description:  number of times a log page has been written to device
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt1                         OffsetCount;
    /*!-------------------------------------------------------------------------
    declaration : EntryLength
    description:  length of a log page entry
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt4                         EntryLength;
    /*!-------------------------------------------------------------------------
    declaration : Label
    description:  page label
    --------------------------------------------------------------------------*/
    typedef Kernel_Page74::MultiUseT1           Label;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Log_Page()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Log_Page();

    /*!-------------------------------------------------------------------------
    function:     Log_Page()
    description:  Constructor; assigns frame

                  - With this constructor, a frame is assigned which contains
                    an initialized or filled log page.
                  - Does not change frame contents.
    arguments:    frame [in]    page frame
    return value: none
    --------------------------------------------------------------------------*/
    Log_Page
        (const PageFrame&               frame);
   
    /*!-------------------------------------------------------------------------
    function:     Log_Page()
    description:  Constructor; assigns frame and initializes frame contents
    arguments:    frame    [in]  page frame
                  quid     [in]  queue id number
                  quseqno  [in]  queue i/o sequence number
    return value: none
    --------------------------------------------------------------------------*/
    Log_Page
        (const PageFrame&               frame, 
         QueueID                        quid,    
         IOSequenceNo                   quseqno);
   
    /*!-------------------------------------------------------------------------
    function:     ~Log_Page()
    description:  might check page
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    ~Log_Page();

    /*!-------------------------------------------------------------------------
    function:     SetAndInitializeFrame()
    description:  Assigns frame and initializes frame contents
    arguments:    frame    [in]  page frame
                  quid     [in]  queue id number
                  quseqno  [in]  queue i/o sequence number
    return value: none
    --------------------------------------------------------------------------*/
    void SetAndInitializeFrame
        (const PageFrame&               frame, 
         QueueID                        quid,    
         IOSequenceNo                   quseqno);

    /*!-------------------------------------------------------------------------
    function:     InitializeFrame()
    description:  Initializes frame contents
    arguments:    quid     [in]  queue id number
                  quseqno  [in]  queue i/o sequence number
    return value: none
    --------------------------------------------------------------------------*/
    void InitializeFrame
        (QueueID                        quid,    
         IOSequenceNo                   quseqno);

    /*!-------------------------------------------------------------------------
    function:     Clear()
    description:  Clears the page

                  - After this operation, the page is considered empty.
                  - Queue i/o sequence number and i/o sequence number 
                    are invalidated. First entry position is reset.
    arguments:    quseqno  [in]  queue i/o sequence number
    return value: none
    --------------------------------------------------------------------------*/
    void Clear();

    /*!-------------------------------------------------------------------------
    function:     Verify()
    description:  checks the page
    arguments:    none
    return value: true, if everything checked on the page is ok.
    --------------------------------------------------------------------------*/
    virtual SAPDB_Bool Verify() const;
    
    /*!-------------------------------------------------------------------------
    function:     Reset()
    description:  Resets the page

                  - After this operation, the page is considered empty.
                  - A new queue i/o sequence number is set and the 
                    i/o sequence number is invalidated.
                    First entry position is reset.
    arguments:    quseqno  [in]  queue i/o sequence number
    return value: none
    --------------------------------------------------------------------------*/
    void Reset
        (IOSequenceNo                   quseqno);

    /*!-------------------------------------------------------------------------
    function:     IsEmpty()
    description:  returns true of no logentry was written into the page.
    arguments:    none
    return value: (bool)
    --------------------------------------------------------------------------*/
    inline bool IsEmpty() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access                                                 */
    /*!-------------------------------------------------------------------------
    function:     WriteWriterIOSequenceNo()
    description:  Writes log writer i/o sequence number to frame
    arguments:    ioseqno [in]  i/o sequence number to be written
    return value: none
    --------------------------------------------------------------------------*/
    void WriteWriterIOSequenceNo
        (IOSequenceNo                   ioseqno);

    /*!-------------------------------------------------------------------------
    function:     ReadFirstWriterIOSequenceNo()
    description:  Reads log writer i/o sequence number from frame
    arguments:    none
    return value: (IOSequenceNo) i/o sequence number
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadFirstWriterIOSequenceNo() const;

    /*!-------------------------------------------------------------------------
    function:     ReadLastWriterIOSequenceNo()
    description:  Reads log writer i/o sequence number from frame
    arguments:    none
    return value: (IOSequenceNo) i/o sequence number
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadLastWriterIOSequenceNo() const;
    
    /*!-------------------------------------------------------------------------
    function:     WriteDate()
    --------------------------------------------------------------------------*/
    void WriteDate (Kernel_Date date);
    /*!-------------------------------------------------------------------------
    function:     ReadDate()
    --------------------------------------------------------------------------*/
    Kernel_Date ReadDate() const;
    /*!-------------------------------------------------------------------------
    function:     WriteTime()
    --------------------------------------------------------------------------*/
    void WriteTime (Kernel_Time time);
    /*!-------------------------------------------------------------------------
    function:     ReadTime()
    --------------------------------------------------------------------------*/
    Kernel_Time ReadTime() const;

    /*!-------------------------------------------------------------------------
    function:     WriteQueueIOSequenceNo()
    description:  Writes queue i/o sequence number to frame
    arguments:    quseqno [in]  queue i/o sequence number to be written
    return value: none
    --------------------------------------------------------------------------*/
    void WriteQueueIOSequenceNo
        (IOSequenceNo                   quseqno);

    /*!-------------------------------------------------------------------------
    function:     ReadQueueIOSequenceNo()
    description:  Reads queue i/o sequence number from frame
    arguments:    none
    return value: (QueueIOSequenceNo) queue i/o sequence number
    --------------------------------------------------------------------------*/
    IOSequenceNo ReadQueueIOSequenceNo() const;

    /*!-------------------------------------------------------------------------
    function:     WriteQueueID()
    description:  Writes queue id to frame
    arguments:    quid [in]  queue id to be written
    return value: none
    --------------------------------------------------------------------------*/
    void WriteQueueID
        (QueueID                        quid);

    /*!-------------------------------------------------------------------------
    function:     ReadQueueID()
    description:  Reads queue id from frame
    arguments:    none
    return value: (QueueID) queue id
    --------------------------------------------------------------------------*/
    QueueID ReadQueueID() const;

    /*!-------------------------------------------------------------------------
    function:     ReadOffsetCount()
    description:  Reads number of offsets stored in page trailer from frame
    arguments:    none
    return value: (OffsetType) actual write position
    --------------------------------------------------------------------------*/
    OffsetCount ReadOffsetCount() const;

    /*!-------------------------------------------------------------------------
    function:     ReadOffsetOnPage()
    description:  Reads offset of specified offset number from frame
    arguments:    writecycle [in]  number between 1 and ReadOffsetCount() 
    return value: (OffsetType) entry offset
    --------------------------------------------------------------------------*/
    OffsetType ReadOffsetOnPage
        (SAPDB_UInt                     offsetno) const;

    /*!-------------------------------------------------------------------------
    function:     WriteFirstFreeOffset()
    description:  Writes offset of actual write position to frame
    arguments:    offset [in]  actual offset to be written
    return value: none
    --------------------------------------------------------------------------*/
    void WriteFirstFreeOffset
        (OffsetType                     offset);

    /*!-------------------------------------------------------------------------
    function:     ReadFirstFreeOffset()
    description:  Reads offset of actual write position from frame
    arguments:    none
    return value: (OffsetType) actual write position
    --------------------------------------------------------------------------*/
    OffsetType ReadFirstFreeOffset() const;

    /*!-------------------------------------------------------------------------
    function:     WriteLabel()
    description:  Writes numeric page label to frame (debug purpose)
    arguments:    label [in]  label to be written
    return value: none
    --------------------------------------------------------------------------*/
    void WriteLabel
        (Label                          label);

    /*!-------------------------------------------------------------------------
    function:     ReadsLabel()
    description:  Reads numeric page label from frame (debug purpose)
    arguments:    none
    return value: (Label) 
    --------------------------------------------------------------------------*/
    Label ReadLabel() const;

    /*!-------------------------------------------------------------------------
    function:     HasFreeSpaceAvailable()
    description:  Checks whether there is at least specified space on the page
    arguments:    size [in]    requested space in bytes
    return value: (SAPDB_Bool) true if space available
    --------------------------------------------------------------------------*/
    SAPDB_Bool HasFreeSpaceAvailable
        (SizeType                       size) const;

    /*!-------------------------------------------------------------------------
    function:     FreeSpaceAvailable()
    description:  Calculates free space that can be reserved "en block".
                  If current offset is greater 0 then the space is calculated from that
                  offset on.
    arguments:    currentoffset
    return value: (SizeType) free space
    --------------------------------------------------------------------------*/
    SizeType FreeSpaceAvailable(OffsetType currentoffset = 0) const;

    /*!-------------------------------------------------------------------------
    function:     MaxFreeSpace()
    description:  Returns the maximum free space that can be reserved on an 
                  empty log page
    arguments:    none
    return value: (SizeType) max. free space
    --------------------------------------------------------------------------*/
    SizeType MaxFreeSpace() const;

    /*--------------------------------------------------------------------------
    function:     MinSpaceOffset()
    description:  Returns size of page header in bytes
                  This is static function which is used in
                  admin-mode, when no Log_Page is accessible
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/
    static SizeType MinSpaceOffset() ;

    /*!-------------------------------------------------------------------------
    function:     Reserve()
    description:  Reserves a specified amount of space on a page and returns
                  pointer to it
    arguments:    restsize     [in]   this size is needed
                  reservesize  [in]   this size is only reserved
    return value: (SAPDB_Byte*) pointer to reserved space
    --------------------------------------------------------------------------*/
    SAPDB_Byte* Reserve
        (SizeType                       restsize,
         SizeType                       reservesize);

    /*!-------------------------------------------------------------------------
    function:     Alignment()
    description:  Returns the page alignment in bytes

                  - Space is reserved with the returned alignment.
    arguments:    none
    return value: (SizeType) page alignment
    --------------------------------------------------------------------------*/
    static SizeType Alignment();

    /*--------------------------------------------------------------------------
    function:     IncOffsetCount()
    description:  Increments i/o write count and writes it to frame
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void IncOffsetCount();

    /*--------------------------------------------------------------------------
    function:     Capacity()
    description:  Returns size of the frame in bytes
    arguments:    none
    return value: (SizeType) frame size
    --------------------------------------------------------------------------*/
    SizeType Capacity() const;

    /*--------------------------------------------------------------------------
    function:     Truncate ()
    description:  truncates the entries on the logpage until the given iosequence 
                  is reached. If the given IOSequence is lower than the lowest
                  IOSequence on this page, then the whole page is cleared.
    arguments:    Log_IOSequenceNo
    return value: none                   
    --------------------------------------------------------------------------*/
    void Truncate (Log_IOSequenceNo newLastIOSeq);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access                                              */
    /* -----------------------------------------------------------------------*/
    
    /*--------------------------------------------------------------------------
    function:     WriteToTrace()
    --------------------------------------------------------------------------*/
    void WriteToTrace() const;

    /* -----------------------------------------------------------------------*/
    /*! chapter: Iterators                                                    */
    /*------------------------------------------------------------------------*/
    
    /*!-------------------------------------------------------------------------
    class:       Iterator
    description: Dereferences parts of split spaces in log pages.
    --------------------------------------------------------------------------*/
    class Iterator
    {
    friend class Log_Page;
    public:
        /*!-------------------------------------------------------------------------
        function:    IsValid
        returns:     true, if it can be dereferenced.
        --------------------------------------------------------------------------*/
        bool IsValid() const
        {
            if ( 0 == m_Page )
                return false;
            if ( ! m_Page->IsAssigned() )
                return false;
            return m_Page->OffsetIsValid(m_Offset);
        }
        /*!-------------------------------------------------------------------------
        function:    Invalidate
        description: resets all internal members. IsValid will return false.
        --------------------------------------------------------------------------*/
        void Invalidate()
        {
            m_Offset = 0;
            m_Page   = 0;
        }
        /*!-------------------------------------------------------------------------
        function:    operator *
        returns:     SAPDB_Byte*
        description: This returns the entry, which is currently referenced.

                     If the iterator is not valid NULL is returned.
        --------------------------------------------------------------------------*/
        SAPDB_Byte* operator * ()
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return m_Page->GetEntry(m_Offset);
        }
        /*!-------------------------------------------------------------------------
        function:    operator ++
        description: This sets the iterator to the next entry in page.

                     If the iterator is not valid NULL is returned.
        --------------------------------------------------------------------------*/
        Iterator& operator ++ ()
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            m_Offset = m_Page->GetOffsetOfNextEntry(m_Offset);
            if ( ! m_Page->OffsetIsValid (m_Offset) )
                Invalidate();
            return *this;
        }
        /*!-------------------------------------------------------------------------
        function:    Length
        returns:     Length of current part of the Split record.
        description: This is the length, which is currently found in the actual page.
        --------------------------------------------------------------------------*/
        EntryLength Length () const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return m_Page->GetEntryLength(m_Offset);
        }
        /*!-------------------------------------------------------------------------
        function:    RestLength
        returns:     EntryLength
        description: The rest length of current entry is given. It may be longer then
                     the current space in page is.
        --------------------------------------------------------------------------*/
        EntryLength RestLength () const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return m_Page->GetRestEntryLength (m_Offset);
        }
        /*!-------------------------------------------------------------------------
        function:    GetIOSequence
        returns:     IOSequenceNo
        description: The corresponding IOSequence to the current referenced
                     logentry is returned. If IsValid() is false, the invalid
                     iosequence is returned.
        --------------------------------------------------------------------------*/
        IOSequenceNo GetIOSequence () const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            SAPDBERR_ASSERT_STATE( ((Length() + sizeof(EntryLength)) & 0xffff0000) == 0);

            return m_Page->GetIOSequence (m_Offset+OffsetType(Length() + sizeof(EntryLength)));
        }
        /*!-------------------------------------------------------------------------
        function:    GetQueueId
        returns:     QueueID
        description: This is the queue id in which this entry was written.
        --------------------------------------------------------------------------*/
        QueueID GetQueueId () const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return m_Page->ReadQueueID();
        }
        /*!-------------------------------------------------------------------------
        function:    GetQueueIOSequenceNo
        returns:     IOSequence
        description: The QueueIOSequence is returned.
        --------------------------------------------------------------------------*/
        IOSequenceNo GetQueueIOSequenceNo () const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return m_Page->ReadQueueIOSequenceNo();
        }
        /*!-------------------------------------------------------------------------
        function:    GetLastEntryTimestamp
        arguments:   date [out]
                     time [out]
        description: At this timestamp the last log entry in this log page was written.
        --------------------------------------------------------------------------*/
        void GetLastEntryTimestamp ( Kernel_Date &date,
                                     Kernel_Time &time ) const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            date = m_Page->ReadDate();
            time = m_Page->ReadTime();
        }
        /*!-------------------------------------------------------------------------
        function:    GetLastEntryTime
        returns:     Kernel_Time
        description: This time is the date of the last log entry in this log page.
        --------------------------------------------------------------------------*/
        Kernel_Date GetLastEntryDate () const
        {
            SAPDBERR_ASSERT_STATE( IsValid() );
            return m_Page->ReadDate();
        }
        /*!-------------------------------------------------------------------------
        function:    GetPosition
        description: This returns the current offset of the iterator.
        returns:     Offset
        --------------------------------------------------------------------------*/
        OffsetType GetPosition () const
        {
            return m_Offset;
        }
        /*!-------------------------------------------------------------------------
        function:    SetPosition
        arguments:   offset [in]
        description: This is sets the iterator to a specific offset in the page.
        --------------------------------------------------------------------------*/
        void SetPosition (OffsetType offset)
        {
            m_Offset = offset;
        }
        /*!-------------------------------------------------------------------------
        function:    SetPage
        arguments:   page [in]
        description: This is sets the iterator to a new page.
        --------------------------------------------------------------------------*/
        void SetPage (Log_Page   &Page)
        {
            m_Page   = &Page;
        }

    protected: // constructor

        /*!-------------------------------------------------------------------------
        function:    Iterator
        description: may only be used by the Log_Page
        --------------------------------------------------------------------------*/
        Iterator ()
        : m_Offset (0),
          m_Page   (0)
        {}

        Iterator (Log_Page &Page,
                  OffsetType      Offset)
        : m_Offset (Offset),
          m_Page   (&Page)
        {}

        Iterator (Log_Page &Page)
        : m_Offset (0),
          m_Page   (&Page)
        {}

    private: // member

        /*! -----------------------------------------------------------------------
        variable:    m_Page
        description: The Logpage used to reference entries in it.
        --------------------------------------------------------------------------*/
        Log_Page *m_Page;

        /*! -----------------------------------------------------------------------
        variable:    m_Offset
        description: This refers to the current record: (m_pFrame + m_RecordOffset)
        --------------------------------------------------------------------------*/
        OffsetType m_Offset;
    };
    /*! endclass: Iterator */

    /*!-------------------------------------------------------------------------
    function:     GetIterator()
    description:  Returns iterator
    arguments:    none
    return value: Iterator
    --------------------------------------------------------------------------*/
    Iterator GetIterator()
    {
        return Iterator (*this);
    }

    /*!-------------------------------------------------------------------------
    function:     Begin()
    description:  Returns iterator on first entry on page
    arguments:    none
    return value: Iterator
    --------------------------------------------------------------------------*/
    void Begin (Iterator &iter)
    {
        iter.SetPage(*this);
        iter.SetPosition ( CalcNearestAlignedSize(GetHeaderSize()) );
    }

    /*!-------------------------------------------------------------------------
    function:     GetOffsetOfNextEntry()
    description:  Returns the offset of the next entry in the page seen
                  from the given offset.
                  If the returned offset is invalid for current page,
                  the next entry or the next part of current entry is found
                  in the following log page.
    arguments:    OffsetType [in]
    return value: OffsetType
    --------------------------------------------------------------------------*/
    OffsetType GetOffsetOfNextEntry(OffsetType entryOffset)
    {
        SAPDBERR_ASSERT_STATE( ((GetRestEntryLength(entryOffset) + sizeof(EntryLength)) & 0xffff0000) == 0);

        return entryOffset + CalcNearestAlignedSize
                             (Log_Page::SizeType(GetRestEntryLength(entryOffset)+sizeof(EntryLength)));
    }
    /*!-------------------------------------------------------------------------
    function:     OffsetIsValid()
    description:  Returns true, if the offset in the valid range.
    arguments:    OffsetType [in]
    return value: SAPDB_Bool
    --------------------------------------------------------------------------*/
    SAPDB_Bool OffsetIsValid (OffsetType offset) const
    {
        return offset >= CalcNearestAlignedSize(GetHeaderSize())
               &&
               offset < ReadFirstFreeOffset();
    }
    /*!-------------------------------------------------------------------------
    function:     GetEntryLength()
    description:  Returns the length of the entry part in this page.
    arguments:    OffsetType [in]
    return value: EntryLength
    --------------------------------------------------------------------------*/
    EntryLength GetEntryLength(OffsetType offset) const
    {
        const EntryLength restlen         = GetRestEntryLength(offset);
        const OffsetType  firstFreeOffset = ReadFirstFreeOffset();
        if ( offset + restlen + sizeof(EntryLength) <= firstFreeOffset ) 
            return restlen;
        else
            return firstFreeOffset - offset - sizeof(EntryLength);
    }
    /*!-------------------------------------------------------------------------
    function:     GetRestEntryLength()
    description:  Returns the length of all entry parts from this one on.
    arguments:    OffsetType [in]
    return value: EntryLength
    --------------------------------------------------------------------------*/
    EntryLength GetRestEntryLength(OffsetType offset) const
    {
        SAPDBERR_ASSERT_STATE( OffsetIsValid(offset) );
        return *(reinterpret_cast<const EntryLength*>(m_Frame.GetPtrAt(offset,sizeof(EntryLength))))
               - sizeof(EntryLength);
    }

    /*!-------------------------------------------------------------------------
    function:     GetEntry()
    description:  Returns a readable pointer to the space referenced by offset.
    arguments:    OffsetType [in]
    return value: SAPDB_Byte*
    --------------------------------------------------------------------------*/
    SAPDB_Byte* GetEntry(OffsetType offset)
    {
        SAPDBERR_ASSERT_STATE( OffsetIsValid(offset) );
        return reinterpret_cast<SAPDB_Byte*>(m_Frame.GetPtrAt(offset,sizeof(SAPDB_Byte)))
               + sizeof(EntryLength);
    }
    /*!-------------------------------------------------------------------------
    function:    GetIOSequence
    returns:     IOSequenceNo
    description: The corresponding IOSequence to the current referenced
                 logentry is returned. If IsValid() is false, the invalid
                 iosequence is returned.
    --------------------------------------------------------------------------*/
    IOSequenceNo GetIOSequence (OffsetType offset) const
    {
    
        OffsetType firstFreeOffset = ReadFirstFreeOffset();
        
        if ( offset < CalcNearestAlignedSize(GetHeaderSize())
             ||
             offset > firstFreeOffset )
            return IOSequenceNo();

        if ( offset == firstFreeOffset )
            return ReadLastWriterIOSequenceNo(); // for continued entries

        const OffsetCount maxoffsetcount = ReadOffsetCount();

        SAPDB_UInt        offsetno       = LOG_PAGE_FIRST_OFFSET;
        OffsetType        testoffset     = ReadOffsetOnPage (offsetno);
        
        while ( offset > testoffset && offsetno <= maxoffsetcount )
        {
            ++offsetno;
            testoffset = ReadOffsetOnPage (offsetno);
        }
        SAPDBERR_ASSERT_STATE( offsetno <= maxoffsetcount );
        
        return ReadFirstWriterIOSequenceNo() + offsetno - LOG_PAGE_FIRST_OFFSET;
    }
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Iterators                                                 */
    /* -----------------------------------------------------------------------*/

private:

    /*--------------------------------------------------------------------------
    declaration: Persistent types
    description: Persistent type definitions with compile time size check
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent     < Kernel_Date,  SAPDB_UInt4, 4 > PersistentDate;
    typedef SAPDB_Persistent     < Kernel_Time,  SAPDB_UInt4, 4 > PersistentTime;
    typedef SAPDB_Persistent     < IOSequenceNo, MultiUse1,   4 > PersistentIOSequenceNo0;
    typedef SAPDB_Persistent     < IOSequenceNo, SAPDB_UInt4, 4 > PersistentIOSequenceNo;
    typedef SAPDB_Persistent     < QueueID,      SAPDB_UInt2, 2 > PersistentQueueID;
    typedef SAPDB_Persistent     < OffsetType,   SAPDB_UInt2, 2 > PersistentOffsetType;
    typedef SAPDB_Persistent     < OffsetCount,  SAPDB_UInt1, 1 > PersistentOffsetCount;

#if defined(AIX) || defined(HPUX) || defined(SUN)
public:
#endif
    
    /*--------------------------------------------------------------------------
    declaration: PersistentHeader
    description: Persistent layout of data within kernel page header
    --------------------------------------------------------------------------*/
    // Must do this redefinition to avoid redundancy (compiler reason):
    // PersistentIOSequenceNo0 reuses MultiUse1 in Kernel_Page74::PersistentHeader
    typedef Kernel_Page74::PersistentHeader             PH1;
    typedef PersistentDate::PersistentFormat            PH2;    // log date
    typedef PersistentTime::PersistentFormat            PH3;    // log time
    typedef PersistentIOSequenceNo::PersistentFormat    PH4;
    typedef PersistentQueueID::PersistentFormat         PH5;
    typedef PersistentOffsetCount::PersistentFormat     PH6;
    typedef SAPDB_UInt1                                 PH7;

protected:

    struct PersistentHeader
    {
        PH1     m_KernelPage74Header;
        PH2     m_LastFlushDate;
        PH3     m_LastFlushTime;
        PH4     m_QueueIOSequenceNo;
        PH5     m_QueueID;
        PH6     m_OffsetCount;
        PH7     m_Filler;
    };

private:

    /*--------------------------------------------------------------------------
    declaration: ConverterPagePersistentHeader
    description: Compile time size check for persistent header
    --------------------------------------------------------------------------*/
    typedef SAPDB_Persistent
        < PersistentHeader, 
          PersistentHeader,
          sizeof(PH1) +
          sizeof(PH2) +
          sizeof(PH3) +
          sizeof(PH4) +
          sizeof(PH5) +
          sizeof(PH6) +
          sizeof(PH7) >        CheckedPersistentHeader;

    typedef Kernel_Page74::PersistentTrailer   PersistentTrailer;

    /*--------------------------------------------------------------------------
    function:     PersistentHeaderPtr()
    description:  Returns pointer to page header on frame
    arguments:    none
    return value: (PersistentHeader*) header pointer
    prototypes:   2
    --------------------------------------------------------------------------*/
    PersistentHeader* PersistentHeaderPtr();

    const PersistentHeader* PersistentHeaderPtr() const;

    /*--------------------------------------------------------------------------
    function:     PersistentTrailerPtr()
    description:  Returns pointer to page trailer on frame
    arguments:    none
    return value: (SAPDB_Byte*) trailer pointer
    prototypes:   2
    --------------------------------------------------------------------------*/
    SAPDB_Byte* PersistentTrailerPtr();

    const SAPDB_Byte* PersistentTrailerPtr() const;

protected:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Frame access (protected)                                     */
    /*--------------------------------------------------------------------------
    function:     GetHeaderSize()
    description:  Returns size of page header in bytes
    arguments:    none
    return value: (SizeType) header size
    --------------------------------------------------------------------------*/
    virtual SizeType GetHeaderSize() const;

    /*--------------------------------------------------------------------------
    function:     GetTrailerSize()
    description:  Returns size of page trailer in bytes

                  - Attention, trailer size is not fixed!
    arguments:    none
    return value: (SizeType) trailer size
    --------------------------------------------------------------------------*/
    virtual SizeType GetTrailerSize() const;

    /*--------------------------------------------------------------------------
    function:     ResetOffsetRelated()
    description:  Sets i/o write count to zero and writes it to frame
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void ResetOffsetCount();

    /*--------------------------------------------------------------------------
    function:     CalcNearestAlignedOffset()
    description:  Calculates aligned offset on the page next to given offset

                  - The returned offset is greater or equal 'offset'.
    arguments:    offset [in]  offset to be analyzed
    return value: (OffsetType) aligned offset
    --------------------------------------------------------------------------*/
    OffsetType CalcNearestAlignedOffset
        (OffsetType                     offset) const;

    /*--------------------------------------------------------------------------
    function:     CalcNearestAlignedSize()
    description:  Calculates aligned size for a given size

                  - The returned offset is greater or equal 'size'.
    arguments:    size [in]  size to be analyzed
    return value: (SizeType) aligned size
    --------------------------------------------------------------------------*/
    SizeType CalcNearestAlignedSize
        (SizeType                       size) const;

    /*--------------------------------------------------------------------------
    function:     MaxAlignedOffset()
    description:  Calculates max. aligned offset on the page for append operations
    arguments:    none
    return value: (OffsetType) maximum aligned offset on the page
    --------------------------------------------------------------------------*/
    OffsetType MaxAlignedOffset() const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Frame access (protected)                                  */
    /* -----------------------------------------------------------------------*/

};
/*!*****************************************************************************

   endclass: Log_Page

*******************************************************************************/

//------------------------------------------------------------------------------

inline Log_Page::Log_Page()
{}

//------------------------------------------------------------------------------

inline Log_Page::Log_Page
    (const PageFrame&               frame)
    : 
    Kernel_Page74(frame)
{}

//------------------------------------------------------------------------------

inline Log_Page::Log_Page
    (const PageFrame&               frame, 
     QueueID                        quid,    
     IOSequenceNo                   quseqno)
    : 
     Kernel_Page74(frame)
{
    InitializeFrame(quid, quseqno);
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::SetAndInitializeFrame
    (const PageFrame&               frame, 
     QueueID                        quid,    
     IOSequenceNo                   quseqno)
{
    SetFrame(frame);
    InitializeFrame(quid, quseqno);
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::InitializeFrame
    (QueueID                        quid,    
     IOSequenceNo                   quseqno)
{
    Kernel_Page74::InitializeFrame (Kernel_LogPageType);
    WriteQueueID                   (quid);
    Reset(quseqno);
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::Clear()
{
    WriteQueueIOSequenceNo (IOSequenceNo());
    WriteWriterIOSequenceNo(IOSequenceNo());
    WriteDate              (0);
    WriteTime              (0);
    ResetOffsetCount       ();
}

//------------------------------------------------------------------------------

inline bool
    Log_Page::IsEmpty() const
{
    return ReadFirstFreeOffset() == CalcNearestAlignedOffset(GetHeaderSize());
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::Reset
    (IOSequenceNo                   quseqno)
{
    WriteQueueIOSequenceNo (quseqno);
    WriteWriterIOSequenceNo(IOSequenceNo());
    WriteDate              (0);
    WriteTime              (0);
    ResetOffsetCount       ();
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::WriteWriterIOSequenceNo
    (IOSequenceNo                   ioseqno)
{
    WriteMultiUse1(ioseqno);
}

//------------------------------------------------------------------------------

inline Log_Page::IOSequenceNo 
    Log_Page::ReadLastWriterIOSequenceNo() const
{
    return ReadMultiUse1();
}

//------------------------------------------------------------------------------

inline Log_Page::IOSequenceNo 
    Log_Page::ReadFirstWriterIOSequenceNo() const
{
    return ReadLastWriterIOSequenceNo().IsValid() ?
           static_cast<SAPDB_UInt4>(ReadLastWriterIOSequenceNo() - ReadOffsetCount() + LOG_PAGE_FIRST_OFFSET) :
           static_cast<SAPDB_UInt4>(ReadLastWriterIOSequenceNo());
}

//------------------------------------------------------------------------------
inline void Log_Page::WriteDate (Kernel_Date date)
{
    PersistentHeaderPtr()->m_LastFlushDate = date;
}

//------------------------------------------------------------------------------
inline Kernel_Date Log_Page::ReadDate() const
{
    return PersistentHeaderPtr()->m_LastFlushDate;
}

//------------------------------------------------------------------------------
inline void Log_Page::WriteTime (Kernel_Time time)
{
    PersistentHeaderPtr()->m_LastFlushTime = time;
}

//------------------------------------------------------------------------------
inline Kernel_Time Log_Page::ReadTime() const
{
    return PersistentHeaderPtr()->m_LastFlushTime;
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::WriteQueueIOSequenceNo
    (IOSequenceNo                   quseqno)
{
    PersistentHeaderPtr()->m_QueueIOSequenceNo = 
        PersistentIOSequenceNo::ConvertInternalToPersistent(quseqno);
}

//------------------------------------------------------------------------------

inline Log_Page::IOSequenceNo 
    Log_Page::ReadQueueIOSequenceNo() const
{
    return PersistentIOSequenceNo::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_QueueIOSequenceNo);
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::WriteQueueID
    (QueueID                        quid)
{
    PersistentHeaderPtr()->m_QueueID = 
        PersistentQueueID::ConvertInternalToPersistent(quid);
}

//------------------------------------------------------------------------------

inline Log_Page::QueueID 
    Log_Page::ReadQueueID() const
{
    return PersistentQueueID::ConvertPersistentToInternal
           (PersistentHeaderPtr()->m_QueueID);
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::IncOffsetCount()
{
    OffsetType actoffset = ReadFirstFreeOffset();
    OffsetCount count = ReadOffsetCount();
    PersistentHeaderPtr()->m_OffsetCount = 
        PersistentOffsetCount::ConvertInternalToPersistent(count+1);
    WriteFirstFreeOffset(actoffset);
}

//------------------------------------------------------------------------------

inline Log_Page::OffsetCount 
    Log_Page::ReadOffsetCount() const
{
    return PersistentOffsetCount::ConvertPersistentToInternal
        (PersistentHeaderPtr()->m_OffsetCount);
}

//------------------------------------------------------------------------------

inline Log_Page::OffsetType 
    Log_Page::ReadOffsetOnPage
    (SAPDB_UInt                     offsetno) const
{
    SAPDB_UInt size = sizeof(PersistentOffsetType::PersistentFormat);

    SAPDBERR_ASSERT_STATE(((    m_Frame.Capacity()
                              - sizeof(PersistentTrailer) 
                              - offsetno * size
                            ) & 0xFFFF0000)
                            == 0);
    
    OffsetType offs = OffsetType (    m_Frame.Capacity()
                                    - sizeof(PersistentTrailer) 
                                    - offsetno * size);

    const OffsetType* offsptr = reinterpret_cast<const OffsetType*>
                                    (m_Frame.GetPtrAt(offs, size));

    return PersistentOffsetType::ConvertPersistentToInternal(*offsptr);
}

//------------------------------------------------------------------------------

inline void Log_Page::WriteLabel
    (Label                          label)
{
    WriteMultiUseT1(label);
}

//------------------------------------------------------------------------------

inline Log_Page::Label 
    Log_Page::ReadLabel() const
{
    return ReadMultiUseT1();
}

//------------------------------------------------------------------------------

inline SAPDB_Bool 
    Log_Page::HasFreeSpaceAvailable
    (SizeType                       size) const
{
    return FreeSpaceAvailable() >= size;
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::FreeSpaceAvailable(OffsetType currentoffset) const
{
    if ( currentoffset == 0 )
        currentoffset = ReadFirstFreeOffset();
    OffsetType firstfreeoffs = currentoffset + sizeof(EntryLength);

    SAPDBERR_ASSERT_STATE(((   m_Frame.Capacity()
                             - GetTrailerSize()
                             - sizeof(OffsetType))
                            & 0xFFFF0000)
                           ==0);
    OffsetType traileroffs   = m_Frame.Capacity() - GetTrailerSize()
                               - (OffsetType)sizeof(OffsetType);         //PTS 1111869 MB 01-09-20
                               
                               // conversion from 'size_t' to 'Kernel_IPage::OffsetType', 
                               
                               // the offset list may be incremented before writing
                               // so we must be pessimistic here
    
    return (firstfreeoffs >= traileroffs) ?
            0 : traileroffs - firstfreeoffs;
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::MaxFreeSpace() const
{
    return FreeSpaceAvailable(GetHeaderSize());
}

//------------------------------------------------------------------------------

inline SAPDB_Byte* Log_Page::Reserve (SizeType restsize,
                                      SizeType reservesize)
{
    SAPDBERR_ASSERT_STATE(HasFreeSpaceAvailable(reservesize));

    SizeType   internReserve = reservesize + sizeof (EntryLength);
    SizeType   lengthval     = restsize + sizeof (EntryLength);
    OffsetType actoffset     = ReadFirstFreeOffset();

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "restsize: "<<restsize<<", reservesize: "<<reservesize);
    
    if ( restsize == reservesize )
    {
        // only try to align, if the requested size fits into one page
        internReserve = CalcNearestAlignedSize(internReserve);
        lengthval     = lengthval;
    }

    WriteFirstFreeOffset(actoffset + internReserve);

    SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "actoffset: "<<actoffset<<" => "<<actoffset+internReserve);

    SAPDB_Byte* space = m_Frame.GetPtrAt(actoffset, internReserve);
    *(reinterpret_cast<EntryLength*> (space)) = lengthval;
    return space + sizeof (EntryLength);
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::Alignment()
{
    return LOG_PAGE_ALIGNMENT;
}

//------------------------------------------------------------------------------

inline Log_Page::PersistentHeader* 
    Log_Page::PersistentHeaderPtr()
{
    return reinterpret_cast<PersistentHeader*>
           (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}

//------------------------------------------------------------------------------

inline const Log_Page::PersistentHeader* 
    Log_Page::PersistentHeaderPtr() const
{
    return reinterpret_cast<const PersistentHeader*>
           (m_Frame.GetPtrAt(0, sizeof(PersistentHeader)));
}

//------------------------------------------------------------------------------

inline SAPDB_Byte* 
    Log_Page::PersistentTrailerPtr()
{
    return m_Frame.GetPtrAt(m_Frame.Capacity() - GetTrailerSize(), 
                            GetTrailerSize());
}

//------------------------------------------------------------------------------

inline const SAPDB_Byte* 
    Log_Page::PersistentTrailerPtr() const
{
    return m_Frame.GetPtrAt(m_Frame.Capacity() - GetTrailerSize(), 
                            GetTrailerSize());
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::GetHeaderSize() const
{
    return sizeof(PersistentHeader);
}

inline Log_Page::SizeType  
Log_Page::MinSpaceOffset() 
{
    return sizeof(PersistentHeader);
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::GetTrailerSize() const
{
    return 
        sizeof(PersistentTrailer) 
        + (ReadOffsetCount()) * sizeof(PersistentOffsetType::PersistentFormat);
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::Capacity() const
{
    return m_Frame.Capacity();
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::WriteFirstFreeOffset
    (OffsetType                     offset)
{
    OffsetType* offsptr = reinterpret_cast<OffsetType*>(PersistentTrailerPtr());
    *offsptr = PersistentOffsetType::ConvertInternalToPersistent(offset);
}

//------------------------------------------------------------------------------

inline Log_Page::OffsetType 
    Log_Page::ReadFirstFreeOffset() const
{
    const OffsetType* offsptr = reinterpret_cast<const OffsetType*>(PersistentTrailerPtr());
    return PersistentOffsetType::ConvertPersistentToInternal(*offsptr);
}

//------------------------------------------------------------------------------

inline void 
    Log_Page::ResetOffsetCount()
{
    PersistentHeaderPtr()->m_OffsetCount = 
        PersistentOffsetCount::ConvertInternalToPersistent(LOG_PAGE_FIRST_OFFSET);
    WriteFirstFreeOffset(CalcNearestAlignedOffset(GetHeaderSize()));
}

//------------------------------------------------------------------------------

inline Log_Page::OffsetType 
    Log_Page::CalcNearestAlignedOffset
    (OffsetType                     offset) const
{
#if LOG_PAGE_ALIGNMENT != 4
#error must adapt alignment in Log_Page::CalcNearestAlignedOffset()
#endif

    // 4 byte aligned offset:
    return ((offset + LOG_PAGE_ALIGNMENT - 1) >> 2) << 2;

    // 8 byte aligned offset:
    // return ((offset + 7) >> 3) << 3;
}

//------------------------------------------------------------------------------

inline Log_Page::SizeType 
    Log_Page::CalcNearestAlignedSize
    (SizeType                       size) const
{
    return CalcNearestAlignedOffset(size);
}

//------------------------------------------------------------------------------

inline Log_Page::OffsetType 
    Log_Page::MaxAlignedOffset() const
{
    static OffsetType offs 
            = CalcNearestAlignedOffset
              (m_Frame.Capacity() - 
               CalcNearestAlignedSize(GetTrailerSize()));
    return offs;
}

#endif // LOG_PAGE_HPP
            
