/*!*****************************************************************************

  module:       Log_Queue.hpp

  ------------------------------------------------------------------------------

  responsible:     UweH

  original author: TillL

  special area:    Logging

  description:     Log queue manager

  see also:

  copyright:       (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_QUEUE_HPP
#define LOG_QUEUE_HPP


#include "gsp00.h" // tsp00_TaskId
#include "heo55k.h"
#include "hgg08.h"

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "Container/Container_Vector.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_RawDeviceIterator.hpp"
#include "Logging/Log_WriterTaskSynchronizer.hpp"
#include "Logging/Log_QueueRingBuffer.hpp"
#include "Logging/Log_AfterImageSpace.hpp"
#include "IOManager/IOMan_LogPages.hpp"


/*!*****************************************************************************

   class: Log_Queue

   description:
   
   Log queue manager class

   Log_Queue provides user tasks (transaction contexts) with 
   memory space to be filled archive log entries.

   Log_Writer asks Log_Queue for filled log pages to be written
   out to log device.

*******************************************************************************/
class Log_Queue
{
public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration : Entry
    description:  log queue entry
    --------------------------------------------------------------------------*/
    typedef Log_QueueEntry                                      Entry;
    /*!-------------------------------------------------------------------------
    declaration : EntryPointer
    description:  pointer to log queue entry
    --------------------------------------------------------------------------*/
    typedef Log_QueueEntry*                                     EntryPointer;
    /*!-------------------------------------------------------------------------
    declaration : EntryCount
    description:  any number of log queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt                                          SizeType;
    /*!-------------------------------------------------------------------------
    declaration : EntryCount
    description:  any number of log queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt                                          EntryCount;
    /*!-------------------------------------------------------------------------
    declaration : IOSequenceNo
    description:  i/o sequence number of log pages
    --------------------------------------------------------------------------*/
    typedef Log_IOSequenceNo                                    IOSequenceNo;
    /*!-------------------------------------------------------------------------
    declaration : QueueID
    description:  identifier for log queues
    --------------------------------------------------------------------------*/
    typedef Log_QueueID                                         QueueID;
    /*!-------------------------------------------------------------------------
    declaration : Page
    description:  log page
    --------------------------------------------------------------------------*/
    typedef Log_Page                                            Page;
    /*!-------------------------------------------------------------------------
    declaration : PageFrame
    description:  page frame handler
    --------------------------------------------------------------------------*/
    typedef Kernel_Page::PageFrame                              PageFrame;
    /*!-------------------------------------------------------------------------
    declaration : PageFrameVector
    description:  vector of page frames 
    --------------------------------------------------------------------------*/
    typedef Container_Vector<PageFrame>                         PageFrameVector;
    /*!-------------------------------------------------------------------------
    declaration : RingBuffer
    description:  log queue ring buffer object
    --------------------------------------------------------------------------*/
    typedef Log_QueueRingBuffer                                 RingBuffer;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  allocator for queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDBMem_IRawAllocator                              Allocator;
    /*!-------------------------------------------------------------------------
    declaration : FrameAllocator
    description:  allocator for page frames
    --------------------------------------------------------------------------*/
    typedef Log_FrameAllocator                                  FrameAllocator;
    /*!-------------------------------------------------------------------------
    declaration : Iterator
    description:  iterator of ring buffer
    --------------------------------------------------------------------------*/
    typedef RingBuffer::Iterator                                Iterator;
    /*!-------------------------------------------------------------------------
    declaration : ConstIterator
    description:  const iterator of ring buffer
    --------------------------------------------------------------------------*/
    typedef RingBuffer::ConstIterator                           ConstIterator;
    /*!-------------------------------------------------------------------------
    declaration : TaskID
    description:  identifier of kernel task
    --------------------------------------------------------------------------*/
    typedef tsp00_TaskId                                        TaskID;
    /*!-------------------------------------------------------------------------
    declaration : SynchronizationMode
    description:  queue synchronization mode 
                  (indicates whether queue requires sync. regions)
    --------------------------------------------------------------------------*/
    enum SynchronizationMode
    {
        UnSynchronized,
        Synchronized
    };
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Log_Queue()
    description:  Constructor

    arguments:    queueid      [in] log queue id
                  syncmode     [in] queue synchronization mode
                  flushmode    [in] append mode for log pages
                  writersync   [in] writer task synchronization object
                  alloc        [in] raw allocator to be used by queue
                  frameres     [in] frame resource from which to allocate page frames
    return value: none
    --------------------------------------------------------------------------*/
    Log_Queue
        (QueueID                        queueid,
         SynchronizationMode            syncmode,
         Log_FlushMode                  flushmode,
         Log_WriterTaskSynchronizer&    writersync,
         Allocator&                     alloc,
         FrameAllocator&                framealloc)
        :
         m_QueueID              (queueid),
         m_Synchronized         (syncmode == Synchronized ? true : false),
         m_FlushMode            (flushmode),
         m_WriterTaskSync       (writersync),
         m_RingBuffer           (alloc, framealloc),
         m_CurrentAppend        (0),
         m_CurrentOutput        (0),
         m_Allocated            (false),
         m_PagesUsed            (1), // PTS 1115332 UH 2002-04-29 added
         m_MaxPagesUsed         (1), // PTS 1115332 UH 2002-04-29 added
         m_QueueFullSuspendCount(0),
         m_ReserveCount         (0),
         m_WaitCount            (0),
         m_GroupCommitCount     (0),
         m_MaxCountOfWaitingTasksPerPage(0),
         m_LogIsFull(true)
    {}

    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  Initializes queue structures

                  - Performs memory allocation of queue entries and page frames.

    arguments:    size   [in]   queue size; number of entries / page frames
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool Initialize 
        (EntryCount                     size)
    {
        // queue needs at least 3 pages to work properly
        SAPDBERR_ASSERT_ARGUMENT(size >= 3);

        if (m_Allocated == false)
        {
            if (!m_RingBuffer.Initialize(size))
                return false;
            m_Allocated = true;
        }

        m_CurrentAppend = m_RingBuffer.Anchor();
        m_CurrentOutput = m_RingBuffer.Anchor();
        m_IOSequenceNo.Invalidate();
        m_IOSequenceNoOfFirstPageOfLastWrittenBlock.Invalidate();
        return true;
    }
    
    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes queue structures

                  - Performs memory deallocation
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Delete()
    {
        m_RingBuffer.Delete();
    }
    
    /*!-------------------------------------------------------------------------
    function:     ~Log_Queue()
    description:  Destructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    ~Log_Queue()
    {
        Delete();
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: User task services                                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function:     Reserve()
    description:  Reserves requested memory space on log pages

                  - The requested memory of 'size' bytes is reserved by the queue 
                    on its internal log pages. A split space description of the
                    memory chunks is returned in 'space'. 
                  - The first chunk must have a size of at least 'unsplitsize'
                    bytes. If this space is not available on the current page, 
                    reservation is extended to the next page(s).
                  - If there is currently not enough log space available to
                    fulfill the request, the calling task identified by 'usertaskid' 
                    is suspended. No space at all is reserved for the task in 
                    this case until resumed.
                  - If the Log-Device is full Reserve() will return false, because
                    normal operations cannot be logged anymore. In all other cases
                    it will return true (PTS 1111924 mb 2001-09-26 )
                  - The method expects 'space' to be unassigned.

    arguments:    usertaskid  [in]      calling user task id
                  size        [in]      requested memory size in bytes
                  unsplitsize [in]      min. unsplit size of first chunk
                  space       [in/out]  split space descriptor
    --------------------------------------------------------------------------*/
    bool Reserve 
        (TaskID                         usertaskid,
         SizeType                       size,
         SizeType                       unsplitsize,
         bool                           useReservedSpace, // PTS 1117320 UH 2002-10-28
         Log_AfterImageSpace&           space)
    {
        if (m_LogIsFull) //PTS 1111924 mb 2001-09-26
        {
            // the utility task must not be suspended
            if (g01pid.taskUtility_gg00 == usertaskid)
                return false;

            if ( ! useReservedSpace )
                // PTS 1117320 UH 2002-10-28
                WaitUntilSpaceIsAvailable(usertaskid,WaitForLogFree);
        }

        Log_Page::OffsetType dummy;
        ReserveSpace(usertaskid, size, unsplitsize, space, dummy);
        return true;
    }
    
    /*!-------------------------------------------------------------------------
    function:     ReserveForEOT()
    description:  Reserves requested memory space for writing out an 
                  end-of-transaction entry

    arguments:    usertaskid [in]      calling user task id
                  size       [in]      requested memory size in bytes
                  space      [out]     split space descriptor
                  entrypos   [out]     position of entry on page
    --------------------------------------------------------------------------*/
    void ReserveForEOT
        (TaskID                         usertaskid,
         SizeType                       size,
         Log_AfterImageSpace&           space,
         Log_Page::OffsetType&          entrypos)
    {
        ReserveSpace(usertaskid, size, size, space, entrypos);
    }
    
    /*!-------------------------------------------------------------------------
    function:     UserTaskReady()
    description:  Tells the queue that an associated user task is ready

                  - A user task calls this method in order to tell the queue 
                    that previously reserved space is ready for i/o.
                  - As a consequence of this call, the writer task is resumed
                    if at least one page is available for output.

    arguments:    usertaskid   [in]      calling user task id
                  space        [in/out]  split space descriptor
    return value: none
    --------------------------------------------------------------------------*/
    void UserTaskReady
        (TaskID                         usertaskid,
         Log_AfterImageSpace&           space);
    
    /*!-------------------------------------------------------------------------
    function:     UserTaskEOTReady()
    description:  Tells the queue that an associated user task is ready with
                  writing an EOT entry

                  - A user task calls this method in order to tell the queue 
                    that previously reserved space is ready for i/o.
                  - As a consequence of this call, the writer task is resumed
                    if at least one page is available for output.
                    
                  - The user task is suspended until the log writer has flushed 
                    the page. The user task is resumed by the writer.

                  - The parameter 'wioseqno' is filled with the writer i/o
                    sequence number of the log page on that the entry has been
                    written.

    arguments:    usertaskid   [in]      calling user task id
                  space        [in/out]  split space descriptor
                  wioseqno     [out]     writer i/o sequence no
                  offset       [out]     offset of page on device
    return value: none
    --------------------------------------------------------------------------*/
    void UserTaskEOTReady
        (TaskID                usertaskid,
         Log_AfterImageSpace  &space,
         IOSequenceNo         &wioseqno,
         Log_RawDeviceOffset  &offset);
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: User task services                                        */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Writer task services                                         */
    /*!-------------------------------------------------------------------------
    function:     GetOutput()
    description:  Puts together filled log pages for write i/o

                  - Log pages that are ready for i/o are assigned to the vector 
                    'output'.
                  - These pages are marked with queue id and queue i/o
                    sequence number.
                    
    arguments:    writertaskid   [in]      calling writer task id
                  output         [in/out]  collection of log pages
                  pagecount      [out]     number of added pages
                  replacefirst   [out]     true if first page returned in 'output'
                                           must replace (overwrite) the last
                                           page returned in the previous call 
                                           to GetOutput()
    return value: none
    --------------------------------------------------------------------------*/
    void GetOutput (TaskID          writertaskid,
                    SAPDB_UInt      maxPages,
                    IOMan_LogPages &output,
                    SAPDB_UInt     &pagecount,
                    SAPDB_Bool     &replacefirst);

    /*!-------------------------------------------------------------------------
    function:     WriterTaskReady()
    description:  Tells the queue that the queue writer has successfully
                  written some pages

                  - As a consequence of this call, the user tasks in the
                    user task wait queue are resumed.

    arguments:    writertaskid      [in]  calling writer task id
                  wioseqno          [in]  contains the writer i/o sequence no
                                          with which the first of the 'pagecount'
                                          pages was written
                  writePosition     [in]  iterator, which points to the first page written
                  pagecount         [in]  number of pages that have been written
    return value: none
    --------------------------------------------------------------------------*/
    void WriterTaskReady
        (TaskID                         writertaskid,
         IOSequenceNo                   wioseqno,
         Log_RawDeviceIterator          writePosition,
         SAPDB_UInt                     pagecount);

    /*!-------------------------------------------------------------------------
    function:     SetLogFullState()
    description:  Sets the state of m_LogIsFull 
                  PTS 1111924 mb 2001-09-26
    arguments:    writertaskid      [in]  calling writer task id
                  state             [in]  new state of m_LogIsFull
    return value: none
    --------------------------------------------------------------------------*/
    void SetLogFullState
        (TaskID                         writerTaskId,
         bool                           newState)
    {
        EnterRegion(writerTaskId);
        if (m_LogIsFull != newState)
        {
            m_LogIsFull = newState;
            if (!m_LogIsFull)
            {
                m_LogFullWaitList.PopAllTaskIdsAndResume(writerTaskId);
            }
        }
        LeaveRegion(writerTaskId);
    }
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Writer task services                                      */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Queue status                                                 */
    /*!-------------------------------------------------------------------------
    function:     GetID()
    description:  Returns the queue id
    arguments:    none
    return:       (QueueID) id of this queue
    --------------------------------------------------------------------------*/
    QueueID GetID() const
    {
        return m_QueueID;
    }

    /*!-------------------------------------------------------------------------
    function:     GetSize()
    description:  Returns the queue size in page frames
    arguments:    none
    return:       (SAPDB_UInt) queue size
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetSize() const
    {
        return m_RingBuffer.GetEntryCount();
    }

    /*!-------------------------------------------------------------------------
    function:     GetMaxPagesUsed()
    description:  Returns number of pages maximal used
    arguments:    none
    return:       (SAPDB_UInt)
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetMaxPagesUsed() const
    {
        return m_MaxPagesUsed;
    }

    /*!-------------------------------------------------------------------------
    function:     GetQueueFullSuspendCount()
    description:  Returns number of user task suspends that have occurred 
                  due to queue full
    arguments:    none
    return:       (SAPDB_UInt) suspend count 
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetQueueFullSuspendCount() const
    {
        return m_QueueFullSuspendCount;
    }

    /*!-------------------------------------------------------------------------
    function:     GetReserveCount()
    description:  Returns number of requests to reserve space for redo entries.
    arguments:    none
    return:       (SAPDB_UInt) reserve count 
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetReserveCount() const
    {
        return m_ReserveCount;
    }

    /*!-------------------------------------------------------------------------
    function:     GetWaitCount()
    description:  Returns number of waits for io.
    arguments:    none
    return:       (SAPDB_UInt) count 
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetWaitCount() const
    {
        return m_WaitCount;
    }

    /*!-------------------------------------------------------------------------
    function:     GetGroupCommitCount()
    description:  Returns number of group commits
    arguments:    none
    return:       (SAPDB_UInt) count 
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetGroupCommitCount() const
    {
        return m_GroupCommitCount;
    }

    /*!-------------------------------------------------------------------------
    function:     GetMaxCountOfWaitingTasksPerPage()
    arguments:    none
    return:       (SAPDB_UInt) 
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetMaxCountOfWaitingTasksPerPage() const
    {
        return m_MaxCountOfWaitingTasksPerPage;
    }

    /*!-------------------------------------------------------------------------
    function:     ResetStatistics()
    description:  resets statistic counter to zero
    arguments:    taskid [in]
    return:       none
    --------------------------------------------------------------------------*/
    void ResetStatistics(tsp00_TaskId taskid)
    {
        EnterRegion(taskid); 
        m_QueueFullSuspendCount         = 0;
        m_ReserveCount                  = 0;
        m_WaitCount                     = 0;
        m_GroupCommitCount              = 0;
        m_MaxCountOfWaitingTasksPerPage = 0;
        LeaveRegion(taskid); 
    }

    /*!-------------------------------------------------------------------------
    function:     WriteToTrace()
    description:  writes the queues information into the trace.
    arguments:    none
    return:       none
    --------------------------------------------------------------------------*/
    void WriteToTrace() const;
    /*!-------------------------------------------------------------------------
    function:     WriteEntryToTrace()
    description:  writes one queue entry to trace
    arguments:    iter    [in]
                  entryno [in]
    return:       none
    --------------------------------------------------------------------------*/
    void WriteEntryToTrace(const Log_QueueEntry &entry,
                           const char           *title           = 0,
                           SAPDB_UInt            entryno         = 0,
                           bool                  isCurrentAppend = false,
                           bool                  isCurrentOutput = false) const;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Queue status                                              */
    /* -----------------------------------------------------------------------*/

private:
    typedef enum WaitQueueSelector
    {
        WaitForWriterAction,
        WaitForLogFree
    } WaitQueueSelector;


    /*-----------------------------------------------------------------------
    function:     ReserveSpace()
    description:  Reserves requested space

    arguments:    usertaskid  [in]  calling user task id
                  size        [in]  requested memory size in bytes
                  unsplitsize [in]  min. unsplit size of first chunk
                  space       [out] space to be filled
                  entrypos    [out] position of entry on page
    --------------------------------------------------------------------------*/
    void ReserveSpace
        (TaskID                         usertaskid,
         SizeType                       size,
         SizeType                       unsplitsize,
         Log_AfterImageSpace&           space,
         Log_Page::OffsetType&          entrypos);
    /*-----------------------------------------------------------------------
    function:     WaitUntilSpaceIsAvailable()
    arguments:    taskid [in]        calling task id
                  queueSelector [in] determines which queue shall be used for enqueueing
    --------------------------------------------------------------------------*/
    void WaitUntilSpaceIsAvailable(tsp00_TaskId usertaskid, WaitQueueSelector queueSelector);
    /*-----------------------------------------------------------------------
    function:     FindSpaceToReserve()
    arguments:    usertaskid  [in]  calling task id
                  size        [in]
                  unsplitsize [in]
    --------------------------------------------------------------------------*/
    void FindSpaceToReserve (tsp00_TaskId usertask,
                             SizeType     size,
                             SizeType     unsplitsize);
    /*-----------------------------------------------------------------------
    function:     FindSpaceToReserve()
    arguments:    usertaskid  [in]  calling task id
                  size        [in]
                  unsplitsize [in]
    --------------------------------------------------------------------------*/
    void FindStartPointToReserve (tsp00_TaskId usertask,
                                  SizeType     size,
                                  SizeType     unsplitsize);
    /*--------------------------------------------------------------------------
    function:     OutputAvailableAt()
    description:  Checks whether referenced log page is available for output

                  - This frame is ready for output only if two
                    conditions are fulfilled:
                  - (1) there are actually no "busy" user tasks writing on the frame
                  - (2a) the current append position is beyond this page, or
                    (2b) there are user tasks in 'commit' or 'savepoint complete' 
                         state waiting for the frame to be flushed
                         (in this case the frame doesn't need to be full)
    arguments:    taskid [in]  calling task id
                  iter   [in]  points to entry that is to be checked
    return value: (SAPDB_Bool) true if available
    --------------------------------------------------------------------------*/
    SAPDB_Bool OutputAvailableAt
        (TaskID                         taskid,
         Iterator                       iter) const
    {
        Log_QueueEntry& entry = *iter;
        
        return (entry.GetBusyTaskCount() == 0)                      // (1)
               &&
               ((iter != m_CurrentAppend) ||                        // (2a)
                (entry.TasksAreWaitingForIO(taskid)));              // (2b)
    }
    /*-----------------------------------------------------------------------
    function:     EnterRegion()
    description:  Enters region
    arguments:    taskid [in]  calling task id
    return value: none
    --------------------------------------------------------------------------*/
    void EnterRegion
        (TaskID                         taskid) const
    {
        if (m_Synchronized)
            vbegexcl(taskid, g08logqueue0 + m_QueueID);
    }
    /*-----------------------------------------------------------------------
    function:     LeaveRegion()
    description:  Exits region
    arguments:    taskid [in]  calling task id
    return value: none
    --------------------------------------------------------------------------*/
    void LeaveRegion
        (TaskID                         taskid) const
    {
        if (m_Synchronized)
            vendexcl(taskid, g08logqueue0 + m_QueueID);
    }
    /*-----------------------------------------------------------------------
    function:     MaxReserveSize()
    description:  Returns the maximum size one single log entry can have
                  based on the actual queue size
    arguments:    none
    return value: (SizeType) maximum reserve size
    --------------------------------------------------------------------------*/
    SizeType MaxReserveSize() const
    {
        return m_RingBuffer.GetSize() * m_CurrentAppend->GetPage().MaxFreeSpace();
    }
    /*-----------------------------------------------------------------------
    function:     CopyCurrentAppend()
    arguments:    none
    description:  Copies the page at m_CurrentAppend to the next queue entry.
                  This is used to if a EOT entry was written by the user or
                  if the user wants to reserve space and the writer will already
                  write the page (IsMarkedForIO()==true)
    --------------------------------------------------------------------------*/
    void CopyCurrentAppend();
    /*-----------------------------------------------------------------------
    function:     UseNextPageCausedByIO()
    arguments:    none
    description:  This can copy page contents to another page in the queue or
                  simply use the next page.
    --------------------------------------------------------------------------*/
    void UseNextPageCausedByIO();
    /*-----------------------------------------------------------------------
    function:     UseNextPage()
    arguments:    none
    description:  This uses the next queue page (current append is incremented).
                  Space must be available.
    --------------------------------------------------------------------------*/
    void UseNextPage();

    /*--------------------------------------------------------------------------
    declaration : m_RingBuffer
    description:  ring buffer for queue entries
    --------------------------------------------------------------------------*/
    RingBuffer                              m_RingBuffer;
    /*--------------------------------------------------------------------------
    declaration : m_CurrentAppend
    description:  iterator at current append position within ring buffer
    --------------------------------------------------------------------------*/
    Iterator                                m_CurrentAppend;
    /*--------------------------------------------------------------------------
    declaration : m_CurrentAppend
    description:  iterator at current output position within ring buffer
                  (references page that is to be written out next)
    --------------------------------------------------------------------------*/
    Iterator                                m_CurrentOutput;
    /*--------------------------------------------------------------------------
    declaration : m_WriterTaskSync
    description:  log writer task synchronization object
    --------------------------------------------------------------------------*/
    Log_WriterTaskSynchronizer&             m_WriterTaskSync;
    /*--------------------------------------------------------------------------
    declaration : m_QueueFullWaitList
    description:  user tasks waiting when queue full 
    --------------------------------------------------------------------------*/
    Kernel_WaitStack                        m_QueueFullWaitList;
    /*--------------------------------------------------------------------------
    declaration : m_LogFullWaitList
    description:  user tasks waiting when log is full 
    --------------------------------------------------------------------------*/
    Kernel_WaitStack                        m_LogFullWaitList;
    /*--------------------------------------------------------------------------
    declaration : m_IOSequenceNo
    description:  queue i/o sequence number
    --------------------------------------------------------------------------*/
    IOSequenceNo                            m_IOSequenceNo;
    /*--------------------------------------------------------------------------
    declaration : m_IOSequenceNoOfFirstPageOfLastWrittenBlock
    description:  last queue i/o sequence number, which was given to the writer.
    --------------------------------------------------------------------------*/
    IOSequenceNo                            m_IOSequenceNoOfFirstPageOfLastWrittenBlock;
    /*--------------------------------------------------------------------------
    declaration : m_QueueID
    description:  ID of the log queue
    --------------------------------------------------------------------------*/
    const QueueID                           m_QueueID;
    /*--------------------------------------------------------------------------
    declaration : m_Synchronized
    description:  indicates whether queue requires synchronization
    --------------------------------------------------------------------------*/
    const SAPDB_Bool                        m_Synchronized;
    /*--------------------------------------------------------------------------
    declaration : m_FlushMode
    description:  log queue flush mode
    --------------------------------------------------------------------------*/
    Log_FlushMode                           m_FlushMode;
    /*--------------------------------------------------------------------------
    declaration : m_Allocated
    description:  indicates whether queue structures have been allocated
    --------------------------------------------------------------------------*/
    SAPDB_Bool                              m_Allocated;
    /*--------------------------------------------------------------------------
    declaration : m_PagesUsed
    description:  number of pages currently used
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_PagesUsed; // PTS 1115332 UH 2002-04-29 added
    /*--------------------------------------------------------------------------
    declaration : m_MaxPagesUsed
    description:  number of pages used maximal
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_MaxPagesUsed; // PTS 1115332 UH 2002-04-29 added
    /*--------------------------------------------------------------------------
    declaration : m_QueueFullSuspendCount
    description:  number of suspends that have occurred due to queue full
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_QueueFullSuspendCount;
    /*--------------------------------------------------------------------------
    declaration : m_ReserveCount
    description:  number of reserves that have occurred
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_ReserveCount;
    /*--------------------------------------------------------------------------
    declaration : m_WaitCount
    description:  number of waits for io that have occurred
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_WaitCount;
    /*--------------------------------------------------------------------------
    declaration : m_GroupCommitCount
    description:  number of group commits
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_GroupCommitCount;
    /*--------------------------------------------------------------------------
    declaration : m_MaxCountOfWaitingTasksPerPage
    --------------------------------------------------------------------------*/
    SAPDB_UInt                              m_MaxCountOfWaitingTasksPerPage;
    /*--------------------------------------------------------------------------
    declaration : m_LogIsFull
    description : If set to TRUE only reserveEOT will be succeed, calls to
                  reserve() will be rejected PTS 1111924 mb 2001-09-26
    --------------------------------------------------------------------------*/
    bool                                    m_LogIsFull;
};
/*!*****************************************************************************

   endclass: Log_Queue

*******************************************************************************/
#endif  // LOG_QUEUE_HPP
