/*!*****************************************************************************

  module:       Log_QueueEntry.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  author:       TillL

  special area: Logging

  description:  QueueEntry (== "control block") of the log queue

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_QUEUE_ENTRY_HPP
#define LOG_QUEUE_ENTRY_HPP


#include "gsp00.h"

#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "Container/Container_Vector.hpp"
#include "Logging/Log_Page.hpp"


/*!*****************************************************************************

   class: Log_QueueEntry

   description: Entry of log queue

*******************************************************************************/
class Log_QueueEntry
{

public:

    /*!-------------------------------------------------------------------------
    declaration : Page
    description:  handler of log page frames
    --------------------------------------------------------------------------*/
    typedef Log_Page                                    Page;
    /*!------------------------------------------------------------------------
    declaration : TaskID
    description:  identifier of kernel task
    --------------------------------------------------------------------------*/
    typedef tsp00_TaskId                                TaskID;
    /*!-------------------------------------------------------------------------
    declaration : TaskCount
    description:  any number of tasks
    --------------------------------------------------------------------------*/
    typedef SAPDB_UInt4                                 TaskCount;
    /*!-------------------------------------------------------------------------
    declaration : Entry
    description:  log queue entry
    --------------------------------------------------------------------------*/
    typedef Log_QueueEntry                              QueueEntry;
    /*!-------------------------------------------------------------------------
    declaration : EntryPointer
    description:  pointer to log queue entry
    --------------------------------------------------------------------------*/
    typedef Log_QueueEntry*                             EntryPointer;
    /*!-------------------------------------------------------------------------
    declaration : PageFrame
    description:  frame containing a log page
    --------------------------------------------------------------------------*/
    typedef Kernel_Page::PageFrame                      PageFrame;
    /*!-------------------------------------------------------------------------
    declaration : Allocator
    description:  raw allocator
    --------------------------------------------------------------------------*/
    typedef SAPDBMem_IRawAllocator                      Allocator;

    /*!-------------------------------------------------------------------------
    function:     Log_QueueEntry()
    description:  Default constructor
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    Log_QueueEntry
        (Allocator&                     alloc)
        :
         m_IOpositions(alloc),
         m_LockedByWriter(false),
         m_BusyTaskCount(0)
    {}

    /*!-------------------------------------------------------------------------
    function:     Log_QueueEntry()
    description:  Default constructor
    arguments:    frame [in] frame where log page is located
    return value: none
    --------------------------------------------------------------------------*/
    Log_QueueEntry
        (Allocator&                     alloc,
         const PageFrame&               frame)
        :
         m_IOpositions(alloc),
         m_LockedByWriter(false),
         m_BusyTaskCount(0),
         m_Page(frame)
    {}

    /*!-------------------------------------------------------------------------
    function:     SetFrame()
    description:  Sets page frame
    arguments:    frame [in] frame where log page is located
    return value: none
    --------------------------------------------------------------------------*/
    void SetFrame
        (const PageFrame&               frame)
    {
        m_Page.SetFrame(frame);
    }

    /*!-------------------------------------------------------------------------
    function:     GetPage()
    description:  Returns reference to associated page
    arguments:    none
    return value: (Page&) reference to page
    --------------------------------------------------------------------------*/
    Page& GetPage()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_Page;
    }

    const Page& GetPage() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_Page;
    }

    /*!-------------------------------------------------------------------------
    function:     PageIsAssigned()
    description:  Checks whether a page has been assigned to this entry
    arguments:    none
    return value: (SAPDB_Bool) true if assigned
    --------------------------------------------------------------------------*/
    SAPDB_Bool PageIsAssigned() const
    {
        return m_Page.IsAssigned();
    }

    /*!-------------------------------------------------------------------------
    function:     GetBusyTaskCount()
    description:  Returns actual busy task count (synchronized)
    arguments:    none
    return value: (TaskCount) actual task count
    --------------------------------------------------------------------------*/
    TaskCount GetBusyTaskCount() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_BusyTaskCount.GetValue();
    }

    /*!-------------------------------------------------------------------------
    function:     IncBusyTaskCount()
    description:  Increments busy task count (synchronized)
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void IncBusyTaskCount()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(m_BusyTaskCount.GetValue() < 1000); // too many tasks indicate error
        m_BusyTaskCount.Increment();
    }

    /*!-------------------------------------------------------------------------
    function:     DecBusyTaskCount()
    description:  Decrements busy task count (synchronized)
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void DecBusyTaskCount()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(m_BusyTaskCount.GetValue() > 0);
        m_BusyTaskCount.Decrement();
    }

    /*!-------------------------------------------------------------------------
    function:     MarkForIO()
    description:  sets a flag, that this page will be written by the logwriter
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void MarkForIO()
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        m_LockedByWriter = true;
    }

    /*!-------------------------------------------------------------------------
    function:     IsMarkedForIO()
    description:  returns true, if this page will be written by the log writer
    arguments:    none
    return value: bool
    --------------------------------------------------------------------------*/
    bool IsMarkedForIO() const
    {
        SAPDBERR_ASSERT_STATE(PageIsAssigned());
        return m_LockedByWriter;
    }

    /*!-------------------------------------------------------------------------
    function:     RegisterForIOWait()
    description:  Registers specified user task as waiting for the i/o of the
                  associated page

                  - The parameter 'wioseqno' contains a reference to an
                    i/o seqno. This variable 

    arguments:    usertaskid  [in]  id of user task to be registered
                  wioseqno    [out] writer i/o sequence no
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool RegisterForIOWait
        (TaskID                         usertaskid,
         Log_IOSequenceNo              &wioseqno,
         Log_RawDeviceOffset           &offset)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_QueueEntry::RegisterForIOWait", LogVolume_Trace, 5);
        
        if (m_IOpositions.IsFull())
            if (!m_IOpositions.Reserve(m_IOpositions.GetSize() + 1))
                return false;

        m_IOpositions.InsertEnd( IOposition(&wioseqno, &offset) );

        m_IOWaitStack.PushTaskId(usertaskid);
        
        return true;
    }
    
    /*!-------------------------------------------------------------------------
    function:     ResumeWaitingTasks()
    description:  Resumes all waiting tasks

                  - This command is to be invoked when the writer task has
                    flushed the page to device.
                  - All user tasks that have been suspended with the 
                    'RegisterForIOWait()' method are resumed.

    arguments:    writertaskid [in]  id of calling writer task
                  wioseqno     [in]  writer i/o seqno with which this entry
                                     was written
    return value: none
    --------------------------------------------------------------------------*/
    void ResumeWaitingTasks
         (TaskID                        writertaskid,
          Log_IOSequenceNo              wioseqno,
          Log_RawDeviceOffset           offset)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_QueueEntry::ResumeWaitingTasks", LogVolume_Trace, 5);

        SAPDBTRACE_WRITELN (LogVolume_Trace, 6, "ioseqno: "<<wioseqno<<", offset: "<<offset);
        
        for (SAPDB_UInt i = 0; i < m_IOpositions.GetSize(); i++)
        {
            *(m_IOpositions[i].pIOsequence)   = wioseqno;
            *(m_IOpositions[i].pDeviceOffset) = offset;
        }    
        if ( ! m_IOpositions.Resize(0) )
            // PTS 1117126 UH 2002-08-07 if-clause added
            RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,
                       "m_IOpositions.Resize(0) failed") );

        m_IOWaitStack.PopAllTaskIdsAndResume(writertaskid);
        
        m_LockedByWriter = false;
    }
    
    /*!-------------------------------------------------------------------------
    function:     TasksAreWaitingForIO()
    description:  Checks whether there are any tasks waiting for the page i/o
    arguments:    writertaskid [in]  id of calling writer task
    return value: (SAPDB_Bool) true if tasks are waiting
    --------------------------------------------------------------------------*/
    SAPDB_Bool TasksAreWaitingForIO
         (TaskID                        writertaskid)
    {
        return !m_IOWaitStack.IsEmpty(writertaskid);
    }
    
    /*!-------------------------------------------------------------------------
    function:     GetCountOfWaitingTasks()
    return value: (SAPDB_UInt) 
    --------------------------------------------------------------------------*/
    SAPDB_UInt GetCountOfWaitingTasks ()
    {
        return m_IOWaitStack.GetCountOfWaitingTasks();
    }
    
protected:

    /*--------------------------------------------------------------------------
    declaration:  IOposition
    description:  information from the writer for the caller
    --------------------------------------------------------------------------*/
    typedef struct IOposition
    {
        Log_IOSequenceNo    *pIOsequence;
        Log_RawDeviceOffset *pDeviceOffset;
        IOposition()
        {
            pIOsequence   = NULL;
            pDeviceOffset = NULL;
        }
        IOposition (Log_IOSequenceNo    *ioseq,
                    Log_RawDeviceOffset *offset)
        {
            pIOsequence   = ioseq;
            pDeviceOffset = offset;
        }
    } IOposition;
    /*--------------------------------------------------------------------------
    declaration:  m_Page
    description:  log page handler
    --------------------------------------------------------------------------*/
    Page                                            m_Page;
    /*--------------------------------------------------------------------------
    declaration:  m_LockedByWriter
    description:  this is true, if the writer will write the page
    --------------------------------------------------------------------------*/
    bool                                            m_LockedByWriter;
    /*--------------------------------------------------------------------------
    declaration:  m_BusyTaskCount
    description:  number of user tasks currently writing on associated frame
    --------------------------------------------------------------------------*/
    RTESync_InterlockedCounter<TaskCount>           m_BusyTaskCount;
    /*--------------------------------------------------------------------------
    declaration : m_IOWaitStack
    description:  user tasks in 'EOT' state waiting for write i/o ready
    --------------------------------------------------------------------------*/
    Kernel_WaitStack                                m_IOWaitStack;
    /*--------------------------------------------------------------------------
    declaration : m_IOpositions
    description:  vector of pointers to i/o sequence numbers and offsets
    --------------------------------------------------------------------------*/
    Container_Vector<IOposition>                    m_IOpositions;
};
/*!*****************************************************************************

   endclass: Log_QueueEntry

*******************************************************************************/




#endif // LOG_QUEUE_ENTRY_HPP
