
/*!**************************************************************************

    module      : Log_ReadWriteActionImage.hpp

    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines a class to read/write before and after images of any action

    last changed: 2001-03-28

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2001 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef LOG_READ_WRITE_ACTION_IMAGE_H
#define LOG_READ_WRITE_ACTION_IMAGE_H

/*===========================================================================*
*  INCLUDES                                                                  *
*============================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "DataAccess/Data_ISplitSpace.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_IAction.hpp"

#include "ggg00.h"

/*===========================================================================*
*  EXTERNAL FUNCTIONS                                                        *
*============================================================================*/

inline SAPDB_UInt4 Log_AlignedImageLen (SAPDB_UInt4 Len)
{
    return ( ((Len + 7) >> 3) << 3 );
}


/*===========================================================================*
*  CLASSES                                                                   *
*============================================================================*/


//===================================================================================
/*! class:       Log_ActionImageReader

    description: reading the image of an action
*/
class Log_ActionImageReader
{

private:
    Data_SplitSpaceReader  &m_SplitSpaceReader;
    SAPDBMem_IRawAllocator &m_IRawAllocator;
    
public:

    //--------------------------------------------------------------Log_ActionImageReader---
    /*! function:    Log_ActionImageReader ctor
    
        arguments:   SplitSpaceReader [in out]
                     IRawAllocator    [in out] is used, if an image part is not contigous in a page
    */
    inline Log_ActionImageReader (
        Data_SplitSpaceReader  &SplitSpaceReader,
        SAPDBMem_IRawAllocator &IRawAllocator);
    
    //--------------------------------------------------------------Log_ActionImageReader---
    /*! function:    CopyImage
    
        description: copies ImageLen bytes from the undo or redo file to the memory specified by pImage
      
        arguments:   pImage   [in]  
                     ImageLen [in]
                     IsOk     [out]
    */
    void CopyImage (
        SAPDB_Byte  *pImage,
        SAPDB_UInt4  ImageLen,
        bool        &IsOk);
    
    //--------------------------------------------------------------Log_ActionImageReader---
    /*! function:    GetAllocator
    
        return value: Allocator
    */
    SAPDBMem_IRawAllocator &GetAllocator ()
    {
        return (m_IRawAllocator);
    }
        
    //--------------------------------------------------------------Log_ActionImageReader---
    /*! function:    PersistentRead
    
        description: locates the image onto a page of the redo/undo file or onto an allocated space

                     an internaly aligned ImageLen is used to locate the image
      
        arguments:   pImage         [in]  
                     ImageLen       [in]
                     pUsedAllocator [out] not null, if pImage points to an allocated space
                     IsOk           [out]
    */
    void PersistentRead (
        SAPDB_Byte             *&pImage,
        SAPDB_UInt4              ImageLen,
        SAPDBMem_IRawAllocator *&pUsedAllocator,
        bool                    &IsOk);
    
};
/*! endclass: Log_ActionImageReader */


//===================================================================================
/*! class:       Log_ActionImageWriter

    description: writing the image of an action
*/
class Log_ActionImageWriter
{

private:
    Data_SplitSpaceWriter         &m_SplitSpaceWriter;
    Data_SplitSpaceWriter::Result  m_WriteResult;   
    
    //--------------------------------------------------------------Log_ActionImageWriter---
    /*! function:     BoolResult

        return value: true, if no errors detected
    */
    bool IsResultOk () const
    {
        return (
            (Data_SplitSpaceWriter::ok                 == m_WriteResult)
            ||
            (Data_SplitSpaceWriter::moreSpaceAvailable == m_WriteResult) );
    }
    
public:

    //--------------------------------------------------------------Log_ActionImageWriter---
    /*! function:    Log_ActionImageWriter ctor
    
        arguments:   SplitSpaceWriter [in out]
    */
    inline Log_ActionImageWriter (Data_SplitSpaceWriter &SplitSpaceWriter);
    

    //--------------------------------------------------------------Log_ActionImageWriter---
    /*! function:    AlignImage
    
        description: appends padding bytes, if the length of the written image is less than the aligned one
      
        arguments:   WrittenLength [in]
                     IsOk          [out]
    */
    void AlignImage (
        SAPDB_UInt4  WrittenLength,
        bool        &IsOk);

    //--------------------------------------------------------------Log_ActionImageWriter---
    /*! function:     IsWrittenCompletely
    
        return value: true, if the image is written completely
    */
    bool IsWrittenCompletely () const
    {
        // return (Data_SplitSpaceWriter::ok == m_WriteResult);
        return ( IsResultOk () );
    }

    //--------------------------------------------------------------Log_ActionImageWriter---
    /*! function:    PersistentWrite
    
        description: copies the source to the undo/redo file

                     If SourceLen is less than the aligned SourceLen, padding bytes are appended
      
        arguments:   pSource   [in]
                     SourceLen [in]
                     IsOk      [out]
    */
    inline void PersistentWrite (
        const void  *pSource,
        SAPDB_UInt4  SourceLen,
        bool        &IsOk);

    //--------------------------------------------------------------Log_ActionImageWriter---
    /*! function:    UnalignedPersistentWrite
    
        description: copies the source to the undo/redo file without any alignment
        
        arguments:   pSource   [in]
                     SourceLen [in]
                     IsOk      [out]
    */
    void UnalignedPersistentWrite (
        const void  *pSource,
        SAPDB_UInt4  SourceLen,
        bool        &IsOk);

    //--------------------------------------------------------------Log_ActionImageWriter---
    /*  function:     WriteResultValue
    */
    SAPDB_Int WriteResultValue () const
    {
        return (m_WriteResult);
    }
};
/*! endclass: Log_ActionImageWriter */


//===================================================================================
/*! class:       Log_ReadWriteActionImage

    description: writing of before and after images of any action
*/
class Log_ReadWriteActionImage: public Log_IAction
{
private:

    Log_ActionType m_ActionType;

protected:

    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    Log_ReadWriteActionImage ctor
    
        arguments:   ActionType [in]
    */
    inline Log_ReadWriteActionImage (Log_ActionType ActionType);

    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    ReadImagePersistent

        description: reads the image of the action

        arguments:   ImageReader [in out]
                     IsOk        [out]
    */
    virtual void ReadImagePersistent (
        Log_ActionImageReader &ImageReader,
        bool                  &IsOk)       = 0;
    
    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    WriteImagePersistent

        description: writes the image of the action

        arguments:   ImageWriter [in out]
                     IsOk        [out]
    */
    virtual void WriteImagePersistent (
        Log_ActionImageWriter &ImageWriter,
        bool                  &IsOk)       const = 0;
    

public:

    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:     GetMinimalPersistentLength
    
        return value: the minimum length of the image which must be contiguous in memory
    */
    virtual SAPDB_UInt GetMinimalPersistentLength () const
    {
        return (0);
    }
    
    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    ReadPersistentFormat

        description: reads the image of the action from the undo or redo file

        arguments:   Reader    [in out]
                     Allocator [in out] is used, if a contents is not contigous in a page
                     IsOk      [out]
    */
    virtual void ReadPersistentFormat (
        Data_SplitSpaceReader  &Reader,
        SAPDBMem_IRawAllocator &Allocator,
        bool                   &IsOk);
    
    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    WriteAfterImage
    
        description: triggers the writing of the after image

        arguments:   TransContext [in out]
    */
    void WriteAfterImage (tgg00_TransContext &TransContext) const;

    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    WriteBeforeImage
    
        description: triggers the writing of the before image

        arguments:   TransContext [in out]
    */
    void WriteBeforeImage (tgg00_TransContext &TransContext) const;

    //--------------------------------------------------------------Log_ReadWriteActionImage---
    /*! function:    WritePersistentFormat
    
        description: writes the image of the action into the undo or redo file
      
        arguments:   Writer [in out]
                     IsOk   [out]
    */
    virtual void WritePersistentFormat (
        Data_SplitSpaceWriter &Writer,
        bool                  &IsOk) const;
 
};
/*! endclass: Log_ReadWriteActionImage */


/*===========================================================================*
*  INLINE METHODS of Log_ActionImageReader                                   *
*============================================================================*/

inline Log_ActionImageReader::Log_ActionImageReader (Data_SplitSpaceReader  &SplitSpaceReader,
                                                     SAPDBMem_IRawAllocator &IRawAllocator):
m_SplitSpaceReader (SplitSpaceReader),
m_IRawAllocator    (IRawAllocator)
{ }


/*===========================================================================*
*  INLINE METHODS of Log_ActionImageWriter                                   *
*============================================================================*/

inline Log_ActionImageWriter::Log_ActionImageWriter (Data_SplitSpaceWriter &SplitSpaceWriter):

m_SplitSpaceWriter (SplitSpaceWriter),
m_WriteResult      (Data_SplitSpaceWriter::moreSpaceAvailable)
{ }

//---------------------------------------------------------------------------

inline void Log_ActionImageWriter::PersistentWrite (const void  *pSource,
                                                    SAPDB_UInt4  SourceLen,
                                                    bool        &IsOk)
{
    UnalignedPersistentWrite (pSource, SourceLen, IsOk);

    if ( ! IsOk ) return;
    
    AlignImage (SourceLen, IsOk);
}

/*===========================================================================*
*  INLINE METHODS of Log_ReadWriteActionImage                                *
*============================================================================*/

inline Log_ReadWriteActionImage::Log_ReadWriteActionImage (Log_ActionType ActionType):
    m_ActionType (ActionType)
{ }

#endif  /* LOG_READ_WRITE_ACTION_IMAGE_H */
