/*!**************************************************************************

  module      : Log_RedoFile.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : defines the classes Log_RedoFile



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef Log_Redo_H
#define Log_Redo_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "DataAccess/Data_ChainSplitSpaceForwardRead.hpp"
#include "Logging/Log_AfterImageSpace.hpp"
#include "Logging/Log_Types.hpp"
#include "File/File_IFile.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_RedoPage
description: 
*/

class Log_RedoPage : public Data_PageSplitSpaceForwardRead
{

public:

    
    /*! ---------------------------------------------------------------------------
    function:    InitializeFrame
    */
    virtual void InitializeFrame ()
    {
        Data_PageSplitSpace::InitializeFrame();
        NodePtr()->nd_trans_id().gg90SetNil();
        SetFirstEntrySequence (0);
        SetLastEntrySequence(Log_EntrySequence());
    }

    /*! --------------------------------------------------------------------------------
    function:     SetTransNo
    arguments:    TransNo [in]
    descriptions: The TransNo is written to the page header.
    */
    void SetTransNo (const tgg91_TransNo& TransNo)
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        NodePtr()->nd_trans_id() = TransNo;
    }
    /*! --------------------------------------------------------------------------------
    function:     GetTransNo
    returns:      TransNo
    descriptions: The TransNo is read from the page header.
    */
    tgg91_TransNo GetTransNo () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return NodePtr()->nd_trans_id();
    }
    /*! --------------------------------------------------------------------------------
    function:     SetFirstEntrySequence
    descriptions: The entry sequence of the first entry in the page is set in
                  the page header.
    arguments:    entrySequence [in]
    */
    void SetFirstEntrySequence (Log_EntrySequence entrySequence)
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        NodePtr()->nd_str_version() = entrySequence;
    }
    /*! --------------------------------------------------------------------------------
    function:     GetFirstEntrySequence
    descriptions: The entry sequence of the first entry is returned.
    returns:      entrySequence
    */
    Log_EntrySequence GetFirstEntrySequence () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return NodePtr()->nd_str_version();
    }
    /*! --------------------------------------------------------------------------------
    function:     SetLastEntrySequence
    descriptions: The entry sequence of the last entry in the page is set in
                  the page header.
    arguments:    entrySequence [in]
    */
    void SetLastEntrySequence (Log_EntrySequence entrySequence)
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        NodePtr()->ndInvUsageCnt_bd00() = entrySequence;
    }
    /*! --------------------------------------------------------------------------------
    function:     GetLastEntrySequence
    descriptions: The entry sequence of the last entry is returned.
    returns:      entrySequence
    */
    Log_EntrySequence GetLastEntrySequence () const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return NodePtr()->ndInvUsageCnt_bd00();
    }
};
/*! endclass: Log_RedoPage */

/*---------------------------------------------------------------------------*/
/*! class:       Log_RedoFile
    description: This is the Interface to Redo log manager.
*/
class Log_RedoFile : public File_IFile
{

public:
    
    /*! --------------------------------------------------------------------------------
    declaration: Container, Iterator, BeforeImageId
    description: These types are used with this class.
    */
    typedef Data_ChainSplitSpaceForwardRead<Log_RedoPage>  Container;
    typedef Container::Iterator                            Iterator;
    typedef Data_ChainIterator<Log_RedoPage>               PageIterator;
    typedef tgg91_TransNo                                  TransNo;
    /*! enddeclaration: */

public:

    /*! --------------------------------------------------------------------------------
    function:    Log_RedoFile
    arguments:   Trans      [in/out]
                 RootPageNo [in]
    description: This constructs an Redo log handle for the given transaction.
                 If the RootPageNo is valid, an existing file is assumed.
    */
    Log_RedoFile (tgg00_TransContext &Trans,
                  Data_PageNo         Root);
              
    /*! --------------------------------------------------------------------------------
    function:    Log_RedoFile
    arguments:   Trans      [in/out]
                 RootPageNo [in]
                 LastPageNo [in]
    description: This constructs an Redo log handle for the given transaction.
                 If the RootPageNo is valid, an existing file is assumed.
    */
    Log_RedoFile (tgg00_TransContext &Trans,
                  Data_PageNo         Root,
                  Data_PageNo         Last);
              
    /*! --------------------------------------------------------------------------------
    function:    ~Log_RedoFile
    description: The file is dropped.
    */
    ~Log_RedoFile ();

    /*! ---------------------------------------------------------------------------
    function:    Create
    arguments:   TransContext    [in]
    returns:     true, if successfull
    description: A new RedoLogfile is created, the transNo is written into the root.
    */
    bool Create (const tgg00_TransContext& Trans);

    /*! ---------------------------------------------------------------------------
    function:    Drop
    description: A RedoLogfile is dropped.
    */
    virtual void Drop ();

    /*! ---------------------------------------------------------------------------
    function:    Verify
    returns:     true, if the file is consistent
    */
    virtual bool Verify(tgg00_TransContext &trans,
                        bool                isCold);
                        
    /*! ---------------------------------------------------------------------------
    function:    WriteToTrace
    description: This traces some basic information about the file to the trace.
    */
    virtual void WriteToTrace() const;

    /*! ---------------------------------------------------------------------------
    function:    IsCreated
    returns:     true, if a page exists for this file.
    */
    bool IsCreated () const
    {
        return m_Container.RootId().PageNo().IsValid();
    }

    /*! --------------------------------------------------------------------------------
    function:    GetEntryInfo
    arguments:   IOSequence [out]
                 Reader     [in]
                 isOK       [out]
    description: the io sequence is extracted.
    */
    static void GetEntryInfo (Data_SplitSpaceReader &Reader,
                              Log_IOSequenceNo      &IOSequence,
                              Log_EntrySequence     &EntrySequence,
                              bool                  &isOK);
                                    
    /*! --------------------------------------------------------------------------------
    function:    ReserveSpace
    description: Space is reserved. the first part has minimum size of MinimalLength.

                 The internal iterator is set to the space.
                 You have to use ReleaseSpace() after writing everything into the reserved space.
    arguments:   Length                [in]
                 MinimalLength         [in]
                 EntrySequence         [in]
                 IOSequence            [in]
    returns:     Data_SplitSpaceWriter [out];
    */
    Data_SplitSpaceWriter ReserveSpace (Data_SplitRecordLength Length,
                                        Data_RecordLength      MinimalLength,
                                        Log_EntrySequence      EntrySequence,
                                        Log_IOSequenceNo       IOSequence);

    /*! --------------------------------------------------------------------------------
    function:    ReleaseSpace
    arguments:   isOK
    description: The internal iterator is deassigned,
                 so all resources are released.
                 If isOK is false, all pareserved pages are released unchanged.
    */
    void ReleaseSpace(bool isOK);

    /*! --------------------------------------------------------------------------------
    function:    GetFirstRedoEntry
    description: Sets the internal iterator to the first record in the file.
    returns:     Iterator&
    */
    Iterator& GetFirstRedoEntry ()
    {
        SAPDBERR_ASSERT_STATE( IsCreated() );

        m_Container.Begin (m_Iterator, Data_ForUpdate); // not Data_ForRead !!

        return m_Iterator;
    }

    /*! --------------------------------------------------------------------------------
    function:    GetLastEntrySequence
    description: determinies the last entry sequence written to this file.
    returns:     entrysequence
    */
    Log_EntrySequence GetLastEntrySequence ();

    /*! --------------------------------------------------------------------------------
    function:    LastPageNo
    description: determines the last page no.
    returns:     Data_PageNo
    */
    Data_PageNo LastPageNo()
    {
        SAPDBERR_ASSERT_STATE( IsCreated() );

        return m_Container.LastPageNo();
    }

    /*! --------------------------------------------------------------------------------
    function:    Truncate
    description: this truncates the file, so the last entry has the given sequence no.
    arguments:   stopSequence [in]
    */
    void Truncate (Log_EntrySequence stopSequence);

private:

    /*! --------------------------------------------------------------------------------
    declaration: EntryHead
    description: this holds additional information for entry handling
    */
    typedef struct EntryHead
    {
        Log_EntrySequence Sequence;
        Log_IOSequenceNo  IOSequence;
    } EntryHead;
    
    /*! --------------------------------------------------------------------------------
    variable:    m_Container
    description: The container for this file.
    */
    Container m_Container;

    /*! --------------------------------------------------------------------------------
    variable:    m_Iterator
    description: The iterator for the container.
    */
    Iterator m_Iterator;

};
/*! endclass: Log_RedoFile */


/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/

/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/
                                         
#endif  /* Log_Redo_H */
