
/*!**************************************************************************

    module      : Log_RedoManager.cpp
    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : parallel undo/redo of transactions in order to restart the database

    last changed: 2001-03-22

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
*  INCLUDES                                                                  *
*============================================================================*/

#include "RunTime/RTE_Message.hpp"
#include "Logging/Log_Exceptions.hpp"
#include "Logging/Log_RedoManager.hpp"
#include "Logging/Log_RedoReadTask.hpp"
#include "Logging/Log_RedoTaskLifeCycle.hpp"
#include "Logging/Log_RedoTrafficControl.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "KernelCommon/Kernel_Trace.hpp"

#include "ggg00.h"

/*===========================================================================*
*  CLASS Log_RedoManager                                                     *
*============================================================================*/

// --------------------------------------------------------------------------

void Log_RedoManager::RedoLog (
                    tgg00_TransContext &TransContext,
			  const Kernel_Date         untilDate,
              const Kernel_Time         untilTime,
              const Log_IOSequenceNo    untilIOSeq,
                    bool                restoreLog)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoManager::RedoLog", LogTrans_Trace, 5);
    
    KERNEL_TRACE_BASIS_ERROR (TransContext.trError_gg00, LogTrans_Trace, 5);
    
    SAPDBMem_IRawAllocator &RawAllocator =
        *(REINTERPRET_CAST (SAPDBMem_IRawAllocator*, TransContext.trAllocator_gg00));

    Log_RedoAllocateReadTask       RedoReadTaskSingleton;
    Log_RedoAllocateTrafficControl RedoTrafficControlSingleton;    
    
    bool IsOk = true;
    
    RedoReadTaskSingleton.Allocate (RawAllocator,
                                    untilDate,
                                    untilTime,
                                    untilIOSeq,
                                    restoreLog,
                                    m_redoIsAborted,
                                    IsOk);
    
    if ( ! IsOk)
    {
        TransContext.trError_gg00 = e_no_more_memory;
        return;
    }
    
    RedoTrafficControlSingleton.Allocate ( RawAllocator,
                                           m_redoIsAborted,
                                           m_transread,
                                           m_transredone,
                                           IsOk);
    
    if ( ! IsOk)
    {
        TransContext.trError_gg00 = e_no_more_memory;
        return;
    }
    
    {   
        // inside this block log reader tasks are running
        Log_RedoReadTaskLifeCycle ReadTaskLifeCycle (TransContext);
        
        {
            // inside this block redo execute tasks are running
            Log_RedoExecuteTaskLifeCycle ExecuteTaskLifeCycle (TransContext);
        }
    }

    if ( TransContext.trRteCommPtr_gg00->to_cancel ) // UH 2002-05-21 added
        m_redoIsAborted = true;                      // UH 2002-05-21 added

    if ( e_ok == TransContext.trError_gg00
         &&
         m_redoIsAborted )
    {
         TransContext.trError_gg00 = m_abortReason;
    }
    
    if (e_ok == TransContext.trError_gg00) // mb 2002-06-06 moved from kb81restart_handling
    {
        Log_SavepointManager.StartSavepointAndWait (TransContext, Log_SVPReasonRestart);
    }
    else if (restoreLog && e_cancelled == TransContext.trError_gg00)
    {
        Log_SavepointManager.StartSavepointAndWait (TransContext, Log_SVPReasonDistance);
        TransContext.trError_gg00 = e_cancelled;
    }
}

// --------------------------------------------------------------------------

void Log_RedoManager::AbortRedo(tgg00_BasisError abortReason)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoManager::AbortRedo", LogTrans_Trace, 5);
    m_redoIsAborted = true;
    m_abortReason = abortReason;
}

// --------------------------------------------------------------------------

void Log_RedoManager::GetProgressInfo( SAPDB_UInt &transread,
                                       SAPDB_UInt &transredone )
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoManager::GetProgressInfo", LogTrans_Trace, 5);
    transread   = m_transread;
    transredone = m_transredone;
    RTE_Message( Log_Exception(__CONTEXT__, LOG_REDO_PROGRESS,
                               SAPDB_ToString(transread),
                               SAPDB_ToString(transredone)) );
}

/*===========================================================================*
*  END OF CODE                                                               *
*============================================================================*/
