/*!**************************************************************************

    module      : Log_RedoReadTask.cpp

    -------------------------------------------------------------------------

    responsible:      UweH

    special area:     Logging
    description :     defines a singleton to handle a task which reads the log
                      and creates redo files

    last changed:     2001-03-22

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
*  INCLUDES                                                                  *
*============================================================================*/

#include "ggg00.h"
#include "ggg92.h"   // used by hkb51.h
#include "heo52.h"   // vclock
#include "hgg01.h"   // g01restart_time
#include "hkb51.h"   // kb51GetRedoTransIndex
#include "hkb50.h"   // k50GetLogTransaction, k50RegisterTransaction
#include "hkb57.h"   // kb57RedoTransNoIncrement
#include "hkb57_1.h" // k57restartrec

#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "Logging/Log_IRedoReadTask.hpp"
#include "Logging/Log_RedoReadTask.hpp"
#include "Logging/Log_IRedoTrafficControl.hpp"
#include "Logging/Log_Transaction.hpp"
#include "Logging/Log_ArchiveLogIterator.hpp"
#include "Logging/Log_SeqTapeReader.hpp"
#include "Logging/Log_ArchiveLogManager.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "Logging/Log_ActionPartialRollback.hpp"
#include "Logging/Log_ActionObject.hpp"
#include "Logging/Log_ActionSavepoint.hpp"
#include "Logging/Log_OpenTransFile.hpp"
#include "Logging/Log_Savepoint.hpp"
/*===========================================================================*
*  DEFINES                                                                   *
*============================================================================*/
 
// #define TEST_DEBUG 1
#define CHECK_ASCENDING_EOT_SEQUENCE            true   // PTS 1113641 UH 2002-01-25 
#define NUMBER_OF_ENTRIES_FOR_NEXT_WRITE        50000

/*===========================================================================*
*  EXTERNAL VARIABLE                                                         *
*============================================================================*/
Log_IRedoReadTask* Log_IRedoReadTask::m_pInstance = NULL;
/*===========================================================================*
*  CLASSES                                                                   *
*============================================================================*/

/*!-------------------------------------------------------------------------
class:        Log_AfterImageHeadSpace
description:  This is used to implement the interface Data_ISplitSpace for
              reading purpose.
--------------------------------------------------------------------------*/
class Log_AfterImageHeadSpace : public Data_ISplitSpace
{

public:

    /*!-------------------------------------------------------------------------
    function:     CurrentPart
    --------------------------------------------------------------------------*/
    virtual SAPDB_Int CurrentPart () const
    {
        return 0;
    }

    /*!-------------------------------------------------------------------------
    function:     GetPart
    arguments:    Part   [in]
                  Length [out]
                  Value  [out]
    --------------------------------------------------------------------------*/
    virtual void GetPart (SAPDB_Int    Part,
                          SAPDB_UInt  &Length,
                          SAPDB_Byte* &Value) const
    {
        Length = m_Length;
        Value  = m_Value;
    }

    /*!-------------------------------------------------------------------------
    function:     Length
    returns:     SAPDB_UInt
    --------------------------------------------------------------------------*/
    virtual SAPDB_UInt Length() const
    {
        return m_TotalLength;
    }

    /*!-------------------------------------------------------------------------
    function:     IsAssigned
    returns:     SAPDB_UInt
    --------------------------------------------------------------------------*/
    virtual bool IsAssigned () const
    {
        return m_Value != NULL;
    }

    /*!-------------------------------------------------------------------------
    function:     Log_AfterImageHeadSpace
    description:
    --------------------------------------------------------------------------*/
    Log_AfterImageHeadSpace ()
    : m_Value       (NULL),
      m_Length      (0),
      m_TotalLength (0)
    {}

    /*!-------------------------------------------------------------------------
    function:     Assign
    arguments:    iter [in]
    description:  assigns a new space
    --------------------------------------------------------------------------*/
    void Assign (Log_ArchiveLogIterator& iter)
    {
        m_Value       = *iter;
        m_Length      = iter.Length();
        m_TotalLength = iter.RestLength();
    }

private:

    /*!-------------------------------------------------------------------------
    variable:     m_Value
    --------------------------------------------------------------------------*/
    SAPDB_Byte *m_Value;
    /*!-------------------------------------------------------------------------
    variable:     m_Length
    --------------------------------------------------------------------------*/
    SAPDB_UInt m_Length;
    /*!-------------------------------------------------------------------------
    variable:     m_TotalLength
    --------------------------------------------------------------------------*/
    SAPDB_UInt m_TotalLength;
};
/*!-------------------------------------------------------------------------
endclass:     Log_AfterImageHeadSpace
--------------------------------------------------------------------------*/

//=======================================================================================
/*! class:       Log_RedoReadTask
    description: a singleton to handle a task which reads the log and creates the redo files
                 This current implementation handles one task only.
*/
class Log_RedoReadTask : public Log_IRedoReadTask
{

public:

    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:    AllocateInstance
        description: allocates the Log_RedoReadTask singleton
        arguments:   RawAllocator [in out]
                     restoreLog   [in]
                     IsOk         [out]
    */
    static void AllocateInstance (      SAPDBMem_IRawAllocator &RawAllocator,
                                  const Kernel_Date             untilDate, 		// PTS 1111525 mb 2001-01-15
                                  const Kernel_Time             untilTime, 		// PTS 1111525 mb 2001-01-15
				                  const Log_IOSequenceNo        untilIOSeq,     // PTS 1121539 mb 2003-05-30
                                        bool                    restoreLog,
                                        volatile bool          &redoIsAborted,
                                        bool                   &IsOk)
    {
        SAPDBTRACE_ROUTINE_DEBUG ("Log_RedoReadTask::AllocateInstance", LogTrans_Trace, 5);

        if (NULL == m_pInstance)
            m_pInstance = new (RawAllocator)
                          Log_RedoReadTask ( RawAllocator,
                                             untilDate, untilTime, untilIOSeq,
                                             restoreLog,
                                             redoIsAborted );

        IsOk = (NULL != m_pInstance);
    }

    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:     GetRawAllocatorUsed

        return value: the allocator used to allocate this singleton
    */
    SAPDBMem_IRawAllocator &GetRawAllocatorUsed ()
    {
        return (m_RawAllocator);
    }

    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:     InstancePtr

        return value: the pointer to the global instance of the singleton
    */
    static Log_RedoReadTask *InstancePtr ()
    {
        return ( REINTERPRET_CAST (Log_RedoReadTask*, m_pInstance) );
    }

    //----------------------------------------------------------------------------------
    /*! function:     GetReadPosition
        description:  The actual read position is determined. e.g. for the savepoint.
        argument:     iosequence  [out]
                      pageoffset  [out]
                      entryoffset [out]
                      entrytype   [out]
                      entrydate   [out]
                      entrytime   [out]
    */
    virtual void GetReadPosition (Log_IOSequenceNo    &iosequence,
                                  Log_RawDeviceOffset &pageoffset,
                                  SAPDB_UInt2         &entryoffset,
                                  Log_ActionType      &entrytype,
                                  Kernel_Date         &entrydate,
                                  Kernel_Time         &entrytime) const
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::GetReadPosition", LogTrans_Trace, 5);
        EntryInfo entryinfo;
        GetCurrentEntry (entryinfo);
        iosequence  = entryinfo.iosequence;
        pageoffset  = entryinfo.pageoffset;
        entryoffset = entryinfo.entryoffset;
        entrytype   = entryinfo.entrytype;
        entrydate   = entryinfo.entrydate;
        entrytime   = entryinfo.entrytime;
        if ( LogVolume_Trace.TracesLevel(6) )
        {
            entryinfo.WriteToTrace("GetReadPosition");
            entryinfo.Message ( "GetReadPosition", 0, 0, 0 );
        }
    }
    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:     ReadLogAndCreateRedoFiles
        description:  Reads log entries from the archive log or the log save and creates redo files.

                      = REDO READER TASK CODE =
                      This is the main function of the redo-reader-task.
                      It is directly called via the pascal-wrapper kb921RedoLogReadTask
                      from the servertask management.
        argument:     trans [in out]
    */
    virtual void ReadLogAndCreateRedoFiles (tgg00_TransContext &readerTrans)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::ReadLogAndCreateRedoFiles", LogTrans_Trace, 5);

        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "trError: " << readerTrans.trError_gg00);

        if (e_ok != readerTrans.trError_gg00) return;

        SAPDBMem_IRawAllocator  &allocator      = *(reinterpret_cast <SAPDBMem_IRawAllocator*>
                                                    (readerTrans.trAllocator_gg00));
        Log_IRedoTrafficControl &coordinator    = Log_IRedoTrafficControl::Instance();
        Log_ArchiveLogManager   &logMgr         = Log_ArchiveLogManager::Instance();
        Log_ArchiveLog          &log            = **(logMgr.ArchiveLogsBegin());
        Log_IOSequenceNo         lastProcessedIOSequence;
        Log_IOSequenceNo         lastIOSeqFromTape;
        bool                     incrementIOSeq = true;

        ReCreateOpenTransactions (readerTrans, allocator);
        if ( readerTrans.trError_gg00 != e_ok )
        {
            SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "trError: " << readerTrans.trError_gg00);
            RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                       SAPDBERR_ASSERT_STATE_FAILED,
                       "Log_RedoReadTask::ReadLogAndCreateRedoFiles() ReCreateOpenTransactions failed") );
        }
        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "open trans are re-created");

        if (m_UntilDate.IsValid() && m_UntilTime.IsValid() && !m_RestoreLog)
        // PTS 1119223 mb 2002-12-10 only for RestartUntil the scan of the Log is done
        {
            log.ScanLogForFirstKnown(readerTrans.trTaskId_gg00);
        }

        log.GetFirstKnownBeforeRestoreLog(m_FirstKnownOffset, lastIOSeqFromTape);

        if (m_RestoreLog)
        {
            if (lastIOSeqFromTape.IsValid())
            // there is something on the log_device
            {
                if (lastIOSeqFromTape > 0)
                {
                    --lastIOSeqFromTape;
                }
            }


            Log_ArchiveLogIterator tapeIter = logMgr.GetTapeIterator(allocator,
                                                                     readerTrans,
                                                                     lastIOSeqFromTape,
                                                                     m_UntilDate,
                                                                     m_UntilTime,
                                                                     m_UntilIOSeq);
            tapeIter.Initialize();
            logMgr.SetToLastSavepointEntry (tapeIter, lastProcessedIOSequence);

            RTE_Message(Log_Exception(__FILE__,__LINE__,
                                      LOG_RECOV_TAPE,
                                      SAPDB_ToString(lastProcessedIOSequence,_T_d),
                                      SAPDB_ToString(lastIOSeqFromTape,_T_d)));
 
            if (lastIOSeqFromTape.IsValid())
            // there is something on the log_device to be redone
            {
                if (lastProcessedIOSequence <= lastIOSeqFromTape)
                // but there has to something read from the backup before 
                {
                    InternReadLogAndCreateRedoFile( readerTrans,
                                                    tapeIter,
                                                    allocator,
                                                    lastProcessedIOSequence,
                                                    true);
                    if (tapeIter.GetState() == Log_ArchiveLogIterator::Iter_untilReached)
                    {
                        m_ContinueWithLogDev = false;
                    }
                }
                else
                {
                    m_RestoreLog = false;
                }
            }
            else
            // nothing has to be read from the log_device, but only from the backup
            {
                m_ContinueWithLogDev=false;
                InternReadLogAndCreateRedoFile( readerTrans,
                                                tapeIter,
                                                allocator,
                                                lastProcessedIOSequence,
                                                true);
            }

            tapeIter.Delete();
        }

        if ( m_redoIsAborted )
        {
            RTE_Message( Log_Exception(__CONTEXT__, LOG_REDO_ABORTED) );
            coordinator.SetEndOfLogRead (readerTrans.trTaskId_gg00);
            return;
        }
            
        Log_ArchiveLogIterator iter = logMgr.GetArchiveLogIterator (allocator,
                                                                    readerTrans.trTaskId_gg00,
                                                                    m_UntilDate,
                                                                    m_UntilTime,
                                                                    m_UntilIOSeq);
        iter.Initialize();
        if (m_RestoreLog && m_ContinueWithLogDev)
        {
            if (m_FirstKnownOffset.IsValid())
            {
                logMgr.SetToBeginOfPage (iter, m_FirstKnownOffset);

                if ( LogTrans_Trace.TracesLevel(7) )
                    iter.WriteToTrace ("ReadIter");

                if ( iter.GetState() == Log_ArchiveLogIterator::Iter_ok )
                // the iterator may be invalidated due to UNTIL-conditions even on the first page
                {
                    ++lastProcessedIOSequence;                       //PTS1115226 mb 2002-04-15
                    if (iter.GetIOSequence() != lastProcessedIOSequence )
                    {
                        Kernel_VTrace() << "lastProcessedIOSequence " << lastProcessedIOSequence
                                        << " != iter.GetIOSequence() " << iter.GetIOSequence()
                                        << NewLine;


                        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                   SAPDBERR_ASSERT_STATE_FAILED,
                                   "Log_RedoReadTask::ReadLogAndCreateRedoFiles() last IOSeq of Tape does not match first IOSeq on LogVol") );
                    }
                    RTE_Message(Log_Exception(__FILE__,__LINE__,
                                              LOG_RECOV_LOG_VOL,
                                              SAPDB_ToString(lastProcessedIOSequence,_T_d)));
                    InternReadLogAndCreateRedoFile( readerTrans,
                                                    iter,
                                                    allocator,
                                                    lastProcessedIOSequence,
                                                    false);
                    }
            }
        }
        else if (!m_RestoreLog && !log.LogIsEmpty()) // PTS 1113550 mb 2002-05-13
        // Redo entries on the log-device only if it hasn't been read already and if there is something on it
        {
            m_ContinueWithLogDev = false;
            logMgr.SetToLastSavepointEntry (iter, lastProcessedIOSequence);

            if ( LogTrans_Trace.TracesLevel(7) )
                iter.WriteToTrace ("ReadIter");

            RTE_Message(Log_Exception(__FILE__,__LINE__,
                                      LOG_RECOV_LOG_VOL,
                                      SAPDB_ToString(lastProcessedIOSequence,_T_d)));
            InternReadLogAndCreateRedoFile( readerTrans,
                                            iter,
                                            allocator,
                                            lastProcessedIOSequence,
                                            true);
        }
        else if (!m_RestoreLog && log.LogIsEmpty())
        {
            // no log has been restored (neither from tape, nor from logVolume)
            // so the log-iosequences has to match the savepoint-iosequence
            // PTS 1116123 mb 2002-06-11
            lastProcessedIOSequence = k57restartrec->rstLastSavept_kb00().svpIOsequence_kb00;
            incrementIOSeq = false;
        }
        else
        {
                SAPDBERR_ASSERT_STATE(m_RestoreLog);
        }

        iter.Delete();

        if ( m_redoIsAborted )
        {
            RTE_Message( Log_Exception(__CONTEXT__, LOG_REDO_ABORTED) );
            coordinator.SetEndOfLogRead (readerTrans.trTaskId_gg00);
            return;
        }
            
        CloseOpenTransactions (readerTrans, lastProcessedIOSequence);
        
        if (m_UntilDate.IsValid() && m_UntilTime.IsValid()
            ||
            m_UntilIOSeq.IsValid())
        {
            if (m_LastReadOffset.IsValid())                    // PTS 1111525 mb 2002-01-24 restore log until
            // some or all pages from the log_device had been read
            {
                SAPDBERR_ASSERT_STATE(m_FirstKnownOffset.IsValid());
                Log_RawDeviceIterator firstToBeCleared = 
                        ++( Log_RawDeviceIterator( log.LogicalLogBeginOffset(),
                                                  m_LastReadOffset,
                                                  log.LogicalLogEndOffset()));
                if (firstToBeCleared.GetPosition() != m_FirstKnownOffset)
                // some pages had NOT been redone => clear them
                {
                    Log_RawDeviceIterator lastToBeCleared = 
                            --(Log_RawDeviceIterator( log.LogicalLogBeginOffset(),
                                                      m_FirstKnownOffset,
                                                      log.LogicalLogEndOffset()));
                    log.ClearLogPartially ( readerTrans.trTaskId_gg00,
                                            firstToBeCleared.GetPosition(),
                                            lastToBeCleared.GetPosition(),
                                            lastProcessedIOSequence+1);
                }
                else
                // in case of a restart until iosequence it may occur, that the last page
                // is not completely redone. In this case adjust log_infopage
                {
                    if (m_UntilIOSeq.IsValid())
                    {
                        if (lastProcessedIOSequence != log.GetLastKnownIOSeq())
                        {
                            log.SetDeviceState(readerTrans.trTaskId_gg00,Log_DeviceStateHistoryLost);
                            log.TruncateLogPage(readerTrans.trTaskId_gg00,
                                                lastProcessedIOSequence,
                                                m_LastReadOffset);
                        }
                    }
                }
            }
            else
            // no pages of the log_device had been read => clear them all
            {
                log.ClearLogComplete (readerTrans.trTaskId_gg00,
                                      lastProcessedIOSequence+1);
            }
        }
        if (incrementIOSeq )                                      // PTS 1116123 mb 2002-06-11
        {
            ++lastProcessedIOSequence;
        }
        log.UpdateIOSequenceNoAfterRedoLog (readerTrans.trTaskId_gg00, lastProcessedIOSequence);
                
        coordinator.SetEndOfLogRead (readerTrans.trTaskId_gg00);

        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "TrError: " << readerTrans.trError_gg00);
    }

private:

    /*! --------------------------------------------------------------------------------
    function:    InternReadLogAndCreateRedoFile(
    arguments:   taskId    [in]
                 allocator [in/out]
    description: all open redo transactions are rollbacked and removed.
    */
    void InternReadLogAndCreateRedoFile(tgg00_TransContext      &readerTrans,
                                        Log_ArchiveLogIterator  &iter,
                                        SAPDBMem_IRawAllocator  &allocator,
                                        Log_IOSequenceNo        &lastIOSequence,
                                        bool                     beginOfRestore)
    {
        Log_AfterImageHeadSpace  entryHeadSpace;
        Log_AfterImage           afterImage;
        bool                     isOK            = true;
        bool                     checkFirstEntry
                                 = beginOfRestore
                                   &&
                                   Log_Savepoint::GetLastSavepointEntryType() != Log_NoOp // for migration
                                   &&
                                   lastIOSequence != 0;         // PTS 1113550 mb 2002-05-06 for recovering from IOSeq 0
        bool                     nextEntryMustBeFinishOpenTrans = false; // PTS 1115065 UH 2002-04-10
        bool                     endOfTransAlreadyProcessed = false; // PTS 1123461 UH 2003-08-28
        tgg00_TransIndex         currTransIndex;
        Log_Transaction         *pTransaction    = NULL;
        SAPDB_UInt               readEntryCount  = 0;
        SAPDB_Int                timeInSec;
        SAPDB_Int                dummyTimeInMSec;
        SAPDB_Int                lastSavepointTriggerTime;
    
        vclock (&lastSavepointTriggerTime, &dummyTimeInMSec);

        while ( iter.GetState() == Log_ArchiveLogIterator::Iter_ok
                &&
                ! m_redoIsAborted )
        {
            entryHeadSpace.Assign           (iter);
            Data_SplitSpaceReader reader    (entryHeadSpace);
            afterImage.ReadPersistentFormat (reader, isOK);

            if ( ! isOK )
            {
                iter.WriteToTrace ("ReadIter");
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Log_RedoReadTask::ReadLogAndCreateRedoFiles() cannot read afterimage") );
            }
            
            if ( checkFirstEntry )
            {
                checkFirstEntry = false;
                if ( afterImage.GetActionType() != Log_Savepoint::GetLastSavepointEntryType() )
                {
                    Kernel_VTrace() << "LastSavepointEntryType: "
                                    << Log_ActionTypeStrings[Log_Savepoint::GetLastSavepointEntryType()];
                    Kernel_VTrace() << "readlog: T" << afterImage.GetTransNo().gg90GetInt4()
                                    <<" (" << afterImage.GetSequence() << ") '"
                                    << Log_ActionTypeStrings[afterImage.GetActionType()] << "' ["
                                    << iter.GetIOSequence() << "]\n";
                    RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__, SAPDBERR_ASSERT_STATE_FAILED,
                               "Log_RedoReadTask::ReadLogAndCreateRedoFiles() redo begin not found (look into vtrace)") );
                }
            }
            
            if ( LogTrans_Trace.TracesLevel(7) )
                Kernel_VTrace() << "readlog: T" << afterImage.GetTransNo().gg90GetInt4()\
                                <<" (" << afterImage.GetSequence() << ") '" \
                                << Log_ActionTypeStrings[afterImage.GetActionType()] << "' ["
                                << iter.GetIOSequence() << "]\n";

            // PTS 1115065 UH 2002-04-10 begin
            if ( nextEntryMustBeFinishOpenTrans )
            {
                if( afterImage.GetActionType() != Log_FinishOpenTrans )
                {
                    // the previous log entry was truncated
                    // this can only happen if the kernel crashed and restarted and now the restart begins
                    // at an earlier time. (e.g. restore data)
                    Kernel_VTrace() << "readlog: T" << afterImage.GetTransNo().gg90GetInt4()
                                    <<" (" << afterImage.GetSequence() << ") '"
                                    << Log_ActionTypeStrings[afterImage.GetActionType()] << "' ["
                                    << iter.GetIOSequence() << "]\n";
                    iter.WriteToTrace ("read iterator");
                    RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,
                                                  "previous logentry was corrupt") );
                }
                nextEntryMustBeFinishOpenTrans = false;
            }
            // PTS 1115065 UH 2002-04-10 end

            lastIOSequence = iter.GetIOSequence();

            if ( ! afterImage.GetTransNo().gg90IsNil() )
                pTransaction = GetOrCreateRedoTransaction ( readerTrans,
                                                            allocator,
                                                            afterImage.GetTransNo(),
                                                            currTransIndex );

            // The read task builds then redo files with a correct transid.
            readerTrans.trWriteTransId_gg00 = afterImage.GetTransNo();
            readerTrans.trIndex_gg00        = currTransIndex;

            // save current read entry position for redo-savepoint
            SetCurrentEntry (iter, afterImage.GetActionType());
            ++readEntryCount;
            
            // write progress messages
            if ( LogTrans_Trace.TracesLevel(9)
                 ||
                 (readEntryCount % NUMBER_OF_ENTRIES_FOR_NEXT_WRITE == 0) )
                m_CurrentEntry.Message ( "redo-read", readEntryCount,
                                         afterImage.GetTransNo().gg90GetInt4(),
                                         afterImage.GetSequence() );

            if ( readEntryCount % NUMBER_OF_ENTRIES_FOR_NEXT_WRITE == 0 )
            {
                vclock (&timeInSec, &dummyTimeInMSec);
                if ( lastSavepointTriggerTime + g01restart_time <= timeInSec )
                {
                    lastSavepointTriggerTime = timeInSec;
                    Log_SavepointManager.StartSavepoint (readerTrans, Log_SVPReasonDistance);
                }
            }
            
            endOfTransAlreadyProcessed = false; // PTS 1123461 UH 2003-08-28
            
            if ( pTransaction != 0
                 &&
                 pTransaction->GetTransType() != Log_Transaction::redoRead )
            {
                // PTS 1123461 UH 2003-08-28 new if-clause
                // The context is already given to the redo tasks.
                // Do not insert it again to the redo traffic control.
                // But check the correct EndOfTrans handling.
                if ( (pTransaction->GetTransState() == Log_Transaction::committed
                      &&
                      afterImage.GetActionType() != Log_Commit)
                     ||
                     (pTransaction->GetTransState() == Log_Transaction::rollbacked
                      &&
                      afterImage.GetActionType() != Log_Rollback)
                     ||
                     (pTransaction->GetTransState() == Log_Transaction::opened) )
                {
                    Kernel_VTrace() << "Found different end of transaction";
                    iter.WriteToTrace ("LogEntryIterator");
                    afterImage.WriteToTrace ("CurrentAction");
                    pTransaction->WriteToTrace ("CurrentTrans");
                    RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,
                                                  "Found different end of transaction") );
                }
                endOfTransAlreadyProcessed = true;
            }

            switch ( afterImage.GetActionType() )
            {
            case Log_Commit:
            case Log_Rollback:
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );

                if ( ! endOfTransAlreadyProcessed ) // PTS 1123461 UH 2003-08-28
                {
                    EndOfTransaction ( readerTrans.trTaskId_gg00,
                                       allocator,
                                       pTransaction,
                                       afterImage.GetActionType() == Log_Commit
                                       ? Log_Transaction::committed
                                       : Log_Transaction::rollbacked,
                                       iter.GetIOSequence(),
                                       currTransIndex,
                                       CHECK_ASCENDING_EOT_SEQUENCE ); // PTS 1113641 UH 2002-01-25
                }
                break;
            case Log_FinishOpenTrans:
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );
                CloseOpenTransactions (readerTrans, lastIOSequence);
                break;
            case Log_PartialRollback:
            {
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );
                Log_ActionPartialRollback Action;
                Action.ReadPersistentFormat (reader, allocator, isOK);
                if ( ! isOK )
                    RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                                  SAPDBERR_ASSERT_STATE_FAILED,
                                                  "Log_RedoReadTask::ReadLogAndCreateRedoFiles: read PartRb failed") );
                if ( Action.GetRedoStopSequence() // PTS 1111 UH 2000-08-10
                     ==
                     afterImage.GetSequence() )
                {
                    // this should be prevented by Log_Transaction,
                    // but for more safety it is checked here again
                    SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "truncate: stopseq: "
                                                           << Action.GetRedoStopSequence()
                                                           << ", nothing to do.");
                    break;
                }
                SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "truncate: inclusive stopseqence: "
                                                       << Action.GetRedoStopSequence());
                // PTS 1113230 UH 2001-12-20 begin
                pTransaction->RedoPartialRollback ( Action.GetUndoStopSequence(),
                                                    Action.GetRedoStopSequence() );
                // PTS 1113230 UH 2001-12-20 end
                break;
            }
            case Log_SavepointEntry:
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );
                 // nothing to do
                break;
            default:
                nextEntryMustBeFinishOpenTrans = // PTS 1115065 UH 2002-04-10
                    ! CopyEntry (*pTransaction, iter, afterImage, readerTrans);
                if ( nextEntryMustBeFinishOpenTrans )
                {
                    if ( LogTrans_Trace.TracesLevel(6) )
                    {
                        Kernel_VTrace() << "nextEntryMustBeFinishOpenTrans:\n";
                        Kernel_VTrace() << "readlog: T" << afterImage.GetTransNo().gg90GetInt4()
                                        <<" (" << afterImage.GetSequence() << ") '"
                                        << Log_ActionTypeStrings[afterImage.GetActionType()] << "' ["
                                        << iter.GetIOSequence() << "]\n";
                        iter.WriteToTrace ("read iterator");
                    }
                }
                break;
            }//switch

            if ( iter.GetState() != Log_ArchiveLogIterator::Iter_ok )
                break;
            else
            {
                m_LastReadOffset = iter.GetPosition(); // PTS 1111525 mb 2002-01-23
				// PTS 1115065 UH 2002-04-10 begin
				// if current entry was truncated and the current queue io sequence is 0
				// then the log writing was beginning here with a restart and the first
				// log entry must be the FinishOpenTrans entry.
				// So the iterator must not be incremented because it is pointing to the beginning.
				if ( ! nextEntryMustBeFinishOpenTrans
					 ||
					 iter.GetQueueIOSequenceNo() > 0 )
					++iter;
				// PTS 1115065 UH 2002-04-10 end
            }
        } //endwhile
        m_CurrentEntry.Message ( "last-redo-read", readEntryCount,
                                 afterImage.GetTransNo().gg90GetInt4(),
                                 afterImage.GetSequence() );
    }


    /*! --------------------------------------------------------------------------------
    function:    CloseOpenTransactions
    arguments:   taskId    [in]
                 allocator [in/out]
    description: all open redo transactions are rollbacked and removed.
    */
    void CloseOpenTransactions (tgg00_TransContext &redoReadTrans,
                                Log_IOSequenceNo    lastIOSequence)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::CloseOpenTransactions", LogTrans_Trace, 5);

        tgg00_TransIndex        openTransIndex       = cgg_nil_transindex;
        Log_Transaction        *pTransaction         = NULL;
        SAPDBMem_IRawAllocator &allocator            =
            *(reinterpret_cast <SAPDBMem_IRawAllocator*>(redoReadTrans.trAllocator_gg00));

		if ( lastIOSequence.IsInvalid() )
        {
			// PTS 1124684 UH 2003-10-16
			lastIOSequence = Log_IRedoTrafficControl::Instance().
				GetLastIOSequence(redoReadTrans.trTaskId_gg00);
            if ( lastIOSequence.IsInvalid() )
                lastIOSequence = 0;
        }
	    
		kb51GetFirstOpenRedoTrans (redoReadTrans.trTaskId_gg00, openTransIndex);

		while ( openTransIndex != cgg_nil_transindex )
        {
            pTransaction = REINTERPRET_CAST(Log_Transaction*,k50GetLogTransaction (openTransIndex));
            if ( NULL == pTransaction )
                RTE_Crash( SAPDBErr_Exception(
                           __FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "CloseOpenTransactions: k50GetLogTransaction() returns NULL") );

            if ( pTransaction->GetRedoEotSequence().IsValid() )
                RTE_Crash( SAPDBErr_Exception(
                           __FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "CloseOpenTransactions: eot sequence is valid") );

            EndOfTransaction ( redoReadTrans.trTaskId_gg00,
                               allocator,
                               pTransaction,
                               Log_Transaction::rollbacked,
                               lastIOSequence,
                               openTransIndex,
                               ! CHECK_ASCENDING_EOT_SEQUENCE ); // PTS 1113641 UH 2002-01-25

            // force search at the beginning of the transactions list
            // until there exists no more open transaction
            openTransIndex = cgg_nil_transindex;
            kb51GetFirstOpenRedoTrans (redoReadTrans.trTaskId_gg00, openTransIndex);
        }//while
    }
    /*! --------------------------------------------------------------------------------
    function:    ReCreateOpenTransactions
    arguments:   redoReaderTrans [in]
                 allocator       [in] - used to allocate transactions
    description: All open transactions of last savepoint are newly registered.

                 This reads the open trans file identified by the root page no.
                 The needed transentries in the locklist are created.
                 The information about what savepoint must be used comes from the
                 restartrecord.
                 The transaction contexts are allocated with the given allocator.
                 This allocator must not be used by the redo tasks itself.
                 The should register their own allocator to work with their own context.
    ----------------------------------------------------------------------------------*/
    void ReCreateOpenTransactions (tgg00_TransContext     &redoReaderTrans,
                                   SAPDBMem_IRawAllocator &allocator)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::ReCreateOpenTransactions", LogTrans_Trace, 5);

        tkb00_SaveptParam &saveptParam = k57restartrec->rstLastSavept_kb00();
        Log_OpenTransFile file ( redoReaderTrans, saveptParam.svpOpenTransRoot_kb00 );

        if ( ! file.IsCreated() )
            return;

        Log_OpenTransFile::Iterator &iter = file.GetFirstOpenTrans(saveptParam.svpId_kb00);

        tgg00_TransIndex currentTransIndex = cgg_nil_transindex;
        tgg00_BasisError trError           = e_ok;

        while ( iter.IsValid() )
        {
            Log_Transaction* newtrans = new (allocator) Log_Transaction ();

            if ( newtrans == NULL )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_RedoReadTask::ReCreateOpenTransactions: allocate failed") );

            newtrans->ReadPersistentFormat( *iter, file.GetSpaceSize() );

            if ( ! newtrans->IsRedoRelevant() )
            {
                // The transaction entry can be deleted with all its empty root pages.
                if ( ! newtrans->InitializeForRedo (redoReaderTrans) )
                    RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                                  SAPDBERR_ASSERT_STATE_FAILED,
                                                  "Log_RedoReadTask::ReCreateOpenTransactions: init for delete failed") );
                newtrans->DeleteAfterRedo();
                destroy (newtrans, allocator);
                ++iter;
                continue;
            }
            
            if ( ! newtrans->InitializeForRedoCopy (redoReaderTrans,
                                                    newtrans->GetRedoTransNo(),
                                                    allocator) )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_RedoReadTask::ReCreateOpenTransactions: init failed") );

            newtrans->UpdateUndoRedoInfoForRestart();
            
            kb51GetRedoTransIndex (redoReaderTrans.trTaskId_gg00,
                                   newtrans->GetRedoTransNo(),
                                   currentTransIndex,
                                   trError);

            if ( trError != e_ok )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_RedoReadTask::ReCreateOpenTransactions: kb51GetRedoTransIndex failed") );
            k50RegisterTransaction (currentTransIndex, newtrans);

            if ( LogTrans_Trace.TracesLevel(6) )
                newtrans->WriteToTrace("OpenTrans");

            if ( newtrans->GetRedoEotSequence().IsValid()
                 && // PTS 1114946 UH 2002-03-20
                 newtrans->GetTransState() != Log_Transaction::opened )
            {
                // The transaction was collected during savepoint before the commit/rollback was written.
                // Between setting TransState and writing eot logentry may the trans been collected by savepoint.
                // Then the trans has a TransState which says that the transaction has ended, but the eot
                // logentry must be found after the savepoint entry.
                // This assumes a fixed sequence of operations in the Log_Savepoint::Coordinator().
                // First Savepoint Entry writing, then transactions collecting.

//                if ( newtrans->GetRedoEotSequence().IsInvalid() )
//                    newtrans->SetEOTInRedo ( newtrans->GetTransState(),
//                                             saveptParam.svpIOsequence_kb00 );

                EndOfTransaction ( redoReaderTrans.trTaskId_gg00,
                                   allocator,
                                   newtrans,
                                   newtrans->GetTransState(),
                                   newtrans->GetRedoEotSequence(),
                                   currentTransIndex,
                                   ! CHECK_ASCENDING_EOT_SEQUENCE ); // PTS 1113641 UH 2002-01-25
            }

            ++iter;
        }
        // PTS 1124684 UH 2003-10-16 added the following two lines
        if ( LogTrans_Trace.TracesLevel(6) )
            Log_IRedoTrafficControl::Instance().
                WriteToTrace("End of Rst_RedoReadTask::ReCreateOpenTransactions()");
        Log_IRedoTrafficControl::Instance().EnableRedoProcessing(redoReaderTrans.trTaskId_gg00);
    }
    /*! --------------------------------------------------------------------------------
    function:    GetOrCreateRedoTransaction()
    arguments:   readerTrans [in]
                 allocator   [in]
                 transno     [in]
    returns:     Log_Transaction*
    description: This gives the trans context to the given transno.

                 If the transaction does not exist, it is created and registered.
    ----------------------------------------------------------------------------------*/
    Log_Transaction* GetOrCreateRedoTransaction ( tgg00_TransContext     &readerTrans,
                                                  SAPDBMem_IRawAllocator &allocator,
                                                  tgg91_TransNo           transno,
                                                  tgg00_TransIndex       &currTransIndex)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::GetOrCreateRedoTransaction", LogTrans_Trace, 5);

        tgg00_BasisError &trError        = readerTrans.trError_gg00;
        tsp00_TaskId     &readerTaskId   = readerTrans.trTaskId_gg00;
        Log_Transaction  *pTransaction   = NULL;

        currTransIndex = cgg_nil_transindex;

        kb51GetRedoTransIndex ( readerTaskId, transno, currTransIndex, trError );

        if ( trError != e_ok || currTransIndex == cgg_nil_transindex )
        {
            SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "trError: " << trError);
            RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                       SAPDBERR_ASSERT_STATE_FAILED,
                       "Log_RedoReadTask::GetOrCreateRedoTransaction() kb51GetRedoTransIndex failed") );
        }

        pTransaction = reinterpret_cast<Log_Transaction*>
                       (k50GetLogTransaction (currTransIndex));

        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "TransNo: " << transno.gg90GetInt4() << \
                                          " => TransIndex: " << currTransIndex << \
                                          (NULL==pTransaction?" NEW":""));

        if ( NULL == pTransaction )
        {
            pTransaction = new (allocator) Log_Transaction ();
            if ( pTransaction == NULL )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_RedoReadTask::GetOrCreateRedoTransaction: allocate failed") );

            k50RegisterTransaction (currTransIndex, pTransaction);

            bool isOK = pTransaction->InitializeForRedoCopy
                                      ( readerTrans, transno, allocator );
            if ( ! isOK )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_RedoReadTask::GetOrCreateRedoTransaction: init failed") );

            kb57RedoTransNoIncrement (readerTaskId, transno);
        }

        return pTransaction;
    }
    /*! --------------------------------------------------------------------------------
    function:    CopyEntry
    arguments:   transaction [in]
                 iter        [in]
                 afterimage  [in]
    returns:     bool
    description: copy the actual redo entry referenced by iter into the redofile
                 of transaction.
                 If the complete entry could be copied true is returned else false.
    ----------------------------------------------------------------------------------*/
    bool CopyEntry (Log_Transaction        &transaction, // PTS 1115065 UH 2002-04-10 return bool
                    Log_ArchiveLogIterator &iter,
                    Log_AfterImage         &afterimage,
                    tgg00_TransContext     &readerTrans)
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::CopyEntry", LogTrans_Trace, 5);

        // PTS 1127732 UH 2004-02-11 added SkipCurrentEntry 
        const bool SkipCurrentEntry = transaction.GetLastRedoEntrySequence().IsValid()
                                      &&
                                      afterimage.GetSequence() <= transaction.GetLastRedoEntrySequence();
        // If SkipCurrentEntry is true then nothing is to do - this is a redo of a redo.
        // This entry was already copied into redo file.
        // PTS 1115065 UH 2002-04-10

        SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "CopyEntry: length:" << iter.RestLength() <<
                                               ", minimal: " << GetMinimalLengthToCopyEntry (afterimage.GetActionType()));

        Log_AfterImageHeadSpace dummySplitSpace;
        Data_SplitSpaceWriter   Writer = SkipCurrentEntry          // PTS 1127732 UH 2004-02-11
            ? Data_SplitSpaceWriter(dummySplitSpace)   // PTS 1127732 UH 2004-02-11 use dummy writer
            : transaction.ReserveSpaceForRedoEntryCopy ( iter.RestLength(),
                                                         GetMinimalLengthToCopyEntry (afterimage.GetActionType()),
                                                         afterimage.GetSequence(),
                                                         iter.GetIOSequence() );

        Data_SplitSpaceWriter::Result writeResult          = Data_SplitSpaceWriter::ok;
        bool                          loopagain            = false;
        bool                          incrementAtLoopBegin = true;
        bool                          entryInterrupted     = false; // PTS 1125914 UH 2003-12-03

        do
        {
            if ( loopagain
                 &&
                 iter.GetState() == Log_ArchiveLogIterator::Iter_ok
                 &&
                 incrementAtLoopBegin )
            {
                m_LastReadOffset = iter.GetPosition(); // PTS 1111525 mb 2002-01-23
                ++iter;
                // PTS 1125914 UH 2003-12-03 added if clause
                if ( iter.GetState() == Log_ArchiveLogIterator::Iter_ok
                     &&
                     iter.GetQueueIOSequenceNo() == 0 )
                {
                    // The current entry was interrupted.
                    entryInterrupted = true;
                }
                if ( iter.GetState() == Log_ArchiveLogIterator::Iter_allRead
                     &&
                     m_RestoreLog
                     &&
                     m_ContinueWithLogDev )
                {
                    // PTS 1125914 UH 2004-03-09
                    Log_ArchiveLogManager   &logMgr         = Log_ArchiveLogManager::Instance();
                    Log_ArchiveLog          &log            = **(logMgr.ArchiveLogsBegin());
                    SAPDBMem_IRawAllocator  &allocator      = *(reinterpret_cast <SAPDBMem_IRawAllocator*>
                                            (readerTrans.trAllocator_gg00));

                    m_ContinueWithLogDev = false;
                    incrementAtLoopBegin = false;

                    iter.SetPageReader(log.GetArchiveLogReader(allocator),allocator);
                    iter.SetLastKnownIOSeq(log.GetLastKnownIOSeq());
                    logMgr.SetToBeginOfPage (iter, m_FirstKnownOffset);
                }
            }
            
            incrementAtLoopBegin = true; // PTS 1127732 UH 2004-02-11
            
            if ( iter.GetState()==Log_ArchiveLogIterator::Iter_ok
                 &&
                 ! entryInterrupted )
            {
                if ( ! SkipCurrentEntry ) // PTS 1127732 UH 2004-02-11
                {
                    SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "CopyEntry: write:" << iter.Length());
    
                    Writer.Write (*iter, iter.Length(), writeResult);
                    
                    if ( writeResult != Data_SplitSpaceWriter::moreSpaceAvailable
                         &&
                         writeResult != Data_SplitSpaceWriter::ok )
                    {
                        Kernel_VTrace() << "writeResult: " << writeResult;
                        iter.WriteToTrace("iterator");
                        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                                      SAPDBERR_ASSERT_STATE_FAILED,
                                                      "Log_RedoReadTask::CopyEntry: copy failed") );
                    }
                }
            }
            else
            {
                loopagain = false;
                if ( LogTrans_Trace.TracesLevel(6) )
                    Kernel_VTrace() << "INFO: *** LOG IS TRUNCATED ***";
                writeResult = Data_SplitSpaceWriter::ok;
                if ( ! SkipCurrentEntry ) // PTS 1127732 UH 2004-02-11
                    transaction.ReleaseSpaceAfterRedoEntryCopy(false); // release resources unchanged
                break;
            }
                
            if ( SkipCurrentEntry ) // PTS 1127732 UH 2004-02-11
                loopagain = iter.RestLength() > iter.Length();
            else
                loopagain = writeResult == Data_SplitSpaceWriter::moreSpaceAvailable
                            &&
                            iter.RestLength() > iter.Length();
        }
        while ( loopagain );
        
        if ( Data_SplitSpaceWriter::ok == writeResult )
        {
            if ( ! SkipCurrentEntry ) // PTS 1127732 UH 2004-02-11
                transaction.ReleaseSpaceAfterRedoEntryCopy(true);
            return true;
        }
        else
        {
            // PTS 1115065 UH 2002-04-10 no abort and no change
            if ( LogTrans_Trace.TracesLevel(6) )
            {
                Kernel_VTrace() << "INFO: *** LOGENTRY TRUNCATED ***, writeResult: " << writeResult;
                iter.WriteToTrace("iterator");
            }
            if ( ! entryInterrupted
                 &&
                 ! SkipCurrentEntry ) // PTS 1127732 UH 2004-02-11
                transaction.ReleaseSpaceAfterRedoEntryCopy(false);
            return false;
        }
    }
    
    /*! --------------------------------------------------------------------------------
    function:    EndOfTransaction
    arguments:   redoReaderTaskid [in]
                 allocator        [in]
                 pTransaction     [in/out]
                 actiontype       [in]
                 eotSequence      [in]
                 transindex       [in]
    description:
    ----------------------------------------------------------------------------------*/
    void EndOfTransaction ( tsp00_TaskId                 redoReaderTaskid,
                            SAPDBMem_IRawAllocator      &allocator,
                            Log_Transaction*            &pTransaction,
                            Log_Transaction::TransState  state,
                            Log_IOSequenceNo             eotSequence,
                            tgg00_TransIndex             transindex,
                            bool                         checkAscendingSequence ) // PTS 1113641 UH 2002-01-25
    {
        SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTask::EndOfTransaction", LogTrans_Trace, 5);

        pTransaction->SetEOTInRedo ( state, eotSequence );

        if ( LogTrans_Trace.TracesLevel(7) )
            pTransaction->WriteToTrace ("EndOfRedoTrans");

        // PTS 1114946 UH 2002-03-21 begin
        bool             transIsAlreadyInRedoList = false;
        Log_Transaction* pPublicTrans = reinterpret_cast<Log_Transaction*>(k50GetLogTransaction(transindex));

        if ( NULL == pPublicTrans )
            RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED, "no public Log_Transaction") );

        if ( pPublicTrans != pTransaction )
        {
            // there is another log transcontext registered in the public transaction entry in locklist
            // Check if this correct.
            if ( pTransaction->GetTransState()      != pPublicTrans->GetTransState()
                 ||
                 pTransaction->GetRedoEotSequence() != pPublicTrans->GetRedoEotSequence() )
            {
                pTransaction->WriteToTrace ("pTransaction");
                pPublicTrans->WriteToTrace ("pPublicTrans");
                RTE_Crash( SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,
                           "transaction end found twice") );
            }
            transIsAlreadyInRedoList = true;
        }
        // PTS 1114946 UH 2002-03-21 end
        
        bool           isOK    = true;
        Log_ActionType eottype = state ==  Log_Transaction::committed
                                           ? Log_Commit
                                           : Log_Rollback; // <- default

        if ( (eottype == Log_Commit
              &&
              ! pTransaction->EntriesToRedoExist())
             ||
             (eottype == Log_Rollback
              &&
              ! pTransaction->WasOpenInLastSavepoint()) )
        {
            tgg00_TransState dummyState;
            dummyState.clear();
            pTransaction->Delete();
            k51remove_locks ( redoReaderTaskid, transindex, dummyState );
            k51del_transentry_from_locklist ( redoReaderTaskid, transindex );
        }
        else
        {
            pTransaction->DeleteAfterRedoCopy();

            if ( ! transIsAlreadyInRedoList ) // PTS 1114946 UH 2002-03-21
                Log_IRedoTrafficControl::Instance().
                    InsertJob ( redoReaderTaskid, transindex,
                                eotSequence, eottype,
                                checkAscendingSequence,// PTS 1113641 UH 2002-01-25
                                *pTransaction,
                                isOK );
            if ( ! isOK )
			{
	            Log_IRedoTrafficControl::Instance().WriteToTrace("Log_RedoReadTask::EndOfTransaction: InsertJob failed");
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_RedoReadTask::EndOfTransaction: InsertJob failed") );
			}

		}

        allocator.Deallocate(pTransaction);
        pTransaction = NULL;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetMinimalLengthToCopyEntry
    returns:     length (SAPDB_UInt)
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt GetMinimalLengthToCopyEntry (Log_ActionType type)
    {
        SAPDB_UInt length = Log_AfterImage::GetPersistentLength();
        switch (type)
        {
            case Log_NewObject:
            case Log_InsertObject:
            case Log_UpdateObject:
            case Log_DeleteObject:
                length += Log_ActionObject::GetMinimalLengthToCopyEntry();
                break;
            case Log_PartialRollback:
                length += Log_ActionPartialRollback::GetMinimalLengthToCopyEntry();
                break;
            case Log_SavepointEntry:
            case Log_FinishOpenTrans:
                length += Log_ActionSavepoint::GetMinimalLengthToCopyEntry();
                break;
            default:
                break;
        }
        return length;
    }

    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:    Log_RedoReadTask ctor
        description: private ctor only used once during initialization
        argument:    RawAllocator [in]
    */
    Log_RedoReadTask (SAPDBMem_IRawAllocator &RawAllocator,
                      bool                    restoreLog,
                      volatile bool          &redoIsAborted)
    :
    m_RawAllocator (RawAllocator),
    m_RestoreLog(restoreLog),
    m_ContinueWithLogDev(true),
    m_redoIsAborted(redoIsAborted)
    {}

private:

    //--------------------------------------------------------------Log_RedoReadTask---
    SAPDBMem_IRawAllocator &m_RawAllocator;
    /*--------------------------------------------------------------------------
    declaration: m_UntilDate
    description: the date until a restore/restart will be done
                 PTS 1111525 mb 2001-01-18
                 
    --------------------------------------------------------------------------*/
    const Kernel_Date       m_UntilDate;

    /*--------------------------------------------------------------------------
    declaration: m_UntilTime
    description: the time until a restore/restart will be done
                 PTS 1111525 mb 2001-01-18
    --------------------------------------------------------------------------*/
    const Kernel_Time       m_UntilTime;

    /*--------------------------------------------------------------------------
    declaration: m_UntilIOSeq
    description: the IOSequencenumber until the redo will process
                 PTS 1121539 mb 2003-05-30
    --------------------------------------------------------------------------*/
    const Log_IOSequenceNo  m_UntilIOSeq;

    /*--------------------------------------------------------------------------
    declaration: m_RestoreLog
    description: if true the log has to be restored from tape
    --------------------------------------------------------------------------*/
    bool                    m_RestoreLog;

    /*--------------------------------------------------------------------------
    declaration: m_ContinueWithLogDev
    description: if true the log-reading has to be continued on the log-device
    --------------------------------------------------------------------------*/
    bool                    m_ContinueWithLogDev;

    /*--------------------------------------------------------------------------
    declaration: m_redoIsAborted
    --------------------------------------------------------------------------*/
    volatile bool          &m_redoIsAborted;
    /*--------------------------------------------------------------------------
    declaration: EntryInfo
    description: this type combines all log entry characteristica
    --------------------------------------------------------------------------*/
    struct EntryInfo
    {
        Log_IOSequenceNo    iosequence;
        Log_RawDeviceOffset pageoffset;
        SAPDB_UInt2         entryoffset;
        Log_ActionType      entrytype;
        SAPDB_UInt1         filler;
        Kernel_Date         entrydate;
        Kernel_Time         entrytime;
        EntryInfo()
        {
            iosequence.Invalidate();
            pageoffset.Invalidate();
            entryoffset = SAPDB_MAX_UINT2;
            entrytype   = Log_NoOp;
            filler      = 0;
            entrydate.Invalidate();
            entrytime.Invalidate();
        }
        void WriteToTrace(const char *title = 0)
        {
            Kernel_VTrace trace;
            if ( title != NULL )
                trace << title << ": ";
            trace << pageoffset << "." << entryoffset
                  << " [" << iosequence << "] '"
                  << Log_ActionTypeStrings[entrytype]
                  << "' " << entrydate << "-" << entrytime;
        }
        void Message (const char *title,
                      SAPDB_UInt  msgcount,
                      SAPDB_UInt  transno,
                      SAPDB_UInt  entrysequence)
        {
            RTE_Message ( Log_Exception ( __FILE__, __LINE__, LOG_ENTRY,
                          title,
                          SAPDB_ToString(msgcount,_T_d),
                          SAPDB_ToString(transno,_T_d),
                          SAPDB_ToString(entrysequence,_T_d),
                          SAPDB_ToString(iosequence,_T_d),
                          SAPDB_ToString(pageoffset,_T_d),
                          SAPDB_ToString(entryoffset,_T_d),
                          Log_ActionTypeStrings[entrytype],
                          SAPDB_ToString(entrydate,_T_d),
                          SAPDB_ToString(entrytime,_T_d) ) );
        }
    };
    /*--------------------------------------------------------------------------
    declaration: m_CurrentEntry
    description: This values is set after it is read and before it is handled.
                 Handled means copied, etc.
    --------------------------------------------------------------------------*/
    EntryInfo            m_CurrentEntry;
    /*! ------------------------------------------------------------------------
    declaration: m_CurrentEntryAccess
    description: This synchronises the access to m_CurrentEntry.
    --------------------------------------------------------------------------*/
    mutable RTESync_Spinlock m_CurrentEntryAccess;

    /*--------------------------------------------------------------------------
    declaration: m_LastReadOffset
    description: the offset of the last page on the logdev which had been read
                 PTS 1111525 mb 2002-01-23
    --------------------------------------------------------------------------*/
    Log_RawDeviceOffset     m_LastReadOffset;

    /*--------------------------------------------------------------------------
    declaration: m_FirstKnownOffset
    description: the offset of the first page on the logdev which has to be read
                 PTS 1111525 mb 2002-01-23
    --------------------------------------------------------------------------*/
    Log_RawDeviceOffset     m_FirstKnownOffset;

private:

    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:    SetCurrentEntry
        description: sets the current entry paramters
        argument:    taskid     [in]
                     iter       [in]
                     actiontype [in]
    */
    void SetCurrentEntry ( const Log_ArchiveLogIterator &iter,
                           Log_ActionType                actiontype )
	{
        m_CurrentEntryAccess.Lock();
        m_CurrentEntry.iosequence = iter.GetIOSequence();
        iter.GetPosition (m_CurrentEntry.pageoffset, m_CurrentEntry.entryoffset);
        m_CurrentEntry.entrytype  = actiontype;
        iter.GetTimeStamp (m_CurrentEntry.entrydate, m_CurrentEntry.entrytime);
        m_CurrentEntryAccess.Unlock();
	}
    //--------------------------------------------------------------Log_RedoReadTask---
    /*! function:    GetCurrentEntry
        description: gets the current entry paramters
        argument:    taskid     [in]
                     EntryInfo  [out]
    */
    void GetCurrentEntry ( EntryInfo &entryinfo ) const
	{
        m_CurrentEntryAccess.Lock();
        entryinfo = m_CurrentEntry;
        m_CurrentEntryAccess.Unlock();
	}

    inline Log_RedoReadTask (
              SAPDBMem_IRawAllocator &RawAllocator,
        const Kernel_Date             untilDate,        // PTS 1111525 mb 2001-01-15
        const Kernel_Time             untilTime,        // PTS 1111525 mb 2001-01-15
        const Log_IOSequenceNo        untilIOSeq,       // PTS 1121539 mb 2003-05-30
              bool                    restoreLog,
              volatile bool          &redoIsAborted)
    :
    m_RawAllocator (RawAllocator),
    m_UntilDate(untilDate),
    m_UntilTime(untilTime),
    m_UntilIOSeq(untilIOSeq),
    m_RestoreLog(restoreLog),
    m_ContinueWithLogDev(true),
    m_redoIsAborted(redoIsAborted)
    {}

};
/*! endclass: Log_RedoReadTask */

/*===========================================================================*
*  METHODS of Log_RedoAllocateReadTask                                       *
*============================================================================*/

Log_RedoAllocateReadTask::~Log_RedoAllocateReadTask ()
{
    Log_RedoReadTask *pRedoReadTask = Log_RedoReadTask::InstancePtr();

    if (NULL == pRedoReadTask) return;


    SAPDBMem_IRawAllocator &RawAllocator = pRedoReadTask->GetRawAllocatorUsed();

	destroy (pRedoReadTask, RawAllocator);
}

//----------------------------------------------------------------------------
void Log_RedoAllocateReadTask::Allocate (      SAPDBMem_IRawAllocator &RawAllocator,
                                         const Kernel_Date             untilDate, 
                                         const Kernel_Time             untilTime,
                                         const Log_IOSequenceNo        untilIOSeq,       // PTS 1121539 mb 2003-05-30
                                               bool                    restoreLog,
                                               volatile bool          &redoIsAborted,
                                               bool                   &IsOk)
{
    Log_RedoReadTask::AllocateInstance (RawAllocator,
                                        untilDate,
                                        untilTime,
                                        untilIOSeq,
                                        restoreLog,
                                        redoIsAborted,
                                        IsOk);
}

/*===========================================================================*
*  END OF CODE                                                               *
*============================================================================*/
