
/*!**************************************************************************

    module      : Log_RedoTaskLifeCycle.cpp
    -------------------------------------------------------------------------

    author      : JuergenA
    responsible : UweH

    special area: Logging
    description : defines classes to start and stop redo tasks

    last changed: 2001-03-21

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
*  INCLUDES                                                                  *
*============================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp" // Routine trace macros
#include "Logging/Log_Types.hpp"
#include "KernelCommon/Kernel_Trace.hpp"
#include "Logging/Log_RedoTaskLifeCycle.hpp"

#include "ggg00.h"
#include "hgg01.h"
#include "hkb53.h"
#include "hkb90.h"

/*===========================================================================*
*  LOCAL FUNCTIONS                                                           *
*============================================================================*/

static void StartTasks (tgg00_TransContext   &TransContext,
                        tgg00_MessType_Param  MessType,
                        SAPDB_Int             MaxTasks,
                        SAPDB_Int            &RunningTasks)
{
    SAPDBTRACE_ROUTINE_DEBUG ("StartTasks", LogTrans_Trace, 5);
    KERNEL_TRACE_BASIS_ERROR (TransContext.trError_gg00, LogTrans_Trace, 5);
    
    tgg00_MessBlock  SendMblock;
    tgg00_TransChild ChildTrans;
    tgg00_BasisError AuxError = TransContext.trError_gg00;
    
    TransContext.trError_gg00 = e_ok;
    RunningTasks              = 0;
    do
    {
        g01mblock_init (TransContext, MessType, mm_nil, SendMblock);
        
        k53child_trans_build (m_restart, RunningTasks+1, ChildTrans);
        
        k90send (SendMblock, ChildTrans);
        
        if (e_ok != TransContext.trError_gg00) return;
        
        ++RunningTasks;
    }
    while (RunningTasks < MaxTasks);
    
    if (e_ok != AuxError) TransContext.trError_gg00 = AuxError;
}

static tgg00_BasisError CombineReturnCodes(const tgg00_BasisError firstErr,
                                           const tgg00_BasisError secondErr)
{
    if (e_ok == firstErr)
    {
        return secondErr;
    }
    else
    {
        if (e_cancelled == firstErr)
        {
            if (   (secondErr != e_ok) 
                && (secondErr != e_cancelled))
            {
                return secondErr;
            }
            return e_cancelled;
        }
        else
        {
            return firstErr;
        }
    }
}
//---------------------------------------------------------------------------

static void StopTasks (tgg00_TransContext &TransContext,
                       SAPDB_Int          &RunningTasks)
{
    SAPDBTRACE_ROUTINE_DEBUG ("StopTasks", LogTrans_Trace, 5);
    KERNEL_TRACE_BASIS_ERROR (TransContext.trError_gg00, LogTrans_Trace, 5);
    
    tgg00_MessBlock  ReceiveMblock;
    tgg00_TransChild ChildTrans;
    tgg00_BasisError AnyChildError = e_ok;
    tgg00_BasisError AuxError      = TransContext.trError_gg00;
    
    TransContext.trError_gg00 = e_ok;
    
    while (RunningTasks > 0)
    {
        g01mblock_init (TransContext, m_nil, mm_nil, ReceiveMblock);
        
        k90rcv_child (ReceiveMblock, ChildTrans);
        
        if (e_ok != TransContext.trError_gg00) return;
        
        AnyChildError = CombineReturnCodes(AnyChildError,
                                           ChildTrans.tcdError_gg00);
        --RunningTasks;
    }
    
    TransContext.trError_gg00 = CombineReturnCodes(AuxError, AnyChildError);
}


/*===========================================================================*
*  CLASS Log_RedoExecuteTaskLifeCycle                                              *
*============================================================================*/

Log_RedoExecuteTaskLifeCycle::Log_RedoExecuteTaskLifeCycle (tgg00_TransContext &TransContext):
m_TransContext (TransContext),
m_RunningTasks (0)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoExecuteTaskLifeCycle::Log_RedoExecuteTaskLifeCycle", LogTrans_Trace, 5);

    const SAPDB_Int ReservedTasks   = 3;
    SAPDB_Int       MaxExecuteTasks = g01maxservertask() - g01no_of_backup_devs() - ReservedTasks;
    
    if (MaxExecuteTasks < 0) MaxExecuteTasks = 1;
    
    StartTasks (TransContext, m_restart, MaxExecuteTasks, this->m_RunningTasks);
}

//---------------------------------------------------------------------------

Log_RedoExecuteTaskLifeCycle::~Log_RedoExecuteTaskLifeCycle ()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoExecuteTaskLifeCycle::~Log_RedoExecuteTaskLifeCycle", LogTrans_Trace, 5);

    StopTasks (this->m_TransContext, this->m_RunningTasks);
}


/*===========================================================================*
*  CLASS Log_RedoReadTaskLifeCycle                                             *
*============================================================================*/

Log_RedoReadTaskLifeCycle::Log_RedoReadTaskLifeCycle (tgg00_TransContext &TransContext):
m_TransContext (TransContext),
m_RunningTasks (0)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTaskLifeCycle::Log_RedoReadTaskLifeCycle", LogTrans_Trace, 5);

    const SAPDB_Int MaxReadTasks = 1;
    
    StartTasks (TransContext, m_restore, MaxReadTasks, this->m_RunningTasks);
}

//---------------------------------------------------------------------------

Log_RedoReadTaskLifeCycle::~Log_RedoReadTaskLifeCycle ()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoReadTaskLifeCycle::~Log_RedoReadTaskLifeCycle", LogTrans_Trace, 5);

    StopTasks (this->m_TransContext, this->m_RunningTasks);
}


/*===========================================================================*
*  END OF CODE                                                               *
*============================================================================*/
