
/*!**************************************************************************

    module      : Log_RedoTrafficControl.cpp

    -------------------------------------------------------------------------

    author      : JuergenA,UweH
    responsible : UweH

    special area: Logging
    description : defines the classes to synchronize the parallel redo

    last changed: 2001-03-22

    -------------------------------------------------------------------------

    copyright:    Copyright by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
*  INCLUDES                                                                  *
*============================================================================*/

#include "Logging/Log_IRedoTrafficControl.hpp"
#include "Logging/Log_RedoTrafficControl.hpp"

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "Logging/Log_Exceptions.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_Exceptions.hpp"
#include "Logging/Log_Transaction.hpp"
#include "gsp00.h"
#include "ggg00.h"
#include "ggg92.h"   // used by hkb51.h
#include "heo55k.h"  // vbegexcl
#include "heo56.h"  // vsuspend
#include "hgg08.h"
#include "hkb50.h"   // k50RegisterTransaction
#include "hkb51.h"
#include "hkb53.h"   // k53server_drop_aux_files
#include "hkb811.h"  // kb811AbortRedo

/*===========================================================================*
*  DEFINES                                                                   *
*============================================================================*/

#define LOG_WAIT_FOR_VACANT_ACTION   231
#define LOG_WAIT_FOR_PRECEDING_TRANS 232


/*===========================================================================*
*  EXTERNAL VARIABLE                                                         *
*============================================================================*/

Log_IRedoTrafficControl* Log_IRedoTrafficControl::m_pInstance = NULL;


/*===========================================================================*
*  LOCAL CLASSES                                                             *
*============================================================================*/

class Log_IoSeqEntry;  // forward declaration


//===================================================================================
/*! class:       Log_DoubleChainEntry

    description: a double chained entry
*/
class Log_DoubleChainEntry
{
private:
    
    Log_DoubleChainEntry *m_pNext;
    Log_DoubleChainEntry *m_pPrev;
    
public:

    //--------------------------------------------------------------Log_DoubleChainEntry---
    /*! function:    Log_DoubleChainEntry ctor
    */
    inline Log_DoubleChainEntry ();

    //--------------------------------------------------------------Log_DoubleChainEntry---
    /*! function:    Insert
    
        description: appends an entry at the end of the entry list
      
        arguments:   pFirstListEntry [in out]
                     pLastListEntry  [in out]
    */
    void Insert (Log_DoubleChainEntry* &pFirstListEntry,
			 Log_DoubleChainEntry* &pLastListEntry,
             Log_DoubleChainEntry*  pPrevListEntry);
    
    //--------------------------------------------------------------Log_DoubleChainEntry---
    /*! function:     GetNext
    
        return value: the following entry
    */
    Log_DoubleChainEntry *GetNext () const
    {
        return (m_pNext);
    }
    
    Log_DoubleChainEntry *GetPrev () const
    {
        return (m_pPrev);
    }
    
    //--------------------------------------------------------------Log_DoubleChainEntry---
    /*! function:    Remove
    
        description: removes this entry from the entry list
      
        arguments:   pFirstListEntry [in out]
                     pLastListEntry  [in out]
    */
    void Remove (
        Log_DoubleChainEntry *&pFirstListEntry,
        Log_DoubleChainEntry *&pLastListEntry);

};
/*! endclass: Log_DoubleChainEntry */


//=======================================================================================
/*! class:       Log_DoubleChainList

    description: a double chained list without any allocation
*/
class Log_DoubleChainList
{
private:
    
    Log_DoubleChainEntry *m_pFirst;
    Log_DoubleChainEntry *m_pLast;
    
public:
    
    //--------------------------------------------------------------Log_DoubleChainList---
    /*! function:    Log_DoubleChainList ctor
    */
    inline Log_DoubleChainList ();

	//--------------------------------------------------------------Log_DoubleChainList---
    void Insert (Log_DoubleChainEntry*  pEntry,
                 Log_DoubleChainEntry*  pPrevEntry)
    {
        pEntry->Insert (m_pFirst, m_pLast, pPrevEntry);
    }
	//--------------------------------------------------------------Log_DoubleChainList---
    /*! function:    Append
    
        description: Inserts an entry at the end of the list
        
        arguments:   pEntry [in] the pointer to the entry to be appended
    */
    void Append (Log_DoubleChainEntry *pEntry)
    {
        pEntry->Insert (m_pFirst, m_pLast, m_pLast);
    }
    
    //--------------------------------------------------------------Log_DoubleChainList---
    /*! function:     GetAndRemoveAny
    
        description:  removes an entry from the list

        return value: the pointer to the removed entry
    */
    Log_DoubleChainEntry *GetAndRemoveAny ();

    //--------------------------------------------------------------Log_DoubleChainList---
    /*! function:     GetFirst
    
        return value: the pointer to the first entry of the list
    */
    inline Log_DoubleChainEntry *GetFirst () const;
    
    //--------------------------------------------------------------Log_DoubleChainList---
    /*! function:     GetLast
    
        return value: the pointer to the last entry of the list
    */
    inline Log_DoubleChainEntry *GetLast () const;    
    
    
    //--------------------------------------------------------------Log_DoubleChainList---
    /*! function:     IsEmpty
    
        return value: true, if the list is empty
    */
    bool IsEmpty () const
    {
        return (NULL == m_pFirst);
    }
  
    //--------------------------------------------------------------Log_DoubleChainList---
    /*! function:    Remove
    
        description: removes the entry from the list
        
        arguments:   pEntry [in] the pointer to the entry to be removed
    */
    void Remove (Log_DoubleChainEntry *pEntry)
    {
        if ( pEntry != NULL )
            pEntry->Remove (m_pFirst, m_pLast);
    }

};
/*! endclass: Log_DoubleChainList */


//===================================================================================
/*! class:       Log_JobEntry

    description: an entry of the job container list
*/

class Log_JobEntry : public Log_DoubleChainEntry
{
private:
    tgg00_TransIndex  m_OwnIndex;
    Log_ActionType    m_UndoOrRedo; 
    Log_IoSeqEntry   *m_pIoSeqEntry;
    Log_Transaction   m_Transaction;

public:
    //--------------------------------------------------------------Log_JobEntry---
    /*! function:    Log_JobEntry ctor
    */
    inline Log_JobEntry ();

    //--------------------------------------------------------------Log_JobEntry---
    /*! function:    GetIoSeqEntry
    
        return value: pointer to the corresponding IoSeqEntry
    */
    Log_IoSeqEntry *GetIoSeqEntry () const
    {
        return (m_pIoSeqEntry);
    }
    
    //--------------------------------------------------------------Log_JobEntry---
    /*! function:     GetNext
    
        return value: the following entry
    */
    Log_JobEntry *GetNext () const
    {
        return ( REINTERPRET_CAST (Log_JobEntry*, Log_DoubleChainEntry::GetNext()) );
    }

    //--------------------------------------------------------------Log_JobEntry---
    /*! function:    GetTransaction
    
        return value: pointer of the transaction handle of this job
    */
    Log_Transaction &GetTransaction ()
    {
        return m_Transaction;
    }

    //--------------------------------------------------------------Log_JobEntry---
    /*! function:    GetTransIndex
    
        return value: TransIndex of this job
    */
    tgg00_TransIndex GetTransIndex () const
    {
        return (m_OwnIndex);
    }

    //--------------------------------------------------------------Log_JobEntry---
    /*! function:     GetUndoOrRedo
    
        return value: the action type commit or rollback
    */
    Log_ActionType GetUndoOrRedo () const
    {
        return (m_UndoOrRedo);
    }
    
    //--------------------------------------------------------------Log_JobEntry---
    /*! function:    Init
    
        description: initializes the job entry
      
        arguments:   UndoOrRedo   [in]
                     pIoSeqEntry  [in]
                     transaction  [in]
    */
    inline void Init (
        Log_ActionType   UndoOrRedo,
        Log_IoSeqEntry  *pIoSeqEntry,
        Log_Transaction &transaction);

    //--------------------------------------------------------------Log_JobEntry---
    /*! function:    SetOwnIndex
    
        description: assigns the OwnIndex
    */
    void SetOwnIndex (tgg00_TransIndex OwnIndex)
    {
        m_OwnIndex = OwnIndex;
    }

};
/*! endclass: Log_JobEntry */


//=======================================================================================
/*! class:       Log_IoSeqEntry

    description: entry of a task wait stack
*/
class Log_IoSeqEntry: public Log_DoubleChainEntry
{
    
private:

    Log_IoSeqUInt       m_IoSeq;
    Log_DoubleChainList m_JobList;
    Kernel_WaitStack    m_WaitStack;
    
public:

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    Log_IoSeqEntry ctor
    */
    inline Log_IoSeqEntry ();
    
    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:     GetFirstJobEntry
    
        return value: the first job entry
    */
    Log_JobEntry *GetFirstJobEntry () const
    {
		return static_cast<Log_JobEntry*>(m_JobList.GetFirst());
    }
    
    //--------------------------------------------------------------Log_IoSeqEntry---
    Log_IoSeqEntry *GetLast() const
    {
		return static_cast<Log_IoSeqEntry*>(m_JobList.GetLast());
    }

	//--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:     GetPrev
    
                      return value: the following entry
    */
    Log_IoSeqEntry *GetPrev () const
    {
        return static_cast<Log_IoSeqEntry*>(Log_DoubleChainEntry::GetPrev());
    }
    
    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:     GetNext
    
                      return value: the following entry
    */
    Log_IoSeqEntry *GetNext () const
    {
        return static_cast<Log_IoSeqEntry*>(Log_DoubleChainEntry::GetNext());
    }

	//--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    GetIoSeq
    
        return value: the I/O sequence contained in the entry
    */
    Log_IoSeqUInt GetIoSeq () const
    {
        return (m_IoSeq);
    }

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    Init
    
        description: initializes the I/O sequence entry
      
        arguments:   IoSeq [in]
    */
    inline void Init (Log_IoSeqUInt IoSeq)
    {
        m_IoSeq = IoSeq;
    }

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    InsertJobEntry
    
        description: inserts a job into the job list
        
        arguments:   pUndoRedoJob [in]
    */
    void InsertJobEntry (Log_JobEntry *pJobEntry)
    {
        m_JobList.Append (pJobEntry);
    }

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    InsertWaitTask
    
        description: inserts a task id into the WaitStack for this IoSeq 
        
        arguments:   TaskId [in]
    */
    void InsertWaitTask (tsp00_TaskId TaskId)
    {
        m_WaitStack.PushTaskId (TaskId);
    }

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:     IsEmptyJobList
    
        return value: true, if the job list is empty
    */
    bool IsEmptyJobList () const
    {
        return ( m_JobList.IsEmpty() );
    }

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    RemoveJobEntry
    
        description: removes a job entry from the job list for this IoSeq 
        
        arguments:   pJobEntry [in]
    */
    void RemoveJobEntry (Log_JobEntry *pJobEntry)
    {
        m_JobList.Remove (pJobEntry);
    }

    //--------------------------------------------------------------Log_IoSeqEntry---
    /*! function:    TryResumeTasks
    
        description: resumes all tasks waiting for this IoSeq 
        
        arguments:   CallingTaskId [in] the id of the task calling this function
    */
    inline void TryResumeTasks (tsp00_TaskId CallingTaskId);
};
/*! endclass: Log_IoSeqEntry */


//=======================================================================================
/*! class:       Log_IoSeqList

    description: a list containing I/O sequence entries 
*/
class Log_IoSeqList: public Log_DoubleChainList
{
private:
    Log_IoSeqEntry *m_First;
    Log_IoSeqEntry *m_Last;
    
public:
    
    //--------------------------------------------------------------Log_IoSeqList---
    /*! function:    Log_IoSeqList ctor
    */
    inline Log_IoSeqList ();

    //--------------------------------------------------------------Log_IoSeqList---
    /*! function:     GetAndRemoveAny
    
        description:  removes an entry from the list

        return value: the pointer to the removed entry
    */
    Log_IoSeqEntry *GetAndRemoveAny ()
    {
        return ( REINTERPRET_CAST (Log_IoSeqEntry*, Log_DoubleChainList::GetAndRemoveAny()) );
    }

    //--------------------------------------------------------------Log_IoSeqList---
    /*! function:     GetFirst
    
        return value: the pointer to the first entry of the list
    */
    inline Log_IoSeqEntry *GetFirst () const
    {
        return ( REINTERPRET_CAST (Log_IoSeqEntry*, Log_DoubleChainList::GetFirst()) );
    }
        
    //--------------------------------------------------------------Log_IoSeqList---
    /*! function:     GetLast
    
        return value: the pointer to the last entry of the list
    */
    Log_IoSeqEntry *GetLast () const
    {
        return ( REINTERPRET_CAST (Log_IoSeqEntry*, Log_DoubleChainList::GetLast()) );
    }
    
    //--------------------------------------------------------------Log_IoSeqList---
    /*! function:     GetLastIoSeq
    
        return value: the I/O sequence of the last entry, zero if the list is empty
    */
    inline Log_IoSeqUInt GetLastIoSeq () const;

    //--------------------------------------------------------------Log_IoSeqList---
    void Insert(Log_IoSeqEntry   *pEntry)
    {
		// PTS 1124684 UH 2003-10-16 new
        Log_DoubleChainList::Insert( pEntry, SearchPrevEntry(pEntry->GetIoSeq()) );
    }
    //--------------------------------------------------------------Log_IoSeqList---
    void Append(Log_IoSeqEntry *pEntry)
    {
        Log_DoubleChainList::Append(pEntry);
    }
    //--------------------------------------------------------------Log_IoSeqList---
    /*! function:     WaitForIoSeq
        
        description:  finds the largest IoSeqEntry which is less IoSeq parameter and waits there
 
        arguments:    RedoTaskId     [in]
                      pOwnEotSeqEntry [in] pointer to IoSeqEntry containing the end of trans entry
                      IoSeq          [in] sequence of the log entry handled by the redo task 
                      WaitInserted   [out] true, if the task id is inserted into a wait list 
    */
    void WaitForIoSeq (
        tsp00_TaskId    RedoTaskId,
        Log_IoSeqEntry *pOwnEotSeqEntry,
        Log_IoSeqUInt   IoSeq,
        bool           &WaitInserted);
    
    //--------------------------------------------------------------Log_IoSeqList---
   /// search an entry which has the greatest ioseq which is smaller then the given
   /// If the list is empty or the given ioseq is smaller then all 0 is returned.
   Log_IoSeqEntry* SearchPrevEntry(const Log_IoSeqUInt IoSeq)
   {
       // PTS 1124684 UH 2003-10-16 new
       if ( GetFirst() == 0 || GetLast() == 0)
           return 0;

       const Log_IoSeqUInt firstIoSeq = GetFirst()->GetIoSeq();
       const Log_IoSeqUInt lastIoSeq  = GetLast()->GetIoSeq();

	   for ( Log_IoSeqEntry* pCurrentEntry = GetLast();
             pCurrentEntry != 0;
             pCurrentEntry = pCurrentEntry->GetPrev() )
       {
           if ( IoSeq > pCurrentEntry->GetIoSeq() )
           return pCurrentEntry;
       }
       return 0;
   }
   /// search an entry which has the same ioseq if none is found 0 is returned.
   Log_IoSeqEntry* SearchEntry(const Log_IoSeqUInt IoSeq)
   {
	// PTS 1124684 UH 2003-10-16 new
       for ( Log_IoSeqEntry* pCurrentEntry = GetLast();
             pCurrentEntry != 0;
             pCurrentEntry = pCurrentEntry->GetPrev() )
       {
           if ( IoSeq == pCurrentEntry->GetIoSeq() )
               return pCurrentEntry;
       }
       return 0;
   }
};
/*! endclass: Log_IoSeqList */


/*===========================================================================*
*  CLASSES                                                                   *
*============================================================================*/

//=======================================================================================
/*! class:       Log_RedoTrafficControl

    description: lists to synchronize the parallel redo (singleton)
*/
class Log_RedoTrafficControl: public Log_IRedoTrafficControl
{
private:
    
    SAPDB_UInt              m_NumberOfTransEntries;
    bool                    m_EndOfLog;
    Kernel_WaitStack        m_VacantJobWait;
    Log_IoSeqList           m_IoSeqList;
    Log_IoSeqList           m_IoSeqFreeList;
    Log_JobEntry           *m_pFirstVacancy;
    Log_IoSeqEntry         *m_pIoSeqContainer;
    Log_JobEntry           *m_pJobContainer;
    SAPDBMem_IRawAllocator &m_RawAllocator;
    volatile bool          &m_redoIsAborted;
	SAPDB_UInt             &m_inserted;
	SAPDB_UInt             &m_removed;
    bool                    m_RedoProcessingDisabled; // PTS 1124684 UH 2003-10-16
    
    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    Log_RedoTrafficControl ctor
    
        description: private ctor only used once during initialization

        argument:    RawAllocator [in]
    */
    inline Log_RedoTrafficControl (SAPDBMem_IRawAllocator &RawAllocator,
                                   volatile bool          &redoIsAborted,
                                   SAPDB_UInt             &m_inserted,
                                   SAPDB_UInt             &m_removed);
    
    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:     GetAndRemoveFromFreeIoSeqList
    
        return value: an IoSeqEntry which is removed from the free list
    */
    Log_IoSeqEntry *GetAndRemoveFromFreeIoSeqList ()
        
    {
        return ( m_IoSeqFreeList.GetAndRemoveAny() );
    }

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:     GetJobEntry
    
        return value: the pointer to the job entry specified by its TransIndex

        arguments:    EntryIndex [in]
    */
    inline Log_JobEntry *GetJobEntry (tgg00_TransIndex TransIndex);
    
    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    PutIntoFreeIoSeqList
    
        description: puts an IoSeqEntry into the free list
    */
    void PutIntoFreeIoSeqList (Log_IoSeqEntry *pFreeEntry)
    {
        m_IoSeqFreeList.Append (pFreeEntry);
    }

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    RemoveJob
    
        description: a redo task removes a job and resumes waiting tasks if possible

        arguments:   RedoTaskId [in]
                     pJobEntry  [in]
    */
    void RemoveJob (
        tsp00_TaskId  RedoTaskId,
        Log_JobEntry *pJobEntry);

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    CheckJobEntry
    */
    void CheckJobEntry (tsp00_TaskId         taskid,
                        tgg00_TransIndex     index,
                        SAPDB_UInt           iosequence,
                        const tgg91_TransNo &transno);

	//--------------------------------------------------------------Log_RedoTrafficControl---
    /*! class:       Log_RedoSync

        description: class to get exclusive access to the redo lists
    */
    
    class Log_RedoSync
    {
    private:
        tsp00_TaskId m_CallingTaskId;
        
    public:
        Log_RedoSync (tsp00_TaskId CallingTaskId);
        
        ~Log_RedoSync ();
    };
    /*! endclass: Log_RedoSync */
    
public:
    
    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    Log_RedoTrafficControl dtor
    
        description: deallocation of containers
    */
    ~Log_RedoTrafficControl ();
    
    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    AllocateInstance
    
        description: allocates the Log_RedoTrafficControl singleton with its containers und supply
      
        arguments:   RawAllocator [in out]
                     IsOk         [out]
    */
    static void AllocateInstance (
        SAPDBMem_IRawAllocator &RawAllocator,
        volatile bool          &redoIsAborted,
        SAPDB_UInt             &inserted,
        SAPDB_UInt             &removed,
        bool                   &IsOk);

    
    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    AllocateContainer
    
        description: allocates the I/O sequence container and the job container
        
                     The transactions of the database are limited by NumberOfTransEntries.
                     if every transaction is closed at a different LogPage, so everey transaction
                     has a different IoSeq. Therefore the IoSeqContainer has NumberOfTransEntries
                     elements. The TransIndex has a range 1..NumberOfTransEntries. To avoid
                     arithmetic expressions when accessing the JobContainer vector, the vector
                     consists of NumberOfTransEntries+1 elements (0 .. n+1).
      
        arguments:   IsOk [out]
    */
    void AllocateContainer (bool &IsOk);

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:     ExecuteJobs
        
        description:  executes vacant jobs. This method is provided for redo tasks.

        argument:     TransContext [in]
    */
    virtual void ExecuteJobs (tgg00_TransContext &TransContext);

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:     GetRawAllocatorUsed
        
        return value: the allocator used to allocate this singleton
    */
    SAPDBMem_IRawAllocator &GetRawAllocatorUsed ()
    {
        return (m_RawAllocator);
    }

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    InsertJob
    
        description: inserts a job. This method is provided for the log reader task.

        arguments:   LogReaderTaskId [in]
                     TransIndex      [in]
                     EndOfTransIoSeq [in]
                     UndoOrRedo      [in]
                     pTransaction    [in]
                     IsOk            [out]
    */
    virtual void InsertJob (
        tsp00_TaskId      LogReaderTaskId,
        tgg00_TransIndex  TransIndex,
        Log_IoSeqUInt     EndOfTransIoSeq,
        Log_ActionType    UndoOrRedo,
        bool              checkAscendingSequence, // PTS 1113641 UH 2002-01-25
        Log_Transaction  &Transaction,
        bool             &IsOk);

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:     InstancePtr
    
        return value: the pointer to the global instance of the singleton
    */
    static Log_RedoTrafficControl *InstancePtr ()
    { 
        return ( REINTERPRET_CAST (Log_RedoTrafficControl*, m_pInstance) );
    }

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:    SetEndOfLogRead
    
        description: the log reader task sets the state: end of log read

        argument:    LogReaderTaskId [in]
    */
    virtual void SetEndOfLogRead (tsp00_TaskId LogReaderTaskId);

    //--------------------------------------------------------------Log_IRedoTrafficControl---
    /*! function:     CheckIoSeq
        description:  a redo task checks if preceding transactions are closed
        arguments:    RedoTaskId [in]
                      TransIndex [in] reference to the transaction handled by the redo task  
                      IoSeq      [in] sequence of the log entry handled by the redo task 
    */
    virtual bool CheckIoSeq (
        tsp00_TaskId     RedoTaskId,
        tgg00_TransIndex TransIndex,
        Log_IoSeqUInt    IoSeq);

    //--------------------------------------------------------------Log_RedoTrafficControl---
    /*! function:     WaitForIoSeq
        description:  a redo task waits until preceding transactions are closed 
        arguments:    RedoTaskId [in]
    */
    virtual void WaitForIoSeq (tsp00_TaskId RedoTaskId);
    /// The log reader calls this after recreation of all open transactions from last savepoint
    virtual void EnableRedoProcessing (tsp00_TaskId LogReaderTaskId)
    // PTS 1124684 UH 2003-10-16 new
    {
        // PTS 1126107 mb 2003-12-10
        SAPDBTRACE_METHOD_DEBUG ("Rst_RedoTraffic::EnableRedoProcessing", LogTrans_Trace, 5);
        Log_RedoSync RedoSync (LogReaderTaskId);
        
        Log_IoSeqEntry *firstIoSeqEntry = m_IoSeqList.GetFirst();
        m_RedoProcessingDisabled        = false;
        m_pFirstVacancy = firstIoSeqEntry == 0
                          ? 0
                          : firstIoSeqEntry->GetFirstJobEntry();
        m_VacantJobWait.PopAnyTaskIdAndResume(LogReaderTaskId);
    }
	/// writes some basic information into the trace 
	virtual void WriteToTrace (const char *title = 0) const;
	/// This returns the youngest iosequence inserted to the redo list
	virtual Log_IOSequenceNo GetLastIOSequence (tsp00_TaskId taskid) const;
};
/*! endclass: Log_RedoTrafficControl */


/*===========================================================================*
*  METHODS of Log_DoubleChainEntry                                           *
*============================================================================*/

inline Log_DoubleChainEntry::Log_DoubleChainEntry ():
m_pNext (NULL),
m_pPrev (NULL)
{ }

//----------------------------------------------------------------------------
inline void Log_DoubleChainEntry::Insert (Log_DoubleChainEntry* &pFirstListEntry,
			 Log_DoubleChainEntry* &pLastListEntry,
             Log_DoubleChainEntry*  pPrevListEntry)
{
	// PTS 1124684 UH 2003-10-16 written from earlier Append()
    if (0 == pFirstListEntry)
    {
        m_pPrev         = NULL;
        m_pNext         = NULL;
        pFirstListEntry = this;
        pLastListEntry  = this;
        return;
    }

    m_pPrev = pPrevListEntry;

    if (pPrevListEntry == 0)
    {
        m_pNext         = pFirstListEntry;
        pFirstListEntry = this;
    }
    else
    {
        m_pNext                 = pPrevListEntry->m_pNext;
        pPrevListEntry->m_pNext = this;
    }

    if ( m_pNext != 0 )
        m_pNext->m_pPrev = this;

    if (pPrevListEntry == pLastListEntry)
        pLastListEntry = this;
}

//----------------------------------------------------------------------------

inline void Log_DoubleChainEntry::Remove (Log_DoubleChainEntry *&pFirstListEntry,
                                          Log_DoubleChainEntry *&pLastListEntry)
{
    if (NULL == this->m_pPrev)
    {
        pFirstListEntry = this->m_pNext;
    }
    else
    {
        this->m_pPrev->m_pNext = this->m_pNext;
    }
    
    if (NULL == this->m_pNext)
    {
        pLastListEntry = this->m_pPrev;
    }
    else
    {
        this->m_pNext->m_pPrev = this->m_pPrev;
    }
    
    this->m_pPrev = NULL;
    this->m_pNext = NULL;
}


/*===========================================================================*
*  METHODS of Log_DoubleChainList                                            *
*============================================================================*/

inline Log_DoubleChainList::Log_DoubleChainList ():
m_pFirst (NULL),
m_pLast  (NULL)
{ }

//----------------------------------------------------------------------------

Log_DoubleChainEntry* Log_DoubleChainList::GetAndRemoveAny ()
{
    Log_DoubleChainEntry *pEntry = this->m_pFirst;
    this->Remove (pEntry);
    return (pEntry);
}

//----------------------------------------------------------------------------

inline Log_DoubleChainEntry* Log_DoubleChainList::GetFirst () const
{
    if ( this->IsEmpty() ) return (NULL);
    
    return (this->m_pFirst);
}

//----------------------------------------------------------------------------

inline Log_DoubleChainEntry* Log_DoubleChainList::GetLast () const
{
    if ( this->IsEmpty() ) return (NULL);
    
    return (this->m_pLast);
}


/*===========================================================================*
*  METHODS of Log_JobEntry                                                   *
*============================================================================*/

inline Log_JobEntry::Log_JobEntry ():
Log_DoubleChainEntry (),
m_OwnIndex           (cgg_nil_transindex),
m_UndoOrRedo         (Log_NoOp),
m_pIoSeqEntry        (NULL),
m_Transaction        ()
{ }

//----------------------------------------------------------------------------

inline void Log_JobEntry::Init (Log_ActionType   UndoOrRedo,
                                Log_IoSeqEntry  *pIoSeqEntry,
                                Log_Transaction &transaction)
{
    this->m_UndoOrRedo   = UndoOrRedo;
    this->m_pIoSeqEntry  = pIoSeqEntry;
    this->m_Transaction  = transaction;
}


/*===========================================================================*
*  METHODS of Log_IoSeqEntry                                                 *
*============================================================================*/

inline Log_IoSeqEntry::Log_IoSeqEntry ():
Log_DoubleChainEntry (),
m_IoSeq              (0),
m_JobList            (),
m_WaitStack          ()
{ }

//----------------------------------------------------------------------------

inline void Log_IoSeqEntry::TryResumeTasks (tsp00_TaskId CallingTaskId)
{
    if ( this->m_JobList.IsEmpty() )
    {
        this->m_WaitStack.PopAllTaskIdsAndResume (CallingTaskId);
    }
}


/*===========================================================================*
*  METHODS of Log_IoSeqList                                                  *
*============================================================================*/

inline Log_IoSeqList::Log_IoSeqList ():
Log_DoubleChainList ()
{ }

//----------------------------------------------------------------------------

inline Log_IoSeqUInt Log_IoSeqList::GetLastIoSeq () const
{
    if ( this->IsEmpty() ) return ( SAPDB_MAX_UINT4 );
    
    return ( this->GetLast()->GetIoSeq() );
}

//----------------------------------------------------------------------------

void Log_IoSeqList::WaitForIoSeq (tsp00_TaskId    RedoTaskId,
                                  Log_IoSeqEntry *pOwnEotSeqEntry,
                                  Log_IoSeqUInt   IoSeq,
                                  bool            &WaitInserted)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_IoSeqList::WaitForIoSeq", LogTrans_Trace, 5);
    
    WaitInserted = false;
    
    if ( LogTrans_Trace.TracesLevel(6) )
		Kernel_VTrace() << "REDOTRAFFIC: check ioseq: " << IoSeq
                        << ", first eotseq: " << GetFirst()->GetIoSeq() << NewLine;

    if ( IoSeq <= GetFirst()->GetIoSeq() )
        return;
        
    Log_IoSeqEntry *pCurr = this->GetFirst();
 
    if ( NULL == pCurr )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                   SAPDBERR_ASSERT_STATE_FAILED,
                   "Log_IoSeqList::WaitForIoSeq() pCurr is nil.") );
 
    Log_IoSeqEntry *pNext = pCurr->GetNext();

    if ( NULL == pNext )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                   SAPDBERR_ASSERT_STATE_FAILED,
                   "Log_IoSeqList::WaitForIoSeq() pNext is nil.") );
                   
    while (pNext->GetIoSeq() < IoSeq)
    {
        pCurr = pNext;
        pNext = pNext->GetNext();
    }

    if (pCurr->GetIoSeq() < IoSeq)
    {
        pCurr->InsertWaitTask (RedoTaskId);
        WaitInserted = true;
    }
    if ( WaitInserted && LogTrans_Trace.TracesLevel(6) )
        Kernel_VTrace() << "REDOTRAFFIC: wait for eotseq: " << pCurr->GetIoSeq();
}


/*===========================================================================*
*  METHODS of Log_RedoTrafficControl                                         *
*============================================================================*/

inline Log_RedoTrafficControl::Log_RedoTrafficControl (SAPDBMem_IRawAllocator &RawAllocator,
                                                       volatile bool          &redoIsAborted,
                                                       SAPDB_UInt             &inserted,
                                                       SAPDB_UInt             &removed):
m_NumberOfTransEntries (0),
m_EndOfLog             (false),
m_VacantJobWait        (),
m_IoSeqList            (),
m_IoSeqFreeList        (),
m_pFirstVacancy        (NULL),
m_pIoSeqContainer      (NULL),
m_pJobContainer        (NULL),
m_RawAllocator         (RawAllocator),
m_redoIsAborted        (redoIsAborted),
m_inserted             (inserted),
m_removed              (removed),
m_RedoProcessingDisabled (true) // PTS 1124684 UH 2003-10-16
{ }

//----------------------------------------------------------------------------

Log_RedoTrafficControl::~Log_RedoTrafficControl ()
{
    if ((NULL != this->m_pIoSeqContainer) && (this->m_NumberOfTransEntries > 0))
    {
        destroyarray (
            this->m_pIoSeqContainer,
            this->m_NumberOfTransEntries,
            this->m_RawAllocator);
    }
    
    if ((NULL != this->m_pJobContainer) && (this->m_NumberOfTransEntries > 0))
    {
        destroyarray (
            this->m_pJobContainer,
            this->m_NumberOfTransEntries,
            this->m_RawAllocator);
    }
    
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::AllocateInstance (SAPDBMem_IRawAllocator &RawAllocator,
                                               volatile bool          &redoIsAborted,
                                               SAPDB_UInt             &inserted,
                                               SAPDB_UInt             &removed,
                                               bool                   &IsOk)
{  
    SAPDBTRACE_ROUTINE_DEBUG ("Log_RedoTraffic::AllocateInstance", LogTrans_Trace, 5);
    
    if (NULL == m_pInstance)
    {
        m_pInstance = new (RawAllocator) Log_RedoTrafficControl
                                         ( RawAllocator,
                                           redoIsAborted,
                                           inserted,
                                           removed );
    }
    
    IsOk = (NULL != m_pInstance);
}

// --------------------------------------------------------------------------------

void Log_RedoTrafficControl::AllocateContainer (bool &IsOk)
{  
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::AllocateContainer", LogTrans_Trace, 5);
    
    SAPDB_UInt i;
    
    IsOk = true;
    
    if (this->m_NumberOfTransEntries > 0)
    {
        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "NOT OK");
        IsOk = false;
        return;
    }
    
    this->m_NumberOfTransEntries = k51number_of_trans_entries();
    
    if (NULL == this->m_pIoSeqContainer)
    {
        newarray (this->m_pIoSeqContainer, this->m_NumberOfTransEntries, this->m_RawAllocator);
        
        if (NULL == this->m_pIoSeqContainer)
        {
            SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "NOT OK");
            IsOk = false;
            return;
        }
    }
    
    if (NULL == this->m_pJobContainer)
    {
        newarray (this->m_pJobContainer, this->m_NumberOfTransEntries, this->m_RawAllocator);
        
        if (NULL == this->m_pJobContainer)
        {
            SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "NOT OK");
            IsOk = false;
            return;
        }
    }
    
    for (i = 0; i < this->m_NumberOfTransEntries; ++i)
    {
        this->m_IoSeqFreeList.Append (this->m_pIoSeqContainer+i);
    }
    
    for (i = 0; i < this->m_NumberOfTransEntries; ++i)
    {
        (this->m_pJobContainer+i)->SetOwnIndex (i+1);
    }
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::ExecuteJobs (tgg00_TransContext &TransContext)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::ExecuteJobs", LogTrans_Trace, 5);
    
    if (e_ok != TransContext.trError_gg00)
    {
        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "trError: " << TransContext.trError_gg00);
        return;
    }
    
    Log_JobEntry *pJob;
    bool          WaitInserted;
    bool          EndOfLog = false;
    
    while ( e_ok == TransContext.trError_gg00
            &&
            ! EndOfLog
            &&
            ! m_redoIsAborted )
    {    
        WaitInserted = false;
        pJob         = NULL;
        
        {
            Log_RedoSync RedoSync (TransContext.trTaskId_gg00);
            
            if ( NULL == m_pFirstVacancy
                 ||
                 m_RedoProcessingDisabled ) // PTS 1124684 UH 2003-10-16
            {
                EndOfLog = this->m_EndOfLog;
                if ( ! EndOfLog )
                {
                    this->m_VacantJobWait.PushTaskId (TransContext.trTaskId_gg00);
                    WaitInserted = true;
                }
            }
            else
            {
                pJob = this->m_pFirstVacancy;
                
                this->m_pFirstVacancy = pJob->GetNext();
                
                if (NULL == this->m_pFirstVacancy)
                {
                    Log_IoSeqEntry *pCurrIoSeq = pJob->GetIoSeqEntry()->GetNext();
                    
                    if (NULL != pCurrIoSeq)
                    {
                        // every IoSeqEntry has at least one job entry (scan of IoSeqList not necessary) 
                        this->m_pFirstVacancy = pCurrIoSeq->GetFirstJobEntry ();
                    }            
                }
            }
        }
        
        // outside RedoSync region
        
        if ( NULL == pJob )
        {
            if (WaitInserted)
                vsuspend (TransContext.trTaskId_gg00, cgg_suspend_lvl_0, LOG_WAIT_FOR_VACANT_ACTION);
        }
        else
        {
            if ( m_redoIsAborted )
                break;
                
            TransContext.trIndex_gg00        = pJob->GetTransIndex();
            TransContext.trWriteTransId_gg00 = pJob->GetTransaction().GetRedoTransNo();
            
            int           checkindex  = TransContext.trIndex_gg00;
            tgg91_TransNo redoTransNo = TransContext.trWriteTransId_gg00;

			pJob->GetTransaction().InitializeForRedo (TransContext);

			if ( LogTrans_Trace.TracesLevel(6) )
				Kernel_VTrace() << "REDOTRAFFIC: start T" << redoTransNo.gg90GetInt4()
								<< ", EOTSEQ: " << pJob->GetIoSeqEntry()->GetIoSeq()
								<< ", KIND: " << (pJob->GetUndoOrRedo()==Log_Commit?"redo":"undo")
                                << NewLine;
                                
            switch ( pJob->GetUndoOrRedo() )
            {
            case Log_Commit:

                if ( pJob->GetTransaction().EntriesToRedoExist() )
                {
                    Log_Transaction::RedoResult RedoResult;
                    
                    RedoResult = pJob->GetTransaction().Redo();
                    
                    if ( TransContext.trError_gg00 == e_ok
                         &&
                         Log_Transaction::redoOk == RedoResult )
                    {
                        // synchronize with other commits
                        if ( CheckIoSeq (TransContext.trTaskId_gg00,
                                         pJob->GetTransIndex(),
                                         pJob->GetIoSeqEntry()->GetIoSeq()) )
                            WaitForIoSeq (TransContext.trTaskId_gg00);
                    }
                    else
                    {
                        RTE_Message( Log_Exception(__CONTEXT__,LOG_REDO_ERROR,
                                                   SAPDB_ToString(TransContext.trWriteTransId_gg00.gg90GetInt4(),_T_d),
                                                   "committed",
                                                   SAPDB_ToString(RedoResult,_T_d),
                                                   SAPDB_ToString(TransContext.trError_gg00,_T_d) ) );
                        if ( TransContext.trError_gg00 == e_ok
                             &&
                             Log_Transaction::redoOk != RedoResult )
                        {
                            TransContext.trError_gg00 = e_log_error;
                        }
                    }
                }
                break;

            case Log_Rollback:

                if ( pJob->GetTransaction().UndoEntriesExist() )
                {
                    Log_Transaction::RollbackResult RollbackResult;

                    if ( CheckIoSeq (TransContext.trTaskId_gg00,
                                     pJob->GetTransIndex(),
                                     pJob->GetIoSeqEntry()->GetIoSeq()) )
                        WaitForIoSeq (TransContext.trTaskId_gg00);
                    
                    RollbackResult = pJob->GetTransaction().Rollback ( Log_Transaction::endOfTrans,
                                                                       LOG_NIL_SUBTRANS_NO );
                    
                    if ( TransContext.trError_gg00 != e_ok
                         ||
                         Log_Transaction::undoOk != RollbackResult )
                    {
                        RTE_Message( Log_Exception(__CONTEXT__,LOG_REDO_ERROR,
                                                   SAPDB_ToString(TransContext.trWriteTransId_gg00.gg90GetInt4(),_T_d),
                                                   "rollbacked",
                                                   SAPDB_ToString(RollbackResult,_T_d),
                                                   SAPDB_ToString(TransContext.trError_gg00,_T_d) ) );
                        if ( TransContext.trError_gg00 == e_ok
                             &&
                             Log_Transaction::undoOk != RollbackResult )
                        {
                            TransContext.trError_gg00 = e_log_error;
                        }
                    }
                }
                break;
                
            default:
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Log_RedoTrafficControl::ExecuteJobs() invalid redo type") );
            }
            
            if ( e_ok == TransContext.trError_gg00
                 &&
                 checkindex != TransContext.trIndex_gg00 )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Log_RedoTrafficControl::ExecuteJobs() transindex mismatch") );

            tgg00_BasisError auxerror = TransContext.trError_gg00;
            TransContext.trError_gg00 = e_ok;

            // The following sequence of CleanUp is important  !!!
            
            this->RemoveJob (TransContext.trTaskId_gg00, pJob);
            
            pJob->GetTransaction().Delete();

            if (e_ok != TransContext.trError_gg00)
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Log_RedoTrafficControl::ExecuteJobs() LogTrans.Delete()") );

            k51remove_locks                 ( TransContext.trTaskId_gg00,
                                              TransContext.trIndex_gg00,
                                              TransContext.trState_gg00 );
            k51del_transentry_from_locklist ( TransContext.trTaskId_gg00,
                                              TransContext.trIndex_gg00 );

            // really drop all dropped files
            k53server_drop_aux_files (TransContext);

            TransContext.trWriteTransId_gg00.gg90SetNil();
            TransContext.trLogContext_gg00 = NULL;
            TransContext.trIndex_gg00      = cgg_nil_transindex;

            if( auxerror != e_ok )
                TransContext.trError_gg00 = auxerror;
        }
    }

    if ( e_ok != TransContext.trError_gg00 )
        kb811AbortRedo(TransContext.trError_gg00 );
}

//----------------------------------------------------------------------------

inline Log_JobEntry* Log_RedoTrafficControl::GetJobEntry (tgg00_TransIndex TransIndex)
{
    if (cgg_nil_transindex == TransIndex)
        return (NULL);
    
    Log_JobEntry* entry = (this->m_pJobContainer+TransIndex-1);
    
    if ( TransIndex != entry->GetTransIndex() )
    {
        RTE_Crash( SAPDBErr_Exception(
                   __FILE__, __LINE__, SAPDBERR_ASSERT_STATE_FAILED,
                   "Log_RedoTrafficControl::GetJobEntry: transindex mismatch") );
    }
    return entry;
}

//----------------------------------------------------------------------------

Log_IOSequenceNo Log_RedoTrafficControl::GetLastIOSequence (tsp00_TaskId taskid) const
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::GetLastIOSequence", LogTrans_Trace, 5);
    Log_RedoSync RedoSync (taskid); // ENTER REGION
    return m_IoSeqList.GetLastIoSeq();
}

//----------------------------------------------------------------------------
void Log_RedoTrafficControl::InsertJob (tsp00_TaskId      LogReaderTaskId,
                                        tgg00_TransIndex  TransIndex,
                                        Log_IoSeqUInt     EndOfTransIoSeq,
                                        Log_ActionType    UndoOrRedo,
                                        bool              checkAscendingSequence, // PTS 1113641 UH 2002-01-25
                                        Log_Transaction  &Transaction,
                                        bool             &IsOk)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::InsertJob", LogTrans_Trace, 5);

    CheckJobEntry (LogReaderTaskId, TransIndex, EndOfTransIoSeq, Transaction.GetRedoTransNo());
    
    Log_RedoSync    RedoSync (LogReaderTaskId);
    Log_JobEntry   *pJobEntry = this->GetJobEntry (TransIndex);
    
	if ( LogTrans_Trace.TracesLevel(6) )
    	Kernel_VTrace() << "REDOTRAFFIC: insert T" << Transaction.GetRedoTransNo().gg90GetInt4()
        	            << ", EOTSEQ: " << EndOfTransIoSeq
            	        << ", KIND: " << (UndoOrRedo==Log_Commit?"redo":"undo")
                        << NewLine;

    IsOk = false;
    
    if (NULL == pJobEntry)
    {
        Kernel_VTrace() << "REDOTRAFFIC: no free job entry";
        return;
    }
    
    Log_IoSeqEntry *pEotIoSeqEntry = m_IoSeqList.SearchEntry(EndOfTransIoSeq);  // PTS 1124684 UH 2003-10-16


    if ( 0 == pEotIoSeqEntry )
    {
        // EndOfTransIoSeq is a new io sequence which must be added to the list
        // at the correct position
        if ( checkAscendingSequence // PTS 1113641 UH 2002-01-25
             &&
             m_IoSeqList.GetLastIoSeq() != SAPDB_MAX_UINT4 // Log_IOSequenceNo() if list is empty this is nil
             // PTS 1120201 UH 2003-02-17
             &&
             m_IoSeqList.GetLastIoSeq() > EndOfTransIoSeq
           )
        {
            Kernel_VTrace() << "EndOfTransIoSeq: " << EndOfTransIoSeq // PTS 1113641 UH 2002-01-25
                            << " > lastIOSequence: " << m_IoSeqList.GetLastIoSeq();
            Kernel_VTrace() << "REDOTRAFFIC: Insert IO SEQUENCE FAILED";
            return;
        }
        
        pEotIoSeqEntry = GetAndRemoveFromFreeIoSeqList ();
        
        if ( 0 == pEotIoSeqEntry )
        {
            Kernel_VTrace() << "REDOTRAFFIC: no free ioseqlist entry";
            return;
        }
        
        pEotIoSeqEntry->Init(EndOfTransIoSeq);
        
        // PTS 1124684 UH 2003-10-16 use Insert() instead of Append()
        m_IoSeqList.Insert (pEotIoSeqEntry);
    }
	
    pJobEntry->Init (UndoOrRedo, pEotIoSeqEntry, Transaction);

    pEotIoSeqEntry->InsertJobEntry (pJobEntry);
    
    if (NULL == this->m_pFirstVacancy)
    {
        this->m_pFirstVacancy = pJobEntry;
    }
    if ( checkAscendingSequence ) // PTS 1124684 UH 2003-10-16
	    this->m_VacantJobWait.PopAnyTaskIdAndResume (LogReaderTaskId);

    IsOk = true;
    
    // The global locklist must use now the Log_Transaction::m_Context of pJobEntry
    // This must be synchronized with the savepoint coordinator.

    kb50EnterTransRegion   (LogReaderTaskId, TransIndex);
    k50RegisterTransaction (TransIndex, &pJobEntry->GetTransaction());
    kb50LeaveTransRegion   (LogReaderTaskId, TransIndex);
    
    ++m_inserted;

	if ( LogTrans_Trace.TracesLevel(7) )
        WriteToTrace("End of InsertJob()");
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::CheckJobEntry (tsp00_TaskId         taskid,
                                            tgg00_TransIndex     index,
                                            SAPDB_UInt           iosequence,
                                            const tgg91_TransNo &transno)
{
    Log_RedoSync    RedoSync (taskid);
    
    Log_IoSeqEntry *pEotIoSeqEntry = NULL;
    Log_JobEntry   *pJobEntry      = NULL;
    
    pEotIoSeqEntry = m_IoSeqList.GetFirst();

	while ( pEotIoSeqEntry != NULL )
	{
		pJobEntry = pEotIoSeqEntry->GetFirstJobEntry();
		while ( pJobEntry != NULL )
		{
            if ( index == pJobEntry->GetTransIndex() )
            {
                Kernel_VTrace() << "T"              << transno.gg90GetInt4()
                                << ", index: "      << index
                                << ", iosequence: " << iosequence << NewLine;
                WriteToTrace("Log_RedoTrafficControl::CheckJobEntry() duplicate index");
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Log_RedoTrafficControl::CheckJobEntry() duplicate index.") );
            }
            if ( transno == pJobEntry->GetTransaction().GetRedoTransNo() )
            {
                Kernel_VTrace() << "T"              << transno.gg90GetInt4()
                                << ", index: "      << index
                                << ", iosequence: " << iosequence << NewLine;
                WriteToTrace("Log_RedoTrafficControl::CheckJobEntry() duplicate transno");
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                           SAPDBERR_ASSERT_STATE_FAILED,
                           "Log_RedoTrafficControl::CheckJobEntry() duplicate transno") );
            }
			pJobEntry = pJobEntry->GetNext();
		}
		pEotIoSeqEntry = pEotIoSeqEntry->GetNext();
	}

        
                   
    
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::WriteToTrace (const char *title) const
{
    Log_IoSeqEntry *pEotIoSeqEntry = NULL;
    Log_JobEntry   *pJobEntry      = NULL;
    
	if ( title != NULL )
		Kernel_VTrace() << title;

    pEotIoSeqEntry = m_IoSeqList.GetFirst();

	Kernel_VTrace() << "REDOLIST:";

	while ( pEotIoSeqEntry != NULL )
	{
        Kernel_VTrace() << "=== IOSEQ ===: " << pEotIoSeqEntry->GetIoSeq();

		pJobEntry = pEotIoSeqEntry->GetFirstJobEntry();
		while ( pJobEntry != NULL )
		{
            Kernel_VTrace() << "T" << pJobEntry->GetTransaction().GetRedoTransNo().gg90GetInt4()
				            << ", index: " << pJobEntry->GetTransIndex()
							<< (pJobEntry->GetUndoOrRedo()==Log_Commit?", redo":", undo")
                            << (m_pFirstVacancy==pJobEntry?" => FirstVacant":"");
			pJobEntry = pJobEntry->GetNext();
		}
		pEotIoSeqEntry = pEotIoSeqEntry->GetNext();
	}

        
                   
    
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::RemoveJob (tsp00_TaskId  RedoTaskId,
                                        Log_JobEntry *pJobEntry)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::RemoveJob", LogTrans_Trace, 5);

	if ( LogTrans_Trace.TracesLevel(6) )
    	Kernel_VTrace() << "REDOTRAFFIC: remove T" << pJobEntry->GetTransaction().GetRedoTransNo().gg90GetInt4()
        	            << ", EOTSEQ: " << pJobEntry->GetIoSeqEntry()->GetIoSeq()
            	        << ", KIND: " << (pJobEntry->GetUndoOrRedo()==Log_Commit?"redo":"undo");
    
    Log_RedoSync RedoSync (RedoTaskId);
    
    Log_IoSeqEntry *pEotSeqEntry = pJobEntry->GetIoSeqEntry();
    
    tgg00_TransIndex transindex = pJobEntry->GetTransIndex();
    kb50EnterTransRegion   (RedoTaskId, transindex);
    k50RegisterTransaction (transindex, NULL);
    kb50LeaveTransRegion   (RedoTaskId, transindex);

    pEotSeqEntry->RemoveJobEntry (pJobEntry);
 
    if ( pEotSeqEntry->IsEmptyJobList() )
    {
        pEotSeqEntry->TryResumeTasks (RedoTaskId);
        this->m_IoSeqList.Remove (pEotSeqEntry);
        this->PutIntoFreeIoSeqList (pEotSeqEntry); // added by UH 2001-07-03
    }

    ++m_removed;
    // WriteToTrace("REDOTRAFFIC: after remove");
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::SetEndOfLogRead (tsp00_TaskId LogReaderTaskId)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::SetEndOfLogRead", LogTrans_Trace, 5);
    
    Log_RedoSync RedoSync (LogReaderTaskId);

    this->m_EndOfLog = true;
    this->m_VacantJobWait.PopAllTaskIdsAndResume (LogReaderTaskId);
}

//----------------------------------------------------------------------------

bool Log_RedoTrafficControl::CheckIoSeq (tsp00_TaskId     RedoTaskId,
                                         tgg00_TransIndex TransIndex,
                                         Log_IoSeqUInt    IoSeq)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::CheckIoSeq", LogTrans_Trace, 5);
    
    bool            WaitInserted;
    Log_IoSeqEntry *pOwnEotSeqEntry;
    
    WaitInserted = false;
    {
        Log_RedoSync RedoSync (RedoTaskId);

        pOwnEotSeqEntry = (m_pJobContainer+TransIndex-1)->GetIoSeqEntry();

        m_IoSeqList.WaitForIoSeq (RedoTaskId, pOwnEotSeqEntry, IoSeq, WaitInserted);
    }

    return WaitInserted;
}

//----------------------------------------------------------------------------

void Log_RedoTrafficControl::WaitForIoSeq (tsp00_TaskId RedoTaskId)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_RedoTraffic::WaitForIoSeq", LogTrans_Trace, 5);
    vsuspend (RedoTaskId, cgg_suspend_lvl_0, LOG_WAIT_FOR_PRECEDING_TRANS);
}


/*===========================================================================*
*  METHODS of Log_RedoSync                                                   *
*============================================================================*/

Log_RedoTrafficControl::Log_RedoSync::Log_RedoSync (tsp00_TaskId CallingTaskId)
{
    this->m_CallingTaskId = CallingTaskId;
    vbegexcl (this->m_CallingTaskId, g08lockwait);
}

//----------------------------------------------------------------------------

Log_RedoTrafficControl::Log_RedoSync::~Log_RedoSync ()
{
    vendexcl (this->m_CallingTaskId, g08lockwait);
}

/*===========================================================================*
*  METHODS of Log_RedoAllocateTrafficControl                                 *
*============================================================================*/

Log_RedoAllocateTrafficControl::~Log_RedoAllocateTrafficControl ()
{
    Log_RedoTrafficControl *pRedoTrafficControl = Log_RedoTrafficControl::InstancePtr();
    
    if (NULL == pRedoTrafficControl) return;
    
    SAPDBMem_IRawAllocator &RawAllocator = pRedoTrafficControl->GetRawAllocatorUsed();
    
    destroy (pRedoTrafficControl, RawAllocator);
}

//----------------------------------------------------------------------------

void Log_RedoAllocateTrafficControl::Allocate (SAPDBMem_IRawAllocator &RawAllocator,
                                               volatile bool          &redoIsAborted,
                                               SAPDB_UInt             &inserted,
                                               SAPDB_UInt             &removed,
                                               bool                   &IsOk)
{    
    Log_RedoTrafficControl::AllocateInstance ( RawAllocator,
                                               redoIsAborted,
                                               inserted,
                                               removed,
                                               IsOk);
    
    if ( !IsOk ) return;
    
    Log_RedoTrafficControl::InstancePtr()->AllocateContainer (IsOk);
}

/*===========================================================================*
*  END OF CODE                                                               *
*============================================================================*/
