/*!**************************************************************************

  module      : Log_Savepoint.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : The savepoint manager.


    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_SAVEPOINT_H
#define LOG_SAVEPOINT_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h" // tsp00_TaskId
#include "ggg00.h" // tgg00_TransContext
#include "gkb00.h" // tkb00_SaveptParam

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "KernelCommon/Kernel_DateTime.hpp"
#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "Logging/Log_Types.hpp"
#include "Converter/Converter_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "DataAccess/Data_Types.hpp"

class SAPDBMem_IRawAllocator;

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*! class:       Log_Savepoint
    description: This is the savepoint manager. It is a Singleton.
*/
class Log_Savepoint
{

public:

    /*! --------------------------------------------------------------------------------
    function:    Log_Savepoint
    ----------------------------------------------------------------------------------*/
    Log_Savepoint()
    : m_IsActive         (false),
      m_CurrentWaitList  (0),
      m_WaitingDBmanager (cgg_nil_pid),
      m_DBmanagerReason  (Log_SVPReasonUnknown),
      m_LogWriteEnabled  (false),
      m_SavepointAllowed(true)
    {}
    /*! --------------------------------------------------------------------------------
    function:    StartSavepointAndWait
    arguments:   TransContext [in/out]
                 Reason       [in]      - special reason used by the DatabaseManager.
    description: This starts a savepoint if none is running.
                 If a savepoint is already running, the caller waits for the next savepoint
                 which is executed automatically.
    ----------------------------------------------------------------------------------*/
    void StartSavepointAndWait (tgg00_TransContext  &Trans,
                                Log_SavepointReason  Reason);
    /*! --------------------------------------------------------------------------------
    function:    StartSavepoint
    arguments:   TransContext [in/out]
                 force        [in] - if true, the savepoint is started
    description: If somebody is waiting for a savepoint, a savepoint is started, if no
                 savepoint is active.
    ----------------------------------------------------------------------------------*/
    void StartSavepoint (tgg00_TransContext  &trans,
                         Log_SavepointReason  reason);
    /*! --------------------------------------------------------------------------------
    function:    IsDBManagerReason
    returns:     true, if the given reason can only be triggered at one time.
                 This reasons normally cause special functions during the savepoint.
                 Examples are DataBackup or Restart.
    ----------------------------------------------------------------------------------*/
    bool IsDBManagerReason (Log_SavepointReason reason);
    /*! --------------------------------------------------------------------------------
    function:    IsActive
    returns:     true, if a savepoint is running.
    ----------------------------------------------------------------------------------*/
    bool IsActive () const
    {
        // without region
        return m_IsActive;
    }
    /*! --------------------------------------------------------------------------------
    function:    Coordinator
    arguments:   TransContext [in]
    description: This is the main servertask method.
    ----------------------------------------------------------------------------------*/
    void Coordinator (tgg00_TransContext &Trans);

    /*! --------------------------------------------------------------------------------
    function:    Verify
    arguments:   trans  [in]
                 isCold [in]
    description: verifies opentrans file in restartrecord
    ----------------------------------------------------------------------------------*/
    static bool Verify (tgg00_TransContext &trans,
                        bool                isCold);
    /*! --------------------------------------------------------------------------------
    function:    GetLastSavepointId
    returns:     Log_EntrySequence
    ----------------------------------------------------------------------------------*/
    static Log_EntrySequence GetLastSavepointId();
    
    /*! --------------------------------------------------------------------------------
    function:    GetLastSavepointEntryType
    description: in online mode this will usually be Log_Savepoint but not during redo.
    returns:     Log_ActionType
    ----------------------------------------------------------------------------------*/
    static Log_ActionType GetLastSavepointEntryType();
    
    /*! --------------------------------------------------------------------------------
    function:    GetStartEntryTimestamp
    ----------------------------------------------------------------------------------*/
    static void GetStartEntryTimestamp( Kernel_Date &date,
                                        Kernel_Time &time );

    /*! --------------------------------------------------------------------------------
    function:    GetLastSavepointStatistics
    arguments:   sec  [out]
                 msec [out]
    ----------------------------------------------------------------------------------*/
    void GetSavepointStatistics (SAPDB_Int4 &lastDurationSec,
                                 SAPDB_Int4 &lastDurationMicroSec,
                                 SAPDB_Int4 &avgWrittenPages,
                                 SAPDB_Int4 &avgIOcount)
    {
        // PTS 1119276 UH 2002-12-10 added
        lastDurationSec      = m_Statistics.last.durationSec;
        lastDurationMicroSec = m_Statistics.last.durationMicroSec;
        m_Statistics.GetAvarage(avgWrittenPages,avgIOcount);
    }

    /*! --------------------------------------------------------------------------------
    function:    DenySavepoint
    arguments:   taskid [in]
    description: This can be used to prevent starting of a new savepoint. If false is returned the call failed.
    ----------------------------------------------------------------------------------*/
    bool DenySavepoint (tsp00_TaskId taskid);

    /*! --------------------------------------------------------------------------------
    function:    PermitSavepoint
    arguments:   taskid [in]
    description: This can be used to allow starting savepoints.
    ----------------------------------------------------------------------------------*/
    void PermitSavepoint (tsp00_TaskId taskid);

    /*! --------------------------------------------------------------------------------
    function:    PermitSavepoint
    returns:     true if savepoint is allowed and restartrecord may be written
    ----------------------------------------------------------------------------------*/
    bool IsSavepointAllowed() const;

    
private: // methods

    /*! --------------------------------------------------------------------------------
    function:    StartCoordinator
    arguments:   TransContext [in/out]
    description: this starts a servertask as a coordinator.
    ----------------------------------------------------------------------------------*/
    void StartCoordinator (tgg00_TransContext  &Trans);
    /*! --------------------------------------------------------------------------------
    function:    BeginRegion
    arguments:   taskid [in]
    ----------------------------------------------------------------------------------*/
    void BeginRegion (tsp00_TaskId taskid);
    /*! --------------------------------------------------------------------------------
    function:    EndRegion
    arguments:   taskid [in]
    ----------------------------------------------------------------------------------*/
    void EndRegion (tsp00_TaskId taskid);
    /*! --------------------------------------------------------------------------------
    function:    SwitchWaitList
    ----------------------------------------------------------------------------------*/
    void SwitchWaitList()
    {
        m_CurrentWaitList = (m_CurrentWaitList == 0) ? 1 : 0;
    }
    /*! --------------------------------------------------------------------------------
    function:    CurrentWaitList
    returns:     current waitlist to use.
    ----------------------------------------------------------------------------------*/
    Kernel_WaitStack& CurrentWaitList()
    {
        return m_Waitlist [m_CurrentWaitList];
    }
    /*! --------------------------------------------------------------------------------
    function:    WaitListForCurrentSavepoint
    returns:     waitlist to use.
    ----------------------------------------------------------------------------------*/
    Kernel_WaitStack& WaitListForCurrentSavepoint()
    {
        return m_Waitlist [m_CurrentWaitList==0?1:0];
    }
    /*! --------------------------------------------------------------------------------
    function:    DetermineRedoStartPositions
    arguments:   CoordinatorTaskId [in]
                 Allocator         [in]
                 Reason            [in]
                 SavepointSequence [in]
                 saveptParam       [in]
    description: This writes the savepoint entry into all queues if no redo.
    ----------------------------------------------------------------------------------*/
    void DetermineRedoStartPositions
            (tsp00_TaskId            CoordinatorTaskId,
             SAPDBMem_IRawAllocator &Allocator,
             Log_SavepointReason     Reason,
             tkb00_SaveptParam      &saveptParam);
    /*! --------------------------------------------------------------------------------
    function:    CreateOpenTransFile
    arguments:   CoordinatorTrans  [in]
                 saveptParam       [out]
    description: This creates a chain file, which contains entries for all open transactions.
    ----------------------------------------------------------------------------------*/
    void CreateOpenTransFile (tgg00_TransContext &CoordinatorTrans,
                              tkb00_SaveptParam  &saveptParam);
    /*! --------------------------------------------------------------------------------
    function:    FlushRestartRecordAndLogInfoPages
    arguments:   CoordinatorTrans  [in]
    description: the loginfo pages are updated and restartrecord and loginfo is flushed.
    ----------------------------------------------------------------------------------*/
    void FlushRestartRecordAndLogInfoPages  (tgg00_TransContext  &coordinatorTrans,
                                             Log_SavepointReason  reason,
                                             tkb00_SaveptParam   &saveptParam,
                                             Converter_Version   &converterVersion,
                                             IOMan_BlockAddress  &converterRootBlockAddress,
                                             Data_PageNo         &maxDynamicPageNo,
                                             Data_PageNo         &maxStaticPageNo);
    /*! --------------------------------------------------------------------------------
    function:    FlushCompleted
    arguments:   TaskId [in] identification of the calling task
                 FBM blocks being in state free after savepoint.
    ----------------------------------------------------------------------------------*/
    void FlushCompleted (tsp00_TaskId   TaskId);
    /*! --------------------------------------------------------------------------------
    function:    ShutdownFilesystem
    arguments:   coordinatorTrans  [in]
    description: This stops all file accesses.
    ----------------------------------------------------------------------------------*/
    void ShutdownFilesystem (tgg00_TransContext &coordinatorTrans);

public: // types only public for sun compiler

    /*! --------------------------------------------------------------------------------
    declaration: Statistic
    ----------------------------------------------------------------------------------*/
    struct Statistic // PTS 1119276 UH 2002-12-10 added
    {
        SAPDB_UInt durationSec;
        SAPDB_UInt durationMicroSec;
        SAPDB_UInt pagesWritten;
        SAPDB_UInt iocount;
        Statistic()
        {
            durationSec      = 0;
            durationMicroSec = 0;
            pagesWritten     = 0;
            iocount          = 0;
        }
        Statistic(SAPDB_UInt newDurationSec,
                  SAPDB_UInt newDurationMicroSec,
                  SAPDB_UInt newPagesWritten,
                  SAPDB_UInt newIocount)
        {
            durationSec      = newDurationSec;
            durationMicroSec = newDurationMicroSec;
            pagesWritten     = newPagesWritten;
            iocount          = newIocount;
        }
        Statistic& operator += ( const Statistic &other )
        {
            this->durationSec      += other.durationSec;
            this->durationMicroSec += other.durationMicroSec;
            this->pagesWritten     += other.pagesWritten;
            this->iocount          += other.iocount;
            return *this;
        }
    };
    /*! --------------------------------------------------------------------------------
    declaration: Statistics
    ----------------------------------------------------------------------------------*/
    struct Statistics // PTS 1119276 UH 2002-12-10 added
    {
        SAPDB_UInt count;
        Statistic  last;
        Statistic  total;
        //-------------------------------------
        //-------------------------------------
        Statistics() { count = 0; }
        void Add(Statistic newValue)
        {
            ++count;
            last  =  newValue;
            total += last;
        }
        void GetAvarage(SAPDB_Int4 &pagesWritten,
                        SAPDB_Int4 &iocount)
        {
            pagesWritten = total.pagesWritten / count;
            iocount      = total.iocount      / count;
        }
    };

private: // member

    /*! --------------------------------------------------------------------------------
    variable:    m_IsActive
    description: It is true, if a savepoint is running or will be running.
    ----------------------------------------------------------------------------------*/
    bool                m_IsActive;
    /*! --------------------------------------------------------------------------------
    variable:    m_CurrentWaitList
    description: Valid values are 0 and 1.
    ----------------------------------------------------------------------------------*/
    SAPDB_UInt          m_CurrentWaitList;
    /*! --------------------------------------------------------------------------------
    variable:    m_Waitlist
    description: This is for tasks waiting for a normal savepoint be finished.
                 There are two lists. One can be filled with requests and the other one
                 can be used, if a savepoint is already running to wait for the next
                 savepoint.
    ----------------------------------------------------------------------------------*/
    Kernel_WaitStack    m_Waitlist [2];
    /*! --------------------------------------------------------------------------------
    variable:    m_WaitingDBmanager
    description: Usually this is the utility task waiting for a backup or shutdown.
    ----------------------------------------------------------------------------------*/
    tsp00_TaskId        m_WaitingDBmanager;
    /*! --------------------------------------------------------------------------------
    variable:    m_DBmanagerReason
    description: This is the reason for the db manager to trigger a savepoint.
    ----------------------------------------------------------------------------------*/
    Log_SavepointReason m_DBmanagerReason;
    /*! --------------------------------------------------------------------------------
    variable:    m_LogWriteEnabled
    description: If this is false the next savepoint will be a ReDo-Savepoint
                 which means, that no log is written, but only the
                 restartRecord is changed                   
                 This member is initialized with false and can only be set
                 to true (only one restart is possible during kernel-lifetime)
                 PTS 1115875 mb 2002-05-28
    ----------------------------------------------------------------------------------*/
    bool         m_LogWriteEnabled;
    /*! --------------------------------------------------------------------------------
    variable:    m_Statistics
    ----------------------------------------------------------------------------------*/
    Statistics   m_Statistics; // PTS 1119276 UH 2002-12-10 added
    /*! --------------------------------------------------------------------------------
    variable:    m_SavepointAllowed
    description: This flag indicates that no savepoint can be written.
    ----------------------------------------------------------------------------------*/
    bool m_SavepointAllowed; // PTS 1121281 UH 2003-03-24
};
/*! endclass: Log_Savepoint */


/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern Log_Savepoint Log_SavepointManager;

/*===========================================================================*
 *  DEFINITION OF INLINE METHODS                                              *
 *===========================================================================*/

#endif  /* LOG_SAVEPOINT_H */
