/*!*****************************************************************************

  module:       Log_SeqLogDevReader.hpp

  ------------------------------------------------------------------------------

  responsible:  MartinB
  
  special area: Archive log

  description:  This reader gives a sequential access to the log pages on a log device

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_SEQ_LOG_DEV_READER_HPP
#define LOG_SEQ_LOG_DEV_READER_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Logging/Log_Types.hpp"
#include "Logging/Log_RawDeviceIterator.hpp"
#include "Logging/Log_ClusterAddress.hpp"
#include "IOManager/IOMan_ILogIO.hpp"
#include "Logging/Log_ISequentialReader.hpp"



/*!-------------------------------------------------------------------------
class:        Log_ArchiveLogIterator
description:  This offers an iterator, to access afterimages from the archive log.
--------------------------------------------------------------------------*/
class Log_SeqLogDevReader : public Log_ISequentialReader
{

public: 
    // methods

    /*!-------------------------------------------------------------------------
    function:     Log_SeqLogDevReader()
    arguments:        
    -------------------------------------------------------------------------*/
    Log_SeqLogDevReader    (Log_RawDeviceOffset     deviceBegin,
                            Log_RawDeviceOffset     deviceEnd,
                            IOMan_ILogIO           &pageReader)
                           :
    m_ClusterAddress       (deviceBegin, 1),
    m_PageIter             (deviceBegin,  deviceBegin, deviceEnd), 
    m_PageReader           (pageReader)
    {}

    
    /*!-------------------------------------------------------------------------
    function:     SetPosition 
    description:  sets the actual page and IOSeq according to the given 
    arguments:    offset         offset, which shall be read next
    return value: ReaderStatus
    --------------------------------------------------------------------------*/
    virtual ReaderStatus  SetPosition(   Log_RawDeviceOffset      offset)
    {
        m_PageIter.SetPosition(offset);
        m_ClusterAddress.SetOffset (offset);
        if (offset.IsValid())
        {
            return invalid_Offset;
        }
        else
        {
            return reader_ok;
        }
        
    }
    
    /*!-------------------------------------------------------------------------
    function:     GetPosition 
    description:  gets the actual offset of the last page read (invalid for a tape)
    arguments:    none
    return value: the offset of the last page read
    --------------------------------------------------------------------------*/
    virtual Log_RawDeviceOffset  GetPosition() const
    {
        return m_ClusterAddress.GetOffset();
    }

protected:
    /*!-------------------------------------------------------------------------
    function:     ReadNextPage 
    description:  reads the actual page and then increments the actual position 
    arguments:    taskId        identification of the running task
                  logPage       location where the read page will be stored
    return value: ReaderStatus
    --------------------------------------------------------------------------*/
    virtual ReaderStatus  ReadNextPage(  tsp00_TaskId             taskId,
                                         Log_Page                &logPage)
    {
        if (m_PageIter.GetPosition().IsValid())
        {
            m_ClusterAddress.SetOffset ( m_PageIter.GetPosition() );
            m_PageReader.ReadLogPage   (taskId, logPage, m_ClusterAddress);

            ++m_PageIter;
            return reader_ok;
        }
        else
        {
            return invalid_Offset;
        }
    }
    
private: // member

    /*--------------------------------------------------------------------------
    declaration : m_ClusterAddress
    description:  address to read from
    --------------------------------------------------------------------------*/
    Log_ClusterAddress             m_ClusterAddress;

    /*--------------------------------------------------------------------------
    declaration : m_PageReader
    description:  handle to read pages from log device.
    --------------------------------------------------------------------------*/
    IOMan_ILogIO                  &m_PageReader;

    /*--------------------------------------------------------------------------
    declaration : m_PageIter
    description:  iterator over log page offsets
    --------------------------------------------------------------------------*/
    Log_RawDeviceIterator          m_PageIter;
};
/*!-------------------------------------------------------------------------
endclass:     Log_SeqLogDevReader
--------------------------------------------------------------------------*/

#endif // LOG_SEQ_LOG_DEV_READER_HPP
