/*!**************************************************************************

  module      : Log_Transaction.hpp

  -------------------------------------------------------------------------

  responsible : UweH

  special area: Logging
  description : This is the interface between the transaction and the logging component.
  
                Log_Transaction is a class which holds all information for a write
                transaction to do a rollback or rollforward.
                It is generated at write trans begin and destroyed at the end of
                commit or rollback.
                The transentry in the KB-locklist has pointer of this object.
                The information can made persistent in a savepoint file, which
                holds all open transactions at savepoint time.


  last changed: 2001-01-15  9:23
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef LOG_TRANSACTION_HPP
#define LOG_TRANSACTION_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#ifndef SUN
#define CHECK_COMMITTED_NEWS 1
#endif


#include "ggg00.h"
#include "ggg92.h"

#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"
#include "DataAccess/Data_Types.hpp"
#include "DataAccess/Data_ISplitSpace.hpp"
#include "Container/Container_Stack.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_UndoFile.hpp"
#include "Logging/Log_RedoFile.hpp"
#include "liveCache/LVC_NewObjectChecker.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

class SAPDBMem_IRawAllocator;
class Log_Queue;
class Log_BeforeImage;
class Log_AfterImage;
class Log_AfterImageSpace;
class Log_IAction;

#define LOG_NIL_SUBTRANS_NO                 0
#define LOG_FIRST_ENTRY_SEQUENCE            0
#define LOG_DEFAULT_SUBTRANS_STACK_CAPACITY 100

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! --------------------------------------------------------------------------------
class:       Log_Transaction
description: 
*/
class Log_Transaction
{

friend class Log_OpenTransFile;

public:

    /*! --------------------------------------------------------------------------------
    declaration: TransType
    description: This Context is used:
                 online      - in a normal user context at online time.
                 redoRead    - by the redo reader task to rebuild the transaction.
                 redoExecute - by a redo executen task to really redo or rollback the transaction.
    ----------------------------------------------------------------------------------*/
    typedef enum TransType
    {
        online,
        redoRead,
        redoExecute,
        illegalTransType
    } TransType;
    /*! --------------------------------------------------------------------------------
    declaration: TransState
    ----------------------------------------------------------------------------------*/
    typedef enum TransState
    {
        opened,
        committed,
        rollbacked,
        illegalTransState
    } TransState;
    /*! --------------------------------------------------------------------------------
    declaration: RollbackKind
    ----------------------------------------------------------------------------------*/
    typedef enum RollbackKind
    {
        endOfTrans,
        command,
        subTrans,
        illegalRollbackKind
    } RollbackKind;
    /*! --------------------------------------------------------------------------------
    declaration: LogState
    ----------------------------------------------------------------------------------*/
    typedef enum LogState
    {
        noError,
        logFull,
        logError
    } LogState;

public:

    /*! --------------------------------------------------------------------------------
    function:    Log_Transaction 
    arguments:   Trans [in]
    description: This is used to create a new write transaction.
    ----------------------------------------------------------------------------------*/
    Log_Transaction ()
    : m_pTrans        (NULL),
      m_pAllocator    (NULL),
      m_MaxSubtransNo (LOG_NIL_SUBTRANS_NO),
      m_pSubtransStack(NULL),
      m_pUndoFile     (NULL),
      m_pRedoFile     (NULL),
      m_pQueue        (NULL),
      m_TransType     (online),
      m_WasOpenInLastSavepoint (false)
    {}
    /*! --------------------------------------------------------------------------------
    function:    IsInitialized
    returns:     true, if initialized
    ----------------------------------------------------------------------------------*/
    bool IsInitialized();
    /*! --------------------------------------------------------------------------------
    function:    Initialize
    arguments:   Trans   [in]
    returns:     true, if no allocation failure.
    ----------------------------------------------------------------------------------*/
    bool Initialize (tgg00_TransContext &Trans);
    /*! --------------------------------------------------------------------------------
    function:    Delete
    description: this deallocates alle resources allocated by Initialize().
    ----------------------------------------------------------------------------------*/
    void Delete ();
    /*! --------------------------------------------------------------------------------
    function:    IsInitializedForRedoCopy
    returns:     true, if initialized
    ----------------------------------------------------------------------------------*/
    bool IsInitializedForRedoCopy();
    /*! --------------------------------------------------------------------------------
    function:    InitializeForRedoCopy
    arguments:   redoReaderTrans [in]
                 transNo         [in]
                 allocator       [in]
    returns:     true, if no allocation failure.
    ----------------------------------------------------------------------------------*/
    bool InitializeForRedoCopy (tgg00_TransContext     &redoReaderTrans,
                                const tgg91_TransNo    &transNo,
                                SAPDBMem_IRawAllocator &allocator);
    /*! --------------------------------------------------------------------------------
    function:    DeleteAfterRedoCopy
    description: this deallocates alle resources allocated by InitializeForRedoCopy().
    ----------------------------------------------------------------------------------*/
    void DeleteAfterRedoCopy ();
    /*! --------------------------------------------------------------------------------
    function:    IsInitializedForRedo
    returns:     true, if initialized
    ----------------------------------------------------------------------------------*/
    bool IsInitializedForRedo();
    /*! --------------------------------------------------------------------------------
    function:    InitializeForRedo
    arguments:   redoExecuteTrans [in]
    returns:     true, if no allocation failure.
    ----------------------------------------------------------------------------------*/
    bool InitializeForRedo (tgg00_TransContext &redoExecuteTrans);
    /*! --------------------------------------------------------------------------------
    function:    DeleteAfterRedo
    description: this deallocates alle resources allocated by InitializeForRedo().
    ----------------------------------------------------------------------------------*/
    void DeleteAfterRedo ();
    /*! --------------------------------------------------------------------------------
    function:    operator =
    description: the pointers are not assigned.
    ----------------------------------------------------------------------------------*/
    Log_Transaction& operator = (const Log_Transaction& srcTrans);
    /*! --------------------------------------------------------------------------------
    function:    WriteToTrace
    arguments:   Title [in]
    description: writes information about this transaction into the trace.
    ----------------------------------------------------------------------------------*/
    void WriteToTrace (const char *Title) const;
    /*! --------------------------------------------------------------------------------
    function:    Allocator
    returns:     Allocator
    ----------------------------------------------------------------------------------*/
    SAPDBMem_IRawAllocator& Allocator() const
    {
        SAPDBERR_ASSERT_STATE (m_pAllocator != NULL);
        return *m_pAllocator;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetRedoTransNo
    returns:     TransNo
    description: used by redo to get the open trans info.
    ----------------------------------------------------------------------------------*/
    tgg91_TransNo GetRedoTransNo() const
    {
        return m_Context.TransNo;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetRedoEotSequence
    returns:     eotSequence
    description: used by redo to synchronize a already finished transaction.
    ----------------------------------------------------------------------------------*/
    Log_IOSequenceNo GetRedoEotSequence() const
    {
        return m_Context.EOT_IOSequenceNo;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetLastRedoEntrySequence
    returns:     Log_EntrySequence
    description: used by redo reader to not copy entries, already copied
    ----------------------------------------------------------------------------------*/
    Log_EntrySequence GetLastRedoEntrySequence() const
    {
        return m_Context.LastCopiedRedoEntrySequence;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetTransState
    returns:     one of: open, committed, rollbacked
    ----------------------------------------------------------------------------------*/
    TransState GetTransState() const
    {
        if ( m_Context.TransStatePersistent <  illegalTransState )
            return TransState(m_Context.TransStatePersistent);
        else
            return illegalTransState;
    }
    /*! --------------------------------------------------------------------------------
    function:    GetTransType
    returns:     online, redoRead, redoExecute
    ----------------------------------------------------------------------------------*/
    TransType GetTransType() const
    {
        return m_TransType;
    }
    /*! --------------------------------------------------------------------------------
    function:    IsOpenInRedo
    returns:     bool true, if it is an open trans not executed by another task.
    ----------------------------------------------------------------------------------*/
    bool IsOpenInRedo() const
    {
      	// PTS 1114125 UH 2002-02-12 new
        return m_TransType == redoRead
               &&
               GetRedoEotSequence().IsInvalid();
    }
    /*! chapter: Subtrans Handling */
    /*! --------------------------------------------------------------------------------
    function:    BeginSubtrans
    description: This begins a sub transaction.
                 A sub transaction can be separately rollbacked.
                 trSubtransNo becomes a new Subtransno.
    ----------------------------------------------------------------------------------*/
    void BeginSubtrans();
    /*! --------------------------------------------------------------------------------
    function:    EndSubtrans
    returns:     false, if subtrans begin is missing.
    description: This ends a sub transaction.
    ----------------------------------------------------------------------------------*/
    bool EndSubtrans();
    /*! endchapter: Subtrans Handling */
    /*! chapter: EndOfTrans Handling */
    /*! --------------------------------------------------------------------------------
    function:    Commit
    description: This TransState becomes committed. And must be open before.
    ----------------------------------------------------------------------------------*/
    void Commit();
    /*! --------------------------------------------------------------------------------
    declaration: RedoResult
    ----------------------------------------------------------------------------------*/
    typedef enum RedoResult
    {
        redoOk,
        redoReadError,
        redoWrongEntry,
        redoExecuteError,
        redoWriteError
    } RedoResult;
    /*! --------------------------------------------------------------------------------
    function:    Redo
    description: The transaction is redone by a redo task.
    ----------------------------------------------------------------------------------*/
    RedoResult Redo();
    /*! --------------------------------------------------------------------------------
    declaration: RollbackResult
    ----------------------------------------------------------------------------------*/
    typedef enum RollbackResult
    {
        undoOk,
        undoNoOpenSubtrans,
        undoReadError,
        undoWrongEntry,
        undoExecuteError,
        undoWriteError,
        undoLogFull
    } RollbackResult;
    /*! --------------------------------------------------------------------------------
    function:    Rollback
    returns:     RollbackResult
    arguments:   rollbackKind     [in] - endOfTrans,command,subTrans
                 SubtransNo       [in]
                 undoStopSequence [in] - only used during redo
    description: The TransState becomes rollbacked and must be open before.
                 If SubtransNo is LOG_NIL_SUBTRANS_NO, then the complete
                 transaction is rollbacked.
    ----------------------------------------------------------------------------------*/
    RollbackResult Rollback (RollbackKind      rollbackKind,
                             tgg00_SubtransNo  SubtransNo,
                             Log_EntrySequence undoStopSequence = Log_EntrySequence()); // PTS 1113230 UH 2001-12-20
    /*! endchapter: EndOfTrans Handling */
    /*! chapter: Logentry Handling */
    /*! --------------------------------------------------------------------------------
    function:    WriteUndoEntry
    arguments:   BeforeImage    [in]
                 Action         [in]
                 isOK           [out]
                 pageno         [out]
                 offset         [out]
    description: The given BeforeImage is stored for rollback purpose.
    ----------------------------------------------------------------------------------*/
    Log_UndoFile::WriteBeforeImageResult
         WriteUndoEntry (Log_BeforeImage   &BeforeImage, // PTS 1121659 UH 2003-04-30
                         const Log_IAction &Action,
                         Data_PageNo       &pageno,
                         Data_PageOffset   &offset);
    /*! --------------------------------------------------------------------------------
    function:    WriteRedoEntry
    arguments:   AfterImage    [in]
                 Action        [in]
    returns:     LogState 
    description: The given AfterImage is stored for redo purpose.
    ----------------------------------------------------------------------------------*/
    LogState WriteRedoEntry (Log_AfterImage    &AfterImage,
                             const Log_IAction &Action);
    /*! --------------------------------------------------------------------------------
    function:    ReserveSpaceForRedoEntryCopy
    description: The space is reserved and must be released after copy.

                 If the file does not exist, it is created.
    arguments:   Length                [in]
                 MinimalLength         [in]
                 EntrySequence         [in]
                 IOSequence            [in]
    returns:     Data_SplitSpaceWriter [out];
    ----------------------------------------------------------------------------------*/
    Data_SplitSpaceWriter ReserveSpaceForRedoEntryCopy
        (Data_SplitRecordLength Length,
         Data_RecordLength      MinimalLength,
         Log_EntrySequence      EntrySequence,
         Log_IOSequenceNo       IOSequence);
    /*! --------------------------------------------------------------------------------
    function:    ReleaseSpaceAfterRedoEntryCopy
    arguments:   isOK
    description: All resources for copying are released.
                 if isOK is false, all resources are released unchanged.
    ----------------------------------------------------------------------------------*/
    void ReleaseSpaceAfterRedoEntryCopy(bool isOK);
    /*! --------------------------------------------------------------------------------
    function:    RedoPartialRollback
    arguments:   UndoStopSequence [in]
                 RedoStopSequence [in]
    description: The redo file is truncated and if necessary the UndoFile is partially
                 rollbacked.
    ----------------------------------------------------------------------------------*/
    void RedoPartialRollback (Log_EntrySequence UndoStopSequence,
                              Log_EntrySequence RedoStopSequence);
    /*! --------------------------------------------------------------------------------
    function:    SetEOTInRedo
    arguments:   eotState    [in]
                 eotSequence [in]
    description: if during redo a transactions end was found. it set here,
                 so a following savepoint can use this on redo/redo.
    ----------------------------------------------------------------------------------*/
    void SetEOTInRedo (TransState        eotState,
                       Log_EntrySequence eotSequence);
    /*! --------------------------------------------------------------------------------
    function:    WasOpenInLastSavepoint
    returns:     true, if trans was recreated from last savepoint.
    description: if during redo a transactions end was found. it set here,
                 so a following savepoint can use this on redo/redo.
    ----------------------------------------------------------------------------------*/
    bool WasOpenInLastSavepoint()
    {
        return m_WasOpenInLastSavepoint;
    }
    /*! endchapter: Logentry Handling */
    /*! chapter: Savepoint Handling */
    /*! --------------------------------------------------------------------------------
    function:    IsSavepointRelevant
    returns:     bool
    description: true, if this transaction needs to be written during savepoint
    ----------------------------------------------------------------------------------*/
    bool IsSavepointRelevant();
    /*! --------------------------------------------------------------------------------
    function:    IsRedoRelevant
    returns:     bool
    description: true, if this transaction entry is needed by redo.
                 This filters the open transactions which are collected after the end
                 of the one transaction and before the open of the new transaction.
    ----------------------------------------------------------------------------------*/
    bool IsRedoRelevant();
    /*! --------------------------------------------------------------------------------
    function:    GetPersistentLength
    returns:     SAPDB_UInt
    description: used for writing the savepoint file
    ----------------------------------------------------------------------------------*/
    static SAPDB_UInt GetPersistentLength()
    {
        return sizeof (Persistent);
    }
    /*! --------------------------------------------------------------------------------
    function:    WritePersistentFormat
    arguments:   Space  [in/out]
                 Length [in]
    description: used for writing the savepoint file
    ----------------------------------------------------------------------------------*/
    void WritePersistentFormat(SAPDB_Byte* Space,
                               SAPDB_UInt  Length);
    /*! --------------------------------------------------------------------------------
    function:    ReadPersistentFormat
    arguments:   Space  [in]
                 Length [in]
    description: Used for recreating a transaction out of a savepoint file.
    ----------------------------------------------------------------------------------*/
    void ReadPersistentFormat(const SAPDB_Byte* Space,
                              SAPDB_UInt        Length);
    /*! endchapter: Savepoint Handling */

    /*! --------------------------------------------------------------------------------
    function:    GetDumpInfo
    arguments:   this*   [in]
                 dataptr [out]
                 length  [out]
    description: this can be used by the locklist to dump with the transentry.
    ----------------------------------------------------------------------------------*/
    static void GetDumpInfo (Log_Transaction *pTrans,
                             void*           &pData,
                             SAPDB_UInt4     &dataLength)
    {
        if ( pTrans == NULL )
        {
            pData      = NULL;
            dataLength = 0;
            return;
        }
        pData      = &(pTrans->m_Context);
        dataLength = sizeof(pTrans->m_Context);
    }
    /*! --------------------------------------------------------------------------------
    function:    UndoEntriesExist
    return:      bool
    description: returns true, if any undo was written by the transaction.
    ----------------------------------------------------------------------------------*/
    bool UndoEntriesExist()
    {
        return m_Context.LastUndoEntrySequence.IsValid();
    }
    /*! --------------------------------------------------------------------------------
    function:    RedoEntriesExist
    return:      bool
    description: returns true, if any redo was written by the transaction.
    ----------------------------------------------------------------------------------*/
    bool RedoEntriesExist()
    {
        return m_Context.LastRedoEntrySequence.IsValid();
    }
    /*! --------------------------------------------------------------------------------
    function:    EntriesToRedoExist
    return:      bool
    description: returns true, if any redo entry was copied for redo by the transaction.
    ----------------------------------------------------------------------------------*/
    bool EntriesToRedoExist()
    {
        return m_Context.LastCopiedRedoEntrySequence.IsValid();
    }
    /*! --------------------------------------------------------------------------------
    function:    UpdateUndoRedoInfoForRestart
    ----------------------------------------------------------------------------------*/
    void UpdateUndoRedoInfoForRestart();
    /*! --------------------------------------------------------------------------------
    function:    GetInfo
    arguments:   undoroot         [out]
                 undolast         [out]
                 undopagecount    [out]
                 lastundosequence [out]
    ----------------------------------------------------------------------------------*/
    void GetInfo ( Data_PageNo      &undoroot,
                   Data_PageNo      &undolast,
                   SAPDB_UInt       &undopagecount,
                   Log_IOSequenceNo &lastundosequence )
    {
        // PTS 1115468 UH 2002-05-13 new
        // MUST BE CALLED IN TRANS REGION FOR THIS TRANSACTION
        undoroot         = m_Context.UndoFileRootPageNo;
        undolast         = m_Context.UndoFileLastPageNo;
        lastundosequence = m_Context.LastUndoEntrySequence;
        if ( m_pUndoFile != NULL
             &&
             m_pUndoFile->IsCreated() )
            undopagecount = m_pUndoFile->GetPageCount();
        else
            undopagecount = 0;
    }
    #ifdef CHECK_COMMITTED_NEWS
    /*! --------------------------------------------------------------------------------
    function:    AddReservedObject
    arguments:   oid [in]
    ----------------------------------------------------------------------------------*/
    void AddReservedObject (const tgg92_KernelOid& oid)
    {
        m_ReservedObjects.Insert(oid);
    }
    /*! --------------------------------------------------------------------------------
    function:    RemoveReservedObject
    arguments:   oid [in]
    ----------------------------------------------------------------------------------*/
    void RemoveReservedObject (const tgg92_KernelOid& oid)
    {
        m_ReservedObjects.Delete(oid);
    }
    #endif

private: // methods

    /*! --------------------------------------------------------------------------------
    function:    SetTransState
    arguments:   State [in]
    ----------------------------------------------------------------------------------*/
    void SetTransState (TransState State)
    {
        m_Context.TransStatePersistent = State;
    }
    /*! --------------------------------------------------------------------------------
    function:    ExecuteUndoAction
    arguments:   BeforeImage [in]
                 isOK        [out]
    ----------------------------------------------------------------------------------*/
    void ExecuteUndoAction (Log_BeforeImage        &BeforeImage,
                            Data_SplitSpaceReader  &Reader,
                            Log_UndoFile::Iterator &iter,
                            bool                   &isOK);
    /*! --------------------------------------------------------------------------------
    function:    ExecuteRedoAction
    arguments:   AfterImage [in]
                 isOK       [out]
    ----------------------------------------------------------------------------------*/
    void ExecuteRedoAction (Log_AfterImage         &AfterImage,
                            Log_IOSequenceNo        IOSequence,
                            Data_SplitSpaceReader  &Reader,
                            Log_RedoFile::Iterator &iter,
                            bool                   &isOK);
    /*! --------------------------------------------------------------------------------
    function:    EndOfTransHandling
    ----------------------------------------------------------------------------------*/
    void EndOfTransHandling (TransState State);
    /*! --------------------------------------------------------------------------------
    function:    PopUntilSubtransFound
    arguments:   SubtransNo       [in]
                 UndoStopSequence [out]
                 RedoStopSequence [out]
    returns:     the new valid SubtransNo
    ----------------------------------------------------------------------------------*/
    tgg00_SubtransNo PopUntilSubtransFound (tgg00_SubtransNo   SubtransNo,
                                            Log_EntrySequence &UndoStopSequence,
                                            Log_EntrySequence &RedoStopSequence);
    /*! --------------------------------------------------------------------------------
    function:    WriteEndOfTransEntry
    returns:     false, on error
    ----------------------------------------------------------------------------------*/
    bool WriteEndOfTransEntry (TransState State);
    /*! --------------------------------------------------------------------------------
    function:    WriteCommitEntry
    arguments:   UndoStopSequence [in]
                 RedoStopSequence [in]
    returns:     LogState
    ----------------------------------------------------------------------------------*/
    LogState WritePartialRollbackEntry (Log_EntrySequence  UndoStopSequence,
                                        Log_EntrySequence  RedoStopSequence);
    /*! --------------------------------------------------------------------------------
    function:    CreateUndoFile
    ----------------------------------------------------------------------------------*/
    bool CreateUndoFile();
    /*! --------------------------------------------------------------------------------
    function:    DropUndoFile
    ----------------------------------------------------------------------------------*/
    void DropUndoFile();
    /*! --------------------------------------------------------------------------------
    function:    CreateRedoFile
    ----------------------------------------------------------------------------------*/
    void CreateRedoFile(tgg00_TransContext &redoReaderTrans);
    /*! --------------------------------------------------------------------------------
    function:    DropRedoFile
    ----------------------------------------------------------------------------------*/
    void DropRedoFile();
    /*! --------------------------------------------------------------------------------
    function:    GiveUndoFileToHistory
    ----------------------------------------------------------------------------------*/
    void GiveUndoFileToHistory();
    /*! --------------------------------------------------------------------------------
    function:    UndoRedoContinue
    ----------------------------------------------------------------------------------*/
    void UndoRedoContinue (Log_IImage             &image,
                           Data_IBreakable        &iter,
                           Data_SplitSpaceReader  &reader,
                           bool                   &isOK);
    /*! --------------------------------------------------------------------------------
    function:    UndoRedoContinue
    ----------------------------------------------------------------------------------*/
    void UndoRedoAction (Log_IImage             &image,
                         Log_IAction            &action,
                         Data_IBreakable        &iter,
                         Data_SplitSpaceReader  &reader,
                         bool                   &isOk);
protected: // types

    /*! --------------------------------------------------------------------------------
    declaration: SubtransStackElem
    ----------------------------------------------------------------------------------*/
    typedef struct SubtransStackElem
    {
        tgg00_SubtransNo  SubtransNo;       // 4 byte
        Log_EntrySequence UndoStopSequence; // 4 byte
        Log_EntrySequence RedoStopSequence; // 4 byte
    } SubtransStackElem;
    /*! --------------------------------------------------------------------------------
    declaration: SubtransStack
    ----------------------------------------------------------------------------------*/
    typedef Container_Stack<SubtransStackElem> SubtransStack;
    /* -------------------------------------------------------------------------------- */
    struct Persistent
    {
        tgg91_TransNo     TransNo;                      // 6  byte
        SAPDB_Byte        TransStatePersistent;         // 1  byte
        SAPDB_Bool1       BeforeImagesNeededByHistory;  // 1  byte
        Log_EntrySequence LastUndoEntrySequence;        // 4  byte
        Log_EntrySequence LastRedoEntrySequence;        // 4  byte
        Log_EntrySequence LastCopiedRedoEntrySequence;  // 4  byte
        SAPDB_Int4        UndoFileRootPageNo;           // 4  byte
        SAPDB_Int4        UndoFileLastPageNo;           // 4  byte
        SAPDB_Int4        RedoFileRootPageNo;           // 4  byte
        SAPDB_Int4        RedoFileLastPageNo;           // 4  byte
        Log_IOSequenceNo  EOT_IOSequenceNo;             // 4  bytes
                                                        // 40
        Persistent()
        {
            init();
        }
        
        void init()
        {
            TransNo.gg90SetNil();
            TransStatePersistent        = 0;
            BeforeImagesNeededByHistory = 0;
            LastUndoEntrySequence.Invalidate();
            LastRedoEntrySequence.Invalidate();
            LastCopiedRedoEntrySequence.Invalidate();
            UndoFileRootPageNo    = SAPDB_MAX_INT4;
            UndoFileLastPageNo    = SAPDB_MAX_INT4;
            RedoFileRootPageNo    = SAPDB_MAX_INT4;
            RedoFileLastPageNo    = SAPDB_MAX_INT4;
            EOT_IOSequenceNo.Invalidate();
        }
    };

private: // member

    /*! --------------------------------------------------------------------------------
    variable:    m_pTrans
    ----------------------------------------------------------------------------------*/
    tgg00_TransContext*    m_pTrans;
    /*! --------------------------------------------------------------------------------
    variable:    m_pAllocator
    ----------------------------------------------------------------------------------*/
    SAPDBMem_IRawAllocator* m_pAllocator;
    /*! --------------------------------------------------------------------------------
    variable:    m_Context
    ----------------------------------------------------------------------------------*/
    Persistent              m_Context;
    /*! --------------------------------------------------------------------------------
    variable:    m_MaxSubtransNo
    description: Subtranshandling
    ----------------------------------------------------------------------------------*/
    tgg00_SubtransNo        m_MaxSubtransNo;
    /*! --------------------------------------------------------------------------------
    variable:    m_pSubtransStack
    description: Subtranshandling
    ----------------------------------------------------------------------------------*/
    SubtransStack          *m_pSubtransStack;
    /*! --------------------------------------------------------------------------------
    variable:    m_pUndofile
    ----------------------------------------------------------------------------------*/
    Log_UndoFile           *m_pUndoFile;
    /*! --------------------------------------------------------------------------------
    variable:    m_pRedofile
    ----------------------------------------------------------------------------------*/
    Log_RedoFile           *m_pRedoFile;
    /*! --------------------------------------------------------------------------------
    variable:    m_pQueue
    ----------------------------------------------------------------------------------*/
    Log_Queue              *m_pQueue;
    /*! --------------------------------------------------------------------------------
    variable:   m_TransType
    ----------------------------------------------------------------------------------*/
    TransType               m_TransType;
    /*! --------------------------------------------------------------------------------
    variable:   m_WasOpenInLastSavepoint
    ----------------------------------------------------------------------------------*/
    bool                    m_WasOpenInLastSavepoint;
    /*! --------------------------------------------------------------------------------
    variable:    m_PersistentAccess
    description: This synchronises parts for WritePersistent.
    ----------------------------------------------------------------------------------*/
    mutable RTESync_Spinlock m_PersistentAccess;
    
    #ifdef CHECK_COMMITTED_NEWS
    /*! --------------------------------------------------------------------------------
    variable:    m_ReservedObjects
    description: A list which holds all objects which are reserved, but not filled.
    ----------------------------------------------------------------------------------*/
    typedef LVC_NewObjectChecker<tgg92_KernelOid> ReservedObjects;
    ReservedObjects m_ReservedObjects;
    #endif
};
/*! endclass: Log_Transaction */


#endif  /* LOG_TRANSACTION_HPP */
