/*!**************************************************************************

  module      : Log_Types.hpp
  special area: Logging
  responsible : UweH
  created     : 2000-09-25  12:00
  last changed: 2000-10-27  12:00
  copyright:    Copyright by SAP AG, 2000
  description : Contains types used from several modules in this component.



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/


#ifndef LOG_TYPES_HPP
#define LOG_TYPES_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/SAPDB_EnhancedTypes.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDBCommon/Diagnostics/SAPDBDiag_CheckTopic.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/
/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

extern SAPDBTrace_Topic     LogVolume_Trace;
extern SAPDBTrace_Topic     LogHistory_Trace;
extern SAPDBTrace_Topic     LogTrans_Trace;
extern SAPDBTrace_Topic     LogAction_Trace;
extern SAPDBTrace_Topic     LogPage_Trace;

extern SAPDBDiag_CheckTopic LogVolume_Check;
extern SAPDBDiag_CheckTopic LogHistory_Check;
extern SAPDBDiag_CheckTopic LogTrans_Check;
extern SAPDBDiag_CheckTopic LogAction_Check;
extern SAPDBDiag_CheckTopic LogPage_Check;

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!-----------------------------------------------------------------------------
declaration : Log_DeviceBlockCount
description:  any number of blocks on a raw log device
------------------------------------------------------------------------------*/
typedef SAPDB_UInt4 Log_DeviceBlockCount;

/*!-----------------------------------------------------------------------------
declaration : Log_RawDeviceOffset
description:  block offset on raw log device
------------------------------------------------------------------------------*/
typedef SAPDB_Invalidatable<Log_DeviceBlockCount, SAPDB_MAX_UINT4> 
        Log_RawDeviceOffset;

/*!-----------------------------------------------------------------------------
declaration : Log_IOSequenceNo
description:  i/o sequence number of log pages
------------------------------------------------------------------------------*/
typedef SAPDB_Invalidatable<SAPDB_UInt4, SAPDB_MAX_UINT4> 
        Log_IOSequenceNo;

/*!-----------------------------------------------------------------------------
declaration : Log_QueueID
description:  log queue identifier
------------------------------------------------------------------------------*/
typedef SAPDB_Invalidatable<SAPDB_UInt2, SAPDB_MAX_UINT2> 
        Log_QueueID;

/*! ----------------------------------------------------------------------------
declaration: Log_EntrySequence
*/
typedef SAPDB_Invalidatable<SAPDB_UInt4, SAPDB_MAX_UINT4> 
        Log_EntrySequence;

/*! ----------------------------------------------------------------------------
declaration: Log_SavepointSequence
*/
typedef SAPDB_Invalidatable<SAPDB_UInt4, SAPDB_MAX_UINT4> 
        Log_SavepointSequence;

/*!-------------------------------------------------------------------------
declaration:  FlushMode
description:  indicates whether partially filled log pages may be flushed
              more times onto the same block on log device, or not
--------------------------------------------------------------------------*/
enum Log_FlushMode
{
    Log_DeviceMinimizeSpace,     // only works in single queue mode
    Log_DeviceMaximizeSafety,
    Log_FlushMode_DO_NOT_USE
};
        
/*!-------------------------------------------------------------------------
declaration:  Log_FlushModeStrings
--------------------------------------------------------------------------*/
extern const char* Log_FlushModeStrings[Log_FlushMode_DO_NOT_USE];

/*!-----------------------------------------------------------------------------
declaration : Log_DeviceState
description:  This describes what was written onto the logdevice and can be found.
              
              If a logdevice is just being cleared the state is:
                  Log_DeviceStateClearing
              The next restart has to clear again.
              
              If a logdevice is new or was cleared it has the state:
                  Log_DeviceStateCleared
              You can find nothing on it.
              
              If a logdevice was written with a savepoint it has the state:
                  Log_DeviceStateHistoryLost
              The logdevice can be normally restarted.
              
              If a complete data backup was made, the logdevice state switches
              to:
                  Log_DeviceStateOkay
                  
              Only with the state Log_DeviceStateOkay a log backup is allowed.
------------------------------------------------------------------------------*/
typedef enum
{
    Log_DeviceStateOkay                    = 0, // after SAVE DATA => SAVE LOG is now allowed
    Log_DeviceStateHistoryLost             = 1, // after writing a savepoint
    Log_DeviceStateCleared                 = 2, // after CLEAR LOG
    Log_DeviceStateClearing                = 3, // during CLEAR LOG
    Log_DeviceStatePartiallyCleared        = 4, // after PARTIALLY CLEAR LOG (restore log until) PTS 1111525 mb 2002-01-29 new
    Log_DeviceStatePartiallyClearing       = 5, // during PARTIALLY CLEAR LOG (restore log until) PTS 1111525 mb 2002-01-29 new
    
    Log_DeviceState_DO_NOT_USE             = 6
}
    Log_DeviceState;        
        
/*!-------------------------------------------------------------------------
declaration:  Log_DeviceStateStrings
--------------------------------------------------------------------------*/
extern const char* Log_DeviceStateStrings[Log_DeviceState_DO_NOT_USE];

/*!-------------------------------------------------------------------------
declaration:  Log_DeviceCheckMode
description:  ArchivLogConfig I   => Log_CheckQueueSequences     (only one log and one queue)
              ArchivLogConfig II  => Log_CheckIOSequences        (one log several queues)
              ArchivLogConfig III => Log_CheckAscendingSequences (several logs and several queues)
--------------------------------------------------------------------------*/
enum Log_DeviceCheckMode
{
    Log_CheckQueueSequences,
    Log_CheckIOSequences,
    Log_CheckAscendingSequences,
    Log_DeviceCheckMode_DO_NOT_USE
};

/*!-------------------------------------------------------------------------
declaration:  Log_DeviceCheckModeStrings
--------------------------------------------------------------------------*/
extern const char* Log_DeviceCheckModeStrings[Log_DeviceCheckMode_DO_NOT_USE];


/*!-----------------------------------------------------------------------------
declaration : Log_ErrorState
description:  constants for different log error reasons

             ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             ++ ATTENTION: YOU MUST KEEP THE ORDER OF THE ENTRIES    ++
             ++            AND YOU MUST UPDATE Log_ErrorStateStrings ++
             ++            in Log_Types.cpp too                      ++
             ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             
------------------------------------------------------------------------------*/
typedef enum {
    Log_Ok = 0,
    Log_LogAndDataIncompatible = 1,
    Log_RawAllocationFailed = 2,
    Log_PageAllocationFailed = 3,
    Log_LogDeviceTooSmall = 4,
    Log_InconsistentLogDeviceIOSequence = 5,
    Log_LogSaveDistanceTooLarge = 6,
    Log_LogSizeHasDecreased = 7,
    Log_InconsistentLogInfoPage = 8,
    Log_SavepointPositionNotFound = 9,
    Log_SaveForAbortedClearLog = 10,
    
    Log_ErrorState_DO_NOT_USE = 11
} 
    Log_ErrorState;

/*!-------------------------------------------------------------------------
declaration:  Log_ActionTypeStrings
description:  for tracing the action types
--------------------------------------------------------------------------*/
extern const char* Log_ErrorStateStrings[Log_ErrorState_DO_NOT_USE];

/*! ----------------------------------------------------------------------------
declaration: Log_ActionType
description: All actions which can be un- and normally re-done are listed here.

             This is used to recognize the persistent entries.

             ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             ++ ATTENTION: YOU MUST KEEP THE ORDER OF THE ENTRIES    ++
             ++            AND YOU MUST UPDATE Log_ActionTypeStrings ++
             ++            in Log_Types.cpp too                      ++
             ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/
typedef enum
{
    Log_NoOp = 0,

    /* transaction related actions */
    Log_Commit          = 1,
    Log_Rollback        = 2,
    Log_PartialRollback = 3,

    /* actions related to OMS */
    Log_NewObject        = 4,
    Log_InsertObject     = 5,
    Log_UpdateObject     = 6,
    Log_DeleteObject     = 7,
    Log_LockObject       = 8,
    Log_CreateObjectFile = 9,
    Log_DropObjectFile   = 10,

    /* actions related to SQL */
    Log_InsertRecord = 11,
    Log_DeleteRecord = 12,
    Log_UpdateRecord = 13,
    //
    Log_ModifyTable = 14,
    //
    Log_ChangeFileType   = 15,
    Log_CreateFile       = 16,
    Log_CreateIndex      = 17,
    Log_CreateTable      = 18,
    Log_RenameFile       = 19,
    Log_SuccFileVersion  = 20,
    Log_UpdFileStatistic = 21,
    //
    Log_CopyBlob     = 22,
    Log_CopySameBlob = 23,
    Log_ExpandBlob   = 24,
    Log_TruncateBlob = 25,
    Log_UpdateBlob   = 26,
    //
    Log_RefreshSequence  = 27,
    Log_RefreshSurrogate = 28,

    /* actions related to restart */
    Log_FinishOpenTrans = 29,
    Log_SavepointEntry  = 30,
    //
    /* additional actions */
    Log_DeleteAllObjects = 31,
    /* load/unload actions */
    Log_ReorganizeFile = 32,
    Log_ReloadFile     = 33,
    Log_UnloadFile     = 34,
    Log_SetFileState   = 35,
    //
    Log_ActionType_DO_NOT_USE = 36
}
    Log_ActionType;

/*! endDeclaration: */

/*!-------------------------------------------------------------------------
declaration:  Log_ActionTypeStrings
description:  for tracing the action types
--------------------------------------------------------------------------*/
extern const char* Log_ActionTypeStrings[Log_ActionType_DO_NOT_USE];

/*!-------------------------------------------------------------------------
declaration: Log_SavepointReason
description: This differs the reasons to trigger a savepoint.

             ++++ CHANGE THE Log_SavepointReasonString variable too. ++++
--------------------------------------------------------------------------*/
typedef enum Log_SavepointReason
{
    Log_SVPReasonUnknown           = 0,
    Log_SVPReasonSaveData          = 1,
    Log_SVPReasonSavePages         = 2,
    Log_SVPReasonRestart           = 3,
    Log_SVPReasonShutdown          = 4,
    Log_SVPReasonDistance          = 5,
    Log_SVPReasonTime              = 6,
    Log_SVPReasonUserRequested     = 7,
    Log_SVPReasonTransRequested    = 8,  // not used
    Log_SVPReasonFBM               = 9,
    Log_SVPReasonDataWriter        = 10,
    Log_SVPReasonBDLockList        = 11, // not used
    Log_SVPReasonEndReadOnly       = 12,
    Log_SVPReasonDataCache         = 13, // not used
    Log_SVPReasonLogWriter         = 14,
    Log_SVPReasonConverter         = 15,
    Log_SVPReasonEndLoad           = 16,
    Log_SVPReasonCreateIndex       = 17,
    Log_SVPReasonCheckTable        = 18,
    Log_SVPReasonCatalogMigration  = 19,
    Log_SavepointReason_DO_NOT_USE = 20
} Log_SavepointReason;
/*! enddeclaration */

/*!-------------------------------------------------------------------------
declaration:  Log_SavepointReasonStrings
description:  for tracing the action types
--------------------------------------------------------------------------*/
extern const char* Log_SavepointReasonStrings[Log_SavepointReason_DO_NOT_USE];

/*!-------------------------------------------------------------------------
declaration:  LOG_INFOPAGEOFFSET
description:  position of log info page on raw log device
--------------------------------------------------------------------------*/
#define LOG_INFOPAGEOFFSET          (Log_RawDeviceOffset)0
/*!-------------------------------------------------------------------------
declaration:  LOG_INFOBLOCKCOUNT
description:  number of log info blocks (pages) on a raw log device
--------------------------------------------------------------------------*/
#define LOG_INFOBLOCKCOUNT          (Log_DeviceBlockCount)1
                                         
#endif  /* LOG_TYPES_HPP */


