/*!*****************************************************************************

  module:       Log_Writer.cpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  author:       TillL

  special area: Archive log

  description:  see .hpp

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/

#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "KernelCommon/Kernel_DateTime.hpp"
#include "Logging/Log_Writer.hpp"
#include "Logging/Log_ArchiveLog.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "Logging/Log_Messages.hpp"
#include "hkb38_1.h"
#include "hkb57_1.h" // k57restartrec
#include "hgg01.h" //g01restart_local
#include "heo52.h"   // vclock


// -----------------------------------------------------------------------------
// suspend reasons
// -----------------------------------------------------------------------------
#define LOG_DEVICE_FULL_SUSPEND    241
// PTS 1114053 UH 2002-10-22 new
#define WAIT_FOR_LOGWRITER_SUSPEND 247
// PTS 1114053 UH 2002-10-22 new
#define LOGWRITER_USER_SUSPEND     248

/*******************************************************************************

   class: Log_Writer

*******************************************************************************/

/* global constant of class Log_Writer */
/*--------------------------------------------------------------------------
declaration : LOGWRITER_MIN_DIST_TO_SAVEPOINT 
description:  describes the minimal distance in percent of the log_size
              from the writer-position to the last written savepoint
--------------------------------------------------------------------------*/    
#define LOGWRITER_MIN_DIST_TO_SAVEPOINT  33

/*--------------------------------------------------------------------------
declaration : LOGWRITER_LOG_FULL_WARNING_LIMIT
description:  describes the limit from which a warning message is written into
              the knldiag in percentage of used log-pages
--------------------------------------------------------------------------*/    
#define LOGWRITER_LOG_FULL_WARNING_LIMIT 95

/// 5000 iosequences must be written before a savepoint must be triggered.
#define MIN_IOSEQUENCES_BEFORE_SAVEPOINT 5000

void Log_Writer::Run()
{
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::Run", LogVolume_Trace, 5);
    
    m_Active = true;

    // ----------------------------------------------------------
    // check if writer has been initialized
    // ----------------------------------------------------------
    if (m_Allocated == false)
    {
        // writer not yet initialized:
        m_WriterTaskSync.Suspend(LOG_WRITER_NOT_INITIALIZED);
        // in the meantime, the writer might have been deactivated:
        if (m_Active == false)
            return;
        // writer may still not be initialized:
        if (m_Allocated == false)
            return;
    }

    // ----------------------------------------------------------
    // set start position for queue loop:
    // ----------------------------------------------------------
    QueueIterator        queueiter         = m_LogQueues.Anchor();
    Log_DeviceBlockCount pagesFreeOnDevice = CalculateFreePages();
    
    CheckAndPrintLogUsageMessage(pagesFreeOnDevice);

    if ( EnoughPagesFreeForUserTaskWrite(pagesFreeOnDevice) )                   // PTS 1115488 mb 2002-05-03
    {
        // enable write for everybody into the Log_Queue
        SetLogFullQueueState(FALSE);
    }

    while (m_Active == true)
    {
        // ----------------------------------------------------------
        // check for log full and suspend if necessary
        // ----------------------------------------------------------
        pagesFreeOnDevice = CalculateFreePages();

        CheckAndPrintLogUsageMessage(pagesFreeOnDevice);

        while ( pagesFreeOnDevice == 0 )
        {
            // absolutely no more write is possible to this log device.
            
            if ( EnableAddedLogDeviceIsPossible() )
                EnableAddedLogDevice();
                
            if ( pagesFreeOnDevice == 0 )
            {
                m_IsSuspended = true;
                m_WriterTaskSync.Suspend(LOG_DEVICE_FULL_SUSPEND);
                m_IsSuspended = false;
                if (m_Active == false)
                    return;
            }
            pagesFreeOnDevice = CalculateFreePages();
    
            CheckAndPrintLogUsageMessage(pagesFreeOnDevice);
        }

        // ----------------------------------------------------------
        // get pages for i/o:
        // ----------------------------------------------------------

        SAPDB_Bool firstflushedagain = false; // indicates whether last written page on device must be overwritten
        #ifdef SAPDB_SLOW
        if ( m_LogQueues.GetSize() > 1 )
            SAPDBERR_ASSERT_STATE(firstflushedagain == false);
        #endif

        QueueIterator fullloop = queueiter;
        
        pagesFreeOnDevice = CalculateFreePages();
    
        CheckAndPrintLogUsageMessage(pagesFreeOnDevice);

        if ( !EnoughPagesFreeForUserTaskWrite(pagesFreeOnDevice) )
        {
            if ( EnableAddedLogDeviceIsPossible() )
            {
                EnableAddedLogDevice();
                pagesFreeOnDevice = CalculateFreePages();
                CheckAndPrintLogUsageMessage(pagesFreeOnDevice);
            }
            else
            {
                if (m_lastPercentageUsed != 100)
                {
                    m_lastPercentageUsed = 100;
                    RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_IS_FULL));
                }
            }
        }

        // ----------------------------------------------------------
        // loop associated log queues; accumulate log pages for i/o
        // ----------------------------------------------------------
        
        if ( ! EnoughPagesFreeForUserTaskWrite(pagesFreeOnDevice) )                   // PTS 1115488 mb 2002-05-03
        {
            // only commit,rollback and savepoint are allowed due to log full
            SetLogFullQueueState(TRUE);
        }
         
		m_RequestPages.Clear();
        m_QueueRequests.Clear();

        do                                                        // PTS 1111918 mb 2001-09-27
        {
            Log_Queue& queue     = **queueiter;
            SAPDB_UInt pagecount = 0;
            
            queue.GetOutput (m_WriterTaskSync.GetTaskID(), 
                             pagesFreeOnDevice - m_RequestPages.GetPageCount(), 
                             m_RequestPages,
                             pagecount,
                             firstflushedagain);
            
            if (pagecount > 0)
                m_QueueRequests.InsertEnd (QueueRequest(&queue,pagecount));

            ++queueiter;
        }
        while ( ! m_RequestPages.IsFull()
                &&
                (queueiter != fullloop)
                &&
                (pagesFreeOnDevice - m_RequestPages.GetPageCount() > 0) );
        
        // ----------------------------------------------------------
        // check whether there is output available:
        // ----------------------------------------------------------

        if (m_RequestPages.IsEmpty())
        {
            if (pagesFreeOnDevice > 0)
            {
                // no output available but free pages on the log
                m_WriterTaskSync.DoSuspend(LOG_QUEUE_EMPTY);
                // in the meantime, the writer might have been deactivated:
                if (m_Active == false)
                    return;
            }
        }
        else
        {
            Log_IOSequenceNo        ioseqnoBeforeFlush = m_IOSequenceNo;
            Log_RawDeviceIterator   firstWrittenPosition; // PTS 1125830 mb 2003-02-12

            // prepare and flush log pages:
            this->PrepareAndFlushPageVector(firstflushedagain, firstWrittenPosition);

            // remind queues that writer is ready:
            for (SAPDB_UInt i=0; i<m_QueueRequests.GetSize(); i++)
            {
                m_QueueRequests[i].pQueue->WriterTaskReady ( m_WriterTaskSync.GetTaskID(), 
                                                             ioseqnoBeforeFlush, 
                                                             firstWrittenPosition,
                                                             m_QueueRequests[i].pageCount );

                ioseqnoBeforeFlush   += m_QueueRequests[i].pageCount;
                firstWrittenPosition += m_QueueRequests[i].pageCount;
                pagesFreeOnDevice    -= m_QueueRequests[i].pageCount;
            }
        }

        // ----------------------------------------------------------
        // check whether logwriter should be suspended by user
        // ----------------------------------------------------------
        if ( m_SuspendedByUser )
            WaitForUserResume();
            
    } // mainloop while (m_Active)
}

// -----------------------------------------------------------------------------

void Log_Writer::PrepareAndFlushPageVector
    (SAPDB_Bool             firstflushedagain,
     Log_RawDeviceIterator &firstWrittenPosition)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::PrepareAndFlushPageVector", LogVolume_Trace, 5);

    // ----------------------------------------------------------
    // determine write position on raw device:
    // ----------------------------------------------------------
    m_FrameCounter += m_RequestPages.GetPageCount();            //PTS 1117709 mb 2002-08-05
    if (firstflushedagain == true)
    {
        // works only in single-queue mode:
        SAPDBERR_ASSERT_STATE(m_LogQueues.GetSize() == 1);
        --m_WriteIterator;
        --m_FrameCounter;                                       //PTS 1117709 mb 2002-08-05
    }
    
    firstWrittenPosition = m_WriteIterator; 

    SAPDB_Bool must_flush_infopage = 
                    m_FrameCounter >= LOG_WRITER_INFOPAGEWRITECYCLE
                    ||
                    m_LogInfoPage.ReadDeviceState() == Log_DeviceStateCleared
                    ||
                    m_LogInfoPage.ReadDeviceState() == Log_DeviceStatePartiallyCleared; //PTS 1111525 mb 2002-01-29

    // ----------------------------------------------------------
    // mark pages with writer i/o seq. numbers:
    // ----------------------------------------------------------

    IOMan_LogPagesIterator  iter( m_RequestPages );
    Kernel_Date             date;
    Kernel_Time             time;

    Kernel_DateTime::GetSystemDateTime (date, time);

    #ifdef SAPDB_QUICK
    Log_RawDeviceIterator checkWriteIter = firstWrittenPosition;
    #endif
    
    for( iter.Begin(); !iter.End(); ++iter )
    {
        Log_Page &page = *iter;
        
        page.WriteWriterIOSequenceNo(m_IOSequenceNo);
        page.WriteDate              (date);
        page.WriteTime              (time);

        #ifdef SAPDB_QUICK
        if ( LogVolume_Trace.TracesLevel(6) )
        {
            Kernel_VTrace() << "offset: " << checkWriteIter.GetPosition()
                            << ", parity: "
                            << page.ReadParityFromHeader()
                            << "/"
                            << page.ReadParityFromTrailer();
        }
        ++checkWriteIter;
        #endif
        ++m_IOSequenceNo;
    }
    
    #ifdef SAPDB_QUICK
    if ( LogVolume_Check.ChecksLevel(5) )
    {
        iter.Begin();
        if ( m_CheckQSequence.IsValid() )
        {
            if ( m_CheckQSequence == (*iter).ReadQueueIOSequenceNo()
                 &&
                 ! firstflushedagain )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_Writer queue seq mismatch") );
        }
        m_CheckQSequence = (*iter).ReadQueueIOSequenceNo();
        
        IOSequenceNo lastWsequence = (*iter).ReadLastWriterIOSequenceNo();
        SAPDB_UInt   offsetcount   = (*iter).ReadOffsetCount();
        if ( LogVolume_Trace.TracesLevel(6) )
            Kernel_VTrace() << "lastWsequence: " << lastWsequence
                            << ", offsetcount: " << offsetcount
                            << ", firstflushedagain: " << firstflushedagain
                            << NewLine;
        
        if ( m_CheckWSequence.IsValid() && firstflushedagain )
        {
            if ( m_CheckWSequence+1 != (*iter).ReadLastWriterIOSequenceNo() )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_Writer writer seq mismatch") );
            if ( m_CheckOffsetCount+1 != (*iter).ReadOffsetCount() )
                RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                              SAPDBERR_ASSERT_STATE_FAILED,
                                              "Log_Writer offsetcount mismatch") );
        }
    
        iter.End();
        m_CheckWSequence   = (*iter).ReadLastWriterIOSequenceNo();
        m_CheckOffsetCount = (*iter).ReadOffsetCount();
    }
    #endif
    
    // ----------------------------------------------------------
    // flush page vector:
    // ----------------------------------------------------------
    Log_DeviceBlockCount freespace = m_WriteIterator.GetEndDistance() + 1;
    if ( freespace > m_RequestPages.GetPageCount() )
    {
        // write whole cluster at one:
        this->FlushPages(m_RequestPages);
    }
    else if ( freespace == m_RequestPages.GetPageCount() )
    {
        if ( EnableAddedLogDeviceIsPossible() )
            EnableAddedLogDevice();
        // write whole cluster at one:
        this->FlushPages(m_RequestPages);
    }
    else if ( EnableAddedLogDeviceIsPossible() )
    {
        EnableAddedLogDevice();
        // write whole cluster at one:
        this->FlushPages(m_RequestPages);
    }
    else
    {
        // must write two chunks:
        SAPDB_UInt             i;
        IOMan_LogPagesIterator iter( m_RequestPages );

        // write first page cluster:
        m_RequestPagesHelper.Clear();

        for (i=0, iter.Begin(); i<freespace; ++i, ++iter)
        {
            m_RequestPagesHelper.Add( *iter );
        }
        this->FlushPages(m_RequestPagesHelper);

        // write second page cluster:
        m_RequestPagesHelper.Clear();
        for (i=freespace; i<m_RequestPages.GetPageCount(); ++i, ++iter)
        {
            m_RequestPagesHelper.Add( *iter );
        }
        this->FlushPages(m_RequestPagesHelper);
    }

    // update and write out info page if required:
    if (must_flush_infopage)
    {
        m_ArchiveLog->UpdateInfoPageFromWriter ( m_WriterTaskSync.GetTaskID(),
                                                 m_IOSequenceNo-1,
                                                 (m_WriteIterator-1).GetPosition() );
        if (m_FrameCounter >= LOG_WRITER_INFOPAGEWRITECYCLE)
            m_FrameCounter -= LOG_WRITER_INFOPAGEWRITECYCLE;
    }
}

// -----------------------------------------------------------------------------
void Log_Writer::EnableAddedLogDevice()
{
    if (m_NewDeviceEndOffset < 
        m_WriteIterator.GetEndOffset() + m_RequestPages.GetPageCount())
    {
        // fatal error: must increase log size for a sufficient amount
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "Log_Writer increase log size failed") );
    }
    // ----------------------------------------------------------
    // - add log devspace has occurred
    // - end of old log device has been reached
    // - expand write cycle to new device now!
    // ----------------------------------------------------------
    m_WriteIterator.SetEndOffset(m_NewDeviceEndOffset);
    // ----------------------------------------------------------
    // make new cycle related log size persistent first:
    // ----------------------------------------------------------
    m_ArchiveLog->ChangeEndOffsetAfterAddDevSpace( m_WriterTaskSync.GetTaskID(),
                                                   m_NewDeviceEndOffset );
    m_NewDeviceEndOffset.Invalidate();
}

// -----------------------------------------------------------------------------

void Log_Writer::FlushPages
    (IOMan_LogPages&                pages)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::FlushPages", LogVolume_Trace, 5);

    Log_ClusterAddress 
        addr(m_WriteIterator.GetPosition(), pages.GetPageCount());
 
    SAPDBTRACE_WRITELN (LogTrans_Trace, 6, ", offset: "    << m_WriteIterator.GetPosition() << \
                                      ", pages: "     << pages.GetPageCount());
    
    m_PageWriter.WriteLogPages (m_WriterTaskSync.GetTaskID(), pages, addr);

    m_WriteIterator += pages.GetPageCount();
}

// -----------------------------------------------------------------------------

Log_Writer::RawDeviceOffset Log_Writer::GetOldestRequiredOffset() const
{
    RawDeviceOffset saveptOffset         = m_LogInfoPage.ReadSavepointEntryOffset();
    RawDeviceOffset oldestNotSavedOffset = m_LogInfoPage.ReadOldestNotSavedOffset();

    if ( saveptOffset.IsInvalid())
    {
        if ( oldestNotSavedOffset.IsValid() )
        {
            return oldestNotSavedOffset;
        }
    }
    else
    {
        if (oldestNotSavedOffset.IsValid())
        {
            if (m_LogInfoPage.ReadOldestNotSavedIOSequenceNo().IsValid())
            {
                if (m_LogInfoPage.ReadSavepointIOSequenceNo().IsValid())
                {

                    // two valid values - return the offset of the lower (older) sequence
                    if (m_LogInfoPage.ReadOldestNotSavedIOSequenceNo()
                        <
                        m_LogInfoPage.ReadSavepointIOSequenceNo())
                    {
                        return oldestNotSavedOffset;
                    }
                    else
                    {
                        return saveptOffset;
                    }
                }
            }                    
        }
    }

    // nothing saved until now
    return m_WriteIterator.GetBeginOffset();

}

// -----------------------------------------------------------------------------
Log_DeviceBlockCount Log_Writer::CalculateFreePages() const
{
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::CalculateFreePages", LogVolume_Trace, 5);

    if ( m_IOSequenceNo.IsInvalid() )
        return m_WriteIterator.GetSize() - LOG_RESTART_RESERVE_SPACE;

    RawDeviceIterator firstFreePage = m_WriteIterator;

    ++firstFreePage;
    
    Log_DeviceBlockCount distance = firstFreePage.GetDistanceTo(GetOldestRequiredOffset());
    
    // PTS 1111862 MB 2001-09-19
    // in case of a restart the LOG_FULL_MARGIN pages are used
    // for writing the first savepoint onto the log_device
    if ( g01restart_local )
        return distance;
    
    // in case of normal operation there are LOG_FULL_MARGIN pages reserved
    // for the restart-case
    if (distance < LOG_RESTART_RESERVE_SPACE)
        return 0;

	return distance - LOG_RESTART_RESERVE_SPACE;
}

// -----------------------------------------------------------------------------
void Log_Writer::CheckAndPrintLogUsageMessage(Log_DeviceBlockCount actualPagesFree)
{
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::CheckAndPrintLogUsageMessage", LogVolume_Trace, 5);

    if ( actualPagesFree <= PagesFreeForUsers(actualPagesFree) )
    {
        if (m_lastPercentageUsed != 100)
        {
            m_lastPercentageUsed = 100;
            RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_IS_FULL));
        }
        return;
    }

    SAPDB_Int newPercentageUsed =
        100 - ( PagesFreeForUsers(actualPagesFree) * 100
                /
                (m_WriteIterator.GetSize() - (m_logQueueReserveSpace+m_logCommitReserveSpace)));

    if (newPercentageUsed > 99)
        newPercentageUsed = 99;

    if (newPercentageUsed - m_lastPercentageUsed != 0)
    {
        if (newPercentageUsed >= LOGWRITER_LOG_FULL_WARNING_LIMIT)
            RTE_Message( Log_Exception(__CONTEXT__, LOG_IS_NEARLY_FULL,
                         SAPDB_ToString(newPercentageUsed,_T_d)));
        m_lastPercentageUsed = newPercentageUsed;
    }
}


// -----------------------------------------------------------------------------

void Log_Writer::CheckForLogSaveAndSavepoint(tgg00_TransContext &trans)
{
    // PTS 1117895 UH 2002-10-21 now called by timeout task
    
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::CheckForLogSaveAndSavepoint", LogVolume_Trace, 5);
    
    //
    // This is called by the timeout task !!
    //
    
    // reading dirty - autosave task may run parallel and change these values
    const RawDeviceIterator copiedWriterIterator = m_WriteIterator;
    const RawDeviceOffset   oldestNotSavedOffset =
                              m_LogInfoPage.ReadOldestNotSavedOffset().IsValid()
                              ? m_LogInfoPage.ReadOldestNotSavedOffset()
                              : copiedWriterIterator.GetBeginOffset();
    const DeviceBlockCount  notsaveddist         = copiedWriterIterator.GetDistanceFrom(oldestNotSavedOffset);
    const SAPDB_UInt        autosaveDistance     = m_ArchiveLog->GetAutosaveDistance();

    SAPDBTRACE_WRITELN (LogTrans_Trace, 6,    "oldestNotSavedOffset: " << oldestNotSavedOffset
                                           << ", notsaveddist: "       << notsaveddist
                                           << ", autosavedist: "       << autosaveDistance);

    if ( notsaveddist >= autosaveDistance ) // PTS 1113361 mb 2002-01-11
    {
        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "trigger autosave");
        if ( k38is_on_autosave (trans, false) && trans.trError_gg00 == e_ok )
            k38trigger_autosave (trans, false);
    }
    
    // check if a savepoint has to be written
    // minDistToSavepoint is recalculated every time, because the log_size may changed

    if (m_lastPercentageUsed > 100 - LOGWRITER_MIN_DIST_TO_SAVEPOINT)
    {
              Log_RawDeviceOffset  startOffset        = k57restartrec->rstLastSavept_kb00().svpStartOffset_kb00;
        const SAPDB_UInt           logSize            = copiedWriterIterator.GetSize();
        const Log_DeviceBlockCount minDistToSavepoint = (logSize * LOGWRITER_MIN_DIST_TO_SAVEPOINT) / 100;

        if (startOffset.IsInvalid())
            startOffset = copiedWriterIterator.GetBeginOffset();
            
        if ( (copiedWriterIterator.GetDistanceTo(startOffset) * 100) / logSize < LOGWRITER_MIN_DIST_TO_SAVEPOINT )
        {
            Log_SavepointManager.StartSavepoint (trans, Log_SVPReasonLogWriter);
            return;
        }
    }

    // if any log was written but not as much as checked before
    // then check if the time defined in RESTART_TIME is reached since last savepoint
    // GetIOSequenceNo() returns the NEXT iosequence which is to be used.
    // writerIOseq - 1 == svpIOseq => nothing was written to the log.

    if ( GetIOSequenceNo()
         >
         k57restartrec->rstLastSavept_kb00().svpIOsequence_kb00
         + 1 + MIN_IOSEQUENCES_BEFORE_SAVEPOINT )
    {
        Kernel_Date currentDate;
        Kernel_Time currentTime;

        Kernel_DateTime::GetSystemDateTime(currentDate, currentTime);
        
        if ( Kernel_DateTime::OlderThan( k57restartrec->rstLastSavept_kb00().svpDate_kb00,
                                         k57restartrec->rstLastSavept_kb00().svpTime_kb00
                                         + g01restart_time,
                                         currentDate,
                                         currentTime ) )
        {
            Log_SavepointManager.StartSavepoint (trans, Log_SVPReasonTime);
        }
    }
}

// -----------------------------------------------------------------------------

void Log_Writer::SetLogFullQueueState(bool allowOnlySavepointCommit )
{
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::SetLogFullQueueState", LogVolume_Trace, 5);

    QueueIterator queueIter = m_LogQueues.Anchor();
    QueueIterator startPoint = queueIter;
    do
    {
        (**queueIter).SetLogFullState(m_WriterTaskSync.GetTaskID(), allowOnlySavepointCommit );
        ++queueIter;
    } while (queueIter != startPoint);
}

// -----------------------------------------------------------------------------
bool Log_Writer::SuspendAndGetLastWrittenIOSequence ( TaskID            taskid,
                                                      Log_IOSequenceNo &iosequence )
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::SuspendAndGetLastWrittenIOSequence", LogVolume_Trace, 5);

    // this only works if only is called by one task (utility)
    // otherwise a criticle section is needed
    SAPDBERR_ASSERT_ARGUMENT (taskid == g01pid.taskUtility_gg00);
    SAPDBERR_ASSERT_ARGUMENT (m_UserWaitingForSuspend == cgg_nil_pid);

    if ( m_SuspendedByUser )
        return false;

    m_SuspendedByUser       = true;
    m_UserWaitingForSuspend = taskid;
    
    m_WriterTaskSync.Resume();
    
    RTE_Message( Log_Exception(__CONTEXT__, LOG_WRITER_USER_SUSPENDED) );

    vsuspend(taskid, cgg_suspend_lvl_0, WAIT_FOR_LOGWRITER_SUSPEND);

    // The logwriter is suspended now.
    iosequence = m_IOSequenceNo;
    
    return true;
}

// -----------------------------------------------------------------------------
void Log_Writer::ResumeByUser ()
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::ResumeByUser", LogVolume_Trace, 5);
    RTE_Message( Log_Exception(__CONTEXT__, LOG_WRITER_USER_RESUMED) );
    m_SuspendedByUser = false;
    m_WriterTaskSync.Resume();
}

// -----------------------------------------------------------------------------
void Log_Writer::WaitForUserResume ()
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_METHOD_DEBUG ("Log_Writer::WaitForUserResume", LogVolume_Trace, 5);
    if ( m_UserWaitingForSuspend != cgg_nil_pid )
    {
        vresume(m_UserWaitingForSuspend, cgg_suspend_lvl_0);
        m_UserWaitingForSuspend = cgg_nil_pid;
    }
    
    RTE_Message( Log_Exception(__CONTEXT__,
                               LOG_WRITER_SUSPENDED_BY_USER,
                               SAPDB_ToString(m_IOSequenceNo)) );
    
    do
        m_WriterTaskSync.Suspend (LOGWRITER_USER_SUSPEND);
    while (m_SuspendedByUser);
}
