/*!*****************************************************************************

  module:       Log_Writer.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH
  
  author:       TillL,UweH,MartinB

  special area: Archive log

  description:  Log writer

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_WRITER_HPP
#define LOG_WRITER_HPP



#include "Container/Container_Vector.hpp"
#include "Container/Container_CircularList.hpp"
#include "Logging/Log_Queue.hpp"
#include "Logging/Log_RawDeviceIterator.hpp"
#include "Logging/Log_InfoPage.hpp"
#include "Logging/Log_WriterTaskSynchronizer.hpp"
#include "IOManager/IOMan_ILogIO.hpp"
#include "hgg01.h" //g01maxuser()


class Log_ArchiveLog;

// -----------------------------------------------------------------------------
// info page is written out every ... cycles ("forced" write)
// -----------------------------------------------------------------------------
#define LOG_WRITER_INFOPAGEWRITECYCLE 50
 
/*!-------------------------------------------------------------------------
declaration:  LOG_RESTART_RESERVE_SPACE
description:  number of logpages reserved for performing a restart with a
              full log
--------------------------------------------------------------------------*/
#define LOG_RESTART_RESERVE_SPACE   10
 
/*!*****************************************************************************

   class: Log_Writer

   description:
   
   Log writer manager class

   Log_Writer retrieves filled log pages from log queues and writes them out
   to the archive log devices.

   Log_Writer asks Log_Queue for filled log pages to be written
   out to log device.

*******************************************************************************/
class Log_Writer
{

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Type definitions                                             */
    /*!-------------------------------------------------------------------------
    declaration:  QueueRing
    description:  ring of log queues
    --------------------------------------------------------------------------*/
    typedef Container_CircularList<Log_Queue*>                  QueueRing;
    /*!-------------------------------------------------------------------------
    declaration:  IOSequenceNo
    description:  i/o sequence number of log pages
    --------------------------------------------------------------------------*/
    typedef Log_IOSequenceNo                                    IOSequenceNo;
    /*!-------------------------------------------------------------------------
    declaration:  RawDeviceOffset
    description:  raw offset on log device
    --------------------------------------------------------------------------*/
    typedef Log_RawDeviceOffset                                 RawDeviceOffset;
    /*!-------------------------------------------------------------------------
    declaration : Log_DeviceBlockCount
    description:  number of blocks on a raw log device
    --------------------------------------------------------------------------*/
    typedef Log_DeviceBlockCount                                DeviceBlockCount;
    /*!-------------------------------------------------------------------------
    declaration:  Allocator
    description:  allocator for queue entries
    --------------------------------------------------------------------------*/
    typedef SAPDBMem_IRawAllocator                              Allocator;
    /*!------------------------------------------------------------------------
    declaration : TaskID
    description:  identifier of kernel task
    --------------------------------------------------------------------------*/
    typedef tsp00_TaskId                                        TaskID;

private:
    /*--------------------------------------------------------------------------
    declaration : QueueIterator
    description:  iterates from queue to queue
    --------------------------------------------------------------------------*/
    typedef QueueRing::Iterator                                 QueueIterator;
    /*--------------------------------------------------------------------------
    declaration : RawDeviceIterator
    description:  block iterator of raw log device
    --------------------------------------------------------------------------*/
    typedef Log_RawDeviceIterator                               RawDeviceIterator;
    /*--------------------------------------------------------------------------
    declaration: QueueRequest
    description: one request of one queue: queueptr, pagecount, pages
    --------------------------------------------------------------------------*/
    struct QueueRequest
    {
        Log_Queue*     pQueue;
        SAPDB_UInt     pageCount;
        void Clear()
        {
            pQueue    = 0;
            pageCount = 0;
        }
        QueueRequest()
        {
            Clear();
        }
        QueueRequest( Log_Queue*     newpQueue,
                      SAPDB_UInt     newpageCount )
        {
            pQueue    = newpQueue;
            pageCount = newpageCount;
        }
    };
    /*--------------------------------------------------------------------------
    declaration: QueueRequest
    description: requests of several queues
    --------------------------------------------------------------------------*/
    typedef Container_Vector<QueueRequest>                      QueueRequests;
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Type definitions                                          */
    /* -----------------------------------------------------------------------*/

public:
    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /*!-------------------------------------------------------------------------
    function:     Log_Writer()
    description:  Constructor
    arguments:    alloc        [in] raw allocator to be used
                  logqueues    [in] circular list of queues that serve this log writer
                  pagewriter   [in] writer of raw log device
                  writersync   [in] writer task synchronization object
    return value: none
    --------------------------------------------------------------------------*/
    Log_Writer
        (Allocator&                     alloc,
         QueueRing&                     logqueues,
         IOMan_ILogIO&                  pagewriter,
         Log_WriterTaskSynchronizer&    writersync)
        :
         m_LogQueues         (logqueues),
         m_PageWriter        (pagewriter),
         m_WriterTaskSync    (writersync),
         m_QueueRequests     (alloc),
         m_RequestPages      (alloc, UTF8( "Log_Writer::WriteLogPages" )),
         m_RequestPagesHelper(alloc, UTF8( "Log_Writer::WriteLogPages" )),
         m_IOSequenceNo      (IOSequenceNo()),  // invalid
         m_Active            (false),
         m_Allocated         (false),
         m_FrameCounter      (0),
         m_IsSuspended       (false),
         m_CheckQSequence    (IOSequenceNo()),  // invalid
         m_CheckWSequence    (IOSequenceNo()),  // invalid
         m_CheckOffsetCount  (0),
         m_lastPercentageUsed(0),
         m_SuspendedByUser   (false),          // PTS 1114053 UH 2002-10-22 new
         m_UserWaitingForSuspend (cgg_nil_pid) // PTS 1114053 UH 2002-10-22 new
    {}
    /*!-------------------------------------------------------------------------
    function:     Initialize()
    description:  Initializes writer structures

                  - Performs memory allocation of writer structures.
                  - The parameter 'pagevecsize' gives the optimum i/o count of
                    page frames to be used by the writer.
                  - 'writeiter' is positioned at the current write position
                    on log device.
                  - 'logfullmargin' is a security margin which must be respected
                    by the log writer. In normal log operation, the log writer 
                    may not write to device blocks within a margin of 
                    'logfullmargin' blocks from the oldest required block.

    arguments:    writeioseqno  [in] ioseqno of the last log page that has been flushed
                  writeiter     [in] iterator at position on device where to write
                                     next block
                  infopage      [in] log info page handler
                  infopageoffs  [in] offset of log info page on raw device
                  logfullmargin [in] log full security margin in blocks
                  pagevecsize   [in] size of the page vector
    return value: (SAPDB_Bool) true if successful
    --------------------------------------------------------------------------*/
    SAPDB_Bool Initialize
        (IOSequenceNo                   writerioseqno,
         RawDeviceIterator              writeiter,
         Log_ArchiveLog                *archivelog,
         Log_InfoPage                   infopage,
         RawDeviceOffset                infopageoffs,
         SAPDB_UInt                     pagevecsize)
    {
        m_IOSequenceNo         = writerioseqno.IsValid() ? IOSequenceNo(writerioseqno+1) : IOSequenceNo(0);
        m_WriteIterator        = writeiter;
        m_ArchiveLog           = archivelog;
        m_LogInfoPage          = infopage;
        m_LogInfoPageOffset    = infopageoffs;
        
        if (m_Allocated == false)
        {
            if (!m_QueueRequests.Initialize(pagevecsize)        ||
                !m_RequestPages.Initialize(pagevecsize) ||
                !m_RequestPagesHelper.Initialize(pagevecsize))
            {
                Delete();
                return false;
            }
            m_Allocated = true;
        }
        return true;
    }

    /*!-------------------------------------------------------------------------
    function:     SetState()
    description:  Updates writer state

    arguments:    taskid        [in] id of calling task
                  writeioseqno  [in] ioseqno to be used for the next page
                  writeoffset   [in] raw device offset to be used for next i/o
                  endoffset     [in] raw device end offset
    return value: none
    --------------------------------------------------------------------------*/
    void SetState
        (TaskID                         taskid,
         IOSequenceNo                   writerioseqno,
         RawDeviceOffset                writeoffset,
         RawDeviceOffset                endoffset)
    {
        m_IOSequenceNo   = writerioseqno.IsValid() ? writerioseqno : IOSequenceNo(0);
        m_WriteIterator.SetPosition(writeoffset);
        SetDeviceEnd(endoffset);
    }

    /*!-------------------------------------------------------------------------
    function:     SetDeviceEnd()
    description:  Sets log device end offset

                  - 'offset' is the raw device offset behind the last block
                    to be accessed by the writer.
                  - It must be at least as large as the current end offset.

    arguments:    offset [in]  device end offset
    return value: none
    --------------------------------------------------------------------------*/
    void SetDeviceEnd (RawDeviceOffset offset)
    {
        if (offset > m_WriteIterator.GetEndOffset())
        {
            m_NewDeviceEndOffset = offset;
        }
    }

    /*!-------------------------------------------------------------------------
    function:     Delete()
    description:  Deletes writer structures

                  - Performs memory deallocation

    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Delete() 
    {
        m_QueueRequests.Delete();
        m_RequestPages.Delete();
        m_RequestPagesHelper.Delete();
    }
    /*!-------------------------------------------------------------------------
    function:     SetIOSequenceNoAfterRestoreLog()
    description:  This must be used, if a log-recovery is started from init-config
                  and not with a data-recovery.
    arguments:    newIOSequenceNo
    return value: none
    --------------------------------------------------------------------------*/
    void SetIOSequenceNoAfterRestoreLog (IOSequenceNo newIOSequence)
    {
        // PTS 1117079 UH 2002-08-06 do not assign invalid but instead assign 0
        m_IOSequenceNo = newIOSequence.IsValid()
                         ? newIOSequence
                         : IOSequenceNo(0);
    }
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    
    /* -----------------------------------------------------------------------*/
    /*! chapter: Writer operation                                             */
    /*!-------------------------------------------------------------------------
    function:     Run()
    description:  Starts the log writer 

                  - The log writer is requesting log pages for output from the
                    associated log queues. Log queues are served in a loop.

    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Run();

    /*!-------------------------------------------------------------------------
    function:     Stop()
    description:  Stops the log writer 

                  - After the current i/o, the writer stops execution.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Stop()
    {
        m_Active = false;
    }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Writer operation                                          */
    /* -----------------------------------------------------------------------*/

    
    /* -----------------------------------------------------------------------*/
    /*! chapter: Writer status                                                */
    /*!-------------------------------------------------------------------------
    function:     GetIOSequenceNo()
    description:  Returns the current i/o sequence number

    arguments:    none
    return value: (IOSequenceNo)
    --------------------------------------------------------------------------*/
    IOSequenceNo GetIOSequenceNo() const
    {
        return m_IOSequenceNo;
    }

    /*!-------------------------------------------------------------------------
    function:     GetWriteOffset()
    description:  Returns the current write offset on the raw device
                    
                  - This is the position where the next log page will be written.
    arguments:    none
    return value: (RawDeviceOffset)
    --------------------------------------------------------------------------*/
    RawDeviceOffset GetWriteOffset() const
    {
        return m_WriteIterator.GetPosition();
    }

    /*!-------------------------------------------------------------------------
    function:     LogIsFull()
    description:  Checks if log is full 
    arguments:    none
    return value: (SAPDB_Bool) true if full
    --------------------------------------------------------------------------*/
	inline SAPDB_Bool LogIsFull()  
	{
		SAPDBTRACE_METHOD_DEBUG ("Log_Writer::LogIsFull", LogVolume_Trace, 5);
		
		return ! EnoughPagesFreeForUserTaskWrite(CalculateFreePages());
	}

    /*!-------------------------------------------------------------------------
    function:     CalculateFreePages()
    description:  gives the count of pages which are not used.
    return value: (Log_DeviceBlockCount) the count of pages
    --------------------------------------------------------------------------*/
    Log_DeviceBlockCount CalculateFreePages() const;
    
    /*!-------------------------------------------------------------------------
    function:     PagesFreeForUsers()
    description:  Gives the amount of pages which cannot be overwritten by users.
                  This is for displaying of log statistics
    return value: (SAPDB_UInt) page count
    --------------------------------------------------------------------------*/
    Log_DeviceBlockCount PagesFreeForUsers(Log_DeviceBlockCount freePages) const
    {
        if( freePages < m_logQueueReserveSpace + m_logCommitReserveSpace )
            return 0;
        return freePages - (m_logQueueReserveSpace + m_logCommitReserveSpace);
    }

    /*!-------------------------------------------------------------------------
    function:     IsSuspended()
    description:  Checks if log writer is suspended due to logfull
    arguments:    none
    return value: (SAPDB_Bool) true if full
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsSuspended() const 
    {
        return m_IsSuspended;
    }

    /*!-------------------------------------------------------------------------
    function:     NumOfPagesReserved()
    description:  returns the number of pages reserved for special purposes
    arguments:    none
    return value: the number of pages not writable by normal operation
    --------------------------------------------------------------------------*/
    inline DeviceBlockCount CalculateReservedPages(
        Log_Queue::EntryCount  numPagesInQueues)
    {
        m_logQueueReserveSpace = numPagesInQueues;
        m_logCommitReserveSpace = g01maxuser()+1;
        m_totalReserveSpace = m_logQueueReserveSpace+m_logCommitReserveSpace+LOG_RESTART_RESERVE_SPACE;

        return m_totalReserveSpace;
    }

    /*--------------------------------------------------------------------------
    function:     CheckForLogSaveAndSavepoint()
    arguments:    trans [in/out]
    description:  Triggers the log save component if enough pages for log save
                  available.
                  This is called by the TimeOut task.
    --------------------------------------------------------------------------*/
    void CheckForLogSaveAndSavepoint(tgg00_TransContext &trans);

    /*--------------------------------------------------------------------------
    function:     SuspendAndGetLastWrittenIOSequence
    arguments:    taskid     [in]
                  iosequence [out]
    returns:      Returns false if this is called more then by one task.
    description:  The logwriter is suspended until Resume() is called.
                  The returned iosequence is the last used iosequence.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    bool SuspendAndGetLastWrittenIOSequence ( TaskID            taskid,
                                              Log_IOSequenceNo &iosequence );

    /*--------------------------------------------------------------------------
    function:     ResumeByUser
    arguments:    none
    description:  The logwriter is resumed after suspended by
                  SuspendAndGetLastWrittenIOSequence.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    void ResumeByUser ();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Writer status                                             */
    /* -----------------------------------------------------------------------*/

private:
    
    /*--------------------------------------------------------------------------
    function:     GetOldestRequiredOffset()
    description:  Returns the oldest offset on the log device that may not
                  be overwritten by the log writer
    return value: (RawDeviceOffset) 
    --------------------------------------------------------------------------*/
    RawDeviceOffset GetOldestRequiredOffset() const;

    /*--------------------------------------------------------------------------
    function:     PrepareAndFlushPageVector()
    description:  All pages get there writer iosequence and a timestamp
                  A consistency check is done.
    @param        replacelastpage     [in]
    @param        firstWrittenPosition    [out]  Iterator, which points to the position of 
                                                 the first written page
    */
    void PrepareAndFlushPageVector (SAPDB_Bool             replacelastpage,
                                    Log_RawDeviceIterator &firstWrittenPosition);

    /*--------------------------------------------------------------------------
    function:     FlushPages()
    description:  The pages are written with the iomanager to the log devspace.
    arguments:    page vector [in]
    --------------------------------------------------------------------------*/
    void FlushPages
        (IOMan_LogPages&                pages);
    
    /*--------------------------------------------------------------------------
    function:     CheckAndPrintLogUsageMessage
    arguments:    actualPagesFree [in]
    description:  This compares the actual filling state and write a percent message of usage
                  To the knldiag.
    --------------------------------------------------------------------------*/
    void CheckAndPrintLogUsageMessage(Log_DeviceBlockCount actualPagesFree);

    /*--------------------------------------------------------------------------
    function:     InitiateSavepoint()
    description:  Initiates a savepoint with a dummy-transaction
    --------------------------------------------------------------------------*/
    void InitiateSavepoint();

    /*--------------------------------------------------------------------------
    function:     EnableAddedLogDevice()
    description:  The added device can now be accessed by the writer.
    --------------------------------------------------------------------------*/
    void EnableAddedLogDevice();

    /*--------------------------------------------------------------------------
    function:     EnableAddedLogDeviceIsPossible()
    description:  Checks if the old cycle is already needed (by backup).
    --------------------------------------------------------------------------*/
    bool EnableAddedLogDeviceIsPossible()
    {
         return m_NewDeviceEndOffset.IsValid()
                &&
                m_WriteIterator.GetPosition() > GetOldestRequiredOffset();
    }

    /*--------------------------------------------------------------------------
    function:     EnoughPagesFreeForUserTaskWrite()
    arguments:    pagesFreeOnDevice [in]
    description:  Checks if pagesFreeOnDevice are enough pages for allowing 
                  user-tasks the writing onto the log
                  new for PTS 1115488 mb 2002-05-03
    --------------------------------------------------------------------------*/
    const bool EnoughPagesFreeForUserTaskWrite(const Log_DeviceBlockCount pagesFreeOnDevice )
    {
         return pagesFreeOnDevice > m_logQueueReserveSpace + m_logCommitReserveSpace;
    }

    /*--------------------------------------------------------------------------
    function:     SetLogFullQueueState
    arguments:    allowOnlySavepointCommit [in]
    description:  if allowOnlySavePointCommit = TRUE then sets all log_queues
                  to Log_Full, so that only commits, rollbacks and savepoints
                  may be written into the queues until a save_log frees log_pages.
                  Otherwise it opens the Log_Queues for every write operation
                  new for PTS 1115488 mb 2002-05-03
    --------------------------------------------------------------------------*/
    void SetLogFullQueueState(bool allowOnlySavepointCommit);

    /*--------------------------------------------------------------------------
    function:     WaitForUserResume
    arguments:    none
    description:  if m_SuspendedByUser is true this method lets wait until
                  m_SuspendedByUser is false.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    void WaitForUserResume();
    
    /*--------------------------------------------------------------------------
    declaration: m_LogQueues
    description: circular list of log queues that serve this writer
    --------------------------------------------------------------------------*/
    QueueRing&                          m_LogQueues;
    /*--------------------------------------------------------------------------
    declaration : m_PageWriter
    description:  writes log pages to raw log device
    --------------------------------------------------------------------------*/
    IOMan_ILogIO&                       m_PageWriter;
    /*--------------------------------------------------------------------------
    declaration : m_WriterTaskSync
    description:  log writer task synchronization object
    --------------------------------------------------------------------------*/
    Log_WriterTaskSynchronizer&         m_WriterTaskSync;
    /*--------------------------------------------------------------------------
    declaration : m_Allocated
    description:  indicates whether writer structures have been allocated
    --------------------------------------------------------------------------*/
    SAPDB_Bool                          m_Allocated;
    /*--------------------------------------------------------------------------
    declaration : m_Active
    description:  indicates whether writer is active
    --------------------------------------------------------------------------*/
    SAPDB_Bool                          m_Active;
    /*--------------------------------------------------------------------------
    declaration : m_IOSequenceNo
    description:  actual writer i/o sequence number: the last page in i/o has 
                  been marked with this seqno 
                  (m_IOSequenceNo may be invalid if not yet any log has been written)
    --------------------------------------------------------------------------*/
    IOSequenceNo                        m_IOSequenceNo;
    /*--------------------------------------------------------------------------
    declaration : m_FrameCounter
    description:  counts frames that have been flushed since info page was flushed
    --------------------------------------------------------------------------*/
    SAPDB_UInt                          m_FrameCounter;
    /*--------------------------------------------------------------------------
    declaration : m_NewDeviceEndOffset
    description:  contains new device offset (if add log devspace has occurred)
    --------------------------------------------------------------------------*/
    RawDeviceOffset                     m_NewDeviceEndOffset;
    /*--------------------------------------------------------------------------
    declaration: m_RequestPages
    description: collection of page frames used for write i/o requests
    --------------------------------------------------------------------------*/
    IOMan_LogPages                      m_RequestPages;
    /*--------------------------------------------------------------------------
    declaration: m_RequestPagesHelper
    description: another collection of page frames (required for split i/o)
    --------------------------------------------------------------------------*/
    IOMan_LogPages                      m_RequestPagesHelper;
    /*--------------------------------------------------------------------------
    declaration : m_QueueRequests
    description:  queues and pagecounts per queue
    --------------------------------------------------------------------------*/
    QueueRequests                       m_QueueRequests;
    /*--------------------------------------------------------------------------
    declaration : m_WriteIterator
    description:  iterator at actual write offset on raw device
    -------------------------------------------------------------------------*/
    RawDeviceIterator                   m_WriteIterator;
    /*--------------------------------------------------------------------------
    declaration: m_ArchiveLog
    description: the arhive log handler for this writer.
    --------------------------------------------------------------------------*/
    Log_ArchiveLog                     *m_ArchiveLog;
    /*--------------------------------------------------------------------------
    declaration: m_LogInfoPage
    description: log infopage handler
    --------------------------------------------------------------------------*/
    Log_InfoPage                        m_LogInfoPage;
    /*--------------------------------------------------------------------------
    declaration: m_LogInfoPageOffset
    description: offset of log info page on raw device
    --------------------------------------------------------------------------*/
    RawDeviceOffset                     m_LogInfoPageOffset;
    /*--------------------------------------------------------------------------
    declaration: m_IsSuspended
    description: Indicates if the log writer task is suspended due to log full.
    --------------------------------------------------------------------------*/
    bool                                m_IsSuspended;
    /*--------------------------------------------------------------------------
    declaration: m_Check....
    description: for checking
    --------------------------------------------------------------------------*/
    IOSequenceNo                        m_CheckQSequence;
    IOSequenceNo                        m_CheckWSequence;
    SAPDB_UInt                          m_CheckOffsetCount;
    /*--------------------------------------------------------------------------
    declaration : m_lastPercentageUsed
    description:  the percentage of used log-pages on the log_device
    --------------------------------------------------------------------------*/    
    SAPDB_UInt                          m_lastPercentageUsed;
    /*--------------------------------------------------------------------------
    declaration : m_logQueueReserveSpace
    description : number of logpages reserved for for assuring that all queue
                  elements may be written onto the log
    --------------------------------------------------------------------------*/
    DeviceBlockCount                    m_logQueueReserveSpace;
    /*--------------------------------------------------------------------------
    declaration : m_logCommitReserveSpace
    description : number of logpages reserved for for assuring that all 
                  usertasks may write a commit/rollback and that a save-
                  point for shutdown is possible
    --------------------------------------------------------------------------*/
    DeviceBlockCount                    m_logCommitReserveSpace;
    /*--------------------------------------------------------------------------
    declaration : m_totalReserveSpace
    description : number of logpages reserved totally for special purposes
    --------------------------------------------------------------------------*/
    DeviceBlockCount                    m_totalReserveSpace;
    /*--------------------------------------------------------------------------
    declaration : m_SuspendedByUser
    description : If this is true, the logwriter must wait until explicitely
                  resumed by user.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    bool                                m_SuspendedByUser;
    /*--------------------------------------------------------------------------
    declaration : m_UserWaitingForSuspend
    description : If this is not nil_task_id there is a user waiting
                  that the logwriter suspends.
    --------------------------------------------------------------------------*/
    // PTS 1114053 UH 2002-10-22 new
    TaskID                              m_UserWaitingForSuspend;    
};
/*!*****************************************************************************

   endclass: Log_Writer

*******************************************************************************/


#endif // LOG_WRITER_HPP


