/*!*****************************************************************************

  module:       Log_WriterTaskSynchronizer.hpp

  ------------------------------------------------------------------------------

  responsible:  UweH

  author:       TillL

  special area: Logging

  description:  

  see also:

  copyright:    (C) 2000 SAP AG

  ------------------------------------------------------------------------------

  ========== licence begin  GPL
  Copyright (C) 2000 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*******************************************************************************/


#ifndef LOG_WRITERTASKSYNCHRONIZER_HPP
#define LOG_WRITERTASKSYNCHRONIZER_HPP


#include "gsp00.h"
#include "ggg00.h"
#include "heo56.h"

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "RunTime/Synchronisation/RTESync_NamedSpinlock.hpp"


// -----------------------------------------------------------------------------
// Suspend reason for log writer:
// -----------------------------------------------------------------------------
#define LOG_QUEUE_EMPTY             235
#define LOG_WRITER_NOT_INITIALIZED  242


/*!*****************************************************************************

   class: Log_WriterTaskSynchronizer

   description:

*******************************************************************************/
class Log_WriterTaskSynchronizer
{
public:

    /*!------------------------------------------------------------------------
    declaration : TaskID
    description:  identifier of kernel task
    --------------------------------------------------------------------------*/
    typedef tsp00_TaskId                                        TaskID;
    /*!------------------------------------------------------------------------
    declaration : SuspendReason
    description:  suspend reason id
    --------------------------------------------------------------------------*/
    typedef SAPDB_Int2                                          SuspendReason;

    /*!-------------------------------------------------------------------------
    function:     Log_WriterTaskSynchronizer()
    description:  Constructor
    arguments:    taskid [in]      caller task id
    --------------------------------------------------------------------------*/
    Log_WriterTaskSynchronizer
        (TaskID                         taskid)
        :
         m_TaskID    (taskid),
         m_Suspended (false),
         m_Spinlock  ((SAPDB_UTF8*)"Log_WriterTaskSynchronizer::m_Spinlock")
    {}

    /*!----------------------------------------------------------------------
    function:     WillBeSuspended()
    description:  save the information, that the writer will be suspended
    arguments:    reason [in]  suspend reason
    --------------------------------------------------------------------------*/
    void WillBeSuspended ()
    {
        RTESync_LockedScope Lock(m_Spinlock);
        m_Suspended = true;
    }

    /*!----------------------------------------------------------------------
    function:     DoSuspend()
    description:  after WillBeSuspend() must follow DoSuspend()
                  it definitely suspends the caller
    arguments:    reason [in]  suspend reason
    --------------------------------------------------------------------------*/
    void DoSuspend  (SuspendReason reason)
    {
        vsuspend(m_TaskID, cgg_suspend_lvl_0, reason);
    }

    /*!----------------------------------------------------------------------
    function:     Suspend()
    description:  Suspends task if necessary

                  - The writer task is suspended if not yet suspended.
    arguments:    reason [in]  suspend reason
    --------------------------------------------------------------------------*/
    void Suspend (SuspendReason reason)
    {
        m_Spinlock.Lock();
        m_Suspended = true;
        m_Spinlock.Unlock();
        vsuspend(m_TaskID, cgg_suspend_lvl_0, reason);
    }

    /*!----------------------------------------------------------------------
    function:     Resume()
    description:  Resumes task if necessary

                  - The writer task is resumed if not yet resumed.

    --------------------------------------------------------------------------*/
    void Resume()
    {
        SAPDB_Bool do_resume = false;
        {
            RTESync_LockedScope Lock(m_Spinlock);

            if (m_Suspended)
            {
                do_resume   = true;
                m_Suspended = false;
            }
        }
        if (do_resume)
            vresume(m_TaskID, cgg_suspend_lvl_0);
    }

    /*!-------------------------------------------------------------------------
    function:     GetID()
    description:  Returns the task id of the writer task
    arguments:    none
    return:       (TaskID) writer task id
    --------------------------------------------------------------------------*/
    TaskID GetTaskID() const
    {
        return m_TaskID;
    }

private:

    /*--------------------------------------------------------------------------
    declaration : m_TaskID
    description:  user task id
    --------------------------------------------------------------------------*/
    TaskID                          m_TaskID;
    /*--------------------------------------------------------------------------
    declaration : m_Suspended
    description:  task suspend state
    --------------------------------------------------------------------------*/
    SAPDB_Bool                      m_Suspended;
    /*--------------------------------------------------------------------------
    declaration : m_Spinlock
    description:  lock for synchronized code parts
    --------------------------------------------------------------------------*/
    mutable RTESync_NamedSpinlock   m_Spinlock;

};
/*!*****************************************************************************

   endclass: Log_WriterTaskSynchronizer

*******************************************************************************/



#endif // LOG_WRITERTASKSYNCHRONIZER_HPP

