/*!================================================================
 module:        OMS_Stream.hpp

 responsible:   ThomasA
 
 special area:  OMS
 
 description:   OMS Streams
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License; or (at your option) any later version.

    This library is distributed in the hope that it will be useful;
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not; write to the Free Software
    Foundation; Inc.; 59 Temple Place; Suite 330; Boston; MA  02111-1307  USA
    ========== licence end

 ===================================================================*/



#ifndef OMS_STREAM_HPP
#define OMS_STREAM_HPP

#include "gco00.h"
#include "livecachetypes.h" // nocheck

class OmsHandle;

class OMS_StreamBody; 

/*! 
  Class: OmsStream

  Description: Base class for OmsIStream and OmsOStream. These classes provide a
               stream oriented access to ABAP tables residing at the client side.

*/

class AFX_EXT_CLASS OmsStream 
{
protected :
  /*!---------------------------------------------------------------------- 
    Function: construtor

    Description: constructs a OmsStream instance


    Arguments:
      sh [in] reference to an ABAP table handle or a stream handle respectively.
              this handle contains information about the size of a stream row
              as well as swap and row construction info.
              This info is needed, if the client layout and the liveCache layout
              of a row differ.
      rsz [in] size of a raw of the stream in liveCache layout
      istream [in] if true, the stream is constructed for an input stream, i.e
                   rows are tranfered from the client to the liveCache.
                   Otherwise the stream is constructed for an output stream, i.e.
                   rows are tranfered from the liveCache to the client.
      
    Return value: none
  ----------------------------------------------------------------------*/

  OmsStream(OmsHandle* h, OmsTypeABAPTabHandle& sh, long rsz, long sz, bool istream = true);
  OmsStream(OmsHandle* h, OmsTypeStreamHandle& sh, long rsz, long sz, bool istream = true);

  /*!---------------------------------------------------------------------- 
    Function: destructor

    Description: destructs an OmsStream instance. In case of an output stream
                 rows that are still buffered are NOT transfered to the client.
                 The stream must therefore closed explicitely.
                 All memory hold by the stream instance is released.


    Arguments: none
  ----------------------------------------------------------------------*/

  ~OmsStream();

  /*!---------------------------------------------------------------------- 
    Function: omsClose

    Description: closes the stream. In case of an output stream, all buffered
                 rows are transfered to the client. After the stream has been
                 closed, no more rows can be read or written.

    Arguments: none

    Return value : none
  ----------------------------------------------------------------------*/

  void           omsClose();

  /*!---------------------------------------------------------------------- 
    Function: omsFlush

    Description: In case of an output stream, all buffered
                 rows are transfered to the client. 

    Arguments: none

    Return value : none
  ----------------------------------------------------------------------*/

  void omsFlush();

  /*!---------------------------------------------------------------------- 
    Function: omsNxt

    Description: yields a pointer to the next row of the stream.  

    Arguments: none

    Return value : pointer to the next row of the stream, NULL if end of stream
                   has been reached
  ----------------------------------------------------------------------*/

  unsigned char* omsNxt();

  /*!---------------------------------------------------------------------- 
    Function: omsWrite

    Description: writes a row identified by a pointer into the stream.  

    Arguments: 
      pRow [in] a pointer to the row to be written into the stream

    Return value : none
  ----------------------------------------------------------------------*/

  void omsWrite (unsigned char* pRow);
private :
  OMS_StreamBody* m_body; // all implementing details are hidden from the user
                          // (body handle idiom)
};

/*! endclass: OmsStream */

#endif