/****************************************************************************

  module      : OMS_VersionDirRegions.cpp

  -------------------------------------------------------------------------

  responsible : MarkusSi

  special area: OMS 
  description : OMS regions for the version directory

  first created: 2003-09-18  

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#include "gco00.h"
#include "Oms/OMS_VersionDirRegions.hpp"
#include "liveCache/LVC_LockRequest.hpp"
#include "hco31-k.h"
#include "Oms/OMS_KernelDefaultInterface.hpp"
#include "gco102.h"

extern LVC_KernelInterface* co10_GetKernelInterface();
extern IliveCacheSink*      co10_GetLcSink();


/*---------------------------------------------------------------------------*/

ExclusiveVersionDirRgn::ExclusiveVersionDirRgn(int lockId, bool useRWLocks) : m_size(1)
{
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "ExclusiveVersionDirRgn: SINGLE");
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Exclusive Region requested: " << lockId);

  m_useRWLocks = useRWLocks;
  m_pLockId[0] = lockId;

  if (useRWLocks){
    // Construct a lock-object and send it to the kernel via kernel-interface
    LVC_LockRequest lock(LVC_LockRequest::RWLOCK_LOCK_EXCLUSIVE, lockId);
    short err = co10_GetKernelInterface()->LockRequest(lock);
    if (err != e_ok)
      throw DbpError(err, "ExclusiveVersionDirRgn::ExclusiveVersionDirRgn-1-");
  }
  else {
    // Enter the critical section via the sink-interface
    co10_GetLcSink()->EnterCriticalSection(lockId);
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Exclusive Region aquired: " << lockId);
}

/*---------------------------------------------------------------------------*/

ExclusiveVersionDirRgn::ExclusiveVersionDirRgn(const int *pLockId, int size, bool useRWLocks) : m_size(size)
{
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "ExclusiveVersionDirRgn: ARRAY");
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Exclusive Regions requested: ");
  //for (int j=0; j<m_size; ++j){
  //  OMS_TRACE(omsTrInterface, co10_GetLcSink(), "  " << pLockId[j]);
  //}

  m_useRWLocks = useRWLocks;
  for (int i=0; i<m_size; ++i){
    m_pLockId[i] = pLockId[i];
    if (i>0 && m_pLockId[i] <= m_pLockId[i-1]){
      // The ids must be sorted ascending; otherwise deadlocks might occur
      throw DbpError(e_invalid_indexorder, "ExclusiveVersionDirRgn::ExclusiveVersionDirRgn-2-");
    }
    if (useRWLocks){
      // Construct a lock-object and send it to the kernel via kernel-interface
      LVC_LockRequest lock(LVC_LockRequest::RWLOCK_LOCK_EXCLUSIVE, m_pLockId[i]);
      short err = co10_GetKernelInterface()->LockRequest(lock);
      if (err != e_ok)
        throw DbpError(err, "ExclusiveVersionDirRgn::ExclusiveVersionDirRgn-3-");
    }
    else {
      // Enter the critical section using the sink interface
      co10_GetLcSink()->EnterCriticalSection(m_pLockId[i]);
    }
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Exclusive Regions aquired: ");
  //for (int k=0; k<m_size; ++k){
  //  OMS_TRACE(omsTrInterface, co10_GetLcSink(), "  " << pLockId[k]);
  //}
}

/*---------------------------------------------------------------------------*/

ExclusiveVersionDirRgn::~ExclusiveVersionDirRgn()
{
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "ExclusiveVersionDirRgn: DTOR");

  for (int i=0; i<m_size; ++i){
    if (m_useRWLocks){
      LVC_LockRequest lock(LVC_LockRequest::RWLOCK_UNLOCK_EXCLUSIVE, m_pLockId[i]);
      short err = co10_GetKernelInterface()->LockRequest(lock);
      if (err != e_ok)
        throw DbpError(err, "ExclusiveVersionDirRgn::~ExclusiveVersionDirRgn");
    }
    else {
      co10_GetLcSink()->LeaveCriticalSection(m_pLockId[i]);
    }
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Exclusive Regions freed: ");
  //for (int j=0; j<m_size; ++j){
  //  OMS_TRACE(omsTrInterface, co10_GetLcSink(), "  " << m_pLockId[j]);
  //}
}

/*===========================================================================*/

SharedVersionDirRgn::SharedVersionDirRgn(int lockId, bool useRWLocks) : m_size(1)
{
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "SharedVersionDirRgn: SINGLE");
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Shared Region requested: " << lockId);

  m_useRWLocks = useRWLocks;
  m_pLockId[0] = lockId;

  if (m_useRWLocks){
    // Construct a lock-object and send it to the kernel via kernel-interface
    LVC_LockRequest lock(LVC_LockRequest::RWLOCK_LOCK_SHARED, lockId);
    short err = co10_GetKernelInterface()->LockRequest(lock);
    if (err != e_ok)
      throw DbpError(err, "SharedVersionDirRgn::SharedVersionDirRgn-1-");
  }
  else {
    // Enter the critical section via the sink-interface
    co10_GetLcSink()->EnterCriticalSection(lockId);
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Shared Region aquired: " << lockId);
}

/*---------------------------------------------------------------------------*/

SharedVersionDirRgn::SharedVersionDirRgn(const int *pLockId, int size, bool useRWLocks) : m_size(size)
{
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "SharedVersionDirRgn: ARRAY");
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Shared Regions requested: ");
  //for (int j=0; j<m_size; ++j){
  //  OMS_TRACE(omsTrInterface, co10_GetLcSink(), "  " << pLockId[j]);
  //}

  m_useRWLocks = useRWLocks;
  for (int i=0; i<m_size; ++i){
    m_pLockId[i] = pLockId[i];
    if (i>0 && m_pLockId[i] <= m_pLockId[i-1]){
      // The ids must be sorted ascending; otherwise deadlocks might occur
      throw DbpError(e_invalid_indexorder, "SharedVersionDirRgn::SharedVersionDirRgn-2-");
    }
    if (m_useRWLocks){
      LVC_LockRequest lock(LVC_LockRequest::RWLOCK_LOCK_SHARED, m_pLockId[i]);
      short err = co10_GetKernelInterface()->LockRequest(lock);
      if (err != e_ok)
        throw DbpError(err, "SharedVersionDirRgn::SharedVersionDirRgn-3-");
    }
    else {
      // Enter the critical section via the sink-interface
      co10_GetLcSink()->EnterCriticalSection(m_pLockId[i]);
    }
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Shared Regions aquired: ");
  //for (int k=0; k<m_size; ++k){
  //  OMS_TRACE(omsTrInterface, co10_GetLcSink(), "  " << pLockId[k]);
  //}
}

/*---------------------------------------------------------------------------*/

SharedVersionDirRgn::~SharedVersionDirRgn()
{
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "SharedVersionDirRgn: DTOR");

  for (int i=0; i<m_size; ++i){
    if (m_useRWLocks){
      LVC_LockRequest lock(LVC_LockRequest::RWLOCK_UNLOCK_SHARED, m_pLockId[i]);
      short err = co10_GetKernelInterface()->LockRequest(lock);
      if (err != e_ok)
        throw DbpError(err, "SharedVersionDirRgn::~SharedVersionDirRgn");
    }
    else {
      co10_GetLcSink()->LeaveCriticalSection(m_pLockId[i]);
    }
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "Shared Regions freed: ");
  //for (int j=0; j<m_size; ++j){
  //  OMS_TRACE(omsTrInterface, co10_GetLcSink(), "  " << m_pLockId[j]);
  //}
}

/*---------------------------------------------------------------------------*/
