/*!**************************************************************************

  module      : Pager_Controller.hpp

  -------------------------------------------------------------------------

  responsible : Henrik

  special area:
  description : Pager instantiation and savepoint control

  last changed: 2002-02-16
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002

  ========== licence begin LGPL
  Copyright (C) 2002 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef PAGER_CONTROLLER_HPP
#define PAGER_CONTROLLER_HPP

class Pager_Controller;

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "ggg00.h"
#include "hgg01.h"          // g01regio_cnt

#include "KernelCommon/Kernel_Common.hpp"
#include "RunTime/Synchronisation/RTESync_NamedSpinlock.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"
#include "Container/Container_Vector.hpp"

#include "Pager/Pager_IRestart.hpp"
#include "Pager/Pager_ISavepoint.hpp"
#include "Pager/Pager_IDump.hpp"
#include "Pager/Pager_Dump.hpp"
#include "Pager/Pager_Writer.hpp"
#include "Pager/Pager_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class Pager_Controller : public Pager_ISavepoint,
    public Pager_IDump, public Pager_IRestart {
public:
    // singleton
    /*!-------------------------------------------------------------------------
    function:     GetInstance
    description:  returns instance of Pager_Controller.

    If no instance exists a new one will be created
    arguments:    none
    return value: (new) singleton instance of Pager_Controller
    --------------------------------------------------------------------------*/
    static Pager_Controller& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     BeginSavepoint 
    description:  initiate Savepoint

      wait for running pager_writer to finish
    arguments:    taskID [in] own taskID 
    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginSavepoint (const tsp00_TaskId taskId);

    /*!-------------------------------------------------------------------------
    function:     FlushDataCacheBeforePrepare 
    description:  flushes DataCache to disk
    arguments:    taskID [in] own taskID 
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FlushDataCacheBeforePrepare (const tsp00_TaskId taskId,
                                                    SAPDB_UInt &numPagesWritten,
                                                    SAPDB_UInt &numIOActions);

    /*!-------------------------------------------------------------------------
    function:     WaitForLastSplit
    description:  
    arguments:    trans [in] own transcontext 
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void WaitForLastSplit (tgg00_TransContext &trans,
                                         SAPDB_UInt &numPagesWritten,
                                         SAPDB_UInt &numIOActions);

    /*!-------------------------------------------------------------------------
    function:     FlushDataCacheAfterPrepare
    description:  flushes DataCache to disk
    arguments:    taskId [in] own taskID 
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FlushDataCacheAfterPrepare (const tsp00_TaskId taskId,
                                                   SAPDB_UInt &numPagesWritten,
                                                   SAPDB_UInt &numIOActions);

    /*!-------------------------------------------------------------------------
    function:     EndSavepoint
    description:  Savepoint finished,
    arguments:    taskID [in] own taskID 
               disablepagerwriter [in] disable pager_writer e.g. for shutdown
    return value: none
    --------------------------------------------------------------------------*/
    virtual void EndSavepoint (const tsp00_TaskId taskId,
                               const SAPDB_Bool   disablePagerWriter);

    /*!-------------------------------------------------------------------------
    function:     Dump
    description:  Dump info
    arguments:    TaskID [in] taskID 
                  Dump [out] dump structure
    return value: none
    --------------------------------------------------------------------------*/
    virtual void Dump(const tsp00_TaskId    taskId,
                      Kernel_Dump           &dump );

    /*!-------------------------------------------------------------------------
    function:     RegisterNextPagerWriter
    description:  Pager_Writer register itself with the Pager_Controller 

      the controller assigns a unique ID to each writer object
    arguments:    TaskID [in] own taskID 
                  writer [in] Pager_Writer
    return value: Pager_Writer ID
    --------------------------------------------------------------------------*/
    SAPDB_Int RegisterNextPagerWriter(const tsp00_TaskId taskId,
                                      Pager_Writer       *writer);

    /*!-------------------------------------------------------------------------
    function:     ReadConverter
    description:  Read converter from devices for restart

    arguments:    TaskID [in] own taskID 
    return value: none
    --------------------------------------------------------------------------*/
    void ReadConverter(const tsp00_TaskId   taskId);

    /*!-------------------------------------------------------------------------
    function:     IsConverterRestart
    description:  check for converter restart order

    return value: true if converter is in method: ReadConverter
    --------------------------------------------------------------------------*/
    SAPDB_Bool IsConverterRestart() const;

    void AddDataIOTotal(const SAPDB_Int io)
    {
        SAPDBERR_ASSERT_ARGUMENT(io >= 0);
        m_dataIOTotal.Increment(io);
    }

    void AddDataPagesTotal(const SAPDB_Int pages)
    {
        SAPDBERR_ASSERT_ARGUMENT(pages >= 0);
        m_dataPagesTotal.Increment(pages);
    }

    void AddConverterIOTotal(const SAPDB_Int io)
    {
        SAPDBERR_ASSERT_ARGUMENT(io >= 0);
        m_convIOTotal.Increment(io);
    }

    void AddConverterPagesTotal(const SAPDB_Int pages)
    {
        SAPDBERR_ASSERT_ARGUMENT(pages >= 0);
        m_convPagesTotal.Increment(pages);
    }

    void SuspendPagerWriter(SAPDB_UInt pagerIndex,
                            tsp00_TaskId taskId);

    SAPDB_Bool IsSvpActive() const;

    SAPDB_Bool IsPreparePagerWriter() const;

    ePagerOrder GetPagerOrder() const;

    void EnablePagerWriter (const tsp00_TaskId taskId);

    /*!-------------------------------------------------------------------------
    function:     FlushDataCacheRegion
    description:  flush one region of the data cache

    arguments:    TaskID [in] own taskID 
                  regOffset [in] region number
    return value: none
    --------------------------------------------------------------------------*/
    void FlushDataCacheRegion (const tsp00_TaskId taskId,
                               const SAPDB_Int   regOffset);

    /*!-------------------------------------------------------------------------
    function:     FlushConverterParallel
    description:  triggers parallel flush of converter segments
    arguments:    trans                     [int] own taskID
    return value: none
    --------------------------------------------------------------------------*/
    void FlushConverterParallel ( const tsp00_TaskId TaskId,
                                                SAPDB_UInt &numPagesWritten,
                                                SAPDB_UInt &numIOActions);


#ifdef CONSTRUCTOR_MAY_NOT_BE_PRIVATE_AND_NO_FRIEND_DEFINED
protected:
#else
private:
#endif

    Pager_Controller(SAPDBMem_IRawAllocator &allocator);

private:

    #define SET_CHECK_FBM  SAPDB_TRUE

    // copy and assignment operators are not supported
    Pager_Controller& operator=( const Pager_Controller &controller);

    Pager_Controller(const Pager_Controller &controller);


    static void CreateInstance();

    void Initialize();

    void ResumeAllPagerWriter (const tsp00_TaskId pid,
                               const SAPDB_Bool1      set_check_fbm,
                               const SAPDB_Bool       resumeForConverter);

    /*
    void ResumePagerWriterForConverter (const tsp00_TaskId pid,
                                        const SAPDB_UInt   pagerIndex);
    */
    void CleanupActiveCaches();

    void ClearConverterCache (tgg00_TransContext &t,
                              SAPDB_Int          &total_io,
                              SAPDB_Int          &total_pages);

    void WaitForPagerWriterReply (const tsp00_TaskId pid);

    void WriteDataPage( tgg00_TransContext &Trans, tbd_nodeptr &pNode) const;

private:

    static Pager_Controller  *m_Instance;

    RTESync_NamedSpinlock            m_SpinLock;
    Container_Vector<Pager_Writer *> m_pagerWriter;
    Container_Vector<SAPDB_Bool>     m_activePagerWriter;
    const SAPDB_UInt                 m_numPagerWriter;
    SAPDB_UInt                       m_registeredPagerWriter;
    tsp00_TaskId                     m_svpCoordinator;
    SAPDB_Bool                       m_svpActive;
    SAPDB_Bool                       m_preparePagerWriter;
    SAPDB_Bool                       m_pagerWriterDisabled;
    ePagerOrder                      m_pagerOrder;

    RTESync_InterlockedCounter<SAPDB_UInt> m_suspendCnt;
    RTESync_InterlockedCounter<SAPDB_UInt> m_dataIOTotal;
    RTESync_InterlockedCounter<SAPDB_UInt> m_dataPagesTotal;
    RTESync_InterlockedCounter<SAPDB_UInt> m_convIOTotal;
    RTESync_InterlockedCounter<SAPDB_UInt> m_convPagesTotal;
};

#endif  /* PAGER_CONTROLLER_HPP */


