/*!**************************************************************************

  module      : Pager_ISavepoint.hpp

  -------------------------------------------------------------------------

  responsible : Henrik

  special area: Pager
  description : Interface for Pager instantiation and savepoint control
  last changed: 2002-02-21
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002

  ========== licence begin LGPL
  Copyright (C) 2002 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef PAGER_ISAVEPOINT_HPP
#define PAGER_ISAVEPOINT_HPP

class Pager_ISavepoint;

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "ggg00.h"

#include "Converter/Converter_Types.hpp"
#include "DataAccess/Data_Types.hpp"
#include "IOManager/IOMan_BlockAddress.hpp"
#include "KernelCommon/Kernel_Common.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class Pager_ISavepoint {
public:
    // singleton

    /*!-------------------------------------------------------------------------
    function:     GetInstance
    description:  returns instance of Pager_ISavepoint/Pager_Controller. If
      no instance exists a new one will be created
    arguments:    none
    return value: (new) singleton instance of Pager_ISavepoint
    --------------------------------------------------------------------------*/
    static Pager_ISavepoint& GetInstance();

    /*!-------------------------------------------------------------------------
    function:     BeginSavepoint 
    description:  initiate Savepoint,
      wait for running pager_writer to finish
    arguments:    taskID [in] own taskID 
    return value: none
    --------------------------------------------------------------------------*/
    virtual void BeginSavepoint (const tsp00_TaskId taskId) = 0;

    /*!-------------------------------------------------------------------------
    function:     FlushDataCacheBeforePrepare 
    description:  flushes DataCache to disk
    arguments:    taskID [in] own taskID 
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FlushDataCacheBeforePrepare (const tsp00_TaskId taskId,
                                                    SAPDB_UInt &numPagesWritten,
                                                    SAPDB_UInt &numIOActions) = 0;

    /*!-------------------------------------------------------------------------
    function:     WaitForLastSplit
    description:  
    arguments:    trans [in] own transcontext 
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void WaitForLastSplit (tgg00_TransContext &trans,
                                         SAPDB_UInt &numPagesWritten,
                                         SAPDB_UInt &numIOActions) = 0;

    /*!-------------------------------------------------------------------------
    function:     FlushDataCacheAfterPrepare
    description:  flushes DataCache to disk
    arguments:    taskId [in] own taskID 
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FlushDataCacheAfterPrepare (const tsp00_TaskId taskId,
                                                   SAPDB_UInt &numPagesWritten,
                                                   SAPDB_UInt &numIOActions) = 0;

    /*!-------------------------------------------------------------------------
    function:     FlushConverterParallel
    description:  triggers parallel flush of converter segments
    arguments:    taskID                     [in] own taskId
                  numPagesWritten [out] number of written pages
                  numIOActions    [out] number of clustered I/O transactions
    return value: none
    --------------------------------------------------------------------------*/
    virtual void FlushConverterParallel ( const tsp00_TaskId TaskId,
                                                SAPDB_UInt &numPagesWritten,
                                                SAPDB_UInt &numIOActions) = 0;

    /*!-------------------------------------------------------------------------
    function:     EndSavepoint
    description:  Savepoint finished,
    arguments:    taskID [in] own taskID 
               disablepagerwriter [in] disable pager_writer e.g. for shutdown
    return value: none
    --------------------------------------------------------------------------*/
    virtual void EndSavepoint (const tsp00_TaskId taskId,
                               const SAPDB_Bool disablepagerwriter) = 0;

};

#endif  /* PAGER_ISAVEPOINT_HPP */


