/*!***************************************************************************

  module      : RTEConf_ParameterAccess.cpp

  -------------------------------------------------------------------------

  responsible : RobinW

  special area: RTE configuration parameters
  description : class for kernel configurable parameter access


  last changed: 2002-04-29  00:00

  first created:2000-03-22  17:49

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000,2002



    ========== licence begin GPL
    Copyright (C) 2000,200 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/
/*PTS 1115041*/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/


#include "SAPDBCommon/SAPDB_Types.hpp"
#include "RunTime/RTE_Types.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.h"
#include "gsp11.h"
#include "heo11.h"
#include "RunTime/Configuration/RTEConf_BuildParamFileName.h"

#include "SAPDBCommon/SAPDB_ToString.hpp"
#include "hsp77.h"
#include "RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"

#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"

extern SAPDBTrace_Topic Runtime_Trace;

#include "gsp01.h"

#include "geo50_0.h"    // KGS
#include "hsp77.h"      // sp77sprintf()

#ifdef _WIN32           /* & if $OSSPEC = WIN32*/
#include "gos00k.h"     /* nocheck */
#else                   /* & else*/
#include "gen81.h"      /* nocheck */
#endif                  /* & endif*/


#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define SIZE_OF_NUMBER_BUFFER_  256

#define MAX_TASKCLUSTER_PARTS 10

#define MAX_EXCLUSIVE     250
#define MAX_REGIONS       40
#define ACT_PARAM_VERSION 0

#define MIN_DISP_LOOPS     1

#define UNKNOWN_PARAMETER_MESSAGE "unknown parameter found in XParam file: "

#define XP_V_UPDATE_OFFLINE 0
#define XP_V_UPDATE_ONLINE  1

#define XP_V_SET_TASKCLUSTER  0
#define XP_V_SET_DEVSP_NAME   1
#define XP_V_SET_DEVSP_SIZE   2
#define XP_V_SET_M_DEVSP_NAME 3

#define XP_V_BOOL_Y   0  /* Bool type (TRUE if Y or y) (for YES   */
#define XP_V_BOOL_N   1  /* Bool type (TRUE if N or n) (for NO)   */
#define XP_V_BOOL_A   2  /* Bool type (TRUE if A or a) (for AUTO) */
#define XP_V_BOOL_S   3  /* Bool type (TRUE if S or s) (for STACK) */
#define XP_V_LONG     4  /* Signed Long type */
#define XP_V_ULONG    5  /* Unsigned Long type */
#define XP_V_REGIONS  6  /* Unsigned Long Region Count */
#define XP_V_STRING   7  /* Any String type */
#define XP_V_IGNORED  8  /* Accepted, but ignored */
#define XP_V_CRYPT    9  /* Crypted Password */
#define XP_V_SPECIAL_STRING  10 /* String Parameter with side effect */
#define XP_V_SPECIAL_LONG    11 /* Long Parameter with side effect */
#define XP_V_SPECIAL_ULONG   12 /* Unsigned Long Parameter with side effect */
#define XP_V_SPECIAL_SPECIAL 13 /* Other Parameter with side effect */

#define XP_PARAM_CHANGED        "__PARAM_CHANGED___"

#ifdef WIN32
#define LAST_SYSTEM_ERROR_          ((SAPDB_ULong)GetLastError())
#else
#define LAST_SYSTEM_ERROR_          errno
#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define ALLOCATE_ARRAY(TypeOfElement,NumberOfElements)                          \
    (TypeOfElement *)AllocateMemory(sizeof(TypeOfElement) * (NumberOfElements))

#define DEALLOCATE_MEMORY(address)                                              \
{                                                                               \
    RTEMem_RteAllocator::Instance().Deallocate((void *)(address));              \
    address = NULL;                                                             \
}                                                                   
  
#define XP_PTR(_x,_t) (((SAPDB_Byte *)_x)+_t->offset)

#define RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE1 \
	(sizeof(RteXparamTable1)/sizeof(struct xparam_strncmp_table))

#define XP_SZ_OFF(_x)  sizeof(((RTE_XPARAM_REC *)0)->_x), \
							((size_t)&(((RTE_XPARAM_REC *)0)->_x))

#define RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE2 \
	(sizeof(RteXparamTable2)/sizeof(struct xparam_strcmp_table))

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

enum xp_paramType
{
    NXP_INT_TYPE,
    NXP_STRING_TYPE,
    NXP_CRYPT_TYPE
};

struct  RTEConf_ParameterValueRecord_
{
    xp_paramType xp_value_t ;
    union           of_xp_value_t
    {
        RTEConf_Parameter::Integer      xp_integer ;
        RTEConf_Parameter::String       xp_string; 
        RTEConf_Parameter::CryptInfo    xp_crypt ;
    }  xp_type ;
}  ;

typedef struct RTEConf_ParameterValueRecord_   RTEConf_ParameterValueRecord ;

static SAPDB_Char* szDevType[ NUM_DEVSP_TYPES ] = 
{ 
    (SAPDB_Char*)"SYS",  
    (SAPDB_Char*)"MSYS",
    (SAPDB_Char*)"ALOG" , 
    (SAPDB_Char*)"MALOG",
    (SAPDB_Char*)"TLOG",
    (SAPDB_Char*)"DATA", 
    (SAPDB_Char*)"MDATA",
    (SAPDB_Char*)"TRACE" 
} ;

struct xparam_strncmp_table
{
    RTEConf_Parameter::Name name;		/* Name found in XParam file */
    size_t                  len;       /* Maximum length to compare */

    SAPDB_Int4              type;      /* Classification of expected type */
    SAPDB_Int4              UpdateMode;/* update param allowed in off/online state of db */

    RTE_DevSpaceType            dt;
    RTE_DevSpaceType            mdt;
};

static  struct xparam_strncmp_table RteXparamTable1[]=
{
 { (RTEConf_Parameter::Name)PAN_TASKCLUSTER_N ,     PAN_TASKCLUSTER_N_LN,      
		XP_V_SET_TASKCLUSTER, XP_V_UPDATE_OFFLINE, RTE_DevSpaceFiller, RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_SYS_N ,             PAN_SYS_N_LN,              
		XP_V_SET_DEVSP_NAME,  XP_V_UPDATE_OFFLINE, RTE_DevSpaceSystem,    RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_ARCHIVE_LOG_N ,     PAN_ARCHIVE_LOG_N_LN,
		XP_V_SET_DEVSP_NAME,  XP_V_UPDATE_OFFLINE, RTE_DevSpaceLog,    RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_DATADEV_N ,         PAN_DATADEV_N_LN,
		XP_V_SET_DEVSP_NAME,  XP_V_UPDATE_OFFLINE, RTE_DevSpaceData,   RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_SYS_N_SIZE ,        PAN_SYS_N_SIZE_LN,
		XP_V_SET_DEVSP_SIZE,  XP_V_UPDATE_OFFLINE, RTE_DevSpaceSystem,    RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_ARCHIVE_LOG_N_SIZE, PAN_ARCHIVE_LOG_N_SIZE_LN,
		XP_V_SET_DEVSP_SIZE,  XP_V_UPDATE_OFFLINE, RTE_DevSpaceLog,    RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_DATADEV_N_SIZE,     PAN_DATADEV_N_SIZE_LN,
		XP_V_SET_DEVSP_SIZE,  XP_V_UPDATE_OFFLINE, RTE_DevSpaceData,   RTE_DevSpaceFiller    },
 { (RTEConf_Parameter::Name)PAN_M_SYS_N ,           PAN_M_SYS_N_LN,
		XP_V_SET_M_DEVSP_NAME,XP_V_UPDATE_OFFLINE, RTE_DevSpaceSystem,    RTE_DevSpaceMirrorSystem  },
 { (RTEConf_Parameter::Name)PAN_M_ARCHIVE_LOG_N ,   PAN_M_ARCHIVE_LOG_N_LN,
		XP_V_SET_M_DEVSP_NAME,XP_V_UPDATE_OFFLINE, RTE_DevSpaceLog,    RTE_DevSpaceMirrorLog  },
 { (RTEConf_Parameter::Name)PAN_M_DATADEV_N ,       PAN_M_DATADEV_N_LN,
		XP_V_SET_M_DEVSP_NAME,XP_V_UPDATE_OFFLINE, RTE_DevSpaceData,   RTE_DevSpaceMirrorData },
};

struct xparam_strcmp_table
{
    RTEConf_Parameter::Name name;		/* Name found in XParam file */

    SAPDB_Int4              type;       /* Claasification of expected type */
    SAPDB_Int4              UpdateMode;/* update param allowed in off/online state of db */

    size_t                  size;       /* Size in RTE struct */
    SAPDB_ULong             offset;     /* Offset in RTE struct */
};

static struct xparam_strcmp_table RteXparamTable2[] =
{
 { (RTEConf_Parameter::Name)PAN_KERNELTRACEFILE,    XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szTraceDevspace)    },
 { (RTEConf_Parameter::Name)PAN_KERNELTRACESIZE,    XP_V_IGNORED,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lTraceDevspaceSize) },
 { (RTEConf_Parameter::Name)PAN_MAXDEVSPACES,       XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxDevspaces)      },
 { (RTEConf_Parameter::Name)PAN_MAXSYSDEVSPACES,    XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxSysDevspaces)   },
 { (RTEConf_Parameter::Name)PAN_MAXARCHIVELOGS,     XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxArchiveLogs)    },
 { (RTEConf_Parameter::Name)PAN_PACKET_SIZE,        XP_V_ULONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulPacketSize)       },
 { (RTEConf_Parameter::Name)PAN_MIN_REPLY_SIZE,     XP_V_ULONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMinReplySize)     },
 { (RTEConf_Parameter::Name)PAN_MAXUSERSTACK,       XP_V_ULONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxUsUtStack)     },
 { (RTEConf_Parameter::Name)PAN_TREE_REGIONS,       XP_V_ULONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumTreeReg)       },
 { (RTEConf_Parameter::Name)PAN_MP_DISP_LOOPS,      XP_V_ULONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulDispLoopsBefSemop)},
 { (RTEConf_Parameter::Name)PAN_LOG_MODE,           XP_V_IGNORED,   XP_V_UPDATE_ONLINE,  0, 0 },
 { (RTEConf_Parameter::Name)PAN_MAXDATADEVSPACES,   XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxDataDevspaces)  },
 { (RTEConf_Parameter::Name)PAN_PAGE_SIZE,          XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, 0, 0 },
   /* PTS 1109850 */
 { (RTEConf_Parameter::Name)PAN_FORMATTING_MODE,    XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, 0, 0 },

#if defined(_WIN32) 

#else
 { (RTEConf_Parameter::Name)PAN_GATEWAYTYPE,        XP_V_IGNORED,   XP_V_UPDATE_OFFLINE, 0, 0 },
#define XP_T_GATEWAYTYPE              15
#endif

/*
 * If there are any other XParam values ignored here, but used in kernel
 * add them here...
 */
 { (RTEConf_Parameter::Name)PAN_SERVERDB,           XP_V_IGNORED,  XP_V_UPDATE_OFFLINE,   0, 0                         },
 { (RTEConf_Parameter::Name)PAN_CONTROLPW,          XP_V_CRYPT,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(crypControlPassw)    },
 { (RTEConf_Parameter::Name)PAN_UNICODE_CATALOG,    XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUnicode)            },
 { (RTEConf_Parameter::Name)PAN_SUBTREE,            XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fSubtree)            },
 { (RTEConf_Parameter::Name)PAN_USE_ASYNC_IO,       XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseAsyncIO)         },
 { (RTEConf_Parameter::Name)PAN_EVENTPRIO,          XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fEventPrio)          },
 { (RTEConf_Parameter::Name)PAN_MAXBACKUPDEVS,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lNoOfBackupDevs)     },
 { (RTEConf_Parameter::Name)PAN_MAXTRANS,           XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxTrans)           },
 { (RTEConf_Parameter::Name)PAN_ALLOW_LBACK_TO_PIPE,XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fLogBackToPipe)      }, /* PTS 1108249 */
 { (RTEConf_Parameter::Name)PAN_DATA_CACHE_PAGS,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lDataCachePages)     },
 { (RTEConf_Parameter::Name)PAN_SEQUENCE_CACHE,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lSequenceCache)      },
 { (RTEConf_Parameter::Name)PAN_CATALOG_CACHE,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lCatalogCachePages)  },
 { (RTEConf_Parameter::Name)PAN_SERVER_DESC_CACHE,  XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lDistribDescCache)   },
 { (RTEConf_Parameter::Name)PAN_SERVER_CMD_CACHE,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lServerCmdCache)     },
 { (RTEConf_Parameter::Name)PAN_MIN_SERVER_DESC,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMinServerDesc)      },
 { (RTEConf_Parameter::Name)PAN_MAXSERVERDB,        XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxServerDB)        },
 { (RTEConf_Parameter::Name)PAN_LOG_IO_QUEUE,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogQueuePages)      },
 { (RTEConf_Parameter::Name)PAN_MAXLOCKS,           XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxLocks)           },
 { (RTEConf_Parameter::Name)PAN_CONV_CACHE_PAGES,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lConvCachePages)     },
 { (RTEConf_Parameter::Name)PAN_FBM_CACHE_PAGES,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lFbmCachePages)      },
 { (RTEConf_Parameter::Name)PAN_USM_CACHE_PAGES,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lUsmCachePages)      },
 { (RTEConf_Parameter::Name)PAN_BACKUP_BLOCK_CNT,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lBackUpBlockCount)   },
 { (RTEConf_Parameter::Name)PAN_DYNDATA,            XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lSharedDynData)      },
 { (RTEConf_Parameter::Name)PAN_DYNPOOL,            XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lSharedDynPool)      },
 { (RTEConf_Parameter::Name)PAN_MAXIOTHREADS,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lDeviPerDevspace)    },
 { (RTEConf_Parameter::Name)PAN_IOTHREADS_FOR_PRIO, XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lPrioDeviPerDevspace) },
 { (RTEConf_Parameter::Name)PAN_MAXDATAPAGES,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMaxDataPages)       },
 { (RTEConf_Parameter::Name)PAN_REALTIME_PRIO,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lRealtimePrio)       },
 { (RTEConf_Parameter::Name)PAN_LOG_SEGMENT_SIZE,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLogSegmentSize)     },
 { (RTEConf_Parameter::Name)PAN_LOCK_TIMEOUT,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lLockTimeout)        },
 { (RTEConf_Parameter::Name)PAN_REQUEST_TIMEOUT,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lRequestTimeout)     },
 { (RTEConf_Parameter::Name)PAN_NET_TIMEOUT,        XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lNetTimeout)         },
 { (RTEConf_Parameter::Name)PAN_INDEXFILE_LIST_SIZE,XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lIndexfileListSize)  },
 { (RTEConf_Parameter::Name)PAN_USE_PROC_SCOPE,     XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseProcessScope)    },  /* PTS 1109046 */
 { (RTEConf_Parameter::Name)PAN_MAXACTIVE,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxUser)           },
 { (RTEConf_Parameter::Name)PAN_MAXSERVER,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxServer)         },
 { (RTEConf_Parameter::Name)PAN_LOCK_SUPPLY_BLOCK,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulLockSupplyBlock)   },
 { (RTEConf_Parameter::Name)PAN_DIAG_SIZE,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulDiagSize)          },
 { (RTEConf_Parameter::Name)PAN_UTILITY_PROTSIZE,   XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulUtilityProtSize)   },
 { (RTEConf_Parameter::Name)PAN_MAXCPU,             XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxCPU)            },
 { (RTEConf_Parameter::Name)PAN_PNOPOOLSIZE,        XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulPnoPoolSize)       },
 { (RTEConf_Parameter::Name)PAN_MAXDATAWRITER,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxDataWriter)     },
 { (RTEConf_Parameter::Name)PAN_MB_DATA_SIZE,       XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMBDataSize)        },
 { (RTEConf_Parameter::Name)PAN_MB_QUAL_SIZE,       XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMBQualSize)        },
 { (RTEConf_Parameter::Name)PAN_MB_STACK_SIZE,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMBStackSize)       },
 { (RTEConf_Parameter::Name)PAN_KB_WORK_ST_SIZE,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulKBWorkStSize)      },
 { (RTEConf_Parameter::Name)PAN_KB_WORKBUF_SIZE,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulKBWorkBufSize)     },
 { (RTEConf_Parameter::Name)PAN_CATALOG_CACHE_SIZE, XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulCatalogCacheSize)  },
 { (RTEConf_Parameter::Name)PAN_XPARAM_CHANGED,     XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulXparamChanged)     },
 { (RTEConf_Parameter::Name)PAN_EVENT_SIZE,         XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulEventSize)         },
 { (RTEConf_Parameter::Name)PAN_MAXEVENTTASKS,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxEventTasks)     },
 { (RTEConf_Parameter::Name)PAN_MAX_BACKUP_TASKS,   XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxBackupTasks)    },
 { (RTEConf_Parameter::Name)PAN_MAXGARBAGE_COLLECTOR, XP_V_ULONG,  XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxGarbageCollector) },
 { (RTEConf_Parameter::Name)PAN_MAXEVENTS,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxEvents)         },
 { (RTEConf_Parameter::Name)PAN_READAHEAD_BLOBS,    XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulReadAheadBlobsGT)  },
 { (RTEConf_Parameter::Name)PAN_DATACACHE_REGIONS,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumDataCacheReg)   },
 { (RTEConf_Parameter::Name)PAN_TRANS_REGIONS,      XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumTransReg)       },
 { (RTEConf_Parameter::Name)PAN_TAB_REGIONS,        XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumTabReg)         },
 { (RTEConf_Parameter::Name)PAN_ROW_REGIONS,        XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumRowReg)         },
 { (RTEConf_Parameter::Name)PAN_OMS_REGIONS,        XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulNumOMSReg)         },
 { (RTEConf_Parameter::Name)PAN_EVENT_ALIVE_CYCLE,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulEventAliveCycle)   },
 { (RTEConf_Parameter::Name)PAN_PARAM_KERNELVERSION,XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szKernelVersion)     },
 { (RTEConf_Parameter::Name)PAN_CONTROLID,          XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szControlUser)       },
 { (RTEConf_Parameter::Name)PAN_DIAGFILE,           XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szDiagFile)          },
 { (RTEConf_Parameter::Name)PAN_BACKUP_HISTFILE,    XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szBackupHistFile)    },
 { (RTEConf_Parameter::Name)PAN_BACKUP_MED_DEF,     XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szBackupMedDefFile)  },
 { (RTEConf_Parameter::Name)PAN_UTILITY_PROTFILE,   XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szUtilityProtFile)   },
 { (RTEConf_Parameter::Name)PAN_RUNDIR,             XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szRunDirectory)      },
 { (RTEConf_Parameter::Name)PAN_KERNELDUMPFILE,     XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szDumpFilename)      },
 { (RTEConf_Parameter::Name)PAN_RTEDUMPFILE,        XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szRTEDumpFilename)   },
 { (RTEConf_Parameter::Name)PAN_EVENTFILE,          XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szEventFile)         },
 { (RTEConf_Parameter::Name)PAN_DEFAULT_CODE,       XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szDefaultCode)       },
 { (RTEConf_Parameter::Name)PAN_MP_RGN_QUEUE,       XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fExclQueue)          },
 { (RTEConf_Parameter::Name)PAN_MAX_MESSAGE_FILES,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(ulMaxMessageFiles)   },
/* PTS 1105389 */
 { (RTEConf_Parameter::Name)PAN_INSTANCE_TYPE,      XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szInstanceType)      },
 { (RTEConf_Parameter::Name)PAN_DEVNO_BIT_COUNT ,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(devnoBitCount)       },
 { (RTEConf_Parameter::Name)PAN_CONVERTER_REGIONS,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(converterRegions)    },
 { (RTEConf_Parameter::Name)PAN_STAT_FREEPOOL_SIZE ,XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(statFreepoolSize)    },
 { (RTEConf_Parameter::Name)PAN_DYN_FREEPOOL_SIZE , XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(dynFreepoolSize)     },
 { (RTEConf_Parameter::Name)PAN_CACHE_SIZE ,        XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(cacheSize)           },
 { (RTEConf_Parameter::Name)PAN_CRASH_HISTORY_NUM,  XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uCrashHistoryNum)    },
 { (RTEConf_Parameter::Name)PAN_CRASH_HISTORY_PATH, XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szCrashHistoryPath)  },
 { (RTEConf_Parameter::Name)PAN_CRASH_SEM,          XP_V_ULONG,    XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uCrashSem)           },

 /* PTS 1108976 */
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_TI,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesTI)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_GC,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesGC)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_AL,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesAL)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_DW,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesDW)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_US,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesUS)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_UT,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesUT)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_SV,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesSV)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_EV,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesEV)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_TW,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesTW)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_CS,     XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesCS)        },
 { (RTEConf_Parameter::Name)PAN_TRACE_PAGES_BUP,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(tracePagesBUP)       },
 { (RTEConf_Parameter::Name)PAN_LOAD_BALANCING_CHK, XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uLoadBalancingCheck) },
 { (RTEConf_Parameter::Name)PAN_LOAD_BALANCING_DIFF,XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uLoadBalancingDiff)  },
 { (RTEConf_Parameter::Name)PAN_LOAD_BALANCING_EQUALITY,XP_V_LONG, XP_V_UPDATE_OFFLINE, XP_SZ_OFF(uLoadBalancingEquality)},

/***************************************************************/
/* OS Depending Parameter                                      */
/***************************************************************/
#if defined(_WIN32)
 { (RTEConf_Parameter::Name)PAN_ALLOW_COMPR_DEVSP,  XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fAllowCompressedDevsp) },
 { (RTEConf_Parameter::Name)PAN_DYNAMIC_SERVER,     XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fDynamicServerTasks)   },
 { (RTEConf_Parameter::Name)PAN_RESTART_SHUTDOWN,   XP_V_BOOL_A,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fAutoRestartShutdown)  },
 { (RTEConf_Parameter::Name)PAN_USE_MEM_ENHANCEMENT,XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseMemEnhancement)    },
 { (RTEConf_Parameter::Name)PAN_MEM_ENHANCEMENT_LIM,XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lMemEnhancementLimit)  },
 { (RTEConf_Parameter::Name)PAN_USE_FIBERS,         XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseFibers)            },
 { (RTEConf_Parameter::Name)PAN_HIRES_TIMER_TYPE,   XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szHiResTimerType)      },
#else
 { (RTEConf_Parameter::Name)PAN_DYNAMIC_TASK_STACK, XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fDynTaskStack)         },
 { (RTEConf_Parameter::Name)PAN_TASK_STACK_AREA,    XP_V_IGNORED,  XP_V_UPDATE_OFFLINE, 0, 0                             },
 { (RTEConf_Parameter::Name)PAN_OPMSG1,             XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szOpMsg1)              },
 { (RTEConf_Parameter::Name)PAN_OPMSG2,             XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szOpMsg2)              },
 { (RTEConf_Parameter::Name)PAN_SHMCHUNK,           XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lShmChunk)             },
 { (RTEConf_Parameter::Name)PAN_SHMKERNEL,          XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(lShmKernel)            },
 { (RTEConf_Parameter::Name)PAN_USE_COROUTINES,     XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseCoroutines)        },
 { (RTEConf_Parameter::Name)PAN_USE_STACK_ON_STACK, XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseStackOnStack)      },
 { (RTEConf_Parameter::Name)PAN_USE_UCONTEXT,       XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fUseUcontext)          },
#endif
/***************************************************************/
/* End of OS Depending Parameter                               */
/***************************************************************/
/***************************************************************/
/* ONLINE changable Parameter     using "diagnose" commands    */
/***************************************************************/
 { (RTEConf_Parameter::Name)PAN_DELAY_LW,           XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDelayLW)            },
 { (RTEConf_Parameter::Name)PAN_FORMAT_DATAVOLUME,  XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fFormatDataVolume)   },
 { (RTEConf_Parameter::Name)PAN_DELAY_COMMIT,       XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fDelayCommit)        },

 { (RTEConf_Parameter::Name)PAN_SERVERDB_FOR_SAP,   XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fServerDBForSAP)     },
 { (RTEConf_Parameter::Name)PAN_DATE_TIME_FORMAT,   XP_V_STRING,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(szDateTimeFormat)    },
 { (RTEConf_Parameter::Name)PAN_MULT_IO_BLOCK_CNT,  XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulMultIOBlockCount)  },
 { (RTEConf_Parameter::Name)PAN_RESTART_TIME,       XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lRestartTime)        },
 { (RTEConf_Parameter::Name)PAN_DEADLOCK_DETECTION, XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulDeadlockDetection) },
 { (RTEConf_Parameter::Name)PAN_SESSION_TIMEOUT,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lSessionTimeout)     },
 { (RTEConf_Parameter::Name)PAN_USE_IOTHREAD_ONLY,  XP_V_BOOL_N,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fAllowSingleIO)      },
 { (RTEConf_Parameter::Name)PAN_DEV_QUEUE,          XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDevQueue)           },
 { (RTEConf_Parameter::Name)PAN_DEFAULT_LRU,        XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fDefaultLRU)         },
 { (RTEConf_Parameter::Name)PAN_MP_RGN_DIRTY_READ,  XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fBegExclDirtyRead)   },
 { (RTEConf_Parameter::Name)PAN_MP_RGN_BUSY_WAIT,   XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fEndExclBusyWait)    },
 { (RTEConf_Parameter::Name)PAN_MP_DISP_PRIO,       XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fDynDispQueSrch)     },
 { (RTEConf_Parameter::Name)PAN_MP_RGN_LOOP,        XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulRegionCollisionLoop) },
 { (RTEConf_Parameter::Name)PAN_MP_RGN_PRIO,        XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fBegExclQueWhenPrio) },
 { (RTEConf_Parameter::Name)PAN_MAXRGN_REQUEST,     XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulRegLockSlice)      },
 { (RTEConf_Parameter::Name)PAN_PRIO_BASE_U2U,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseU2U)        },
 { (RTEConf_Parameter::Name)PAN_PRIO_BASE_IOC,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseIOC)        },
 { (RTEConf_Parameter::Name)PAN_PRIO_BASE_RAV,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseRAV)        },
 { (RTEConf_Parameter::Name)PAN_PRIO_BASE_COM,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseCOM)        },
 { (RTEConf_Parameter::Name)PAN_PRIO_BASE_REX,      XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioBaseREX)        },
 { (RTEConf_Parameter::Name)PAN_PRIO_FACTOR,        XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lPrioFactor)         },
 { (RTEConf_Parameter::Name)PAN_SVP_1_CONV_FLUSH,   XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fSvpFstConvFlush)    },
 { (RTEConf_Parameter::Name)PAN_DW_COPY_BOUNDARY,   XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwCopyBoundary)     },
 { (RTEConf_Parameter::Name)PAN_DW_IO_AREA_SIZE,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwIOAreaSize)       },
 { (RTEConf_Parameter::Name)PAN_DW_IO_AREA_FLUSH,   XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwIOAreaFlush)      },
 { (RTEConf_Parameter::Name)PAN_FBM_DEVSP_COMPR,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFbmCompression)     },
 { (RTEConf_Parameter::Name)PAN_FBM_DEVSP_BALANCE,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFbmBalance)         },
 { (RTEConf_Parameter::Name)PAN_FBM_LOW_IO_RATE,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lFbmLowIORate)       },
 { (RTEConf_Parameter::Name)PAN_COPY_IO_BLOCK_CNT,  XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulDwIOBufSize)       },
 { (RTEConf_Parameter::Name)PAN_DW_LRU_TAIL_FLUSH,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lDwLRUTailFlush)     },
 { (RTEConf_Parameter::Name)PAN_JOIN_SEARCH_LEVEL,  XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinSearchLevel)    },
 { (RTEConf_Parameter::Name)PAN_JOIN_MAXTAB_LEVEL4, XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinMaxTabLevel4)   },
 { (RTEConf_Parameter::Name)PAN_JOIN_MAXTAB_LEVEL9, XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinMaxTabLevel9)   },
 { (RTEConf_Parameter::Name)PAN_OPTIM_MAX_MERGE,    XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lMaxMerge)           },
 { (RTEConf_Parameter::Name)PAN_OPTIM_INV_ONLY,     XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fOptimInvOnly)       },
 { (RTEConf_Parameter::Name)PAN_OPTIM_CACHE,        XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fOptimCache)         },
 { (RTEConf_Parameter::Name)PAN_OPTIM_JOIN_FETCH  , XP_V_LONG,     XP_V_UPDATE_ONLINE,  XP_SZ_OFF(lJoinPrefetchPercent)},
 { (RTEConf_Parameter::Name)PAN_AK_DUMP_ALLOWED,    XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(fAKDumpAllowed)     },
 { (RTEConf_Parameter::Name)PAN_OMS_HEAP_LIMIT  ,   XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOMSHeapLimit) },
 /* PTS 1110149 */
 { (RTEConf_Parameter::Name)PAN_OMS_HEAP_THRESHOLD, XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOmsHeapUsageThreshold) },
 { (RTEConf_Parameter::Name)PAN_OMS_VERS_THRESHOLD, XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOmsVersionFreeThreshold) },
 { (RTEConf_Parameter::Name)PAN_HEAP_CHECK_LEVEL,   XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulHeapCheckLevel) },
 { (RTEConf_Parameter::Name)PAN_OMS_HEAP_COUNT,     XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOmsHeapCount) },
 { (RTEConf_Parameter::Name)PAN_OMS_HEAP_BLOCKSIZE, XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOmsHeapBlocksize) },
 { (RTEConf_Parameter::Name)PAN_SHOW_MAX_STACK_USE, XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(showMaxStackUse)     },
 { (RTEConf_Parameter::Name)PAN_MMAP_MEM_POOL,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(mmapMemPoolSize)     },
 { (RTEConf_Parameter::Name)PAN_SHM_MEM_POOL,       XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(shmMemPoolSize)     },
 { (RTEConf_Parameter::Name)PAN_VALLOC_MEM_POOL,    XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(vallocMemPoolSize)     },
 { (RTEConf_Parameter::Name)PAN_HEAP_MEM_POOL,      XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(heapMemPoolSize)     },
 { (RTEConf_Parameter::Name)PAN_MEMORY_SLOT_SIZE,   XP_V_LONG,     XP_V_UPDATE_OFFLINE, XP_SZ_OFF(memorySlotSize)     },
 /* PTS 1111789 */
 { (RTEConf_Parameter::Name)PAN_COMMENT,            XP_V_IGNORED,  XP_V_UPDATE_OFFLINE,   0, 0                         },
 /* PTS 1113405 */
 { (RTEConf_Parameter::Name)PAN_SUPPRESS_CORE,      XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(suppressCore) },
 /* PTS 1115635 */
 { (RTEConf_Parameter::Name)PAN_OMS_STREAM_TIMEOUT, XP_V_ULONG,    XP_V_UPDATE_ONLINE,  XP_SZ_OFF(ulOmsStreamTimeout) },
 /* PTS 1116151 */
 { (RTEConf_Parameter::Name)PAN_OFFICIAL_NODE,      XP_V_STRING,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(szOfficialNode) },
 /* PTS 1118987 */
 { (RTEConf_Parameter::Name)PAN_ALLOW_DEMANGLE,     XP_V_BOOL_Y,   XP_V_UPDATE_ONLINE,  XP_SZ_OFF(allowDemangle)     },
 /* PTS 1120119 */
 { (RTEConf_Parameter::Name)PAN_LOG_MIRRORED,       XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fLogMirrored)        },
 { (RTEConf_Parameter::Name)PAN_MIRR_DATA,          XP_V_BOOL_Y,   XP_V_UPDATE_OFFLINE, XP_SZ_OFF(fMirrData)           },
 };


/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

static inline void *AllocateMemory(size_t size);

static SAPDB_Bool RTEConf_AnalyzeParam  ( RTE_XPARAM_REC               *XParam,
                                          RTEConf_Parameter::Name      name,
                                          RTEConf_ParameterValueRecord *xpr ,
                                          SAPDB_Bool                   inOnlineMode,
                                          SAPDB_Bool                   &fGotControlPw,
                                          SAPDBErr_MessageList         &err) ;

static SAPDB_Int4 RTEConf_FindParam ( RTEConf_Parameter::Name             name ,
                                      RTEConf_ParameterValueRecord        *xpr ,
                                      SAPDBErr_MessageList &err);

static void   RTEConf_InitParam          ( RTE_XPARAM_REC        *XParam  );

static void   RTEConf_ReadSpecialParameters(RTE_XPARAM_REC *XParam);

static void RTEConf_ModifyParameters(RTE_XPARAM_REC *XParam);

static void   RTEConf_SetTaskCluster ( RTE_XPARAM_REC            *XParam,
                                       RTEConf_Parameter::String szTaskCluster,
                                       SAPDB_ULong               ulTCNo );

static SAPDB_Bool RTEConf_SetDevspName  ( RTE_XPARAM_REC               *XParam,
                                          RTEConf_Parameter::Name      DevspName ,
                                          RTEConf_ParameterValueRecord *xpr ,
                                          SAPDB_Int                    lDevType ,
                                          SAPDB_ULong                  ulDevspNo ,
                                          SAPDBErr_MessageList         &err);

static SAPDB_Bool RTEConf_SetDevspSize  ( RTE_XPARAM_REC *XParam,
                                          SAPDB_Long     lDevType ,
                                          SAPDB_ULong    ulDevspNo ,
                                          SAPDB_Int      ulDevspSize );

static RTEConf_Parameter::Integer  RTEConf_XParamInt ( RTEConf_ParameterValueRecord  *xpr ,
                                                       SAPDBErr_MessageList          &err) ;

static void  RTEConf_AllocDevspDescr ( RTE_XPARAM_REC *XParam,
                                       SAPDB_Long     lDevspType,
                                       SAPDB_Int      ulNumDevsp);

externC SAPDB_Bool RTEConf_CheckDuplicateDevSpaceName( DEVSPACES_DESCR *devspacesArray,
                                                       SAPDB_Char *szDevspaceName );



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/


/*!
  Function: RTEConf_CheckDuplicateDevSpaceName
  Description: Check if given names has duplicate in devspacesArray
  Arguments: devspacesArray [in] Array of devspace description arrays
             szDevspaceName [in] Devspacename to check
  Return value: true is no duplicate, false if duplicate
 */
externC SAPDB_Bool RTEConf_CheckDuplicateDevSpaceName( DEVSPACES_DESCR *devspacesArray,
                                                       SAPDB_Char      *szDevspaceName )
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_CheckDuplicateDevSpaceName", Runtime_Trace, 5);

    SAPDB_Bool   Ok = true;;
    SAPDB_Int4   devSpaceIterator;
    for(devSpaceIterator = (SAPDB_Int4)RTE_DevSpaceSystem ; devSpaceIterator <= (SAPDB_Int4)RTE_DevSpaceTrace ; devSpaceIterator++)
    {
        DEVSPACES_DESCR   *pDevspDesc;
        SAPDB_UInt4        ulDevNo;
        SAPDB_Char      *pDevspaceName;

        pDevspDesc = devspacesArray + devSpaceIterator;

        for ( ulDevNo = 0; ulDevNo < pDevspDesc->ulNumDevs; ulDevNo++ )
        {
            if (pDevspDesc->Devspaces[ulDevNo].lDevspaceType == DEVSPACE_UNUSED) 
                continue;

            pDevspaceName = (SAPDB_Char *)pDevspDesc->Devspaces[ulDevNo].szDevspaceName;

#if defined (_WIN32)
            Ok = (_stricmp( pDevspaceName, szDevspaceName ) != 0);
#else
            Ok = (strcmp( pDevspaceName, szDevspaceName ) != 0);
#endif

            if ( !Ok )
            {
                return Ok;
            }
        }
    }

    return Ok;
} // RTEConf_CheckDuplicateDevSpaceName

/* some parameters have to be read before others can be read, for example MAXARCHIVELOGS must be read before ARCHIVE_LOG_* can be read */

static void RTEConf_ReadSpecialParameters(RTE_XPARAM_REC *XParam)
{
    RTEConf_Parameter::Integer valueInt;
    RTEConf_Parameter::String valueString;
    SAPDBErr_MessageList err;
    
/*
    if((RTEConf_ParameterAccess::Instance())->GetInteger((RTEConf_Parameter::Name)PAN_PAGE_SIZE,valueInt,err))
    {
		if ( valueInt != XParam->ulDBPageSize )
		{
        // put a trace output here...
        }
    }
 */

    if((RTEConf_ParameterAccess::Instance())->GetString((RTEConf_Parameter::Name)PAN_FORMATTING_MODE,valueString,err))
    {
		switch ( valueString[0] )
		{
            case 'a' :
            case 'A' : 
                XParam->ulFormattingMode = RTE_FORMATTING_MODE_AUTO;
                break ;
            case 'p' :
            case 'P' : 
                XParam->ulFormattingMode = RTE_FORMATTING_MODE_PARALLEL;
                break ;
            case 's' :
            case 'S' : 
                XParam->ulFormattingMode = RTE_FORMATTING_MODE_SERIAL;
                break ;

            default  : 
                RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_XP_INVALID_FORMATTING_MODE,(SAPDB_Char*)(valueString)));
            break;
		}
    }
    if((RTEConf_ParameterAccess::Instance())->GetInteger((RTEConf_Parameter::Name)PAN_MAXDEVSPACES,valueInt,err))
    {
        XParam->lMaxDevspaces = (SAPDB_Int)valueInt + 1;/* Add TRACE_DEVSPACE */
    }

    /* Still needed, but only for inplace migration... */
    if((RTEConf_ParameterAccess::Instance())->GetInteger((RTEConf_Parameter::Name)PAN_MAXSYSDEVSPACES,valueInt,err))
    {
        XParam->lMaxSysDevspaces = (SAPDB_Int)valueInt;
    }
    if ( XParam->lMaxSysDevspaces > 0 )
    {
        RTEConf_AllocDevspDescr ( XParam, RTE_DevSpaceSystem,
            XParam->lMaxSysDevspaces ) ;
        if ( XParam->fMirrData )
        {
            RTEConf_AllocDevspDescr ( XParam, RTE_DevSpaceMirrorSystem,
                XParam->lMaxSysDevspaces);
        }
    }

    if((RTEConf_ParameterAccess::Instance())->GetInteger((RTEConf_Parameter::Name)PAN_MAXARCHIVELOGS,valueInt,err))
    {
        XParam->lMaxArchiveLogs = (SAPDB_Int)valueInt;
        RTEConf_AllocDevspDescr ( XParam, 
                                  RTE_DevSpaceLog,
                                  XParam->lMaxArchiveLogs);
        RTEConf_AllocDevspDescr ( XParam,
                                  RTE_DevSpaceMirrorLog,
                                  XParam->lMaxArchiveLogs );
    }

    if((RTEConf_ParameterAccess::Instance())->GetInteger((RTEConf_Parameter::Name)PAN_MAXDATADEVSPACES,valueInt,err))
    {
        XParam->lMaxDataDevspaces = (SAPDB_Int)valueInt;
        RTEConf_AllocDevspDescr ( XParam,
                                  RTE_DevSpaceData,
                                  XParam->lMaxDataDevspaces ) ;
        RTEConf_AllocDevspDescr ( XParam,
                                  RTE_DevSpaceMirrorData,
                                  XParam->lMaxDataDevspaces);
    }
    if((RTEConf_ParameterAccess::Instance())->GetString((RTEConf_Parameter::Name)PAN_LOG_MODE,valueString,err))
    {
        switch ( valueString[1] )
        {
        case 'u' :
        case 'U' : 
            XParam->lLogMode = csp2_dual_log  ;
            break ;
        case 'i' :
        case 'I' : 
            XParam->lLogMode = csp2_single_log ;
            break ;
        case 'e' :
        case 'E' : 
            XParam->lLogMode = csp2_demo_log  ;
            break ;
        default  : 
            RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_XP_INVALID_LOG_MODE,(SAPDB_Char *)(valueString)));
            break;
        }
    }

    RTEConf_AllocDevspDescr ( XParam, RTE_DevSpaceTrace, 1 ) ;

    if((RTEConf_ParameterAccess::Instance())->GetInteger((RTEConf_Parameter::Name)PAN_KERNELTRACESIZE,valueInt,err))
    {
        XParam->lTraceDevspaceSize = (SAPDB_Int)valueInt;
        if ( XParam->DevspDescr [ RTE_DevSpaceTrace ].Devspaces == NULL )
            RTEConf_AllocDevspDescr ( XParam, RTE_DevSpaceTrace, 1 ) ;
        XParam->DevspDescr [ RTE_DevSpaceTrace ].Devspaces[0].lDevspaceSize =
            (SAPDB_Int)valueInt;
    }
    
    if((RTEConf_ParameterAccess::Instance())->GetString((RTEConf_Parameter::Name)PAN_KERNELTRACEFILE,valueString,err))
    {
        strncpy(XParam->szTraceDevspace,(char *)valueString,sizeof(tsp2_devnamec));
        if ( XParam->DevspDescr [ RTE_DevSpaceTrace ].Devspaces == NULL )
            RTEConf_AllocDevspDescr ( XParam, RTE_DevSpaceTrace, 1 ) ;
        strncpy(XParam->DevspDescr [ sp2dt_trace ].Devspaces[0].szDevspaceName,(char *)valueString,sizeof(tsp2_devnamec));
        XParam->DevspDescr [ RTE_DevSpaceTrace ].Devspaces[0].lDevspaceType =
            RTE_DevSpaceTrace ;
    }
} // RTEConf_ReadSpecialParameters

//---------------------------------------------------------------

static void RTEConf_ModifyParameters(RTE_XPARAM_REC *XParam)
{
    XParam->ulMaxUsUtStack  <<= 10;
    
    XParam->ulPacketSize    &= ~7;
    XParam->ulMinReplySize  &= ~7;

	if ( XParam->ulDispLoopsBefSemop < 2 )
		XParam->ulDispLoopsBefSemop = 2;

    XParam->ulNoOfRegions += XParam->ulNumDataCacheReg ;
    XParam->ulNoOfRegions += XParam->ulNumTransReg ;
    XParam->ulNoOfRegions += XParam->ulNumTabReg ;
    XParam->ulNoOfRegions += XParam->ulNumRowReg ;
    XParam->ulNoOfRegions += XParam->ulNumOMSReg ;
    XParam->ulNoOfRegions += XParam->converterRegions ;
    
    XParam->ulNoOfRegions += XParam->ulNumTreeReg ;
					/* NumSplitt = NumTree */
	XParam->ulNoOfRegions += XParam->ulNumTreeReg ;

    if (XParam->fLogMirrored || (csp2_dual_log == XParam->lLogMode) )
    {
        SAPDB_ULong numberOfLogDevSpaces = XParam->DevspDescr[RTE_DevSpaceLog].ulNumDevs;
        for(SAPDB_ULong currentDevSpace = 0;currentDevSpace < numberOfLogDevSpaces;currentDevSpace++)
        {
            XParam->DevspDescr[RTE_DevSpaceMirrorLog].Devspaces[currentDevSpace].lDevspaceSize
                = XParam->DevspDescr[RTE_DevSpaceLog].Devspaces[currentDevSpace].lDevspaceSize;
        }
    }
    else
    {
        RTEMem_RteAllocator::Instance().Deallocate(XParam->DevspDescr[RTE_DevSpaceMirrorLog].Devspaces);
        XParam->DevspDescr[RTE_DevSpaceMirrorLog].Devspaces = 0;
        XParam->DevspDescr[RTE_DevSpaceMirrorLog].ulNumDevs = 0 ;
    }

    if (XParam->fMirrData)
    {
        SAPDB_ULong numberOfSysDevSpaces = XParam->lMaxSysDevspaces;
        for(SAPDB_ULong currentSysDevSpace = 0;currentSysDevSpace < numberOfSysDevSpaces;currentSysDevSpace++)
        {
            XParam->DevspDescr[RTE_DevSpaceMirrorSystem].Devspaces[currentSysDevSpace].lDevspaceSize
                = XParam->DevspDescr[RTE_DevSpaceSystem].Devspaces[currentSysDevSpace].lDevspaceSize;
        }
        SAPDB_ULong numberOfDataDevSpaces = XParam->lMaxDataDevspaces;
        for(SAPDB_ULong currentDataDevSpace = 0;currentDataDevSpace < numberOfDataDevSpaces;currentDataDevSpace++)
        {
            XParam->DevspDescr[RTE_DevSpaceMirrorData].Devspaces[currentDataDevSpace].lDevspaceSize
                = XParam->DevspDescr[RTE_DevSpaceData].Devspaces[currentDataDevSpace].lDevspaceSize;
        }
    }
    else
    {
        RTEMem_RteAllocator::Instance().Deallocate(XParam->DevspDescr[RTE_DevSpaceMirrorSystem].Devspaces);
        XParam->DevspDescr[RTE_DevSpaceMirrorSystem].Devspaces = 0;
        XParam->DevspDescr[RTE_DevSpaceMirrorSystem].ulNumDevs = 0;
        RTEMem_RteAllocator::Instance().Deallocate(XParam->DevspDescr[RTE_DevSpaceMirrorData].Devspaces);
        XParam->DevspDescr[RTE_DevSpaceMirrorData].Devspaces = 0;
        XParam->DevspDescr[RTE_DevSpaceMirrorData].ulNumDevs = 0 ;
    }
} // RTEConf_ModifyParameters

//---------------------------------------------------------------

externC SAPDB_Bool RTEConf_ReadAndAnalyzeParam ( tsp00_DbNamec VAR_ARRAY_REF    ServerDB,
                                                 RTE_XPARAM_REC                 *XParam ,
                                                 SAPDB_Bool *fGotControlPw,
                                                 SAPDB_Char **errText)
{
    SAPDBErr_MessageList err;
    memset (XParam, 0, sizeof(RTE_XPARAM_REC) );

    /* --- Set XPARAM default values */
    RTEConf_InitParam ( XParam );

    /* read the special parameters other parameters depend on (MAXARCHIVELOGS etc.) */
    RTEConf_ReadSpecialParameters(XParam);

    RTEConf_ParameterReader paramReader;
    paramReader.StartIteration();

    RTEConf_Parameter::Name name;
    RTEConf_Parameter::Type type;
    RTEConf_ParameterValueRecord valueRec;
    while(paramReader.Next(name,type,err))
    {
        switch (type)
        {
        case RTEConf_Parameter::t_Integer:
        {
            RTEConf_Parameter::Integer value;
            RTEConf_ParameterAccess::Instance()->GetInteger(name,value,err);
            valueRec.xp_value_t = NXP_INT_TYPE;
            valueRec.xp_type.xp_integer = (SAPDB_Int4)value;
            break;
        }
        case RTEConf_Parameter::t_Real:
            break;
        case RTEConf_Parameter::t_String:
        {
            RTEConf_Parameter::String value;
            RTEConf_ParameterAccess::Instance()->GetString(name,value,err);
            valueRec.xp_value_t = NXP_STRING_TYPE;
            valueRec.xp_type.xp_string = value;
            break;
        }
        case RTEConf_Parameter::t_CryptInfo:
        {
            RTEConf_Parameter::CryptInfo value;
            RTEConf_ParameterAccess::Instance()->GetCryptInfo(name,value,err);
            valueRec.xp_value_t = NXP_CRYPT_TYPE;
            memcpy(valueRec.xp_type.xp_crypt,(SAPDB_Byte *)value,sizeof(RTEConf_Parameter::CryptInfo));
            break;
        }
        default:
            continue;
        }
        err.ClearMessageList();
        if(!RTEConf_AnalyzeParam ( XParam , name , &valueRec ,false, *fGotControlPw,err))
        {
            *errText = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
            return false;
        }
    }

    RTEConf_ModifyParameters(XParam);

    if(!(err.IsEmpty() || SAPDBErr_MessageList::Info == err.Type()) || !(paramReader.EndIteration(err)))
    {
        *errText = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
        return false;
    }
    else
    {
        return true;
    }
} // RTEConf_ReadAndAnalyzeParam

//-------------------------------------------

externC void   RTEConf_FreeTaskClusterArrayInXParamStructure( RTE_XPARAM_REC *XParam)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_FreeTaskClusterArrayInXParamStructure", Runtime_Trace, 5);

    if ( XParam->TaskClusterArray != NULL )
    {
        int i ;
        for ( i=0; XParam->TaskClusterArray[ i ] ; i++ )
        {
            RTEMem_RteAllocator::Instance().Deallocate ( XParam->TaskClusterArray[ i ] ) ;
        }
        RTEMem_RteAllocator::Instance().Deallocate ( XParam->TaskClusterArray ) ;
    }
}

externC void RTEConf_FreeDevDescrInXParamStructure( RTE_XPARAM_REC *XParam,int idx )
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_FreeDevDescrInXParamStructure", Runtime_Trace, 5);

    RTEMem_RteAllocator::Instance().Deallocate(XParam->DevspDescr[idx].Devspaces);
}

static void  RTEConf_CopyStringOutOfValueRecord ( unsigned char *s, RTEConf_ParameterValueRecord * xpr ,size_t sizeOfDestinationArea )

{
    size_t lengthOfSourceString = strlen((char *)xpr->xp_type.xp_string);
    if( lengthOfSourceString > sizeOfDestinationArea)
    {
        RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_CONF_NEVER_VERIFIED));
    }
    if ( NXP_STRING_TYPE != xpr->xp_value_t  )
    {
        RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_CONF_NEVER_VERIFIED));
    }

    strncpy  ((char *) s , (char *)xpr->xp_type.xp_string , lengthOfSourceString + 1 );
}

static inline void *AllocateMemory(size_t size)
{
    void *adr = RTEMem_RteAllocator::Instance().Allocate(size);
    if(!adr)
    {
        RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_CONF_NEVER_VERIFIED));
    }
    return adr;
}

static SAPDB_Bool RTEConf_AnalyzeParam ( RTE_XPARAM_REC                 *XParam  ,
                                         RTEConf_Parameter::Name        name ,
                                         RTEConf_ParameterValueRecord   *xpr ,
                                         SAPDB_Bool                     inOnlineMode,
                                         SAPDB_Bool                     &fGotControlPw,
                                         SAPDBErr_MessageList           &err)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_AnalyzeParam", Runtime_Trace, 5);
    SAPDB_Bool fOk = true;
    SAPDB_Int4 which;
    RTEConf_Parameter::String valueString = xpr->xp_type.xp_string;
    struct xparam_strcmp_table *pE;

	which = RTEConf_FindParam(name, xpr,err);
	if ( which < 0 )
	{
        if(inOnlineMode)
        {
            return false;
        }
        else
        {
/*IST on request of Juergen Primsch 1993-01-05: suppress this warning */
            SAPDB_UInt8 lengthOfMessage = (SAPDB_UInt8)strlen((SAPDB_Char *)name) + strlen(UNKNOWN_PARAMETER_MESSAGE) + 1;
            SAPDB_Char *message = (SAPDB_Char *)alloca((int)lengthOfMessage);
            sp77sprintf(message,(int)lengthOfMessage,UNKNOWN_PARAMETER_MESSAGE"%s",(SAPDB_Char *)name);
            SAPDBTRACE_WRITELN (Runtime_Trace, 1, message);
            return true;
        }
	}

	if ( which < RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE1 )
	{
		struct xparam_strncmp_table *pN;
        SAPDB_UInt2 indexOfElement;

		pN = &RteXparamTable1[which];

        if(pN->UpdateMode==XP_V_UPDATE_OFFLINE && inOnlineMode)
        {
            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_CANT_CHANGE_PARAMETER_ONLINE, (SAPDB_Char *)name);
            return false;
        }

		indexOfElement = atoi( (SAPDB_Char *)name + pN->len + 1 ) - 1;

		switch( pN->type )
		{
		case XP_V_SET_TASKCLUSTER:
            RTEConf_SetTaskCluster( XParam, xpr->xp_type.xp_string, indexOfElement ) ;
            // if anything had gone wrong, RTEConf_SetTaskCluster() would have called RTE_Crash()...
            break;

		case XP_V_SET_DEVSP_NAME:
            fOk = RTEConf_SetDevspName ( XParam, name, xpr, pN->dt, indexOfElement ,err);
			break;

        case XP_V_SET_M_DEVSP_NAME:
            fOk = RTEConf_SetDevspName ( XParam, name, xpr, pN->mdt, indexOfElement ,err);
			break;

        case XP_V_SET_DEVSP_SIZE:
			fOk = RTEConf_SetDevspSize ( XParam, pN->dt,
									indexOfElement, (SAPDB_Int)RTEConf_XParamInt ( xpr ,err) ) ;
			break;

		default:
			break;
		}
	}
	else
	{

		/*
		 * Reduced to offset into RteXparamTable2
		 */
		which -= RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE1;
		pE = &RteXparamTable2[which];

        if(inOnlineMode && (XP_V_UPDATE_OFFLINE == pE->UpdateMode))
        {
            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_CANT_CHANGE_PARAMETER_ONLINE, (SAPDB_Char *)name);
            return false;
        }
		switch( pE->type )
		{
		case XP_V_BOOL_Y:
			*(tsp00_Bool *)XP_PTR(XParam, pE) =
					(tsp00_Bool)( *valueString == 'y' || *valueString == 'Y' );
			break;

		case XP_V_BOOL_N:
			*(tsp00_Bool *)XP_PTR(XParam, pE) =
					(tsp00_Bool)( *valueString == 'n' || *valueString == 'N' );
			break;

		case XP_V_BOOL_A:
			*(tsp00_Bool *)XP_PTR(XParam, pE) =
					(tsp00_Bool)( *valueString == 'a' || *valueString == 'A' );
			break;

		case XP_V_BOOL_S:
			*(tsp00_Bool *)XP_PTR(XParam, pE) =
					(tsp00_Bool)( *valueString == 's' || *valueString == 'S' );
			break;

		case XP_V_LONG:
			*(tsp00_Int *)XP_PTR(XParam, pE) = (tsp00_Int)RTEConf_XParamInt( xpr , err);
			break;

		case XP_V_ULONG:
			*(tsp00_Uint *)XP_PTR(XParam, pE) = (tsp00_Uint)RTEConf_XParamInt( xpr ,err);
			break;

        case XP_V_STRING:
            RTEConf_CopyStringOutOfValueRecord (  XP_PTR(XParam, pE), xpr, pE->size);
            break;

		case XP_V_CRYPT:
			fGotControlPw = TRUE ;
			COPY ( XP_PTR(XParam,pE), xpr->xp_type.xp_crypt , pE->size ) ;
			break;

		case XP_V_IGNORED:
			break;

        case XP_V_SPECIAL_LONG:
			*(tsp00_Int *)XP_PTR(XParam, pE) = (tsp00_Int)RTEConf_XParamInt( xpr ,err);
			/* Currently, there are no special longs */
            break;

        case XP_V_SPECIAL_ULONG:
			*(tsp00_Uint *)XP_PTR(XParam, pE) = (tsp00_Uint)RTEConf_XParamInt( xpr ,err);
            break;

		case XP_V_SPECIAL_STRING:
            break;
        case XP_V_SPECIAL_SPECIAL:
			break;

		default:
			break;
		}
	}
    return fOk;
}

/*
 * ===========================================================================
 */

static  SAPDB_Bool RTEConf_SetDevspSize ( RTE_XPARAM_REC            *XParam,
                                          SAPDB_Long                lDevType ,
                                          SAPDB_ULong               ulDevspNo ,
                                          SAPDB_Int                 ulDevspSize )
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_SetDevspSize", Runtime_Trace, 5);

    SAPDB_Bool fOk ;

    if ( ulDevspSize == 0 )
        return ( TRUE ) ;

    if ( ( fOk = XParam->DevspDescr [ lDevType ]. ulNumDevs > ulDevspNo ) )
    { 
        DEV_DESCR *ActDevsp = XParam->DevspDescr[lDevType].Devspaces + ulDevspNo;
        ActDevsp->lDevspaceSize = ulDevspSize ;
    }
    else
    { 
        RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_XP_TOO_MANY_DEVSPACES, szDevType[lDevType] ,SAPDB_ToString(XParam->DevspDescr [ lDevType ]. ulNumDevs )));
    }

    return (fOk) ;
}

static  void   RTEConf_SetTaskCluster( RTE_XPARAM_REC            *XParam,
                                       RTEConf_Parameter::String szTaskCluster,
                                       SAPDB_ULong               ulTCNo )
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_SetTaskCluster", Runtime_Trace, 5);

    size_t TaskClusterLen = STRLEN_UTF8( szTaskCluster ) + 1;

    if ( XParam->TaskClusterArray == NULL )
    {
        XParam->TaskClusterArray = (char **)RTEMem_RteAllocator::Instance().Allocate(MAX_TASKCLUSTER_PARTS * sizeof ( char * ) );
        if ( 0 == XParam->TaskClusterArray)
        { 
            RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_MALLOC_FAILED_FOR,"set TASKCLUSTER" ));
        } 
        else
        {
            memset ( XParam->TaskClusterArray, 0, 
                MAX_TASKCLUSTER_PARTS * sizeof ( char * ) ) ;
        }
    } 

    if (ulTCNo >= MAX_TASKCLUSTER_PARTS-1 )
    { 
        RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_XP_TOO_MANY_TASKCLUSTER,SAPDB_ToString(ulTCNo),SAPDB_ToString(MAX_TASKCLUSTER_PARTS)));
    } 
    XParam->TaskClusterArray[ulTCNo] = (char *)RTEConf_MakeACopyOfString(szTaskCluster);
}


/*
 * ===========================================================================
 */

static void RTEConf_InitParam ( RTE_XPARAM_REC *XParam  )
{
  SAPDBTRACE_ROUTINE_DEBUG("RTEConf_InitParam", Runtime_Trace, 5);

  /* All parameters not set here are automatically set to 0 / FALSE */

  XParam->ulRegLockSlice        = MAX_EXCLUSIVE;
  XParam->ulDispLoopsBefSemop   = MIN_DISP_LOOPS;
  XParam->fExclQueue            = TRUE;
  XParam->ulNumDataCacheReg     = 1;
  XParam->ulNumTreeReg          = 1;
  XParam->ulNumTransReg         = 1;
  XParam->ulNumTabReg           = 1;
  XParam->ulNumRowReg           = 1;
  XParam->ulNumOMSReg           = 8;
  XParam->lSequenceCache        = 1;
  XParam->lPrioBaseU2U          = 100;
  XParam->lPrioBaseRAV          = 80;
  XParam->lPrioBaseIOC          = 80;
  XParam->lPrioBaseCOM          = 10;
  XParam->lPrioBaseREX          = 40;
  XParam->lPrioFactor           = 80;
  XParam->lFbmCompression       = 50;
  XParam->lFbmBalance           = 10;
  XParam->lFbmLowIORate         = 10;
  XParam->fDefaultLRU           = TRUE;
  XParam->ulMaxCPU              = 1;
  XParam->lJoinMaxTabLevel4     = 16;
  XParam->lJoinMaxTabLevel9     = 5;
  XParam->fOptimCache           = FALSE;
  XParam->fOptimInvOnly         = TRUE;

  XParam->lDwCopyBoundary       = 50;
  XParam->lDwIOAreaSize         = 20;
  XParam->lDwIOAreaFlush        = 50;
  XParam->lDwLRUTailFlush       = 5;
  XParam->ulDwIOBufSize         = 1;
  XParam->ulPnoPoolSize         = 10000;
  XParam->ulMultIOBlockCount    = 8;
  XParam->lBackUpBlockCount     = 8;

  XParam->ulMBDataSize          = 4096 ;
  XParam->ulMBQualSize          = 4096 ;
  XParam->ulMBStackSize         = 4 * 4096 ;
  XParam->ulKBWorkStSize        = 8 * 1024 ;
  XParam->ulKBWorkBufSize       = 4096 ;
  XParam->ulCatalogCacheSize    = 64 * 1024 ;
  XParam->ulLockSupplyBlock     = 100 ;
  XParam->fFormatDataVolume     = TRUE;

  XParam->ulMaxEvents           = 100;
  XParam->ulMaxMessageFiles     = 5;
  XParam->ulOmsHeapUsageThreshold   = 100; /* PTS 1110149 */ 
  XParam->ulOmsHeapCount            = 1;    /* PTS 1110828 */ 
  XParam->ulOmsHeapBlocksize        = 10000;/* PTS 1110828 */
  XParam->lMaxSysDevspaces      = 1;  /* for migration only */

/***************************************************************/
#if defined(_WIN32)
/***************************************************************/
#if defined ( DYNAMIC_SERVER )
   XParam->fDynamicServerTasks  = TRUE;
#endif

  XParam->fUseMemEnhancement    = FALSE;
  XParam->lMemEnhancementLimit  = 0;
  XParam->fUseFibers            = TRUE;
  XParam->fUseAsyncIO           = 1;
  XParam->lMaxServerDB          = 1;

/***************************************************************/
#else
/***************************************************************/
  XParam->fUseThreadForTask     = true;
#ifdef OSF1
  XParam->fUseAsyncIO           = 1 ;
#endif
/***************************************************************/
#endif
/***************************************************************/

  XParam->ulDBPageSize          = sizeof ( tsp00_Page ) ; /* XXX default value, overwritten by XParam... Will detect if tsp00_Page is really gone.. */
  XParam->ulNoOfRegions         = MAX_REGIONS ;

  XParam->ulReadAheadBlobsGT    = 25;

  XParam->memorySlotSize        = 1; /* 1 MB memory slot size */

  /* Specify fastest allocation on the system */
#if defined(_WIN32)
  XParam->vallocMemPoolSize     = -1; /* Use valloc allocated memory */
#elif defined(HPUX)
  XParam->shmMemPoolSize        = -1; /* Use 'SYSV IPC Shm' allocated memory */
#else /* all other UNIX */
  XParam->mmapMemPoolSize       = -1; /* Use mmap allocated memory */
#endif

  XParam->allowDemangle         = true;

  return;
}

static  void    RTEConf_AllocDevspDescr ( RTE_XPARAM_REC         *XParam,
                                          SAPDB_Long             lDevspType,
                                          SAPDB_Int              ulNumDevsp)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_AllocDevspDescr", Runtime_Trace, 5);

    SAPDB_Int i ;

    if ( ulNumDevsp > 0 )
    {
        XParam->DevspDescr[lDevspType].Devspaces = (struct dev_descr *)RTEMem_RteAllocator::Instance().Allocate(ulNumDevsp * sizeof ( DEV_DESCR ) );
        if(0 == XParam->DevspDescr[lDevspType].Devspaces )
        {
            RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_MALLOC_FAILED_FOR,"DEVSPACE_DESCR" ));
        }

        for ( i=0; i < ulNumDevsp; i++ )
        {
            XParam->DevspDescr[lDevspType].Devspaces[i].lDevspaceType = DEVSPACE_UNUSED;
            XParam->DevspDescr[lDevspType].Devspaces[i].physicalDevID = RTE_UNDEF_ID;
        }
    }
    XParam->DevspDescr[lDevspType].ulNumDevs = ulNumDevsp ;
}

/*
 * ===========================================================================
 */

static  SAPDB_Bool RTEConf_SetDevspName ( RTE_XPARAM_REC                  *XParam,
                                          RTEConf_Parameter::Name         DevspName ,
                                          RTEConf_ParameterValueRecord    *xpr ,
                                          SAPDB_Int                       lDevType ,
                                          SAPDB_ULong                     ulDevspNo ,
                                          SAPDBErr_MessageList            &err)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_SetDevspName", Runtime_Trace, 5);

    SAPDB_Bool fOk ;

    RTEConf_Parameter::String szDevName = xpr->xp_type.xp_string;
    if ( szDevName[0] == '\0' )
        fOk = FALSE;
    else
    {
        if ( FALSE && /* UH 2001-04-24 ++++ for Testing Logging74 ++++ */
            ( ulDevspNo > 0 ) &&
            ( XParam->DevspDescr[lDevType].Devspaces[ulDevspNo - 1].lDevspaceType == DEVSPACE_UNUSED ))
        {
   		    err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_INVALID_DEVSPACE_NO, SAPDB_ToString(ulDevspNo));
            fOk = FALSE;
        }
        else
        {
            fOk = RTEConf_CheckDuplicateDevSpaceName( XParam->DevspDescr, (SAPDB_Char *)szDevName );

            if ( fOk )
            {
                if (( fOk = XParam->DevspDescr [ lDevType ]. ulNumDevs > ulDevspNo ))
                {
                    DEV_DESCR *ActDevsp = XParam->DevspDescr[lDevType].Devspaces + ulDevspNo;
                    size_t count = strlen((const char*) szDevName);
                    if (count > sizeof(tsp2_devname)) {
                      memcpy (ActDevsp->szDevspaceName, szDevName, sizeof(tsp2_devname));
                      ActDevsp->szDevspaceName[sizeof(tsp2_devname)] = 0;
                    } else {
                      memcpy (ActDevsp->szDevspaceName, szDevName, count + 1);
                    }
                    ActDevsp->lDevspaceType = lDevType ;
                }
                else
                {
                    RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_XP_TOO_MANY_DEVSPACES, szDevType[lDevType] ,SAPDB_ToString(XParam->DevspDescr [ lDevType ]. ulNumDevs )));
                }
            }
        }
    }
    return (fOk) ;
}

static RTEConf_Parameter::Integer  RTEConf_XParamInt ( RTEConf_ParameterValueRecord  *xpr ,SAPDBErr_MessageList &err)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_XParamInt", Runtime_Trace, 5);

    if ( NXP_INT_TYPE  == xpr->xp_value_t )
    {
        return ( xpr->xp_type.xp_integer );
    }
    else
    {
   		err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ILLEGAL_XP_TYPE, SAPDB_ToString( xpr->xp_value_t ));
        return ( -1 );
    }

}

/*
 * Search parameter in table and return index if found. -1 if not found
 * The returned index can be used as index into RteXparamTable1.
 */
static SAPDB_Int4 RTEConf_FindParam( RTEConf_Parameter::Name             name ,
                                     RTEConf_ParameterValueRecord        *xpr ,
                                     SAPDBErr_MessageList                &err)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEConf_FindParam", Runtime_Trace, 5);

    SAPDB_Int4 which;

	/*
     * First entry is special! All other are compared with strcmp()...
     */
	for ( which = 0; which < RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE1; which ++ )
	{
		struct xparam_strncmp_table *pN;

		pN = &RteXparamTable1[which];
		if ( !STRNCMP_UTF8(name, pN->name, pN->len) )
		{
          if ( xpr )
			switch( pN->type )
            {
			case XP_V_SET_TASKCLUSTER:
			case XP_V_SET_DEVSP_NAME:
			case XP_V_SET_M_DEVSP_NAME:
                if( NXP_STRING_TYPE != xpr->xp_value_t )
                {
   		            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ILLEGAL_XP_TYPE, SAPDB_ToString(  xpr->xp_value_t));
					which = -1;
				}
				break;

			case XP_V_SET_DEVSP_SIZE:
                if( NXP_INT_TYPE != xpr->xp_value_t )
                {
   		            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ILLEGAL_XP_TYPE, SAPDB_ToString( xpr->xp_value_t ));
					which = -1;
				}
				break;
			}
			return which;
		}
	}

    for ( which = 0; which < RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE2; which++ )
	{
		struct xparam_strcmp_table *pE;
		pE = &RteXparamTable2[which];

        if ( !STRNCMP_UTF8( name, pE->name, STRLEN_UTF8(pE->name) ) )
		{
          if ( xpr )
			switch( pE->type )
			{
			case XP_V_LONG:
			case XP_V_ULONG:
			case XP_V_REGIONS:
			case XP_V_SPECIAL_LONG:
			case XP_V_SPECIAL_ULONG:
                if( NXP_INT_TYPE != xpr->xp_value_t )
                {
   		            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ILLEGAL_XP_TYPE ,SAPDB_ToString( xpr->xp_value_t ));
					which = -1;
				}
                break;

			case XP_V_BOOL_Y:
			case XP_V_BOOL_N:
			case XP_V_BOOL_A:
			case XP_V_BOOL_S:
			case XP_V_STRING:
			case XP_V_SPECIAL_STRING:
                if( NXP_STRING_TYPE != xpr->xp_value_t )
                {
   		            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ILLEGAL_XP_TYPE ,SAPDB_ToString( xpr->xp_value_t ));
					which = -1;
				}
				break;

			case XP_V_IGNORED:
			case XP_V_SPECIAL_SPECIAL:
				break;

			case XP_V_CRYPT:
                if( NXP_CRYPT_TYPE != xpr->xp_value_t )
                {
   		            err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ILLEGAL_XP_TYPE, SAPDB_ToString( xpr->xp_value_t ));
					which = -1;
				}
				break;

			default:
				which = -1;
				break;
			}
			break;
		}
	}

	if ( which == RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE2 )
	{
        err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_XP_UNKNOWN_PARAM,(SAPDB_Char *)(name));
		which = -1;
	}

	/*
	 * Add number of elements in first table (subtracted in analyze_param...
	 */
	if ( which >= 0 )
	{
		which += RTE_NUMBER_OF_ELEMENTS_IN_XPARAM_TABLE1;
	}

	return which;
}


/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                       class RTEConf_ParameterAccess                       */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/

RTEConf_ParameterAccess *RTEConf_ParameterAccess::m_Instance=NULL;

/*---------------------------------------------------------------------------*/

void RTEConf_ParameterReader::StartIteration(void)
{
    m_NumberBuffer = ALLOCATE_ARRAY(SAPDB_Char,SIZE_OF_NUMBER_BUFFER_);
    m_Iterator = new RTEConf_ParameterIterator(*(RTEConf_ParameterAccess::Instance())->m_Conf_Parameter);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterReader::EndIteration(SAPDBErr_MessageList &err)
{
	if(NULL == m_Iterator)
	{
        err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ITERATION_HAD_NOT_BEEN_STARTED);
        return false;
	}
    else
    {
		DEALLOCATE_MEMORY(m_NumberBuffer);
		delete m_Iterator;
		m_Iterator = NULL;
		return true;
	}
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterReader::Next(RTEConf_Parameter::Name &parameterName,SAPDB_Char *&valueString,SAPDBErr_MessageList &err)
{
	if(NULL == m_Iterator)
	{
        err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ITERATION_HAD_NOT_BEEN_STARTED);
        return false;
	}
    else
    {
		RTEConf_Parameter::Type parameterType;
		SAPDB_Bool endOfListReached;
        SAPDBErr_MessageList err;
		while(m_Iterator->Next(parameterName,parameterType,endOfListReached,err))	
		/* this is done in a loop, because it is possible that the Iterator::Next() gives elements that are actually not in the parameter set anymore 
           (this happens when elements are Delete()d after the construction of the iterator. If such an element is detected, the next element is searched */
		{
			switch(parameterType)
			{
			case RTEConf_Parameter::t_String:
				{
					RTEConf_Parameter::String stringValue;
					if((RTEConf_ParameterAccess::Instance())->GetString(parameterName,stringValue,err))
					{
						valueString = (SAPDB_Char *)stringValue;
						return true;
					}
				}
				break;
			case RTEConf_Parameter::t_Integer:
				{
					RTEConf_Parameter::Integer integerValue;
					if((RTEConf_ParameterAccess::Instance())->GetInteger(parameterName,integerValue,err))
					{
						sp77sprintf(m_NumberBuffer,SIZE_OF_NUMBER_BUFFER_,"%qd",integerValue);
						valueString = m_NumberBuffer;
						return true;
					}
				}
				break;
			case RTEConf_Parameter::t_Real:
				{
					RTEConf_Parameter::Real realValue;
					if((RTEConf_ParameterAccess::Instance())->GetReal(parameterName,realValue,err))
					{
						sp77sprintf(m_NumberBuffer,SIZE_OF_NUMBER_BUFFER_,"%f",realValue);
						valueString = m_NumberBuffer;
						return true;
					}
				}
				break;
			case RTEConf_Parameter::t_CryptInfo:
				valueString = (SAPDB_Char *)"";
				return true;
				break;
			}
		}
	}
	return false;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterReader::Next(RTEConf_Parameter::Name &parameterName,RTEConf_Parameter::Type &parameterType,SAPDBErr_MessageList &err)
{
	if(NULL == m_Iterator)
	{
        err = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_ITERATION_HAD_NOT_BEEN_STARTED);
        return false;
	}
    else
	{
		SAPDB_Bool endOfListReached;
		return m_Iterator->Next(parameterName,parameterType,endOfListReached,err);
	}
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterAccess::SetInteger(RTEConf_Parameter::Name name, RTEConf_Parameter::Integer  value,SAPDBErr_MessageList &err)
{
    return m_Conf_Parameter->SetValue(name,value,err);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterAccess::SetReal(RTEConf_Parameter::Name name, RTEConf_Parameter::Real  value,SAPDBErr_MessageList &err)
{
    return m_Conf_Parameter->SetValue(name,value,err);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterAccess::SetString(RTEConf_Parameter::Name name, RTEConf_Parameter::String  value,SAPDBErr_MessageList &err)
{
    return m_Conf_Parameter->SetValue(name,value,err);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool RTEConf_ParameterAccess::SetCryptInfo(RTEConf_Parameter::Name name, RTEConf_Parameter::CryptInfo  &value,SAPDBErr_MessageList &err)
{
    RTEConf_Parameter::CryptInfo *CryptInfoValue = &value;

    return m_Conf_Parameter->SetValue(name,*CryptInfoValue,err);
}

/*---------------------------------------------------------------------------*/

RTEConf_ParameterAccess::RTEConf_ParameterAccess(SAPDB_Char *dbName)
{
    SAPDB_Bool dataRead;
    SAPDB_Bool fileReadSuccess;

    m_Conf_Parameter = new (AllocateMemory(sizeof(RTEConf_Parameter))) RTEConf_Parameter(dbName,0,RTECONF_MAXNAMELENGTH,RTECONF_MAXSTRINGLENGTH);
    SAPDBErr_MessageList err;
    fileReadSuccess = m_Conf_Parameter->Read(dataRead,err);

    // if the file does not exist, Read() returns true and sets dataRead to false. For the kernel, this is an error case, so 
    // an error message has to be generated and m_FileReadOk must be false.
    if( !fileReadSuccess )
    {
        RTE_Crash(err);
    }
    if( !dataRead )
    {
        RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_CONF_PARAMFILE_NOT_FOUND,dbName,SAPDB_ToString(LAST_SYSTEM_ERROR_)));
    }
}

/*---------------------------------------------------------------------------*/

void *RTEConf_ParameterAccess::operator new(size_t size)
{
    return AllocateMemory(size);
}

/*---------------------------------------------------------------------------*/

externC void vconf_param_get_int4      ( tsp11_ConfParamName        VAR_ARRAY_REF conf_param_name,
									     tsp00_Int4                 VAR_VALUE_REF conf_param_value,
									     tsp00_ErrText              VAR_ARRAY_REF errtext,
									     tsp11_ConfParamReturnValue VAR_VALUE_REF conf_param_ret )
{
    SAPDB_Int4 lengthOfName = conf_param_name.length();
    SAPDB_UTF8 *nameAsCString = (SAPDB_UTF8 *)AllocateMemory(lengthOfName+1);
    RTEConf_Parameter::Integer value;
    memcpy(nameAsCString,conf_param_name.asCharp(),lengthOfName);
    nameAsCString[lengthOfName]=0;
    SAPDBErr_MessageList err;
    if(RTEConf_ParameterAccess::Instance()->GetInteger(nameAsCString,value,err))
	{
		conf_param_ret.becomes(ok_sp11);
        conf_param_value = (tsp00_Int4)value;
	}
	else
	{
		conf_param_ret.becomes(not_ok_sp11);
		errtext.c2p(err.Message());
	}
    DEALLOCATE_MEMORY(nameAsCString);
}

/*---------------------------------------------------------------------------*/

externC void vconf_param_get_integer   ( tsp11_ConfParamName        VAR_ARRAY_REF conf_param_name,
									     tsp00_Int8                 VAR_VALUE_REF conf_param_value,
									     tsp00_ErrText              VAR_ARRAY_REF errtext,
									     tsp11_ConfParamReturnValue VAR_VALUE_REF conf_param_ret )
{
    SAPDB_Int4 lengthOfName = conf_param_name.length();
    SAPDB_UTF8 *nameAsCString = (SAPDB_UTF8 *)AllocateMemory(lengthOfName+1);
    memcpy(nameAsCString,conf_param_name.asCharp(),lengthOfName);
    nameAsCString[lengthOfName]=0;
    SAPDBErr_MessageList err;
    if(RTEConf_ParameterAccess::Instance()->GetInteger(nameAsCString,conf_param_value,err))
	{
		conf_param_ret.becomes(ok_sp11);
	}
	else
	{
		conf_param_ret.becomes(not_ok_sp11);
		errtext.c2p(err.Message());
	}
    DEALLOCATE_MEMORY(nameAsCString);
}

/*---------------------------------------------------------------------------*/

externC void vconf_param_get_real      ( tsp11_ConfParamName        VAR_ARRAY_REF conf_param_name,
									     tsp00_Longreal             VAR_VALUE_REF conf_param_value,
									     tsp00_ErrText              VAR_ARRAY_REF errtext,
									     tsp11_ConfParamReturnValue VAR_VALUE_REF conf_param_ret )
{
    SAPDB_Int4 lengthOfName = conf_param_name.length();
    SAPDB_UTF8 *nameAsCString = (SAPDB_UTF8 *)AllocateMemory(lengthOfName+1);
    memcpy(nameAsCString,conf_param_name.asCharp(),lengthOfName);
    nameAsCString[lengthOfName]=0;
    SAPDBErr_MessageList err;

    if(RTEConf_ParameterAccess::Instance()->GetReal(nameAsCString,conf_param_value,err))
	{
		conf_param_ret.becomes(ok_sp11);
	}
	else
	{
		conf_param_ret.becomes(not_ok_sp11);
		errtext.c2p(err.Message());
	}
    DEALLOCATE_MEMORY(nameAsCString);
}

/*---------------------------------------------------------------------------*/

externC void vconf_param_get_string    ( tsp11_ConfParamName        VAR_ARRAY_REF conf_param_name,
									     tsp11_ConfParamValue       VAR_ARRAY_REF conf_param_value,
									     tsp00_Int4                 VAR_VALUE_REF conf_param_length,
									     tsp00_ErrText              VAR_ARRAY_REF errtext,
									     tsp11_ConfParamReturnValue VAR_VALUE_REF conf_param_ret )
{
	RTEConf_Parameter::String StringValue;
    SAPDB_Int4 lengthOfName = conf_param_name.length();
    SAPDB_UTF8 *nameAsCString = (SAPDB_UTF8 *)RTEMem_RteAllocator::Instance().Allocate(lengthOfName+1);
    memcpy(nameAsCString,conf_param_name.asCharp(),lengthOfName);
    nameAsCString[lengthOfName]=0;
    SAPDBErr_MessageList err;

    if(RTEConf_ParameterAccess::Instance()->GetString(nameAsCString,StringValue,err))
	{
		conf_param_ret.becomes(ok_sp11);
        conf_param_length = (SAPDB_Int4)strlen((SAPDB_Char *)StringValue);
		conf_param_value.c2p((SAPDB_Char *)StringValue);
	}
	else
	{
		conf_param_ret.becomes(not_ok_sp11);
		errtext.c2p(err.Message());
	}
    DEALLOCATE_MEMORY(nameAsCString);
}

/*---------------------------------------------------------------------------*/

externC void vconf_param_get_cryptinfo ( tsp11_ConfParamName        VAR_ARRAY_REF conf_param_name,
									     tsp00_CryptName            VAR_ARRAY_REF conf_param_value,
									     tsp00_ErrText              VAR_ARRAY_REF errtext,
									     tsp11_ConfParamReturnValue VAR_VALUE_REF conf_param_ret )
{
    SAPDB_Int4 lengthOfName = conf_param_name.length();
    SAPDB_UTF8 *nameAsCString = (SAPDB_UTF8 *)RTEMem_RteAllocator::Instance().Allocate(lengthOfName+1);
    memcpy(nameAsCString,conf_param_name.asCharp(),lengthOfName);
    nameAsCString[lengthOfName]=0;
    SAPDBErr_MessageList err;

    if(RTEConf_ParameterAccess::Instance()->GetCryptInfo(nameAsCString,*((RTEConf_Parameter::CryptInfo *)(conf_param_value.asCharp())),err))
	{
		conf_param_ret.becomes(ok_sp11);
	}
	else
	{
		conf_param_ret.becomes(not_ok_sp11);
		errtext.c2p(err.Message());
	}

    DEALLOCATE_MEMORY(nameAsCString);
}

/*---------------------------------------------------------------------------*/

externC SAPDB_Bool RTEConf_ParameterAccessGetInteger   ( SAPDB_UTF8  *  conf_param_name,
									                     SAPDB_Int8  *  conf_param_value,
									                     SAPDB_Char  ** errtext)
{
    SAPDBErr_MessageList err;

    RTEConf_ParameterAccess* instance = RTEConf_ParameterAccess::Instance();
    if(!instance)
    {
        *errtext = (SAPDB_Char *)"database name not yet available";
        return false;
    }
    
    if(instance->GetInteger(conf_param_name,*conf_param_value,err))
	{
		return true;
	}
	else
	{
		*errtext = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
        return false;
    }
}

/*---------------------------------------------------------------------------*/

externC SAPDB_Bool RTEConf_ParameterAccessGetReal      ( SAPDB_UTF8  *  conf_param_name,
									                     SAPDB_Real8 *  conf_param_value,
									                     SAPDB_Char  ** errtext)
{
    SAPDBErr_MessageList err;

    RTEConf_ParameterAccess* instance = RTEConf_ParameterAccess::Instance();
    if(!instance)
    {
        *errtext = (SAPDB_Char *)"database name not yet available";
        return false;
    }
    
    if(instance->GetReal(conf_param_name,*conf_param_value,err))
	{
		return true;
	}
	else
	{
		*errtext = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
        return false;
    }
}

/*---------------------------------------------------------------------------*/

externC SAPDB_Bool RTEConf_ParameterAccessGetString    ( SAPDB_UTF8  *  conf_param_name,
									                     SAPDB_Char  ** conf_param_value,
									                     SAPDB_Char  ** errtext)
{
	RTEConf_Parameter::String StringValue = (RTEConf_Parameter::String)conf_param_value;
    SAPDBErr_MessageList err;

    RTEConf_ParameterAccess* instance = RTEConf_ParameterAccess::Instance();
    if(!instance)
    {
        *errtext = (SAPDB_Char *)"database name not yet available";
        return false;
    }
    
    if(instance->GetString(conf_param_name,StringValue,err))
	{
        *conf_param_value = (SAPDB_Char *)StringValue;
		return true;
	}
	else
	{
		*errtext = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
        return false;
    }
}

/*---------------------------------------------------------------------------*/

externC SAPDB_Bool RTEConf_ParameterAccessGetCryptInfo ( SAPDB_UTF8  *  conf_param_name,
									                     SAPDB_Byte  *  conf_param_value,
									                     SAPDB_Char  ** errtext)
{
    SAPDBErr_MessageList err;

    RTEConf_ParameterAccess* instance = RTEConf_ParameterAccess::Instance();
    if(!instance)
    {
        *errtext = (SAPDB_Char *)"database name not yet available";
        return false;
    }
    
    if(instance->GetCryptInfo(conf_param_name,*((RTEConf_Parameter::CryptInfo *)conf_param_value),err))
	{
		return true;
	}
	else
	{
		*errtext = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
        return false;
    }
}

/*---------------------------------------------------------------------------*/


externC void vconf_param_put           ( tsp11_ConfParamName        VAR_ARRAY_REF conf_param_name,
                                         tsp11_ConfParamValue       VAR_ARRAY_REF conf_param_value,
                                         boolean                                  is_numeric,
                                         tsp00_ErrText              VAR_ARRAY_REF errtext,
                                         tsp11_ConfParamReturnValue VAR_VALUE_REF conf_param_ret )
{
    SAPDB_Int4 lengthOfName = conf_param_name.length();
    SAPDB_UTF8 *nameAsCString = (SAPDB_UTF8 *)AllocateMemory(lengthOfName+1);

    memcpy(nameAsCString,conf_param_name.asCharp(),lengthOfName);
    nameAsCString[lengthOfName]=0;

    SAPDB_Int4 lengthOfValue = conf_param_value.length();
    SAPDB_UTF8 *valueAsCString = (SAPDB_UTF8 *)AllocateMemory(lengthOfValue+1);
    
    memcpy(valueAsCString,conf_param_value.asCharp(),lengthOfValue);
    valueAsCString[lengthOfValue]=0;

    RTE_XPARAM_REC *XParam = KGS->XParam ;

    RTEConf_ParameterValueRecord xpr;

    if(is_numeric)
    {
        xpr.xp_value_t = NXP_INT_TYPE;
        xpr.xp_type.xp_integer = atoi( (SAPDB_Char *)valueAsCString );
    }
    else
    {
        xpr.xp_value_t = NXP_STRING_TYPE;
        xpr.xp_type.xp_string = valueAsCString;
    }
  
    SAPDB_Bool fGotControlPw;
    SAPDBErr_MessageList err;
    if(!RTEConf_AnalyzeParam  ( XParam, nameAsCString, &xpr, true, fGotControlPw, err) )
    {
        if(RTEERR_ILLEGAL_XP_TYPE_ID == err.ID())
        {
            SAPDB_Char errorMessage[sizeof(tsp00_ErrText)];
            if(is_numeric)
            {
                sp77sprintf(errorMessage,sizeof(tsp00_ErrText),"param. %s must not be numeric",nameAsCString);
            }
            else
            {
                sp77sprintf(errorMessage,sizeof(tsp00_ErrText),"param. %s must be numeric",nameAsCString);
            }
            errtext.c2p(errorMessage);
        }
        else
        {
            errtext.c2p(RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message()));
        }
        conf_param_ret.becomes(not_ok_sp11);
    }
    else
    {
        if(is_numeric)
        {
            (RTEConf_ParameterAccess::Instance())->SetInteger(nameAsCString,(RTEConf_Parameter::Integer)atoi((SAPDB_Char *)valueAsCString),err);
        }
        else
        {
            (RTEConf_ParameterAccess::Instance())->SetString(nameAsCString,valueAsCString,err);
        }
        conf_param_ret.becomes(ok_sp11);
    }
    DEALLOCATE_MEMORY(nameAsCString);
    DEALLOCATE_MEMORY(valueAsCString);
}

/*---------------------------------------------------------------------------*/
 
externC SAPDB_Bool RTEConf_ParameterAccessSetCrashSem( SAPDB_Int8 conf_param_value,
                                                       SAPDB_Char ** errtext)
{
    SAPDBErr_MessageList err;
    if((RTEConf_ParameterAccess::Instance())->MarkCrashed(conf_param_value ,err))
    {
        return true;
    }
    *errtext = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
    return false;
}

/*---------------------------------------------------------------------------*/
 
externC SAPDB_Bool RTEConf_ParameterAccessResetParamChangedFlag ( SAPDB_Char ** errtext )
{
    SAPDBErr_MessageList err;
    if((RTEConf_ParameterAccess::Instance())->MarkAsVerified(err))
    {
        return true;
    }
    else
    {   
        *errtext = RTEConf_MakeACopyOfString((SAPDB_Char *)err.Message());
        return false;
    }
}

/*---------------------------------------------------------------------------*/
 
externC SAPDB_Bool RTEConf_ParameterAccessIsVerified ( )
{
    if(!RTEConf_ParameterAccess::Instance()->FileInOldFormat())
    {
      return RTEConf_ParameterAccess::Instance()->IsVerified();
    } 
    else 
    {
      return true;
    } // end if
}

/*---------------------------------------------------------------------------*/
 
externC SAPDB_Bool RTEConf_ParameterAccessGetCrashSem ( )
{
    if(!RTEConf_ParameterAccess::Instance()->FileInOldFormat())
    {
      return RTEConf_ParameterAccess::Instance()->IsCrashed();
    } 
    else 
    {
      return true;
    } // end if
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/

