/*!***************************************************************************

  module      : RTEConf_ParameterAccessKernelInterface.hpp

  -------------------------------------------------------------------------

  responsible : FrankH, JoergM, RobinW

  special area: RTE configuration parameters
  description : class RTEConf_ParameterAccess for accessing parameters from kernel C++-code


  last changed: 2002-04-29  10:30

  first created:2000-03-22  17:49

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef CONF_PARAMETER_ACCESS_KERNEL_INTERFACE_HPP
#define CONF_PARAMETER_ACCESS_KERNEL_INTERFACE_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "RunTime/RTE_Types.hpp"
#include "RunTime/RTE_ISystem.hpp"
#include "RunTime/RTE_MessageList.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/RTE_Messages.hpp"

#include "RunTime/Configuration/RTEConf_ParameterAccess.hpp"

#ifdef _WIN32           /* & if $OSSPEC = WIN32*/
#include "gos00k.h"     /* nocheck */
#else                   /* & else*/
#include "gen81.h"      /* nocheck */
#endif                  /* & endif*/

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



 /*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
  Class: RTEConf_ParameterAccess
  Description: class for Parameter access from kernel. 
  
  Realized as a singleton as there is exactly one parameter set
  in the kernel context

 */
class RTEConf_ParameterAccess
{
public:
    /*!----------------------------------------------
      Function: Instance
      Description: returns the one and only instance of the class

      Creates the instance if it does not exist
  
      Return value: the instance
     */
	static RTEConf_ParameterAccess *Instance(void)
	{
		if(NULL == m_Instance)
        {
            SAPDB_Char *dbName;
#ifdef _WIN32
            dbName = kgs.szServerDB;
#else
            dbName = en81_dbname;
#endif
            if(dbName[0])
                m_Instance = new RTEConf_ParameterAccess(dbName);
        }
		return m_Instance;
	};
	
    /*!----------------------------------------------
      Function: GetInteger
      Description: returns the value of an integer parameter

  
      Arguments: name [in] parameter name
                 integerValue [out] value reference
                 err [out] contains an error message if unsuccessful

      Return value: true if successful, false if not. 
     */
	SAPDB_Bool GetInteger(RTEConf_Parameter::Name name,RTEConf_Parameter::Integer &integerValue,SAPDBErr_MessageList &err)
	{
		return m_Conf_Parameter->GetValue(name,integerValue,err);
	};
	
    /*!----------------------------------------------
      Function: GetReal
      Description: returns the value of a real parameter

  
      Arguments: name [in] parameter name
                 realValue [out] value reference
                 err [out] contains an error message if unsuccessful

      Return value: true if successful, false if not. 
     */
	SAPDB_Bool GetReal(RTEConf_Parameter::Name name,RTEConf_Parameter::Real &realValue,SAPDBErr_MessageList &err)
	{
		return m_Conf_Parameter->GetValue(name,realValue,err);
	};

    /*!----------------------------------------------
      Function: GetString
      Description: returns the value of a string parameter

  
      Arguments: name [in] parameter name
                 stringValue [out] value reference
                 err [out] contains an error message if unsuccessful

      Return value: true if successful, false if not. 
     */
	SAPDB_Bool GetString(RTEConf_Parameter::Name name,RTEConf_Parameter::String &stringValue,SAPDBErr_MessageList &err)
	{
		return m_Conf_Parameter->GetReference(name,stringValue,err);
	};
	
    /*!----------------------------------------------
      Function: GetCryptInfo
      Description: returns the value of a cryptinfo parameter

  
      Arguments: name [in] parameter name
                 cryptInfoValue [out] value reference
                 err [out] contains an error message if unsuccessful

      Return value: true if successful, false if not. 
     */
	SAPDB_Bool GetCryptInfo(RTEConf_Parameter::Name name,RTEConf_Parameter::CryptInfo &cryptInfo,SAPDBErr_MessageList &err)
	{
		return m_Conf_Parameter->GetValue(name,cryptInfo,err);
	};
	
    /*!----------------------------------------------
      Function: IsVerified
      Description: returns the verified flag

      Return value: true if verified, false if not. 
     */
	SAPDB_Bool IsVerified( )
	{
      RTEConf_Parameter::TimeStamp oTimeStamp;
      return m_Conf_Parameter->GetLastVerifyTime(oTimeStamp);
	};

    /*!----------------------------------------------
      Function: IsCrashed
      Description: returns the crash flag

      Return value: true if crashed or online/admin, false if not. 
     */
	SAPDB_Bool IsCrashed( )
	{
      return m_Conf_Parameter->GetOriginalCrashFlag();
	};
    /*!----------------------------------------------
      Function: MarkCrashed
      Description: sets the crash flag

      Return value: true if successfull, false if not. 
     */
	SAPDB_Bool MarkCrashed(SAPDB_Int8 state, SAPDBErr_MessageList &err)
	{
    if (state == 0)
       return m_Conf_Parameter->MarkStoppedCorrectly(err);
    else
       return m_Conf_Parameter->SetCrashFlag(err);
	};

	SAPDB_Bool FileInOldFormat( )
	{
      return m_Conf_Parameter->FileInOldFormat();
	};

    SAPDB_Bool SetInteger(RTEConf_Parameter::Name name, RTEConf_Parameter::Integer  value,SAPDBErr_MessageList &err);


    /*!----------------------------------------------
      Function: SetReal
      Description: sets the value of the parameter identified by name.

      Arguments: name [in] parameter name reference
                 value [in] parameter value reference
                 err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 
     */
    SAPDB_Bool SetReal(RTEConf_Parameter::Name name, RTEConf_Parameter::Real  value,SAPDBErr_MessageList &err);


    /*!----------------------------------------------
      Function: SetString
      Description: sets the value of the parameter identified by name.

      Arguments: name [in] parameter name reference
                 value [in] parameter value reference
                 err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 
     */
    SAPDB_Bool SetString(RTEConf_Parameter::Name name, RTEConf_Parameter::String  value,SAPDBErr_MessageList &err);


    /*!----------------------------------------------
      Function: SetCryptInfo
      Description: sets the value of the parameter identified by name.

      Arguments: name [in] parameter name reference
                 value [in] parameter value reference
                 err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 
     */
    SAPDB_Bool SetCryptInfo(RTEConf_Parameter::Name name, RTEConf_Parameter::CryptInfo  &value,SAPDBErr_MessageList &err);


    /*!----------------------------------------------
      Function: MarkAsVerified
      Description: Marks the configuration as verified

      Arguments: err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 
                    In this case, the error message can be retrieved by calling GetLastErrorMessage
     */
    SAPDB_Bool MarkAsVerified(SAPDBErr_MessageList &err)
    {
        return m_Conf_Parameter->MarkAsVerified(err);
    };
    
    /*!----------------------------------------------
      Function: operator new
      Description: operator new

      uses RTEMem_RteAllocator to allocate size bytes.
      If allocation fails, RTECrash() is called

      Argument: size [in] number of bytes to allocate
     */
    void *operator new(size_t size);

    static SAPDB_Bool InstanceExists(void)
    {
        return (NULL != m_Instance);
    };

private:
    RTEConf_ParameterAccess(SAPDB_Char *dbName);
    RTEConf_Parameter *m_Conf_Parameter;
    static RTEConf_ParameterAccess *m_Instance;
friend class RTEConf_ParameterReader;
};

class RTEConf_ParameterReader
{
public:
    RTEConf_ParameterReader()
    {
        m_Iterator = NULL;
        m_NumberBuffer = NULL;
    };
    /*!----------------------------------------------
      Function: StartIteration
      Description: start an iteration through all parameters. 

     */
	void StartIteration(void);

    /*!----------------------------------------------
      Function: EndIteration
      Description: end the iteration. Frees the memory that has been allocated for the iterator

      Arguments: err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 

    */
	SAPDB_Bool EndIteration(SAPDBErr_MessageList &err);

    /*!----------------------------------------------
      Function: Next
      Description: returns the next parameter. integer and real parameters are given as strings.
                   For CryptInfo parameters, an empty string is returned

      Arguments: name [out] parameter name reference
                 valueString [out] value reference as string
      Return value: true if successful, false if not. 
                    In this case, the error message can be retrieved by calling GetLastErrorMessage
     */
	SAPDB_Bool Next(RTEConf_Parameter::Name &name,SAPDB_Char *&valueString,SAPDBErr_MessageList &err);

    /*!----------------------------------------------
      Function: Next
      Description: returns the name and type of the next parameter.

      Arguments: parameterName [out] parameter name reference
                 parameterType [out] parameter type reference
                 err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 
     */
    SAPDB_Bool Next(RTEConf_Parameter::Name &parameterName,RTEConf_Parameter::Type &parameterType,SAPDBErr_MessageList &err);

    /*!----------------------------------------------
      Function: SetInteger
      Description: sets the value of the parameter identified by name.

      Arguments: name [in] parameter name reference
                 value [in] parameter value reference
                 err [out] contains an error message if unsuccessful
      Return value: true if successful, false if not. 
     */
    RTEConf_ParameterIterator *m_Iterator;
    SAPDB_Char *m_NumberBuffer;
};

#endif // CONF_PARAMETER_ACCESS_KERNEL_INTERFACE_HPP
