/*!**************************************************************************

  module      : RTEHSS_PascalKernelInterface.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Kernel RunTime
  description : HotStandbyStorage Kernel Interface

  last changed: 2001-09-24  14:52
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef RTEHSS_PASCALINTERFACE_H
#define RTEHSS_PASCALINTERFACE_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/RTE_CompilerFeatures.h"
#include "SAPDBCommon/SAPDB_Types.h"
#include "gsp00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
  Function:    RTEHSS_GetMappingInfo
  Description: Get the mapping info string

  Returns a pointer to own mapping info string. This string contains all
  information needed to allow establishing a mirror of all data devspaces.

  Mapping info is string build of Hexdigits only [0-9A-F].

  Arguments:    ownMappingInfoSize [out] number of hex digits returned
				ownMappingInfo [out] own mapping info string
  Return value: true if call suceeded, false if call failed
 */
externC SAPDB_Bool
RTEHSS_GetMappingInfo(
	SAPDB_Int4    VAR_VALUE_REF ownMappingInfoSize,
	SAPDB_Char *  VAR_VALUE_REF ownMappingInfo );

/*!
  Function:    RTEHSS_EstablishMirror
  Description: Establish the mirror using given mapping info

  This routine triggers establishing a mirror of the data devspaces using
  the 

  Arguments:    otherMappingInfoSize [in] number of hex digits given
				otherMappingInfo [in] other machine mapping info 
  Return value: true if call suceeded, false if call failed
 */
externC SAPDB_Bool
RTEHSS_EstablishMirror(
	SAPDB_Int4      otherMappinginfoSize,
	SAPDB_Char *    otherMappingInfo );

/*!
  Function:    RTEHSS_SplitMirror
  Description: Split a previously established mirror of data devspaces

  This routine triggers splitting the previouly established mirror
  of the data devspaces.

  Return value: true if call suceeded, false if call failed
 */
externC SAPDB_Bool
RTEHSS_SplitMirror();

/*!
  Function:    RTEHSS_TimeUntilSynchronized
  Description: Get time to wait until storage is synchronized
  
  Synchronization means either a mirror is synchronized or a split
  has completed all outstanding I/O.

  Arguments:    timeToWaitInSeconds [out]
  Return value: true if call suceeded, false if call failed
 */
externC SAPDB_Bool
RTEHSS_TimeUntilSynchronized(
	SAPDB_Int4    VAR_VALUE_REF timeToWaitInSeconds );
	
/*!
  Function:     RTEHSS_SetLogAccess
  Description:  Set the log devspace access status.
 
  This is a pure protectional call. It is not absolutely necessary
  to have the LOG-devspaces readonly for the standby database server.
  But if the storage system provides such a feature, that only the
  master has read/write access and it can prevent the standby database
  server from write access, it is usefull to call it.
 
  Both master and standby can issue this call.
 
  Arguments:    readOnly [in] if false, set to read/write else set to read only.
  Return value: true if call suceeded, false if call failed
 */ 
externC SAPDB_Bool
RTEHSS_SetLogAccess (
	SAPDB_Bool	                readOnly );

/*!
  Function:     RTEHSS_SendBeginStandby
  Description:  Inform node that standby mode may begin

  The given node is informed, that its standby mode may begin now. This call includes
  informing all other nodes, which are already in standby mode about a new node in
  standby mode (but only if beginning standby mode is confirmed). Any hosts, that did
  not react on this information are removed from local list of standby nodes.

  Arguments:    nodeName [in] node to call
  Return value: true if call suceeded, false if call failed
 */ 
externC SAPDB_Bool
RTEHSS_SendBeginStandby (
	tsp00_NodeId  VAR_ARRAY_REF nodeName );

/*!
  Function:     RTEHSS_SendPrepareStandby
  Description:  Inform node that standby mode may begin

  The given node is informed, that its standby mode may begin now. This call includes
  informing all other nodes, which are already in standby mode about a new node in
  standby mode (but only if beginning standby mode is confirmed). Any hosts, that did
  not react on this information are removed from local list of standby nodes.

  Arguments:    nodeName [in] node to call
  Return value: true if call suceeded, false if call failed
 */ 
externC SAPDB_Bool
RTEHSS_SendPrepareStandby (
	tsp00_NodeId  VAR_ARRAY_REF nodeName );

/*!
  Function:     RTEHSS_AddStandbyNode
  Description:  Add standby node to internal list of standby nodes

  This call is used if a 'NEW STANDBY' command was received.

  Arguments:    nodeToAdd [in] node name to add
  Return value: true if call suceeded, false if call failed
 */ 
externC SAPDB_Bool
RTEHSS_AddStandbyNode(
	tsp00_NodeId  VAR_ARRAY_REF nodeToAdd );

/*!
  Function:     RTEHSS_RemoveStandbyNode
  Description:  Remove standby node from internal list of standby nodes

  Arguments:    nodeToRemove [in] node name to remove
 */ 
externC void
RTEHSS_RemoveStandbyNode(
	tsp00_NodeId  VAR_ARRAY_REF nodeToRemove );

/*!
  Function:     RTEHSS_GetFirstStandbyNode
  Description:  Get first node from internal list of standby nodes

  Call will return false if another routine is already iterating node list.
  If only a single node was i9n list, endOfList is set to true. 
  If more node names are in list, endOfList is set to false and list is locked.

  Arguments:    firstNode [out] first node name
                endOfList [out] set to true if end of list reached
  Return value: true if call suceeded, false if call failed
 */ 
externC SAPDB_Bool
RTEHSS_GetFirstStandbyNode(
	tsp00_NodeId  VAR_ARRAY_REF firstNode,
	tsp00_Bool    VAR_VALUE_REF endOfList);

/*!
  Function:     RTEHSS_GetNextStandbyNode
  Description:  Get next node from internal list of standby nodes

  If last node is returned, endOfList is set to true. List is unlocked
  and next call of this function would return false!

  Arguments:    nextNode [in] next node name in list
                endOfList [out] set to true if end of list reached
  Return value: true if call suceeded, false if call failed
 */ 
externC SAPDB_Bool
RTEHSS_GetNextStandbyNode(
	tsp00_NodeId  VAR_ARRAY_REF nextNode,
	tsp00_Bool    VAR_VALUE_REF endOfList);

#endif /* RTEHSS_PASCALINTERFACE_H */
