/*!**************************************************************************

  module      : RTEHSS_PascalKernelInterface.cpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Kernel RunTime
  description : HotStandbyStorage Kernel Interface

  last changed: 2001-09-24  14:52
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/HotStandbyStorage/RTEHSS_StandbyNodeList.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

RTEHSS_StandbyNodeList & RTEHSS_StandbyNodeList::Instance( )
{
  SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTEHSS_StandbyNodeList));

  if ( !m_Instance )
  {
	  m_Instance = new (Space) RTEHSS_StandbyNodeList();
  }

  SAPDBERR_ASSERT_STATE(m_Instance != NULL);

  return *m_Instance;
}

RTEHSS_StandbyNodeList * RTEHSS_StandbyNodeList::m_Instance = 0;

RTEHSS_StandbyNodeList::RTEHSS_StandbyNodeList()
: m_NodeCount(0),
  m_NodeList(0),
  m_Spinlock((const SAPDB_UTF8 *)"RTEHSS_StandbyNodeList"),
  m_ListIsParsed(false),
  m_ParsedNodeCount(0),
  m_ParsedNodeList(0),
  m_NextParsedIndex(0)
{}

  /*!
    Function: RTEHSS_StandbyNodeList::AddNode
    Description: Add a node to list of known nodes
    Arguments: nodeName [in] node to add
    Return value: true if call succeeded, false else
   */
SAPDB_Bool RTEHSS_StandbyNodeList::AddNode(NodeName nodeName)
{
  RTESync_LockedScope Lock(m_Spinlock);

  for ( SAPDB_Int4 nodeIndex = 0; nodeIndex < m_NodeCount; nodeIndex++ )
  {
	  if ( strcmp(m_NodeList[nodeIndex], nodeName) == 0 )
	  {
		  return true;
	  }
  }
  NodeName *newList = (NodeName *)RTEMem_RteAllocator::Instance().Allocate( 
	  (m_NodeCount+1) * sizeof(NodeName) );
  if ( !newList )
  {
	  return false;
  }

  SAPDB_Char *newName = (SAPDB_Char *)RTEMem_RteAllocator::Instance().Allocate( strlen(nodeName) );
  if ( !newName )
  {
      RTEMem_RteAllocator::Instance().Deallocate( newList );
	  return false;
  }

  strcpy(newName, nodeName);
  memcpy(newList, m_NodeList, m_NodeCount * sizeof(NodeName));
  newList[m_NodeCount] = newName;

  NodeName *oldList = m_NodeList;
  m_NodeList = newList;
  if ( oldList )
  {
	  RTEMem_RteAllocator::Instance().Deallocate( oldList );
  }
  return true;
}
  /*!
    Function: RTEHSS_StandbyNodeList::RemoveNode
    Description: Remove a node from list of known nodes
    Arguments: nodeName [in] node to remove
   */
void RTEHSS_StandbyNodeList::RemoveNode(NodeName nodeName)
{
  RTESync_LockedScope Lock(m_Spinlock);

  for ( SAPDB_Int4 nodeIndex = 0; nodeIndex < m_NodeCount; nodeIndex++ )
  {
	  if ( strcmp(m_NodeList[nodeIndex], nodeName) == 0 )
	  {
	      RTEMem_RteAllocator::Instance().Deallocate(m_NodeList[nodeIndex]);

		  for ( SAPDB_Int4 copyIndex = nodeIndex+1; copyIndex < m_NodeCount; copyIndex++ )
		  {
			  m_NodeList[copyIndex-1] = m_NodeList[copyIndex];
		  }

		  --m_NodeCount;
		  if ( 0 == m_NodeCount )
		  {
		      RTEMem_RteAllocator::Instance().Deallocate(m_NodeList);
			  m_NodeList = 0;
		  }
		  
		  break;
	  }
  }
}

  /*!
    Function: RTEHSS_StandbyNodeList::FirstNode
    Description: Get first node from list
	
	Arguments: nodeName [out] first node in list
	           moreNodes [out] true if more than one node in list
	Return value: true if got node, false if list is empty
   */
SAPDB_Bool RTEHSS_StandbyNodeList::FirstNode(NodeName &nodeName,
	                    SAPDB_Bool &moreNodes)
{
  RTESync_LockedScope Lock(m_Spinlock);
  if ( m_ListIsParsed )
  {
	  return false;
  }
  m_ListIsParsed = true;
  m_ParsedNodeCount = m_NodeCount;

  /* Copy Node List */
  m_ParsedNodeList = (NodeName *)RTEMem_RteAllocator::Instance().Allocate( m_NodeCount * sizeof(NodeName) );
  if ( 0 == m_ParsedNodeList )
  {
	return false;
  }

  for ( SAPDB_Int4 nodeIndex = 0; nodeIndex < m_NodeCount; nodeIndex++ )
  {
	  m_ParsedNodeList[nodeIndex] = 
		  (NodeName)RTEMem_RteAllocator::Instance().Allocate( strlen(m_NodeList[nodeIndex]) + 1 );
	  if ( 0 == m_ParsedNodeList[nodeIndex] )
	  {
		return false;
	  }
	  strcpy(m_ParsedNodeList[nodeIndex], m_NodeList[nodeIndex]);
  }

  m_NextParsedIndex = 0;

  return NextNode(nodeName, moreNodes);
}

  /*!
    Function: RTEHSS_StandbyNodeList::NextNode
    Description: Get next node from list
	
	Arguments: nodeName [in] next node in list
	           moreNodes [out] true if more nodes in list
	Return value: true if got node, false if list is empty
   */
SAPDB_Bool RTEHSS_StandbyNodeList::NextNode(NodeName &nodeName,
	                   SAPDB_Bool &moreNodes)
{
  RTESync_LockedScope Lock(m_Spinlock);

  if ( !m_ListIsParsed )
  {
	  return false;
  }

  nodeName = m_ParsedNodeList[m_NextParsedIndex];

  ++m_NextParsedIndex;

  moreNodes = m_ListIsParsed = (m_NextParsedIndex < m_ParsedNodeCount) ;

  if ( !m_ListIsParsed )
  {
	for ( SAPDB_Int4 nodeIndex = 0; nodeIndex < m_NodeCount; nodeIndex++ )
	{
      RTEMem_RteAllocator::Instance().Deallocate( m_ParsedNodeList[nodeIndex] );
    }
    RTEMem_RteAllocator::Instance().Deallocate( m_ParsedNodeList );

	m_ParsedNodeList  = 0;
	m_ParsedNodeCount = 0;
	m_NextParsedIndex = 0;
  }

  return true;
}

SAPDB_Bool RTEHSS_StandbyNodeList::SynchronizeLogPosition( Log_RawDeviceOffset lastCommittedLogPosition,
	                              Log_RawDeviceOffset &minimumLastReadLogPosition,
	                              Log_RawDeviceOffset &maximumLastReadLogPosition,
	                              SAPDB_Int4          &numberOfNodesThatAnswered )
{
	/* dummy .... */
	minimumLastReadLogPosition = lastCommittedLogPosition;
	maximumLastReadLogPosition = lastCommittedLogPosition;
    numberOfNodesThatAnswered  = 1;
	return true;
}

