/*!**************************************************************************

  module      : RTEHSS_StandbyNodeList.hpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Kernel RunTime
  description : HotStandbyStorage Kernel Interface

  last changed: 2001-09-24  14:52
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef RTEHSS_STANDBYNODELIST_HPP
#define RTEHSS_STANDBYNODELIST_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.h"
#include "RunTime/Synchronisation/RTESync_NamedSpinlock.hpp"
#include "Logging/Log_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! 
  Class: RTEHSS_StandbyNodeList

  Description: Container for standby node handling

  This container handles a list of known standby nodes and all actions
  to be applied on individual as well as all standby nodes.
 */
class RTEHSS_StandbyNodeList
{
public:
  typedef SAPDB_Char *NodeName;
  /*!
    Function: Instance

    Description: Return reference to single instance of RTEHSS_StandbyNodeList

    Return value: Reference to RTEMem_StandbyNodeList instance
   */
   static RTEHSS_StandbyNodeList & Instance();

  /*!
    Function: AddNode
    Description: Add a node to list of known nodes
    Arguments: nodeName [in] node to add
    Return value: true if call succeeded, false else
   */
   SAPDB_Bool AddNode(NodeName nodeName);

  /*!
    Function: RemoveNode
    Description: Remove a node from list of known nodes
    Arguments: nodeName [in] node to remove
   */
   void RemoveNode(NodeName nodeName);

  /*!
    Function: FirstNode
    Description: Get first node from list
	
	Arguments: nodeName [out] first node in list
	           moreNodes [out] true if more than one node in list
	Return value: true if got node, false if list is empty
   */
   SAPDB_Bool FirstNode(NodeName &nodeName,
	                    SAPDB_Bool &moreNodes);

  /*!
    Function: NextNode
    Description: Get next node from list
	
	Arguments: nodeName [in] next node in list
	           moreNodes [out] true if more nodes in list
	Return value: true if got node, false if list is empty
   */
   SAPDB_Bool NextNode(NodeName &nodeName,
	                   SAPDB_Bool &moreNodes);

  /*!
    Function: SynchronizeLogPosition
	Description: Exchange Log Position Information with all Standby Hosts

    The actual logical LogPosition value is sent to all known standby nodes and the answer
	is the minimum and maximum logical LogPosition retrieved from all those hosts.

    The number of hosts that answered is also returned. If a host did not answer due to a
	communication problem, the host state is checked and if this check also fails, it
	is discarded from the list of known standby hosts...

    The call fails only if not a signle host answered the call.

	Arguments: lastCommittedLogPosition [in] last commited log position
               minimumLastReadLogPosition [out]	       
               maximumLastReadLogPosition [out]	       
			   numberOfNodesThatAnswered [out]
	Return value: true if at least one standby node answered, false if no one answered
   */
   SAPDB_Bool SynchronizeLogPosition( Log_RawDeviceOffset lastCommittedLogPosition,
	                                  Log_RawDeviceOffset &minimumLastReadLogPosition,
	                                  Log_RawDeviceOffset &maximumLastReadLogPosition,
	                                  SAPDB_Int4          &numberOfNodesThatAnswered );

private:
	RTEHSS_StandbyNodeList();
	SAPDB_Int4              m_NodeCount;
	NodeName            *   m_NodeList;
	RTESync_NamedSpinlock   m_Spinlock;
	SAPDB_Bool              m_ListIsParsed;
	SAPDB_Int4              m_ParsedNodeCount;
	NodeName            *   m_ParsedNodeList;
	SAPDB_Int4              m_NextParsedIndex;
	static RTEHSS_StandbyNodeList *m_Instance;
};

#endif /* RTEHSS_STANDBYNODELIST_HPP */
