/*!**************************************************************************

  module      : RTEMem_ProtectingAllocator.cpp

  -------------------------------------------------------------------------

  responsible : Robin Wibrock

  special area:
  description :


  last changed: 2002-01-16  12:25
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002

  ========== licence begin  GPL
  Copyright (C) 2002 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/RTE_MessageList.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/RTE_Messages.hpp"

#include "RunTime/MemoryManagement/RTEMem_ProtectingAllocator.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"

#ifndef _WIN32          /*&if !$OS in [ WIN32 ]*/
#include <unistd.h>     /*nocheck*/
#include <sys/mman.h>   /*nocheck*/
#include "gen41.h"      /*nocheck*/
#endif                  /*&endif*/

extern SAPDBTrace_Topic Memory_Trace;
extern SAPDBTrace_Topic Runtime_Trace;

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define SIXTYFOUR_KB 0x10000
#define ONE_MB      (1024*1024)
#define ONE_GB      (1024*1024*1024)

#if defined (_WIN32)
#define RESERVE_BLOCK_SIZE ((SAPDB_ULong)(52 * ONE_MB))
#elif defined (LINUX)
#define RESERVE_BLOCK_SIZE 93 * ONE_MB
#elif defined (OSF1)
#define RESERVE_BLOCK_SIZE 16 * ONE_MB
#elif defined (HPUX)
#define RESERVE_BLOCK_SIZE 17 * ONE_MB
#elif defined (SUN)
#define RESERVE_BLOCK_SIZE 1 * ONE_MB
#endif

//#define INSTALL_MERGER_THREAD
//#define USE_MERGE_WAIT_SEMAPHORE

/* the literal constants used in scanning the expressions that describe the memory configuration */

#define DESC_PHYSMEM	"PHYSMEM"
#define DESC_MB			"MB"
#define DESC_GB			"GB"

/* the granularity for allocations to use if none has been given by the user */
#define INITIAL_CHUNKSIZE ONE_MB

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define IS_A_LETTER_(character)  (('a' <= character  && 'z' >= character) || ('A' <= character && 'Z' >= character))
#define IS_PART_OF_NUMBER_(character)  (('0' <= character  && '9' >= character) || '.' == character || ',' == character)

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/


/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                         class VallocMemoryGrabber                         */
/*                                                                           */
/* memory grabber that uses VirtualAlloc(). Windows only, uses its own       */
/* algorithm to make sure that all possible memory can be used               */
/*                                                                           */
/*****************************************************************************/

VallocMemoryGrabber::VallocMemoryGrabber()
{
	m_Name = (SAPDB_Char *)"virtual";
}

#ifdef _WIN32

SAPDB_ULong VallocMemoryGrabber::GrabMemoryBlock(SAPDB_ULong SizeInBytes)
{
	MEMORY_BASIC_INFORMATION mem_basic_info;

    SAPDB_ULong sizeOfBiggestRegion = 0;
    SAPDB_Byte *startOfBiggestRegion;  

	SAPDB_Byte *currentAddress = 0;
    /* just calling VirtualAlloc to get memory sometimes obviously doesnt work, e.g. on 32-bit systems with 3GB option */
    /* memory above 2GB sometimes cannot be allocated by calling VirtualAlloc without a given address */
    /* To avoid this, the whole address space is scanned through using VirtualQuery(). The memory block to allocate */
    /* is finally taken from the largest free area found */
    do
	{
		memset(&mem_basic_info,0,sizeof(mem_basic_info));
		
		if(0==VirtualQuery(currentAddress,&mem_basic_info,sizeof(mem_basic_info)))
			break;

		if(MEM_FREE == mem_basic_info.State && mem_basic_info.RegionSize >= SIXTYFOUR_KB )
		{
			SAPDB_Byte *startOfUseableRegion = currentAddress;
			SAPDB_ULong sizeOfUseableRegion  = mem_basic_info.RegionSize;

            /* VirtualAlloc() automatically rounds given addresses _down_ to the next lower 64KB boundary. */
            /* If we get an address from VirtualQuery(), the space at the next lower 64KB boundary is normally */
            /* not free, but the space at the next higher 64 KB area is if the area itself is larger than 64KB */
            /* So the start address is rounded up to the next higher 64KB boundary berfore giving it to */
            /* VirtualAlloc() */
            SAPDB_ULong spaceToNext64Kboundary = SIXTYFOUR_KB  - (((SAPDB_ULong)currentAddress) & (SIXTYFOUR_KB - 1));
			if(SIXTYFOUR_KB  != spaceToNext64Kboundary)
			{
				startOfUseableRegion += spaceToNext64Kboundary;
				sizeOfUseableRegion  -= spaceToNext64Kboundary;
			}

			/* If the memory area found is larger than any area previously found, save its address */
            if(sizeOfUseableRegion > sizeOfBiggestRegion)
			{
				startOfBiggestRegion = startOfUseableRegion;
                sizeOfBiggestRegion  = sizeOfUseableRegion;
			}
        }
		currentAddress += mem_basic_info.RegionSize;
	}
	while(1);

	/* If absolutely nothing has been found, we cannot allocate anything */
    if(0 == sizeOfBiggestRegion)
		return 0;
	
	
	SAPDB_ULong SizeToUse = min(SizeInBytes,sizeOfBiggestRegion);
	
	void *addressOfAllocatedBlock = VirtualAlloc(startOfBiggestRegion,SizeToUse,MEM_RESERVE|MEM_COMMIT,PAGE_READWRITE);
	if(addressOfAllocatedBlock == NULL)
	{
		DWORD dwErr;
		dwErr = GetLastError();
		return 0;		
	}
	((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).AddBlock(addressOfAllocatedBlock,SizeToUse);
	return SizeToUse;	
}

#endif	

/* for UNIX platforms: define the AllocMemBlock() routines for shared and mmap memory   */


/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                          class ShmMemoryGrabber                           */
/*                                                                           */
/* memory grabber that uses shared memory. Unix only.                        */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

ShmMemoryGrabber::ShmMemoryGrabber()
{
	m_Name = (SAPDB_Char *)"shared";
}

#ifndef _WIN32
void *ShmMemoryGrabber::AllocMemBlock(SAPDB_ULong SizeInBytes)
{
	int ShmSegmentId = sql41_create_shm ( (key_t)0, 100 , 0 , SizeInBytes, 0664 , (int)getuid() , (char *)"db" , (char *)"db");
	if(ShmSegmentId == -1)
		return NULL;
	else
		return sql41AttachShm ( ShmSegmentId, (char *)0, 0 );
}
#endif


/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                          class MmapMemoryGrabber                          */
/*                                                                           */
/* memory grabber that uses mmap(). Unix only.                               */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

MmapMemoryGrabber::MmapMemoryGrabber()
{
	m_Name = (SAPDB_Char *)"mmap";
}

#ifndef _WIN32
void *MmapMemoryGrabber::AllocMemBlock(SAPDB_ULong SizeInBytes)
{
#if defined(SUN) || defined(NMP)
	int devZero = open("/dev/zero",O_RDWR);
	return mmap(NULL,SizeInBytes,PROT_READ|PROT_WRITE,MAP_PRIVATE,devZero,0);
#else
	return mmap(NULL,SizeInBytes,PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
#endif
}

#endif




/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                          class HeapMemoryGrabber                          */
/*                                                                           */
/* memory grabber that uses malloc(). Available on all platforms.            */
/*                                                                           */
/*****************************************************************************/

HeapMemoryGrabber::HeapMemoryGrabber()
{
	m_Name = (SAPDB_Char *)"heap";
}

void *HeapMemoryGrabber::AllocMemBlock(SAPDB_ULong SizeInBytes)
{
#ifdef AIX
	if(SizeInBytes <= ONE_MB)
		return 0;
#endif
	return malloc(SizeInBytes);
}

/*------------------------------*/

#if defined(LINUX)	
/* on linux, mallocating more than 2 GB doesnt give an error - but actually doesnt work. */
/* So the maximum size must be set to 2GB-1MB */
SAPDB_ULong HeapMemoryGrabber::GetMaxRealisticSize(double SizeInBytes)
{			
	if(SizeInBytes > 2047.0 * (double)ONE_MB)
		return 2047 * ONE_MB;
	else
		return (SAPDB_ULong)SizeInBytes;
}
#endif




/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                           class ScanExpression                            */
/*                                                                           */
/*  used scan through an expression that describes an amount of memory. The  */
/*  expression may be a number followed by "MB" or "GB", the word "PHYSMEM", */
/*  or something like "PHYSMEM+100MB", "PHYSMEM-100MB" or "PHYSMEM*0.5".     */
/*  Currently only used internally by class IMemoryGrabber                   */
/*                                                                           */
/*****************************************************************************/

class ScanExpression
{
private:
	SAPDB_Char *m_currentCharPosition;
public:
    typedef enum 
    { 
        TOK_NONE,
        TOK_NUMBER,
        TOK_STAR,
        TOK_MINUS,
        TOK_PLUS,
        TOK_PHYSMEM
    } tokenType;

	ScanExpression(SAPDB_Char *pcExpression)
	{
		m_currentCharPosition = pcExpression;
	}

    /*------------------------------*/

	tokenType getNextToken(double *Value)
	{
		if(*m_currentCharPosition == '*')
		{
			m_currentCharPosition++;
			return TOK_STAR;
		}
		else if('-' == *m_currentCharPosition)
		{
			m_currentCharPosition++;
			return TOK_MINUS;
		}
		else if('+' == *m_currentCharPosition)
		{
			m_currentCharPosition++;
			return TOK_PLUS;
		}
		else if(IS_PART_OF_NUMBER_ (*m_currentCharPosition))
		{
			SAPDB_Char *startOfNumber = m_currentCharPosition;
			while(IS_PART_OF_NUMBER_ (*m_currentCharPosition))
			{
				m_currentCharPosition++;
			}
			SAPDB_Char *copyOfNumber = (SAPDB_Char *)malloc(m_currentCharPosition-startOfNumber+1);
			strncpy(copyOfNumber,startOfNumber,m_currentCharPosition-startOfNumber+1);

			*Value = strtod(copyOfNumber,NULL);
			free(copyOfNumber);			

			SAPDB_Char *startOfString = m_currentCharPosition;
            if(IS_A_LETTER_(*m_currentCharPosition))
			{
				while(IS_A_LETTER_(*m_currentCharPosition))
				{
					m_currentCharPosition++;
				}
				if(0 == strncmp(startOfNumber,(SAPDB_Char *)DESC_MB,sizeof(DESC_MB)-1))
					*Value *= ONE_MB;
				else if(0 == strncmp(startOfNumber,(SAPDB_Char *)DESC_GB,sizeof(DESC_GB)-1))
					*Value *= ONE_GB;
			}
			return TOK_NUMBER;
		}
		else if(IS_A_LETTER_(*m_currentCharPosition))
		{
			SAPDB_Char *startOfString = m_currentCharPosition;
			m_currentCharPosition++;
			while(IS_A_LETTER_(*m_currentCharPosition))
			{
				m_currentCharPosition++;
			}
			if(0 == strncmp(startOfString,(SAPDB_Char *)DESC_PHYSMEM,sizeof(DESC_PHYSMEM)-1))
				return TOK_PHYSMEM;

		}
		return TOK_NONE;
	}
}
;



/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                           class IMemoryGrabber                            */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

IMemoryGrabber::IMemoryGrabber()
{
	m_Name = NULL;
	m_State = MEMGRABBER_DIDNOTHING;
	m_MemoryAllocatedInBytes = 0;
}

/*------------------------------*/

SAPDB_Bool IMemoryGrabber::successful(void)
{
	return m_State == MEMGRABBER_SUCCESSFUL || m_State == MEMGRABBER_PARTLY_SUCCESSFUL;
}
  
/*------------------------------*/

void IMemoryGrabber::SetParameters(double physicalMemoryInBytes,SAPDB_ULong minMemAreaSizeInBytes)
{
	m_PhysicalMemoryInBytes = physicalMemoryInBytes;
	m_MinMemAreaSizeInBytes = minMemAreaSizeInBytes;
}

/*------------------------------*/

void *IMemoryGrabber::AllocMemBlock(SAPDB_ULong sizeInBytes)
{
	m_State = MEMGRABBER_NOTAVAILABLE;
	return NULL;
}

/*------------------------------*/

SAPDB_ULong IMemoryGrabber::GetMaxRealisticSize(double SizeInBytes)
{			
#if !defined(BIT64) && (defined(LINUX)||defined(HPUX))
	if(SizeInBytes > (double)((SAPDB_ULong)-1))				
		return (SAPDB_ULong)-1;
#endif
	return (SAPDB_ULong)SizeInBytes;
}

/*------------------------------*/

void IMemoryGrabber::GrabAmountOfMemory(SAPDB_Char *expression)
{
    m_Expression = expression;

    printf("Grab: %s\n",expression);
	double sizeInBytes = interpret(expression);

	if(sizeInBytes > 0)
	{
		SAPDB_ULong memoryAllocatedInBytesThisTime = 0;
		double remainingSizeInBytes = sizeInBytes;
		while(remainingSizeInBytes > 0)
		{
			printf("GrabMemoryBlock (%s): desired %f\n",m_Name,remainingSizeInBytes);
/* remainingSizeInBytes is of type double and so possibly much bigger than the biggest SAPDB_ULong, which theoretically is the */
/* maxmimum size a memory block can have. In this case, on some platforms (sun32 and aix32) a simple cast to SAPDB_ULong would */
/* give the biggest SAPDB_ULong, but on other platforms, it just gives 0. So GetMaxRealisticSize() is called, which returns    */
/* the maximum size a memory block can have on all platforms */

			memoryAllocatedInBytesThisTime = GrabMemoryBlock(GetMaxRealisticSize(remainingSizeInBytes));

            printf("GrabMemoryBlock (%s) got %lu of %f bytes\n",m_Name,memoryAllocatedInBytesThisTime,remainingSizeInBytes);
			remainingSizeInBytes -= memoryAllocatedInBytesThisTime;
			m_MemoryAllocatedInBytes += memoryAllocatedInBytesThisTime;
			if(memoryAllocatedInBytesThisTime == 0)
				break;
		}
		printf("GrabAmountOfMemory (%s): m_MemoryAllocatedInBytes=%lu\n",m_Name,m_MemoryAllocatedInBytes);
		if (0 == m_MemoryAllocatedInBytes)
		{
			m_State = MEMGRABBER_NOTAVAILABLE;
		}
		else if(m_MemoryAllocatedInBytes == sizeInBytes)
		{
			m_State = MEMGRABBER_SUCCESSFUL;
		}
		else
		{
			m_State = MEMGRABBER_PARTLY_SUCCESSFUL;
		}
	}
}

/*------------------------------*/

SAPDB_ULong IMemoryGrabber::GrabMemoryBlock(SAPDB_ULong sizeInBytes)
{
	SAPDBERR_ASSERT_STATE(m_Name != 0);	/* name is set in every derived class,so name == 0 would mean that an object of the base class has been created which makes no sense */

	SAPDB_ULong memoryAllocatedInBytesThisTime = 0;
	
	printf("GrabMemoryBlock (%s): desired %lu\n",m_Name,sizeInBytes);
	
    if(sizeInBytes < m_MinMemAreaSizeInBytes)
		return 0;

    /* first attempt: try to allocate the number of bytes desired */
    void *addressOfallocatedBlock = AllocMemBlock(sizeInBytes);	/*  allocate memory using the method of the concrete memory grabber */
	if(m_State == MEMGRABBER_NOTAVAILABLE)
		return 0;

	if(addressOfallocatedBlock != NULL && addressOfallocatedBlock != (void *)-1)
	{
        /* if the attempt to allocate the desired number of bytes succeeded, we are happy */
        printf("first attempt successful\n");
		((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).AddBlock(addressOfallocatedBlock,sizeInBytes);
		memoryAllocatedInBytesThisTime = sizeInBytes;
	}
	else
	{
        /* if we could not get the the whole number of bytes in one single block, we must try to get as much as we can */
        printf("first attempt unsuccessful\n");
		SAPDB_ULong blockSizeDesired = sizeInBytes;
		do
		{
            /* We always try to allocate blocks with sizes that are powers of 2*/
            SAPDB_ULong blockSizeToUse = ~(((SAPDB_ULong)-1) >> 1);
			{
				while(blockSizeToUse > blockSizeDesired)
				{
					blockSizeToUse >>= 1;
				}
			}
			
			addressOfallocatedBlock = AllocMemBlock(blockSizeToUse);
            if(addressOfallocatedBlock != NULL && addressOfallocatedBlock != (void *)-1)
			{
    			/* if allocation succeeded, make the block useable */
				((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).AddBlock(addressOfallocatedBlock,blockSizeToUse);
                
                memoryAllocatedInBytesThisTime += blockSizeToUse;
                /* repeat allocating with the number of bytes remaining */
                blockSizeDesired = sizeInBytes - memoryAllocatedInBytesThisTime;
			}
			else
			{
                /* if allocation did not succeed, repeat allocating with the next smaller size  */
                /* blockSizeToUse is always calculated as the biggest number that is a power of */
                /* two and not bigger than blockSizeDesired (e.g.: if blockSizeDesired is 20, blockSizeToUse will be 16) */
                /* So, by setting blockSizeDesired to blockSizeToUse-1, blockSizeToUse will be blockSizeToUse/2 in the next */
                /* iteration. */
                blockSizeDesired = blockSizeToUse - 1;
			}
		}
		while(blockSizeDesired > ((SAPDB_ULong)m_MinMemAreaSizeInBytes));
	}
	return memoryAllocatedInBytesThisTime;
}

/*------------------------------*/

void IMemoryGrabber::ProcessState()
{
    SAPDBTRACE_WRITELN (Runtime_Trace, 0, "this is " << m_Name << "allocator");
	switch(m_State)
	{
	case MEMGRABBER_NOTAVAILABLE:
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_SYS_MEM_TYPE_NOT_AVAILABLE,m_Name));
		break;
	case MEMGRABBER_FAILEDTOALLOCATEPHYSMEM:
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_SYS_COULD_NOT_CALCULATE_MEMSIZE));
		break;
	case MEMGRABBER_SYNTAXERROR:
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_SYNTAX_ERROR,m_Name,m_Expression));
		break;
	case MEMGRABBER_PARTLY_SUCCESSFUL:
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEWARN_SYS_COULD_NOT_ALLOCATE_DESIRED_MEMORY));
	case MEMGRABBER_SUCCESSFUL:
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEINFO_SYS_PROTECT_ALLOC_SUCCESSFUL,SAPDB_ToString(m_MemoryAllocatedInBytes/ONE_MB),m_Name));
		break;
	}
}

/*------------------------------*/

double IMemoryGrabber::interpret(SAPDB_Char *expression)
{
    /* the expression must match one of three cases cases: */
    /* (a) a numeric value */
    /* (b) the string "PHYSMEM" */
    /* (c) the string "PHYSMEM", followed by an operator and a numeric value  */
	ScanExpression scan(expression);
    ScanExpression::tokenType currentToken;
	double value;
	
	currentToken = scan.getNextToken(&value);
    
    /* if the expression starts with the string "PHYSMEM", it is checked if it is followed by an operator or not */
    if(currentToken == ScanExpression::TOK_PHYSMEM)
	{
		if(0 == m_PhysicalMemoryInBytes)
		{
			m_State = MEMGRABBER_FAILEDTOALLOCATEPHYSMEM;
			return 0;
		}
		currentToken = scan.getNextToken(&value);
		switch(currentToken)
		{
		case ScanExpression::TOK_STAR:
            currentToken = scan.getNextToken(&value);
            if(currentToken == ScanExpression::TOK_NUMBER)
	        {
			    return value * m_PhysicalMemoryInBytes;
            }
			break;
		case ScanExpression::TOK_MINUS:
            currentToken = scan.getNextToken(&value);
            if(currentToken == ScanExpression::TOK_NUMBER)
	        {
			    return m_PhysicalMemoryInBytes - value;
            }
			break;
		case ScanExpression::TOK_PLUS:
            currentToken = scan.getNextToken(&value);
            if(currentToken == ScanExpression::TOK_NUMBER)
	        {
			    return m_PhysicalMemoryInBytes + value;
            }
			break;
		case ScanExpression::TOK_NONE:
            /* if no operator follows, the value of the expression is given by m_PhysicalMemoryInBytes. Nothing more has to be done. */
			return m_PhysicalMemoryInBytes;
        }
	}
    else if(currentToken == ScanExpression::TOK_NUMBER) 
    {
        /* If the expression starts with a number, the value of the expression is just the value of this number */
    	return value;
    }

    /* If we come here, something must have gone wrong */
	m_State = MEMGRABBER_SYNTAXERROR;
	return 0;
}


double IMemoryGrabber::m_PhysicalMemoryInBytes;

SAPDB_ULong  IMemoryGrabber::m_MinMemAreaSizeInBytes;



/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                     class RTEMem_ProtectingAllocator                      */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

RTEMem_ProtectingAllocator::memoryAreaProtectionLevel RTEMem_ProtectingAllocator::m_protectionLevel = MEMORYAREA_PROTECTIONLEVEL_NONE;
MemoryAreaDescriptor *RTEMem_ProtectingAllocator::m_veryLastBlock = 0;
FreeMemoryChain *RTEMem_ProtectingAllocator::m_freeMemoryChain = 0;
teo07_Mutex RTEMem_ProtectingAllocator::m_mergeMutex;
SAPDB_Bool RTEMem_ProtectingAllocator::m_mergeIsRunning = false;
SAPDB_Int RTEMem_ProtectingAllocator::m_mergeWaitingCounter = 0;
teo07_Mutex RTEMem_ProtectingAllocator::m_mergeEntryMutex;
teo07_ThreadSemaphore RTEMem_ProtectingAllocator::m_mergeSemaphore;
teo07_Mutex RTEMem_ProtectingAllocator::m_protectMutex;
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
teo07_Mutex RTEMem_ProtectingAllocator::m_outMutex;
#endif
SAPDB_ULong RTEMem_ProtectingAllocator::m_totalMemorySize = 0;
MemoryAreaDescriptor *RTEMem_ProtectingAllocator::m_NextAvailableMemoryAreaDescriptor=NULL;
SAPDB_ULong RTEMem_ProtectingAllocator::m_ChunkSizeInBytes = 0;
SAPDB_ULong RTEMem_ProtectingAllocator::m_PageSize = 0;


void *RTEMem_ProtectingAllocator::getAvailableMemoryAreaDescriptor(void)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::getAvailableMemoryAreaStruct", Memory_Trace, 1);

	MemoryAreaDescriptor *foundMemoryAreaDescriptor;
	SAPDB_Int4 numberOfMemoryAreaDescriptorsInSpace = 5;
	if(m_NextAvailableMemoryAreaDescriptor == NULL)
	{
		m_allocations++;
		void *spaceForMemoryAreaDescriptors = malloc(numberOfMemoryAreaDescriptorsInSpace*sizeof(MemoryAreaDescriptor));
		MemoryAreaDescriptor *currentMemoryAreaDescriptor = (MemoryAreaDescriptor *)spaceForMemoryAreaDescriptors;
		for(SAPDB_Int4 counter=0;counter<numberOfMemoryAreaDescriptorsInSpace-1;counter++)
		{
			currentMemoryAreaDescriptor->m_nextElement = currentMemoryAreaDescriptor+1;
			currentMemoryAreaDescriptor++;
		}
		currentMemoryAreaDescriptor->m_nextElement = NULL;
		m_NextAvailableMemoryAreaDescriptor = (MemoryAreaDescriptor *)spaceForMemoryAreaDescriptors;
	}
	foundMemoryAreaDescriptor = m_NextAvailableMemoryAreaDescriptor;
	m_NextAvailableMemoryAreaDescriptor = m_NextAvailableMemoryAreaDescriptor->m_nextElement;

	return (void *)foundMemoryAreaDescriptor;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::makeMemoryAreaDescriptorAvailable(MemoryAreaDescriptor *theMemoryAreaDescriptor)
{
#ifndef INSTALL_MERGER_THREAD
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::makeMemoryAreaStructAvailable", Memory_Trace, 1);
#endif

	theMemoryAreaDescriptor->m_nextElement = m_NextAvailableMemoryAreaDescriptor;
	m_NextAvailableMemoryAreaDescriptor = theMemoryAreaDescriptor;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::SetParameters(SAPDB_Long chunkSizeInMB,SAPDB_ULong pageSize,SAPDB_Bool protectMemory)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::SetParameters", Memory_Trace, 1);

	m_PageSize = pageSize;
	m_ChunkSizeInBytes = chunkSizeInMB*ONE_MB;
	if(protectMemory)
		m_protectionLevel = MEMORYAREA_PROTECTIONLEVEL_FULL;
	else
		m_protectionLevel = MEMORYAREA_PROTECTIONLEVEL_NONE;

	m_FullyInitialized = true;
}


RTEMem_ProtectingAllocator * RTEMem_ProtectingAllocator::m_Instance = NULL;

/*------------------------------*/

SAPDBMem_IBlockAllocator & RTEMem_ProtectingAllocator::Instance()
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::Instance", Memory_Trace, 1);

	if(m_Instance == NULL)
	{
		SAPDBMEM_STATIC_RAW_ALLOCATE(pSpace,sizeof(RTEMem_ProtectingAllocator));

		m_Instance = new ((void *)pSpace) RTEMem_ProtectingAllocator();
	}
	return *m_Instance;
}

/*------------------------------*/

RTEMem_ProtectingAllocator::RTEMem_ProtectingAllocator(SAPDB_ULong BlockSize):
    m_BytesControlled(0),
    m_CountAlloc(0),
    m_CountDealloc(0),
	m_allocations(0),
	m_FullyInitialized(false)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::RTEMem_ProtectingAllocator", Memory_Trace, 1);

	SAPDBMEM_STATIC_RAW_ALLOCATE(spaceForFreeMemoryChain,sizeof(FreeMemoryChain));
	m_freeMemoryChain = new ((void *)spaceForFreeMemoryChain) FreeMemoryChain();

    if(BlockSize == 0)
		m_ChunkSizeInBytes = INITIAL_CHUNKSIZE;
	else
		m_ChunkSizeInBytes = BlockSize;

	m_PageSize = RTE_ISystem::Instance().GetSystemPageSize();
	m_protectionLevel = MEMORYAREA_PROTECTIONLEVEL_FULL;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::InitializeMultithreading(void)
{
#ifdef INSTALL_MERGER_THREAD

    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::InitializeMultithreading", Memory_Trace, 1);

	tsp00_ErrTextc errText;
	teo07_ThreadErr err;
	teo07_Thread mergerThread;

	sqlcreatemutex(&m_mergeMutex);
	sqlcreatemutex(&m_mergeEntryMutex);
	sqlcreatemutex(&m_protectMutex);
	sqlcreatesem(&m_mergeSemaphore,0,errText,&err);
	sqlbeginthread(512*1024,(teo07_ThreadCallback)(RTEMem_ProtectingAllocator::mergerThreadproc), NULL, 0,&mergerThread,errText, &err);

	sqlsetthreadpriority(mergerThread,THR_PRIO_IDLE_EO07);
#endif
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
	sqlcreatemutex(&m_outMutex);
#endif
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::GetBaseAllocatorCallStatistics", Memory_Trace, 1);

    CountAlloc   = m_CountAlloc;
    CountDealloc = m_CountDealloc;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::GetCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::GetCallStatistics", Memory_Trace, 1);

    CountAlloc   = m_CountAlloc;
    CountDealloc = m_CountDealloc;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::CalcStatistics( SAPDB_ULong &BytesUsed,
                               SAPDB_ULong &BytesControlled)
{
	BytesUsed       = m_BytesControlled;
    BytesControlled = m_BytesControlled;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::AddBlock(void *address,SAPDB_ULong size)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::AddBlock", Memory_Trace, 1);

	if(address == NULL)
		return;
	prepareBlock(&address,&size);


/* create a free memory area that contains the whole block */
	MemoryAreaDescriptor *memoryAreaDescriptorForBlock;

    memoryAreaDescriptorForBlock = new (getAvailableMemoryAreaDescriptor()) MemoryAreaDescriptor();
	initMemoryArea(memoryAreaDescriptorForBlock,address,size,m_veryLastBlock);
	m_freeMemoryChain->add(memoryAreaDescriptorForBlock);

/*	create a dummy memory area at the end of the block that only consists of the info space.    */
/*	It has the following purposes:                                                              */
/*	1) to give a link to the last memory area in the block. This is used in merge() to walk     */
/*	   through all blocks                                                                       */
/*	2) to mark the end of the block to avoid an attempt to merge the last area of a block with  */
/*	   the first area of the next block                                                         */
	MemoryAreaDescriptor *dummyMemoryAreaDescriptorForBlockEnd;

	dummyMemoryAreaDescriptorForBlockEnd = new (getAvailableMemoryAreaDescriptor()) MemoryAreaDescriptor();
	initMemoryArea(dummyMemoryAreaDescriptorForBlockEnd,(SAPDB_Byte *)address+size,sizeofinfospace(),memoryAreaDescriptorForBlock);

	m_veryLastBlock = dummyMemoryAreaDescriptorForBlockEnd;

	m_totalMemorySize += size;
}

/*------------------------------*/

void * RTEMem_ProtectingAllocator::Allocate(SAPDB_ULong BlockCount)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::Allocate", Memory_Trace, 1);

	SAPDBERR_ASSERT_STATE(m_ChunkSizeInBytes != 0);
	SAPDB_ULong desiredBytes = BlockCount * m_ChunkSizeInBytes;
	SAPDB_ULong neededBytes = calcNeededBytes(desiredBytes);
	SAPDB_ULong newSize;
	void *pNewAddr;

    if(!m_FullyInitialized)
	{
        SAPDB_ULong ulBlockSize = neededBytes+3*sizeofinfospace();	/* 1 chunk (1 MB) + 1 guard page am Anfang + 1 guard page am Ende + 1 page frs Alignment */
#if 0
        if(m_FirstBlock.m_BaseAddr == NULL)	/* only one block may be allocated before the allocator is completely initialized */
		{
			m_FirstBlock.m_BaseAddr = malloc(ulBlockSize);	
			m_Instance->AddBlock(m_FirstBlock.m_BaseAddr,ulBlockSize);
			m_FirstBlock.m_AddrOfUserSpace = m_freeMemoryChain->m_firstElement->m_AddressOfUserSpace;
			m_FirstBlock.m_Size = ulBlockSize;
		}
		else if(m_SecondBlock.m_BaseAddr == NULL)
		{
			m_SecondBlock.m_BaseAddr = malloc(ulBlockSize);	
			m_Instance->AddBlock(m_SecondBlock.m_BaseAddr,ulBlockSize);
			m_SecondBlock.m_AddrOfUserSpace = m_freeMemoryChain->m_firstElement->m_AddressOfUserSpace;
			m_SecondBlock.m_Size = ulBlockSize;
		}
		else if(m_ThirdBlock.m_BaseAddr == NULL)
		{
			m_ThirdBlock.m_BaseAddr = malloc(ulBlockSize);	
			m_Instance->AddBlock(m_ThirdBlock.m_BaseAddr,ulBlockSize);
			m_ThirdBlock.m_AddrOfUserSpace = m_freeMemoryChain->m_firstElement->m_AddressOfUserSpace;
			m_ThirdBlock.m_Size = ulBlockSize;
		}
		else if(m_FourthBlock.m_BaseAddr == NULL)
		{
			m_FourthBlock.m_BaseAddr = malloc(ulBlockSize);	
			m_Instance->AddBlock(m_FourthBlock.m_BaseAddr,ulBlockSize);
			m_FourthBlock.m_AddrOfUserSpace = m_freeMemoryChain->m_firstElement->m_AddressOfUserSpace;
			m_FourthBlock.m_Size = ulBlockSize;
		}
		else
        {
            DebugBreak();
            return NULL;
        }
#else
        SAPDB_UInt4 firstBlockCounter;
        for(firstBlockCounter = 0;firstBlockCounter < NUMBER_OF_FIRST_BLOCKS;firstBlockCounter++)
        {
            if(m_FirstBlocks[firstBlockCounter].m_BaseAddr == NULL)	
		    {
			    m_FirstBlocks[firstBlockCounter].m_BaseAddr = malloc(ulBlockSize);	
			    m_Instance->AddBlock(m_FirstBlocks[firstBlockCounter].m_BaseAddr,ulBlockSize);
			    m_FirstBlocks[firstBlockCounter].m_AddrOfUserSpace = m_freeMemoryChain->m_firstElement->m_AddressOfUserSpace;
			    m_FirstBlocks[firstBlockCounter].m_Size = ulBlockSize;
                break;
            }
        }
        if(NUMBER_OF_FIRST_BLOCKS == firstBlockCounter)
        {
#ifdef _WIN32
//            DebugBreak();
#endif
            return NULL;
        }
#endif
	}
	bool merged = false;

#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
	sqlbeginmutex(&m_outMutex);
#endif
	while(1)
	{
#ifdef INSTALL_MERGER_THREAD
		if(NULL != m_mergeMutex)
			sqlbeginmutex(&m_mergeMutex);
#endif
/* search for a free block that can be allocated */
        MemoryAreaDescriptor *foundBlock = m_freeMemoryChain->searchForSuitableList(neededBytes);
		if(foundBlock != NULL && foundBlock->m_Size >= neededBytes)
		{
			m_freeMemoryChain->remove(foundBlock);
			foundBlock->m_ReserveSpaceSize = foundBlock->m_Size - desiredBytes;
			foundBlock->m_Size = desiredBytes;
			MemProtect(foundBlock->m_AddressOfUserSpace,foundBlock->m_Size,MEMORYPROTECTION_READWRITE);

			SAPDB_ULong remainingSize = foundBlock->m_Size + foundBlock->m_ReserveSpaceSize - neededBytes;
			if(remainingSize >= m_ChunkSizeInBytes + sizeofinfospace())
			{
				MemoryAreaDescriptor *splitOffBlock = new (getAvailableMemoryAreaDescriptor()) MemoryAreaDescriptor();
				SAPDBERR_ASSERT_STATE(splitOffBlock != NULL);
				pNewAddr = (SAPDB_Byte *)foundBlock->m_AddressOfUserSpace + neededBytes;
				newSize = remainingSize;

				SAPDBERR_ASSERT_STATE(pNewAddr != NULL && newSize != 0);
				initMemoryArea(splitOffBlock,pNewAddr,newSize,foundBlock);

	            MemoryAreaDescriptor *nextNeighbour = getNextNeighbour(foundBlock);
	            if(nextNeighbour != NULL)
	            {
		            nextNeighbour->m_prevNeighbour = splitOffBlock;
	            }
				foundBlock->m_ReserveSpaceSize = neededBytes - foundBlock->m_Size;

				m_freeMemoryChain->add(splitOffBlock);
			}
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
			sqlendmutex(&m_outMutex);
#endif
#ifdef INSTALL_MERGER_THREAD
			if(NULL != m_mergeMutex)
				sqlendmutex(&m_mergeMutex);
#endif
			return foundBlock->m_AddressOfUserSpace;
		}
#ifdef INSTALL_MERGER_THREAD
		if(NULL != m_mergeMutex)
			sqlendmutex(&m_mergeMutex);
#endif
		if(merged)
		{
#ifdef __GENERATE_OUTPUT
			printf("failed\n");
#endif
			break;
		}
#ifdef __GENERATE_OUTPUT
		printf("trying to merge...");
#endif
		merge();
//		out();
		merged = true;
	}
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
	sqlendmutex(&m_outMutex);
#endif
	return NULL;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::Deallocate(
     void * BlockStartAddress,
     SAPDB_ULong BlockCount )
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::Deallocate", Memory_Trace, 1);

	if(BlockStartAddress == NULL)
		return;

#if 0
    if(BlockStartAddress == m_FirstBlock.m_AddrOfUserSpace)
	{
		free(m_FirstBlock.m_BaseAddr);
	}
	else if(BlockStartAddress == m_SecondBlock.m_AddrOfUserSpace)
	{
		free(m_SecondBlock.m_BaseAddr);
	}
	else if(BlockStartAddress == m_ThirdBlock.m_AddrOfUserSpace)
	{
		free(m_ThirdBlock.m_BaseAddr);
	}
	else if(BlockStartAddress == m_FourthBlock.m_AddrOfUserSpace)
	{
		free(m_FourthBlock.m_BaseAddr);
	}
    else
#else
    SAPDB_UInt4 firstBlockCounter;
    for(firstBlockCounter = 0;firstBlockCounter < NUMBER_OF_FIRST_BLOCKS;firstBlockCounter++)
    {
        if(m_FirstBlocks[firstBlockCounter].m_AddrOfUserSpace == BlockStartAddress)	
		{
    		free(m_FirstBlocks[firstBlockCounter].m_BaseAddr);
            return;
		}
    }
#endif
	{
		MemoryAreaDescriptor *foundMemoryAreaDescriptor = getMemoryAreaDescriptor(BlockStartAddress);
		SAPDBERR_ASSERT_STATE(foundMemoryAreaDescriptor != NULL);
		SAPDBERR_ASSERT_STATE(foundMemoryAreaDescriptor->m_AddressOfUserSpace == BlockStartAddress);
		SAPDBERR_ASSERT_STATE(foundMemoryAreaDescriptor->m_Size == BlockCount * m_ChunkSizeInBytes);
		foundMemoryAreaDescriptor->m_Size += foundMemoryAreaDescriptor->m_ReserveSpaceSize;
		foundMemoryAreaDescriptor->m_ReserveSpaceSize = 0;
#ifdef INSTALL_MERGER_THREAD
		if(NULL != m_mergeMutex)
			sqlbeginmutex(&m_mergeMutex);
#endif
		MemProtect(foundMemoryAreaDescriptor->m_AddressOfUserSpace,foundMemoryAreaDescriptor->m_Size,MEMORYPROTECTION_NOACCESS);
		m_freeMemoryChain->add(foundMemoryAreaDescriptor);
#ifdef INSTALL_MERGER_THREAD
		if(NULL != m_mergeMutex)
			sqlendmutex(&m_mergeMutex);
#endif
	}
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::ProtectBlock(void * BlockStartAddress,SAPDB_ULong ProtectionLevel)
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::ProtectBlock", Memory_Trace, 1);

	if(BlockStartAddress == NULL)
		return;

	MemoryAreaDescriptor *memoryAreaDescriptor = getMemoryAreaDescriptor(BlockStartAddress);
	SAPDBERR_ASSERT_STATE(memoryAreaDescriptor != NULL);
	SAPDBERR_ASSERT_STATE(memoryAreaDescriptor->m_AddressOfUserSpace == BlockStartAddress);

    if(ProtectionLevel == MEMORYAREA_PROTECTIONLEVEL_FULL)
		MemProtect(memoryAreaDescriptor->m_AddressOfUserSpace,memoryAreaDescriptor->m_Size,MEMORYPROTECTION_NOACCESS);
	else if(ProtectionLevel == MEMORYAREA_PROTECTIONLEVEL_NONE)
		MemProtect(memoryAreaDescriptor->m_AddressOfUserSpace,memoryAreaDescriptor->m_Size,MEMORYPROTECTION_READWRITE);
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::merge(void)
{
#ifndef INSTALL_MERGER_THREAD
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::merge", Memory_Trace, 1);
#endif

#ifdef INSTALL_MERGER_THREAD
	if(NULL != m_mergeEntryMutex)
		sqlbeginmutex(&m_mergeEntryMutex);
#endif
	if(!m_mergeIsRunning)
	{
		m_mergeIsRunning = true;
#ifdef INSTALL_MERGER_THREAD
		if(NULL != m_mergeEntryMutex)
			sqlendmutex(&m_mergeEntryMutex);
#endif
    	MemoryAreaDescriptor *currentMemoryArea;

/* we start with the very last memory block, because a pointer to the previous memory area is stored directly in each memory area       */
/* structure. So, walking from the end of the chain of memory areas to the beginning is much cheaper than walking from the beginning    */
/* to the end.                                                                                                                          */
        currentMemoryArea = m_veryLastBlock;
		while(currentMemoryArea != NULL)
		{
#ifdef INSTALL_MERGER_THREAD
			if(NULL != m_mergeMutex)
				sqlbeginmutex(&m_mergeMutex);
#endif
			currentMemoryArea = currentMemoryArea->m_prevNeighbour;
			if(currentMemoryArea == NULL)
			{
#ifdef INSTALL_MERGER_THREAD
				if(NULL != m_mergeMutex)
					sqlendmutex(&m_mergeMutex);
#endif
				break;
			}
        	MemoryAreaDescriptor *precedingMemoryArea;
		
            precedingMemoryArea = currentMemoryArea->m_prevNeighbour;
			if(precedingMemoryArea == NULL)
			{
#ifdef INSTALL_MERGER_THREAD
				if(NULL != m_mergeMutex)
					sqlendmutex(&m_mergeMutex);
#endif
				break;
			}
			if(currentMemoryArea->isFree() && precedingMemoryArea->isFree())
			{
				removeMemoryAreaFromNeighbourhood(currentMemoryArea);
				m_freeMemoryChain->remove(currentMemoryArea);
				m_freeMemoryChain->remove(precedingMemoryArea);
				precedingMemoryArea->m_Size += currentMemoryArea->m_Size;
				precedingMemoryArea->m_Size += sizeofinfospace();
				makeMemoryAreaDescriptorAvailable(currentMemoryArea);
				m_freeMemoryChain->add(precedingMemoryArea);
			}
#ifdef INSTALL_MERGER_THREAD
			if(NULL != m_mergeMutex)
				sqlendmutex(&m_mergeMutex);
#endif
		}

#ifdef USE_MERGE_WAIT_SEMAPHORE
		for(;m_mergeWaitingCounter>0;m_mergeWaitingCounter--)
			sqlsignalsem(m_mergeSemaphore);
#endif

		m_mergeIsRunning = false;
	}
	else
	{
		m_mergeWaitingCounter++;
#ifdef INSTALL_MERGER_THREAD
		if(NULL != m_mergeEntryMutex)
    		sqlendmutex(&m_mergeEntryMutex);
#endif
#ifdef USE_MERGE_WAIT_SEMAPHORE
        sqlwaitsem(m_mergeSemaphore);
#endif
	}
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::initMemoryArea(MemoryAreaDescriptor *memoryAreaToInit,void *address,SAPDB_ULong size,MemoryAreaDescriptor *previousNeighbour)
{
/* with the current implementation:         */
/* calculate the user available addr and usable size */
/* protect the whole area                   */
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::initMemoryArea", Memory_Trace, 1);


	SAPDBERR_ASSERT_STATE(address != NULL);
	MemProtect(address,sizeofinfospace(),MEMORYPROTECTION_READWRITE);

	*(MemoryAreaDescriptor **)address = memoryAreaToInit;

	memoryAreaToInit->m_AddressOfUserSpace = (SAPDB_Byte *)address + sizeofinfospace();	/* leave the space for the info field */
	memoryAreaToInit->m_Size = size - sizeofinfospace();
	memoryAreaToInit->m_prevNeighbour = previousNeighbour;
	MemProtect(address,size,MEMORYPROTECTION_NOACCESS);
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::removeMemoryAreaFromNeighbourhood(MemoryAreaDescriptor *memoryAreaToRemove)
{
#ifndef INSTALL_MERGER_THREAD
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::removeMemoryAreaFromNeighbourhood", Memory_Trace, 1);
#endif
	MemoryAreaDescriptor *nextNeighbour = getNextNeighbour(memoryAreaToRemove);
	if(nextNeighbour != NULL)
		nextNeighbour->m_prevNeighbour = memoryAreaToRemove->m_prevNeighbour;
	memoryAreaToRemove->m_prevNeighbour = NULL;
}

/*------------------------------*/

void *RTEMem_ProtectingAllocator::mergerThreadproc(void *dummyParameter)
{
#ifndef INSTALL_MERGER_THREAD
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::mergerThreadProc", Memory_Trace, 1);
#endif

/* as this is a thread-main-function, this function must return a value, but we want it to loop forever.    */
/* If a simple "while(1)" would be used, some compilers would complain that the return statement would      */
/* never be reached. If a volatile variable is used, this warning is avoided                                */
    volatile SAPDB_Bool alwaysTrue = true;
	while(alwaysTrue)
	{
		printf("-mt-");
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
		sqlbeginmutex(&m_outMutex);
#endif
		merge();
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
		sqlendmutex(&m_outMutex);
#endif
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
		out();
#endif
	}
	return 0;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::MemProtect(void *address,SAPDB_ULong size,int type)
{
#ifndef INSTALL_MERGER_THREAD
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::MemProtect", Memory_Trace, 1);
#endif

	if(m_protectionLevel == MEMORYAREA_PROTECTIONLEVEL_FULL)
	{
		SAPDB_ULong result = RTE_ISystem::Instance().MemProtect(address,size,type);
		SAPDBERR_ASSERT_STATE(0 == result);
	}
}

/*------------------------------*/

SAPDB_ULong RTEMem_ProtectingAllocator::GetAlignmentSize() const
{
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::GetAlignmentSize", Memory_Trace, 1);

	return m_PageSize;
}

/*------------------------------*/

SAPDB_ULong RTEMem_ProtectingAllocator::sizeofinfospace(void)
{
#ifndef INSTALL_MERGER_THREAD
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::sizeofinfospace", Memory_Trace, 1);
#endif

#ifdef RETURN_NOT_ALIGNED_BLOCKS
	if(m_protectionLevel == MEMORYAREA_PROTECTIONLEVEL_NONE)
		return sizeof(MemoryArea *);
#else
	return m_PageSize;
#endif
}

/*------------------------------*/

SAPDB_ULong RTEMem_ProtectingAllocator::GetBlockSize() const
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::GetBlockSize", Memory_Trace, 1);

	return m_ChunkSizeInBytes;
}

/*------------------------------*/

SAPDB_ULong RTEMem_ProtectingAllocator::CalcBlockCount(SAPDB_ULong NumberOfBytes) const
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::CalcBlockCount", Memory_Trace, 1);

	return ( NumberOfBytes + m_ChunkSizeInBytes - 1 ) / m_ChunkSizeInBytes;
}

/*------------------------------*/

const SAPDB_UTF8 * RTEMem_ProtectingAllocator::GetIdentifier() const
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::GetIdentifier", Memory_Trace, 1);

	return (const SAPDB_UTF8 *)"RTEMem_ProtectingAllocator";
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::out(void)
{

	MemoryAreaDescriptor *currentMemoryArea;

	SAPDB_ULong totalfree = 0;
	currentMemoryArea = m_veryLastBlock;
#ifdef __GENERATE_OUTPUT
	printf("beginning of free memory chain: %lx\n",m_freeMemoryChain->m_firstElement);
#endif
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
	sqlbeginmutex(&m_outMutex);
#endif
	while(currentMemoryArea != NULL)
	{
		currentMemoryArea->out();
		totalfree += currentMemoryArea->m_Size;
		currentMemoryArea = currentMemoryArea->m_prevNeighbour;
		if(currentMemoryArea == NULL)
			break;
	}
	SAPDBERR_ASSERT_STATE(totalfree/1024/1024/16 == m_totalMemorySize/1024/1024/16);

	m_freeMemoryChain->checkAndOut();
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
	sqlendmutex(&m_outMutex);
#endif
#ifdef __GENERATE_OUTPUT
	printf("\n==================================================================\n");
#endif
}

/*------------------------------*/

SAPDB_ULong RTEMem_ProtectingAllocator::calcNeededBytes(SAPDB_ULong desiredSize)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::calcNeededBytes", Memory_Trace, 1);

	SAPDB_ULong reserveBytes = (((desiredSize-1)/m_ChunkSizeInBytes) * sizeofinfospace());
	if(reserveBytes >= m_ChunkSizeInBytes + sizeofinfospace())
		return desiredSize;
	else
		return desiredSize + reserveBytes;
}

/*------------------------------*/

void RTEMem_ProtectingAllocator::prepareBlock(void **address,SAPDB_ULong *size)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::prepareBlock", Memory_Trace, 1);

/* with the current implementation: align the block to page size, install a guard page at the end of the block  */
	SAPDB_ULong spaceBefore = sizeofinfospace() - ((SAPDB_ULong)*address % sizeofinfospace());
	SAPDB_ULong spaceAfter = (SAPDB_ULong)(((SAPDB_Byte *)*address+*size)) % sizeofinfospace();

	if(spaceBefore != 0)
	{
		*address = (SAPDB_Byte *)*address + spaceBefore;
		*size -= spaceBefore;
	}
	if(spaceAfter != 0)
	{
		*size -= spaceAfter;
	}
	*size -= sizeofinfospace();
}


/*------------------------------*/

MemoryAreaDescriptor *RTEMem_ProtectingAllocator::readMemoryAreaDescriptorFromProtectedAddr(void *address)
{
#ifndef INSTALL_MERGER_THREAD
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::readMemoryAreaDescriptorFromProtectedAddr", Memory_Trace, 1);
#endif
	MemoryAreaDescriptor *addressOfMemoryAreaDescriptor;
#ifdef INSTALL_MERGER_THREAD
	if(NULL != m_protectMutex)
		sqlbeginmutex(&m_protectMutex);
#endif
	MemProtect(address,sizeofinfospace(),MEMORYPROTECTION_READONLY);

	addressOfMemoryAreaDescriptor = *((MemoryAreaDescriptor **)address);

	MemProtect(address,sizeofinfospace(),MEMORYPROTECTION_NOACCESS);
#ifdef INSTALL_MERGER_THREAD
	if(NULL != m_protectMutex)
		sqlendmutex(&m_protectMutex);
#endif

	return addressOfMemoryAreaDescriptor;
}

/*------------------------------*/

MemoryAreaDescriptor *RTEMem_ProtectingAllocator::getMemoryAreaDescriptor(void *userSpaceAddr)
{
//    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::getMemoryAreaDescriptor", Memory_Trace, 1);

	return readMemoryAreaDescriptorFromProtectedAddr((SAPDB_Byte *)userSpaceAddr-sizeofinfospace());
}

/*------------------------------*/

MemoryAreaDescriptor *RTEMem_ProtectingAllocator::getNextNeighbour(MemoryAreaDescriptor *theMemoryArea)
{
#ifndef INSTALL_MERGER_THREAD
    SAPDBTRACE_ROUTINE_DEBUG("RTEMem_ProtectingAllocator::getNextNeighbour", Memory_Trace, 1);
#endif
	return readMemoryAreaDescriptorFromProtectedAddr((SAPDB_Byte *)theMemoryArea->m_AddressOfUserSpace
        + theMemoryArea->m_Size
        + theMemoryArea->m_ReserveSpaceSize);
}


/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                             class MemoryAreaDescriptor                    */
/*                                                                           */
/* instances of this class contain information about exactly one memory block*/
/* that is either allocated or free.                                         */
/*                                                                           */
/*****************************************************************************/

MemoryAreaDescriptor::MemoryAreaDescriptor():
    m_prevElement(NULL),
    m_nextElement(NULL),
    m_prevList(NULL),
    m_nextList(NULL),
    m_ReserveSpaceSize(0),
    m_State(MEMORYAREA_STATE_INUSE)

{
}


/*------------------------------*/

void MemoryAreaDescriptor::markInUse(void)
{
	SAPDBERR_ASSERT_STATE(m_State == MEMORYAREA_STATE_FREE);
	m_State = MEMORYAREA_STATE_INUSE;
}

/*------------------------------*/

bool MemoryAreaDescriptor::isFree(void)
{
	return m_State == MEMORYAREA_STATE_FREE;
}

/*------------------------------*/

void MemoryAreaDescriptor::markFree(void)
{
	SAPDBERR_ASSERT_STATE(m_State == MEMORYAREA_STATE_INUSE);
	m_State = MEMORYAREA_STATE_FREE;
}

/*------------------------------*/

void MemoryAreaDescriptor::out(void)
{
#ifdef __GENERATE_OUTPUT
	printf("MemoryArea %lx @%lx, size %ld (%ld MB)",this,m_AddressOfUserSpace,m_Size,m_Size/1024/1024);
	if(m_State == MEMORYAREA_STATE_FREE)
		printf("free ");
	else
		printf("in use, rSS %ld ",m_ReserveSpaceSize);
	printf("pn %lx\n",m_prevNeighbour);
	printf("next element %lx, next list %lx, prev. element %lx, prev. list %lx\n",m_nextElement,m_nextList,m_prevElement,m_prevList);
	printf("-\n");
#endif
}



/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*                           class FreeMemoryChain                           */
/*                                                                           */
/*                                                                           */
/*  the FreeMemoryChain is a list whose elements are "sub"lists containing   */
/*  MemoryAreaDescriptors. Each sublist contains the memory area descriptors */
/*  for all memory areas of the same size. The sublists are sorted according */
/*  to the size of the memory areas they contain the descriptors for.        */
/*                                                                           */
/*  In the following comments, "chain" means the FreeMemoryChain itself,     */
/*  whereas "list" means one of the sublists.                                */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/



FreeMemoryChain::FreeMemoryChain(void)
{
//    SAPDBTRACE_ROUTINE_DEBUG("FreeMemoryChain::FreeMemoryChain", Memory_Trace, 1);

	m_firstElement = NULL;
}

/*------------------------------*/

void FreeMemoryChain::add(MemoryAreaDescriptor *memoryAreaDescriptor)
{
#ifndef INSTALL_MERGER_THREAD
//    SAPDBTRACE_ROUTINE_DEBUG("FreeMemoryChain::add", Memory_Trace, 1);
#endif

	memoryAreaDescriptor->markFree();
    /* search for a "suitable" list. Suitable means: If a list exists that contains descriptors for memory areas with exactly  */
    /* the same size as the one that is to be added, return the head of this list */
	MemoryAreaDescriptor *foundList = searchForSuitableList(memoryAreaDescriptor->m_Size);	

	if(foundList == NULL)
	{
        /* If there is not a single sublist in the free memory chain, we add the new element at the top of the chain */
        m_firstElement = memoryAreaDescriptor;
		memoryAreaDescriptor->m_nextElement = NULL;
		memoryAreaDescriptor->m_nextList = NULL;
		memoryAreaDescriptor->m_prevElement = NULL;
		memoryAreaDescriptor->m_prevList = NULL;
	}
	else
	{
    	MemoryAreaDescriptor *precedingList;
		if(foundList == m_firstElement)	/* in some cases, f->m_prevList is set to an (invalid) value != 0 even if f is the first element of the list */
			precedingList = NULL;
		else
			precedingList = foundList->m_prevList;
		if(foundList->m_Size == memoryAreaDescriptor->m_Size)	/* a list with the size already exists */
		{
            /* If a list was found that contains descriptors for memory areas with the same size as the one that is to be added, */
            /* The new element is added at the top of this list and all references to the list must now point to the added element */
            if(precedingList != NULL)
			{
				precedingList->m_nextList = memoryAreaDescriptor;
			}
			else
			{
				m_firstElement = memoryAreaDescriptor;
			}
			memoryAreaDescriptor->m_nextElement = foundList;
			foundList->m_prevElement = memoryAreaDescriptor;
			memoryAreaDescriptor->m_nextList = foundList->m_nextList;
			memoryAreaDescriptor->m_prevList = precedingList;
			if(foundList->m_nextList != NULL)
				foundList->m_nextList->m_prevList = memoryAreaDescriptor;
		}
		else if(foundList->m_Size > memoryAreaDescriptor->m_Size)
		{
            /* If a list was found that contains descriptors with a bigger size than the one that is to be added */
            /* a new list has to created that contains the descriptor that is to be added. The new list has to be */
            /* enchained in the FreeMemoryChain before the one that was found */
            if(precedingList != NULL)
			{
				memoryAreaDescriptor->m_nextElement = NULL;
				memoryAreaDescriptor->m_prevElement = NULL;
				memoryAreaDescriptor->m_nextList = precedingList->m_nextList;
				precedingList->m_nextList = memoryAreaDescriptor;
				memoryAreaDescriptor->m_prevList = precedingList;
				foundList->m_prevList = memoryAreaDescriptor;
			}
			else
			{
				/* If the list that was found is the first list of the chain, the newly created list will be the first list */
                m_firstElement = memoryAreaDescriptor;
				memoryAreaDescriptor->m_nextElement = NULL;
				memoryAreaDescriptor->m_prevElement = NULL;
				memoryAreaDescriptor->m_prevList = NULL;
				memoryAreaDescriptor->m_nextList = foundList;
				foundList->m_prevList = memoryAreaDescriptor;
			}
		}
		else
		{
            /* If only a list was found that contains descriptors with a smaller size than the one that is to be added */
            /* a new list has to created that contains the descriptor that is to be added. The new list has to be */
            /* enchained in the FreeMemoryChain after the one that was found */
			foundList->m_nextList = memoryAreaDescriptor;
			memoryAreaDescriptor->m_prevList = foundList;
			memoryAreaDescriptor->m_nextList = NULL;
			memoryAreaDescriptor->m_nextElement = NULL;
			memoryAreaDescriptor->m_prevElement = NULL;
		}
	}
}

/*------------------------------*/

void FreeMemoryChain::remove(MemoryAreaDescriptor *memoryAreaDescriptor)
{
#ifndef INSTALL_MERGER_THREAD
//    SAPDBTRACE_ROUTINE_DEBUG("FreeMemoryChain::remove", Memory_Trace, 1);
#endif

	memoryAreaDescriptor->markInUse();
	if(memoryAreaDescriptor == m_firstElement)
	{
        /* If the memory area descriptor that is to be removed is at the top of the FreeMemoryChain, m_firstElement must get */
        /* a new value */
        if(memoryAreaDescriptor->m_nextElement != NULL)
		{
            /* If there are more memory areas with the same size as the one to be removed, m_firstElement will be  */
            m_firstElement = memoryAreaDescriptor->m_nextElement;
			if(memoryAreaDescriptor->m_nextList != NULL)
				memoryAreaDescriptor->m_nextList->m_prevList = m_firstElement;
			m_firstElement->m_nextList = memoryAreaDescriptor->m_nextList;
			m_firstElement->m_prevElement = NULL;
		}
		else
		{
			m_firstElement = memoryAreaDescriptor->m_nextList;
			if(m_firstElement != NULL)
			{
				m_firstElement->m_prevList = NULL;
				m_firstElement->m_prevElement = NULL;
			}
		}
	}
	else	/* remove from middle of chain  */
	{
		MemoryAreaDescriptor *previousMemoryArea;
		MemoryAreaDescriptor *nextMemoryArea;

		previousMemoryArea = memoryAreaDescriptor->m_prevElement;
		nextMemoryArea = memoryAreaDescriptor->m_nextElement;
		if(nextMemoryArea == NULL && previousMemoryArea == NULL)
		{
	        /* If the descriptor to be removed is the only element in the list, let the previous list have the following one */
            /* as successor and vice versa */
            SAPDBERR_ASSERT_STATE(memoryAreaDescriptor->m_prevList != NULL);
			memoryAreaDescriptor->m_prevList->m_nextList = memoryAreaDescriptor->m_nextList;
			if(memoryAreaDescriptor->m_nextList != NULL)
			{
				memoryAreaDescriptor->m_nextList->m_prevList = memoryAreaDescriptor->m_prevList;
			}
		}
		else
		{
			/* If the descriptor to be removed is not at the top of the chain and there are more memory areas with the same */
            /* size, just remove the element from its list */
            if(nextMemoryArea != NULL)	
			{
				nextMemoryArea->m_prevElement = previousMemoryArea;
			}
			if(previousMemoryArea != NULL)	/* not the first element of the list?  ->  */
			{
				previousMemoryArea->m_nextElement = memoryAreaDescriptor->m_nextElement;
			}
			else
			{
				if(memoryAreaDescriptor->m_prevList != NULL)
					memoryAreaDescriptor->m_prevList->m_nextList = nextMemoryArea;
				if(memoryAreaDescriptor->m_nextList != NULL)
					memoryAreaDescriptor->m_nextList->m_prevList = nextMemoryArea;
				if(nextMemoryArea != NULL)
				{
					nextMemoryArea->m_nextList = memoryAreaDescriptor->m_nextList;
					nextMemoryArea->m_prevList = memoryAreaDescriptor->m_prevList;
				}
			}
		}
	}
	memoryAreaDescriptor->m_prevElement = NULL;
	memoryAreaDescriptor->m_nextElement = NULL;
	memoryAreaDescriptor->m_prevList = NULL;
	memoryAreaDescriptor->m_nextList = NULL;
}

/*------------------------------*/

MemoryAreaDescriptor *FreeMemoryChain::searchForSuitableList(SAPDB_ULong neededSize)
{
/* search for the most suitable list to a given size: */
/* If a list exists whose elements are descriptors for memory area with the exactly the given size, return the head of this list */
/* If a such a list does not exist, but a least one list exists whose elements are descriptors for memory areas with a bigger    */
/* size than the given one, return the head of that list whose elements are descriptors for memory areas with the smallest size  */
/* (that size is yet bigger than the smallest one).   */
/* If there are only lists in the chain whose elements are descriptors to memory areas that are smaller than the given size,     */
/* return the head of the list whose elements are descriptors for memory areas with the biggest size. If there is not a single   */
/* list in the chain, return 0. */


#ifndef INSTALL_MERGER_THREAD
//    SAPDBTRACE_ROUTINE_DEBUG("FreeMemoryChain::search", Memory_Trace, 1);
#endif

	MemoryAreaDescriptor *currentElement = m_firstElement;
	MemoryAreaDescriptor *previousList = NULL;

	while(currentElement != NULL)
	{
		SAPDBERR_ASSERT_STATE(currentElement->isFree());
		if(currentElement->m_Size >= neededSize)
		{
			return currentElement;
		}
		previousList = currentElement;
		currentElement = currentElement->m_nextList;
	}
	return previousList;
}

/*------------------------------*/

void FreeMemoryChain::checkAndOut(void)
{
    SAPDBTRACE_ROUTINE_DEBUG("FreeMemoryChain::checkAndOut", Memory_Trace, 1);

	MemoryAreaDescriptor *nextList;
	MemoryAreaDescriptor *currentElement;
	SAPDB_ULong totalfree = 0;

	if(m_firstElement != NULL)
	{
		currentElement = m_firstElement;
		SAPDBERR_ASSERT_STATE(currentElement->isFree());
		nextList = currentElement->m_nextList;
		while(1)
		{
#ifdef __GENERATE_OUTPUT
//			pCurrElement->out();
#endif
			totalfree += currentElement->m_Size;
			currentElement = currentElement->m_nextElement;
			if(currentElement == NULL)
			{
				currentElement = nextList;
				if(currentElement == NULL)
					break;
				else
					nextList = nextList->m_nextList;
			}
		}
	}
#ifdef __GENERATE_OUTPUT
	else
	{
		printf("free memory chain is empty\n");
	}
#endif
}



/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*        global function RTEMem_InitializeProtectingAllocator()             */
/*                                                                           */
/*                                                                           */
/*  this routine initializes the allocator. The strings that contain the     */
/*  memory sizes to allocate are interpreted and the memory blocks are       */
/*  allocated. This is a global function so it can be called from C code.    */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

extern "C" bool RTEMem_InitializeProtectingAllocator(SAPDB_Long memoryChunkSizeInMB,SAPDB_Char *memSizeShared,SAPDB_Char *memSizeMmap,SAPDB_Char *memSizeVirtual,SAPDB_Char *memSizeHeap, SAPDB_ULong pageSize,SAPDB_Bool protectMemory)
{
    return ((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).Initialize(
        memoryChunkSizeInMB,
        memSizeShared,
        memSizeMmap,
        memSizeVirtual,
        memSizeHeap,
        pageSize,
        protectMemory);
}

bool RTEMem_ProtectingAllocator::Initialize(SAPDB_Long memoryChunkSizeInMB,SAPDB_Char *memSizeShared,SAPDB_Char *memSizeMmap,SAPDB_Char *memSizeVirtual,SAPDB_Char *memSizeHeap, SAPDB_ULong pageSize,SAPDB_Bool protectMemory)
{
	SAPDB_Long physicalMemoryInMB = RTE_ISystem::Instance().GetPhysicalMemorySize();

	printf("lPhysicalMemoryInMB: %ld\n",physicalMemoryInMB);

	IMemoryGrabber::SetParameters((double)physicalMemoryInMB * ONE_MB,memoryChunkSizeInMB * ONE_MB);

	if(0==*memSizeShared && 0==*memSizeMmap && 0==*memSizeVirtual && 0==*memSizeHeap)
	{
#if defined(_WIN32)
		memSizeVirtual = "PHYSMEM*2";
		memSizeMmap = "0";
		memSizeShared = "0";
#else
		memSizeMmap = (SAPDB_Char *)"PHYSMEM";
		memSizeShared = (SAPDB_Char *)"0";
		memSizeVirtual = (SAPDB_Char *)"0";
#endif
		memSizeHeap = (SAPDB_Char *)"0";
	}

	
	SAPDBTRACE_ROUTINE_DEBUG("RTEMem_InitializeProtectingAllocator", Memory_Trace, 1);

	((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).InitializeMultithreading();
	((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).SetParameters(memoryChunkSizeInMB,pageSize,protectMemory);


	void *reserveBlock;
#if defined(RESERVE_BLOCK_SIZE)

#if defined(OSF1)
	reserveBlock = malloc(RESERVE_BLOCK_SIZE);
	printf("OSF1 ReserveBlock: %ld\n",RESERVE_BLOCK_SIZE);
#else
	reserveBlock = RTE_ISystem::Instance().AllocSystemPages(RESERVE_BLOCK_SIZE,true);
	printf("ReserveBlock: %ld\n",RESERVE_BLOCK_SIZE);
#endif

#endif  
	VallocMemoryGrabber vallocMemoryGrabber;
	MmapMemoryGrabber mmapMemoryGrabber;
	HeapMemoryGrabber heapMemoryGrabber;
	ShmMemoryGrabber shmMemoryGrabber;

#ifdef _WIN32
	DWORD dwErr;
	if(NULL == reserveBlock)
	{
		dwErr = GetLastError();
	}
#endif

	shmMemoryGrabber.GrabAmountOfMemory(memSizeShared);
	vallocMemoryGrabber.GrabAmountOfMemory(memSizeVirtual);
	mmapMemoryGrabber.GrabAmountOfMemory(memSizeMmap);
	heapMemoryGrabber.GrabAmountOfMemory(memSizeHeap);

#if defined(RESERVE_BLOCK_SIZE)
#if defined(OSF1)
	printf("OSF1 freeing reserve block\n");
	free(reserveBlock);
#else
	printf("freeing reserve block\n");
	RTE_ISystem::Instance().FreeSystemPages(reserveBlock,RESERVE_BLOCK_SIZE);
#endif

#endif
	SAPDB_Bool emergency=false;

    /* if no memory was allocated, an emergency message is produced. To do this, a certain amount of memory is necessary... */
    if(!vallocMemoryGrabber.successful() && !mmapMemoryGrabber.successful() && !heapMemoryGrabber.successful() && !shmMemoryGrabber.successful())	
	{
    	void *emergencyMemoryBlock;
		emergencyMemoryBlock = malloc(ONE_MB + 3*RTEMem_ProtectingAllocator::sizeofinfospace());
		((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).AddBlock(emergencyMemoryBlock,ONE_MB + 3*RTEMem_ProtectingAllocator::sizeofinfospace());
		emergency = true;
	}
	if(0 == physicalMemoryInMB)
	{
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEWARN_SYS_COULD_NOT_DETERMINE_PHYSMEM));
	}
#if 0
    RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEINFO_SYS_PROTECT_ALLOC_SUCCESSFUL,SAPDB_ToString(((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).m_FirstBlock.m_Size/ONE_MB),"first"));
	RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEINFO_SYS_PROTECT_ALLOC_SUCCESSFUL,SAPDB_ToString(((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).m_SecondBlock.m_Size/ONE_MB),"second"));
	RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEINFO_SYS_PROTECT_ALLOC_SUCCESSFUL,SAPDB_ToString(((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).m_ThirdBlock.m_Size/ONE_MB),"third"));
	RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEINFO_SYS_PROTECT_ALLOC_SUCCESSFUL,SAPDB_ToString(((RTEMem_ProtectingAllocator &)RTEMem_ProtectingAllocator::Instance()).m_FourthBlock.m_Size/ONE_MB),"fourth"));
#endif
    vallocMemoryGrabber.ProcessState();
	mmapMemoryGrabber.ProcessState();
	heapMemoryGrabber.ProcessState();
	shmMemoryGrabber.ProcessState();

	if(emergency)
	{
		RTE_Message(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_SYS_COULD_NOT_ALLOCATE_MEMORY));
		RTE_Crash(SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_SYS_COULD_NOT_ALLOCATE_MEMORY));
	}
	return true;
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
