#ifndef RTEMEM_PROTECTINGALLOCATOR_HPP
#define RTEMEM_PROTECTINGALLOCATOR_HPP


#include "heo07.h"

#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IBlockAllocator.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"

#include <stdio.h>

#define NUMBER_OF_FIRST_BLOCKS 20

class MemoryAreaDescriptor
{
public:
	MemoryAreaDescriptor();

	void markFree(void);

	void markInUse(void);

	void out(void);
	
	bool isFree(void);

	MemoryAreaDescriptor *m_nextElement,*m_nextList,*m_prevElement,*m_prevList,*m_prevNeighbour;
	void *m_AddressOfUserSpace;
	SAPDB_ULong m_Size;

	SAPDB_ULong m_ReserveSpaceSize;
	
private:
    typedef enum
    {
        MEMORYAREA_STATE_FREE,
        MEMORYAREA_STATE_INUSE
    }
    memoryAreaState;
    memoryAreaState m_State;
}
;

class FreeMemoryChain
{
public:
	void add(MemoryAreaDescriptor *memoryAreaDescriptor);

	MemoryAreaDescriptor *searchForSuitableList(SAPDB_ULong neededSize);

	void remove(MemoryAreaDescriptor *memoryArea);

	MemoryAreaDescriptor *m_firstElement;

	void checkAndOut(void);

	FreeMemoryChain(void);

private:
	teo07_Mutex m_mutex;
};

class IMemoryGrabber
{
public:
	IMemoryGrabber();
	virtual void GrabAmountOfMemory(SAPDB_Char *expression);
	virtual SAPDB_Bool successful(void);

	virtual void ProcessState();

	static void SetParameters(double physicalMemoryInBytes,SAPDB_ULong minMemAreaSizeInBytes);

protected:
    typedef enum
    {
        MEMGRABBER_SYNTAXERROR,
        MEMGRABBER_FAILEDTOALLOCATEPHYSMEM,
        MEMGRABBER_NOTAVAILABLE,
        MEMGRABBER_DIDNOTHING,
        MEMGRABBER_SUCCESSFUL,
        MEMGRABBER_PARTLY_SUCCESSFUL
    }
    memoryGrabberState;

    memoryGrabberState m_State;
	SAPDB_Char *m_Name;
	SAPDB_ULong m_MemoryAllocatedInBytes;
	virtual SAPDB_ULong GrabMemoryBlock(SAPDB_ULong SizeInBytes);
	virtual SAPDB_ULong GetMaxRealisticSize(double RemainingSizeInBytes);
private:
	double interpret(SAPDB_Char *expression);
	virtual void *AllocMemBlock(SAPDB_ULong sizeInBytes);

	static double m_PhysicalMemoryInBytes;
	static SAPDB_ULong m_MinMemAreaSizeInBytes;

    SAPDB_Char *m_Expression;
};

class VallocMemoryGrabber : public IMemoryGrabber
{
public:
	VallocMemoryGrabber();
#ifdef _WIN32
	virtual SAPDB_ULong GrabMemoryBlock(SAPDB_ULong SizeInBytes);
#endif
};

class MmapMemoryGrabber : public IMemoryGrabber
{
public:
	MmapMemoryGrabber();
#ifndef _WIN32
	virtual void *AllocMemBlock(SAPDB_ULong SizeInBytes);
#endif
};


class ShmMemoryGrabber : public IMemoryGrabber
{
public:
	ShmMemoryGrabber();
#ifndef _WIN32
	virtual void *AllocMemBlock(SAPDB_ULong SizeInBytes);
#endif
};


class HeapMemoryGrabber : public IMemoryGrabber
{
public:
	HeapMemoryGrabber();
	virtual void *AllocMemBlock(SAPDB_ULong SizeInBytes);
#ifdef LINUX
	SAPDB_ULong GetMaxRealisticSize(double RemainingSizeInBytes);
#endif
};


/*! 
  Class:       BasicMemoryBlock
  
  Description: 

    this class contains just the basic information about a memory block. m_AddrOfUserSpace is the address returned to the caller
    in a malloc() call. There is usually housekeeping data stored just before this address. m_BaseAddr is the start address
    of these data.
    This class is only used internally, currently only to describe blocks that are allocated before the allocator is fully
    initialized.

 */
struct BasicMemoryBlock
{
    void *m_BaseAddr;
    void *m_AddrOfUserSpace;
    SAPDB_ULong m_Size;
	
    BasicMemoryBlock():
        m_BaseAddr(NULL),
        m_AddrOfUserSpace(NULL),
        m_Size(0)
    {
    };
};

/*! 
  Class:       RTEMem_ProtectingAllocator
  
  Description: 

    Only a singleton instances of the class is creatable. It is used
    to allocate chunks...

    To get an RTEMem_ProtectingAllocator::Instance() no argument is needed.
 */
class RTEMem_ProtectingAllocator : public SAPDBMem_IBlockAllocator
{
public:
	
    /*! 
      Function:    Instance
  
      Description: Returns the singleton RTE Protecting Allocator working with system memory pages

      Return value: The protecting allocator instance for working with memory pages
     */
	static SAPDBMem_IBlockAllocator & Instance();

    /*! 
      Function: GetBlockSize
	  
      Description: Returns the size in bytes of the blocks that can be allocated by this allocator.

       Returns blocksize setup by the constructor. Returns size of a system memory page in bytes.
                 
      Return value: The system memory page in bytes used for block count parameter
     */
	virtual SAPDB_ULong GetBlockSize() const ;

    /*! 
      Function: CalcBlockCount
	  
      Description: Returns the number of blocks needed to allocate given number of bytes
    
       Since Allocate() needs the number of blocks instead of number of bytes, this helper 
       routine exists to calculate the right argument. When used calling the Routine 
       GetBlockSize() is not needed.

      Arguments: NumberOfBytes [in] Number of bytes instead of number of blocks

      Return value: The number of system memory page needed for given number of bytes  
    */
	virtual SAPDB_ULong CalcBlockCount(SAPDB_ULong NumberOfBytes) const ;

    /*!
      Function: CalcStatistics
      Description:  Outputs statistics about the memory usage of the allocator.
      Arguments:    BytesUsed [out]
                    The number of bytes used by this allocater.
                    BytesControlled [out]
                    The number of bytes controlled by this allocator.

                    The default implementation returns GetInvalidSize() for 
                    both values.
      Return value: none
    */
	virtual void CalcStatistics( SAPDB_ULong &BytesUsed,
                               SAPDB_ULong &BytesControlled);

    /*! 
      Function: GetAlignmentSize
    
      Description: Returns the alignment size in bytes of the blocks
        that can be allocated by this allocator.

        An implementation must ensure that the alignment size A and
        the block size B satisfy B = N * A, where N is a non-zero
        integer number.

      Return value: The system page size as defined by the underlying operating system
     */
	virtual SAPDB_ULong GetAlignmentSize() const ;

    /*! 
      Function: Allocate

      Description: Allocates contiguous memory for 'count' blocks of memory.

        An implementation grants that a pointer returned will be an
        integral multiple of the system page size.

        The memory is not initialized.  

      Arguments: BlockCount [ in ] Number of blocks (NOT bytes!) to allocate.
    
      Return value: pointer to the block allocated
     */
	virtual void * Allocate(SAPDB_ULong BlockCount);

    /*! 
      Function: Deallocate
    
      Description: Frees the memory. 

       None of the objects in the memory block will be deinitialised
       by this call.  The size argument must match the value passed
       to 'allocate' to obtain this memory. The pointer argument
       must not be null.

      Arguments: BlockStartAddress [ in ] the pointer returned by a previous allocate call.
                 BlockCount [ in ] Number of blocks (NOT bytes!) to deallocate.
     */
	virtual void Deallocate(
     void * BlockStartAddress, 
     SAPDB_ULong BlockCount );

	virtual void ProtectBlock(void * BlockStartAddress,SAPDB_ULong ProtectionLevel);
	
    /*!
     function: GetBaseAllocatorCallStatistics
     description:  Outputs statistics about the number of Allocate and Deallocate calls of
                   the allocator to the underlying base allocator.
     arguments:    CountAlloc [out]
                   The number of calls of method Allocate of the base allocator.
                   CountDealloc [out]
                   The number of calls of method Deallocate of the base.

     return value: none
    */

	virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;

    /*!
     Function: GetCallStatistics
     Description:  Outputs statistics about the number of Allocate and Deallocate calls.
     Arguments:    CountAlloc [out]
                   The number of calls of method Allocate.
                   CountDealloc [out]
                   The number of calls of method Deallocate.

     return value: none
    */

	virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;


     /*!
      Function: GetIdentifier
      Description: Return identifier
      Return value: const SAPDB_UTF8 * "RTE_RawAllocator"
     */

	virtual const SAPDB_UTF8 * GetIdentifier() const ;


    bool Initialize(SAPDB_Long memoryChunkSizeInMB,SAPDB_Char *memSizeShared,SAPDB_Char *memSizeMmap,SAPDB_Char *memSizeVirtual,SAPDB_Char *memSizeHeap, SAPDB_ULong pageSize,SAPDB_Bool protectMemory);

private:
	virtual void AddBlock(void *addr,SAPDB_ULong Size);

	void InitializeMultithreading(void);

	void SetParameters(SAPDB_Long chunkSizeInMB,SAPDB_ULong pageSize,SAPDB_Bool ProtectMemory);

	static SAPDB_ULong sizeofinfospace(void);
#ifdef __OUTPUT_MEMAREAS_AFTER_EACH_STEP
	static teo07_Mutex m_outMutex;
#endif
	
	static void merge(void);

	static void out(void);

	static teo07_Mutex m_mergeMutex;
	static teo07_Mutex m_mergeEntryMutex;
	static teo07_ThreadSemaphore m_mergeSemaphore;
	static teo07_Mutex m_protectMutex;
	static SAPDB_Bool m_mergeIsRunning;
	static SAPDB_Int m_mergeWaitingCounter;
	static SAPDB_ULong m_totalMemorySize;
	
    static void *mergerThreadproc(void *egal);

    typedef enum
    {
        MEMORYAREA_PROTECTIONLEVEL_FULL,
        MEMORYAREA_PROTECTIONLEVEL_NONE
    }
    memoryAreaProtectionLevel;

    static memoryAreaProtectionLevel m_protectionLevel;

	static void MemProtect(void *address,SAPDB_ULong Size,int type);

	static MemoryAreaDescriptor *readMemoryAreaDescriptorFromProtectedAddr(void *address);

    SAPDB_ULong calcNeededBytes(SAPDB_ULong userSize);

	MemoryAreaDescriptor *getMemoryAreaDescriptor(void *userSpaceAddr);

	void prepareBlock(void **address,SAPDB_ULong *size);

	RTEMem_ProtectingAllocator(SAPDB_ULong BlockSize=0);  /* Private constructor */

	static FreeMemoryChain *m_freeMemoryChain;

	static SAPDB_ULong m_ChunkSizeInBytes;                       /* The size of a block defined by the user */
	static SAPDB_ULong m_PageSize;                       

	RTESync_InterlockedCounter<SAPDB_ULong> m_BytesControlled;
	RTESync_InterlockedCounter<SAPDB_ULong> m_CountAlloc;
	RTESync_InterlockedCounter<SAPDB_ULong> m_CountDealloc;
	static RTEMem_ProtectingAllocator *m_Instance; /* The Memory Page Block allocator instance */
	static MemoryAreaDescriptor *m_veryLastBlock; /* must be static because it is used in merge() which can be called from a thread-main-function which must be static */

	static void removeMemoryAreaFromNeighbourhood(MemoryAreaDescriptor *theMemoryArea);

	int m_allocations;

	static void makeMemoryAreaDescriptorAvailable(MemoryAreaDescriptor *theMemoryAreaDescriptor);
	void *getAvailableMemoryAreaDescriptor(void);

	void initMemoryArea(MemoryAreaDescriptor *memoryAreaToInit,void *addr,SAPDB_ULong size,MemoryAreaDescriptor *previousNeighbour);

	static MemoryAreaDescriptor *getNextNeighbour(MemoryAreaDescriptor *theMemoryArea);

	static MemoryAreaDescriptor *m_NextAvailableMemoryAreaDescriptor;

	SAPDB_Bool m_FullyInitialized;

    /* The allocator must be initialized before it can be used to allocate memory blocks. But unfortunately, memory has to be */
    /* allocated to determine the values the allocator is to be initialized with... To solve this dilemma, the allocator simply */
    /* calls malloc() when Allocate() is called before initialization. The sizes and addresses of blocks allocated this way */
    /* are stored in BasicMemoryBlock structures. As with the current implementation, Allocate() is called no more than two */
    /* times before initialization, two simple BasicMemoryBlocks are used */
#if 0
    BasicMemoryBlock m_FirstBlock;
	BasicMemoryBlock m_SecondBlock;
	BasicMemoryBlock m_ThirdBlock;
	BasicMemoryBlock m_FourthBlock;
#else
    BasicMemoryBlock m_FirstBlocks[NUMBER_OF_FIRST_BLOCKS];
#endif

friend class IMemoryGrabber;
friend class VallocMemoryGrabber;
};


#endif