/*!***************************************************************************

  module      : RTE_ISystem.hpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: System Interface
  description : Interface class to hide system dependencys from RTE_Global class

  last changed: 2000-10-23  10:49
  see also    : example.html ...
  first created:2000-02-28  15:52

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef RTE_ISYSTEM_HPP
#define RTE_ISYSTEM_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/RTE_Types.hpp"
#include "RunTime/System/RTESys_MemProtect.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! Class:       RTE_ISystem
    Description: Interface class to access to system dependend information

                 This pure virtual class abstracts from system information.
                 The implementation must derive from this class.
                 Caller must not depend on actual size of RTE_ISystem pointer.
 */
class RTE_ISystem 
{
public:
  /*!---------------------------------------------------------------------------
    Declaration:  Public types
    Description:  Public types used by public member functions.
  */
  typedef struct { SAPDB_Int2   Year;
                   SAPDB_Int2   Month;
                   SAPDB_Int2   DayOfWeek;
                   SAPDB_Int2   Day;
                   SAPDB_Int2   Hour;
                   SAPDB_Int2   Minute;
                   SAPDB_Int2   Second;
                   SAPDB_Int2   Milliseconds;} DateTime;
  /*! EndDeclaration:  Public types */

  static const SAPDB_ULong       UnlimitedMemory;

  static const RTE_SystemRc NoError;

public:
  /*!
    Function: Instance
    description: Return the system instance

    Since there is only one system, implementation of this is done in the system
    specific implementation module.
   */
  static RTE_ISystem & Instance();

  /*!
    Function: GiveUpTimeSlice
    Descriptions: Give up the time slice

    In multithreaded environment this allows the scheduler to select the next thread in the
    current process.

    In singlethreaded environment this allows the system to schedule the next process.

    It makes no difference weather this is called within a task context (in term of internal tasking)
    or within a 'normal' context.
   */
  static void   GiveUpTimeSlice();

  /*!
    Function: GetSystemPageSize
    Descriptions: Return the size of a memory page in bytes
    
    Size of memory pages handled in the operating system depends on architecture.
    The value maybe not fixed, so this call returns the system configuration value.

    Return value: The system memory page size in bytes
   */
  virtual const SAPDB_ULong GetSystemPageSize() const =0;

  /*!
    Function: RoundUpToSystemPageSize
    Description: Round up to multiple size of a memory page in bytes
    
    Return value: The size round up to system memory page size in bytes
   */
  virtual const SAPDB_ULong RoundUpToSystemPageSize(SAPDB_ULong numberOfBytes) const = 0;

  /*!
    Function: GetPhysicalMemorySize
    Description: Return size of phsical memory in bytes
    Return value: Physical memory size in bytes
   */
  virtual const SAPDB_ULong GetPhysicalMemorySize() const =0;

  /*!
    Function: MemProtect
    Description: Protect the given memory range

    Arguments: pvAdr [in] Start address (must be system memory page aligned...)
               ulSize [in] size of memory area to protect (must be multiple of system memory page)
               type [in] MEMORYPROTECTION_READWRITE, MEMORYPROTECTION_READONLY, MEMORYPROTECTION_NOACCESS
    Return value: RTE_ISystem::NoError if call succeeded, else system error code directly returned
   */
  virtual RTE_SystemRc MemProtect(void *pvAdr,SAPDB_ULong ulSize,SAPDB_Int type) =0;

  /*!
    Function: GetNumberOfCPU
    Descriptions: Return the maximum number of CPUs available on the system
    
    This call only returns the maximum number of CPUs available on the machine. If an external tool assign
    specific CPUs to specific processes or assign a group of CPUs to the program, it will not be detected.
    The call is mainly for deciding if Spinlocks should do some hot loops before giving up time slice.

    Return value: Maximum number of CPUs avaliable in the machine
   */
  virtual const SAPDB_UInt4 GetNumberOfCPU() const =0;

  /*!
    Function: AllocSystemPages
    Descriptions: Allocate to given number of bytes as SystemPages
    
    This function allocates the given number of bytes as system pages .
    Size of a system page is returned by GetSystemPageSize().
    The rest of a page is returned to the caller too...
    
    argument: SizeInBytes [ in ] Number of consecutive bytes to allocate
              DoCommit [ in ] Flag to indicate commit wanted
    Return value: Start address of system pages or NULL if no more pages
   */
  virtual void * AllocSystemPages(SAPDB_ULong SizeInBytes, bool DoCommit=true) =0;

  /*!
    Function: ReserveSystemPages
    Descriptions: Reserve the virtual address space for the given number of bytes as SystemPages
    
    This function looks up space for the given number of bytes as system pages .
    Size of a system page is returned by GetSystemPageSize().
    The rest of a page is reserved for the caller too, but no physical pages are allocated, not even uncommitted.
    This function works only since the address space on the 64Bit maschines allows huge amount 
    The address can be used later in a call to AllocSystemPagesAtFixedAddress()

    argument: SizeInBytes [ in ] Number of consecutive bytes to allocate
    Return value: Start address of the first reserved system page or NULL if no more pages
   */
  virtual void * ReserveSystemPages(SAPDB_ULong SizeInBytes) =0;

  /*!
    Function: AllocSystemPagesAtFixedAddress
    Descriptions: Allocate to given number of bytes as SystemPages at given address
    
    This function allocates the given number of bytes as system pages.
    Size of a system page is returned by GetSystemPageSize().
    The rest of a page is returned to the caller too...
    If the FixedAddress is already occupied, the function fails and returns NULL.
    
    argument: FixedAddress [ in ] Address where to allocate the system pages or NULL if any address
              SizeInBytes [ in ] Number of consecutive bytes to allocate
              DoCommit [ in ] Flag to indicate commit wanted
    Return value: Fixed address of system pages or NULL if no more pages or address already used
   */
  virtual void * AllocSystemPagesAtFixedAddress(void *FixedAddress, SAPDB_ULong SizeInBytes, bool DoCommit=true) =0;

  /*!
    Function: CommitSystemPages
    Descriptions: Commit the given number of bytes as SystemPages
    
    This function commits the given number of bytes as system pages at the given start address.
    The start address is returned by a previous call to AllocSystemPages() with DoCommit
    flag set to false. If call fails NULL is returned.

    argument: FirstPageAddress [ in ] Start address of block in pages
              NumberOfBytes [ in ] Number of consecutive bytes to commit
    Return value: FirstPageAddress or NULL if call failed.
   */
  virtual void * CommitSystemPages(void *FirstPageAddress, SAPDB_ULong NumberOfBytes) =0;

  /*!
    Function: FreeSystemPages
    Descriptions: Allocate to given number of bytes as SystemPages
    
    Size of a system page is returned by GetSystemPageSize().
    NumberOfBytes is always rounded up, so a last partial page is freed too!

    Arguments: FirstPageAddress [ in ] Start address of block in pages
              NumberOfBytes [ in ] Number of consecutive bytes to free
              
    Return value: true if call succeeded, false else.
   */
  virtual bool   FreeSystemPages(void *FirstPageAddress, SAPDB_ULong NumberOfBytes) =0;

  /*!
    Function: SetMemoryLimitInBytes
    Descriptions: Set the total limit of Memory
    
    If memory limit cannot be set during initialization, it is possible to delay setting
    the limit until the really wanted / configured limit is known.

    If memory limit is already exceeded, the current used memory is used as new limit,
    so next allocation would fail!!!!

    Arguments: NewMemoryLimit [in] New memory limit to set
               OldMemoryLimit [out] Former memory limit found
  
    Return value: Now used memory limit (can differ from wanted limit, see above)
   */
  virtual SAPDB_ULong SetMemoryLimitInBytes(SAPDB_ULong NewMemoryLimit, SAPDB_ULong & OldMemoryLimit) =0;

  /*!
    Function: GetMemoryStatistics
    Descriptions: Get the current limit of MemoryUsage
    
    If MemoryLimit cannot be set during initialization, it is possible to delay setting
    the limit until the really wanted / configured limit is known.

    Arguments: MemoryLimitInBytes    [out] actual MemoryLimit (or RTE_ISystem::UnlimitedMemory)
               MemoryCommitedInBytes [out] actual amount of memory commited
               MemoryUsedInBytes     [out] actual amount of memory used
               MemorySuccessfullAllocs   [out] actual number of successfull allocation requests
               MemoryFailedAllocs        [out] actual number of failed allocation requests
               MemoryCommits             [out] actual number of commit requests
               MemoryFrees               [out] actual number of deallocation requests
   */
  virtual void GetMemoryStatistics( SAPDB_ULong & MemoryLimitInBytes,
                                 SAPDB_ULong & MemoryCommitedInBytes,
                                 SAPDB_ULong & MemoryUsedInBytes,
                                 SAPDB_UInt8 & MemorySuccessfullAllocs,
                                 SAPDB_UInt8 & MemoryFailedAllocs,
                                 SAPDB_UInt8 & MemoryCommits,
                                 SAPDB_UInt8 & MemoryFrees) =0;

  /*!
    Function: InitLock
    description: Initialization of Atomic lock area

    Arguments: lock [ in ] The lock space
   */
  static void InitLock( RTE_Lock &lock );

  /*!
    Function: TestAndLock
    description: Atomic test and lock

    Arguments: lock [ in ] The lock space
    Return value: true if the lock space now locked by caller, false if not
   */
  static bool TestAndLock( RTE_Lock &lock );

  /*!
    Function: Unlock
    description: Unlock Atomic lock area

    Arguments: lock [ in ] The lock space
   */
  static void Unlock( RTE_Lock &lock );


  /*!
    Function: GetLocalDateTime
    Description: GetLocalDateTime returns the local time

    Arguments: LocalDateTime [ out ] - local date and time
   */
  static void GetLocalDateTime( RTE_ISystem::DateTime  &LocalDateTime );

  /*!
    Function: DoSleep
    Description: Sleep the specified number of milliseconds
    
    This actually calls the system function, that reschedules the current thread.
    
    Arguments: NumberOfMilliseconds [ in ] - Number of milliseconds to sleep
   */
  static void DoSleep(SAPDB_UInt4 NumberOfMilliseconds);

private:
  /*!
    Function: Initialize
    description: Initialize the system instance

    Since there is only one system, implementation of this is done in the system
    specific implementation module. Parameters given are the user imposed limits
    on the system. These parameters are known by the RTE_IGlobal

    Arguments: MemoryLimit [in] Limitation to dynamic memory usage
   */
  static void Initialize(SAPDB_ULong MemoryLimitInBytes=UnlimitedMemory);

  friend class RTE_IGlobal;
}; 
/*! EndClass: RTE_ISystem */

#endif  /* RTE_ISYSTEM_HPP */
