/*!***************************************************************************

  module      : RTE_KGS.hpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Kernel
  description : RTE Kernel Global Section

  Singleton used to reference all global information of the Kernel Runtime
  to ease debugging tasks.

  last changed: 2001-04-24  13:36

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef RTE_KGS_HPP
#define RTE_KGS_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Singleton.hpp"
#include "geo002.h"
#define KERNEL_LZU
#include "geo50_0.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

// forward declerations
class RTE_IGlobal;
class RTE_ISystem;          // OS specific singleton
class RTEMem_Allocator;     // Synchronized allocator for common use
class RTEMem_PageAllocator; // Pageallocator used for I/O Manager Cache
class SAPDBMem_IRawAllocator; // Raw allocator interface class

/*!---------------------------
  Class: RTE_KGS
  Description: Container class for all global references

  To ease debugging all global singletons etc. should be added to this class.
  The header file itself should only use forward declerations and never be 
  dependend on the real implementation of the items referenced. Same is for structures
  referenced etc.

  The static member RTE_KGS::Instance() allows anybody who included this header to access
  this singleton.
 */
class RTE_KGS : public SAPDB_Singleton
{
public:
#if defined(WIN32)
    typedef KERNEL_GLOBAL_SEC     KernelGlobalSection;
#else
    typedef KERNEL_GLOBAL_SECTION KernelGlobalSection;
#endif

    /*!
      Function: Instance
      Description: Returns the singleton instance
      Return value: Reference to the one and only 
     */
    static RTE_KGS & Instance();

    //---------------------------------------------------------------------
    // Singletons, that need no initialization code for instantiation
    // Public interface is
    //   name & Get##name()
    // i.e. RTEMem_Allocator & GetRTEMem_Allocator()
#define RTE_KGS_REFERENCED_SINGLETON(name_) \
public: \
    name_ & Get##name_() { return m_##name_; } \
private: \
    name_ & m_##name_

    RTE_KGS_REFERENCED_SINGLETON(RTE_IGlobal);
    RTE_KGS_REFERENCED_SINGLETON(RTE_ISystem);
#undef RTE_KGS_REFERENCED_SINGLETON

    // RTEMem_Allocator::Instance returns a different type, so macro above cannot be used
public:
    SAPDBMem_IRawAllocator & GetRTEMem_Allocator() { return m_RTEMem_Allocator; }
private:
    SAPDBMem_IRawAllocator & m_RTEMem_Allocator;

    // Singletons, that cannot exist without initialization code.
    // Public interfaces are
    //   name & Get##name()
    //   void   Set##name(name *pointerToInstance)
    // i.e. RTEMem_PageAllocator & GetRTEMem_PageAllocator()
    //      void                   SetRTEMem_PageAllocator(RTEMem_PageAllocator *pointerToInstance)
    //
#define RTE_KGS_POINTER_REFERENCED_SINGLETON(name_) \
public: \
    void Set##name_(name_ * val) { m_##name_ = val; } \
public: \
    name_ & Get##name_() { return *m_##name_; } \
private: \
    name_ * m_##name_

    RTE_KGS_POINTER_REFERENCED_SINGLETON(RTEMem_PageAllocator);
#undef RTE_KGS_POINTER_REFERENCED_SINGLETON

public:
    /*!---------------------------------------------------------------
      Function: GetKGS
      Description: Return pointer to old C KernelGlobalSection
      Return value: Pointer to old C KernelGlobalSection
     */
    KernelGlobalSection *GetKGS() { return m_KGS; }

    /*!---------------------------------------------------------------
      Function: GetXParam
      Description: Return pointer to old XParam record
      Return value: Pointer to old C XParam record
     */
    struct rte_xparam_record *GetXParam() { return m_KGS->XParam; }

    /*!---------------------------------------------------------------
      Function: SetKGS
      Description: Set the pointer to old C KernelGlobalSection
      Arguments: pointerToKGS [in] pointer to old C KernelGlobalSection
     */
    void SetKGS(KernelGlobalSection *pointerToKGS) { m_KGS = pointerToKGS; }

private:
    RTE_KGS();
    static RTE_KGS *m_Instance;
    KernelGlobalSection *m_KGS;
}; /*! Endclass: RTE_KGS */

#endif  /* RTE_KGS_HPP */
