/*!***************************************************************************

  module      : RTE_Types.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: RunTime Typedefinitions
  description : Contains all common RunTime Type definitions

  last changed: 2000-06-15  14:25
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef RTE_TYPES_H
#define RTE_TYPES_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.h"

/* Special include files for atomic operations */

#if !defined(_WIN32)
#include <sys/types.h>
#endif /* !WIN32 */

#if defined(AIX)
# include  <sys/atomic_op.h>
#endif /*AIX*/

#if defined(NMP)
# include <ulocks.h>
#endif /* NMP */

#if defined (_WIN32) && !defined (NO_WINDOWSH)
#include <windows.h>
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/* --- mapped system return codes (see also: RTE_SystemRc) */
#if defined(_WIN32)
# define RTE_SYSTEMRC_NO_ERROR                   NO_ERROR
# define RTE_SYSTEMRC_FILE_NOT_FOUND             ERROR_FILE_NOT_FOUND
# define RTE_SYSTEMRC_PATH_NOT_FOUND             ERROR_PATH_NOT_FOUND
# define RTE_SYSTEMRC_ALREADY_EXISTS             ERROR_ALREADY_EXISTS
# define RTE_SYSTEMRC_ACCESS_DENIED              ERROR_ACCESS_DENIED
# define RTE_SYSTEMRC_NOT_ENOUGH_MEMORY          ERROR_NOT_ENOUGH_MEMORY
# define RTE_SYSTEMRC_OUTOFMEMORY                ERROR_OUTOFMEMORY
#else /* UNIX */
# define RTE_SYSTEMRC_NO_ERROR                   0
# define RTE_SYSTEMRC_FILE_NOT_FOUND             ENOENT
# define RTE_SYSTEMRC_PATH_NOT_FOUND             ENOENT
# define RTE_SYSTEMRC_ALREADY_EXISTS             EEXIST
# define RTE_SYSTEMRC_ACCESS_DENIED              EACCES
# define RTE_SYSTEMRC_NOT_ENOUGH_MEMORY          ENOMEM
# define RTE_SYSTEMRC_OUTOFMEMORY                ENOMEM
#endif

/* --- formatting modes for devspaces */
#define RTE_FORMATTING_MODE_AUTO        0
#define RTE_FORMATTING_MODE_PARALLEL    1
#define RTE_FORMATTING_MODE_SERIAL      2


#if !defined(_WIN32) && defined(BIT64)
#define RTE_UNDEF_ID            0x7FFFFFFFFFFFFFF0
#else
#define RTE_UNDEF_ID            0x7FFFFFF0
#endif

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

typedef SAPDB_Int4   RTE_TaskId;             /* Identification of an SAPDB Kernel Task         */
#if defined(__cplusplus)
const   RTE_TaskId   RTE_UnknownTaskId = 0;  /* Reserved for returning unknown Task Identifier */
#else
/* this prevents C compiler from producing redefinition warnings.... */
#define RTE_UnknownTaskId ((const RTE_TaskId)0)
#endif
typedef SAPDB_UInt2  RTE_Reason;             /* Suspend Reason used in Suspend/Resume calls    */
typedef SAPDB_UInt4  RTE_TLSKey;             /* Thread Local Storage Key                       */

#if defined(_WIN32)
    typedef SAPDB_UInt4     RTE_DevID;
#else /* UNIX */
#if defined (BIT64)
    typedef SAPDB_Int8      RTE_DevID;      
#else
    typedef SAPDB_Int4      RTE_DevID;      
#endif
#endif


/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_Lock
  Description: Memory object needed as core of a SpinLock

  The following system depend typedefinitions are needed to support backward compatibility
  with Interlocks created i.e. on shared memory segments allocated by older clients (<7.3...)

  The RTE_Lock should not be used directly with one exception: For communication with older
  clients via shared memory segments. In such a case this type is used to describe the lock.

  New implementation should use the classes defined in RTESync_Spinlock.hpp
 */
#if defined(_WIN32)
  typedef long       RTE_Lock;
#endif

#if defined(ALPHA)
  typedef SAPDB_Long RTE_Lock[1]; /* atomic_p is a pointer to a long on ALPHA */
#endif

#if defined(AIX) || defined(SUN) || defined(LINUX)  || (defined (HPUX) && defined (HP_IA64))
  typedef SAPDB_Int4 RTE_Lock[1]; /* atomic_p is a pointer to a 32bit word on AIX */
#endif

#if defined(PA20W) || defined(PA11)
  /* Warning: If HP uses new architecture, cacheline size may change... */
# define CACHELINE_SIZE 16          /* Spinlock must be in seperate chacheline to */
                                    /* guarantie read/write consistency           */
  /* 32bit words are ok in 64bit mode too */
  typedef SAPDB_Int4 RTE_Lock[CACHELINE_SIZE*2];
#endif /* PA20W || PA11 */

#if defined(NMP)
  typedef spinlock_t RTE_Lock;
#endif

/*
  End of this ugly RTE_Lock...
 *-------------------------------------------------------------------------------------------*/

  /* C Strings are Zero byte terminate character arrays. The type definitions for C strings       */
  /* derived from a guarantuid PASCAL string length, must always use this macro. Otherwise        */
  /* on DEC i.e. we generated alignment problems, that produced core dumps due to unaligned data. */
  /* This phenomenon seems no longer reproducable...                                              */

#define SAPDB_ALIGNED_CSTRING_SIZE(StringLength_) ((StringLength_+4)&(~3))

  typedef       SAPDB_Char     RTE_Path     [ SAPDB_ALIGNED_CSTRING_SIZE( 256 ) ];
  typedef       SAPDB_Char     RTE_Filename [ SAPDB_ALIGNED_CSTRING_SIZE(  64 ) ];
  typedef       SAPDB_Char     RTE_DBName   [ SAPDB_ALIGNED_CSTRING_SIZE(  18 ) ];
  typedef       SAPDB_Char     RTE_Version  [ SAPDB_ALIGNED_CSTRING_SIZE(  40 ) ];


/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_SystemRc
  Description: System return code

  Common to operating system interface is the type of the last error.
  Examples are 'int errno' on UNIX and 'DWORD(unsiged long) GetLastError()' on NT.
 */
#if defined(_WIN32) || defined(WIN32)
  typedef unsigned long   RTE_SystemRc;
#else
  typedef int             RTE_SystemRc;
#endif

# define RTE_UNDEF_ERROR        ((RTE_SystemRc)-1)   

/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_SpinlockStatistic
  Description: SpinLoop instance statistic record

  The statistic contains the number of locks and collisions. For each collision the
  number of spin loops (if any) and the number of yield loops are counted and the
  total number as well as the maximum number is kept. All these values are written
  after the lock succeeded, so they are consistent, even if another task is already
  spinning. The currentLoops count is the only value that is 'dirty written', that
  means each time a spinning tasks tests the lock, it updates this counter to its
  individual value. This value is consistent only for the task.
 */
typedef struct RTE_SpinlockStatistic_ {
  SAPDB_Int8 locks;           /* total number of locks */
  SAPDB_Int8 collisions;      /* total number of collisions */
  SAPDB_Int4 totalSpinLoops;  /* total number of spin loops needed */
  SAPDB_Int4 totalYieldLoops; /* total number of yield loops needed */
  SAPDB_Int4 maxSpinLoops;    /* maximum number of spin loops needed */
  SAPDB_Int4 maxYieldLoops;   /* maximum number of yield loops needed */
  SAPDB_Int4 currentLoops;    /* current number of loops needed (spin loops or yield loops) */
} RTE_SpinlockStatistic;

 
#if defined(_WIN32)         /* PTS 1110642 */
typedef MEMORYSTATUSEX  RTE_NTMEMORYSTATUSEX; 
#endif

/*!--------------------------------------------------------------------------------------------
  Declaration: Version strings
  Description: Version strings
 */
#define     RTE_MAXSPEEDLEN     6     
#define     RTE_FAST    "FAST"
#define     RTE_QUICK   "QUICK"
#define     RTE_SLOW    "SLOW"

#if defined(_WIN32)
typedef HANDLE     RTE_FileHandle;
typedef SAPDB_Int8 RTE_FileOffset;
#else
typedef int        RTE_FileHandle;
typedef off_t      RTE_FileOffset;
typedef void       SECURITY_ATTRIBUTES; /* this NT system type is defined */
                                        /* here for UNIX to simplify the */ 
                                        /* interfaces */
typedef SECURITY_ATTRIBUTES *   PSECURITY_ATTRIBUTES;
#endif

/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_LoadMode
  Description: Describes how to handle symbols in DLL.
  
  RTE_LoadSymbolsGlobal means all symbols are tried to be imported into global scope.

  RTE_LoadSymbolsLocal means all symbols are DLL local only. This allows to load two DLLs
  with the same symbols defined.
 */
typedef enum { RTE_LoadSymbolsGlobal = 1, RTE_LoadSymbolsLocal = 2 } RTE_LoadMode;

/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_SharedLibHandle
  Description: Handle for dynamically loaded shared library
 */
#if defined(_WIN32)
typedef HANDLE     RTE_SharedLibHandle;
#else
typedef void *     RTE_SharedLibHandle;
#endif
#define RTE_ILLEGAL_SHARED_LIB_HANDLE ((RTE_SharedLibHandle)0)

/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_SharedLibProcAddress
  Description: Function address used to call inside a dynamically loaded shared library (extern "C" !!!)
 */
typedef void (*RTE_SharedLibProcAddress)();

/*!--------------------------------------------------------------------------------------------
  Declaration: RTE_DevSpaceType
  Description: Describes the type of a DevSpace
  
 */
typedef enum 
{
    RTE_DevSpaceSystem       = 0,
    RTE_DevSpaceMirrorSystem = 1,
    RTE_DevSpaceLog          = 2,
    RTE_DevSpaceMirrorLog    = 3,
    RTE_DevSpaceFiller       = 4,
    RTE_DevSpaceData         = 5,
    RTE_DevSpaceMirrorData   = 6,
    RTE_DevSpaceTrace        = 7 
}
RTE_DevSpaceType ;


typedef SAPDB_Int4   RTE_RegionId;             /* Identification of an SAPDB Kernel Task         */

#endif  /* RTE_TYPES_H */
