/****************************************************************************

  module      : RTESys_IO.c
  
  -------------------------------------------------------------------------
    
  responsible : RobinW/JoergM
      
  special area: platform-independent file I/O routines
  description : Contains simple routines for opening files, reading from them etc..
                The routines work on all platforms. They were extracted from veo15.cpp
                in order to make them available outside the kernel, too.
        
  last changed: 2002-05-07  12:29
  see also    : 
          
  -------------------------------------------------------------------------
            
  copyright:    Copyright by SAP AG, 2002
              


    ========== licence begin LGPL
    Copyright (C) 2002 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/

#include "gsp00.h"
#include "RunTime/System/RTESys_IO.h"
#include "geo007_1.h"
#include "hsp77.h" /* sp77sprintf() */

/*===========================================================================*
*  DEFINES                                                                  *
*===========================================================================*/

#define TEST_PRINT(fmt)
#define TEST_PRINT1(fmt,arg1)
#define TEST_PRINT3(fmt,arg1,arg2,arg3)
#define TEST_DPRINT3(fmt,arg1,arg2,arg3)

                                                
# define COPY_SIZE_RTESYS_IO ((16*1024)/sizeof(long))

/*===========================================================================*
*  MACROS                                                                   *
*===========================================================================*/

/*===========================================================================*
*  LOCAL VARIABLES                                                          *
*===========================================================================*/

/*===========================================================================*
*  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
*===========================================================================*/

#define LAST_ERROR_MESSAGE_SIZE_RTESYS_IO 256
typedef char RTESys_IOLastErrorMessage[LAST_ERROR_MESSAGE_SIZE_RTESYS_IO];
#define RTESYS_ERROR_MESSAGE_SIZE 256
typedef SAPDB_Char RTESys_ErrorMessage[RTESYS_ERROR_MESSAGE_SIZE];

/*===========================================================================*
*  PUBLIC FUNCTIONS (CODE)                                                  *
*===========================================================================*/

externC SAPDB_Char *RTESys_StrError(SAPDB_Int4 errCode, RTESys_ErrorMessage spaceToFill)
{
#if defined(_WIN32)
    LPVOID lpMsgBuf;
    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
      NULL, errCode, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL );
    strncpy(&spaceToFill[0], (char *)lpMsgBuf, sizeof(RTESys_ErrorMessage)-1);
    LocalFree(lpMsgBuf);
#else
    /* all platform seem to have it now..., */
    /* Otherwise this should return sys_errlist[errCode] if errCode is in range [0...sys_nerrlist] */
    strncpy(&spaceToFill[0], strerror(errCode), sizeof(RTESys_ErrorMessage)-1);
#endif
    spaceToFill[sizeof(RTESys_ErrorMessage)-1] = 0;
    return &spaceToFill[0];
}

/*---------------------------------------------------------------------------*/

/* PTS 1105164 */
externC void RTESys_IOCheckExist( const tsp00_Char * Filename,
                            bool  *DoesExist,
                            bool  *IsDirectory,
                            tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOCheckExist"));

#if defined(_WIN32)
  DWORD Attributes = GetFileAttributes(Filename);
  
  *DoesExist = ( Attributes != 0xFFFFFFFF );
  
  if ( *DoesExist )
  {
    *ReturnStatus = vf_ok;
    *IsDirectory = ( (Attributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY );
  }
  else
  {
    *ReturnStatus = (GetLastError() == ERROR_FILE_NOT_FOUND ? vf_ok : vf_notok);
    *IsDirectory = false;
  }
#else
  struct stat testBuf;
  
  /* stat follows symbolic links... */
  *DoesExist = ( stat( Filename, &testBuf ) == 0 );
  
  if ( *DoesExist )
  {
    *ReturnStatus = vf_ok;
    *IsDirectory = ( testBuf.st_mode == S_IFDIR );
  }
  else
  {
    *ReturnStatus = ( errno == ENOENT ? vf_ok : vf_notok );
    *IsDirectory = false;
  }
#endif
  
  return;
}

/*---------------------------------------------------------------------------*/
/* PTS 1105164 */
externC void RTESys_IOGetWorkingDirectory( tsp00_Pathc          WorkingDirectory,
                                     tsp00_VfReturn_Param *ReturnStatus )
{
    /* PTS 1114436 */
  static tsp00_Pathc currentWorkingDirectory; /* never changed after first getcwd... */
  static int         wasSet = 0;

  ROUTINE_DBG_MEO60 (_T("RTESys_IOGetWorkingDirectory"));

  if ( 1 == wasSet )
  {
      *ReturnStatus = vf_ok;
  }
  else
  {
      *ReturnStatus = ( 
#if defined(_WIN32)
        GetCurrentDirectory(sizeof(tsp00_Pathc), &currentWorkingDirectory[0]) == 0xFFFFFFFF 
#else
        getcwd(&currentWorkingDirectory[0], sizeof(tsp00_Pathc)) == NULL 
#endif
        ? vf_notok : vf_ok );
  }
  if ( vf_ok == *ReturnStatus )
  {
      wasSet = 1;
      strcpy( &WorkingDirectory[0], &currentWorkingDirectory[0]);
  }
}
/*===========================================================================*
*  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
*===========================================================================*/

/*---------------------------------------------------------------------------*/

externC void RTESys_IOOpen ( tRTESys_IOHandle     *FileHandle,
                        const tsp00_Char     *Path,
                        const tRTESys_IOOpenMode  FileMode,
                        const bool            DoCreate,
                        tRTESys_IOSecurityAttrib  Security,
                        tsp00_VfReturn_Param *ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOOpen"));

#if defined(_WIN32)
  DWORD OpenMode;
  switch( FileMode )
  {
  case RTESys_IOReadWrite:
    OpenMode = GENERIC_READ|GENERIC_WRITE;
    break;
  case RTESys_IOReadOnly:
    OpenMode = GENERIC_READ;
    break;
  case RTESys_IOWriteOnly:
    OpenMode = GENERIC_WRITE;
    break;
  default:
    OpenMode = GENERIC_READ|GENERIC_WRITE;
    break;
  }
  *FileHandle = CreateFile( Path,
    OpenMode,
    FILE_SHARE_READ,
    Security,
    ( DoCreate ? CREATE_ALWAYS : OPEN_EXISTING ),
    FILE_ATTRIBUTE_NORMAL, /* WRITE_THROUGH would only be needed to protect againt system crash */
    INVALID_HANDLE_VALUE);
  if ( *FileHandle == INVALID_HANDLE_VALUE )
#else
  int OpenMode;
  switch( FileMode )
  {
  case RTESys_IOReadWrite:
    OpenMode = O_RDWR;
    break;
  case RTESys_IOReadOnly:
    OpenMode = O_RDONLY;
    break;
  case RTESys_IOWriteOnly:
    OpenMode = O_WRONLY;
    break;
  default:
    OpenMode = O_RDWR;
    break;
  }
  *FileHandle = open( Path, OpenMode | ( DoCreate ? O_CREAT : 0 ), ( Security == 0 ? 0666 : Security ) );
  if ( *FileHandle < 0 )
#endif
  {
    *ReturnStatus = vf_notok;
  }
  else
  {
    *ReturnStatus = vf_ok;
  }
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IOClose ( const tRTESys_IOHandle     FileHandle,
                         tsp00_VfReturn_Param      *ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOClose"));

#if defined(_WIN32)
  BOOL Closed;

  Closed = CloseHandle( FileHandle );
  if ( !Closed )
#else
  if ( close( FileHandle ) == -1 )
#endif
  {
    *ReturnStatus = vf_notok;
  }
  else
  {
    *ReturnStatus = vf_ok;
  }
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IOChangeMode ( const SAPDB_Char *filePath,
                                   tRTESys_IOOpenMode newMode,
                                   tsp00_VfReturn_Param      *ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOChangeMode"));

#if defined(_WIN32)
  DWORD currentAttributes;
  DWORD newAttributes;
  currentAttributes = GetFileAttributes(filePath);
  
  if ( ((DWORD)0xFFFFFFFF) == currentAttributes )
  {
      *ReturnStatus = (GetLastError() == ERROR_FILE_NOT_FOUND ? vf_ok : vf_notok);
  }
  else
  {
      *ReturnStatus = vf_ok;
      newAttributes = currentAttributes;

      switch( newMode ) {
      case RTESys_IOReadOnly:
          if ( FILE_ATTRIBUTE_READONLY != ( FILE_ATTRIBUTE_READONLY & currentAttributes) )
          {
              newAttributes = (currentAttributes | FILE_ATTRIBUTE_READONLY);
          }
          break;
      case RTESys_IOReadWrite:
          if ( FILE_ATTRIBUTE_READONLY == ( FILE_ATTRIBUTE_READONLY & currentAttributes) )
          {
              newAttributes = (currentAttributes & ~FILE_ATTRIBUTE_READONLY);
          }
          break;
      case RTESys_IOWriteOnly:
          /* Windows does not support explicit WriteOnly mode */
          if ( FILE_ATTRIBUTE_READONLY == ( FILE_ATTRIBUTE_READONLY & currentAttributes) )
          {
              newAttributes = (currentAttributes & ~FILE_ATTRIBUTE_READONLY);
          }
          break;          
      default:
          break;
      }

      if ( newAttributes != currentAttributes )
      {
          *ReturnStatus = ( SetFileAttributes(filePath, newAttributes) ? vf_ok : vf_notok );
      }
  }
#else
  struct stat statBuf;
  mode_t      newAttributes;

  if ( 0 != stat( filePath, &statBuf ) )
  {
      *ReturnStatus = (errno == ENOENT ? vf_ok : vf_notok);
  }
  else
  {
      *ReturnStatus = vf_ok;
      newAttributes = statBuf.st_mode;

      switch( newMode ) {
      case RTESys_IOReadOnly:
          if ( 0 != ( (S_IWUSR|S_IWGRP|S_IWOTH) & statBuf.st_mode ) )
          {
              /* ReadOnly means no write access for anybody and at least read acccess for user */
              newAttributes = ((~(S_IWUSR|S_IWGRP|S_IWOTH)) & statBuf.st_mode );
          }
          newAttributes |= S_IRUSR;
          break;
      case RTESys_IOReadWrite:
          /* Read write for user */
          newAttributes |= (S_IRUSR|S_IWUSR);
          break;
      case RTESys_IOWriteOnly:
          if ( 0 != ( (S_IRUSR|S_IRGRP|S_IROTH) & statBuf.st_mode ) )
          {
              /* WriteOnly means no read access for anybody and at least write acccess for user */
              newAttributes = ((~(S_IRUSR|S_IRGRP|S_IROTH)) & statBuf.st_mode );
          }
          newAttributes |= S_IWUSR;
          break;          
      default:
          break;
      }

      if ( newAttributes != statBuf.st_mode )
      {
          *ReturnStatus = ( (0 == chmod(filePath, newAttributes)) ? vf_ok : vf_notok );
      }
  }
#endif
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IOWrite ( const tRTESys_IOHandle    FileHandle,
                         const void           *Buffer,
                         const tRTESys_IOPosition  WantedSizeInBytes,
                         tRTESys_IOPosition       *WrittenSizeInBytes,
                         tsp00_VfReturn_Param *ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOWrite"));
TEST_DPRINT3("Writing to %d 0x%lx (0x%lx bytes)\n", FileHandle, Buffer, WantedSizeInBytes )
#if defined(_WIN32)
  DWORD Wanted = (DWORD)WantedSizeInBytes;
  DWORD Written;
#else
  size_t Wanted = WantedSizeInBytes;
  ssize_t Written;
#endif

  *WrittenSizeInBytes = 0;
  do {
#if defined(_WIN32)
  
    if ( !WriteFile( FileHandle,
      ((char *)Buffer)+*WrittenSizeInBytes,
      Wanted,
      &Written,
      NULL) )
    {
      break;
    }
    
    if ( Written > 0 )
    {
      *WrittenSizeInBytes += Written;
      Wanted -= Written;
    }
#else
    Written = write( FileHandle, ((char *)Buffer)+*WrittenSizeInBytes, Wanted );

    /* PTS 1109920 */
    if ( (Written == (ssize_t)-1) && (errno != EINTR) )
    {
        break;
    }

    if ( Written > 0 )
    {
      *WrittenSizeInBytes += Written;
      Wanted -= Written;
    }
#endif
  } while ( Wanted > 0 );
  
  if ( *WrittenSizeInBytes != WantedSizeInBytes )
  {
    *ReturnStatus = vf_notok;
  }
  else
  {
    *ReturnStatus = vf_ok;
  }
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IORead ( const tRTESys_IOHandle      FileHandle,
                        void                   *Buffer,
                        const tRTESys_IOPosition    WantedSizeInBytes,
                        tRTESys_IOPosition         *ReceivedSizeInBytes,
                        tsp00_VfReturn_Param   *ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IORead"));

#if defined(_WIN32)
  DWORD BytesToRead = (DWORD)WantedSizeInBytes;
  DWORD ReceivedBytes;
  bool  ReadOk;

  ReadOk = ( ReadFile( FileHandle,
    Buffer,
    BytesToRead,
    &ReceivedBytes,
    NULL) != 0 );
  
  if ( ReadOk )
  {
      *ReceivedSizeInBytes = ReceivedBytes;
  }
  
  if ( !ReadOk )
  {
      if ( GetLastError() == ERROR_HANDLE_EOF 
        || GetLastError() == ERROR_BROKEN_PIPE )
      {
          ReadOk = true;
          *ReceivedSizeInBytes = 0;
      }
  }
 
  if ( !ReadOk )
#else
    *ReceivedSizeInBytes = read( FileHandle, Buffer, WantedSizeInBytes );
  
  if ( *ReceivedSizeInBytes < 0 )
#endif
  {
    *ReturnStatus = vf_notok;
  }
  else
  {
    if ( *ReceivedSizeInBytes == 0 )
    {
      *ReturnStatus = vf_eof;
    }
    else
    {
      *ReturnStatus = vf_ok;
    }
  }
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IOSeek ( const tRTESys_IOHandle      FileHandle,
                        const tRTESys_IOPosition    Position,
                        const tRTESys_IOSeekMode    SeekMode,
                        tRTESys_IOPosition         *NewPosition,
                        tsp00_VfReturn_Param   *ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOSeek"));

#if defined(_WIN32)
  DWORD whence;
  LONG lowPosition = (LONG)(Position & 0xFFFFFFFFU);
  LONG highPosition = (LONG)(Position >> 32);
  DWORD lowNewPosition;

  switch ( SeekMode )
  {
  case RTESys_IOSeekSet:
    whence = FILE_BEGIN;
    break;
  case RTESys_IOSeekEnd:
    whence = FILE_END;
    break;
  case RTESys_IOSeekCurrent:
    whence = FILE_CURRENT;
    break;
  default:
    whence = FILE_BEGIN;
    break;
  }
  lowNewPosition = SetFilePointer( FileHandle,
    lowPosition,
    &highPosition,
    whence);
  if ( ((DWORD)-1) != lowNewPosition )
  {
     *NewPosition = lowNewPosition + ( ((tRTESys_IOPosition)highPosition) << 32 );
  }
  else
  {
     *NewPosition = (tRTESys_IOPosition)-1;
  }
#else
  int whence;
  switch ( SeekMode )
  {
  case RTESys_IOSeekSet:
    whence = SEEK_SET;
    break;
  case RTESys_IOSeekEnd:
    whence = SEEK_END;
    break;
  case RTESys_IOSeekCurrent:
    whence = SEEK_CUR;
    break;
  default:
    whence = SEEK_SET;
    break;
  }
  *NewPosition = lseek( FileHandle,
    Position,
    whence);
#endif
  if ( ((tRTESys_IOPosition)-1) == *NewPosition )
  {
    *ReturnStatus = vf_noseek;
  }
  else
  {
    *ReturnStatus = vf_ok;
  }
}
/*---------------------------------------------------------------------------*/
externC void RTESys_IOCopyFile ( const tsp00_Char     *Original,
                            const tsp00_Char     *Destination,
                            tsp00_VfReturn_Param *ReturnStatus )
{
#if !defined(_WIN32)
  tRTESys_IOPosition BytesWritten;
  tRTESys_IOPosition BytesRead;
  long Buffer[COPY_SIZE_RTESYS_IO];
#endif

  ROUTINE_DBG_MEO60 (_T("RTESys_IOCopyFile"));

#if defined(_WIN32)
  *ReturnStatus = ( CopyFile( Original, Destination, false ) ? vf_ok : vf_notok );
#else
  tRTESys_IOHandle OriginalFd;
  tRTESys_IOHandle DestinationFd;
  tsp00_VfReturn_Param IgnoredError; /* close errors of temporary files are ignored here */
  
  RTESys_IOOpen( &OriginalFd, Original, RTESys_IOReadOnly, false, DEFAULT_SECURITY_RTESYS_IO, ReturnStatus );
  if ( vf_ok == *ReturnStatus )
  {
    RTESys_IOOpen( &DestinationFd, Destination, RTESys_IOWriteOnly, true, DEFAULT_SECURITY_RTESYS_IO, ReturnStatus );
    if ( vf_ok != *ReturnStatus )
    {
      RTESys_IOOpen( &DestinationFd, Destination, RTESys_IOWriteOnly, false, DEFAULT_SECURITY_RTESYS_IO, ReturnStatus );
      if ( vf_ok != *ReturnStatus )
      {
        RTESys_IOShowLastOsError("Open(Destination)", Destination );
        RTESys_IOClose( OriginalFd, &IgnoredError );
        return;
      }
    }
    
    
    do {
      RTESys_IORead( OriginalFd, Buffer, COPY_SIZE_RTESYS_IO, &BytesRead, ReturnStatus );
      if ( vf_ok == *ReturnStatus )
      {
        RTESys_IOWrite( DestinationFd, Buffer, BytesRead, &BytesWritten, ReturnStatus );
        if ( vf_ok != *ReturnStatus )
        {
          RTESys_IOShowLastOsError("Write(Destination)", Destination );
        }
      }
      else if ( vf_eof != *ReturnStatus ) /* 1110223 */
      {
        RTESys_IOShowLastOsError("Read(Original)", Original );
      }
    } while ( vf_ok == *ReturnStatus && 0 != BytesRead );
    RTESys_IOClose( OriginalFd, &IgnoredError );
    RTESys_IOClose( DestinationFd, &IgnoredError );

    if ( *ReturnStatus == vf_eof && BytesRead == 0 )
    {
      *ReturnStatus = vf_ok;
    }

    if ( vf_ok == *ReturnStatus )
    {
        /* Ignore any errors if mode setting fails... */
        struct stat orgBuf;
        if ( 0 == stat(Original, &orgBuf) )
        {
            if ( 0 != chmod(Destination, orgBuf.st_mode) )
            {
                RTESys_IOShowLastOsError("Ignoring bad chmod(Destination)", Destination );
            }
        }
        else
        {
            RTESys_IOShowLastOsError("Ignoring bad stat(Original)", Original );
        }
    }
  }
  else
  {
    RTESys_IOShowLastOsError("Open(Original)", Original );
  }
#endif
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IOUnlink ( const tsp00_Char                 *FileName,
                           tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus )
{
  ROUTINE_DBG_MEO60 (_T("RTESys_IOUnlink"));

#if defined(_WIN32)
    *ReturnStatus = ( DeleteFile(FileName) ? vf_ok : vf_notok );
#else
    *ReturnStatus = ( unlink(FileName) == 0 ? vf_ok : vf_notok );
#endif /* WIN32 */
    if ( vf_ok != *ReturnStatus )
    {
      RTESys_IOShowLastOsError("Unlink", FileName );
    }
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IOShowLastOsError( const tsp00_Char *    Routine,
                                       const tsp00_Char *    Filename )
{
  RTESys_IOLastErrorMessage LastErrorMessage;
  RTESys_ErrorMessage       SysErrorMessage;

#ifdef _WIN32
  sp77sprintf( &LastErrorMessage[0], sizeof(RTESys_IOLastErrorMessage)-1, "[%lu] %s", GetLastError(), RTESys_StrError(GetLastError(), SysErrorMessage) );
#else
  sp77sprintf( &LastErrorMessage[0], sizeof(RTESys_IOLastErrorMessage)-1, "[%d] %s", errno, RTESys_StrError(errno, SysErrorMessage) );
#endif
  MSGD(( ERR_TRACEIO_OS_ERROR, 
         Routine, 
         Filename == NULL ? "" : Filename, 
         &LastErrorMessage[0] ));
}

/*---------------------------------------------------------------------------*/

externC bool RTESys_IOCheckFilename(const tsp00_Char *Filename)
{
  bool ReturnStatus = true;

  if ( (Filename == NULL) || (*Filename == '\0') || strlen(Filename) > sizeof(tsp00_Path) )
  {
    ReturnStatus = false;

    if ( Filename == NULL )
    {
      MSGD((ERR_TRACEIO_CHECK_FILENAME, "<NULL>", "invalid file name"));
    } else if ( *Filename == '\0' ) {
      MSGD((ERR_TRACEIO_CHECK_FILENAME, "<Empty>", "zero length file name"));
    } else {
      MSGD((ERR_TRACEIO_CHECK_FILENAME, Filename, "file name too long"));
    }
  }

  return ReturnStatus;
}

/*---------------------------------------------------------------------------*/

externC void RTESys_IORename(const tsp00_Char *OldFileName,
                             const tsp00_Char *NewFileName,
                             tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus)
{
    ROUTINE_DBG_MEO60 (_T("RTESys_IORename"));

#if defined(_WIN32)
    *ReturnStatus = ( MoveFile(OldFileName, NewFileName) ? vf_ok : vf_notok );
#else
    *ReturnStatus = ( rename(OldFileName, NewFileName) == 0 ? vf_ok : vf_notok );
#endif /* WIN32 */
    if ( vf_ok != *ReturnStatus )
    {
      RTESys_IOShowLastOsError("Rename", OldFileName );
    }
}
/*===========================================================================*
*  END OF CODE                                                              *
*===========================================================================*/
