/*!***************************************************************************

  module      : RTESys_IO.h

  -------------------------------------------------------------------------

  responsible : RobinW/JoergM

  special area: platform-independent file I/O routines
  description : Contains simple routines for opening files, reading from them etc..
                The routines work on all platforms. They were extracted from veo15.cpp
                in order to make them available outside the kernel, too.
        
  last changed: 2002-05-07  19:55
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002



    ========== licence begin LGPL
    Copyright (C) 2002 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef RTESYS_IO_H
#define RTESYS_IO_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "geo00_0.h"


#if defined(_WIN32)

    typedef HANDLE tRTESys_IOHandle;
# define INVALID_HANDLE_RTESYS_IO   ((tRTESys_IOHandle)INVALID_HANDLE_VALUE)
    typedef teo00_Int8 tRTESys_IOPosition;
    typedef LPSECURITY_ATTRIBUTES tRTESys_IOSecurityAttrib;
# define DEFAULT_SECURITY_RTESYS_IO NULL

#else

   typedef int   tRTESys_IOHandle;
# define INVALID_HANDLE_RTESYS_IO   ((tRTESys_IOHandle)-1)
    typedef off_t tRTESys_IOPosition;
    typedef int tRTESys_IOSecurityAttrib;
# define DEFAULT_SECURITY_RTESYS_IO ((tRTESys_IOSecurityAttrib)0)

#endif

typedef enum { RTESys_IOReadWrite,
               RTESys_IOReadOnly, 
               RTESys_IOWriteOnly } tRTESys_IOOpenMode;

typedef enum { RTESys_IOSeekSet,
               RTESys_IOSeekEnd, 
               RTESys_IOSeekCurrent } tRTESys_IOSeekMode;

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/


/*!
  Chapter: Simple Kernel File Interface
 */

externC void RTESys_IOCheckExist( const tsp00_Char * Filename,
                            bool  *DoesExist,
                            bool  *IsDirectory,
                            tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus );
/*!
  Function:     RTESys_IOOpen
  description:  Open a file
        
  arguments:
    FileHandle [out] The file handle returned by open
    Path [in] The file name and path information
    FileMode [in] The mode how to open the file (readonly,readwrite,writeonly)
    DoCreate [in] Flag showing to return an error file exists
    Security [inout] The scerurity descrptior on NT ( on UNIX the access mask )
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IOOpen ( tRTESys_IOHandle           VAR_VALUE_REF FileHandle,
                             const tsp00_Char           *Path,
                             const tRTESys_IOOpenMode   FileMode,
                             const bool                 DoCreate,
                             tRTESys_IOSecurityAttrib   Security,
                             tsp00_VfReturn_Param       VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOClose
  description:  Close a file
    
  arguments:
    FileHandle [in] The file handle returned by a previous call to RTESys_IOOpen
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IOClose ( const tRTESys_IOHandle FileHandle,
                              tsp00_VfReturn_Param   VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOWrite
  description:  Write a fixed size buffer to a file
    
  arguments:
    FileHandle [in] The file handle returned by a previous call to RTESys_IOOpen
    Buffer [in] The buffer containing the information to write
    WantedSizeInBytes [in] Number of bytes to write
    WrittenSizeInBytes [out] Number of Bytes written
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IOWrite ( const tRTESys_IOHandle   FileHandle,
                              const void               *Buffer,
                              const tRTESys_IOPosition WantedSizeInBytes,
                              tRTESys_IOPosition       VAR_VALUE_REF WrittenSizeInBytes,
                              tsp00_VfReturn_Param     VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IORead
  description:  Read a fixed amount of data into a buffer from a file
    
  arguments:
    FileHandle [in] The file handle returned by a previous call to RTESys_IOOpen
    Buffer [out] The buffer containing the information read if sucessfull
    WantedSizeInBytes [in] Maximum number of bytes to read
    ReceivedSizeInBytes [out] Number of Bytes read
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IORead ( const tRTESys_IOHandle   FileHandle,
                             void                     *Buffer,
                             const tRTESys_IOPosition WantedSizeInBytes,
                             tRTESys_IOPosition       VAR_VALUE_REF ReceivedSizeInBytes,
                             tsp00_VfReturn_Param     VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOSeek
  description:  Seek to the given position in file
    
  arguments:
    FileHandle [in] The file handle returned by a previous call to RTESys_IOOpen
    Position [in] The position where to seek
    SeekMode [in] How to seek (Absolut,From End of File, Relativ to current position)
    NewPosition [out] New absolut position after seek
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_noseek if failed )
    
  return value: none
 */
externC void RTESys_IOSeek ( const tRTESys_IOHandle   FileHandle,
                             const tRTESys_IOPosition Position,
                             const tRTESys_IOSeekMode SeekMode,
                             tRTESys_IOPosition       VAR_VALUE_REF NewPosition,
                             tsp00_VfReturn_Param     VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOCopyFile
  description:  Make a copy of a given file

    If copy file does not exist, it is created. If copy file already exists, it is overwritten!
     
  arguments:
    Original [in] The path to the original file
    Destination [in] The path to the destination
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok or vf_noseek if failed )
    
  return value: none
 */
externC void RTESys_IOCopyFile ( const tsp00_Char     *Original,
                                 const tsp00_Char     *Destination,
                                 tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOUnlink
  description:  Unlink a file 
  
    If file is still open, vf_notok is returned.

    If file does exist and is successfull removed, vf_ok is returned.
  
  arguments:
    FileName [in] The path to the file to remove
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IOUnlink ( const tsp00_Char     *FileName,
                               tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOGetWorkingDirectory
  description:  Get path to directory where message files are stored
    
    This returns the path to the directory that is used for all relativ specified message files.
    It should be identical to the WorkingDirectory of the database.
    
  arguments:
    WorkingPath [out] The path to the root of all relativ specified message files
    ReturnStatus [out] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IOGetWorkingDirectory( tsp00_Pathc          VAR_ARRAY_REF WorkingDirectory,
                                           tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus );

/*!
  Function:     RTESys_IOShowLastOsError
  description:  
    
    This returns the path to the directory that is used for all relativ specified message files.
    It should be identical to the WorkingDirectory of the database.
    
  arguments:
    Routine  [in] The path to the root of all relativ specified message files
    Filename [in] Status enumeration ( vf_ok if ok, vf_notok if failed )
    
  return value: none
 */
externC void RTESys_IOShowLastOsError( const tsp00_Char *    Routine,
                                       const tsp00_Char *    Filename );

/*!
  Function:     RTESys_IOCheckFilename
  description:  Check if Filename is a valid string
    
    This function checks if the parameter Filename is a valid string, that means if it is not NULL and does not
    point to an empty string.
    
  arguments:
    Filename [in] The path to the root of all relativ specified message files
    
  return value: true if Filename is a valid string, false if not
 */
externC bool RTESys_IOCheckFilename(const tsp00_Char *Filename);

/*!
  Function:     RTESys_IORename
  description:  Rename given file
    
    This function renames the given file
    
  arguments:
    OldFilename [in] The path to the root of all relativ specified message files
    NewFilename [in] The new filename
    ReturnStatus [out] The result of the rename action (vf_ok if renamed, vf_notok else)
 */
externC void RTESys_IORename(const tsp00_Char *OldFilename,
                             const tsp00_Char *NewFilename,
                             tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus);

/*!
  Function:     RTESys_IOChangeMode
  description:  Change access mode of given file
    
    This function changes access mode of the given file
    
  arguments:
    Filename [in] The path to the file to change
    newMode [in] The requested mode (RTESys_IOReadOnly,RTESys_IOReadWrite or RTESys_IOWriteOnly )
    ReturnStatus [out] The result of the rename action (vf_ok if now in that mode, vf_notok else)
 */
externC void RTESys_IOChangeMode ( const SAPDB_Char     *filePath,
                                   tRTESys_IOOpenMode  newMode,
                                   tsp00_VfReturn_Param VAR_VALUE_REF ReturnStatus );
/*!
  EndChapter: Simple Kernel File Interface
 */
#endif  /* RTESYS_IO_H */
