/*!**************************************************************************

  module      : RTESys_NTAWE.hpp

  -------------------------------------------------------------------------

  responsible : RaymondR

  special area: Address Windowing Extensions (AWE)
  description : AWE provides 32-bit applications to address more than 32-bit regions
                of physical memory.

                AWE is supported on all platforms supported by Windows 2000, including
                the 64-bit version of Windows 2000. 

                Windows 2000 (x86) provides applications with a flat 32-bit virtual 
                address space, which describes 4 GB of virtual memory. This address space 
                is usually splitted into two parts. The first part (2GB of address space) 
                is directly accessible to the application, the other 2GB memory is reserved 
                for the operating system.
                On Windows 2000 Advanced Server x86 systems an 'boot.ini' option (/3GB) 
                is provided, so that the operating system reserves only 1 GB of the address 
                space, and the application can access up to 3 GB.

                AWE solves the problem accessing more than 3GB of memory by allowing 
                applications to directly address huge amounts of physical memory while 
                continuing to use 32-bit pointers (where sufficient physical memory is 
                present). 

  last changed: 2001-11-21  12:39
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef RTESYS_NTAWE_HPP
#define RTESYS_NTAWE_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#define _WIN32_WINNT  0x0500
#include <windows.h>
#include "SAPDBCommon/SAPDB_Types.h"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_MessageList.hpp"
#include "RunTime/System/RTESys_NTLoadDLLFunc.h"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IAllocatorInfo.hpp"
#include "RunTime/MemoryManagement/RTEMem_AllocatorRegister.hpp"
#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class RTESys_NTAWEPhysMem;
class RTESys_NTAWEMapAreaHashList;

/*!---------------------------------------------------------------------------
  Class:        RTESys_NTAWE
  Description:  Address Windowing Extensions

                This class allows the application to use up to 64 GB of
                physical non-paged memory in a 32-bit virtual address space on 32-bit 
                platforms, with windowed views to this physical memory from within 
                the applications virtual address space.
*/

class RTESys_NTAWE : public SAPDBMem_IAllocatorInfo 
{
public:
         RTESys_NTAWE();
virtual ~RTESys_NTAWE();

public:

  /*!---------------------------------------------------------------------------
   Function:       Initialize

   Description:    Initialize the RTESys_NTAWE instance. 
                   This method always returns a 'MessageList'. Its filled with an success 
                   or an error message.

   Arguments:      DBPageSize               [in] - Database page size in bytes
                   NumberOfMapAreaPages     [in] - Number of mapping pages to reserve, which
                                                    are used to map physical pages into them.
                   NumberOfPhysPages    [in/out] - Number of physical pages to allocate.
                   ReservedPhysMemory       [in] - The size of the reserved physical memory, 
                                                   which is used for applications virtual 
                                                   address space. Zero means, the size of
                                                   the complete virtual address space 
                                                   ( excluding the mapping area) is reserved for
                                                   for application space. If the /3GB option is
                                                   set the application reserves up to 3 GB of 
                                                   physical memory which is accessed via the
                                                   virtual address space

    Return value:  none
  */
  SAPDB_UInt4 Initialize            ( SAPDB_UInt4 const     DBPageSize,
                                      SAPDB_UInt4 const     NumberOfMapAreaPages,
                                      SAPDB_UInt4           &NumberOfPhysPages,
                                      SAPDBErr_MessageList  &MessageList,
                                      SAPDB_UInt8           ReservedPhysMemory = 0);  // Use this parameter for tests on small systems only


  /*!
   Function:       AWEAvailable
   Description:    Function returns true if AWE memory is available

   Arguments:      none

   Return value:   True if AWE memory is available otherwise false.
  */
  bool        AWEAvailable          () const  { return m_Initalized; };

  /*! 
   Function:        AllocatMappingPage
   Description:     AllocatMappingPage one mapping page and mappes a physical page into it.
                    To map a new physical page into the mapping page call 'Map'.

                    The memory is suitably aligned for all types.

                    The memory is not initialised.  

   Arguments:       ppMappingPageAddr [out] - Position where mapping page is allocated.
                    pPhysPageNo [out]       - A reserved physical page number;
                                              If NULL no physical page is mapped.
                    MessageList [out]       - Error and/or Info messages.
    Return value:   System return code
  */
  SAPDB_UInt4 AllocatMappingPage    ( void **               ppMappingPageAddr,
                                      SAPDB_UInt4*          pPhysPageNo,
                                      SAPDBErr_MessageList  &MessageList );

  /*!
    Function:       FreeMappingPage
    Description:    Frees the memory. 

                    The pointer must have been obtained by a previous call to
                    'AllocatMappingPage'.

    Arguments:      pMappingPageAddr [in] - Pointer to memory to deallocate
                    MessageList [out] - Error and/or Info messages.
    Return value:   System return code
  */
  SAPDB_UInt4 FreeMappingPage       ( void **               ppMappingPageAddr,
                                      SAPDBErr_MessageList  &MessageList );


  /*!
    Function:       Map
    Description:    Map a physical page into a specified mapping page.
                   
                    The pointer must have been obtained by a previous call to
                    'allocate'. The physical page number must have been 
                    obtained by a call to 'allocate' or 'ReservePhysPage'

    Arguments:      pMappingPageAddr [in] - Pointer to mapping page.
                    PhysPageNo [in]       - Physical page number.
                    MessageList [out]     - Error and/or Info messages.
    Return value:   System return code
  */
  SAPDB_UInt4 Map                   ( void * const          pMappingPageAddr,
                                      SAPDB_UInt4 const     PhysPageNo,
                                      SAPDBErr_MessageList  &MessageList );

  /*!
    Function:       UnMap
    Description:    UnMap a physical page from a specified mapping page.
                   
                    The pointer must have been obtained by a previous call to
                    'allocate' and physical page has to be mapped.

    Arguments:      pMappingPageAddr [in] - Pointer to mapping page.
                    MessageList [out]     - Error and/or Info messages.
    Return value:   System return code
  */
  SAPDB_UInt4 UnMap                 ( void * const           pMappingPageAddr,
                                      SAPDBErr_MessageList   &MessageList);

  /*!
    Function:       ReservePhysPage
    Description:    Reserve a physical page.
                   
    Arguments:      PhysPageNo [out]      - Reserved physical page. The page can be accessed
                                            by mapping it into a mapping page. 
                    MessageList [out]     - Error and/or Info messages.

    Return value:   System return code
  */
  SAPDB_UInt4 ReservePhysPage       ( SAPDB_UInt4           &PhysPageNo,
                                      SAPDBErr_MessageList  &MessageList );
  /*!
    Function:       ReleasePhysPage
    Description:    Release a physical page.
                   
    Arguments:      PhysPageNo [in]      - Physical page to release.
                    MessageList [out]     - Error and/or Info messages.
    Return value:   System return code
  */
  SAPDB_UInt4 ReleasePhysPage       ( SAPDB_UInt4 const     PhysPageNo,
                                      SAPDBErr_MessageList  &MessageList );

  /*!
    Function:       NumOfPhysicalPages
    Description:    Maximum number of physical pages.

    Arguments:      none
    Return value:   Maximum number of physical pages
  */
  SAPDB_UInt4 NumOfPhysicalPages    ()  const;

  /*!
    Function:       NumOfFreePhysicalPages
    Description:    Available number of physical pages.

    Arguments:      none
    Return value:   Available number of physical pages
  */
  SAPDB_UInt4 NumOfFreePhysicalPages()  const;

    /*!
    Function:       NumOfMapAreaPages
    Description:    Maximum number of mapping pages.

    Arguments:      none
    Return value:   Maximum number of mapping pages
  */
  SAPDB_UInt4 NumOfMapAreaPages     ()  const { return m_NumberOfMapAreaPages; };


  /*!
    Function:       NumOfFreeMapAreaPages
    Description:    Available number of mapping pages.

    Arguments:      none
    Return value:   Available number of mapping pages
  */
  SAPDB_UInt4 NumOfFreeMapAreaPages ()  const { RTESync_LockedScope LockedScope( (RTESync_Spinlock&)m_Lock );
                                                return m_NumberOfFreeMapAreaPages; };

  /*!
    Function:       DBPageSize
    Description:    Size of a database page.

    Arguments:      none
    Return value:   Size of a database page.
  */
  SAPDB_UInt4 DBPageSize            ()  const { return m_DBPageSize; };

  /*!
    Function:       NumOfMaps
    Description:    Number of Map calls.

    Arguments:      none
    Return value:   Number of Map calls.
  */
  SAPDB_UInt4 NumOfMaps             ()  const;

  /*!
    Function:       NumOfUnMaps
    Description:    Number of UnMap calls.

    Arguments:      none
    Return value:   Number of UnMap calls.
  */
  SAPDB_UInt4 NumOfUnMaps           ()  const;

  /*!
    Function:       GetFirstMapPageInfo / GetNextMapPageInfo
    Description:    Get info of a physical page.

    Arguments:      Handle [out]          - (GetFirstMapPageInfo)use for iteration
                           [in/out]       - (GetNextMapPageInfo) use for iteration
                    pMappingPageAddr[out] - Address of the map page
                    PhysPageNo[out]       - physical page assign to the map page
                    Reserved[out]         - If page has been reserved
                    MapCount[out]         - Number of map in calls
    Return value:   false - if no info available or wrong page number
                    true  - info available
  */
  bool        GetFirstMapPageInfo   ( SAPDB_UInt4            &Idx,
                                      void const *           &pMappingPageAddr,
                                      SAPDB_UInt4            &PhysPageNo,
                                      bool                   &Reserved,
                                      SAPDB_UInt8            &MapCount )  const;

  bool        GetNextMapPageInfo   ( SAPDB_UInt4            &Idx,
                                     void const *           &pMappingPageAddr,
                                     SAPDB_UInt4            &PhysPageNo,
                                     bool                   &Reserved,
                                     SAPDB_UInt8            &MapCount )  const;


  /*!
    Function:       GetPhysPageInfo
    Description:    Get info of a physical page.

    Arguments:      PhysPageNo [in]      - Physical page to release.
                    Reserved[out]        - If page has been reserved
                    MapCount[out]        - Number of map in calls
    Return value:   false - if no info available or wrong page number
                    true  - info available
  */
  bool        GetPhysPageInfo       ( SAPDB_UInt4  const     PhysPageNo,
                                      bool                   &Reserved,
                                      SAPDB_UInt8            &MapCount )  const;

private:
 /*-----------------------------------------------------------------------
  Function:     GetCallStatistics
  Description:  Outputs statistics about the number of Allocate and Deallocate calls.
  -----------------------------------------------------------------------*/
  virtual void  GetCallStatistics          ( SAPDB_ULong            &CountAlloc,
                                             SAPDB_ULong            &CountDealloc ) const;
 /*-----------------------------------------------------------------------
  Function:     GetBaseAllocatorCallStatistics
  Description:  Outputs statistics about the number of Allocate and Deallocate calls of
                the allocator to the underlying base allocator.
  -----------------------------------------------------------------------*/
  virtual void  GetBaseAllocatorCallStatistics( SAPDB_ULong         &CountAlloc,
                                                SAPDB_ULong         &CountDealloc ) const
                                              { CountAlloc   = 0; CountDealloc = 0; };

 /*-----------------------------------------------------------------------
  Function:     CalcStatistics
  Description:  Outputs statistics about the memory usage of the allocator.
  -----------------------------------------------------------------------*/
  virtual void  CalcStatistics                ( SAPDB_ULong         &BytesUsed,
                                                SAPDB_ULong         &BytesControlled );

 /*-----------------------------------------------------------------------
  Function:     GetIdentifier
  Description:  Class identifier.
  -----------------------------------------------------------------------*/
  virtual const SAPDB_UTF8 * GetIdentifier    () const  
                               { return (const SAPDB_UTF8 *)"RTESys_NTAWE"; }

  /*!----------------------------------------------------------------------
   Function: GetErrorCount
   Description:  Returns the number of errors found and corrected by the allocator
  -----------------------------------------------------------------------*/
 
  virtual int GetErrorCount() const;

private:
  bool                          m_Initalized;

  RTESync_Spinlock              m_Lock;                      // used for locked scopes

  RTEMem_AllocatorInfo          m_AllocatorInfo;

  SAPDB_UInt4                   m_NumberOfFreeMapAreaPages;  // DB page size
  SAPDB_UInt4                   m_NumberOfMapAreaPages;      // DB page size
  SAPDB_UInt4                   m_DBPageSize;

  SAPDB_ULong                   m_CountAlloc;
  SAPDB_ULong                   m_CountDealloc;

  RTESys_NTAWEPhysMem*          m_pAWEPhysMem;
  RTESys_NTAWEMapAreaHashList*  m_pMapAreaHashList;
};
/*! EndClass: RTESys_NTAWE */


#endif  /* RTESYS_NTAWE_HPP */
