/*!**************************************************************************

  module      : RTESys_NTHighResTimer.hpp

  -------------------------------------------------------------------------

  responsible : RaymondR

  special area: High Resolution Timer
  description : A counter is a general term used in programming to refer to an incrementing 
                variable. Some systems include a high-resolution performance counter that 
                provides high-resolution elapsed times. 

                If you need a timer with higher precision, use the high-resolution timer.
                For more information, see the class RTESys_NTHighResTimer below

  last changed: 2002-03-21  16:40
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2002

  ========== licence begin  GPL
  Copyright (C) 2002 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/


#ifndef RTESYS_NTHIGHRESTIMER_HPP
#define RTESYS_NTHIGHRESTIMER_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#define _WIN32_WINNT  0x0500
#include <windows.h>
#include "SAPDBCommon/SAPDB_Types.h"
#if defined(__cplusplus)
# include "SAPDBCommon/SAPDB_Singleton.hpp"
#endif  /* __cplusplus */

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

#if defined(__cplusplus)

/*!---------------------------------------------------------------------------
  Class:        RTESys_NTHighResTimer
  Description:  Windows High Resolution Timer

                This class implements a high-resolution timer. 
                
                Use the 'Initialize' methode first to initialize the instance and
                to choose one of the high resolution counter types.

                The 'Frequency' methode can be used to express the frequency, 
                in counts per second. The value of the count is timer type or 
                processor dependent.  On some processors, for example, the 
                count might be the cycle rate of the processor clock. 

                The 'Counter' methode retrieves the current value of the 
                high-resolution counter. By calling this methode at the 
                beginning and end of a section of code, an application essentially uses
                the counter as a high-resolution timer. For example, suppose that 
                'Frequency' indicates that the frequency of the high-resolution timer 
                is 50,000 counts per second. If the application calls 'Counter' 
                immediately before and immediately after the section of code to be timed, 
                the counter values might be 1500 counts and 3500 counts, respectively.
                These values would indicate that .04 seconds (2000 counts) elapsed while 
                the code executed. 
                An easier way to compute counter time differences is to call one of the 
                'TimeDiff' methods like 'TimeDiffMicroSec', 'TimeDiffSecAndMilliSec' etcetera.

.
*/
class RTESys_NTHighResTimer
{
public:
  typedef              enum { CPUTimer, QueryPerfCntr, TickCount } TimerType;

public:
  /*!
   Function:       Instance
   Description:    Reference to single instance of RTESys_NTHighResTimer

   Arguments:      none
   Return value:   Reference to RTESys_NTHighResTimer instance
  */
  static RTESys_NTHighResTimer & Instance();

  /*!
   Function:       Initialize
   Description:    Use the 'Initialize' methode first to initialize the instance and
                   to choose one of the high resolution counter types.

   Arguments:      Timer [in] - Type of high resolution timer
   Return value:   'False' if something went wrong.
  */
  static bool          Initialize ( TimerType  &Timer );

  /*!
   Function:       StartCount
   Description:    Returns the 'RTESys_NTHighResTimer' instance startup counter value.

   Arguments:      none
   Return value:   start counter value.
  */
  SAPDB_UInt8          StartCount() { return m_StartCount; }

  /*!
   Function:       Counter
   Description:    The 'Counter' methode retrieves the current value of the 
                   high-resolution counter.

   Arguments:      none
   Return value:   Current value of the high-resolution counter.
  */
  virtual SAPDB_UInt8  Counter()   = 0;

  /*!
   Function:       Frequency
   Description:    The 'Frequency' methode can be used to express the frequency, 
                   in counts per second. 

   Arguments:      none
   Return value:   Frequency value of the high-resolution counter.
  */
  SAPDB_UInt8          Frequency()  { return m_Frequency; }

  /*!
   Function:       TimeDiffNanoSec, TimeDiffMicroSec, TimeDiffMilliSec, TimeDiffSec
   Description:    If the application calls 'Counter' immediately before and immediately
                   after the section of code to be timed, the 'TimeDiff' methods can be
                   used to calculate the elapsed time while the code executed.  

   Arguments:      StartCounter     [in] - Start counter value
                   StopCounter      [in] - Stop counter value
                   CallTimeAdjusted [in] - Indicates if the call time of the 'Counter' 
                                           methodes should be subtracted.
   Return value:   Time difference.

   Prototypes: 4
  */
  SAPDB_UInt8          TimeDiffNanoSec ( SAPDB_UInt8 StartCounter,
                                         SAPDB_UInt8 StopCounter,
                                         bool        CallTimeAdjusted = true )
                                    { SAPDB_UInt8 CounterDiff = StopCounter - StartCounter - (CallTimeAdjusted ? m_CallTime : 0);
                                      return ( CounterDiff * 1000 * 1000 * 1000 ) / m_Frequency; }

  SAPDB_UInt8          TimeDiffMicroSec( SAPDB_UInt8 StartCounter,
                                         SAPDB_UInt8 StopCounter,
                                         bool        CallTimeAdjusted = true )
                                    { SAPDB_UInt8 CounterDiff = StopCounter - StartCounter - (CallTimeAdjusted ? m_CallTime : 0);
                                      return ( CounterDiff * 1000 * 1000 ) / m_Frequency; }
    
  SAPDB_UInt8          TimeDiffMilliSec( SAPDB_UInt8 StartCounter,
                                         SAPDB_UInt8 StopCounter,
                                         bool        CallTimeAdjusted = true )
                                    { SAPDB_UInt8 CounterDiff = StopCounter - StartCounter - (CallTimeAdjusted ? m_CallTime : 0);
                                      return (CounterDiff * 1000 ) / m_Frequency; }

  SAPDB_UInt8          TimeDiffSec     ( SAPDB_UInt8 StartCounter,
                                         SAPDB_UInt8 StopCounter,
                                         bool        CallTimeAdjusted = true )
                                    { SAPDB_UInt8 CounterDiff = StopCounter - StartCounter - (CallTimeAdjusted ? m_CallTime : 0);
                                      return CounterDiff / m_Frequency; }

  /*!
   Function:       TimeDiffSecAndMilliSec, TimeDiffSecAndMicroSec
   Description:    If the application calls 'Counter' immediately before and immediately
                   after the section of code to be timed, the 'TimeDiff' methods can be
                   used to calculate the elapsed time while the code executed.  

   Arguments:      StartCounter       [in]  - Start counter value
                   StopCounter        [in]  - Stop counter value
                   Sec                [out] - Time difference (Seconds)
                   MilliSec/MicroSec  [out] - Time difference (Milliseconds/Microseconds)
                   CallTimeAdjusted   [in]  - Indicates if the call time of the 'Counter' 
                                              methodes should be subtracted.
   Return value:   none

   Prototypes: 2
  */
  void                 TimeDiffSecAndMilliSec ( SAPDB_UInt8 StartCounter,
                                                SAPDB_UInt8 StopCounter,
                                                SAPDB_UInt8 &Sec,
                                                SAPDB_UInt8 &MilliSec,
                                                bool        CallTimeAdjusted = true )
                                    { SAPDB_UInt8 CounterDiff = StopCounter - StartCounter - (CallTimeAdjusted ? m_CallTime : 0);
                                      Sec      = CounterDiff / m_Frequency;
                                      MilliSec = ((CounterDiff * 1000 ) / m_Frequency ) % 1000; }

  void                 TimeDiffSecAndMicroSec ( SAPDB_UInt8 StartCounter,
                                                SAPDB_UInt8 StopCounter,
                                                SAPDB_UInt8 &Sec,
                                                SAPDB_UInt8 &MicroSec,
                                                bool        CallTimeAdjusted = true )
                                    { SAPDB_UInt8 CounterDiff = StopCounter - StartCounter - (CallTimeAdjusted ? m_CallTime : 0);
                                      Sec      = CounterDiff / m_Frequency;
                                      MicroSec = ((CounterDiff * 1000 * 1000 ) / m_Frequency ) % (1000 * 1000); }



  /*!
   Function:       CounterToSecAndMilliSec, CounterToSecAndMicroSec
   Description:    Converts a counter value into seconds and milli- or microseconds.

   Arguments:      Counter            [in]  - Counter value
                   Sec                [out] - Time difference (Seconds)
                   MilliSec/MicroSec  [out] - Time difference (Milliseconds/Microseconds)
   Return value:   none

   Prototypes: 2
  */
  void                 CounterToSecAndMilliSec( SAPDB_UInt8 Counter,
                                                SAPDB_UInt8 &Sec,
                                                SAPDB_UInt8 &MilliSec )
                                    { Sec      = Counter / m_Frequency;
                                      MilliSec = ((Counter * 1000 ) / m_Frequency ) % 1000; }

  void                 CounterToSecAndMicroSec( SAPDB_UInt8 Counter,
                                                SAPDB_UInt8 &Sec,
                                                SAPDB_UInt8 &MicroSec )
                                    { Sec      = Counter / m_Frequency;
                                      MicroSec = ((Counter * 1000 * 1000 ) / m_Frequency ) % (1000 * 1000); }

  /*!
   Function:       CounterToMicroSeconds
   Description:    Converts a counter value into microseconds.

   Arguments:      Counter            [in]  - Counter value
   Return value:   number of microseconds
  */
  SAPDB_UInt8          CounterToMicroSeconds( SAPDB_UInt8 Counter )
                                    { return (SAPDB_UInt8)( ( ((SAPDB_Int8)Counter) * (double)1000000. ) / ((SAPDB_Int8)m_Frequency) ); }


  /*!
   Function:       CallTime
   Description:    Returns the call time of the 'Counter' methode.

   Arguments:      none
   Return value:   call time of the 'Counter' methode
  */
  SAPDB_UInt8          CallTime() { return m_CallTime; }

  /*!
   Function:       OverrunUpTime
   Description:    Returns the maximum system uptime until a counter overrun occurs.

   Arguments:      Years  [out] - Uptime (years)
                   Month  [out] - Uptime (month)
   Return value:   none
  */
  void                 OverrunUpTime ( SAPDB_UInt8 &Years,
                                       SAPDB_UInt4 &Month );

protected:
  RTESys_NTHighResTimer ();

  
  bool                 LeapCheck(); // check time values for strictly monotonous increasing

private:
  virtual bool         InitTimer() = 0;
  SAPDB_UInt8          CallcCallTime();

protected:
  SAPDB_UInt8          m_StartCount;
  SAPDB_UInt8          m_Frequency;
  bool                 m_Available;
  SAPDB_UInt8          m_CallTime;

private:
  SAPDB_Singleton      m_ThisIsASingleton; // prevents copy and assignment

  static RTESys_NTHighResTimer *m_Instance;
};
#endif  /* __cplusplus */
/*! EndClass: RTESys_NTHighResTimer */


#endif  /* RTESYS_NTHIGHRESTIMER_HPP */
