/*!***************************************************************************

  module      : RTESys_SharedLib.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: System
  description : Class for loading DLL during runtime
                
  last changed: 2001-06-05  10:30
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef RTESYS_SHAREDLIB_H
#define RTESYS_SHAREDLIB_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/RTE_CompilerFeatures.h"

#if defined (__cplusplus)
#  include "RunTime/RTE_MessageList.hpp"
#else
#  include "RunTime/RTE_Types.h"
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  PROTOTYPES                                                               *
 *===========================================================================*/

/* Since these function are used from C coding also, here is the 'basic' interface used in C++ implementation too... */
/*!
  Function: RTESys_LoadSharedLibrary
  Description: This function maps the specified .DLL file into the address space of the calling process.

  If the sharedLibName ends without an extension, the platform specifc extension is added
  automatically. If an extension is given, no other extension is added...

  Arguments:  sharedLibName [in] Name of the DLL to be load 
                                 Optionally without platform specific extension!
              fullPath [out] completed path as it was used in loading call
              loadMode [in] mode describing how to handle symbols
              errorText [inout] Pointer to error text buffer
              maxErrorTextLengthInBytes [in] maximum length of output including zero termination
              sharedLibHandle [out] filled with handle to shared library on success

  Return value: true if DLL is loaded, false else (see errorText for details)
 */
externC SAPDB_Bool RTESys_LoadSharedLibrary( SAPDB_Char const *sharedLibName, 
                                             RTE_LoadMode loadMode,
                                             SAPDB_Char *errorText,
                                             SAPDB_Int4  maxErrorTextLengthInBytes,
                                             RTE_Path VAR_ARRAY_REF fullPath, 
                                             RTE_SharedLibHandle VAR_VALUE_REF sharedLibHandle );

/*!
  Function: RTESys_UnloadSharedLibrary
  Description: Unload a shared library loaded with a previous call to RTESys_LoadSharedLibrary

  Arguments:  sharedLibHandle [in] handle to shared library
              errorText [inout] Pointer to error text buffer
              maxErrorTextLengthInBytes [in] maximum length of output including zero termination

  Return value: true if DLL was unloaded, false else (see errorText for details)
 */
externC  SAPDB_Bool RTESys_UnloadSharedLibrary( RTE_SharedLibHandle sharedLibHandle,
                                                SAPDB_Char *errorText,
                                                SAPDB_Int4  maxErrorTextLengthInBytes );

/*!
  Function: RTESys_GetProcAddressFromSharedLibrary
  Description: Get a procedure address from the given shared library opened by a previous call to RTESys_LoadSharedLibrary

  Arguments:  sharedLibHandle [in] handle to shared library
              symbolicName [in] Pointer to the symbolic name of procedure address to search
              errorText [inout] Pointer to error text buffer
              maxErrorTextLengthInBytes [in] maximum length of output including zero termination
              procAddress [out] filled with procedure address if symbolic name was resolved

  Return value: true if symbol was resolved, false else (see errorText for details)
 */
externC  SAPDB_Bool RTESys_GetProcAddressFromSharedLibrary( RTE_SharedLibHandle sharedLibHandle,
                                                            SAPDB_Char const *symbolicName,
                                                            SAPDB_Char *errorText,
                                                            SAPDB_Int4  maxErrorTextLengthInBytes,
                                                            RTE_SharedLibProcAddress VAR_VALUE_REF procAddress );

#if defined (__cplusplus)

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
  class: RTESys_SharedLib
  Description: This class loads and releases dynamic shared libraries (DLL's)

  This class is introduced to load system DLL's of NT such as KERNEL32.DLL during runtime..

  NOTE: A copy constructor is forbidden, as well as an assignment operation...
 */
class RTESys_SharedLib 
{
public:
    /*!
    Function: ctor

    description:    Construct object handling loaded shared library
    
    This function maps the specified .DLL file into the address space of the calling process.
    If the sharedLibName ends without an extension, the platform specifc extension is added
    automatically. If an extension is given, no other extension is added...

    Arguments:  sharedLibName [in] Name of the DLL to be load 
                                   Optionally without platform specific extension!
                loadMode [in] mode describing how to handle symbols
                ok [out] flag to indicate loading succeeded
                errList [out] fill if ok is returned as false
    Prototypes: 2
     */
    RTESys_SharedLib ( SAPDB_Char const * sharedLibName, 
                       RTE_LoadMode loadMode,
                       SAPDB_Bool &ok,
                       SAPDBErr_MessageList &errList )
    : m_LibHandle(0)
    {
        ok = LoadSharedLib(sharedLibName, loadMode, errList);
    }
    
    RTESys_SharedLib ( SAPDB_Char const * sharedLibName, 
                       SAPDB_Bool &ok,
                       SAPDBErr_MessageList &errList)
    : m_LibHandle(0)
    {
        ok = LoadSharedLib(sharedLibName, RTE_LoadSymbolsLocal, errList);
    }

    /*!
      Function: dtor
      Description: Unload the given DLL

      Since external reference counting would be needed, it is discouraged to destruct without exactly
      knowing and synchronizing all users of the DLL...

      Since it is a dtor errors are silently ignored. If you are intersted in reporting errors during unload operation,
      call the the Unload() member function before.
    */
    ~RTESys_SharedLib()
    {
        (void)UnloadSharedLib((SAPDBErr_MessageList *)0);
    }

    /*!
      Function: Unload
      Description: Unload the given DLL

      Since external reference counting would be needed, it is discouraged to destruct without exactly
      knowing and synchronizing all users of the DLL...

      Arguments: errList [out] fill if ok is returned as false

      Return value: True if unload without problems, false else (errList is filled)
    */
    SAPDB_Bool Unload(SAPDBErr_MessageList & errList)
    {
        return UnloadSharedLib(&errList);
    }

    /*!
      Function: GetProcAddress
      Description: Get a procedure calling address of the DLL

      The DLL symbols are not type save, so the returned pointer must be casted into a proper type or a union
      must be used if casting fails... Only the caller knows how to interpret correctly the address returned...
    
      Arguments:  symbolicName [in] Name of procedure to search
                  procAddress [out] procedure address if found
                  errList [out] fill if ok is returned as false

      Return value: True if procedure found, false else
    */
    SAPDB_Bool GetProcAddress( SAPDB_Char const * symbolicName,
                               RTE_SharedLibProcAddress & procAddress,
                               SAPDBErr_MessageList     & errList);

private:
    SAPDB_Bool LoadSharedLib( SAPDB_Char const *sharedLibName,
                              RTE_LoadMode loadMode,
                              SAPDBErr_MessageList &errList);

    /* forbidden, since it would have problems without reference counting... */
    RTESys_SharedLib( const RTESys_SharedLib & ) {}
    RTESys_SharedLib & operator=( const RTESys_SharedLib & ) { return *this; }

    SAPDB_Bool UnloadSharedLib(SAPDBErr_MessageList * pErrList); /* error list is optional since used in dtor too... */

    RTE_SharedLibHandle m_LibHandle;
    RTE_Path            m_FullSharedLibName;
}; /*! Endclass: RTESys_SharedLib */

#endif /* __cplusplus */

#endif /* RTESYS_SHAREDLIB_H */

