/*!***************************************************************************

  module      : RTESys_SystemInfo.c

  -------------------------------------------------------------------------

  responsible : JoergM/RobinW

  special area: System Calls
  description : Retrieving various machine information


  last changed: 2001-04-03  16:25
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/System/RTESys_SystemInfo.h"
#include "geo60.h"      /* MSGD */
#include "geo007_1.h"   /* error messages */
#include "SAPDBCommon/SAPDB_sprintf.h"
#include <errno.h>

#if defined(WIN32)
#  include <windows.h>
#else
#  include <stdlib.h>
#  include <unistd.h>
#  include <stdio.h>
#  include <string.h>
#  include <sys/utsname.h>
#endif /* WIN32 */

#if defined(HPUX)
#  include <sys/param.h>
#  include <sys/pstat.h>
#endif /* HPUX */

#if defined(NMP)
#  include <sys/types.h>
#  include <sys/mpcntl.h>
#endif /* NMP */

#if defined(AIX)
#  include <sys/vminfo.h>
#  include <dirent.h>
#endif /* AIX */

#if defined(OSF1)
#  include <sys/sysinfo.h>
#  include <machine/hal_sysinfo.h>
#endif /* OSF1 */

#if defined(LINUX)
#  include <sys/sysinfo.h>
#endif /* LINUX */

#if defined(NMP)||defined(SUN)||defined(OSF1)
#  include <sys/stat.h>

/* The combination of a 32 bit environment and _FILE_OFFSET_BITS=32 does not work properly on Solaris.      */
/* This can be solved by redefining swapent and swaptable (see under DEFINES). Unfortunately, in Solaris 8  */
/* swap.h cannot be included under such circumstances. As swap.h is needed for the definition of SC_GETNSWP */
/* and SC_LIST, this nasty workaround has to be done...*/

#  if defined(SUN) && !defined(BIT64) && _FILE_OFFSET_BITS == 64
#    undef _FILE_OFFSET_BITS 
#    define _FILE_OFFSET_BITS 32

#    include <sys/swap.h>

#    undef _FILE_OFFSET_BITS 
#    define _FILE_OFFSET_BITS 64

#  else
#    include <sys/swap.h>
#  endif

#endif /* NMP || SUN || OSF1 */



/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#ifdef HPUX
#define USE_HPUX_PSTAT
#endif /* HPUX */

#ifdef NMP
#define USE_SNI_MPCNTL
#endif /* NMP */


#if defined(SUN) || defined(OSF1) || defined(NMP) || defined(AIX)
#define MAXSTRSIZE 128
#endif

#if defined(SUN) || defined(OSF1) || defined(NMP)

#if !defined(BIT64)

/* in <sys/swap.h> swapent and swaptable is defined like this, but ste_start and ste_length are of type off_t. Unfortunately, on 32 bit systems */
/* (Sun AND Sni!) the size of off_t is 8 bytes if _FILE_OFFSET_BITS is defined to be 64, while the library assumes the size of the fields to be */
/* 4 bytes. */

typedef struct myswapent {
        char    *ste_path;      /* get the name of the swap file */
        long    ste_start;      /* starting block for swapping */
        long    ste_length;     /* length of swap area */
        long    ste_pages;      /* numbers of pages for swapping */
        long    ste_free;       /* numbers of ste_pages free */
        int     ste_flags;      /* see below */
} myswapent_t;

typedef struct myswaptable {
        int     swt_n;                  /* number of swapents following */
        struct  myswapent swt_ent[1];     /* array of swt_n swapents */
} myswaptbl_t;

#else

#define myswapent swapent
#define myswapent_t swapent_t
#define myswaptable swaptable
#define myswaptbl_t swaptbl_t

#endif /* !BIT64 */

#endif /* SUN || NMP || OSF1 */

#if defined(WIN32)
#  define USE_GLOBAL_MEMORY_STATUS_FOR_FREE_MEMORY_
#elif defined(NMP) || defined(OSF1) || defined(AIX)
#  define USE_VMSTAT_FOR_FREE_MEMORY_
#  if defined(NMP)
#    define VMSTAT_MULTIPLIER_  BYTES_PER_KILOBYTE
#  elif defined(OSF1) || defined(AIX)
#    define VMSTAT_MULTIPLIER_  (4*BYTES_PER_KILOBYTE)
#  endif
#elif defined(LINUX)
#  define USE_SYSINFO_FOR_FREE_MEMORY_
#elif defined(SUN)
#  define USE_SYSCONF_FOR_FREE_MEMORY_
#elif defined(HPUX)
#  define USE_PSTAT_FOR_FREE_MEMORY_
#  define PSTAT_MULTIPLIER_  (4*BYTES_PER_KILOBYTE)
#endif

#define BYTES_PER_MEGABYTE      1048576
#define KILOBYTES_PER_MEGABYTE  1024
#define BYTES_PER_KILOBYTE      1024
#define BYTES_PER_GIGABYTE      (1024*1024*1024)

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  STRUCTURES, TYPES, UNIONS ...                                            *
 *===========================================================================*/
/* The field mem_unit in the sysinfo structure had been introduced with linux 2.4 */
/* It gives a multiplier for the elements in the structure that describe amounts  */
/* of memory. To make the code compilable under all releases of the linux kernel, */
/* the structure is definde here and not taken from /usr/include/linux/kernel.h   */

#ifdef LINUX
    struct new_sysinfo {
        long uptime;             /* Seconds since boot */
        unsigned long loads[3];  /* 1, 5, and 15 minute load averages */
        unsigned long totalram;  /* Total usable main memory size */
        unsigned long freeram;   /* Available memory size */
        unsigned long sharedram; /* Amount of shared memory */
        unsigned long bufferram; /* Memory used by buffers */
        unsigned long totalswap; /* Total swap space size */
        unsigned long freeswap;  /* swap space still available */
        unsigned short procs;    /* Number of current processes */
        unsigned long totalhigh; /* Total high memory size */
        unsigned long freehigh;  /* Available high memory size */
        unsigned int mem_unit;   /* Memory unit size in bytes */
        char _f[20-2*sizeof(long)-sizeof(int)]; /* Padding for libc5 */
    };
#endif

/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/

#if defined(WIN32)

static SYSTEM_INFO *getSystemInfo();

#endif /* WIN32 */

#if defined(AIX)
/* unfortunately, there is no headerfile for swapqry() */
int swapqry(char *, struct pginfo *);

#endif /* AIX */
/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

externC SAPDB_UInt4 RTESys_NumberOfCPU()
{
    SAPDB_UInt4 nProc = 0;
#if defined(WIN32)
    nProc = (SAPDB_UInt4)(getSystemInfo()->dwNumberOfProcessors);
#elif defined(_SC_NPROCESSORS_ONLN)
    /* AIX + SUN + DEC + LINUX */
    nProc = (SAPDB_UInt4)sysconf(_SC_NPROCESSORS_ONLN);
#elif defined(USE_HPUX_PSTAT)
    /* HPUX */
    struct pst_processor buf;

    while ( 0 != pstat_getprocessor(&buf, sizeof(struct pst_processor), 1, nProc) )
        nProc++;
#elif defined(USE_SNI_MPCNTL)
    /* NMP */
    nProc = (SAPDB_UInt4)mpcntl(MPCNTL_CPUCNT, (void *)0);
#endif

    return ( nProc > 0 ? nProc : (SAPDB_UInt4)1 );
}

/*-----------------*/

externC SAPDB_UInt4 RTESys_SystemPageSize()
{
  SAPDB_UInt4 pageSize;
  
#if defined(WIN32)
  pageSize = (SAPDB_UInt4)(getSystemInfo()->dwPageSize);
#elif defined(_SC_PAGESIZE)
  pageSize = (SAPDB_UInt4)sysconf(_SC_PAGESIZE);
#elif defined(_SC_PAGE_SIZE)
  pageSize = (SAPDB_UInt4)sysconf(_SC_PAGE_SIZE);
#endif
  return pageSize;
}

/*-----------------*/

externC SAPDB_ULong RTESys_PhysicalMemorySize()
{
#if defined(AIX)
    char receiveBuffer[MAXSTRSIZE];
    FILE *fileHandle;
    char *beginOfNumber,*endOfNumber;

    fileHandle = popen("lsattr -E -l sys0 -a realmem","r");
    if(NULL == fileHandle)
    {
        MSGD (( ERR_POPEN_SYSCALL, "lsattr -E -l sys0 -a realmem", errno));
        return 0;
    }
    fgets(receiveBuffer,MAXSTRSIZE,fileHandle);
    if(-1 == pclose(fileHandle))
    {
        MSGD (( ERR_SYSCALL, "pclose", errno));
        return 0;
    }

    /* the lsattr command gives the size of the memory in a string like */
    /* "realmem 8388608 Amount of usable physical memory in Kbytes False" */
    /* so the output has to be parsed */
    beginOfNumber = strchr(receiveBuffer,' ')+1;
    if(0 == beginOfNumber)
    {
        MSGD (( ERR_COULD_NOT_FIND_NUMBER_IN_STRING ));
        return 0;
    }
    endOfNumber = strchr(beginOfNumber,' ');
    if(0 == endOfNumber)
    {
        MSGD (( ERR_COULD_NOT_FIND_NUMBER_IN_STRING ));
        return 0;
    }
    *endOfNumber = 0;
    return strtoul(beginOfNumber,NULL,10)/KILOBYTES_PER_MEGABYTE;
#endif /*AIX*/

#if defined(HPUX)
    struct pst_static pst;
    union pstun pu;
    
    pu.pst_static = &pst;
    if ( pstat( PSTAT_STATIC, pu,(size_t)sizeof(pst), (size_t)0, 0 ) != -1 ) 
    {
        return (unsigned long)( (double)pst.physical_memory * pst.page_size / BYTES_PER_MEGABYTE ) ;
    }
    else 
    {
        MSGD (( ERR_SYSCALL, "pstat", errno));
        return 0;
    }
#endif /*HPUX*/

#if defined(LINUX)
    struct new_sysinfo info;

    unsigned long totalBytes;
    memset(&info,0,sizeof(info));
    if(-1 == sysinfo((struct sysinfo *)&info))
    {
        MSGD (( ERR_SYSCALL, "sysinfo",errno));
        return 0;
    }

    totalBytes = info.totalram;

    if(0 != info.mem_unit)
    {
        totalBytes *= info.mem_unit;
    }
    return (totalBytes / BYTES_PER_MEGABYTE);
#endif /*LINUX*/

#if defined(NMP)
    char receiveBuffer[MAXSTRSIZE];
    FILE *fileHandle;
    char *beginOfNumber;

    *totalMB = 0;
    *freeMB  = 0;   

    fileHandle = popen("uname -M","r");
    if(NULL == fileHandle)
    {
        MSGD (( ERR_POPEN_SYSCALL, "uname -M", errno));
        return 0;
    }
    fgets(receiveBuffer,MAXSTRSIZE,fileHandle);
    if(-1 == pclose(fileHandle))
    {
        MSGD (( ERR_SYSCALL, "pclose", errno));
        return;
    }

    beginOfNumber = strchr(receiveBuffer,'/')+1;
    if(0 == beginOfNumber)
    {
        MSGD (( ERR_COULD_NOT_FIND_NUMBER_IN_STRING ));
        return 0;
    }
    else
    {
        return strtoul(beginOfNumber,NULL,10);
    }
#endif /*NMP*/

#if defined(OSF1)
    int memsize;
    getsysinfo(GSI_PHYSMEM,(char *)&memsize,sizeof(int),0,0,0);
    return memsize/KILOBYTES_PER_MEGABYTE;
#endif /*OSF1*/

#if defined(SUN)
    return (unsigned long)( (double)sysconf(_SC_PHYS_PAGES) * sysconf(_SC_PAGE_SIZE) / BYTES_PER_MEGABYTE );
#endif /*SUN*/

#if defined(_WIN32)

    MEMORYSTATUSEX memoryStatus;
    memoryStatus.dwLength = sizeof(MEMORYSTATUSEX);

    if(0==GlobalMemoryStatusEx (&memoryStatus))
    {
        return 0;
    }
    else
    {
        return (SAPDB_ULong)(memoryStatus.ullTotalPhys/BYTES_PER_MEGABYTE);
    }
  
#endif /*_WIN32*/
}

/*-----------------*/

externC SAPDB_ULong RTESys_PhysicalMemoryAvailable()
{
#if defined(USE_GLOBAL_MEMORY_STATUS_FOR_FREE_MEMORY_)
    MEMORYSTATUSEX memoryStatus;
    memoryStatus.dwLength = sizeof(MEMORYSTATUSEX);

    if(0==GlobalMemoryStatusEx (&memoryStatus))
    {
        return 0;
    }
    else
    {
        return (SAPDB_ULong)(memoryStatus.ullAvailPhys/BYTES_PER_MEGABYTE);
    }
#elif defined (USE_VMSTAT_FOR_FREE_MEMORY_)
    FILE *fileHandle;
    char receiveBuffer[MAXSTRSIZE];
    char *positionOfFre;    /* */
    char *beginningOfNumber;
    char *endOfNumber;
    SAPDB_ULong number;

    fileHandle = popen("vmstat","r");
    if(NULL == fileHandle)
    {
        MSGD (( ERR_POPEN_SYSCALL, "vmstat", errno));
        return 0;
    }

    do
    {
        fgets(receiveBuffer,MAXSTRSIZE,fileHandle);
    }
    while(0 == (positionOfFre = strstr(receiveBuffer,"fre")));

    fgets(receiveBuffer,MAXSTRSIZE,fileHandle);

    printf("%x\n",*(receiveBuffer));
    beginningOfNumber = positionOfFre;
    if(' ' == *beginningOfNumber)
    {
        while(' ' == *beginningOfNumber)
        {
            beginningOfNumber++;
        }
    }
    else
    {
        while(' ' != *beginningOfNumber)
        {
            beginningOfNumber--;
        }
        beginningOfNumber++;
    }

    endOfNumber = beginningOfNumber;
    while(' ' != *endOfNumber)
    {
        endOfNumber++;
    }
    *endOfNumber = 0;

    number = strtoul(beginningOfNumber,NULL,10);
    
    switch(*(endOfNumber-1))
    {
    case 'k':
    case 'K':
        number *= BYTES_PER_KILOBYTE;
        break;
    case 'm':
    case 'M':
        number *= BYTES_PER_MEGABYTE;
        break;
    case 'g':
    case 'G':
        number *= BYTES_PER_GIGABYTE;
        break;
    }

    number *= VMSTAT_MULTIPLIER_;
    number /= BYTES_PER_MEGABYTE;

    if(-1 == pclose(fileHandle))
    {
        printf("pclose\n");
        MSGD (( ERR_SYSCALL, "pclose", errno));
        return 0;
    }
    return number;
    
#elif defined(USE_PSTAT_FOR_FREE_MEMORY_)
    struct pst_dynamic dynamic;
	union pstun pu;
        
	pu.pst_dynamic = &dynamic;
	if ( pstat( PSTAT_DYNAMIC, pu,(size_t)sizeof(struct pst_dynamic), (size_t)0, 0 ) != -1 ) 
	{
        return (unsigned long)( (double)dynamic.psd_free * ((double)PSTAT_MULTIPLIER_ / BYTES_PER_MEGABYTE)) ;
	}
#elif defined(USE_SYSCONF_FOR_FREE_MEMORY_)
    return (unsigned long)( ((double)sysconf(_SC_AVPHYS_PAGES) * sysconf(_SC_PAGE_SIZE)) / BYTES_PER_MEGABYTE);
#elif defined (USE_SYSINFO_FOR_FREE_MEMORY_)
    SAPDB_ULong numberAccordingToSysInfo;

    struct new_sysinfo info;
	memset(&info,0,sizeof(info));
	sysinfo((struct sysinfo *)&info);

	numberAccordingToSysInfo = (info.freeram);
    printf("%lu\n",numberAccordingToSysInfo);

	if(0 != info.mem_unit)
	{
		numberAccordingToSysInfo *= info.mem_unit;
	}
    return numberAccordingToSysInfo / BYTES_PER_MEGABYTE;
#else
    return 0;
#endif

}

/*-----------------*/

externC void RTESys_SwapSize(SAPDB_ULong *totalMB,SAPDB_ULong *freeMB)
{
#if defined(AIX)
    /* AIX only offers a function to get details about a given swap devices */
    /* So all block devices in /dev/ are enumerated and given to this function */
    struct dirent *dirEntry;
    DIR *directoryHandle;

    static struct pginfo    pginfo;
    double totalPages=0.0,freePages=0.0;
    char *deviceName;

    *totalMB = 0;
    *freeMB  = 0;   

    if (NULL == (directoryHandle = opendir("/dev")))
    {
        MSGD (( ERR_OPENDIR_SYSCALL, "/dev/", errno));
        return;
    }
    else
    {
        struct stat statStruct;

        /* read all entries from /dev/ */
        while (dirEntry = readdir(directoryHandle))
        {
            deviceName = malloc(strlen(dirEntry->d_name) + strlen("/dev/") + 1);
            if(NULL == deviceName)
            {    /* NULL == deviceName --> could not allocate memory */
                MSGD (( ERR_ALLOC_MEMORY, errno));
                break;
            }
            else 
            {
                strcpy(deviceName,"/dev/");
                strcat(deviceName,dirEntry->d_name);
                /* check if it is a block device. The function would work without this check, but the swapqry() call seems to */
                /* take much more time than the stat() call */
                if(-1 == stat(deviceName,&statStruct))
                {
                    MSGD (( ERR_SYSCALL, "stat", errno));
                    return;
                }
                if(_S_IFBLK == (statStruct.st_mode & _S_IFBLK))
                {
                    if (swapqry(deviceName, &pginfo) != -1) 
                    {
                        /* If we reach here, we have really found a swap device and can add the size */
                        totalPages += pginfo.size;
                        freePages  += pginfo.free;
                    }
                }
                free(deviceName);
            }
        }
        if(-1 == closedir(directoryHandle))
        {
            MSGD (( ERR_SYSCALL, "closedir", errno));
        }
    }

    *totalMB = (SAPDB_UInt8)(totalPages * (double)RTESys_SystemPageSize() / (double)BYTES_PER_MEGABYTE);
    *freeMB  = (SAPDB_UInt8)(freePages  * (double)RTESys_SystemPageSize() / (double)BYTES_PER_MEGABYTE);
#endif /*AIX*/

#if defined(HPUX)
    struct pst_swapinfo info;

    int counter;

    counter = 0;
    while(pstat_getswap(&info,sizeof(info),1,counter))
    {
        *totalMB += info.pss_un.Pss_blk.Pss_nblks;
        *freeMB  += info.pss_un.Pss_blk.Pss_nblksavail;
        counter++;
    }
    *totalMB /= BYTES_PER_KILOBYTE;
    *freeMB  /= BYTES_PER_KILOBYTE;
#endif /*HPUX*/

#if defined(LINUX)
    struct new_sysinfo info;
    double totalMemUnits=0.0,freeMemUnits=0.0;

    *totalMB = 0;
    *freeMB  = 0;   

    memset(&info,0,sizeof(info));
    if(-1 == sysinfo((struct sysinfo *)&info))
    {
        MSGD (( ERR_SYSCALL, "sysinfo", errno));
        return;
    }
    totalMemUnits = info.totalswap;
    freeMemUnits  = info.freeswap;

    /* in older linux kernels, the entries in the sysinfo struct were always given in bytes */
    /* in newer kernels, the entries are given in "mem_unit"s and a field mem_unit was added */
    /* in older versions there is a 0 at the space of this field. */
    if(0 != info.mem_unit)
    {
        totalMemUnits *= info.mem_unit;
        freeMemUnits  *= info.mem_unit;
    }
    *totalMB = totalMemUnits / (double)BYTES_PER_MEGABYTE;
    *freeMB  = freeMemUnits  / (double)BYTES_PER_MEGABYTE;
#endif /*LINUX*/

#if defined(SUN) || defined(OSF1) || defined(NMP)
    /* the interface for swapctl() is a little difficult to use: When called with SC_LIST, it fills an array of structures. */
    /* Each swap device has an entry in this array. Unfortunately, the space for the array as well as space for the path    */
    /* entries in this structure has to be allocated before calling swapctl(). So swaptctl() has to be called with SC_GETNSWP */
    /* in order to determine the number of swap devices. Theoretically, the number of swap devices can change between */
    /* swapctl(SC_GETNSWP,..) and swapctl(SC_GETLIST,..). The latter call returns the number of swap devices actually found. */
    /* If this number differs from the one returned by swapctl(SC_GETNSWP..), the space for the array has to be allocated */
    /* once again and the whole process has to be repeated */
    myswaptbl_t    *swaptable;
    int            numberOfSwapTableEntries;        /* return value from swapctl(SC_GETNSWP,..) */
    int            actualNumberOfSwapTableEntries;  /* return value from swapctl(SC_LIST,..) */
    int            counter;
    char           *stringTable;    /* string table for path names */

    double totalPages=0.0,freePages=0.0;

    *totalMB = 0;
    *freeMB  = 0;   

again:
    if ((numberOfSwapTableEntries = swapctl(SC_GETNSWP, 0)) <= 0) 
    {
        MSGD (( ERR_SYSCALL, "swapctl", errno));
        return;
    }
    /* allocate swaptable for num+1 entries */
    if ((swaptable = (myswaptbl_t *)malloc(numberOfSwapTableEntries * sizeof(myswapent_t) + sizeof(struct myswaptable))) == (void *) 0) 
    {
        MSGD (( ERR_ALLOC_MEMORY, errno));
        return;
    }

    memset(swaptable,0,numberOfSwapTableEntries * sizeof(myswapent_t) + sizeof(struct myswaptable));
    /* allocate num+1 string holders */
    if ((stringTable = (char *)malloc((numberOfSwapTableEntries + 1) * MAXSTRSIZE)) == (void *) 0) 
    {
        MSGD (( ERR_ALLOC_MEMORY, errno));
        return;
    }
    memset(stringTable,0,(numberOfSwapTableEntries + 1) * MAXSTRSIZE);
    /* initialize string pointers */
    for (counter = 0; counter < (numberOfSwapTableEntries + 1); counter++) 
    {
        swaptable->swt_ent[counter].ste_path = stringTable + (counter * MAXSTRSIZE);
    }

    swaptable->swt_n = numberOfSwapTableEntries + 1;
    if ((actualNumberOfSwapTableEntries = swapctl(SC_LIST, swaptable)) < 0) 
    {
        MSGD (( ERR_SYSCALL, "swapctl", errno));
        return;
    }
    if (actualNumberOfSwapTableEntries > numberOfSwapTableEntries) 
    {        /* more were added */
        free(swaptable);
        free(stringTable);
        goto again;
    }
    /* If we have reached here, we finally have all swap devices in our swaptable...*/
    for (counter = 0; counter < actualNumberOfSwapTableEntries; counter++)
    {
        totalPages += swaptable->swt_ent[counter].ste_pages;
        freePages  += swaptable->swt_ent[counter].ste_free;
    }

    *totalMB = (SAPDB_UInt8)(totalPages * (double)RTESys_SystemPageSize() / (double)BYTES_PER_MEGABYTE);
    *freeMB  = (SAPDB_UInt8)(freePages  * (double)RTESys_SystemPageSize() / (double)BYTES_PER_MEGABYTE);

    free(swaptable);
    free(stringTable);
#endif /*SUN || OSF1 || NMP*/

#if defined(_WIN32)
    MEMORYSTATUSEX memoryStatus;
    memoryStatus.dwLength = sizeof(MEMORYSTATUSEX);

    if(0==GlobalMemoryStatusEx (&memoryStatus))
    {
        *totalMB = 0;
        *freeMB  = 0;
    }
    else
    {
        *totalMB = (SAPDB_ULong)(memoryStatus.ullTotalPageFile/BYTES_PER_MEGABYTE);
        *freeMB  = (SAPDB_ULong)(memoryStatus.ullAvailPageFile/BYTES_PER_MEGABYTE);
    }
#endif /*_WIN32*/
}

/*-----------------*/

externC void RTESys_MemInfo(SAPDB_ULong *totalPhysicalMB,SAPDB_ULong *freePhysicalMB,SAPDB_ULong *totalSwapMB,SAPDB_ULong *freeSwapMB)
{
    RTESys_SwapSize(totalSwapMB,freeSwapMB);
    *totalPhysicalMB = RTESys_PhysicalMemorySize();
    *freePhysicalMB  = RTESys_PhysicalMemoryAvailable();
}

/*-----------------*/

externC SAPDB_Bool RTESys_GetCPUType(SAPDB_Char *string,SAPDB_UInt4 length)
{
#ifdef _WIN32
    switch (getSystemInfo()->wProcessorArchitecture)
    {
    case PROCESSOR_ARCHITECTURE_INTEL:
        SAPDB_sprintf(string,length,"Intel IA32");
        break;
    case PROCESSOR_ARCHITECTURE_IA64:
        SAPDB_sprintf(string,length,"Intel IA64");
        break;
    case PROCESSOR_ARCHITECTURE_IA32_ON_WIN64:
        SAPDB_sprintf(string,length,"Intel IA32 on WIN64");
        break;
    case PROCESSOR_ARCHITECTURE_AMD64:
        SAPDB_sprintf(string,length,"AMD64");
        break;
    }
    SAPDB_sprintf(string,length,"%s level %x revision %x",
        string,
        getSystemInfo()->wProcessorLevel,
        getSystemInfo()->wProcessorRevision);
#else   /* UNIX: take the info from uname */
    struct utsname utsNameStruct;
    uname(&utsNameStruct);
#if defined (AIX)
    SAPDB_sprintf(string,length,"IBM RS 6000 (%s)",
        utsNameStruct.machine);
#elif defined(HP9)
#  if defined(PA20W)||defined(PA11)
    SAPDB_sprintf(string,length,"HP PA-RISC (%s)",
        utsNameStruct.machine);
#  elif defined(HP_IA64)
    SAPDB_sprintf(string,length,"HP IA64 (%s)",
        utsNameStruct.machine);
#  endif
#elif defined (SUN)
    SAPDB_sprintf(string,length,"Sparc (%s)",
        utsNameStruct.machine);
#else
    SAPDB_sprintf(string,length,"%s",
        utsNameStruct.machine);
#endif
#endif  /* WIN32 / UNIX */
    return true;
}    

/*===========================================================================*
 *  STATIC FUNCTIONS                                                         *
 *===========================================================================*/

#if defined(WIN32)

static SYSTEM_INFO *getSystemInfo()
{
    static SYSTEM_INFO   systemInfo;
    static SYSTEM_INFO *pSystemInfo = 0;

    if ( !pSystemInfo )
    {
        GetSystemInfo(&systemInfo);
        pSystemInfo = &systemInfo;
    }
    return pSystemInfo;
}

#endif


/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
