/*!***************************************************************************

  module      : RTESys_Time.c

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: System Calls
  description : System Calls for Time


  last changed: 2001-04-03  16:25
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "RunTime/System/RTESys_Time.h"

#if defined(_WIN32)
#include <windows.h>
#else
#include <poll.h>
#endif
#include <time.h>

#ifdef AIX
#   include     <sys/time.h>
#endif


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  STRUCTURES, TYPES, UNIONS ...                                            *
 *===========================================================================*/

static SAPDB_Int4 RTESys_InCrashSignalHandler = 0;

/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/

/*
  Function: FillProtocolTimeStamp
  Description: Fill the given buffer with local time information
   Format YYYY-MM-DD HH:MM:SS
  Argument: timeBuffer [out] the time buffer to fill
            isLocalTime [in] flag to indicate local time usage or GMT (UTC)
  Return value: Pointer to first character
 */
static SAPDB_Char *FillProtocolTimeStamp(RTESys_Timestamp timeBuffer, SAPDB_Int4 isLocalTime);

/*
  Function: FillDigits
  Description: Fill the given value into buffer

  The given pointer points behind last character to write.
  Number of digits is given (producing lots of leading zeros).
  If separator is not 0 (binary, not ascii...) a separator is
  written after ndigits are written.
  
  Argument: value [in] the value to convert into digits
            ptrBehindLastDigit [in] pointer behind last digit to write
            digitCount [in] number of digits to write
            separator [in] optional separator character
  Return value: Pointer to last character written (most significant digit or separator)
 */
static SAPDB_Char *FillDigits(SAPDB_UInt8 value,
                              SAPDB_Char *ptrBehindLastDigit,
                              SAPDB_Int4 digitCount, 
                              SAPDB_Char separator);

/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

/*
  Function: RTESys_FillProtocolTimeStamp
  Description: Fill the given buffer with local time information
   Format YYYY-MM-DD HH:MM:SS
  Argument: timeBuffer [out] the time buffer to fill
  Return value: Pointer to first character
 */
SAPDB_Char *RTESys_FillProtocolTimeStamp(RTESys_Timestamp timeBuffer)
{
    return FillProtocolTimeStamp(timeBuffer, 1);
}

/*
  Function: RTESys_FillProtocolGMTTimeStamp
  Description: Fill the given buffer with local time information
   Format YYYY-MM-DD HH:MM:SS
  Argument: timeBuffer [out] the time buffer to fill
  Return value: Pointer to first character
 */
SAPDB_Char *RTESys_FillProtocolGMTTimeStamp(RTESys_Timestamp timeBuffer)
{
    return FillProtocolTimeStamp(timeBuffer, 0);
}

/*
  Function: RTESys_Time
  Description: Returns seconds since 1.1.1990
  Argument: none
  Return value: Seconds since 1.1.1990
 */
SAPDB_UInt8 RTESys_Time()
{
    time_t clk ;
	clk = time ( (time_t *) 0 );

    return (SAPDB_UInt8)clk;
}

/*
  Function: RTESys_SetInCrashSignalHandler
  Description: Set flag to indicate we are running inside a crash signal handler

  Some of the function we use are not async signal save. To prevent calling them the signal
  handler must call this flag before issuing any of the system calls
 */
void RTESys_SetInCrashSignalHandler()
{
    RTESys_InCrashSignalHandler = 1;
}

/*
  Function: RTESys_ResetInCrashSignalHandler
  Description: Reset flag to indicate we are running outside a crash signal handler

  Some of the function we use are not async signal save. To prevent calling them the signal
  handler must call this global flag before issuing any of the system calls
 */
void RTESys_ResetInCrashSignalHandler()
{
    RTESys_InCrashSignalHandler = 0;
}

/*
  Function: RTESys_IsInCrashSignalHandler
  Description: Check flag to indicate we are running inside a crash signal handler
  Return value: true if flag set, false if flag is reset
 */
SAPDB_Bool RTESys_IsInCrashSignalHandler()
{
    return (SAPDB_Bool)(0 != RTESys_InCrashSignalHandler);
}

/*===========================================================================*
 *  STATUC FUNCTIONS                                                         *
 *===========================================================================*/

/*
  Function: FillDigits
  Description: Fill the given buffer with ascii representation of given number of digits


  Argument: value [in] number to convert
            ptrBehindLastDigit [in] pointer behind last digit to write
			digitCount [in] number of digits to write
			separator [in] seprator to add after last digit written
  Return value: Pointer to last written character
 */
static SAPDB_Char *FillDigits(SAPDB_UInt8 value,
                              SAPDB_Char *ptrBehindLastDigit,
                              SAPDB_Int4 digitCount, 
                              SAPDB_Char separator)
{
    SAPDB_Int4 nDigits = digitCount;
    SAPDB_UInt8 tmpValue = value;
    SAPDB_Char *ptr = ptrBehindLastDigit;

    while ( nDigits > 1 )
    {
        *--ptr = '0'+(SAPDB_Char)(value % 10);
        value /= 10;
        --nDigits;
    }
    *--ptr = '0'+(SAPDB_Char)value;
    if ( 0 != separator )
    {
        *--ptr = separator;
    }

    return ptr;
}

/*--------------------------------*/

#if !defined(_WIN32)
static struct tm lastTime_buffer = { 0 }; /* PTS 1112083 */
static time_t    lastClk = 0; 

static struct tm *EmergencyLocaltime(time_t clk, SAPDB_Int4 isLocalTime )
{
	struct tm *tim;

    if ( !RTESys_InCrashSignalHandler )
    {
#ifdef _REENTRANT
        struct tm   tm_buffer;
        if ( isLocalTime )
        {
            tim = localtime_r( &clk, &tm_buffer ); /* Not async signal save!!! */
        }
        else
        {
            tim = gmtime_r( &clk, &tm_buffer ); /* Not async signal save!!! */
        }
#else
        if ( isLocalTime )
        {
            tim = localtime( &clk ); /* Not async signal save!!! */
        }
        else
        {
            tim = gmtime( &clk ); /* Not async signal save!!! */
        }
#endif
		memcpy(&lastTime_buffer, tim, sizeof(struct tm));
		tim = &lastTime_buffer;
		lastClk = clk;
	}
	else
	{
        /*
         * We have no way to detect, that a 'free()' is crashing the database.
         * Since 'localtime_r' uses free() internally, we have to prevent from using
         * it during crash. Therefore we only output 'clk' as seconds since 1.1.1970
         */

		/* PTS 1112083 since stack back trace is needed for support, 'clk' output was visible in knldiag.err */
		/* To make it reasonable the last timestamp produced is used and the time difference is used to update */
		/* A difference in weekdays is not correctly handled, but... */
		long timeDiff = clk - lastClk;
		tim = &lastTime_buffer;

		if ( timeDiff > 0 )
		{
			tim->tm_sec += (timeDiff % 60);
			timeDiff /= 60;
			if ( timeDiff > 0 )
			{
				tim->tm_min += (timeDiff % 60);
                timeDiff /= 60;
				if ( timeDiff > 0 )
				{
					tim->tm_hour += (timeDiff % 24);
                    timeDiff /= 24;
					if ( timeDiff > 0 )
					{
						tim->tm_mday += timeDiff;
					}
				}
			}

            /* work on overflows... */
			if ( tim->tm_sec >= 60 )
			{
			   tim->tm_min += tim->tm_sec/60;
			   tim->tm_sec %= 60;
			}

			if ( tim->tm_min >= 60 )
			{
			   tim->tm_hour += tim->tm_min/60;
			   tim->tm_min %= 60;
			}

			if ( tim->tm_hour >= 60 )
			{
			   tim->tm_mday += tim->tm_hour/24;
			   tim->tm_hour %= 24;
			}

            /* overflow on mday are accepted... */

			lastClk = clk;
		}
	}
    return tim;
}
#endif

/*--------------------------------*/

static SAPDB_Char *FillProtocolTimeStamp(RTESys_Timestamp timeBuffer, SAPDB_Int4 isLocalTime )
{
    SAPDB_Char  *ptr = &timeBuffer[sizeof(RTESys_Timestamp)-1];
    /* ptr starts at end of buffer and is used to fill beginning with least significant digits */
    *ptr = 0;
    {
#if defined(_WIN32)

    /*----------------*/
        /* Windows */
    /*----------------*/

        SYSTEMTIME   DateTime;

        if ( isLocalTime )
        {
            GetLocalTime(&DateTime);
        }
        else
        {
            GetSystemTime(&DateTime);
        }

        ptr = FillDigits( (SAPDB_UInt8)DateTime.wSecond, ptr, 2, (SAPDB_Char)':');
        ptr = FillDigits( (SAPDB_UInt8)DateTime.wMinute, ptr, 2, (SAPDB_Char)':');
        ptr = FillDigits( (SAPDB_UInt8)DateTime.wHour,   ptr, 2, (SAPDB_Char)' ');
        ptr = FillDigits( (SAPDB_UInt8)DateTime.wDay,    ptr, 2, (SAPDB_Char)'-');
        ptr = FillDigits( (SAPDB_UInt8)DateTime.wMonth,  ptr, 2, (SAPDB_Char)'-');
        ptr = FillDigits( (SAPDB_UInt8)DateTime.wYear,   ptr, 4, (SAPDB_Char)0  );

#else

    /*----------------*/
         /* UNIX */
    /*----------------*/

	/*
	 *  Create date and time stamp
	 */
        struct tm * tim = EmergencyLocaltime( time ( (time_t *) 0 ), isLocalTime );

		ptr = FillDigits( (SAPDB_UInt8)tim->tm_sec,         ptr, 2, (SAPDB_Char)':');
        ptr = FillDigits( (SAPDB_UInt8)tim->tm_min,         ptr, 2, (SAPDB_Char)':');
        ptr = FillDigits( (SAPDB_UInt8)tim->tm_hour,        ptr, 2, (SAPDB_Char)' ');
        ptr = FillDigits( (SAPDB_UInt8)tim->tm_mday,        ptr, 2, (SAPDB_Char)'-');
        ptr = FillDigits( (SAPDB_UInt8)(tim->tm_mon+1),     ptr, 2, (SAPDB_Char)'-');
        ptr = FillDigits( (SAPDB_UInt8)(tim->tm_year+1900), ptr, 4, (SAPDB_Char)0  );
#endif
    }
    return ptr;
}

/*--------------------------------*/

SAPDB_UInt4 RTESys_Timeout(bool *pFlag, SAPDB_UInt4 waitingIntervalsOf100msec)
{
    SAPDB_UInt intervalsLeft = ( waitingIntervalsOf100msec > 0 ? waitingIntervalsOf100msec : 1 );

    while ( ( false == *pFlag )
         && ( intervalsLeft > 1 ) )
    {
        --intervalsLeft;
        /* 100 milliseconds polling period */
#if defined(_WIN32)
        Sleep( 100 );
#else
        {
            struct pollfd dummy[1];
            dummy[0].fd = -1;
            dummy[0].events = 0;

            poll(dummy, 1, 100);
        }
#endif
    }

    return intervalsLeft;
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/