/*!***************************************************************************
  module:       SAPDBAlgo_Quicksort

  responsible:  Henrik

  special area: Basic Algorithms

  description:  Implementation of Quicksort

  see:          http://pwww.bea.sap-ag.de/Kern/v72/BD-Sortieralgorithmen.doc

  copyright:    Copyrighted by SAP AG, 1999

  created:      2002-04-19

  updated:

    ========== licence begin  GPL
    Copyright (C) 2002 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

#ifndef SAPDBALGO_QUICKSORT_H
#define SAPDBALGO_QUICKSORT_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

// maximal field length for the QuickSort algorithm and its stack
// ( log2(MAX_ELEMS_QUICKS)*2 )
#define MAX_ELEMS_QUICKS         2000000
#define MAX_STACK_ELEMS_QUICKS   32

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/


struct QuickSortItem
{
    tsp00_Int4  qsiLeft;
    tsp00_Int4  qsiRight;
};

/*===========================================================================*
 * DECLARATIONS                                                              *
 *===========================================================================*/

/*!
  Chapter: In-memory sorting

  description: Generic sort algorithms suitable for in-memory data.
*/

/*!
  function: SAPDBAlgo_QuickSort

  template arguments:
    ElementClass  The type of the elements to be sorted.
    CompareClass  The comparison operator's type.

  arguments:
    Elements         [in/out]
    NumberOfElements [in]
    Compare          [in] optional

  return value:    Success or failure.

  description:
    Non recursive quicksort implementation for sorting lists of elements.

    A list consists of an array of 'NumberOfElements' elements of type
    'ElementClass'. A direct memory access via 'Elements[Index]' must be
    possible.

    Version 1 of the function needs a user defined comparison-operator which
    has to be derived from the abstract class 'cgg200ComparisonOperator'.
    Version 2 uses the C++-defined operators '<' and '>' and is suitable
    for elements of basic types (int, float, char).

  code:
    // Call example 1 (basic type)
    int MyArray[500];
    SAPDBAlgo_QuickSort(MyArray, 500);

*/

template <class ElementClass>
bool SAPDBAlgo_QuickSort(
					ElementClass		*Elements,
					const tsp00_Int4	NumberOfElements);

/*---------------------------------------------------------------------------*/

// Quicksort implementation using the default compare operators (<,>).
template <class ElementClass>
bool SAPDBAlgo_QuickSort(
					ElementClass		*Elements,
					const tsp00_Int4	NumberOfElements)
{
    tsp00_Int4              i;
    tsp00_Int4	            j;
    tsp00_Int4	            Left;
    tsp00_Int4	            Right;
    tsp00_Int4              PivotIndex;
    tsp00_Int4              StackIndex = 0;
    ElementClass            AuxElem;
    ElementClass            PivotElem;
    QuickSortItem    Stack[MAX_STACK_ELEMS_QUICKS];

    if (MAX_ELEMS_QUICKS <= NumberOfElements) return false;
    if (0 > NumberOfElements) return false;

    if ((0 == NumberOfElements) || (Elements == NULL)) return true;

    Stack [StackIndex].qsiLeft = 0;
    Stack [StackIndex].qsiRight = NumberOfElements - 1;
    do
    {
        Left  = Stack [StackIndex].qsiLeft;
        Right = Stack [StackIndex].qsiRight;
        StackIndex --;
        do
        {
            i          = Left;
            j          = Right;
            PivotIndex = (Left + Right) >> 1;
            PivotElem  = Elements [PivotIndex];
            do
            {
                while (Elements [i] < PivotElem) i++;
                while (PivotElem    < Elements [j]) j--;

                if (i <= j)
                {
                    AuxElem      = Elements [i];
                    Elements [i] = Elements [j];
                    Elements [j] = AuxElem;
                    i++;
                    j--;
                }
            }
            while (i <= j);

            if ((j - Left) < (Right - i))
            {
                if (i < Right)
                {
                    if (MAX_STACK_ELEMS_QUICKS <= StackIndex) return false;
                    StackIndex++;
                    Stack [StackIndex].qsiLeft  = i;
                    Stack [StackIndex].qsiRight = Right;
                }
                Right = j;
            }
            else
            {
                if (Left < j)
                {
                    if (MAX_STACK_ELEMS_QUICKS <= StackIndex) return false;
                    StackIndex++;
                    Stack [StackIndex].qsiLeft  = Left;
                    Stack [StackIndex].qsiRight = j;
                }
                Left = i;
            }
        }
        while (Left < Right);
    }
    while (StackIndex >= 0);
    return true;
}

/*---------------------------------------------------------------------------*/

#endif  /* GGG200_H */



