/*!***************************************************************************

  module      : SAPDBConfig_Parameters.cpp

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: SAPDB configuration parameters
  description : class for kernel configurable parameter access


  last changed: 2000-06-19  10:20
  see also    : example.html ...
  first created:2000-03-22  17:49

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/SAPDB_ToString.hpp"
#include "SAPDBCommon/Configuration/SAPDBConfig_Parameters.hpp"
#include "SAPDBCommon/Configuration/SAPDBConfig_Messages.hpp" /* nocheck */
#include "heo11.h"
#include "heo46.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define __CONTEXT__  __FILE__,__LINE__


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/


/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

SAPDBConfig_ConfigurationParameters::SAPDBConfig_ConfigurationParameters(const SAPDB_Char *ParamFileName)
{

  tsp11_XpReturnCode XpRet ;
  tsp01_RteError     RteError ;
  tsp00_DbNamec      DBName ;
       
  DBName.rawAssign(ParamFileName);
  XpRet = sqlXparamOpen   ( DBName, &m_plXParamFd, XP_OPEN_READONLY_SP11, &RteError );

  if ( XpRet != xp_ok_sp11 )
  {
    m_plXParamFd = -1 ;
    throw SAPDBConfig_ParameterAccessFailure( __CONTEXT__,
                                             SAPDBCONFIG_OPEN_PARAMFILE_FAILED, 
                                             SAPDB_ToString(ParamFileName), 
                                             SAPDB_ToString(RteError.RteErrText));
  }
}

/*!*****************************************************************************
*******************************************************************************/
SAPDBConfig_ConfigurationParameters::~SAPDBConfig_ConfigurationParameters()
{
  if ( m_plXParamFd >= 0 ) 
  {
      tsp01_RteError     RteError ;

      sqlXparamClose( m_plXParamFd, false, &RteError ) ;
  }
}

/*!*****************************************************************************
*******************************************************************************/
SAPDB_Int4 SAPDBConfig_ConfigurationParameters::GetIntParameter(const SAPDB_Char *Idenfifier) const 
{

  tsp11_XpKeyTypec ParamKey ;
  XP_VALUE_REC     ParamValue ;
  tsp01_RteError   RteError ;
  SAPDB_Int4       IntParameter ;

  ParamKey.rawAssign(Idenfifier);

  if ( sqlXparamGet ( m_plXParamFd, ParamKey, reinterpret_cast<tsp00_XpValueRec *>(&ParamValue), &RteError ) == xp_ok_sp11 )
  {
    switch ( ParamValue.xp_value_t )
    {
      case xp_int2_type_sp11   : IntParameter = (SAPDB_Int4 ) ParamValue.xp_type.xp_int2;
                                 break;
      case xp_int4_type_sp11   : IntParameter = (SAPDB_Int4 ) ParamValue.xp_type.xp_int4;
                                 break;
      default                  : throw SAPDBConfig_ParameterAccessFailure( __CONTEXT__,
                                             SAPDBCONFIG_WRONG_PARAMETER_TYPE, 
                                             SAPDB_ToString(Idenfifier), 
                                             SAPDB_ToString("integer"),
                                             SAPDB_ToString("non integer"));
    }
  }
  else
  {
    throw SAPDBConfig_ParameterAccessFailure( __CONTEXT__,
                                              SAPDBCONFIG_GET_PARAMETER_FAILED, 
                                              SAPDB_ToString(Idenfifier), 
                                              SAPDB_ToString(RteError.RteErrText));
   }
  return( IntParameter ) ;
}

/*!*****************************************************************************
*******************************************************************************/
SAPDB_Real8 SAPDBConfig_ConfigurationParameters::GetRealParameter(const SAPDB_Char *Idenfifier) const
{
  if ( Idenfifier != 0 )
  {
    throw("Not Implemented");
  }
  return 0.;
}

#ifdef STLPORT_ALLOWED
/*!*****************************************************************************
*******************************************************************************/
SAPDB_Int4 SAPDBConfig_ConfigurationParameters::GetIntParameter(const SAPDBSTL::string Idenfifier) const 
{
  return GetIntParameter( Idenfifier.c_str() ) ;
}

/*!*****************************************************************************
*******************************************************************************/
SAPDB_Char *SAPDBConfig_ConfigurationParameters::GetStringParameter(const SAPDBSTL::string   Idenfifier ,
                                                                    SAPDBConfig_ParamValueAsString &ParamAsString) const
{
  return GetStringParameter( Idenfifier.c_str(), ParamAsString ) ;
}
#endif

/*!*****************************************************************************
*******************************************************************************/
SAPDB_Char* SAPDBConfig_ConfigurationParameters::GetStringParameter(const SAPDB_Char *Idenfifier,
                                                                    SAPDBConfig_ParamValueAsString &ParamAsString) const
{
  tsp11_XpKeyTypec ParamKey ;
  XP_VALUE_REC     ParamValue ;
  tsp01_RteError   RteError ;

  ParamKey.rawAssign(Idenfifier);

  if ( sqlXparamGet ( m_plXParamFd, ParamKey, reinterpret_cast<tsp00_XpValueRec *>(&ParamValue), &RteError ) == xp_ok_sp11 )
  {
    switch ( ParamValue.xp_value_t )
    {
      case xp_c8_type_sp11    :  eo46PtoC ( ParamAsString , (SAPDB_Byte *)ParamValue.xp_type.xp_c8 .asCharp(),  8 ) ;
                                 break;
      case xp_c18_type_sp11   :  eo46PtoC ( ParamAsString , (SAPDB_Byte *)ParamValue.xp_type.xp_c18.asCharp(), 18 ) ;
                                 break;
      case xp_c24_type_sp11   :  eo46PtoC ( ParamAsString , (SAPDB_Byte *)ParamValue.xp_type.xp_c24.asCharp(), 24 ) ;
                                 break;
      case xp_c40_type_sp11   :  eo46PtoC ( ParamAsString , (SAPDB_Byte *)ParamValue.xp_type.xp_c40.asCharp(), 40 ) ;
                                 break;
      case xp_c64_type_sp11   :  eo46PtoC ( ParamAsString , (SAPDB_Byte *)ParamValue.xp_type.xp_c64.asCharp(), 64 ) ;
                                 break;
      default                 :  throw SAPDBConfig_ParameterAccessFailure( __CONTEXT__,
                                             SAPDBCONFIG_WRONG_PARAMETER_TYPE, 
                                             SAPDB_ToString(Idenfifier), 
                                             SAPDB_ToString("string"),
                                             SAPDB_ToString("non string"));
    }
  }
  else
  {
    throw SAPDBConfig_ParameterAccessFailure( __CONTEXT__,
                                              SAPDBCONFIG_GET_PARAMETER_FAILED, 
                                              SAPDB_ToString(Idenfifier), 
                                              SAPDB_ToString(RteError.RteErrText));
  
  }
  return ( ParamAsString ) ;
}
/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
