/*!***************************************************************************

  module      : SAPDBConfig_Parameters.hpp

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: SAPDB configuration parameters
  description : class for kernel configurable parameter access


  last changed: 2000-06-19  10:20
  see also    : 
  first created:2000-03-22  17:49

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef SAPDBCONFIG_PARAMETERS_HPP
#define SAPDBCONFIG_PARAMETERS_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/Configuration/SAPDBConfig_Exception.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************
    typedef:      SAPDBConfig_ParamValueAsString

    description:  Type for configuration parameters "string" 
*******************************************************************************/

typedef SAPDB_Char SAPDBConfig_ParamValueAsString[128];

/*!*****************************************************************************
   
     class: SAPDBConfig_ParameterAccessFailure

*******************************************************************************/

class SAPDBConfig_ParameterAccessFailure : public SAPDBConfig_Exception
{
  public:
      SAPDBConfig_ParameterAccessFailure() : SAPDBConfig_Exception () {;}
      SAPDBConfig_ParameterAccessFailure( SAPDB_Char const * const   FileName,
                              SAPDB_UInt4 const          LineNumber,
                              MessageType const          Type,
                              MessageID const            ID,
                              SAPDB_Char const * const   Message,
                              SAPDB_UInt4 const          NumOfArgs,
                              SAPDB_Char const * const   Arg0 = NULL,
                              SAPDB_Char const * const   Arg1 = NULL,
                              SAPDB_Char const * const   Arg2 = NULL,
                              SAPDB_Char const * const   Arg3 = NULL,
                              SAPDB_Char const * const   Arg4 = NULL,
                              SAPDB_Char const * const   Arg5 = NULL,
                              SAPDB_Char const * const   Arg6 = NULL,
                              SAPDB_Char const * const   Arg7 = NULL,
                              SAPDB_Char const * const   Arg8 = NULL,
                              SAPDB_Char const * const   Arg9 = NULL ) : 
                                        SAPDBConfig_Exception ( FileName,
                                                               LineNumber, Type,
                                                               ID, Message, NumOfArgs,
                                                               Arg0, Arg1, Arg2,
                                                               Arg3, Arg4, Arg5, Arg6,
                                                               Arg7, Arg8,Arg9 ) {;} 
};

/*!*****************************************************************************
   
     endclass: SAPDBConfig_ParameterAccessFailure

*******************************************************************************/


/*!*****************************************************************************
   
     class: SAPDBConfig_ConfigurationParameters

*******************************************************************************/

class SAPDBConfig_ConfigurationParameters
{
public:


    /*!-------------------------------------------------------------------------
    function:     SAPDBConfig_ConfigurationParameters()
    description:  Constructor; 
                  creates a new SAPDBConfig_ConfigurationParameters object and
                  reads the parameter file containing the configuration

    arguments:    ParamFileName [in] reference to the filename containing the 
                  parameters 

                  If the read of the parameter file fails, an exception is thrown.
    return value: none
    exceptions:   SAPDBConfig_ParameterAccessFailure()
    --------------------------------------------------------------------------*/
    SAPDBConfig_ConfigurationParameters(const SAPDB_Char *ParamFileName) ;

    /*!-------------------------------------------------------------------------
    function:     ~SAPDBConfig_ConfigurationParameters()
    description:  Destructor; 
                  destroy SAPDBConfig_ConfigurationParameters object and
                  closes the parameterfile 

    return value: none
    --------------------------------------------------------------------------*/
   ~SAPDBConfig_ConfigurationParameters();

    /*!-------------------------------------------------------------------------
    function:     GetIntParameter()
    description:  get an integer parameter from the configuration file
                  given by the "Identifier"

    arguments:    Idenfifier [in] reference to the parameter wanted

    return value: value of the parameter given by the "Identifier"

    exceptions:   SAPDBConfig_ParameterAccessFailure if get of the parameter fails
    --------------------------------------------------------------------------*/
   SAPDB_Int4        GetIntParameter(const SAPDB_Char *Idenfifier) const ;

    /*!-------------------------------------------------------------------------
    function:     GetRealParameter()
    description:  get an real parameter from the configuration file
                  given by the "Identifier"

    arguments:    Idenfifier [in] reference to the parameter wanted

    return value: value of the parameter given by the "Identifier"

    exceptions:   SAPDBConfig_ParameterAccessFailure if get of the parameter fails
    --------------------------------------------------------------------------*/
   SAPDB_Real8       GetRealParameter(const SAPDB_Char *Idenfifier) const ;

    /*!-------------------------------------------------------------------------
    function:     GetStringParameter()
    description:  get an string parameter from the configuration file
                  given by the "Identifier"

    arguments:    Idenfifier [in] reference to the parameter wanted 

    return value: value of the parameter given by the "Identifier"

    exceptions:   SAPDBConfig_ParameterAccessFailure if get of the parameter fails
    --------------------------------------------------------------------------*/
   SAPDB_Char       *GetStringParameter(const SAPDB_Char *Idenfifier,
                                        SAPDBConfig_ParamValueAsString &ParamAsString ) const ;
 
private:
    SAPDB_Int4                      m_plXParamFd ;
    SAPDBConfig_ParamValueAsString  m_ParamFileName ;

};
/*!*****************************************************************************
   
     endclass: SAPDBConfig_ConfigurationParameters

*******************************************************************************/

/*
#ifdef STLPORT_ALLOWED
 \#include "SAPDBCommon/STL/SAPDBSTL_string.hpp"

   SAPDB_Int4        GetIntParameter(const SAPDBSTL::string Idenfifier) const ;
   SAPDB_Char       *GetStringParameter(const SAPDBSTL::string Idenfifier, 
                                        SAPDBConfig_ParamValueAsString &ParamAsString ) const ;

#endif
*/

#endif // SAPDBCONFIG_PARAMETERS_HPP


