/*!******************************************************************************

  module:		 SAPDBFields_BasicRAMBuffer.hpp

  ------------------------------------------------------------------------------

  responsible:  TillL

  special area: SAPDBCommon/Fields

  description:  Template class for memory buffers with random access

                - A buffer is a container which is based on a resource R
                  of elements of type R::BasisType.

  see also:     

  ------------------------------------------------------------------------------

                         Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


#ifndef SAPDBFIELDS_BASICRAMBUFFER_HPP
#define SAPDBFIELDS_BASICRAMBUFFER_HPP

#include "SAPDBCommon/Fields/SAPDBFields_Field.hpp"


/*!*****************************************************************************

   class: SAPDBFields_BasicRAMBuffer

   description:
   - This class is modelling a random access buffer for use in SAPDB.
   - The buffer is associated with a memory block.
   - The buffer allows random read/write operations to be done within 
     the limits of the associated memory block.
   - If the buffer is used with resources other than byte resources, 
     e.g. R = SAPDB_FieldResource<T>, with T = SAPDB_UInt2, SAPDB_UInt4, etc.,
     be careful when indexing positions within the buffer with IndexType:
     Like in a vector of T, IndexType  p h y s i c a l l y  does not index 
     byte positions, but sizeof(T)-positions!
   
*******************************************************************************/
template <class R>
class SAPDBFields_BasicRAMBuffer
{
public:
    /*!-------------------------------------------------------------------------
    declaration : ResourceType
    description:  Type of the buffer resource
    --------------------------------------------------------------------------*/
    typedef R                                       ResourceType;
    /*!-------------------------------------------------------------------------
    declaration : SizeType
    description:  The range for buffer lengths (in sizeof(BasisType))
    --------------------------------------------------------------------------*/
    typedef typename R::SizeType                    SizeType;
    /*!-------------------------------------------------------------------------
    declaration : IndexType
    description:  The range for positions within buffers
    --------------------------------------------------------------------------*/
    typedef typename R::IndexType                   IndexType;
    /*!-------------------------------------------------------------------------
    declaration : BasisType
    description:  Type of data contained in buffer
    --------------------------------------------------------------------------*/
    typedef typename R::BasisType                   BasisType;      
    /*!-------------------------------------------------------------------------
    declaration : Pointer
    description:  Pointer to buffer data
    --------------------------------------------------------------------------*/
    typedef typename R::Pointer                     Pointer;
    /*!-------------------------------------------------------------------------
    declaration : Iterator
    description:  Buffer iterator
    --------------------------------------------------------------------------*/
    typedef Pointer                                 Iterator;
    /*!-------------------------------------------------------------------------
    declaration : ConstPointer
    description:  Pointer to constant buffer data
    --------------------------------------------------------------------------*/
    typedef typename R::ConstPointer                ConstPointer;
    /*!-------------------------------------------------------------------------
    declaration : ConstIterator
    description:  Buffer iterator to constant data
    --------------------------------------------------------------------------*/
    typedef ConstPointer                            ConstIterator;
    /*!-------------------------------------------------------------------------
    declaration : AllocatorType
    description:  Accepted allocator type
    --------------------------------------------------------------------------*/
    typedef typename R::AllocatorType               AllocatorType;
    /*!-------------------------------------------------------------------------
    declaration : BasisFieldType
    description:  Field handle type for fields of BasisType
    --------------------------------------------------------------------------*/
    typedef SAPDBFields_BasicField<BasisType>       BasisFieldType;
    /*!-------------------------------------------------------------------------
    declaration : CommonFieldType
    description:  Field handle type for common SAPDB fields
    --------------------------------------------------------------------------*/
    typedef SAPDBFields_Field                       CommonFieldType;


    /*!-------------------------------------------------------------------------
    function:     SAPDBFields_BasicRAMBuffer()
    description:  Constructors
    
                  - Allocator based buffers:
                      - 'alloc' allows to specify a dedicated allocator for the buffer. 
                      - If 'alloc' is not specified, the constructed string uses 
                        AllocatorType::Instance().

                  - Field-based buffers:
                      - 'src' and 'len' specify the source for the buffer.

    arguments:    src   [in]  pointer to memory block
                  len   [in]  length of memory block
                  alloc [in]  resource allocator
    return value: none
    prototypes:   3
    --------------------------------------------------------------------------*/
    SAPDBFields_BasicRAMBuffer
        ();

    SAPDBFields_BasicRAMBuffer
        (void*                          src, 
         SizeType                       len);

    SAPDBFields_BasicRAMBuffer
        (AllocatorType&                 alloc);

    /*!-------------------------------------------------------------------------
    function:     Assign()
    description:  Assigns memory block 
    
                  - Field-based buffers only!
                  - Assigns a buffer pointing to memory block at 'src'.
                  - 'src' must be sizeof(BasisType)-aligned.
                  - 'len' gives the buffer size in sizeof(BasisType).
    arguments:    src [in]   pointer to memory block
                  len [in]   length of memory block
    return value: none
    --------------------------------------------------------------------------*/
    void Assign
        (void*                          src, 
         SizeType                       len);

    /*!-------------------------------------------------------------------------
    function:     Deassign()
    description:  Deassigns the buffer resource

                  - Has  n o  effect on  s t a t i c  buffer resources.
                  - After this operation the buffer is unassigned.
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Deassign();

    /*!-------------------------------------------------------------------------
    function:     ToBasisField()
    description:  Constructs and returns a field resource handle 
    
                  - Field comprises the actually filled portion of the buffer.
                  - Contents is not copied.
    arguments:    none
    return value: (BasisFieldType) field handle of BasisType
    --------------------------------------------------------------------------*/
    BasisFieldType ToBasisField() const;

    /*!-------------------------------------------------------------------------
    function:     ToField()
    description:  Constructs and returns a field resource handle 
    
                  - Field comprises the actually filled portion of the buffer.
                  - Contents is not copied.
    arguments:    none
    return value: (CommonFieldType) field handle of common SAPDB type
    --------------------------------------------------------------------------*/
    CommonFieldType ToField() const;

    /*!-------------------------------------------------------------------------
    function:     GetByteAt()
    description:  Returns the byte at a given offset within the buffer
    arguments:    offs [in]    position within the buffer
    return value: (SAPDB_Byte)
    --------------------------------------------------------------------------*/
    SAPDB_Byte GetByteAt
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function:     GetCharAt()
    description:  Returns the char at a given offset within the buffer
    arguments:    offs [in]    position within the buffer
    return value: (SAPDB_Char)
    --------------------------------------------------------------------------*/
    SAPDB_Char GetCharAt
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function:     GetUInt2At()
    description:  Returns the uint2 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 2-byte aligned)
    return value: (SAPDB_UInt2)
    --------------------------------------------------------------------------*/
    SAPDB_UInt2 GetUInt2At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetInt2At()
    description:  Returns the int2 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 2-byte aligned)
    return value: (SAPDB_Int2)
    --------------------------------------------------------------------------*/
    SAPDB_Int2 GetInt2At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetUInt4At()
    description:  Returns the uint4 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 4-byte aligned)
    return value: (SAPDB_UInt4)
    --------------------------------------------------------------------------*/
    SAPDB_UInt4 GetUInt4At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetInt4At()
    description:  Returns the int4 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 4-byte aligned)
    return value: (SAPDB_Int4)
    --------------------------------------------------------------------------*/
    SAPDB_Int4 GetInt4At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetUInt8At()
    description:  Returns the uint8 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 8-byte aligned)
    return value: (SAPDB_UInt8)
    --------------------------------------------------------------------------*/
    SAPDB_UInt8 GetUInt8At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetInt8At()
    description:  Returns the int8 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 8-byte aligned)
    return value: (SAPDB_Int8)
    --------------------------------------------------------------------------*/
    SAPDB_Int8 GetInt8At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetReal4At()
    description:  Returns the real4 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 4-byte aligned)
    return value: (SAPDB_Real4)
    --------------------------------------------------------------------------*/
    SAPDB_Real4 GetReal4At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: GetReal8At()
    description:  Returns the real8 at a given offset within the buffer
    arguments:    offs [in]    position within the buffer (must be 8-byte aligned)
    return value: (SAPDB_Real8)
    --------------------------------------------------------------------------*/
    SAPDB_Real8 GetReal8At
        (IndexType                      offs) const;

    /*!-------------------------------------------------------------------------
    function: SetByteAt()
    description:  Writes a byte at a given offset into the buffer
    arguments:    offs [in]     position within the buffer
                  byte [in]     byte value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetByteAt
        (IndexType                      offs, 
         SAPDB_Byte                     byte);

    /*!-------------------------------------------------------------------------
    function: SetCharAt()
    description:  Writes a char at a given offset into the buffer
    arguments:    offs [in]     position within the buffer
                  ch   [in]     char value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetCharAt
        (IndexType                      offs, 
         SAPDB_Char                     ch);

    /*!-------------------------------------------------------------------------
    function: SetUInt2At()
    description:  Writes an uint2 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 2-byte aligned)
                  val  [in]     uint2 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetUInt2At
        (IndexType                      offs, 
         SAPDB_UInt2                    val);

    /*!-------------------------------------------------------------------------
    function: SetInt2At()
    description:  Writes an int2 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 2-byte aligned)
                  val  [in]     int2 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetInt2At
        (IndexType                      offs, 
         SAPDB_Int2                     val);

    /*!-------------------------------------------------------------------------
    function: SetUInt4At()
    description:  Writes an uint4 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 4-byte aligned)
                  val  [in]     uint4 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetUInt4At
        (IndexType                      offs, 
         SAPDB_UInt4                    val);

    /*!-------------------------------------------------------------------------
    function: SetInt4At()
    description:  Writes an int4 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 4-byte aligned)
                  val  [in]     int4 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetInt4At
        (IndexType                      offs, 
         SAPDB_Int4                     val);

    /*!-------------------------------------------------------------------------
    function: SetUInt8At()
    description:  Writes an uint8 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 8-byte aligned)
                  val  [in]     uint8 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetUInt8At
        (IndexType                      offs, 
         SAPDB_UInt8                    val);

    /*!-------------------------------------------------------------------------
    function: SetInt8At()
    description:  Writes an int8 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 8-byte aligned)
                  val  [in]     int8 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetInt8At
        (IndexType                      offs, 
         SAPDB_Int8                     val);

    /*!-------------------------------------------------------------------------
    function: SetReal4At()
    description:  Writes an real4 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 4-byte aligned)
                  val  [in]     real4 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetReal4At
        (IndexType                      offs, 
         SAPDB_Real4                    val);

    /*!-------------------------------------------------------------------------
    function: SetReal8At()
    description:  Writes an real8 at a given offset into the buffer
    arguments:    offs [in]     position within the buffer (must be 8-byte aligned)
                  val  [in]     real8 value to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetReal8At
        (IndexType                      offs, 
         SAPDB_Real8                    val);

    /*!-------------------------------------------------------------------------
    function: SetFieldAt()
    description:  Writes the contents of a field into the buffer
    arguments:    offs  [in]   position within the buffer (must be 8-byte aligned)
                  field [in]   field to be written
    return value: none
    --------------------------------------------------------------------------*/
    void SetFieldAt
        (IndexType                      offs, 
         const CommonFieldType  &       field);

    /*!-------------------------------------------------------------------------
    function: operator[]
    description:  Returns a reference to an element at a given offset
                  
                  - It is checked in an assertion, whether 'offs' is located 
                    within the buffer.
    arguments:    offs      [in] position in sizeof(BasisType) within the buffer
    return value: none
    prototypes:   2
    --------------------------------------------------------------------------*/
    const BasisType& operator[]
        (IndexType                      offs) const;

    BasisType& operator[]
        (IndexType                      offs);

    /*!-------------------------------------------------------------------------
    function: GetPtrAt()
    description:  Returns a pointer to an element at a given offset
                  
                  - It is checked in an assertion, whether 'offs' and 'offs' + 'len' 
                    are located within the buffer.
    arguments:    offs      [in] position in sizeof(BasisType) within the buffer
                  len       [in] size in sizeof(BasisType) of the object that is 
                                 to be pointed to
    return value: none
    prototypes:   2
    --------------------------------------------------------------------------*/
    ConstPointer GetPtrAt
        (IndexType                      offs, 
         SizeType                       len) const;
   
    Pointer GetPtrAt
        (IndexType                      offs, 
         SizeType                       len);
   
    /*!-------------------------------------------------------------------------
    function:     Capacity()
    description:  Returns capacity of buffer in elements of BasisType

                  - For dynamic buffer resources, this returns the actual count
                    of elements that can be stored within the buffer without
                    reallocation.
                  - For static or field buffer resources the actual (fixed) size
                    of the buffer is returned.
    arguments:    none
    return value: (SizeType) capacity of buffer in elements of BasisType
    --------------------------------------------------------------------------*/
    SizeType Capacity() const;

    /*!-------------------------------------------------------------------------
    function:     IsAssigned()
    description:  Checks whether buffer is assigned a resource
    arguments:    none
    return value: (bool) true if assigned
    --------------------------------------------------------------------------*/
    bool IsAssigned() const;

    /*--------------------------------------------------------------------------
    function:     ProvideCapacity
    description:  Reserves resource space for elements

                  - Tries to reserve resource space for at least
                    the requested amount of elements.
                  - All elements are preserved on reallocation.
    arguments:    requested [in]  requested resource size in elements
    return value: (SAPDB_Bool) true if enough space available
    // exceptions:   SAPDBFields_ResourceExhausted
    --------------------------------------------------------------------------*/
    SAPDB_Bool ProvideCapacity
        (SizeType                       requested);

protected:

    /*--------------------------------------------------------------------------
    function:     SpaceRequired()
    description:  Returns the number of BasisType elements covered 
                  (fully or partially) by an aligned object of given size

                  Example: sizeof(BasisType) = 4
                           objsize  1 2 3 4 5 6 7 8 9
                           returns  1 1 1 1 2 2 2 2 3
    arguments:    objsize  [in]     size of object in bytes
    return value: (SAPDB_Int) number of basis elements
    --------------------------------------------------------------------------*/
    static SAPDB_Int SpaceRequired
        (SAPDB_Int                      objsize);

    /*!-------------------------------------------------------------------------
    function:     IsAligned()
    description:  Checks whether object of given size is aligned if placed at
                  given offset within buffer

                  - Precondition: offs 0 is aligned
    arguments:    objsize  [in]     size of object in bytes
    return value: (bool) true if assigned
    --------------------------------------------------------------------------*/
    static bool IsAligned
        (IndexType                      offs, 
         SAPDB_Int                      objsize);

private:
    /*--------------------------------------------------------------------------
    declaration : m_Resource
    description:  Memory resource of the buffer
    --------------------------------------------------------------------------*/
    ResourceType m_Resource;

};
/******************************************************************************/
/*                                                                            */
/*! endclass: SAPDBFields_BasicRAMBuffer                                      */
/*                                                                            */
/******************************************************************************/



template <class R>
inline SAPDBFields_BasicRAMBuffer<R>::SAPDBFields_BasicRAMBuffer()
    {}

//------------------------------------------------------------------------------

template <class R>
inline SAPDBFields_BasicRAMBuffer<R>::SAPDBFields_BasicRAMBuffer
    (void*                          src, 
     SizeType                       len)
    : 
    m_Resource(src, len)
{}

//------------------------------------------------------------------------------

template <class R>
inline SAPDBFields_BasicRAMBuffer<R>::SAPDBFields_BasicRAMBuffer
    (AllocatorType&                 alloc)
    : 
    m_Resource(alloc) 
{}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::Assign
    (void*                          src, 
     SizeType                       len) 
{
    m_Resource.Assign(src, len);
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::Deassign
    ()
{
    m_Resource.Deassign();
}

//------------------------------------------------------------------------------

template <class R>
inline typename SAPDBFields_BasicRAMBuffer<R>::BasisFieldType 
SAPDBFields_BasicRAMBuffer<R>::ToBasisField
    () const
{
    return BasisFieldType(m_Resource.Data(), m_Resource.Capacity());
}

//------------------------------------------------------------------------------

template <class R>
inline typename SAPDBFields_BasicRAMBuffer<R>::CommonFieldType 
SAPDBFields_BasicRAMBuffer<R>::ToField
    () const
{
    return CommonFieldType
            (m_Resource.Data(), 
             m_Resource.Capacity() * (sizeof(BasisType)/sizeof(CommonFieldType::ValueType))
            );
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Byte SAPDBFields_BasicRAMBuffer<R>::GetByteAt
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE(IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Byte)));
    return *(reinterpret_cast<const SAPDB_Byte*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Char SAPDBFields_BasicRAMBuffer<R>::GetCharAt
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE(IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Char)));
    return *(reinterpret_cast<const SAPDB_Char*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_UInt2 SAPDBFields_BasicRAMBuffer<R>::GetUInt2At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_UInt2)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_UInt2)));
    return *(reinterpret_cast<const SAPDB_UInt2*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Int2 SAPDBFields_BasicRAMBuffer<R>::GetInt2At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Int2)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Int2)));
    return *(reinterpret_cast<const SAPDB_Int2*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_UInt4 SAPDBFields_BasicRAMBuffer<R>::GetUInt4At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_UInt4)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_UInt4)));
    return *(reinterpret_cast<const SAPDB_UInt4*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Int4 SAPDBFields_BasicRAMBuffer<R>::GetInt4At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Int4)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Int4)));
    return *(reinterpret_cast<const SAPDB_Int4*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_UInt8 SAPDBFields_BasicRAMBuffer<R>::GetUInt8At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_UInt8)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_UInt8)));
    return *(reinterpret_cast<const SAPDB_UInt8*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Int8 SAPDBFields_BasicRAMBuffer<R>::GetInt8At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Int8)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Int8)));
    return *(reinterpret_cast<const SAPDB_Int8*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Real4 SAPDBFields_BasicRAMBuffer<R>::GetReal4At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Real4)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Real4)));
    return *(reinterpret_cast<const SAPDB_Real4*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Real8 SAPDBFields_BasicRAMBuffer<R>::GetReal8At
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Real8)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Real8)));
    return *(reinterpret_cast<const SAPDB_Real8*>(m_Resource.PtrTo(offs)));
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetByteAt
    (IndexType                      offs, 
     SAPDB_Byte                     byte)
{
    SAPDBERR_ASSERT_STATE(IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Byte)));
    *reinterpret_cast<SAPDB_Byte*>(m_Resource.PtrTo(offs)) = byte;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetCharAt
    (IndexType                      offs, 
     SAPDB_Char                     ch)
{
    SAPDBERR_ASSERT_STATE(IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Char)));
    *reinterpret_cast<SAPDB_Char*>(m_Resource.PtrTo(offs)) = ch;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetUInt2At
    (IndexType                      offs, 
     SAPDB_UInt2                    val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_UInt2)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_UInt2)));
    *reinterpret_cast<SAPDB_UInt2*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetInt2At
    (IndexType                      offs, 
     SAPDB_Int2                     val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Int2)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Int2)));
    *reinterpret_cast<SAPDB_Int2*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetUInt4At
    (IndexType                      offs, 
     SAPDB_UInt4                    val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_UInt4)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_UInt4)));
    *reinterpret_cast<SAPDB_UInt4*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetInt4At
    (IndexType                      offs, 
     SAPDB_Int4                     val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Int4)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Int4)));
    *reinterpret_cast<SAPDB_Int4*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetUInt8At
    (IndexType                      offs, 
     SAPDB_UInt8                    val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_UInt8)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_UInt8)));
    *reinterpret_cast<SAPDB_UInt8*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetInt8At
    (IndexType                      offs, 
     SAPDB_Int8                     val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Int8)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Int8)));
    *reinterpret_cast<SAPDB_Int8*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetReal4At
    (IndexType                      offs, 
     SAPDB_Real4                    val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Real4)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Real4)));
    *reinterpret_cast<SAPDB_Real4*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetReal8At
    (IndexType                      offs, 
     SAPDB_Real8                    val)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(IsAligned(offs, sizeof(SAPDB_Real8)));
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(sizeof(SAPDB_Real8)));
    *reinterpret_cast<SAPDB_Real8*>(m_Resource.PtrTo(offs)) = val;
}

//------------------------------------------------------------------------------

template <class R>
inline void SAPDBFields_BasicRAMBuffer<R>::SetFieldAt
    (IndexType                      offs, 
     const CommonFieldType&         field)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - field.GetByteLength());
    memmove(m_Resource.PtrTo(offs), field.GetDataPtr(), field.GetByteLength());
}

//------------------------------------------------------------------------------

template <class R>
inline const typename SAPDBFields_BasicRAMBuffer<R>::BasisType& 
SAPDBFields_BasicRAMBuffer<R>::operator[]
    (IndexType                      offs) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity());
    return *m_Resource.PtrTo(offs);
}

//------------------------------------------------------------------------------

template <class R>
inline typename SAPDBFields_BasicRAMBuffer<R>::BasisType& 
SAPDBFields_BasicRAMBuffer<R>::operator[]
    (IndexType                      offs)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity());
    return *m_Resource.PtrTo(offs);
}

//------------------------------------------------------------------------------

template <class R>
inline typename SAPDBFields_BasicRAMBuffer<R>::ConstPointer 
SAPDBFields_BasicRAMBuffer<R>::GetPtrAt
    (IndexType                      offs, 
     SizeType                       len) const
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(len));
    return m_Resource.PtrTo(offs);
}

//------------------------------------------------------------------------------

template <class R>
inline typename SAPDBFields_BasicRAMBuffer<R>::Pointer 
SAPDBFields_BasicRAMBuffer<R>::GetPtrAt
    (IndexType                      offs, 
     SizeType                       len)
{
    SAPDBERR_ASSERT_STATE   (IsAssigned());
    SAPDBERR_ASSERT_ARGUMENT(offs <= Capacity() - SpaceRequired(len));
    return m_Resource.PtrTo(offs);
}

//------------------------------------------------------------------------------

template <class R>
inline typename SAPDBFields_BasicRAMBuffer<R>::SizeType 
       SAPDBFields_BasicRAMBuffer<R>::Capacity
    () const
{
    return m_Resource.Capacity();
}

//------------------------------------------------------------------------------

template <class R>
inline bool SAPDBFields_BasicRAMBuffer<R>::IsAssigned() const
{
    return m_Resource.IsAssigned();
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Bool SAPDBFields_BasicRAMBuffer<R>::ProvideCapacity
    (SizeType                       requested) 
{
    return m_Resource.ProvideCapacity(requested, Capacity());
}

//------------------------------------------------------------------------------

template <class R>
inline SAPDB_Int SAPDBFields_BasicRAMBuffer<R>::SpaceRequired
    (SAPDB_Int                      objsize)
{
    return objsize <= sizeof(BasisType) ? 
        1 : (objsize + sizeof(BasisType) - 1) / sizeof(BasisType);
}

//------------------------------------------------------------------------------

template <class R>
inline bool SAPDBFields_BasicRAMBuffer<R>::IsAligned
    (IndexType                      offs, 
     SAPDB_Int                      objsize)
{
    return sizeof(BasisType)*offs % objsize == 0;
}

//------------------------------------------------------------------------------


#endif // SAPDBFIELDS_BASICRAMBUFFER_HPP

