/*!******************************************************************************

  module:		 SAPDBFields_Field.hpp

  ------------------------------------------------------------------------------

  responsible:  StefanB

  special area: Field class

  description:  class declarations for Field

  ------------------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


#ifndef SAPDBFIELDS_FIELD_HPP
#define SAPDBFIELDS_FIELD_HPP

// Include for memmove
#include <string.h>

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"


/*!******************************************************************************
   class: SAPDBFields_Field

   description: A handler to a contiguous memory block of certain size.

                The memory block is fixed in content and size, that means 
                SAPDBFields_Field has only methods for reading the content and no 
                methods for changing the content.
********************************************************************************/

template <class T>
class SAPDBFields_BasicField
{
public:
   /*!-------------------------------------------------------------------------
   declaration : SizeType
   description:  The range for field lengths
   --------------------------------------------------------------------------*/
   typedef SAPDB_UInt                  SizeType;
   /*!-------------------------------------------------------------------------
   declaration : IndexType
   description:  The range for positions within fields
   --------------------------------------------------------------------------*/
   typedef SizeType                    IndexType;
   /*!-------------------------------------------------------------------------
   declaration : ValueType
   description:  Type of data contained in field
   --------------------------------------------------------------------------*/
   typedef T                           ValueType;      

protected:
   /*!-------------------------------------------------------------------------
   declaration : ConstPointer
   description:  Pointer to constant field data
   --------------------------------------------------------------------------*/
   typedef const ValueType*            ConstPointer;
   /*!-------------------------------------------------------------------------
   declaration : ConstIterator
   description:  Field iterator referencing constant data
   --------------------------------------------------------------------------*/
   typedef ConstPointer                ConstIterator;

  /*! enddeclaration: */

private:
   /*--------------------------------------------------------------------------
   declaration : m_DataPtr
   description:  Pointer to memory block
   --------------------------------------------------------------------------*/
   ConstPointer   m_DataPtr;
   /*--------------------------------------------------------------------------
   declaration : m_DataLen
   description:  Count of elements of ValueType in memory block
   --------------------------------------------------------------------------*/
   SizeType       m_DataLen;

public:
   /*!-------------------------------------------------------------------------
   function: Default Constructor
   description:  Constructs an empty field
   arguments:    none
   return value: none
   --------------------------------------------------------------------------*/
   SAPDBFields_BasicField() : m_DataPtr(0), m_DataLen(0) {}

   /*!-------------------------------------------------------------------------
   function: Constructor
   description:  Constructs a field pointing to Src with length Len.
   arguments:    Src [in]
                 The address of the memory block
                 Len [in]
                 The length of the memory block in sizeof(ValueType)
   return value: none
   --------------------------------------------------------------------------*/
   SAPDBFields_BasicField(const void *Src, SizeType Len) 
      : m_DataPtr(reinterpret_cast<ConstPointer>(Src)), m_DataLen(Len) {}

   /*!-------------------------------------------------------------------------
   function: Assign
   description:  Assigns values for the memory block pointer and length either
                 from parameters or from another field.
   arguments:    Src [in]
                 Pointer to memory block
                 Len [in]
                 Length of memory block
                 Field [in]
                 Another field
   return value: none
   prototypes:   2
   --------------------------------------------------------------------------*/
   void Assign(const void *Src, SizeType Len)
   {
       m_DataPtr = reinterpret_cast<ConstPointer>(Src);
       m_DataLen = Len;
   }

   void Assign(const SAPDBFields_BasicField<T>& Field)
   {
       m_DataPtr = Field.m_DataPtr;
       m_DataLen = Field.m_DataLen;
   }

   /*!-------------------------------------------------------------------------
   function: Deassign
   description:  Deassigns a Field
   arguments:    none
   return value: none
   --------------------------------------------------------------------------*/
   void Deassign()
      { m_DataPtr = 0; m_DataLen = 0; }

   /*!-------------------------------------------------------------------------
   function: AssignEmpty
   description:  Assigns an empty field.
                 
                 The field is simply assigned to the this-pointer with length
                 zero.
   arguments:    none
   return value: none
   --------------------------------------------------------------------------*/
   void AssignEmpty() 
      { Assign(this, 0); }

   /*!-------------------------------------------------------------------------
   function: IsAssigned
   description:  Indicates, if the instance is already assigned to a memory 
                 block.
   arguments:    none
   return value: bool
   --------------------------------------------------------------------------*/
   bool IsAssigned() const
      { return (m_DataPtr != 0); }

   /*!-------------------------------------------------------------------------
   function: IsEmpty
   description:  Indicates, if the instance points to a (valid) memory block 
                 with a length of zero.
   arguments:    none
   return value: bool
   --------------------------------------------------------------------------*/
   bool IsEmpty() const
   { 
      SAPDBERR_ASSERT_STATE(IsAssigned());
      return (m_DataLen == 0);
   }

   /*!-------------------------------------------------------------------------
   function: GetLength
   description:  Returns the length of the memory block in sizeof(ValueType)
   arguments:    none
   return value: SizeType
   --------------------------------------------------------------------------*/
   SizeType GetLength() const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      return m_DataLen;
   }

   /*!-------------------------------------------------------------------------
   function: GetByteLength
   description:  Returns the length of the memory block in bytes
   arguments:    none
   return value: SAPDB_UInt
   --------------------------------------------------------------------------*/
   SAPDB_UInt GetByteLength() const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      return m_DataLen * sizeof(ValueType);
   }

   /*!-------------------------------------------------------------------------
   function: Peek
   description:  Returns the byte at a certain position within the memory block
   arguments:    pos [in]
                 The position within the memory block
   return value: const ValueType
   --------------------------------------------------------------------------*/
   const ValueType Peek(IndexType pos) const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      SAPDBERR_ASSERT_ARGUMENT(pos < m_DataLen);         
      return *(m_DataPtr + pos);
   }

   /*--------------------------------------------------------------------------
   function: Compare
   description:  Compares two fields and returns on of the results
                 Smaller, Greater or Equal

                 !!! Not yet implemented
   arguments:    Field [in]
                 The field to compare with
   return value: SAPDB_CompareType
   --------------------------------------------------------------------------*/
   //SAPDB_CompareType Compare(const SAPDBFields_BasicField<T>& Field) const;

   /*!-------------------------------------------------------------------------
   function: IsEqualTo
   description:  Compares two fields.
   arguments:    Field [in]
                 The field to compare with
   return value: bool
   --------------------------------------------------------------------------*/
   bool IsEqualTo(const SAPDBFields_BasicField<T>& Field) const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      SAPDBERR_ASSERT_STATE(Field.IsAssigned());
      return ( (m_DataLen == Field.m_DataLen) && 
               (memcmp(m_DataPtr, Field.m_DataPtr, GetByteLength())==0) );
   }

   /*!-------------------------------------------------------------------------
   function: IsSmallerThan
   description:  Compares two fields.
   arguments:    Field [in]:
                 The field to compare with
   return value: bool
   --------------------------------------------------------------------------*/
   bool IsSmallerThan(const SAPDBFields_BasicField<T>& Field) const;

   /*!-------------------------------------------------------------------------
   operator: ==
   description:  Checks, if two fields point to the same memory block
   arguments:    Field [in]
                 The instance for the comparison
   return value: bool
   --------------------------------------------------------------------------*/    
   bool operator ==(const SAPDBFields_BasicField<T>& Field) const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      SAPDBERR_ASSERT_STATE(Field.IsAssigned());
      return (m_DataPtr == Field.m_DataPtr) && (m_DataLen == Field.m_DataLen);
   }

   /*!-------------------------------------------------------------------------
   operator: !=
   description:  Checks the opposite of ==
   arguments:    Field [in]
                 The instance for the comparison
   return value: bool
   --------------------------------------------------------------------------*/    
   bool operator !=(const SAPDBFields_BasicField<T>& Field) const
   {
      return !operator==(Field);
   }

   /*!-------------------------------------------------------------------------
   function: Begin
   description:  Returns an iterator which points to the begin of the memory block
   arguments:    none
   return value: ConstIterator
   --------------------------------------------------------------------------*/
   ConstIterator Begin() const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      return m_DataPtr;
   }

   /*!-------------------------------------------------------------------------
   function: End
   description:  Returns an iterator which points behind the end of the memory
                 block.
   arguments:    none
   return value: ConstIterator
   --------------------------------------------------------------------------*/
   ConstIterator End() const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      return m_DataPtr + m_DataLen;
   }

   /*!-------------------------------------------------------------------------
   function: GetDataPtr
   description:  Returns the address of the memory block
   arguments:    none
   return value: ConstPointer
   --------------------------------------------------------------------------*/
   ConstPointer GetDataPtr() const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      return m_DataPtr;
   }

   /*!-------------------------------------------------------------------------
   function: GetPointer
   description:  Returns the address of a certain position within the memory 
                 block

                 The field must be not empty.
   arguments:    pos [in]
                 The position within the memory block
   return value: ConstPointer
   --------------------------------------------------------------------------*/
   ConstPointer GetPointer(IndexType pos) const
   {
      SAPDBERR_ASSERT_STATE(IsAssigned());
      SAPDBERR_ASSERT_ARGUMENT(pos < m_DataLen);
      return m_DataPtr + pos;
   }

};

/*! endclass: */

//------------------------------------------------------------------------------
// Definition of methods of class SAPDBFields_BasicField<T>
//------------------------------------------------------------------------------

template <class T>
bool SAPDBFields_BasicField<T>::IsSmallerThan
    (const SAPDBFields_BasicField<T>& Field) const
{
   SAPDBERR_ASSERT_STATE(IsAssigned());
   
   SizeType Len = (Field.m_DataLen < m_DataLen) ? Field.m_DataLen : m_DataLen;

   int result = memcmp(m_DataPtr, Field.m_DataPtr, Len * sizeof(ValueType));

   return ( (result < 0) || ( (result == 0) && (m_DataLen < Field.m_DataLen) ) );
}

/*!-------------------------------------------------------------------------
declaration : SAPDBFields_Field
--------------------------------------------------------------------------*/

typedef SAPDBFields_BasicField<SAPDB_Byte> SAPDBFields_Field;

/*! enddeclaration: */




#endif // SAPDBFIELDS_FIELD_HPP

