/*!******************************************************************************

  module:		 SAPDBFields_FieldResource.hpp

  ------------------------------------------------------------------------------

  responsible:  TillL

  special area: SAPDBCommon/Fields

  description:  Field based resource of elements of T

  see also:     

  ------------------------------------------------------------------------------

                         Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/


#ifndef SAPDBFIELDS_FIELDRESOURCE_HPP
#define SAPDBFIELDS_FIELDRESOURCE_HPP

#include "SAPDBCommon/Fields/SAPDBFields_IResource.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"


/*!*****************************************************************************

    class: SAPDBFields_FieldResource

*******************************************************************************/
template <class T>
class SAPDBFields_FieldResource : public SAPDBFields_IResource<T>
{
public:
    /*!-------------------------------------------------------------------------
    declaration : P
    description:  Parent class
    --------------------------------------------------------------------------*/
    typedef SAPDBFields_IResource<T>            P;

    typedef typename P::SizeType                SizeType;
    typedef typename P::IndexType               IndexType;
    typedef typename P::BasisType               BasisType;      
    typedef typename P::Reference               Reference;      
    typedef typename P::ConstReference          ConstReference;
    typedef typename P::Pointer                 Pointer;
    typedef typename P::ConstPointer            ConstPointer;

    /*!-------------------------------------------------------------------------
    function:     SAPDBFields_FieldResource()
    description:  Default Constructor
    
                  - Constructs an empty resource
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    SAPDBFields_FieldResource();

    /*!-------------------------------------------------------------------------
    function:     SAPDBFields_FieldResource()
    description:  Constructor 
                    
                  - Constructs a resource pointing to memory block at 'src'.
                  - 'src' must be sizeof(BasisType)-aligned.
                  - 'len' is the resource capacity in sizeof(BasisType)
    arguments:    src  [in]   pointer to memory block
                  len  [in]   capacity of memory block in elements of BasisType
    return value: none
    --------------------------------------------------------------------------*/
    SAPDBFields_FieldResource
        (void*                          src, 
         SizeType                       len);

    /*!-------------------------------------------------------------------------
    function:     Assign()
    description:  Assigns resource
    
                  - Assigns a resource pointing to memory block at 'src'.
                  - 'src' must be sizeof(BasisType)-aligned.
                  - 'len' is the resource capacity in sizeof(BasisType)
    arguments:    src  [in]   pointer to memory block
                  len  [in]   capacity of memory block in elements of BasisType
    return value: none
    --------------------------------------------------------------------------*/
    void Assign
        (void*                          src, 
         SizeType                       len);

    /*!-------------------------------------------------------------------------
    function:     Deassign()
    description:  Deassigns the resource
    arguments:    none
    return value: none
    --------------------------------------------------------------------------*/
    void Deassign();

    /*!-------------------------------------------------------------------------
    function:     Capacity()
    description:  Returns the capacity of the resource

                  - The capacity gives the maximum count of elements that can be 
                    stored in the actual memory block associated with the resource.
    arguments:    none
    return value: (SizeType) capacity in elements of BasisType
   --------------------------------------------------------------------------*/
    SizeType Capacity() const;

    /*!-------------------------------------------------------------------------
    function:     MaxCapacity()
    description:  Returns the maximum count of elements that can be stored 
                  in the resource

                  - For field resources, this is equal to capacity
    arguments:    none
    return value: (SizeType) maximum capacity of the resource
   --------------------------------------------------------------------------*/
    SizeType MaxCapacity() const;

    /*!-------------------------------------------------------------------------
    function:     IsAssigned()
    description:  Checks whether resource is assigned
    arguments:    none
    return value: (bool) true if assigned
    --------------------------------------------------------------------------*/
    bool IsAssigned() const;

    /*--------------------------------------------------------------------------
    function:     Data
    description:  Returns the address of the associated memory block
    arguments:    none
    return value: (Pointer/ConstPointer) pointer to memory block
    prototypes:   2
    --------------------------------------------------------------------------*/
    Pointer Data()
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return m_Data;
    }

    ConstPointer Data() const
    {
        SAPDBERR_ASSERT_STATE(IsAssigned());
        return m_Data;
    }

    /*--------------------------------------------------------------------------
    function:     PtrTo
    description:  Returns pointer to element

                  - Returns a pointer to a specified element within the 
                    associated memory block.
                  - 'pos' may range from 0 to Capacity(). 
                    'pos' == Capacity() enables use of iterator End().
    arguments:    pos  [in]    element index [0..Capacity()]
    return value: (Pointer/ConstPointer) pointer to element
    prototypes:   2
    --------------------------------------------------------------------------*/
    Pointer PtrTo
        (IndexType                      pos)
    {
        SAPDBERR_ASSERT_STATE   (IsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(pos <= m_Capacity);

        return m_Data + pos;
    }

    ConstPointer PtrTo
        (IndexType                      pos) const
    {
        SAPDBERR_ASSERT_STATE   (IsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(pos <= m_Capacity);

        return m_Data + pos;
    }


    /*--------------------------------------------------------------------------
    function:     Element
    description:  Returns reference to element

                  - Returns a reference to a specified element within the 
                    associated memory block.
    arguments:    pos  [in]    element index [0..Capacity()-1]
    return value: (Reference/ConstReference) reference to element
    --------------------------------------------------------------------------*/
    Reference Element
        (IndexType                      pos)
    {
        SAPDBERR_ASSERT_STATE   (IsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(pos < m_Capacity);

        return *(m_Data + pos);
    }

    ConstReference Element
        (IndexType                      pos) const
    {
        SAPDBERR_ASSERT_STATE   (IsAssigned());
        SAPDBERR_ASSERT_ARGUMENT(pos < m_Capacity);

        return *(m_Data + pos);
    }

    /*--------------------------------------------------------------------------
    function:     IsResizable
    description:  Checks whether resource is resizable

                  - A field resource has a fixed length and is not resizable.
    arguments:    none
    return value: (bool) false
    --------------------------------------------------------------------------*/
    bool IsResizable() const;

    /*--------------------------------------------------------------------------
    function:     ProvideCapacity
    description:  Reserves resource space for elements

                  - A field resource cannot reserve more space than the actual 
                    field capacity. Therefore, the method does nothing, except
                    for checking the requested capacity against the available
                    capacity. If the available capacity is not sufficient, 
                    an exception is thrown.

    arguments:    requested [in]  requested resource size in elements of BasisType
                  used      [in]  number of elements of BasisType to be saved on realloc
                                  (obsolete for field resources)
    return value: (SAPDB_Bool) true if enough space available
    // exceptions:   SAPDB_ResourceExhausted
    --------------------------------------------------------------------------*/
    SAPDB_Bool ProvideCapacity
        (SizeType                       requested, 
         SizeType                       used        = 0);

protected:

    Pointer   m_Data;                // Pointer to the memory block of the resource
    SizeType  m_Capacity;            // Capacity of the memory block in elements of BasisType
};
/*!*****************************************************************************

    endclass: SAPDBFields_FieldResource

*******************************************************************************/



//------------------------------------------------------------------------------
// SAPDBFields_FieldResource
//------------------------------------------------------------------------------

template <class T>
inline SAPDBFields_FieldResource<T>::SAPDBFields_FieldResource() 
    : 
    m_Data(0), 
    m_Capacity(0) 
{}

//------------------------------------------------------------------------------

template <class T>
inline SAPDBFields_FieldResource<T>::SAPDBFields_FieldResource
    (void*                          src, 
     SizeType                       len) 
    : 
    m_Data    (reinterpret_cast<Pointer>(src)), 
    m_Capacity(len)
{}

//------------------------------------------------------------------------------

template <class T>
inline void SAPDBFields_FieldResource<T>::Assign
    (void*                          src, 
     SizeType                       len) 
{
    m_Data     = reinterpret_cast<Pointer>(src);
    m_Capacity = len;
}

//------------------------------------------------------------------------------

template <class T>
inline void SAPDBFields_FieldResource<T>::Deassign
    ()
{
    m_Data      = 0;
    m_Capacity  = 0;
}

//------------------------------------------------------------------------------

template <class T>
inline typename SAPDBFields_FieldResource<T>::SizeType SAPDBFields_FieldResource<T>::Capacity
    () const
{
    return IsAssigned() ? m_Capacity : 0;
}

//------------------------------------------------------------------------------

template <class T>
inline typename SAPDBFields_FieldResource<T>::SizeType SAPDBFields_FieldResource<T>::MaxCapacity
    () const
{
    return Capacity();
}

//------------------------------------------------------------------------------

template <class T>
inline bool SAPDBFields_FieldResource<T>::IsAssigned
    () const
{
    return (m_Data != 0);
}



//------------------------------------------------------------------------------

template <class T>
inline bool SAPDBFields_FieldResource<T>::IsResizable
    () const
{
    return false;
}

//------------------------------------------------------------------------------

template <class T>
inline SAPDB_Bool SAPDBFields_FieldResource<T>::ProvideCapacity
    (SizeType                       requested, 
     SizeType                       used)
{
    SAPDBERR_ASSERT_STATE(IsAssigned());

    /* 8.0 version:
    (change returntype from SAPDB_Bool to void)
    if (requested > Capacity())
    {
        throw SAPDBFields_ResourceExhausted(__CONTEXT__, Capacity(), requested);
    }
    */

    // 7.4 version:
    return requested <= Capacity();
}

//------------------------------------------------------------------------------
// SAPDBFields_FieldResource
//------------------------------------------------------------------------------



#endif // SAPDBFIELDS_FIELDRESOURCE_HPP

