/*!***************************************************************************
  module:       SAPDBMem_CallDirectory.hpp

  responsible:  ThomasA

  special area: Memory management

  description:  directory of call stacks of memory requestors

  copyright:    Copyrighted by SAP AG, 1999

  created:      Wed Jan 12 14:35:56  2000

  updated:      Wed Feb  2 12:26:43  2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef SAPDBMEM_CALLDIRECTORY_HPP
#define SAPDBMEM_CALLDIRECTORY_HPP

#include <string.h>
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"
#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"

/*! 
  Class: SAPDBMem_CallDirectory

  Description: this class realizes a singleton, that stores call stacks
               given by an array of pointers. For each call stack a
               unique integer is assigned, which serves a surrogate for
               the call stack. 
               If call stack monitoring of the global allocator is enabled,
               each memory chunk contains the call stack surrogate of the
               memory requestor. Together with an iterator over the allocator
               and over this call stack directory, this allows a better
               analysis of memory leak problems.
               The system table sysdd.heap_content yields the corresponding information
 */

class SAPDBMem_CallDirectory
{
  struct CallStackDescriptor;
public :
  class Iterator;

  friend class SAPDBMem_CallDirectory::Iterator;

  typedef void* VoidPtr;
  /*!
  class : Iterator

  description : defines an iterator over call directoy.   
*/
  class Iterator {
    friend class SAPDBMem_CallDirectory;
  public :
    operator bool() const;
    void operator++();
    void Current(int& callStackId, int& stackLength, void*** pStack) const;
  private :
    Iterator();
    SAPDBMem_CallDirectory& m_callStackDirectory;
    int                     m_headIndex;
    CallStackDescriptor*    m_curr;
  };
/*! endclass: HeapIterator */

  /*!----------------------------------------------------------------------
   function: Begin

   description:  returns an iterator over the call stack directory
   arguments:    none
   return value: call stack iterator 
   -----------------------------------------------------------------------*/

  Iterator Begin() const; 

  /*!----------------------------------------------------------------------
   function: Instance

   description:  returns a reference to the call director instance, If the 
                 does not already exist, the singleton in created impilcitly.
   arguments:    none
   return value: reference to the call directory instance 
   -----------------------------------------------------------------------*/

  static SAPDBMem_CallDirectory& Instance();

  /*!----------------------------------------------------------------------
   function: GetCallId

   description:  returns a surrogate (4 Bytes integer) for a given call stack
                 identified by an array of pointers.
                 The method checks, if the call stack is already stored in the
                 directory. If true, the corresponding surrogate is returned.
                 Otherwise a new surrogate is generated and the given call
                 stack is interted into the directory. 
   arguments:
      stackLength [in] number of pointers stored in pStack
      pStack [in] pointer array of PC adresses.
     
   return value: surrogate for the given call stack 
   -----------------------------------------------------------------------*/

  SAPDB_UInt4 GetCallId (int stackLength, VoidPtr* pStack);

  /*!----------------------------------------------------------------------
   function: GetMaxCallStackId

   description:  return the maximal call stack surrogate

   arguments: none
   
   return value: the maximal call stack surrogate 
   -----------------------------------------------------------------------*/

  inline SAPDB_UInt4 GetMaxCallStackId() const;
  
private :
  
  enum { HEADER_SIZE = 217 };

  struct CallStackDescriptor
  {
    CallStackDescriptor* m_next;
    SAPDB_Int4           m_stackLength;
    SAPDB_UInt4          m_callId;
    void*                m_stack[1024];
    CallStackDescriptor(SAPDB_UInt4          callId,
                        int                  stackLength, 
                        void*                pStack,
                        CallStackDescriptor* next);

    void* operator new(size_t sz, int stackLength);

    void  operator delete (void* p, int stackLength);
  };

  SAPDBMem_CallDirectory();

  inline SAPDB_UInt4 FindCallId (int hashIndex, int stackLength, void* pStack);
  SAPDB_UInt4        Insert (int hashIndex, SAPDB_UInt4 callId, int stackLength, void* pStack);

  typedef CallStackDescriptor* CallStackDescriptorPtr;

  static SAPDBMem_CallDirectory*          m_pInstance;
  int                                     m_symbOffset;
  RTESync_InterlockedCounter<SAPDB_UInt4> m_callIdGenerator;
  RTESync_Spinlock                        m_slotLock[HEADER_SIZE];
  volatile CallStackDescriptorPtr         m_head[HEADER_SIZE];
}; /*! endclass: SAPDBMem_CallDirectory */

/*-------------------------------------------------------------------*/

inline SAPDB_UInt4 SAPDBMem_CallDirectory::FindCallId (int hashIndex, int stackLength, void* pStack)
{
  CallStackDescriptor* pCurr = m_head[hashIndex];
  while (NULL != pCurr)
  {
    if (stackLength == pCurr->m_stackLength)
    {
      if (0 == memcmp (pStack, &pCurr->m_stack[0], stackLength * sizeof(void*)))
      {
        return pCurr->m_callId;
      }
    }
    pCurr = pCurr->m_next;
  }
  return 0;
}

/*-------------------------------------------------------------------*/

inline SAPDB_UInt4 SAPDBMem_CallDirectory::GetMaxCallStackId() const
{
  return m_callIdGenerator;
}

#endif