/*!*****************************************************************************

  module:       SAPDBMem_DefaultRawAllocator.hpp

  ------------------------------------------------------------------------------

  responsible:  Stefan Baier

  special area: Memory management

  description:  Memory allocator using standard 'operator new'
                and 'operator delete' functions.
                
  changes:      31.May 2000
                SAPDBMem_DefaultRawAllocator becomes a singleton with
                the method Instance().

  ------------------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#ifndef SAPDBMem_DefaultRawAllocator_HPP
#define SAPDBMem_DefaultRawAllocator_HPP

#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"

#include "SAPDBCommon/MemoryManagement/SAPDBMem_Exceptions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*!****************************************************************************
  class: SAPDBMem_DefaultRawAllocator

  description: Simple implementation of the 'SAPDBMem_IRawAllocator'
               interface.

               This class uses the 'operator new' and 'operator delete' functions
               for memory management. It's purpose is to have a simple
               implementation of the 'SAPDBMem_IRawAllocator' interface at hand
               that can be used for testing purposes.
*****************************************************************************/

class SAPDBMem_DefaultRawAllocator : public SAPDBMem_IRawAllocator
{
public:
   /*!----------------------------------------------------------------------
	function: Instance
	description: Returns a reference to the singleton of this class
   -----------------------------------------------------------------------*/
   static SAPDBMem_DefaultRawAllocator& Instance(); 

   /*!----------------------------------------------------------------------
	function: Allocate
	description: See 'SAPDBMem_IRawAllocator'

                These allocate methods use 'operator new'.
   exceptions:  SAPDBMem_BadAllocException:
                Thrown, if the memory cannot be allocated
   prototypes:  2
   -----------------------------------------------------------------------*/
   virtual void* Allocate(SAPDB_ULong Count);

   virtual void* Allocate(SAPDB_ULong Count, const void* Hint)
      { return Allocate(Count); }

   /*!----------------------------------------------------------------------
	function: Deallocate
	description: See 'SAPDBMem_IRawAllocator'.

                This method uses 'operator delete'.
   -----------------------------------------------------------------------*/
   virtual void Deallocate(void *p);

   /*!----------------------------------------------------------------------
   function: GetBaseAllocatorCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls of
                 the allocator to the underlying base allocator.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate of the base allocator.
                 CountDealloc [out]
                 The number of calls of method Deallocate of the base.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;


   /*!----------------------------------------------------------------------
   function: GetCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate.
                 CountDealloc [out]
                 The number of calls of method Deallocate.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;

   /*!----------------------------------------------------------------------
    function: GetErrorCount
    description:  Returns the number of errors found and corrected by the allocator

                  this allocator is not able to provide the information, i.e. returns 0
 
    return value: 0
    -----------------------------------------------------------------------*/
  
    virtual int GetErrorCount() const;

private:
   /*----------------------------------------------------------------------
	function: Default constructor
	description: Does nothing. This constructor is private, because this
                class is a singleton.
   -----------------------------------------------------------------------*/
    SAPDBMem_DefaultRawAllocator() 
        : m_CountAlloc(0)
        , m_CountDealloc(0)
    {}

   /*-----------------------------------------------------------------------
	function: copy constructor
	description: Private copy constructor avoids copying.
                This copy constructor need not to be defined, because
                noone can construct a SAPDBMem_DefaultRawAllocator and call
                the copy constructor thereafter.
                Besides, a private definition of the copy constructor would 
                result in a warning on several compilers.
   -----------------------------------------------------------------------*/
   /*
   SAPDBMem_DefaultRawAllocator(const SAPDBMem_DefaultRawAllocator &)
      {}
   */

   /*-----------------------------------------------------------------------
	operator: =	
	description: Private assignment operator avoids copying.
   -----------------------------------------------------------------------*/
   SAPDBMem_DefaultRawAllocator operator=(SAPDBMem_DefaultRawAllocator &)
      { return *this; }

   SAPDB_ULong m_CountAlloc;
   SAPDB_ULong m_CountDealloc;

   static SAPDBMem_DefaultRawAllocator* m_Instance;
};

/*!****************************************************************************
  endclass: SAPDBMem_DefaultRawAllocator
*****************************************************************************/


#endif // SAPDBMem_DefaultRawAllocator_HPP
