/*!***************************************************************************

  module      : SAPDBMem_IAllocatorInfo.hpp

  -------------------------------------------------------------------------

  responsible : ThomasA

  special area: Memory Management
  description : Interface for allocator statistics   
  
 
  last changed: 2001-01-10  11:04
  see also    : 
  first created:2000-03-09  18:21

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/

#ifndef SAPDBMEM_ALLOCATORINFO_HPP
#define SAPDBMEM_ALLOCATORINFO_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class SAPDBMem_IAllocatorInfo
{
public : 
  /*!----------------------------------------------------------------------
   function: CalcStatistics
   description:  Outputs statistics about the memory usage of the allocator.
   arguments:    BytesUsed [out]
                 The number of bytes used by this allocater.
                 BytesControlled [out]
                 The number of bytes controlled by this allocator.

                 The default implementation returns GetInvalidSize() for 
                 all values.
   return value: none
   -----------------------------------------------------------------------*/
   virtual void CalcStatistics(SAPDB_ULong &BytesUsed,
                               SAPDB_ULong &BytesControlled)
   { 
       BytesUsed       = GetInvalidSize();
       BytesControlled = GetInvalidSize();
   }

   /*!----------------------------------------------------------------------
   function: GetBaseAllocatorCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls of
                 the allocator to the underlying base allocator.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate of the base allocator.
                 CountDealloc [out]
                 The number of calls of method Deallocate of the base.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const = 0;

  /*!----------------------------------------------------------------------
   function: GetCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate.
                 CountDealloc [out]
                 The number of calls of method Deallocate.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const = 0;

   /*!----------------------------------------------------------------------
   function: GetErrorCount
   description:  Returns the number of errors found and corrected by the allocator

   The derived class is responsible to implement this feature (but must not do so...)

   return value: integer
   -----------------------------------------------------------------------*/
  
   virtual int GetErrorCount() const = 0;

  /*!----------------------------------------------------------------------
   function: GetIdentifier
   description:  Returns the identifier associated with this allocator

   This function is called to identify an allocator instance. The derived
   class is responsible to implement this feature (but must not do so...)

   return value: const SAPDB_UTF8 *
   -----------------------------------------------------------------------*/
  
   virtual const SAPDB_UTF8 *GetIdentifier() const
      { return (const SAPDB_UTF8 *)""; }

   /*!----------------------------------------------------------------------
   function: GetInvalidSize
   description:  Returns the value for an invalid size.
                 The is defined as 0xFFFF... .
   return value: SAPDB_ULong
   -----------------------------------------------------------------------*/
   static SAPDB_ULong GetInvalidSize()
      { return (((SAPDB_ULong)((SAPDB_Long)(-1)))); }
};

#endif 