/*!***************************************************************************
  module:       SAPDBMem_IBlockAllocator.hpp

  responsible:  StefanB

  special area: Memory management

  description:  Block allocator interface for memory management of blocked memory.

  copyright:    Copyright by SAP AG, 2000

  created:      Thu Feb 10 14:24:40  2000

  updated:      Thu Apr 06 14:24:40  2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef SAPDBMem_IBlockAllocator_HPP
#define SAPDBMem_IBlockAllocator_HPP

#include "SAPDBCommon/MemoryManagement/SAPDBMem_IAllocatorInfo.hpp"

/*!***************************************************************************
  class:       SAPDBMem_IBlockAllocator

  description: Allocator interface for aligned blocks of memory of
               fixed size.

    Classes implementing this interface provide methods for reserving
    and freeing memory in integral multiples of a fixed block
    size. The reserved memory is always aligned to integral multiples
    of a fixed alignment size, which may differ from the block size.

    The interface serves the abstraction from low level allocation
    mechanisms like, e.g., the 'mmap' system call on UNIX platforms.
    Application programmers will most likely prefer STL allocators or
    raw allocators in the sense of the C++ Standard.

    While this interface is similar to the one required by the
    Standard Template Library, the block size and the alignment size
    is not determined by a C++ data type at compile time, but instead
    is determined at runtime.

    Implementations of this interface must not be confused neither
    with STL allocators nor with raw storage allocation functions as
    described in the C++ Standard, par.3.7.3 'Dynamic storage duration
    [basic.stc.dynamic]'.
    
    Every class implementing this interface must assure, that the
    copy constructor and the assignment operator is inaccessible.
*****************************************************************************/

class SAPDBMem_IBlockAllocator : public SAPDBMem_IAllocatorInfo
{
public:
   /*!----------------------------------------------------------------------
   function: CalcBlockCount
   description:  Returns the number of blocks needed to allocate 
                 <NumberOfBytes> bytes.

                 Since Allocate() needs the number of blocks instead of number 
                 of bytes, this helper routine exists to calculate the right 
                 argument. When used calling the Routine GetBlockSize() is not needed.

   arguments:    NumberOfBytes [in] 
                 Number of bytes instead of number of blocks.

   return value: SAPDB_ULong
                 Number of blocks needed for a given number of bytes.
   -----------------------------------------------------------------------*/
   virtual SAPDB_ULong CalcBlockCount(SAPDB_ULong NumberOfBytes) const
      { return ( NumberOfBytes + GetBlockSize() - 1 ) / GetBlockSize(); }

   /*!----------------------------------------------------------------------
   function: GetBlockSize
   description: Returns the size in bytes of the blocks that can
                be allocated by this allocator.
   -----------------------------------------------------------------------*/
   virtual SAPDB_ULong GetBlockSize() const = 0;

   /*!----------------------------------------------------------------------
   function: GetAlignmentSize
   description: Returns the alignment size (in bytes) of the blocks
                that can be allocated by this allocator.

                An implementation must ensure that the alignment size A and
                the block size B satisfy B = N * A, where N is a non-zero
                integer number.
   -----------------------------------------------------------------------*/
   virtual SAPDB_ULong GetAlignmentSize() const = 0;

   /*!----------------------------------------------------------------------
   function: Allocate
   description:  Allocates contiguous memory for 'count' blocks of
                 memory.
                 An implementation grants that a pointer returned will be an
                 integral multiple of the alignment size.

                 The memory is not initialised.
   arguments:    BlockCount [in]
                 The number of blocks to be allocated. This argument must be 
                 greater zero.
   return value: void*
                 A pointer to the allocated raw memory block of <BlockCount> 
                 blocks. If exception handling is disabled, a NULL pointer will
                 be returned in case of insufficient memory.
   exceptions:   From 'SAPDBMem_IBlockAllocator' derived classes may throw a
                 'SAPDBMem_BadAllocException' if the allocation fails.
   -----------------------------------------------------------------------*/
   virtual void* Allocate(SAPDB_ULong BlockCount) = 0;

   /*!----------------------------------------------------------------------
   function: Deallocate
   description: Frees the memory.
                None of the objects in the memory block will be deinitialised
                by this call.  The size argument must match the value passed
                to 'allocate' to obtain this memory. The pointer argument
                must not be null.
   -----------------------------------------------------------------------*/
   virtual void Deallocate(void* p, SAPDB_ULong BlockCount) = 0;


   /*!----------------------------------------------------------------------
   function: GetErrorCount
   description:  Returns the number of errors found and corrected by the allocator
                 Block allocator normally aren't able to provide this information

   return value: 0
      -----------------------------------------------------------------------*/
  
   virtual int GetErrorCount() const
   {
       return 0;
   }
};

/*!***************************************************************************
  endclass:       SAPDBMem_IBlockAllocator
***************************************************************************/

#endif // SAPDBMem_IBlockAllocator_HPP
