/*!*****************************************************************************

  module:       SAPDBMem_IRawAllocator.hpp

  ------------------------------------------------------------------------------

  responsible:  Stefan Baier

  special area: Memory management

  description:  - Allocator interface for memory management of raw memory.
                - Macro for reserving static data space.
                
  ------------------------------------------------------------------------------

  copyright:             Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#ifndef SAPDBMem_IRawAllocator_HPP
#define SAPDBMem_IRawAllocator_HPP

#include "SAPDBCommon/MemoryManagement/SAPDBMem_IAllocatorInfo.hpp"

/*!****************************************************************************
  class: SAPDBMem_IRawAllocator

  description: Allocator interface for 'raw' memory. 
               
               Classes implementing this interface provide methods for reserving
               and freeing 'raw' memory. The reserved memory always is suitably
               (platform dependent) aligned such that any object type can be put
               there. Implementations shall meet the requirements of the C++
               Standard, par.3.7.3 'Dynamic storage duration
               [basic.std.dynamic]'.
               Every class implementing this interface must assure, that the
               copy constructor and the assignment operator is inaccessible.
******************************************************************************/

class SAPDBMem_IRawAllocator : public SAPDBMem_IAllocatorInfo 
{
public: 
   /*!----------------------------------------------------------------------
   declaration: AlignType
   description: Variables of this type are properly aligned on all platforms.
   -----------------------------------------------------------------------*/
   struct AlignType {
      union { void* p; long l; double d; };
   };
   /*! enddeclaration: */

   /*!----------------------------------------------------------------------
   function: GetAlignmentSize
   description:  Returns the alignment size in bytes of this allocator.
   return value: SAPDB_ULong
   -----------------------------------------------------------------------*/
   static SAPDB_ULong GetAlignmentSize()
      { return SAPDB_ALIGNMENT; }

   /*!----------------------------------------------------------------------
   function: GetInvalidSize
   description:  Returns the value for an invalid size.
                 The is defined as 0xFFFF... .
   return value: SAPDB_ULong
   -----------------------------------------------------------------------*/
   static SAPDB_ULong GetInvalidSize()
      { return (((SAPDB_ULong)((SAPDB_Long)(-1)))); }

   /*!----------------------------------------------------------------------
   function: CalcAlignmentOffset
   description:  Returns the offset of <Address> to the next address which is 
                 greater than <Address> and is aligned.

                 This offset can be used, to get an aligned address with the 
                 formula 
                    NewAddress = Address + Offset;
   return value: SAPDB_ULong
   -----------------------------------------------------------------------*/
   static SAPDB_ULong CalcAlignmentOffset(void *Address)
   {
      return GetAlignmentSize()-
             (reinterpret_cast<SAPDB_Byte*>(Address)-reinterpret_cast<SAPDB_Byte*>(0)) 
              % GetAlignmentSize();
   }

   /*!----------------------------------------------------------------------
   function: IsAligned
   description:  Checks, if the given address is aligned.
   return value: bool
   -----------------------------------------------------------------------*/
   static bool IsAligned(void *Address)
   {
      return CalcAlignmentOffset(Address) == GetAlignmentSize();
   }

   /*!----------------------------------------------------------------------
   function: Destructor
   description: Does nothing
   -----------------------------------------------------------------------*/
   virtual ~SAPDBMem_IRawAllocator() {}

   /*!----------------------------------------------------------------------
   function: Allocate
   description:  Allocates contiguous memory for <ByteCount> bytes.
                 The memory is aligned as indicated by the 'GetAlignmentSize' 
                 method.
                 The memory is not initialised.
   arguments:    ByteCount [in]
                 The size in bytes of the memory to be allocated. This 
                 argument must be greater zero.
                 Hint [in]
                 If possible, the allocator tries to reserve the memory at
                 the same position as given by <Hint>. This is useful for
                 resizing a memory block. If the reallocation was succesfull
                 the return value and <Hint> are identical.
   return value: void*
                 A pointer to the allocated raw memory block of <ByteCount> 
                 bytes. If exception handling is disabled, a NULL pointer will
                 be returned in case of insufficient memory.
   exceptions:   From 'SAPDBMem_IRawAllocator' derived classes may throw a
                 'SAPDBMem_BadAllocException' if the allocation fails.
   prototypes:   2
   -----------------------------------------------------------------------*/
   virtual void* Allocate(SAPDB_ULong ByteCount) = 0;
   virtual void* Allocate(SAPDB_ULong ByteCount, const void* Hint) = 0;

   /*!----------------------------------------------------------------------
   function: Deallocate
   description:  Frees the memory behind <p>.
                 
                 The memory must have been allocated by a previous call to
                 'Allocate'. This is checked by assertions.
                 !!! An implementation of this method may not throw an 
                 exception !!!
   arguments:    p [in]
   return value: none
   -----------------------------------------------------------------------*/
   virtual void Deallocate(void* p) = 0;

   /*!----------------------------------------------------------------------
   function: CheckConsistency
   description:  Checks the integrity of the allocator.

                 This function checks the inner state of the allocator 
                 concerning consistency. Therefore, the total memory controlled 
                 by this allocator is checked.
                 If an inconsistency is detected, an excpetion is thrown.
                 The default implementation does nothing.
   arguments:    none
   return value: bool (only in the non-exception case)
   exceptions:   SAPDBMem_ConsistencyException
   -----------------------------------------------------------------------*/
   virtual bool CheckConsistency() { return true; }

   /*!----------------------------------------------------------------------
   function: GetChunkSize
   description:  returns the size af a chunk.
   arguments:    p [int]
                 The address of the chunk. This must be the result of
                 a previous call of the Allocate method of the allocator.
                 The default implementation returns GetInvalidSize().
   return value: the internal size of the chunk in bytes. The size may differ
                 from the required size of the corresponding Allocate call.
   -----------------------------------------------------------------------*/

   virtual SAPDB_ULong GetChunkSize(void* p)
   {
       return GetInvalidSize();
   }

   /*!----------------------------------------------------------------------
   operator: ==
   description:  Compares two raw allocator instances.

                 Two instances of raw allocators are equal, if they lie at 
                 the same memory position.
   arguments:    Other [in]
                 The other instance to compare with.
   return value: bool
   -----------------------------------------------------------------------*/
   bool operator ==(SAPDBMem_IRawAllocator& Other)
      { return (this == &Other); }

   /*!----------------------------------------------------------------------
   operator: !=
   description:  Returns the opposite of operator ==.
   return value: bool
   -----------------------------------------------------------------------*/
   bool operator !=(SAPDBMem_IRawAllocator& Other)
      { return !(operator==(Other)); }
};

/*! endclass: SAPDBMem_IRawAllocator */


/*!****************************************************************************
  declaration: SAPDBMEM_STATIC_RAW_ALLOCATE
  description: Macro used to allocate static space aligned properly
  
               If a modules needs space even before any allocator is available, 
               this macro should be used, so space is always properly aligned. 
               There is no corresponding free.
  arguments:   _name [in]
               Name of variable used i.e. in a later placement new
               _size_in_bytes [in]
               Number of bytes needed
  example:

        SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTEMem_RawAllocator));

        m_Instance = new (Space) RTEMem_RawAllocator();
*******************************************************************************/

#define SAPDBMEM_STATIC_RAW_ALLOCATE(_name, _size_in_bytes)                  \
   static SAPDBMem_IRawAllocator::AlignType                                  \
       _name[((_size_in_bytes)+sizeof(SAPDBMem_IRawAllocator::AlignType)-1)/ \
              sizeof(SAPDBMem_IRawAllocator::AlignType)]

/*! enddeclaration: SAPDBMEM_STATIC_RAW_ALLOCATE */



#endif // SAPDBMem_IRawAllocator_HPP
