/*!*****************************************************************************

  module:        SAPDBMem_IncrementalBufferAllocator.hpp

  ------------------------------------------------------------------------------

  responsible:  Stefan Baier

  special area: Memory management

  description:  Simple implementation of 'SAPDBMem_IRawAllocator' interface.

  ------------------------------------------------------------------------------

  copyright:                 Copyright by SAP AG, 2000



     ========== licence begin LGPL
     Copyright (C) 2000 SAP AG

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
     ========== licence end

*******************************************************************************/

#ifndef SAPDBMem_IncrementalBufferAllocator_HPP
#define SAPDBMem_IncrementalBufferAllocator_HPP


#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*!***************************************************************************
  class: SAPDBMem_IncrementalBufferAllocator

  description: Simple implementation of the 'SAPDBMem_IRawAllocator'
                    interface.

                    This class manages a (large) memory block and allocates smaller
                    blocks inside in an aligned way.  It can only free the large
                    memory block as a whole.
                    The user of this allocator has to take care about the memory
                    that comes with the arguments in the constructor.
*****************************************************************************/

class SAPDBMem_IncrementalBufferAllocator
    : public SAPDBMem_IRawAllocator
{
    friend class SAPDBMem_IncrementalRawAllocator;

protected:
    SAPDB_Byte* m_Memory;       // Address of the memory block managed by this allocator
    SAPDB_ULong m_Size;         // Length of the memory block in bytes
    SAPDB_Byte* m_FreeMemory;   // Address of the free memory block
                                        // This address is always in the range m_Memory...m_Memory+m_Capacity-1
    SAPDB_ULong m_FreeSize;     // Length of the free memory in bytes
    SAPDB_Byte* m_LastPointer; // Hint for reallocating memory
    SAPDB_ULong m_LastSize;     // Hint for reallocating memory
    SAPDB_ULong m_TotalAllocatedSize;    // The total size of the memory allocated from the user.
    SAPDB_ULong m_CountAlloc;            // number of calls of method Allocate
    SAPDB_ULong m_CountDealloc;          // number of calls of method Deallocate
public:
    /*!----------------------------------------------------------------------
    function: Constructor
    description: Constructs an raw allocator with an initial memory block.
                     
                 Address must be aligned.
    arguments:   Address [in]
                 The address of the memory block.
                 Length [in]
                 The length of the memory block.
    -----------------------------------------------------------------------*/
    SAPDBMem_IncrementalBufferAllocator(void* Address, SAPDB_ULong Length);
    
    /*!----------------------------------------------------------------------
    function: Destructor
    description: Does nothing.
    -----------------------------------------------------------------------*/
    virtual ~SAPDBMem_IncrementalBufferAllocator() {}

    /*!----------------------------------------------------------------------
    function: IsAssigned
    description: Checks, if the allocator manages a valid memory block.
    -----------------------------------------------------------------------*/
    bool IsAssigned()
        { return ((m_Memory != 0) && (m_Size != 0)); }

    /*!----------------------------------------------------------------------
    function: Allocate
    description: see 'SAPDBMem_IRawAllocator'
    prototypes:  2
    -----------------------------------------------------------------------*/
    virtual void* Allocate(SAPDB_ULong ByteCount) ;
    
    virtual void* Allocate(SAPDB_ULong ByteCount, const void* Hint);

    /*!----------------------------------------------------------------------
    function: Dellocate
    description: see 'SAPDBMem_IRawAllocator'
    -----------------------------------------------------------------------*/
    virtual void Deallocate(void* Ptr);

    /*!----------------------------------------------------------------------
    function: CheckConsistency
    description: See 'SAPDBMem_IRawAllocator
    -----------------------------------------------------------------------*/
    virtual bool CheckConsistency();

    /*!----------------------------------------------------------------------
    function: CalcStatistics
    description: See 'SAPDBMem_IRawAllocator
    -----------------------------------------------------------------------*/
    virtual void CalcStatistics(SAPDB_ULong &BytesUsed,
                                SAPDB_ULong &BytesControlled);

    /*!----------------------------------------------------------------------
    function: FreeAll
    description:  Frees the total memory controlled by this allocator.
    arguments:    none
    return value: none
    -----------------------------------------------------------------------*/
    void FreeAll();

    /*!----------------------------------------------------------------------
    function: GetBaseAllocatorCallStatistics
    description: See 'SAPDBMem_IAllocatorInfo'
    -----------------------------------------------------------------------*/

    virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;

    /*!----------------------------------------------------------------------
    function: GetCallStatistics
    description: See 'SAPDBMem_IRawAllocator
    -----------------------------------------------------------------------*/
 
    virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                   SAPDB_ULong &CountDealloc) const;

    /*!----------------------------------------------------------------------
    function: GetCapacity
    description: Returns the size of bytes, that can be allocated with the
                 next call to Allocate().
    -----------------------------------------------------------------------*/
    SAPDB_ULong GetCapacity()
        { return m_FreeSize; }


    /*!----------------------------------------------------------------------
    function: GetErrorCount
    description:  Returns the number of errors found and corrected by the allocator

                  this allocator is not able to provide the information, i.e. returns 0
 
    return value: 0
    -----------------------------------------------------------------------*/
  
    virtual int GetErrorCount() const;

private:
    /*-----------------------------------------------------------------------
    function: copy constructor
    description: Private copy constructor avoids copying.
    -----------------------------------------------------------------------*/
    SAPDBMem_IncrementalBufferAllocator(const SAPDBMem_IncrementalBufferAllocator &)
        {}

    /*-----------------------------------------------------------------------
    operator: = 
    description: Private assignment operator avoids copying.
    -----------------------------------------------------------------------*/
    SAPDBMem_IncrementalBufferAllocator operator=(SAPDBMem_IncrementalBufferAllocator &)
        { return *this; }

    /*-----------------------------------------------------------------------
    function: Initialize
    description: Initialize all members.
    -----------------------------------------------------------------------*/
    void Initialize(void* Address, SAPDB_ULong Length);

};
/*!***************************************************************************
  endclass: SAPDBMem_IncrementalBufferAllocator
****************************************************************************/

#endif // SAPDBMem_IncrementalBufferAllocator_HPP

