/*!*****************************************************************************

  module:       SAPDBMem_IncrementalRawAllocator.hpp

  ------------------------------------------------------------------------------

  responsible:  Stefan Baier

  special area: Memory management

  description:  Simple implementation of 'SAPDBMem_IRawAllocator' interface.

  ------------------------------------------------------------------------------

  copyright:                 Copyright by SAP AG, 2000



     ========== licence begin LGPL
     Copyright (C) 2000 SAP AG

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
     ========== licence end

*****************************************************************************/
#ifndef SAPDBMem_IncrementalRawAllocator_HPP
#define SAPDBMem_IncrementalRawAllocator_HPP


#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IncrementalBufferAllocator.hpp"


/*!***************************************************************************
  class: SAPDBMem_IncrementalRawAllocator
  description: Implementation of the 'SAPDBMem_IRawAllocator' interface.

                    This class manages large memory blocks obtained from a raw 
                    allocator and allocates smaller blocks inside sequentially in 
                    an aligned way. It can only free the large memory block as a 
                    whole.
*****************************************************************************/

class SAPDBMem_IncrementalRawAllocator : public SAPDBMem_IRawAllocator
{
private:
    SAPDBMem_IRawAllocator& m_RawAllocator;     // The underlying raw allocator.
    SAPDB_ULong m_MinSize;                      // The minimum size for each large block.
                                                // from the underlying raw allocator
    void *m_FirstBlock;                         // The address of the first block in the chain.
    SAPDB_ULong m_ActualBlockSize;              // The size of the last block in the chain
    void *m_ActualBlock;                        // The address of the last block in the chain.
    SAPDB_ULong m_HistoryUsedSize;              // The total size of all large blocks allocated 
                                                // from the underlying raw allocator without the
                                                // size of the actual large block.
    SAPDB_ULong m_CountAlloc;                   // number of calls of method Allocate
    SAPDB_ULong m_CountDealloc;                 // number of calls of method Deallocate
    // The incremental raw allocator uses an incremental buffer allocator, which
    // manages the actual block.
    SAPDBMem_IncrementalBufferAllocator m_BufferAllocator;

public:
    /*!----------------------------------------------------------------------
    function: Constructor
    description: Constructs an incremental raw allocator with memory blocks 
                     obtained from a raw allocator.

                     This allocator manages an unlimitid number of blocks obtained 
                     from the underlying raw allocator. Within these blocks the 
                     allocator allocates smaller blocks sequentially in an aligned 
                     way. 
                     The minimum size of each block is given at construct time.
                     The allocator starts with one block of the minimum size.
                     If the remaining space within the first block is not sufficient 
                     to fulfill the next Allocate() request, additional blocks are
                     requested fomr the underlying raw allocator.
                     The maximum size depends on the argument in the Allocate() 
                     request.
    arguments:   RawAlloc [in]
                     The underlying raw allocator.
                     MinSize [in]
                     The minimum and initial size of the managed blocks.
    -----------------------------------------------------------------------*/
    SAPDBMem_IncrementalRawAllocator(
        SAPDBMem_IRawAllocator& RawAlloc,
        SAPDB_ULong MinSize
    );

    /*!----------------------------------------------------------------------
    function: Destructor
    description: 
    -----------------------------------------------------------------------*/
    virtual ~SAPDBMem_IncrementalRawAllocator();

    /*!----------------------------------------------------------------------
    function: Allocate
    description: See 'SAPDBMem_IRawAllocator'
    prototypes:  2
    -----------------------------------------------------------------------*/
    virtual void* Allocate(SAPDB_ULong ByteCount);
    virtual void* Allocate(SAPDB_ULong ByteCount, const void* Hint);

    /*!----------------------------------------------------------------------
    function: Deallocate
    description: See 'SAPDBMem_IRawAllocator'
    -----------------------------------------------------------------------*/
    virtual void Deallocate(void* p) {}

    /*!----------------------------------------------------------------------
    function: CheckConsistency
    description: See 'SAPDBMem_IRawAllocator
    -----------------------------------------------------------------------*/

    virtual bool CheckConsistency();

	/*!----------------------------------------------------------------------
    function: GetBaseAllocatorCallStatistics
    description: See 'SAPDBMem_IAllocatorInfo'
    -----------------------------------------------------------------------*/

    virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;

    /*!----------------------------------------------------------------------
    function: CalcStatistics
    description: See 'SAPDBMem_IRawAllocator
    -----------------------------------------------------------------------*/
    virtual void CalcStatistics(SAPDB_ULong &BytesUsed,
                                SAPDB_ULong &BytesControlled);
    
    /*!----------------------------------------------------------------------
    function: GetCallStatistics
    description: See 'SAPDBMem_IAllocatorInfo'
    -----------------------------------------------------------------------*/
    virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                   SAPDB_ULong &CountDealloc) const;

    /*!----------------------------------------------------------------------
    function: GetErrorCount
    description:  Returns the number of errors found and corrected by the allocator

                  this allocator is not able to provide the information, i.e. returns 0
 
    return value: 0
    -----------------------------------------------------------------------*/
  
    virtual int GetErrorCount() const;
                                   
    /*!----------------------------------------------------------------------
    function: FreeAll
    description:  Frees the total memory controlled by this allocator.
    arguments:    none
    return value: none
    -----------------------------------------------------------------------*/
    void FreeAll();

private:
    /*-----------------------------------------------------------------------
    function: copy constructor
    description: Private copy constructor avoids copying.
    -----------------------------------------------------------------------*/
    SAPDBMem_IncrementalRawAllocator(const SAPDBMem_IncrementalRawAllocator &InitAlloc)
     : m_RawAllocator(InitAlloc.m_RawAllocator), 
        m_BufferAllocator(0, 0)
    {}

    /*-----------------------------------------------------------------------
    operator: = 
    description: Private assignment operator avoids copying.
    -----------------------------------------------------------------------*/
    SAPDBMem_IncrementalRawAllocator operator=(SAPDBMem_IncrementalRawAllocator &)
        { return *this; }

    /*-----------------------------------------------------------------------
    function: GetNewBlock
    description: Get a new block from the raw allocator and chain it.
    arguments:   Size [in]
                 The size of the new block.
    -----------------------------------------------------------------------*/
    void* GetNewBlock(SAPDB_ULong Size);

    /*-----------------------------------------------------------------------
    function: InitializeBufferAllocator
    description: Initialize the buffer allocator with a memory block.
    arguments:   Block [in]
                 The address of the new memory block.
                 BlockSize [in]
                 The size of the memory block.
    -----------------------------------------------------------------------*/
    void InitializeBufferAllocator(void *Block, SAPDB_ULong BlockSize);
};

/*!***************************************************************************
  endclass: SAPDBMem_IncrementalRawAllocator
****************************************************************************/

#endif // SAPDBMem_IncrementalRawAllocator_HPP

