/*!*****************************************************************************

  module:       SAPDBMem_PageAllocator.hpp

  ------------------------------------------------------------------------------

  responsible:  Stefan Baier

  special area: Memory management

  description:  Raw allocator for allocating memory pages of constant size.
                     
  ------------------------------------------------------------------------------

  copyright:                 Copyright by SAP AG, 2000



     ========== licence begin LGPL
     Copyright (C) 2000 SAP AG

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
     ========== licence end

*******************************************************************************/

#ifndef SAPDBMem_PageAllocator_HPP
#define SAPDBMem_PageAllocator_HPP

#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*!***************************************************************************
  class: SAPDBMem_PageAllocator
  description: Implementation of the 'SAPDBMem_IRawAllocator' interface.

                    (Wer mchte, kann das nachfolgende ja bersetzen)
                    Diese Klasse verwaltet groe Speicherblcke, die ihr bergeben 
                    werden. Sie teilt diese Speicherblcke vollstndig in kleinere 
                    Blcke einheitlicher Gre, sogenannte Pages auf, und kann diese 
                    auf Anforderung vergeben.
*****************************************************************************/

class SAPDBMem_PageAllocator : public SAPDBMem_IRawAllocator
{
private:
    struct Page {
        Page *NextPage;
        Page *PreviousPage;
    };
    SAPDB_ULong m_PageSize;                 // The size of one page.
    SAPDB_ULong m_FreePages;                // The number of free pages.
    Page *m_FirstFreePagePtr;               // The address of the first free page in the chain.
    SAPDB_ULong m_CountAlloc;               // number of Allocate calls.
    SAPDB_ULong m_CountDealloc;             // number of Deallocate calls.       
public:
    /*!----------------------------------------------------------------------
    function: Constructor
    description: Constructs an page allocator

                 The page allocator is empty at the beginning. 
    arguments:   PageSize [in] page size in bytes
                 The size of one page.
    -----------------------------------------------------------------------*/
    SAPDBMem_PageAllocator(SAPDB_ULong PageSize);

    /*!----------------------------------------------------------------------
    function: Destructor
    description: 
    -----------------------------------------------------------------------*/
    virtual ~SAPDBMem_PageAllocator() {}

    /*!----------------------------------------------------------------------
    function: Allocate
    description: See 'SAPDBMem_IRawAllocator'
    prototypes:  2
    -----------------------------------------------------------------------*/
    virtual void* Allocate(SAPDB_ULong ByteCount);
    virtual void* Allocate(SAPDB_ULong ByteCount, const void* Hint)
        { return Allocate(ByteCount); }

    /*!----------------------------------------------------------------------
    function: Deallocate
    description: See 'SAPDBMem_IRawAllocator'
    -----------------------------------------------------------------------*/
    virtual void Deallocate(void* p);

    /*!----------------------------------------------------------------------
    function: AddMemoryBlock
    description:  This method lets an memory block be controlled by the page
                  allocator. The page allocator splits the memory block into
                  smaller pieces of the page size.
    arguments:    Ptr [in] pointer to the given memory block
                  Len [in] number of bytes belonging to the given memory block
    return value: none
    -----------------------------------------------------------------------*/
    void AddMemoryBlock(void* Ptr, SAPDB_ULong Len);

   /*!----------------------------------------------------------------------
   function: GetBaseAllocatorCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls of
                 the allocator to the underlying base allocator.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate of the base allocator.
                 CountDealloc [out]
                 The number of calls of method Deallocate of the base.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;

   /*!----------------------------------------------------------------------
   function: GetCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate.
                 CountDealloc [out]
                 The number of calls of method Deallocate.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;


    /*!----------------------------------------------------------------------
    function: GetFreePages
    description:  Returns the number of free pages, that is the number of
                  subsequent sucessful calls to Allocate().
    arguments:    none
    return value: SAPDB_ULong
    -----------------------------------------------------------------------*/
    SAPDB_ULong GetFreePages() const
        { return m_FreePages; }

    /*!----------------------------------------------------------------------
    function: GetErrorCount
    description:  Returns the number of errors found and corrected by the allocator

                  this allocator is not able to provide the information, i.e. returns 0
 
    return value: 0
    -----------------------------------------------------------------------*/
  
    virtual int GetErrorCount() const
    {
        return 0;
    }

private:
    /*-----------------------------------------------------------------------
    function: copy constructor
    description: Private copy constructor avoids copying.
    -----------------------------------------------------------------------*/
    SAPDBMem_PageAllocator(const SAPDBMem_PageAllocator &InitAlloc)
        : m_PageSize(InitAlloc.m_PageSize), m_FirstFreePagePtr(0)
    {}

    /*-----------------------------------------------------------------------
    operator: = 
    description: Private assignment operator avoids copying.
    -----------------------------------------------------------------------*/
    SAPDBMem_PageAllocator operator=(SAPDBMem_PageAllocator &)
        { return *this; }
    
};

/*!**************************************************************************
  endclass: SAPDBMem_PageAllocator
****************************************************************************/

#endif //SAPDBMem_PageAllocator_HPP

