/*!***************************************************************************

  module:       SADPB_ToString.hpp

  responsible : FrankH

  special area: Common

  description:  Value to String conversion i.e. for argument passing

  This class supports a set of constructors to create a string from argument.
  The number formats are named as those known from C++ IOStreams. For ease of
  use some macro definition are given, that allows 'smaller' argument line.
  This macros do not have the common 'SAPDB_' prefix, otherwise, they wouldn't
  be 'small' anymore. I.e. thne format specifier have to be used with prefix SAPDB_ToString

            noformat   = 0x0000, //
         //  internal   = 0x0001, // pad between sign and value (currently not supported)
            left       = 0x0002, // pad after value
            right      = 0x0004, // pad before value (default)
            boolalpha  = 0x0008, // use symbolic representation of true and false
            dec        = 0x0010, // integer base: base 10 output (decimal) (default)
            oct        = 0x0020, // base 8 output (octal)
            hex        = 0x0040, // base 16 output (hexadecimal)
            showbase   = 0x0080, // on output prefix oct by 0 and hex by 0x
            showpoint  = 0x0100, // print trailing zeros
            uppercase  = 0x0200, // 'E','X' rather than 'e','x'
            showpos    = 0x0400, // explicit '+' for positiv ints
            scientific = 0x0800, // floating-point notation: d.dddddddEdd
            fixed      = 0x1000, // ddd.dd
         //  skipws     = 0x2000, // to skip leading white space (currently not supported)
            ptr        = 0x8000  // Pointer

  Following default conversion formats are used

  SAPDB_Int1, SAPDB_Int2, SAPDB_Int4, SAPDB_Int8, SAPDB_Long
  SAPDB_UInt1, SAPDB_UInt2, SAPDB_UInt4, SAPDB_UInt8, SAPDB_ULong -> SAPDB_ToString;;dec

  Number in hexadecimal representation show only as much digits as they consist of.
  Example ToStr((SAPDB_Int2)-1, SAPDB_ToString::hex) -> "0xffff" !

  SAPDB_Char * -> no modification (passed)

  SAPDB_Real4,SAPDB_Real8 -> SAPDB_ToString;;noformat (_T_g) Default precision for floating point
  numbers is 6 digits.

  Any pointer (not SAPDB_Char *) -> SAPDB_ToString;;ptr, SAPDB_ToString;;showbase, SAPDB_ToString;;hex

  To specifier Outputput of an string pointer as pointer use
  
    SAPDB_ToString( stringPtr, SAPDB_ToString;;ptr, SAPDB_ToString;;showbase, SAPDB_ToString;;hex )

  ToStr() can be used instead of the ctor SAPDB_ToString()

  Since this is too long, a short form can be used

    ToStr( stringPtr, _T_p )

  last changed: 2000-07-14  17:17
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef SAPDB_TOSTRING_HPP
#define SAPDB_TOSTRING_HPP

#define MAX_CHARACTER_ALLOWED_IN_PRESENTATION 127
#define MAX_CHARACTER_ALLOWED_IN_FORMAT 40

#include "SAPDBCommon/SAPDB_Types.hpp"

/*!------------------------------------------------------------
    Declaration: Format specifier macro definitions
    Description: This is the list of predefined format specifier macro definitions
    <li>_T_b bool as text 'true' or 'false'
    <li>_T_p pointer as hexadecimal number uppercase with prefix '0X'
    <li>_T_d decimal number
    <li>_T_x hexadecimal number using lowercase for a-f with prefix '0x' 
    <li>_T_o octal number with prefix '0'
    <li>_T_e real value in scientific format, exponent 'e' lowercase
    <li>_T_f real value in fixed format
    <li>_T_g real value in scientifc or fixed format whatever suits with lowercase exponent 'e'
    <\li>
 */
#define _T_b SAPDB_ToString::boolalpha
#define _T_d SAPDB_ToString::dec
#define _T_x SAPDB_ToString::hex,SAPDB_ToString::showbase
#define _T_o SAPDB_ToString::oct,SAPDB_ToString::showbase
#define _T_e SAPDB_ToString::scientific
#define _T_f SAPDB_ToString::fixed
#define _T_g SAPDB_ToString::noformat
#define _T_p SAPDB_ToString::ptr,SAPDB_ToString::hex,SAPDB_ToString::showbase

#define ToStr SAPDB_ToString

class SAPDB_ToString
{
private:
    char m_FormatBuffer[MAX_CHARACTER_ALLOWED_IN_FORMAT+1];
    char m_PresentationBuffer[MAX_CHARACTER_ALLOWED_IN_PRESENTATION+1];
    const char *m_OutputString ;
          char *m_Buf          ;

public:
    enum fmtflags
    {
            noformat   = 0x0000, // no specific format
         //  internal   = 0x0001, // pad between sign and value (not supported)
            left       = 0x0002, // pad after value
            right      = 0x0004, // pad before value (default)
            boolalpha  = 0x0008, // use symbolic representation of true and false
            dec        = 0x0010, // integer base: base 10 output (decimal) (default)
            oct        = 0x0020, // base 8 output (octal)
            hex        = 0x0040, // base 16 output (hexadecimal)
            showbase   = 0x0080, // on output prefix oct by 0 and hex by 0x
            showpoint  = 0x0100, // print trailing zeros
            uppercase  = 0x0200, // 'E','X' rather than 'e','x'
            showpos    = 0x0400, // explicit '+' for positiv ints
            scientific = 0x0800, // floating-point notation: d.dddddddEdd
            fixed      = 0x1000, // ddd.dd
         //  skipws     = 0x2000, // to skip leading white space (not supported)
            ptr        = 0x8000  // Pointer
    } ;


  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - string only
    description: This function formats the given string

    return value: pointer to the given string (exactly the same)
    --------------------------------------------------------------------------*/

    explicit SAPDB_ToString(const char *String)
    { m_Buf = 0; m_OutputString = String ; } ;

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - string with specified flags

    supported fmtflags: SAPDB_ToString::left

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
             SAPDB_ToString(const char *String, const fmtflags )
    { m_Buf = 0; m_OutputString = String ; } ;

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - string with width and specified flags

    supported fmtflags: SAPDB_ToString::left

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
             SAPDB_ToString( const char *String,
                    const SAPDB_UInt2 Width,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat);

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - string with width and precision and specified flags

    supported fmtflags: SAPDB_ToString::left

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
             SAPDB_ToString(const char *String,
                    const SAPDB_UInt2 Width,
                    const SAPDB_UInt2 Precision,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat);

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - SAPDB_Int[1|2|4|8],SAPDB_Long with specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::showbase (default - not set)
                        SAPDB_ToString::dec (default)
                        SAPDB_ToString::oct
                        SAPDB_ToString::hex
                        SAPDB_ToString::uppercase (default - not set)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/

    explicit SAPDB_ToString(const SAPDB_Int1 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - SAPDB_Int[1|2|4|8],SAPDB_Long with width and specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::showbase (default - not set)
                        SAPDB_ToString::dec (default)
                        SAPDB_ToString::oct
                        SAPDB_ToString::hex
                        SAPDB_ToString::uppercase (default - not set)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/

             SAPDB_ToString(const SAPDB_Int1 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- UINT 1 -----------------

    explicit SAPDB_ToString(const SAPDB_UInt1 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_UInt1 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- INT 2 -----------------

    explicit SAPDB_ToString(const SAPDB_Int2 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_Int2 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- UINT 2 -----------------

    explicit SAPDB_ToString(const SAPDB_UInt2 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_UInt2 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- INT 4-----------------

    explicit SAPDB_ToString(const SAPDB_Int4 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_Int4 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- UINT 4 -----------------

    explicit SAPDB_ToString(const SAPDB_UInt4 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_UInt4 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

#if !defined (BIT64) && !defined (_WIN64)
    // ---------------- INT8 -----------------

    explicit SAPDB_ToString(const SAPDB_Int8 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_Int8 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- UINT8 -----------------

    explicit SAPDB_ToString(const SAPDB_UInt8 Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_UInt8 Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);
#endif

    // ---------------- LONG-----------------

    explicit SAPDB_ToString(const SAPDB_Long Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_Long Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

    // ---------------- ULONG-----------------

    explicit SAPDB_ToString(const SAPDB_ULong Number,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

             SAPDB_ToString(const SAPDB_ULong Number,
                    const SAPDB_UInt2 Width ,
                    const fmtflags Flag1=noformat,
                    const fmtflags Flag2=noformat,
                    const fmtflags Flag3=noformat,
                    const fmtflags Flag4=noformat);

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - SAPDB_Real[4|8] with specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::scientific
                        SAPDB_ToString::fixed
                        SAPDB_ToString::showpoint
                        SAPDB_ToString::uppercase (default - not set)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/

    explicit SAPDB_ToString(const SAPDB_Real8  Real,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat) ;

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - SAPDB_Real[4|8] with width and specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::scientific
                        SAPDB_ToString::fixed
                        SAPDB_ToString::showpoint
                        SAPDB_ToString::uppercase (default - not set)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
             SAPDB_ToString(const SAPDB_Real8  Real,
                            const SAPDB_UInt2 Width ,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat) ;

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - SAPDB_Real[4|8] with width and precision and specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::scientific
                        SAPDB_ToString::fixed
                        SAPDB_ToString::showpoint
                        SAPDB_ToString::uppercase (default - not set)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
             SAPDB_ToString(const SAPDB_Real8  Real,
                            const SAPDB_UInt2 Width ,
                            const SAPDB_UInt2 Precision,
                            const fmtflags Flag1=noformat,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat) ;

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - Pointer (address) with specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::ptr (default)
                        SAPDB_ToString::showbase (default - set)
                        SAPDB_ToString::uppercase (default)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
    explicit SAPDB_ToString(const void *Addr,
                            const fmtflags Flag1=SAPDB_ToString::showbase,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - Pointer (address) with width and specified flags

    supported fmtflags: SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::ptr (default)
                        SAPDB_ToString::showbase (default - set)
                        SAPDB_ToString::uppercase (default - not set)

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/
             SAPDB_ToString(const void *Addr,
                            const SAPDB_UInt2 Width ,
                            const fmtflags Flag1=SAPDB_ToString::showbase,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

#ifndef CPP_BOOL_NOT_IMPLEMENTED
  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - Bool with specified flags

    supported fmtflags: SAPDB_ToString::boolalpha (default "true", "false" )
                        SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::uppercase (default - not set)

    Macros: _T_b for default _T_B for uppercase strings

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/

    explicit SAPDB_ToString(const SAPDB_Bool Number,
                            const fmtflags Flag1=boolalpha,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);
  /*!-------------------------------------------------------------------------
    function: SAPDB_ToString - Bool with width and specified flags

    supported fmtflags: SAPDB_ToString::boolalpha (default "true", "false" )
                        SAPDB_ToString::left
                        SAPDB_ToString::right (default)
                        SAPDB_ToString::uppercase (default - not set)

    Macros: _T_b for default _T_B for uppercase strings

    return value: pointer to the formatted string
    --------------------------------------------------------------------------*/

             SAPDB_ToString(const SAPDB_Bool Number,
                            const SAPDB_UInt2 Width,
                            const fmtflags Flag1=boolalpha,
                            const fmtflags Flag2=noformat,
                            const fmtflags Flag3=noformat,
                            const fmtflags Flag4=noformat);

#endif /* CPP_BOOL_NOT_IMPLEMENTED */

    ~SAPDB_ToString();

    operator       char * () { return ( const_cast<char *>(m_OutputString) ) ; } ;
    operator const char * () { return ( m_OutputString ) ; } ;

private:

    void FillBoolAlphaFormatBuffer( const SAPDB_Bool isTrue,
                                    const SAPDB_UInt2 Width,
                                    const int         Flags );

    void FillNumberFormatBuffer( const SAPDB_UInt2 Width,
                                 const int         Flags,
                                 const bool        QuadWord = false);

    void FillSignedNumberFormatBuffer( const SAPDB_Int8  Number,
                                   const SAPDB_UInt2 Width,
                                   const int         Flags,
                                   const bool        QuadWord=false );

    void FillUnsignedNumberFormatBuffer( const SAPDB_UInt8  Number,
                                     const SAPDB_UInt2 Width,
                                     const int         Flags,
                                     const bool        QuadWord=false );

    void FillStringFormatBuffer( const char       *String,
                                 const SAPDB_UInt2 Width,
                                 const SAPDB_UInt2 Precision,
                                 const int         Flags=noformat );

    void FillRealFormatBuffer  ( const SAPDB_UInt2 Width,
                                 const SAPDB_UInt2 Precision,
                                 const int         Flags,
                                 const SAPDB_Real8 Number) ;
};


#endif // SAPDB_TOSTRING_HPP
