/*!*********************************************************************

  module:       SAPDBTrace_Usage.hpp

  ----------------------------------------------------------------------

  responsible:  DanielD

  special area: SAPDB Tracing

  description:  macros und utility

  ----------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

  *********************************************************************/

#ifndef SAPDBTRACE_USAGE_HPP
#define SAPDBTRACE_USAGE_HPP

#include "SAPDBCommon/Tracing/SAPDBTrace_Stream.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_MethodBrace.hpp"

#if defined (SAPDB_SLOW)
    /*! declaration: SAPDBTRACE_ROUTINE_DEBUG
        description: traces routine entry and exit

            Insert this macro as the first line of a routine:

            - <Param>name</> is the name of the routine
            - <Param>topic</> is a <Class>SAPDBTrace_Topic</> to test against
            - <Param>level</> is the level to indicate when this routine
                will be traced
     */
    #define SAPDBTRACE_ROUTINE_DEBUG(name, topic, level) \
        SAPDBTrace_MethodBrace methodBrace (name, topic, level)
    /*! enddeclaration */

    /*! declaration: SAPDBTRACE_METHOD_DEBUG
        description: traces method entry and exit

            Insert this macro as the first line of a instance method.

            Parameters are the same as SAPDBTRACE_ROUTINE_DEBUG.
            Additionally, the value of the <Param>this</> pointer
            will appear in the trace.

     */
    #define SAPDBTRACE_METHOD_DEBUG(name, topic, level) \
        SAPDBTrace_MethodBrace methodBrace (name, this, topic, level)
    /*! enddeclaration */


    /*! declaration: SAPDBTRACE_WRITELN
        description: write expressions to the trace.

            - output is supressed when level <Param>level</> is not
                switched on for <Param>topic</>
            - the output will end in a new line
            - the trace stream will be flushed

        example:
            SAPDBTRACE_WRITELN (SQLParserTrace, 5, var1 << var2);
     */
    #define SAPDBTRACE_WRITELN(topic, level, traceExpr)         \
        if (topic.TracesLevel (level)) {                         \
            SAPDBTrace_Stream stream (&topic.GetTracePrinter ()); \
            stream << traceExpr << NewLine;                      \
        }
    /*! enddeclaration */

    #define SAPDBTRACE_INDENTLN(topic, level, indent, traceExpr) \
        SAPDBTRACE_WRITELN (topic, level, SAPDB_Indent (indent) << traceExpr)
    /*! declaration: SAPDBTRACE_IF
        description: <Param>stmt</> is executed dependent on the trace level
        example:
            SAPDBTRACE_IF (SQLParserTrace, 5, HexDump (SQLParserTrace.GetTracePrinter (), buffer, sizeof (buffer));
     */
    #define SAPDBTRACE_IF(topic, level, stmt)                   \
        if (topic.TracesLevel (level)) {                             \
            stmt;                                               \
        }
        /*! enddeclaration */
#else
    #define SAPDBTRACE_ROUTINE_DEBUG(name, topic, level)
    #define SAPDBTRACE_METHOD_DEBUG(name, topic, level)
    #define SAPDBTRACE_WRITELN(topic, level, traceExpr)
    #define SAPDBTRACE_IF(topic, level, stmt)
#endif

#endif

