/*!**************************************************************************

  module      : SysView_ITableObj.hpp

  -------------------------------------------------------------------------

  responsible : UlrichJ

  special area: SystemViews
  description : This contains an abstract interface for the system view 
                table object class.

  last changed: 2002-08-13
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef SYSVIEW_ITABLEOBJ_HPP
#define SYSVIEW_ITABLEOBJ_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SystemViews/SysView_Defines.hpp"

 /*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

const bool							C_TO_UNICODE    = true;
const pasbool						C_IS_UNDEF      = true;
const pasbool						C_ZERO_AS_UNDEF = true;


struct tak40_show_glob;
struct tgg00_Rec;
//class tgg00_Surrogate;

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!***************************************************************************

   class: SysView_ITableObj

   description: This interface class provides an interface for the system view 
                table object class.

******************************************************************************/

class SysView_ITableObj
{

public:


    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : ~SysView_ITableObj()
    description : Class destructor. Is not used in the moment.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    virtual ~SysView_ITableObj()
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : GetPascalObj()
    description : Returns the underlying pascal object. Should be avoided!
    arguments   : none
    return value: [tak40_show_glob] - AKA a41v
    --------------------------------------------------------------------------*/

    virtual tak40_show_glob * GetPascalObjPtr()
    { return (NULL); }

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table creation                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function(s) : AppendCol()
    description : These functions append a column to the table object. The
                  type depends on the first parameter.
    arguments   : [iType]       - Type of the column to be added.
                  [sColName]    - Column title
                  [iColNameSize]- length of the column title string. (This just
                                  saves some time, since the title is static and
                                  we already know the strlen(name) result.)
                  [iColSize]    - Size of the column in bytes
    return value: none
    --------------------------------------------------------------------------*/

    virtual void AppendCol (SysView_ITOColType  iType,
                            const char          *sColName, 
                            int                 iColNameSize,
                            int                 iColSize)
    {}
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table creation                                            */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table access                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : GetCatalogTable()
    description : This function initialises the table catalog. Must be called
                  before you can add data to columns.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    virtual void GetCatalogTable ( void )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table access                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table information                                            */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : GetShowKind()
    description : This function returns the showkind of the table object.
    arguments   : none
    return value: [SAPDB_Int4] - showkind
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int4 GetShowKind( void )
    { return(0); }

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table information                                         */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table filling                                                */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : MoveToCol()
    description : This function has many overloads and moves a variable of
                  any kind to a table column.
    arguments   : [val] - a value of any kind.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void MoveToCol ( SysView_ITOVarType iType,
                             void               *pVal,
                             SAPDB_Int          iLength)
    {}

    /*!-------------------------------------------------------------------------
    function    : CreateLongColContainer()
    description : This function initialises a new long col container.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    virtual void CreateLongColContainer ( void )
    {}

    /*!-------------------------------------------------------------------------
    function    : MoveToLongColContainer()
    description : This function move a long record to a long col container.
    arguments   : [rec] - Reference to the record
                  [erg] - Resulting reference to the used long col container
    return value: none
    --------------------------------------------------------------------------*/

    virtual void MoveToLongColContainer ( tgg00_Rec &rec, tgg00_Surrogate &erg )
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table filling                                             */
    /* -----------------------------------------------------------------------*/

protected:

    /*!-------------------------------------------------------------------------
    function    : SysView_ITableObj()
    description : This constructor ensures that no-one will instanciate this
                  class.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

	SysView_ITableObj()
    {}


};


#endif  /* SYSVIEW_ITABLEOBJ_HPP */

