/*!**************************************************************************

  module      : SysView_PascalTableObj.hpp

  -------------------------------------------------------------------------

  responsible : UlrichJ

  special area: SystemViews
  description : This contains an abstract interface of all system view classes.

  last changed: 2002-08-13
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef SYSVIEW_PASCALTABLEOBJ_HPP
#define SYSVIEW_PASCALTABLEOBJ_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"
#include "gak40.h"
#include "vak001.h"
#include "gkb05.h"
#include "SAPDBCommon/SAPDB_UTF8Basis.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "SystemViews/SysView_Defines.hpp"
#include "SystemViews/SysView_ITableObj.hpp"

 /*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!***************************************************************************

   class: SysView_PascalTableObj

   description: This interface class provides an interface for the SystemView
                table object.

******************************************************************************/

class SysView_PascalTableObj : public SysView_ITableObj
{

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : SysView_OldPascalObj()
    description : 
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

	SysView_PascalTableObj( tak_all_command_glob	&acv,
                            tak40_show_glob         &a41v);

    /*!-------------------------------------------------------------------------
    function    : ~SysView_ISysView()
    description : Class destructor. Is not used in the moment.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    ~SysView_PascalTableObj()
    {}


    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : GetPascalObj()
    description : Returns the underlying pascal object. Should be avoided!
    arguments   : none
    return value: [tak40_show_glob] - AKA a41v
    --------------------------------------------------------------------------*/

    tak40_show_glob * GetPascalObjPtr();

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table creation                                               */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function(s) : AppendCol()
    description : This function append a column to the table object. The
                  type depends on the first parameter.
    arguments   : [iType]       - Type of the column to be added.
                  [sColName]    - Column title
                  [iColNameSize]- length of the column title string. (This just
                                  saves some time, since the title is static and
                                  we already know the strlen(name) result.)
                  [iColSize]    - Size of the column in bytes
    return value: none
    --------------------------------------------------------------------------*/

    void AppendCol (SysView_ITOColType  iType,
                    const char          *sColName, 
                    int                 iColNameSize,
                    int                 iColSize);
    
    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table creation                                            */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table access                                                 */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : GetCatalogTable()
    description : This function initialises the table catalog. Must be called
                  before you can add data to columns.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    void GetCatalogTable ( void );


    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table access                                              */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table information                                            */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : GetShowKind()
    description : This function returns the showkind of the table object.
    arguments   : [ccode] - new result code
    return value: none
    --------------------------------------------------------------------------*/

    SAPDB_Int4 GetShowKind( void );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table information                                         */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Table filling                                                */
    /* -----------------------------------------------------------------------*/

    void MoveToCol ( SysView_ITOVarType iType,
                     void               *pVal,
                     SAPDB_Int          iLength);

    /*!-------------------------------------------------------------------------
    function    : CreateLongColContainer()
    description : This function initialises a new long col container.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    void CreateLongColContainer ( void );

    /*!-------------------------------------------------------------------------
    function    : MoveToLongColContainer()
    description : This function move a long record to a long col container.
    arguments   : [rec] - Reference to the record
                  [erg] - Resulting reference to the used long col container
    return value: none
    --------------------------------------------------------------------------*/

    void MoveToLongColContainer ( tgg00_Rec         &rec, 
                                  tgg00_Surrogate   &erg );

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Table filling                                             */
    /* -----------------------------------------------------------------------*/

private:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Private tool functions                                       */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : MoveToCol()
    description : This function has many overloads and moves a variable of
                  any kind to a table column.
    arguments   : [val] - a value of any kind.
    return value: none
    --------------------------------------------------------------------------*/

    void MoveInt8 ( SAPDB_Int8 &val );

    void MoveUTF8Ptr ( SAPDB_UTF8 *val,
                       SAPDB_Int  length );

    void MoveTransNo ( tgg91_TransNo &val );

    void MoveTransLockMode ( tkb05_TransLockMode_Enum &val );

    int HandleUTF8Error(SAPDB_UTF8Basis::ConversionResult &res);

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Private tool functions                                    */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Local Variables                                              */
    /* -----------------------------------------------------------------------*/

    tak_all_command_glob	&m_ACV;
    tak40_show_glob         &m_Show;

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Local Variables                                           */
    /* -----------------------------------------------------------------------*/

};


#endif  /* SYSVIEW_PASCALTABLEOBJ_HPP */

