/*!**************************************************************************

  module      : SysView_SysViewBase.hpp

  -------------------------------------------------------------------------

  responsible : UlrichJ

  special area: SystemViews
  description : This contains the basic functionality for all system views.

  last changed: 2002-08-13
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin LGPL
  Copyright (C) 2001 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end

*****************************************************************************/


#ifndef SYSVIEW_SYSVIEWBASE_HPP
#define SYSVIEW_SYSVIEWBASE_HPP



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SystemViews/SysView_ISysView.hpp"
#include "SystemViews/SysView_ITableObj.hpp"
#include "SystemViews/SysView_ISessionObj.hpp"

 /*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

struct tak_all_command_glob;
struct tak40_show_glob;

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!***************************************************************************

   class: SysView_SysViewBase

   description: This class is designed to be an interface class for all system
                views. However, this class is a bit more than that, since it
                also provides some additional functions which are useful for
                all derived instances.
                Note: This class should be the only one being used from outside.

******************************************************************************/

class SysView_SysViewBase : public SysView_ISysView
{

public:

    /* -----------------------------------------------------------------------*/
    /*! chapter: Constructors and initialization                              */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : SysView_ISysView()
    description : The only real existing constructor (all derived classes just 
                  have overload constructors which call this one again) which 
                  just sets some local variables.
    arguments   : [acv] - a reference to the globally used acv structure.
    return value: none
    --------------------------------------------------------------------------*/

	SysView_SysViewBase( tak_all_command_glob	&acv,
                         tak40_show_glob       &a41v);

    /*!-------------------------------------------------------------------------
    function    : ~SysView_ISysView()
    description : Class destructor. Is not used in the moment.
    arguments   : none
    return value: none
    --------------------------------------------------------------------------*/

    ~SysView_SysViewBase();

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Constructors and initialization                           */
    /* -----------------------------------------------------------------------*/

    /* -----------------------------------------------------------------------*/
    /*! chapter: Virtual functions                                            */
    /* -----------------------------------------------------------------------*/

    /*!-------------------------------------------------------------------------
    function    : Ceate()
    description : Creates columns. Has to be overloaded by its derived class.
    arguments   : [a41v] - a reference to the table object.
    return value: none
    --------------------------------------------------------------------------*/

    virtual void	Create (void)
    {}

    /*!-------------------------------------------------------------------------
    function    : GetColCount()
    description : Returns no. of columns. Has to be overloaded by its derived 
                  class.
    arguments   : none
    return value: [int] - always 0 (dummy)
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int   GetColCount (void)
    { return(0); }

    /*!-------------------------------------------------------------------------
    function    : EstimateRows()
    description : Returns the estimated number of rows for this table. Has to be 
                  overloaded by its derived class.
    arguments   : none
    return value: [int] - always 0 (dummy)
    --------------------------------------------------------------------------*/

    virtual SAPDB_Int   EstimateRows (void)
    { return(0); }

    /*!-------------------------------------------------------------------------
    function    : Execute()
    description : Fills this table with information. Has to be overloaded by its 
                  derived class.
    arguments   : [a41v] - reference to the pascal table object
    return value: none
    --------------------------------------------------------------------------*/

    virtual void	Execute (void)
    {}

    /* -----------------------------------------------------------------------*/
    /*! endchapter: Virtual functions                                         */
    /* -----------------------------------------------------------------------*/

protected:


    SysView_ITableObj       *m_Table;
    SysView_ISessionObj     *m_Session;

};


#endif  /* SYSVIEW_SYSVIEWBASE_HPP */

