/*!**************************************************************************

  module      : Trace_Manager.hpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area: Interface of the kernel trace manager.
  description : This class defines the singleton component trace manager and 
                offers the possibility to insert trace entries into the UKT 
                specific memory trace buffers. Furthermore it's possible to
                write the materialized trace entries into the kernel trace 
                file.
                The writeable trace entries are derived from the abstract
                classes Trace_IBaseEntry and Trace_ILargeEntry found in 
                Trace_Types.hpp. The concretely implementations of the trace
                entries are in Trace_[A-Z*]Entry.hpp, e.g.Trace_PacketEntry.hpp.

  last changed: 2000-11-27  18:15
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef TRACE__MANAGER_HPP
#define TRACE__MANAGER_HPP


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"   // PASCAL: SP_basic_constants_and_types
#include "ggg00.h"   // PASCAL: GG_kernel_constants_and_types

#include "KernelCommon/Kernel_WaitQueue.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "Trace/Trace_DevSpace.hpp"
#include "Trace/Trace_Types.hpp"
#include "Trace/Trace_SlotPool.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!*****************************************************************************

   class: Trace_Manager

   description: This class is the trace interface.
   
              
*******************************************************************************/

class Trace_Manager
{
public:

    /*!-------------------------------------------------------------------------
    function:     CreateInstance
    description:  This method creates the singleton trace manager.It is assumed 
                  that this method will be callled only one time by the timeout 
                  task and therefore no synchonization mechanism is used.
    arguments:    TaskId [in] current task id
    return value: true means instance could be created; else false
    --------------------------------------------------------------------------*/

    static bool CreateInstance( tsp00_TaskId TaskId );
    
    /*!-------------------------------------------------------------------------
    function:     GetInstance
    description:  This method returns the access to the trace manager. It is 
                  assumed that the trace manager singleton is available! No
                  CreateInstance call is will be done.
    arguments:    TaskId [in] current task id
    return value: Access to the trace manager singleton
    --------------------------------------------------------------------------*/

    static Trace_Manager& GetInstance(){return( *m_Instance );};

    /*!-------------------------------------------------------------------------
    function:     Flush
    description:  This method is used to flush the memory trace buffers into
                  the kernel trace file. The flush will be done by the trace 
                  writer task, triggered by the diagnose vtrace flush command,
                  the diagnose clear vtrace command or a kernel abort.
    arguments:    TaskId [in] current task id
    return value: none
    --------------------------------------------------------------------------*/

    void Flush( tsp00_TaskId TaskId );

    /*!-------------------------------------------------------------------------
    function:     GetDumpFile
    description:  This method provides the name of the kernel dump file. It is
                  used by the trace writer to write the memory structures, e.g.
                  data cache into the kernel dump file. During the flush of an
                  UKT specific trace buffer, it is locked via a spinlock, so that
                  no other task within the same UKT is able to write trace entries
                  into the trace buffer.
    arguments:    none
    return value: Name of the kernel dump file
    --------------------------------------------------------------------------*/

    tsp00_VFilename GetDumpFile();
    
    /*!-------------------------------------------------------------------------
    function:     Insert
    description:  This method is used to write a trace entry smaller than 132
                  bytes into the kernel trace. Trace entries smaller than 0 and
                  and greater than 132 won't be inserted. No errors are reported.
    arguments:    Trans [in] current transaction context
                  Entry [in] trace entry handler for plain entries
    return value: none
    --------------------------------------------------------------------------*/

    void Insert( 
        tgg00_TransContext  &Trans,
        Trace_IBaseEntry    &Entry );

    /*!-------------------------------------------------------------------------
    function:     InsertLarge
    description:  This method is used to write a large trace entry into the kernel 
                  trace. Trace entries smaller than 0 won't be inserted. No errors 
                  are reported. The trace entry will be splitted into separate parts.
                  Each part has the layout of an ordinary trace entry, but inside
                  the info part it contains a internal header. This internal header
                  is used by the diagnose tool the combine the splitted parts to one
                  trace entry.
    arguments:    Trans [in] current transaction context
                  Entry [in] trace entry handler for huge entries.
    return value: none
    --------------------------------------------------------------------------*/

    void InsertLarge( 
        tgg00_TransContext  &Trans,
        Trace_ILargeEntry   &Entry );

    /*!-------------------------------------------------------------------------
    function:     PrepareAndWaitForFlush
    description:  This method is used to force the flushing of the memory trace 
                  buffers into the kernel trace file. The calling user task will
                  be suspended as long as the trace is not flushed by the trace 
                  writer task. The flush order is locked via the trace region to
                  avoid parallel access. After the complete flush the waiting task
                  will be resumed by the trace writer task. This method furthermore
                  offers the possibility to reinitialize the kernel trace file. This
                  means that all information will be lost!
    arguments:    TaskId [in] current task id
                  bReInitialization [in] true means to reinitialize the trace file.
    return value: none
    --------------------------------------------------------------------------*/

    void PrepareAndWaitForFlush( 
        tsp00_TaskId    TaskId,
        SAPDB_Bool      bReInitialization = false );
   
    /*!-------------------------------------------------------------------------
    function:     TraceWriterRegistration
    description:  This method is used one time by the trace writer task to register 
                  themself by the trace manager component after the own creation.
    arguments:    TaskId [in] current task id
    return value: none
    --------------------------------------------------------------------------*/

    void TraceWriterRegistration( tsp00_TaskId TaskId );
   
private:
    
    Trace_Manager ( SAPDB_Int4  NumUKTs );
    
    bool Initialize( tsp00_TaskId TaskId );
    
    bool WriteEntryPossible( 
        tgg00_TransContext   &Trans,
        SAPDB_Int4           EntrySize );

private:

    static  Trace_Manager                   *m_Instance;

    bool                                    m_TraceWriterRegistered;
    SAPDB_Int4                              m_NumUKTs;
    SAPDB_UInt4                             m_TraceWriterRequested; // PTS 1112933 TS 2001-12-04
    Kernel_WaitStack                        m_UserWaitFlushQueue;   // PTS 1112933 TS 2001-12-04
    Trace_SlotPool                          m_SlotPool;
    Trace_DevSpace                          m_DevSpace;
    RTESync_InterlockedCounter<Trace_SeqNo> m_SeqNoGenerator;
    tsp00_VFilename                         m_DumpFile;
};

/*!*****************************************************************************

   endclass: Trace_Manager

*******************************************************************************/


#endif  /* TRACE__MANAGER_HPP */
