/****************************************************************************

  module      : Trace_Slot.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  special area:
  description : 

  last changed: 2000-11-27  18:15
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg90.h"  // CPP   : GG_kernel_templates
#include "hgg01.h"  // PASCAL: Configuration
#include "hsp77.h"

#include "Trace/Trace_Slot.hpp"
#include "Trace/Trace_Entry.hpp"
#include "Trace/Trace_InfoPage.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define TRUE    1

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

Trace_Slot::Trace_Slot()
:
m_UKTId( 0 ),
m_NumPages( 0 ),
m_UsableSize( 0 ),
m_PageListOffset( 0 ),
m_PageNo( TRACE_FIRST_PNO ),
m_pPageList( NULL ),
m_SpinLock(( SAPDB_UTF8* )"Trace_Slot::WriteEntry" )
{
};

/*---------------------------------------------------------------------------*/

bool
Trace_Slot::Initialize( SAPDB_Int4 UKTId )
{
    if ( !Trace_InfoPage::GetUktTraceBuffer( UKTId, m_pPageList, m_NumPages) )
    {
        return ( false );
    }

    m_UKTId = UKTId;

    SAPDB_UTF8 Identifier[40];
    sp77sprintf((char *)Identifier, sizeof(Identifier), "TraceSlot::UKT%03d", m_UKTId);
    m_SpinLock.SetIdentifier( Identifier );

    if( 0 < m_NumPages )
    {
        m_Page.AssignNewPage( m_pPageList, m_PageNo );
        m_UsableSize = m_NumPages * m_Page.BodySize();
    }
    else
        m_UsableSize = 0;
    
    return( true );
};

/*---------------------------------------------------------------------------*/

void
Trace_Slot::Flush(
                  tsp00_TaskId      TaskId, 
                  Trace_DevSpace    &DevSpace )
{
    if( 0 >= m_UsableSize ) // PTS 1120308 TS 2003-01-24
        return;

    RTESync_LockedScope Lock(m_SpinLock);
    

    DevSpace.Write( TaskId, m_UKTId, m_pPageList, m_NumPages );
};

/*---------------------------------------------------------------------------*/

void 
Trace_Slot::Insert(
                   tgg00_TransContext                       &Trans,
                   SAPDB_Int4                               UKTId,
                   Trace_IBaseEntry                         &Entry,
                   RTESync_InterlockedCounter<Trace_SeqNo>  &SeqNoGenerator )
{
    if ( m_UsableSize < Entry.GetSize() ) return;

    RTESync_LockedScope Lock(m_SpinLock);
    
    // Get unique number for the current trace entry
    
    Trace_SeqNo SequenceNo = 0;
    SeqNoGenerator.Increment( 1, &SequenceNo );
    
    Trace_KernelFullEntry   FullEntry( Trans, Entry, SequenceNo );
    
    InsertIntoTracePage( FullEntry );
};

/*---------------------------------------------------------------------------*/

void 
Trace_Slot::InsertLarge(
                        tgg00_TransContext                       &Trans,
                        SAPDB_Int4                               UKTId,
                        Trace_ILargeEntry                        &Entry,
                        RTESync_InterlockedCounter<Trace_SeqNo>  &SeqNoGenerator )
{
    if ( m_UsableSize < Entry.GetSize() ) return;

    RTESync_LockedScope Lock(m_SpinLock);
  
    // Get unique number(s) for the current trace entry

    Trace_SeqNo SequenceNo = 0;
    
    SeqNoGenerator.Increment( Entry.GetNumberOfParts(), &SequenceNo );
    
    Trace_LargeEntryIterator Iter( Entry );
    
    for( Iter.Begin(); !Iter.End(); ++Iter )
    {
        Trace_OverflowEntry OverflowEntry( Trans, (tgg00_Debug::fromConst( bd_oflw )),
            Entry.GetLayer(), Entry.GetType(), Iter.GetPart(), Iter.GetPartSize(),
            Iter.GetPartNo() );
        
        Trace_KernelFullEntry   FullEntry( Trans, OverflowEntry, SequenceNo + Iter.GetPartNo() -1 );
        
        InsertIntoTracePage( FullEntry );
    }
};

/*---------------------------------------------------------------------------*/

void 
Trace_Slot::ReInitialization()
{
    m_PageListOffset = 0;
    m_PageNo         = TRACE_FIRST_PNO;

    if( 0 >= m_UsableSize ) // PTS 1120308 TS 2003-01-24
        return;

    RTESync_LockedScope Lock(m_SpinLock);


    for( SAPDB_Int4 PageCount = 0; PageCount < m_NumPages; ++PageCount)
    {
        m_Page.AssignNewPageAndFormat( (m_pPageList + PageCount), TRACE_NIL_PNO );
    };

    // Set trace page handler onto the first trace page
    m_Page.AssignNewPage( m_pPageList, m_PageNo );
};

/*---------------------------------------------------------------------------*/

void
Trace_Slot::InsertIntoTracePage( Trace_IFullEntry &FullEntry )
{
    if (! m_Page.IsSpaceAvailable( FullEntry.GetSize() ))
    {
        // Prepare next trace page to store current trace entry
        
        if (m_PageListOffset < ( m_NumPages-1 ))
        {
            m_Page.AssignNewPage(( m_pPageList + ( ++m_PageListOffset )), ++m_PageNo );
        }
        else
        {
            // The end of the trace buffers is reached. Next trace page is the
            // first allocated page
            m_PageListOffset = 0;
            m_Page.AssignNewPage( m_pPageList, ++m_PageNo );
        }
    }
    m_Page.Insert( FullEntry );
};

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
