/****************************************************************************

  module      : LVC_KernelInterfaceInstance.cpp

  -------------------------------------------------------------------------

  responsible : ThomasA

  special area: liveCache  
  description : kernel interface provided for OMS


  last changed: 2000-07-11  13:54
  see also    : example.html ...
  first created:2000-05-26  19:32

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#include "liveCache/LVC_KernelInterfaceInstance.hpp"
#include "liveCache/MemoryManagement/LVCMem_Allocator.hpp"
#include "liveCache/MemoryManagement/LVCMem_Wrapper.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "SAPDB/RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"
#include "KernelCommon/Kernel_OpMsg.hpp"
#include "Oms/OMS_StackTraceCallbackInterface.hpp"

#include "vak001.h"
#include "hgg01.h"
#include "hgg01_3.h"
#include "heo52.h"  // vclock
#include "heo56.h"  // vsignal
#include "heo58.h"  // vGetAcvPtrFromCurrentTask
#include "geo573.h" // bad alloc
#include "hgg11.h"
#include "hgg01.h"
#include "heo51.h"  // vgetpid
#include "heo670.h"

LVC_KernelInterfaceInstance* LVC_KernelInterfaceInstance::m_Instance = 0;
int                          LVC_KernelInterfaceInstance::m_omsVersionUnloads = 0;

/*-----------------------------------------------------*/

LVC_KernelInterface& LVC_KernelInterfaceInstance::Instance(bool isLiveCache)
{
  SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(LVC_KernelInterfaceInstance));
  if (0 == m_Instance)
  {
    m_Instance = new(Space) LVC_KernelInterfaceInstance(isLiveCache);
  }
  return *m_Instance;
}

/*-----------------------------------------------------*/
// New with PTS 1124170
LVC_KernelInterfaceInstance::LVC_KernelInterfaceInstance(bool isLiveCache)
  : m_pSpinlockPool(NULL)
  , m_ppRWLocks(NULL)
  , m_isLiveCacheInstance(isLiveCache)
{
  if (isLiveCache){
    {
      SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTESync_SpinlockPool));
      m_pSpinlockPool = new(Space) RTESync_SpinlockPool((const SAPDB_UTF8 *) "OMS R/W locks", LVC_LockRequest::MAX_RW_SPINLOCKS);
    }

    {
      SAPDBMEM_STATIC_RAW_ALLOCATE(Space, LVC_LockRequest::MAX_RW_REGION * sizeof(RTESync_IRWRegion));
      m_ppRWLocks = reinterpret_cast<RTESync_IRWRegion**>(Space); 
      memset(m_ppRWLocks, 0, LVC_LockRequest::MAX_RW_REGION * sizeof(void*));
    }
  }
}

/*-----------------------------------------------------*/

void* LVC_KernelInterfaceInstance::Allocate(SAPDB_ULong count)
{
    return TheLiveCacheAllocator.AllocateImplementation (count);
}

/*-----------------------------------------------------*/

void* LVC_KernelInterfaceInstance::AllocateArray(SAPDB_ULong count)
{
  return TheLiveCacheAllocator.AllocateImplementation (count);
}

/*-----------------------------------------------------*/

tsp00_Int4 LVC_KernelInterfaceInstance::Clock()
{
  tsp00_Int4 Sec;
  tsp00_Int4 MSec; 
  vclock(&Sec, &MSec); 
  return Sec;
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::Deallocate(void* p)
{
    TheLiveCacheAllocator.DeallocateImplementation (p);
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::DeallocateArray(void* p)
{
  TheLiveCacheAllocator.DeallocateImplementation (p);
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::DeregisterAllocator(RTEMem_AllocatorInfo& allocatorInfo)
{
    RTEMem_AllocatorRegister::Instance().Deregister(allocatorInfo); 
}

/*-----------------------------------------------------*/

SAPDBMem_IRawAllocator* LVC_KernelInterfaceInstance::GetOmsAllocator()
{
  return &TheLiveCacheOmsAllocator; 
}

/*-----------------------------------------------------*/

bool LVC_KernelInterfaceInstance::HeapThresholdReached()
{
  SAPDB_ULong BytesUsed;
  SAPDB_ULong BytesControlled;
  if (0 == g01oms_heap_limit())
  {
	  return false;
  }
  TheLiveCacheAllocator.CalcStatistics(BytesUsed, BytesControlled);
  double percent = (double) BytesUsed * 100.0 / (double) (g01oms_heap_limit() * 1024);
  return(percent > g01oms_heap_usage_threshold());
}

/*-----------------------------------------------------*/

void* LVC_KernelInterfaceInstance::GetKnlVersion(void* pBuf, int bufSize)
{
  if (bufSize >= sizeof(tsp_version) + 1)
  {
    g11kernel_version(*REINTERPRET_CAST(tsp_version*, pBuf));
    char* pVersion = REINTERPRET_CAST(char*, pBuf) + sizeof(tsp_version);
    do
      --pVersion;
    while (' ' == *pVersion);
    *pVersion = 0;
    return pBuf;
  }
  return NULL;
}

/*-----------------------------------------------------*/

tsp00_Int4 LVC_KernelInterfaceInstance::GetLockRequestTimeout()
{
  return g01timeout.timeReq_gg00;
}

/*-----------------------------------------------------*/

int LVC_KernelInterfaceInstance::GetOmsRegionCnt()
{
  return g01region_cnt(rgnOMS_egg00);
}

/*-----------------------------------------------------*/

IliveCacheSink* LVC_KernelInterfaceInstance::GetSinkPtr()
{
   tak_all_command_glob* pAcv = REINTERPRET_CAST(tak_all_command_glob*, vGetAcvPtrFromCurrentTask());
   return REINTERPRET_CAST(IliveCacheSink*, &pAcv->a_kernel_sink);
}

/*-----------------------------------------------------*/

tsp00_Uint4 LVC_KernelInterfaceInstance::GetOmsVersionThreshold()
{
  return g01oms_vers_free_threshold();
}

/*-----------------------------------------------------*/

tsp00_Int4 LVC_KernelInterfaceInstance::GetPacketMaxUsableArea()
{
  return g01packet_size() - sizeof (tsp1_packet_header) -
    sizeof (tsp1_segment_header) - sizeof (tsp1_part_header);
}

/*-----------------------------------------------------*/
   
int LVC_KernelInterfaceInstance::GetOmsVersionUnloads()
{
    return m_Instance->m_omsVersionUnloads;
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::IncOmsVersionUnloadCounter()
{
  ++m_omsVersionUnloads;
}

/*-----------------------------------------------------*/

bool LVC_KernelInterfaceInstance::IsUnicodeInstance()
{
    return ( g01unicode != 0 );
}

/*-----------------------------------------------------*/

void* LVC_KernelInterfaceInstance::OmsAllocate(SAPDB_ULong count)
{
    return TheLiveCacheAllocator.OmsAllocate (count);
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::OmsDeallocate(void* p)
{
    TheLiveCacheAllocator.OmsDeallocate (p);
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::RegisterAllocator(RTEMem_AllocatorInfo& allocatorInfo)
{
    RTEMem_AllocatorRegister::Instance().Register(allocatorInfo); 
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::Signal(tsp00_TaskId taskId, bool timeout)
{
    vsignal(taskId, !timeout);
}

/*-----------------------------------------------------*/

void  LVC_KernelInterfaceInstance::ResetOmsVersionUnloadCounter()
{
    m_Instance->m_omsVersionUnloads = 0;
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::SignalUnloadLiboms()
{
    TheLiveCacheAllocator.DisableClientInquiry();
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::TestBadAlloc()
{
    if (LVCMem_Wrapper::BadAllocRequired(vGetAcvPtrFromCurrentTask()))
    {
        THROW_BAD_ALLOC_GEO573;
    }
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::SimCtlAttachSink(OmsHandle &handle)
{
}


/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::ConsoleMessage(const char *msg)
{
  	Kernel_OpInfo(csp3_n_obj) << msg;
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::DumpStackTrace()
{
  eo670_CTraceStack();
}

/*-----------------------------------------------------*/

unsigned int LVC_KernelInterfaceInstance::SaveStackTrace(
  void **buffer, unsigned int levels)
{
  unsigned int saved = eo670_FillCallArray(levels, buffer);
  levels -= saved;
  buffer += saved;
  while (levels--) {
    *(buffer++) = NULL;
  }
  return saved;
}

/*-----------------------------------------------------*/

void LVC_KernelInterfaceInstance::EvalStackTrace(
  void **buffer, unsigned int levels,
  OmsStackTraceCallbackInterface *cb)
{
  void *context = NULL;
  if (levels && eo670_GetAddressInfoProlog(&context)) {
    eo670AddressInfo addrInfo;
    while (levels--) {
      if (eo670_GetAddressInfo(*buffer, &addrInfo, context)) {
        cb->omsStackTrace(*buffer, addrInfo.symbol, 
          addrInfo.demangledSymbol,
          addrInfo.offsetToSymbol, addrInfo.moduleName,
          addrInfo.sourceFileName, addrInfo.lineNumber);
        eo670_FreeAddressInfo(&addrInfo);
      } else {
        cb->omsStackTrace(*buffer, "unknown", "unknown", 0,
          "unknown", "", -1);
      }
      ++buffer;
    }
    eo670_GetAddressInfoEpilog(context);
  }
}

/*-----------------------------------------------------*/

short LVC_KernelInterfaceInstance::LockRequest(LVC_LockRequest &lockReq)
{
  // TODO: proper error codes

  // R/W locks are only initialized if instance is a livecache (see constructor)
  if (!m_isLiveCacheInstance){
    if (lockReq.m_action == LVC_LockRequest::RWLOCK_CREATE){
      // ignore the construction
      return e_ok;
    }
    else {
      // return error if a lock is really requested
      return e_invalid_command;
    }
  }

  RTESync_IRWRegion *lock;
  if (lockReq.m_action <= LVC_LockRequest::RWLOCK_IS_LOCKED) {
    // R/W lock action, check ID and get proper lock
    if (lockReq.m_id < 0 || lockReq.m_id >= LVC_LockRequest::MAX_RW_REGION) {
      return e_invalid_parameter;
    }
    lock = m_ppRWLocks[lockReq.m_id];
    if (lock == NULL && lockReq.m_action != LVC_LockRequest::RWLOCK_CREATE) {
      return e_invalid_parameter;
    }
  }
  else {
    return e_invalid_command;
  }

  switch (lockReq.m_action) {
  case LVC_LockRequest::RWLOCK_CREATE:
    if (lock == NULL) {
      // TODO: synchronize
      m_ppRWLocks[lockReq.m_id] = RTESync_CreateRWRegion(lockReq.m_id, *m_pSpinlockPool,
        RTEMem_RteAllocator::Instance());
      if (m_ppRWLocks[lockReq.m_id] == NULL) {
        return e_no_more_memory;
      }
    }
    return 0;
  case LVC_LockRequest::RWLOCK_LOCK_EXCLUSIVE:
    lock->enter(true);
    return 0;
  case LVC_LockRequest::RWLOCK_LOCK_SHARED:
    lock->enter(false);
    return 0;
  case LVC_LockRequest::RWLOCK_TRYLOCK_EXCLUSIVE:
    if (lock->tryEnter(true))
      return 0;
    else
      return e_request_timeout;
  case LVC_LockRequest::RWLOCK_TRYLOCK_SHARED:
    if (lock->tryEnter(false))
      return 0;
    else
      return e_request_timeout;
  case LVC_LockRequest::RWLOCK_UNLOCK_EXCLUSIVE:
    lock->leave(true);
    return 0;
  case LVC_LockRequest::RWLOCK_UNLOCK_SHARED:
    lock->leave(false);
    return 0;
  case LVC_LockRequest::RWLOCK_IS_LOCKED:  // ONLY FOR DEBUGGING
    {
      tsp00_TaskId ownpid; 
      vgetpid(ownpid);
      short pid = lock->getLastEntered();
      if (pid == -1)
        return -1;
      else if (pid == ownpid)
        return 0;
      else
        return pid;
    }
  default:
    return e_invalid_command;
  }
}

/*-----------------------------------------------------*/

bool LVC_KernelInterfaceInstance::UseReaderWriterLocks(){
  return g01cnstview_rw_sync() ? true : false;
}
