/*!***************************************************************************

  module      : LVC_KernelInterfaceInstance.hpp

  -------------------------------------------------------------------------

  responsible : ThomasA

  special area: OMS
  description : context free kernel interface to liboms.

  last changed: 2000-12-06  19:04
  see also    : example.html ...
  first created:2000-03-09  18:21

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef LVC_KERNELINTERFACEINSTANCE_HPP
#define LVC_KERNELINTERFACEINSTANCE_HPP

#include "liveCache/LVC_KernelInterface.hpp"
#include "liveCache/LVC_LockRequest.hpp"
#include "RunTime/Synchronisation/RTESync_RWRegion.hpp"


class LVC_KernelInterfaceInstance : public LVC_KernelInterface
{
  public :
    LVC_KernelInterfaceInstance(bool isLiveCache);
   
    static LVC_KernelInterface& Instance(bool isLiveCache);

    /*!-------------------------------------------------------------------------
    function:     Allocate
	  description:  Allocates contiguous memory for 'count' bytes from LVCMem_Allocator.

    arguments:    count [in] number of requested bytes

    
  return value: pointer to memory chunk
  --------------------------------------------------------------------------*/
  
  virtual void* Allocate(SAPDB_ULong count);
  
  virtual tsp00_Int4 Clock();
  
  /*!-------------------------------------------------------------------------
  function:     Deallocate
  description:  Deallocates contiguous memory identified by p from LVCMem_Allocator.
  
  arguments:    p [in] memory to be deallocated
    
  return value: none
  --------------------------------------------------------------------------*/
  
  virtual void Deallocate(void* p);

  /*!-------------------------------------------------------------------------
	function:     DeregisterAllocator
	description:  deregisters an allocator in the kernel allocator registry.  
		   
	arguments:    allocatorInfo [in] a structure containing allocator info.
	return value: none
   --------------------------------------------------------------------------*/

  virtual void DeregisterAllocator(RTEMem_AllocatorInfo& allocatorInfo);

  /*!-------------------------------------------------------------------------
  function:     GetOmsAllocator
  description:  returns a pointer to the global liveCache oms allocator 
  
  arguments:    none
  return value: pointer to the liveCache oms allocator
  --------------------------------------------------------------------------*/
  
  virtual SAPDBMem_IRawAllocator* GetOmsAllocator();
  
  /*!-------------------------------------------------------------------------
  function:     GetKnlVersion
  description:  fills a buffer with liveCache version information 
  
  arguments:    pBuffer [out] address of the buffer to be filled
  bufferLength [in] the size of the buffer
  return value: pointer to the buffer, if successful, NULL otherwise
  --------------------------------------------------------------------------*/

  virtual void* GetKnlVersion(void*,int);

  /*!-------------------------------------------------------------------------
	function:     GetLockRequestTimeout
	description:  yields the value of the parameter REQUEST_TIMEOUT 
		   
	arguments:    none
	return value: the request timeout value
  --------------------------------------------------------------------------*/

  virtual tsp00_Int4 GetLockRequestTimeout();
 
  /*!-------------------------------------------------------------------------
	function:     GetOmsRegionCnt
	description:  yields the value of the parameter _OMS_REGIONS 
		   
	arguments:    none
	return value: the value of oms regions
  --------------------------------------------------------------------------*/

  virtual int   GetOmsRegionCnt();

  /*!-------------------------------------------------------------------------
	function:     GetPacketMaxUsableArea
	description:  yields the maximal usable area of the packet, i.e. the
                number of bytes, that can be transferred via a packet
                containing one data part.
		   
	arguments:    none
	return value: the number of bytes, that can be transferred via a packet
                containing one data part.
  --------------------------------------------------------------------------*/

  virtual tsp00_Int4 GetPacketMaxUsableArea();

  /*!-------------------------------------------------------------------------
	function:     GetSinkPtr
	description:  yields the address of the current kernel sink, i.e. the
                kernel callback of the current session.
		   
	arguments:    none
	return value: the address of the kernel sink
  --------------------------------------------------------------------------*/

  virtual IliveCacheSink* GetSinkPtr();

  /*!-------------------------------------------------------------------------
	function:     GetOmsVersionThreshold
	description:  yields the value of the parameter OMS_VERS_THRESHOLD 
		   
	arguments:    none
	return value: the oms version threshold value
  --------------------------------------------------------------------------*/

  virtual tsp00_Uint4 GetOmsVersionThreshold();

  /*!-------------------------------------------------------------------------
	function:     GetOmsVersionUnloads
	description:  returns the number of oms version unloads 
		   
	arguments:    none
	return value: the number of oms version unloads
  --------------------------------------------------------------------------*/

  static  int   GetOmsVersionUnloads();

  /*!-------------------------------------------------------------------------
	function:     HeapThresholdReached
	description:  returns true, if the heap threshold has been reached, i.e.
                OMS_HEAP_THRESHOLD percent of the heap are in use 
		   
	arguments:    none
	return value: true, if threshold value has been reached, false otherwise
  --------------------------------------------------------------------------*/
  
  virtual bool  HeapThresholdReached();

  /*!-------------------------------------------------------------------------
    function:     IncOmsVersionUnloadCounter
    description:  increments the counter counting number of oms version unloads
   
    arguments:    none
     
    return value: none
   --------------------------------------------------------------------------*/
 
  virtual void IncOmsVersionUnloadCounter();

  /*!-----------------------------------------------------------------------
   function:     IsUnicodeInstance
   description:  tells whether liveCache is runnung in UNICODE mode
   arguments:    none
   return value: true, if parameter _UNICODE='YES', otherwise false
   -------------------------------------------------------------------------*/  
  
  virtual bool IsUnicodeInstance();

  /*!-------------------------------------------------------------------------
   function:     OmsAllocate
   description:  Allocates contiguous memory for 'count' bytes from oms allocator.
  
   arguments:    count [in] number of requested bytes
    
   return value: pointer to memory chunk
  --------------------------------------------------------------------------*/

  virtual void* OmsAllocate(SAPDB_ULong count);

  /*!-------------------------------------------------------------------------
   function:     Deallocate
   description:  Deallocates contiguous memory identified by p from LVCMem_Allocator.
  
   arguments:    p [in] memory to be deallocated
    
   return value: none
  --------------------------------------------------------------------------*/
  
  virtual void OmsDeallocate(void* p);

  /*!-------------------------------------------------------------------------
	function:     RegisterAllocator
	description:  registers an allocator in the kernel allocator registry.  
		   
	arguments:    allocatorInfo [in] a structure containing allocator info.
	return value: none
   --------------------------------------------------------------------------*/

  virtual void RegisterAllocator(RTEMem_AllocatorInfo& allocatorInfo);  

  /*!-------------------------------------------------------------------------
    function:     ResetOmsVersionUnloadCounter
    description:  sets m_omsVersionUnloads to 0.  
 		   
    arguments:    none.
    return value: none
   --------------------------------------------------------------------------*/
 
   static void ResetOmsVersionUnloadCounter();

  /*!-------------------------------------------------------------------------
	 function:     Signal
   description:  wakes up or timeouts a task waiting for an OMS Lock.  
   
     arguments:    taskId [in] the task identification of the waiting task.
     timeout[in] if true, waitung task is timed out
     return value: none
  --------------------------------------------------------------------------*/
  
  virtual void Signal(tsp00_TaskId, bool timeout);
  
  /*!-------------------------------------------------------------------------
	 function:     SignalUnloadLiboms
   description:  the liboms calls this method, when it starts to become unloaded, i.e.
   if it starts to destruct the globals. 
   
     arguments:    none
     return value: none
  --------------------------------------------------------------------------*/
  
  virtual void SignalUnloadLiboms();
  
  /*!-------------------------------------------------------------------------
	 function:     TestBadAlloc
   description:  tests, if bad alloc count down has reached 0. (for test
   purposes only (quickknl))
   
     arguments:    none
     return value: none
  --------------------------------------------------------------------------*/
  
  virtual void TestBadAlloc();
  
  /*!-------------------------------------------------------------------------
	function:	  SimCtlAttachSink
	description:  attach a new sink to a handle in simulator to ease
				  writing tests

	arguments:	  handle [in]	OMS handle to attach new sink to
	return value: none
  --------------------------------------------------------------------------*/

  virtual void SimCtlAttachSink(OmsHandle &handle);

  /*!-------------------------------------------------------------------------
	function:	  ConsoleMessage
	description:  print a console message

	arguments:	  msg [in]		message to print
	return value: none
  --------------------------------------------------------------------------*/

  virtual void ConsoleMessage(const char *msg);

  /*!-------------------------------------------------------------------------
	function:     AllocateArray
	description:  Allocates contiguous memory for 'count' bytes from LVCMem_Allocator
                for use by array new[].

    arguments:    count [in] number of requested bytes
    
	return value: pointer to memory chunk
  --------------------------------------------------------------------------*/
  
  virtual void* AllocateArray(SAPDB_ULong count);

  /*!-------------------------------------------------------------------------
	function:     DeallocateArray
	description:  Deallocates contiguous memory identified by p from LVCMem_Allocator
                for use by array delete[].

    arguments:    p [in] memory to be deallocated
    
	return value: none
  --------------------------------------------------------------------------*/

  virtual void DeallocateArray(void* p);

  virtual void DumpStackTrace();

  virtual unsigned int SaveStackTrace(void **buffer, unsigned int levels);

   /*!-------------------------------------------------------------------------
	 function:     LockRequest
	 description:  Interface for reader-/writer-locks (new with PTS 1124170)
		   
	 arguments:    none
	 return value: none
   --------------------------------------------------------------------------*/
   virtual short LockRequest(LVC_LockRequest &lockReq);

   /*!-------------------------------------------------------------------------
	 function:     UseReaderWriterLocks
	 description:  Use readerWriterLocks or CriticalSections for synchonization
                 of the version dictionary (new with PTS 1124170)
		   
	 arguments:    none
	 return value: none
   --------------------------------------------------------------------------*/
   virtual bool UseReaderWriterLocks();

   virtual void EvalStackTrace(void **buffer, unsigned int levels,
     OmsStackTraceCallbackInterface *cb);

  private :
    /// Spinlock pool for R/W regions.
    RTESync_SpinlockPool  *m_pSpinlockPool;
    /// Reader/writer locks for R/W regions.
    RTESync_IRWRegion    **m_ppRWLocks;
    bool                   m_isLiveCacheInstance;

  static LVC_KernelInterfaceInstance* m_Instance;
  static int                          m_omsVersionUnloads;
};

#endif
