/*!*****************************************************************************

  module:       LVCMem_DefaultAllocator.hpp

  ------------------------------------------------------------------------------

  responsible:  ThomasA

  special area: OMS Memory management

  description:  default allocator of liboms
                
  changes:      31.May 2000
                LVCMem_DefaultAllocator becomes a singleton with
                the method Instance().

  ------------------------------------------------------------------------------

                         Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#ifndef LVCMEM_DEFAULTALLOCATOR_HPP
#define LVCMEM_DEFAULTALLOCATOR_HPP

#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*!****************************************************************************
  class: LVCMem_DefaultAllocator

  description: Default allocator of liboms. 

               This allocator is active, until a call of OMS_LibOmsInterfaceInstance::InitLibOms
			   has been made.
			   This call is made only, if the liboms runs in the kernel environment
*****************************************************************************/

class LVCMem_DefaultAllocator : public SAPDBMem_IRawAllocator
{
public:

   /*!----------------------------------------------------------------------
	function: Allocate
	description: See 'SAPDBMem_IRawAllocator'

                These allocate methods use 'operator new'.
   exceptions:  SAPDBMem_BadAllocException:
                Thrown, if the memory cannot be allocated
   prototypes:  2
   -----------------------------------------------------------------------*/
   virtual void* Allocate(SAPDB_ULong Count);

   virtual void* Allocate(SAPDB_ULong Count, const void* Hint)
      { return Allocate(Count); }

   /*!----------------------------------------------------------------------
	function: Deallocate
	description: See 'SAPDBMem_IRawAllocator'.

                This method uses 'operator delete'.
   -----------------------------------------------------------------------*/
   virtual void Deallocate(void *p);

   /*!----------------------------------------------------------------------
   function: GetBaseAllocatorCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls of
                 the allocator to the underlying base allocator.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate of the base allocator.
                 CountDealloc [out]
                 The number of calls of method Deallocate of the base.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetBaseAllocatorCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;

   /*!----------------------------------------------------------------------
   function: GetCallStatistics
   description:  Outputs statistics about the number of Allocate and Deallocate calls.
   arguments:    CountAlloc [out]
                 The number of calls of method Allocate.
                 CountDealloc [out]
                 The number of calls of method Deallocate.

   return value: none
   -----------------------------------------------------------------------*/

   virtual void GetCallStatistics(SAPDB_ULong &CountAlloc,
                                  SAPDB_ULong &CountDealloc) const;


   /*!----------------------------------------------------------------------
    function: GetErrorCount
    description:  Returns the number of errors found and corrected by the allocator

                  this allocator is not able to provide the information, i.e. returns 0
 
    return value: 0
    -----------------------------------------------------------------------*/
  
   virtual int GetErrorCount() const;

   /*----------------------------------------------------------------------
	function: Default constructor
	description: Does nothing.
   -----------------------------------------------------------------------*/
    LVCMem_DefaultAllocator();

private:
   /*-----------------------------------------------------------------------
	function: copy constructor
	description: Private copy constructor avoids copying.
                This copy constructor need not to be defined, because
                noone can construct a LVCMem_DefaultAllocator and call
                the copy constructor thereafter.
                Besides, a private definition of the copy constructor would 
                result in a warning on several compilers.
   -----------------------------------------------------------------------*/
   /*
   LVCMem_DefaultAllocator(const LVCMem_DefaultAllocator &)
      {}
   */

   /*-----------------------------------------------------------------------
	operator: =	
	description: Private assignment operator avoids copying.
   -----------------------------------------------------------------------*/
   LVCMem_DefaultAllocator operator=(LVCMem_DefaultAllocator &)
      { return *this; }

};

/*!****************************************************************************
  endclass: LVCMem_DefaultAllocator
*****************************************************************************/


#endif // LVCMem_DefaultAllocator_HPP
