/*!***************************************************************************

  module      : LVCMem_EmergencyAllocator.hpp

  -------------------------------------------------------------------------

  responsible : ThomasA

  special area: OMS memory management
  description : emergency allocator which provides memory for clean up tasks

                This allocator wrapps internally a SAPDBMem_DougLeaAllocator 
                It is also a singleton.

  last changed: 2000-12-06  19:04
  see also    : example.html ...
  first created:2000-03-09  18:21

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef LVCMEM_EMERGENCYALLOCATOR_HPP
#define LVCMEM_EMERGENCYALLOCATOR_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/MemoryManagement/SAPDBMem_RawAllocator.hpp"
#include "gsp00.h"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*! 
  Class: LVCMem_EmergencyAllocator

  Description: implements the OMS emergency allocator. This allocator
               is attached to a user task, if a memory request of the
               OMS allocator fails due to lack of memory. All following
               memory requests of that task are fulfilled by the emergeny allocator
               until the user tasks leaves the calling db-procedure.
               Note : At one time at most one user task uses the emergeny allocator.
                      This is realized by use of a region.
*/
class LVCMem_EmergencyAllocator : public SAPDBMem_RawAllocator
{
public:
  typedef tsp00_TaskId TaskId;

  /*!----------------------------------------------------------------------
    Function: Instance

    description: Returns reference to single instance of LVCMem_EmergencyAllocator
    
    Return value: Reference to LVCMem_EmergencyAllocator instance
   */

  static LVCMem_EmergencyAllocator& Instance();

  /*!---------------------------------------------------------------------- 
    Function: Allocate

    Description: Allocates contiguous memory for 'count' bytes.

      The memory is suitably aligned for all types.

      The memory is not initialised.  

      std::bad_alloc is thrown on failure.

    Arguments: count [in] Number of Bytes to allocate

    Return value: Position where memory is allocated
  ----------------------------------------------------------------------*/

  virtual void* Allocate(SAPDB_ULong count);

  virtual void* Allocate(SAPDB_ULong count, const void* hint)
  {
    return Allocate(count);
  }

  /*!---------------------------------------------------------------------- 
    Function: GetEmergencyAllocatorInfo

    Description: returns allocator statistics 
    Arguments: BytesUsed [out] bytes in use
               BytesControlled [out] memory controlled by this allocator
               BytesControlled [out] max(BytesUsed)

    Return value: none
  ----------------------------------------------------------------------*/

  void GetEmergencyAllocatorInfo(SAPDB_ULong &BytesUsed,
                                 SAPDB_ULong &BytesControlled,
                                 SAPDB_ULong &MaxBytesUsed) const; 

  /*!---------------------------------------------------------------------- 
    Function: Release

    Description: releases the emergency allocator. 
                 If the given taskId identifies the owner of the emergency
                 allocator, the region blocking the emergeny alloctor
                 is unlocked.
    Arguments: taskId [in] taskId of the caller

    Return value: none
  ----------------------------------------------------------------------*/

  void Release(TaskId taskId);

  /*!---------------------------------------------------------------------- 
    Function: Reserve

    Description: reserves the emergency allocator for the specified task. 
    Arguments: taskId [in] taskId of the caller

    Return value: true, if emergency allocator has been granted to task, false
                  if task has already been owner of emergency allocator
  ----------------------------------------------------------------------*/

  bool Reserve(TaskId taskId);

private:
  LVCMem_EmergencyAllocator();
  TaskId                            m_Owner;     // current owner of emergency allocator
  SAPDB_ULong                       m_maxUsed;   // max memory used
  static LVCMem_EmergencyAllocator* m_Instance;  // the emergency allocator instance
}; /*! endclass: LVCMem_Allocator */

#endif



