/*!
  -----------------------------------------------------------------------------
  module: vcn10.cpp
  -----------------------------------------------------------------------------

  responsible:  BerndV

  special area: DBMServer

  description:  "main" file for DBMServer

                implementing ctrlservopen()
                             ctrlservclose()
                             ctrlservcommand()

  copyright:    by SAP AG, 2001
 
  -----------------------------------------------------------------------------

    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "vsp001.h"
#include "vsp009c.h"

#include "hsp100.h"
#include "heo02.h"
#include "heo06.h"

#include "hcn20.h"
#include "hcn30.h"
#include "hcn31.h"
#include "hcn32.h"
#include "hcn40.h"
#include "hcn41.h"
#include "hcn42.h"
#include "hcn43.h"
#include "hcn45.h"
#include "hcn50.h"
#include "hcn51.h"
#include "hcn80.h"
#include "hcn81.h"
#include "hcn82.h"
#include "hcn83.h"
#include "hcn84.h"
#include "hcn90.h"

#include "SAPDB/DBM/Srv/BackupHistory/DBMSrvBHist_RestorePlanner.hpp"
/*!  
  -----------------------------------------------------------------------------
  Chapter: Private types
  -----------------------------------------------------------------------------
 */

/*! Declaration: DBMServer command structure */
typedef struct DBMServerCommand {
  const _TCHAR             * name;
  bool                       bLogon;
  bool                       bSysLogon;
  tsp_int2                   nDBNameState;
  tsp_int2                   nMinParam;
  tsp_int2                   nMaxParam;
  tsp_int2                   nTokens2Log;
  cnFunctionPrototype_CN00 * cnFunction;
  const _TCHAR             * shortHelp;
  tsp_uint4                  nRights;
} DBMServerCommand;
/*! EndDeclaration: DBMServer command structure */

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Private types
  -----------------------------------------------------------------------------
 */

/*!  
  -----------------------------------------------------------------------------
  Chapter: Private macros
  -----------------------------------------------------------------------------
 */

/*! Declaration: DBMServer command protocol values */
#define CMD_PROT_NO            (-1)
#define CMD_PROT_ALL           (-2)
#define CMD_PROT_NOERR         (-3)
#define CMD_PROT_YES           (0)
/*! EndDeclaration: DBMServer command protocol values */

#define OBSOLETE_MARK           _T("(obsolete")

#ifndef DBROOT_ENV_VAR
  #define DBROOT_ENV_VAR         "DBROOT"
#endif

const int CommandT::InvalidIndex = -1;

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Private macros
  -----------------------------------------------------------------------------
 */

/*!  
  -----------------------------------------------------------------------------
  Chapter: Private functions
  -----------------------------------------------------------------------------
 */

/*!
  -------------------------------------------------------------------------
  function:     cn10_AnalyzeCommand
  -------------------------------------------------------------------------
  description:  DBMServer command analyzer

                This function searches in an array for the properties
                of the specified DBMServer command.

  arguments:    szData     [INOUT] - buffer with the command and parameters
                nDataLen   [IN]    - the name of the database
                DBMCommand [INOUT] - structur for the analyzed command

  -------------------------------------------------------------------------
*/
static void cn10_AnalyzeCommand
      ( const _TCHAR * szData,
        const int      nDataLen,
        CommandT     & DBMCommand );

/*!
  -------------------------------------------------------------------------
  function:     cn10_showHelp
  -------------------------------------------------------------------------
  description:  Help

                Here will be executed the help command.

  arguments:    vcontrol    [INOUT]  - DBMServer session data
                command     [IN]     - DBMServer command
                replyData   [OUT]    - reply buffer
                replyLen    [OUT]    - lenght of reply
                replyLenMax [IN]     - max. length of reply buffer

  return value: error code specified in gcn00c.h

  -------------------------------------------------------------------------
*/
static tcn00_Error cn10_showHelp (
    VControlDataT * vcontrol,
    CommandT      * command,
    char          * replyData,
    int           * replyLen,
    int             replyLenMax);

/*!
  -------------------------------------------------------------------------
  function:     cn10_dbmRelease
  -------------------------------------------------------------------------
  description:  Release

                Here will be executed the release command.

  arguments:    vcontrol    [INOUT]  - DBMServer session data
                command     [IN]     - DBMServer command
                replyData   [OUT]    - reply buffer
                replyLen    [OUT]    - lenght of reply
                replyLenMax [IN]     - max. length of reply buffer

  return value: error code specified in gcn00c.h

  -------------------------------------------------------------------------
*/
static tcn00_Error cn10_dbmRelease (
    VControlDataT * vcontrol,
    CommandT      * command,
    char          * replyData,
    int           * replyLen,
    int             replyLenMax);

/*!
  -------------------------------------------------------------------------
  function:     cn10_LogCommand
  -------------------------------------------------------------------------
  description:  Command log

                This functions logs the specified command.

  arguments:    command     [IN]     - DBMServer command
                
  -------------------------------------------------------------------------
*/
static void cn10_LogCommand (
    CommandT      * command );

/*!
  -------------------------------------------------------------------------
  function:     cn10_LogError
  -------------------------------------------------------------------------
  description:  Error log

                This functions logs the specified error text.

  arguments:    szError     [IN]     - error text
                nCmdRc      [IN]     - numeric error code
                
  -------------------------------------------------------------------------
*/
static void cn10_LogError (
      const _TCHAR      * szError,
      const tcn00_Error   nCmdRc );

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Private functions
  -----------------------------------------------------------------------------
 */

static DBMServerCommand namesC[] = {
// name                        Logon Sys    DBNameState       Min  Max  Protocol      Function
//                             short help
//                             rights
{"archive_stage"              ,true ,false ,DBNAME_YES_CN00 ,   2,   6, CMD_PROT_ALL   ,cn31BackupStage
                              ,"<medium> <stage> [verify|noverify] [remove|keep]\n                         [FileNumberList <list>]"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"archive_stage_repeat"       ,true ,false ,DBNAME_YES_CN00 ,   1,   3, CMD_PROT_ALL   ,cn31BackupStageRepeat
                              ,"<medium> [verify|noverify] [remove|keep]"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autolog_cancel"             ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_YES   ,cn31AutosaveCancel
                              ,"                              (needs util session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autosave_cancel"            ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_YES   ,cn31AutosaveCancel
 /*** obsolete ***/           ,"(obsolete version of autolog_cancel)              "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autolog_off"                ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_YES   ,cn31AutosaveEnd
                              ,"                              (needs util session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autosave_off"               ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_YES   ,cn31AutosaveEnd
 /*** obsolete ***/           ,"(obsolete version of autolog_off)                 "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autolog_on"                 ,true ,false ,DBNAME_YES_CN00  ,  0 ,  1 ,CMD_PROT_YES   ,cn31AutosaveOn
                              ,"[<medium>]                    (needs util session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autosave_on"                ,true ,false ,DBNAME_YES_CN00  ,  0 ,  1 ,CMD_PROT_YES   ,cn31AutosaveOn
 /*** obsolete ***/           ,"(obsolete version of autolog_on)                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"autolog_show"               ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_NO    ,cn31AutosaveShow
                              ,"                              (needs util session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"autosave_show"              ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_NO    ,cn31AutosaveShow
 /*** obsolete ***/           ,"(obsolete version of autolog_show)                "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"autorecover"                ,true ,false ,DBNAME_YES_CN00  ,  0,   8 ,CMD_PROT_ALL   ,DBMSrvBHist_RestorePlanner::AutomaticRecovery
                              ,"[LOG] [UNTIL <date> <time>] [IBID <int_backup_ID>]\n"
                               "                                                            (needs util session)"
                              ,DBMRightBackupRestore_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00 ,  0,   0 ,CMD_PROT_NO    ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"backup_cancel"              ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveCancel
                              ,"                            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_cancel"         ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveCancel
 /*** obsolete ***/           ,"(obsolete version of autolog_on)                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_cancel_req"          ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveCancelRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_cancel_req"     ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveCancelRequest
 /*** obsolete ***/           ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_command"             ,true ,false ,DBNAME_YES_CN00 ,   1,   1 ,CMD_PROT_ALL   ,cn31BackupCommand
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_command_req"         ,true ,false ,DBNAME_YES_CN00 ,   1,   1 ,CMD_PROT_ALL   ,cn31BackupCommandRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_ext_ids_forget",      true, false, DBNAME_YES_CN00,    0,   0, CMD_PROT_ALL   ,cn32ForgetEBIDs,
                               "                                                  ",
                               DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_ext_ids_get",         true, false, DBNAME_YES_CN00,    1,   3, CMD_PROT_ALL   ,cn32GetEBIDs,
                               "<medium> [<dbname>] [<node>]                      ",
                               DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_ext_ids_list",        true, false, DBNAME_YES_CN00,    0,   0, CMD_PROT_ALL   ,cn32ListNextEBIDs,
                               "                                                  ",
                               DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_ext_ids_listnext",    true, false, DBNAME_YES_CN00,    0,   0, CMD_PROT_ALL   ,cn32ListNextEBIDs,
                               "                                                  ",
                               DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_history_close"       ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO    ,cn32BackupHistoryClose
                              ,"                                                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_history_date"        ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO    ,cn32BackupHistoryDate
                              ,"                                                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_history_list"        ,true ,false ,DBNAME_YES_CN00  ,  0,  10 ,CMD_PROT_NO    ,cn32BackupHistoryList
                              ,"[-c <col,col,...>] [-k <key>] [-l <label>]\n                         [-a <action>] [-r <restore_flag>] [-m] [-e]\n                         [-Inverted]"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_history_listnext"    ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO    ,cn32BackupHistoryListNext
                              ,"                                                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_history_open"        ,true ,false ,DBNAME_YES_CN00  ,  0,   1 ,CMD_PROT_NO    ,cn32BackupHistoryOpen
                              ,"[-e]                                              "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_ignore"              ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveIgnore
                              ,"                            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_ignore"         ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveIgnore
 /*** obsolete ***/           ,"(obsolete version of backup_ignore)               "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_ignore_req"          ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveIgnoreRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_ignore_req"     ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn31BackupSaveIgnoreRequest
 /*** obsolete ***/           ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_replace"             ,true ,false ,DBNAME_YES_CN00 ,   1,   2 ,CMD_PROT_ALL   ,cn31BackupSaveReplace
                              ,"<medium> [<loc>]            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_replace"        ,true ,false ,DBNAME_YES_CN00 ,   1,   2 ,CMD_PROT_ALL   ,cn31BackupSaveReplace
 /*** obsolete ***/           ,"(obsolete version of backup_replace)              "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_replace_req"         ,true ,false ,DBNAME_YES_CN00 ,   1,   2 ,CMD_PROT_ALL   ,cn31BackupSaveReplaceRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_replace_req"    ,true ,false ,DBNAME_YES_CN00 ,   1,   2 ,CMD_PROT_ALL   ,cn31BackupSaveReplaceRequest
 /*** obsolete ***/           ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_reply_available"     ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO    ,cn31BackupReplyAvailable
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_reply_receive"       ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL   ,cn31BackupReplyReceive
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_req"                 ,true ,false ,DBNAME_YES_CN00 ,   1,   3 ,CMD_PROT_ALL   ,cn31BackupSaveRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_req"            ,true ,false ,DBNAME_YES_CN00 ,   1,   3 ,CMD_PROT_ALL   ,cn31BackupSaveRequest
 /*** obsolete ***/           ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_stage"               ,true ,false ,DBNAME_YES_CN00 ,   2,   6, CMD_PROT_ALL   ,cn31BackupStage
                              ,"(obsolete version of archive_stage)     "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_stage_repeat"        ,true ,false ,DBNAME_YES_CN00 ,   1,   3, CMD_PROT_ALL   ,cn31BackupStageRepeat
                              ,"(obsolete version of archive_stage_repeat)     "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_start"               ,true ,false ,DBNAME_YES_CN00 ,   1,   4 ,CMD_PROT_ALL   ,cn31BackupSave
                              ,"<medium> [RECOVERY|MIGRATION] [DATA|PAGES|LOG]    \n                         [AUTOIGNORE]                  (needs util session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save"                ,true ,false ,DBNAME_YES_CN00 ,   1,   4 ,CMD_PROT_ALL   ,cn31BackupSave
 /*** obsolete ***/           ,"(obsolete version of backup_start)                "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_state"               ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO    ,cn31BackupSaveState
                              ,"                            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_save_state"          ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO    ,cn31BackupSaveState
 /*** obsolete ***/           ,"(obsolete version of backup_state)                "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"bye"                        ,false,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO    ,cn10_dbmRelease
                              ,NULL                                                 
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO    ,NULL
                              ,"                                                  " 
                              ,DBMMaskNothing_CN50 },
{"crash"                      ,false,false ,DBNAME_EGAL_CN00,   0,  -1 ,CMD_PROT_NO    ,cn40Crash
                              ,NULL
                              ,DBMMaskNothing_CN50 },
{"db_addvolume"               ,true ,false ,DBNAME_YES_CN00 ,   4,   6 ,CMD_PROT_ALL   ,cn40DbAddDevice
                              ,"<mod> <nam> <typ> <siz> <mnam> <mtyp>             "
                              ,DBMRightAccessUtility_CN50 | DBMRightParamFull_CN50},
{"db_adddevspace"             ,true ,false ,DBNAME_YES_CN00 ,   4,   6 ,CMD_PROT_ALL   ,cn40DbAddDevice
 /*** obsolete ***/           ,"(obsolete version of db_addvolume)                "
                              ,DBMRightAccessUtility_CN50 | DBMRightParamFull_CN50},
{"db_adddevice"               ,true ,false ,DBNAME_YES_CN00 ,   4,   6 ,CMD_PROT_ALL   ,cn40DbAddDevice
 /*** obsolete ***/           ,"(obsolete version of db_addvolume)                "
                              ,DBMRightAccessUtility_CN50 | DBMRightParamFull_CN50},
{"db_clear"                   ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn40DBClear
                              ,"                                    (needs dbname)"
                              ,DBMRightDBStop_CN50 },
{"db_admin"                   ,true ,false ,DBNAME_YES_CN00 ,   0,   2 ,CMD_PROT_ALL   ,cn40DBCold
                              ,"[-immediate|-fast|-quick|-slow]      (needs dbname)"
                              ,DBMRightDBStop_CN50 },
{"db_cold"                    ,true ,false ,DBNAME_YES_CN00 ,   0,   2 ,CMD_PROT_ALL   ,cn40DBCold
/*** obsolete ***/            ,"(obsolete version of db_admin)                     "
                              ,DBMRightDBStop_CN50 },
{"dbcold"                     ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_ALL   ,cn40DBCold
 /*** obsolete ***/           ,"(obsolete version of db_admin)                     "
                              ,DBMRightDBStop_CN50 },
{"db_cons"                    ,true ,false ,DBNAME_YES_CN00 ,   1,  -1 ,CMD_PROT_NO    ,cn40DBConsole
                              ,"<console command>                                 "
                              ,DBMRightDBStop_CN50 },
{"db_create"                  ,false,false ,DBNAME_NO_CN00  ,   2,   4 ,1              ,cn40DBCreate
                              ,"[-f|-q|-s] [-a] <dbname> <usr>,<pwd> [<osusr>,<pwd>]"
                              ,DBMMaskNothing_CN50 },
{"db_drop"                    ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_NO    ,cn40DBDrop
                              ,"[WITHOUTFILES]                      (needs dbname)"
                              ,DBMRightInstallMgm_CN50 },
{"db_enum"                    ,false,false ,DBNAME_EGAL_CN00,   0,   1 ,CMD_PROT_NO    ,cn40DBEnum
                              ,"[-s]                                              "
                              ,DBMMaskNothing_CN50 },
{"enum_dbs"                   ,false,false ,DBNAME_EGAL_CN00,   0,   1 ,CMD_PROT_NO    ,cn40DBEnum
 /*** obsolete ***/           ,"(obsolete version of db_enum)                     "
                              ,DBMMaskNothing_CN50 },
{"db_migratecatalog"          ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL   ,cn45MigrateCatalog
                              ,"                                                  "
                              ,DBMRightAccessUtility_CN50},
{"db_offline"                 ,true ,false ,DBNAME_YES_CN00 ,   0,   2 ,CMD_PROT_YES   ,cn40DBOffline
                              ,"[-immediate] [-dump]                (needs dbname)"
                              ,DBMRightDBStop_CN50 },
{"dboffline"                  ,true ,false ,DBNAME_YES_CN00 ,   0,   2 ,CMD_PROT_ALL   ,cn40DBOffline
 /*** obsolete ***/           ,"(obsolete version of db_offline)                   "
                              ,DBMRightDBStop_CN50 },
{"db_online"                  ,true ,false ,DBNAME_YES_CN00 ,   0,   4 ,CMD_PROT_ALL   ,cn40DBWarm
                              ,"[(-f|-q|-s)] [-u <yyyymmdd> <hhmmss>]             "
                              ,DBMRightDBStart_CN50 },
{"db_warm"                    ,true ,false ,DBNAME_YES_CN00 ,   0,   4 ,CMD_PROT_ALL   ,cn40DBWarm
 /*** obsolete ***/           ,"(obsolete version of db_online)                   "
                              ,DBMRightDBStart_CN50 },
{"dbwarm"                     ,true ,false ,DBNAME_YES_CN00 ,   0,   4 ,CMD_PROT_ALL   ,cn40DBWarm
 /*** obsolete ***/           ,"(obsolete version of db_online)                    "
                              ,DBMRightDBStart_CN50 },
{"db_reg"                     ,true ,false ,DBNAME_YES_CN00 ,   0,   3 ,CMD_PROT_ALL   ,cn40DBRegister
                              ,"[-f|-q|-s][-R <newinstroot>]         (needs dbname)"
                              ,DBMRightInstallMgm_CN50 },
{"db_register"                ,true ,false ,DBNAME_YES_CN00 ,   0,   3 ,CMD_PROT_ALL   ,cn40DBRegister
 /*** obsolete ***/           ,"(obsolete version of db_reg)                       "
                              ,DBMRightInstallMgm_CN50 },
{"db_reinstall"               ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL ,cn45Reinstall     
                              ,"                                                  "
                              ,DBMRightAccessUtility_CN50},
{"db_restart"                 ,true ,false ,DBNAME_YES_CN00 ,   0,   6 ,CMD_PROT_ALL ,cn40DBRestart
                              ,"[(-f|-q|-s)] [-i] [-d] [-u <yyyymmdd> <hhmmss>]   "
                              ,DBMRightDBStop_CN50 },
{"dbrestart"                  ,true ,false ,DBNAME_YES_CN00 ,   0,   6 ,CMD_PROT_ALL ,cn40DBRestart
 /*** obsolete ***/           ,"(obsolete version of db_restart)                   "
                              ,DBMRightDBStop_CN50 },
{"db_restartinfo"             ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn31BackupRestartInfo
                              ,"                              (needs util session)"
                              ,DBMRightBackupRestore_CN50 },
{"backup_restart_info"        ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn31BackupRestartInfo
 /*** obsolete ***/           ,"(obsolete version of db_restartinfo)              "
                              ,DBMRightBackupRestore_CN50 },
{"db_speed"                   ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn40DBSpeed
                              ,"                                    (needs dbname)"
                              ,DBMRightDBInfoRead_CN50 },
{"db_state"                   ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn40DBState
                              ,"                                    (needs dbname)"
                              ,DBMRightDBInfoRead_CN50 },
{"db_start"                   ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_ALL ,cn40DBStart
                              ,"[(-fast|-quick|-slow)]              (needs dbname)"
                              ,DBMRightDBStart_CN50 },
{"start"                      ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_ALL ,cn40DBStart
 /*** obsolete ***/           ,"(obsolete version of db_start)                     "
                              ,DBMRightDBStart_CN50 },
{"db_stop"                    ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_ALL ,cn40DBStop
                              ,"[-dump]                             (needs dbname)"
                              ,DBMRightDBStop_CN50 },
{"stop"                       ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_ALL ,cn40DBStop
 /*** obsolete ***/           ,"(obsolete version of db_stop)                      "
                              ,DBMRightDBStop_CN50 },
{"db_unreg"                   ,true ,false ,DBNAME_YES_CN00 ,   1,   1 ,CMD_PROT_ALL ,cn40DBUnregister
                              ,"-fast|-quick|-slow                  (needs dbname)"
                              ,DBMRightInstallMgm_CN50 },
{"db_unregister"              ,true ,false ,DBNAME_YES_CN00 ,   1,   1 ,CMD_PROT_ALL ,cn40DBUnregister
 /*** obsolete ***/           ,"(obsolete version of db_unreg)                    "
                              ,DBMRightInstallMgm_CN50 },
{"dban_delete"                ,true ,false ,DBNAME_YES_CN00  ,  1 , 1 ,CMD_PROT_NO  ,cn43DBanalyzerDelete
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"dban_start"                 ,true ,false ,DBNAME_YES_CN00  ,  0 , -1 ,CMD_PROT_NO  ,cn43DBanalyzerStart
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"dban_stop"                  ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_NO  ,cn43DBanalyzerStop
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"dban_state"                 ,true ,false ,DBNAME_YES_CN00  ,  0 ,  0 ,CMD_PROT_NO  ,cn43DBanalyzerState
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"dbm_configset"              ,true ,false ,DBNAME_YES_CN00 ,   2,   3 ,1            ,cn41DBMSetValue
                              ,"[-raw] <name> <value>                (needs dbname)"
                              ,DBMRightInstallMgm_CN50 },
{"dbm_configget"              ,true ,false ,DBNAME_YES_CN00 ,   1,   2 ,CMD_PROT_NO  ,cn41DBMGetValue
                              ,"[-raw] <name>                        (needs dbname)"
                              ,DBMRightInstallMgm_CN50 },
{"dbm_getpath"                ,false,false ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn40DBMGetPath
                              ,"[IndepDataPath|IndepProgPath]                     "
                              ,DBMMaskNothing_CN50 },
{"dbm_next"                   ,false,false ,DBNAME_EGAL_CN00 ,  0 ,  0 ,CMD_PROT_NO , cn40DBMNext
                              ,NULL
                              ,DBMMaskNothing_CN50 },
{"dbm_setpath"                ,false,false ,DBNAME_EGAL_CN00,   2,   2 ,CMD_PROT_NO  ,cn40DBMSetPath
                              ,"[IndepDataPath|IndepProgPath] <path>              "
                              ,DBMMaskNothing_CN50 },
{"dbm_version"                ,false,false ,DBNAME_EGAL_CN00,   0,   1 ,CMD_PROT_NO  ,cn40DBMVersion
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"diag_histlist"              ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_NO  ,cn42DiagHistList
                              ,"                                    (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
{"diag_pack"                  ,true ,false ,DBNAME_YES_CN00 ,   0,   3, CMD_PROT_NO  ,cn42DiagPack
                              ,"<content spec>                      (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
//{"dir_close"                  ,true ,true  ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn42DirClose
//                              ,"<directory handle>                                "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
//{"dir_create"                 ,true ,true  ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn42DirCreate
//                              ,"<directory name>                                  "
//                              ,DBMRightFileFull_CN50 },
//{"dir_open"                   ,true ,true  ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn42DirOpen
//                              ,"<directory name>                                  "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
//{"dir_readnext"               ,true ,true  ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn42DirReadNext
//                              ,"<directory handle>                                "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"event_list"                 ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO ,cn82EventList
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"event_set"                  ,true ,false ,DBNAME_YES_CN00 ,   1,   4 ,CMD_PROT_ALL ,cn82EventSet
                              ,"<event> [<prio> [<values>]]   (needs util session)"
                              ,DBMRightDBInfoRead_CN50 },
{"event_delete"               ,true ,false ,DBNAME_YES_CN00 ,   1,   3 ,CMD_PROT_ALL ,cn82EventDelete
                              ,"<event> [<values>]            (needs util session)"
                              ,DBMRightDBInfoRead_CN50 },
{"event_release"              ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn82EventRelease
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"event_wait"                 ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn82EventWait
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
//{"exec_command"               ,true ,true  ,DBNAME_EGAL_CN00,   0,  -1 ,CMD_PROT_NO  ,cn40ExecCommand
//                              ,"<command>                                         "
//                              ,DBMRightSystemCmd_CN50 },
{"exec_lcinit"                ,true ,false ,DBNAME_YES_CN00 ,   0,  -1 ,CMD_PROT_YES ,cn40ExecLcInit
                              ,"                                                  "
                              ,DBMRightSystemCmd_CN50 },
{"exec_nipingsrv"             ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_NO  ,cn40ExecNiPingSrv
                              ,"[<idletime>]                                      "
                              ,DBMRightSystemCmd_CN50 },
{"exec_xpu"                   ,true ,false ,DBNAME_YES_CN00 ,   0,  -1 ,CMD_PROT_YES ,cn40ExecXPU
                              ,"                                                  "
                              ,DBMRightAccessSQL_CN50 },
{"exec_load"                  ,true ,false ,DBNAME_YES_CN00,    2,   4 ,CMD_PROT_YES ,cn40ExecLoad
                              ,"<user>,<pw> <file> [\"<options>\"]    (needs dbname)"
                              ,DBMRightExecLoad_CN50 },
{"exec_sdbinfo"               ,false,false ,DBNAME_EGAL_CN00,   0,  -1 ,CMD_PROT_NO ,cn40ExecSDBInfo
                              ,"[<parameters>]                                    "
                              ,DBMMaskNothing_CN50 },
{"exec_xload"                 ,true ,false ,DBNAME_YES_CN00,    2,   4 ,CMD_PROT_YES ,cn40ExecXLoad
                              ,"<user>,<pw> <file> [\"<options>\"]    (needs dbname)"
                              ,DBMRightExecLoad_CN50 },
{"exit"                       ,false,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,cn10_dbmRelease
                              ,NULL                                                 
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"file_backup"                ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn42FileBackup
                              ,"<fileid>                            (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
//{"file_close"                 ,true ,true  ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn42FileClose
//                              ,"<file handle>                                     "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
//{"file_erase"                 ,true ,true  ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn42FileErase
//                              ,"<filename>                                        "
//                              ,DBMRightFileFull_CN50 },
{"file_getfirst"              ,true ,false ,DBNAME_YES_CN00 ,   1,   3 ,CMD_PROT_NO  ,cn42FileGet
                              ,"<fileid> [LINE=<n>|DATE=<yyyymmdd>] (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
{"file_getlist"               ,true ,false ,DBNAME_YES_CN00 ,   0,   1 ,CMD_PROT_NO  ,cn42FileGetList
                              ,"[<showlevel>]                       (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
{"file_getnext"               ,true ,false ,DBNAME_YES_CN00 ,   2,   2 ,CMD_PROT_NO  ,cn42FileGetNext
                              ,"<fileid> <file handle>              (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
//{"file_info"                  ,true ,false ,DBNAME_EGAL_CN00,   1,  -1 ,CMD_PROT_NO  ,cn42FileInfo
//                              ,"[-m] <filename> [<filename> ...]                  "
//                              ,DBMRightDBInfoRead_CN50 },
//{"file_open"                  ,true ,true  ,DBNAME_EGAL_CN00,   1,   3 ,CMD_PROT_NO  ,cn42FileOpen
//                              ,"[-(a|b)] [-(r|w|rw|x)] <file name>                "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
{"file_operation"             ,true ,false ,DBNAME_YES_CN00 ,   2,   3 ,CMD_PROT_ALL ,cn42FileOperation
                              ,"<fileid> OP=<operid> [DATE=<date>|LINE=<line>]    "
                              ,DBMRightDBFileRead_CN50 },
//{"file_readfirst"             ,true ,true  ,DBNAME_EGAL_CN00,   1,   3 ,CMD_PROT_NO  ,cn42FileRead
//                              ,"[-(a|b)] <filename> [NOCLOSE]                     "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
//{"file_readnext"              ,true ,true  ,DBNAME_EGAL_CN00,   1,   3 ,CMD_PROT_NO  ,cn42FileReadNext
//                              ,"[-(a|b)] <filehandle> [NOCLOSE]                   "
//                              ,DBMRightFileFull_CN50 | DBMRightFileRead_CN50},
{"file_restore"               ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn42FileRestore
                              ,"<fileid>                            (needs dbname)"
                              ,DBMRightDBFileRead_CN50 },
//{"file_seek"                  ,true ,true  ,DBNAME_EGAL_CN00,   2,   3 ,CMD_PROT_NO  ,cn42FileSeek
//                              ,"<file handle> <seek pos> [CUR]                    "
//                              ,DBMRightFileFull_CN50 },
//{"file_writefirst"            ,true ,true  ,DBNAME_EGAL_CN00,   0,  -1 ,CMD_PROT_NO  ,cn42FileWrite
//                              ,"[-(a|b)] <filename> <NL> <length> <NL> <output>   "
//                              ,DBMRightFileFull_CN50 },
//{"file_writenext"             ,true ,true  ,DBNAME_EGAL_CN00,   0,  -1 ,CMD_PROT_NO  ,cn42FileWriteNext
//                              ,"[-(a|b)] <filehandle> <NL> <length> <NL> <output> "
//                              ,DBMRightFileFull_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"help"                       ,false,false ,DBNAME_EGAL_CN00,   0,   2 ,CMD_PROT_NO  ,cn10_showHelp
                              ,"[<command>]                                       "
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"info"                       ,true ,false ,DBNAME_YES_CN00,    1,  -1 ,CMD_PROT_NO  ,cn81Info
                              ,"<name>                         (needs sql session)"
                              ,DBMRightDBInfoRead_CN50 },
{"info_next"                  ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn81InfoNext
                              ,"                               (needs sql session)"
                              ,DBMRightDBInfoRead_CN50 },
{"inst_enum"                  ,false,false ,DBNAME_EGAL_CN00,   0,   1 ,CMD_PROT_NO  ,cn40InstEnum
                              ,"[-k]                                              "
                              ,DBMMaskNothing_CN50 },
{"installation_enum"          ,false,false ,DBNAME_EGAL_CN00,   0,   1 ,CMD_PROT_NO  ,cn40InstEnum
 /*** obsolete ***/           ,"(obsolete version of inst_enum)                   "
                              ,DBMMaskNothing_CN50 },
{"enum_installations"         ,false,false ,DBNAME_EGAL_CN00,   0,   1 ,CMD_PROT_NO  ,cn40InstEnum
 /*** obsolete ***/           ,"(obsolete version of inst_enum)                   "
                              ,DBMMaskNothing_CN50 },
{"inst_reg"                   ,false,true  ,DBNAME_NO_CN00,     0,   5 ,CMD_PROT_ALL ,cn40InstReg
                              ,"[<dbroot>] [[-v]  <version>] [-k <key>]           "
                              ,DBMMaskNothing_CN50 },
{"installation_register"      ,false,true  ,DBNAME_NO_CN00,     0,   5 ,CMD_PROT_ALL ,cn40InstReg
 /*** obsolete ***/           ,"(obsolete version of inst_reg)                    "
                              ,DBMMaskNothing_CN50 },
{"register_installation"      ,false,true  ,DBNAME_NO_CN00,     0,   5 ,CMD_PROT_ALL ,cn40InstReg
 /*** obsolete ***/           ,"(obsolete version of inst_reg)                    "
                              ,DBMMaskNothing_CN50 },
{"inst_unreg"                 ,false,true  ,DBNAME_NO_CN00,     0,   1 ,CMD_PROT_ALL ,cn40InstUnreg
                              ,"[<dbroot>]                                        "
                              ,DBMMaskNothing_CN50 },
{"installation_unregister"    ,false,true  ,DBNAME_NO_CN00,     0,   1 ,CMD_PROT_ALL ,cn40InstUnreg
 /*** obsolete ***/           ,"(obsolete version of inst_unreg)                  "
                              ,DBMMaskNothing_CN50 },
{"unregister_installation"    ,false,true  ,DBNAME_NO_CN00,     0,   1 ,CMD_PROT_ALL ,cn40InstUnreg
 /*** obsolete ***/           ,"(obsolete version of inst_unreg)                  "
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"load_systab"                ,true ,false ,DBNAME_YES_CN00,    0,   4 ,CMD_PROT_YES ,cn40LoadSysTab
                              ,"[-u <user>,<pwd>] [-ud <domainpwd>] (needs dbname)"
                              ,DBMRightLoadSysTab_CN50 | DBMRightExecLoad_CN50},
{"load_r3tab"                 ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_YES ,cn40LoadR3Tab
                              ,"                                    (needs dbname)"
                              ,DBMRightLoadSysTab_CN50 | DBMRightExecLoad_CN50},
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"medium_date"                ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn30BackupMediaDate
                              ,"                                                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_media_date"          ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn30BackupMediaDate
 /*** obsolete ***/           ,"(obsolete version of medium_date)                 "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"medium_delete"              ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn30BackupMediaDelete
                              ,"<name>                                            "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_media_delete"        ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn30BackupMediaDelete
 /*** obsolete ***/           ,"(obsolete version of medium_delete)               "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"medium_get"                 ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_NO  ,cn30BackupMediaGet
                              ,"<name>                                            "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_media_get"           ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_NO  ,cn30BackupMediaGet
 /*** obsolete ***/           ,"(obsolete version of medium_get)                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"medium_getall"              ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn30BackupMediaGetAll
                              ,"                                                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"backup_media_getall"        ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn30BackupMediaGetAll
 /*** obsolete ***/           ,"(obsolete version of medium_getall)               "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{"medium_label"               ,true ,false ,DBNAME_YES_CN00  ,  1,   2 ,CMD_PROT_NO  ,cn31BackupLabel
                              ,"<medium> [<nnn>]              (needs util session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_label"               ,true ,false ,DBNAME_YES_CN00  ,  1,   2 ,CMD_PROT_NO  ,cn31BackupLabel
 /*** obsolete ***/           ,"(obsolete version of medium_label)                "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"medium_labeloffline"        ,true ,false ,DBNAME_YES_CN00  ,  1,   2 ,CMD_PROT_NO  ,cn31BackupLabelOffline
                              ,"<medium> [<nnn>]                                  "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_labeloffline"        ,true ,false ,DBNAME_YES_CN00  ,  1,   2 ,CMD_PROT_NO  ,cn31BackupLabelOffline
 /*** obsolete ***/           ,"(obsolete version of medium_labeoffline)          "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"medium_put"                 ,true ,false ,DBNAME_YES_CN00  ,  2,  10 ,CMD_PROT_ALL ,cn30BackupMediaPut
                              ,"<name> <loc> <type> <stype> <size> <blksize> <owr>\n                         <autoldr> <oscmd>"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_media_put"           ,true ,false ,DBNAME_YES_CN00  ,  2,  10 ,CMD_PROT_ALL ,cn30BackupMediaPut
 /*** obsolete ***/           ,"(obsolete version of media_put)                   "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"medium_migrate"             ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_NO  ,cn30BackupMediaMigrate
                              ,NULL
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 | DBMRightDBInfoRead_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"package_info"               ,false,false ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO ,cn40ExecSDBInst
                              ,"<directory>                                         "
                              ,DBMMaskNothing_CN50 },
{"param_abortsession"         ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_ALL ,cn20ParamAbortSession
                              ,"                             (needs param session)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_addvolume"            ,true ,false ,DBNAME_YES_CN00,    4,   5 ,CMD_PROT_ALL ,cn20ParamAddDevSpace
                              ,"<num> <mod> <nam> <typ> <siz>       (needs dbname)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_adddevspace"          ,true ,false ,DBNAME_YES_CN00,    4,   5 ,CMD_PROT_ALL ,cn20ParamAddDevSpace
 /*** obsolete ***/           ,"(obsolete version of param_addvolume)             "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_checkall"             ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_ALL ,cn20ParamCheckAll
                              ,"[<mode>]                            (needs dbname)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_commitsession"        ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_ALL ,cn20ParamCommitSession
                              ,"                             (needs param session)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_copy"                 ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamCopy
                              ,"<sourcedb>                   (needs param session)"
                              ,DBMRightInstallMgm_CN50 },
{"param_delvolume"            ,true ,false ,DBNAME_YES_CN00,    2,   2 ,CMD_PROT_ALL ,cn20ParamDelDevSpace
                              ,NULL
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_deldevspace"          ,true ,false ,DBNAME_YES_CN00,    2,   2 ,CMD_PROT_ALL ,cn20ParamDelDevSpace
                              ,NULL
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_directdel"            ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_ALL ,cn20ParamDirectDel
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 },
{"param_directget"            ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamDirectGet
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamRead_CN50 },
{"param_directgetall"         ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn20ParamDirectGetAll
                              ,"                                                  "
                              ,DBMRightParamFull_CN50 | DBMRightParamRead_CN50 },
{"param_directgetallnext"     ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn20ParamDirectGetAllNext
                              ,"                                                  "
                              ,DBMRightParamFull_CN50 | DBMRightParamRead_CN50 },
{"param_directput"            ,true ,false ,DBNAME_YES_CN00,    2,   2 ,CMD_PROT_ALL ,cn20ParamDirectPut
                              ,"<keyname> <value>                                 "
                              ,DBMRightParamFull_CN50 },
{"param_extget"               ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamExtGet
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_extgetall"            ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_NO  ,cn20ParamExtGetAll
                              ,"[<property>=<value>]                              "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_extgetallnext"        ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_NO  ,cn20ParamExtGetAllNext
                              ,"[<property>=<value>]                              "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getdefault"           ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetDefault
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getexplain"           ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetExplain
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getfull"              ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetFull
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_gethelp"              ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetHelp
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_gethistory"           ,true ,false ,DBNAME_YES_CN00,    0,   5 ,CMD_PROT_NO  ,cn20ParamHistory 
                              ,"                                                  "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_gethistorynext"       ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn20ParamHistoryNext 
                              ,"                                                  "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getproperties"        ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetProperties
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_gettype"              ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetType
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getvalue"             ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn20ParamGetValue
                              ,"<keyname>                                         "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getvolsall"           ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_NO  ,cn20ParamGetDevSpaceAll
                              ,"[<mod>]                                           "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getdevsall"           ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_NO  ,cn20ParamGetDevSpaceAll
 /*** obsolete ***/           ,"(obsolete version of param_getvolsall)            "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getvolume"            ,true ,false ,DBNAME_YES_CN00,    2,   2 ,CMD_PROT_NO  ,cn20ParamGetDevSpace
                              ,"<num> <mod>                                       "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_getdevspace"          ,true ,false ,DBNAME_YES_CN00,    2,   2 ,CMD_PROT_NO  ,cn20ParamGetDevSpace
 /*** obsolete ***/           ,"(obsolete version of param_getvolume)             "
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_init"                 ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_NO  ,cn20ParamInit
                              ,"[<instance type>]            (needs param session)"
                              ,DBMRightInstallMgm_CN50 },
{"param_put"                  ,true ,false ,DBNAME_YES_CN00 ,   2,   3 ,CMD_PROT_ALL ,cn20ParamPut
                              ,"<keyname> <value>            (needs param session)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_restore"              ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_ALL ,cn20ParamRestore
                              ,"<version number>                    (needs dbname)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 },
{"param_rmfile"               ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_ALL ,cn20ParamRmFile
                              ,"                                    (needs dbname)"
                              ,DBMRightInstallMgm_CN50 },
{"param_startsession"         ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_ALL ,cn20ParamStartSession
                              ,"                                    (needs dbname)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"param_versions"             ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn20ParamVersions
                              ,"                                    (needs dbname)"
                              ,DBMRightParamFull_CN50 | DBMRightParamCheckWrite_CN50 | DBMRightParamRead_CN50 },
{"pure_execute"               ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_ALL ,cn80PureExecute
                              ,"<statement>                                       "
                              ,DBMRightAccessUtility_CN50 | DBMRightAccessSQL_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"quit"                       ,false,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,cn10_dbmRelease
                              ,NULL
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"recover_req"                ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestoreRequest
                              ,NULL
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore_req"         ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestoreRequest
 /*** obsolete ***/           ,NULL
                              ,DBMRightBackupRestore_CN50 },
{"recover_cancel"             ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreCancel
                              ,"                            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore_cancel"      ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreCancel
 /*** obsolete ***/           ,"(obsolete version of recover_cancel)              "
                              ,DBMRightBackupRestore_CN50 },
{"recover_cancel_req"         ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreCancelRequest
                              ,NULL
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore_cancel_req"  ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreCancelRequest
 /*** obsolete ***/           ,NULL
                              ,DBMRightBackupRestore_CN50 },
{"recover_check"              ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestoreCheck
                              ,"<medium> <savetype> [ExternalBackupID <ext_backup_ID>]\n                         [<nnn>] [UNTIL <date> <time>] (needs service session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_check"       ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestoreCheck
 /*** obsolete ***/           ,"(obsolete version of recover_check)               "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_check_req"          ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestoreCheckRequest
                              ,NULL
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_check_req"   ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestoreCheckRequest
 /*** obsolete ***/           ,NULL
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_config"             ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn31BackupRestoreConfig
                              ,"<medium>                   (needs service session)"
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore_config"      ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn31BackupRestoreConfig
 /*** obsolete ***/           ,"(obsolete version of recover_config)              "
                              ,DBMRightBackupRestore_CN50 },
{"recover_ignore"             ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreIgnore
                              ,"                            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_ignore"      ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreIgnore
 /*** obsolete ***/           ,"(obsolete version of recover_ignore)              "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_ignore_req"         ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreIgnoreRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_ignore_req"  ,true ,false ,DBNAME_YES_CN00  ,  0,   0 ,CMD_PROT_ALL ,cn31BackupRestoreIgnoreRequest
 /*** obsolete ***/           ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_replace"            ,true ,false ,DBNAME_YES_CN00  ,  1,   4 ,CMD_PROT_ALL ,cn31BackupRestoreReplace
                              ,"<medium> [ExternalBackupID <ext_backup_ID>]\n                         [<loc> [<nnn>]]    (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_replace"     ,true ,false ,DBNAME_YES_CN00  ,  1,   4 ,CMD_PROT_ALL ,cn31BackupRestoreReplace
 /*** obsolete ***/           ,"(obsolete version of recover_replace)             "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_replace_req"        ,true ,false ,DBNAME_YES_CN00 ,   1,   3 ,CMD_PROT_ALL ,cn31BackupRestoreReplaceRequest
                              ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_replace_req" ,true ,false ,DBNAME_YES_CN00 ,   1,   3 ,CMD_PROT_ALL ,cn31BackupRestoreReplaceRequest
 /*** obsolete ***/           ,NULL                                                 
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_start"              ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestore
                              ,"<medium> <savetype> [ExternalBackupID <ext_backup_ID>]\n                         [<nnn>] [UNTIL <date> <time>]\n                         [AUTOIGNORE]                  (needs util session)"
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore"             ,true ,false ,DBNAME_YES_CN00  ,  1,   8 ,CMD_PROT_ALL ,cn31BackupRestore
 /*** obsolete ***/           ,"(obsolete version of recover_start)               "
                              ,DBMRightBackupRestore_CN50 },
{"recover_state"              ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn31BackupRestoreState
                              ,"                            (needs backup session)"
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore_state"       ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn31BackupRestoreState
 /*** obsolete ***/           ,"(obsolete version of revocer_state)               "
                              ,DBMRightBackupRestore_CN50 },
{"recover_state_check"        ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn31BackupRestoreStateCheck
                              ,"                           (needs backup session)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"backup_restore_state_check" ,true ,false ,DBNAME_YES_CN00 ,   0,   0 ,CMD_PROT_NO  ,cn31BackupRestoreStateCheck
 /*** obsolete ***/           ,"(obsolete version of recover_state_check)         "
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"recover_volume"             ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn31BackupRestoreDevspace
                              ,"<devicename>                  (needs util session)"
                              ,DBMRightBackupRestore_CN50 },
{"recover_devspace"           ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn31BackupRestoreDevspace
 /*** obsolete ***/           ,"(obsolete version of recover_volume)              "
                              ,DBMRightBackupRestore_CN50 },
{"backup_restore_devspace"    ,true ,false ,DBNAME_YES_CN00  ,  1,   1 ,CMD_PROT_ALL ,cn31BackupRestoreDevspace
 /*** obsolete ***/           ,"(obsolete version of recover_volume)              "
                              ,DBMRightBackupRestore_CN50 },
{"release"                    ,false,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,cn10_dbmRelease
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"show"                       ,true ,false ,DBNAME_YES_CN00,    1,  -1 ,CMD_PROT_NO  ,cn40ShowSomething
                              ,"<SHOW command>                                    "
                              ,DBMRightDBInfoRead_CN50 },
{"show_list"                  ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn40ShowList
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"show_next"                  ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn40ShowNext
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 },
{"service_connect"            ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO    ,cn80ServiceConnect
                              ,"                                    (needs dbname)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"service_release"            ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO    ,cn80ServiceRelease
                              ,"                                    (needs dbname)"
                              ,DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50 },
{"sql_connect"                ,true ,false ,DBNAME_YES_CN00,    0,   2 ,CMD_PROT_NO    ,cn80SQLUserConnect
                              ,"[<user>,<password>]                 (needs dbname)"
                              ,DBMRightAccessSQL_CN50 | DBMRightDBInfoRead_CN50 },
{"sql_execute"                ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_NOERR ,cn80SQLUserExecute
                              ,"<statement>                                       "
                              ,DBMRightAccessSQL_CN50 },
{"sql_fetch"                  ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn80SQLUserFetch
                              ,"                                                  "
                              ,DBMRightAccessSQL_CN50 },
{"sql_info"                   ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_NO  ,cn80SQLUserInfo
                              ,"<statement>                                       "
                              ,DBMRightAccessSQL_CN50 },
{"sql_recreateindex"          ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_ALL  ,cn83RecreateIndex
                              ,"\"<scheme>\".\"<table>\".\"<index>\"              "
                              ,DBMRightAccessSQL_CN50 },
{"sql_release"                ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn80SQLUserRelease
                              ,"                                                  "
                              ,DBMRightAccessSQL_CN50 | DBMRightDBInfoRead_CN50 },
{"sql_updatestat"             ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_ALL  ,cn83UpdateStat
                              ,"[<table or column spec>]                          "
                              ,DBMRightAccessSQL_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"test"                       ,false,false ,DBNAME_EGAL_CN00,   0,  -1 ,CMD_PROT_ALL  ,cn40Test
                              ,NULL
                              ,DBMMaskNothing_CN50 },
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"user_create"                ,true ,false ,DBNAME_YES_CN00,    1,   2 ,CMD_PROT_YES ,cn41UserCreate
                              ,"<username>,<pwd> [<templateuser>]   (needs dbname)"
                              ,DBMRightUserMgm_CN50 },
{"user_delete"                ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_ALL ,cn41UserDelete
                              ,"<username>                          (needs dbname)"
                              ,DBMRightUserMgm_CN50 },
{"user_get"                   ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn41UserGet
                              ,"<username>                          (needs dbname)"
                              ,DBMRightUserMgm_CN50 },
{"user_getall"                ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn41UserGetAll
                              ,"                                    (needs dbname)"
                              ,DBMRightUserMgm_CN50 },
{"user_getrights"             ,true ,false ,DBNAME_YES_CN00,    2,   2 ,CMD_PROT_NO  ,cn41UserGetRights
                              ,"<username> SERVERRIGHTS|GUIRIGHTS   (needs dbname)"
                              ,DBMRightUserMgm_CN50 },
{"user_logon"                 ,false,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_NO  ,cn41UserLogon
                              ,"<username>,<pwd>                    (needs dbname)"
                              ,DBMMaskNothing_CN50 },
{"user_put"                   ,true ,false ,DBNAME_YES_CN00,    2,  -1, 1            ,cn41UserPut
                              ,"<username> <property>=<value> ...   (needs dbname)"
                              ,DBMRightUserMgm_CN50 },
{"user_sap"                   ,false,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_YES  ,cn41UserSap
                              ,NULL                                                
                              ,DBMRightUserMgm_CN50 },
{"user_system"                ,false,false ,DBNAME_EGAL_CN00,   1,   1 ,CMD_PROT_NO  ,cn41UserSystem
                              ,NULL                                                
                              ,DBMMaskNothing_CN50 },
{"util_activate"              ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_YES ,cn80UtilActivate
                              ,"<sysdba>,<password>           (needs util session)"
                              ,DBMRightAccessUtility_CN50 },
{"util_connect"               ,true ,false ,DBNAME_YES_CN00,    0,   2 ,CMD_PROT_NO  ,cn80UtilConnect
                              ,"[<user>,<password>]                 (needs dbname)"
                              ,DBMRightAccessUtility_CN50 | DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50},
{"util_execute"               ,true ,false ,DBNAME_YES_CN00,    0,  -1 ,CMD_PROT_ALL ,cn80UtilExecute
                              ,"<statement>                   (needs util session)"
                              ,DBMRightAccessUtility_CN50 },
{"util_getpage"               ,true ,false ,DBNAME_YES_CN00,    1,   3 ,CMD_PROT_NO  ,cn80UtilGetPage
                              ,"[<devmode> <devno>] <pageno>  (needs util session)"
                              ,DBMRightAccessUtility_CN50 },
{"util_putpage"               ,true ,false ,DBNAME_YES_CN00,    1,  -1 ,CMD_PROT_YES ,cn80UtilPutPage
                              ,"[<devmode> <devno> <pageno>] <size> <NL> <page>   "
                              ,DBMRightAccessUtility_CN50 },
{"util_release"               ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_NO  ,cn80UtilRelease
                              ,"                              (needs util session)"
                              ,DBMRightAccessUtility_CN50 | DBMRightBackupRestore_CN50 | DBMRightBackupSave_CN50},
{""                           ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"trace_show"                 ,true ,false ,DBNAME_YES_CN00,    0,   1 ,CMD_PROT_YES ,cn84TraceShow
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{"trace_on"                   ,true ,false ,DBNAME_YES_CN00,    1,  -1 ,CMD_PROT_YES ,cn84TraceOn
                              ,"<option> [<level>]                                "
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{"trace_off"                  ,true ,false ,DBNAME_YES_CN00,    1,  -1 ,CMD_PROT_YES ,cn84TraceOff
                              ,"<option>                                          "
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{"trace_flush"                ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_YES ,cn84TraceFlush
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{"trace_clear"                ,true ,false ,DBNAME_YES_CN00,    0,   0 ,CMD_PROT_YES ,cn84TraceClear
                              ,"                                                  "
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{"trace_protopt"              ,false ,false ,DBNAME_EGAL_CN00,  0,   0 ,CMD_PROT_NO ,cn84TraceProtOpt
                              ,NULL
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{"trace_prot"                 ,true ,false ,DBNAME_YES_CN00,    1,   1 ,CMD_PROT_YES ,cn84TraceProt
                              ,"[a[b[k[m[e[n[s[t[x]]]]]]]]]         (needs dbname)"
                              ,DBMRightDBInfoRead_CN50 | DBMRightDBFileRead_CN50 },
{""                           ,true ,false ,DBNAME_YES_CN00,   0,    0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{"version"                    ,false,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,cn40GetVersion
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 },
{NULL                         ,true ,false ,DBNAME_EGAL_CN00,   0,   0 ,CMD_PROT_NO  ,NULL
                              ,"                                                  "
                              ,DBMMaskNothing_CN50 }
};

/*!  
  -----------------------------------------------------------------------------
  Chapter: static variables
  -----------------------------------------------------------------------------
 */
 VControlDataT * cn00DBMServerData::m_vcontrol = NULL;
/*!  
  -----------------------------------------------------------------------------
  EndChapter: static variables
  -----------------------------------------------------------------------------
 */

/*!  
  -----------------------------------------------------------------------------
  Chapter: Public functions
  -----------------------------------------------------------------------------
 */

/*!
  -------------------------------------------------------------------------
  function:     ctrlservopen
  -------------------------------------------------------------------------
  description:  DBMServer startup

                This function initializes the DBMServer session data. The 
                function will be called by the DBMServer framework at the
                startup of the DBMServer.

  arguments:    vcontrolData   [OUT]  - pointer for DBMServer session data
                dbname         [IN]   - name of database
                dbroot         [IN]   - dbroot path
                pfCancelFlag   [IN]   - the cancel flag
                loggedOn       [IN]   - the logged on flag

  -------------------------------------------------------------------------
*/
void ctrlservopen (
            void      ** vcontrolData,
            const char * dbname,
            const char * dbroot,
            tsp00_Bool * pfCancelFlag,
            tsp00_Bool   loggedOn)
{
  static char szEnvVar[512];

  VControlDataT * vcontrol  = NULL;

  vcontrol = new VControlDataT;

  if (vcontrol != NULL) {
    vcontrol->bSysLogon       = (loggedOn != false);
    vcontrol->bLocalSysLogon  = (loggedOn != false);
    vcontrol->active          = true;
    strcpy (vcontrol->dbname, (dbname == NULL) ? "" : dbname);
    cn90StrUpperCopy(vcontrol->dbname, vcontrol->dbname, false);
    strcpy (vcontrol->dbroot, (dbroot == NULL) ? "" : dbroot);
    vcontrol->pXParamSection  = NULL;
    vcontrol->pUtilSession    = NULL;
    vcontrol->pSQLUserSession = NULL;
    vcontrol->pServiceSession = NULL;
    vcontrol->pBackupSession  = NULL;
    vcontrol->pBackupStageSession=0;
    vcontrol->pHistorySession = 0;
    vcontrol->pEBIDList       = 0;
    vcontrol->pEventSession   = NULL;
    vcontrol->pCommandObj     = NULL;
    vcontrol->pParamHistory   = NULL;
    vcontrol->pfCancelFlag    = (bool *) pfCancelFlag;
    vcontrol->pCurrentUser    = new cn50DBMUser;
    vcontrol->bUnicode        = false;
    vcontrol->szNextCommand.rawAssign("");
    vcontrol->nNextCommandSkip= 0;
    vcontrol->szLastParameter.rawAssign("");

    cn00DBMServerData::setVcontrol(vcontrol);

    // enable global protocol
    if (vcontrol->dbname.length() == 0) {
      tin100_GlobalLog * aLog = tin100_GlobalLog::createObject();
      if (aLog != NULL) {
        tsp00_Pathc szLogFile;
        cn42GetFileName(vcontrol->dbname, FGET_DBMSRV_CN42, szLogFile);
        tin105_SequentialLogFile * aLogFile = new tin105_SequentialLogFile(szLogFile, LOG_ALL_MIN105);

        if (aLogFile != NULL) {
          aLog->addFile(aLogFile, true);
          aLogFile->checkFile();
#ifndef _WIN32
          chmod(szLogFile, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
#endif
        } // end if
      } // end if
    } // end if

    //set DBROOT environment variable (PTS: 1107753)
    if (vcontrol->dbroot.length() > 0) {
      strcpy(szEnvVar, DBROOT_ENV_VAR);
      strcat(szEnvVar, "=");
      strcat(szEnvVar, vcontrol->dbroot);
      putenv(szEnvVar);
    }

  } // end if

  *vcontrolData = (void *) vcontrol;

} /* end ctrlservopen */

/*!
  -------------------------------------------------------------------------
  function:     ctrlserverclose
  -------------------------------------------------------------------------
  description:  DBMServer shutdown

                This function cleans up the DBMServer session data. The 
                function will be called by the DBMServer framework at the
                shutdown of the DBMServer.

  arguments:    vcontrolData   [INOUT]  - pointer for DBMServer session data

  -------------------------------------------------------------------------
*/
void ctrlservclose (
      void * vcontrolData)
{
  VControlDataT * vcontrol = (VControlDataT *) vcontrolData;

  if (vcontrol != NULL) {
    cn20FreeXParamSection(vcontrol);
    cn31FreeBackupSession(vcontrol);
    cn31FreeBackupStageSession(vcontrol);
    cn80FreeUtilSession(vcontrol);
    cn80FreeSQLUserSession(vcontrol);
    cn80FreeServiceSession(vcontrol);
    cn32FreeHistorySession(vcontrol);
    cn32FreeEBIDList(vcontrol);
    cn82FreeEventSession(vcontrol);
    cn40FreeCommandObj(vcontrol);
    cn20FreeParamHistory(vcontrol);
    if (vcontrol->pCurrentUser != NULL) {
      delete vcontrol->pCurrentUser;
      vcontrol->pCurrentUser = NULL;
    } // end if

    delete vcontrol;
  } // end if

  tin100_GlobalLog::deleteObject ();

} // end ctrlserverclose

/*!
  -------------------------------------------------------------------------
  function:     ctrlservcommand
  -------------------------------------------------------------------------
  description:  DBMServer command

                This function executes one DBMServer command. The 
                function will be called by the DBMServer framework.

  arguments:    vcontrolData   [INOUT]  - pointer for DBMServer session data
                requestData    [IN]     - pointer to the request buffer
                requestLen     [IN]     - length of request buffer
                replyData      [INOUT]  - pointer to the reply buffer
                replyLen       [OUT]    - length of replied data buffer
                replyLenMax    [IN]     - max. length of reply buffer
                
  return value: csp9_ctrl_release for stopping the DBMServer, otherwise
                csp9_ctrlcmd_ok

  -------------------------------------------------------------------------
*/
int ctrlservcommand (
      void   * vcontrolData,
      char   * requestData,
      int      requestLen,
      char   * replyData,
      int    * replyLen,
      int      replyLenMax)
{
  FUNCTION_DBG_MCN00_1(_T("vcn10::ctrlservcommand"));

  VControlDataT      * vcontrol  = (VControlDataT*) vcontrolData;;
  CommandT             command;
  tcn00_Error          nCmdRc    = OK_CN00;

  *replyLen = 0;

  // analyze request command
  requestData[requestLen] = CHAR_STRINGTERM_CN90;
  replyLenMax = replyLenMax - requestLen - 200;
  cn10_AnalyzeCommand (requestData, requestLen, command);

  if (command.nIndex != CommandT::InvalidIndex) {

    // log command
//    namesC[command.nIndex].nTokens2Log = CMD_PROT_ALL;
    if ((namesC[command.nIndex].nTokens2Log != CMD_PROT_NO   ) &&
        (namesC[command.nIndex].nTokens2Log != CMD_PROT_NOERR)    ) {
      cn10_LogCommand(&command);
    } // end if

    if (namesC[command.nIndex].name == NULL) {
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_COMMAND_CN00);
    } else if  (namesC[command.nIndex].nDBNameState == DBNAME_YES_CN00 &&
                vcontrol->dbname[0] == CHAR_STRINGTERM_CN90 ) {
      // dbname required
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_NODBNAME_CN00);
    } else if  (namesC[command.nIndex].nDBNameState == DBNAME_NO_CN00  &&
                vcontrol->dbname[0] != CHAR_STRINGTERM_CN90 ) {
      // no dbname
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_DBNAME_CN00);
    } else if (namesC[command.nIndex].bLogon && !vcontrol->pCurrentUser->isLoggedOn()) {
      // Logon check
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_NEEDLOGON_CN00);
    } else if  (namesC[command.nIndex].bSysLogon   && !vcontrol->bSysLogon) {
      // system logon
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_NEEDSYSLOGON_CN00);
    } else if  (!vcontrol->pCurrentUser->isServerRight(namesC[command.nIndex].nRights)) {
      // check rights
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_MISSRIGHT_CN00);
    } else if  (namesC[command.nIndex].nMinParam > 0 &&
                !cn90GetToken(command.args, NULL, namesC[command.nIndex].nMinParam)) {
      // min param count
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_TOOFEWPARAM_CN00);
    } else if  (namesC[command.nIndex].nMaxParam > -1 && cn90GetToken(command.args, NULL, namesC[command.nIndex].nMaxParam + 1)) {
      // max param count
      nCmdRc = cn90AnswerIError(replyData, replyLen, ERR_TOOMANYPARAM_CN00);
    } else {
      if ( namesC[command.nIndex].cnFunction != 0 ) {
        // remove old next command
        if (namesC[command.nIndex].cnFunction != cn40DBMNext) {
          vcontrol->szNextCommand.rawAssign("");
          vcontrol->nNextCommandSkip = 0;
        } // end if
        // call function pointer
        nCmdRc = namesC[command.nIndex].cnFunction (vcontrol, &command, replyData, replyLen, replyLenMax);
      } else {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_COMMAND_CN00_1, namesC[command.nIndex].name);
        nCmdRc = cn90AnswerEvent(replyData, replyLen, aEvent);
      } // end if
    } // end if

    // in case of error log command always
    if ((namesC[command.nIndex].nTokens2Log == CMD_PROT_NO     ) && 
        (nCmdRc                             != OK_CN00         ) && 
        (nCmdRc                             != ERR_NOREPLY_CN00) && 
        (vcontrol->active                                      )    ) {
      cn10_LogCommand(&command);
    } // end if

  } else {
    _TCHAR  szToken[PARAMETER_MAXLEN_CN90];

    cn90GetToken(requestData, szToken, 1);
    teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_COMMAND_CN00_1, szToken);
    nCmdRc = cn90AnswerEvent(replyData, replyLen, aEvent);
  } // end if

  // in case of error log answer
  if ((nCmdRc                             != OK_CN00         ) &&
      (nCmdRc                             != ERR_NOREPLY_CN00) &&  
      (namesC[command.nIndex].nTokens2Log != CMD_PROT_NOERR  )    ) {
    cn10_LogError(replyData, nCmdRc);
  } // end if

  return (vcontrol->active) ? csp9_ctrlcmd_ok : csp9_ctrl_release;
} // end ctrlservcommand

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Public functions
  -----------------------------------------------------------------------------
 */

/*
  -----------------------------------------------------------------------------
  IMPLEMENTATION PRIVATE FUNCTION cn10_LogCommand
  -----------------------------------------------------------------------------
*/
static void cn10_LogCommand
      ( CommandT   * command )
{
  int nPosition =  0;
  int nMaxLen   =  0;
  int nToken    =  0;

  if (namesC[command->nIndex].nTokens2Log == CMD_PROT_ALL) {
    nPosition = command->argLen + 1;
  } else {
    while (nPosition < command->argLen && nToken < namesC[command->nIndex].nTokens2Log) {
      ++nPosition;
      if (isspace(command->args[nPosition])     != 0 &&
          isspace(command->args[nPosition - 1]) == 0    ) {
        ++nToken;
      } // end if
    } // end while
  } // end if

  nMaxLen = MAX_ARGUMENT_LENGTH_EO200 - (int)_tcslen((namesC[command->nIndex].name == NULL) ? _T("") : namesC[command->nIndex].name) - 1;
  nPosition = (nPosition > nMaxLen) ? nMaxLen : nPosition;

  // Log Command
  teo200_EventList aEvent(_T(""), INF_DBMCMD_CN00_1,
                                  (namesC[command->nIndex].name == NULL) ? _T("") : namesC[command->nIndex].name,
                                  nPosition ,
                                  command->args);

  tin100_GlobalLog::writeEntry (aEvent);

} // end cn10_LogCommand

/*
  -----------------------------------------------------------------------------
  IMPLEMENTATION PRIVATE FUNCTION cn10_LogError
  -----------------------------------------------------------------------------
*/
static void cn10_LogError
      ( const _TCHAR      * szError,
        const tcn00_Error   nCmdRc ) 
{
  const _TCHAR * pError     = strstr(szError, LINE_SEPSTRING_CN00);
  const _TCHAR * pSecond    = NULL;

  if (pError != NULL) {
    pError = pError + strlen(LINE_SEPSTRING_CN00);
    pError = strstr(pError, TOKEN_SEPSTRING_CN00);
    if (pError != NULL) {
      pError++;

      int nPosition  = (int)(strstr(pError, LINE_SEPSTRING_CN00) - pError);

      nPosition = (nPosition < 0) ? 80 : nPosition;

      pSecond = strstr(pError, LINE_SEPSTRING_CN00);
      if (pSecond != NULL) {
        pSecond = pSecond + strlen(LINE_SEPSTRING_CN00);
        if (_tcslen(pSecond) > 0) {
          int nSecond = (int)(strstr(pSecond, LINE_SEPSTRING_CN00) - pSecond);
          nSecond = (nSecond < 0) ? 80 : nSecond;

          // Log Command
          teo200_EventList aSecond(_T(""), nCmdRc,  TERR_CN00_1, _T("DBM"), _T("%.*s"), nSecond, pSecond);
          teo200_EventList aEvent (aSecond, _T(""), nCmdRc,  TERR_CN00_1, _T("DBM"), _T("%.*s"), nPosition, pError);
          tin100_GlobalLog::writeEntry (aEvent);
        } else {
          // Log Command
          teo200_EventList aEvent(_T(""), nCmdRc,  TERR_CN00_1, _T("DBM"), _T("%.*s"), nPosition, pError);
          tin100_GlobalLog::writeEntry (aEvent);
        } // end if

      } else {
        // Log Command
        teo200_EventList aEvent(_T(""), nCmdRc,  TERR_CN00_1, _T("DBM"), _T("%.*s"), nPosition, pError);
        tin100_GlobalLog::writeEntry (aEvent);
      } // end if

    } // end if
  } // end if

} // end cn10_LogError

/*
  -----------------------------------------------------------------------------
  IMPLEMENTATION PRIVATE FUNCTION cn10_AnalyzeCommand
  -----------------------------------------------------------------------------
*/
static void cn10_AnalyzeCommand
      ( const _TCHAR * szData,
        const int      nDataLen,
        CommandT     & DBMCommand )
{
  int            nIndex = 0;
  const _TCHAR * pArgs  = NULL;

  DBMCommand.nIndex = CommandT::InvalidIndex;
  pArgs             = szData;
  DBMCommand.argLen = nDataLen;

  while (isspace(*pArgs) != 0) {
    ++pArgs;
    --DBMCommand.argLen;
  } // end while

  for (nIndex = 0;namesC[nIndex].name != NULL && DBMCommand.nIndex == -1; ++nIndex) {

    if (cn90CheckKeyword(pArgs, namesC[nIndex].name, false)) {
      DBMCommand.nIndex = nIndex;
      pArgs             = pArgs             + _tcslen (namesC[nIndex].name);
      DBMCommand.argLen = DBMCommand.argLen - (int)_tcslen (namesC[nIndex].name);
      while (isspace(*pArgs) != 0) {
        ++pArgs;
        --DBMCommand.argLen;
      } // end while
      DBMCommand.args = pArgs;
    } // end if

  } // end for

} // end cn10_AnalyzeCommand

/*
  -----------------------------------------------------------------------------
  IMPLEMENTATION PRIVATE FUNCTION cn10_showHelp
  -----------------------------------------------------------------------------
*/
static tcn00_Error cn10_showHelp (
    VControlDataT * vcontrol,
    CommandT      * command,
    char          * replyData,
    int           * replyLen,
    int             replyLenMax)
{
  _TCHAR       * pCurrPos  = NULL;
  bool           bCommand  = false;
  bool           bObsolete = false;
  int            nIndex    = 0;
  _TCHAR         szCommand[PARAMETER_MAXLEN_CN90];

  // Init Output
  _stprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
  pCurrPos = replyData + _tcslen (replyData);

  // read requested command(s)
  cn90GetToken(command->args, szCommand, 1);
  if (_tcscmp(szCommand, "-obsolete") == 0) {
    bObsolete = true;
    bCommand     = cn90GetToken(command->args, szCommand, 2);
  } else {
    bCommand = true;
  } // end if
  
  // search command(s)
  while (namesC[nIndex].name != NULL) {

    if ( (!bCommand || ( _tcsnicmp(szCommand, namesC[nIndex].name, _tcslen(szCommand)) == 0)) &&
         ( namesC[nIndex].shortHelp                                      != NULL )               ) {
        if ( (  bObsolete && (_tcsnicmp(OBSOLETE_MARK, namesC[nIndex].shortHelp, _tcslen(OBSOLETE_MARK)) == 0) ) ||
             ( !bObsolete && (_tcsnicmp(OBSOLETE_MARK, namesC[nIndex].shortHelp, _tcslen(OBSOLETE_MARK)) != 0) )    ) {
          _stprintf (pCurrPos, "%-25s%s%s", namesC[nIndex].name, namesC[nIndex].shortHelp, LINE_SEPSTRING_CN00);
        } else {
          // always show exakt match
          if ( bCommand && ( _tcsicmp(szCommand, namesC[nIndex].name) == 0) ) {
            _stprintf (pCurrPos, "%-25s%s%s", namesC[nIndex].name, namesC[nIndex].shortHelp, LINE_SEPSTRING_CN00);
          } // end if
        } // end if
    } // end if

    pCurrPos = pCurrPos + _tcslen (pCurrPos);

    ++nIndex;
  } // end while

  *replyLen = (int)_tcslen(replyData);

  return OK_CN00;
} // end cn10_showHelp

/*
  -----------------------------------------------------------------------------
  IMPLEMENTATION PRIVATE FUNCTION cn10_dbmRelease
  -----------------------------------------------------------------------------
*/
static tcn00_Error cn10_dbmRelease (
    VControlDataT * vcontrol,
    CommandT      * command,
    char          * replyData,
    int           * replyLen,
    int             replyLenMax)
{
  vcontrol->active = false;
  *replyLen        = 0;

  return OK_CN00;
} // end cn10_dbmRelease
