/*!
  -----------------------------------------------------------------------------

  module: vcn28.cpp

  -----------------------------------------------------------------------------

  responsible:  BerndV

  special area: DBMServer Parameter File Migration

  description:  DBMServer Parameter File Migration - Implementation

  version:      7.4.*

  -----------------------------------------------------------------------------

  copyright: by SAP AG, 2001

    ========== licence begin LGPL
    Copyright (C) 2001 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

  -----------------------------------------------------------------------------
*/

/* 
  -------------------------------------------------------------------------
  includes
  -------------------------------------------------------------------------
 */

#include <stdlib.h>

#include "hsp100.h"
#include "hcn28.h"
#include "hcn29.h"
#include "hcn90.h"

/*
  -------------------------------------------------------------------------
  function:     cn20_Migrate_ArchiveSize
  -------------------------------------------------------------------------
 */
static tcn00_Error cn20_Migrate_ArchiveSize 
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        char              * replyData,
        int               * replyLen)

{
  FUNCTION_DBG_MCN00_1("cn20_Migrate_ArchiveSize");

  tcn00_Error          nFuncReturn = OK_CN00;
  SAPDBErr_MessageList oMsg;
  SAPDB_Bool           bFileExist  = false;
  RTEConf_Parameter    oConfigFile(szDbName.asCharp(), 
                                   0, 
                                   RTECONF_MAXNAMELENGTH,
                                   RTECONF_MAXSTRINGLENGTH);

  // open param file
  if (!oConfigFile.Read(bFileExist, oMsg)) {
    SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
    aDBM.AppendNewMessage(oMsg);
    nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
  } else {

    // read LOG Volumes
    RTEConf_Parameter::Integer nSize;
    RTEConf_Parameter::Integer nVolumes;
    int                        nVolumeIndex;
    char                       szName[RTECONF_MAXNAMELENGTH+1];
      
    tcn002_XParameter * pParameter = NULL;
    tcn002_XParameter * pDynHelp   = NULL;
    bool                bMallocOK  = true;

    if (!oConfigFile.GetValue ((RTEConf_Parameter::Name) PAN_MAXARCHIVELOGS, nVolumes, oMsg)) {
      SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
      aDBM.AppendNewMessage(oMsg);
      nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
    } // end if

    for (nVolumeIndex = 1; (nVolumeIndex <= nVolumes) && (nFuncReturn == OK_CN00); ++nVolumeIndex) {

      sprintf(szName, PAN_ARCHIVE_LOG_N_SIZE"%02d", nVolumeIndex);
      if (oConfigFile.GetValue ((RTEConf_Parameter::Name) szName, nSize, oMsg)) {
        sprintf(szName, PAN_ARCHIVE_LOG_N_SIZE"%03d", nVolumeIndex);
        // write log size
        if (nFuncReturn == OK_CN00) {

          pDynHelp = cn29SearchDynamic(pXParamData->pFirst, szName);
          if (pDynHelp != NULL) {
            cn00MyMalloc (sizeof (tcn002_XParameter), &pParameter, &bMallocOK);
            if ((pParameter != NULL) && (bMallocOK)) {
              /* copy data from dynamic record */
              nFuncReturn = cn29CopyParameter(pParameter, pDynHelp);
              /* clear the dynamic/intern property */
              pParameter->xpProperties[XPPropClassDynamic] = XPPropValueNo;
              pParameter->xpProperties[XPPropClassIntern] = XPPropValueNo;

              strcpy(pParameter->szParameterName, szName);

              /* list management (insert after dynamic) */
              pParameter->pNext = pDynHelp->pNext;
              pDynHelp->pNext   = pParameter;
              pParameter->pPrev = pDynHelp;
              if (pParameter->pNext != NULL) {
                pParameter->pNext->pPrev = pParameter;
              } else {
                pXParamData->pLast = pParameter;
              } // end if
            } // end if
          } // end if

          if (pParameter != NULL) {
            pParameter->xpRecord.xpValue.xpValueInt = nSize;
            if (!oConfigFile.SetValue ((RTEConf_Parameter::Name) pParameter->szParameterName, pParameter->xpRecord.xpValue.xpValueInt, oMsg)) {
              SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
              aDBM.AppendNewMessage(oMsg);
              nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
            } else {
              cn29ParamValueToString(pParameter->szNewValue, &(pParameter->xpRecord));
              cn29WriteParamHistory(szDbName, pParameter, PAHI_CHANGE_CN002);
            } // end if
          } else {
            nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPNOTFOUND_CN00);
          } // end if
        } // end if

      } // end if
    } // end for

    // close param file
    if (nFuncReturn == OK_CN00) {
      if (!oConfigFile.Write(oMsg)) {
        SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
        aDBM.AppendNewMessage(oMsg);
        nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
      } // end if
    } // end if

  } // end if
  
  return nFuncReturn;
} // end cn20_Migrate_ArchiveSize

/*
  -------------------------------------------------------------------------
  function:     cn20_MigrateFrom_7_2_5_or_7_3_0
  -------------------------------------------------------------------------
 */
static tcn00_Error cn20_MigrateFrom_7_2_5_or_7_3_0 
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        char              * replyData,
        int               * replyLen)

{
  FUNCTION_DBG_MCN00_1("cn20_MigrateFrom_7_2_5_or_7_3_0");

  tcn00_Error          nFuncReturn = OK_CN00;
  SAPDBErr_MessageList oMsg;
  SAPDB_Bool           bFileExist  = false;
  RTEConf_Parameter    oConfigFile(szDbName.asCharp(), 
                                   0, 
                                   RTECONF_MAXNAMELENGTH,
                                   RTECONF_MAXSTRINGLENGTH);

  // open param file
  if (!oConfigFile.Read(bFileExist, oMsg)) {
    SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
    aDBM.AppendNewMessage(oMsg);
    nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
  } else {

    // Parameter to write
    tcn002_XParameter * pParameter = NULL;
    
    // CACHE_SIZE = DATA_CACHE + CONVERTER_CACHE
    RTEConf_Parameter::Integer nDatacache;
    RTEConf_Parameter::Integer nConvertercache;

    // read DATA_CACHE
    if (nFuncReturn == OK_CN00) {
      if (!oConfigFile.GetValue ((const SAPDB_UTF8 *) PAN_DATA_CACHE_PAGS, nDatacache, oMsg)) {
        SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
        aDBM.AppendNewMessage(oMsg);
        nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
      } // end if
    } // end if

    // read CONVERTER_CACHE
    if (nFuncReturn == OK_CN00) {
      if (!oConfigFile.GetValue ((const SAPDB_UTF8 *) PAN_CONV_CACHE_PAGES, nConvertercache, oMsg)) {
        SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
        aDBM.AppendNewMessage(oMsg);
        nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
      } // end if
    } // end if

    // write CACHE_SIZE
    if (nFuncReturn == OK_CN00) {
      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_CACHE_SIZE);
      if (pParameter != NULL) {
        pParameter->xpRecord.xpValue.xpValueInt = nDatacache + nConvertercache;
        if (!oConfigFile.SetValue ((const SAPDB_UTF8 *) pParameter->szParameterName, pParameter->xpRecord.xpValue.xpValueInt, oMsg)) {
          SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
          aDBM.AppendNewMessage(oMsg);
          nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
        } else {
          cn29ParamValueToString(pParameter->szNewValue, &(pParameter->xpRecord));
          cn29WriteParamHistory(szDbName, pParameter, PAHI_CHANGE_CN002);
        } // end if
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPNOTFOUND_CN00);
      } // end if
    } // end if

    // ready stamp kernelversion
    if (nFuncReturn == OK_CN00) {
      pParameter = cn29SearchParameter(pXParamData->pFirst, PAN_PARAM_KERNELVERSION);
      if (pParameter != NULL) {
        // get own version
        tsp00_Versionc      VersionString;
        sp100_GetVersionString(COMP_NAME_KERNEL_SP100, s100buildnumber, VersionString);
        
        cn29StringToParamValue(&(pParameter->xpRecord), VersionString.asCharp());

        RTEConf_Parameter::String pString = &(pParameter->xpRecord.xpValue.xpValueString[0]);

        if (!oConfigFile.SetValue ((const SAPDB_UTF8 *) pParameter->szParameterName, pString, oMsg)) {
          SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
          aDBM.AppendNewMessage(oMsg);
          nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
        } else {
          cn29ParamValueToString(pParameter->szNewValue, &(pParameter->xpRecord));
          cn29WriteParamHistory(szDbName, pParameter, PAHI_CHANGE_CN002);
        } // end if
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_XPNOTFOUND_CN00);
      } // end if
    } // end if

    // close param file
    if (nFuncReturn == OK_CN00) {
      if (!oConfigFile.Write(oMsg)) {
        SAPDBErr_MessageList aDBM(DBMSrv_DBMError(RTE), 0);
        aDBM.AppendNewMessage(oMsg);
        nFuncReturn = cn90AnswerMessage (replyData, replyLen, aDBM);
      } // end if
    } // end if

  } // end if
  
  return nFuncReturn;
} // end cn20_MigrateFrom_7_2_5_or_7_3_0
/*
  -------------------------------------------------------------------------
  function:     cn28_AnalyzeVersion
  -------------------------------------------------------------------------
 */
static bool cn28_AnalyzeVersion 
      ( const tsp00_C256c  &  szVersionString,
        tsp100_VersionID1  &  VersionID1)
{
  tsp00_C64c szVersion;
  tsp00_C64c szBuild;
  tsp00_C64c szTemp;
  char *     pBegin;
  char *     pEnd;
  
  // extract version information from string
  strncpy(szVersion, &szVersionString[BEG_OF_1ST_VERS_SUBSTR_SP100], END_OF_1ST_VERS_SUBSTR_SP100 - BEG_OF_1ST_VERS_SUBSTR_SP100 + 1);
  szVersion[END_OF_1ST_VERS_SUBSTR_SP100 - BEG_OF_1ST_VERS_SUBSTR_SP100 + 1] = 0;

  // extract build information from string
  strncpy(szBuild, &szVersionString[BEG_OF_2CND_VERS_SUBSTR_SP100], END_OF_2CND_VERS_SUBSTR_SP100 - BEG_OF_2CND_VERS_SUBSTR_SP100 + 1);
  szBuild[END_OF_2CND_VERS_SUBSTR_SP100 - BEG_OF_2CND_VERS_SUBSTR_SP100 + 1] = 0;

  // major version  
  pBegin = &szVersion[0];
  pEnd   = strchr(pBegin, '.');
  if (pBegin != NULL && pEnd != NULL) {
    strncpy(szTemp, pBegin, pEnd - pBegin);
    szTemp[(int) (pEnd - pBegin)] = 0;
    VersionID1.MajorVersion_sp100 = (unsigned char) atoi(szTemp);
    pBegin = pEnd + 1;
    pEnd   = strchr(pBegin, '.');
  } else {
    return false;
  } // end if

  // minor version  
  if (pBegin != NULL && pEnd != NULL) {
    strncpy(szTemp, pBegin, pEnd - pBegin);
    szTemp[(int) (pEnd - pBegin)] = 0;
    VersionID1.MinorVersion_sp100 = (unsigned char) atoi(szTemp);
    pBegin = pEnd + 1;
  } else {
    return false;
  } // end if

  // correction level
  if (pBegin != NULL) {
    VersionID1.CorrLevel_sp100 = (unsigned char) atoi(pBegin);
  } else {
    return false;
  } // end if
  
  // build
  VersionID1.BuildNumberPrefix_sp100 = (unsigned char) atoi(szBuild);

  return true;
} // end cn28_AnalyzeVersion

/*
  -------------------------------------------------------------------------
  function:     cn28MigrateParamFile
  -------------------------------------------------------------------------
 */
tcn00_Error cn28MigrateParamFile 
      ( tsp00_DbNamec       szDbName,
        tcn002_XParamData * pXParamData,
        char              * replyData,
        int               * replyLen)

{
  FUNCTION_DBG_MCN00_1("cn28MigrateParamFile");

  tcn00_Error nFuncReturn = OK_CN00;


  SAPDBErr_MessageList oMsg;
  SAPDB_Bool           bFileExist  = false;
  tsp00_C256c          szVersion;
  RTEConf_Parameter    oConfigFile(szDbName.asCharp(), 
                                   0, 
                                   RTECONF_MAXNAMELENGTH,
                                   RTECONF_MAXSTRINGLENGTH);
  tcn002_XPRecord      xpRecord;

  // open param file
  if (!oConfigFile.Read(bFileExist, oMsg)) {
    // key not found or paramfile not exist 
    // don't migrate but success
    return nFuncReturn;
  } else {
    if (!oConfigFile.GetValue((SAPDB_UTF8 *) PAN_PARAM_KERNELVERSION ,xpRecord.xpValue.xpValueString, oMsg)) {
      // key not found or paramfile not exist 
      // don't migrate but success
      return nFuncReturn;
    } else {
      strcpy(szVersion, (char *) xpRecord.xpValue.xpValueString);
    } // end if
  } // end if

  // analyse version string
  tsp100_VersionID1   ParamVersionID1;
  if (!cn28_AnalyzeVersion(szVersion, ParamVersionID1)) {
    teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_BADVERSION_CN00_1, szVersion.asCharp());
    nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
  } // end if
  
  if (nFuncReturn == OK_CN00) {
    // get own version
    tsp100_VersionID1   DBMVersionID1;
    sp100_GetVersionID ( VersionIDType1_esp100, s100buildnumberEx, &DBMVersionID1 );

    // I am should be 7.4.3 (target)
    if ((DBMVersionID1.MajorVersion_sp100        == 7) &&
        (DBMVersionID1.MinorVersion_sp100        == 4) &&
        (DBMVersionID1.CorrLevel_sp100           == 3)    ) {
   
      // check valid sources
      if ((ParamVersionID1.MajorVersion_sp100        == 7) &&
          (ParamVersionID1.MinorVersion_sp100        == 4)    ) {
        // OK can I can use all 7.4 with no change
        nFuncReturn = cn20_Migrate_ArchiveSize(szDbName, pXParamData, replyData, replyLen);
      } else if ((ParamVersionID1.MajorVersion_sp100        ==  7) &&
                 (ParamVersionID1.MinorVersion_sp100        ==  2) &&
                 (ParamVersionID1.CorrLevel_sp100           ==  5) &&
                 (ParamVersionID1.BuildNumberPrefix_sp100   >  14)    ) {
        // Ooops 7.2.5 I love it!
        // place here a conversion 
        nFuncReturn = cn20_MigrateFrom_7_2_5_or_7_3_0(szDbName, pXParamData, replyData, replyLen);

      } else if ((ParamVersionID1.MajorVersion_sp100        ==  7) &&
                 (ParamVersionID1.MinorVersion_sp100        ==  3) &&
                 (ParamVersionID1.CorrLevel_sp100           ==  0) &&
                 (ParamVersionID1.BuildNumberPrefix_sp100   >  20)    ) {
        // 7.3.0 same conversion like 7.2.5
        // place here a conversion 
        nFuncReturn = cn20_MigrateFrom_7_2_5_or_7_3_0(szDbName, pXParamData, replyData, replyLen);

      } else {
        // go away
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_MIGRATE_CN00_1, szVersion.asCharp());
        nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
      } // 
    
    } else {

      // If I am not 7.4.3 make a default check
      if ((DBMVersionID1.MajorVersion_sp100        == ParamVersionID1.MajorVersion_sp100) &&
          (DBMVersionID1.MinorVersion_sp100        == ParamVersionID1.MinorVersion_sp100) /* PTS 1113473 &&
          (DBMVersionID1.CorrLevel_sp100           == ParamVersionID1.CorrLevel_sp100   )  */  ) {
        // OK I can use my own minor release
        // -> Do Nothing!
      } else {
        // But nothing else
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_MIGRATE_CN00_1, szVersion.asCharp());
        nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
      } // end if
    } // end if

  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} // end cn28MigrateParamFile
