/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(CN72) && defined(_WIN32)
  #line 3 "d:\\v72\\sys\\src\\cn\\vcn83.cpp"
#endif 

/*! 
  -----------------------------------------------------------------------------
 
  module: vcn83.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: DBMServer
 
  description:  implementation module for special sql access
                - index handling
                - update statistics
  
  version:      7.2./7.3

  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 1998
 
  -----------------------------------------------------------------------------
*/

/* 
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/
#include "hin110.h"

#include "hcn51.h"
#include "hcn80.h"
#include "hcn90.h"

#include "hcn83.h"

/* 
  -----------------------------------------------------------------------------
  specification private types and macros
  -----------------------------------------------------------------------------
*/

#define SQL_COMMANDLENGTH      1024

#define SQL_BADINDEXBYNAME     _T("SELECT * FROM SYSDD.BAD_INDEXES WHERE OWNER='%s' AND TABLENAME='%s' AND INDEXNAME='%s'")
#define SQL_BADINDEXBYCOL      _T("SELECT * FROM SYSDD.BAD_INDEXES WHERE OWNER='%s' AND TABLENAME='%s' AND COLUMNNAME='%s'")

#define SQL_INDEXBYCOL         _T("SELECT * FROM SYSDD.INDEXES WHERE OWNER='%s' AND TABLENAME='%s' AND COLUMNNAME='%s'")
#define SQL_INDEXBYNAME        _T("SELECT * FROM SYSDD.INDEXES WHERE OWNER='%s' AND TABLENAME='%s' AND INDEXNAME='%s' ORDER BY COLUMNNO ASC")
#define SQL_INDEXCOUNT         _T("SELECT COUNT(*) FROM SYSDD.INDEXES WHERE OWNER='%s' AND TABLENAME='%s' AND INDEXNAME='%s'")

#define SQL_DROPINDEXBYNAME    _T("DROP INDEX \"%s\" ON \"%s\".\"%s\"")
#define SQL_DROPINDEXBYCOL     _T("DROP INDEX \"%s\".\"%s\".\"%s\"")


#define SQL_CREATEINDEXBYNAME1 _T("CREATE %s INDEX \"%s\" ON \"%s\".\"%s\" (%s) ")
#define SQL_CREATEINDEXBYNAME2 _T("\"%s\" %s,")

#define SQL_CREATEINDEXBYCOL   _T("CREATE %s INDEX \"%s\".\"%s\".\"%s\" %s")

#define SQL_RESTART            _T("RESTART INDEX")

#define SQLF_TYPE              _T("TYPE")
#define SQLF_SORT              _T("SORT")
#define SQLF_COLUMNNAME        _T("COLUMNNAME")

#define SQL_UPDATESTAT         _T("UPDATE STAT %s")
#define SQL_UPDATESTAT_ALL     _T("UPDATE STAT *")

/* 
  -----------------------------------------------------------------------------
  specification private functions
  -----------------------------------------------------------------------------
*/
static tcn00_Error cn83_RecreateIndex
      ( const tsp00_DbNamec & pDBName,
        const _TCHAR  * pScheme,
        const _TCHAR  * pTable,
        const _TCHAR  * pIndex,
        char          * replyData,
        int           * replyLen);

static tcn00_Error cn83_RestartIndex
      ( const tsp00_DbNamec & pDBName,
        char          * replyData,
        int           * replyLen);

static tcn00_Error cn83_UpdateStat
      ( const tsp00_DbNamec & pDBName,
        const _TCHAR  * pSpec,
        char          * replyData,
        int           * replyLen );
        
/* 
  -----------------------------------------------------------------------------
  function:     cn83RecreateIndex
  -----------------------------------------------------------------------------
 */
tcn00_Error cn83RecreateIndex
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn   = OK_CN00;

  _TCHAR              szIndexSpec[PARAMETER_MAXLEN_CN90];
  _TCHAR            * pScheme       = NULL;
  _TCHAR            * pTable        = NULL;
  _TCHAR            * pIndex        = NULL;

  // analyze parameters
  if (cn90GetToken(command->args, szIndexSpec, 1)) {
    pScheme = &szIndexSpec[0];
    pTable  = strstr(szIndexSpec, ".");
    if (pTable == NULL) {
      nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } else {
      *pTable = CHAR_STRINGTERM_CN90;
      ++pTable;
      pIndex  = strstr(pTable, ".");
      if (pIndex == NULL) {
        nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } else {
        *pIndex = CHAR_STRINGTERM_CN90;
        ++pIndex;
      } // end if
    } // end if

    if (nFuncReturn == OK_CN00) {
      cn90StripLeadingChar(cn90StripTrailingChar(pScheme, '"'), '"');
      cn90StripLeadingChar(cn90StripTrailingChar(pTable,  '"'), '"');
      cn90StripLeadingChar(cn90StripTrailingChar(pIndex,  '"'), '"');

      nFuncReturn = cn83_RecreateIndex (vcontrol->dbname, pScheme, pTable, pIndex, replyData, replyLen);

    } // end if

  } else {
    // no index specified -> restart index
    nFuncReturn = cn83_RestartIndex (vcontrol->dbname, replyData, replyLen);
  } // end if



  if (nFuncReturn == OK_CN00) {
    //cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} // end cn83RecreateIndex

/* 
  -----------------------------------------------------------------------------
  private function:     cn83_RecreateIndex
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn83_RecreateIndex
      ( const tsp00_DbNamec & pDBName,
        const _TCHAR  * pScheme,
        const _TCHAR  * pTable,
        const _TCHAR  * pIndex,
        char          * replyData,
        int           * replyLen )
{
  FUNCTION_DBG_MCN00_1(_T("cn83_RecreateIndex"));

  tcn00_Error             nFuncReturn    = OK_CN00;

  tin01_sql_session       SQLSession;
  tin110_SQLResultTable   aTable(&SQLSession);
  tin110_SQLRecord        aRecord;
  
  _TCHAR                  szCommand     [SQL_COMMANDLENGTH];
  _TCHAR                  szDropCommand [SQL_COMMANDLENGTH];
  _TCHAR                * pCreateCommand = NULL;
  _TCHAR                * pColumns       = NULL;
  _TCHAR                * pCurrent       = NULL;

  _TCHAR                  szType  [10]   = _T("");
  _TCHAR                  szSort  [10]   = _T("");
  tsp00_KnlIdentifierc    szColumn;

  bool                    bNamed = false;
  tsp00_Int4              nColumns = 0;
    
  nFuncReturn = cn80ConnectSQL(pDBName, &SQLSession, replyData, replyLen);

  // check sysdd.bad_indexes for existence and named property
  if (nFuncReturn == OK_CN00) {
    _stprintf(szCommand, SQL_BADINDEXBYNAME, pScheme, pTable, pIndex);
    aTable.execute(szCommand);
    if (aTable.sqlRc() == cin01_db_ok) {
      bNamed = true;
    } else if (aTable.sqlRc() == cin01_db_row_not_found) {
      _stprintf(szCommand, SQL_BADINDEXBYCOL, pScheme, pTable, pIndex);
      aTable.execute(szCommand);
      if (aTable.sqlRc() == cin01_db_row_not_found) {
        // oops index not found
        teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
        nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
      } else if (aTable.sqlRc() != cin01_db_ok) {
        // oops other sql error
        teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
        nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
      } // end if
    } else {
      // oops sql error
      teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
      nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
    } // end if

    // generate drop command
    if (nFuncReturn == OK_CN00) {
      if (bNamed) {
        _stprintf(szDropCommand, SQL_DROPINDEXBYNAME, pIndex, pScheme, pTable);
      } else {
        _stprintf(szDropCommand, SQL_DROPINDEXBYCOL, pScheme, pTable, pIndex);
      } // end if
    } // end if

    // generate create command
    if (nFuncReturn == OK_CN00) {

      if (!bNamed) { // for unamed index

        _stprintf(szCommand, SQL_INDEXBYCOL, pScheme, pTable, pIndex);
        aTable.execute(szCommand);

        if (aTable.sqlRc() == cin01_db_ok) {
          aRecord = aTable.record();

          _tcscpy(szType, aRecord.fieldAsChar(SQLF_TYPE));
          _tcscpy(szSort, aRecord.fieldAsChar(SQLF_SORT));
        
          pCreateCommand = new _TCHAR[SQL_COMMANDLENGTH];
          _stprintf(pCreateCommand, SQL_CREATEINDEXBYCOL, szType, pScheme, pTable, pIndex, szSort);
        } else {
          teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
          nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
        } // end if

      } else {  // for named index

        // check field count
        _stprintf(szCommand, SQL_INDEXCOUNT, pScheme, pTable, pIndex);
        aTable.execute(szCommand);
        if (aTable.sqlRc() == cin01_db_ok) {
          aRecord = aTable.record();
          nColumns = aRecord.fieldAsInt(1);
        } else {
          teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
          nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
        } // end if

        // check sysdd.indexes
        if (nFuncReturn == OK_CN00) {
          _stprintf(szCommand, SQL_INDEXBYNAME, pScheme, pTable, pIndex);
          aTable.execute(szCommand);

          if (aTable.sqlRc() == cin01_db_ok) {
            aRecord = aTable.record();

            // save type
            _tcscpy(szType, aRecord.fieldAsChar(SQLF_TYPE));

            // allocate memory for fieldlist
            pColumns  = new _TCHAR[nColumns * (sizeof(tsp00_KnlIdentifier) + _tcslen(SQL_CREATEINDEXBYNAME2)) + 1];
            *pColumns = CHAR_STRINGTERM_CN90;
            pCurrent  = pColumns;

            // create fieldlist
            while (aTable.sqlRc() == cin01_db_ok) {
              _tcscpy(szSort,   aRecord.fieldAsChar(SQLF_SORT));
              _tcscpy(szColumn, aRecord.fieldAsChar(SQLF_COLUMNNAME));
              _stprintf(pCurrent, SQL_CREATEINDEXBYNAME2, szColumn.asCharp(), szSort);
              pCurrent = pCurrent + _tcslen(pCurrent);
          
              aRecord = aTable.nextRecord();
            } // end while

            // create command
            if (aTable.sqlRc() == cin01_db_row_not_found) {
              --pCurrent;
              *pCurrent = CHAR_STRINGTERM_CN90;
              pCreateCommand = new _TCHAR[SQL_COMMANDLENGTH + _tcslen(pColumns)];
              _stprintf(pCreateCommand, SQL_CREATEINDEXBYNAME1, szType, pIndex, pScheme, pTable, pColumns);

              // free field list
              delete pColumns;
            } else {
              teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
              nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
            } // end if

          } else {
            teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
            nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
          } // end if

        } // end if
      } // end if

      // drop index
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn80ExecuteSQL  ( &SQLSession,
                                        szDropCommand,
                                        replyData,
                                        SQL_COMMANDLENGTH,
                                        *replyLen );
      } // end if

      // create index
      if (nFuncReturn == OK_CN00) {
        nFuncReturn = cn80ExecuteSQL  ( &SQLSession,
                                        pCreateCommand,
                                        replyData,
                                        SQL_COMMANDLENGTH,
                                        *replyLen );
      } // end if

      // free command buffer
      if (pCreateCommand != NULL) {
        delete pCreateCommand;
      } // end if
    } // end if

    // release session with commit or not
    cn80ReleaseSQL  ( &SQLSession, nFuncReturn == OK_CN00);

  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;

} // end cn83_RecreateIndex

/* 
  -----------------------------------------------------------------------------
  private function:     cn83_RestartIndex
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn83_RestartIndex
      ( const tsp00_DbNamec & pDBName,
        char          * replyData,
        int           * replyLen )
{
  tcn00_Error             nFuncReturn    = OK_CN00;
  tin01_sql_session       SQLSession;

  nFuncReturn = cn80ConnectSQL(pDBName, &SQLSession, replyData, replyLen);

  // restart index
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80ExecuteSQL  ( &SQLSession,
                                    SQL_RESTART,
                                    replyData,
                                    SQL_COMMANDLENGTH,
                                    *replyLen );
    // release session with commit or not
    cn80ReleaseSQL  ( &SQLSession, nFuncReturn == OK_CN00);
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;

} // end cn83_RestartIndex

/*
static tcn00_Error cn83_AnalyzeIndex
      ( const tsp00_DbNamec & pDBName,
        const _TCHAR  * pScheme,
        const _TCHAR  * pTable,
        const _TCHAR  * pIndex,
        char          * replyData,
        int           * replyLen )
{
  FUNCTION_DBG_MCN00_1(_T("cn83_AnalyzeIndex"));

  tcn00_Error         nFuncReturn   = OK_CN00;
  tin01_sql_session   SQLSession;

  nFuncReturn = cn80ConnectSQL(pDBName, &SQLSession);

  if (nFuncReturn == OK_CN00) {

    tin110_SQLResultTable aTable(&SQLSession, "select * from indexes");

    if (aTable.sqlRc() == cin01_db_ok) {
      tin110_SQLRecord aRecord = aTable.record();
      int i;
      _TCHAR * pCurrent = replyData;

      // init OK Answer
      _stprintf (pCurrent, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);

      // print column names
      for (i = 0; i < aRecord.fields(); ++i) {
        _stprintf (pCurrent, "%s;", aRecord.fieldName (i+1) );
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if
      --pCurrent;

      _stprintf (pCurrent, "%s", LINE_SEPSTRING_CN00 );
      pCurrent = pCurrent + _tcslen(pCurrent);

      // print records
      while (aTable.sqlRc() == cin01_db_ok) {
        for (i = 0; i < aRecord.fields(); ++i) {
          if (aRecord.isNull(i + 1)) {
            _stprintf (pCurrent, "(null);");
          } else {
            _stprintf (pCurrent, "%s;", aRecord.fieldAsChar (i+1, '\'') );
          } // end if
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if
        --pCurrent;

        _stprintf (pCurrent, "%s", LINE_SEPSTRING_CN00 );
        pCurrent = pCurrent + _tcslen(pCurrent);

        aRecord = aTable.nextRecord();
      } // end if

      *replyLen = _tcslen(replyData);

      if (aTable.sqlRc() != cin01_db_row_not_found && aTable.sqlRc() != cin01_db_ok) {
        teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
        nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
      } // end if
    
    } else {
      teo200_EventList lEvtList(aTable.lastEvent(), FUNCTION_NAME_MCN00_1, ERR_SQL_CN00_1);
      nFuncReturn = STATIC_CAST(tcn00_Error, cn90AnswerEvent(replyData, replyLen, lEvtList));
    } // end if

    cn80ReleaseSQL(&SQLSession, false);
  } // end if

  return nFuncReturn;

} // end cn83_AnalyzeIndex
*/

/* 
  -----------------------------------------------------------------------------
  function:     cn83UpdateStat
  -----------------------------------------------------------------------------
 */
tcn00_Error cn83UpdateStat
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn   = OK_CN00;

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn83_UpdateStat(vcontrol->dbname, command->args, replyData, replyLen);
  }

  if (nFuncReturn == OK_CN00) {
    //cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} // end cn83UpdateStat

/* 
  -----------------------------------------------------------------------------
  private function:     cn83_UpdateStat
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn83_UpdateStat
      ( const tsp00_DbNamec & pDBName,
        const _TCHAR  * pSpec,
        char          * replyData,
        int           * replyLen )
{
  _TCHAR                  szCommand     [SQL_COMMANDLENGTH];

  tcn00_Error             nFuncReturn    = OK_CN00;
  tin01_sql_session       SQLSession;

  nFuncReturn = cn80ConnectSQL(pDBName, &SQLSession, replyData, replyLen);

  // update statistics
  if (nFuncReturn == OK_CN00) {
    if (strlen(pSpec) == 0) {
      _stprintf(szCommand, SQL_UPDATESTAT_ALL);
    } else {
      _stprintf(szCommand, SQL_UPDATESTAT, pSpec);
    }

    nFuncReturn = cn80ExecuteSQL  ( &SQLSession,
                                    szCommand,
                                    replyData,
                                    SQL_COMMANDLENGTH,
                                    *replyLen );

    // release session with commit or not
    cn80ReleaseSQL  ( &SQLSession, nFuncReturn == OK_CN00);
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} // end cn83_UpdateStat
