/*!================================================================
 module:        gco10.h
 
 responsible:   ThomasA
 
 special area:  OMS
 
 description:   C++ Persistent Object Interface
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/



#if !defined (__HCO101OmsObjByKeyIterBase)
#define __HCO101OmsObjByKeyIterBase

#include "gco00.h"
#include "livecachetypes.h" // nocheck

class  OmsAbstractObject;
class  OmsHandle;
struct KeyRangeIterBody;

/*!
  class :  OmsObjByKeyIterBase

  description: defines a key range iterator, i.e. an iterator that
               yields keyed objects with keys >= a start key and <= a
               stop key.
               The iterators yields a sequence of keyed objects, that is
               sorted by the key using memcmp for comparism.
*/

class AFX_EXT_CLASS OmsObjByKeyIterBase 
{
  friend class OmsHandle;
public :
  /*!-----------------------------------------------------------------------

    function:     copy constructor
  
    description:  constructs a OmsObjByKeyIterBase instance as a copy of
                  another OmsObjByKeyIterBase instance.
                  As the result both instances share the same m_pBody.
                  The copy constructor is used only for returning an iterator
                  from a method call.

    arguments:
      source [in] the instance to be copied.
      
    return value: none 
      
  -------------------------------------------------------------------------*/

  OmsObjByKeyIterBase(const OmsObjByKeyIterBase& source);
  
  /*!-----------------------------------------------------------------------

    function:     constructor
  
    description:  constructs a OmsObjByKeyIterBase instance.

    arguments:    
      handle      [in,out] the OmsHandle instance of the caller
      guid        [in] the guid of the persistent keyed object, this is
                       a part of the container identification of the container
                       to be traversed
      Schema      [in] the schema identification of the schema containing
                       the container
      ContainerNo [in] the container number of the container to be traversed  
      pStartKey   [in] a pointer to the start key
      pStopKey    [in] a pointer to the stop key
    
    return value: none 
      
  -------------------------------------------------------------------------*/

  OmsObjByKeyIterBase
    (
    OmsHandle&      handle, 
    const ClassIDRef guid, 
    OmsSchemaHandle Schema, 
    OmsContainerNo  ContainerNo,
    const void*     pStartKey,
    const void*     pStopKey
    );

  /*!-----------------------------------------------------------------------

    function:     Destructor
  
    description:  destructs the OmsObjByKeyIterBase instance and releases all
                  memory used by it.
    arguments:    none
    
    return value: none 
      
  -------------------------------------------------------------------------*/
  ~OmsObjByKeyIterBase();
protected:
  OmsObjByKeyIterBase() {}  // required to compile properly with GCC

  /*!-----------------------------------------------------------------------

    function:     operator bool
  
    description:  yields true, if the iterator points to an keyed object;
                  otherwise false.
    arguments:    none
    
    return value: none 
      
  -------------------------------------------------------------------------*/
  
  operator bool () const;
  
  /*!-----------------------------------------------------------------------

    function:     omsDeRef
  
    description:  returns a pointer to the keyed object the iterator is pointing to.
                  Has no meaning, if operator bool returns false.
    arguments:    
      forUpd [in] if true, the returned keyed object may be updated by the caller.
      doLock [in] if true, the current keyed object is locked
    
    return value: pointer to the current keyed object 
      
  -------------------------------------------------------------------------*/
  
  OmsAbstractObject* omsDeRef(bool forUpd, bool doLock) const;
  
  /*!-----------------------------------------------------------------------

    function:     operator++
  
    description:  steps to the next keyed object.

    arguments:    none    
    
    return value: pointer to the current keyed object 
      
  -------------------------------------------------------------------------*/

  void operator++ ();

private:
    KeyRangeIterBody* m_pBody;
};

/*! endclass: OmsObjByKeyIterBase */

#endif