/*!================================================================
 module:        hco17.h
 
 responsible:   MarkusSi
 
 special area:  OMS
 
 description:   OmsVersion Dictionary   
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#ifndef HCO17_H
#define HCO17_H


#include "hco12.h"
#include "gco001.h"
#include "Oms/OMS_LibOmsInterfaceInstance.hpp"

enum LockMode {SHARED, EXCLUSIVE, NO_LOCK};

class CVersionDictionary_co17;

//////////////////////////////////////////////////////////////////////////
// Class: CUnloadableVersions                                           //
//////////////////////////////////////////////////////////////////////////
class CUnloadableVersions
{
private :
  CUnloadableVersions();
  void InsertVersion(OmsContext_co12*, bool callFromDestructor);
  OmsContext_co12* RemoveVersion(OmsContext_co12*);
  OmsContext_co12* m_first;
  friend class CVersionDictionary_co17;
};


//////////////////////////////////////////////////////////////////////////
// Class: CVersionDictionary_co17                                       //
//////////////////////////////////////////////////////////////////////////
class CVersionDictionary_co17 {
public :

  /// Iterator for the version dictionary
  /*! 
  ** This class implements an iterator to read the version dictionary.
  **
  ** \attention Even if the access is synchronized, the iterator might not return a 'snapshort'
  ** as always only a single lock on the current slot is hold while iterating over the entries
  ** and not a global lock on all slots. This is done because of performance and as it is not
  ** considered as so important to get a real 'snapshot'.
  */
  class Iter {
  private:
    /// Slot of the hashbucket, where the iterator is currently positioned to. 
    int                            m_slot;
    /// How should the entries be locked while iteration over them?
    LockMode                       m_lockMode;
    /// The version the iteration is currently pointing to.
    OmsContext_co12               *m_pCurr;
    /// Pointer to the version dictionary.
    CVersionDictionary_co17       *m_pVD;
    /// Enters the region (either reader-writer of critical section) which belongs to the current slot  
    void enterRgn();
    /// Leaves the region (either reader-writer of critical section) which belongs to the current slot  
    void leaveRgn();

    Iter(const Iter&);                  ///< Not allowed, therefore private
    Iter& operator=(const Iter &iter);  ///< Not allowed, therefore private

  public :
    /// Default-constructor
    Iter();
    /// Constructor
    /*! 
    ** A iterator is created and the iterator is positioned on the first entry.
    ** if the lockMode does not equal NO_LOCK, then a lock on the slot of the first
    ** entry is aquired and is hold.
    **
    ** \param lockMode          [in] should the access to the dictionary be synchronized.
    ** \param versionDictionary [in] Reference to the version dictionary over which iterator
    **                               should run
    */
    Iter(LockMode lockMode, CVersionDictionary_co17 &versionDictionary);
    
    /// Destructor
    /*! 
    ** The iterator is destructed. If it still holds some locks, than these will be
    ** released.
    */
    ~Iter();

    /// Increment
    /*! 
    ** Increments the iterator to the next entry. If this entry is locacted in another
    ** hash-bucket, then the old lock will be released an the lock for the next slot
    ** will be aquired. This means, that a pointer to a version, which is remembered
    ** might not be locked anymore, after the iterator has been incremented. 
    **
    ** \attention Do not drop a version while iterating over the version dictionary
    **    with this operator. If the deletion of the current version is needed use
    **    the method CVersionDictionary_co17::getNext instead.
    */
    void operator++() { getNext(false); }

    /// Increment with possibility to delete the current version
    /*! 
    ** Increments the iterator to the next entry and if specified the entry of the
    ** current version is deleted. If this entry is locacted in another hash-bucket, 
    ** then the old lock will be released an the lock for the next slot will be aquired.
    **
    ** \param dropCurr [in] If this parameter is specified, then the version, the iterator
    **        is currently pointing to, is deleted and the iterator is then moved to the 
    **        next entry. 
    */
    void getNext(bool dropCurr=false);

    /// Returns true if the iterator does not have reached its end.
    operator bool() const;

    /// Returns the version the iterator is currently pointing to.
    OmsContext_co12* operator()() const;

    /// Positions an existing iterator onto the frist entry.
    /*! 
    ** Positions an existing iterator onto the frist entry.
    ** 
    ** \param lockMode          [in] should the access to the dictionary be synchronized.
    ** \param versionDictionary [in] Reference to the version dictionary over which iterator
    **                               should run
    */
    void first(LockMode lockMode, CVersionDictionary_co17 &versionDictionary);
    
    /// Stops an iterator
    /*! 
    **  Stops an iterator. This function is needed, if an iterator should stop processing
    **  without having run to the end. In this case the iterator might hold a lock which must
    **  be released to prevent lock contention.
    **  A call of this method set the iterator to invalid mode. Therefore it is not possible 
    **  to continue iterating after having called 'stop'.
    */
    void stop();
  };

  /// Constructor
  CVersionDictionary_co17();

  /// Returns the version corrsponding to the specified version-id
  OmsContext_co12* FindVersion  (const OmsVersionId&);

  /// Inserts a new version in the directory
  /*!
  ** Inserts a new version in the directory.
  **
  ** \param versionId [in] Name of the version (key in the search-structure), 
  **                       which should be inserted
  ** \param pContext  [in] Pointer to the version, which should be inserted
  **
  ** \return If a version with the same name already exists, then the error code
  **         e_duplicate_key is returned; otherwise e_ok
  */
  tgg00_BasisError InsertVersion(const OmsVersionId& versionId, OmsContext_co12* pContext); 

  /// Deletes a version out of the directory
  /*!
  ** Deletes a version out of the directory.
  ** 
  ** \param versionId [in] Name of the version, which which should be deleted
  **
  ** \return If no version with the specified name is found, then e_unknown_version
  **         is returned; otherwise e_ok
  */
  tgg00_BasisError DropVersion  (const OmsVersionId& versionId);

  void             MarkUnloadable(OmsContext_co12* pContext, bool callFromDestructor);
  void             MarkNotUnloadable(IliveCacheSink* lcSink, OmsContext_co12* pContext);
  void             Shutdown();
  bool             UnloadOldestVersion(IliveCacheSink* lcSink);

  /// Determines and returns the hash-slot of a given version-id  
  int              GetSingleLockId(const OmsVersionId &versionId);

  /// Determines and returns the hash-slots of all given version-ids
  /*! 
  ** Determines and returns the hash-slots of all given version-ids.
  **
  ** \param ppVersionId [in]  Array with the given version-ids.
  ** \param size        [in,out] in: number of version-ids in the array ppVersionId
  **                             out: number of slot-ids returned in pLockId
  **  \param pLockId    [out] Array with the hash-slots which corresponds to the given version-ids
  **
  ** \attention The entries in the array pLockId are sorted in ascending order
  ** to prevent deadlocks. Therefore does not change this sequence manually!
  */
  void             GetMultiLockIds(const OmsVersionId **ppVersionId, int &size, int  *pLockId);
  
  /// Returns all lock-ids of the version-dictionary
  void             GetAllLockIds(int *pLockId, int &size);

  /// Are reader-writer-locks or critical section used for synchonization
  bool             UseRWLocks() { return m_useRWLocks; }

private :
  /// Compute the hash-slot for a given version-id
  inline int       HashValue (const OmsVersionId*) const;
  /// Return the offset of the lock-ids.
  /*! 
  ** In the version-dictionary coding it is assumed that the lock-ids start with
  ** the number zero, but in the kernel these lock-ids might be mapped to other 
  ** ids. Therefore this offset is needed.
  */
  inline int       GetVDirStartCnt() const;

  /// Does the own session own a lock on the given slot
  bool isLockedByOwn(int slot) const;   

  /// Hash-array. Collisions are resolved by chaining 
  OmsContext_co12      *m_pVersionDictionary[OMS_VDIR_SIZE];
  /// Datastructure which contains all versions which could be unloaded
  CUnloadableVersions   m_UnloadableVersions;
  /// Flag which indicates whether to use reader-writer-locks or critical sections for synchonization
  bool                  m_useRWLocks;

  friend void OMS_LibOmsInterfaceInstance::AdviseKernelInterface(LVC_KernelInterface& knlInterface, bool); // Initializes the variable m_useRWLocks
  friend class Iter;
};

/*---------------------------------------------------------------------------*/

inline int CVersionDictionary_co17::GetVDirStartCnt() const {
  return m_useRWLocks ? OMS_VDIR_RW_START_CNT : OMS_VDIR_CS_START_CNT; 
}

/*---------------------------------------------------------------------------*/

inline int CVersionDictionary_co17::HashValue (const OmsVersionId *versionId) const{
  unsigned long sum = 0;
  for (int i=0; i<sizeof(OmsVersionId); ++i){
    sum += (sum ^ (*versionId)[i]);
  }
  return (sum % OMS_VDIR_SIZE);
}

#endif
