/*!**************************************************************************

  module      : vco101OmsObjByKeyIterBase.cpp

  -------------------------------------------------------------------------

  responsible : d0254850

  special area: oms
  description : implementation of key range iterator


  last changed: 2001-01-15  15:00
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2001

  ========== licence begin  GPL
  Copyright (C) 2001 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "hco101OmsObjByKeyIterBase.h"
#include "gco10.h"
#include <new>

/*===========================================================================*
 *  LOCAL CLASSES                                                            *
 *===========================================================================*/

/*!
  class :  KeyRangeUnloadedVersionIter

  description: defines a key iterator over all keyed objects, that have
               been created in a version and have been unloaded
*/

class KeyRangeUnloadedVersionIter
{
public :
  KeyRangeUnloadedVersionIter(
    OmsSession_co12*  pSession,
    tco13_ClsIdEntry* pContainerInfo,
    const void*       pStartKey,
    const void*       pStopKey
    );
  ~KeyRangeUnloadedVersionIter();
  operator bool() const;
  void operator++();
  OmsObjectContainer* GetObjContainerPtr();     
private :
  bool                  m_first;
  int                   m_keyLen;
  int                   m_objectSize;
  void*                 m_pCurrKey;
  void*                 m_pStopKey;
  OmsObjectContainerPtr m_pCurrObj;
  OmsSession_co12*      m_pSession;
  tco13_ClsIdEntry*     m_pContainerInfo;
};

/*! endclass: KeyRangeUnloadedVersionIter */

/*----------------------------------------------------------------------*/

/*!
  class :  KeyRangeVersionInMemIter

  description: defines a key iterator over all keyed objects, that have
               been created in a version and have not been unloaded
*/

class KeyRangeVersionInMemIter
{
public :
  KeyRangeVersionInMemIter(
    OmsSession_co12*  pSession,
    tco13_ClsIdEntry* pContainerInfo,
    const void*       pStartKey,
    const void*       pStopKey
    );
  ~KeyRangeVersionInMemIter();
  operator bool() const;
  void operator++();
  OmsObjectContainer* GetObjContainerPtr();     
private :
  bool                         m_end;
  int                          m_keyLen;
  int                          m_objectSize;
  void*                        m_pStopKey;
  OmsSession_co12*             m_pSession;
  tco13_ClsIdEntry*            m_containerInfo;  // PTS 1117571
  cgg250AvlTree<unsigned char*, tco13_ClsIdEntry,OmsContext_co12>::Iterator m_iter;
};

/*! endclass: KeyRangeVersionInMemIter */

/*----------------------------------------------------------------------*/

/*!
  class :  KeyRangeVersionIter

  description: defines a key iterator over all keyed objects, that have
               been created in a version.
               This is a merge of KeyRangeVersionInMemIter and
               KeyRangeUnloadedVersionIter
*/

class KeyRangeVersionIter
{
public :
  KeyRangeVersionIter(
    OmsSession_co12*  pSession,
    tco13_ClsIdEntry* pContainerInfo,
    const void*       pStartKey,
    const void*       pStopKey
    );
  ~KeyRangeVersionIter();
  operator bool() const;
  void operator++();
  const OmsObjectContainer* GetObjContainerPtr();     
private :
  OmsSession_co12*             m_pSession;
  tco13_ClsIdEntry*            m_pContainerInfo;
  KeyRangeUnloadedVersionIter* m_pUnloadedVersionIter;
  KeyRangeVersionInMemIter     m_versionIter;
};

/*! endclass: KeyRangeVersionIter */

/*----------------------------------------------------------------------*/

/*!
  class :  KeyRangeIterBody

  description: defines the implementation details of OmsObjByKeyIterBase
               Is used to implement a merge of KeyRangeVersionIter and
               a KB key range iterator. 
*/

struct KeyRangeIterBody
{
  OmsSession_co12*         m_pSession;
  tco13_ClsIdEntry*        m_containerInfo;
  OmsObjectContainerPtr    m_curr;
  OmsObjectContainerPtr    m_next;
  OmsObjectContainerPtr    m_nextVersion;
  void*                    m_pKBIterator;
  tgg00_OmsVersionContext* m_versionContext;
  KeyRangeVersionIter*     m_pVersionIterator;
  int                      m_refCnt;
  int                      m_keyLen;
  bool                     m_firstInVersion;
  KeyRangeIterBody(
    OmsSession_co12*  pSession,
    tco13_ClsIdEntry* pContainerInfo)
    : m_pSession(pSession)
    , m_containerInfo(pContainerInfo)
    , m_curr(NULL)
    , m_next(NULL)
    , m_nextVersion(NULL)
    , m_pKBIterator(NULL)
    , m_pVersionIterator(NULL)
    , m_versionContext(REINTERPRET_CAST(tgg00_OmsVersionContext*, m_pSession->CurrentContext()->VersionContext()))
    , m_refCnt(1)
    , m_keyLen(pContainerInfo->GetKeyLen())
    , m_firstInVersion(true)
  {
  }
  void IncRefCnt()
  {
    ++m_refCnt;
  }
  int DecRefCnt()
  {
    --m_refCnt;
    return m_refCnt;
  }
};

/*! endclass: KeyRangeIterBody */

/*===========================================================================*
 *  Implementation of  KeyRangeVersionInMemIter                              *
 *===========================================================================*/

// PTS 1117571
// Changed routine is such a way, that only objects, which are newly 
// created in a version are considered (Objects which are stored in the
// structure m_index because of 'cached keys' are not considered. 
KeyRangeVersionInMemIter::KeyRangeVersionInMemIter
(
 OmsSession_co12*  pSession,
 tco13_ClsIdEntry* pContainerInfo,
 const void*     pStartKey,
 const void*     pStopKey
 )
 : m_end(true)
 , m_keyLen(pContainerInfo->GetKeyLen())
 , m_objectSize(pContainerInfo->GetObjectSize())
 , m_pStopKey(NULL)
 , m_pSession(pSession)
 , m_iter(pContainerInfo->m_index.First())
 , m_containerInfo(pContainerInfo)
{
  const char* msg = "KeyRangeVersionInMemIter::KeyRangeVersionInMemIter";

  bool isVersionOid = true;         
  OmsObjectContainer* pCurrObject;  

  // Looking for smallest object which is larger than spezified start-key
  while (m_iter)
  {
    // If cached keys are used, then there might be entries in the structure
    // m_index which do not belong to objects which are newly created in
    // a version. Therefore for the iteration consider only those objects 
    // which belong to a version.
    if (pContainerInfo->UseCachedKeys()){
      pCurrObject = pContainerInfo->VersionGetInfoFromNode(m_iter());          
      if (!pCurrObject) {
        throw DbpError (e_system_error, msg); 
      }

      isVersionOid = m_pSession->CurrentContext()->IsVersionOid(pCurrObject->m_oid);
    }
    else {
      isVersionOid = true;
    }

    if (memcmp(*m_iter(), pStartKey, m_keyLen) < 0)
    {
      ++m_iter;
    }
    else
    {
      if (isVersionOid){
        break;
      }
      else {
        ++m_iter;
      }
    }
  }

  // Check if found object is already larger than the stop-key. If so 
  // there are no matching entries 
  if (m_iter)
  {
    if (memcmp(*m_iter(), pStopKey, m_keyLen) <= 0)
    {
      m_end = false;
    }
  }

  if (!m_end)
  {
    m_pStopKey = pSession->allocate(m_keyLen);
    memcpy(m_pStopKey, pStopKey, m_keyLen);
  }
}

/*----------------------------------------------------------------------*/

KeyRangeVersionInMemIter::operator bool() const
{
  return (!m_end);
}

/*----------------------------------------------------------------------*/

OmsObjectContainer* KeyRangeVersionInMemIter::GetObjContainerPtr()
{
  unsigned char* const * p = m_iter();
  OmsObjectContainer* pObj = REINTERPRET_CAST(OmsObjectContainer*, CONST_CAST(unsigned char*, (*p) + 
    m_keyLen  - m_objectSize));
  return pObj;
}

/*----------------------------------------------------------------------*/
// PTS 1117571
// Changed routine is such a way, that only objects, which are newly 
// created in a version are considered (Objects which are stored in the
// structure m_index because of 'cached keys' are not considered. 
void KeyRangeVersionInMemIter::operator++()
{
  const char* msg = "KeyRangeVersionInMemIter::operator++";

  bool isVersionOid = true;         
  OmsObjectContainer* pCurrObject;  

  do {
    ++m_iter;
    if (m_iter)
    {
      unsigned char* p = *m_iter();

      // If cached keys are used, then there might be entries in the structure
      // m_index which do not belong to objects which are newly created in
      // a version. Therefore for the iteration consider only those objects 
      // which belong to a version.
      if (m_containerInfo->UseCachedKeys()){
        pCurrObject = m_containerInfo->VersionGetInfoFromNode(m_iter());          
        if (!pCurrObject) {
          throw DbpError (e_system_error, msg); 
        }
        isVersionOid = m_pSession->CurrentContext()->IsVersionOid(pCurrObject->m_oid);
      }
      else{
        isVersionOid = true;
      }

      if (memcmp(*m_iter(), m_pStopKey, m_keyLen) > 0){
        m_end = true;
        break;
      }
    }
    else {
      m_end = true;
      break;
    }
  } while (!isVersionOid);
}

/*----------------------------------------------------------------------*/

KeyRangeVersionInMemIter::~KeyRangeVersionInMemIter()
{
  if (NULL != m_pStopKey)
  {
    m_pSession->deallocate(m_pStopKey);
  }
}

/*===========================================================================*
 *  Implementation of KeyRangeUnloadedVersionIter                            *
 *===========================================================================*/

KeyRangeUnloadedVersionIter::KeyRangeUnloadedVersionIter
(
 OmsSession_co12*  pSession,
 tco13_ClsIdEntry* pContainerInfo,
 const void*       pStartKey,
 const void*       pStopKey
 )
 : m_first(true)
 , m_keyLen(pContainerInfo->GetKeyLen())
 , m_objectSize(pContainerInfo->GetObjectSize())
 , m_pStopKey(NULL)
 , m_pCurrKey(NULL)
 , m_pCurrObj(NULL)
 , m_pSession(pSession)
 , m_pContainerInfo(pContainerInfo)
{
  m_pCurrKey = pSession->allocate(m_keyLen);
  memcpy(m_pCurrKey, pStartKey, m_keyLen);
  m_pStopKey = pSession->allocate(m_keyLen);
  memcpy(m_pStopKey, pStopKey, m_keyLen);
  this->operator++();
}

/*----------------------------------------------------------------------*/

OmsObjectContainer* KeyRangeUnloadedVersionIter::GetObjContainerPtr()
{
  return m_pCurrObj;
}

/*----------------------------------------------------------------------*/

KeyRangeUnloadedVersionIter::operator bool() const
{
  return (m_pCurrObj != NULL);
}

/*----------------------------------------------------------------------*/

void KeyRangeUnloadedVersionIter::operator++()
{
  OmsObjectId Oid;
  tsp00_Int2 DBError;
  do
  {
    m_pSession->m_lcSink->VersionNextKey(
      REINTERPRET_CAST(tgg00_OmsVersionContext*, m_pSession->CurrentContext()->VersionContext()),
      m_pContainerInfo->GetFileId(),
      m_pCurrKey,
      m_keyLen,
      m_first,
      m_pStopKey,
      *REINTERPRET_CAST(OmsTypeOid*, &Oid),
      DBError);
    m_first = false;
    if (e_ok == DBError)
    {
      m_pCurrObj = m_pSession->CurrentContext()->GetObj(Oid, false);
      if (NULL != m_pCurrObj)
      {
        break;
      }
    }
    else
    {
      if (e_no_next_object == DBError)
      {
        m_pCurrObj = NULL;
        break;
      }
      else
      {
        m_pSession->ThrowDBError (DBError, "KeyRangeUnloadedVersionIter::operator++");
      }
    }
  }
  while (true);
}

/*----------------------------------------------------------------------*/

KeyRangeUnloadedVersionIter::~KeyRangeUnloadedVersionIter()
{
  if (NULL != m_pCurrKey)
  {
    m_pSession->deallocate(m_pCurrKey);
  }
  if (NULL != m_pStopKey)
  {
    m_pSession->deallocate(m_pStopKey);
  }
}

/*===========================================================================*
 *  Implementation of KeyRangeVersionIter                                    *
 *===========================================================================*/

KeyRangeVersionIter::KeyRangeVersionIter
(
 OmsSession_co12*  pSession,
 tco13_ClsIdEntry* pContainerInfo,
 const void*     pStartKey,
 const void*     pStopKey
 )
 : m_pSession(pSession)
 , m_pContainerInfo(pContainerInfo)
 , m_versionIter(pSession, pContainerInfo, pStartKey, pStopKey)
 , m_pUnloadedVersionIter(NULL)
{
  if (m_pSession->CurrentContext()->IsUnloaded())
  {
    m_pUnloadedVersionIter = REINTERPRET_CAST(KeyRangeUnloadedVersionIter*, 
      pSession->allocate(sizeof(KeyRangeUnloadedVersionIter)));
    new(m_pUnloadedVersionIter) KeyRangeUnloadedVersionIter(pSession,
       pContainerInfo, pStartKey, pStopKey);
  }
}

/*----------------------------------------------------------------------*/

KeyRangeVersionIter::operator bool() const
{
  if (NULL != m_pUnloadedVersionIter)
  {
    return (m_versionIter || *m_pUnloadedVersionIter);
  }
  else
  {
    return m_versionIter;
  }
}

/*----------------------------------------------------------------------*/

void KeyRangeVersionIter::operator++()
{
  if (NULL != m_pUnloadedVersionIter)
  {
    if (*m_pUnloadedVersionIter && m_versionIter)
    {
      int cmpResult = memcmp(m_pContainerInfo->GetKeyPtr(m_versionIter.GetObjContainerPtr()),
                 m_pContainerInfo->GetKeyPtr(m_pUnloadedVersionIter->GetObjContainerPtr()),
                 m_pContainerInfo->GetKeyLen());
      if (cmpResult < 0)
      {
        ++m_versionIter;
      }
      else
      {
        ++(*m_pUnloadedVersionIter);
        if (0 == cmpResult)
        {
          ++m_versionIter;
        }
      }
    }
    else
    {
      if (m_versionIter)
      {
        ++m_versionIter;
      }
      else
      {
        ++(*m_pUnloadedVersionIter);
      }
    }
  }
  else
  {
    ++m_versionIter;
  }
}

/*----------------------------------------------------------------------*/

const OmsObjectContainer* KeyRangeVersionIter::GetObjContainerPtr()
{
  if (NULL != m_pUnloadedVersionIter)
  {
    if (*m_pUnloadedVersionIter && m_versionIter)
    {
      if (memcmp(m_pContainerInfo->GetKeyPtr(m_versionIter.GetObjContainerPtr()),
                 m_pContainerInfo->GetKeyPtr(m_pUnloadedVersionIter->GetObjContainerPtr()),
                 m_pContainerInfo->GetKeyLen()) < 0)
      {
        return m_versionIter.GetObjContainerPtr();
      }
      else
      {
        return m_pUnloadedVersionIter->GetObjContainerPtr();
      }
    }
    else
    {
      if (m_versionIter)
      {
        return m_versionIter.GetObjContainerPtr();
      }
      else
      {
        return m_pUnloadedVersionIter->GetObjContainerPtr();
      }
    }
  }
  else
  {
    return m_versionIter.GetObjContainerPtr();
  }
}

/*----------------------------------------------------------------------*/

KeyRangeVersionIter::~KeyRangeVersionIter()
{
  if (NULL != m_pUnloadedVersionIter)
  {
    m_pSession->deallocate(m_pUnloadedVersionIter);
  }
}

/*===========================================================================*
 *  Implementation of OmsObjByKeyIterBase                                    *
 *===========================================================================*/

OmsObjByKeyIterBase::OmsObjByKeyIterBase
(
 OmsHandle&      h, 
 const ClassIDRef guid, 
 OmsSchemaHandle Schema, 
 OmsContainerNo  ContainerNo,
 const void*     pStartKey,
 const void*     pStopKey
 )
 : m_pBody(NULL)
{
  m_pBody = new(h.m_pSession->allocate(sizeof(KeyRangeIterBody))) 
    KeyRangeIterBody(h.m_pSession, h.m_pSession->GetClsInfo (guid, Schema, ContainerNo));
  tsp00_Int4 size = m_pBody->m_containerInfo->GetObjectSize();
  /* PTS 1111569 FF   check if (StartKey <= Stopkey) */
  if ( 0 >= memcmp( pStartKey, pStopKey, m_pBody->m_containerInfo->GetKeyLen() )) 
  {
    /*--- PTS 1111569 FF          --- */
    m_pBody->m_curr = REINTERPRET_CAST(OmsObjectContainerPtr, h.m_pSession->allocate (size));
    m_pBody->m_curr->InitObjContainer(*m_pBody->m_containerInfo);
    m_pBody->m_curr->m_oid = OmsObjectId();
    m_pBody->m_next          = REINTERPRET_CAST(OmsObjectContainerPtr, h.m_pSession->allocate (size));
    m_pBody->m_next->InitObjContainer(*m_pBody->m_containerInfo);
    m_pBody->m_next->m_oid = OmsObjectId();
    
    tsp00_Int2 DBError;
    h.m_pSession->m_lcSink->CreateKeyRangeIterator(
      h.m_pSession->m_context->m_consistentView,
      REINTERPRET_CAST(tgg00_OmsVersionContext*, h.m_pSession->m_context->VersionContext()),
      m_pBody->m_containerInfo->GetFileId(),
      m_pBody->m_containerInfo->GetKeyLen(),
      pStartKey,
      pStopKey,
      &m_pBody->m_pKBIterator,
      DBError);
    if (0 != DBError)
    {
      h.m_pSession->ThrowDBError (DBError, "OmsObjByKeyIterBase::OmsObjByKeyIterBase");
    }
    this->operator++();
    if (h.m_pSession->InVersion())
    {
      m_pBody->m_pVersionIterator = new(h.m_pSession->allocate(sizeof(KeyRangeVersionIter)))
        KeyRangeVersionIter(h.m_pSession, m_pBody->m_containerInfo, pStartKey, pStopKey);
      if (m_pBody->m_pVersionIterator)
      {
        m_pBody->m_nextVersion = REINTERPRET_CAST(OmsObjectContainerPtr, h.m_pSession->allocate (size));
        m_pBody->m_nextVersion->InitObjContainer(*m_pBody->m_containerInfo);
        m_pBody->m_nextVersion->m_oid = OmsObjectId();
        OmsObjectContainerPtr aux = m_pBody->m_next; 
        m_pBody->m_next = NULL;
        this->operator++();
        m_pBody->m_next = aux;
      }
    }
    this->operator++();
  }
}

/*----------------------------------------------------------------------*/

OmsObjByKeyIterBase::OmsObjByKeyIterBase(const OmsObjByKeyIterBase& source)
{
  *this = source;
  m_pBody->IncRefCnt();
}

/*----------------------------------------------------------------------*/

OmsObjByKeyIterBase::operator bool () const
{
  return (NULL != m_pBody->m_curr) ? true : false;
}

/*----------------------------------------------------------------------*/

void OmsObjByKeyIterBase::operator++()
{
  OmsObjectContainerPtr curr;
  while (true)
  {
    if ((NULL != m_pBody->m_nextVersion) && (NULL != m_pBody->m_next))
    { 
      if (memcmp(m_pBody->m_containerInfo->GetKeyPtr(m_pBody->m_next), 
        m_pBody->m_containerInfo->GetKeyPtr(m_pBody->m_nextVersion), 
        m_pBody->m_keyLen) < 0)
      {
        curr = m_pBody->m_next;
      }
      else
        curr = m_pBody->m_nextVersion;
    }
    else
    {
      if (NULL != m_pBody->m_next)
      {
        curr = m_pBody->m_next;       
      }
      else
      {
        curr = m_pBody->m_nextVersion;
      }
    }
    if (NULL == curr)
    {
      m_pBody->m_pSession->deallocate(m_pBody->m_curr);
      m_pBody->m_curr = NULL;
      return;
    }
    if (curr == m_pBody->m_next)
    {
      // next object from livecache 
      tsp00_Int4 objHistReadCount;
      tsp00_Int2 DBError;
      m_pBody->m_next = m_pBody->m_curr;
      m_pBody->m_curr = curr;
      m_pBody->m_next->InitObjContainer(*m_pBody->m_containerInfo); /* PTS 1113209 */
      m_pBody->m_pSession->m_lcSink->NextObjFromKeyIterator (
        m_pBody->m_pKBIterator,
        m_pBody->m_versionContext,
        m_pBody->m_containerInfo->GetFileId(),
        m_pBody->m_containerInfo->GetPersistentSize(),
        ((unsigned char*) &m_pBody->m_next->m_pobj) + sizeof(void*),
        *REINTERPRET_CAST(OmsTypeOid*, &m_pBody->m_next->m_oid),
        m_pBody->m_next->m_objseq,
        objHistReadCount,
        DBError); 
      if (0 != DBError)
      {
        if (e_no_next_object == DBError)
        {
          m_pBody->m_pSession->deallocate(m_pBody->m_next);
          m_pBody->m_next = NULL;
        }
        else
        {
          m_pBody->m_pSession->ThrowDBError (DBError, "OmsObjByKeyIterBase::operator++");
        }
      }
      else
      {
        m_pBody->m_pSession->IncLogHop(objHistReadCount);
      }
    }
    else
    {
      // next object from version
      m_pBody->m_nextVersion = m_pBody->m_curr;
      m_pBody->m_curr        = curr;
      if (!m_pBody->m_firstInVersion)
      {
        ++(*m_pBody->m_pVersionIterator);
      }
      else
      {
        m_pBody->m_firstInVersion = false;
      }
      if (*m_pBody->m_pVersionIterator)
      {
        memcpy(m_pBody->m_nextVersion,
          m_pBody->m_pVersionIterator->GetObjContainerPtr(),
          m_pBody->m_containerInfo->GetObjectSize());
      }
      else
      {
        m_pBody->m_pSession->deallocate(m_pBody->m_nextVersion);
        m_pBody->m_nextVersion = NULL;
      }
    }
    if (NULL != m_pBody->m_curr)
    {
      OmsObjectContainerPtr found = 
        m_pBody->m_pSession->m_context->FindObjInContext (&m_pBody->m_curr->m_oid);
      if (found && found->DeletedFlag())
      {
        continue;
      }
    }
    break;
  }
}

/*----------------------------------------------------------------------*/

OmsAbstractObject* OmsObjByKeyIterBase::omsDeRef (bool forUpd, bool doLock) const{
  if (m_pBody->m_curr) 
  {
    OmsSession_co12& Session = *m_pBody->m_pSession;
    OmsContext_co12& Context = *Session.m_context;
    Session.IncDeref();

    OmsObjectContainerPtr found = 
      Context.FindObjInContext (&m_pBody->m_curr->m_oid);
    if (found) {
      if (found->DeletedFlag()) {
        Session.ThrowDBError (e_object_not_found, 
          "OmsObjByKeyIterBase::omsDeRef", m_pBody->m_curr->m_oid);
        return NULL;
      }
      else {
        if ((doLock) && (!found->LockedFlag()) && (!Context.IsVersion()))
        {
          Context.LockObj(found->m_oid);
        }
        if (forUpd) {
          return Session.ForUpdPtr(found);
        }

        return &found->m_pobj;
      }
    }
    else 
    {
      tco13_ClsIdEntry& ContainerInfo = *m_pBody->m_containerInfo;
      OmsObjectContainerPtr obj = Session.GetMemory (ContainerInfo); 
      memcpy (obj, m_pBody->m_curr,  ContainerInfo.GetObjectSize());
      Context.PutObjectIntoContext (obj, ContainerInfo.GetContainerHandle());
      if ((doLock) && (!Context.IsVersion()))
      {
        Context.LockObj(obj->m_oid);
      }
      if (forUpd) {
        return Session.ForUpdPtr(obj);
      }

      return &obj->m_pobj;
    }
  }
  return NULL;
}

/*----------------------------------------------------------------------*/

OmsObjByKeyIterBase::~OmsObjByKeyIterBase()
{
  if (0 == m_pBody->DecRefCnt())
  {
    if (NULL != m_pBody->m_curr)
    {
      m_pBody->m_pSession->deallocate (m_pBody->m_curr);
    }
    if (NULL != m_pBody->m_next)
    {
      m_pBody->m_pSession->deallocate (m_pBody->m_next);
    }
    if (NULL != m_pBody->m_pKBIterator)
    {
      m_pBody->m_pSession->m_lcSink->DestroyKeyRangeIterator(m_pBody->m_pKBIterator);
    }
    if (NULL != m_pBody->m_pVersionIterator)
    {
      m_pBody->m_pVersionIterator->~KeyRangeVersionIter();
      m_pBody->m_pSession->deallocate (m_pBody->m_pVersionIterator);
    }
    m_pBody->m_pSession->deallocate(m_pBody);
  } 
}