/*!================================================================
 module:        vco17.cpp
 
 responsible:   MarkusSi
 
 special area:  OMS
 
 description:   OmsVersion Dictionary   
 
 see:
 
 copyright:     SAP AG 1998 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/


#include "gco00.h"
#include "geo00.h"
#include "gco10.h"
#include "hco12.h"
#include "hco17.h"
#include "liveCache/LVC_LockRequest.hpp"
#include "Oms/OMS_VersionDirRegions.hpp"
#include "Oms/OMS_KernelDefaultInterface.hpp"




/*--------------------------------------------------------------------------*/
/*       Implementation of  CVersionDirctionary_co17                        */
/*--------------------------------------------------------------------------*/

CUnloadableVersions::CUnloadableVersions()
: m_first(NULL)
{
}

/*---------------------------------------------------------------------------*/

void CUnloadableVersions::InsertVersion(OmsContext_co12* pVersion, bool callFromDestructor)
{
    OmsContext_co12* curr = m_first;
    while (curr)
    {
        if (curr == pVersion)
        {
            DbpBase opMsg(co10_GetLcSink());
            opMsg.dbpOpError("ignored duplicate insert into unloadable versions");
            char versionId[sizeof(OmsVersionId)+1];
            memcpy(&versionId[0], &pVersion->GetVersionId()[0], sizeof(OmsVersionId));
            versionId[sizeof(OmsVersionId)] = 0;
            opMsg.dbpOpError("%s", &versionId[0]);
            if (callFromDestructor)
            {
                opMsg.dbpOpError("call from ~OmsSession_co12");           
            }
            else
            {
                opMsg.dbpOpError("call from ClearTransVersion");
            }
            return;
        }
        curr = curr->GetNextUnloaded();
    }
    pVersion->SetNextUnloaded(m_first);
    m_first = pVersion;
}

/*---------------------------------------------------------------------------*/

OmsContext_co12* CUnloadableVersions::RemoveVersion(OmsContext_co12* pVersion)
{
  OmsContext_co12* prev = NULL;
  OmsContext_co12* curr = m_first;
  while (NULL != curr)
  {
    if ((curr == pVersion) ||
       ((NULL == pVersion) && (NULL == curr->GetNextUnloaded()))) // remove last
    {
      if (NULL == prev)
      {
        m_first = curr->GetNextUnloaded();
      }
      else
      {
        prev->SetNextUnloaded(curr->GetNextUnloaded());
      }
      return curr;
    }
    else
    {
      prev = curr;
      curr = curr->GetNextUnloaded();
    }
  }
  return NULL;
}



/*--------------------------------------------------------------------------*/
/*       Implementation of  CVersionDirctionary_co17                        */
/*--------------------------------------------------------------------------*/

CVersionDictionary_co17::CVersionDictionary_co17() 
  : m_UnloadableVersions(),
    m_useRWLocks(true)
{
  for (int i=0; i<OMS_VDIR_SIZE; ++i){
    // Initialize buckets of hash-array
    m_pVersionDictionary[i] = NULL;
  }
}

/*--------------------------------------------------------------------------*/

tgg00_BasisError CVersionDictionary_co17::DropVersion (const OmsVersionId& versionId) {
  int slot = HashValue(&versionId);
  OmsContext_co12 *pCurr = m_pVersionDictionary[slot];
  OmsContext_co12 *pPred = NULL;
  while (pCurr){
    if (!memcmp(pCurr->GetVersionId(), versionId, sizeof(OmsVersionId))){
      if (pPred != NULL)
        pPred->SetNext(pCurr->GetNext());
      else
        m_pVersionDictionary[slot] = pCurr->GetNext();

      // PTS 1125426
      pCurr->SetNext(NULL);

      return e_ok;
    }
    pPred = pCurr;
    pCurr = pCurr->GetNext();
  }
  return e_unknown_version;
}

/*---------------------------------------------------------------------------*/

OmsContext_co12* CVersionDictionary_co17::FindVersion (const OmsVersionId& versionId) {
  OmsContext_co12 *pCurr = m_pVersionDictionary[HashValue(&versionId)];
  while (pCurr){
    if (!memcmp(pCurr->GetVersionId(), versionId, sizeof(OmsVersionId))){
      return pCurr;
    }
    pCurr = pCurr->GetNext();
  }
  return NULL;
}

/*---------------------------------------------------------------------------*/

tgg00_BasisError CVersionDictionary_co17::InsertVersion(const OmsVersionId& versionId,OmsContext_co12* pContext) {
  int slot = HashValue(&versionId);
  if (m_pVersionDictionary[slot] == NULL){
    m_pVersionDictionary[slot] = pContext;
  }
  else {
    OmsContext_co12* pCurr = m_pVersionDictionary[slot];
    while (true) {
      // Check if there is already a version with the same name
      if (!memcmp(pCurr->GetVersionId(), versionId, sizeof(OmsVersionId))){
        return e_duplicate_key;
      }

      if (pCurr->GetNext() == NULL){
        // Append new entry at end of list
        pCurr->SetNext(pContext);
        return e_ok;
      }
      else {
        pCurr = pCurr->GetNext();
      }
    }
  }

  return e_ok;
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::Shutdown() {
  for (int i=0; i<OMS_VDIR_SIZE; ++i){
    m_pVersionDictionary[i] = NULL;
  }
}

/*---------------------------------------------------------------------------*/

bool CVersionDictionary_co17::UnloadOldestVersion(IliveCacheSink* lcSink) 
{
  OmsContext_co12* pContext = NULL;
  { // Begin of exclusive critical section.   // PTS 1124170
    // Aquire all locks on the version dictionary.
    int         pLockId[OMS_VDIR_SIZE];
    int         size;
    GetAllLockIds(&pLockId[0], size);
    ExclusiveVersionDirRgn rgn(&pLockId[0], size, UseRWLocks());

    CriticalSection_co001 unloadRegion(lcSink, RGN_UNLOAD_VERSION);
    unloadRegion.Enter();
    pContext = m_UnloadableVersions.RemoveVersion(NULL);
    if (NULL != pContext)
    {
        DbpBase b(lcSink);
        char versionId[sizeof(OmsVersionId)+1];
        memcpy(&versionId[0], &pContext->GetVersionId()[0], sizeof(OmsVersionId));
        versionId[sizeof(OmsVersionId)] = 0;
        b.dbpOpMsg("unloading oms version %s", &versionId[0]);
        pContext->AssignLcSink(lcSink);
        pContext->UnLoad();
        pContext->ResetLcSink();
    }
  }  // End of exclusive critical section.

  return pContext == NULL ? false : true;
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::MarkNotUnloadable(IliveCacheSink* lcSink, OmsContext_co12* pContext)
{
  CriticalSection_co001 region(lcSink, RGN_UNLOAD_VERSION);
  region.Enter();
  m_UnloadableVersions.RemoveVersion(pContext);
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::MarkUnloadable(OmsContext_co12* pContext, bool callFromDestructor)
{
  CriticalSection_co12 region(pContext->m_session, RGN_UNLOAD_VERSION);
  region.Enter();
  m_UnloadableVersions.InsertVersion(pContext, callFromDestructor);
}

/*---------------------------------------------------------------------------*/

int CVersionDictionary_co17::GetSingleLockId(const OmsVersionId &versionId){
  return HashValue(&versionId) + GetVDirStartCnt();
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::GetMultiLockIds(const OmsVersionId** ppVersionId, int &size, int *pLockId){
  int i, j;
  int tmp[OMS_VDIR_SIZE];

  // Initialize temporary array
  for (i=0; i<OMS_VDIR_SIZE; ++i){
    tmp[i] = -1;
  }

  // Determine hash slots which corresponds to the given version-ids. 
  for (i=0; i<size; ++i){
    int slot = HashValue(ppVersionId[i]);
    if (tmp[slot] != -1)
      tmp[slot] = slot + GetVDirStartCnt();
  }

  // "Compress" the array
  for (i=0,j=0; i<OMS_VDIR_SIZE; ++i){
    if (tmp[i] != -1)
      pLockId[j++] = tmp[i];
  }

  // Return the number of lock-ids
  size = j;
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::GetAllLockIds(int *pLockId, int &size){
  for (int i=0; i<OMS_VDIR_SIZE; ++i)
    pLockId[i] = i + GetVDirStartCnt();
  size   = OMS_VDIR_SIZE;
}

/*---------------------------------------------------------------------------*/

bool CVersionDictionary_co17::isLockedByOwn(int slot) const {  
  int lockId = slot + GetVDirStartCnt();
  if (m_useRWLocks){
    LVC_LockRequest lock(LVC_LockRequest::RWLOCK_IS_LOCKED, lockId);
    short err = co10_GetKernelInterface()->LockRequest(lock);
    return (err == 0 ? true : false);
  }
  else {
    bool inRgn = false;
    co10_GetLcSink()->IsInRegion(lockId, inRgn);
    return inRgn;
  }
}

/*--------------------------------------------------------------------------*/
/*       Implementation of  CVersionDirctionary_co17::Iter                  */
/*--------------------------------------------------------------------------*/


CVersionDictionary_co17::Iter::Iter()
//: m_slot(0),
//  m_lockMode(NO_LOCK),
//  m_pCurr(NULL),
//  m_pVD(NULL)
{
  m_slot     = 0;
  m_lockMode = NO_LOCK;
  m_pCurr    = NULL;
  m_pVD      = NULL;

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "CVersionDictionary_co17::Iter::CTOR-1-");
}
  
/*---------------------------------------------------------------------------*/

CVersionDictionary_co17::Iter::Iter(LockMode lockMode, CVersionDictionary_co17 &versionDictionary)
//: m_lockMode(lockMode),
//  m_slot(0),
//  m_pVD(&versionDictionary),
//  m_pCurr(NULL)
{ 
  m_lockMode  = lockMode;
  m_slot      = 0;
  m_pVD       = &versionDictionary;
  m_pCurr     = NULL;

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "BEGIN CVersionDictionary_co17::Iter::CTOR-2-");

  while (m_slot < OMS_VDIR_SIZE){
    // Get appropriate lock on the current slot
    enterRgn();

    if (m_pVD->m_pVersionDictionary[m_slot] != NULL){
      m_pCurr = m_pVD->m_pVersionDictionary[m_slot];
      break;
    }

    // Release lock on the current slot
    leaveRgn();

    ++m_slot;
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "END CVersionDictionary_co17::Iter::CTOR-2-" << (m_pCurr==0 ? " = EMPTY" : ""));
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::Iter::first(LockMode lockMode, CVersionDictionary_co17 &versionDictionary){
  m_lockMode   = lockMode;
  m_slot       = 0;
  m_pVD        = &versionDictionary;
  m_pCurr      = NULL;

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "BEGIN CVersionDictionary_co17::Iter::first");
  
  while (m_slot < OMS_VDIR_SIZE){
    // Get appropriate lock on the current slot
    enterRgn();

    if (m_pVD->m_pVersionDictionary[m_slot] != NULL){
      m_pCurr = m_pVD->m_pVersionDictionary[m_slot];
      break;
    }

    // Release lock on the current slot
    leaveRgn();

    ++m_slot;
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "END CVersionDictionary_co17::Iter::first" << (m_pCurr==0 ? " = EMPTY" : ""));
}

/*---------------------------------------------------------------------------*/

CVersionDictionary_co17::Iter::Iter(const Iter &iter){
  // Copying is not allowed
  throw DbpError(e_invalid_command, "CVersionDictionary_co17::Iter::Iter");
}

/*---------------------------------------------------------------------------*/

CVersionDictionary_co17::Iter& CVersionDictionary_co17::Iter::operator=(const CVersionDictionary_co17::Iter &iter){
  // Assignment is not allowed
  throw DbpError(e_invalid_command, "CVersionDictionary_co17::Iter::operator==");
  return *this;
}

/*---------------------------------------------------------------------------*/
CVersionDictionary_co17::Iter::~Iter(){
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "BEGIN CVersionDictionary_co17::Iter::DTOR");

  if (m_pCurr != NULL){
    // Release lock on the current slot
    leaveRgn();
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "END CVersionDictionary_co17::Iter::DTOR");
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::Iter::getNext(bool dropCurr){
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "BEGIN CVersionDictionary_co17::Iter::getNext" << (dropCurr ? " DropCurr" : ""));

  if (m_pCurr == NULL){
    throw DbpError(e_invalid_iterator, "CVersionDictionary_co17::Iter::getNext");
    return;
  }

  if (m_pCurr->GetNext() != NULL){
    OmsContext_co12 *pNext = m_pCurr->GetNext();
    if (dropCurr)
      m_pVD->DropVersion(m_pCurr->GetVersionId());
    m_pCurr = pNext;
  }
  else {
    if (dropCurr)
      m_pVD->DropVersion(m_pCurr->GetVersionId());

    // Release lock on the slot which was aquired by the previous iterator step
    leaveRgn();
    m_pCurr = NULL;

    //Searching for an entry starting from the next slot
    ++m_slot;
    while (m_slot < OMS_VDIR_SIZE){
      // Get appropriate lock on the current slot
      enterRgn();

      if (m_pVD->m_pVersionDictionary[m_slot] != NULL){
        m_pCurr = m_pVD->m_pVersionDictionary[m_slot];
        break;
      }

      // Release lock on the current slot
      leaveRgn();

      ++m_slot;
    }
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "END CVersionDictionary_co17::Iter::getNext" << (m_pCurr==0 ? " = EMPTY" : ""));
}

/*---------------------------------------------------------------------------*/

CVersionDictionary_co17::Iter::operator bool() const {
  return (m_pCurr != NULL);
}

/*---------------------------------------------------------------------------*/

OmsContext_co12* CVersionDictionary_co17::Iter::operator()() const {
  return m_pCurr;
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::Iter::enterRgn(){
  if (m_pCurr != NULL || m_slot >= OMS_VDIR_SIZE){
    throw DbpError(e_invalid_iterator, "CVersionDictionary_co17::Iter::enterRgn");
    return;
  }

  int lockId = m_slot + m_pVD->GetVDirStartCnt();

  //if (m_pVD->isLockedByOwn(m_slot)){
  //  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), 
  //  //  "Try to enter already locked region " << lockId);
  //  throw DbpError(e_invalid_iterator, "CVersionDictionary_co17::Iter::enterRgn-2-");
  //  return;
  //}

  if (m_lockMode == NO_LOCK)
    return;


  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), 
  //  (m_lockMode == EXCLUSIVE ? "Exclusive " : "Shared ") << "Region requested: " << lockId);

  if (m_pVD->m_useRWLocks){
    LVC_LockRequest lock(m_lockMode == SHARED ? LVC_LockRequest::RWLOCK_LOCK_SHARED : LVC_LockRequest::RWLOCK_LOCK_EXCLUSIVE, 
                         lockId);
    short err = co10_GetKernelInterface()->LockRequest(lock);
    if (err != e_ok)
      throw DbpError(err, "CVersionDictionary_co17::Iter::enterRgn");
  }
  else {
    co10_GetLcSink()->EnterCriticalSection(lockId);
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), 
  //  (m_lockMode == EXCLUSIVE ? "Exclusive " : "Shared ") << "Region aquired: " << lockId);
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::Iter::leaveRgn(){
  if (m_slot >= OMS_VDIR_SIZE){
    throw DbpError(e_invalid_iterator, "CVersionDictionary_co17::Iter::leaveRgn");
    return;
  }

  int lockId = m_slot + m_pVD->GetVDirStartCnt();

  //if (!m_pVD->isLockedByOwn(m_slot)){
  //  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), 
  //  //  "Try to free unlocked region " << lockId);
  //  throw DbpError(e_invalid_iterator, "CVersionDictionary_co17::Iter::leaveRgn-2-");
  //  return;
  //}

  if (m_lockMode == NO_LOCK)
    return;

  if (m_pVD->m_useRWLocks){
    LVC_LockRequest lock(m_lockMode == SHARED ? LVC_LockRequest::RWLOCK_UNLOCK_SHARED : LVC_LockRequest::RWLOCK_UNLOCK_EXCLUSIVE, 
                         lockId);
    short err = co10_GetKernelInterface()->LockRequest(lock);
    if (err != e_ok)
      throw DbpError(err, "CVersionDictionary_co17::Iter::leaveRgn");
  }
  else {
    co10_GetLcSink()->LeaveCriticalSection(lockId);
  }

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), 
  //  (m_lockMode == EXCLUSIVE ? "Exclusive " : "Shared ") << "Region freed: " << lockId);
}

/*---------------------------------------------------------------------------*/

void CVersionDictionary_co17::Iter::stop(){
  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "BEGIN CVersionDictionary_co17::Iter::stop");

  if (m_pCurr == NULL){
    return;
  }

  leaveRgn();
  
  m_pCurr    = NULL;
  m_slot     = 0;
  m_pVD      = NULL;

  //OMS_TRACE(omsTrInterface, co10_GetLcSink(), "END CVersionDictionary_co17::Iter::stop");
}

/*---------------------------------------------------------------------------*/

