/*!***************************************************************************

  module      : gen71.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: dispatcher
  description : Dispatcher and Queue Management


  last changed: 1999-01-27  17:45
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef GEN71_H
#define GEN71_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "gen50.h"


/* ---- long term move task locks */
#define NEVER_MOVE_TASK                  0x0001
#define CLIENT_NOT_TASK_MOVE_ENABLED     0x0002


/* - used by 'en71AlterTaskCnt' */
#define ALL_COUNTER                0x00FF
#define ACTIVE_COUNTER             0x0001
#define TOTAL_COUNTER              0x0002

#define COUNTER_ADD                0x0100
#define COUNTER_SUB                0x0200

#define ADD_ALL_COUNTER            ALL_COUNTER | COUNTER_ADD
#define SUB_ALL_COUNTER            ALL_COUNTER | COUNTER_SUB
#define ADD_ACTIVE_COUNTER         ACTIVE_COUNTER | COUNTER_ADD
#define SUB_ACTIVE_COUNTER         ACTIVE_COUNTER | COUNTER_SUB
#define ADD_TOTAL_COUNTER          TOTAL_COUNTER | COUNTER_ADD
#define SUB_TOTAL_COUNTER          TOTAL_COUNTER | COUNTER_SUB


/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/
/*!
  Chapter: Dispatcher
 */
/*!
  Function:     en71Dispatcher
  description:  Select the next running task in a UKT
    
  This function is called within the RTE code, whenever a new task selection
  is performed. It therefore scans the UKT specific queues for candidates and
  moves them into the running queue if they are allowed to run.

  It begins scanning the queue that queues request from other UKTs (XXX)

  Second scan is done on the queue of outstanding asynchronous IO (AIOC)

  Third scan is done on the queue of outstanding communication request (COM)

  Forth scan is done on the queue of internal requests to UKT (IOC)

  Then the top element of the run queue is selected.

  If run queue was empty after all these scans UKT starts sleeping on its semaphore.

  arguments:
    this_ukt [in] Pointer to current UKT Control structure

  return value: none
 */
externC void en71Dispatcher ( ten50_UKT_Control *this_ukt );


/*!
   @brief          Set temporary lock to prevent moving task from one UKT to another.
   @param          tcb [in] Pointer to current Task Control structure
   @param          cnt [in] temporary lock count
   @param          setLock [in] true if the lock should be set, else false
   @return         none

 */

externC void en71TempMoveLock( struct TASK_TYPE   *tcb,
                               SAPDB_UInt2        cnt,
                               SAPDB_Bool         setLock );


/*!
   @brief          Set long term lock to prevent moving task from one UKT to another.
   @param          tcb [in] Pointer to current Task Control structure
   @param          lockType [in] type of lock (e.g. client is not able to handle task moves)
   @param          setLock [in] true if the lock should be set, else false
   @return         none

 */

externC void en71LongTermMoveLock( struct TASK_TYPE   *tcb,
                                   SAPDB_UInt2        lockType,
                                   SAPDB_Bool         setLock );

/*@)*/
/*! @name Queuing */
 /*@(*/
/*!
  Function:     e71_ioc
  description:  Enqueue element in IOC queue
  arguments:
    tid [in] Pointer to current Task Control structure
    quu [in] Pointer to request element to be enqueued
  return value: none
 */

externC void en71_ioc ( ten50_UKT_Control    * this_ukt,
                        struct TASK_TYPE     * tid ,
                        struct DOUBLY_LINKED * quu );
 
 /*!
  Function:     en71_iocEnquAndWake
  description:  Enqueue element in IOC queue and wakeup UKT
  arguments:
	this_ukt [in] Pointer to current UKT Control structure
    tid [in] Pointer to current Task Control structure
    quu [in] Pointer to request element to be enqueued
  return value: none
 
  */
 
 externC void en71_iocEnquAndWake ( ten50_UKT_Control    * this_ukt,
                                    struct TASK_TYPE     * tid ,
                                    struct DOUBLY_LINKED * quu );

/*!
  Function:     e71_com
  description:  Enqueue task in COM queue
  arguments:
    tcb [in] Pointer to current Task Control Block
  return value: none
 */
externC void e71_com ( struct TASK_TYPE     * tcb );

/*!
  Function:     e71_EnquRav
  description:  Enqueue task in RUN queue using BASE_PRIO_RAV

  If Task was waiting in a vsuspend() call, this call is used.

  arguments:
    tcb [in] Pointer to current Task Control Block
    Prio [in] If set increase dynamic priority
  return value: none
 */
externC void e71_EnquRav ( struct TASK_TYPE * tcb ,
                           tsp00_Bool         Prio );

/*!
  Function:     e71_EnquRex
  description:  Enqueue task in RUN queue using BASE_PRIO_REX

  If Task was waiting rescheduling itself, this call is used.
  Base priority is lower than BASE_PRIO_RAV.

  arguments:
    tcb [in] Pointer to current Task Control Block
    Prio [in] If set increase dynamic priority
  return value: none
 */
externC void e71_EnquRex ( struct TASK_TYPE * tcb ,
                           tsp00_Bool         Prio );

/*!
  Function:     e71_FastEnquRex
  description:  Enqueue task in RUN queue using BASE_PRIO_REX

  If Task was waiting rescheduling itself while waiting for a spinlock, this call is used.
  Base priority is lower than BASE_PRIO_RAV.

  arguments:
    this_ukt [in] Pointer to current user kernel thread
    Prio [in] If set increase dynamic priority
  return value: none
 */
externC void e71_FastEnquRex ( ten50_UKT_Control * this_ukt
                              , tsp00_Bool          Prio );

/*!

  Enqueue a function call to be executed on other UKT.
  Function is called with two argument

    void Func(void *FuncParam, struct TASK_TYPE *FromTask)

  arguments:
    SourceUkt [in] Pointer to source User Kernel Thread
    TargetTcb [in] Pointer to target Task Control Block
    FromTcb   [in] Pointer to source Task Control Block
    Func [in] Function to execute in target UKT
    FuncParam [in] Function func parameter
  return value: none
 */

externC void en71EnqueueUkt2Ukt ( struct ten50_UKT_Control * SourceUkt ,
                                  struct TASK_TYPE     * TargetTcb ,
                                  struct TASK_TYPE     * FromTcb ,
                                  VOIDFUNC_PTR           Func ,
                                  void                 * FuncParam );


/*!
   @brief          Alter UKTs task counter values
   @param          tid [in] Pointer to current Task Control structure
   @param          alterType [in] Use one of the following defines:
                  ADD_ALL_COUNTER 
                  SUB_ALL_COUNTER
                  ADD_ACTIVE_COUNTER
                  SUB_ACTIVE_COUNTER
                  ADD_TOTAL_COUNTER 
                  SUB_TOTAL_COUNTER
   @return         none

 */

externC void  en71AlterTaskCnt( struct TASK_TYPE   *tcb,
                                SAPDB_UInt2        alterType );


/*@)*/


/*!
   @brief          Check load balancing
   @param          none
   @return         next check wait time

 */
time_t  en71LoadBalancing();


#endif  /* GEN71_H */
