/*!

  module      : hen09.h

  -------------------------------------------------------------------------

  mesponsible : Joerg Mensing - D027923

  special area: Registry for UNIX
  description : Handling of '.ini' files on UNIX

    Each .ini File consists of sections, which are identified by a section header.
    The section header consists of an name framed by brackets.
    Each entry in the section consists of a key an value separated by an equal sign.
    The section is seperated from the next section by an empty line.

    Example  SAPDB_Tech.ini

    <UL>
    <LI>[Database]
    <LI>SUT72=/home/remuser/releases/V72/usr
    <LI>MUT7264=/home/remuser/releases/V7264/usr
    </UL>

  first created: 1998-06-23  12:30
  last changed:  1999-12-14  16:48
  -------------------------------------------------------------------------
  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

#ifndef HEN09_H
#define HEN09_H

#ifdef REL62
#define MAXERRORTEXT 40
typedef char tsp00_ErrText[MAXERRORTEXT];
#else
#ifndef GSP00_H
#include "gsp00.h"
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* error codes used by config file functions */
#define config_error_chen09         0
#define config_error_ok_chen09      (config_error_chen09)
#define config_error_open_chen09    (config_error_chen09 + 1)
#define config_error_close_chen09   (config_error_chen09 + 2)
#define config_error_temp_chen09    (config_error_chen09 + 3)
#define config_error_write_chen09   (config_error_chen09 + 4)
#define config_error_section_chen09 (config_error_chen09 + 5)
#define config_error_entry_chen09   (config_error_chen09 + 6)	
#define config_error_key_chen09     (config_error_chen09 + 7)
#define config_error_end_chen09     (config_error_chen09 + 8)
#define config_error_read_chen09    (config_error_chen09 + 9)
#define config_error_lock_chen09    (config_error_chen09 + 10)
#define config_error_trunc_chen09   (config_error_chen09 + 11)
#define config_error_seek_chen09    (config_error_chen09 + 12)
#define config_error_param_chen09   (config_error_chen09 + 13)

/*
PTS 1103527
typedef struct then09_ConfigHandleStruct * then09_ConfigHandle;
 */
typedef int then09_ConfigHandle;

/*!
  Function:     sqlGetConfigString ()

  description:  Read Config Entry
  
    Gives back a string from ini-file 'szFile' from the position descripted by
    'szSection' and 'szEntry'. It fails if either the file, the section or the
    entry does not exist.

  arguments:
    szFile [in] Name of registry file
    szSection [in] Given section to read from
    szEntry [in] Given key of 'key=value' entry
    szString [out] Place to store the value of 'key=value' entry
    MaxStringLen [in] Maximum value buffer size
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

    Possible values for Ok
    <UL>
      <LI>config_error_open_chen09    failed to open registry
      <LI>config_error_lock_chen09    failed to lock registry
      <LI>config_error_read_chen09    failed to read registry
      <LI>config_error_section_chen09 failed to find section
      <LI>config_error_trunc_chen09   returned truncated value
      <LI>config_error_param_chen09   NULL pointer given
      <LI>config_error_ok_chen09    No error
    </UL>

  Return value: none, but err is updated to
    <UL>
        <LI>0: failed ( inspect Ok )
        <LI>1: ok
    </UL>
 */
int sqlGetConfigString (
    const char * szFile,
    const char * szSection,
    const char * szEntry,
    char       * szString,
    const int    StringSize,
    tsp00_ErrText VAR_ARRAY_REF ErrText,
    unsigned char VAR_VALUE_REF Ok);

/*!
  Function:     sqlPutConfigString ()
  description:  Write or update Registry Entry
  
   Inserts or replaces a string into the ini-file name 'szString'
   to a position descripted by 'section'-name and 'entry'-name.
 
   If the file does not exists, it will be created
   files are located in 'INI_PATH', which is '/usr/spool/sql/ini'
   If this directory does not exists, the directory 'ini' will be 
   created in '/usr/spool/sql'
   If '/usr/spool/sql' does not exists the function failed
  
   This function needs write permission in 'INI_PATH'
   to create a temporary file, otherwise it fails 
    
  arguments:
    szFile [in] Name of registry file
    szSection [in] Section to write into
    szEntry [in] Given key of 'key=value' entry
    szString [in] Given value of 'key=value' entry
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

    Possible values for Ok
    <UL>
      <LI>config_error_key_chen09   bad key given (contains '=') or bad entry found (no '=')
      <LI>config_error_temp_chen09  failed to create temp file or directory
      <LI>config_error_open_chen09  failed to open registry
      <LI>config_error_lock_chen09  failed to lock registry
      <LI>config_error_read_chen09  failed to read registry
      <LI>config_error_write_chen09 failed to write temp file
      <LI>config_error_seek_chen09  failed to seek to begin of file
      <LI>config_error_param_chen09   NULL pointer given
      <LI>config_error_ok_chen09  No error
    </UL>

  return value: none, but err is updated to
    <UL>
        <LI>0: failed ( inspect Ok )
        <LI>1: ok
    </UL>
 */
int sqlPutConfigString (
    const char * szFile,
    const char * szSection,
    const char * szEntry,
    const char * szString,
    tsp00_ErrText VAR_ARRAY_REF ErrText,
    unsigned char VAR_VALUE_REF Ok);

/*!
  Function:     sqlRemoveConfigString ()
  description:  Remove entry from Registry
                
    Removes the entry specified by 'szSection' and 'szEntry' from config file 

    This function needs write permission in 'INI_PATH'
    to create a temporary file, otherwise it fails .
    
  arguments:
    szFile [in] Name of registry file
    szSection [in] Section to write into
    szEntry [in] Given key of 'key=value' entry
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

    Possible values for Ok
    <UL>
      <LI>config_error_key_chen09   bad key given (contains '=') or bad entry found (no '=')
      <LI>config_error_temp_chen09  failed to create temp file or directory
      <LI>config_error_open_chen09  failed to open registry
      <LI>config_error_lock_chen09  failed to lock registry
      <LI>config_error_read_chen09  failed to read registry
      <LI>config_error_write_chen09 failed to write temp file
      <LI>config_error_seek_chen09  failed to seek to begin of file
      <LI>config_error_entry_chen09 failed to find entry
      <LI>config_error_param_chen09   NULL pointer given
      <LI>config_error_ok_chen09  No error
    </UL>

  return value: none, but err is updated to
    <UL>
        <LI>0: failed ( inspect Ok )
        <LI>1: ok
    </UL>
 */
int sqlRemoveConfigString (
    const char * szFile,
    const char * szSection,
    const char * szEntry,
    tsp00_ErrText VAR_ARRAY_REF ErrText,
    unsigned char VAR_VALUE_REF Ok);

/*!
  Function:     sqlOpenConfigEnum ()
  description:  Open registry for enumeration of a section
                
    Open the registery and position to the given section.
    It is expected that a loop over 'sqlNextConfig()' is used to retrieve the
    entries.

    Note: The returned handle has no state information. The complete Registry is locked.
    Therefore you must call sqlCloseConfigEnum() after last usage, or the Registry is locked
    until end of process.
  
  arguments:
    szFile [in] Name of registry file
    szSection [in] Section to write into
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

    Possible values for Ok
    <UL>
      <LI>config_error_open_chen09    failed to open registry
      <LI>config_error_lock_chen09    failed to lock registry
      <LI>config_error_read_chen09    failed to read registry
      <LI>config_error_section_chen09 failed to find section
      <LI>config_error_param_chen09   NULL pointer given
      <LI>config_error_ok_chen09    No error
    </UL>

  return value:
    <UL>
        handle  used in subsequent sqlNextConfig calls
    </UL>
 */
then09_ConfigHandle sqlOpenConfigEnum (
    const char * szFile,
    const char * szSection,
    tsp00_ErrText VAR_ARRAY_REF ErrText, 
    unsigned char VAR_VALUE_REF Ok);

/*!
  Function:     sqlCloseConfigEnum ()
  description:  Closes registry opened by sqlOpenConfigFile ()
                
    Close the registry. This invalidates the given handle.
    Subsequent calls to sqlNextConfig() must fail.
  
  arguments:
    hEnum [in] Handle of registry file
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

    Possible values for Ok
    <UL>
      <LI>config_error_close_chen09   failed to close registry
      <LI>config_error_ok_chen09    No error
    </UL>

  return value:
    <UL>
        <LI>0: failed ( inspect Ok )
        <LI>1: ok
    </UL>
 */
int sqlCloseConfigEnum (
    then09_ConfigHandle hEnum,
    tsp00_ErrText VAR_ARRAY_REF ErrText,
    unsigned char VAR_VALUE_REF Ok);

/*!
  Function:     sqlNextConfig ()
  description:  Reads next entry from registry opened by sqlOpenConfigFile ()
                
   Reads next 'szEntry' and 'szString' from section of config file opened by 
   sqlOpenConfigEnum (). If caller like, he can pass NULL pointer, which make
   it possible ie. just to count the entries of a section...

   Gives back 'config_error_end_chen09' in 'Ok' on end of section or end of file.
  
  arguments:
    hEnum [in] Handle of registry file
    szEntry [out] Place to store to key of 'key=value' entry (NULL means dont store)
    MaxEntryLen [in] Maximum key buffer size
    szString [out] Place to store the value of 'key=value' entry (NULL means dont store)
    MaxStringLen [in] Maximum value buffer size
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

    Possible values for Ok
    <UL>
      <LI>config_error_read_chen09    failed to read registry
      <LI>config_error_key_chen09     entry contains no seperator
      <LI>config_error_trunc_chen09   returned truncated value
      <LI>config_error_end_chen09     at end of section or file
      <LI>config_error_ok_chen09    No error
    </UL>

  return value:
    <UL>
        <LI>0: failed or at end of section ( inspect Ok )
        <LI>1: ok
    </UL>
 */
int sqlNextConfig (
    then09_ConfigHandle hEnum,
    char *szEntry,
    const int MaxEntryLen, 
    char *szString,
    const int MaxStringLen, 
    tsp00_ErrText VAR_ARRAY_REF ErrText,
    unsigned char VAR_VALUE_REF Ok);

#ifdef __cplusplus
}
#endif

#endif

