/****************************************************************************

  module      : ven09.c

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Registry for UNIX
  description : Handling of '.ini' files on UNIX

    Each .ini File consists of sections, which are identified by a section header.
    The section header consists of an name framed by brackets.
    Each entry in the section consists of a key an value separated by an equal sign.
    The section is seperated from the next section by an empty line.

    Example

    [Database]
    SUT72=/home/remuser/releases/V72/usr
    MUT7264=/home/remuser/releases/V7264/usr

  created:      1998-06-23  12:30
  last changed: 2000-11-30  14:02
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/

/*
 * first check if any UNIX platform is defined
 */
#if !(defined SUN || defined AIX || defined OSF1 || defined NMP || defined HPUX || defined(LINUX))
#error SUN, AIX, OSF1, NMP, HPUX or LINUX not defined !
#endif

#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "hen09.h"
#include "gen00.h"
#include "hen40.h"

#ifdef NMP
void *alloca(size_t);
#endif /* NMP */

/*
 *  declarations for config file processing
 */
#define LINE_SIZE_MAX 800
#define INI_PATH "/usr/spool/sql/ini"
#define MOD__ "ven09.c:"

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*
 * This must be implemented as a Macro, since it uses alloca...
 */
#define READLINE_EN09(_Fd, _Buffer, _OldSize, _Size, _MoreData, _Result) \
  _OldSize = 0; \
  for ( _MoreData = true; _MoreData; ) \
  { \
    _Result = ReadLine ( _Fd, _Buffer+_OldSize, (_Size-_OldSize), &_MoreData ); \
    if ( _Result != 1 ) \
    { \
      break; \
    } \
\
    if ( _MoreData ) \
    { \
      char *SecondBuffer; \
      _OldSize = _Size-1; \
      _Size += LINE_SIZE_MAX; \
\
      SecondBuffer = (char *)alloca(_Size); \
      strcpy( SecondBuffer, _Buffer); \
      _Buffer = SecondBuffer; \
    } \
  }

/*===========================================================================*
 *  LOCAL FUNCTIONS (PROTOTYPES)                                             *
 *===========================================================================*/

static const char *GetRegistryPath();
static int ReadLine (int, char *, int /* PTS 1104679 */, bool * );
static int FindSection (int, const char *);
static int WriteSection (int, const char*, int);
static int WriteEntry (int, const char *, const char *);
static int WriteLine (int, const char*);
static int UpdateConfigString (const char *szFile, 
                               const char *szSection, 
                               const char *szEntry,
                               const char *szString,
                               const bool DeleteFlag,
                               tsp00_ErrText ErrText,
                               unsigned char *pOk);
typedef struct _LockFile {
    bool m_IsTemp;
    const char *m_Path;
    int Fd;
    bool m_IsOpen;
    bool m_IsLocked;
} LockFile;

static void RegistryFile_Init(LockFile *this, bool IsTemp);
static bool RegistryFile_Open(LockFile *this, const char *Path);
static bool RegistryFile_Create(LockFile *this, const char *Path);
static bool RegistryFile_Lock(LockFile *this);
static bool RegistryFile_Reset(LockFile *this);
static void RegistryFile_Close(LockFile *this);

/*---------------------------------------------------------------------------*/

/*
  Class:     LockFile
  description:  Helper class to make sure file lock is removed and file is closed after return

    LockFile object is created without any file handle attached.
    After call to Open() the file handle is stored internally.
    The member function 'Lock()' is called afterwards.
    Destructor will remove the lock automatically if function returns and close the File.

  arguments: none
  return value: none
 */
static void RegistryFile_Init(LockFile *this, bool IsTemp)
{
  this->m_IsTemp = IsTemp;
  this->m_Path   = "";
  this->Fd     = -1;
  this->m_IsOpen = false;
  this->m_IsLocked = false;
}

static bool RegistryFile_Open(LockFile *this, const char *Path)
{
  if ( !this->m_IsOpen )
  {
    this->m_Path = Path;
    /* To obtain an exclusive lock file must be opened read/write */
    this->Fd = open(Path, O_RDWR);
    this->m_IsOpen = ( this->Fd != -1 );
  }
  return this->m_IsOpen;
}

static bool RegistryFile_Create(LockFile *this, const char *Path)
{
  if ( !this->m_IsOpen )
  {
    /* PTS 1105151 */
    int  umask_old ;
    umask_old = umask(0);  
    this->m_Path = Path;
    this->Fd = open(Path, O_RDWR|O_CREAT, 0666);
    this->m_IsOpen = ( this->Fd != -1 );
    umask(umask_old);  
  }
  return this->m_IsOpen;
}

static bool RegistryFile_Lock(LockFile *this)
{
  if ( !this->m_IsLocked 
     && RegistryFile_Reset(this) )
  {
    this->m_IsLocked = ( lockf(this->Fd, F_LOCK, (off_t)0) != -1 ); /* PTS 1108386 */
  }
  return this->m_IsLocked;
}

static bool RegistryFile_Reset(LockFile *this)
{
  if ( !this->m_IsOpen
    || lseek(this->Fd, (off_t)0, SEEK_SET) == -1 ) /* CTS 1108386 */
  {
      return false;
  }
  return true;
}

static void RegistryFile_Close(LockFile *this)
{
  if ( this->m_IsOpen )
  {
    this->m_IsLocked = this->m_IsOpen = ( close(this->Fd) == -1 );
  }
  if ( this->m_IsTemp )
  {
    unlink(this->m_Path);
  }
}

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

/*
  Function:     sqlGetConfigString ()
  description:  Read Config Entry
  
    Gives back a string from ini-file 'szFile' from the position descripted by
    'szSection' and 'szEntry'. It fails if either the file, the section or the
    entry does not exist.
    
  arguments:
    szFile [in] Name of registry file
    szSection [in] Given section to read from
    szEntry [in] Given key of 'key=value' entry
    szString [out] Place to store the value of 'key=value' entry
    MaxStringLen [in] Maximum value buffer size
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

  return value: none, but err is updated to
    <UL>
        <LI>0: Write failed ( inspect errno )
        <LI>1: Write ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"sqlGetConfigString"
int sqlGetConfigString (
                        const char *    szFile,
                        const char *    szSection,
                        const char *    szEntry, 
                        char       *    szString,
                        const int       MaxStringLen,
                        tsp00_ErrText   ErrText,
                        unsigned char * pOk)
{
  
  char * pVal;
  char * ptr;
  char * szPath;
  int    ReadResult;
  LockFile RegistryFile;

  RegistryFile_Init(&RegistryFile, false);

  if ( !szFile || !szSection || !szEntry || !szString )
  {
    *pOk = config_error_param_chen09;
    strcpy ( (char *)ErrText, "GetConfigString NULL pointer passed" );
    return 0;
  }

  if ( *szEntry == 0 )
  {
    *pOk = config_error_param_chen09;
    strcpy ( (char *)ErrText, "GetConfigString no key passed" );
    return 0;
  }

  *szString = 0;

  szPath = (char *)alloca(strlen(GetRegistryPath())+1+strlen(szFile)+1);
  strcpy (szPath, GetRegistryPath());
  strcat (szPath, "/");
  strcat (szPath, szFile);

  if ( !RegistryFile_Open (&RegistryFile, szPath) )
  {
    *pOk = config_error_open_chen09;
    strcpy ( (char *)ErrText, "Open Registry:" );
    strncat ( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    return 0;
  }

  if ( !RegistryFile_Lock(&RegistryFile) )
  {
    *pOk = config_error_lock_chen09;
    strcpy ( (char *)ErrText, "Lock(Registry) failed:" );
    strncat ( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    return 0;
  }

  ReadResult = FindSection (RegistryFile.Fd, szSection);
  if ( ReadResult == -1 )
  {
    szString[0] = '\0';
    *pOk = config_error_read_chen09;
    strcpy( (char *)ErrText, "Read section(Registry) '" );
    strncat( (char *)ErrText, szSection, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
    return 0;
  }

  if ( ReadResult == 0) 
  {
    szString[0] = '\0';
    *pOk = config_error_section_chen09;
    strcpy( (char *)ErrText, "Section not in Registry:" );
    strncat( (char *)ErrText, szSection, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
    return 0;
  }
  {
  size_t EntryLen = strlen(szEntry);
  bool FoundEntry = false;

  char *Buffer = (char *)alloca(LINE_SIZE_MAX);
  int   OldBufferSize = 0;
  int   BufferSize = LINE_SIZE_MAX;
  bool  MoreData;

  do
  {
    READLINE_EN09(RegistryFile.Fd, Buffer, OldBufferSize, BufferSize, MoreData, ReadResult);
    if ( ReadResult == -1 )
    {
      szString[0] = '\0';
      *pOk = config_error_read_chen09;
      strcpy( (char *)ErrText, "Read problem(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
      return 0;
    }

    if ( ReadResult == 0 
      || Buffer[0] == '\0'
      || Buffer[0] == '[' )
    {
      szString[0] = '\0';
      *pOk = config_error_section_chen09;
      strcpy( (char *)ErrText, "Entry not in Registry:" );
      strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
      return 0;
    }

    /* PTS 1105650 */
    pVal = strchr (Buffer, '=');

  } while ( pVal == NULL 
         || ( EntryLen != ( pVal - Buffer ) )
         || ( strncasecmp (Buffer, szEntry, EntryLen) != 0 ) );
                         
/* PTS 1105122 */
  if ( pVal != NULL )
  {
    ++pVal;
    strncpy(szString, pVal, MaxStringLen-1);
    szString[MaxStringLen-1] = 0;

    if ( strlen(pVal) >= MaxStringLen )
    {
      *pOk = config_error_trunc_chen09;
      sprintf( (char *)ErrText, "Value [%d/%d] truncated:", MaxStringLen, strlen(pVal)+1 );
      strncat( (char *)ErrText, Buffer, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    }
    else
    {
      *pOk = config_error_ok_chen09;
      ErrText[0] = '\0';
    }
  }
  else
  {
    *pOk = config_error_key_chen09;
    strcpy( (char *)ErrText, "Bad entry:");
    strncat( (char *)ErrText, Buffer, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
    return 0;
  }
  }
RegistryFile_Close(&RegistryFile);
  return (strlen (szString));
}

/*---------------------------------------------------------------------------*/

/*
  Function:     sqlPutConfigString ()
  description:  Write or update Registry Entry
  
   Inserts or replaces a string into the ini-file name 'szString'
   to a position descripted by 'section'-name and 'entry'-name.
 
   If the file does not exists, it will be created
   files are located in 'INI_PATH', which is '/usr/spool/sql/ini'
   If this directory does not exists, the directory 'ini' will be 
   created in '/usr/spool/sql'
   If '/usr/spool/sql' does not exists the function failed
  
   This function needs write permission in 'INI_PATH'
   to create a temporary file, otherwise it fails 
    
  arguments:
    szFile [in] Name of registry file
    szSection [in] Section to write into
    szEntry [in] Given key of 'key=value' entry
    szString [in] Given value of 'key=value' entry
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

  return value: none, but err is updated to
    <UL>
        <LI>0: failed ( inspect errno )
        <LI>1: ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"sqlPutConfigString"
int sqlPutConfigString (
    const char *szFile,
    const char *szSection,
    const char *szEntry, 
    const char *szString,
    tsp00_ErrText ErrText,
    unsigned char *pOk)
{
    if ( !szFile || !szSection || !szEntry || !szString )
    {
      *pOk = config_error_param_chen09;
      strcpy( (char *)ErrText, "PutConfigString NULL pointer passed" );
      return 0;
    }
    return ( UpdateConfigString (szFile, szSection, szEntry, szString, false, ErrText, pOk) );
}

/*---------------------------------------------------------------------------*/

/*
  Function:     sqlRemoveConfigString ()
  description:  Remove entry from Registry
                
    Removes the entry specified by 'szSection' and 'szEntry' from config file 

    This function needs write permission in 'INI_PATH'
    to create a temporary file, otherwise it fails .
    
  arguments:
    szFile [in] Name of registry file
    szSection [in] Section to write into
    szEntry [in] Given key of 'key=value' entry
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

  return value: none, but err is updated to
    <UL>
        <LI>0: failed ( inspect errno )
        <LI>1: ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"sqlRemoveConfigString"
int sqlRemoveConfigString (
    const char *szFile,
    const char *szSection,
    const char *szEntry, 
    tsp00_ErrText ErrText,
    unsigned char *pOk)
{
    if ( !szFile || !szSection || !szEntry )
    {
      *pOk = config_error_param_chen09;
      strcpy( (char *)ErrText, "RemoveConfigString NULL pointer passed" );
      return 0;
    }
    return ( UpdateConfigString (szFile, szSection, szEntry, NULL, true, ErrText, pOk) );
}

/*---------------------------------------------------------------------------*/

/*
  Function:     sqlOpenConfigEnum ()
  description:  Open registry for enumeration of a section
                
    Open the registery and position to the given section.
    It is expected that a loop over 'sqlNextConfig()' is used to retrieve the
    entries.

    Note: The returned handle has no state information. The complete Registry is locked.
    Therefore you must call sqlCloseConfigEnum() after last usage, or the Registry is locked
    until end of process.
  
  arguments:
    szFile [in] Name of registry file
    szSection [in] Section to write into
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

  return value:
    <UL>
        handle  used in subsequent sqlNextConfig calls
    </UL>
 */
#undef MF__
#define MF__ MOD__"sqlOpenConfigEnum"
then09_ConfigHandle sqlOpenConfigEnum (
    const char *szFile,
    const char *szSection,
    tsp00_ErrText ErrText, 
    unsigned char *pOk)
{
    int      Fd;
    char *   szPath;
    int      ReadResult;

    if ( !szFile || !szSection )
    {
      *pOk = config_error_param_chen09;
      strcpy( (char *)ErrText, "OpenConfigEnum NULL pointer passed" );
      return 0;
    }

    szPath = (char *)alloca(strlen(GetRegistryPath())+1+strlen(szFile)+1);
    strcpy (szPath, GetRegistryPath());
    strcat (szPath, "/");
    strcat (szPath, szFile);

    Fd = open (szPath, O_RDWR);
    if (Fd < 0) 
    {
        *pOk = config_error_open_chen09;
        strcpy( (char *)ErrText, "Open Registry failed:" );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return (then09_ConfigHandle)0;
    }

    if ( lockf(Fd, F_LOCK, (off_t)0) == -1 ) /* PTS 1108386 */
    {
        close(Fd);
        *pOk = config_error_lock_chen09;
        strcpy( (char *)ErrText, "Lock Registry failed:" );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return (then09_ConfigHandle)0;
    }

    ReadResult = FindSection (Fd, szSection);
    if ( ReadResult == -1 )
    {
        close(Fd);
        *pOk = config_error_read_chen09;
        strcpy( (char *)ErrText, "Read Registry:");
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return (then09_ConfigHandle)0;
    }
    if ( ReadResult == 0) 
    {
        close(Fd);
        *pOk = config_error_section_chen09;
        strcpy( (char *)ErrText, "Can't find Registry section:" );
        strncat( (char *)ErrText, szSection, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return (then09_ConfigHandle)0;
    }

    ErrText[0] = '\0';
    *pOk = config_error_ok_chen09;
    return (then09_ConfigHandle)Fd;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     sqlCloseConfigEnum ()
  description:  Closes registry opened by sqlOpenConfigFile ()
                
    Close the registry. This invalidates the given handle.
    Subsequent calls to sqlNextConfig() must fail.
  
  arguments:
    hEnum [in] Handle of registry file
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

  return value:
    <UL>
        <LI>0: failed ( inspect errno )
        <LI>1: ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"sqlCloseConfigEnum"
int sqlCloseConfigEnum (
    then09_ConfigHandle hEnum,
    tsp00_ErrText ErrText,
    unsigned char *pOk)
{
    int rc;
    
    rc = close ((int)hEnum);
    if (rc < 0) 
    {
        *pOk = config_error_close_chen09;
        strcpy ( (char *)ErrText, "Close Registry:");
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return 0;
    }
    ErrText[0] = '\0';
    *pOk = config_error_ok_chen09;
    return 1;
}

/*
  Function:     sqlNextConfig ()
  description:  Reads next entry from registry opened by sqlOpenConfigFile ()
                
   Reads next 'szEntry' and 'szString' from section of config file opened by 
   sqlOpenConfigEnum ().
   Gives back 'config_error_end_chen09' on 'Ok' on end of section or end of file.
  
  arguments:
    hEnum [in] Handle of registry file
    szEntry [out] Place to store to key of 'key=value' entry
    MaxEntryLen [in] Maximum key buffer size
    szString [out] Place to store the value of 'key=value' entry
    MaxStringLen [in] Maximum value buffer size
    ErrText [out] Place for descriptive error message
    Ok [out] Error return value

  return value:
    <UL>
        <LI>0: failed ( inspect errno )
        <LI>1: ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"sqlNextConfig"
int sqlNextConfig (
    then09_ConfigHandle hEnum,
    char *szEntry,
    const int MaxEntryLen,
    char *szString,
    const int MaxStringLen,
    tsp00_ErrText ErrText,
    unsigned char *pOk)
{
    int i, j, State;
    int Fd = (int)hEnum;

    char *Buffer = (char *)alloca(LINE_SIZE_MAX);
    int   OldBufferSize = 0;
    int   BufferSize = LINE_SIZE_MAX;
    bool  MoreData;
    int   ReadResult;

    if ( szEntry != NULL ) *szEntry = 0;
    if ( szString != NULL ) *szString = 0;

    for (;;) 
    {
        READLINE_EN09(Fd, Buffer, OldBufferSize, BufferSize, MoreData, ReadResult);
        if ( ReadResult == -1 )
        {
          *pOk = config_error_read_chen09;
          strcpy( (char *)ErrText, "Read head Registry:" );
          strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
          return 0;
        }

        if ( ReadResult == 0 )
        {
          *pOk = config_error_end_chen09;
          strcpy( (char *)ErrText, "End of Registry" );
          return 0;
        }

        if (Buffer[0] == '[') 
        {
          *pOk = config_error_end_chen09;
          strcpy( (char *)ErrText, "End of Section" );
          return 0;
        }

        /* skip empty lines */
        if (Buffer[0] != '\0')
            break;
    }

    {
    char *EqualSign;

    EqualSign = strchr(Buffer, '=');
    if ( !EqualSign )
    {
      close(Fd);
      *pOk = config_error_key_chen09;
      strcpy( (char *)ErrText, "No '=' found in:" );
      strncat( (char *)ErrText, Buffer, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      return 0;
    }

    *EqualSign = '\0';

    if ( szEntry != NULL )
    {
      strncpy(szEntry, Buffer, MaxEntryLen-1);
      szEntry[MaxEntryLen-1] = '\0';

      if ( strlen(Buffer) >= MaxEntryLen )
      {
        *pOk = config_error_trunc_chen09;
        sprintf( (char *)ErrText, "Key [%d/%d] truncated:", MaxEntryLen, strlen(Buffer)+1 );
        strncat( (char *)ErrText, Buffer, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return 0;
      }
    }

    if ( szString != NULL )
    {
      strncpy(szString, EqualSign+1, MaxStringLen-1);
      szString[MaxStringLen-1] = '\0';

      if ( strlen(EqualSign+1) >= MaxStringLen )
      {
        *pOk = config_error_trunc_chen09;
        sprintf( (char *)ErrText, "Value [%d/%d] truncated:", MaxStringLen, strlen(EqualSign+1)+1 );
        strncat( (char *)ErrText, Buffer, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        return 0;
      }
    }
    }

    ErrText[0] = '\0';
    *pOk = config_error_ok_chen09;
    return 1;
}

/*===========================================================================*
 *  LOCAL FUNCTIONS (CODE)                                                   *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/
/*
  Function:     GetRegistryPath ()
  description:  Return the constant path to registry files directory

  arguments: none

  return value: 
     Path to registry directory
 */
static const char *GetRegistryPath()
{
  return INI_PATH;
}

/*---------------------------------------------------------------------------*/
/*
  Function:     ReadLine ()
  description:  Read from given filedescriptor until EOL('\n') or EOF
  
    Puts the next line as a zero terminated string to szBuffer.
    Any nonprintable characters are ignored (this works on dos files too).
    If given buffer space is not sufficient, 'MoreData' returns true.

    An empty line read is returned as a Buffer with first character set to 0, MoreData is false.

    If MoreData flag is set, the outer function must loop to read the rest of the line. Since buffer
    is always 0 terminated, it must use 'BufSize-1' bytes of each call to ReadLine().

    Single character I/O is used since no buffering is used.

  arguments:
    Fd [in] Opened Filedescriptor
    pBuffer [out] Pointer to Buffer space
    BufSize[in] Size of Bufferspace
    MoreData [out] Flag to indicate more data follows

  return value:
    <UL>
        <LI>0: EOL or EOF already reached (Buffer not modified)
        <LI>-1: Read error
        <LI>1: Buffer content filled
    </UL>
 */
#undef MF__
#define MF__ MOD__"ReadLine"
static int ReadLine (
                     int    Fd,
                     char * pBuffer,
                     int    BufSize,
                     bool * pMoreData )
{
  char Ch;
  int Val;
  char *ptr;
  int ReadLen ;
  int ReadResult ;

  *pMoreData = true;
  ptr = pBuffer;
  for (ReadLen=0; ReadLen<(BufSize-1); ReadLen++ ) 
  {
    Val = read (Fd, &Ch, 1);
    if (Val == 0)
    {
      *pMoreData = false;
      if (ptr > pBuffer)
      {
        DBG1 ((MF__,"eof at end of line", ptr )) ;
        break;
      }
      else
      {
        DBG1 ((MF__,"eof at begin of line", ptr )) ;
        return 0;
      }
    }

    if ( Val != 1 )
    {
      DBG1 ((MF__,"Read error:%s", sqlerrs() )) ;
      *pMoreData = false;
      *ptr = 0;
      return -1;
    }

    /* end of line */
    if (Ch == '\n')
    {
      *pMoreData = false;
      break;
    }
    /* ignore any other nonpritable character */
    if (Ch < ' ')
    {
      ReadLen--;
      continue;
    }
    *ptr++ = Ch;
  }
  *ptr = '\0';
  DBG1 ((MF__,"'%s'", pBuffer )) ;
  return 1;
}


/*---------------------------------------------------------------------------*/

/*
  Function:     FindSection ()
  description:  Read from given filedescriptor until given section is reached
  
   Next char to read is the first char of the first line of this section

  arguments:
    Fd [in] Opened Filedescriptor
    szSection [in] Given Section Identifier

  return value:
    <UL>
        <LI>0: EOF reached (Section not found)
        <LI>1: Section found
        <LI>-1: Read error
    </UL>
 */
#undef MF__
#define MF__ MOD__"FindSection"
static int FindSection (
                        int          Fd,
                        const char * szSection)
{
  char *szTempSection = (char *)alloca(strlen(szSection) + strlen("[]") + 1);
  
  char *Buffer = (char *)alloca(LINE_SIZE_MAX);
  int   BufferSize = LINE_SIZE_MAX;
  int   OldBufferSize = 0;
  bool  MoreData;
  int   ReadResult;

  strcpy (szTempSection, "[");
  strcat (szTempSection, szSection);
  strcat (szTempSection, "]");
  DBG1 ((MF__,"searching for section:'%s'", szTempSection )) ;

  do 
  {
    READLINE_EN09(Fd, Buffer, OldBufferSize, BufferSize, MoreData, ReadResult);
    if ( ReadResult != 1 )
    {
        DBG1 ((MF__,"read: no more data")) ;
        return ReadResult;
    }
  } while ( strcasecmp (Buffer, szTempSection) );
  return 1;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     WriteSection ()
  description:  Write a section header
  
   To seperate each section, the section header can be prepended by a newline.

   "[szSection]\n"   without NewLineFlag

   or

   "\n[szSection]\n" with NewLineFlag

  arguments:
    Fd [in] Opened Filedescriptor
    szSection [in] Given Section Identifier
    NewLineFlag [in] Flag to indicate additional Newline prepended

  return value:
    <UL>
        <LI>0: Write failed ( inspect errno )
        <LI>1: Write ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"WriteSection"
static int WriteSection (
                         int Fd,
                         const char *szSection,
                         int NewLineFlag)
{
  char *szTempSection;
  int cbLen, rc;

  szTempSection = (char *)alloca(strlen(szSection)+strlen("\n[]\n") + 1);
  if (NewLineFlag == 0) 
  {
    strcpy (szTempSection, "[");
  }
  else 
  {
    strcpy (szTempSection, "\n[");
  }
  strcat (szTempSection, szSection);
  strcat (szTempSection, "]\n");
  cbLen = strlen (szTempSection); 

  DBG1 ((MF__,"write: '%s'", szTempSection )) ;
  rc = write (Fd, szTempSection, cbLen);
  if (cbLen != rc) 
  {
    return 0;
  }
  return 1;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     WriteEntry ()
  description:  Write a entry line
  
    Write a line formed as 

    "szEntry=szString\n"

  arguments:
    Fd [in] Opened Filedescriptor
    szEntry [in] Given entry key
    szString [in] Given entry value
    
  return value:
    <UL>
        <LI>0: Write failed ( inspect errno )
        <LI>1: Write ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"WriteEntry"
static int WriteEntry (
                       int Fd,
                       const char *szEntry,
                       const char *szString)
{
  char *szTempEntry;
  int cbLen, rc;

  szTempEntry = (char *)alloca(strlen(szEntry) + strlen("=\n") + strlen(szString) + 1);
  strcpy (szTempEntry, szEntry);
  strcat (szTempEntry, "=");
  strcat (szTempEntry, szString);
  strcat (szTempEntry, "\n");
  
  cbLen = strlen (szTempEntry); 
  rc = write (Fd, szTempEntry, cbLen);
  if ( cbLen != rc) 
  {
    return 0;
  }
  return 1;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     WriteLine ()
  description:  Write a line formed read by ReadLine
  
    Used to copy Registry content into a temporary file. Need to add a newline on each line.
    
  arguments:
    Fd [in] Opened Filedescriptor
    szString [in] Given line
    
  return value:
    <UL>
        <LI>0: Write failed ( inspect errno )
        <LI>1: Write ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"WriteLine"
static int WriteLine (
                      int Fd,
                      const char *szString)
{
  char *szLine;
  int cbLen, rc;

  szLine = (char *)alloca(strlen(szString) + 2);
  strcpy (szLine, szString);
  strcat (szLine, "\n");
  cbLen = strlen (szLine);
  rc = write (Fd, szLine, cbLen);
  if (cbLen != rc)
  {
    return 0;
  }
  return 1;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     OpenTempConfigPath ()
  description:  Open the path to the temporary config file

    First create directory used for registry file.

  arguments:
    szTempPath [in] Complete path to temporary configuration file
    ErrText [out] Errormessage if something went wrong
    Ok [out] Error enumeration

  return value:
    <UL>
        <LI>true:  Open ok,
        <LI>false: Open failed (Ok and ErrText are set). 
    </UL>
 */
#undef MF__
#define MF__ MOD__"OpenTempConfigPath"
static bool OpenTempConfigPath( tsp00_ErrText ErrText,
                                unsigned char *pOk)
{
  struct stat StatBuff;
  int  umask_old ;

  umask_old = umask(0);

  if (stat (GetRegistryPath(), &StatBuff) != 0)
  {
    if ( mkdir (GetRegistryPath(), 0775) != 0 )
    {
      *pOk = config_error_temp_chen09;
      strcpy( (char *)ErrText, "Mkdir(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      umask(umask_old);
      return false;
    }
  }

  umask(umask_old);

  return true;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     OpenTempConfigFile ()
  description:  Open the temporary config file

    First create directory used for registry file. Then try to create the temporary file.

  arguments:
    szTempPath [in] Complete path to temporary configuration file
    ErrText [out] Errormessage if something went wrong
    Ok [out] Error enumeration

  return value:
    <UL>
        <LI>true:  Open ok,
        <LI>false: Open failed (Ok and ErrText are set). 
    </UL>
 */
#undef MF__
#define MF__ MOD__"OpenTempConfigFile"
static bool OpenTempConfigFile(LockFile *pTempFile,
                              const char *szTempPath,
                              tsp00_ErrText ErrText,
                              unsigned char *pOk)
{
  struct stat StatBuff;
  int  umask_old ;

  umask_old = umask(0);

  if ( !RegistryFile_Open(pTempFile, szTempPath) )
  {
    *pOk = config_error_temp_chen09;
    strcpy( (char *)ErrText, "Open(TempRegistry):" );
    strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    return false;
  }

  umask(umask_old);

  return true;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     CopyBackTempFile ()
  description:  Copy Temp file Back to Original

    A simple rename() would create a critical section since unlocking of registry 
    would takes place before rename() could start. So we have to copy content...
  
  arguments:
    TempFile [in] LockFile of Tempfile
    RegistryFile [in] LockFile of Registryfile
    ErrText [out] Errormessage if something went wrong
    Ok [out] Error enumeration

  return value:
    <UL>
        <LI>0: Write failed ( inspect Ok and ErrText )
        <LI>1: Write ok
    </UL>
 */
#undef MF__
#define MF__ MOD__"CopyBackTempFile"
static int CopyBackTempFile(  LockFile *pTempFile,
                              LockFile *pRegistryFile,
                              tsp00_ErrText ErrText,
                              unsigned char *pOk)
{
  char Buffer[4096];
  int  BytesRead = 0;
  int  BytesWritten = 0;
  int  TotalSize = 0;

  if ( !RegistryFile_Reset(pTempFile) )
  {
      *pOk = config_error_seek_chen09;
      strcpy( (char *)ErrText, "Seek(TempRegistry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(pRegistryFile);
/* PTS 1107324 */
RegistryFile_Close(pTempFile);
      return 0;
  }
  if ( !RegistryFile_Reset(pRegistryFile) )
  {
      *pOk = config_error_seek_chen09;
      strcpy( (char *)ErrText, "Seek(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(pRegistryFile);
/* PTS 1107324 */
RegistryFile_Close(pTempFile);
      return 0;
  }

  while ( (BytesRead = read(pTempFile->Fd, Buffer, 4096)) > 0 )
  {
    size_t BytesOffset;

    for ( BytesOffset = 0; 
          BytesOffset < BytesRead; 
          BytesOffset += BytesWritten, TotalSize += BytesWritten )
    {
      BytesWritten = write(pRegistryFile->Fd, Buffer+BytesOffset, BytesRead-BytesOffset);
      if ( (BytesWritten < 0) 
        && (errno != EINTR) )
      {
        break;
      }
    }
  }

  if ( BytesRead < 0 )
  {
      *pOk = config_error_read_chen09;
      strcpy( (char *)ErrText, "Read(TempRegistry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(pRegistryFile);
/* PTS 1107324 */
RegistryFile_Close(pTempFile);
      return 0;
  }

  if ( BytesWritten < 0 )
  {
      *pOk = config_error_write_chen09;
      strcpy( (char *)ErrText, "Writeback(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(pRegistryFile);
/* PTS 1107324 */
RegistryFile_Close(pTempFile);
      return 0;
  }

  ftruncate( pRegistryFile->Fd, TotalSize );

  ErrText[0] = '\0';
  *pOk = config_error_ok_chen09;
RegistryFile_Close(pRegistryFile);
/* PTS 1107324 */
RegistryFile_Close(pTempFile);
  return 1;
}

/*---------------------------------------------------------------------------*/

/*
  Function:     UpdateConfigString ()
  description:  Write or delete a complete config string
  
    Used to write a new entry into a config file if DeleteFlag == 0 or
    to delete an entry from config file if DeleteFlag != 0.
    
  arguments:
    szFile [in] Name of the configuration file
    szSection [in] Given section
    szEntry [in] Given entry key in 'key=value' string
    szString [in] Given entry value in 'key=value' string
    DeleteFlag [in] Flag to indicate delete instead of write
    ErrText [out] Errormessage if something went wrong
    Ok [out] Error enumeration

  return value:
    <UL>
        <LI>0: Write failed ( inspect ErrText )
        <LI>1: Write ok
    </UL>
 */
#undef  MF__
#define MF__ MOD__"UpdateConfigString"
static int UpdateConfigString (
                              const char *szFile,
                              const char *szSection,
                              const char *szEntry, 
                              const char *szString,
                              const bool DeleteFlag,
                              tsp00_ErrText ErrText,
                              unsigned char *pOk)
{
  int   BufferSize = LINE_SIZE_MAX;
  int   OldBufferSize = 0;
  char *Buffer = (char *)alloca(BufferSize);
  bool  MoreData;
  LockFile RegistryFile;
  LockFile TempFile;
  char *szPath;
  char *szTempPath;

  DBG1 ((MF__,"section:'%s', entry: '%s'", szSection, szEntry)) ;
  
  RegistryFile_Init(&RegistryFile, false);
  RegistryFile_Init(&TempFile, true);

  /* Validate given szEntry */
  if (strchr (szEntry, '=') != 0) 
  {
    *pOk = config_error_key_chen09;
    strcpy( (char *)ErrText, "Found '=' in key for Registry:" );
    strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
    return 0;
  }

  *pOk = config_error_ok_chen09;

  szPath = (char *)alloca(strlen(GetRegistryPath())+1+strlen(szFile)+1);
  strcpy (szPath, GetRegistryPath()); strcat (szPath, "/"); strcat (szPath, szFile);
  
  /* Create directory for temporary registry file */
  if ( !OpenTempConfigPath(ErrText, pOk) )
  {
    /* Ok & ErrText already setup... */
    return 0;
  }

  szTempPath = (char *)alloca(strlen(szPath) + strlen(".XXXXXX") + 1);
  strcpy (szTempPath, szPath); strcat (szTempPath, ".XXXXXX"); 
  /* PTS 1100015 */
  if ( mkstemp (szTempPath) == -1 )
  {
      *pOk = config_error_open_chen09;
      strcpy( (char *)ErrText, "mkstemp failed:" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      return 0;
  }

  /* Open temporary registry file */
  if ( !OpenTempConfigFile(&TempFile, szTempPath, ErrText, pOk) )
  {
    unlink(szTempPath);
    /* Ok & ErrText already setup... */
    return 0;
  }

  /* Open and lock Registry (create if does not exist */
  if ( !RegistryFile_Open (&RegistryFile, szPath ) )
  {
    if ( errno != ENOENT )
    {
      *pOk = config_error_open_chen09;
      strcpy( (char *)ErrText, "Open(Registry) failed:" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&TempFile);
      return 0;
    }

    /* Registry.ini-file did not exist before... */

    if ( !RegistryFile_Create(&RegistryFile, szPath) )
    {
      *pOk = config_error_open_chen09;
      strcpy( (char *)ErrText, "Create(Empty Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&TempFile);
      return 0;
    }

    if ( !RegistryFile_Lock(&RegistryFile) )
    {
      *pOk = config_error_lock_chen09;
      strcpy( (char *)ErrText, "Lock(Empty Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&TempFile);
      return 0;
    }

    if ( DeleteFlag )
    {
      /* nothing to delete */ 
      *pOk = config_error_entry_chen09;
      strcpy( (char *)ErrText, "Registry empty" );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }

    /* Write initial section */

    if (WriteSection (TempFile.Fd, szSection, 0) == 0) 
    {
      *pOk = config_error_write_chen09;
      strcpy( (char *)ErrText, "First write(Registry) section '" );
      strncat( (char *)ErrText, szSection, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }

    if (WriteEntry (TempFile.Fd, szEntry, szString) == 0) 
    {
      *pOk = config_error_write_chen09;
      strcpy( (char *)ErrText, "First write(Registry) entry '" );
      strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }

    return CopyBackTempFile(&TempFile, &RegistryFile, ErrText, pOk);

  }

  /*-------------------------------------
   * Registry File exists and is opened
   */
  if ( !RegistryFile_Lock(&RegistryFile) )
  {
    *pOk = config_error_lock_chen09;
    strcpy( (char *)ErrText, "Lock(Registry) failed:" );
    strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
    return 0;
  }

  /*------------------------------------------------
     Copy each line until searched section is found 
     If EOF is reached before, append new Section.
   */
  {
  int ReadResult;
  char *szTempSection = (char *)alloca( strlen("[]") + strlen(szSection) + 1 );
  /*  ini-file already exists */
  strcpy (szTempSection, "[");
  strcat (szTempSection, szSection);
  strcat (szTempSection, "]");

  do
  {
    READLINE_EN09(RegistryFile.Fd, Buffer, OldBufferSize, BufferSize, MoreData, ReadResult);
    if ( ReadResult == -1 )
    {
      *pOk = config_error_read_chen09;
      strcpy( (char *)ErrText, "Read head(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }

    if ( ReadResult == 0 )
    {
      /* EOF reached before section was found */
      if ( DeleteFlag )
      {
        /* nothing to delete */ 
        *pOk = config_error_entry_chen09;
        strcpy( (char *)ErrText, "No Registry section '" );
        strncat( (char *)ErrText, szSection, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, "'", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }

      /* no such section found, append it */
      if (WriteSection (TempFile.Fd, szSection, 1) == 0) 
      {
        *pOk = config_error_write_chen09;
        strcpy( (char *)ErrText, "Write(Registry) section '" );
        strncat( (char *)ErrText, szSection, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }

      if (WriteEntry (TempFile.Fd, szEntry, szString) == 0) 
      {
        *pOk = config_error_write_chen09;
        strcpy( (char *)ErrText, "Write(Registry) entry '" );
        strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }

      return CopyBackTempFile(&TempFile, &RegistryFile, ErrText, pOk);

    }

    if ( WriteLine (TempFile.Fd, Buffer) == 0 )
    {
      *pOk = config_error_write_chen09;
      strcpy( (char *)ErrText, "Copy write(Registry) failed:" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }

  } while ( strcasecmp (Buffer, szTempSection) != 0 );

  /*--------------------------------------------------
     searched section found,
     copy each line until searched entry key is found 
   */
  for (;;) 
  {
    READLINE_EN09(RegistryFile.Fd, Buffer, OldBufferSize, BufferSize, MoreData, ReadResult);
    if ( ReadResult == -1 )
    {
      *pOk = config_error_read_chen09;
      strcpy( (char *)ErrText, "Read(Registry) entries:" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }

    if ( ReadResult == 0 )
    {
      /* end of file, no such entry found */
      if ( DeleteFlag )
      {
        /* nothing to delete */ 
        *pOk = config_error_entry_chen09;
        strcpy( (char *)ErrText, "No Registry entry '" );
        strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, "'", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }

      /* append entry */
      if (WriteEntry (TempFile.Fd, szEntry, szString) == 0) 
      {
        *pOk = config_error_write_chen09;
        strcpy( (char *)ErrText, "Append(Registry) entry '" );
        strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }

      return CopyBackTempFile(&TempFile, &RegistryFile, ErrText, pOk);
    }

    {
    char *EqualSign = strchr(Buffer, '=');

    if ( EqualSign != NULL )
    {
      *EqualSign = '\0';
      if ( strcasecmp (Buffer, szEntry) == 0) 
      {
        /* entry found */ 
        if ( !DeleteFlag ) 
        {
          /* replace it */
          if (WriteEntry (TempFile.Fd, szEntry, szString) == 0) 
          {
            *pOk = config_error_write_chen09;
            strcpy( (char *)ErrText, "Replace(Registry) entry '" );
            strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
            strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
            strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
            return 0;
          }
        }   
        break;
      }
      *EqualSign = '=';
    }

    if ( Buffer[0] == '[' 
      || Buffer[0] == '\0' )
    {
      if ( DeleteFlag )
      {
        /* nothing to delete */ 
        *pOk = config_error_entry_chen09;
        strcpy( (char *)ErrText, "No Registry entry '" );
        strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
        strncat( (char *)ErrText, "'", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }
      else
      {
        /* insert it */
        if (WriteEntry (TempFile.Fd, szEntry, szString) == 0) 
        {
          *pOk = config_error_write_chen09;
          strcpy( (char *)ErrText, "Insert(Registry) entry '" );
          strncat( (char *)ErrText, szEntry, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
          strncat( (char *)ErrText, "':", sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
          strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
          return 0;
        }
      }

      if ( Buffer[0] == '[' )
      {
        if ( WriteLine (TempFile.Fd, "") == 0 )
        {
          *pOk = config_error_write_chen09;
          strcpy( (char *)ErrText, "Insert(Registry) empty line:" );
          strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
          return 0;
        }
      }

      if ( WriteLine (TempFile.Fd, Buffer) == 0 )
      {
        *pOk = config_error_write_chen09;
        strcpy( (char *)ErrText, "Insert(Registry) next section:" );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }

      break;
    }
    else
    {
      if ( !EqualSign )
      {
        *pOk = config_error_key_chen09;
        strcpy( (char *)ErrText, "No '=' found in:" );
        strncat( (char *)ErrText, Buffer, sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }
    }
    }

    if (WriteLine (TempFile.Fd, Buffer) == 0)
    {
      *pOk = config_error_write_chen09;
      strcpy( (char *)ErrText, "Copy entrys(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }
  }

  do
  {
    READLINE_EN09(RegistryFile.Fd, Buffer, OldBufferSize, BufferSize, MoreData, ReadResult);
    if ( ReadResult == -1 )
    {
      *pOk = config_error_read_chen09;
      strcpy( (char *)ErrText, "Read rest(Registry):" );
      strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
      return 0;
    }
    if ( ReadResult == 1 )
    {
      if ( WriteLine (TempFile.Fd, Buffer) == 0 )
      {
        *pOk = config_error_write_chen09;
        strcpy( (char *)ErrText, "Copy rest(Registry):" );
        strncat( (char *)ErrText, sqlerrs(), sizeof(tsp00_ErrText) - strlen((char *)ErrText) - 1 );
RegistryFile_Close(&RegistryFile);
RegistryFile_Close(&TempFile);
        return 0;
      }
    }
  } while( ReadResult != 0 );
  }

  return CopyBackTempFile(&TempFile, &RegistryFile, ErrText, pOk);
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/

#ifdef TEST

int main(int argc, char *argv)
{
  tsp00_ErrText ErrText;
  unsigned char Ok = 0;
  char          String[20];
  int rc;
  then09_ConfigHandle handle;
  char Entry[20];
  char Value[20];
  
  setbuf(stdout, NULL);

  unlink("/usr/spool/sql/ini/TestRegistry.ini");

  rc = sqlGetConfigString ( "TestRegistry.ini",
                            "Section",
                            "Key",
                            String,
                            3,
                            ErrText,
                            &Ok );
  printf("Get [Section] Key ->%s %s: Code %d | %s\n", 
          String, rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);

  rc = sqlPutConfigString ( "TestRegistry.ini",
                            "Section",
                            "Key",
                            "Value",
                            ErrText,
                            &Ok );
  printf("Put [Section] Key=Value %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  rc = sqlPutConfigString ( "TestRegistry.ini",
                            "Section2",
                            "Key2",
                            "Value2",
                            ErrText,
                            &Ok );
  printf("Put [Section2] Key2=Value2 %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  rc = sqlPutConfigString ( "TestRegistry.ini",
                            "Section",
                            "Key2",
                            "Value2",
                            ErrText,
                            &Ok );
  printf("Put [Section] Key2=Value2 %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  rc = sqlGetConfigString ( "TestRegistry.ini",
                            "Section",
                            "Key",
                            String,
                            sizeof(String),
                            ErrText,
                            &Ok );
  printf("Get [Section] Key ->%s %s: Code %d | %s\n", 
          String, rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  handle = sqlOpenConfigEnum ("TestRegistry.ini",
                              "Section",
                              ErrText,
                              &Ok );

  do {
    rc = sqlNextConfig ( handle,
                         Entry,
                         20,
                         Value,
                         20,
                         ErrText,
                         &Ok );
    printf("Next [Section] ->(%s=%s) %s: Code %d | %s\n",
          Entry, Value, rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);

  } while ( rc != 0 );
  
  rc = sqlRemoveConfigString ( "TestRegistry.ini",
                               "Section",
                               "Key",
                               ErrText,
                               &Ok );
  printf("Remove [Section] Key %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  sqlCloseConfigEnum ( handle,
                       ErrText,
                       &Ok );

  rc = sqlRemoveConfigString ( "TestRegistry.ini",
                               "Section",
                               "Key",
                               ErrText,
                               &Ok );
  printf("Remove [Section] Key %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  rc = sqlRemoveConfigString ( "TestRegistry.ini",
                            "Section2",
                            "Key2",
                            ErrText,
                            &Ok );
  printf("Remove [Section2] Key2 %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  rc = sqlRemoveConfigString ( "TestRegistry.ini",
                            "Section",
                            "Key2",
                            ErrText,
                            &Ok );
  printf("Remove [Section] Key2 %s: Code %d | %s\n",
          rc == 0 ? "failed" : "succeeded", Ok, &ErrText[0]);
  system("cat /usr/spool/sql/ini/TestRegistry.ini");

  return 0;
}

#endif
