/*!
  @file           ven88.cpp
  @author         JoergM
  @brief          Kernel RunTime: User Kernel Thread
  @see            

\if EMIT_LICENCE
  ========== licence begin LGPL
  Copyright (C) 2001-2003 SAP AG

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  ========== licence end
\endif
*/



#define MOD__ "ven88.cpp:"

#if defined(BIT64)
#  undef HAS_VERY_LIMITED_VIRTUAL_ADDRESS_SPACE
#  if defined(OSF1)
#    define HAS_LIMITED_NUMBER_OF_MMAP_SECTIONS
#  else
#    undef HAS_LIMITED_NUMBER_OF_MMAP_SECTIONS
#  endif /* OSF1 */
#else
/* Especially true for 32bit LINUX ... */
#  define HAS_VERY_LIMITED_VIRTUAL_ADDRESS_SPACE
#  if defined(LINUX)
#     define USE_VMALLOC_NOT_MMAP
#  endif
#endif /* BIT64 */

/* exported functions */

#include "gen88.h"
#include "hen51.h"

/* imported functions */
#include "geo007_1.h"
#include "geo50_0.h"
#include "heo69.h"  /* vsReinit() */
#include "geo002.h" /* XPARAM(showMaxStackUse/fUseProcessScope) */
#include "heo92.h"  /* eo92GetTaskTypeName() */
#include "heo07.h"  /* Thread functions */
#include "hen40.h"  /* sqlerrs() */ 
#include "hen50.h"  /* vabort() */ 
#include "gen71.h"  /* en71Dispatcher() */
#include "gen73.h"  /* en73_dl_enqu() */
#include "gen74.h"  /* e74_init_big_comseg() */
#include "gen81.h"
#ifdef DEBUG_RTE
#include "SAPDBCommon/SAPDB_sprintf.h" /*nocheck*/
#endif

#ifdef DEBUG_RTE
#include "gen38.h" /* e38_type_name() */
#endif /* DEBUG_RTE */

#include "RunTime/RTE_Crash.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"
#include "RunTime/System/RTESys_AtomicOperation.hpp"
#include "RunTime/RTE_Coroutine.h"
#include <alloca.h>

/*
 * Interface function of AK layer pascal procedure (for which no header file will ever exist...)
 */
externC void a91mainprogram ( void );
externC void en88_CallCoroutineKernelTaskMain ();

/*
   local structure describing task chunk parameters
 */
class RTE_TaskStackControl
{
protected:
    /*
       @brief container for a mapping of a single task stack array
     */
    struct TaskStackMapping {
        struct TaskStackMapping *m_Next;     ///< next task stack in chain of task stack mappings
        SAPDB_Byte *m_StartOfFirstTaskStack; ///< address of first byte of first task stack
        SAPDB_Byte *m_EndOfLastTaskStack;    ///< address behind end of last task stack
        SAPDB_ULong m_SingleStackSize;       ///< Sizeof single task stack
        SAPDB_ULong m_SingleGapSize;         ///< Sizeof single gap between task stacks
        RTE_TaskId  m_FirstId;               ///< TaskId associated with first task stack
    };

public:
    static RTE_TaskStackControl &Instance();

    /*
      Function: Initialize
      Description: Once called function to initialize task stack control
      Arguments: numberOfTasks [in] number of tasks
                 pFirst [in] pointer to first task control record
                 pLast [in] pointer to last task control record
     */
    virtual void Initialize( SAPDB_Long numberOfTasks, 
                             struct TASK_TYPE *pFirst,
                             struct TASK_TYPE *pLast) = 0;

    /*
      @brief Using current stack context return task control block of calling task or 0 if not inside task stack
      @return Task control block pointer of calling task or 0 if not on task stack
    */
    struct TASK_TYPE * GetTaskControl() const
    {
        // atomic reading of top of queue
        TaskStackMapping *pMapping = m_FirstMapping;
        SAPDB_Byte *CompareAddress = (SAPDB_Byte *)&pMapping;
#ifdef DEBUG_RTE
char eBuffer[256];
#endif
        while ( pMapping )
        {
            if ( CompareAddress >= pMapping->m_StartOfFirstTaskStack
              && CompareAddress <  pMapping->m_EndOfLastTaskStack )
            {
                SAPDB_ULong totalStackSize = pMapping->m_SingleStackSize + pMapping->m_SingleGapSize;

                SAPDB_Long indexInStackArray = (CompareAddress - pMapping->m_StartOfFirstTaskStack)
                                             / totalStackSize;

                if (             0 == pMapping->m_SingleGapSize
                  || CompareAddress < ( pMapping->m_StartOfFirstTaskStack 
                                      + ( indexInStackArray * totalStackSize )
                                      + pMapping->m_SingleStackSize ) )
                {
                    struct TASK_TYPE * tcb = (KGS->pFirstTaskCtrl + pMapping->m_FirstId - 1 + indexInStackArray);
#ifdef DEBUG_RTE
SAPDB_sprintf(eBuffer, sizeof(eBuffer), "CompareAddress 0x%p Mapping 0x%p ---> Task%d", CompareAddress, pMapping, tcb->index);
MSGD(( INFO_TEST_WILDCARD, eBuffer ));
#endif
#ifdef SAPDB_SLOW
                    if ( CompareAddress >= tcb->StackChunkEnd )
                    {
                        vabort(WRITE_CORE); // internal error : address not within task stack range
                    }
#endif /* SAPDB_SLOW */
                    return tcb;
                }
            }
            pMapping = pMapping->m_Next;
        }
#ifdef DEBUG_RTE
SAPDB_sprintf(eBuffer, sizeof(eBuffer), "CompareAddress 0x%p ---> NO MATCH!", CompareAddress );
MSGD(( ERR_TEST_WILDCARD, eBuffer ));
#endif
        return (struct TASK_TYPE *)0;
    }

    /*
      @brief The tasks may have the stack on stack of UKT thread. In that case the mapping information
      is known at a very late moment. This call defines the interface used in such a case.
      The call allows to associate a large number of task stack with a single call. This is used in a
      user task thread, where lots of identical tasks are added.

      If stack address space is not used for task stack, this call is not implemented.

      @param newMapping [in] structure filled with mapping
    */
    void  AddTaskStackMapping( TaskStackMapping &newMapping )
    {
        // atomic replacement of top of queue
        TaskStackMapping *pLastFirstMapping = m_FirstMapping;
        do {
            newMapping.m_Next = pLastFirstMapping;
        } while ( !RTESys_CompareAndExchange( m_FirstMapping, 
#ifdef OSF1
                                              pLastFirstMapping,
                                              &newMapping,
#else
                                              const_cast<TaskStackMapping * const>(pLastFirstMapping),
                                              const_cast<TaskStackMapping * const>(&newMapping),
#endif
                                              pLastFirstMapping ) );

#ifdef DEBUG_RTE
char eBuffer[256];
TaskStackMapping *pMapping = &newMapping;
SAPDB_sprintf(eBuffer, sizeof(eBuffer), "Added newMapping 0x%p Start 0x%p End 0x%p FirstId %d SingleStackSize %ld SingleGapSize %ld", 
            pMapping,
            pMapping->m_StartOfFirstTaskStack,
            pMapping->m_EndOfLastTaskStack,
            pMapping->m_FirstId,
            pMapping->m_SingleStackSize, 
            pMapping->m_SingleGapSize );
MSGD(( INFO_TEST_WILDCARD, eBuffer ));
#endif
    }

    /*
      @brief The tasks may have the stack on stack of UKT thread. In that case the mapping information
      is known at a very late moment. This call defines the interface used in such a case.
      The call allows to associate a large number of task stack with a single call. This is used in a
      user task thread, where lots of identical tasks are added.

      If stack address space is not used for task stack, this call is not implemented.

      @param firstStack [in] address of first stack to map
      @param singleStackSize [in] size of a single stack size in bytes
      @param firstId [in] the first task id associated with the first task stack
      @param numberOfTasksToAdd [in] the number of tasks added with identical task stack size
      @return true if task stack mapping was added, false if not
    */
    SAPDB_Bool AddTaskStackMapping( SAPDB_Byte *firstStack,
                                    SAPDB_ULong singleStackSize,
                                    RTE_TaskId  firstId,
                                    SAPDB_UInt4 numberOfTasksToAdd )
    {
        TaskStackMapping *pNewMapping = 
            (TaskStackMapping *)RTEMem_RteAllocator::Instance().Allocate(sizeof(TaskStackMapping));
        if ( !pNewMapping )
        {
            return false;
        }

        pNewMapping->m_StartOfFirstTaskStack = firstStack;
        pNewMapping->m_EndOfLastTaskStack = firstStack + (singleStackSize * numberOfTasksToAdd);
        pNewMapping->m_SingleStackSize = singleStackSize ;
        pNewMapping->m_SingleGapSize = 0;
        pNewMapping->m_FirstId = firstId;

        AddTaskStackMapping( *pNewMapping );
        return true;
    }

protected:
    /*-------------------------------------------------------------
      Function: CalculateMaxTaskStack
      Description: Calculate the maximum task stack size for all tasks
  
      Loop over all tasks and get the maximum. Take into account that maximum cannot be smaller than minimum thread stack size
      Arguments: numberOfTasks [in] number of tasks defined
                 pFirst[in] first task control
                 pLast [in] last task control
     */
    void CalculateMaxTaskStack( SAPDB_Long numberOfTasks,
                                struct TASK_TYPE *pFirst, 
                                struct TASK_TYPE *pLast);

    /*-------------------------------------------------------------
      Function: TryToGuard
      Description: Try to allocate and guard specified memory range

      If allacation at fixed address fails, it is fatal. If protection fails, it is a warning only...

      Arguments: startAddress [in] start address of memory area to protect
                 guardSize [in] Size of memory area to protect
     */
    void TryToGuard(SAPDB_Byte *startAddress, SAPDB_Long guardSize);

protected:
    RTE_TaskStackControl();

    void ReserveOrAllocateBigChunk(SAPDB_Bool tryReserve);

    void ConsumedMemoryInfo();

    SAPDB_Byte *m_StartAddress;             /* start address of complete task stack chunk */
    SAPDB_Byte *m_StartOfFirstTaskStack;    /* start address of first task stack in chunk */
    SAPDB_Long  m_TotalSize;                /* total size of task stack chunk including all guard pages */
    SAPDB_Long  m_SingleAlignedStackSize;   /* aligned stack size of single task */
    SAPDB_Long  m_SingleGuardSize;          /* size of a single guard area */
    SAPDB_Long  m_SingleProtectedStackSize; /* size of a single stack area including guard pages */
    SAPDB_Long  m_NumberOfTasks;            /* Number of tasks contained in chunk */
    SAPDB_Long  m_consumedPhysicalMemory;   /* Number of bytes used for stack pages */
    SAPDB_Long  m_consumedGuardedMemory;    /* Number of bytes used for guard pages */
    SAPDB_ULong m_maxTaskStackSize;         /* Maximum task stack size */

protected:
    SAPDB_Bool  m_useReservedAreaAndGaps;   /* Flag to show mprotect calls are not needed */
private:
    TaskStackMapping            *m_FirstMapping; /* first mapping */
    static RTE_TaskStackControl *m_Instance;
}; //  class RTE_TaskStackControl

RTE_TaskStackControl * RTE_TaskStackControl::m_Instance = 0;

class RTE_LimitedMmapSectionsTaskStackControl: public RTE_TaskStackControl
{
public:
    RTE_LimitedMmapSectionsTaskStackControl();
    /*---------------------------------------------------------------
    Function: UpdateTaskControlRecords
    Description: For limited address space, the single chunk allocation is all we do
    so only an update of task control resords is needed
      Arguments: pFirst[in] first task control
                 pLast [in] last task control

     */
    void UpdateTaskControlRecords(struct TASK_TYPE *pFirst, struct TASK_TYPE *pLast);

    virtual void Initialize( SAPDB_Long numberOfTasks,
                             struct TASK_TYPE *pFirst,
                             struct TASK_TYPE *pLast );
private:
    TaskStackMapping m_Mapping; ///< The only mapping needed mapping
}; // class RTE_LimitedMmapSectionsTaskStackControl: public RTE_TaskStackControl


class RTE_LimitedAddressSpaceTaskStackControl: public RTE_TaskStackControl
{
public:
    RTE_LimitedAddressSpaceTaskStackControl();
    virtual void Initialize( SAPDB_Long numberOfTasks,
                             struct TASK_TYPE *pFirst,
                             struct TASK_TYPE *pLast );
private:
    /*--------------------------------------------------------------------------------
        Function: TryToGuardStartPages
        Description: Try to guard all pages of a task stack

        If 64Bit address room allows it, the guard pages in front of task stack area are setup here

      Arguments: pFirst[in] first task control
                 pLast [in] last task control
        Return value: The number of bytes consumed by guard pages
     */
    SAPDB_Long TryToGuardStartPages(struct TASK_TYPE *pFirst, struct TASK_TYPE *pLast);

    /*-------------------------------------------------------------
    Function: AllocateAndProtectTaskStacks
    Description: Allocate and protect task stacks and guardpages (guardpages only if needed) 

    Take into account, that two different sizes are used for user task stack and non user task stacks

    Arguments: pFirst[in] first task control
               pLast [in] last task control
     */
    void AllocateAndProtectTaskStacks(struct TASK_TYPE *pFirst, struct TASK_TYPE *pLast);
    /*-------------------------------------------------------------
      Function: CalculateMaxNonUserTaskStack
      Description: Calculate the maximum task stack size for non user tasks

      Loop over all tasks and get the maximum task stack for non user tasks.
      Count the number of these non user tasks too.
  
      Arguments: pFirst[in] first task control
                 pLast [in] last task control
      Return value: maximum task stack size for non user tasks in bytes
     */
    SAPDB_ULong CalculateMaxNonUserTaskStack( struct TASK_TYPE *pFirst,
                                              struct TASK_TYPE *pLast );
    SAPDB_Byte *m_StartOfFirstUserTaskStack;/* start address of first user task stack in chunk */
    SAPDB_Long  m_SingleAlignedNonUserStackSize; /* aligned stack size of single non user task */
    SAPDB_Long  m_SingleProtectedNonUserStackSize; /* size of a single stack area including guard pages */
    SAPDB_Long  m_NumberOfNonUserTasks;     /* Number of non user tasks contained in chunk */
    TaskStackMapping m_NonUserTaskMapping; ///< The mapping for non user tasks
    TaskStackMapping m_UserTaskMapping; ///< The mapping for user tasks
}; // class RTE_LimitedAddressSpaceTaskStackControl: public RTE_TaskStackControl

class RTE_BigAddressSpaceTaskStackControl: public RTE_TaskStackControl
{
public:
    RTE_BigAddressSpaceTaskStackControl();
    virtual void Initialize( SAPDB_Long numberOfTasks,
                             struct TASK_TYPE *pFirst,
                             struct TASK_TYPE *pLast );
private:
    /*--------------------------------------------------------------------------------
        Function: TryToGuardStartPages
        Description: Try to guard all pages of a task stack

        If 64Bit address room allows it, the guard pages in front of task stack area are setup here

      Arguments: pFirst[in] first task control
                 pLast [in] last task control
        Return value: The number of bytes consumed by guard pages
     */
    SAPDB_Long TryToGuardStartPages(struct TASK_TYPE *pFirst, struct TASK_TYPE *pLast);

    /*-------------------------------------------------------------
    Function: AllocateAndProtectTaskStacks
    Description: Allocate and protect task stacks and guardpages (guardpages only if needed) 
    Arguments: pFirst[in] first task control
               pLast [in] last task control
     */
    void AllocateAndProtectTaskStacks(struct TASK_TYPE *pFirst, struct TASK_TYPE *pLast);

    TaskStackMapping m_Mapping; ///< The only mapping needed mapping
}; // class RTE_BigAddressSpaceTaskStackControl: public RTE_TaskStackControl

class RTE_StackAddressSpaceTaskStackControl: public RTE_TaskStackControl
{
public:
    RTE_StackAddressSpaceTaskStackControl()
    {}

    virtual void Initialize( SAPDB_Long,
                             struct TASK_TYPE *,
                             struct TASK_TYPE * )
    {}
}; // class RTE_LimitedAddressSpaceTaskStackControl: public RTE_TaskStackControl


/* local functions */

//  Must be visible only because C++ does not allow static extern "C" but pthread main entry must be extern "C"
externC void *      en88_RunPthreadTask(void *arg);
static  void        en88_InitializePthreadTaskUKT( struct  TASK_TYPE * firstTask );
static  void        en88_CallKernelTaskMain ( struct  TASK_TYPE * tcb );
static  void        en88_InitPthreadTasks ( ten50_UKT_Control *this_ukt );

volatile int en88_NeverStop = 1; /* used to prevent compiler warnings on wanted endless loops */

/*
  Function: en88CreateAndPrepareTaskStacks
  Description: Create and prepare all kernel task stacks

  There are several possible implementation variations:

  a) The kernel task stacks occupy a single contigous section in memory which is not separated by guard pages or gaps.
     Each task stack size is system page aligned. This model is used for HAS_VERY_LIMITED_VIRTUAL_ADDRESS_SPACE
     (default model for 32Bit maschines)

  b) The kernel task stacks occupy a single contigous section in memory which but are separated by guard pages or gaps.
     Each task stack size is system page aligned. This model is used for HAS_LIMITED_VIRTUAL_ADDRESS_SPACE

  c) The kernel task stacks occupy a single contigous section in memory which is not separated by guard pages or gaps
     but each task stack separated by the maximum task stack size of all task stacks HAS_LIMITED_NUMBER_OF_MMAP_SECTIONS

  d) Each kernel task stack gets its own mmap section, protected by gaps and with a fixed size spacing allowing gaps
     of unmapped pages in between (default model for 64Bit maschines).

  The fixed separation allows quiet fast calculation of task id using current task stack addresses. To allow this in model
  a) and b) only two kind of task stacks are allowed: UserTaskStacks and NonUserTaskStacks. In model c) and d) the maximum
  task stack which is allowed for all tasks is used as spacing distance. In model c) and d) the task stacks are allowed
  to grow up beginning with a definied minimum up to a defined maximum, while in model a) and b) the task stack are allocated
  once and forever at start.
 */
extern "C" void en88CreateAndPrepareTaskStacks()
{
#undef  MF__
#define MF__ MOD__"en88CreateAndPrepareTaskStacks"

    RTE_TaskStackControl::Instance().Initialize(KGS->ulNumOfTasks, KGS->pFirstTaskCtrl, KGS->pLastTaskCtrl);
}

/*---------------------------------------------------------------------------
  Function: vgetpid
  Description: Return task id

  Using the address of a local variable the stack area is searched and the task index returned.
  If called from non task thread, a zero is returned.

  Argument: taskId [out] task id set with current task id
 */
extern "C" void vgetpid ( 
tsp00_TaskId &taskId )
{
#undef  MF__
#define MF__ MOD__"vgetpid"
    struct TASK_TYPE *tcb = RTE_TaskStackControl::Instance().GetTaskControl();

    taskId = ( 0 == tcb ? 0 : tcb->index );
}

/*---------------------------------------------------------------------------
  @brief Return task id

  Using the address of a local variable the stack area is searched and the task index returned.
  If called from non task thread, a zero is returned.
 */
RTE_TaskId RTE_GetOwnTaskId()
{
    struct TASK_TYPE *tcb = RTE_TaskStackControl::Instance().GetTaskControl();
    return (RTE_TaskId)( 0 == tcb ? 0 : tcb->index );
}

/*---------------------------------------------------------------------------
  @brief Return task context pointer
  @param the task id 
 */
RTE_CoroutineContext *RTE_GetTaskContext(RTE_TaskId id)
{
    return (KGS->pFirstTaskCtrl + (id-1))->taskContext;
}

/*---------------------------------------------------
  Function: en88GetThisUkt
  Description: Returns current UKT control structure pointer or NULL if not called from within UKT

  This implements the THIS_UKT macro
  Return value: pointer to current UKT control structure or NULL if not called from within a UKT
 */
extern "C" ten50_UKT_Control *en88GetThisUkt()
{
#undef MF__
#define MF__ MOD__ "en88GetThisUkt"
    struct TASK_TYPE *tcb = RTE_TaskStackControl::Instance().GetTaskControl();

    return ( 0 == tcb ? (ten50_UKT_Control *)0 : tcb->ukt );
}

/*---------------------------------------------------------------------------
  Function: en88InitCoroutineTask
  Description: Initialize tasks using coroutines for task

  Arguments: tcb [in] Pointer to current task control block
 */
extern "C" tsp00_Int4 en88InitCoroutineTask ( struct  TASK_TYPE * tcb )
{
#undef MF__
#define MF__ MOD__ "en88InitCoroutineTask"
    DBGIN_T(tcb->index);

    tcb->taskContext = RTE_ContextAllocate(RTEMem_RteAllocator::Instance());
    if ( !tcb->taskContext )
    {
        vabort(WRITE_CORE);
        return false;
    }
    DBGOUT_T(tcb->index);

    return true;
}

/*---------------------------------------------------------------------------
  Function: en88_InitDynPthreadTasks
  Description: Initialize tasks using pthread for task

  The given tcb is recreated dynamically

  Arguments: tcb [in] Pointer to current task control block
 */
extern "C" tsp00_Int4 en88_InitDynPthreadTasks ( struct  TASK_TYPE * tcb )
{
#undef MF__
#define MF__ MOD__ "en88_InitDynPthreadTasks"
    DBGIN_T(tcb->index);

    tsp00_ErrTextc threadErrText;
    teo07_ThreadErr threadCallErr;

    sqlbeginthread_on_own_stack( (SAPDB_Int4)(tcb->StackChunkEnd - tcb->StackChunkStart),
                                 tcb->StackChunkStart,
                                 en88_RunPthreadTask,
                                 tcb,
                                 THR_CREATE_SUSPENDED_EO07 
                               | ( XPARAM(fUseProcessScope) ? THR_CREATE_PROCESS_SCOPE_EO07 : 0 ),
                                &tcb->taskThread,
                                 threadErrText,
                                &threadCallErr );
    if ( THR_OK_EO07 != threadCallErr )
    {
        MSGD(( ERR_CREATING_THREAD, threadErrText.asCharp(), threadCallErr )); /* PTS 1115185 */
        return false;
    }

    DBGOUT_T(tcb->index);

    return true;
}

/*===========================================================================*
 *  LOCAL/INLINE FUNCTIONS, LOCAL CLASS METHODS (IMPLEMENTATION)             *
 *===========================================================================*/

RTE_TaskStackControl & RTE_TaskStackControl::Instance()
{
    if ( KGS && XPARAM(fUseCoroutines) && XPARAM(fUseStackOnStack) )
    {
        SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTE_StackAddressSpaceTaskStackControl));
        if ( !m_Instance )
        {
            m_Instance = new (Space) RTE_StackAddressSpaceTaskStackControl();
        }
    }
    else
    {
#if defined(HAS_VERY_LIMITED_VIRTUAL_ADDRESS_SPACE)
        SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTE_LimitedAddressSpaceTaskStackControl));
        if ( !m_Instance )
        {
            m_Instance = new (Space) RTE_LimitedAddressSpaceTaskStackControl();
        }
#elif defined(HAS_LIMITED_NUMBER_OF_MMAP_SECTIONS)
        SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTE_LimitedMmapSectionsTaskStackControl));
        if ( !m_Instance )
        {
            m_Instance = new (Space) RTE_LimitedMmapSectionsTaskStackControl();
        }
#else
        SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTE_BigAddressSpaceTaskStackControl));
        if ( !m_Instance )
        {
            m_Instance = new (Space) RTE_BigAddressSpaceTaskStackControl();
        }
#endif
    }
    SAPDBERR_ASSERT_STATE(m_Instance != NULL);
    return *m_Instance;
}

RTE_TaskStackControl::RTE_TaskStackControl()
: m_StartAddress(0),
  m_StartOfFirstTaskStack(0),
  m_TotalSize(0),
  m_SingleAlignedStackSize(0),
  m_SingleGuardSize(0),
  m_SingleProtectedStackSize(0),
  m_NumberOfTasks(0),
  m_useReservedAreaAndGaps(false),
  m_FirstMapping(0)
{}

RTE_LimitedMmapSectionsTaskStackControl::RTE_LimitedMmapSectionsTaskStackControl()
{}

RTE_LimitedAddressSpaceTaskStackControl::RTE_LimitedAddressSpaceTaskStackControl()
: m_StartOfFirstUserTaskStack(0),
  m_SingleAlignedNonUserStackSize(0),
  m_SingleProtectedNonUserStackSize(0),
  m_NumberOfNonUserTasks(0)
{}

RTE_BigAddressSpaceTaskStackControl::RTE_BigAddressSpaceTaskStackControl()
{}

//-----------------------------------------------------------------------------
//--------- Reserve or allocate the big chunk ---------------------------------
//-----------------------------------------------------------------------------

void RTE_TaskStackControl::CalculateMaxTaskStack(
    SAPDB_Long numberOfTasks,
    struct TASK_TYPE *pFirst,
    struct TASK_TYPE *pLast)
{
    m_NumberOfTasks = numberOfTasks;

    for ( struct TASK_TYPE *tcb = pFirst; tcb <= pLast; ++tcb )
    {
        if ( tcb->stack_size > m_maxTaskStackSize )
        {
            m_maxTaskStackSize = tcb->stack_size;
        }
    }

    if ( m_maxTaskStackSize < sqlget_thread_min_stack() )
    {
        m_maxTaskStackSize = sqlget_thread_min_stack();
    }

    MSGD (( INFO_MAX_TASK_STACK, m_maxTaskStackSize/1024 ));

    m_SingleAlignedStackSize   = RTE_ISystem::Instance().RoundUpToSystemPageSize(m_maxTaskStackSize);
} // RTE_TaskStackControl::CalculateMaxTaskStack

// -----------------------------------------------------

void RTE_TaskStackControl::TryToGuard(SAPDB_Byte *startAddress, SAPDB_Long guardSize)
{
#ifndef USE_VMALLOC_NOT_MMAP
    if ( !RTE_ISystem::Instance().AllocSystemPagesAtFixedAddress(startAddress, guardSize, true) )
    {
        MSGD(( ERR_REALLOCATING_TASK_STACK, startAddress, startAddress + guardSize - 1 ));
        vabort(false);
    }
#endif /* USE_VMALLOC_NOT_MMAP */

    if ( RTE_ISystem::Instance().MemProtect( startAddress, guardSize, MEMORYPROTECTION_NOACCESS ) )
    {
        MSGD(( WRN_PROTECTING_TASK_STACK, startAddress, startAddress + guardSize - 1 ));
    }
} // RTE_TaskStackControl::TryToGuard

// -----------------------------------------------------------

void RTE_TaskStackControl::ReserveOrAllocateBigChunk(SAPDB_Bool tryReserve)
{
    if ( tryReserve )
    {
        // to be replaced by usage of RTEMem_ProtectedAllocator ????....
        m_StartAddress = (SAPDB_Byte *)RTE_ISystem::Instance().ReserveSystemPages(m_TotalSize);
        if ( m_StartAddress )
        {
            MSGD(( INFO_RESERVED_TASK_STACK, m_TotalSize/1024, m_StartAddress ));
        }
    }
    else
    {
        m_StartAddress             = (SAPDB_Byte *)0; // Any address possible
        MSGD(( INFO_SYSTEM_CHOOSEN_TASK_STACK ));
    }

    if ( 0 == m_StartAddress )
    {
        m_StartAddress = (SAPDB_Byte *)RTE_ISystem::Instance().AllocSystemPagesAtFixedAddress((void *)0, m_TotalSize, true);
    }
    else
    {
        m_useReservedAreaAndGaps = true;
        MSGD(( INFO_ALLOCATING_TASK_STACK_GAPS ));
    }

    MSGD(( INFO_ALLOCATING_TASK_STACK1, m_NumberOfTasks ));
    MSGD(( INFO_ALLOCATING_TASK_STACK2, m_SingleAlignedStackSize/1024 ));
    MSGD(( INFO_ALLOCATING_TASK_STACK3, m_SingleGuardSize/1024 ));

    if ( !m_StartAddress )
    {
        MSGD(( ERR_ALLOCATING_TASK_STACK1 ));
        MSGD(( ERR_ALLOCATING_TASK_STACK2 ));
        vabort(false);
    }
    else
    {
        m_StartOfFirstTaskStack = m_StartAddress + m_SingleGuardSize;
        MSGD(( INFO_ALLOCATING_TASK_STACK4,
               m_StartAddress,
               m_StartAddress + m_TotalSize - 1 ));
    }
} // RTE_TaskStackControl::ReserveOrAllocateBigChunk

// -------------------------------------------------------------

void RTE_LimitedMmapSectionsTaskStackControl::UpdateTaskControlRecords(
    struct TASK_TYPE *pFirst, 
    struct TASK_TYPE *pLast)
{
    //-----------------------------------------------------------------------------
    // Update individual task control record stack chunk information
    //-----------------------------------------------------------------------------

    SAPDB_Byte *stackStartAddress = m_StartOfFirstTaskStack;

    for ( struct TASK_TYPE * tcb = pFirst; tcb <= pLast; ++tcb )
    {
        SAPDB_Int4 alignedStackSize = RTE_ISystem::Instance().RoundUpToSystemPageSize(tcb->stack_size);
 
        tcb->StackChunkStart = stackStartAddress;
        tcb->StackChunkEnd   = stackStartAddress + alignedStackSize;
        tcb->stack_size      = alignedStackSize;

        tcb->GuardPageBegin  = 0;
        tcb->GuardPageEnd    = 0;

        m_consumedPhysicalMemory += m_SingleProtectedStackSize;

        // now we can step to the next gap
        stackStartAddress        += m_SingleProtectedStackSize;
    }
} // RTE_LimitedMmapSectionsTaskStackControl::UpdateTaskControlRecords

// --------------------------------------------------------------------------

void RTE_BigAddressSpaceTaskStackControl::AllocateAndProtectTaskStacks(
    struct TASK_TYPE *pFirst,
    struct TASK_TYPE *pLast)
{
    if ( !m_useReservedAreaAndGaps )
    {
#ifndef USE_VMALLOC_NOT_MMAP
        // We got the big chunk, but now we free it again, to reduce the amount of memory used...
        if ( !RTE_ISystem::Instance().FreeSystemPages(m_StartAddress, m_TotalSize) )
        {
            MSGD(( ERR_FREE_BIG_TASK_STACK_CHUNK ));
            vabort(false);
        }
#endif /* USE_VMALLOC_NOT_MMAP */

        if ( NUM_GUARDPAGES_EO70 > 0 )
        {
            m_consumedGuardedMemory = TryToGuardStartPages(pFirst, pLast);
        }
    }

    //-----------------------------------------------------------------------------
    // Allocate the stack pages at fixed addresses and allocate guard pages if wanted and no reservation was possible
    //-----------------------------------------------------------------------------

    SAPDB_Byte *stackStartAddress = m_StartOfFirstTaskStack;

    for ( struct TASK_TYPE *tcb = pFirst; tcb <= pLast; ++tcb )
    {
        SAPDB_Int4 alignedStackSize = RTE_ISystem::Instance().RoundUpToSystemPageSize(tcb->stack_size);
 
        tcb->StackChunkStart = stackStartAddress;
        tcb->StackChunkEnd   = stackStartAddress + alignedStackSize;
        tcb->stack_size      = alignedStackSize;

#ifndef USE_VMALLOC_NOT_MMAP
        if ( !RTE_ISystem::Instance().AllocSystemPagesAtFixedAddress(stackStartAddress, alignedStackSize, true) )
        {
            MSGD(( ERR_REALLOCATING_TASK_STACK, 
                   stackStartAddress, 
                   stackStartAddress + alignedStackSize - 1 ));
            vabort(false);
        }
#endif /* USE_VMALLOC_NOT_MMAP */

        // check for extra guard page needed
        if ( (NUM_GUARDPAGES_EO70 > 0) 
          && (alignedStackSize < m_maxTaskStackSize) )
        {
            // the stack chunk is smaller than maximum
            // the gap is unprotected! So we have to add an extra guard page
            // if the gap is smaller than the wanted guard page size, the 
            // smaller gap is completely protected and the adjacing guard
            // page guaranties that the guard page size is sufficient
            SAPDB_Long extraGuardSize = m_maxTaskStackSize - alignedStackSize;
            
            if ( extraGuardSize > m_SingleGuardSize )
            {
                extraGuardSize = m_SingleGuardSize;
            }

            tcb->GuardPageBegin = stackStartAddress + alignedStackSize;

            if ( extraGuardSize < m_SingleGuardSize )
            {
                tcb->GuardPageEnd   = tcb->GuardPageBegin 
                                    + m_maxTaskStackSize - alignedStackSize // see comment above
                                    + m_SingleGuardSize;
            }
            else
            {
                tcb->GuardPageEnd   = tcb->GuardPageBegin + m_SingleGuardSize;
            }

            if ( !m_useReservedAreaAndGaps )
            {
                TryToGuard( m_StartOfFirstTaskStack 
                          + m_SingleProtectedStackSize * (tcb->index-1)
                          + alignedStackSize,
                            extraGuardSize);
                m_consumedGuardedMemory += extraGuardSize;
            }
        }

        m_consumedPhysicalMemory += alignedStackSize;

        // now we can step to the next gap
        stackStartAddress += m_SingleProtectedStackSize;
    }
} // BigAddressSpaceTaskStackControl::AllocateAndProtectTaskStacks

// --------------------------------------------------------------------------

void RTE_LimitedAddressSpaceTaskStackControl::AllocateAndProtectTaskStacks(
    struct TASK_TYPE *pFirst,
    struct TASK_TYPE *pLast)
{
    if ( !m_useReservedAreaAndGaps )
    {
#ifndef USE_VMALLOC_NOT_MMAP
        // We got the big chunk, but now we free it again, to reduce the amount of memory used...
        if ( !RTE_ISystem::Instance().FreeSystemPages(m_StartAddress, m_TotalSize) )
        {
            MSGD(( ERR_FREE_BIG_TASK_STACK_CHUNK ));
            vabort(false);
        }
#endif /* USE_VMALLOC_NOT_MMAP */

        if ( NUM_GUARDPAGES_EO70 > 0 )
        {
            m_consumedGuardedMemory = TryToGuardStartPages(pFirst, pLast);
        }
    }

    //-----------------------------------------------------------------------------
    // Allocate the stack pages at fixed addresses and allocate guard pages if wanted and no reservation was possible
    //-----------------------------------------------------------------------------

    SAPDB_Byte *stackStartAddress = m_StartOfFirstTaskStack;

    for ( struct TASK_TYPE *tcb = pFirst; tcb <= pLast; ++tcb )
    {
        SAPDB_Int4 alignedStackSize = RTE_ISystem::Instance().RoundUpToSystemPageSize(tcb->stack_size);
 
        tcb->StackChunkStart = stackStartAddress;
        tcb->StackChunkEnd   = stackStartAddress + alignedStackSize;
        tcb->stack_size      = alignedStackSize;

#ifndef USE_VMALLOC_NOT_MMAP
        if ( !RTE_ISystem::Instance().AllocSystemPagesAtFixedAddress(stackStartAddress, alignedStackSize, true) )
        {
            MSGD(( ERR_REALLOCATING_TASK_STACK, 
                   stackStartAddress, 
                   stackStartAddress + alignedStackSize - 1 ));
            vabort(false);
        }
#endif /* USE_VMALLOC_NOT_MMAP */

        // non user max task stack size and user max task stack size differ
        SAPDB_Long taskTypeMaxStackSize;

        if ( tcb->type != TT_US_EO00
          && tcb->type != TT_UT_EO00 )
        {
            taskTypeMaxStackSize = m_SingleAlignedNonUserStackSize;
        }
        else
        {
            taskTypeMaxStackSize = m_SingleAlignedStackSize;
        }

        // check for extra guard page needed
        if ( (NUM_GUARDPAGES_EO70 > 0) 
          && (alignedStackSize < taskTypeMaxStackSize) )
        {
            // the stack chunk is smaller than maximum
            // the gap is unprotected! So we have to add an extra guard page
            // if the gap is smaller than the wanted guard page size, the 
            // smaller gap is completely protected and the adjacing guard
            // page guaranties that the guard page size is sufficient
            SAPDB_Long extraGuardSize;
            
            extraGuardSize = taskTypeMaxStackSize - alignedStackSize;
            
            if ( extraGuardSize > m_SingleGuardSize )
            {
                extraGuardSize = m_SingleGuardSize;
            }

            tcb->GuardPageBegin = stackStartAddress + alignedStackSize;

            if ( extraGuardSize < m_SingleGuardSize )
            {
                tcb->GuardPageEnd   = tcb->GuardPageBegin 
                                    + taskTypeMaxStackSize - alignedStackSize // see comment above
                                    + m_SingleGuardSize;
            }
            else
            {
                tcb->GuardPageEnd   = tcb->GuardPageBegin + m_SingleGuardSize;
            }

            if ( !m_useReservedAreaAndGaps )
            {
                TryToGuard( stackStartAddress + alignedStackSize,
                            extraGuardSize);
                m_consumedGuardedMemory += extraGuardSize;
            }
        }

        m_consumedPhysicalMemory += alignedStackSize;

        // now we can step to the next gap
        if ( tcb->type != TT_US_EO00
          && tcb->type != TT_UT_EO00 )
        {
            stackStartAddress += m_SingleProtectedNonUserStackSize;
        }
        else
        {
            stackStartAddress += m_SingleProtectedStackSize;
        }

    }
} // RTE_LimitedAddressSpaceTaskStackControl::AllocateAndProtectTaskStacks

// -------------------------------------------------------------

void RTE_TaskStackControl::ConsumedMemoryInfo()
{
    MSGD(( INFO_ALLOCATING_TASK_STACK5, m_consumedPhysicalMemory/1024 ));
    MSGD(( INFO_ALLOCATING_TASK_STACK6, m_consumedGuardedMemory/1024 ));
} // RTE_TaskStackControl::ConsumedMemoryInfo()

// -------------------------------------------------------------

SAPDB_Long RTE_BigAddressSpaceTaskStackControl::TryToGuardStartPages(
   struct TASK_TYPE *pFirst,
   struct TASK_TYPE *pLast)
{
    SAPDB_Byte *guardStartAddress = m_StartAddress;

    TryToGuard( guardStartAddress, m_SingleGuardSize);

    for ( struct TASK_TYPE *tcb = pFirst; tcb <= pLast; ++tcb )
    {
        guardStartAddress += m_SingleProtectedStackSize;

        tcb->GuardPageBegin = guardStartAddress;
        tcb->GuardPageEnd   = guardStartAddress + m_SingleGuardSize;

        TryToGuard( guardStartAddress, m_SingleGuardSize);
    }

    return m_SingleGuardSize * (1 + m_NumberOfTasks);
} // RTE_TaskStackControl::TryToGuardStartPages

// -------------------------------------------------------------

SAPDB_Long RTE_LimitedAddressSpaceTaskStackControl::TryToGuardStartPages(
   struct TASK_TYPE *pFirst,
   struct TASK_TYPE *pLast)
{
    SAPDB_Byte *guardStartAddress = m_StartAddress;

    TryToGuard( guardStartAddress, m_SingleGuardSize);

    for ( struct TASK_TYPE *tcb = pFirst; tcb <= pLast; ++tcb )
    {
        if ( tcb->type != TT_US_EO00
          && tcb->type != TT_UT_EO00 )
        {
            guardStartAddress += m_SingleProtectedNonUserStackSize;
        }
        else
        {
            guardStartAddress += m_SingleProtectedStackSize;
        }

        tcb->GuardPageBegin = guardStartAddress;
        tcb->GuardPageEnd   = guardStartAddress + m_SingleGuardSize;

        TryToGuard( guardStartAddress, m_SingleGuardSize);
    }

    return m_SingleGuardSize * (1 + m_NumberOfTasks);
} // RTE_TaskStackControl::TryToGuardStartPages

// -----------------------------------------------------

SAPDB_ULong RTE_LimitedAddressSpaceTaskStackControl::CalculateMaxNonUserTaskStack(
 struct TASK_TYPE *pFirst,
 struct TASK_TYPE *pLast )
{
    struct  TASK_TYPE * tcb;
    SAPDB_ULong         maxNonUserTaskStackSize = 0;

    m_NumberOfNonUserTasks = 0;

    for ( tcb = pFirst; tcb <= pLast; ++tcb )
    {
        if ( tcb->type != TT_US_EO00
          && tcb->type != TT_UT_EO00 )
        {
            ++m_NumberOfNonUserTasks;     /* Number of non user tasks contained in chunk */

            if ( tcb->stack_size > maxNonUserTaskStackSize )
            {
                maxNonUserTaskStackSize = tcb->stack_size;
            }
        }
    }

    if ( maxNonUserTaskStackSize < sqlget_thread_min_stack() )
    {
        maxNonUserTaskStackSize = sqlget_thread_min_stack();
    }

    MSGD (( INFO_MAX_NON_USER_TASK_STACK, maxNonUserTaskStackSize/1024 ));

    return maxNonUserTaskStackSize;
} // RTE_LimitedMmapSectionsTaskStackControl::CalculateMaxNonUserTaskStack()

// -----------------------------------------------------
// RTE_TaskStackControl::Initialize implementations
// -----------------------------------------------------

void RTE_LimitedMmapSectionsTaskStackControl::Initialize(
    SAPDB_Long numberOfTasks, 
    struct TASK_TYPE *pFirst, 
    struct TASK_TYPE *pLast)
{
    CalculateMaxTaskStack(numberOfTasks, pFirst, pLast);

    m_SingleGuardSize          = 0;
    m_SingleProtectedStackSize = m_SingleAlignedStackSize;
    m_TotalSize                = m_NumberOfTasks * m_SingleProtectedStackSize; 

    ReserveOrAllocateBigChunk(false); // false means do not try to reserve

    UpdateTaskControlRecords(pFirst, pLast);

    m_Mapping.m_Next = 0;
    m_Mapping.m_StartOfFirstTaskStack = m_StartOfFirstTaskStack;
    m_Mapping.m_EndOfLastTaskStack = m_StartOfFirstTaskStack + m_TotalSize;
    m_Mapping.m_SingleStackSize = m_SingleAlignedStackSize;
    m_Mapping.m_SingleGapSize = 0;
    m_Mapping.m_FirstId = 1;

    AddTaskStackMapping( m_Mapping );

    ConsumedMemoryInfo();

} // RTE_LimitedMmapSectionsTaskStackControl::Initialize()

// -------------------------------------------------------------

void RTE_LimitedAddressSpaceTaskStackControl::Initialize(
    SAPDB_Long numberOfTasks, 
    struct TASK_TYPE *pFirst, 
    struct TASK_TYPE *pLast)
{
    m_SingleAlignedNonUserStackSize = 
        RTE_ISystem::Instance().RoundUpToSystemPageSize( CalculateMaxNonUserTaskStack(pFirst, pLast) );
    CalculateMaxTaskStack(numberOfTasks, pFirst, pLast);

    m_SingleGuardSize                 = RTE_ISystem::Instance().GetSystemPageSize() * NUM_GUARDPAGES_EO70;
    m_SingleProtectedStackSize        = m_SingleAlignedStackSize        + m_SingleGuardSize;
    m_SingleProtectedNonUserStackSize = m_SingleAlignedNonUserStackSize + m_SingleGuardSize;
    // calculate max task stack for non user tasks and count the number of non user tasks in the system
    m_TotalSize = ( m_NumberOfNonUserTasks * m_SingleProtectedNonUserStackSize )
                + ( (m_NumberOfTasks - m_NumberOfNonUserTasks) * m_SingleProtectedStackSize );

    ReserveOrAllocateBigChunk(true); // true means try to reserve, which on LINUX will always fail...

    AllocateAndProtectTaskStacks( pFirst, pLast );

    m_StartOfFirstUserTaskStack = m_StartOfFirstTaskStack
                                + (m_NumberOfNonUserTasks * m_SingleProtectedNonUserStackSize);

    m_NonUserTaskMapping.m_Next = 0;
    m_NonUserTaskMapping.m_StartOfFirstTaskStack = m_StartOfFirstTaskStack;
    // protected size times non user tasks minus gap at end
    m_NonUserTaskMapping.m_EndOfLastTaskStack = m_StartOfFirstTaskStack 
                                    + ( m_SingleProtectedNonUserStackSize * m_NumberOfNonUserTasks )
                                    - m_SingleGuardSize;
    m_NonUserTaskMapping.m_SingleStackSize = m_SingleAlignedNonUserStackSize;
    m_NonUserTaskMapping.m_SingleGapSize = m_SingleGuardSize;
    m_NonUserTaskMapping.m_FirstId = 1;

    AddTaskStackMapping( m_NonUserTaskMapping );

    m_UserTaskMapping.m_Next = 0;
    m_UserTaskMapping.m_StartOfFirstTaskStack = m_StartOfFirstUserTaskStack;
    // protected size times user tasks minus gap at end
    m_UserTaskMapping.m_EndOfLastTaskStack = m_StartOfFirstUserTaskStack
                                    + ( m_SingleProtectedStackSize * (m_NumberOfTasks - m_NumberOfNonUserTasks) )
                                    - m_SingleGuardSize;
    m_UserTaskMapping.m_SingleStackSize = m_SingleAlignedStackSize;
    m_UserTaskMapping.m_SingleGapSize = m_SingleGuardSize;
    m_UserTaskMapping.m_FirstId = m_NumberOfNonUserTasks + 1;

    AddTaskStackMapping( m_UserTaskMapping );

    ConsumedMemoryInfo();

} // RTE_LimitedAddressSpaceTaskStackControl::Initialize

// ------------------------------------------------------

void RTE_BigAddressSpaceTaskStackControl::Initialize(
    SAPDB_Long numberOfTasks, 
    struct TASK_TYPE *pFirst, 
    struct TASK_TYPE *pLast)
{
    CalculateMaxTaskStack(numberOfTasks, pFirst, pLast);

    m_SingleGuardSize          = RTE_ISystem::Instance().GetSystemPageSize() * NUM_GUARDPAGES_EO70;
    m_SingleProtectedStackSize = m_SingleAlignedStackSize + m_SingleGuardSize;
    m_TotalSize                = m_SingleGuardSize        + ( m_NumberOfTasks * m_SingleProtectedStackSize ); 

    ReserveOrAllocateBigChunk(true); // true means try to reserve

    AllocateAndProtectTaskStacks( pFirst, pLast );

    m_Mapping.m_Next = 0;
    m_Mapping.m_StartOfFirstTaskStack = m_StartOfFirstTaskStack;
    // total size minus gap at front and gap at end
    m_Mapping.m_EndOfLastTaskStack = m_StartOfFirstTaskStack + m_TotalSize - ( 2 * m_SingleGuardSize );
    m_Mapping.m_SingleStackSize = m_SingleAlignedStackSize;
    m_Mapping.m_SingleGapSize = m_SingleGuardSize;
    m_Mapping.m_FirstId = 1;

    AddTaskStackMapping( m_Mapping );

    ConsumedMemoryInfo();

} // RTE_BigAddressSpaceTaskStackControl::Initialize

/*-------------------------------------------------------------*/
/*   U K T       Code                                          */
/*-------------------------------------------------------------*/

/*
   @brief Continue executing coroutine based UKT
   Separated to allow stack allocation via alloca() in calling subroutine

   @param this_ukt [in] the running ukt
   @return always 0
 */
extern "C" void *en88ExecuteCoroutineUKT(ten50_UKT_Control *this_ukt)
{
    // starting task of UKT has to complete initialization of UKT structure
    this_ukt->state      = KT_RUNNING;
    this_ukt->start_time = time ( (time_t *) 0 );
 
    // Initialize communication segment of this UKT
    e74_init_big_comseg (this_ukt);

    /*
     * If timer task was the initializing task, dispatcher is not needed
     */
    if ( KGS->ti->ukt == this_ukt )
    {
        this_ukt->curr_task = KGS->ti;
    }
    else
    {
        this_ukt->curr_task = (struct TASK_TYPE*) 0 ;
        en71Dispatcher (this_ukt);
    }

//    do {
        RTE_ContextSwitch( *(this_ukt->removeTaskContext),
                           *(this_ukt->curr_task->taskContext) );
        // finalize....
//        this_ukt->curr_task = (struct TASK_TYPE*) 0 ;
//        en71Dispatcher (this_ukt);
//    } while ( this_ukt->curr_task );
    return (void *)0;
}

//---------------------------------------------------------------

static struct TASK_TYPE *
en88_RelatedTaskStackInfo( struct TASK_TYPE *pTask,
                           SAPDB_ULong &singleStackSize,
                           RTE_TaskId  &firstTaskId,
                           SAPDB_ULong &numberOfTasksToAdd )
{
struct TASK_TYPE *pNextTask;
RTE_TaskId nextTaskId;

    singleStackSize = RTE_ISystem::Instance().RoundUpToSystemPageSize(pTask->stack_size);
    firstTaskId = pTask->index;
    nextTaskId = firstTaskId + 1;

    for ( pNextTask = pTask->next_task; pNextTask; pNextTask = pNextTask->next_task )
    {
        if ( pNextTask->index != nextTaskId
          || RTE_ISystem::Instance().RoundUpToSystemPageSize(pNextTask->stack_size) != singleStackSize )
        {
            break;
        }
        ++nextTaskId;
    }
    numberOfTasksToAdd = nextTaskId - firstTaskId;
    return pNextTask;
}

/*-------------------------------------------------------------
  Function: en88RunCoroutineUKT
  Description: Common entry point for all user kernel threads using coroutines

  This routine is non static only because C++ does not allow extern "C" functions to be static...

  en88_CallKernelTaskMain is the C++ exception catching code.

  Arguments: arg [in] the user kernel thread control block for this ukt

  Return value: always 0, since no exit code transferred
 */
extern "C" void *en88RunCoroutineUKT(void *arg)
{
#undef MF__
#define MF__ MOD__ "en88RunCoroutineUKT"
    ten50_UKT_Control *this_ukt = (ten50_UKT_Control *)arg;

    MSGD(( INFO_UNIX_UKT_STARTED , this_ukt->index )) ;

    this_ukt->removeTaskContext = RTE_ContextAllocate(RTEMem_RteAllocator::Instance());
    if ( !this_ukt->removeTaskContext )
    {
        vabort(WRITE_CORE);
    }

    RTE_ContextPrepare( *(this_ukt->removeTaskContext), 0, 0, 0);

    SAPDB_Int4 totalAlignedStackSize;
    struct TASK_TYPE *pTask;
    for ( totalAlignedStackSize = 0, pTask = this_ukt->pFirstTaskCtrl; 
          pTask;
          pTask = pTask->next_task )
    {
#ifdef DEBUGCOROUTINE
        printf("UKT%d T%d\n", this_ukt->index, pTask->index);
        fflush(stdout);
#endif
        totalAlignedStackSize += RTE_ISystem::Instance().RoundUpToSystemPageSize(pTask->stack_size);
    }

    /* as long as it need tasks on stack of pthread this will be true... */
    if ( XPARAM(fUseStackOnStack) )
    {
        SAPDB_Byte *UKTtaskStack = (SAPDB_Byte *)alloca(totalAlignedStackSize+RTE_ISystem::Instance().GetSystemPageSize());

        SAPDB_Byte *stackStartAddress = (SAPDB_Byte *)0 +
            RTE_ISystem::Instance().RoundUpToSystemPageSize(UKTtaskStack - (SAPDB_Byte *)0);

        SAPDB_Int   alignedStackSize;
        for ( pTask = this_ukt->pFirstTaskCtrl; pTask; pTask = pTask->next_task )
        {
            alignedStackSize = RTE_ISystem::Instance().RoundUpToSystemPageSize(pTask->stack_size);
            pTask->StackChunkStart = stackStartAddress;
            pTask->StackChunkEnd   = stackStartAddress + alignedStackSize;
            pTask->stack_size      = pTask->stack_size;

            pTask->GuardPageBegin  = 0;
            pTask->GuardPageEnd    = 0;
            stackStartAddress     += alignedStackSize;
        }

        SAPDB_Byte *taskStack = (SAPDB_Byte *)0 +
            RTE_ISystem::Instance().RoundUpToSystemPageSize(UKTtaskStack - (SAPDB_Byte *)0);

        for ( pTask = this_ukt->pFirstTaskCtrl; pTask; )
        {
            SAPDB_ULong singleStackSize;
            RTE_TaskId  firstTaskId;
            SAPDB_ULong numberOfTasksToAdd;

            struct TASK_TYPE *pNextTask = en88_RelatedTaskStackInfo(pTask, singleStackSize, firstTaskId, numberOfTasksToAdd);

            if ( !RTE_TaskStackControl::Instance().AddTaskStackMapping( taskStack
                                                                       ,singleStackSize
                                                                       ,firstTaskId
                                                                       ,numberOfTasksToAdd ) )
            {
                // XXX Message..
                vabort(WRITE_CORE);
            }
            taskStack += (singleStackSize * numberOfTasksToAdd);
            pTask = pNextTask;
        }

        for ( pTask = this_ukt->pFirstTaskCtrl; pTask; pTask = pTask->next_task )
        {
#ifdef DEBUGCOROUTINE
            printf("Preparing 0x%p for Task T%d\n", pTask->taskContext, pTask->index);
            fflush(stdout);
#endif
            RTE_ContextPrepare( *(pTask->taskContext),
                                pTask->StackChunkStart,
                                pTask->stack_size,
                                (void *)en88_CallCoroutineKernelTaskMain);
        }
        return en88ExecuteCoroutineUKT(this_ukt);

    }
    else
    {
        for ( pTask = this_ukt->pFirstTaskCtrl; pTask; pTask = pTask->next_task )
        {
#ifdef DEBUGCOROUTINE
            printf("Preparing 0x%p for Task T%d\n", pTask->taskContext, pTask->index);
            fflush(stdout);
#endif
            RTE_ContextPrepare( *(pTask->taskContext),
                                pTask->StackChunkStart,
                                pTask->stack_size,
                                (void *)en88_CallCoroutineKernelTaskMain);
        }
        return en88ExecuteCoroutineUKT(this_ukt);
    }
}

/*-------------------------------------------------------------
  Function: en88_RunPthreadTask
  Description: Common entry point for all task threads

  This routine is non static only because C++ does not allow extern "C" functions to be static...

  en88_CallKernelTaskMain is the C++ exception catching code.

  Arguments: arg [in] the task control block for this task

  Return value: always 0, since no exit code transferred
 */
extern "C" void *en88_RunPthreadTask(void *arg)
{
#undef MF__
#define MF__ MOD__ "en88_RunPthreadTask"
    struct  TASK_TYPE * tcb = (struct TASK_TYPE *)arg;

    DBGIN_T(tcb->index);

    tcb->taskThreadId = sqlgetthreadid();
 
    if ( KT_INITIALIZED == tcb->ukt->state  )
    {
        // starting task of UKT has to complete initialization of UKT structure
        en88_InitializePthreadTaskUKT(tcb);
    }

    en88_CallKernelTaskMain( tcb );

    DBGOUT_T(tcb->index);

    return (void *)0;
}

/*----------------------------------------------------*/
extern "C" void en88TaskContextSwitch( struct TASK_TYPE * pOldTask,
                                       struct TASK_TYPE * pNewTask )
{
   if ( XPARAM(fUseCoroutines) )
   {
       if ( !RTE_ContextSwitch( *(pOldTask->taskContext),
                                *(pNewTask->taskContext) ) )
       {
           MSGD(( 11900, ERR_TYPE, "TASKING ", "switching to next task failed" ));
           vabort(true);
       }
   }
   else
   {
      tsp00_ErrTextc   threadErrtext;
      teo07_ThreadErr  threadErr;
      teo07_Thread     oldThread = pOldTask->taskThread;

      sqlresumethread( pNewTask->taskThread, threadErrtext, &threadErr );
      if ( threadErr != THR_OK_EO07 )
      {
          MSGD(( 11900, ERR_TYPE, "TASKING ", "resuming next task failed" ));
          MSGD(( 11900, ERR_TYPE, "TASKING ", &threadErrtext[0] ));
          vabort(true);
      }

      sqlsuspendthread( oldThread, threadErrtext, &threadErr );
      if ( threadErr != THR_OK_EO07 )
      {
          MSGD(( 11901, ERR_TYPE, "TASKING ", "suspend task itself failed" ));
          MSGD(( 11901, ERR_TYPE, "TASKING ", &threadErrtext[0] ));
          vabort(true);
      }
   }
}

/*-------------------------------------------------------------
   Function: en88_InitializePthreadTaskUKT
   Description: Initialize an UKT 

   Called from first active task in UKT (state == KT_INITIALIZED)
   Arguments: firstTask [in] 
 */
static void en88_InitializePthreadTaskUKT( struct  TASK_TYPE * firstTask )
{
#undef MF__
#define MF__ MOD__"en88_InitializePthreadTaskUKT"

    ten50_UKT_Control *this_ukt = firstTask->ukt;

    this_ukt->state = KT_RUNNING;
    this_ukt->start_time = time ( (time_t *) 0 );

    MSGD(( INFO_UNIX_UKT_STARTED , this_ukt->index )) ;

    DBG1 (( MF__,"ukt     0x%lx", this_ukt ));
    DBG1 (( MF__,"semid   %d", this_ukt->semid ));

    /*
     * Calling thread is UKT thread
     */
    memcpy( &this_ukt->thread, &firstTask->taskThread, sizeof(teo07_Thread) );
    this_ukt->tid = firstTask->taskThreadId;

    // Initialize communication segment of this UKT
    e74_init_big_comseg (this_ukt);

    /*
     * If timer task was the initializing task, dispatcher is not needed
     */
    if ( KGS->ti == firstTask )
    {
        DBG1 (( MF__,"starting the timeout task T%d", this_ukt->curr_task->index ));
        this_ukt->curr_task = firstTask ;
    }
    else
    {
        DBG1 (( MF__,"entering dispatcher to find first task" ));
        this_ukt->curr_task = (struct TASK_TYPE*) 0 ;
        en71Dispatcher (this_ukt);
    }
}

/*---------------------------------------------------------------------------
  Function: en88_CallCoroutineKernelTaskMain
  Description: All UKT C++ main function

  Check for initial request. If not found call pascal main in endless loop.
  Will never return. If exception is caught, it is printed via RTE_Crash()

  Arguments; tcb [in] task control block of this task
 */
extern "C" void en88_CallCoroutineKernelTaskMain ()
{
    en88_CallKernelTaskMain(RTE_TaskStackControl::Instance().GetTaskControl());
}

/*---------------------------------------------------------------------------
  Function: en88_CallKernelTaskMain
  Description: All UKT C++ main function

  Check for initial request. If not found call pascal main in endless loop.
  Will never return. If exception is caught, it is printed via RTE_Crash()

  Arguments; tcb [in] task control block of this task
 */
static void en88_CallKernelTaskMain ( struct  TASK_TYPE * tcb )
{
#undef MF__
#define MF__ MOD__"en88_CallKernelTaskMain"
    try {
        ten50_UKT_Control *this_ukt = tcb->ukt;
#ifdef DEBUG_RTE 
char eBuffer[256];
SAPDB_sprintf(eBuffer, sizeof(eBuffer),"---> Task%d(%s) UKT%d (curr T%d) curr_quu 0x%p first_free 0x%p last_free 0x%p",
    tcb->index,
    e38_type_name(this_ukt->curr_task->type),
    this_ukt->index,
    this_ukt->curr_task->index,
    this_ukt->curr_quu,
    KGS->first_free,
    KGS->last_free);
MSGD(( INFO_TEST_WILDCARD, eBuffer ));
#endif

        if ( this_ukt->curr_quu )
        {
            DBG1((MF__, "T%d found curr_quu entry", this_ukt->curr_task->index ));
            if ( (this_ukt->curr_quu < KGS->first_free) 
              || (this_ukt->curr_quu > KGS->last_free) )
            {
                MSGALL (( ERR_WRONG_SEMAPHORE_IDX ));
                vabort ( WRITE_CORE );
            }
            if (    (this_ukt->curr_quu->req_type != REQ_INITOK)
                 && (this_ukt->curr_quu->req_type != REQ_CONNECT) )
            {
                MSGALL (( ERR_ILLEGAL_REQUEST, "initial", this_ukt->curr_quu->req_type ));
                vabort ( WRITE_CORE );
            }
            if ( this_ukt->curr_quu->req_type == REQ_INITOK )
            {
                e73_dl_enqu ( & KGS-> freelist, this_ukt->curr_quu );
                this_ukt->curr_quu = 0 ;
            }
        }

        while ( en88_NeverStop )
        {
            /* PTS 1110953 */
            if ( XPARAM(showMaxStackUse) )
            {
                vsReinit(this_ukt->curr_task->index,0);
            }

/*-----------------------------------------------*/
/*            Entering kernel zone               */
/*-----------------------------------------------*/
            MSGD(( INFO_UNIX_TASK_STARTED , tcb->index )) ;
            a91mainprogram ();

/*-----------------------------------------------*/
/*            Leaving kernel zone                */
/*-----------------------------------------------*/
            en51Finish();

        }
    }
    catch(SAPDBErr_MessageList &lastMsg)
    {
        RTE_Crash(lastMsg);
    }

}


