/*!***************************************************************************

  module      : geo43.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: ceo43_CmdPipe/x_cons show API
  description : class interface to access X_CONS show output


  last changed: 1999-01-19  10:13

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef GEO43_H
#define GEO43_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "geo47.h"
#include <stdio.h>
#include "gsp00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
 Class: ceo43_CmdPipe
 */
class ceo43_CmdPipe
{
private:
    FILE     * m_CommandPipe_eo43;
    tsp00_Int4 m_LastExitCode_eo43;
    TCHAR    * m_OpenedCommand_eo43;
public:
    /*!
     function: ceo43_CmdPipe
     description: Construct the object, but does NOT create the command pipe
     arguments:
        none
     return value:
        none
     */
    ceo43_CmdPipe();
    /*!
     function: ~ceo43_CmdPipe
     description: Cleans up command pipe if exists
     arguments:
        none
     return value:
        none
     */
    ~ceo43_CmdPipe();
    /*!
     function: eo43OpenCommand
     description: Create the output pipe and start the given command to fill it
     arguments:
        command [in] complete command line to execute
        openOk [out] Flag to indicate starting command succeeded
     return value:
        none
     */
    void eo43OpenCommand(const TCHAR *command, tsp00_Bool & openOk );
    /*!
     function: eo43CloseCommand
     description: Close the command output pipe and free internal buffer if any
     arguments:
        ExitCode [out] The exit code of the program executing the command
     return value:
        none
     */
    void eo43CloseCommand( tsp00_Int4 & ExitCode );
    /*!
     function: eo43NextLine
     description: Get the next line from the output pipe
     
                  This call blocks until data is available in the pipe.
                  The given buffer is filled. If buffer is too small for
                  the next input line, a flag is set.
                  The buffer returned is always zero terminated.
     arguments: 
        Buffer     [out] The buffer to be filled
        SizeInChars [in] Buffer size in number of characters
        NoNewline  [out] Flag to indicate that no newline at end of returned buffer content
        ExitCode   [out] Set to exit code of process, if 0 bytes are returned. Untouched otherwise
     return value:
        number of characters filled into buffer. 0 if end of pipe (ExitCode valid).
     */
    tsp00_Int4 eo43NextLine( TCHAR *Buffer, 
                             const tsp00_Int4 SizeInChars, 
                             bool & NoNewline, 
                             tsp00_Int4 & ExitCode );
};
/*!
 EndClass: ceo43_CmdPipe
 */

/*!
 Class: ceo43_xshow
 */
class ceo43_xshow: public ceo43_CmdPipe
{
private:
    TCHAR    * m_Command_eo43;
public:
    /*!
     function: ceo43_xshow
     description: x_cons show command output class derived from ceo43_CmdPipe.
     
                  The class simplifies the interface for ceo43_CmdPipe. Only the
                  database name and the specific show command are needed.

                  The usage is siplified too. Only the constructor is needed and
                  a loop over eo43NextLine() until 0 is returned:

                    ... sqlxinit(dbname, dbroot); // Mandatory for all sqlx... calls

                    ... ceo43_xshow xshow_cmd("all", openOk);

                    ... if ( !openOk ) command did not start

                    ... while ( xshow_cmd.eo43NextLine(Buffer, BufferSize, NoNewline, ExitCode) != 0 )
                    
                    ........ do something with buffered output of show all....

                    ... if ( ExitCode != 0 ) command reported problem...

     arguments:
        showCommand [in] The specific show command
        openOk     [out] Flag to indicate starting command succeeded
     return value:
        none
     */
    ceo43_xshow( const TCHAR *showCommand, tsp00_DbNamec DBName, tsp00_Bool & openOk );
};
/*!
 EndClass: ceo43_xshow
 */

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/


/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/



#endif  /* GEO43_H */
