/*!================================================================
 module:        heo06.h

 responsible:   DanielD
 
 special area:  RTE | Client | Virtual File
 
 description:   Handling of files and directories
 
 see:           http://p26326.bea.sap-ag.de/LZU/specs/VirtualFile.htm

 copyright:     SAP AG 1998
 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/


#ifndef HEO06_H
#define HEO06_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp01.h"
#include "gsp05.h"
#include "hsp77.h"

/*!
  Notes: Conventions

  description:
          Most routines are provided both with Pascal and C "bindings": 

          - Pascal routines end with 'p' 
          - C routines end with 'c' 
          - the length of a Pascal string is determined by searching for the last non blank character 
          - the length of a C string is determined by searching for the first zero byte 
          - calls to Pascal and C version can be intermixed, even on the same file handle

          All filenames may contain references to environment variables of the form

          - %varname% on NT
          - ${varname} on UNIX
 */

/*!
  Notes: Features no longer supported

  description:
          - An sqlfopen for write will overwrite an existing file. If this behaviour is not wanted, the user can check
                  the existence of a file using sqlfinfo 
          - no support for records 
          - no support for variable length records 
          - The RTE cannot detect the file type (binary, text) from the file contents

 */

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

/*!
  Chapter: Reading and Writing
 */

/*!
  Function: sqlfopen[pc]

  description: Opens a operating system file for reading or writing.

      As far as possible with the particular operating system, the
      runtime environment should be implemented to guarantee the the same
      file cannot be opened simultaneously by different users for writing
      purposes.

  arguments:
      filename     [in] The name of the file to be opened. If
          the file is opened for reading, a name of "-" or "stdin" will use
          standard input. If the file is opened for writing, a name of "-" or
          "stdout" will use standard output, "stderr" will use standard error.
          
      dataKind     [in]
          <UL>
          <LI>sp5vf_text: the file is interpreted as a sequence of text
           lines. End of line markers are stripped in read mode. In write mode,
           an end of line marker is inserted after every write operation.  *
          <LI>sp5vf_binary: the file is read and written without any
            interpretation.
          </UL>
          
      fileMode     [in]
          <UL>
          <LI>sp5vf_read: the file is read starting with the first byte
          <LI>sp5vf_write: the file is truncated (or created), writing starts at
            the first byte
          <LI>sp5vf_append: writing starts at the end of the file,
            the file is created if it does not exists.
          </UL>
           
      buffering    [in]
          <UL>
          <LI>sp5bk_buffered: 
          <LI>sp5bk_unbuffered: 
          <LI>sp5bk_disksynced: the runtime environment must guarantee that
            the data is flushed to disk when sqlfopen returns
          </UL>
            
      fileHandle   [out] if the file could be opened, fhandle contains a
          value to be used in subsequent calls to sqlfread, sqlfwrite,
          sqlfseek, sqlfclose.
          
      ferr         [out] error information

  return value: none

  prototypes: 2

 */

externC void
sqlfopenc (
    const char *rawFName,
    tsp5_rte_datakind dataKind,
    tsp5_rte_filemode fileMode,
    tsp5_rte_buffering_kind buffering,
    tsp_int4 *fileHandle,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfopenp (
    const tsp_vfilename rawFName,
    tsp5_rte_datakind dataKind,
    tsp5_rte_filemode fileMode,
    tsp5_rte_buffering_kind buffering,
    tsp_int4 *fileHandle,
    tsp5_rte_fileerror *ferr);
 

/*!
  Function: sqlfopenEncodedc

  description: Opens operating system file for reading or writing
        using a character encoding.

      Remarks for sqlfopoen[pc] apply.

  arguments:
      filename     [in] The name of the file to be opened. If
          the file is opened for reading, a name of "-" or "stdin" will use
          standard input. If the file is opened for writing, a name of "-" or
          "stdout" will use standard output, "stderr" will use standard error.

      fileEncoding  [in]

        - on read: the file is assumed to have this encoding. When
            <Param>fileEncoding</> is NULL, sqlfopenEncodedc will
            try to detect the encoding from the file content. Console
            input will be read according to the current locale.

        - on write: data is written using this encoding. Console output
            will be written according to the current locale.

      clientEncoding [in] buffers specified by the client are assumed to
        have this encoding

      fileMode     [in]
          <UL>
          <LI>sp5vf_read: the file is read starting with the first byte
          <LI>sp5vf_write: the file is truncated (or created), writing starts at
            the first byte
          <LI>sp5vf_append: writing starts at the end of the file,
            the file is created if it does not exists.
          </UL>

      buffering    [in]
          <UL>
          <LI>sp5bk_buffered:
          <LI>sp5bk_unbuffered:
          <LI>sp5bk_disksynced: the runtime environment must guarantee that
            the data is flushed to disk when sqlfopen returns
          </UL>

      fileHandle   [out] if the file could be opened, fhandle contains a
          value to be used in subsequent calls to sqlfread, sqlfwrite,
          sqlfseek, sqlfclose.

      ferr         [out] error information

 */
externC void
sqlfopenEncodedc (
    const char *filename,
    const tsp77encoding * fileEncoding,
    const tsp77encoding * clientEncoding,
    tsp5_rte_filemode fileMode,
    tsp5_rte_buffering_kind buffering,
    tsp00_Int4 *fileHandle,
    tsp5_rte_fileerror *ferr);

/*!
  Function: sqlfsaveopenc

  description: Opens a database save for reading or writing.

      The file is always opend binary and unbuffered. This is currently
          the only reliable way to work with tape files on NT.

      Although this routine can be used to open devspaces, handling of
          the magic record is not transparent. Seeking to
          position 0 and writing data will destroy the magic record. 

  arguments:
      filename     [in] The name of the file to be opened.
          
      fileMode     [in]
          <UL>
          <LI>sp5vf_read: the file is opened for reading, the "magic record" is skipped
          <LI>sp5vf_write: the file is opened for writing, a "magic record" is written
          <LI>sp5vf_append: the file is opened for reading and writing,
            the "magic record" is skipped 
          </UL>
          
      buffersize   [in/out] the assumed buffer size of the save. If the file is opened for
          reading and this parameter does not match the buffer size inside the
          "magic record", an error is returned and buffersize contains the size
          read from the "magic record" 
                           
      fileHandle   [out] if the file could be opened, fhandle contains a
          value to be used in subsequent calls to sqlfread, sqlfwrite,
          sqlfseek, sqlfclose.

      bufptr       [out] all reading/writing on the filehandle must be
          done through this buffer, as some environments have 
          
      ferr         [out] error information

  return value: none

 */

externC void
sqlfsaveopenc (
    const char *rawFName,
    tsp5_rte_filemode fileMode,
    tsp_int4* buffersize,
    tsp_int4 *fileHandle,
    void** bufptr,
    tsp5_rte_fileerror *ferr) ;

/*!
  Function: sqlfclose[pc]

  description: Closes the specified file.

     Any associated ressources (buffers) are released. The runtime
     environment is free to reuse the value of fhandle for
     a subsequent sqlfopen. 

  arguments:
      fileHandle [in] the handle to the file to be closed
      
      option     [in]
          <UL>
          <LI>sp5vf_close_normal: no special processing
          <LI>sp5vf_close_rewind: tape files are rewound
          <LI>sp5vf_close_next_tape: tape files are rewound,
            the tape is ejected and the next tape is loaded. If
            the loading of the next tape fails (or the device does
            not support autoloading),
            err.sp5fe_warning is set to sp5vfw_no_next_tape.
          </UL>
      err        [out] error information
      
  prototypes: 2

 */

externC void
sqlfclosec (
    tsp_int4 fileHandle,
    tsp5_rte_closeoption option,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfclosep (
    tsp_int4 fileHandle,
    tsp5_rte_closeoption option,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlfread[pc]

  description: Reads data from the specified file.

  When the file was opened as sp5vf_binary, 
     sqlfreadp and sqlfreadp read at most bufsize bytes info buf.
     Fewer bytes may be transferred if the end of
     the file is reached, but err.sp5fe_result will still be set to vf_ok.

  [sqlfopenc] When the file was opened as sp5vf_text, 
     sqlfreadc reads at most (bufsize - 1) characters into buf up to
     the next end of line marker and ends the data with a zero byte.
     The end of line marker is not transferred. If buf is to small to
     hold the line, then err.sp5fe_warning is set to sp5vwf_no_eol_found.
     The rest of the line can be read with a subsequent sqlfreadc.

  [sqlfopenp] When the file was opened as sp5vf_text, 
     sqlfreadp reads at most bufsize characters into buf up to the
     next end of line marker. The end of line
     marker is not transferred. If buf is to small to hold the line,
     then err.sp5fe_warning is set to sp5vwf_no_eol_found.
     The rest of the line can be read with a subsequent sqlfreadp.
     
  arguments:
     fhandle    [in]  the file to read from
     buf        [out] the data will be read into this buffer
     bufsize    [in]  the size of buffer
     outlen     [out] the number of bytes actually read
     err        [out] error information
          
  prototypes: 2

 */

externC void
sqlfreadc (
    tsp_int4 fileHandle,
    void* buf,
    tsp_longint bufSize,
    tsp_longint* outLen,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfreadp (
    tsp_int4 fileHandle,
    void* buf,
    tsp_longint bufSize,
    tsp_longint* outLen,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlfwrite[pc]

  description: Write data from the specified file.

  When the file has been opened as sp5vf_binary 
     sqlfwritec and sqlfwritep write inlen bytes from buf to the file 

  When the file has been opened as sp5vf_text 
     sqlfwritec and sqlfwritep write inlen bytes from buf to the file
     followed by an end of line marker. If inlen
     is less than zero, then inlen is calculated using the C library
     function strlen. 
     
  arguments:
     fhandle    [in]  the file to write to
     buf        [out] the data will be written from this buffer
     inlen      [in]  the data len
     err        [out] error information
      

  prototypes: 2

 */
externC void
sqlfwritec (
    tsp_int4 fileHandle,
    const void* buf,
    tsp_longint inLen,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfwritep (
    tsp_int4 fileHandle,
    const void* buf,
    tsp_longint inLen,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlfseek[pc]

  description: Moves the file pointer.

     A negative distance and reccount is allowed. 

     If the device is not capable of seeking, the runtime environment
         sets err.fe_result to vf_noseek. 

     There is no difference between the C (*c) and Pascal (*p) versions.

  arguments:
      fileHandle    [in]
      distance      [in]
      whence        [in]
          <UL>
          <LI>sp5vf_seek_begin: 
            sqlfseek positions the file pointer at byte distance.
            The first byte is byte "0".
          <LI>sp5vf_seek_cur: 
            sqlfseek advances the file pointer by distance bytes.
          <LI>sp5vf_seek_end: 
            sqlfseek positions the file pointer at byte distance
            relative to the end of the file.
          </UL>
      err           [out] error information
     
  prototypes: 2

 */

externC void
sqlfseekc (
    tsp_int4 fileHandle,
    tsp_longint distance,
    tsp5_rte_seekkind whence,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfseekp (
    tsp_int4 fileHandle,
    tsp_longint distance,
    tsp5_rte_seekkind whence,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlflongseek[pc]

  description: Moves the file pointer (suitable for 64bit file systems).

     A negative distance and reccount is allowed. 

     If the device is not capable of seeking, the runtime environment
         sets err.fe_result to vf_noseek. 

     There is no difference between the C (*c) and Pascal (*p) versions.

  arguments:
      fileHandle    [in]
      distance      [in]
      whence        [in]
          <UL>
          <LI>sp5vf_seek_begin :
            sqlflongseek positions the file pointer at byte
            (recsize * reccount). The first byte is byte "0".
          <LI>sp5vf_seek_cur:
            sqlflongseek advances the file pointer by
            (recsize * reccount) bytes.
          <LI>sp5vf_seek_end: 
            sqlflongseek positions the file
            pointer at byte (recsize * reccount) relative to the
            end of the file.
          </UL>
      err           [out] error information
     
  prototypes: 2

 */

externC void
sqlflongseekc (
    tsp_int4 fileHandle,
    tsp_longint distance,
    tsp_longint recordCount,
    tsp5_rte_seekkind whence,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlflongseekp (
    tsp_int4 fileHandle,
    tsp_longint distance,
    tsp_longint recordCount,
    tsp5_rte_seekkind whence,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlflongseek[pc]

  description: Gets the current file position.

      This position may be used in a subsequent call to sqlfseek.
      If the value cannot be represented by an tsp_longint,
      sp5vf_notok is returned in err. 

  arguments:
      fileHandle    [in]
      distance      [out] the current file position from the beginning
          of the file
      err           [out] error information
      
  prototypes: 2

 */

externC void
sqlftellc (
    tsp_int4 fileHandle,
    tsp_longint *pos,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlftellp (
    tsp_int4 fileHandle,
    tsp_longint *pos,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlftruncate[pc]

  description: Truncate the given file, but do not modify current write position.

      If the file does not exists, err.sp5fe_result is set to
      vf_notok. If the file exists, but cannot be truncated,
      err.sp5fe_result is set to vf_notok.

      If newSize is larger than current size, file is expanded, but
      the new space is left uninitialized. Otherwise it is truncated and additional
      space is freed.

      NOTE On NT after this call the current write position is reset to write position before
      this call ! This means the write position is not modified by this call! 
      
      Do not expect, that next write action is done at beginning of file or at the new end of
      the file! If you must ensure this, do so by an explicit sqlfseek[pc]() call in advance.
      Such behaviour is implemented identical for NT and UNIX ( it actually implements the
      SingleUNIX ftruncate() specification). 

      Again... sqlftruncate() does not modify current write position.

  arguments:
      fileHandle  [in]
      newSize     [in] New size of file
      err         [out] error information

  prototypes: 2

 */

externC void
sqlftruncatec (
    tsp_int4 fileHandle,
    tsp_longint newSize,
    tsp5_rte_fileerror *ferr);

externC void
sqlftruncatep (
    tsp_int4 fileHandle,
    tsp_longint newSize,
    tsp5_rte_fileerror *ferr);

/*!
  EndChapter: Reading and Writing
 */

/*!
  Chapter: File Functions
 */

/*!
  Function: sqlferase[pc]

  description: Erases the given file.

      If the file does not exists, err.sp5fe_result is set to
          vf_ok. If the file exists, but cannot be erased,
          err.sp5fe_result is set to vf_notok. 

  arguments:
      filename    [in]
      err         [out] error information
      
  prototypes: 2

 */

externC void
sqlferasec (
    const char *rawFName,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlferasep (
    const tsp_vfilename rawFName,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlfinfo[pc]

  description: Returns information about the given file.

     If the file does not exists,
         finfo.sp5fi_exists is set to false and err.sp5fe_result is set to vf_ok.
         The runtime environment may set err.sp5fe_result to vf_notok to
         signal other errors (e.g. device not accessible), but
         finfo.sp5fi_exists must be set to false.
         
     finfo.fi_size and finfo.fi_size_x are set such that the file size
         can be calculated as
         (finfo.fi_size_x * csp_maxlongint) + finfo.fi_size (to support 64 bit file systems).

  arguments:
      fname       [in]
      finfo       [out] a struct containing file information
      err         [out] error information
      
  prototypes: 2

 */

externC void
sqlfinfoc (
    const char *rawFName,
    tsp5_rte_fileinfo *finfo,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfinfop (
    const tsp_vfilename rawFName,
    tsp5_rte_fileinfo *finfo,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlfilecopyc

  description: Copies a file from Src to Dst.

  arguments:
      Src      [in]    the name of the source file
      Dst      [in]    the name of the destination file
      RteError [out]   error information

  returns: true if the file could be copied, false otherwise

  new: 28.07.98
  
 */

externC boolean
sqlfilecopyc( char           *Src,
              char           *Dst,
              tsp01_RteError *RteError );

#if !defined(_WIN32)
/*!
  Function: sqlmkfifo[pc]

  description: Create a named pipe (UNIX only)
  
      While NT uses a specific naming convention, named pipes in UNIX file systems
      are not identified by their name. So a special function call is needed to create
      a named pipe in an UNIX environment.
      
  arguments:
      rawFName     [in] The name of the named pipe to be created
          
      ferr         [out] error information

  return value: none

  prototypes: 2

 */

externC void
sqlmkfifoc (
    const char *rawFName,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlmkfifop (
    const tsp_vfilename rawFName,
    tsp5_rte_fileerror *ferr);
#endif /* !_WIN32 */ 

/*!
  EndChapter: File Functions
 */

/*!
  Chapter: Directory Functions
 */

/*!
  Function: sqlfopendirc

  description: Opens a directory for iterating over the filenames.

      The first filename is read.
      Wildcards in parameter dirname are not permitted. 

  arguments:
      dirname    [in]  the directory name
      dirHandleP [out] the resulting handle to be used in subsequent calls
          to sqlclosedirc
      firstname  [out] the name of the first file
      ferr       [out] error information

 */

externC void
sqlfopendirc (
    tsp_c256           *dirname,
    void              **dirHandleP,
    tsp_c256           *firstname,
    tsp5_rte_fileerror *ferr);

/*!
  Function: sqlfclosedirc

  description: Closes the given directory handle.

  arguments:
      dirHandle   [in]  the directory handle
      ferr        [out] error information

 */

externC void
sqlfclosedirc (
    void               *dirHandle,
    tsp5_rte_fileerror *ferr);

/*!
  Function: sqlfreaddirc

  description: Reads the next name from a directory

  arguments:
      dirHandle   [in]  the directory handle
      fname       [out] the resulting filename
      ferr        [out] error information

 */

externC void
sqlfreaddirc (
    void               *dirHandle,
    tsp_c256           *fname,
    tsp5_rte_fileerror *ferr);

/*!
  Function: sqlgetcwdc

  description: Retrieves the name of the current directory.

  arguments:
      dirbuf   [out]  the directory name is writtin into this buffer
      buflen   [in]   the size of dirbuf
      ferr     [out]  error information

  returns: true if the directory name could be copied, false otherwise.
  
  new: 28.07.98
  
 */

externC boolean
sqlgetcwdc (
    char           * dirbuf,
    int              buflen,
    tsp01_RteError * RteError);

/*!
  Function: sqlchdirc

  description: Changes the current directory.

  arguments:
      dirname  [out]  the directory name
      RteError [out]  error information

  returns: true if the directory name could be changed, false otherwise.
  
  new: 28.07.98
  
*/

externC boolean
sqlchdirc (
    const char     * dirname,
    tsp01_RteError * RteError);

/*!
  Function: sqljoindirc

  description: Concatenates two path names to a new pathname.

      When the second part is an absolute path, the first part is ignored.

  arguments:
      target     [out]  the resulting path name
      part1      [in]   the first path (e.g. directory)
      part2      [in]   the second path
      targetSize [in]  the size of target

  returns: a pointer to target if target contains the result, NULL otherwise.
  
  new: 28.07.98
  
*/

externC const char *
sqljoindirc (
    char           * target,
    const char     * part1,
    const char     * part2,
    unsigned int     targetSize);

/*!
  Function: sqlcreate_dirc

  description: Creates a new directory.

  arguments:
      PathToCreate  [in]  the directory name
      RteError      [out]  error information

  returns: true if the directory could be created, false otherwise.
  
  new: 28.07.98
  
 */

externC boolean
sqlcreate_dirc ( char           *PathToCreate,
                 tsp01_RteError *RteError ) ;

/*!
  Function: sqlremove_empty_dirc

  description: Deletes a empty directory.

  arguments:
      PathToRemove  [in]  the directory name
      RteError      [out]  error information

  returns: true if the directory could be deleted, false otherwise.
  
  new: 07.01.2000
  
 */
externC boolean
sqlremove_empty_dirc ( const char     *PathToRemove,
                       tsp01_RteError *RteError );

/*!
  Function: sqlspecialenum

  description: Creates a new enumeration to iterate over special devices.

  arguments:
      devKind         [in]
          <UL>
          <LI>sdk_tapes_sp05: tape devices
          <LI>sdk_drives_sp05: drives (suitable only on NT)
          </UL>

      longInfo        [in]  whether the enumeration shows only the
          item names (False) or additional information (True)

      additionalInfo  [in]  some enumerations require/allow additional infomation
          (e.g. a filename pattern for UNIX tapes)
      listHandle      [out] a handle to be used in subsequent calls to
          sqlspecialnext and sqlspecialclose
      header          [out] will contain the field names 
      ferr     [out]  error information

  new: 28.07.98
  
 */

externC void
sqlspecialenum (
    tsp05_SpecialdevKind        devKind,
    boolean                     longInfo,
    const char                * additionalInfo,
    tsp05_SpecialEnumeration ** listHandle,
    char                      * header,
    int                         headerSize,
    tsp5_rte_fileerror        * ferr);

/*!
  Function: sqlspecialnext

  description: Copies information about the next item into buf.

  arguments:
      listHandle   [in]  the handle to the enumeration
      buf          [out] the buffer containing the result
      bufSize      [in]  the size of buf
      ferr         [in]  error information
      
  new: 28.07.98
  
 */

externC void
sqlspecialnext (
    tsp05_SpecialEnumeration  * listHandle,
    char                      * buf,
    int                         bufSize,
    tsp5_rte_fileerror        * ferr);
   
/*!
  Function: sqlspecialclose

  description: Closes the enumeration handle.

  arguments:
      listHandle    [in] the handle to the enumeration
      
  new: 28.07.98
  
 */

externC void
sqlspecialclose (
    tsp05_SpecialEnumeration  * listHandle);

/*!
  EndChapter: Directory Functions
 */

/*!
  Chapter: Misc.
 */

/*!
  Function: sqlffinish

  description: Closes all open file handles.

 */

externC void
sqlffinishc (void);
 
externC void
sqlffinishp (void);

/*!
  Function: sqlfunlock[pc]

  description: Erases the given file.

      If the file does not exists, err.sp5fe_result is set to
          vf_ok. If the file exists, but cannot be erased,
          err.sp5fe_result is set to vf_notok. 

  arguments:
      filename    [in]
      err         [out] error information
      
  prototypes: 2

  obsolete: use <A HREF="#sqlferase">sqlferase</A> instead

 */

externC void
sqlfunlockc (
    const char *rawFName,
    tsp5_rte_fileerror *ferr);
 
externC void
sqlfunlockp (
    const tsp_vfilename rawFName,
    tsp5_rte_fileerror *ferr);
 
/*!
  Function: sqlgetpass

  description: Reads password from console input
  
      The password is prompted and read without echo. The result is NUL terminated C string
      stripped from Linefeed and CarriageReturn.
      If the password was read, err.sp5fe_result is set to vf_ok. If the password cannot be
      read err.sp5fe_result is set to vf_notok. 

  arguments:
      Prompt    [in]  Prompt to be written before password is read
      Password  [out] password read
      err       [out] error information
      
  returns: true if a password was read, false otherwise.
 */

externC tsp00_Bool
sqlgetpass(const char *                     Prompt, 
           tsp00_Namec        VAR_ARRAY_REF Password, 
           tsp5_rte_fileerror VAR_VALUE_REF err);

/*!
  EndChapter: Misc.
 */

/*!
  Chapter: Examples
 */

/*!
  Example: Read a text file

  description: sqlfopenc, sqlfreadc, sqlfclosec

      Show how a file is opened, read until an error or end
      of file occurs and closed

  code:
    include <stdio.h>
    include "heo06.h"

    int
    main (
        int argc,
        char * argv [])
    {
        tsp_int4            fhandle;
        char                buf [4096];
        tsp_longint         outLen;
        tsp_longint         outMax;
        tsp5_rte_fileerror  ferr;

        sqlfopenc (argv [1], SP5VF_TEXT, SP5VF_READ, SP5BK_BUFFERED,
                   &fhandle, &ferr);
        while (ferr.sp5fe_result == vf_ok) {
            sqlfreadc (fhandle, buf, sizeof (buf), &outLen, &ferr);
            if (ferr.sp5fe_result == vf_ok) {
                fprintf (stdout, "'%s' ", buf);
                if (ferr.sp5fe_warning == SP5VFW_NO_EOL_FOUND) {
                    fprintf (stdout, "(incomplete)");
                }
                fprintf (stdout, "\n");
            }
        }
        sqlfclosec (fhandle, SP5VF_CLOSE_NORMAL, &ferr);
        return 0;
    }

 */

/*!
  Example: Show all tapes

  description: sqlspecialenum, sqlspecialnext, sqlspecialclose

  code:
    include <stdio.h>
    include "heo06.h"
    
    int
    main ()
    {
        tsp5_rte_fileerror ferr;
        tsp05_SpecialEnumeration * enumHandle;
        
        sqlspecialenum (sdk_tapes_sp05, 1, NULL, &enumHandle, buf, sizeof (buf), &ferr);
        while (ferr.sp5fe_result == vf_ok) {
            fprintf (stdout, "%s", buf);
            sqlspecialnext (enumHandle, buf, sizeof (buf), &ferr);
        }
        sqlspecialclose (enumHandle);
        return 0;
    }
 */

/*!
  EndChapter: Examples
 */



#endif  /* HEO06_H */
