/*!***************************************************************************

  module      : heo07.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Multithreading Runtime API
  description : Routines for Creating and managing Threads,Mutex and 
                Semaphores.

                Routines are system independend. Including
                heo07.h or geo07.h does not force to include
                system specific header files.

  Original Author: Burkhard Diesing
  last changed: 1999-07-08  18:07
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO07_H
#define HEO07_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "geo07.h"

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

/*!
  Chapter: Thread
 */

/*!
  Function:     sqlinit_multithreading
  description:  This function initialize global data needed. 
    
    It must to called before
    any thread is created or any other function of this API is called.
  arguments:
    thread [out] pointer to thread structure for calling thread
    errtext [out] verbose error message
    err [out] see below

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
 */
void sqlinit_multithreading(teo07_Thread *thread,
                            tsp00_ErrTextc VAR_ARRAY_REF errtext,
                            teo07_ThreadErr *err);

/*!
  Function: sqlget_thread_min_stack
  
  description:
    Get the minimum thread stack size.
    
    If the given thread size for sqlbeginthread is smaller than this amount,
    it is silently increased (as the OS does too...).
  return value:
    minimum stack size in bytes
*/
tsp00_Int4 sqlget_thread_min_stack();

/*!
  Function: sqlbeginthread
  
  description:
    Creates a thread.
    
    NOTE: If the given thread is started SUSPENDED, it will suspend itself.
    In UNIX pthread implementation only the thread itself can suspend itself.
    So be carefull not to misuse SUSPENDING...
    Another thread should NOT allowed be suspended from outside.

    Replaces fork() function in a process environment, but SHARES all data.

  arguments:
        stack_size_in_bytes [in] stack size wanted for thread
        proc [in] thread main function 
        arg [in] argument passed to thread main function
        flags [in] flags specifying thread
        pThread [out] pointer storing returned thread
        errtext [out] error message if error occured
        err [out] error return value

  NOTE: errtext is set only if returned *err is != THR_OK_EO07

  return value:
    none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If linked with no thread library
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlbeginthread(tsp00_Int4 stack_size_in_bytes, 
                    teo07_ThreadCallback proc, 
                    void *arg, 
                    tsp00_Int4 flags, 
                    teo07_Thread *pThread,
                    tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                    teo07_ThreadErr *err);

/*!
  Function: sqlbeginthread_on_own_stack
  
  description:
    Creates a thread, but using thread stack from outside.
    
    NOTE: If the given thread is started SUSPENDED, it will suspend itself.
    In UNIX pthread implementation only the thread itself can suspend itself.
    So be carefull not to misuse SUSPENDING...
    Another thread should NOT allowed be suspended from outside.

    Replaces fork() function in a process environment, but SHARES all data.

    The stack start address is always given as the lowest address of the stack area even
    if the stack is unusually growing to higher addresses...
    Additional guardpages are not added by this call, so they have to be added outside
    if wanted. WIN32 does not support own thread stack, so that parameter is ignored.

  arguments:
        stack_size_in_bytes [in] stack size wanted for thread
        stack_start_address [in] stack start address
        proc [in] thread main function 
        arg [in] argument passed to thread main function
        flags [in] flags specifying thread
        pThread [out] pointer storing returned thread
        errtext [out] error message if error occured
        err [out] error return value

  NOTE: errtext is set only if returned *err is != THR_OK_EO07

  return value:
    none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If linked with no thread library
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlbeginthread_on_own_stack(tsp00_Int4 stack_size_in_bytes, 
                     void *stack_start_address,
                     teo07_ThreadCallback proc, 
                     void *arg, 
                     tsp00_Int4 flags, 
                     teo07_Thread *pThread,
                     tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                     teo07_ThreadErr *err);

/*!
  Function:     sqlsetthreadpriority
  description:  Modify the thread priority
    
  arguments:
      thread   [in] Thread to modify
      priority [in] Wanted priority

                    Windows NT:
                    
                    THREAD_PRIORITY_ABOVE_NORMAL 
                    Indicates 1 point above normal priority
                    
                    THREAD_PRIORITY_BELOW_NORMAL 
                    Indicates 1 point below normal priority
                    
                    THREAD_PRIORITY_HIGHEST 
                    Indicates 2 points above normal priority
                    
                    THREAD_PRIORITY_IDLE 
                    Indicates a base priority level of 1
                    
                    THREAD_PRIORITY_LOWEST 
                    Indicates 2 points below normal priority
                    
                    THREAD_PRIORITY_NORMAL 
                    Indicates normal priority for the priority class

                    THREAD_PRIORITY_TIME_CRITICAL 
                    Indicates a base priority level of 15
                    
                    UNIX/PTHREADS:
                    
                    absolute value
                    1..127 (1 least favoured, 127 most favoured)

                    UNIX/CTHREADS:

                    absolute value
                    0..127 (0 least favoured, 127 most favoured)

  return value:
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If priority cannot be set
        <LI>THR_NOT_OK_EO07: 
    </UL>
 */
teo07_ThreadErr sqlsetthreadpriority( teo07_Thread thread, 
                                      tsp00_Int4 priority );

/*!
  Function:     sqlsetmythreadpriority
  description:  Modify the current thread priority
    
  arguments:
      thread   [in] Thread to modify
      priority [in] Wanted priority

                    Windows NT:
                    
                    THREAD_PRIORITY_ABOVE_NORMAL 
                    Indicates 1 point above normal priority
                    
                    THREAD_PRIORITY_BELOW_NORMAL 
                    Indicates 1 point below normal priority
                    
                    THREAD_PRIORITY_HIGHEST 
                    Indicates 2 points above normal priority
                    
                    THREAD_PRIORITY_IDLE 
                    Indicates a base priority level of 1
                    
                    THREAD_PRIORITY_LOWEST 
                    Indicates 2 points below normal priority
                    
                    THREAD_PRIORITY_NORMAL 
                    Indicates normal priority for the priority class

                    THREAD_PRIORITY_TIME_CRITICAL 
                    Indicates a base priority level of 15
                    
                    UNIX/PTHREADS:
                    
                    absolute value
                    1..127 (1 least favoured, 127 most favoured)

                    UNIX/CTHREADS:

                    absolute value
                    0..127 (0 least favoured, 127 most favoured)

  return value:
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If priority cannot be set
        <LI>THR_NOT_OK_EO07: 
    </UL>
 */
teo07_ThreadErr sqlsetmythreadpriority( tsp00_Int4 priority );

/*!
  Function:     sqlgetthreadpriority
  description:  Return the thread priority
    
  arguments:
      thread   [in] Thread 
      pPriority [out] Priority return value (see sqlsetthreadpriority for values)
  return value:
    
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If priority cannot be set
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
 */
teo07_ThreadErr sqlgetthreadpriority( teo07_Thread thread, 
                                      tsp00_Int4 VAR_VALUE_REF pPriority );

/*!
  Function:     sqlgetmythreadpriority
  description:  Return the current thread priority
    
  arguments:
      pPriority [out] Priority return value (see sqlsetthreadpriority for values)
  return value:
    
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If priority cannot be set
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
 */
teo07_ThreadErr sqlgetmythreadpriority( tsp00_Int4 VAR_VALUE_REF pPriority );

/*!
  Function:     sqlendthread
  description:  Allows a thread to terminate itself. 
    
    Equivalent to exit() function in a process environment.
  arguments:    returncode [in] Returncode passed to sqljointhread
  return value: none
 */
void sqlendthread(tsp00_Int4 returncode);

/*!
  Function:     sqlkillthread
  description:  Allows to send a SIGTERM signal to a specific thread.

    Replaces kill(SIGTERM) function. In WINDOWS environment this kill
    terminates the thread.
  arguments:
    thread [in] Pointer to the thread structure
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
 */
void sqlkillthread(teo07_Thread thread, 
                   tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                   teo07_ThreadErr *err);

/*!
  Function:     sqlsigthread
  description:  Allows to send a given signal to a specific thread.

    Added for KERNEL interface. Replaces kill() function.
  arguments:
    thread [in] Pointer to the thread structure
    signal [in] The signal to be send
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
 */
void sqlsigthread( teo07_Thread thread, 
                   tsp00_Int4 signal,
                   tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                   teo07_ThreadErr *err);

/*!
  Function:    sqlcancelthread
  description: Allows to cancel a specific thread. 
    
    Canceling has the
    same effect as an sqlendthread call of the cancelled thread.
    NOTE: Cancelling effectivly happens if the thread reaches a
    cancelpoint, which is normally a call to sqlyield() or a system
    call.
    Added for KERNEL interface.
  arguments:
    thread [in] Pointer to the thread structure
    returncode [in] Returncode received by sqljointhread call (Used in WIN32 only)
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned

   NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlcancelthread(teo07_Thread thread, 
                     tsp00_Int4 returncode,
                     tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                     teo07_ThreadErr *err);

/*!
  Function:     sqlforgetthread
  description:  Detaches the given thread. 

    No sqljointhread is needed to reclaim storage allocated for the thread.
    NOTE: After this call use of 'thread' is a fatal mistake...
  arguments:    thread [in] Pointer to the thread structure
  return value: none
*/
void sqlforgetthread (teo07_Thread thread);

/*!
  Function:    sqlyieldthread
  description: Allows a thread to reschedule itself. 
    
    Needed if explicit blocking (i.e. via sqllockmutex) would lead to block all
    messages (i.e. window messages would not arrive at client,
    so no update or repaint is possible...).
  arguments:    none
  return value: none
*/
void sqlyieldthread(void);

/*!
  Function:     sqlgetthread
  description:  Returns the current thread structure, usable in subsequent calls
  arguments:    none
  return value: The current thread of the calling procedure
*/
teo07_Thread   sqlgetthread();

/*!
  Function:     sqlgetthreadid
  description:  Returned TID uniquely identifies this thread within process.
  arguments:    none
  return value: The unique thread id of the current thread
*/
teo07_ThreadId sqlgetthreadid();

/*!
  Function:     sqlthreadid
  description:  Returned TID uniquely identifies the given thread within process.

    Added for KERNEL interface.
  arguments:    thread [in] Pointer to the thread structure
  return value: The unique thread id of the given thread or 0 if no such thread.
*/
teo07_ThreadId sqlthreadid(teo07_Thread thread);

/*!
  Function:    sqlsuspendthread
  description: Allows a thread to suspend itself. 

    NOTE: Only the thread itself can suspend itself.
    In UNIX environment (POSIX Threads) threads are 
    NOT allowed to suspend other threads but themself.
  arguments:   thread  [in]  Pointer to the thread structure
               errtext [out] Place for error message to be returned
               err     [out] Place for error code to be returned

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: If thread cannot suspend
    </UL>
*/
void sqlsuspendthread(teo07_Thread thread, 
                      tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                      teo07_ThreadErr *err);

/*!
  Function:     sqlresumethread
  description:  Allows to resume a formerly suspended thread.

    NOTE: If the given thread is not in suspended state
    resume does not have any effect, but a subsequent
    sqlsuspendthread will return THR_NOT_OK_EO07.
  arguments:
    thread [in] Pointer to the thread structure
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlresumethread(teo07_Thread thread, 
                     tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                     teo07_ThreadErr *err);

/*!
  Function:    sqlIsThreadSuspended
  description: Check if given thread has suspended itself

    For 'thread grouping' there is a little time period where two
	threads are running. The one that gave up its slice has to wake
	up and suspend itself afterwards. The second thread can use this
	call, to check that the first has entered its suspend state.

    If alone, it is i.e. possible to protect its task stack etc...
  
	NT: No API function that allows this...

  arguments:
    thread   [in] Pointer to the thread structure

  return value:
    <UL>
        <LI>true: Thread has suspended itself
		<LI>false: Thread has not suspended itself
    </UL>
*/
tsp00_Bool sqlIsThreadSuspended(teo07_Thread thread);

/*!
  Function:    sqljointhread
  description: Allows to join a formerly created thread.

    Joining means
    that it will wait until the given thread exits, or is 
    terminated. Joining includes reclaiming all memory that
    was allocated for the thread.

    NOTE: If the given thread is detached, joining cannot
    be done.
  arguments:
    thread   [in] Pointer to the thread structure
    status  [out] Exit status of the thread
    errtext [out] Place for error message to be returned
    err     [out] Place for error code to be returned

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NO_DATA_FOUND_EO07: If thread was not found
        <LI>THR_NOT_OK_EO07: If thread cannot be joined
    </UL>
*/
void sqljointhread(teo07_Thread thread, 
                   tsp00_Int4 *status, 
                   tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                   teo07_ThreadErr *err);

/*!
  Function:      sqlthreadalive
  description:   Allows to check whether a thread is still running.
  arguments:     thread   [in] Pointer to the thread structure
                 errtext [out] Place for error message to be returned
                 err     [out] Place for error code to be returned

  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: TRUE if thread is alive
                FALSE if thread is dead
*/
int sqlthreadalive( teo07_Thread thread, 
                     tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                     teo07_ThreadErr *err);

/*!
  Function:     sqlonce
  description:  Allows to execute a function exactly once. 
    
    The first call
    for a given initialized flag will execute the callback
    with the given argument. All subsequent calls will not
    get executed.
  arguments:    thread [in] Pointer to the thread structure
                callback [in] The function to execute
                arg [out] The argument for that function
  return value: none
*/
void sqlonce (tsp00_Int4* initialized,
              teo07_ThreadCallback callback,
              void* arg);

/*!
  EndChapter: Thread
 */

/*!
  Chapter: Mutex
 */
/*!  
  Function:     sqlcreatemutex
  description:  Creates a mutual exlusive (mutex) structure.

    Such a structure is used to protect a code region from being entered
    by several threads at once.

    NOTE: Must be destroyed by a subsequent call to 
    sqldestroymutex    if no longer needed.
  arguments:    hMutex [in] The mutex to destroy
  return value: none
*/
void sqlcreatemutex(teo07_Mutex *hMutex);
/*!
  Function:     sqldestroymutex
  description:  Destroyes a mutex structure.
  arguments:    hMutex [in] The mutex to destroy
  return value: none
*/
void sqldestroymutex(teo07_Mutex *hMutex);

/*!  
  Function: sqlbeginmutex
  description:  Blocking version of entering a critical section.

    NOTE: Some systems do not allow messages to be passed
    while waiting to enter a critical section (NT/Windows95).
    
  arguments:    hMutex [in] The mutex
  return value: none
*/
void sqlbeginmutex(teo07_Mutex *hMutex);

/*!  
  Function:     sqlissetmutex
  description:  Test whether critical section is entered
  arguments:    hMutex [in] The mutex
  return value: none
*/
int sqlissetmutex(teo07_Mutex *hMutex);

/*!  
  Function:     sqltrybeginmutex
  description:  Non blocking version of entering a critical section.

    Used in combination with sqlyieldthread() to allow
    messages to be passed while waiting to enter a critical
    section.

    A special macro SYNCHRONIZED_EO07 can be used to form a block
    of statements protected by a pair of sqlbeginmutex()/sqlendmutex()
    calls.
  arguments:    hMutex [in] The mutex
  return value: 1 (true) on success, 0 (false) on error
*/
tsp00_Int4 sqltrybeginmutex(teo07_Mutex *hMutex);
/*!  
  Function:     sqlendmutex
  description:  Leave a critical section
  arguments:    hMutex [in] The mutex
  return value: none
*/
void sqlendmutex(teo07_Mutex *hMutex);
/*!
  EndChapter: Mutex
 */

/*!
  Chapter: Local Storage
 */
/*!  
  Function:     sqlcreatetls
  description:  Allocates a Thread Local Storage key.

    NOTE: Must be destroyed by a subsequent call to 
    sqldestroytls if no longer needed.
  arguments:
    hKey [in] The Key to be allocated
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned
  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_IMPLEMENTED_EO07: If linked with no thread library
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlcreatetls(tsp00_Int4 *hKey, 
                  tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                  teo07_ThreadErr *err);
/*!  
  Function:     sqldestroytls
  description:  Frees a Thread Local Storage key.
  arguments:
    hKey [in] The Key to be deallocated
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned
  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqldestroytls(tsp00_Int4 hKey, 
                   tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                   teo07_ThreadErr *err);
/*!  
  Function:      sqlsettls
  description:   Set the value for a given Thread Local Storage key.

    Each thread will see always only its on TLS Key value.
  arguments:
    hKey [in] The Key to be allocated
    pObjPtr [in] The pointer to store.
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned
  NOTE: errtext is set only if returned *err is != THR_OK_EO07
  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlsettls(tsp00_Int4 hKey, 
               void*  pObjPtr, 
               tsp00_ErrTextc VAR_ARRAY_REF errtext, 
               teo07_ThreadErr *err);
/*!  
  Function:     sqlgettls
  description:  Get the value for a given Thread Local Storage key.

    Each thread will see always only its on TLS Key value.
  arguments: hKey [in] The Key to be allocated

  NOTE: Since NULL can be a valid value, no error check can be done...
  return value: the value set by a previous sqlsettls call or NULL.
*/
void *sqlgettls(tsp00_Int4 hKey);
/*!
  EndChapter: Local Storage
 */

/*!
  Chapter: Semaphore
 */

/*!  
  Function: sqlcreatesem
  description: Create a thread global semaphore.

    NOTE: These semaphore are NOT used to synchronize between processes. 
    Only usefull for interthread synchronization. You cannot use them 
    i.e. for shared memory synchronization between processes!

  arguments:
    sem [out] The pointer to the semaphore to create
    initval [in] Initial value
    errtext [out] Place for error message to be returned
    err [out] Place for error code to be returned
  NOTE: errtext is set only if returned *err is != THR_OK_EO07

  return value: none, but err is updated to
    <UL>
        <LI>THR_OK_EO07: On success
        <LI>THR_NOT_OK_EO07: See errtext for details
    </UL>
*/
void sqlcreatesem (teo07_ThreadSemaphore* sem,
                   tsp00_Int4 initval,
                   tsp00_ErrTextc VAR_ARRAY_REF errtext, 
                   teo07_ThreadErr* err);
/*!  
  Function:     sqldestroysem
  description:  Delete a semaphore if no longer used
  arguments:    sem [in] The semaphore to destroy
  return value: none
*/
void sqldestroysem (teo07_ThreadSemaphore sem);

/*!  
  Function:     sqlwaitsem
  description:  Blocking wait on semaphore.

    Decreasing counter by 1. If already 0 then wait 
    blocking on the semaphore to get signalled.
  arguments:    sem [in] The semaphore to wait for 
  return value: none
*/
void sqlwaitsem (teo07_ThreadSemaphore sem);
/*!  
  Function:     sqltimedwaitsem
  description:  Nonblocking wait on semaphore.
    
    Decreasing counter by 1. If already 0 then wait on 
    the semaphore to get signalled within a specified
    timeout.
  arguments:
    sem [in] The semaphore to wait for 
    seconds [in] The maximum number of seconds to wait
    err [out] Set to THR_TIMEOUT_EO07 or THR_OK_EO07 respectivly.
  return value: none, but err is modified according to result.
*/
void sqltimedwaitsem (teo07_ThreadSemaphore sem, 
                      tsp00_Int4 seconds, 
                      teo07_ThreadErr *err);
/*!  
  Function:     sqlsignalsem
  description:  Signals the given semaphore, inceasing counter by 1
  arguments:    sem [in] The semaphore to trigger
  return value: none
*/
void sqlsignalsem (teo07_ThreadSemaphore sem);
/*!
  EndChapter: Semaphore
 */

/* _m is of type (teo07_Mutex *) */
#define SYNCHRONIZED_EO07(_m) for ( sqlbeginmutex(_m); \
                                    sqlissetmutex(_m); \
                                    sqlendmutex(_m) )

#if defined(__cplusplus)
} /* extern "C" */

class Synchronized_eo07
{
private:
    teo07_Mutex *m_mutex;
public:
    Synchronized_eo07(teo07_Mutex *m):m_mutex(m) { sqlbeginmutex(m); }
    ~Synchronized_eo07()                         { sqlendmutex(m_mutex); }
};

#endif /* __cplusplus */

#endif  /* HEO07_H */
