/*!***************************************************************************

  module      : heo61.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Simple Kernel TRACE
  description : TRACE stack functions for kernel tasks

  last changed: 1998-11-11  16:47

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO61_H
#define HEO61_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "geo47.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

 /*!
  Chapter: TrackBuffer
 
  description: The TrackBuffer implements a simple Tracking Mechanism
  
      This module implements the track buffer. It can be used using wrapper 
	  functions or directly by the class Interface or a mixture of both.
	  The implementation uses a double linked list of class TrackBufferEntry 
	  elements. The TrackBufferEntry elements are handled by the LZU only.

	  The basic idea is not to spent too much brain in not consuming time
	  for copying messages.

	  1. NO OUTPUT is done (must be done from outside)

			Output routine is wanted only in case of error traceback needed.
			So there is no need to know how output is done in this API.

	  2. MEMORY is allocated within class

			NOTE: By calling vtrack_insert() the complete message is
			copied, regardless of its contents. But for convienience of the
			error output routine, all inserted lines should not be terminated
			by newlines...
			The amount of memory allocated is set from outside in the call to
			vtrack_alloc(size) in Bytes. This size includes all
			TrackBuffer internal overhead.

	  3. NO LIMIT on number of elements to push

			If element does not fit, the first elements are overwritten.

	  4. NO CHECKS on pointer validity are performed

			No Message passing is done on the entries.
 */

/*!
  EndChapter: TrackBuffer
 */

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
  Chapter: TrackBuffer API Functions
 */
/*!
  Function: vtrack_alloc
  
  description: Allocate a TrackBuffer with a given size
  
	This function is used to allocate a track buffer for usage inside
	a specific task.

  arguments:
	size [in] Number of bytes to allocate for track buffer

  return value: 
	FALSE if failed
	TRUE  if succeeded
 */
externC tsp00_Bool vtrack_alloc ( tsp00_Int4 size );

/*!
  Function: vtrack_free
  
  description: Free a previously allocated TrackBuffer
  
	This function is used to free a track buffer allocated for the calling
	task. If no track buffer was allocated, nothing is done.

  arguments:
	none

  return value: 
	none
 */
externC void vtrack_free( );

/*!
  Function: vtrack_insert
  
  description: Insert an element in the TrackBuffer
  
	This function is used to copy the message in the TrackBuffer and
	doing the pointer updates. If no more space is available a special
	overflow element is added, that contiains an overflow counter...

  arguments:
	vtrace [in] Pointer to TrackBuffer content to add
	len    [in] Size of content in bytes

  return value: 
    none
 */
externC void vtrack_insert ( void *vtrace, tsp00_Int4 len );

/*!
  Function: vtrack_remove_last
  
  description: RemoveLast element in the TrackBuffer
  
	This function is used to remove the last element in the track buffer.
	If last element is an overflow element, its counter is decreased.

  arguments:
	none

  return value: 
	none
 */
externC void vtrack_remove_last ( );

/*!
  Function: vtrack_empty
  
  description: Clear all entries in TrackBuffer
  
	This function is simple removing the reference to the first element.
	After this call the TrackBuffer is empty.

  arguments:
	none

  return value: 
	none
 */
externC void vtrack_empty( );

/*!
  Function: vtrack_first
  
  description: Return the first element in the TrackBuffer
  
	This function is used to return a pointer to the first internally stored 
	message to the caller. The first element is the first element ever stored
	or the first element stored after the last vtrack_Empty() call.

  arguments:
	pMsg [out] Pointer to pointer to message
	pLen [out] Pointer to size of message line

  return value: 
	FALSE if no elements
	TRUE  if valid first element returned
 */
externC tsp00_Bool vtrack_first ( void **pVtrace, tsp00_Int4 *pLen );

/*!
  Function: vtrack_next
  
  description: Return the next element in the TrackBuffer
  
	This function is used to return a pointer to the internally stored 
	message to the caller. If no more message is stored, FALSE is returned
	and the values are set to NULL and 0 respectivly.

	The internal state of the track buffer changes, so subsequent calls to
	this function return followon messages.

  arguments:
	pVtrace [out] Pointer to pointer to message
	pLen [out] Pointer to size of message line

  return value: 
	FALSE if no more elements
	TRUE  if valid element returned
 */
externC tsp00_Bool vtrack_next ( void **pVtrace, tsp00_Int4 *pLen );

/*!
  Function: vtrack_check
  
  description: Return the counter of the overflow element if overflow
    occured. 
  
	This function is used to return the number of overflows stored in the 
	TrackBuffer. A overflow occures if no more space is in the track buffer
	for a given message. A overflow can be removed by vtrack_remove_last
	like a normal message. A overflow count is maintained to keep track of the
	number of overflows that occured. This way only the message gets lost, but
	the number of messages stored is always correct.

  arguments:
	none

  return value: 
	number of overflows occured
 */
externC tsp00_Int4 vtrack_check ( );

/*!
  Function: vtrack_dump_first
  
  description: Return the first element of the first task that has a TrackBuffer
  
	This function is used to return a pointer to the first internally stored 
	message of the first task to the caller. This call works like vtrack_first
	but does not necessary return an element stored by the caller.

  arguments:
	pMsg [out] Pointer to pointer to message
	pLen [out] Pointer to size of message line
	pId  [out] Pointer to TaskId

  return value: 
	FALSE if no valid task/element returned
	TRUE  if valid first element or first task returned
 */
externC tsp00_Bool vtrack_dump_first ( void **pVtrace, tsp00_Int4 *pLen, tsp00_TaskId *pId );

/*!
  Function: vtrack_dump_next
  
  description: Return the next element in TrackBuffer of given Task, skipping to next task if necessary
  
	This function is used to return a pointer to the next internally stored 
	message of the first task to the caller. This call works like vtrack_next
	but does not stop before all task track buffers are dumped.

  arguments:
	pMsg [out] Pointer to pointer to message
	pLen [out] Pointer to size of message line
	pId  [inout] Pointer to TaskId (updated on success)

  return value: 
	FALSE if no valid task/element returned
	TRUE  if valid first element or first task returned
 */
externC tsp00_Bool vtrack_dump_next ( void **pVtrace, tsp00_Int4 *pLen, tsp00_TaskId *pId );

/*!
  Function: vtrack_dump_check
  
  description: Return the count of overflows (if any) occured in task specific TrackBuffer. 
  
	This function is used to return the number of overflows stored in the 
	TrackBuffer for a specific task. A overflow occures if no more space is in 
	the track buffer for a given message. A overflow can be removed by 
	vtrack_remove_last like a normal message. A overflow count is maintained 
	to keep track of the number of overflows that occured. This way only the message 
	gets lost, but the number of messages stored is always correct.

  arguments:
	id [in] The Task Id for which TrackBuffer is tested

  return value: 
	number of overflows occured for task specific TrackBuffer
 */
externC tsp00_Int4 vtrack_dump_check ( tsp00_TaskId id );

/*!
  EndChapter: TrackBuffer API Functions
 */

/*!
  Chapter: Examples
 */

/*!
  Example: TrackBuffer usage example

  description:
      vtrack_alloc vtrack_free vtrack_insert vtrack_remove_last

  code:

#define MY_MESSAGE(_s) (_s),(strlen(_s)+1)
		int
	task_main()
	{
		vtrack_alloc(1000);

		vtrack_insert(MY_MESSAGE("main calling sub1"));
		sub1();
		vtrack_empty();
		vtrack_insert(MY_MESSAGE("main calling sub2"));
		sub2();

		vtrack_free();
		return 0;
	}

	void
	sub1()
	{
		{ // First interesting section
			vtrack_insert(MY_MESSAGE("sub1 calling sub2 first time"));
			sub2();
			vtrack_remove_last( );
		}
		{ // Second interesting section
			vtrack_insert(MY_MESSAGE("sub1 calling sub2 second time"));
			sub2();
			vtrack_remove_last( );
		}
		return;
	}

	void
	sub2() // Passing of to interesting points...
	{
		vtrack_insert(MY_MESSAGE("sub2 successfully called"));
		error_call();
		vtrack_remove_last( );
		return;
	}

	void
	error_call()
	{
	tsp00_Bool ok;
	tsp00_Int4 cnt;
	void *msg;
	tsp00_Int4 len;

		cout << _T("Error BIG PROBLEM... Traceback follows:") << endl;
		for ( ok = vtrack_first( &msg, &len ); 
			  ok != FALSE; 
			  ok = vtrack_next( &msg, &len ) )
		{
			cout << (char *)msg << endl;
		}
		cnt = vtrack_check( );
		if ( cnt > 0 )
		{
			cout << "Sorry OVERFLOW in TrackBuffer. Last " 
				 << cnt 
				 << " messages not stored"
				 << endl;
		}
	}
	
	Output

		Error Traceback...
		main
		sub1_1
		sub2_1
		sub2_2
		Error Traceback...
		main
		sub1_2
		sub2_1
		sub2_2		
 */

/*!
  Example: Output of the TrackBuffer in order

  description: vtrack_first,vtrack_next
  
			Shows how a track buffer content is dumped in order.
			NOTE: Content of TrackBuffer is opaque. For this example
			strings have been inserted...

  code:
		To do it in order use a simple loop:

		tsp00_Bool ok;
		void *msg;
		tsp00_Int4 len;
		tsp00_Int4 cnt;

			cout << _T("Error BIG PROBLEM... Traceback follows:") << endl;
			for ( ok = vtrack_first( &msg, &len ); 
				  ok != FALSE; 
				  ok = vtrack_next( &msg, &len ) )
			{
			    cout << (char *)msg << endl;
			}

			cnt = vtrack_check( );
			if ( cnt > 0 )
			{
				cout << "Sorry OVERFLOW in TrackBuffer. Last " 
					 << cnt 
					 << " messages not stored"
					 << endl;
			}
 */

/*!
  Example: Output of all TrackBuffers of all Tasks in order

  description: vtrack_dump_first,vtrack_dump_next
  
			Shows how all track buffer contents are dumped in order.
			NOTE: Content of TrackBuffer is opaque. For this example
			strings have been inserted...

  code:
	To do it in order use a simple loop:

	WARNING: This example is not multi thread save! To make it multi thread save
			 enclose it in a critical section or make sure no task is writing in
			 its TrackBuffer anymore (should be true during trace writer dump...).

	tsp00_Bool ok;
	void *msg;
	tsp00_Int4 len;
	tsp00_Int4 cnt;
	tsp00_TaskId last_id;
	tsp00_TaskId id;

		cout << "TRACKBUFFER DUMP FOLLOWS:" << endl;
		ok = vtrack_dump_first( &msg, &len, &id );
		while ( ok != FALSE )
		{
			last_id = id;

			cout << "----------------------------------" << endl;
			cout << "Dumping Trackbuffer for Task T" << (long)last_id << endl;

			do {
				cout << (char *)msg << endl;
				ok = vtrack_dump_next( &msg, &len, &id ) )
			} while ( ok != FALSE && id == last_id );

			cnt = vtrack_dump_check( last_id );
			if ( cnt > 0 )
			{
				cout << "Sorry OVERFLOW in TrackBuffer. Last " 
					 << cnt 
					 << " messages not stored"
					 << endl;
			}
		}

 */

/*!
  EndChapter: Examples
 */

#endif  /* HEO61_H */
