/*!***************************************************************************

  module      : heo69.h

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Accounting
  description : Accounting functions (Stack,Timing,IOs etc.)
  
  last changed: 1999-10-23  2:27
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef HEO69_H
#define HEO69_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "gsp00.h"
#include "RunTime/RTE_Types.h"

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/

/*!
  Function:     vcinit
  description:  Reset Millisecond counter

	This call gets the current number of ticks used in subsequent
	calls (vccheck) to get the number of milliseconds needed.
	User and System times are added to a total time.
  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
  return value: none
 */
externC void vcinit ( RTE_TaskId  taskid );

/*!
  Function:     vccheck
  description:  Get number of milliseconds since last call to vcinit()

	This call gets the current number of ticks and returns the number
	of milliseconds used since last call to vcinit().
	User and System times are added to a total time.
  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
	time   [out] Number of milliseconds since last vcinit
  return value: none
 */
externC void vccheck ( RTE_TaskId  taskid,
					   SAPDB_Int4    VAR_VALUE_REF time );

/*!
  Function:     vioinit
  description:  Reset task specific IO counter

	This call resets the global IO counter for the specific task
  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
  return value: none
 */
externC void vioinit ( RTE_TaskId  taskid );

/*!
  Function:     viocheck
  description:  Get direct IO counter since last call to vioinit()

	This call gets the global direct IO count for the specific task

  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
	no   [out] Number of direct IO calls
  return value: none
 */
externC void viocheck ( RTE_TaskId  taskid ,
						SAPDB_Int4    VAR_VALUE_REF no );

/*!
  Function:     vmonitor
  description:  Get task specific counter

	This call gets the counter for the specific task.

  arguments: 
	taskid [in] TaskId (only used in kernel RTE)
	phys_ios [out] Sum of SelfIORead/Write and DevIORead/Write
	suspends [out] Number of vsuspend() calls
	waits    [out] Number of vwait() calls
  return value: none
 */
externC void vmonitor ( RTE_TaskId  taskid ,
						SAPDB_Int4    VAR_VALUE_REF phys_ios ,
						SAPDB_Int4    VAR_VALUE_REF suspends ,
						SAPDB_Int4    VAR_VALUE_REF waits );

/*!
  Function:     vsinit

  description:  Initialization of stack checking code

  Initialization of stack checking code (only needed for Client RTE)
  This function is a dummy for kernel RTE.

  arguments: 
	none
  return value: none
 */
externC void vsinit ( /* void */ );

/*!
  Function:     vscheck

  description:  Returns the number of bytes left on the stack

  This call returns the current number of bytes left on the stack.

  In kernel RTE the current task stack is used.

  arguments: 
	LeftStackSize [out] Number of bytes left on stack

  return value: none
 */
externC void vscheck ( SAPDB_Int4 VAR_VALUE_REF LeftStackSize );

/*!
  Function:     vsfill

  description:  Fill stack with pattern

  Fill the remaning stack with a pattern.
  Leave the two-word chgstack()-test-pattern intact.
  Current function's additional stack requirements
  estimated as eight words.

  In kernel RTE the current task stack is used.

  arguments: 
    none

  return value: none
 */
externC void vsfill ( /* void */ );

/*!
  Function:     vsscan

  description:  Return number of bytes used on stack since last call to vsfill()

  This call uses the first difference from the pattern setup with last
  vsfill() call. This difference is used to return the maximum number
  of bytes used on the stack.

  In kernel RTE the current task stack is used.

  arguments: 
	MaxStackLeft [out] Number of bytes left on stack

  return value: none
 */
externC void vsscan ( SAPDB_Int4 VAR_VALUE_REF MaxStackLeft );

/*!
  Function:     vsoverflow

  description:  Calls vabort of stack overflow was detected
  
  This call uses the vsscan() function and compares the result with
  the RTE internal known task stack size. If overflow occured, a
  vabort() call is done (and a Stack Overflow message written to knldiag).

  arguments: 
    none

  return value: 
    none
 */
#if COMPILEMODE_MSP00 >= SLOW_MSP00	
#if defined(_WIN32)
#define SECURITY_BUFFER_MXEO69 (kgs.ulSystemPageSize)
#else
extern long en81_MemPageSize;
#define SECURITY_BUFFER_MXEO69 en81_MemPageSize
#endif
#else
#define SECURITY_BUFFER_MXEO69 0
#endif
externC void vsoverflow ( /* void */ );

/*!
  Function:     vsShowMaxStackUse
  
  description:  Show maximum stack usage of givben task in knldiag

  Arguments: taskId [in] task id of task for which the value is requested
             taskTypeName [in] readable type name of task
 */
externC void vsShowMaxStackUse(RTE_TaskId taskId, const char *taskTypeName);

/*!
  Function:     vsResetCmdMax
  
  description:  Reset the currently known max stack usage of given task.

  The maxstackusage is set by all vscheck() calls, if current stack usage is higher
  that the value stored in the task context.

  Arguments: taskId [in] task id of task for which the value is requested
 */
externC void vsResetCmdMax(RTE_TaskId taskId);

/*!
  Function:     vsGetCmdMax
  
  description:  Returns the currently known max stack usage of given task.

  The maxstackusage is set by all vscheck() calls, if current stack usage is higher
  that the value stored in the task context.

  Arguments: taskId [in] task id of task for which the value is requested
             cmdMax [out] current maximum of stack usage for command
  Return value: Stored maximum number of bytes used on stack of given task
 */
externC void vsGetCmdMax(RTE_TaskId taskId, SAPDB_Int4 VAR_VALUE_REF cmdMax);

/*!
  Function:     vsReinit

  description:  Reinitialize the current stack

  This call does a vsfill() but beginning with current stack position up to a known
  max stack size or up to stack limit if not known. The maxUsed value can be retrieved
  with a call to vsscan().

  Arguments:  taskId [in] task id for which to reinitialize stack
              maxUsed [in] maximum amount of bytes used on stack of task
 */
externC void vsReinit( RTE_TaskId taskId, SAPDB_Int4 maxUsed );

#endif  /* HEO69_H */
