/*! 
  -----------------------------------------------------------------------------
 
  module: hin101.h
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: XDB interface
 
  description:  Specification of XDB class
                
                This class provide a Interface for filling the R/3-XDB-tables in
                the right matter.
                
  version:      min 7.2.

  copyright:    by SAP AG, 1998

  -----------------------------------------------------------------------------
 
    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end
 
  -----------------------------------------------------------------------------
*/
 

#ifndef _HIN101_H_
#define _HIN101_H_

/* 
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/
#include "gsp00.h"
#include "geo200.h"
#include "hin29.h"

/*!  
  -----------------------------------------------------------------------------
  Chapter: Macros and Constants
  -----------------------------------------------------------------------------
 */

/*! Declaration: XDB function ids */

// Autosave
#define XDBOP_DISAUTO_IN101      "ase"           // disable autosave log
#define XDBOP_ENAUTO_IN101       "aso"           // enable autosave log

// Backup
#define XDBOP_SAVEDATAC_IN101    "sdc"           // Complete database backup  (consistent)
#define XDBOP_SAVEDATA_IN101     "sda"           // Complete database backup
#define XDBOP_SAVEPAGESC_IN101   "spc"           // Incremental database backup (consistent)
#define XDBOP_SAVEPAGES_IN101    "spa"           // Incremental database backup
#define XDBOP_SAVELOG_IN101      "slo"           // Save archive logs
#define XDBOP_SAVESTAGE_IN101    "ssa"           // Save stanging area

// Restore
#define XDBOP_RESTDATA_IN101     "rda"           // Complete database restore
#define XDBOP_RESTPAGES_IN101    "rpa"           // Incremental database restore 
#define XDBOP_RESTLOG_IN101      "rlo"           // Restore archive logs
#define XDBOP_CHECKDATA_IN101    "cda"           // check complete database restore
#define XDBOP_CHECKPAGES_IN101   "cpa"           // check incremental database restore 
#define XDBOP_CHECKLOG_IN101     "clo"           // check restore archive logs

// Config
#define XDBOP_ADDDATDEV_IN101    "add"           // add data devspace
#define XDBOP_ADDLOGDEV_IN101    "ald"           // add log devspace
#define XDBOP_CHANGECFG_IN101    "cco"           // change DB Konfiguration
#define XDBOP_LOADSYSTAB_IN101   "lst"           // load system tables

// check
#define XDBOP_CHECKDB_IN101      "cdb"           // check database structure
#define XDBOP_TABLECHECK_IN101   "cta"           // Check database structure (only tables) (xpu -v)

// statistics
#define XDBOP_OPTCHECK_IN101     "och"           // Check optimizer statistics
#define XDBOP_OPTCREATE_IN101    "osc"           // Create new optimizer statistics
#define XDBOP_OPTREFR_IN101      "osr"           // Refresh optimizer statistics
#define XDBOP_TABLESTAT_IN101    "cts"           // Create table statistics

/*! EndDeclaration: XDB function ids */

/* Declaration: XDB topics (operation group) */

#define XDBTP_BACKUP_IN101       "BACKUP"        // Backup
#define XDBTP_CONFIG_IN101       "CONFIGURATION" // Configuration
#define XDBTP_CHECK_IN101        "CHECK"         // Check
#define XDBTP_STRUCTURE_IN101    "STRUCTURE"     // Structure
#define XDBTP_PERF_IN101         "PERFORMANCE"   // Performance

/* EndDeclaration: XDB topics (operation group) */

/*! Declaration: XDB components (programs) */

#define XDBCO_DBM_IN101          "DBMSERVER"     // Database Manager
#define XDBCO_DISPWORK_IN101     "DISP+WORK"     // R/3 DISP+WORK
#define XDBCO_XPU_IN101          "XPU"           // XPU

/*! EndDeclaration: XDB components (programs) */

/*! Declaration: XDB states */

#define XDBST_OK_IN101                   0           // OK
#define XDBST_WARNING_IN101              1           // Warning
#define XDBST_ERROR_IN101                2           // Error

/*! EndDeclaration: XDB states */

/*! Declaration: XDB limitations */

#define XDBMAX_FUNCID_IN101              3           // max. length of function id
#define XDBMAX_COMPONENT_IN101          30           // max. length of component name
#define XDBMAX_TIMESTAMP_IN101          14           // max. length of timestamp YYYYMMDDHHMMSS
#define XDBMAX_BAOBJECT_IN101           64           // max. length of backup object type
#define XDBMAX_BATYPE_IN101             15           // max. length of backup type
#define XDBMAX_BAMETHOD_IN101           15           // max. length of bckup method
#define XDBMAX_STATE_IN101            9999           // upper limit for state 
#define XDBMIN_USRSTATE_IN101            3           // lower limit for user defined state
#define XDBMIN_STATE_IN101               0           // lower limit for state

/*! EndDeclaration: XDB limitations */

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Macros and Constants
  -----------------------------------------------------------------------------
 */

/*!
  -----------------------------------------------------------------------------
  Class:        tin101_XDBOperation
  -----------------------------------------------------------------------------

  description:  XDB Operation Object

                An instance of this class represents a XDB Operation.

                There is now only "online" ("warm") support of XDB. The
                "offline" support ist not yet implemeted.

                <B>
                On default instances of this class are disabled. That means,
                nothing is written to the XDB tables. You have to explicit 
                enable  XDB by tin101_XDBOperation::enable(). This is a static
                member, so you need only one call for all present and future
                instances of this class.
                </B>
                
  -----------------------------------------------------------------------------
 */
class tin101_XDBOperation {

  public:

    /*!  
      Chapter: Construction and destruction
     */

    /*!
      -------------------------------------------------------------------------
      function:     tin101_XDBOperation
      -------------------------------------------------------------------------
      description:  Constructor

                    This constructor opens a user session to the specified
                    database with the user from the DEFAULT-Key of the 
                    local XUser-File.

                    For errors during the contruction phase check lastEvent.

      arguments:    szNode   [IN] - the name of the node
                    szDB     [IN] - the name of the database
                    szUsr    [IN] - the name of the user
                    szPwd    [IN] - the password (clear) of the user
                    bUnicode [IN] - is this a unicode database?

      prototypes:   2

      -------------------------------------------------------------------------
     */
    tin101_XDBOperation  ( const tsp00_DbNamec        & szDb,  
                           const tsp00_KnlIdentifierc & szUsr,
                           const tsp00_Namec          & szPwd,
                                 bool                   bUnicode );

    tin101_XDBOperation  ( const tsp00_NodeIdc        & szNode,  
                           const tsp00_DbNamec        & szDb,  
                           const tsp00_KnlIdentifierc & szUsr,
                           const tsp00_Namec          & szPwd,
                                 bool                   bUnicode );

    /*!
      -------------------------------------------------------------------------
      function:     ~tin101_XDBOperation
      -------------------------------------------------------------------------
      description:  Destructor

                    The destructor releases the user session and frees all
                    allocated resources.

      -------------------------------------------------------------------------
     */
    virtual ~tin101_XDBOperation ( );

    /*!  
      EndChapter: Construction and destruction
     */

    /*!  
      Chapter: Error handling
     */

    /*!
      -------------------------------------------------------------------------
      function:     lastEvent
      -------------------------------------------------------------------------
      description:  get last error description

                    For more in informatio see <a href="http://pwww/LZU/specs/event-list-handling.html">Event List Handling</a>.

      return value: A Object of teo200_EventList. This object might contains
                    more then one entry.

      -------------------------------------------------------------------------
     */
    teo200_EventList & lastEvent ( );


    /*!
      -------------------------------------------------------------------------
      function:     isXDBInstalled
      -------------------------------------------------------------------------
      description:  XDB installation state

                    This function shows the XDB installation state

      return value: A value of false shows the absence of the XDB Tables

      -------------------------------------------------------------------------
     */
    bool isXDBInstalled();

    /*!
      -------------------------------------------------------------------------
      function:     reset
      -------------------------------------------------------------------------
      description:  reset object

                    This function sends a ROLLBACK to the database, releases 
                    and reconnects the session and clears the object status.

                    Use this function moving the operation object in a defined
                    state after errors.

      return value: A boolean value of false indicates an error. Use lastEvent
                    for analyzing this error.

      -------------------------------------------------------------------------
     */
    bool reset     ( );

    /*!  
      EndChapter: Error handling
     */

    /*!  
      Chapter: Enable/Disable XDB
     */

    /*!
      -------------------------------------------------------------------------
      function:     enable
      -------------------------------------------------------------------------
      description:  enable XDB 

                    All subsequent calls to XDB will write to database.

                    This is a static member function writing to a static
                    member variable, so this call concerns all present and
                    future instances of tin101_XDBOperation of this process.


      arguments:    bEnable     [IN] - set the state

      -------------------------------------------------------------------------
     */
    static void enable     (const bool bEnable = true );

    /*!
      -------------------------------------------------------------------------
      function:     disable
      -------------------------------------------------------------------------
      description:  disable XDB 

                    All subsequent calls to XDB will do nothing

                    This is a static member function writing to a static
                    member variable, so this call concerns all present and
                    future instances of tin101_XDBOperation of this process.

      -------------------------------------------------------------------------
     */
    static void disable     ( );

    /*!
      -------------------------------------------------------------------------
      function:     isEnabled
      -------------------------------------------------------------------------
      description:  get state of XDB 

      -------------------------------------------------------------------------
     */
    static bool isEnabled     ( );

    /*!  
      EndChapter: Enable/Disable XDB
     */

    /*!  
      Chapter: XDB Operation handling
     */

    /*!
      -------------------------------------------------------------------------
      function:     start
      -------------------------------------------------------------------------
      description:  write STARTED entry

                    This function writes a new record to table DBAML. The
                    state ist set to started all other data (except the 
                    parameters of this function) will be provided by the
                    XDB Operation object.

      arguments:    szFunction  [IN] - the function id (see <A HREF="#XDB%20function%20ids" >XDB function ids</A>)
                    szComponent [IN] - the name of the component (program) (see <A HREF="#XDB%20components%20%28programs%29" >XDB components</A>)
                    szTimeStamp [OUT]- the generated timestamp (OPERID)
                    bCommit     [IN] - commit changes to database?

      return value: A boolean value of false indicates an error. Use lastEvent
                    for analyzing this error.

      -------------------------------------------------------------------------
     */
    bool start     ( const char        * szFunction,
                     const char        * szComponent,
                     tsp00_Timestampc  & szTimeStamp,
                     const bool          bCommit = true );

    /*!
      -------------------------------------------------------------------------
      function:     run
      -------------------------------------------------------------------------
      description:  write RUNNING entry

                    This function updates the record written by start() to the
                    state RUNNING. The application can provide the current
                    operation return code and a approximated end "

      arguments:    nState      [IN] - current return code of operation
                    szApprEnd   [IN] - approximated end " YYYYMMDDHHMMSS
                    bCommit     [IN] - commit changes to database?

      return value: A boolean value of false indicates an error. Use lastEvent
                    for analyzing this error.

      -------------------------------------------------------------------------
     */
    bool run       ( const int           nState,
                     const char        * szApprEnd,
                     const bool          bCommit = true );
    /*!
      -------------------------------------------------------------------------
      function:     complete
      -------------------------------------------------------------------------
      description:  write COMPLETED entry

                    This function updates the record written by start() and 
                    possible updated by run() to the state COMPLETED. 

      arguments:    nState      [IN] - return code of operation
                    bCommit     [IN] - commit changes to database?

      return value: A boolean value of false indicates an error. Use lastEvent
                    for analyzing this error.

      -------------------------------------------------------------------------
     */
    bool complete  ( const int           nState,
                     const bool          bCommit = true );

    /*!  
      EndChapter: XDB Operation handling
     */

  protected:
    bool start46B        ( const tsp00_Timestampc & szTimeStamp,
                           const bool               bCommit );
    bool start46C        ( const tsp00_Timestampc & szTimeStamp,
                           const bool               bCommit );

    bool run46B          ( const int                nState,
                           const char             * szApprEnd, 
                           const bool               bCommit );
    bool run46C          ( const int                nState,
                           const char             * szApprEnd, 
                           const bool               bCommit );

    bool complete46B     ( const int                nState,
                           const bool               bCommit );    
    bool complete46C     ( const int                nState,
                           const bool               bCommit );    

    void initObject      ( const tsp00_NodeIdc          & szNode,  
                           const tsp00_DbNamec          & szDb,  
                           const tsp00_KnlIdentifierc   & szUsr,
                           const tsp00_Namec            & szPwd,
                                 bool                     bUnicode );

    bool connect         ( const tsp00_NodeIdc          & szNode,  
                           const tsp00_DbNamec          & szDb,  
                           const tsp00_KnlIdentifierc   & szUsr,
                           const tsp00_Namec            & szPwd,
                                 bool                     bUnicode,
                                 tin01_sql_session      & aSession,
                                 teo200_EventList       & aEvent      ) const;

    bool connect         (       tin01_sql_session & aSession,
                                 teo200_EventList  & aEvent      ) const;

    void disconnect      (       tin01_sql_session & aSession    ) const;

    bool execute         (       tin01_sql_session & aSession,
                                 const char        * pStatement, 
                                 teo200_EventList  & aEvent,
                                 const bool          bLog = true ) const;
            
    bool isXDBInstalled  (       tin01_sql_session & aSession,
                                 teo200_EventList  & aEvent      );

    void timestamp       (       tsp00_Timestampc  & szTimestamp ) const;

    bool checkFunc46B    (       tin01_sql_session & aSession,
                           const char              * szFunction,
                           const char              * szComponent,
                           const char              * szTable,
                                 teo200_EventList  & aEvent      ) const;
    bool checkFunc46C    (       tin01_sql_session & aSession,
                           const char              * szFunction,
                           const char              * szComponent,
                                 teo200_EventList  & aEvent      ) const;

    bool checkTrial      (       tin01_sql_session & aSession,
                           const char              * szFunction,
                           const char              * szComponent,
                           const int                 nTrial, 
                                 teo200_EventList  & aEvent      ) const;

    bool checkErr        (       tin01_sql_session & aSession,
                           const char              * szComponent,
                           const int                 nError,
                                 teo200_EventList  & aEvent      ) const;

    // current operation state
    const char   *    m_pszState;
    // active operation id - timestamp (for record update)
    tsp00_Timestampc  m_szOperId;
    // active function id (for record update)
    char              m_szFuncId[XDBMAX_FUNCID_IN101 + 1];
    // active component name (for record update)
    char              m_szComponent[XDBMAX_COMPONENT_IN101 + 1];
    // active timestamp
    tsp00_Timestampc  m_szNowStamp;

    // current sql session
    tin01_sql_session m_Session;

    // object with the last error description
    teo200_EventList  m_EvtLast;  

     // R/3 System ID
    tsp00_Namec       m_R3Sid;

    // XDB enabled state
    static bool         m_bEnabled;
    // active R/3 Version
    static tsp00_Int4   m_R3Version;

};

/*! EndClass: tin101_XDBOperation */

/*!
  -----------------------------------------------------------------------------
  Class:        tin101_XDBBackupOperation
  -----------------------------------------------------------------------------

  description:  XDB Backup Operation Object

                An instance of this class represents a XDB Backup Operation.
                
                <B>
                On default instances of this class are disabled. That means,
                nothing is written to the XDB tables. You have to explicit 
                enable  XDB by tin101_XDBOperation::enable(). This is a static
                member, so you need only one call for all present and future
                instances of this class.
                </B>

  -----------------------------------------------------------------------------
 */

/*!  
  -----------------------------------------------------------------------------
  Chapter: Macros and Constants (tin101_XDBBackupOperation)
  -----------------------------------------------------------------------------
 */

/*! Declaration: XDB Backup Objects */

#define XDB_BAOBJ_PAGES_IN101    "PAGES"
#define XDB_BAOBJ_DATA_IN101     "DATA"
#define XDB_BAOBJ_LOG_IN101      "LOG"

/*! EndDeclaration: XDB Backup Objects */

/*! Declaration: XDB Backup Type */

#define XDB_BATYP_WARM_IN101     "WARM"
#define XDB_BATYP_COLD_IN101     "COLD"

/*! EndDeclaration: XDB Backup Type */

/*! Declaration: XDB Backup Method */

#define XDB_BAMTH_MIGR_IN101     "WITH CHECKPOINT"
#define XDB_BAMTH_REC_IN101      "NO CHECKPOINT"

/*! EndDeclaration: XDB Backup Method */

/*!  
  -----------------------------------------------------------------------------
  EndChapter: Macros and Constants
  -----------------------------------------------------------------------------
 */

class tin101_XDBBackupOperation : public tin101_XDBOperation {

  public:
    /*!  
      Chapter: Construction and destruction
     */

    /*!
      -------------------------------------------------------------------------
      function:     tin101_XDBBackupOperation
      -------------------------------------------------------------------------
      description:  Constructor

                    This constructor opens a user session to the specified
                    database with the user from the DEFAULT-Key of the 
                    local XUser-File.

                    For errors during the contruction phase check lastEvent.

      arguments:    szNode   [IN] - the name of the node
                    szDB     [IN] - the name of the database
                    szUsr    [IN] - the name of the user
                    szPwd    [IN] - the password (clear) of the user
                    bUnicode [IN] - is this a unicode database?

      prototypes:   2
      -------------------------------------------------------------------------
     */
    tin101_XDBBackupOperation  ( tsp00_DbNamec        szDb,  
                                 tsp00_KnlIdentifierc szUsr,
                                 tsp00_Namec          szPwd,
                                 bool                 bUnicode );

    tin101_XDBBackupOperation  ( tsp00_NodeIdc        szNode,  
                                 tsp00_DbNamec        szDb,  
                                 tsp00_KnlIdentifierc szUsr,
                                 tsp00_Namec          szPwd,
                                 bool                 bUnicode );

    /*!  
      EndChapter: Construction and destruction
     */


    /*!  
      Chapter: XDB Operation handling
     */

    /*!
      -------------------------------------------------------------------------
      function:     start
      -------------------------------------------------------------------------
      description:  write STARTED entry

                    This function writes a new record to table DBAML. The
                    state ist set to started all other data (except the 
                    parameters of this function) will be provided by the
                    XDB Operation object.

      arguments:    szFunction  [IN] - the function id (see <A HREF="#XDB%20function%20ids" >XDB function ids</A>)
                    szComponent [IN] - the name of the component (program) (see <A HREF="#XDB%20components%20%28programs%29" >XDB components</A>)
                    szTimeStamp [OUT]- the generated timestamp (OPERID)
                    szObject    [IN] - Object to save/restore (see <A HREF="#XDB%20Backup%20Objects" >XDB Backup Objects</A>)
                    szType      [IN] - database state during backup (see <A HREF="#XDB%20Backup%20Type" >XDB Backup Type</A>)
                    szMethod    [IN] - with checkpoint (see <A HREF="#XDB%20Backup%20Method" >XDB Backup Method</A>)

      return value: A boolean value of false indicates an error. Use lastEvent
                    for analyzing this error.

      -------------------------------------------------------------------------
     */
    bool start     ( const char        * szFunction,
                     const char        * szComponent,
                     tsp00_Timestampc  & szTimeStamp,
                     const char        * szObject,
                     const char        * szType,
                     const char        * szMethod);

    /*!  
      EndChapter: XDB Operation handling
     */

  protected:
    bool start46B  ( const char        * szFunction,
                     const char        * szComponent,
                     tsp00_Timestampc  & szTimeStamp,
                     const char        * szObject,
                     const char        * szType,
                     const char        * szMethod);
    bool start46C  ( const char        * szFunction,
                     const char        * szComponent,
                     tsp00_Timestampc  & szTimeStamp,
                     const char        * szObject,
                     const char        * szType,
                     const char        * szMethod);

};

/*! EndClass: tin101_XDBBackupOperation */

#endif  // _HIN101_H_
