/*!================================================================
 module:        hin202.h

 -------------------------------------------------------------------

 responsible:   DanielD

 special area:  SQL

 description:   Segment

 -------------------------------------------------------------------

 copyright:     SAP AG 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#ifndef HIN202_H
#define HIN202_H

#include "vsp001.h"
#include "hin201.h"

/*----------------------------------------*/

/*!
  Class: cin202_Segment
 */
class cin202_Segment
{
protected:
    /*! Function: constructor

        description: creates a new Segment object.

            This can be called only by subclasses.

     */
    cin202_Segment (tsp1_segment * segmentArg = NULL);

public:
    /*! Function: destructor

        description: releases any ressources
     */
    ~cin202_Segment ();

    /*! Function: TraceOn

        description: dumps content of Segment to @param{trace}.
     */
    void TraceOn (cin97_TraceStream * trace, int buflimit = -1);

    /*! Function: Size

        return value: the maximal size of the Segment
     */
    int Size () const;

    /*! Function: Length

        return value: the actual length of the Segment
     */
    int Length () const;

    /*! Function: IsValid

        return value: true if Segment points to an actual tsp1_segment
     */
    bool IsValid () const;

    /*! Function: Invalidate

        description: makes the reference invalid.
     */
    void Invalidate ();

    /*! Function: GetRawSegment

        return value: the native tsp1_segment.

        prototypes: 2
     */
    tsp1_segment * GetRawSegment ();
    const tsp1_segment * GetRawSegment () const;

    /*! Function: GetRawHeader

        return value: the native tsp1_segment_header

        prototypes: 2
     */
    tsp1_segment_header * GetRawHeader ();
    const tsp1_segment_header * GetRawHeader () const;


protected:
    tsp1_segment * rawSegment;
};

/*!
  EndClass: cin202_Segment
 */
/*----------------------------------------*/

/*!
  Class: cin202_RequestSegment
 */
class cin202_RequestSegment : public cin202_Segment
{
public:

    /*! Function: constructor

        description: creates a new RequestSegment

            This should never be called directly. ReplySegments
            are created by RequestPacket::AddSegment.
     */
    cin202_RequestSegment (tsp1_segment * segmentArg = NULL);

    /*! Function: Init

        description: initializes a RequestSegment.

            This should never be called directly. It is
            called implicitly by RequestPacket::AddSegment.
    */

    /* PTS 1108446 sqlMode added */
    /* PTS 1112690 parsingAgain added */
    void Init (tsp1_cmd_mess_type_Param messType, const tsp1_sqlmode_Enum &sqlMode,
      const pasbool parsingAgain,
      int offset, int index);

    /*! Function: ExtendLength

        description: grows the segment by @param{increment} bytes.

            The packet grows by the same amount.
     */
    void ExtendLength (int increment);

    /*! Function: AddPart

        description: creates a new @param{partKind} part.
     */
    cin201_Part AddPart (tsp1_part_kind_Param partKind);

    /*! Function: ClosePart

        description: closes @param{part}
     */
    void ClosePart (cin201_Part & part);

};

/*!
  EndClass: cin202_RequestSegment
 */

/*----------------------------------------*/

/*!
  Class: cin202_ReplySegment
 */
class cin202_ReplySegment : public cin202_Segment
{
public:
    /*! Function: constructor

        description: creates a new ReplySegment.

            This should never be called directly. ReplySegments
            are created by RequestPacket::nextSegment.

        arguments:
            segmentArg [in, optional] the segment pointer
     */
    cin202_ReplySegment (tsp1_segment * segmentArg = NULL);

    /*! Function: ErrorCode

        return value: the sql error code
     */
    int ErrorCode () const;

    /*! Function: FunctionCode

        return value: information about the executed command
     */
    int FunctionCode () const;

    /*! Function: GetFirstPart

        return value: the first Part of the Segment.
     */
    const cin201_Part GetFirstPart () const;

    /*! Function: GetNextPart

        return value: the next Part.

            When @param{part} is the last part, an invalid Part is returned
            When @param{part} is an invalid Part, then the first Part is used.
     */
    const cin201_Part GetNextPart (const cin201_Part & part) const;

    /*! Function: FindPart

        return value: the first Part of the given kind.

            When no Part of this kind is found, an invalid Part is returned.
     */
    const cin201_Part FindPart (tsp1_part_kind_Param partKind) const;

    /*! Function: FindParts

        description: allows to retrieve several parts in one operation.

        arguments:
            count [in]      the number of requested parts
            partKinds [in]  an array of tsp1_part_kind_Param
            parts [out]     each entry will contain either a Part
                of the kind in the corresponding slot of @param{partKinds}
                or an invalid Part.

        return value: the number of requested parts actually found
     */
    int FindParts (int count, const  tsp1_part_kind_Param partKinds [],
        cin201_Part parts []) const;

};

/*!
  EndClass: cin202_ReplySegment
 */

/*----------------------------------------*/

#endif

