/*!================================================================
 module:        vin203.cpp

 -------------------------------------------------------------------

 responsible:   DanielD

 special area:  SQL

 description:   Packet

 -------------------------------------------------------------------

 copyright:     SAP AG 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#include "hin203.h"

/*----------------------------------------*/

cin203_SQLPacket::cin203_SQLPacket (tsp1_packet * packetPtr)
{
    this->rawPacket = packetPtr;
}

/*----------------------------------------*/

cin203_SQLPacket::~cin203_SQLPacket ()
{
}

/*----------------------------------------*/

static const char * messCodeStrings [] = {
    "ascii",
    "ebcdic",
    "codeneutral"
};

static const tin97_EnumDesc messCodeDesc = {
    csp_codeneutral,
    "Invalid messcode %d",
    messCodeStrings
};

static const char * swapTypeStrings [] = {
    "dummy",
    "normal",
    "full",
    "part"
};

static const tin97_EnumDesc swapTypeDesc = {
    sw_part_swapped,
    "Unknwon %d",
    swapTypeStrings
};

/*----------------------------------------*/

void cin203_SQLPacket::TraceOn (cin97_TraceStream * trace, int buflimit)
{
    tsp1_packet_header * header = &this->rawPacket->sp1_header;
    char tmpBuf [tmpBufSize_cin97] [2];

    if (!this->IsValid ()) {
        trace->writeln ("Invalid Packet");
    }
    trace->printf ("<Packet> %s  %s swap  %.3s-%.5s (transfer len %d)\n",
        in97EnumDump (&messCodeDesc, header->sp1h_mess_code, tmpBuf [0]),
        in97EnumDump (&swapTypeDesc, header->sp1h_mess_swap, tmpBuf [1]),
        header->sp1h_application.asCharp (),
        header->sp1h_appl_version.asCharp (),
        this->Length ());
    trace->printf ("    %3d Segment(s) (%d of %d bytes)\n",
        header->sp1h_no_of_segm, header->sp1h_varpart_len,
        header->sp1h_varpart_size);

    cin203_ReplyPacket tmpPacket (this->rawPacket);
    cin202_ReplySegment segment = tmpPacket.GetFirstSegment ();
    while (segment.IsValid ()) {
        segment.TraceOn (trace, buflimit);
        segment = tmpPacket.GetNextSegment (segment);
    }
    trace->writeln ("</Packet>");
}

/*----------------------------------------*/

int cin203_SQLPacket::Length () const
{
    return this->rawPacket->sp1_header.sp1h_varpart_len
        + sizeof (tsp1_packet_header);
}

/*----------------------------------------*/

bool cin203_SQLPacket::IsValid () const
{
    return this->rawPacket != NULL;
}

/*----------------------------------------*/

void cin203_SQLPacket::Invalidate ()
{
    this->rawPacket = NULL;
}

/*----------------------------------------*/

tsp1_packet * cin203_SQLPacket::GetRawPacket ()
{
    return this->rawPacket;
}

/*----------------------------------------*/

const tsp1_packet * cin203_SQLPacket::GetRawPacket () const
{
    return this->rawPacket;
}

/*----------------------------------------*/

tsp1_packet_header * cin203_SQLPacket::GetRawHeader ()
{
    return &this->rawPacket->sp1_header;
}

/*----------------------------------------*/

const tsp1_packet_header * cin203_SQLPacket::GetRawHeader () const
{
    return &this->rawPacket->sp1_header;
}

/*----------------------------------------*/

cin203_RequestPacket::cin203_RequestPacket (tsp1_packet * packetPtr, 
                                            int size, 
                                            int unicode)
    : cin203_SQLPacket (packetPtr)
{
    tsp1_packet_header * header = &this->rawPacket->sp1_header;
    union {
        tsp00_Int4 asInt;
        char       asChar [4];
    } swapTest;

    memset (header, '\0', sizeof (tsp1_packet_header));
    if  ( unicode ) {
      union { tsp00_Uint2 s; tsp00_Byte c[2]; } ucs2_blank;
      ucs2_blank.s = 32;
      if ((ucs2_blank.c[0] ? 1 : 0 )) {
        header->sp1h_mess_code = csp_unicode_swap;
      }
      else {
        header->sp1h_mess_code = csp_unicode;
      };
    } 
    else {
      
      if (' ' == 32) {
        header->sp1h_mess_code = csp_ascii;
      }
      else if (' ' == 20) {
        header->sp1h_mess_code = csp_ebcdic;
      }
      else {
        header->sp1h_mess_code = csp_codeneutral + 1; // unknown
      }
    }

    swapTest.asInt = 1;
    if (swapTest.asChar [3] == 1) {
        header->sp1h_mess_swap.becomes (sw_normal);
    }
    else if (swapTest.asChar [0] == 1) {
        header->sp1h_mess_swap.becomes (sw_full_swapped);
    }
    else {
        header->sp1h_mess_swap.becomes (sw_part_swapped);
    }
    header->sp1h_varpart_len = 0;
    header->sp1h_no_of_segm = 0;
    header->sp1h_varpart_size = size - sizeof (tsp1_packet_header);
    this->InitVersion ("CPC");
}

/*----------------------------------------*/

cin203_RequestPacket::cin203_RequestPacket ()
    : cin203_SQLPacket (NULL)
{
}

/*----------------------------------------*/

cin203_RequestPacket::~cin203_RequestPacket ()
{
    // nothing to do
}


/*----------------------------------------*/

void cin203_RequestPacket::InitVersion (const char * component, const char * version)
{
    this->rawPacket->sp1_header.sp1h_application.c2p (component);
    if (version != NULL) {
        this->rawPacket->sp1_header.sp1h_appl_version.c2p (version);
    }
    else {
        this->rawPacket->sp1_header.sp1h_appl_version.c2p ("70201");
    }
}

/*----------------------------------------*/

void cin203_RequestPacket::Reset ()
{
    tsp1_packet_header * header = &this->rawPacket->sp1_header;

    header->sp1h_varpart_len = 0;
    header->sp1h_no_of_segm = 0;
}

/*----------------------------------------*/

void cin203_RequestPacket::CloseSegment (cin202_RequestSegment & segment)
{
    // nothing to do because segment keeps packet up to date
}

/*----------------------------------------*/

/* PTS 1108446 sqlMode added      */
/* PTS 1112690 parsingAgain added */
cin202_RequestSegment cin203_RequestPacket::AddSegment (tsp1_cmd_mess_type_Param messType,
                                                        const tsp1_sqlmode_Enum &sqlMode,
                                                        const pasbool parsingAgain )
{
    int newOffset = this->rawPacket->sp1_header.sp1h_varpart_len;
    tsp1_segment * segmPtr = REINTERPRET_CAST (tsp1_segment *,
        this->rawPacket->sp1_varpart ().asCharp () + newOffset);
    cin202_RequestSegment result (segmPtr);
    ++this->rawPacket->sp1_header.sp1h_no_of_segm;
    /* PTS 1108446 sqlMode added */
    result.Init (messType, sqlMode, parsingAgain, newOffset, this->rawPacket->sp1_header.sp1h_no_of_segm);
    return result;
}

/*----------------------------------------*/

cin203_ReplyPacket::cin203_ReplyPacket (tsp1_packet * packetPtr)
    : cin203_SQLPacket (packetPtr)
{
}

/*----------------------------------------*/

cin203_ReplyPacket::~cin203_ReplyPacket ()
{
    // nothing to do
}

/*----------------------------------------*/

const cin202_ReplySegment cin203_ReplyPacket::GetFirstSegment () const
{
    tsp1_segment * rawSegment;

    rawSegment = &this->rawPacket->sp1_segm ();
    cin202_ReplySegment result (rawSegment);
    return result;
}

/*----------------------------------------*/

const cin202_ReplySegment cin203_ReplyPacket::GetNextSegment (
    const cin202_ReplySegment & segment) const
{
    const tsp1_segment * segmPointer;
    const teo00_Byte * segmAddr;
    int nextSegmPos;
    const teo00_Byte * nextSegmAddr;
    const tsp1_segment * nextSegmPointer;

    segmPointer = segment.GetRawSegment ();
    segmAddr = REINTERPRET_CAST (const teo00_Byte *, segmPointer);
    nextSegmAddr = NULL;
   	if ( segmPointer->sp1s_segm_header().sp1s_no_of_parts() ) {
      nextSegmPos = segmPointer->sp1p_segm_offset ()
        + sizeof (tsp1_segment_header)
        + segment.Length ();
      if (this->rawPacket->sp1_header.sp1h_varpart_len > nextSegmPos) {
        nextSegmAddr = segmAddr + segmPointer->sp1p_segm_offset ();
      }
    }
    nextSegmPointer = REINTERPRET_CAST (const tsp1_segment*, nextSegmAddr);
    cin202_ReplySegment result (CONST_CAST (tsp1_segment*, nextSegmPointer));
    return result;
}

/*----------------------------------------*/


