/*!***************************************************************************

    module      : vkb391.cpp
    responsible : MartinB,UweH
    special area: KB_Logging
    last changed: 2001-05-03
    copyright   : Copyright by SAP AG, 2001
    description : Interface to new Logging component Log_SaveIterator
                  added RTEConf_Parameter handling (UH)

    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Logging/Log_Page.hpp"
#include "Logging/Log_SaveIterator.hpp"
#include "Logging/Log_ArchiveLog.hpp"
#include "Logging/Log_ArchiveLogManager.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccess.hpp"

#include "hkb391.h"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *   LOCAL FUNCTIONS                                                         *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/


/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

externPascal void kb391_InitLogSaveIterator (
    tgg00_TransContext    VAR_VALUE_REF  trans,
    tsp00_Int2                           maxToBeRead,
    pasbool                              dbIsCold,
    pasbool                              forAutosave,
    pasbool                              completeLog,
    Log_SaveIterator*&                   iterator)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_InitLogSaveIterator", LogVolume_Trace, 5);

    trans.trError_gg00 = e_ok;

    if (maxToBeRead < 0)
    {
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                      SAPDBERR_ASSERT_STATE_FAILED,
                      "kb391_InitLogSaveIterator:max_to_be_read < 0"));
    }

    SAPDBMem_IRawAllocator& allocator =
              *(reinterpret_cast <SAPDBMem_IRawAllocator*>(trans.trAllocator_gg00));

    Log_ArchiveLogManager    &log_manager = Log_ArchiveLogManager::Instance(); 
    Log_ArchiveLog           *first_log   = *log_manager.ArchiveLogsBegin();
    Log_ArchiveLog::SaveInfo  saveInfo;
    
    first_log->GetSaveLogIterator(  trans.trTaskId_gg00,
                                    allocator,
                                    dbIsCold?true:false,
                                    forAutosave?true:false,
                                    completeLog?true:false,
                                    maxToBeRead,
                                    iterator,
                                    saveInfo);

    if ( iterator == NULL )
    {
        switch (saveInfo)
        {
            case Log_ArchiveLog::okay:
                    trans.trError_gg00 = e_ok;
                    break;
            case Log_ArchiveLog::noLogToSave:
					if ( forAutosave )
						trans.trError_gg00 = e_incomplete_logsegm;
					else
						trans.trError_gg00 = e_no_log_to_save;
                    break;
            case Log_ArchiveLog::historyIsLost:
                    trans.trError_gg00 = e_backup_history_lost; 
                    break;
            case Log_ArchiveLog::incompleteSegment:
                    trans.trError_gg00 = e_incomplete_logsegm;
                    break;
            default:
                    trans.trError_gg00 = e_ok;
                    break;
        }
    }
}
/*---------------------------------------------------------------------------*/
externPascal void kb391_ReadNextLogpages (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tkb3_block_ptr                       queue_buffer,
    tsp00_Int2            VAR_VALUE_REF  num_read)

{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_ReadNextLogpages", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );
    Log_DeviceBlockCount    unsigned_numPagesRead;
    
    iterator->GetNextLogPages(t.trTaskId_gg00,queue_buffer,unsigned_numPagesRead);
    num_read = unsigned_numPagesRead;
}
/*---------------------------------------------------------------------------*/
externPascal void kb391_PrepareFreeLogForPipe (// PTS 1114791 mb 2002-08-22 new
    Log_SaveIterator*                    iterator,
    tgg00_BasisError      VAR_VALUE_REF  TrError)
{
    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
    Log_ArchiveLog::PipeHandlingResult result;
    result = log.PrepareFreeLogForPipe(          iterator->GetStartIOSequence(),
                                                 iterator->GetEndIOSequence(),
                                                 iterator->GetStartPageOffset(),
                                                 iterator->GetEndPageOffset());
    switch (result)
    {
        case Log_ArchiveLog::noFreeLogCalledInbetween:
            TrError = e_invalid; // same error as in V7.3
            break;
        case Log_ArchiveLog::pipeOk:
            TrError = e_ok; 
            break;
        default:
            TrError = e_log_error;
    }
}
/*---------------------------------------------------------------------------*/

externPascal void kb391_DeleteIterator (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator**                   iterator)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_DeleteIterator", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    // deallocate memory */
    SAPDBMem_IRawAllocator& Allocator =
        *(REINTERPRET_CAST (SAPDBMem_IRawAllocator*,
                            t.trAllocator_gg00));

    (*iterator)->Delete();
    destroy(*iterator, Allocator);
    *iterator = NULL;
}

/*---------------------------------------------------------------------------*/


externPascal void kb391_GetStartTime (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_Int4            VAR_VALUE_REF  startTime,
    tsp00_Int4            VAR_VALUE_REF  startDate)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetStartTime", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );
    
    Kernel_Date    date;
    Kernel_Time    time;
    
    iterator->GetStartTime(date,time);
    
    startTime=time;
    startDate=date;
}

/*---------------------------------------------------------------------------*/

externPascal void kb391_GetEndTime (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_Int4            VAR_VALUE_REF  endTime,
    tsp00_Int4            VAR_VALUE_REF  endDate)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetEndTime", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    Kernel_Date    date;
    Kernel_Time    time;
    
    iterator->GetEndTime(date,time);

    endTime=time;
    endDate=date;
}


/*---------------------------------------------------------------------------*/

externPascal void kb391_GetStartIOSeq (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_PageNo          VAR_VALUE_REF  startIOSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetStartIOSeq", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    startIOSeq = iterator->GetStartIOSequence();
}


/*---------------------------------------------------------------------------*/

externPascal void kb391_GetEndIOSeq (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_PageNo          VAR_VALUE_REF  endIOSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetEndIOSeq", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    endIOSeq = iterator->GetEndIOSequence();
}


/*---------------------------------------------------------------------------*/

externPascal void kb391_GetStartPageOffset (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_Uint4           VAR_VALUE_REF  startIOSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetStartPageOffset", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    startIOSeq = iterator->GetStartPageOffset();
}


/*---------------------------------------------------------------------------*/

externPascal void kb391_GetEndPageOffset (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_Uint4           VAR_VALUE_REF  endIOSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetEndPageOffset", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    endIOSeq = iterator->GetEndPageOffset();
}
/*---------------------------------------------------------------------------*/

externPascal void kb391_GetFirstNotSavedIOSeq (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_PageNo          VAR_VALUE_REF  endIOSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetEndIOSeq", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    endIOSeq = ++(iterator->GetEndIOSequence());
}

/*---------------------------------------------------------------------------*/

externPascal void kb391_GetFirstNotSavedPageOffset (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_Uint4           VAR_VALUE_REF  endOffset)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetEndPageOffset", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );



    Log_ArchiveLogManager    &log_manager=Log_ArchiveLogManager::Instance(); 

    Log_ArchiveLog             *first_log  = *log_manager.ArchiveLogsBegin();

    Log_RawDeviceIterator endIter = Log_RawDeviceIterator(  first_log->LogicalLogBeginOffset(),
                                                            iterator->GetEndPageOffset(),
                                                            first_log->LogicalLogEndOffset());
    ++endIter;
    endOffset = endIter.GetPosition();
}

/*---------------------------------------------------------------------------*/
externPascal void kb391_GetNumOfPagesLeftForSegment (
    Log_SaveIterator*                    iterator,
    tsp00_Uint4           VAR_VALUE_REF  numOfUnsavedPages)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetNumOfSavedPages", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );
    
    if (NULL == iterator)
    {
        numOfUnsavedPages = 0;
    }
    else
    {
        numOfUnsavedPages = iterator->GetNumPagesLeft();
    }
}

/*---------------------------------------------------------------------------*/
externPascal void kb391_GetNumOfPagesLeftForLogDev (
    Log_SaveIterator*                    iterator,
    tsp00_Uint4           VAR_VALUE_REF  numOfUnsavedPages)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetNumOfPagesLeftForLogDev", LogVolume_Trace, 5);

    Log_ArchiveLogManager    &log_manager=Log_ArchiveLogManager::Instance(); 
    Log_ArchiveLog           *firstLog  = *log_manager.ArchiveLogsBegin();
    
    if (NULL == iterator)
    {
        numOfUnsavedPages = firstLog->GetNumberOfUnsavedPages();
    }
    else
    {
        Log_DeviceBlockCount autosaveDistance = firstLog->GetAutosaveDistance();
        if (firstLog->GetNumberOfUnsavedPages() >= autosaveDistance)
        {
            numOfUnsavedPages = firstLog->GetNumberOfUnsavedPages()
                              - autosaveDistance
                              + iterator->GetNumPagesLeft();
        }
        else
        // the last segment to be saved
        {
            numOfUnsavedPages = iterator->GetNumPagesLeft();
        }
    }
}

/*---------------------------------------------------------------------------*/
externPascal void kb391_GetBackupCount (
    tgg00_TransContext    VAR_VALUE_REF  t,
    Log_SaveIterator*                    iterator,
    tsp00_Uint4           VAR_VALUE_REF  backupCount)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb391_GetBackupCount", LogVolume_Trace, 5);
    SAPDBERR_ASSERT_STATE( iterator != NULL );

    backupCount = iterator->GetLogBackupCount();
}

/*---------------------------------------------------------------------------*/
externPascal tsp00_PageNo kb391_ReadFirstWriterIOSequenceNo (
    tkb00_Page            VAR_VALUE_REF  page)
{
    Kernel_IPage::PageFrame frame(&page, g01page_size);
    Log_Page log_page(frame);
    return log_page.ReadFirstWriterIOSequenceNo();
}


/*---------------------------------------------------------------------------*/
externPascal tsp00_PageNo kb391_ReadLastWriterIOSequenceNo (
    tkb00_Page            VAR_VALUE_REF  page)
{
    Kernel_IPage::PageFrame frame(&page, g01page_size);
    Log_Page log_page(frame);
    return log_page.ReadLastWriterIOSequenceNo();
}

/*---------------------------------------------------------------------------*/
externPascal pasbool kb391_IsCompleteSegment ( Log_SaveIterator* iterator)
{
    if (iterator != NULL)
    {
        return iterator->IsCompleteSegment();
    }
    else
    {
        return false;
    }
}
/*---------------------------------------------------------------------------*/
externPascal void kb391Create_RTEConf_ParameterAccess (
    tgg00_TransContext    VAR_VALUE_REF  trans,
    tsp00_Addr            VAR_VALUE_REF  params)
{
    SAPDBMem_IRawAllocator& allocator    = *(reinterpret_cast<SAPDBMem_IRawAllocator*>(trans.trAllocator_gg00));
    const SAPDB_Char *      dummyDbname  = "";
    const SAPDB_Int4        dummyVersion = 0;

    if ( params != 0 )
        RTE_Crash(SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_ARGUMENT_FAILED, "params == 0"));

    RTEConf_Parameter* auxParams = new (allocator)
        RTEConf_Parameter(dummyDbname, dummyVersion, RTECONF_MAXNAMELENGTH, RTECONF_MAXSTRINGLENGTH);

    if ( auxParams == 0 )
        trans.trError_gg00 = e_no_more_memory;

    params = reinterpret_cast<tsp00_Addr>(auxParams);
}
/*---------------------------------------------------------------------------*/
externPascal void kb391AppendTo_RTEConf_ParameterAccess (
    RTEConf_Parameter&                   params,
    tkb3_xparam_page      VAR_VALUE_REF  xpPage)
{
    SAPDB_Byte * data = reinterpret_cast<SAPDB_Byte*>(&(xpPage.xp_data()));
    SAPDBErr_MessageList  error;
    if ( ! params.AddBinaryData(data, xpPage.xp_length(), error) )
    {
        RTE_Message(error);
        return;
    }
}
/*---------------------------------------------------------------------------*/
externPascal tgg00_BasisError kb391CheckParameterForRestore (
    RTEConf_Parameter&                   params)
{
    SAPDBErr_MessageList error;
    SAPDB_Bool           dataRead;

    if ( ! params.InterpretBinaryData(dataRead, error) )
    {
        RTE_Message(error);
        return e_conf_parameter_error;
    }

    // Add the checks here
    
    RTEConf_Parameter::String valueFromBackup;
    RTEConf_Parameter::String valueFromInstance;
    const SAPDB_UTF8 * paramName = UTF8("_UNICODE");

    if ( ! RTEConf_ParameterAccess::Instance()->GetString(paramName, valueFromInstance, error) ) 
    {
        RTE_Message(error);
        return e_conf_parameter_error;
    }

    if ( ! params.GetReference(paramName, valueFromBackup, error) ) 
    {
        RTE_Message(error);
        return e_conf_parameter_error;
    }
    
    if ( valueFromInstance[0] != valueFromBackup[0] )
    {
        RTE_Message(SAPDBErr_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,"_UNICODE must be equal"));
        return e_wrong_configuration;
    }

    return e_ok;
}
/*---------------------------------------------------------------------------*/
externPascal void kb391Destroy_RTEConf_ParameterAccess (
    tgg00_TransContext    VAR_VALUE_REF  trans,
    tsp00_Addr            VAR_VALUE_REF  params)
{
    SAPDBMem_IRawAllocator& allocator = *(reinterpret_cast<SAPDBMem_IRawAllocator*>(trans.trAllocator_gg00));
    destroy(params, allocator);
    params = 0;
}
