/*!***************************************************************************

  module      : vkb560.cpp
  responsible : UweH
  special area: KB_Logging
  last changed: 2000-03-15
  copyright   : Copyright by SAP AG, 2000
  see also    :
  description : interface to new Logging component



    ========== licence begin  GPL
    Copyright (C) 2000 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "heo47.h"   // sql47_ptoc
#include "heo56.h"   // vsleep
#include "heo58.h"
#include "hgg01.h"   // g01is_livecache
#include "hgg01_3.h" // g01pid
#include "hgg11.h"   // g11ParseVersion      PTS 1122558 FF 11-JUN-200
#include "hbd91.h"   // bd91StartOMSGarbageCollection
#include "hkb57_1.h" // k57restartrec
#include "hkb57.h"   // save_restart_rec
#include "SAPDBCommon/SAPDB_ToString.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "KernelCommon/Kernel_DateTime.hpp"
#include "KernelCommon/Kernel_Trace.hpp"
#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "Logging/Log_ArchiveLogManager.hpp"
#include "Logging/Log_Transaction.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "Logging/Log_History.hpp"
#include "Logging/Log_Exceptions.hpp"
#include "IOManager/IOMan_ILogInfo.hpp"

#include "hkb560.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *   LOCAL FUNCTIONS (PROTOTYPES)                                            *
 *===========================================================================*/

/*-----------------------------------------------------------------------------------*/

static inline Log_SavepointReason MapToSavepointReason (tgg00_MessType2_Param MessType2)
{
    switch (MessType2)
    {
    case mm_database:
        return Log_SVPReasonSaveData;
    case mm_pages:
        return Log_SVPReasonSavePages;
    case mm_down:
        return Log_SVPReasonShutdown;
    case mm_end_read_only:
        return Log_SVPReasonEndReadOnly;
    case mm_read:
        return Log_SVPReasonEndLoad;
    case mm_user:
        return Log_SVPReasonUserRequested;
    case mm_index:
        return Log_SVPReasonCreateIndex;
    case mm_test:
        return Log_SVPReasonCheckTable;
    case mm_config:
        return Log_SVPReasonCatalogMigration;
    case mm_lock:
        return Log_SVPReasonBDLockList;
    default:
        return Log_SVPReasonUnknown;
    }
}

/*-----------------------------------------------------------------------------------*/

static inline Log_Transaction&
GetLogTrans (tgg00_TransContext &Trans,
                   const char *        Msg)
{
    Log_Transaction* LogTrans = REINTERPRET_CAST(Log_Transaction*,Trans.trLogContext_gg00);
    if ( LogTrans == NULL )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED, Msg) );
    return *LogTrans;
}

/*===========================================================================*
 *  GLOBAL FUNCTIONS                                                         *
 *===========================================================================*/

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560ResetStatistics (tsp00_TaskId taskid)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560ResetStatistics", LogVolume_Trace, 5);
    Log_ArchiveLogManager &alman   = Log_ArchiveLogManager::Instance();
    Log_ArchiveLog        &archlog = alman.GetOneArchiveLog(0);
    archlog.ResetStatistics(taskid);
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560GetStatistics (tsp00_TaskId         taskId,
                    tkb00_LogDeviceInfo &LogDeviceInfo,
                    tkb00_LogQueueInfo  &LogQueueInfo)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetStatistics", LogVolume_Trace, 5);

    LogDeviceInfo.ldiPagesUsed_kb00     = 0;
    LogDeviceInfo.ldiPagesNotSaved_kb00 = 0;
    LogDeviceInfo.ldiPagesSinceBup_kb00 = 0;
    LogDeviceInfo.ldiSaveptCount_kb00   = 0;

    LogQueueInfo.lqiCountGroup_kb00      = 0;
    LogQueueInfo.lqiCountInsert_kb00     = 0;
    LogQueueInfo.lqiCountRequest_kb00    = 0;
    LogQueueInfo.lqiCountWait_kb00       = 0;
    LogQueueInfo.lqiAvgSumPagewaits_kb00 = 0;
    LogQueueInfo.lqiLogQueueSize_kb00    = 0;
    LogQueueInfo.lqiMaxQueueUsed_kb00    = 0;
    LogQueueInfo.lqiMaxWaitLogpage_kb00  = 0;

    pasbool     available;
    SAPDB_UInt4 deviceSize;
    SAPDB_UInt4 auxUsed;
    SAPDB_UInt4 auxNotSaved;

    kb560GetArchiveLogDeviceInfo ( taskId, 0, available,
                                   deviceSize,
                                   auxUsed,
                                   auxNotSaved );
                                   
    LogDeviceInfo.ldiPagesUsed_kb00     = auxUsed;
    LogDeviceInfo.ldiPagesNotSaved_kb00 = auxNotSaved;
    LogDeviceInfo.ldiSaveptCount_kb00   = Log_Savepoint::GetLastSavepointId();

    if ( available == 0 )
        return;
        
    SAPDB_UInt4 auxWaitForQueueFree;
    SAPDB_UInt4 auxQueueSize;
    SAPDB_UInt4 auxMaxQueueUsed;
    SAPDB_UInt4 auxCountRequest;
    SAPDB_UInt4 auxCountWaitForIO;
    SAPDB_UInt4 auxMaxWait;
    SAPDB_UInt4 auxCountGroupCommit;
    
    kb560GetArchiveLogQueueInfo ( taskId, 0, available,
                                  auxQueueSize,
                                  auxMaxQueueUsed,
                                  auxCountRequest,
                                  auxCountWaitForIO,
                                  auxCountGroupCommit,
                                  auxWaitForQueueFree,
                                  auxMaxWait );

    LogQueueInfo.lqiLogQueueSize_kb00   = auxQueueSize;
    LogQueueInfo.lqiMaxQueueUsed_kb00   = auxMaxQueueUsed;
    LogQueueInfo.lqiCountInsert_kb00    = auxCountRequest;
    LogQueueInfo.lqiCountWait_kb00      = auxCountWaitForIO;
    LogQueueInfo.lqiCountRequest_kb00   = auxWaitForQueueFree;
    LogQueueInfo.lqiMaxWaitLogpage_kb00 = auxMaxWait;

    LogQueueInfo.lqiCountGroup_kb00     = auxCountGroupCommit;
    if ( auxCountRequest > 0 )
        LogQueueInfo.lqiAvgSumPagewaits_kb00 = auxCountWaitForIO / auxCountRequest;
} 

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560GetArchiveLogDeviceInfo (tsp00_TaskId   taskId,
                              SAPDB_UInt4    logno,
                              pasbool       &available,
                              SAPDB_UInt4   &deviceSize,
                              SAPDB_UInt4   &pagesUsed,
                              SAPDB_UInt4   &pagesNotSaved)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetArchiveLogDeviceInfo", LogVolume_Trace, 5);

    available             = 1;
    deviceSize            = 0;
    pagesUsed             = 0;
    pagesNotSaved         = 0;

    if ( logno >= 1 )
    {
        available = 0;
        return;
    }

    Log_ArchiveLogManager &alman   = Log_ArchiveLogManager::Instance();
    Log_ArchiveLog        &archlog = alman.GetOneArchiveLog(logno);
    
    archlog.GetDeviceUsage (deviceSize, pagesUsed, pagesNotSaved);
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560GetArchiveLogQueueInfo (tsp00_TaskId    taskId,
                             SAPDB_UInt4     logno,
                             pasbool        &available,
                             SAPDB_UInt4    &queueSize,
                             SAPDB_UInt4    &maxPagesUsedInQueue,
                             SAPDB_UInt4    &countReserveRequests,
                             SAPDB_UInt4    &countWaitForIO,
                             SAPDB_UInt4    &auxCountGroupCommit,
                             SAPDB_UInt4    &countWaitForQueueFree,
                             SAPDB_UInt4    &maxWaitForPageIO)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetArchiveLogQueueInfo", LogVolume_Trace, 5);

    available             = 1;
    queueSize             = 0;
    maxPagesUsedInQueue   = 0;
    countReserveRequests  = 0;
    countWaitForIO        = 0;
    countWaitForQueueFree = 0;
    maxWaitForPageIO      = 0;

    if ( logno >= 1 )
    {
        available = 0;
        return;
    }
    Log_ArchiveLogManager &alman   = Log_ArchiveLogManager::Instance();
    Log_ArchiveLog        &archlog = alman.GetOneArchiveLog(logno);
    
    archlog.GetQueueUsage ( queueSize,
                            maxPagesUsedInQueue,
                            countReserveRequests,
                            countWaitForIO,
                            auxCountGroupCommit,
                            countWaitForQueueFree,
                            maxWaitForPageIO );
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560GetRestartVersion (
    tsp00_TaskId                         taskid,
    tsp00_Int4            VAR_VALUE_REF  firstiosequence,
    tsp00_Line            VAR_ARRAY_REF  dbident,
    pasbool               VAR_VALUE_REF  log_devspace_enabled)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetRestartVersion", LogVolume_Trace, 5);
    Log_ArchiveLogManager &alman   = Log_ArchiveLogManager::Instance();
    Log_ArchiveLog        &archlog = alman.GetOneArchiveLog(0);
    Log_IOSequenceNo       auxfirstiosequence;
    SAPDB_Char            *auxdbident = reinterpret_cast<SAPDB_Char*>(&dbident);

    archlog.GetRestartVersion (taskid, auxfirstiosequence, auxdbident);
    // The cleared log device:
    // FIRST_USED_LPNO: intern "invalid"/MAX_UINT4 => 0
    // FIRST_LOG_LPNO:   intern "invalid"/MAX_UINT4 => MAX_INT4 (nil_page)
    // LAST_LOG_LPNO:    intern "invalid"/MAX_UINT4 => MAX_INT4 (nil_page)

    firstiosequence = auxfirstiosequence;

    log_devspace_enabled =  archlog.IsDevspaceEnabled();
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560GetLogDBIdent (             // PTS 1113550 mb 2002-03-22 new
    tsp00_Line            VAR_ARRAY_REF  dbident)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetLogDBIdent", LogVolume_Trace, 5);

    Log_ArchiveLogManager &alman   = Log_ArchiveLogManager::Instance();
    Log_ArchiveLog        &archlog = alman.GetOneArchiveLog(0);
    SAPDB_Char            *auxdbident = reinterpret_cast<SAPDB_Char*>(&dbident);

    archlog.GetLogDBIdent (auxdbident);
}    

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560DumpTransaction (Log_Transaction *pTrans,
                      void*           &pData,
                      SAPDB_UInt4     &dataLength)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560DumpTransaction", LogTrans_Trace, 5);
    Log_Transaction::GetDumpInfo (pTrans, pData, dataLength);
}
                            
/*-----------------------------------------------------------------------------------*/
externPascal pasbool
kb560VerifyOpenTransAndHistory (tgg00_TransContext &trans,
                                pasbool             isCold)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560VerifyOpenTransAndHistory", LogTrans_Trace, 5);
    bool result = Log_Savepoint::Verify (trans, isCold == 1);
    if ( g01is_livecache() == 1 )
    {
        result &= Log_History::GetInstance().Verify(trans, isCold == 1);
    }
    return result;
}
                            
/*-----------------------------------------------------------------------------------*/
externPascal void
kb560CreateHistoryManager (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CreateHistoryManager", LogHistory_Trace, 5);
    if ( ! Log_History::CreateInstance (Trans) )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "Log_History::CreateInstance() failed") );
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560DropHistoryManager ()
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560DropHistoryManager", LogHistory_Trace, 5);
	Log_History::DropInstance();
}

/*-----------------------------------------------------------------------------------*/
externPascal tsp00_Int4 kb560GetNumOfUserUKTs ()
{
    tsp0058_RTEInfo RTEInfo;

    vgetrteinfo (RTEInfo);

    return RTEInfo.NumOfUserUKTs;
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560FreeLog (
    tsp00_TaskId                         taskId,
    tsp00_Uint4                          oldLastSavedIOsequence,
    tsp00_Uint4                          oldLastSavedOffset,
    tsp00_Uint4                          newLastSavedIOsequence,
    tsp00_Uint4                          newLastSavedOffset)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560FreeLog", LogVolume_Trace, 5);
    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();
    if ( ! alman.FreeLog (taskId, oldLastSavedIOsequence, oldLastSavedOffset,
                                  newLastSavedIOsequence, newLastSavedOffset) )
    {
        alman.WriteToTrace();
        RTE_Crash( Log_Exception
                   (__FILE__, __LINE__,
                    LOG_FREE, SAPDB_ToString(oldLastSavedIOsequence,_T_d),
                              SAPDB_ToString(newLastSavedIOsequence,_T_d),
                              SAPDB_ToString(oldLastSavedOffset,_T_d),
                              SAPDB_ToString(newLastSavedOffset,_T_d)) );
    }
}
/*-----------------------------------------------------------------------------------*/
externPascal void kb560ExecuteFreeLogForPipe (                    // PTS 1114791 mb 2002-08-22 new
    tsp00_TaskId                         taskId,
    tsp00_Uint4                          firstSavedIOsequence,
    tsp00_Uint4                          lastSavedIOsequence,
    tgg00_BasisError      VAR_VALUE_REF  trError)
{
    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());

    Log_ArchiveLog::PipeHandlingResult result;
    result = log.ExecuteFreeLogForPipe(taskId, firstSavedIOsequence, lastSavedIOsequence);
    switch (result)
    {
        case Log_ArchiveLog::pipeOk:
            trError = e_ok;
            break;
        case Log_ArchiveLog::IOSequenceDoesNotMatch:
            trError = e_invalid;
            break;
        default:
            trError = e_log_error;
    }
}

/*-----------------------------------------------------------------------------------*/
externPascal pasbool kb560TransactionIsOpen (void* TransPtr)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560TransactionIsOpen", LogTrans_Trace, 5);
    if ( NULL == TransPtr )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560TransactionIsOpen(): TransPtr is null") );
	// PTS 1114125 UH 2002-02-12 use of IsOpenInRedo()
    Log_Transaction &trans = *(reinterpret_cast<Log_Transaction*>(TransPtr));
    return trans.IsOpenInRedo();
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560InitSavepointParam (tkb00_SaveptParam &param)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560InitSavepointParam", LogTrans_Trace, 5);

    param.svpId_kb00               = 0;
    param.svpReason_kb00           = Log_SVPReasonUnknown;
    param.svpFiller1_kb00          = 0;
    param.svpOpenTransCnt_kb00     = 0;

    param.svpIOsequence_kb00       = SAPDB_MAX_UINT4;
    param.svpStartOffset_kb00      = SAPDB_MAX_UINT4;

    param.svpStartEntryOffset_kb00 = 0;
    param.svpStartEntryType_kb00   = Log_NoOp;

    param.svpFiller2_kb00          = 0;
    param.svpFiller3_kb00          = 0;

    param.svpOpenTransRoot_kb00    = NIL_PAGE_NO_GG00;
    param.svpHistoryRoot_kb00      = NIL_PAGE_NO_GG00;

    param.svpDate_kb00             = KERNEL_INVALID_DATE;
    param.svpTime_kb00             = KERNEL_INVALID_TIME;

    param.svpStartEntryDate_kb00   = KERNEL_INVALID_DATE;
    param.svpStartEntryTime_kb00   = KERNEL_INVALID_TIME;

    memset (&param.svpFiller4_kb00, 0, sizeof(param.svpFiller4_kb00));

    param.svpFiller5_kb00          = 0;
    param.svpFiller6_kb00          = 0;
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560StartSavepoint (tgg00_TransContext    &Trans,
                     tgg00_MessType2_Param  MessType2)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560StartSavepoint", LogTrans_Trace, 5);
    Log_SavepointManager.StartSavepoint (Trans, MapToSavepointReason(MessType2));
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560StartSavepointAndWait (tgg00_TransContext    &Trans,
                            tgg00_MessType2_Param  MessType2)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560StartSavepointAndWait", LogTrans_Trace, 5);
    Log_SavepointManager.StartSavepointAndWait (Trans, MapToSavepointReason(MessType2));
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560SavepointCoordinator (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560SavepointCoordinator", LogTrans_Trace, 5);
    Log_SavepointManager.Coordinator (Trans);
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560AllocateArchiveLogManager (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560AllocateArchiveLogManager", LogVolume_Trace, 5);
    Log_ArchiveLogManager::CreateInstance
                            (RTEMem_Allocator::Instance(),
                             Log_FrameAllocator::Instance());   // !!!provisorisch!!!

    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();

    /*
       for the parameters of AddArchiveLog() and InitArchiveLog() 
       (see kb560InitArchiveLogManager() below) there are two constellations:

       1.)    Log_Queue::Unsynchronized, Log_Queue::MinimizeSpace,  1 (single!) log queue
       2.)    Log_Queue::Synchronized,   Log_Queue::MaximizeSafety, any number of queues

       (Test) Log_Queue::Synchronized,   Log_Queue::MinimizeSpace,  1 log queue
    */
    
    // 7.4: add one single archive log:
    if (!alman.AddArchiveLog(TaskId,
                             g01pid.taskLogWriter_gg00, 
                             Log_Queue::Synchronized,     // or Unsynchronized
                             Log_DeviceMinimizeSpace))   // or MaximizeSafety
    {
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560CreateArchiveLogManager failed") );
    }

    /* 
    >>> Future version with may archive logs (draft):

    IOMan_Manager::ArchiveLogIterator iter = ioman.ArchiveLogsBegin();
    IOMan_Manager::ArchiveLogIterator end  = ioman.ArchiveLogsEnd();
    while (iter != end)
    {
        if (!alman.AddArchiveLog((*iter).GetWriterTaskID(), 
                                 Log_Queue::Synchronized,
                                 Log_Queue::MaximizeSafety)
        {
            // FATAL: failed to create archive log
        }
        ++iter;
    }
    */
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560InitLogForNewDatabase (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560InitLogForNewDatabase", LogVolume_Trace, 5);

    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();

    const IOMan_BlockCount  LogCapacity = IOMan_ILogInfo::GetInstance().TotalUsableLogPages();

    // 7.4: add one single archive log:
    Log_ErrorState result = 
        alman.CreateArchiveLogForNewDatabase
            (TaskId,
             g01pid.taskLogWriter_gg00,
             0, LogCapacity - 1, g01autosave_distance,
             1, // no of log queues [or g01kbfunction_tasks()]
             g01cache_size (cachLogQueue_egg00), // page frames per log queue
             g01io_block_count());

    if (result != Log_Ok)
    {
        alman.WriteToTrace();
        RTE_Crash( Log_Exception
                   (__FILE__, __LINE__,
                    LOG_INIT_ERROR, Log_ErrorStateStrings[result]) );
    }
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560InitLogForRestart (tsp00_TaskId      TaskId,
                        tgg00_BasisError &TrError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560InitLogForRestart", LogVolume_Trace, 5);
    
    TrError = e_ok;
    
    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();

    const IOMan_BlockCount  LogCapacity = IOMan_ILogInfo::GetInstance().TotalUsableLogPages();

    // 7.4: add one single archive log:
    Log_ErrorState result = 
        alman.RestartArchiveLog
            (TaskId,
             g01pid.taskLogWriter_gg00, 
             0, LogCapacity - 1, g01autosave_distance,
             g01cache_size (cachLogQueue_egg00), // page frames per log queue
             g01io_block_count());

    if ( result != Log_Ok )
    {
        alman.WriteToTrace();
        if ( Log_LogAndDataIncompatible    == result 
             ||
             Log_SavepointPositionNotFound == result )
        {
            TrError = e_incompatible_log;
            RTE_Message( Log_Exception
                         (__FILE__, __LINE__,
                         LOG_RESTART_ERROR, Log_ErrorStateStrings[result]) );
        }
        else
            RTE_Crash( Log_Exception
                       (__FILE__, __LINE__,
                        LOG_RESTART_ERROR, Log_ErrorStateStrings[result]) );
    }
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560InitLogForRestore (
    tsp00_TaskId                         TaskId,
    tgg00_BasisError      VAR_VALUE_REF  TrError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560InitLogForRestore", LogVolume_Trace, 5);
    
    TrError = e_ok;
    
    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();

    const IOMan_BlockCount  LogCapacity = IOMan_ILogInfo::GetInstance().TotalUsableLogPages();

    // 7.4: add one single archive log:
    Log_ErrorState result = 
        alman.RestartForRestoreLog
            (TaskId,
             g01pid.taskLogWriter_gg00, 
             0, LogCapacity - 1, g01autosave_distance,
             g01cache_size (cachLogQueue_egg00), // page frames per log queue
             g01io_block_count());

    if ( result != Log_Ok )
    {
        alman.WriteToTrace();
        if ( Log_LogAndDataIncompatible    == result 
             ||
             Log_SavepointPositionNotFound == result )
        {
            TrError = e_incompatible_log;
            RTE_Message( Log_Exception
                         (__FILE__, __LINE__,
                         LOG_RESTART_ERROR, Log_ErrorStateStrings[result]) );
        }
        else
            RTE_Crash( Log_Exception
                       (__FILE__, __LINE__,
                        LOG_RESTART_ERROR, Log_ErrorStateStrings[result]) );
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560InitLogForLogSave (
    tsp00_TaskId                         TaskId,
    tgg00_BasisError      VAR_VALUE_REF  TrError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560InitLogForRestart", LogVolume_Trace, 5);
    
    TrError = e_ok;
    
    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();

    const IOMan_BlockCount  LogCapacity = IOMan_ILogInfo::GetInstance().TotalUsableLogPages();

    // 7.4: add one single archive log:
    Log_ErrorState result = 
        alman.RestartForSaveLogCold( TaskId,
                                     g01pid.taskLogWriter_gg00, 
                                     0, LogCapacity - 1, g01autosave_distance);
                                                                 

    if ( result != Log_Ok )
    {
        alman.WriteToTrace();
        if ( Log_LogAndDataIncompatible == result )
        {
            TrError = e_incompatible_log;
            RTE_Message( Log_Exception
                         (__FILE__, __LINE__,
                         LOG_RESTART_ERROR, Log_ErrorStateStrings[result]) );
        }
        else
            RTE_Crash( Log_Exception
                       (__FILE__, __LINE__,
                        LOG_RESTART_ERROR, Log_ErrorStateStrings[result]) );
    }
}

/*-----------------------------------------------------------------------------------*/
externPascal void
kb560ClearArchiveLog (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560ClearArchiveLog", LogVolume_Trace, 5);
    
    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();

    const IOMan_BlockCount  LogCapacity = IOMan_ILogInfo::GetInstance().TotalUsableLogPages();

    Log_ErrorState result = 
        alman.ClearArchiveLog
            (TaskId,
             g01pid.taskLogWriter_gg00, 
             0, LogCapacity - 1, g01autosave_distance,
             g01cache_size (cachLogQueue_egg00), // page frames per log queue
             g01io_block_count());

    if (result != Log_Ok)
    {
        alman.WriteToTrace();
        RTE_Crash( Log_Exception
                   (__FILE__, __LINE__,
                    LOG_CLEAR_ERROR, Log_ErrorStateStrings[result]) );
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560LogWriter (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560LogWriter", LogVolume_Trace, 5);
    Log_ArchiveLogManager& alman = Log_ArchiveLogManager::Instance();
    Log_Writer& writer = alman.GetWriter(TaskId);
    writer.Run();
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560CreateLogTransaction (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CreateLogTransaction", LogTrans_Trace, 5);

    SAPDBMem_IRawAllocator& Allocator =
        *(REINTERPRET_CAST (SAPDBMem_IRawAllocator*, Trans.trAllocator_gg00));

    Log_Transaction* LogTrans = new (Allocator) Log_Transaction ();

    if ( LogTrans == NULL )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560CreateLogTransaction: allocate") );

    if ( ! LogTrans->Initialize(Trans) )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560CreateLogTransaction: init") );
    
    Trans.trLogContext_gg00 = LogTrans;
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560DeleteLogTransaction (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560DeleteLogTransaction", LogTrans_Trace, 5);

    Log_Transaction* LogTrans = REINTERPRET_CAST(Log_Transaction*,Trans.trLogContext_gg00);
    if ( LogTrans == NULL )
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560DeleteLogTransaction: pointer not valid") );

    SAPDBMem_IRawAllocator& allocator = LogTrans->Allocator();
    LogTrans->Delete();
    destroy(LogTrans, allocator);
    Trans.trLogContext_gg00 = NULL;
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560CommitLogTransaction (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CommitLogTransaction", LogTrans_Trace, 5);
    GetLogTrans (Trans, "CommitLogTrans: ptr=nil ").Commit();
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560RollbackLogTransAction (tgg00_TransContext    &Trans,
                             tgg00_MessType2_Param  RollbackKind,
                             tgg00_SubtransNo       SubtransNo)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560RollbackLogTransaction", LogTrans_Trace, 5);

    Log_Transaction::RollbackKind newRollbackKind;
    switch ( RollbackKind )
    {
    case mm_rollback:
        newRollbackKind = Log_Transaction::endOfTrans;
        break;
    case mm_nil:
        newRollbackKind = Log_Transaction::command;
        break;
    case mm_subtrans:
        newRollbackKind = Log_Transaction::subTrans;
        break;
    default:
        SAPDBTRACE_WRITELN (LogTrans_Trace, 6, "RollbackKind: " << RollbackKind);
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560RollbackLogTransAction: unknown RollbackKind") );
    }

    Trans.trState_gg00.addElement (tsRollback_egg00); // PTS XXXX TS 2003-04-11

    Log_Transaction::RollbackResult Result =
        GetLogTrans (Trans, "RollbackLogTrans:ptr=nil").
            Rollback (newRollbackKind, SubtransNo);

    Trans.trState_gg00.delElement (tsRollback_egg00); // PTS XXXX TS 2003-04-11

    Trans.trError_gg00 = kb560ConvertRollbackResultToBasisError (Result);
}

/*-----------------------------------------------------------------------------------*/

externPascal tgg00_BasisError kb560ConvertRollbackResultToBasisError (
    tsp00_Int4                           result)
{
    switch ( Log_Transaction::RollbackResult(result) )
    {
    case Log_Transaction::undoOk:
        return e_ok;
    case Log_Transaction::undoNoOpenSubtrans:
        return e_no_open_subtrans;
    case Log_Transaction::undoLogFull:
        return e_log_full;
    default:
        // case undoReadError, undoWrongEntry, undoExecuteError, undoWriteError
        return e_log_error;
    }
}

/*-----------------------------------------------------------------------------------*/
externPascal tgg00_BasisError kb560ConvertWriteRedoEntryResultToBasisError (
    tsp00_Int4                           result)
{
    switch ( Log_Transaction::LogState(result) )
    {
    case Log_Transaction::noError:
        return e_ok;
    case Log_Transaction::logFull:
        return e_log_full;
    case Log_Transaction::logError:
        return e_log_error;
    default:
        return e_log_error;
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560BeginSubtransLogTransaction (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560BeginSubtransLogTransaction", LogTrans_Trace, 5);
    GetLogTrans (Trans, "BeginSubtLogTrns:ptr=nil").BeginSubtrans();
}

/*-----------------------------------------------------------------------------------*/

externPascal void
kb560EndSubtransLogTransaction (tgg00_TransContext &Trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560EndSubtransLogTransaction", LogTrans_Trace, 5);
    if ( ! GetLogTrans (Trans, "EndSubtrLogTrans:ptr=nil").EndSubtrans() )
        Trans.trError_gg00 = e_no_open_subtrans;
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560GetFirstKnownIOSeqBeforeRestoreLog (
    tsp00_PageNo          VAR_VALUE_REF  FirstIOSeq,
    pasbool               VAR_VALUE_REF  firstIOSeqIsValid)
{
    Log_ArchiveLogManager    &log_manager=Log_ArchiveLogManager::Instance(); 
    Log_ArchiveLog           *first_log  = *log_manager.ArchiveLogsBegin();
    
    Log_IOSequenceNo        ioSeq;
    Log_RawDeviceOffset     offset;
    first_log->GetFirstKnownBeforeRestoreLog(offset,ioSeq);
    
    FirstIOSeq = ioSeq;
    firstIOSeqIsValid = ioSeq.IsValid();
}    

/*-----------------------------------------------------------------------------------*/

externPascal pasbool
kb560GetLogInfo ( tsp00_TaskId   taskid,
                  tsp00_Int4    &date,
                  tsp00_Int4    &time,
                  tsp00_Int4    &flushmode,
                  tsp00_Int4    &devicestate,
                  tsp00_Uint4   &queuecount,
                  tsp00_Uint4   &cycleendoffset,
                  tsp00_Uint4   &oldestnotsavedoffset,
                  tsp00_Uint4   &oldestnotsavedsequence,
                  tsp00_Uint4   &lastknownoffset,
                  tsp00_Uint4   &lastknownsequence,
                  tsp00_Uint4   &backupcount,
                  tsp00_Int4    &untildate,
                  tsp00_Int4    &untiltime,
                  tsp00_Uint4   &untilclearloweroffset,
                  tsp00_Uint4   &untilclearupperoffset,
                  tsp00_Uint4   &untilsequenceafterredo,
                  tsp00_Uint4   &savepointsequence,
                  tsp00_Uint4   &savepointdevoffset,
                  tsp00_Uint4   &savepointoffset )
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetLogInfo", LogVolume_Trace, 5);
    Log_ArchiveLogManager &alman   = Log_ArchiveLogManager::Instance();
    Log_ArchiveLog        &archlog = alman.GetOneArchiveLog(0);

    Kernel_Date              auxdate;
    Kernel_Time              auxtime;
    Log_FlushMode            auxflushmode;
    Log_DeviceState          auxdevicestate;
    Log_InfoPage::QueueCount auxqueuecount;
    Log_RawDeviceOffset      auxcycleendoffset;
    Log_RawDeviceOffset      auxoldestnotsavedoffset;
    Log_IOSequenceNo         auxoldestnotsavedsequence;
    Log_RawDeviceOffset      auxlastknownoffset;
    Log_IOSequenceNo         auxlastknownsequence;
    SAPDB_UInt4              auxbackupcount;
    Kernel_Date              auxuntildate;
    Kernel_Time              auxuntiltime;
    Log_RawDeviceOffset      auxuntilclearloweroffset;
    Log_RawDeviceOffset      auxuntilclearupperoffset;
    Log_IOSequenceNo         auxuntilsequenceafterredo;
    Log_IOSequenceNo         auxsavepointsequence;
    Log_RawDeviceOffset      auxsavepointdevoffset;
    Log_InfoPage::PageOffset auxsavepointoffset;

    bool isvalid = archlog.GetLogInfo ( taskid,
                         auxdate,
                         auxtime,
                         auxflushmode,
                         auxdevicestate,
                         auxqueuecount,
                         auxcycleendoffset,
                         auxoldestnotsavedoffset,
                         auxoldestnotsavedsequence,
                         auxlastknownoffset,
                         auxlastknownsequence,
                         auxbackupcount,
                         auxuntildate,
                         auxuntiltime,
                         auxuntilclearloweroffset,
                         auxuntilclearupperoffset,
                         auxuntilsequenceafterredo,
                         auxsavepointsequence,
                         auxsavepointdevoffset,
                         auxsavepointoffset );
    
    if ( ! isvalid )
        return false;
        
    date                   = auxdate;
    time                   = auxtime;
    flushmode              = auxflushmode;
    devicestate            = auxdevicestate;
    queuecount             = auxqueuecount;
    cycleendoffset         = auxcycleendoffset;
    oldestnotsavedoffset   = auxoldestnotsavedoffset;
    oldestnotsavedsequence = auxoldestnotsavedsequence;
    lastknownoffset        = auxlastknownoffset;
    lastknownsequence      = auxlastknownsequence;
    backupcount            = auxbackupcount;
    untildate              = auxuntildate;
    untiltime              = auxuntiltime;
    untilclearloweroffset  = auxuntilclearloweroffset;
    untilclearupperoffset  = auxuntilclearupperoffset;
    untilsequenceafterredo = auxuntilsequenceafterredo;
    savepointsequence      = auxsavepointsequence;
    savepointdevoffset     = auxsavepointdevoffset;
    savepointoffset        = auxsavepointoffset;
    
    return true;
}

/*-----------------------------------------------------------------------------------*/

externPascal pasbool kb560CheckLogFull (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CheckLogFull", LogVolume_Trace, 5);

    if (g01pid.taskUtility_gg00 != TaskId)
        return kb560LogIsFull(TaskId);
    else
        return false;
}

/*-----------------------------------------------------------------------------------*/
externPascal pasbool kb560LogIsFull (tsp00_TaskId TaskId)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560LogIsFull", LogVolume_Trace, 5);

    Log_ArchiveLogManager &log_manager =  Log_ArchiveLogManager::Instance(); 
    Log_ArchiveLog        *first_log   = *log_manager.ArchiveLogsBegin();
    
    return first_log->LogIsFull(TaskId);
}


/*-----------------------------------------------------------------------------------*/

externPascal pasbool kb560SupressDBIdentCheckForMigration (    // PTS 1116030 mb 2002-06-03 new
    tsp00_Version         VAR_ARRAY_REF  backupVersion)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560SupressDBIdentCheckForMigration", LogVolume_Trace, 5);

    SAPDB_Int4  majorVersion;
    SAPDB_Int4  minorVersion;
    SAPDB_Int4  correctionLevel;
    SAPDB_Int4  buildNo;
    
    /* PTS 1122558 FF 11-JUN-2003 */
    g11ParseVersion (backupVersion, majorVersion, minorVersion, correctionLevel, buildNo);

    /* Begin PTS 1122558 FF 11-JUN-2003 
    char *auxVersion = reinterpret_cast <char *> (&(backupVersion[BEG_OF_1ST_VERS_SUBSTR_SP100]));
    sscanf (auxVersion , "%d.%d.%d", &majorVersion, &minorVersion, &correctionLevel);
    auxVersion = reinterpret_cast <char *> (&(backupVersion[BEG_OF_2CND_VERS_SUBSTR_SP100]));
    sscanf (auxVersion , "%d-",    &buildNo                                );
    * end PTS 1122558 FF 11-JUN-2003 */
    
    // Migration: DBIdentifier may not be correct initialized in 7.4.1 or 7.4.2 with build > 7
    if (  ( 7 == majorVersion  && 4 == minorVersion && 1 == correctionLevel )
          ||
          ( ( 7 == majorVersion  && 4 == minorVersion && 2 == correctionLevel )
            &&
            (buildNo < 8)))
    {
        if ( LogVolume_Trace.TracesLevel(6) )
            Kernel_VTrace() << "kb560CheckLogTapeAndDataCompatibility: wrong DBIDENT ignored due to kernel version of log-backup";
        return true;
    }
    else
    {
        return false;
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560CheckLogTapeAndDataCompatibility (
    tgg00_TransContext    VAR_VALUE_REF  trans,
    const tkb00_Page&                    restartRecord,
    tsp00_Version         VAR_ARRAY_REF  backupVersion,                  // PTS 1116030 mb 2002-06-03
    tsp00_Line            VAR_ARRAY_REF  dbident,
    tsp00_PageNo                         firstIOSeq,
    tsp00_PageNo                         lastIOSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CheckLogTapeAndDataCompatibility", LogVolume_Trace, 5);

    if (Log_IOSequenceNo(restartRecord.rstLastSavept_kb00().svpIOsequence_kb00).IsInvalid() &&     //PTS1113550 mb 2002-05-06
        0 == firstIOSeq)     //PTS1113550 mb 2002-05-06
    // for a restart from the Log_IOSequence 0 into an empty DB: Copy DB-Identifier
    {
        k57restartrec->rstDbIdent_kb00() = dbident;
        k57save_restartrecord (trans.trTaskId_gg00);

        Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
        log.SetLogDBIdent (trans.trTaskId_gg00,dbident);
    }
    else
    {
        const Log_IOSequenceNo svpIOSeq = restartRecord.rstLastSavept_kb00().svpIOsequence_kb00;
        
        if ( svpIOSeq.IsInvalid() && firstIOSeq != 0 )
        {
            // RESTORE LOG directly after INIT CONFIG without RESTORE DATA
            trans.trError_gg00 = e_incompatible_log;
            return;
        }
    
        /* PTS 1113550 mb 2002-22-03 */
        if (memcmp(dbident, &(restartRecord.rstDbIdent_kb00()), sizeof(tsp00_Line)))
        {
            if (!kb560SupressDBIdentCheckForMigration(backupVersion))
            {
                trans.trError_gg00 = e_invalid_label;
                return;
            }
        }

        if ( svpIOSeq.IsValid() && ((svpIOSeq < SAPDB_UInt4(firstIOSeq)) || (svpIOSeq > SAPDB_UInt4(lastIOSeq))) )
        {
            trans.trError_gg00 = e_incompatible_log;
            return;
        }
    }
    trans.trError_gg00 = e_ok;
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560NewLogSize (tsp00_Int4 NewLogSize)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560NewLogSize", LogVolume_Trace, 5);

    if ( NewLogSize <= 0 )
    {
        Kernel_VTrace() << "kb560NewLogSize: Invalid new logsize: " << NewLogSize;
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "kb560NewLogSize: Invalid new logsize ") );
    }
    Log_ArchiveLogManager::Instance().GetOneArchiveLog(0).SetNewDeviceEndOffset(NewLogSize - 1);
}

/*-----------------------------------------------------------------------------------*/

externPascal void k560check_logsave_allowed (
    tsp00_TaskId                         TaskId,
    tgg00_BasisError      VAR_VALUE_REF  BasisError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("k560check_logsave_allowed", LogVolume_Trace, 5);

    Kernel_Date              date;
    Kernel_Time              time;
    Log_FlushMode            flushmode;
    Log_DeviceState          devicestate;
    Log_InfoPage::QueueCount queuecount;
    Log_RawDeviceOffset      cycleendoffset;
    Log_RawDeviceOffset      oldestnotsavedoffset;
    Log_IOSequenceNo         oldestnotsavedsequence;
    Log_RawDeviceOffset      lastknownoffset;
    Log_IOSequenceNo         lastknownsequence;
    SAPDB_UInt4              backupcount;
    Kernel_Date              untildate;
    Kernel_Time              untiltime;
    Log_RawDeviceOffset      untilclearloweroffset;
    Log_RawDeviceOffset      untilclearupperoffset;
    Log_IOSequenceNo         untilsequenceafterredo;
    Log_IOSequenceNo         savepointsequence;
    Log_RawDeviceOffset      savepointdevoffset;
    Log_InfoPage::PageOffset savepointoffset;

    Log_ArchiveLogManager &logManager =  Log_ArchiveLogManager::Instance(); 
    Log_ArchiveLog        *firstLog   = *logManager.ArchiveLogsBegin();

    if (firstLog->GetLogInfo (TaskId,
                              date,
                              time,
                              flushmode,
                              devicestate,
                              queuecount,
                              cycleendoffset,
                              oldestnotsavedoffset,
                              oldestnotsavedsequence,
                              lastknownoffset,
                              lastknownsequence,
                              backupcount,
                              untildate,
                              untiltime,
                              untilclearloweroffset,
                              untilclearupperoffset,
                              untilsequenceafterredo,
                              savepointsequence,
                              savepointdevoffset,
                              savepointoffset ))
    {
        if (Log_DeviceStateOkay == devicestate)
        {
            BasisError = e_ok;
        }
        else
        {
            BasisError = e_backup_history_lost;
            RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_AUTOSAVE_ABORTED_HIST_LOST));
        }
    }
    else
    {
        RTE_Crash( SAPDBErr_Exception(__FILE__, __LINE__,
                                      SAPDBERR_ASSERT_STATE_FAILED,
                                      "k560check_logsave_allowed: ArchiveLog not initialized ") );
    }

    // PTS 1128223 mb 2004-04-03 check for AutoOverWrite 
    if (firstLog->IsAutomaticOverwrite())
    {
        BasisError = e_log_autooverwrite_on;
        RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_SAVE_DENIED_AUTOOVERWRITE));
    }

    // PTS 1128223 mb 2004-04-03 check for LogWriting switched off
    if (! firstLog->IsDevspaceEnabled())
    {
        BasisError = e_log_writing_disabled;
        RTE_Message(Log_Exception(__FILE__,__LINE__,LOG_SAVE_DENIED_WRITING_OFF));
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal tsp00_Int4 kb560GetLogDeviceSize ()
{
    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
    return log.GetDeviceSize();
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560SetLogDevspace (
    tsp00_TaskId                         TaskId,
    pasbool                              isCold,
    tgg00_BasisError      VAR_VALUE_REF  TrError,
    pasbool                              on)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560SetLogDevspace", LogVolume_Trace, 5);

    Log_ArchiveLogManager &logManager =  Log_ArchiveLogManager::Instance(); 

    if ( isCold )
    {
        if ( logManager.SetDevspace (TaskId, isCold, on) )
        {
            TrError = e_ok;
        }
        else
        {
            if (on)
            {
                RTE_Message( SAPDBErr_Exception(__FILE__, __LINE__,
                                                SAPDBERR_ASSERT_STATE_FAILED,
                                                "kb560SetLogDevspace on failed") );
            }
            else
            {
                RTE_Message( SAPDBErr_Exception(__FILE__, __LINE__,
                                                SAPDBERR_ASSERT_STATE_FAILED,
                                                "kb560SetLogDevspace off failed") );
            }
            TrError = e_log_error;
        }
    }
    else   
    {
        RTE_Message( SAPDBErr_Exception(__FILE__, __LINE__,
                                        SAPDBERR_ASSERT_STATE_FAILED,
                                        "kb560SetLogDevspace: may only be used BEFORE restart") );
        TrError = e_log_error;
    }
}
/*-----------------------------------------------------------------------------------*/
externPascal pasbool kb560IsLogDeviceEnabled (    tsp00_TaskId                         taskid)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560IsLogDeviceEnabled", LogVolume_Trace, 5);
    return Log_ArchiveLogManager::Instance().IsDevspaceEnabled(taskid);
}
/*-----------------------------------------------------------------------------------*/

externPascal void kb560SetDeviceStateOkay (
    tsp00_TaskId                         taskid,
    pasbool                              isCold,
    tgg00_BasisError      VAR_VALUE_REF  TrError)
{
    // PTS 1113190 UH 2001-12-19 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560SetDeviceStateOkay", LogVolume_Trace, 5);
    if (Log_ArchiveLogManager::Instance().IsDevspaceEnabled(taskid)) // PTS 1118565 mb 2002-11-1
    {
        if ( ! Log_ArchiveLogManager::Instance().SetDeviceState(taskid, 1==isCold, Log_DeviceStateOkay) )
            TrError = e_log_error;
        else
            TrError = e_ok;
    }
}

/* -------------------------------------------------------------------------------- */
externPascal void kb560SetDeviceStateHistoryLost (
    tsp00_TaskId                         taskid,
    pasbool                              isCold,
    tgg00_BasisError      VAR_VALUE_REF  TrError)
{
    // PTS 1117498 UH 2002-08-30 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560SetDeviceStateHistoryLost", LogVolume_Trace, 5);
    if ( ! Log_ArchiveLogManager::Instance().SetDeviceState(taskid, 1==isCold, Log_DeviceStateHistoryLost) )
        TrError = e_log_error;
    else
        TrError = e_ok;
}

/* -------------------------------------------------------------------------------- */
externPascal void kb560GetHistFileInfo (
    tgg00_TransContext    VAR_VALUE_REF  Trans,
    tsp00_Int4            VAR_VALUE_REF  HistFileNo,
    tgg91_TransNo         VAR_VALUE_REF  FirstTransId,
    pasbool               VAR_VALUE_REF  FirstTransIsUsed,
    tgg91_TransNo         VAR_VALUE_REF  LastTransId,
    pasbool               VAR_VALUE_REF  LastTransIsUsed,
    tsp00_Int4            VAR_VALUE_REF  PageCount,
    pasbool               VAR_VALUE_REF  bFileIsEmpty)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetHistFileInfo", LogVolume_Trace, 5);
    Log_History::HistoryFileInfo  info;
    Log_History                  &history = Log_History::GetInstance();
    SAPDB_UInt                    pagecount;

    if ( HistFileNo >= tsp00_Int4(history.GetNumberOfHistoryFiles()) )
    {
        HistFileNo = -1;
        return;
    }

    history.GetInfo (Trans.trTaskId_gg00, HistFileNo, info, pagecount);

    FirstTransId     = info.oldestTransNo;
    FirstTransIsUsed = 0;
    LastTransId      = info.youngestTransNo;
    LastTransIsUsed  = 0;
    PageCount        = tsp00_Int4(pagecount); // ++++++++++
    bFileIsEmpty     = info.oldestTransNo.gg90IsNil()
                       &&
                       info.youngestTransNo.gg90IsNil();
                       
    HistFileNo       = HistFileNo + 1;
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560CheckKernelVersionForMigration (
    tsp00_Version         VAR_ARRAY_REF  version)
{
    // PTS 1114877 UH 2002-03-18 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CheckKernelVersionForMigration", LogHistory_Trace, 5);

    SAPDB_Int4  majorVersion;
    SAPDB_Int4  minorVersion;
    SAPDB_Int4  correctionLevel;
    SAPDB_Int4  buildNo;
    
    char *auxVersion = reinterpret_cast <char *> (&(version[BEG_OF_1ST_VERS_SUBSTR_SP100]));
    
    sscanf (auxVersion , "%d.%d.%d", &majorVersion, &minorVersion, &correctionLevel);
    
    auxVersion = reinterpret_cast <char *> (&(version[BEG_OF_2CND_VERS_SUBSTR_SP100]));
    
    sscanf (auxVersion , "%d-",    &buildNo                                );
    
    // ----------- set g01ignore_rootid_check if version <= 7.4.2.3 -----------

    if ( majorVersion > 7 )
        return;
    
    if (( majorVersion == 7 ) && ( minorVersion > 4 ))
        return;
    
    if (( majorVersion == 7 ) && ( minorVersion == 4 ) && ( correctionLevel >= 3 ))
        return;
    
    if (( majorVersion == 7 ) && ( minorVersion == 4 ) && ( correctionLevel == 2 ) && ( buildNo > 3 ))
        return;
    
    g01ignore_rootid_check = true;
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560RemoveCompleteHistory (
    tgg00_TransContext    VAR_VALUE_REF  Trans)
{
    // PTS 1114877 UH 2002-03-18 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560RemoveCompleteHistory", LogHistory_Trace, 5);
    Log_History::GetInstance().RemoveCompleteHistory(Trans);
}

/*-----------------------------------------------------------------------------------*/
externPascal pasbool kb560AnyHistoryExists ()
{
    // PTS 1114877 UH 2002-03-18 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560AnyHistoryExists", LogHistory_Trace, 5);
    return Log_History::GetInstance().AnyHistoryExists();
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560StartGCAndWaitForNoHistoryExists (
    tgg00_TransContext    VAR_VALUE_REF  Trans)
{
    // PTS 1114877 UH 2002-03-18 new
    // The call of this routine assumes that the caller knows that all history can be removed now.
    // Otherwise it would loop endless here, but the loop can be cancelled by request.
    
    SAPDBTRACE_ROUTINE_DEBUG ("kb560StartGCAndWaitForNoHistoryExists", LogHistory_Trace, 5);

    pasbool &isCancelled = Trans.trRteCommPtr_gg00->to_cancel;
    
    bd91StartOMSGarbageCollection (Trans, false);
    
    RTE_Message( Log_Exception(__CONTEXT__,LOG_WAIT_FOR_REMOVE_HISTORY) );

    while ( Log_History::GetInstance().AnyHistoryExists() && ! isCancelled )
        vsleep (Trans.trTaskId_gg00, 20);

    if ( isCancelled )
    {
        RTE_Message( Log_Exception(__CONTEXT__,LOG_WAIT_FOR_REMOVE_HISTORY_CANCELLED,
                                   SAPDB_ToString(Log_History::GetInstance().AnyHistoryExists(),_T_d)) );
        isCancelled = false;
    }
}

/*-----------------------------------------------------------------------------------*/

void kb560InternNewDBIdent (Log_ArchiveLog                      &log, 
                       tgg00_TransContext    VAR_VALUE_REF  trans, 
                       boolean                              isCold)
// new for PTS 1117498 mb 2002-08-30
{
    tsp00_Line		newDBIdent;      
    vget_uniqe_id  (newDBIdent);

    k57restartrec->rstDbIdent_kb00() = newDBIdent;
    if (   (Log_DeviceStateCleared == log.GetDeviceState())
        && (isCold))
    // adjust the offset if the log is empty, because the
    // first entry is written onto offset 24 in page 1
    // PTS 1116123 mb 2002-06-12
    {
        tkb00_SaveptParam& SaveptParam = k57restartrec->rstLastSavept_kb00();
        SaveptParam.svpStartEntryType_kb00 = Log_FinishOpenTrans;
        SaveptParam.svpStartEntryOffset_kb00= Log_Page::MinSpaceOffset();
    }

    k57save_restartrecord (trans.trTaskId_gg00);
    log.SetLogDBIdent (trans.trTaskId_gg00,newDBIdent);
    
    SAPDB_Char newLogDBIdent[sizeof(tsp00_Line)+1];
    sql47_ptoc(newLogDBIdent, k57restartrec->rstDbIdent_kb00(), sizeof(tsp00_Line));
    RTE_Message(Log_Exception(__FILE__,__LINE__,
                              LOG_NEW_DBIDENT,
                              newLogDBIdent));
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560NewDBIdent (tgg00_TransContext    VAR_VALUE_REF  trans)
// new for PTS 1117498 mb 2002-08-30
{
    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
    kb560InternNewDBIdent (log, trans, false);
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560NewDBIdentIfHistLost (	// PTS1113550 mb 2002-04-30 new
    tgg00_TransContext    VAR_VALUE_REF  trans,
    pasbool                              isCold)
{
    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());

    if (0 != isCold)
    {
        if ( ! log.DiagnoseOpenInfoPage(trans.trTaskId_gg00, isCold) )
        {
            return;
        }
    }
    
    if ((   Log_DeviceStateCleared          == log.GetDeviceState()  // PTS 1115889 mb 2002-05-27  
         || Log_DeviceStatePartiallyCleared == log.GetDeviceState() 
         || Log_DeviceStateHistoryLost      == log.GetDeviceState())
        &&
         log.IsDevspaceEnabled())
    // a new backuphistory has began
    {
        kb560InternNewDBIdent (log, trans, 0 != isCold);
    }

    if (0 != isCold )
    {
        log.DiagnoseCloseDevice(trans.trTaskId_gg00);
    }
}
/*-----------------------------------------------------------------------------------*/
externPascal void kb560GetLogTransactionInfo (
    Log_Transaction*                     ptrans,
    tsp00_PageNo          VAR_VALUE_REF  undoroot,
    tsp00_PageNo          VAR_VALUE_REF  undolast,
    tsp00_Int4            VAR_VALUE_REF  undopagecount,
    tsp00_Int4            VAR_VALUE_REF  lastundosequence)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560GetLogTransactionInfo", LogTrans_Trace, 5);

    undoroot         = NIL_PAGE_NO_GG00;
    undolast         = NIL_PAGE_NO_GG00;
    undopagecount    = 0;
    lastundosequence = -1;

    if ( ptrans == NULL ) return;

    Data_PageNo      auxundoroot;
    Data_PageNo      auxundolast;
    SAPDB_UInt       auxundopagecount;
    Log_IOSequenceNo auxlastundosequence;

    ptrans->GetInfo ( auxundoroot,
                      auxundolast,
                      auxundopagecount,
                      auxlastundosequence );

    undoroot         = auxundoroot;
    undolast         = auxundolast;
    undopagecount    = auxundopagecount;
    lastundosequence = auxlastundosequence;
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560ConvertInvalidIOSeqToLPNo (
    tsp00_PageNo          VAR_VALUE_REF  ioSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560ConvertInvalidIOSeqToLPNo", LogVolume_Trace, 5);

    // extern (i.e. DBMGui, the representation of INVALID is signed int
    // so the invalid value is also different (0xffffffff != 0x7fffffff)
    Log_IOSequenceNo cppIOSeq(ioSeq);
    if (cppIOSeq.IsInvalid())
    {
        ioSeq = MAX_INT4_SP00;
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560InvalidateIOSeq (           // PTS 1116063 mb 2002-06-05 new
    tsp00_PageNo          VAR_VALUE_REF  ioSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560InvalidateIOSeq", LogVolume_Trace, 5);

    Log_IOSequenceNo cppIOSeq;
    ioSeq = (tsp00_PageNo) cppIOSeq();
}

/*-----------------------------------------------------------------------------------*/

externPascal pasbool kb560IsValidIOSeq (           // PTS 1116063 mb 2002-06-05 new
    tsp00_PageNo                         ioSeq)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560IsValidIOSeq", LogVolume_Trace, 5);

    Log_IOSequenceNo cppIOSeq((SAPDB_UInt4) ioSeq);
    return cppIOSeq.IsValid();
}

/*-----------------------------------------------------------------------------------*/

externPascal pasbool kb560IsSuccessor (
    Log_IOSequenceNo&                    predecessor,
    Log_IOSequenceNo&                    successor)
// new with PTS 1126639 mb 2003-12-22
{
    if ((predecessor+1) == successor)
    {
        return true;
    }
    else
    {
        RTE_Message(Log_Exception(__CONTEXT__, LOG_NON_CONTINUOUS_IOSEQ,
                                  SAPDB_ToString(predecessor),
                                  SAPDB_ToString(successor)));
        return false;
    }
}

/*-----------------------------------------------------------------------------------*/

externPascal pasbool kb560IsGreaterIOSeq (          // PTS 1116063 mb 2002-06-05 new
    tsp00_PageNo                         leftOperand,
    tsp00_PageNo                         rightOperand)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560IsGreaterIOSeq", LogVolume_Trace, 5);

    Log_IOSequenceNo cppLeftOperand((SAPDB_UInt4) leftOperand);
    Log_IOSequenceNo cppRightOperand((SAPDB_UInt4) rightOperand);

    SAPDBERR_ASSERT_ARGUMENT(cppLeftOperand.IsValid());
    SAPDBERR_ASSERT_ARGUMENT(cppRightOperand.IsValid());

    return cppLeftOperand() > cppRightOperand();
}

/*-----------------------------------------------------------------------------------*/

externPascal void kb560CheckForLogSaveAndSavepoint (
    tgg00_TransContext    VAR_VALUE_REF  trans)
{
    SAPDBTRACE_ROUTINE_DEBUG ("kb560CheckForLogSaveAndSavepoint", LogVolume_Trace, 5);
    
    if ( ! Log_ArchiveLogManager::IsRestarted() )
        return;

    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
    log.CheckForLogSaveAndSavepoint(trans);
}

/*-----------------------------------------------------------------------------------*/
externPascal pasbool kb560SuspendAndGetLastWrittenIOSequence (
    tsp00_TaskId                         taskid,
    tsp00_Uint4           VAR_VALUE_REF  iosequence)
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_ROUTINE_DEBUG ("kb560SuspendAndGetLastWrittenIOSequence", LogVolume_Trace, 5);

    if ( ! Log_ArchiveLogManager::IsRestarted()
         ||
         ! Log_SavepointManager.DenySavepoint (taskid) ) // PTS 1121281 UH 2003-03-24
    {
        iosequence = Log_IOSequenceNo();
        return false;
	}

    Log_ArchiveLog   &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
    Log_IOSequenceNo  seq;

    if ( ! log.SuspendAndGetLastWrittenIOSequence(taskid,seq) )
    {
        iosequence = seq;
        return false;
    }

    iosequence = seq;
    return true;
}

/*-----------------------------------------------------------------------------------*/
externPascal void k560ResumeLogWriterByUser (
    tsp00_TaskId                         taskid)
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_ROUTINE_DEBUG ("k560ResumeLogWriterByUser", LogVolume_Trace, 5);

    if ( ! Log_ArchiveLogManager::IsRestarted() )
        return;
        
	Log_SavepointManager.PermitSavepoint (taskid);

    Log_ArchiveLog &log = **(Log_ArchiveLogManager::Instance().ArchiveLogsBegin());
    log.ResumeByUser();
}

/*-----------------------------------------------------------------------------------*/
externPascal pasbool k560IsSavepointAllowed ()
{
    // PTS 1114053 UH 2002-10-22 new
    SAPDBTRACE_ROUTINE_DEBUG ("k560IsSavepointAllowed", LogVolume_Trace, 5);
	return Log_SavepointManager.IsSavepointAllowed();
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560GetSavepointStatistics (
    tsp00_Int4            VAR_VALUE_REF  lastDurationSec,
    tsp00_Int4            VAR_VALUE_REF  lastDurationMicroSec,
    tsp00_Int4            VAR_VALUE_REF  avgWrittenPages,
    tsp00_Int4            VAR_VALUE_REF  avgIOcount)
{
    // PTS 1119276 UH 2002-12-10 new
    Log_SavepointManager.GetSavepointStatistics (lastDurationSec,
                                                 lastDurationMicroSec,
                                                 avgWrittenPages,
                                                 avgIOcount);
}
/*-----------------------------------------------------------------------------------*/
externPascal void kb560SetLogAutoOverwrite (
    tsp00_TaskId                         taskid,
    pasbool                              isAdminMode,
    pasbool                              on)
{
    if ( ! Log_ArchiveLogManager::Instance().SetAutomaticOverwrite ( taskid,
                                                                     isAdminMode,
                                                                     on ) )
	{
        RTE_Crash( SAPDBErr_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED, "kb560SetLogAutoOverwrite") );
	}
}
/*-----------------------------------------------------------------------------------*/
externPascal pasbool kb560IsLogAutoOverwriteEnabled (
    tsp00_TaskId                         taskid,
    pasbool                              isAdminMode)
{
    return Log_ArchiveLogManager::Instance().IsAutomaticOverwrite ( taskid,
                                                                    isAdminMode );
}

/*-----------------------------------------------------------------------------------*/
externPascal void kb560SetTraceMaxBufferLength (
    tsp00_Int4                           length)
{
    if ( length > 0 )
        Kernel_TraceMaxBufferLength = length;
}
