/*! 
  -----------------------------------------------------------------------------
 
  module: vls22.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS and d033893

  special area: SAP DB LOADER
 
  description:  Fastload - filling db pages
                
  version:      7.4.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1999-2004
 
  -----------------------------------------------------------------------------



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/

#include "gls00.h"
#include "hls03.h"
#include "hls04.h"
#include "hls07.h"
#include "hls13.h"
#include "hls16_condition.h"
#include "hls18.h"
#include "hls20.h"
#include "hls22.h"
#include "hls23.h"
#include "hls30.h"
#include "hls98msg.h"
#include "hls99.h"
#include "gbd00.h"      // database bd layer (data page layout)
#include "ggg00.h"      // database bd layer (data page layout)

#define _INTERNAL_FILE_ID_   _T("vls22.cpp")

/*!
  -----------------------------------------------------------------------------
  Chapter: local functions
  -----------------------------------------------------------------------------
*/

/*!
  -----------------------------------------------------------------------------
 
  function:     ls22LoadPages
  -----------------------------------------------------------------------------

  description:  Loads data pages to a given db table (FASTLOAD).
                Initializes file structure for the given data file, opens it,
                reads data records, builds up a db data page and sends the
                pages to the db.
  
  arguments:    FloadCmd  [IN]   - fastload structure built by the parser from
                                    given command
                FloadRec  [IN]   - fastload structure built by the server
                DBInfo    [IN]   - dbinfo structure
                lPcktNo   [IN]   - number of request packet to fill
                VFile     [IN]   - data file structure
                ErrText   [OUT]  - error text

  returnvalue:  errOK_els00             - data loaded
                errors when loading data failed

                Logs the error in log file.
  -----------------------------------------------------------------------------
*/
int ls22_FillPacket(tls00_FloadCmd *FloadCmd,
                    tls00_FloadRec *FloadRec,
                    tls00_DBInfo   *DBInfo,
                    tsp00_Int4      lPcktNo,
                    tls00_VFile    *VFile,
                    tsp00_Addr      ErrText);



/*!
  -----------------------------------------------------------------------------
 
  function:     ls22_PageLoadLoop
  -----------------------------------------------------------------------------

  description:  Main loop of fastload.
                Fills data pages ,sends those pages and receives answers from db.
  
  arguments:    FloadCmd  [IN]   - fastload command structure
                FloadRec  [IN]   - fastload record structure
                DBInfo    [IN]   - dbinfo structure
                VFile     [IN]   - vfile structure
                ErrText   [OUT]  - error text

  returnvalue:  errOK_els00      - data successfully processed
                errors from db system

                Logs the error in log file.
  -----------------------------------------------------------------------------
*/
static int
ls22_PageLoadLoop(tls00_FloadCmd &FloadCmd,
                  tls00_FloadRec &FloadRec,
                  tls00_DBInfo   &DBInfo,
                  tls00_VFile    &VFile,
                  tsp00_Addr      ErrText);


/*!
  -----------------------------------------------------------------------------
  End Chapter: local functions
  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
   function:     ls22_FillPacket
  -----------------------------------------------------------------------------
*/
int ls22_FillPacket(tls00_FloadCmd *FloadCmd,
                    tls00_FloadRec *FloadRec,
                    tls00_DBInfo   *DBInfo,
                    tsp00_Int4      lPcktNo,
                    tls00_VFile    *VFile,
                    tsp00_Addr      ErrText)
{
    ROUTINE_DBG_MEO00 ("ls22_FillPacket");

    tsp00_Int4   ErrCode       = errOK_els00;
    tsp00_Int4   rcREAD        = errOK_els00;


    tsp1_packet *SqlPacket     = NULL;
    tsp1_part   *SqlPart_page  = NULL;
    tsp00_Int4   ListTop       = 0;
    tsp00_Int4   Bottom        = 0;
    tsp00_Int4   Space         = 0;
    tsp00_Int4   PageNo        = 0;
    tsp00_Int4   nPageRecCount = 0;
    tsp00_Int4   RecordLen     = 0;
    tsp00_Addr   Buffer        = NULL;

    tsp00_SwapKind_Enum SSwap   = FloadRec->flrSwapKind_ls00;
    tsp00_SwapKind_Enum DSwap   = FloadRec->flrTableInfo_ls00.tdKernelSwap_ls00;

    tls00_FileSpec *pInFileSpec = &FloadCmd->flcInfile_ls00;
    tls00_String    DataRecord;
    DataRecord.strAddr_ls00     = STATIC_CAST(char*, FloadRec->flrDataSource_ls00.dasInputBuffer.pBufPtr_ls00);

    tls00_String   *RawData     = FloadRec->flrDataSource_ls00.dasConvertedData;

    tbd_node       pDataPage;
    tls00_PageInfo PGI           = FloadRec->flrPageInfo_ls00;
    SAPDB_Bool     bLimitReached = SAPDB_FALSE;



    //*
    //*     Prepare packet to send
    //*
    ls13Prepare_FL_TL_Packet(DBInfo, lPcktNo, SqlPacket, SqlPart_page);


    SAPDB_Bool bOfNoInterest = SAPDB_TRUE;

    //*
    //*     2 while loops to fill the data pages
    //*
    //* - 1st loop goes over the count of pages to fill
    //* - 2nd loop is used to fill a single page
    //*
    while ( (errOK_els00 == ErrCode) && (PageNo < FloadRec->flrPgPerPckt_ls00) )
    {
        pDataPage.nd_full_header().SetZero();               // Initialize page header

        Bottom        = PGI.pgiDataStartPos;
        ListTop       = PGI.pgiDataEndPos;
        Space         = PGI.pgiDataSpace;
        Buffer        = STATIC_CAST(char*, pDataPage.nd_body());  // Offset is 0 for nd_body() !!!
        nPageRecCount = 0;
        bLimitReached = SAPDB_FALSE;

        //*
        //*     Read data from file and insert into page
        //*
        while ( (Space > FloadRec->flrRecordLen_ls00) && (errOK_els00 == ErrCode) && (SAPDB_FALSE == bLimitReached) )
        {
            rcREAD = ls20ReadData(VFile,
                                  &FloadCmd->flcInfile_ls00.fsFormat_ls00,
                                  &FloadRec->flrPattern_ls00,
                                  NULL,
                                  &FloadRec->flrDataSource_ls00.dasInputBuffer,
                                  &bOfNoInterest,
                                  (0 == FloadRec->flrRecordCount_ls00),
                                  ErrText);
            if (rcREAD < 0)
            {
                // following values may happen:
                // - STOP_LAST_DATA_LS00    --> process read record and then stop processing (only for RECORDS format of data file)
                // - STOP_NOMORE_DATA_LS00  --> nothing more to process - get outa here
                // - error code             --> nothing more to process - get outa here
                if (STOP_NOMORE_DATA_LS00 != rcREAD)
                {   // In case of error we stop immediately.
                    ErrCode = rcREAD;
                }
                else// if (STOP_NOMORE_DATA_LS00 == rcREAD)
                {   // In case of EOF ErrCode remains OK and the started page will be finished.

                    if (0 == nPageRecCount )
                    {   // eof while reading first data record
                        ErrCode = STOP_NODATA_LS22;
                    }
                    else
                    {
                        ErrCode = STOP_LS22;
                    }
                }
                break;
            }

            //*
            //*     Check for condition in command
            //*
            if (NULL != FloadCmd->flcTable_ls00.tsCondSpec_ls00)
            {
                if ( false == ls16CheckCondition(*FloadCmd->flcTable_ls00.tsCondSpec_ls00,
                                                 DataRecord,
                                                 pInFileSpec->fsFormat_ls00) )
                {
                    continue;
                }
            }

            //*
            //*     Insert record into data page
            //*
            RecordLen = ls23BuildRecord(//DataSource,
                                        //RawData,
                                        Buffer,
                                        FloadCmd,
                                        FloadRec,
                                        DBInfo->dbiSession_ls00.siConversionTables_ls00,
                                        DBInfo->dbiSession_ls00.siSQLMode_ls00,
                                        ErrText);
            
            ls18RecordInit(FloadRec->flrDataSource_ls00.lReadFieldsCnt_ls00,
                           RawData,
                           STATIC_CAST(char*, FloadRec->flrDataSource_ls00.dasInputBuffer.pBufPtr_ls00),
                           FloadRec->flrDataSource_ls00.dasInputBuffer.lBufPos_ls00);

            if (0 > RecordLen)      // Error while processing record
            {
                if (sp5vf_text == VFile->vfDataType_ls00)
                {
                    tsp00_Int8 ErrorLine = (UNDEFINED_LS00 == pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00) ?
                                            VFile->vfFileLineNum_ls00 :
                                            VFile->vfFileLineNum_ls00 + pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00 - 1;
                    RepServMessages::Instance()->ls98PrintMessage(layGRec_els98, errAtRow2_els98, ErrorLine);
                }
                else
                {
                    tsp00_Int8 ErrorPos = VFile->vfFilePos_ls00.lnu2GB_Blocks_ls00*MAX_INT4_SP00 +
                                           VFile->vfFilePos_ls00.lnuRemainder_ls00;
                    RepServMessages::Instance()->ls98PrintMessage(layGRec_els98, errAtPos2_els98, ErrorPos);
                }

                FloadRec->flrLRejected_ls00 -= 1;
                if (FloadRec->flrLRejected_ls00 <= 0)
                {
                    ErrCode = RecordLen;
                    break;
                }
                else
                {
                    continue;
                }
            }

            ListTop -= POINTERSIZE_BD00;    // update positions

            //  Write pointer to the end of the data list
            ls07Swap4BIntToChar(Bottom + 1, SSwap, DSwap,
                                REINTERPRET_CAST( char*, &pDataPage.nd_body()[ListTop - PGI.pgiDataStartPos]) );

            Space  -= POINTERSIZE_BD00;     // update available space
        
            Bottom += RecordLen;
            Space  -= RecordLen;
            if ( 0 != (RecordLen%2) )       // 2 byte alignment of record (only the record pointer is changed
            {                               // not the record length)
                Bottom += 1;
                Space  -= 1;
            }
            nPageRecCount += 1;             // Update number of records written to table

            //*
            //*     Check for user specified filling level for a single database page (in rows)
            //*
            if (true == FloadCmd->flcUsage_ls00.uUsingRows_ls00)
            {
                if (nPageRecCount == FloadCmd->flcUsage_ls00.uUsageValue_ls00)
                {
                    bLimitReached = SAPDB_TRUE;
                }
            }

            //*
            //*     Check for user specified START position in data file
            //*
            //* If the user specified a START position he may have defined a count of
            //* rows to insert. This is check here as a stop option.
            //*
            if (UNDEFINED_LS00 != pInFileSpec->fsExtract_ls00.feEnd_ls00.lnuRemainder_ls00)
            {
                if ( (FloadRec->flrRecordCount_ls00 + nPageRecCount) == pInFileSpec->fsExtract_ls00.feEnd_ls00.lnuRemainder_ls00 )
                {
                    ErrCode       = STOP_LS22;
                    bLimitReached = SAPDB_TRUE;
                }
            }

            // Reassign Buffer to the next free byte in nd_body() - ATTENTION: Initial offset of nd_body
            // is 0, so we have to substact the bytes for the header from the new offset all the time
            Buffer = REINTERPRET_CAST( char*, &pDataPage.nd_body()[Bottom - PGI.pgiDataStartPos] );
        
        }   // end while ( (Space > FloadRec->flrRecordLen_ls00) && (errOK_els00 == ErrCode) )

        TRACE_PRNF_MLS99(("ls22_FillPacket", "%d rows at page.\n", nPageRecCount));

        if ( (ErrCode >= 0) && (ErrCode != STOP_NODATA_LS22) )
        {
            //*
            //*     update record header values
            //*
            pDataPage.nd_record_cnt() = ls07Swap4BInt(nPageRecCount, SSwap, DSwap);
            pDataPage.nd_bottom()     = ls07Swap4BInt(Bottom + 1, SSwap, DSwap);
            pDataPage.nd_sorted()     = true;

            //*
            //*     Write filled page to part
            //*
            ls04BufToPart(SqlPart_page, REINTERPRET_CAST(char*, &pDataPage), DBInfo->dbiPageSize_ls00);
            PageNo += 1;

            FloadRec->flrRecordCount_ls00 += nPageRecCount; // adapt count of inserted records
            FloadRec->flrPageCount_ls00   += 1;             // adapt count of sent pages
        }

        //TRACE_BUFF_MLS99 ("ls22_FillPacket", "PAGE HEADER", sizeof(tbd_node_header), pDataPage.nd_full_header());
        //TRACE_BUFF_MLS99 ("ls22_FillPacket", "PAGE BODY", PGI.pgiDataEndPos - PGI.pgiDataStartPos, pDataPage.nd_body());
    
    }   // end while ( (EOF_LS00 != VFile->vfFilePos_ls00) && ...

    
    if (ErrCode >= 0)
    {
        if (0 != PageNo)            // no EOF without data
        {
            tsp00_Int4 rc2 = ls04FinishPart(SqlPacket, SqlPart_page);
            if (errOK_els00 != rc2)
            {
                ErrCode = rc2;
            }
            else
            {
                if (STOP_NODATA_LS22 == ErrCode)
                {
                    ErrCode = STOP_LS22;
                }
            }
        }
    }

    return ErrCode;
}
// ls22_FillPacket()


/*
  -----------------------------------------------------------------------------
  function:     ls22_PageLoadLoop
  -----------------------------------------------------------------------------
*/
static int
ls22_PageLoadLoop(tls00_FloadCmd &FloadCmd,
                  tls00_FloadRec &FloadRec,
                  tls00_DBInfo   &DBInfo,
                  tls00_VFile    &VFile,
                  tsp00_Addr      ErrText)
{
    ROUTINE_DBG_MEO00 ("ls22_PageLoadLoop");
    
    tsp00_Int4 rc          = errOK_els00;
    tsp00_Int4 rcFill      = errOK_els00;
    tsp00_Int4 lSndPcktNo  = 0;
    tsp00_Int4 lRcvPcktNo  = 0;
    tsp00_Int4 nSndPcktCnt = 0;

    //*  MAIN part of fastload
    while ( (errOK_els00 == rcFill) && (errOK_els00 == rc) )
    {
        //*
        //*     Fill SQL-Packet
        //*
        rcFill = ls22_FillPacket(&FloadCmd, &FloadRec, &DBInfo, lSndPcktNo, &VFile, ErrText);
        if (lSndPcktNo != lRcvPcktNo)
        {
            //*
            //*     Get answer packet
            //*
            rc = ls13ReceiveBdinfoPckt(&DBInfo, lRcvPcktNo, FloadRec.flrBdInfo_ls00, ErrText);
            if (errOK_els00 == rc)
            {
                lRcvPcktNo = (lRcvPcktNo + 1) % 2;
            }
        }
        
        //*
        //*     Send packet
        //*
        if ( (rcFill >= 0) && (rcFill != STOP_NODATA_LS22) && (errOK_els00 == rc) )
        {
            rc = ls13SendBdinfoPckt(&DBInfo, lSndPcktNo, FloadRec.flrBdInfo_ls00, ErrText);
            if (errOK_els00 == rc)
            {
                nSndPcktCnt += 1;

                lSndPcktNo = (lSndPcktNo + 1) % 2;
            }
        }
    }

    // Create an error message if something happened during sending/receiving the packages
    if (errOK_els00 != rc)
    {
        RepServMessages::Instance()->ls98Msg(layFastload_els98, ErrText, errInternal_els98,
                                                                         _INTERNAL_FILE_ID_,
                                                                         __LINE__,
                                                                         STATIC_CAST(ErrorNumbers, rc));
        rc = errInternal_els98;
    }


    //*
    //*     Check in case of no (more) data found in data file if there was any and
    //*     adapt the return code respectively
    //*
    if (STOP_NODATA_LS22 == rcFill)
    {                                           
        if (0 == FloadRec.flrRecordCount_ls00)
        {   // Nothing sent - error
            RepServMessages::Instance()->ls98Msg(layFastload_els98, ErrText,
                                                                    errEmptyDataFile_els98,
                                                                    VFile.vfFileName_ls00);
            rc = errEmptyDataFile_els98;
        }
        else
        {
            rc = rcFill = errOK_els00;   // Ok, already data sent; end with ok
        }
    }


    //*
    //*      last packet to receive
    //* Because the communication to the db is synchronous we MUST receive a packet
    //*
    if ( (0 != nSndPcktCnt) && (lSndPcktNo != lRcvPcktNo) )
    {
        tsp00_Int4 rc2 = ls13ReceiveBdinfoPckt(&DBInfo, lRcvPcktNo, FloadRec.flrBdInfo_ls00, ErrText);

//TODODTODO  problem with error handling: if receiving packet fails but we have already an error from
        // handling the data - which error to log?
        // At the moment we log the error from last receive only if no other error occurred
        if (errOK_els00 != rc2)
        {
            if (errOK_els00 == rc)
            {
                RepServMessages::Instance()->ls98Msg(layFastload_els98, ErrText, errInternal_els98,
                                                                                 _INTERNAL_FILE_ID_,
                                                                                 __LINE__,
                                                                                 STATIC_CAST(ErrorNumbers, rc2));
                rc = errInternal_els98;
            }
        }
    }

    if (errOK_els00 == rc)
    {
        if ( (errOK_els00 != rcFill) && (STOP_LS22 != rcFill) )
        {
            rc = rcFill;
        }
    }

    return rc;
}
//  ls22_PageLoadLoop()


//*********************** G L O B A L   F U N C T I O N S *********************


/*
  -----------------------------------------------------------------------------
  function:     ls22LoadPages
  -----------------------------------------------------------------------------
*/
int ls22LoadPages(tls00_FloadCmd *FloadCmd,
                  tls00_FloadRec *FloadRec,
                  tls00_DBInfo   *DBInfo,
                  tsp00_Addr      ErrText)
{
    ROUTINE_DBG_MEO00 ("ls22LoadPages");

    tsp00_Int4 rc = errOK_els00;
    
    tls00_FileSpec *pInFileSpec = &FloadCmd->flcInfile_ls00;
    
    
    // Initialize file structure
    tls00_VFile VFile;
    memset(&VFile,0,sizeof(tls00_VFile));

    if (false == pInFileSpec->fsFormat_ls00.ffoBinary_ls00)
    {
        ls30VFInitialize(VFile,
                         pInFileSpec->fsFileName_ls00,
                         sp5vf_text,
                         sp5vf_read,
                         FloadCmd->flcInfile_ls00.fsFormat_ls00.ffoCodeType_ls00);
    }
    else
    {
        ls30VFInitialize(VFile, pInFileSpec->fsFileName_ls00);
    }

    // Open file and calculate record length
    rc = ls30VFOpen(VFile, ErrText);
    if (errOK_els00 == rc)
    {
        tsp00_Int4 lByteCount = ls18GetRowLength(FloadCmd->flcColSpec_ls00,
                                                 FloadRec->flrTableInfo_ls00,
                                                 pInFileSpec->fsFormat_ls00.ffoFormatted_ls00);

        if (true == pInFileSpec->fsFormat_ls00.ffoBinary_ls00)
        {
            if (0 == pInFileSpec->fsFormat_ls00.ffoBytesPerLine_ls00)       // NOHEADER not found in command
            {
                rc = ls30VFReadHeader(VFile, pInFileSpec->fsFormat_ls00.ffoBytesPerLine_ls00, ErrText);
                if (errOK_els00 != rc)
                {
                    if (errNoDatafileHeader_ls98 == rc)
                    {
                        pInFileSpec->fsFormat_ls00.ffoBytesPerLine_ls00 = lByteCount;
                        RepServMessages::Instance()->ls98PrintMessage(layFastload_els98,
                                                                      errNoDatafileHeader_ls98,
                                                                      pInFileSpec->fsFormat_ls00.ffoBytesPerLine_ls00);
                        rc = errOK_els00;
                    }
                    // In case of file read error we stop at this point;
                    // if only no header log it and calculate length
                }
            }
            // else is done because value is already set
        }
        else
        {
            pInFileSpec->fsFormat_ls00.ffoBytesPerLine_ls00 = lByteCount;
        }
    }

    if (errOK_els00 == rc)
    {
        //*
        //*     Check for loading a part of the file only
        //*
        if (errOK_els00 == rc)
        {
            if ( (UNDEFINED_LS00 != pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00) &&
                 (0              != pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00) )   // in case of 0 we're already on the right pos
            {
                tls00_LargeNumber lFilePos;
                lFilePos.lnu2GB_Blocks_ls00 = 0;

                if (false == pInFileSpec->fsFormat_ls00.ffoBinary_ls00)  // text file
                {
                    lFilePos.lnuRemainder_ls00 = pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00;
                }
                else
                {
                    lFilePos.lnuRemainder_ls00 = pInFileSpec->fsFormat_ls00.ffoBytesPerLine_ls00 * (pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00 - 1);
                }
                rc = ls30VFMoveToStartPosition(VFile, &lFilePos, ErrText);
            }
        }

    
        //*
        //*     Main working routine for loading fast
        //*
        //* We need to catch the case when the user specified a START position in infile and
        //* a count of rows to load but this count is 0.
        //*
        if (errOK_els00 == rc)
        {
            if ( (UNDEFINED_LS00 == pInFileSpec->fsExtract_ls00.feEnd_ls00.lnuRemainder_ls00) ||
                 (0               < pInFileSpec->fsExtract_ls00.feEnd_ls00.lnuRemainder_ls00) )
            {
                rc = ls22_PageLoadLoop(*FloadCmd, *FloadRec, *DBInfo, VFile, ErrText);
                if (errOK_els00 == rc)
                {
                    FloadRec->flrLastCommLine_ls00 = VFile.vfFileLineNum_ls00;
                    if (UNDEFINED_LS00 != pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00)
                    {
                        FloadRec->flrLastCommLine_ls00 += pInFileSpec->fsExtract_ls00.feStart_ls00.lnuRemainder_ls00 - 1;
                    }
                }
            }
            else if (0 == pInFileSpec->fsExtract_ls00.feEnd_ls00.lnuRemainder_ls00)
            {
                FloadRec->flrLastCommLine_ls00 = 0;
            }
        }
    }

//TODOTODO  catch error from closing file (not that important
    tsp00_Int4 rc2 = ls30VFClose(VFile, ErrText);

    return rc;
}
//  ls22LoadPages()

