/*! 
  -----------------------------------------------------------------------------
 
  module: vls40bload.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  SteffenS
 
  special area: SAP DB Loader
 
  description:  implementation of command line Loader access

  version:      7.4

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1997-2003
 
  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*!
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "hls05.h"

#include "gsp00.h"

#include "gsp01.h"
#include "vsp001.h"
#include "gsp09.h"

#include "geo47.h"

#include "heo01.h"      // function to define db root directory
#include "heo02.h"
#include "heo03.h"
#include "hsp100.h"
#include "hcn13.h"
#include "hcn14.h"
#include "hls40cli.h"


/*! 
  -----------------------------------------------------------------------------
  Specification
  -----------------------------------------------------------------------------
*/

externPascal int s02applencrypt( char * pw_clear, char * pw_crypt);     // modul vsp002

ls40_Options *g_pOptions = NULL;

/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_Initialize

  -----------------------------------------------------------------------------

  description:  Initializes the structure with all information
                for execution.

  arguments:    opt         [IN]  - structure to initialize

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls40_Initialize(ls40_Options& opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_Usage

  -----------------------------------------------------------------------------

  description:  Prints usage information.

  arguments:    No

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void ls40_Usage();


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ShowPrgVersion

  -----------------------------------------------------------------------------

  description:  Print version info of client program to specified outfile.

  arguments:    OutFile     [IN]   - file pointer of file to print to

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void
ls40_ShowPrgVersion(FILE *OutFile);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ShowServerVersion

  -----------------------------------------------------------------------------

  description:  Print version info of loader server.
                To get the version information the function builds a command and
                sends it to the server. Then it 'analyzes' the reply.

  arguments:    options     [IN]    - structure with all information needed

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error getting version, error message is written
                PRG_SERVER_ERROR    - fatal server error getting version,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int
ls40_ShowServerVersion(ls40_Options & options);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ParseOptions

  -----------------------------------------------------------------------------

  description:  Parses command line.

  arguments:    options     [IN]        - structure to initialize
                argc        [IN]        - command line argument
                argv        [IN]        - command line argument

  returnvalue:  PRG_OK      - success
                PRG_FATAL   - either error in command line or only the help
                              information should be displayed
  -----------------------------------------------------------------------------
*/
static int
ls40_ParseOptions(ls40_Options *options, int argc, char *argv[]);


/*!
  -----------------------------------------------------------------------------
   function:     ls40_ReadUser
  -----------------------------------------------------------------------------

  description:  Reads user name and password hidden from console

  arguments:    options     [IN/OUT]  - structure that holds the user information

  returnvalue:  PRG_OK      - success
                PRG_FATAL   - error reading password
  -----------------------------------------------------------------------------
*/
static int
ls40_ReadUser(ls40_Options *pOptions);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ServerConnect

  -----------------------------------------------------------------------------

  description:  Connects the client to the loader server.
                The server node to connect to is chosen in the following way:
                1. No server node given in command line -> the local host is used
                or
                2. command option indicates to use the database server node. If there
                   is no db server node given in the command line anyway the local host
                   is used for the loader server.
                or
                3. An own server node is given for the loader server which will
                   be used.

                In case there are more than one installations of db software of the
                same version on the server where to start the loader server the
                user has to give parameter DBROOT in the command line call which will
                be given through the connect to the C-server.

                Additionally the function fills in the max. packet length in the
                option structure.

  arguments:    *options        [IN]    - structure containing all needed information
                                          to connect to loader server

  returnvalue:  PRG_OK      - success
                PRG_FATAL   - could not connect to server; error message is given
  -----------------------------------------------------------------------------
*/
static int ls40_ServerConnect(ls40_Options *pOptions);


static void
ls40_DisconnectServer(ls40_Options *pOptions);

/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_UserConnect

  -----------------------------------------------------------------------------

  description:  Connect user to db over the loader manager. User name
                and db name are delivered as options to the program.
                In case either the user name or the db name options are not
                specified the function returns simply PRG_OK without doing anything.
                In this case the user must make sure to specify the user name
                and db name in the script file to the loader server.

  arguments:    opt     [IN]        - structure containing all needed information
                                       to logon user

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error executing the command, error message is written
                PRG_SERVER_ERROR    - fatal server error executing the command,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int ls40_UserConnect(ls40_Options &opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_ServerCommand

  -----------------------------------------------------------------------------

  description:  Executes the command delivered in command buffer of options 
                structure. Checks the answer of server.

  arguments:    opt     [IN]        - structure containing all needed information
                                       to send the command

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error executing the command
                PRG_SERVER_ERROR    - fatal server error executing the command,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int ls40_ServerCommand(ls40_Options *options);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_CheckAnswer

  -----------------------------------------------------------------------------

  description:  Checks the answer of the server after command execution. Removes
                the server answer header from the reply string. Extracts in case
                of error the error number from the reply string and returns it.

  arguments:    opt     [IN]    - structure containing all needed information

  returnvalue:  PRG_OK          - no error on execution of the command
                value > 0       - error executing the command, return value is 
                                  the error number returned by the server
                The reply string contains the reply from the server after returning.
  -----------------------------------------------------------------------------
*/
static int ls40_CheckAnswer(ls40_Options & options, tsp00_Int4 ReplyLen);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_Batch

  -----------------------------------------------------------------------------

  description:  Reads the loader server command file, separates commands and
                sends them to the server.

                Reads 32k from command file in a buffer and works on buffer;
                leaves free space of at least ReplyBufferLen byte in packet for
                answer of server
                At the beginning allocates a buffer of size packet size. For every
                command the buffer is initialized.

  arguments:    opt     [IN]        - structure containing all needed information

  returnvalue:  PRG_OK              - success
                PRG_CMD_ERROR       - error executing the command, error message is written
                PRG_SERVER_ERROR    - fatal server error executing the command,
                                      error message is written
  -----------------------------------------------------------------------------
*/
static int ls40_Batch(ls40_Options &opt);


/*
  -----------------------------------------------------------------------------
  function:     ls40_ProcessInteractively
  -----------------------------------------------------------------------------
*/
static int ls40_ProcessInteractively(ls40_Options *pOptions);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_InitBuffers

  -----------------------------------------------------------------------------

  description:  Initializes used command and reply buffers in options structure.

  arguments:    opt     [IN]        - structure to initialize

  returnvalue:  No
  -----------------------------------------------------------------------------
*/
static void ls40_InitBuffers(ls40_Options & opt);


/*!
  -----------------------------------------------------------------------------
 
  function:     ls40_IsCmdSeparator

  -----------------------------------------------------------------------------

  description:  Checks if an null terminated buffer (read line from an command
                file) evaluates to an command separator.
                The only recognized separator is a double slash at the beginning
                of a line.

  arguments:    p       [IN]        - buffer

  returnvalue:  true  = comment
                false = no comment
  -----------------------------------------------------------------------------
*/
static bool ls40_IsCmdSeparator(_TCHAR *p);


/*
  -----------------------------------------------------------------------------
 
  function:     ls40_Substitute

  -----------------------------------------------------------------------------

  description:  Substitutes parameters of the form &<digit> if such parameter
                was defined at the command line when calling the client.

  arguments:    opt           [IN]      - structure holding all information
                pszBuffer     [IN|OUT]  - string where to substitute the
                                          parameters
                lBufferLength [IN|OUT]  - length of buffer, will be
                                          adapted if parameters are found

  returnvalue:  0  = successful substitution or nothing to substitute; the string
                     can be used as command to send to the server
                -1 = substititution creates a string that exceeds the possible
                     length of command strings
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls40_Substitute(ls40_Options &opt, _TCHAR *pszBuffer, tsp00_Longint & lBufferLength);

/*
  -----------------------------------------------------------------------------
 
  function:     ls40_ReadNextArgument

  -----------------------------------------------------------------------------

  description:  

  arguments:    

  returnvalue:  
  -----------------------------------------------------------------------------
*/
static char*
ls40_ReadNextArgument(char *szArgs[],
                      int  &nCurArg,
                      int   nMaxArgs,
                      int   nMode = 0);


/****************************************************************************** 
  -----------------------------------------------------------------------------
  START OF IMPLEMENTATION
  -----------------------------------------------------------------------------
******************************************************************************/

/* 
  -----------------------------------------------------------------------------
  function: ls40_ShowPrgVersion
  -----------------------------------------------------------------------------
*/
static void ls40_ShowPrgVersion(FILE *OutFile)
{
    tsp100_VersionID1 Version;

    sp100_GetVersionID(VersionIDType1_esp100, s100buildnumberEx, &Version );

    fprintf(OutFile, "Version: %d.%d.%.2d.%.2d\n",
                     STATIC_CAST(tsp00_Int4, Version.MajorVersion_sp100),
                     STATIC_CAST(tsp00_Int4, Version.MinorVersion_sp100),
                     STATIC_CAST(tsp00_Int4, Version.CorrLevel_sp100),
                     STATIC_CAST(tsp00_Int4, Version.BuildNumberPrefix_sp100));
}
//  ls40_ShowPrgVersion()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ShowServerVersion
  -----------------------------------------------------------------------------
*/
static int ls40_ShowServerVersion(ls40_Options & options)
{
    short rc        = PRG_OK;

    ls40_InitBuffers(options);

    options.CmdBufferLen = 7;
    memcpy(options.pszCmdBuffer, "version", 7);

    rc = ls40_ServerCommand(&options);
    if (PRG_OK == rc)
    {
        fprintf(options.outfile, "Server version information:\n%s\n", options.pszReplyBuffer);
    }
    else if (PRG_CMD_ERROR == rc)
    {
        if (false == options.bSuppressOutput_ls40)
        {
            fprintf(options.outfile, "Error getting server version information: %s\n", options.pszReplyBuffer);
        }
    }
    // else means PRG_SERVER_ERROR error msg is already written by ls40_ServerCommand

    return rc;
}


/*
  -----------------------------------------------------------------------------
  function:     ls40_Usage
  -----------------------------------------------------------------------------
*/
static void ls40_Usage()
{
    char szString[3096];
  
    sprintf(szString,   "usage: loadercli [<options>]\n"
                        " <options>:\n"
                        "  -u <username>,<passwd>           :: connect user\n"
// PTS 1107858
//                      "  -U <userkey>            :: instead of Username/Password\n"
//                      "                          :: combination\n"
                        "  -d <serverdb>                    :: database name\n"
                        "  -n <db node>                     :: database server node (default localhost)\n"
                        "  -r [<node>]                      :: Loader server node (default localhost);\n"
                        "                                      remote communication not supported;\n"
                        "                                      omit this option to use statically linked Loader client/server\n"
                        "  -R <INSTROOT>                    :: set INSTROOT directory (meaningless if omitting option -r)\n"
                        "  -b <file name>                   :: command input file; omitting starts client in interactive mode\n"
                        "  -p<number> <substitution string> :: string that substitutes place holder '&<number>'\n"
                        "                                      in command input file (up to 9 parameters possible)\n"
                        "  -o[w|a] <file name>              :: output protocol file (default: stdout)\n"
                        "  -E <number>                      :: defines the number of errors after that the client\n"
                        "                                      stops executing (1 = default and 0 = ignore errors)\n"
                        "  -V                               :: show version of Loader Server\n"
                        "  -v                               :: show version of Loader Server client\n"
                        "\n"
                        "Examples:  loadercli -u MyName,MyPassword -d MyDB -b MyFile\n"
// PTS 1107858          "           repmcli -U MyKey -n <db server node> -r <repm server node> -d MyDB -b MyFile\n");
                        "           loadercli -u MyName,MyPassword -n <db server node> -r <repm server node> -d MyDB -b MyFile\n");
    printf(szString);

    return;
}
//  ls40_Usage()



/*
  -----------------------------------------------------------------------------
   function:     ls40_Initialize
  -----------------------------------------------------------------------------
*/
static void ls40_Initialize(ls40_Options& opt)
{
    opt.dbuser[0]               = 0;
    opt.dbname                  = NULL;
    
    opt.szDBRoot_ls40[0]        = 0;

    opt.db_node                 = "";

    opt.NodeLoc                 = rncLocal_els40;
    opt.repm_node               = NULL;
    
    opt.serverpgm               = DEFAULT_PGM_L;

    opt.infile                  = stdin;
    opt.outfile                 = stdout;
    opt.showServerVersion       = false;
    opt.session                 = NULL;

    opt.pszCommBuffer_ls40      = NULL;
    opt.pszCmdBuffer            = NULL;
    opt.CmdBufferLen            = 0;

    opt.pszReplyBuffer          = NULL;
    opt.ReplyBufferLen          = RPM_REPLY_BUFFER_LEN_CN14 + RPM_ERR_PROTOCOL_SIZE_CN14;

    opt.CmdExecError            = 0;
    opt.BatchErrorCnt_ls40      = 1;     // default: stop after 1 error

    opt.bNoParam_ls40           = true;  // default: no parameters to consider
    
    opt.bSuppressOutput_ls40    = false;   // hidden option

    tsp00_Int2 i = 0;
    
    for (i; i < MAX_PARAM_LS40; ++i)
    {
        opt.pszParamList_ls40[i] = NULL;
    }

    // Initialize to static server (client and server linked together)
    opt.bStaticServer_ls40      = true;

    opt.bInteractive_ls40       = true;
    opt.bRelease_ls40           = false;
    opt.bCancelByte_ls40        = false;
}
// ls40_Initialize()


/*
  -----------------------------------------------------------------------------
   function:     ls40_InitBuffers
  -----------------------------------------------------------------------------
*/
static void ls40_InitBuffers(ls40_Options & opt)
{

    if (NULL == opt.pszCommBuffer_ls40)
    {
        // First thing to do - allocate space

        //*
        //*     Allocate space for command buffer
        //*
        tsp00_Int4 blen = MAX_COMMAND_LENGTH_LS40;
        if (true == opt.bStaticServer_ls40)
        {
            blen += RPM_PROTOCOL_SIZE_CN14 + sizeof(char);  
        }

        opt.pszCommBuffer_ls40 = new _TCHAR[blen];
        memset(opt.pszCommBuffer_ls40, 0, blen);

        //*
        //*     In case of static server we need to leave some space for
        //*     client/server protocol
        //*
        if (true == opt.bStaticServer_ls40)
        {
            opt.pszCmdBuffer = opt.pszCommBuffer_ls40 + RPM_PROTOCOL_SIZE_CN14;
        }
        else
        {
            opt.pszCmdBuffer = opt.pszCommBuffer_ls40;
        }

        //*
        //*     Allocate space for (server) reply buffer
        //*
        opt.pszReplyBuffer = new _TCHAR[opt.ReplyBufferLen];
    }

    //*
    //*     In any case - initialize the buffers
    //*
    if (true == opt.bStaticServer_ls40)
    {
        // Initialize communication protocol
        memset(opt.pszCommBuffer_ls40, 0, RPM_PROTOCOL_SIZE_CN14);
    }
    opt.pszCmdBuffer[0]   = 0;
    opt.CmdBufferLen      = 0;
    opt.pszReplyBuffer[0] = 0;
}
// ls40_InitBuffers()


/*
  -----------------------------------------------------------------------------
  function:     ls40_Aborthandler
  -----------------------------------------------------------------------------
 */
extern "C" void
ls40_Aborthandler(int sig)
{
    switch (sig)
    {
        case SIGINT:
        case SIGTERM:
        case SIGABRT:
#if defined (_WIN32)
        case SIGBREAK:
#endif
        {
            // disconnect
            if (g_pOptions != NULL)
            {
                if (g_pOptions->session != NULL)
                {
                    ls40_DisconnectServer(g_pOptions);
                }
            }
            exit(0);
            break;
        }
        default:
        {
            // do nothing
            break;
        }
    } // end switch
}
// ls40_Aborthandler()


/*
  -----------------------------------------------------------------------------
   function:     ls40_ReadNextArgument
  -----------------------------------------------------------------------------
*/
static char*
ls40_ReadNextArgument(char *szArgs[], int  &nCurArg, int  nMaxArgs, int  nMode)
{
    int    nCurrent  = nCurArg + 1;
    char * pArgument = NULL;
    if (nCurrent < nMaxArgs)
    {
        if ( (1 == nMode) && ('-' == szArgs[nCurrent][0]) )
        {
            pArgument = &(szArgs[nCurrent][1]);
            nCurArg   = nCurrent;
        }
        else if ( (0 == nMode) && ('-' != szArgs[nCurrent][0]) )
        {
            pArgument = szArgs[nCurrent];
            nCurArg   = nCurrent;
        }
    }
    return pArgument;
}
// ls40_ReadNextArgument()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ParseOptions
  -----------------------------------------------------------------------------
*/
static int ls40_ParseOptions(ls40_Options *options, int argc, char *argv[])
{
    short rc        = PRG_OK;

    bool  inOptions = true;
    char  cOptChar  = '\0';
    int   argI      = 0;
    char *pSwitch   = NULL;
    bool  bReadUser = false;

    options->application = argv[0];     // get name of application for prompting in interactive mode
    if (1 != argc)
    {
        pSwitch = ls40_ReadNextArgument(argv, argI, argc, 1);

        while ( (true == inOptions) && ( NULL != pSwitch ) )
        {
            // User name and Password
            if (0 == strcmp(pSwitch, "u"))
            {
                char* pUser = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == pUser)
                {
                    bReadUser = true;
                }
                else
                {
                    options->dbuser[0] = 0;
                    _tcscpy(options->dbuser, pUser);
                }
            }
            // DB name
            else if (0 == strcmp(pSwitch, "d"))
            {
                options->dbname = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == options->dbname)
                {
                    printf("Database name missing (d option)\n");
                    inOptions = false;
                }
                else
                {
                    size_t l = strlen(options->dbname);
                    for (char *p = options->dbname; p < options->dbname + l; *p = _totupper(*p), ++p)
                        ;
                }
            }
            // Database node name
            else if (0 == strcmp(pSwitch, "n"))
            {
                options->db_node = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == options->db_node)
                {
                    printf("Database node name missing (n option)\n");
                    inOptions = false;
                }
            }
            // loader server node name
            else if (0 == strcmp(pSwitch, "r"))
            {
                options->repm_node = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == options->repm_node)
                {
                    // lonely -r points to the db server as Loader server node
                    options->NodeLoc = rncDBNode_els40;
                }
                else
                {
                    options->NodeLoc = rncDifferent_els40;
                }

                // User wishes to deploy a client server model
                options->bStaticServer_ls40 = false;
            }
            // name of dbroot
            else if (0 == strcmp(pSwitch, "R"))
            { 
                char *pTmp = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL != pTmp)
                {
                    // Is the given path name longer than var generate error
                    size_t lDBRootLen = strlen(pTmp);
                    if (lDBRootLen <= (sizeof(tsp00_Pathc)-1) )
                    {
                        memcpy(options->szDBRoot_ls40, pTmp, lDBRootLen);
                        options->szDBRoot_ls40[lDBRootLen] = '\0';
                    }
                    else
                    {
                        printf("Path name of dbroot path too long.\n\n");
                        inOptions = false;
                    }
                }
            }
            // command file
            else if (0 == strcmp(pSwitch, "b"))
            {
                inOptions = false;

                // Open file binary for later operations like fseek
                options->pszInfileName = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == options->pszInfileName)
                {
                    printf("Input file name missing (b option)\n");
                }
                else
                {
                    if (0 == _tcsicmp(options->pszInfileName, "stdin"))
                    {
                        printf("stdin cannot be used as batch file; omit b option to run Loader in interactive mode\n");
                    }
                    else
                    {
                        options->infile = fopen(options->pszInfileName, "rb");
                        if (NULL != options->infile)
                        {
                            options->bInteractive_ls40 = false;
                            inOptions = true;
                        }
                        else
                        {
                            printf("Cannot open input file %s\n", options->pszInfileName);
                        }
                    }
                }
            }
            // output file
            else if ('o' == pSwitch[0])
            {
                inOptions = false;

                options->pszOutfileName = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == options->pszOutfileName)
                {
                    printf("Output file name missing (o option)\n");
                }
                else
                {
                    if (0 == _tcsicmp(options->pszOutfileName, "stdout"))
                    {
                        // Nothing to do - user specified default
                        inOptions = true;
                    }
                    else
                    {
                        char mode[2] = {'w'};
                        if (strlen(pSwitch) > 1)
                        {
                            if ('a' == pSwitch[1])
                            {
                                mode[0] = 'a';
                            }
                        }

                        options->outfile = fopen(options->pszOutfileName, mode);
                        if (NULL != options->outfile)
                        {
                            inOptions = true;
                        }
                        else
                        {
                            printf("Cannot open output file %s\n", options->pszOutfileName);
                            inOptions = false;
                        }
                    }
                }
            }
            // error number
            else if (0 == strcmp(pSwitch, "E"))
            {
                inOptions = false;          // initialize to error

                char   *pszNumber = ls40_ReadNextArgument(argv, argI, argc);
                if (NULL == pszNumber)
                {
                    printf("Number specification missing (E option)\n");
                }
                else
                {
                    bool    bInvalid   = false;
                    size_t  lNumberLen = strlen(pszNumber);
        
                    if (lNumberLen < 10)
                    {
                        if ( (1 == lNumberLen) && ('0' == pszNumber[0]) )
                        {
                            options->BatchErrorCnt_ls40 = 0;
                            inOptions = true;
                        }
                        else
                        {
                            options->BatchErrorCnt_ls40 = atoi(pszNumber);
                            if (0 == options->BatchErrorCnt_ls40)
                            {
                                bInvalid = true;
                            }
                            else
                            {
                                inOptions = true;
                            }
                        }
                    }
                    else if (lNumberLen == 10)
                    {
                        if (0 < memcmp("2147483647", pszNumber, 10))
                        {
                            options->BatchErrorCnt_ls40 = atoi(pszNumber);
                            if (0 == options->BatchErrorCnt_ls40)
                            {
                                bInvalid = true;
                            }
                            else
                            {
                                inOptions = true;
                            }
                        }
                    }
                    // Is the given value longer than 10 bytes or GT or EQ 2147483647? --> error
                    if (false == inOptions)
                    {
                        if (true == bInvalid)
                        {
                            printf("The value '%s' for the number of errors is not a valid number.\n\n",
                                   pszNumber);
                        }
                        else
                        {
                            printf("Error numbers may not exceed 2147483646\n\n");
                        }
                        ls40_Usage();
                    }
                }   // end if (NULL == pszNumber)
            }
            // client version
            else if (0 == strcmp(pSwitch, "v"))
            {
                ls40_ShowPrgVersion(options->outfile);
                inOptions = false;
            }
            // server version
            else if (0 == strcmp(pSwitch, "V"))
            {
                options->showServerVersion = true;
            }
            // parameter substitution
            else if ('p' == pSwitch[0])
            {
                tsp00_Int4 Index = atoi(&pSwitch[1]);
                if ( (0 == Index) || (Index > MAX_PARAM_LS40) )
                {
                    printf("Wrong parameter specification\n");
                    inOptions = false;
                }
                else
                {
                    options->bNoParam_ls40 = false;

                    Index -= 1;

                    char *pszParam = ls40_ReadNextArgument(argv, argI, argc);
                    if (NULL == pszParam)
                    {
                        inOptions = false;
                    }
                    else
                    {
                        size_t lStrLen = strlen(pszParam);
                        if (NULL != options->pszParamList_ls40[Index])
                        {
                            // specifying parameters twice not allowed
                            printf("Substitution parameter p%c specified twice\n", pSwitch[1]);
                            inOptions = false;
                        }
                        else
                        {
                            options->pszParamList_ls40[Index] = new _TCHAR[lStrLen + 1];
                            memcpy(options->pszParamList_ls40[Index], pszParam, lStrLen);
                            options->pszParamList_ls40[Index][lStrLen] = '\0';
                        }
                    }
                }
            }
            // suppress output - hidden option
            else if (0 == strcmp(pSwitch, "x"))
            {
                options->bSuppressOutput_ls40 = true;
            }
            else
            {
                ls40_Usage();
                inOptions = false;
            }

            if (true == inOptions)
            {
                pSwitch = ls40_ReadNextArgument(argv, argI, argc, 1);
            }
        }       // while ( (true == inOptions) && (argI < argc) )


        // Special handling of user name and password
        if (true == inOptions)
        {
            // Check if special handling of reading users name and password is requested
            if ( (true == bReadUser) && (0 == options->dbuser[0]) )
            {
                rc = ls40_ReadUser(options);
                if (PRG_FATAL == rc)
                {
                    inOptions = false;
                }
                else
                {
                    // Returned without any user information => error. May also be caused by CTRL-C.
                    if (0 == options->dbuser[0])
                    {
                        printf("User data missing\n");
                        inOptions = false;
                    }
                }
            }

            //*
            //*     Another special handling for option user: make it all upper case
            //*
            if (true == inOptions)
            {
                size_t l = _tcslen(options->dbuser);
                for (char *p = options->dbuser; p < options->dbuser + l; *p = _totupper(*p), ++p)
                    ;
            }
        }   // end if (true == inOptions)
    }   // if ( (1 == argc) || ...

    if ( (true == inOptions) && (argc > (argI + 1)) )
    {
        // There are still plenty of arguments to process but the user made obviously
        // a mistake with one of them - print usage information
        ls40_Usage();
        inOptions = false;
    }

    if (false == inOptions)
    {
        rc = PRG_FATAL;
    }
    else
    {
        // Some additional checks: interactive mode and another outfile than stdout does not match
        if ( (stdout != options->outfile) && (options->bInteractive_ls40) )
        {
            printf("Interactive mode impossible with an outfile other than stdout\n");
            rc = PRG_FATAL;
        }
    }

    return rc;
}
//  ls40_ParseOptions()


/*
  -----------------------------------------------------------------------------
  function: ls40_ReadUser
  -----------------------------------------------------------------------------
*/
static int
ls40_ReadUser(ls40_Options *pOptions)
{
    int rc = PRG_OK;

    char         szUser[128];
    szUser[0] = 0;
    tsp00_Namec  szPwd;
    bool         bGiven = false;

    tsp5_rte_fileerror rteError;

    tsp00_Int4    lCnt = 0;
    tsp00_Longint slen = 0;

    while ( (false == bGiven) && (PRG_OK == rc) && (lCnt < 3) )     // limit: 3 trials
    {
        lCnt++;
        if (0 == slen)
        {
            //*
            //*     Read the user name from stdin
            //*
            cn13fputs(stdout, "Enter user for authorization: ", true);
            cn13fgets(szUser, 99, stdin, true);

            slen = _tcslen(szUser);
            
            ls05StrTrimRight(szUser, slen, '\n');
            ls05StrTrimRight(szUser, slen, '\r');
            szUser[slen] = 0;
        }   // end if (0 == slen)

        if (slen > 0)
        {
            //*
            //*     Read the password from stdin
            //*
            if (!sqlgetpass("Enter password: ", szPwd, rteError))
            {
                if (vf_ok != rteError.sp5fe_result)
                {
                    printf("Error reading password from console: %s", rteError.sp5fe_text.asCharp());
                }
                rc = PRG_FATAL;
            }
            else
            {
                if (szPwd.length() > 0)
                {
                    bGiven = true;
                    pOptions->dbuser[0] = 0;
                    sp77sprintf(pOptions->dbuser, 128, "%s,%s", szUser, szPwd.asCharp());
                }
            }
        }   // end if (slen > 0)
    }   // end while ( false == bNameGiven)

    return rc;
}
// ls40_ReadUser()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ServerConnect
  -----------------------------------------------------------------------------
*/
static int ls40_ServerConnect(ls40_Options *pOptions)
{
    tsp00_ErrTextc   errtext;
    tsp00_Int4       rc = PRG_OK;
    tsp00_NodeIdc    szNode;
    tsp00_DbNamec    szDbName;
    tsp00_VFilenamec szDbRoot;

    if (true == pOptions->bStaticServer_ls40)
    {
        // Use statically linked client/server
        sqlxinit("","");
        ctrlservopen (&(pOptions->session), pOptions->dbname, pOptions->szDBRoot_ls40, &pOptions->bCancelByte_ls40, true);
    }
    else
    {
        switch (pOptions->NodeLoc)       // repm_node-option empty ? --> log on at local host
        {
            case rncLocal_els40:
            {
                pOptions->repm_node = "";                        // use local machine
                break;
            }
            case rncDBNode_els40:
            {
                if ('\0' != pOptions->db_node[0])
                {
                    pOptions->repm_node = pOptions->db_node;      // use node running the db for repm server
                }
                else
                {
                    pOptions->repm_node = "";                    // use local machine
                }
                break;
            }
            default:
            {
                break;  // Default is rncDifferent_els40 -> repm_node has already right value
            }
        }



        if (0 != pOptions->repm_node[0])
        {
// We do TEMPORARILY not support remote communication between client and server to prevent vulnerability
// that result from the fact that the server runs as user root. In this case it is necessary to pass username
// and passwort as parameters to the server to make a user change possible.
// Even better would be to not install the server as user root.
// Until this is done the loadercli will not be able to remotely communicate with the server.

            tsp01_CommErr commerr;
            tsp00_NodeId  localnode;
            tsp00_ErrText ErrText;
        
            sqlhostname(localnode, ErrText, &commerr);
            if (0 != commerr)
            {
                // could not get local host name;
                fprintf(pOptions->outfile, "Getting local hostname failed (errno = %d)\n", STATIC_CAST(tsp00_Int4, commerr));
                rc = PRG_FATAL;
            }
            else
            {
                // Make the hostname a c string for comparison
                tsp00_NodeIdc Nodec;
                Nodec.p2c(localnode);
                if (0 != _tcsicmp(Nodec.asCharp(), pOptions->repm_node) )
                {
                    fprintf(pOptions->outfile, "\n** ATTENTION **\n"
                                              "Remote communication not supported (option -r);\n"
                                              "use the local Loader server instead\n\n");
                    rc = PRG_SHOW;      // This is a 'misuse' of a special error message to prevent the client
                                        // from writing too much error messages to stdout/output file
                }
            }
        }
    
        if (PRG_OK == rc)
        {
            char pszNodeString[1024] = "local host";
            if ('\0' != pOptions->repm_node[0])
            {
                strcpy(&pszNodeString[0], pOptions->repm_node);
            }

            szDbName.Init();
            szNode.Init();
            szDbRoot.Init();

            if ('\0' != pOptions->repm_node[0])
            {
                szNode.rawAssign(pOptions->repm_node);
            }

            //*
            //*     Adjust calling parameter at connect to server
            //*
            //* - user may have specified: repm_node, db_node, dbname, dbroot
            //* - name of loader server host is ALWAYS specified; empty string means <local host>
            //*
            //* 1. if db_node != repm_node
            //*     ---> dbname (ALWAYS)= leer, dbroot = [<name>]
            //*
            //* 2. if db_node == repm_node
            //*     - dbname given, no dbroot given     ---> dbname = <name>, dbroot = leer
            //*     - dbname given, dbroot given        ---> dbname = leer,   dbroot = <name>
            //*     - no dbname given, dbroot given     ---> dbname = leer,   dbroot = <name>
            //*     - no dbname given, no dbroot given  ---> dbname = leer,   dbroot = leer

            if (0 != _tcsicmp(pOptions->repm_node, pOptions->db_node) )
            {
                if ('\0' != pOptions->szDBRoot_ls40[0])
                {
                    szDbRoot.rawAssign(pOptions->szDBRoot_ls40.asCharp());
                }
            }
            else
            {
                if ('\0' != pOptions->szDBRoot_ls40[0])
                {
                    szDbRoot.rawAssign(pOptions->szDBRoot_ls40.asCharp());
                }
                else
                {
                    if (NULL != pOptions->dbname)
                    {
                        szDbName.rawAssign(pOptions->dbname);
                    }
                }
            }

            char pszReply[2*PATH_MXSP00];
            rc = cn14connectRPM(szNode, szDbName, szDbRoot, pszReply, &(pOptions->session), errtext);
            if (PRG_OK == rc)
            {
                if (false == pOptions->bSuppressOutput_ls40)
                {
                    fprintf(pOptions->outfile,
                            "Opened connection to Loader SERVER at node %s.\n", pszNodeString);
                    if (NULL != pszReply)
                    {
                        fprintf(pOptions->outfile, "%s\n", pszReply);
                    }
                }
            } 
            else 
            {
                if (false == pOptions->bSuppressOutput_ls40)
                {
                    fprintf(pOptions->outfile, "Connection to Loader SERVER at node %s failed: %s\n",
                                              pszNodeString, STATIC_CAST(char*,errtext) );
                }
                rc = PRG_FATAL;
            }

            fflush(pOptions->outfile);

        }   // end if (PRG_OK == rc)

    }   // end else of if (true == pOptions->bStaticServer_ls40)


    //*
    //*     Get maximum packet length and build command buffer;
    //* depends on static or not
    //*
    if (PRG_OK == rc)
    {
        ls40_InitBuffers(*pOptions);
    }


    if (true == pOptions->bStaticServer_ls40)
    {
        // Before doing anything else give some state information; this includes the
        // info on the server log file location
        strcpy(pOptions->pszCmdBuffer, "SET");
        pOptions->CmdBufferLen = 3 * sizeof(char);

        rc = ls40_ServerCommand(pOptions);
        if (PRG_OK == rc)
        {
            if ('\0' != pOptions->pszReplyBuffer[0])
            {
                fprintf(pOptions->outfile, "%s\n", pOptions->pszReplyBuffer);
            }
        }
    }

    return rc;
}
//  ls40_ServerConnect()


/*
  -----------------------------------------------------------------------------
  function:     ls40_DisconnectServer
  -----------------------------------------------------------------------------
*/
static void
ls40_DisconnectServer(ls40_Options *pOptions)
{
    if (NULL != pOptions->session)
    {
        if (true == pOptions->bStaticServer_ls40)
        {
            ctrlservclose(pOptions->session);
        }
        else
        {
            cn14release(&pOptions->session);
        }
    }
}
// ls40_DisconnectServer()


/*
  -----------------------------------------------------------------------------
  function:     ls40_UserConnect
  -----------------------------------------------------------------------------
*/
static int ls40_UserConnect(ls40_Options &opt)
{
    int         rc = PRG_OK;

    const char *pszPassword;
    size_t      nUserNameLength;
    
    if (NULL == opt.dbname)      // db name given ?
    {
        if ( (0 != opt.dbuser[0]) && (false == opt.bSuppressOutput_ls40) )
        {
            fprintf(opt.outfile, "Error connecting to database: database name missing\n");
        }
        return rc;
    }
    
    ls40_InitBuffers(opt);

    if (0 == opt.dbuser[0])
    {
        // connects only the loader server to db but does not logon user
        // (password and user name missing)
        sprintf(opt.pszCmdBuffer, "USE SERVERDB %s", opt.dbname);
        if ('\0' != opt.db_node[0])
        {
            strcat(opt.pszCmdBuffer, " ON ");
            strcat(opt.pszCmdBuffer, opt.db_node);
        }
    }
    else
    {
        // Separate user name from password
        pszPassword = strchr (opt.dbuser, ',');
        if (NULL != pszPassword)
        {
            nUserNameLength = pszPassword - opt.dbuser;
            pszPassword++;
        }
        else
        {
            nUserNameLength = strlen(opt.dbuser);
            pszPassword = "";
        }

        // Mask DB name with double quotes to make special identifiers (like
        // hyphen -) possible in DB names

        sprintf(opt.pszCmdBuffer, "USE USER \"%.*s\" %s SERVERDB \"%s\"",
                                  nUserNameLength, opt.dbuser, pszPassword, opt.dbname);
        if ('\0' != opt.db_node[0])
        {
            strcat(opt.pszCmdBuffer, " ON ");
            strcat(opt.pszCmdBuffer, opt.db_node);
        }
    }

    char pszNode[1024] = "local host";
    if ('\0' != opt.db_node[0])
    {
        strcpy(&pszNode[0], opt.db_node);
    }


    opt.CmdBufferLen = _tcslen(opt.pszCmdBuffer);

    rc = ls40_ServerCommand(&opt);
    if (false == opt.bSuppressOutput_ls40)      // output only if requested
    {
        if (PRG_OK == rc)
        {
            if (0 == opt.dbuser[0])
            {
                fprintf(opt.outfile, "Connection to database %s on %s established.\n",
                                     opt.dbname, pszNode);
            }
            else
            {
                fprintf(opt.outfile, "User %.*s connected to database %s on %s.\n",
                                     nUserNameLength, opt.dbuser, opt.dbname, pszNode);
            }
        }
        else if (PRG_CMD_ERROR == rc)
        {
            if (0 == opt.dbuser[0])
            {
                fprintf(opt.outfile, "Error establishing connection to database %s on %s: %s\n",
                                     opt.dbname, pszNode, opt.pszReplyBuffer);
            }
            else
            {
                fprintf(opt.outfile, "Error connecting user %.*s to database %s on %s: %s\n",
                                     nUserNameLength, opt.dbuser, opt.dbname, pszNode, opt.pszReplyBuffer);
            }
        }
        // else means PRG_SERVER_ERROR error msg is already written by ls40_ServerCommand
        
        fflush(opt.outfile);
    }    

    return rc;
}
//  ls40_UserConnect()


/*
  -----------------------------------------------------------------------------
  function:     ls40_IsCmdSeparator
  -----------------------------------------------------------------------------
*/
static bool ls40_IsCmdSeparator(_TCHAR *p)
{
    bool rc             = false;

    size_t len = strlen(p);
    size_t i   = 0;

    for (i = 0; i < len; ++i)
    {
        if ( (0 == isspace(p[i])) && (0 == iscntrl(p[i])) )
        {
            if ('/' == p[i])
            {
                if ('/' == p[i+1])
                {
                    rc = true;
                }
            }
            break;
        }
    }

    return rc;
}
//  ls40_IsCmdSeparator()


/*
  -----------------------------------------------------------------------------
  function:     ls40_Substitute
  -----------------------------------------------------------------------------
*/
static tsp00_Int4
ls40_Substitute(ls40_Options &opt, _TCHAR *pszBuffer, tsp00_Longint & lBufferLength)
{
    tsp00_Int4 rc          = 0;

    _TCHAR        *pszBufStart = pszBuffer;
    tsp00_Longint  lMyBufLen   = lBufferLength;

    
    tsp00_Int4    lIndex    = 0;
    tsp00_Longint lRestLen  = 0;
    tsp00_Longint lParamLen = 0;

    _TCHAR *pszFound  = NULL;
    bool    bNotFound = true;

    while (NULL != (pszFound = strchr(pszBuffer, '&') ) )
    {
        bNotFound = true;

        if ( 0 != isdigit(pszFound[1]) )            // We treat only parameters with one digit
        {
            _TCHAR pszTemp[2] = {'\0'};
            pszTemp[0] = pszFound[1];
            if (0 != (lIndex = atoi(pszTemp) ) )    // the char following & must be a digit
            {
                lIndex -= 1;                        // lIndex moves by -1, because we start at 0

                if (NULL != opt.pszParamList_ls40[lIndex])  // No parameter defined -> jump over it
                {
                    bNotFound = false;

                    lParamLen = _tcslen(opt.pszParamList_ls40[lIndex]);
                    
                    if ( (lMyBufLen + lParamLen - 3) > MAX_COMMAND_LENGTH_LS40)
                    {
                        rc = -1;                     // entire length of string to generate may not exceed
                        break;                       // allowed length (considering terminating 0, too)
                    }

                    lRestLen = lMyBufLen - ((pszFound - pszBufStart) + 2);

                    memmove(&pszFound[lParamLen], pszFound + 2, lRestLen);
                    memcpy(pszFound, opt.pszParamList_ls40[lIndex], lParamLen);
                    
                    pszBuffer = pszFound + lParamLen;       // move buf ptr to new position behind inserted
                                                            // parameter string
                    lMyBufLen += (lParamLen - 2);           // adapt length
                    
                    pszBufStart[lMyBufLen] = '\0';          // memmove destroys '\0'
                }
            }
        }

        if (true == bNotFound)
        {
            pszBuffer = &pszFound[1];
        }
    }

    if (lMyBufLen != lBufferLength)
    {
        lBufferLength = lMyBufLen;
    }

    return rc;
}
// ls40_Substitute()


/*
  -----------------------------------------------------------------------------
  function:     ls40_Batch
  -----------------------------------------------------------------------------
*/
static int
ls40_Batch(ls40_Options &opt)
{
    int rc = PRG_OK;
    
    _TCHAR        *pszBuffer  = new _TCHAR[MAXBUFFERSIZE_LS40];
    tsp00_Longint  lBufferLen = 0;

    int            lErrNum    = 0;
    char          *dummy      = NULL;
    bool           cmdfound   = false;

    //*
    //*     Read command file until eof is reached and process found commands
    //*

    ls40_InitBuffers(opt);      // Initialize buffer
    while ( (PRG_OK == rc) && ( 0 == feof(opt.infile) ) )
    {
        opt.CmdBufferLen = 0;

        //*
        //*     Find rows belonging to command until we find a comment or the end of file is
        //*     reached or the command becomes too long and must be sent in parts
        //*
        while ( (PRG_OK == rc) && (NULL   != (dummy = fgets(pszBuffer, MAXBUFFERSIZE_LS40, opt.infile) ) ) )
        {
            lBufferLen = strlen(pszBuffer);
            if ( true == ls40_IsCmdSeparator(pszBuffer) )
            {
                if (true == cmdfound)
                {
                    break;
                }
            }
            else
            {
                opt.CmdBufferLen += lBufferLen;
                if (opt.CmdBufferLen > MAX_COMMAND_LENGTH_LS40)
                {
                    //*
                    //*     Error: command becomes too long
                    //* Log the command to file and exit
                    //*
                    if (false == opt.bSuppressOutput_ls40)      // output only if requested
                    {
                        if ('\n' == opt.pszCmdBuffer[opt.CmdBufferLen - lBufferLen - 1])
                        {
                            fprintf(opt.outfile, "%s", opt.pszCmdBuffer);
                        }
                        else
                        {
                            fprintf(opt.outfile, "%s\n", opt.pszCmdBuffer);
                        }
                        fprintf(opt.outfile, "Error processing command; command is too long\n"
                                             "Maximum command length: %d\n", MAX_COMMAND_LENGTH_LS40);
                    }
                    rc = PRG_FATAL;

                }
                else
                {
                    //*
                    //*     Copy command extract to command buffer
                    //*
                    strcat(opt.pszCmdBuffer, pszBuffer);
                    if (false == cmdfound)
                    {
                        cmdfound = true;
                    }
                }
            }

        }   // end while ( (PRG_OK == rc) && ...

        //*
        //*     Check for reading error
        //*
        if ( (PRG_OK == rc) && (NULL == dummy) )
        {
            int nError = ferror(opt.infile);
            if (0 != nError)
            {
                if (false == opt.bSuppressOutput_ls40)      // output only if requested
                {
                    fprintf(opt.outfile, "Error reading from command file: %d", nError);
                }
                rc = PRG_FATAL;
            }
        }

        //*
        //*     Process found command
        //*
        if ( (PRG_OK == rc) && (true == cmdfound) )
        {
            //* Remove trailing newline chars
            if ('\n' == opt.pszCmdBuffer[opt.CmdBufferLen - 1])
            {
                opt.CmdBufferLen -= 1;
                if ('\r' == opt.pszCmdBuffer[opt.CmdBufferLen - 1])
                {
                    opt.CmdBufferLen -= 1;
                }
                opt.pszCmdBuffer[opt.CmdBufferLen] = 0;
            }

            //*
            //*     Substitute parameters and log command to outfile
            //*
            if (false == opt.bNoParam_ls40)
            {
                ls40_Substitute(opt, opt.pszCmdBuffer, opt.CmdBufferLen);
            }

            if (false == opt.bSuppressOutput_ls40)      // output only if requested
            {
                if ('\n' == opt.pszCmdBuffer[opt.CmdBufferLen - 1])
                {
                    fprintf(opt.outfile, "%s", opt.pszCmdBuffer);   // Log command to file
                }
                else
                {
                    fprintf(opt.outfile, "%s\n", opt.pszCmdBuffer);   // Log command to file
                }
            }

            //*
            //*     Send command to server and wait for reply
            //*
            rc = ls40_ServerCommand(&opt);
            if (PRG_OK == rc)
            {
                if (false == opt.bSuppressOutput_ls40)      // output only if requested
                {
                    if ('\0' != opt.pszReplyBuffer[0])
                    {
                        fprintf(opt.outfile, "%s\n", opt.pszReplyBuffer);
                    }
                    fprintf(opt.outfile, "Successfully executed\n\n");
                }
            }
            else if (PRG_CMD_ERROR == rc)
            {
                if (false == opt.bSuppressOutput_ls40)      // output only if requested
                {
                    fprintf(opt.outfile, "Error during execution\n-->%s\n\n", opt.pszReplyBuffer);
                }
            
                ++lErrNum;
                if ( (0 != opt.BatchErrorCnt_ls40) && (lErrNum == opt.BatchErrorCnt_ls40) )
                {
                    if (false == opt.bSuppressOutput_ls40)      // output only if requested
                    {
                        fprintf(opt.outfile, "Defined maximum number of errors (%d) reached\n",
                                             opt.BatchErrorCnt_ls40);
                    }
                }
                else
                {
                    rc = PRG_OK;
                }
            }
        }   // end if ( (PRG_OK == rc) && (true == cmdfound) )

        cmdfound = false;           // reinitialize
        ls40_InitBuffers(opt);      // reinitialize buffer

        fflush(opt.outfile);
    }   // end while ( (PRG_OK == rc) && ( 0 == feof(opt.infile) ) )


    if (NULL != pszBuffer)
    {
        delete [] pszBuffer;
        pszBuffer = NULL;
    }

    // In case of error while processing commands print summary to stderr 
    if ( (false == opt.bSuppressOutput_ls40) && (0 != lErrNum) )
    {
        fprintf(stderr, "\n\n%d error(s) during execution of command file '%s'.\n"
                        "---> See log file(s) for more information.\n",
                        lErrNum, opt.pszInfileName);
    }
    fflush(opt.outfile);

    return rc;
}
//  ls40_Batch()


/*
  -----------------------------------------------------------------------------
  function:     ls40_ProcessInteractively
  -----------------------------------------------------------------------------
*/
static int
ls40_ProcessInteractively(ls40_Options *pOptions)
{
    int rc = PRG_OK;
    
    ls40_InitBuffers(*pOptions);      // Initialize buffer

    //*
    //*     Print prompt
    //*
    cn13fputs(pOptions->outfile, pOptions->application, true);
    cn13fputs(pOptions->outfile, ">", true);

    int                 linelen;
    int                 connected = true;
    int                 concat    = true;
    char                *bufpos;

    int availbuf = MAX_COMMAND_LENGTH_LS40;

    // read command(s) from file (e.g. stdin)
    while ( (false  == pOptions->bRelease_ls40) &&
            (PRG_OK == rc) &&
            (NULL   != cn13fgets(pOptions->pszCmdBuffer, MAX_COMMAND_LENGTH_LS40, pOptions->infile, true) ) )
    {
        linelen = (int)strlen(pOptions->pszCmdBuffer);

        availbuf -= linelen;

        concat = ((linelen > 1) && (pOptions->pszCmdBuffer[linelen-2] == '/'));

        while (concat)
        {
            --linelen;               // throw away '/'
            ++availbuf;

//!!! Make sure no buffer overflow
            pOptions->pszCmdBuffer[linelen-1] = '\n';

            bufpos = pOptions->pszCmdBuffer + linelen;

            if (NULL != (cn13fgets(bufpos, availbuf, pOptions->infile, true) ) )
            {
                linelen = (int)strlen(pOptions->pszCmdBuffer);
                availbuf = MAX_COMMAND_LENGTH_LS40 - linelen;

                concat = ((linelen > 1) && (pOptions->pszCmdBuffer[linelen-2] == '/'));
            }
            else
            {
                concat = false;
            }
        }   // end while  (concat)

        // remove new line character
        if (pOptions->pszCmdBuffer[linelen - 1] == '\n')
        {
            --linelen; 
        }

        pOptions->CmdBufferLen = linelen;

        // comment or empty line ?
        if ((linelen == 0) || (pOptions->pszCmdBuffer[0] == '#'))
        {
            continue;
        }

        //*
        //*     Finally process the command
        //*
        rc = ls40_ServerCommand(pOptions);
        if (PRG_OK == rc)
        {
            if ('\0' != pOptions->pszReplyBuffer[0])
            {
                fprintf(pOptions->outfile, "%s\n", pOptions->pszReplyBuffer);
            }
            fprintf(pOptions->outfile, "OK\n");
        }
        else if (PRG_CMD_ERROR == rc)
        {
            fprintf(pOptions->outfile, "ERR %s\n", pOptions->pszReplyBuffer);
            
            rc = PRG_OK;
        }


        //*
        //*     Print prompt
        //*
        cn13fputs(pOptions->outfile, "\n---\n", true);
        
        if (false == pOptions->bRelease_ls40)
        {
            cn13fputs(pOptions->outfile, pOptions->application, true);
            cn13fputs(pOptions->outfile, ">", true);
        }

        ls40_InitBuffers(*pOptions);      // reinitialize buffer
    }   // end while ( (false  == pOptions->bRelease_ls40) && ...

    return rc;
}
//  ls40_ProcessInteractively()


/*
  -----------------------------------------------------------------------------
  function: ls40_CheckAnswer
  -----------------------------------------------------------------------------
*/
static int ls40_CheckAnswer(ls40_Options & options, tsp00_Int4 ReplyLen)
{
    short   i               = 0;
    int     nErrorNumber    = PRG_OK;


    if (0 == strncmp(options.pszReplyBuffer, REPMAN_ERR_STR, strlen(REPMAN_ERR_STR)))       // error
    {
        // Skip indicator "err" and read error number. The error is given in
        // the form "E<error number>:" . The first and the last char need to be removed.
        nErrorNumber = atol(&options.pszReplyBuffer[REPLY_HEADER_SIZE_LS40 + 1]);

        options.CmdExecError = (nErrorNumber == 0) ? PRG_FATAL : nErrorNumber;
        nErrorNumber = PRG_CMD_ERROR;
    }
    
    // Remove answer header in any case
    tsp00_Int4 len = ReplyLen - REPLY_HEADER_SIZE_LS40;

    memmove(options.pszReplyBuffer, &options.pszReplyBuffer[REPLY_HEADER_SIZE_LS40], len);
    options.pszReplyBuffer[len] = '\0';
    
    return nErrorNumber;
}
//  ls40_CheckAnswer()


/* 
  -----------------------------------------------------------------------------
  function: ls40_ServerCommand
  -----------------------------------------------------------------------------
*/
static int ls40_ServerCommand(ls40_Options  *options)
{
    tsp00_Int4      rc        = PRG_OK;
    
    tsp00_ErrTextc  errtext;
    tsp00_Int4      commErr   = CtrlCmdOk_csp09;
    tsp00_Int4      nReplyLen = options->ReplyBufferLen;
    

    options->bRelease_ls40 = (0 == _tcsnicmp(options->pszCmdBuffer, "release", 7) ) ||
                             (0 == _tcsnicmp(options->pszCmdBuffer, "bye", 3)     ) ||     
                             (0 == _tcsnicmp(options->pszCmdBuffer, "exit", 4)    ) ||     
                             (0 == _tcsnicmp(options->pszCmdBuffer, "quit", 4)    )  ;

    if (false == options->bRelease_ls40)
    {
        if (true == options->bStaticServer_ls40)
        {
            // bei statischem betrieb kann das kommando beliebig lang sein, da es einfach
            // als funktionsparameter uebergeben wird

            options->pszCommBuffer_ls40[1] = 1;     // give server a hint to run statically linked
            
            options->CmdBufferLen += RPM_PROTOCOL_SIZE_CN14;

            commErr = ctrlservcommand(options->session,
                                      options->pszCommBuffer_ls40,
                                      STATIC_CAST(tsp00_Int4, options->CmdBufferLen),
                                      options->pszReplyBuffer,
                                      &nReplyLen,
                                      RPM_REPLY_BUFFER_LEN_CN14);
        }
        else
        {
            commErr = cn14ExecuteLoaderCmd(options->session,
                                           options->pszCommBuffer_ls40,
                                           STATIC_CAST(tsp00_Int4, options->CmdBufferLen),     // IA64
                                           options->pszReplyBuffer,
                                           &nReplyLen,
                                           errtext);
        }

        // return codes of server: CtrlCmdOk_csp09      0
        //                         CtrlCmdErr_csp09     1
        //                         CtrlRelease_csp09    2

        if (CtrlCmdOk_csp09 == commErr)         // no server error; check for command
        {                                       // processing return code
            rc = ls40_CheckAnswer(*options, nReplyLen);
        }
        else if (CtrlRelease_csp09 == commErr)      // relaese command; nothing to do
        {
            ;        
        }
        else                                        // server error
        {
            if (false == options->bSuppressOutput_ls40)
            {
                fprintf (options->outfile, "Loader server error (%d : '%s')",
                                           STATIC_CAST(int, commErr),
                                           STATIC_CAST(char*, errtext) );
            }
            rc = PRG_SERVER_ERROR;
        }

    }   // end if (false == options->bRelease_ls40)


    return rc;
}
//  ls40_ServerCommand()


/* 
  -----------------------------------------------------------------------------
  function: main
  -----------------------------------------------------------------------------
*/
int main (int argc, char* argv[])
{
    int             MainRetCode = PRG_OK;
    ls40_Options    options;

    g_pOptions = &options;


    tsp00_CompName c_component;           // initialize connection to Xserver
    strcpy(c_component, C_COMPONENT);
    sqlinit(c_component, NULL);

    ls40_Initialize(options);

    MainRetCode = ls40_ParseOptions(&options, argc, argv);
    if (PRG_OK != MainRetCode)
    {
        exit(PRG_FATAL);
    }

    // Connect to loader server;
    // this function fills options->nMaxDataSize und options-session and allocates
    // space for command buffer;
    // in case of error we need to release this connection
    if (PRG_OK == MainRetCode)
    {
        MainRetCode = ls40_ServerConnect(&options);
        if (PRG_OK != MainRetCode)
        {
            // we couldn't even log on to the requested server for what reason o'ever
            // ---> suppress any other error output than the one the function already did
            MainRetCode = PRG_SHOW;     // suppress any other error output than the one the function already did
        }
    }

    if ( (MainRetCode == PRG_OK) && (true == options.showServerVersion) )
    {
        MainRetCode = ls40_ShowServerVersion(options);
    }


    //*
    //*     User logon
    //*
    if (PRG_OK == MainRetCode)
    {
        MainRetCode = ls40_UserConnect(options);
        if (PRG_OK != MainRetCode)
        {
            // we couldn't even log on the user for what reason o'ever
            // ---> suppress any other error output than the one the function already did
            MainRetCode = PRG_SHOW;
        }
    }

    // setup signal managment
    signal(SIGINT,  ls40_Aborthandler);
    signal(SIGTERM, ls40_Aborthandler);
    signal(SIGABRT, ls40_Aborthandler);
#if defined (_WIN32)
    signal(SIGBREAK, ls40_Aborthandler);
#endif
    
    
    //*     Command file
    if (PRG_OK == MainRetCode) 
    {
        if (true == options.bInteractive_ls40)
        {
            // Process commands interactively
            MainRetCode = ls40_ProcessInteractively(&options);
        }
        else// if (NULL != options.infile)
        {
            MainRetCode = ls40_Batch(options);
        }
        
    }



    // Print general error msg to stderr in case of error
    if ( (PRG_OK != MainRetCode) && (PRG_SHOW != MainRetCode) )
    {
        if (false == options.bSuppressOutput_ls40)
        {
            fprintf(stderr, "\n\nSome error(s) occurred running command.\n"
                            "---> See log file(s) for more information.\n");
        }
    }
    
    //*     Disconnect Server
    ls40_DisconnectServer(&options);

    sqlfinish(false);           // Undo LZU-initialization (sqlinit() )

    if ( NULL != options.pszCommBuffer_ls40)
    {
        options.pszCmdBuffer = NULL;

        delete [] options.pszCommBuffer_ls40;
        options.pszCommBuffer_ls40 = NULL;
    }
    if ( NULL != options.pszReplyBuffer)
    {
        delete [] options.pszReplyBuffer;
        options.pszReplyBuffer = NULL;
    }

    if ( NULL != options.infile )
    {
        fclose (options.infile);
        options.infile = NULL;
    }
    if ( (options.outfile != stdout) && (NULL != options.outfile) )
    {
        fclose (options.outfile);
        options.outfile = NULL;
    }

    tsp00_Int2 i = 0;
    
    for (i; i < MAX_PARAM_LS40; ++i)
    {
        if (NULL != options.pszParamList_ls40[i])
        {
            delete [] options.pszParamList_ls40[i];
            options.pszParamList_ls40[i] = NULL;
        }
    }

    exit(MainRetCode);
    return MainRetCode;
}
//  main()
